Package: reporter
Type: Package
Title: Creates Statistical Reports
Version: 1.4.7
Authors@R: c( 
    person("David", "Bosak", email = "dbosak01@gmail.com", role = c("aut", "cre")),
    person("Bill", "Huang", email = "Bill.Huang@Toastmasters.org.tw", role = c("aut")),
    person("Kevin", "Kramer", email = "kkrame02@amgen.com", role = "ctb"),
    person("Duong", "Tran", email = "trand000@aol.com", role = "ctb"),
    person("Raphael", "Huang", email = "chi-hua.huang@astellas.com", role = "ctb"),
    person("Archytas Clinical Solutions", role = "cph")
    )
Maintainer: David Bosak <dbosak01@gmail.com>
Description: Contains functions to create regulatory-style statistical reports.
    Originally designed to create tables, listings, and figures for the 
    pharmaceutical, biotechnology, and medical device industries, these
    reports are generalized enough that they could be used in any industry.
    Generates text, rich-text, PDF, HTML, and Microsoft Word file formats.  
    The package specializes 
    in printing wide and long tables with automatic page wrapping and splitting.  
    Reports can be produced with a minimum of function calls, and without 
    relying on other table packages.  The package supports titles, footnotes, 
    page header, page footers, spanning headers, page by variables, 
    and automatic page numbering.
License: CC0
Encoding: UTF-8
URL: https://reporter.r-sassy.org, https://github.com/dbosak01/reporter
BugReports: https://github.com/dbosak01/reporter/issues
Depends: R (>= 3.6), common (>= 1.1.0)
Suggests: testthat, magrittr, dplyr, tidyr, readr, knitr, rmarkdown,
        ggplot2, gridExtra, survminer, utils, logr, covr, libr, devEMF,
        patchwork, tools
Imports: fmtr(>= 1.5.8), stringi, crayon, jpeg, graphics, grDevices,
        zip, withr, glue
RoxygenNote: 7.3.3
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-15 04:21:35 UTC; dbosa
Author: David Bosak [aut, cre],
  Bill Huang [aut],
  Kevin Kramer [ctb],
  Duong Tran [ctb],
  Raphael Huang [ctb],
  Archytas Clinical Solutions [cph]
Repository: CRAN
Date/Publication: 2026-05-15 04:50:02 UTC
Built: R 4.6.0; ; 2026-05-15 06:00:10 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,plot_spec)
S3method(print,report_spec)
S3method(print,style_spec)
S3method(print,table_spec)
S3method(print,text_spec)
S3method(render,default)
S3method(render,pdf_array)
S3method(render,pdf_dictionary)
S3method(render,pdf_document)
S3method(render,pdf_image_stream)
S3method(render,pdf_info)
S3method(render,pdf_object)
S3method(render,pdf_text_stream)
export(add_content)
export(add_style)
export(cell_style)
export(column_defaults)
export(create_plot)
export(create_report)
export(create_style)
export(create_table)
export(create_text)
export(define)
export(footer_image)
export(footnotes)
export(ftn)
export(get_theme)
export(header_image)
export(lowcase_n)
export(lowcase_parens)
export(options_fixed)
export(page_by)
export(page_footer)
export(page_header)
export(set_margins)
export(span)
export(spanning_header)
export(stub)
export(title_header)
export(titles)
export(ttl)
export(upcase_n)
export(upcase_parens)
export(write_registration_file)
export(write_report)
import(common)
import(crayon)
import(fmtr)
import(glue)
import(grDevices)
import(graphics)
import(jpeg)
import(stats)
import(stringi)
import(withr)
import(zip)

<!-- badges: start -->

[![reporter version](https://www.r-pkg.org/badges/version/reporter)](https://cran.r-project.org/package=reporter)
[![reporter lifecycle](https://img.shields.io/badge/lifecycle-maturing-blue.svg)](https://cran.r-project.org/package=reporter)
[![reporter downloads](https://cranlogs.r-pkg.org/badges/reporter)](https://cran.r-project.org/package=reporter)
[![reporter total downloads](https://cranlogs.r-pkg.org/badges/grand-total/reporter)](https://cran.r-project.org/package=reporter)
[![R-CMD-check](https://github.com/dbosak01/reporter/workflows/R-CMD-check/badge.svg)](https://github.com/dbosak01/reporter/actions)
[![Codecov test coverage](https://codecov.io/gh/dbosak01/reporter/branch/master/graph/badge.svg)](https://app.codecov.io/gh/dbosak01/reporter?branch=master)

<!-- badges: end -->

# Introduction to **reporter**
<img src="man/images/reporter_new.png" align="left" height="138px" style="margin-right:10px;height:138px"
   alt="Reporter package logo"/>

Historically, R has not been very strong on reporting.  The **reporter**
package aims to fill that gap.  

Using **reporter**, you can create a report in just a few lines of code. 
Not only is it easy to create a report, but the **reporter** package can handle 
all sorts of situations that other packages struggle with.

For example, unlike other packages, the **reporter** package creates the *entire
report*: page header and footer, titles, footnotes, tables - everything.
The end result of a reporter call is a complete, printable report. 

In addition, **reporter** can handle page breaking, page wrapping, and 
automatic sizing of column widths.  The package
offers a choice of output file types.  And it supports the inclusion of 
tables, text and graphics into a report.

What is more, the package does not expect you to know R Markdown, knitr, or pandoc.
You do not need to learn Latex, HTML, or any other intermediate language.
With **reporter**, you send your data into a create function, assign 
titles and footnotes, and write the report.  That's it!

If you are familiar with SAS® software, you may notice some 
similarity between **reporter** functions and `proc report`.  This similarity,
however, is only on the surface.  The implementation of **reporter** is
done entirely in R, and, internally, is modeled in a different way. 
However, SAS® users will find the **reporter** 
functions very convenient and easy to understand compared to the
alternatives.

### Installation

The **reporter** package can be installed from the console.  Simply run 
the following command: 

    install.packages("reporter")
    
Or if you want the latest development version, you can install it directly
from github:

    devtools::install_github("https://github.com/dbosak01/reporter")


Then put the following line at the top of your program or script:

    library(reporter)

The **reporter** package will give you access to a number of functions
to help create, lay out, and write your report to the file system. 
For examples and usage information, visit the **reporter** documentation
site [here](https://reporter.r-sassy.org/articles/reporter.html).


### Introductory Video

Here is a YouTube video that gives a brief introduction to the main features
of the **reporter** package:

[![Reporter Package Introduction](https://img.youtube.com/vi/PuqQFqOrFXQ/0.jpg)](https://www.youtube.com/watch?v=PuqQFqOrFXQ)

### Getting Help

If you need help, the first place 
to turn to is the [reporter](https://reporter.r-sassy.org/) web site. The web site
has full documentation on all **reporter** functions.

If you want to look at the code for the **reporter** package, visit the
github page [here](https://github.com/dbosak01/reporter/).

If you encounter a bug or have a feature request, please submit an issue 
[here](https://github.com/dbosak01/reporter/issues/).

### See Also

The **reporter** package is part of the **sassy** meta-package. 
The **sassy** meta-package includes several packages that help make R
easier for SAS® programmers.  You can read more about the **sassy** package
[here](https://sassy.r-sassy.org/).

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  
#  # Create temporary path
#  tmp <- file.path(tempdir(), "example5.pdf")
#  
#  # Prepare Data
#  dat <- mtcars[1:10, ]
#  df <- data.frame(vehicle = rownames(dat), dat)
#  
#  # Define Table with spanning headers
#  tbl <- create_table(df) %>%
#    titles("Table 1.0", "MTCARS Spanning Headers") %>%
#    spanning_header(from = "mpg", to = "hp", label = "Span 1", n = 10) %>%
#    spanning_header(from = "drat", to = "qsec", label = "Span 2", n = 10) %>%
#    spanning_header(from = "vs", to = "carb", label = "Span 3", n = 10) %>%
#    spanning_header(from = "drat", to = "carb",
#                    label = "Super Span", level = 2) %>%
#    define(vehicle, label = "Vehicle") %>%
#    define(mpg, format = "%.1f") %>%
#    define(disp, visible = FALSE) %>%
#    define(am, visible = FALSE) %>%
#    footnotes("* Motor Trend, 1974")
#  
#  # Create Report and add table
#  rpt <- create_report(tmp, output_type = "PDF",
#                       font = "Courier", font_size = 12) %>%
#    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>%
#    add_content(tbl) %>%
#    page_footer(left = Sys.time(),
#              center = "Confidential",
#              right = "Page [pg] of [tpg]")
#  
#  # Write the report
#  res <- write_report(rpt)
#  
#  # file.show(tmp)
#  

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  
#  # Create temp file name
#  tmp <- file.path(tempdir(), "example7.pdf")
#  
#  # Prepare data
#  dat <- mtcars[order(mtcars$cyl), ]
#  dat <- data.frame(vehicle = rownames(dat), dat)
#  
#  # Define table
#  tbl <- create_table(dat, show_cols = 1:8) %>%
#    page_by(cyl, label="Cylinders: ")
#  
#  # Create the report
#  rpt <- create_report(tmp, output_type = "PDF",
#                       font = "Courier", font_size = 12) %>%
#    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>%
#    titles("Listing 3.0", "MTCARS Data Listing with Page By") %>%
#    set_margins(top = 1, bottom = 1) %>%
#    add_content(tbl) %>%
#    page_footer(left = Sys.time(),
#            center = "Confidential",
#            right = "Page [pg] of [tpg]")
#  
#  
#  # Write the report
#  write_report(rpt)
#  
#  # file.show(tmp)
#  

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  
#  # Create temporary path
#  tmp <- file.path(tempdir(), "example4.pdf")
#  
#  # Dummy text
#  cnt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
#                "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
#                "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
#                "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ",
#                "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
#                "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
#                "qui officia deserunt mollit anim id est laborum.")
#  
#  # Create text content
#  txt <- create_text(cnt, width = 6)
#  
#  # Prepare data
#  dat <- mtcars[1:10, ]
#  
#  # Create table content
#  tbl <- create_table(dat) %>%
#    titles("Table 1.0", "MTCARS Sample Data") %>%
#    footnotes("* Motor Trend, 1974")
#  
#  # Create report and add both table and text content
#  rpt <- create_report(tmp, font_size = 12,
#                       output_type = "PDF") %>%
#    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>%
#    add_content(tbl, page_break = FALSE) %>%
#    add_content(txt) %>%
#    page_footer(left = Sys.time(),
#                center = "Confidential",
#                right = "Page [pg] of [tpg]")
#  
#  # Write the report
#  write_report(rpt)
#  
#  # file.show(tmp)
#  

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  library(ggplot2)
#  
#  # Get image path
#  dir <- system.file("extdata", package = "reporter")
#  ipth <- file.path(dir, "Logo_full.png")
#  
#  # create dummy data
#  data <- data.frame(
#    name=letters[1:5],
#    value=sample(seq(4,15),5),
#    sd=c(1,0.2,3,2,4)
#  )
#  
#  # Define Revenue Plot
#  p <- ggplot(data) +
#    geom_bar( aes(x=name, y=value), stat="identity", fill="maroon", alpha=0.7) +
#    geom_errorbar( aes(x=name, ymin=value-sd, ymax=value+sd), width=0.4, colour="orange",
#                   alpha=0.9, linewidth=1.3)
#  
#  
#  # Create plot object
#  plt <- create_plot(p, height = 4.5, width = 6) |>
#    titles("Figure 1.0", "Projected Revenue by Quarter", bold = TRUE, font_size = 12)
#  
#  # Create report
#  rpt <- create_report("./output/myreport.docx", output_type = "DOCX",
#                       font = "Arial", font_size = 10) |>
#    page_header("",                          # Put nothing in left column
#      c("Archytas Solutions\nChicago, IL, 60601"), # Combine right column text into same cell
#                width = c(6, 0, 3)) |>       # Remove middle cell with zero width
#    header_image(ipth,                       # Assign image file path
#                 height = .75, width = 3.5,  # Set height and width
#                 align = "left") |>          # Align to left header column
#    add_content(plt) |>
#    page_footer(paste0("Date: ", Sys.Date()), right = "Page [pg] of [tpg]")
#  
#  # Write report to file
#  res <- write_report(rpt)
#  
#  # View report
#  # file.show(res$path)
#  

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  
#  # Create temporary path
#  tmp <- file.path(tempdir(), "example10.rtf")
#  
#  # Read in prepared data
#  df <- read.table(header = TRUE, text = '
#        var     label        A             B
#        "ampg"   "N"          "19"          "13"
#        "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)"
#        "ampg"   "Median"     "16.4"        "21.4"
#        "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
#        "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
#        "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)"
#        "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)"
#        "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
#  
#  # Create table
#  tbl <- create_table(df, first_row_blank = TRUE, borders = "all") %>%
#    stub(c("var", "label")) %>%
#    column_defaults(width = 1.25) %>%
#    define(var, blank_after = TRUE, label_row = TRUE,
#           format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>%
#    define(label, indent = .25) %>%
#    define(A, label = "Group A", align = "center", n = 19) %>%
#    define(B, label = "Group B", align = "center", n = 13)  %>%
#    titles("Table 1.0", "MTCARS Summary Table", borders = "outside",
#           bold = TRUE, font_size = 14) %>%
#    footnotes("* Motor Trend, 1974", borders = "outside")
#  
#  # Create report and add content
#  rpt <- create_report(tmp, output_type = "RTF", font = "Arial",
#                       font_size = 12) %>%
#    set_margins(top = 1, bottom = 1) %>%
#    add_content(tbl)
#  
#  # Write out report
#  write_report(rpt)
#  
#  # View report
#  # file.show(tmp)
#  

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  
#  # Create temp file name
#  tmp <- file.path(tempdir(), "example6.pdf")
#  
#  # Prepare data
#  dat <- mtcars[1:10, ]
#  dat <- data.frame(vehicle = rownames(dat), dat)
#  
#  # Define table
#  tbl <- create_table(dat, show_cols = 1:8) %>%
#    define(vehicle, label = "Vehicle", width = 3, id_var = TRUE, align = "left") %>%
#    define(mpg, label = "Miles per Gallon", width = 1) %>%
#    define(cyl, label = "Cylinders", format = "%.1f") %>%
#    define(disp, label = "Displacement") %>%
#    define(hp, label = "Horsepower", page_wrap = TRUE) %>%
#    define(drat, visible = FALSE) %>%
#    define(wt, label = "Weight") %>%
#    define(qsec, label = "Quarter Mile Time", width = 1.5)
#  
#  
#  # Create the report
#  rpt <- create_report(tmp, output_type = "PDF",
#                       font = "Courier", font_size = 12) %>%
#    titles("Listing 2.0", "MTCARS Data Listing with Page Wrap") %>%
#    set_margins(top = 1, bottom = 1) %>%
#    add_content(tbl) %>%
#    page_footer(right = "Page [pg] of [tpg]")
#  
#  # Write the report
#  write_report(rpt)
#  
#  # file.show(tmp)
#  

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  
#  # Create temp file name
#  tmp <- file.path(tempdir(), "example16a.pdf")
#  
#  # Create data
#  arm <- c(rep("A", 3), rep("B", 2), rep("C", 3), rep("D", 2))
#  subjid <- 100:109
#  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
#            "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma",
#            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
#            "Smith, Shaileigh")
#  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
#  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
#  
#  df <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)
#  df <- rbind(df, df, df, df)
#  
#  # Output with group_border
#  tbl1 <- create_table(df, first_row_blank = FALSE, borders = "outside") %>%
#    define(subjid, label = "Subject ID for a patient", n = 10, align = "left",
#           width = 1) %>%
#    define(name, label = "Subject Name", width = 1) %>%
#    define(sex, label = "Sex", n = 10, align = "center") %>%
#    define(age, label = "Age", n = 10) %>%
#    define(arm, label = "Arm",
#           dedupe = TRUE,
#           group_border = TRUE)
#  
#  
#  rpt <- create_report(tmp, output_type = "pdf", font = "Arial",
#                       font_size = 10) %>%
#    titles(c("Table 1.0", "This is a table with group border"), align = "center") %>%
#    add_content(tbl1) %>%
#    set_margins(top = 1, bottom = 1)
#  
#  
#  res <- write_report(rpt)
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  
#  fp <- file.path(tempdir(), "example16b.pdf")
#  
#  # Prepare data
#  adae <- data.frame(
#    total = rep("Number of subject with adverse events", 53),
#    AESOC = c(NA,
#              rep("General disorders and administration site conditions", 9),
#              rep("Hepatobiliary disorders", 5),
#              rep("Infections and infestations", 25),
#              rep("Injury, poisoning and procedural complications", 13)),
#    AEDECOD = c(NA,
#                NA, rep("APPLICATION SITE ERYTHEMA", 4), rep("INFLUENZA LIKE ILLNESS", 4),
#                NA, rep("GALLBLADDER DISORDER", 4),
#                NA, rep("BRONCHITIS", 4), rep("NAIL INFECTION", 4), rep("PARONYCHIA", 4),
#                rep("RASH PUSTULAR", 4), rep("RESPIRATORY TRACT INFECTION", 4),
#                rep("RHINITIS", 4),
#                NA, rep("FALL", 4), rep("HEAT CRAMPS", 4), rep("MUSCLE INJURY", 4)
#    )
#  )
#  
#  adae$aedecod_seq <- rep(NA, nrow(adae))
#  non_na_idx <- !is.na(adae$AEDECOD)
#  adae$aedecod_seq[non_na_idx] <- ave(adae$AEDECOD[non_na_idx], adae$AEDECOD[non_na_idx], FUN = seq_along)
#  
#  adae$AESEV <- adae$aedecod_seq
#  adae$AESEV[adae$AESEV == "1"] <- NA
#  adae$AESEV[adae$AESEV == "2"] <- "Mild"
#  adae$AESEV[adae$AESEV == "3"] <- "Moderate"
#  adae$AESEV[adae$AESEV == "4"] <- "Severe"
#  
#  adae$trt1 <- "x ( x.x%)"
#  adae$trt2 <- "x ( x.x%)"
#  adae$trt3 <- "x ( x.x%)"
#  
#  adae$aedecod_id <- rep(NA, nrow(adae))
#  adae$aedecod_id[non_na_idx] <- ave(adae$AEDECOD[non_na_idx], adae$AESOC[non_na_idx],
#                                     FUN = function(x){as.numeric(factor(x, levels = unique(x)))})
#  
#  adae$blank_grp <- paste(adae$total,
#                          adae$AESOC,
#                          ifelse(adae$aedecod_id == 1,NA,adae$AEDECOD),
#                          sep = "|")
#  
#  adae <- adae[, setdiff(names(adae), c("aedecod_seq", "aedecod_id"))]
#  
#  # Output
#  custom_n_format <- function(x){
#    return(paste0("\n(N = ",x,")\nn(%)"))
#  }
#  
#  current_date <- gsub(" ","",toupper(format(Sys.Date(),"%d %b %Y")))
#  current_time <- substr(Sys.time(),12,19)
#  output_name <- "example16c.pdf"
#  program_path <- "/dummy/tables/t-teae-soc-pt-gr-saf.R"
#  
#  total_width <- 6
#  trt_width <- 1.04
#  item_width <- total_width - (trt_width*3)
#  
#  tbl <- create_table(adae,
#                      borders = "outside",
#                      n_format = custom_n_format) %>%
#  
#    # ----- Column setting -----#
#    column_defaults(from = trt1, to = trt3, align = "center", width = trt_width) %>%
#    define(blank_grp, blank_before = T, visible = FALSE) %>%
#    stub(vars = c("total", "AESOC", "AEDECOD", "AESEV"),
#         label = "System Organ Class\n  Preferred Term\n    Grade",
#         width = item_width) %>%
#    define(AESOC, break_label = "(Continued)") %>%
#    define(AEDECOD, indent = 0.16, break_label = "(Continued)") %>%
#    define(AESEV, indent = 0.32) %>%
#    define(trt1, label = "Treatment A", n=11) %>%
#    define(trt2, label = "Treatment B", n=7) %>%
#    define(trt3, label = "Treatment C", n=15) %>%
#  
#    # ----- Footnote setting -----#
#    footnotes("Page [pg] of [tpg]", align = "right", blank_row = "none", valign = "top") %>%
#    footnotes("Safety Analysis Set includes all subjects who are received at least one dose of study drug.",
#              align = "left", blank_row = "none", valign = "top") %>%
#    footnotes("The analysis is performed on the maximum severity grade.",
#              align = "left", blank_row = "none", valign = "top") %>%
#    footnotes("Coded using MedDRA version 22.1.",
#              align = "left", blank_row = "none", valign = "top") %>%
#    footnotes(paste0("Program: ", program_path), italics = TRUE)
#  
#  sty <- create_style(font_name = "Arial",
#                      font_size = 9)
#  
#  rpt <- create_report(fp,
#                       orientation = "landscape",
#                       output_type = "PDF")  %>%
#    add_style(style = sty) %>%
#    set_margins(top = 1, bottom = 0.75, right = 1, left = 1.5) %>%
#    add_content(tbl) %>%
#  
#    # ----- Title setting -----#
#    titles("Table 14-6.1.3.1.  Treatment-emergent Adverse Events by System Organ Class,",
#           "Preferred Term and Grade",
#           "(Safety Analysis Set)",
#           bold = T,
#           font_size = 11)
#  
#  # Write the report
#  res <- write_report(rpt)
#  
#  # Uncomment to view
#  # file.show(res$path)
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  define(AESOC, break_label = "(Continued)", group_cohesion = TRUE)
#  define(AEDECOD, indent = 0.16, break_label = "(Continued)", group_cohesion = TRUE)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  
#  fp <- file.path(tempdir(), "example16b.pdf")
#  
#  # Create data
#  dat <- iris[1:91,]
#  
#  dat$group_1 <- c(
#    rep("A", 14),
#    rep("B", 6),
#    rep("C", 22),
#    rep("D", 18),
#    rep("E", 31)
#  )
#  
#  dat$group_2 <- c(
#    rep("Flower A\nSubgroup A1", 14),
#    rep("Flower B\nSubgroup B1", 3),
#    rep("Flower B\nSubgroup B2", 3),
#    c(rep("Flower C1", 17)),
#    c(rep("Flower C2", 5)),
#    c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)),
#    c(rep("Flower E1", 2)),
#    c(rep("Flower E2", 29))
#  )
#  
#  dat <- dat[, c("group_1", "group_2", "Sepal.Length",
#                 "Sepal.Width", "Petal.Length","Petal.Width")]
#  
#  # Create table
#  tbl <- create_table(dat, borders = "outside") %>%
#    define(group_1, group_cohesion = TRUE, label = "Region", blank_after = T,
#           width = 1.2) %>%
#    define(group_2, group_cohesion = TRUE, label = "Sector")
#  
#  # Create report
#  rpt <- create_report(fp, output_type = "pdf", font = "Arial",
#                       font_size = 10, orientation = "landscape") %>%
#    titles("Table 1.0", "Analysis of Iris Flowers") %>%
#    set_margins(top = 1, bottom = 1) %>%
#    add_content(tbl) %>%
#    footnotes("Program: /test/example16b.R",
#              "Output: /test/example16b.pdf")
#  
#  # Output report
#  res <- write_report(rpt)
#  

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  library(magrittr)
#  
#  # Create temp file name
#  tmp <- file.path(tempdir(), "example.pdf")
#  
#  # Create report content
#  tbl <- create_table(mtcars) %>%
#    titles("MTCARS Sample Data") %>%
#    footnotes("* Motor Trend, 1974")
#  
#  # Create report and add content
#  rpt <- create_report(tmp, output_type = "PDF",
#                       font = "Courier", font_size = 12) %>%
#    add_content(tbl)
#  
#  # Write out the report
#  write_report(rpt)
#  
#  # Un-comment to view report
#  # file.show(tmp)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  library(magrittr)
#  
#  # Create temp file path
#  tmp <- file.path(tempdir(), "example13.rtf")
#  
#  # Prepare Data
#  dat <- airquality[sample(1:153, 15), ]
#  dat$Month <-  as.Date(paste0("1973-", dat$Month, "-01"))
#  
#  # Define table
#  tbl <- create_table(dat, show_cols = c("Month", "Day", "Wind",
#                                         "Temp", "Ozone")) %>%
#    titles("Table 9.6", "Air Quality Sample Report{supsc('1')}",   # Superscript 1
#           borders = c("top", "bottom"), blank_row = "none") %>%
#    column_defaults(width = .7) %>%
#    define(Month, format = "%B", align = "left", width = 1) %>%
#    define(Temp, format = "%.0f") %>%
#    footnotes("{supsc('1')}New York, May to September 1973",       # Superscript 1
#              borders = c("top", "bottom"), blank_row = "none")
#  
#  # Define report
#  rpt <- create_report(tmp, output_type = "RTF", font = "Arial",
#                       font_size = 12, missing = "-") %>%
#    page_header("Sponsor: EPA", "Study: B34958", blank_row = "below") %>%
#    add_content(tbl) %>%
#    page_footer(Sys.Date(), right = "Page [pg] of [tpg]")
#  
#  # Write the report to the file system
#  write_report(rpt)
#  
#  # View report
#  # file.show(tmp)
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  
#  # Create complex superscript expression
#  ex1 <- "x" %p% supsc("(n+4)")
#  
#  # View expression
#  ex1
#  # [1] "x⁽ⁿ⁺⁴⁾"
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  
#  # Append superscript using paste0()
#  df[2, "A"] <- df[2, "A"] %p% supsc('2')
#  
#  # View data frame
#  df
#  #    var      label           A           B
#  # 1 ampg          N          19          13
#  # 2 ampg       Mean 18.8 (6.5)²  22.0 (4.9)  # Observe superscript in column A
#  # 3 ampg     Median        16.4        21.4
#  # 4 ampg    Q1 - Q3 15.1 - 21.2 19.2 - 22.8
#  # 5 ampg      Range 10.4 - 33.9 14.7 - 32.4
#  # 6  cyl 8 Cylinder 10 ( 52.6%)  4 ( 30.8%)
#  # 7  cyl 6 Cylinder  4 ( 21.1%)  3 ( 23.1%)
#  # 8  cyl 4 Cylinder  5 ( 26.3%)  6 ( 46.2%)
#  

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  
#  # Create temp file name
#  tmp <- file.path(tempdir(), "example1.pdf")
#  
#  # Create the report
#  rpt <- create_report(tmp, orientation = "portrait", output_type = "PDF") %>%
#    page_header(left = "Client: 1-800-Flowers", right = "Study: Iris") %>%
#    titles("Listing 1.0", "IRIS Data Listing") %>%
#    add_content(create_table(iris)) %>%
#    footnotes("* Bulletin of the American Iris Society, 1935") %>%
#    page_footer(left = Sys.time(),
#                center = "Confidential",
#                right = "Page [pg] of [tpg]")
#  
#  # Write the report
#  write_report(rpt)
#  
#  # file.show(tmp)
#  

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  
#  # Create temp file name
#  tmp <- file.path(tempdir(), "example8.rtf")
#  
#  tbl <- create_table(mtcars[1:10, ], width = 7) %>%
#         title_header("Listing 4.0", "MTCARS Data Listing with Title Header",
#                 right = c("Page [pg] of [tpg]",  "Study: Cars"),
#                 blank_row = "both")
#  
#  # Create the report
#  rpt <- create_report(tmp, output_type = "RTF",
#                       font = "Courier", font_size = 12) %>%
#    page_header(left = "Client: Motor Trend", right = "Study: Cars",
#                blank_row = "below") %>%
#    add_content(tbl) %>%
#    set_margins(top = 1, bottom = 1) %>%
#    footnotes("* Motor Trend, 1974") %>%
#    page_footer(left = Sys.time(),
#          center = "Confidential",
#          right = "Page [pg] of [tpg]")
#  
#  # Write the report
#  write_report(rpt)
#  
#  # file.show(tmp)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  library(ggplot2)
#  
#  # Create temporary path
#  tmp <- file.path(tempdir(), "example10.pdf")
#  
#  # Prepare data
#  dat <- mtcars[order(mtcars$cyl), ]
#  
#  # Generate plot
#  p <- ggplot(dat, aes(x=disp, y=mpg)) + geom_point()
#  
#  # Define plot object
#  plt <- create_plot(p, height = 4, width = 8) %>%
#    titles("Figure 1.0", "MTCARS Mileage By Displacement", blank_row = "none") %>%
#    footnotes("* Motor Trend, 1974")
#  
#  # Add plot to report
#  rpt <- create_report(tmp, output_type = "PDF") %>%
#    set_margins(top = 1, bottom = 1) %>%
#    options_fixed(font_size = 12) %>%
#    page_header("Sponsor", "Study: cars") %>%
#    add_content(plt) %>%
#    page_footer(Sys.time(), "Confidential", "Page [pg] of [tpg]")
#  
#  # Write out report
#  res <- write_report(rpt)
#  
#  
#  # View report
#  # file.show(tmp)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  
#  # Create temporary path
#  tmp <- file.path(tempdir(), "example12a.html")
#  
#  # Read in prepared data
#  df <- read.table(header = TRUE, text = '
#      var     label        A             B
#      "ampg"   "N"          "19"          "13"
#      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)"
#      "ampg"   "Median"     "16.4"        "21.4"
#      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
#      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
#      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)"
#      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)"
#      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
#  
#  # Create custom style
#  sty <- create_style(font_name = "Arial",
#                      font_size = 10,
#                      background_color = "WhiteSmoke",
#                      border_color = "Grey",
#                      title_font_size = 12,
#                      title_font_bold = TRUE,
#                      title_font_color = "SteelBlue",
#                      table_header_background = "Tan",
#                      table_header_font_bold = TRUE,
#                      table_header_font_color = "White",
#                      table_body_background = "White",
#                      table_body_stripe = "Wheat",
#                      table_stub_background = "Tan",
#                      table_stub_font_color = "White",
#                      table_stub_font_bold = TRUE
#                   )
#  
#  # Create table
#  tbl <- create_table(df, first_row_blank = TRUE, borders = c("all")) %>%
#    stub(c("var", "label")) %>%
#    column_defaults(width = 1.25) %>%
#    define(var, blank_after = TRUE, label_row = TRUE,
#           format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>%
#    define(label, indent = .25) %>%
#    define(A, label = "Group A", align = "center", n = 19) %>%
#    define(B, label = "Group B", align = "center", n = 13)  %>%
#    titles("Table 1.0", "MTCARS Summary Table with Custom Style",
#           borders = "none") %>%
#    footnotes("* Motor Trend, 1974", borders = "outside")
#  
#  # Create report and add custom style
#  rpt <- create_report(tmp, output_type = "HTML", font = "Arial") %>%
#    add_content(tbl) %>%
#    add_style(sty)
#  
#  # Write out report
#  write_report(rpt)
#  
#  # View report
#  # file.show(tmp)
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  
#  # Create temporary path
#  tmp <- file.path(tempdir(), "example12b.html")
#  
#  # Read in prepared data
#  df <- read.table(header = TRUE, text = '
#        var     label        A             B
#        "ampg"   "N"          "19"          "13"
#        "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)"
#        "ampg"   "Median"     "16.4"        "21.4"
#        "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
#        "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
#        "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)"
#        "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)"
#        "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
#  
#  # Create table
#  tbl <- create_table(df, first_row_blank = FALSE, borders = "all") %>%
#    stub(c("var", "label")) %>%
#    column_defaults(width = 1.25) %>%
#    define(var, blank_after = TRUE, label_row = TRUE,
#           format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>%
#    define(label, indent = .25) %>%
#    define(A, label = "Group A", align = "center", n = 19) %>%
#    define(B, label = "Group B", align = "center", n = 13)  %>%
#    titles("Table 1.0", "MTCARS Summary Table with SteelBlue Theme",
#           borders = "none") %>%
#    footnotes("* Motor Trend, 1974", borders = "outside")
#  
#  # Create report and add theme
#  rpt <- create_report(tmp, output_type = "HTML", font = "Arial",
#                       font_size = 12) %>%
#    set_margins(top = 1, bottom = 1) %>%
#    add_content(tbl) %>%
#    add_style(theme = "SteelBlue")
#  
#  # Write out report
#  write_report(rpt)
#  
#  # View report
#  # file.show(tmp)
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  library(magrittr)
#  
#  tm <- get_theme("SteelBlue")
#  
#  print(tm)
#  # # A style specification:
#  # - font_name: 'Arial'
#  # - font_size: 10
#  # - text_color: 'DimGrey'
#  # - title_font_size: 11
#  # - title_font_bold: TRUE
#  # - title_font_color: 'SteelBlue'
#  # - border_color: 'Grey'
#  # - table_header_background: 'SteelBlue'
#  # - table_header_font_bold: TRUE
#  # - table_header_font_color: 'LightGrey'
#  # - table_body_background: 'White'
#  # - table_body_stripe: 'WhiteSmoke'
#  # - table_stub_background: 'SteelBlue'
#  # - table_stub_font_color: 'LightGrey'
#  # - table_stub_font_bold: TRUE
#  
#  # Modify theme
#  tm$table_stub_background <- "Orange"
#  tm$table_stub_font_color <- "White"
#  
#  # Create temporary path
#  tmp <- file.path(tempdir(), "example12c.html")
#  
#  # Read in prepared data
#  df <- read.table(header = TRUE, text = '
#        var     label        A             B
#        "ampg"   "N"          "19"          "13"
#        "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)"
#        "ampg"   "Median"     "16.4"        "21.4"
#        "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
#        "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
#        "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)"
#        "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)"
#        "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
#  
#  # Create table
#  tbl <- create_table(df, first_row_blank = FALSE, borders = "all") %>%
#    stub(c("var", "label")) %>%
#    column_defaults(width = 1.25) %>%
#    define(var, blank_after = TRUE, label_row = TRUE,
#           format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>%
#    define(label, indent = .25) %>%
#    define(A, label = "Group A", align = "center", n = 19) %>%
#    define(B, label = "Group B", align = "center", n = 13)  %>%
#    titles("Table 1.0", "MTCARS Summary Table with Modified Theme",
#           borders = "none") %>%
#    footnotes("* Motor Trend, 1974", borders = "outside")
#  
#  # Create report and add modified theme
#  rpt <- create_report(tmp, output_type = "HTML", font = "Arial",
#                       font_size = 12) %>%
#    set_margins(top = 1, bottom = 1) %>%
#    add_content(tbl) %>%
#    add_style(style = tm)
#  
#  # Write out report
#  write_report(rpt)
#  
#  # View report
#  # file.show(tmp)
#  
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  
#  # Create temporary path
#  tmp <- file.path(tempdir(), "example12j.rtf")
#  
#  # Read in prepared data
#  df <- read.table(header = TRUE, text = '
#      var     label        A             B
#      "ampg"   "N"          "19"          "13"
#      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)"
#      "ampg"   "Median"     "16.4"        "21.4"
#      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
#      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
#      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)"
#      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)"
#      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
#  
#  
#  # Create table
#  tbl <- create_table(df, first_row_blank = TRUE, borders = c("all"),
#                      header_bold = TRUE) %>%
#    stub(c("var", "label"),
#         style = cell_style(bold = TRUE, indicator = "labelrow")) %>%
#    column_defaults(width = 1.25) %>%
#    define(var, blank_after = TRUE, label_row = TRUE,
#           format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>%
#    define(label, indent = .25) %>%
#    define(A, label = "Group A", align = "center", n = 19) %>%
#    define(B, label = "Group B", align = "center", n = 13)  %>%
#    titles("Table 1.0", "MTCARS Summary Table with Cell Style",
#           borders = "none", bold = TRUE, font_size = 11) %>%
#    footnotes("* Motor Trend, 1974", borders = "none", blank_row = "none")
#  
#  # Create report and add custom style
#  rpt <- create_report(tmp, output_type = "RTF", font = "Arial") %>%
#    add_content(tbl)
#  
#  # Write out report
#  write_report(rpt)
#  
#  # View report
#  # file.show(tmp)
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  
#  # Create temporary path
#  tmp <- file.path(tempdir(), "example12k.rtf")
#  
#  # Prepare data
#  df <- data.frame(names = rownames(mtcars), mtcars[, 1:3])
#  
#  # Set style indicator variable
#  df$mpgind <- ifelse(df$mpg > 20, TRUE, FALSE)
#  
#  # Create table
#  tbl <- create_table(df, first_row_blank = TRUE,
#                      header_bold = TRUE, borders = c("top", "bottom")) %>%
#    define(names, label = "Car Name") %>%
#    define(mpg, label = "Miles Per Gallon",
#           style = cell_style(bold = TRUE, indicator = "mpgind")) %>%
#    define(cyl, label = "Cylinders") %>%
#    define(disp, label = "Displacement")  %>%
#    define(mpgind, visible = FALSE) %>%
#    titles("Table 1.0", "MTCARS with Indicator Variable",
#           borders = "none", bold = TRUE, font_size = 11) %>%
#    footnotes("* Motor Trend, 1974", borders = "none", blank_row = "none")
#  
#  # Create report and add custom style
#  rpt <- create_report(tmp, output_type = "RTF", font = "Arial") %>%
#    add_content(tbl)
#  
#  # Write out report
#  write_report(rpt)
#  
#  # View report
#  # file.show(tmp)
#  

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  
#  # Create temporary path
#  tmp <- file.path(tempdir(), "example9.docx")
#  
#  # Read in prepared data
#  df <- iris[1:25, ]
#  
#  # Create table
#  tbl <- create_table(df, borders = "bottom")
#  
#  
#  # Create report and add content
#  rpt <- create_report(tmp, output_type = "DOCX", orientation = "portrait",
#                       font = "Times", font_size = 12) %>%
#    set_margins(top = 1, bottom = 1) %>%
#    page_header(left = "Client: 1-800-FLOWERS", right = "Study: Iris") %>%
#    titles("Listing 1.0", "Iris Data Listing") %>%
#    add_content(tbl) %>%
#    footnotes("* Bulletin of the American Iris Society, 1935") %>%
#    page_footer(left = Sys.time(),
#                center = "Confidential",
#                right = "Page [pg] of [tpg]")
#  
#  # Write out report
#  write_report(rpt)
#  
#  # View report
#  # file.show(tmp)
#  

# Create Tables ----------------------------------------------------------


#' @noRd
create_table_pages_html <- function(rs, cntnt, lpg_rows) {
  
  ts <- cntnt$object
  content_blank_row <- cntnt$blank_row
  
  pgby_var <- NA
  pgby_cnt <- 0
  pgby_fmt <- NULL
  if (!is.null(rs$page_by)) {
    pgby_var <- rs$page_by$var
    pgby_fmt <- rs$page_by$format
  } else if (!is.null(ts$page_by)) {
    pgby_var <- ts$page_by$var
    pgby_fmt <- ts$page_by$format
  }
  
  
  if (all(ts$show_cols == "none") & length(ts$col_defs) == 0) {
    
    stop("ERROR: At least one column must be defined if show_cols = \"none\".")
  }
  
  font_name <- rs$font
  
  # Set up control columns
  dat <- as.data.frame(ts$data, stringsAsFactors = FALSE)  
  dat$..blank <- ""
  dat$..row <- 1
  dat$..page_by <- NA
  
  # If page_break variable has been defined, use it
  if (is.null(ts$page_var)) {
    if (is.na(pgby_var))
      dat$..page <- NA
    else 
      dat$..page <-  dat[[pgby_var]]
    
  } else {
    if (any(class(dat[[ts$page_var]]) == "factor"))
      dat$..page <- as.character(dat[[ts$page_var]])
    else 
      dat$..page <- dat[[ts$page_var]]
  }
  
  # If page by is defined, use it
  if (!is.na(pgby_var)) {
    
    # Clear out factors on page by if they exist
    # Needed because split cells used to convert everything to characters
    if (any(class(dat[[pgby_var]]) == "factor")) {
      dat[[pgby_var]] <- as.character(dat[[pgby_var]] )
      dat$..page_by <- dat[[pgby_var]] 
      
      if (any(class(dat$..page) == "factor")) 
        dat$..page <- as.character(dat[[pgby_var]])
      
    } else {
      
      if (!is.null(pgby_fmt))
        dat$..page_by <-  fapply(dat[[pgby_var]], pgby_fmt)
      else 
        dat$..page_by <-  dat[[pgby_var]]
    
    }
    
    pgby_cnt <- get_pgby_cnt(dat$..page_by)
    
    # Commenting out for now
    # if (is.unsorted(dat[[pgby_var]], strictly = FALSE))
    #   message("Page by variable not sorted.")
  }
  
  # Deal with invisible columns
  if (!is.null(ts$col_defs)) {
    for (def in ts$col_defs) {
      if (def$visible == FALSE) {
        nnm <- paste0("..x.", def$var_c)
        dat[[nnm]] <- dat[[def$var_c]]
      }
    }
    
    # Deal with Group Cohesion
    # Prepare ..group_cohesion and ..min_page_prop
    group_cohesion <- c()
    min_page_prop <- c()
    
    i <- 1
    for (def in ts$col_defs) {
      if (def$group_cohesion == TRUE) {
        group_cohesion <- c(group_cohesion, def$var_c)
        min_page_prop <- c(min_page_prop, def$min_page_prop)
        
        dat$..temp <- dat[[def$var_c]]
        names(dat)[names(dat) == "..temp"] <-  paste0("..group_cohesion",i)
        
        dat$..temp <- def$min_page_prop
        names(dat)[names(dat) == "..temp"] <-  paste0("..min_page_prop",i)
        i <- i + 1
      }
    }
    
    if (length(group_cohesion) == 0) {
      group_cohesion <- NULL
    }
    
    if (length(min_page_prop) == 0) {
      min_page_prop <- NULL
    }
    
    # Prepare empty ..break_label and ..break_label_lines for later use
    i <- 1
    for (def in ts$col_defs) {
      if (!is.null(def$break_label)) {
        
        dat$..temp <- ifelse(!is.na(dat[[def$var_c]]) & dat[[def$var_c]] != "",
                             paste0(dat[[def$var_c]], " ", def$break_label),
                             dat[[def$var_c]])
        names(dat)[names(dat) == "..temp"] <-  paste0("..break_label",i)
        
        dat$..temp <- rep(NA, nrow(dat))
        names(dat)[names(dat) == "..temp"] <-  paste0("..break_label_lines",i)
        i <- i + 1
      }
    }
    # Prepare repeated label occurrence for later use
    if (any(grepl("..break_label", names(dat)))) {
      dat$..break_occur <- rep(NA, nrow(dat))
    }
  }
  
  # Get control column names 
  control_cols <- names(dat)[is.controlv(names(dat))]
  
  # Get vector of all included column names
  # Not all columns in dataset are necessarily included
  # depends on show_cols parameter on create_table and
  # visible parameter on column definitions
  keys <- get_table_cols(ts, control_cols)
  # print("keys")
  # print(keys)
  
  # Filter dataset by included columns
  dat <- get_data_subset(dat, keys, rs$preview)
  # print("Key columns:")
  # print(dat)
  
  # Update column definitions with column defaults
  ts$col_defs <- set_column_defaults(ts, keys)
  # print("col_defs:")
  # print(ts$col_defs)
  
  # Deal with styles
  styles <- get_styles(ts)
  
  # Get labels
  labels <- get_labels(dat, ts)
  # print("Labels:")
  # print(labels)
  
  # Get column alignments
  aligns <- get_aligns(dat, ts)
  
  # Get alignment for labels
  # Follows column alignment by default
  label_aligns <- get_label_aligns(ts, aligns)
  # print("Label Aligns:")
  # print(label_aligns)
  
  # Clear out existing formats
  cdat <- clear_formats(dat)
  
  # Get column formats
  formats(cdat) <- get_col_formats(dat, ts)
  # print("formats:")
  # print(formats(cdat))
  
  # Apply formatting
  fdat <- fdata(cdat)
  # print("fdata:")
  # print(fdat)
  
  # Prep data for blank lines, indents, and stub columns
  fdat <- prep_data(fdat, ts, rs$char_width, rs$missing) # OK
  # print("prep_data")
  # print(fdat)
  # str(fdat)
  
  if ("..stub_var" %in% names(fdat)){
    control_cols <- c(control_cols, "..stub_var")
  }
  
  if ("..group_border" %in% names(fdat)){
    control_cols <- c(control_cols, "..group_border")
  }
  
  # Reset keys, since prep_data can add/remove columns for stub
  keys <- names(fdat)
  # print("Keys")
  # print(keys)
  # print("Aligns")
  # print(aligns)
  
  # Copy any width attributes to formatted data frame
  if ("width" %in% ts$use_attributes)
    widths(fdat) <- widths(dat)
  # print("Original Widths")
  # print(widths(dat))
  
  # Don't merge if there is a stub with background color
  merge_label_row <- TRUE
  if (has_style(rs, "table_stub_background") & "stub" %in% keys)
    merge_label_row <- FALSE
  
  # Get column widths
  widths_uom <- get_col_widths_variable(fdat, ts, labels, 
                                        rs$font, rs$font_size, rs$units, 
                                        rs$gutter_width, merge_label_row) 
  # print("Widths UOM")
  # print(widths_uom)
  
  # Split long text strings into multiple rows. Number of rows are stored in
  # ..row variable. If too slow, may need to be rewritten in C
  fdat <- split_cells_variable(fdat, widths_uom, rs$font,
                               rs$font_size, rs$units, rs$output_type, 
                               rs$char_width, ts)$data
  # print("split_cells")
  # print(fdat)
  
  
  # Break columns into pages
  wraps <- get_page_wraps(rs$line_size, ts, 
                          widths_uom, 0, control_cols)  # No gutter width for RTF
  # print("wraps")
  # print(wraps)
  
  
  # Create a temporary page info to pass into get_content_offsets
  tmp_pi <- list(keys = keys, col_width = widths_uom, label = labels,
                 label_align = label_aligns, table_align = cntnt$align, data = fdat)
  # print("Temp PI")
  # print(tmp_pi)
  
  
  # Offsets are needed to calculate splits and page breaks
  content_offset <- get_content_offsets_html(rs, ts, tmp_pi, 
                                             content_blank_row, pgby_cnt)
  
  
  # split rows
  splits <- get_splits_text(fdat, widths_uom, rs$body_line_count, 
                            lpg_rows, content_offset$lines, ts, TRUE)  
  # print("splits")
  # print(splits)
  
  # Subset splits by preview, if requested
  if (!is.null(rs$preview)) {
    if (rs$preview < length(splits))
      splits <- splits[seq(1, rs$preview)] 
  }
  
  tot_count <- length(splits) * length(wraps)
  counter <- 0
  wrap_flag <- FALSE
  blnk_ind <- "none"
  
  pg_lst <- list()
  for(s in splits) {
    for(pg in wraps) {
      counter <- counter + 1
      
      if (counter < tot_count)
        wrap_flag <- TRUE
      else 
        wrap_flag <- FALSE
      
      #print(s)
      if (!is.na(pgby_var))
        pgby <- trimws(s[1, "..page_by"])
      else 
        pgby <- NULL
      
      # Ensure content blank rows are added only to the first and last pages
      blnk_ind <- get_blank_indicator(counter, tot_count, content_blank_row,
                                      rs$body_line_count, content_offset$lines, 
                                      nrow(s), pgby)
      #print(blnk_ind)
      
      pi <- page_info(data= s[, pg], keys = pg, label=labels[pg],
                      col_width = widths_uom[pg], col_align = aligns[pg],
                      font_name = font_name, label_align = label_aligns[pg],
                      pgby, cntnt$align)
      pg_lst[[length(pg_lst) + 1]] <- create_table_html(rs, ts, pi, 
                                                       blnk_ind, wrap_flag,
                                                       lpg_rows, styles)
    }
  }
  
  ret <- list(widths = widths_uom, page_list = pg_lst)
  
  return(ret)
  
  
}



#' @noRd
create_table_html <- function(rs, ts, pi, content_blank_row, wrap_flag, 
                             lpg_rows, styles) {
  rh <- rs$row_height
  shdrs <- list(lines = 0, twips = 0)
  hdrs <- list(lines = 0, twips = 0)
  
  # Default to content width
  ls <- rs$content_size[["width"]]
  
  # Get table width
  if (!is.null(pi$col_width))
    ls <- sum(pi$col_width, na.rm = TRUE)
  
  if (!is.null(ts$title_hdr))
    ttls <- get_title_header_html(ts$title_hdr, ls, rs, pi$table_align)
  else
    ttls <- get_titles_html(ts$titles, ls, rs, pi$table_align) 
  
  
  if (!is.null(rs$page_by)) {
    pgby <- get_page_by_html(rs$page_by, rs$content_size[["width"]], 
                            pi$page_by, rs, pi$table_align, ttls$border_flag)
  } else if(!is.null(ts$page_by)) {
    pgby <- get_page_by_html(ts$page_by, ls, pi$page_by, rs, 
                             pi$table_align, ttls$border_flag)
  } else 
    pgby <- c()
  
  exbrdr <- ttls$border_flag
  if (length(pgby) > 0)
    exbrdr <- pgby$border_flag
  
  if (ts$headerless == FALSE) {
    
    # Get table header also includes spanning header
    hdrs <- get_table_header_html(rs, ts, pi, exbrdr)  
  }
  
  # rs, ts, widths,  algns, halgns, talgn
  rws <- get_table_body_html(rs, pi$data, pi$col_width, 
                             pi$col_align, pi$table_align, ts$borders, 
                             !ts$headerless,
                             ts$first_row_blank, styles, ts)
  
  a <- NULL
  if (content_blank_row %in% c("above", "both"))
    a <- "<br>"
  
  
  blnks <- c()

  
  # Determine sum of all lines
  rc <- sum(ttls$lines, pgby$lines, shdrs$lines, 
            hdrs$lines, rws$lines,
            length(a))
  
  # Get footnotes, passing in sum of all current lines
  ftnts <- get_page_footnotes_html(rs, ts, ls, lpg_rows, rc,
                                  wrap_flag, content_blank_row,  pi$table_align, 
                                  rws$border_flag)
  
  # Replace the page number for table footnote
  # temp_footnote <- gsub("[pg]", unique(pi$data$..page), ftnts$html, fixed = TRUE) 
  # ftnts$html <- temp_footnote

  
  u <- rs$units
  if (u == "inches")
    u <- "in"
  
  # ds <- paste0("<div ", ta, ">")
  ts <- paste0("<table class=\"tbs\" style=\"width:", 
               round(sum(pi$col_width, 
                         na.rm = TRUE), 3), u,";\">")
  
  ret <- list(html = c(a, ttls$html, pgby$html, ts, shdrs$html, 
                      hdrs$html, rws$html, "</table>", ftnts$html),
              lines = rc  + ftnts$lines)
  
  return(ret) 
}

# Haven't quite decided what to do with this function in HTML
# For now, just return the footnote and any blank lines above or below
get_page_footnotes_html <- function(rs, spec, spec_width, lpg_rows, row_count,
                                   wrap_flag, content_blank_row, talgn, 
                                   ex_brdr = FALSE) {
  
  ftnts <- list(lines = 0, twips = 0, border_flag = FALSE)
  vflag <- "none"

  # Deal with valign parameter
  if (!is.null(spec$footnotes)) {
    if (!is.null(spec$footnotes[[length(spec$footnotes)]])) {
      if (spec$footnotes[[length(spec$footnotes)]]$valign == "bottom") {
        
        vflag <- "bottom"
        ftnts <- get_footnotes_html(spec$footnotes, 
                                   spec_width, rs, 
                                   talgn, FALSE) 
      } else {
        vflag <- "top"
        ftnts <- get_footnotes_html(spec$footnotes, spec_width, rs, talgn, ex_brdr) 
      }
      
    }
  } else {
    
    if (!is.null(rs$footnotes[[1]])) {
      if (!is.null(rs$footnotes[[1]]$valign)) {
        if (rs$footnotes[[1]]$valign == "top") {
          vflag <- "top"
          ftnts <- get_footnotes_html(rs$footnotes, 
                                     spec_width, rs, 
                                     talgn, ex_brdr) 
        } else {
          
          if (wrap_flag)
            vflag <- "bottom" 
        }
      }
    }
  }
  
  b <- NULL
  blen <- 0
  if (content_blank_row %in% c("below", "both")) {
    b <- "<br>"
    blen <- 1
  }
  
  
  # Add extra offsets if table has a lot of borders turned on
  # to avoid undesired page wraps
  boff <- 0
  if (any(class(spec) == "table_spec") &
      any(spec$borders %in% c("all", "inside"))) {

    #boff <- round(row_count * rs$border_height / rs$row_height)
    boff <- 1
  }

  ublnks <- c()
  lblnks <- c()

  if (rs$paper_size != "none") {
    
    # Determine number of filler lines needed
    len_diff <- rs$body_line_count - row_count - ftnts$lines - lpg_rows - blen - boff
    
    if (vflag == "bottom" & len_diff > 0) {
  
      ublnks <- c(b, rep("<br>", len_diff))
  
    } else {
  
      if ((wrap_flag & len_diff > 0)) {
        if (vflag == "bottom" | has_page_footer(rs))
          lblnks <- c(rep("<br>", len_diff), b)
      } else {
        lblnks <- b
      }
    }
  }
  
  tlns <- sum(ftnts$lines, length(ublnks), length(lblnks))
  ret <- list(html = c(ublnks, ftnts$html, lblnks),
              lines = tlns)
  
  # ret <- list(html =  ftnts$html,
  #             lines = ftnts$lines)
  
  return(ret)
}

# Sub-Functions ---------------------------------------------------------


#' Get content offsets for table header, titles, footnotes, and content blanks.
#' Needed to calculate page breaks accurately.
#' @return A vector of upper and lower offsets
#' @noRd
get_content_offsets_html <- function(rs, ts, pi, content_blank_row, pgby_cnt = NULL) {
  
  cnt <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
  
  # Width is normally the width of the table, not the page
  wdth <- rs$content_size[["width"]]
  if (!is.null(pi$col_width))
    wdth <- sum(pi$col_width)
  
  # Default to zero
  shdrs <- list(lines = 0, twips = 0)
  hdrs <- list(lines = 0, twips = 0)
  
  
  if (ts$headerless == FALSE) {
    
    # Spanning headers now inside get_table_header_html
    hdrs <- get_table_header_html(rs, ts, pi)  
  }
  
  # Get title headers or titles
  if (is.null(ts$title_hdr))
    ttls <- get_titles_html(ts$titles, wdth, rs) 
  else 
    ttls <- get_title_header_html(ts$title_hdr, wdth, rs)
  
  # Get page by if it exists
  # pgb <- list(lines = 0, twips = 0)
  # if (!is.null(ts$page_by))
  #   pgb <- get_page_by_html(ts$page_by, wdth, NULL, rs, pi$table_align, pgby_cnt = pgby_cnt)
  # else if (!is.null(rs$page_by))
  #   pgb <- get_page_by_html(rs$page_by, wdth, NULL, rs, pi$table_align, pgby_cnt = pgby_cnt)
  page_by_info <- NULL
  if (!is.null(ts$page_by)) {
    page_by_info <- ts$page_by
  } else if (!is.null(rs$page_by)) {
    page_by_info <- rs$page_by
  }
  
  # Get lines for each page by value
  if (!is.null(page_by_info)) {
    cnt <- c(upper = list(), lower = 0, blank_upper = 0, blank_lower = 0)
    
    pgby_unique <- unique(pi$data$..page_by)
    
    for (i in 1:length(pgby_unique)) {
      pgby_temp <- get_page_by_html(page_by_info, wdth, pgby_unique[i], rs, pi$table_align, pgby_cnt = pgby_cnt)
      
      # Add everything up
      cnt[["upper"]][[i]] <- shdrs$lines + hdrs$lines + ttls$lines + pgby_temp$lines
    }
    
    names(cnt[["upper"]]) <- pgby_unique
    
  } else {
    pgb <- list(lines = 0)
    pgb <- get_page_by_html(page_by_info, wdth, NULL, rs, pi$table_align, pgby_cnt = pgby_cnt)
    cnt <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
    
    # Add everything up
    cnt[["upper"]] <- shdrs$lines + hdrs$lines + ttls$lines + pgb$lines
    
  }
  
  # Add everything up
  # cnt[["upper"]] <- shdrs$lines + hdrs$lines + ttls$lines + pgb$lines
  
  if (content_blank_row %in% c("above", "both")) {
    #ret[["blank_upper"]] <- rs$line_height
    cnt[["blank_upper"]] <- 1 
  }
  
  ftnts <- get_footnotes_html(ts$footnotes, wdth, rs) 
  rftnts <- get_footnotes_html(rs$footnotes, wdth, rs)
  
  if (has_top_footnotes(rs)) {
    cnt[["lower"]] <- ftnts$lines + rftnts$lines
  } else {
    cnt[["lower"]] <- ftnts$lines
  }
  
  # Add extra offsets if table has a lot of borders turned on
  # to avoid undesired page wraps
  if (any(ts$borders %in% c("all", "inside"))) {
    #ret[["lower"]] <- ret[["lower"]] + (rs$row_height * 2)
    cnt[["lower"]] <- cnt[["lower"]] + 1
  }
  
  if (content_blank_row %in% c("both", "below")) {
    cnt[["blank_lower"]] <- 1 
  }
  
  res <- list(lines = cnt)
  
  return(res)
  
}


#' @description Return a vector of strings for the table header
#' @details Basic idea of this function is to create a list
#' of string vectors of label words sized according to the column 
#' widths, then combine by line/row, and concatenate everything and justify.
#' @noRd
get_table_header_html <- function(rs, ts, pi, ex_brdr = FALSE) {
  
  ret <- c()
  cols <- c()
  cnt <- 0
  rh <- rs$row_height
  tbl <- ts$data

  
  widths <- pi$col_width[!is.na(pi$col_width)]
  lbls <- pi$label
  halgns <- pi$label_align
  talgn <- pi$table_align
  exclude_top <- NULL
  if (ex_brdr)
    exclude_top <- "top"
  
  
  #conv <- rs$twip_conversion
  nms <- names(lbls)
  
  u <- rs$units
  if (rs$units == "inches")
    u <- "in"
  
  # Get cell widths
  sz <- c()
  for (k in seq_along(widths)) {
    if (!is.control(nms[k])) {
        sz[k] <- paste0(widths[k], u)
    }
  }
  

  brdrs <- ts$borders
  if (length(ts$col_spans) > 0) {
    
    if (any(ts$borders %in% c("outside")))
      brdrs <- c("left", "right")
      
    if (any(ts$borders %in% "top"))
      brdrs <- brdrs[!brdrs %in% "top"]
    
  }
  
  # Header Cell alignment
  ha <- c()
  for (k in seq_along(halgns)) {
    if (!is.control(nms[k])) {
      if (halgns[k] == "left")
        ha[k] <- "tdl"
      else if (halgns[k] == "right")
        ha[k] <- "tdr"
      else if (halgns[k] %in% c("center", "centre"))
        ha[k] <- "tdc"
    }
  }
  
  # Table Header
  ret[1] <- "<tr>\n"
  cols[1] <- "<colgroup>\n"
  

  
  cnt <-  1 
  if (has_style(rs, "border_color")) {
    bd <- paste0("border-bottom: thin solid ", get_style(rs, "border_color"), ";")
  } else {
    bd <- "border-bottom: thin solid black;"
  }
  
  # Loop for column names
  pdf(NULL)
  par(family = get_font_family(rs$font), ps = rs$font_size)
  
  for(k in seq_along(widths)) {
    if (!is.control(nms[k])) {
      
      cols[1] <- paste0(cols[1], "<col style=\"width:", sz[k], ";\">\n")
      
      
      b <- get_cell_borders_html(1, k, 2, length(widths), brdrs, 
                                 exclude = exclude_top,
                                 border_color = get_style(rs, "border_color"))
      
      # Split label strings if they exceed column width
      tmp <- split_string_html(lbls[k], widths[k], rs$units)
      

      if (ts$header_bold)
        tstr <- paste0("<b>", encodeHTML(tmp$html), "</b>")
      else 
        tstr <- encodeHTML(tmp$html)

      if (b == "") {
        ret[1] <- paste0(ret[1], "<td class=\"thdr ", ha[k], "\">", 
                         tstr, "</td>\n")
      } else {
        ret[1] <- paste0(ret[1], "<td class=\"thdr ", ha[k], "\" ", 
                                 "style=\"", b, "\">", 
                         tstr, "</td>\n")
        
      }
      
      # Add in extra lines for labels that wrap
      xtr <- tmp$lines
      if (xtr > cnt)
         cnt <- xtr
    }
  }
  dev.off()
  
  cols[1] <- paste0(cols[1], "</colgroup>\n")
  ret[1] <- paste0(ret[1], "</tr>\n")
  
  if (ts$first_row_blank == TRUE) {
    
    sflg <- nms[1] == "stub" &  has_style(rs, "table_stub_background")

    tstub <- ""
    if (sflg)
      tstub <- " ts"

    if (sflg) {
      
      if (any(brdrs == "body")) {
        
        b1 <- get_cell_borders_html(2, 1, 3, 1, "left",
                                    border_color = get_style(rs, "border_color"), 
                                    stub_flag = TRUE)
        
        b2 <- get_cell_borders_html(2, 1, 3, 1, "right",
                                    border_color = get_style(rs, "border_color"), 
                                    stub_flag = FALSE)
        
      } else {
        
        b1 <- get_cell_borders_html(2, 1, 3, 1, brdrs,
                                   border_color = get_style(rs, "border_color"), 
                                   stub_flag = TRUE)
        
        b2 <- get_cell_borders_html(2, 1, 3, 1, brdrs,
                                    border_color = get_style(rs, "border_color"), 
                                    stub_flag = FALSE)
      
      }

      ret[1] <- paste0(ret[1], "<tr><td class=\"tc ts\"", 
                       "\" style=\"", b1, "\">&nbsp;</td>",
                       "<td class=\"tc\" colspan=\"", length(sz) - 1, 
                       "\" style=\"", b2, "\">&nbsp;</td></tr>")

    } else {
    
      if (any(brdrs == "body"))
        b <- get_cell_borders_html(2, 1, 3, 1, c("left", "right"),
                                   border_color = get_style(rs, "border_color"))
      else {
        b <- get_cell_borders_html(2, 1, 3, 1, brdrs,
                                   border_color = get_style(rs, "border_color"))
      }
  
      if (b == "") {
        ret[1] <- paste0(ret[1], "<tr><td class=\"tc\" colspan=\"", length(sz), 
                       "\">&nbsp;</td></tr>")
      } else {
        ret[1] <- paste0(ret[1], "<tr><td class=\"tc\" colspan=\"", length(sz), 
                         "\" style=\"", b, "\">&nbsp;</td></tr>")
      }
    }
    cnt <- cnt + 1
  }
  
  # Get spanning headers
  sphdrs <- get_spanning_header_html(rs, ts, pi,
                                     ifelse(is.null(exclude_top), FALSE, TRUE))
  
  res <- list(html = paste0(cols, "<thead>\n", paste0(sphdrs$html, collapse=""), 
                            ret, "</thead>\n"),
              lines = cnt + sphdrs$lines)
  
  return(res)
  
}


#' @description Return a vector of html strings for the table spanning headers
#' @details Basic idea of this function is to figure out which columns 
#' the header spans, add widths, then call get_table_header.  Everything
#' from there is the same.  
#' @import stats
#' @noRd
get_spanning_header_html <- function(rs, ts, pi, ex_brdr = FALSE) {
  
  spns <- ts$col_spans
  cols <- pi$keys
  cols <- cols[!is.controlv(cols)]
  w <- pi$col_width 
  w <- w[cols]
  gutter <- 0
  cnt <- c()
  exclude_top <- NULL
  if (ex_brdr)
    exclude_top <- "top"
  
  # print("Cols:")
  # print(cols)
  #print(w)
  wlvl <- get_spanning_info(rs, ts, pi, w, gutter)
  lvls <- sort(seq_along(wlvl), decreasing = TRUE)
  
  # At this point we have a data frame with labels, etc. and spanning
  # column widths, and are ready to create spanning header rows
  #print(wlvl)
  
  # Add information for gap
  if (length(wlvl) > 0) {
    wlvl <- get_spanning_gap_html(wlvl)
    
    # Do not create gaps for all and inside border
    if (any(ts$borders %in% c("all", "inside"))) {
      for (i in 1:length(wlvl)) {
        wlvl[[i]]$gap_width <- NA
      }
    }
  }
  
  # Get borders
  brdrs <- ts$borders
  
  brdrcolor <- get_style(rs, "border_color")
  if (brdrcolor == "")
    brdrcolor <- "black"
  
  # Format labels for each level
  ln <- c()
  for (l in lvls) {
    
    s <- wlvl[[l]]
   # print(s)
    widths <- s$width
    names(widths) <- s$name
    algns <- s$align
    names(algns) <- s$name
    lbls <- s$label
    names(lbls) <- s$name
    cs <- s$col_span
    
    # Header Cell alignment
    ha <- c()
    for (k in seq_along(algns)) {
      
      if (algns[k] == "right")
        ha[k] <- "text-align:right;"
      else if (algns[k] %in% c("center", "centre"))
        ha[k] <- "text-align:center;"
      else
        ha[k] <- "text-align:left;"
      
    }
    
    r <- ""
    cnt[length(cnt) + 1] <- 1 
    
    # Start row
    r <-  "<tr>\n"
    
    # Open device context
    pdf(NULL)
    par(family = get_font_family(rs$font), ps = rs$font_size)
    
    # Loop for labels
    for(k in seq_along(lbls)) {
      
      # Split label strings if they exceed column width
      tmp <- split_string_html(lbls[k], widths[k], rs$units)
      
      
      b <- get_cell_borders_html(length(lvls) - l + 1, k, length(lvls) + 1, 
                                 length(widths), brdrs, 
                                 exclude = exclude_top,
                                 border_color = get_style(rs, "border_color"))
      
      # Add colspans
      vl <- tmp$html
      
      # Special handling of borders for different situations.  
      # I hate this, but can't see a way around it.
      if (any(brdrs %in% c("all", "inside"))) {
        sflg <- "B"
        if (k > 1) {
          if (cs[k - 1] > 1)
            sflg <- ""
        }
        
        if (vl == "") {
          bb <- get_cell_borders_html(length(lvls) - l + 1, k, length(lvls), 
                                                 length(widths), brdrs, 
                                                 flag = sflg,
                                                 exclude = exclude_top,
                                      border_color = get_style(rs, "border_color"))
        } else 
          bb <- b
      } else if (all(brdrs == "outside")) {
        
        if (vl == "") {
          bb <- b
        } else if (s$underline[k] == TRUE) {
          bb <- paste0(b, "border-bottom:thin solid ", brdrcolor)
        } else {
          bb <- b
        }
      } else {
        
        if (vl == "") {
          bb <- get_cell_borders_html(length(lvls) - l + 1, k, length(lvls), 
                                                      length(widths), brdrs, 
                                                      flag = "",
                                                      exclude = c("bottom", exclude_top),
                                      border_color = get_style(rs, "border_color"))
        } else if (s$underline[k] == TRUE) {
          bb <- paste0(b, "border-bottom:thin solid ", brdrcolor)
        } else {
          bb <- b
        }
      }
      
      if (s$bold[k])
        tstr <- paste0("<b>", encodeHTML(vl), "</b>")
      else 
        tstr <- encodeHTML(vl)
      
      # Check gap information
      gap <- ""
      if (!is.na(s$gap_width[k])){
        # This div appears to create a small white gap in the black bottom border of the cell
        gap <- sprintf(
          "<div style=\"position:absolute;left:-3px;width:%sin;height:1px;bottom:-1px;background-color:white;\"></div>",
          s$gap_width[k]
        )
        
        #  Sets the positioning context for absolute positioning within the cell
        bb <- paste0(bb, ";position:relative;")
      }
      
      r <- paste0(r, "<td class=\"shdr\" colspan=\"", cs[k], 
                  "\" style=\"vertical-align:bottom;", ha[k], bb, "\">",
                  gap, tstr, "</td>\n")
      # print(lbls[k])
      # print(widths[k])
      # Add in extra lines for labels that wrap
      xtr <- tmp$lines
      
      if (xtr > cnt[length(cnt)])
        cnt[length(cnt)] <- xtr
      
    }
    dev.off()
    
    
    r <- paste0(r, "</tr>\n")
    
    ln[[length(ln) + 1]] <- r
    
  }
  
  
  ret <- unlist(ln)
  
  res <- list(html = ret, 
              lines = sum(cnt))
  
  return(res)
}

#' @description Check gap and insert gap information for HTML
#' @details Check gap and insert gap information for HTML
#' @noRd
get_spanning_gap_html <- function(wlvl, gap_twips = 100){
  
  ret <- list()
  
  for (k in 1:length(wlvl)) {
    
    df <- wlvl[[k]]
    
    # Prepare gap variables
    df$gap_width <- rep(NA, nrow(df))
    
    # Extract spanning and underline only for detection
    df_span <- df[df$span > 0 & df$underline == TRUE,] 
    
    if (nrow(df_span) < 2) {
      ret[[k]] <- df
    } else {
      # Convert twips to inch (Default 100 twips for consistency)
      gap_width <- gap_twips/1440
      
      df_span$order <- as.numeric(stri_extract_first_regex(df_span$name, "\\d+"))
      
      for (i in 2:nrow(df_span)) {
        
        cur_order <- df_span$order[i]
        
        # Gap is processed only for next cell
        pre_order <- df_span$order[i - 1]
        
        # Continuous underline detection
        if (cur_order - pre_order == 1) {
          
          # Add indenting information for gap
          df_span$gap_width[i] <- gap_width
        }
        
      } # End of data row loop
      
      # Drop order
      df_span <- df_span[, setdiff(names(df_span), "order")]
      
      # Stack together and sort
      df_ret <- rbind(df_span, df[!(df$span > 0 & df$underline == TRUE),])
      df_ret <- df_ret[order(df_ret$name),]
      
      ret[[k]] <- df_ret
      
    } # End of gap process
  }
  
  return(ret)
}

#' @description This function counts lines per row independently because
#' the ..row field does not account for page wrapping.  Need number
#' of lines on this particular page.
#' @noRd
get_table_body_html <- function(rs, tbl, widths, algns, talgn, tbrdrs, 
                                ex_brdr = FALSE, frb = FALSE, styles,
                                ts) {
  
  if ("..blank" %in% names(tbl))
    flgs <- tbl$..blank
  else 
    flgs <- NA
  
  # Count lines per row
  rws <- c()
  border_flag <- FALSE
  exclude_top <- NULL
  if (ex_brdr)
    exclude_top <- "top"
  
  
  nms <- names(widths)
  nms <- nms[!is.na(nms)]
  nms <- nms[!is.controlv(nms)]
  wdths <- widths[nms]
  if (length(nms) == 1) {
    t <- as.data.frame(tbl[[nms]])
    names(t) <- nms
  } else 
    t <- tbl[ , nms]
  
  brdrs <- tbrdrs
  if (all(tbrdrs == "body"))
      brdrs <- c("top", "bottom", "left", "right")

  # Cell alignment
  ca <- c()
  castr <- c()
  for (k in seq_along(algns)) {
    if (!is.control(nms[k])) {
      if (is.na(nms[k]) | is.null(nms[k])) {
        tcls <- ""
      } else {
        if (nms[k] == "stub")
          tcls <- "ts"
        else 
          tcls <- ""
      }
      
      if (algns[k] == "left")
        ca[k] <- paste0("tc tdl ", tcls)
      else if (algns[k] == "right")
        ca[k] <- paste0("tc tdr ", tcls)
      else if (algns[k] %in% c("center", "centre"))
        ca[k] <- paste0("tc tdc ", tcls)
      
      
      if (algns[k] == "left")
        castr[k] <- paste0("tc tdl tbstr ", tcls)
      else if (algns[k] == "right")
        castr[k] <- paste0("tc tdr tbstr ", tcls)
      else if (algns[k] %in% c("center", "centre"))
        castr[k] <- paste0("tc tdc tbstr ", tcls)
    }
  }
  
  ret <- c()
  
  strpmod <- 1
  if (frb)
    strpmod <- 0
  
  merge_label_row <- TRUE
  if (has_style(rs, "table_stub_background") & "stub" %in% nms)
    merge_label_row <- FALSE
  
  pdf(NULL)
  par(family = get_font_family(rs$font), ps = rs$font_size)
  
  # For break labels, prepare the break label columns for later use
  defs <- ts$col_defs
  break_label_col <- c()
  if (any(grepl("..break_label", names(tbl)))) {
    
    for (d in 1:length(defs)) {
      if (!is.null(defs[[d]]$break_label)){
        break_label_col <- c(break_label_col, defs[[d]]$var_c)
      }
    }
  }
  
  # Table Body
  for(i in seq_len(nrow(t))) {
    
    
    if (i == 1)
      ret[i] <- "<tbody>\n<tr>"
    else
      ret[i] <- "<tr>"
    
    mxrw <- 1
    
    # Loop for cell values
    for(j in seq_len(ncol(t))) {
      
      
      if (!is.control(nms[j])) {
        
        sflg <- nms[j] == "stub" &  has_style(rs, "table_stub_background")
        
        cell_border <- NULL
        if ("..group_border" %in% names(tbl)) {
          cell_border <- tbl[["..group_border"]][i]
        }
        
        # Don't draw group line for first blank row
        if (flgs[i] %in% c("B", "A") & i == 1){
          cell_border <- NULL
        }
        
        b <- get_cell_borders_html(i, j, nrow(t), ncol(t), brdrs, flgs[i], 
                                   exclude = exclude_top, 
                                   border_color = get_style(rs, "border_color"),
                                    stub_flag = sflg, cell_border = cell_border)
        
        # Put indent information into style
        if (rs$units == "inches") {
          ind_unit <- "in"
        } else if (rs$units == "cm") {
          ind_unit <- "cm"
        }
        pad <- ""
        
        # For break labels, use the indenting info from original column
        repeat_break_label <- FALSE
        if (any(grepl("..break_label", names(tbl)))) {
          if (!is.na(tbl$..break_occur[i]) & flgs[i] == "L") {
            cur_break_label_col <- break_label_col[as.numeric(tbl$..break_occur[i])]
            if (!is.null(defs[[cur_break_label_col]]$indent)) {
              pad <- sprintf("padding-left: %s%s;",
                             defs[[cur_break_label_col]]$indent,
                             ind_unit)
            }
            repeat_break_label <- TRUE
          }
        }
        
        if (!repeat_break_label) {
          if (!is.null(defs[[nms[j]]]$indent)) {
            
            pad <- sprintf("padding-left: %s%s;",
                           defs[[nms[j]]]$indent,
                           ind_unit)
            
          } else if (nms[j] == "stub" & !is.null(ts$stub)) {
            
            stub_var <- tbl$..stub_var[i]
            if (!is.null(defs[[stub_var]]$indent)) {
              pad <- sprintf("padding-left: %s%s;",
                             defs[[stub_var]]$indent,
                             ind_unit)
            }
          }
        }
        
        b <- paste0(b, pad)
        
        
        lrflg <- ""
        if ( nms[j] == "stub" & flgs[i] == "L")
          lrflg <- " tlr"
        
        # Stripe class
        if (i %% 2 == strpmod) {
          scls <- ca[j]
        } else {
          scls <- castr[j]
        }
        
        vl <- t[i, j]
        if (all(class(vl) != "character"))
          vl <- as.character(vl)
        else 
          vl <- encodeHTML(vl)
        
        if (merge_label_row  & flgs[i] %in% c("B", "A", "L")) {
          if (j == 1) {
            
            # # Strip out line feeds for label rows
            # vl <- gsub("\n", " ", vl, fixed = TRUE)
            # 
            # # Recalculate based on total width of table
            # tmp <- split_string_html(vl, sum(wdths), rs$units)
            # 
            # vl <- tmp$html
            
            stl <- get_cell_styles(nms[j], styles, flgs, i, tbl)
            
            if ("bold" %in% stl) {
              vl <- paste0("<b>", vl, "</b>") 
            }
            
          
            if (b == "") {
              
              ret[i] <- paste0(ret[i], "<td class=\"", scls, lrflg, "\"",
                               " colspan = \"", ncol(t), "\">", 
                               vl, "</td>")
            } else { 
              
              ret[i] <- paste0(ret[i], "<td class=\"", scls, lrflg, "\"",
                               " colspan = \"", ncol(t), "\"",
                               " style=\"", b, "\">", 
                               vl, "</td>")
            }
          
          
          }
        } else {
          
          stl <- get_cell_styles(nms[j], styles, flgs, i, tbl)
          
          if ("bold" %in% stl) {
            vl <- paste0("<b>", vl, "</b>") 
          } 
          
          # Construct html
          if (b == "") {
  
            ret[i] <- paste0(ret[i], "<td class=\"", scls, lrflg, "\">", 
                             vl, "</td>")
          } else { 
            
            ret[i] <- paste0(ret[i], "<td class=\"", scls, lrflg, 
                             "\" style=\"", b, "\">", 
                             vl, "</td>")
          }
        }
        
        # Count lines in cell
        cl <- strsplit(vl, "<br>", fixed = TRUE)[[1]]
        if (length(cl) > mxrw)
          mxrw <- length(cl) 
      }
      
    }
    
    rws[i] <- mxrw
    
    if (i == nrow(t))
      ret[i] <- paste0(ret[i], "</tr>\n</tbody>")
    else 
      ret[i] <- paste0(ret[i], "</tr>")
    
    
  }
  
  dev.off()
  
  if ("bottom" %in% get_outer_borders(brdrs))
    border_flag <- TRUE
  
  res <- list(html = ret,
              lines = sum(rws), 
              border_flag = border_flag)
  
  return(res)
  
  
}

#' @description Have to check wrapping on a lot of files.  May have unintended 
#' results.
#' @noRd
encodeHTML <- function(strng) {
  
  ret <- strng
  
  ret <- gsub("&", "&amp;", ret , fixed = TRUE)
  ret <- gsub(">", "&gt;", ret , fixed = TRUE)
  ret <- gsub("<", "&lt;", ret , fixed = TRUE)
  ret <- gsub("\n", "<br>", ret , fixed = TRUE)
  ret <- gsub(" ", "&nbsp;", ret, fixed = TRUE)
  if (ret == "")
    ret <- "&nbsp;"
  
  return(ret)
}

# Create Tables ----------------------------------------------------------


#' @noRd
create_table_pages_pdf <- function(rs, cntnt, lpg_rows) {
  
  ts <- cntnt$object
  content_blank_row <- cntnt$blank_row
  
  pgby_var <- NA
  pgby_cnt <- 0
  pgby_fmt <- NULL
  if (!is.null(rs$page_by)) {
    pgby_var <- rs$page_by$var
    pgby_fmt <- rs$page_by$format
  } else if (!is.null(ts$page_by)) {
    pgby_var <- ts$page_by$var
    pgby_fmt <- ts$page_by$format
  }
  
  
  if (all(ts$show_cols == "none") & length(ts$col_defs) == 0) {
    
    stop("ERROR: At least one column must be defined if show_cols = \"none\".")
  }
  
  font_name <- rs$font
  
  # Set up control columns
  dat <- as.data.frame(ts$data, stringsAsFactors = FALSE)  
  dat$..blank <- ""
  dat$..row <- 1
  dat$..page_by <- NA
  
  # If page_break variable has been defined, use it
  if (is.null(ts$page_var)) {
    if (is.na(pgby_var))
      dat$..page <- NA
    else 
      dat$..page <-  dat[[pgby_var]]
    
  } else {
    if (any(class(dat[[ts$page_var]]) == "factor"))
      dat$..page <- as.character(dat[[ts$page_var]])
    else 
      dat$..page <- dat[[ts$page_var]]
  }
  
  # If page by is defined, use it
  if (!is.na(pgby_var)) {
    
    # Clear out factors on page by if they exist
    # Needed because split cells used to convert everything to characters
    if (any(class(dat[[pgby_var]]) == "factor")) {
      dat[[pgby_var]] <- as.character(dat[[pgby_var]] )
      dat$..page_by <- dat[[pgby_var]] 
      
      if (any(class(dat$..page) == "factor")) 
        dat$..page <- as.character(dat[[pgby_var]])
      
    } else {
      
      if (!is.null(pgby_fmt))
        dat$..page_by <-  fapply(dat[[pgby_var]], pgby_fmt)
      else 
        dat$..page_by <- dat[[pgby_var]]
    }
    
    pgby_cnt <- get_pgby_cnt(dat$..page_by)
    
    # Commenting out for now
    # if (is.unsorted(dat[[pgby_var]], strictly = FALSE))
    #   message("Page by variable not sorted.")
  }
  
  # Deal with invisible columns
  if (!is.null(ts$col_defs)) {
    for (def in ts$col_defs) {
      if (def$visible == FALSE) {
        nnm <- paste0("..x.", def$var_c)
        dat[[nnm]] <- dat[[def$var_c]]
      }
    }
    
    # Deal with Group Cohesion
    # Prepare ..group_cohesion and ..min_page_prop
    group_cohesion <- c()
    min_page_prop <- c()
    
    i <- 1
    for (def in ts$col_defs) {
      if (def$group_cohesion == TRUE) {
        group_cohesion <- c(group_cohesion, def$var_c)
        min_page_prop <- c(min_page_prop, def$min_page_prop)
        
        dat$..temp <- dat[[def$var_c]]
        names(dat)[names(dat) == "..temp"] <-  paste0("..group_cohesion",i)
        
        dat$..temp <- def$min_page_prop
        names(dat)[names(dat) == "..temp"] <-  paste0("..min_page_prop",i)
        i <- i + 1
      }
    }
    
    if (length(group_cohesion) == 0) {
      group_cohesion <- NULL
    }
    
    if (length(min_page_prop) == 0) {
      min_page_prop <- NULL
    }
    
    # Prepare empty ..break_label and ..break_label_lines for later use
    i <- 1
    for (def in ts$col_defs) {
      if (!is.null(def$break_label)) {
        
        dat$..temp <- ifelse(!is.na(dat[[def$var_c]]) & dat[[def$var_c]] != "",
                             paste0(dat[[def$var_c]], " ", def$break_label),
                             dat[[def$var_c]])
        names(dat)[names(dat) == "..temp"] <-  paste0("..break_label",i)
        
        dat$..temp <- rep(NA, nrow(dat))
        names(dat)[names(dat) == "..temp"] <-  paste0("..break_label_lines",i)
        i <- i + 1
      }
    }
    # Prepare repeated label occurrence for later use
    if (any(grepl("..break_label", names(dat)))) {
      dat$..break_occur <- rep(NA, nrow(dat))
    }
  }
  
  # Get control column names 
  control_cols <- names(dat)[is.controlv(names(dat))]
  
  # Get vector of all included column names
  # Not all columns in dataset are necessarily included
  # depends on show_cols parameter on create_table and
  # visible parameter on column definitions
  keys <- get_table_cols(ts, control_cols)
  # print("keys")
  # print(keys)
  
  # Filter dataset by included columns
  dat <- get_data_subset(dat, keys, rs$preview)
  # print("Key columns:")
  # print(dat)
  
  # Update column definitions with column defaults
  ts$col_defs <- set_column_defaults(ts, keys)
  # print("col_defs:")
  # print(ts$col_defs)
  
  # Deal with styles
  styles <- get_styles(ts)
  
  # Get labels
  labels <- get_labels(dat, ts)
  # print("Labels:")
  # print(labels)
  
  # Get column alignments
  aligns <- get_aligns(dat, ts)
  
  # Get alignment for labels
  # Follows column alignment by default
  label_aligns <- get_label_aligns(ts, aligns)
  # print("Label Aligns:")
  # print(label_aligns)
  
  # Clear out existing formats
  cdat <- clear_formats(dat)
  
  # Get column formats
  formats(cdat) <- get_col_formats(dat, ts)
  # print("formats:")
  # print(formats(cdat))
  
  # Apply formatting
  fdat <- fdata(cdat)
  # print("fdata:")
  # print(fdat)
  
  # Prep data for blank lines, indents, and stub columns
  fdat <- prep_data(fdat, ts, rs$char_width, rs$missing) # OK
  # print("prep_data")
  # print(fdat)
  # str(fdat)
  
  # Add stub_var into control_cols for adding indentation
  if ("..stub_var" %in% names(fdat)) {
    control_cols <- c(control_cols, "..stub_var")
  }
  
  if ("..group_border" %in% names(fdat)) {
    control_cols <- c(control_cols, "..group_border")
  }
  
  # Reset keys, since prep_data can add/remove columns for stub
  keys <- names(fdat)
  # print("Keys")
  # print(keys)
  # print("Aligns")
  # print(aligns)
  
  # Copy any width attributes to formatted data frame
  if ("width" %in% ts$use_attributes)
    widths(fdat) <- widths(dat)
  # print("Original Widths")
  # print(widths(dat))
  
  
  # Get column widths
  widths_uom <- get_col_widths_variable(fdat, ts, labels, 
                                        rs$font, rs$font_size, rs$units, 
                                        rs$gutter_width) 
  # print("Widths UOM")
  # print(widths_uom)
  
  # Split long text strings into multiple rows. Number of rows are stored in
  # ..row variable. If too slow, may need to be rewritten in C
  sp <- split_cells_variable(fdat, widths_uom, rs$font, 
                               rs$font_size, rs$units, rs$output_type, rs$char_width, ts) 
  fdat <- sp$data
  wdat <- sp$widths
  
  # print("split_cells")
  # print(fdat)
  
  
  # Break columns into pages
  wraps <- get_page_wraps(rs$line_size, ts, 
                          widths_uom, 0, control_cols)  # No gutter width for RTF
  # print("wraps")
  # print(wraps)
  
  
  # Create a temporary page info to pass into get_content_offsets
  tmp_pi <- list(keys = keys, col_width = widths_uom, label = labels,
                 label_align = label_aligns, table_align = cntnt$align, data = fdat)
  # print("Temp PI")
  # print(tmp_pi)
  
  
  # Offsets are needed to calculate splits and page breaks
  content_offset <- get_content_offsets_pdf(rs, ts, tmp_pi, 
                                            content_blank_row, pgby_cnt)
  
  # Need to do something with widths
  # split rows
  splits <- get_splits_text(fdat, widths_uom, rs$body_line_count, 
                            lpg_rows, content_offset$lines, ts, TRUE)  
  # print("splits")
  # print(splits)
  
  # Subset splits by preview, if requested
  if (!is.null(rs$preview)) {
    if (rs$preview < length(splits))
      splits <- splits[seq(1, rs$preview)] 
  }
  
  tot_count <- length(splits) * length(wraps)
  counter <- 0
  wrap_flag <- FALSE
  blnk_ind <- "none"
  spstart <- 1
  spend <- 1
  fp_offset <- lpg_rows
  
  pg_lst <- list()
  for(s in splits) {
    
    # Subset text widths by current split rows
    #spend <- spstart + nrow(s) - 1
    
    # If the data is changed because of break labels, grab the original row labels
    break_num <- NULL
    if (any(grepl("break_label", rownames(s)))) {
      rnms <- rownames(s)[!grepl("break_label", rownames(s))]
      break_num <- sum(grepl("break_label", rownames(s)))
    } else {
      rnms <- rownames(s)
    }
    
    spstart <- rnms[1]
    spend <- rnms[length(rnms)]
    spwidths <- wdat[seq(spstart, spend)]
    
    # Add additional objects for width data to align with new break label data
    if (!is.null(break_num)) {
      for (b in 1:break_num) {
        spwidths <- append("", spwidths)
      }
    }
    
    
    for(pg in wraps) {
      counter <- counter + 1
      
      if (counter < tot_count) {
        wrap_flag <- TRUE
      } else {
        wrap_flag <- FALSE
      } 
      
      #print(s)
      if (!is.na(pgby_var)) {
        pgby <- trimws(s[1, "..page_by"])
      } else {
        pgby <- NULL
      } 
      
      # Ensure content blank rows are added only to the first and last pages
      blnk_ind <- get_blank_indicator(counter, tot_count, content_blank_row,
                                      rs$body_line_count, content_offset$lines, 
                                      nrow(s), pgby)
      #print(blnk_ind)
      
      pi <- page_info(data= s[, pg], keys = pg, label=labels[pg],
                      col_width = widths_uom[pg], col_align = aligns[pg],
                      font_name = font_name, label_align = label_aligns[pg],
                      pgby, cntnt$align)
      pg_lst[[length(pg_lst) + 1]] <- create_table_pdf(rs, ts, pi, 
                                                       blnk_ind, wrap_flag,
                                                       fp_offset, spwidths, styles)
      
      fp_offset <- 0
    }
    
    #spstart <- spend + 1
  }
  
  ret <- list(widths = widths_uom, page_list = pg_lst)
  
  return(ret)
  
  
}



#' @noRd
create_table_pdf <- function(rs, ts, pi, content_blank_row, wrap_flag, 
                             lpg_rows, spwidths, styles) {
  rh <- rs$row_height
  ys <- lpg_rows * rh
  conv <- rs$point_conversion
  bf <- FALSE
  
  # Default to content width
  ls <- rs$content_size[["width"]]
  
  ys <- sum(ys, rs$page_template$page_header$points, 
            rs$page_template$titles$points, rs$page_template$title_hdr$points)
  
  # Get table width
  if (!is.null(pi$col_width))
    ls <- sum(pi$col_width, na.rm = TRUE)
  
  if (content_blank_row %in% c("above", "both"))
    ys <- ys + rh
  
  if (!is.null(ts$title_hdr)) {
    ttls <- get_title_header_pdf(ts$title_hdr, ls, rs, pi$table_align, 
                                 ystart = ys)
    
  } else {
    ttls <- get_titles_pdf(ts$titles, ls, rs, pi$table_align, ystart = ys) 
  }
  
  ys <-  ys + ttls$points 
  if (ttls$lines > 0)
    bf <- ttls$border_flag
  
  if (!is.null(rs$page_by)) {
    pgby <- get_page_by_pdf(rs$page_by, rs$content_size[["width"]], 
                            pi$page_by, rs, pi$table_align, ystart = ys,
                            brdr_flag = bf)
  } else if(!is.null(ts$page_by)) {
    pgby <- get_page_by_pdf(ts$page_by, ls, pi$page_by, rs, pi$table_align, 
                            ystart = ys, brdr_flag = bf)
  } else {
    pgby <- c()
  } 
  
  if (length(pgby) > 0) {
    ys <- ys + pgby$points
    bf <- pgby$border_flag 
  }
  
  shdrs <- list(lines = 0, points = 0)
  hdrs <- list(lines = 0, points = 0)
  
  if (ts$headerless == FALSE) {
    shdrs <- get_spanning_header_pdf(rs, ts, pi, ystart = ys,
                                     brdr_flag = bf)
    
    ys <- ys + shdrs$points
    
    hs <- FALSE
    if (shdrs$lines > 0) {
      bf <- shdrs$border_flag
      hs <- TRUE
    }
    
    hdrs <- get_table_header_pdf(rs, ts, pi$col_width, 
                                 pi$label, pi$label_align, 
                                 pi$table_align, ystart = ys,
                                 brdr_flag = bf, has_spans = hs)
    
    ys <- ys + hdrs$points
  }
  
  # rs, ts, widths,  algns, halgns, talgn
  bdy <- get_table_body_pdf(rs, pi$data, pi$col_width, 
                            pi$col_align, pi$table_align, ts$borders,
                            ystart = ys,
                            spwidths, frb = ts$first_row_blank, styles = styles,
                            ts = ts)
  ys <- ys + bdy$points


  if (bdy$lines > 0)
    bf <- bdy$border_flag
  
  # blnks <- c()
  
  # Get row count by summing ..row variable
  # if ("..row" %in% names(pi$data)) {
  #   rcnt <- sum(pi$data$..row)
  # } else {
  #   rcnt <- nrow(pi$data) 
  # }
  
  
  # Get footnotes, passing in sum of all current lines
  ftnts <- get_page_footnotes_pdf(rs, ts, ls, lpg_rows, ys,
                                  wrap_flag, content_blank_row,  pi$table_align,
                                  brdr_flag = bf)
  
  # Deal with cell padding.  Don't count this in line count.
  # cp <- paste0("\\li", rs$cell_padding, "\\ri", rs$cell_padding, rs$spacing_multiplier)
  
  # # On LibreOffice, have to protect the table from the title width or
  # # the table row will inherit the title row width. Terrible problem.
  # tpt <- "{\\pard\\fs1\\sl0\\par}"
  # if (any(ts$borders %in% c("all", "top", "outside"))) {
  #   if (ttls$border_flag | rs$page_template$titles$border_flag |  
  #       rs$page_template$title_hdr$border_flag)
  #     tpt <- ""
  #   
  #   if (length(pgby) > 0) {
  #     if (pgby$border_flag)
  #       tpt <- ""
  #   }
  # }
  # 
  # # Same thing as above with titles.  If footnote block is contiguous
  # # with table and footnotes are wider than table, row width of footnotes
  # # will infect row width of table.  On libre only.  So this is to protect
  # # the table width.
  # bpt <- "{\\pard\\fs1\\sl0\\par}"
  # if (any(ts$borders %in% c("all", "top", "outside", "body"))) {
  #   if (!is.null(ftnts)) {
  #     if (ftnts$border_flag)
  #       bpt <- ""
  #   }
  #   
  #   if (!is.null(rs$page_template$footnotes)) {
  #     if (rs$page_template$footnotes$border_flag)
  #       bpt <- ""
  #   }
  # }
  
  # Determine sum of all lines
  rc <- sum(ttls$lines, pgby$lines, shdrs$lines, 
            hdrs$lines, bdy$lines, ftnts$lines)
  
  ret <- list(pdf = c(ttls$pdf, pgby$pdf, shdrs$pdf, 
                      hdrs$pdf, bdy$pdf, ftnts$pdf),
              lines = rc,
              points = rc * rh)
  
  return(ret) 
}

get_page_footnotes_pdf <- function(rs, spec, spec_width, lpg_rows, ystart,
                                   wrap_flag, content_blank_row, talgn,
                                   brdr_flag = FALSE) {
  
  ftnts <- list(lines = 0, twips = 0, border_flag = FALSE)
  vflag <- "none"
  
  fl <- rs$page_template$page_footer$lines
  
  # Deal with valign parameter
  if (!is.null(spec$footnotes)) {
    if (!is.null(spec$footnotes[[length(spec$footnotes)]])) {
      if (spec$footnotes[[length(spec$footnotes)]]$valign == "bottom") {
        
        vflag <- "bottom"
        ftnts <- get_footnotes_pdf(spec$footnotes, 
                                   spec_width, rs, 
                                   talgn, footer_lines = fl) 
      } else {
        vflag <- "top"
        ftnts <- get_footnotes_pdf(spec$footnotes, spec_width, rs, talgn,
                                   ystart = ystart,
                                   brdr_flag = brdr_flag) 
      }
      
    }
  } else {
    
    if (!is.null(rs$footnotes[[1]])) {
      if (!is.null(rs$footnotes[[1]]$valign)) {
        if (rs$footnotes[[1]]$valign == "top") {
          vflag <- "top"
          ftnts <- get_footnotes_pdf(rs$footnotes, 
                                     spec_width, rs, 
                                     talgn, ystart = ystart,
                                     brdr_flag = brdr_flag) 
        } else {
          
          if (wrap_flag)
            vflag <- "bottom" 
        }
      }
    }
  }
  

  blen <- 0
  if (content_blank_row %in% c("below", "both")) {
    blen <- 1
  }
  # } else {
  #   b <- paste0("\\fs1\\sl0\\par\\pard", rs$font_rtf, rs$spacing_multiplier)
  # }
  
  # Add extra offsets if table has a lot of borders turned on
  # to avoid undesired page wraps
  # boff <- 0
  # if (any(class(spec) == "table_spec") &
  #     any(spec$borders %in% c("all", "inside"))) {
  #   
  #   boff <- round(row_count * rs$border_height / rs$row_height)
  # }
  
  tlns <- ftnts$lines + blen
  ret <- list(pdf = ftnts$pdf,
              lines = tlns,
              points = tlns * rs$line_height,
              border_flag = ftnts$border_flag)
  
  return(ret)
}

# Sub-Functions ---------------------------------------------------------


#' Get content offsets for table header, titles, footnotes, and content blanks.
#' Needed to calculate page breaks accurately.
#' @return A vector of upper and lower offsets
#' @noRd
get_content_offsets_pdf <- function(rs, ts, pi, content_blank_row, pgby_cnt = NULL) {
  
  ret <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
  cnt <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
  
  # Width is normally the width of the table, not the page
  wdth <- rs$content_size[["width"]]
  if (!is.null(pi$col_width))
    wdth <- sum(pi$col_width)
  
  # Default to zero
  shdrs <- list(lines = 0, points = 0)
  hdrs <- list(lines = 0, points = 0)
  
  if (ts$headerless == FALSE) {
    shdrs <- get_spanning_header_pdf(rs, ts, pi)   
    
    hdrs <- get_table_header_pdf(rs, ts, pi$col_width, 
                                 pi$label, pi$label_align, 
                                 pi$table_align)  
  }
  
  # Get title headers or titles
  if (is.null(ts$title_hdr))
    ttls <- get_titles_pdf(ts$titles, wdth, rs) 
  else 
    ttls <- get_title_header_pdf(ts$title_hdr, wdth, rs)
  
  # Get page by if it exists
  # if (!is.null(ts$page_by))
  #   pgb <- get_page_by_pdf(ts$page_by, wdth, NULL, rs, pi$table_align, pgby_cnt = pgby_cnt)
  # else if (!is.null(rs$page_by))
  #   pgb <- get_page_by_pdf(rs$page_by, wdth, NULL, rs, pi$table_align, pgby_cnt = pgby_cnt)
  
  
  #print(length(pgb))
  
  # print(paste("Table titles:", ttls))
  
  # Add everything up
  # ret[["upper"]] <- shdrs$points + hdrs$points + ttls$points + pgb$points
  # cnt[["upper"]] <- shdrs$lines + hdrs$lines + ttls$lines + pgb$lines
  
  page_by_info <- NULL
  if (!is.null(ts$page_by)) {
    page_by_info <- ts$page_by
  } else if (!is.null(rs$page_by)) {
    page_by_info <- rs$page_by
  }
  
  # Get lines for each page by value
  if (!is.null(page_by_info)) {
    ret <- c(upper = list(), lower = 0, blank_upper = 0, blank_lower = 0)
    cnt <- c(upper = list(), lower = 0, blank_upper = 0, blank_lower = 0)
    
    pgby_unique <- unique(pi$data$..page_by)
    
    for (i in 1:length(pgby_unique)) {
      pgby_temp <- get_page_by_pdf(page_by_info, wdth, pgby_unique[i], rs, pi$table_align, pgby_cnt = pgby_cnt)
      
      # Add everything up
      ret[["upper"]][[i]] <- shdrs$points + hdrs$points + ttls$points + pgby_temp$points
      cnt[["upper"]][[i]] <- shdrs$lines + hdrs$lines + ttls$lines + pgby_temp$lines
    }
    
    names(ret[["upper"]]) <- pgby_unique
    names(cnt[["upper"]]) <- pgby_unique
    
  } else {
    pgb <- list(lines = 0, points = 0)
    pgb <- get_page_by_pdf(page_by_info, wdth, NULL, rs, pi$table_align, pgby_cnt = pgby_cnt)
    ret <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
    cnt <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
    
    # Add everything up
    ret[["upper"]] <- shdrs$points + hdrs$points + ttls$points + pgb$points
    cnt[["upper"]] <- shdrs$lines + hdrs$lines + ttls$lines + pgb$lines
    
  }
  
  if (content_blank_row %in% c("above", "both")) {
    ret[["blank_upper"]] <- rs$line_height
    cnt[["blank_upper"]] <- 1 
  }
  
  ftnts <- get_footnotes_pdf(ts$footnotes, wdth, rs) 
  rftnts <- get_footnotes_pdf(rs$footnotes, wdth, rs)
  
  if (has_top_footnotes(rs)) {
    ret[["lower"]] <- ftnts$points + rftnts$points
    cnt[["lower"]] <- ftnts$lines + rftnts$lines
  } else {
    ret[["lower"]] <- ftnts$points
    cnt[["lower"]] <- ftnts$lines
  }
  
  # Add extra offsets if table has a lot of borders turned on
  # to avoid undesired overlay
  # Not perfect but good enough
  brdrs <- strip_borders(ts$borders )
  if (any(brdrs %in% c("all", "inside"))) {
    if (is.null(page_by_info)) {
      
      epnts <- floor(rs$body_line_count - cnt[["lower"]] - cnt[["upper"]]) * rs$border_height 
      ret[["lower"]] <- ret[["lower"]] + epnts - rs$line_height
      cnt[["lower"]] <- cnt[["lower"]] + floor(epnts / rs$row_height) - 1
      
    } else {
      
      # Convert lower to list because of different upper
      ret_lower <- ret[["lower"]]
      cnt_lower <- cnt[["lower"]]
      ret[["lower"]] <- list()
      cnt[["lower"]] <- list()
      
      for (i in 1:length(pgby_unique)) {
        epnts <- floor(rs$body_line_count - cnt_lower - cnt[["upper"]][[i]]) * rs$border_height 
        ret[["lower"]][[i]] <- ret_lower + epnts - rs$line_height
        cnt[["lower"]][[i]] <- cnt_lower + floor(epnts / rs$row_height) - 1
      }
      
      names(ret[["lower"]]) <- pgby_unique
      names(cnt[["lower"]]) <- pgby_unique
    }
  }
  
  if (content_blank_row %in% c("both", "below")) {
    ret[["blank_lower"]] <- rs$line_height
    cnt[["blank_lower"]] <- 1 
  }
  
  res <- list(lines = cnt,
              points = ret)
  
  return(res)
  
}


#' @description Return a vector of strings for the table header
#' @details Basic idea of this function is to create a list
#' of string vectors of label words sized according to the column 
#' widths, then combine by line/row, and concatenate everything and justify.
#' @noRd
get_table_header_pdf <- function(rs, ts, widths, lbls, halgns, talgn, 
                                 ystart = 0, brdr_flag = FALSE, 
                                 has_spans = FALSE) {
  
  ret <- c()
  cnt <- 0
  rh <- rs$row_height 
  bs <- rs$border_spacing
  bh <- rs$border_height
  tbl <- ts$data
  conv <- rs$point_conversion
  nms <- names(lbls)[is.controlv(names(lbls)) == FALSE]
  unts <- rs$units
  wdths <- widths[nms]
  brdrs <- strip_borders(ts$borders)
  pnts <- 0
  # print("ystart header")
  # print(ystart)
  
  # Sum up widths 
  width <- sum(wdths, na.rm = TRUE)
  
  # Get content alignment codes
  if (talgn == "right") {
    tlb <- rs$content_size[["width"]] - width
    trb <- rs$content_size[["width"]]
  } else if (talgn %in% c("center", "centre")) {
    tlb <- (rs$content_size[["width"]] - width) / 2
    trb <- width + tlb
  } else {
    tlb <- 0
    trb <- width
  }

  cnt <-  1
 
  # Loop for column names
  pdf(NULL)
  par(family = get_font_family(rs$font), ps = rs$font_size)
  
  
  tmplst <- list()
  mxlns <- 0
  for(k in seq_along(nms)) {
    
    tmplst[[k]] <- split_string_text(lbls[k], wdths[k], rs$units)
    
    if (tmplst[[k]]$lines > mxlns)
      mxlns <- tmplst[[k]]$lines
    
    pnts <- mxlns * rh
  }
  
  if (any(brdrs %in% c("all", "outside", "top"))) {
    if (!brdr_flag & !has_spans) {
      tbs <- ystart + bs - rh 
      pnts <- pnts + bs + 1
    } else {
      tbs <- ystart - rh + 1
      pnts <- pnts + 3
    }
  } else {
    tbs <- ystart - rh + 1
    pnts <- pnts + bh
  }

  for(k in seq_along(nms)) {
    
    # Split label strings if they exceed column width
    tmp <- tmplst[[k]]
    

    yline <- ystart + (rh * (mxlns - tmp$lines)) 
    if (any(brdrs %in% c("all", "outside", "top")) & !brdr_flag & !has_spans) {
      yline <- yline + bh

    }
    
    #print(yline)
    
    if (k == 1) {
      lb <- tlb
      rb <- lb + wdths[k]
    } else {
      lb <- rb
      rb <- lb + wdths[k]
    }
      
    for (ln in seq_len(tmp$lines)) {
      
      ret[[length(ret) + 1]] <- page_text(tmp$text[ln], rs$font_size, 
                                          bold = ts$header_bold,
                                          xpos = get_points(lb, 
                                                            rb,
                                                            tmp$widths[ln],
                                                            units = unts,
                                                            align = halgns[k]),
                                          ypos = yline)
      yline <- yline + rh
    }
    
    if (any(brdrs %in% c("all", "inside"))) {
      
      
      ret[[length(ret) + 1]] <- page_vline(rb * conv, 
                                           tbs, 
                                           (rh * mxlns) + bs) 
      
    }
    
    # Add in extra lines for labels that wrap
    xtr <- tmp$lines
    if (xtr > cnt)
      cnt <- xtr
    
  }
  
  dev.off()

  if (ts$first_row_blank == TRUE) {
    cnt <- cnt + 1
    pnts <- pnts + rh
  }
  
  # Top border
  if ((any(brdrs == "all")) | (any(brdrs %in% c("outside", "top")) & !has_spans)) {
    
    ret[[length(ret) + 1]] <- page_hline(tlb * conv, 
                                         tbs, 
                                         (trb - tlb) * conv) 
    
  }
  
  # Left border
  if (any(brdrs %in% c("all", "outside", "left"))) {
    
    
    ret[[length(ret) + 1]] <- page_vline(tlb * conv, 
                                         tbs, 
                                         (rh * cnt) + bs) 
    
  }
  
  # Right border
  if (any(brdrs %in% c("all", "outside", "right"))) {
    
    
    ret[[length(ret) + 1]] <- page_vline(trb * conv, 
                                         tbs, 
                                         (rh * cnt) + bs) 
    
  }
  

  yline <- tbs + (rh * mxlns) + bs 
  if (!any(brdrs %in% c("all", "outside", "top")) & brdr_flag) {
    # yline <- yline + bs
    # pnts <- pnts + bs
  }
  
  
  # First row blank
  if (ts$first_row_blank == TRUE) {
    if (any(brdrs %in% c("all", "inside")) ) {
      
      ret[[length(ret) + 1]] <- page_hline(tlb * conv, 
                                           yline + rh, 
                                           (trb - tlb) * conv)
      
    }
  }
  
  # print("yline")
  # print(yline)
  
  # Bottom header border always present
  ret[[length(ret) + 1]] <- page_hline(tlb * conv, 
                                       yline, 
                                       (trb - tlb) * conv)
  
  
  
  res <- list(pdf = ret,
              lines = pnts / rh,
              points = pnts )
  
  return(res)
  
}

#' @description Return a vector of pdf codes for the table spanning headers
#' @details Basic idea of this function is to figure out which columns 
#' the header spans, add widths, then call get_table_header.  Everything
#' from there is the same.  Border adjustments are the closest thing to spaghetti
#' in this package. But it works.
#' @import stats
#' @noRd
get_spanning_header_pdf <- function(rs, ts, pi, ystart = 0, brdr_flag = FALSE) {
  
  spns <- ts$col_spans
  cols <- pi$keys
  cols <- cols[!is.controlv(cols)]
  w <- pi$col_width 
  w <- w[cols]
  gutter <- 0
  gap <- 3
  cnt <- c()
  rh <- rs$row_height
  bs <- rs$border_spacing
  bh <- rs$border_height
  border_flag <-  FALSE
  
  # print("Cols:")
  # print(cols)
  #print(w)
  wlvl <- get_spanning_info(rs, ts, pi, w, gutter)
  lvls <- sort(seq_along(wlvl), decreasing = TRUE)
  
  # At this point we have a data frame with labels, etc. and spanning
  # column widths, and are ready to create spanning header rows
  #print(wlvl)
  
  conv <- rs$point_conversion
  talgn <- pi$table_align
  
  # Table alignment
  # ta <- "\\trql"
  # if (talgn == "right")
  #   ta <- "\\trqr"
  # else if (talgn %in% c("center", "centre"))
  #   ta <- "\\trqc"
  
  # Get borders
  brdrs <- strip_borders(ts$borders)
  
  rh <- rs$row_height
  if (any(brdrs %in% c("all", "inside")))
    rh <-  rh + bh
  
  width <- sum(w)
  
  if (talgn == "right") {
    tlb <- rs$content_size[["width"]] - width
    trb <- rs$content_size[["width"]]
  } else if (talgn %in% c("center", "centre")) {
    tlb <- (rs$content_size[["width"]] - width) / 2
    trb <- width + tlb
  } else {
    tlb <- 0
    trb <- width
  }
  
  ret <- list()
  
  tbrdrs <- any(brdrs %in% c("top", "outside", "all"))
  
  # Nightmare
  lline <- ystart
  if (brdr_flag & !tbrdrs) {
    lline <- ystart + bs
  } else if (brdr_flag & tbrdrs) {
    if (all(brdrs %in% "outside"))
      lline <- ystart + 1
    else 
      lline <- ystart + bs 
  } else if (!brdr_flag & tbrdrs) {
    lline <- ystart + bs + bs
  }
  
  ybegin <- lline
  
  # Open device context
  pdf(NULL)
  par(family = get_font_family(rs$font), ps = rs$font_size)
  
  # Format labels for each level
  ln <- c()
  for (l in lvls) {
    
    s <- wlvl[[l]]
    
    widths <- s$width
    names(widths) <- s$name
    algns <- s$align
    names(algns) <- s$name
    lbls <- s$label
    names(lbls) <- s$name
    cs <- s$col_span
    
    r <- ""
    cnt[length(cnt) + 1] <- 1 
  
    

    
   mxyl <- 0
   mxlns <- 0
   lnadj <- 0
   tmps <- list()
   
   # Get max lines for this spanning level
   for (k in seq_along(lbls)) {
     
     # Split label strings if they exceed column width
     tmps[[k]] <- split_string_text(lbls[k], widths[k], rs$units) 
     
     if (tmps[[k]]$lines > mxlns)
       mxlns <- tmps[[k]]$lines
   }
    
    # Loop for labels
   for(k in seq_along(lbls)) {

     yline <- lline 
     
     # Get left and right bounds for this span
     # Will accumulate as it goes across 
     if (k == 1) {
       lb <- tlb
       rb <- lb + widths[k]
     } else {
       lb <- rb
       rb <- lb + widths[k]
     }
     
     if (lbls[k] != "") {
      
        tmp <- tmps[[k]]
        
        for (ln in seq_len(tmp$lines)) {
          
          # Adjust yline to push down labels
          if (ln == 1)
            yline <- yline + ((mxlns - tmp$lines) * rh)
          
          adj <- 0
          if (any(brdrs %in% c("all", "inside")))
              adj <- 4
          
          # Get pdf text for label
          ret[[length(ret) + 1]] <- page_text(tmp$text[ln], rs$font_size, 
                                              bold = s$bold[k],
                                              xpos = get_points(lb, 
                                                                rb,
                                                                tmp$widths[ln],
                                                                units = rs$units,
                                                                align = algns[k]),
                                              ypos = yline - adj)
          yline <- yline + rh
          if (yline > mxyl)
            mxyl <- yline
        }
        
        if (any(brdrs %in% c("all", "inside"))) { 
          

          ret[[length(ret) + 1]] <- page_vline(lb * conv,
                                               lline - rh,
                                               (mxlns * rh) +  .5)
          
          ret[[length(ret) + 1]] <- page_vline(rb * conv,
                                               lline - rh,
                                               (mxlns * rh) +  .5)
          
          
        } else {
          # Get underline for label if requested
          if (s$span[k] > 0 & s$underline[k]) {
            
            lnadj <- .5
            
            yline <- mxyl + lnadj - (rh * .75) + 1 
    
            ret[[length(ret) + 1]] <- page_hline((lb * conv) + gap,
                                                 yline,
                                                 ((rb - lb) * conv) - (gap * 2))
  
          }
        }
        

        # Add in extra lines for labels that wrap
        xtr <- tmp$lines + lnadj
        if (xtr > cnt[length(cnt)])
          cnt[length(cnt)] <- xtr
      
     }
      
   }
   
   if (any(brdrs %in% c("all", "inside"))) {
     
     ret[[length(ret) + 1]] <- page_hline(tlb * conv, 
                                          lline - rh, 
                                          (trb - tlb) * conv) 
     
   }

    lline <- mxyl + (rh * lnadj)
    
    
  }
  dev.off()
  
  cnts <- sum(cnt)
  
  if (length(lvls) > 0) {

    
    ypos <- ybegin - rh #+ bs 
    
    if (any(brdrs %in% c("top", "outside"))) {
      
      ret[[length(ret) + 1]] <- page_hline(tlb * conv,
                                           ypos,
                                           (trb - tlb) * conv)
      
    }
    
    # if (any(brdrs %in% c("all"))) {
    #   
    #   ret[[length(ret) + 1]] <- page_hline(tlb * conv, 
    #                                        ystart - rh  + (cnts * rh) + (1.5 * bs), 
    #                                        (trb - tlb) * conv) 
    #   
    # }
    
    if (any(brdrs %in% c("all", "outside", "left"))) {
      
      
      ret[[length(ret) + 1]] <- page_vline(tlb * conv, 
                                           ypos, 
                                           (cnts * rh)) 
      
    }
    
    if (any(brdrs %in% c("all", "outside", "right"))) {
      
      
      ret[[length(ret) + 1]] <- page_vline(trb * conv, 
                                           ypos, 
                                           (cnts * rh)) 
      
    }
  }
  
  pnts <- cnts * rh
  if (!brdr_flag & tbrdrs)
    pnts <- pnts + bs
  
  res <- list(pdf = ret, 
              lines = cnts, 
              points = pnts,
              border_flag = border_flag)
  
  return(res)
}

#' @description This function counts lines per row independently because
#' the ..row field does not account for page wrapping.  Need number
#' of lines on this particular page.
#' @noRd
get_table_body_pdf <- function(rs, tbl, widths, algns, talgn, tbrdrs, 
                               ystart = 0, spwidths = list(), 
                               brdr_flag = FALSE, frb = FALSE, styles, ts) {
  
  border_flag <- FALSE
  
  if ("..blank" %in% names(tbl)) {
    flgs <- tbl$..blank
  } else {
    flgs <- NA
  } 
  
  # Count lines per row
  rws <- c()
  cnt <- 0
  
  nms <- names(widths)
  nms <- nms[!is.na(nms)]
  nms <- nms[!is.controlv(nms)]
  wdths <- widths[nms]
  
  #saveRDS(spwidths, "spwidths0a.rds")
  
  if (!"..blank" %in% names(tbl)) {
    blnks <- rep("", nrow(tbl))
  } else {
    blnks <- tbl$..blank
  } 
  
  # Deal with one column situation
  if (length(nms) == 1) {
    t <- as.data.frame(tbl[[nms]])
    names(t) <- nms
  } else {
    t <- tbl[ , nms]
  }
  
  conv <- rs$point_conversion
  unts <- rs$units
  bs <- rs$border_spacing
  bh <- rs$border_height
  cp <- rs$cell_padding
  
  brdrs <- strip_borders(tbrdrs)
  if (all(tbrdrs == "body"))
    brdrs <- c("top", "bottom", "left", "right")

  # Get line height.  Don't want to leave editor default.
  if (any(brdrs %in% c("all", "inside"))) {
    rh <- rs$row_height + bh
  } else {
    rh <- rs$row_height
  } 
  
  # Sum up widths 
  width <- sum(wdths, na.rm = TRUE)
  
  # Get content alignment codes
  if (talgn == "right") {
    tlb <- rs$content_size[["width"]] - width
    trb <- rs$content_size[["width"]]
  } else if (talgn %in% c("center", "centre")) {
    tlb <- (rs$content_size[["width"]] - width) / 2
    trb <- width + tlb
  } else {
    tlb <- 0
    trb <- width
  }
  
  rline <- ystart + 1
  
  ret <- c()
  fs <- rs$font_size
  
  defs <- ts$col_defs
  
  # For break labels, prepare the break label columns for later use
  break_label_col <- c()
  if (any(grepl("..break_label", names(tbl)))) {
    
    for (d in 1:length(defs)) {
      if (!is.null(defs[[d]]$break_label)){
        break_label_col <- c(break_label_col, defs[[d]]$var_c)
      }
    }
  }
  
  pdf(NULL)
  par(family = get_font_family(rs$font), ps = fs)
  
  # Loop for rows
  for(i in seq_len(nrow(t))) {
    
    yline <- rline
    mxrw <- yline
    cnt <- cnt + 1
  
    # Loop for columns 
    for(j in nms) {
      
      # Seems like this should be done already
      # Need to get widths from split_cells
      if (all(class(tbl[i, j]) != "character")) {
        vl <- as.character(tbl[i, j])
      } else {
        vl <- tbl[i, j]
      } 
      
      if (flgs[i] %in% c("B", "A", "L")) {
        
        # Strip out line feeds for label rows
        #vl <- gsub("\n", " ", vl, fixed = TRUE)
        
        #browser()
        # Recalculate based on total width of table
        #stmp <- split_string_text(vl, sum(wdths), rs$units)
        
        #tmp <- stmp$text
        
        tmp <- strsplit(vl, "\n", fixed = TRUE)[[1]]
        
      } else {
      
        tmp <- strsplit(vl, "\n", fixed = TRUE)[[1]]
      }
        
      
      if (j == nms[1]) {
        lb <- tlb
        rb <- lb + wdths[j]
      } else {
        lb <- rb
        rb <- lb + wdths[j]
      }
      
      # If value is empty or blank
      if (length(tmp) == 0) {
        yline <- yline + rh
      } else if (length(trimws(tmp)) == 0) {
        yline <- yline + rh
      } else {
        
        stl <- get_cell_styles(j, styles, flgs, i, tbl)
        
        bflg <- FALSE
        if ("bold" %in% stl) {
          bflg <- TRUE 
        }
      
        # Loop for cell wraps
        for (ln in seq_len(length(tmp))) {
          lb_cell <- lb
          # Indenting with lb
          if (flgs[i] == "L" & any(grepl("..break_label", names(tbl)))) {
            if (!is.na(tbl$..break_occur[i])) {
              cur_break_label_col <- break_label_col[as.numeric(tbl$..break_occur[i])]
              if (!is.null(defs[[cur_break_label_col]]$indent)) {
                lb_cell <- lb_cell + defs[[cur_break_label_col]]$indent
              }
            }
            width_cell <- width
          } else if (!is.null(defs[[j]]$indent)) {
            lb_cell <- lb_cell + defs[[j]]$indent
            width_cell <- spwidths[[i]][[j]][ln]
          } else if (j == "stub" & !is.null(ts$stub)) {
            stub_var <- tbl$..stub_var[i]
            if (!is.null(defs[[stub_var]]$indent)) {
              lb_cell <- lb_cell + defs[[stub_var]]$indent
            }
            width_cell <- spwidths[[i]][[j]][ln]
          } else {
            width_cell <- spwidths[[i]][[j]][ln]
          }
          
          ret[[length(ret) + 1]] <- page_text(tmp[ln], fs, 
                                              bold = bflg,
                                              xpos = get_points(lb_cell, 
                                                                rb,
                                                                width_cell,
                                                                units = unts,
                                                                align = algns[j]),
                                              ypos = yline)
          yline <- yline + rh
        }
      }
      
      if (yline > mxrw)
        mxrw <- yline
      
      yline <- rline
      
    }
    

    # Loop for columns 
    for(j in nms) {
      
      # Applies inside vertical borders
      if (any(brdrs %in% c("all", "inside")) & j != nms[length(nms)] 
          & !blnks[i] %in% c("B", "A", "L")) {
        
        if (j == nms[1]) {
          lb <- tlb
          rb <- lb + wdths[j]
        } else {
          lb <- rb
          rb <- lb + wdths[j]
        }
        
        if (i == 1) { 
          
          if (is.null(frb)) {
            ret[[length(ret) + 1]] <- page_vline(rb * conv, (rline + bs) - rh - 1, 
                                               mxrw - rline + 1)
           } else if (frb == FALSE) {
            ret[[length(ret) + 1]] <- page_vline(rb * conv, (rline + bs) - rh - 1, 
                                               mxrw - rline + 1)
           } else {
             ret[[length(ret) + 1]] <- page_vline(rb * conv, (rline + bs) - rh, 
                                                  mxrw - rline + 1)
           }
          
        } else if (i == nrow(t)) {
          
          ret[[length(ret) + 1]] <- page_vline(rb * conv, (rline + bs) - rh, 
                                               mxrw - rline + 1)
        
        } else {
          
          ret[[length(ret) + 1]] <- page_vline(rb * conv, (rline + bs) - rh, 
                                               mxrw - rline )
          
        }
        
      }
      
    }

    group_border <- ""
    if ("..group_border" %in% names(tbl)) {
      group_border <- tbl[["..group_border"]][i]
    }
    
    # Don't draw group line for first blank row
    if (blnks[i] %in% c("B", "A") & i == 1){
      group_border <- ""
    }
    
    # Don't draw group line for last record when table border is all, bottom, outside
    if (i == nrow(t) & any(brdrs %in% c("all", "outside", "bottom"))) {
      group_border <- ""
    }
    
    # Applies inside horizontal borders
    if ((any(brdrs %in% c("all", "inside")) & i < nrow(t)) |
        group_border == "bottom"
        ) {
    
      ret[[length(ret) + 1]] <- page_hline(tlb * conv, mxrw -rh + bs, 
                                           (trb - tlb) * conv)
    
    }
    
    rline <- mxrw
    
    # Add border height if the line is drawn by group border
    if (!any(brdrs %in% c("all", "inside")) & group_border == "bottom") {
      rline <- rline + bh
    }
  }
  
  dev.off()
  
  if (frb & "body" %in% tbrdrs) {
    
    
    ypos <- ystart - rs$row_height - rs$row_height + bh 
    
    #ylen <- cnt * rh
    ylen <- rline - rh + rs$row_height + bs + 1
    
  } else {

    ypos <- ystart - rs$row_height + bh - 1
    
    #ylen <- cnt * rh
    ylen <- rline - rh + bs + 1
  }
  
  if (any(brdrs %in% c("all", "left", "outside"))) {
    
    ret[[length(ret) + 1]] <- page_vline(tlb * conv, ypos, ylen - ypos)
  }
  if (any(brdrs %in% c("all", "right", "outside"))) {
    
    ret[[length(ret) + 1]] <- page_vline(trb * conv, ypos, ylen - ypos)
  }
  
  # pnts <- cnt * rh
  pnts <- ylen - ystart + rh 
  
  if (any(brdrs %in% c("all", "bottom", "outside"))) {
    
    
    # ret[[length(ret) + 1]] <- page_hline(tlb * conv, ypos + ylen, 
    #                                      (trb - tlb) * conv)
    
    ret[[length(ret) + 1]] <- page_hline(tlb * conv, ylen, 
                                         (trb - tlb) * conv)
    
    border_flag <- TRUE
    if (any(brdrs %in% c("all"))) {
      
      pnts <- pnts - bh 
    }
   # pnts <- pnts + bh
    
  }

  # fcntr <- fcntr + 1
  # saveRDS(ret, paste0("ret", fcntr, ".rds"))

  
  rws <- rline
  
  res <- list(pdf = ret,
              lines = pnts / rs$row_height,
              points = pnts ,
              border_flag = border_flag)
  
  return(res)
  
  
}

fcntr <- 0

# Globals -----------------------------------------------------------------

pointsize <- 1/72
inchsize <- 72
in2cm <- 2.54
binchars <- charToRaw("\xe2\xe3\xcf\xd3")



# Create PDF --------------------------------------------------------------

#' @noRd
create_pdf <- function(filename = NULL, 
                       page_height = 11,
                       page_width = 8.5,
                       fontname = "Courier", 
                       fontsize = 10,
                       margin_top = 1,
                       margin_left = 1, 
                       orientation = "landscape",
                       units = "inches",
                       conversion = 72,
                       info = TRUE) {
  
  # Check font size is valid
  if (!fontsize %in% c(8, 9, 10, 11, 12))
    stop(paste0("Fontsize ", fontsize, " not valid."))
  
  
  rpt <- structure(list(), class = c("pdf_report", "list"))
  
  rpt$filename <- filename
  rpt$page_height <- page_height
  rpt$page_width <- page_width
  rpt$fontname <- fontname
  rpt$fontsize <- fontsize
  rpt$margin_top <- margin_top
  rpt$margin_left <- margin_left
  rpt$info <- info
  rpt$author <- Sys.info()[["user"]]
  rpt$title <- ""
  rpt$subject <- ""
  rpt$keywords <- ""
  rpt$orientation <- orientation
  rpt$units <- units
  rpt$conversion <- conversion
  rpt$pages <- list()
    
  
  return(rpt)
}


#' @param x The pdf_report object to add a page to.
#' @param ... The page_text or page_image content to add.
#' @noRd
add_page <- function(x, ...) {
  
  
  pg <- structure(list(), class = c("pdf_page", "list"))
  
  inc <- list(...)
  
  for (itm in inc) {
    
    if (all(class(itm) == "list")) {
      
      for (sub in itm) {
        pg[[length(pg) + 1]] <- sub 
        
      }
      
      
    } else 
      pg[[length(pg) + 1]] <- itm
  }

  
  
  x$pages[[length(x$pages) + 1]] <- pg
  
  return(x)
  
}


add_info <- function(x,
                     author = "",
                     title = "",
                     subject = "",
                     keywords = "") {
  
  x$info <- TRUE
  x$author <- author
  x$title <- title
  x$subject <- subject
  x$keywords <- keywords
  
  return(x)
  
}


#' @noRd
page_text <- function(text, font_size = NULL, 
                      xpos = NULL, ypos = NULL, bold = FALSE,
                      align = NULL, alignx = NULL, has_page_numbers = NULL,
                      italics = FALSE, footnotes = FALSE, titles = FALSE) {
  
  txt <- structure(list(), class = c("page_text", "page_content", "list"))
  
  txt$text <- text
  txt$font_size <- font_size
  txt$xpos <- xpos
  txt$ypos <- ypos
  txt$bold <- bold
  txt$italics <- italics
  txt$align <- align
  txt$alignx <- alignx  # In units of measure
  txt$footnotes <- footnotes
  txt$titles <- titles
  
  res1 <- grepl("[pg]", text, fixed = TRUE)
  res2 <- grepl("[tpg]", text, fixed = TRUE)
  
  txt$has_page_numbers <- FALSE
  if (any(res1 == TRUE) | any(res2 == TRUE))
    txt$has_page_numbers <- TRUE
  
  return(txt)
  
}




#' Either pass align parameter and line_start, or pass specific xpos and ypos
#' All measurements in units specified.  Processing function will convert
#' appropriately.
#' @noRd
page_image <- function(filename, height, width,  
                       align = "center", line_start = NULL,
                       xpos = NULL, ypos = NULL, units = "inches",
                       dpi = 300) {
  
  img <- structure(list(), class = c("page_image", "page_content", "list"))
  
  if (!is.null(xpos) & !is.null(ypos)) {
    
    img$xpos <- xpos
    img$ypos <- ypos

  } else if (!is.null(align) * !is.null(line_start)) {
    
    img$align <- align 
    img$line_start <- line_start
    
  } else {
    
    stop("Must set either align parameter or xpos.") 
  }
  
  img$filename <- filename
  img$height <- height
  img$width <- width
  img$units <- units
  img$dpi <- dpi

  
  return(img)
  
}

page_line <- function(startx, starty, endx, endy) {
  
  ln <- structure(list(), class = c("page_line", "page_content", "list"))
  
  
  ln$startx <- startx
  ln$starty <- starty
  ln$endx <- endx
  ln$endy <- endy
  
  return(ln)
}

page_hline <- function(startx, starty, pwidth) {
  
  ln <- structure(list(), class = c("page_line", "page_content", "list"))
  
  
  ln$startx <- startx
  ln$starty <- starty
  ln$endx <- startx + pwidth
  ln$endy <- starty
  
  return(ln)
}

page_vline <- function(startx, starty, pheight) {
  
  ln <- structure(list(), class = c("page_line", "page_content", "list"))
  
  
  ln$startx <- startx
  ln$starty <- starty
  ln$endx <- startx
  ln$endy <- starty + pheight
  
  return(ln)
}


page_box <- function(startx, starty, pheight, pwidth, except = "") {
  
  
  bx <- structure(list(), class = c("page_box", "page_content", "list"))
  
  bx$startx <- startx
  bx$starty <- starty
  bx$pheight <- pheight
  bx$pwidth <- pwidth
  bx$except <- except
  
  return(bx)
  
}

page_grid <- function(startx, starty, rows, cols, pheights, pwidths) {
  
  
  grd <- structure(list(), class = c("page_grid", "page_content", "list"))
  
  grd$startx <- startx
  grd$starty <- starty
  grd$nrow <- nrow
  grd$rows <- rows
  grd$cols <- cols
  grd$pheights <- pheights
  grd$pwidths <- pwidths
  
  return(grd)
  
}

# Write PDF ---------------------------------------------------------------

#' A function to write out a PDF file
#' @param rpt The pdf_report object to write.
#' @param filename An optional file name.  If no filename is supplied,
#' it will use the filename from the report object. 
#' @noRd
write_pdf <- function(rpt, filename = NULL) {
  

  if (is.null(filename)) {
    if (is.null(rpt$filename))
      stop("Filename cannot be NULL.")
    
    filename <- rpt$filename
    
  }
  
  if (rpt$orientation == "landscape") {
    tmp <- rpt$page_height
    page_height <- rpt$page_width
    page_width <- tmp
    
  } else {
    
    page_height <- rpt$page_height
    page_width <- rpt$page_width
  }
  
  if (rpt$units == "cm") {
    
   page_height <- round(page_height / in2cm, 2)
   page_width <- round(page_width / in2cm, 2)
   margin_top <- round(rpt$margin_top / in2cm, 2)
   margin_left <- round(rpt$margin_left / in2cm, 2)
  } else {
    
   margin_top <- rpt$margin_top
   margin_left <- rpt$margin_left
  }
    
  
  # Remove existing file if needed
  if (file.exists(filename))
    file.remove(filename)
  
  bp <- dirname(filename)
  
  # Check that base path exists
  if (!file.exists(bp))
    stop(paste0("Base path '", bp, "' does not exist."))
  

  bdy <- get_pages(rpt$pages, margin_left, margin_top, 
                   page_height, page_width, rpt$fontsize, units = rpt$units,
                   fontname = rpt$fontname, conversion = rpt$conversion)
  
  rpt$pages <- length(rpt$pages)
                
  kids <- bdy$page_ids
  pgs <- bdy$objects


  # Get header objects
  hd <- get_header(font_name = rpt$fontname,
                   page_height = (page_height * inchsize),
                   page_width = (page_width * inchsize),
                   page_count = length(kids), 
                   page_ids = kids)
  

  # Add info if desired
  if (rpt$info) {
    
    idno <- length(hd) + length(pgs) + 1
    
    inf <- pdf_info(idno, author = rpt$author,
                    title = rpt$title, 
                    subject = rpt$subject,
                    keywords = rpt$keywords)
    
    doc <- pdf_document(hd, pgs, inf)
    
  } else {

    doc <- pdf_document(hd, pgs)
  }
  
  # Render document
  rdoc <- render.pdf_document(doc)
  
  # Write document to file system
  # Encoding is important.  Has to be UTF-8.
  # This is the only way to do it on Windows.
  f <- file(filename, open="wb", encoding = "native.enc")
  
  
  writeBin(rdoc, con = f, useBytes = TRUE)
  
  
  close(f)
  
  
  return(rpt)
}



#' A function to create a list of objects for the PDF header.
#' This includes the catalog, font, and pages.
#' @noRd
get_header <- function(page_count = 1, 
                       font_name = "Courier", 
                       page_height = 612, 
                       page_width = 792, 
                       page_ids = c()) {
  
  lst <- list()
  
  lst[[1]] <- pdf_object(1, pdf_dictionary(Type = "/Catalog",
                                           Pages = ref(5)))

  # 
  fn <- "Courier"
  fb <- "Courier-Bold"
  fi <- "Courier-Oblique"
  if (tolower(font_name) == "times") {
    fn <- "Times-Roman"
    fb <- "Times-Bold"
    fi <- "Times-Italic"
  } else if (tolower(font_name) == "arial") {
    fn <- "Helvetica"
    fb <- "Helvetica-Bold"
    fi <- "Helvetica-Oblique"
  }

  
  lst[[2]] <- pdf_object(2, pdf_dictionary(Type = "/Font", 
                                           Subtype = "/Type1", 
                                           BaseFont = paste0("/", fn),
                                           Encoding = "/WinAnsiEncoding"))
  
  lst[[3]] <- pdf_object(3, pdf_dictionary(Type = "/Font", 
                                           Subtype = "/Type1", 
                                           BaseFont = paste0("/", fb),
                                           Encoding = "/WinAnsiEncoding"))
  
  lst[[4]] <- pdf_object(4, pdf_dictionary(Type = "/Font", 
                                           Subtype = "/Type1", 
                                           BaseFont = paste0("/", fi),
                                           Encoding = "/WinAnsiEncoding"))

  
  if (page_count > 10)
    kds <- paste(page_ids, "0 R\n", collapse = " ")
  else 
    kds <- paste(page_ids, "0 R", collapse = " ")
  
  lst[[5]] <- pdf_object(5, pdf_dictionary(Type = "/Pages",
                                           Kids = pdf_array(kds),
                                           Count = page_count,                                                 
                                           MediaBox = pdf_array(0, 0, 
                                                        page_width, 
                                                        page_height)))
  
  
  return(lst)
  
}


#' Purpose of this function is to create the appropriate pdf objects
#' based on the pages added to the report.  Each page can have 1 or more 
#' pieces of content.  For instance, a page can have text and 2 images.
#' Object to pages is not 1 to 1.  There are at least two objects for a single
#' page: one for the page, and one for the page content.  If there is an image
#' on the page, there is another object to hold the image stream.  Note
#' that the image has to be referenced in both the page object and the 
#' the content object.  That is why the function creates the page and image
#' objects last: you can't really complete these objects until you examine
#' all the page contents.
#' @return A two part list, with a list of objects and a vector of ids for the 
#' pages.
#' @noRd
get_pages <- function(pages, margin_left, margin_top, page_height, page_width,
                      fontsize, units = "inches", fontname = "Courier", 
                      conversion = 72) {
  
  # Vector for object IDs of pages only
  kids <- c()
  
  # List for all objects
  ret <- list()
  
  # Determined by trial and error
  if (tolower(fontname) == "fixed")
    fontscale <- 87  # 87
  else 
    fontscale <- 100
  
  # Hate this.  Don't know why it is needed.  Width estimates should be the same.
  pg_adj <- 0
  if (tolower(fontname) == "arial" | tolower(fontname) == "times") {
    pg_adj <- -1
  } else if (tolower(fontname) == "courier" & units == "inches") {
    pg_adj <- 1 
  }
  
  # Get x starting position in points
  stx <- (margin_left * inchsize) - 5  # This was for fixed width
  
  # Get y starting position in points
  sty <- ((page_height * inchsize) -  (margin_top * inchsize)) - 5
  
  # Calculate reasonable line height
  # Also trial and error
  lh <- fontsize  + round(fontsize * .19, 2) 
  
  # Starting ID is 6 because of standard header objects.
  # This id variable will be incremented along the way 
  # as needed to get unique ids for the objects. 
  id <- 6
  
  pgnum <- 0
  tpg <- length(pages)
  
  # Loop through added pages
  for (pg in pages) {
    
    pgnum <- pgnum + 1
    
    # Set current page id
    page_id <- id
    
    # Add this page to the kids list
    kids <- append(kids, page_id)
    
    # There will always be one content object per page
    content_id <- id + 1
    
    # Increment id in preparation for next object that needs an id
    id <- content_id + 1
    
    # May or may not be image ids
    img_ids <- c()
    
    # Create content object
    # Content will be appended as we go along
    cnto <- pdf_text_stream(content_id, "")
    
    # Create a list of image streams for this page
    imgs <- list()
    
    for (cnt in pg) {
    
      if ("page_text" %in% class(cnt)) {
        
        # If fixed width PDF, will not have x/y positions
        if (is.null(cnt$xpos) | is.null(cnt$ypos)) {

          tmp <- get_byte_stream(cnt$text,
                                 stx, sty, lh, fontsize, fontscale)
        } else {
          
          # For PDF2 with page number flag set
          if (cnt$has_page_numbers & !is.null(cnt$align) & !cnt$footnotes) {
            
            # Need to adjust x position if there are page numbers,
            # because when the page number tokens are replaced, 
            # the string width has changed and can throw off the alignment.
            # Not working perfectly, but better.
            txt <- get_page_numbers_pdf(cnt$text, pgnum, tpg)
            w <- get_text_width(txt, fontname, 
                                ifelse(is.null(cnt$font_size), 
                                                      fontsize, cnt$font_size), 
                                units,
                                multiplier = 1.03) # 1.03 to match split_strings
            if (cnt$align == "left")
              nx <- cnt$alignx * conversion
            else if (cnt$align == "right")
              nx <- ((cnt$alignx - w) * conversion) + pg_adj  # No idea why pg_adj is needed
            else
              nx <- (cnt$alignx - (w / 2)) * conversion + pg_adj
            
            # For PDF2
            tmp <- get_byte_stream(txt, 
                                   stx + nx, sty - cnt$ypos, 
                                   lh, ifelse(is.null(cnt$font_size), 
                                              fontsize, cnt$font_size),
                                   fontscale, cnt$bold, cnt$italics)
            
          } else if (cnt$has_page_numbers & cnt$footnotes) {
            
            nx <- 0
            
            if (cnt$align %in% c("right", "center", "centre")) {
              w1 <- get_text_width(cnt$text, fontname, 
                                  ifelse(is.null(cnt$font_size), 
                                         fontsize, cnt$font_size), 
                                  units,
                                  multiplier = 1.03) 
              
              txt <- get_page_numbers_pdf(cnt$text, pgnum, tpg)
              
              w2 <- get_text_width(txt, fontname, 
                                  ifelse(is.null(cnt$font_size), 
                                         fontsize, cnt$font_size), 
                                  units,
                                  multiplier = 1.03) 
              
              nx <- (w1 - w2) * conversion
              
              if (cnt$align %in% c("center", "centre")) {
                
                nx <- nx/2
              }
            
            } else {
              
              txt <- get_page_numbers_pdf(cnt$text, pgnum, tpg) 
            }
            
            tmp <- get_byte_stream(txt, 
                                   stx + cnt$xpos + nx, sty - cnt$ypos, 
                                   lh, ifelse(is.null(cnt$font_size), 
                                              fontsize, cnt$font_size),
                                   fontscale, cnt$bold, cnt$italics)
            
          } else if (cnt$titles & cnt$has_page_numbers) {
            
            txt <- get_page_numbers_pdf(cnt$text, pgnum, tpg)
            tmp <- get_byte_stream(txt, 
                                   stx + cnt$xpos, sty - cnt$ypos, 
                                   lh, ifelse(is.null(cnt$font_size), 
                                              fontsize, cnt$font_size),
                                   fontscale, cnt$bold, cnt$italics)
            
          }else {
            
            # For other PDF2
            tmp <- get_byte_stream(cnt$text, 
                                   stx + cnt$xpos, sty - cnt$ypos, 
                                   lh, ifelse(is.null(cnt$font_size), 
                                              fontsize, cnt$font_size),
                                   fontscale, cnt$bold, cnt$italics)
            
          }
          

        }
    
      
      } else if ("page_image" %in% class(cnt)) {
        
        # Don't know how many there will be
        img_ids <- append(img_ids, id)
        
        # Convert measurements to points
        d <- calc_points(cnt, margin_left, margin_top, page_height, 
                             page_width, cnt$units, lh)
        
        # print(paste("wth:", d$wth))
        # print(paste("hgt:", d$hgt))
        # print(paste("xpos:", d$xpos))
        # print(paste("ypos:", d$ypos))


        # Every image needs a "Do" command on the content page
        tmp <- get_image_text(img_ref = id,
                              width = d$wth,
                              height = d$hgt,
                              xpos = d$xpos,
                              ypos = d$ypos)
        
        # Add stream to the list
        imgs[[length(imgs) + 1]] <- pdf_image_stream(id, 
                                                     height = d$phgt,
                                                     width = d$pwth,
                                       get_image_stream(cnt$filename))
          
        # Increment id in preparation for next object
        id <- id + 1
        
      } else if ("page_line" %in% class(cnt)) {
        
        tmp <- get_line_segment(startx = stx + cnt$startx,
                                starty = sty - cnt$starty,
                                endx = stx + cnt$endx,
                                endy = sty - cnt$endy)
        
      } else if ("page_box" %in% class(cnt)) {
        
        tmp <- get_box(startx = stx + cnt$startx,
                       starty = sty - cnt$starty, 
                       pheight = cnt$pheight,
                       pwidth = cnt$pwidth, 
                       except = cnt$except)
        
      } else if ("page_grid" %in% class(cnt)) {
        
        tmp <- get_grid(startx = stx + cnt$startx,
                        starty = sty - cnt$starty, 
                        rows = cnt$rows,
                        cols = cnt$cols,
                        pheights = cnt$pheight,
                        pwidths = cnt$pwidth)
        
      }
      
      # Append or replace content as appropriate
      if (all(cnto$contents == ""))
        cnto$contents <- tmp
      else 
        cnto$contents <- append(cnto$contents, tmp)
    
    
    }
    
    # Now can finally create all objects
    ret[[length(ret) + 1]] <- pdf_page(page_id, content_id, img_ids)
    ret[[length(ret) + 1]] <- cnto
    if (length(imgs) > 0)
      ret <- append(ret, imgs)
    
  }
  
  
  res <- list()
  res[["page_ids"]] <- kids
  res[["objects"]] <- ret
  
  return(res)
  
}



get_pages_back <- function(pages, margin_left, margin_top, page_height, page_width,
                      fontsize, units = "inches") {
  
  # Vector for object IDs of pages only
  kids <- c()
  
  # List for all objects
  ret <- list()
  
  # Determined by trial and error
  fontscale <- 87
  
  # Get x starting position in points
  stx <- (margin_left * inchsize) - 5
  
  # Get y starting position in points
  sty <- ((page_height * inchsize) -  (margin_top * inchsize)) - 5
  
  # Calculate reasonable line height
  # Also trial and error
  lh <- fontsize  + round(fontsize * .19, 2) 
  
  # Starting ID is 5 because of standard header objects.
  # This id variable will be incremented along the way 
  # as needed to get unique ids for the objects. 
  id <- 4
  
  # Loop through added pages
  for (pg in pages) {
    
    # Set current page id
    page_id <- id
    
    # Add this page to the kids list
    kids <- append(kids, page_id)
    
    # There will always be one content object per page
    content_id <- id + 1
    
    # Increment id in preparation for next object that needs an id
    id <- content_id + 1
    
    # May or may not be image ids
    img_ids <- c()
    
    # Create content object
    # Content will be appended as we go along
    cnto <- pdf_text_stream(content_id, "")
    
    # Create a list of image streams for this page
    imgs <- list()
    
    for (cnt in pg) {
      
      if ("page_text" %in% class(cnt)) {
        
        # Under current logic, there should only be one of these
        # In the future, will need to account for multiple pieces
        # of text, and may define their own positions and font size.
        # Right now everything is Courier from top left.
        tmp <- get_byte_stream(cnt$text,
                               stx, sty, lh, fontsize, fontscale)
        
        
      } else if ("page_image" %in% class(cnt)) {
        
        # Don't know how many there will be
        img_ids <- append(img_ids, id)
        
        # Convert measurements to points
        d <- calc_points(cnt, margin_left, margin_top, page_height, 
                         page_width, cnt$units, lh)
        
        # print(paste("wth:", d$wth))
        # print(paste("hgt:", d$hgt))
        # print(paste("xpos:", d$xpos))
        # print(paste("ypos:", d$ypos))
        
        
        # Every image needs a "Do" command on the content page
        tmp <- get_image_text(img_ref = id,
                              width = d$wth,
                              height = d$hgt,
                              xpos = d$xpos,
                              ypos = d$ypos)
        
        # Add stream to the list
        imgs[[length(imgs) + 1]] <- pdf_image_stream(id, 
                                                     height = d$phgt,
                                                     width = d$pwth,
                                                     get_image_stream(cnt$filename))
        
        # Increment id in preparation for next object
        id <- id + 1
        
      }
      
      # Append or replace content as appropriate
      if (all(cnto$contents == ""))
        cnto$contents <- tmp
      else 
        cnto$contents <- append(cnto$contents, tmp)
      
      
    }
    
    # Now can finally create all objects
    ret[[length(ret) + 1]] <- pdf_page(page_id, content_id, img_ids)
    ret[[length(ret) + 1]] <- cnto
    if (length(imgs) > 0)
      ret <- append(ret, imgs)
    
  }
  
  
  res <- list()
  res[["page_ids"]] <- kids
  res[["objects"]] <- ret
  
  return(res)
  
}

#' Convert all measurement to points, so they can be sent to pdf.
#' @import jpeg
#' @noRd
calc_points <- function(cnt, margin_left, margin_top, 
                        page_height, page_width, 
                        units, row_height) {
  
  pw <- page_width
  ph <- page_height
  ml <- margin_left
  mt <- margin_top
  
  if (units == "inches") {
    cw <- cnt$width
    ch <- cnt$height
    xp <- cnt$xpos 
    yp <- cnt$ypos

  } else if (units == "cm") {
    
    cw <- cnt$width / in2cm
    ch <- cnt$height / in2cm
    xp <- round(cnt$xpos / in2cm, 2)
    yp <- round(cnt$ypos/ in2cm, 2)
    
  } else 
    stop("Specified units not supported.")
  
  ret <- list()
  
  
  img <- readJPEG(cnt$filename)
  d <- dim(img)
  
  ret$pwth <- d[2]
  ret$phgt <- d[1]
  ret$wth <- round(cw  * inchsize, 2)
  ret$hgt <- round(ch * inchsize, 2)
  
  if (!is.null(cnt$align)) {
    if (cnt$align == "left")
      ret$xpos <- ml * inchsize
    else if (cnt$align == "right")
      ret$xpos <- (pw * inchsize) - ret$wth - (ml * inchsize)
    else 
      ret$xpos <- ((pw * inchsize)/2) - (ret$wth / 2)
    
    ret$ypos <- (ph * inchsize) - ret$hgt - 
      (row_height * (cnt$line_start - 1)) - (mt * inchsize)
    
  } else {
    ret$xpos <- xp * inchsize
    ret$ypos <- (ph * inchsize) - ret$hgt - (yp * inchsize) 
  }
  
  return(ret)
  
}

# Render Functions ---------------------------------------------------------



#' Generic render function
#' @noRd
render <- function(x) {
  
  UseMethod("render", x)
  
}

#' @exportS3Method render default
render.default <- function(x) {
  
  
  return(as.character(x))
  
}


#' @exportS3Method render pdf_array
render.pdf_array <- function(x) {
  
  
  ret <- paste0("[", paste(x, collapse = " "), "]") 
  
  
  return(ret)
  
}


#' @exportS3Method render pdf_object
render.pdf_object <- function(x) {
  
  ret <- paste0(x$id, " ", x$version, " obj")
  
  if (!is.null(x$parameters)) {
    ret <- paste0(ret, render(x$parameters), "\n")
  }
  
  if (!is.null(x$contents)) {
    
    ret <- paste0(ret, render(x$contents)) 
  }
  
  ret <- paste0(ret, "endobj\n")
  
  return(ret)
  
}

#' @exportS3Method render pdf_dictionary
render.pdf_dictionary <- function(x) {
  
  cnts <- c()
  
  for (itm in names(x)) {

    cnts[length(cnts) + 1] <- paste0("/", itm, " ", render(x[[itm]]))
  }
  
  cnts <- paste(cnts, collapse = " ")
  
  ret <- paste0("<<", cnts , ">>")
  
  return(ret)
  
}

#' @exportS3Method render pdf_text_stream
render.pdf_text_stream <- function(x) {
  

  if (length(x$contents) > 1) {
    cnts <- paste0(x$contents, collapse = "\n")
    cnts <- paste0(cnts, "\n")
  } else {
    
    cnts <- paste0(x$contents, "\n") 
  }
  
  strm <-   paste0("stream\n", cnts, "endstream\n")
  
  obj <- pdf_object(x$id, pdf_dictionary(Length = chars(cnts)), strm) 

                                         
  ret <- render.pdf_object(obj)
  
  return(ret)
  
}


# This was hard to make work. Took a lot of research and trial
# and error.  PDF has a lot of different way to render an image, and most
# are very complicated.  This is the simplest one. Requires a JPEG.  
#' @exportS3Method render pdf_image_stream
render.pdf_image_stream <- function(x) {
  
  view = FALSE
  
  if (length(x$contents) > 1) {
    cnts <- unlist(x$contents)
  } 
  
  
  if (rawToChar(cnts[[length(cnts)]]) != "\n")
    cnts[[length(cnts) + 1]] <- charToRaw("\n")
    
  strm <- list()
  strm[[1]] <- paste0(x$id, " ", x$version, " obj\n")
  strm[[2]] <- paste0(render(pdf_dictionary(Type = "/XObject",
                                     Subtype = "/Image",
                                     Width = x$width, 
                                     Height = x$height,
                                     BitsPerComponent = 8,
                                     ColorSpace = "/DeviceRGB",
                                     Filter = "/DCTDecode",
                                     # BitsPerComponent = 8,
                                     # ColorSpace = pdf_array("/Indexed", 
                                     #                        "/DeviceRGB",
                                     #                        111,
                                     #                        ref(10)),
                                     Length = chars(cnts))), "\n")
  strm[[3]] <- "stream\n"
  strm[[4]] <- cnts
  strm[[5]] <- "endstream\n"
  strm[[6]] <- "endobj\n"
  

  if (view) 
    ret <- strm
  else 
    ret <- unlist(vraw(strm))
  
  
  return(ret)
  
}


#' @exportS3Method render pdf_info
render.pdf_info <- function(x) {
  
  
  # Get current time
  tm <- Sys.time()
  cdt <- paste0(format(tm, format = "D:%Y%m%d%H%M%S"), 
                stri_sub(format(tm, format = "%z"), 1, 3), "'",
                stri_sub(format(tm, format = "%z"), 4, 5), "'")
  
  
  dict <- pdf_dictionary(Producer = paste0("(", x$producer, ")"),
                         Author = paste0("(", x$author, ")"),
                         Title = paste0("(", x$title, ")"),
                         Subject = paste0("(", x$subject, ")"),
                         Creator = paste0("(", x$creator, ")"),
                         Keywords = paste0("(", paste(x$keywords, collapse = " ")
                                           , ")\n"),
                         CreationDate = paste0("(", cdt, ")"),
                         ModDate = paste0("(", cdt, ")"))
  
  obj <- pdf_object(x$id, dict)
  
  ret <- render.pdf_object(obj)
  
  
  return(ret)
  
}

# Returns a raw byte array that can be written
# directly to disk with writeBin
#' @encoding UTF-8
#' @exportS3Method render pdf_document
render.pdf_document <- function(x) {
  
  
  #cnts <- c("%PDF-1.7\n", "%âãÏÓ\n")
  
  cnts <- list()
  # First line identifies this a a PDF file of a particular version.
  cnts[[1]] <- "%PDF-1.7\n"
  
  # Second line is comment of binary characters to identify the content as binary,
  # so editors or operating systems won't mess with it.
  cnts[[2]] <- paste0("%", rawToChar(binchars), "\n")
  xrefs <- c()
  infoid <- NULL
  
  for (itm in x) {
    
    if (!is.null(itm)) {
      
      # Sum up the number of bytes for all previous objects
      # Plus 1 to offset to first character of block
      xrefs[length(xrefs) + 1] <- chars(cnts) 
      
      #print(itm$id)
      tmp <-  render(itm) 
      #print(tmp)
      cnts[[length(cnts) + 1]] <- tmp
      
      if ("pdf_info" %in% class(itm))
        infoid <- itm$id
    
    }
    
  }


  cnts[[length(cnts) + 1]] <- render_xref(xrefs, x[[1]]$id, infoid, chars(cnts)) 
  cnts[[length(cnts) + 1]] <- "%%EOF"
  
  ret <- unlist(vraw(cnts))
  
  
  return(ret)
  
}

#' Create cross-reference table
#' @noRd
render_xref <- function(xrefs, rootID, infoID, startpos) {
  
  
  # Create first entry of cross reference table
  ret <- paste0("xref\n", "0 ", length(xrefs) + 1, "\n")
  
  
  # # Dynamically create subsequent entries
  # # Windows has two character end character, 
  # # and therefore doesn't need an extra space.
  # # Total width has to be 20 characters exactly.
  # if (Sys.info()[["sysname"]] == "Windows") {
  #   ret <- paste0(ret, "0000000000 65535 f\n")
  #   refs <- paste0(sprintf("%010d", xrefs), " 00000 n\n", collapse = "")
  # } else {
    ret <- paste0(ret, "0000000000 65535 f \n")
    refs <- paste0(sprintf("%010d", xrefs), " 00000 n \n", collapse = "")
  # }
    
  # Combine first and subsequent entries
  ret <- paste0(ret, refs)
  
  # Create dictionary for trailer
  dict <- pdf_dictionary(Size = length(xrefs) + 1, 
                         Root = ref(rootID))
  
  if (!is.null(infoID)) {

    dict$Info <- ref(infoID)    
  } 
  
  # Create trailer
  ret <- paste0(ret, "trailer ", 
                render(dict),
                "\n", 
                "startxref\n",
                startpos, "\n")
  
  return(ret)
  
}


# Class Definitions -------------------------------------------------------

#' Class definition for the Document 
#' Accepts a set of objects
#' @noRd
pdf_document <- function(...) {
  
  lst <- list(...)
  

  doc <- structure(list(), class = c("pdf_document", "list"))
  
  
  for (itm in lst) {
    
    # Input can be a single object or a list of objects
    if ("pdf_object" %in% class(itm)) {
      doc[[itm$id]] <- itm
    } else if (all("list" %in% class(itm))){
      
      for (sub in itm) {
        if (!"pdf_object"  %in% class(sub))
          stop("Document subitem must be of class pdf_object.")
        else 
          doc[[sub$id]] <- sub
      }
        
    } else 
      stop("Document item must be of class pdf_object.")
      
  }
  
  
  return(doc)
  
}


#' Define a class to contain a PDF array
#' This is a list of numbers, strings, or object references
#' @noRd
pdf_array <- function(...) {
  

  arr <- structure(list(...), class = c("pdf_array", "list"))
  
  
  return(arr)
  
}

#' Define a general object class
#' @param id The id number for the object
#' @param params A dictionary of parameters for this object
#' @param contents Contents for this object.  May be null.
#' @noRd
pdf_object <- function(id, params = NULL, contents = NULL) {
  
  
  if (!class(id) %in% c("integer", "numeric"))
    stop("Class of id must be integer or numeric.")
  
  if (!"pdf_dictionary" %in% class(params))
    stop("Class of params must be pdf_dictionary.")
  
  obj <- structure(list(), class = c("pdf_object", "list"))
  
  obj$id <- id
  obj$version <- 0
  obj$parameters <- params
  obj$contents <- contents
  
  return(obj)
}


#' Define a dictionary class, which is a list of name/value pairs
#' @noRd
pdf_dictionary <- function(...) {
  
  
  d <- list(...)
  
  if (length(d) > 0 & length(names(d)) == 0)
    stop("Dictionary entries must have names.")
  

  dict <- structure(d, class = c("pdf_dictionary", "list"))
  
  
  return(dict)
  
}

#' @param id The id for this page object
#' @param content_id The id of the content object for this page
#' @param graphic_ids The ids for the graphics streams for this page
#' @noRd
pdf_page <- function(id, content_id, graphic_ids = NULL) {
  
  
  pg <- structure(list(), class = c("pdf_page", "pdf_object", "list"))
  
  # If there are graphics, expand the procedures
  if (length(graphic_ids) > 0) {
    
    procs <- pdf_array("/PDF", 
                      "/Text", 
                      "/ImageB", 
                      "/ImageC", 
                      "/ImageI")
    
    xobj <- pdf_dictionary()
    
    for(g in graphic_ids) {
      
      xobj[[paste0("X", g)]] <- ref(g) 
      
    }
    
    
    res <- pdf_dictionary(Font = pdf_dictionary(F1 = ref(2),
                                                F2 = ref(3),
                                                F3 = ref(4)), 
                          ProcSet = procs,
                          XObject = xobj)
    
    parms <-  pdf_dictionary(Type = "/Page",
                             Parent = ref(5),
                             Contents = ref(content_id),
                             Resources = res)
                             
  } else {
    
    procs <- pdf_array("/PDF", "/Text")
    
    res <- pdf_dictionary(Font = pdf_dictionary(F1 = ref(2),
                                                F2 = ref(3),
                                                F3 = ref(4)), 
                          ProcSet = procs)
    
    parms <-  pdf_dictionary(Type = "/Page",
                             Parent = ref(5),
                             Contents = ref(content_id),
                             Resources = res)
  }

  
  # Assign properties
  pg$id <- id
  pg$version <- 0
  pg$parameters <-parms
  pg$content_id <- content_id
  pg$graphic_ids <- graphic_ids
  
  
  return(pg)
  
}

#' Define a stream for text content
#' @noRd
pdf_text_stream <- function(id, contents = NULL) {

  
  strm <- structure(list(), class = c("pdf_text_stream", "pdf_object", "list"))

  strm$id <- id
  strm$contents <- contents
  
  return(strm)
}

#' Define a stream for image content
#' @noRd
pdf_image_stream <- function(id, height, width, contents = NULL) {
  
  
  strm <- structure(list(), class = c("pdf_image_stream", "pdf_object", "list"))
  
  strm$id <- id
  strm$contents <- contents
  strm$version <- 0
  strm$height <- height
  strm$width <- width
  
  return(strm)
}

#' A function to create an info object.  The info object contains the 
#' author, etc.  of the document.  The create date in this is desirable
#' for production files, but make development a pain because it guarantees
#' that the PDF file will change every time a test case is run, and therefore
#' it is hard to tell which files have actually changed.  The only way to 
#' tell is to open them all up and look at them.
#' @noRd
pdf_info <- function(id, 
                     author = NULL, title = NULL,
                     subject = NULL, creator = NULL,
                     keywords = NULL) {
  
  
  info <- structure(list(), class = c("pdf_info", "pdf_object", "list"))
  
  info$id <- id
  info$producer <- paste0("reporter v", getNamespaceVersion("reporter"))
  info$author <- author
  info$title <- title
  info$subject <- subject
  info$creator <- R.Version()["version.string"]
  info$keywords <- keywords
  info$create_date <- Sys.time()
  info$mod_date <- Sys.time()
  
  return(info)
}




# Utilities ---------------------------------------------------------------



#' A function to cast lists of mixed content to lists of raw vectors
#' @noRd
vraw <- Vectorize(function(line) {
  
  
 if (any(class(line) == "raw")) {
   ret <- line
 } else {
   
  ret <- charToRaw(line)
 }
  
  return(ret)
  
})

#' Count actual bytes as will exist in the file
#' @noRd
chars <- function(lines) {
  
  ret <- length(unlist(vraw(lines)))
  
  return(ret)
}


#' Return a reference.  Version numbers are all zero.  Not dealing with
#' versions in these documents. 
#' @noRd
ref <- function(id) {
  
 ret <- paste(id, "0 R")
 
 return(ret)
  
}

#' Utility function to create content for a text stream.  This will 
#' take a vector of strings, and create pdf statements to display each one.
#' Multiple lines are separated by the distance specified in lineheight.
#' Would like to add compression to this function to reduce size of PDF file.
#' This function largely replaces get_text_stream() because it supports more special 
#' characters.
#' @noRd
get_byte_stream <- function(contents, startx, starty, 
                            lineheight, fontsize, fontscale, bold = FALSE, 
                            italics = FALSE) {
  
  # Calculate y positions
  ypos <- seq(from = starty, length.out = length(contents), by = -lineheight)
  
  # lns <- stri_encode(contents, from = stri_enc_detect2(contents),
  #                    to = "Adobe-Standard-Encoding")
  
  
  #print(stri_enc_detect(contents))
  
  cnts <- c()
  
  # Convert text characters to byte codes
  for (ln in contents) {
    cnts[length(cnts) + 1] <- paste0(charToRaw(viconv(ln)), collapse = "")
  
  }
  
  bld <- "/F1 "
  if (bold == TRUE)
    bld <- "/F2 "
  
  if (italics == TRUE)
    bld <- "/F3 "
  
  # Create report line
  ret <- paste0("BT ", bld , fontsize, 
                " Tf ", fontscale, " Tz ", startx, " ", ypos, " Td <", 
                cnts, ">Tj ET")
  
  return(ret)
}


#' Utility function to create content for a text stream.  This will 
#' take a vector of strings, and create pdf statements to display each one.
#' Multiple lines are separated by the distance specified in lineheight.
#' Would like to add compression to this function to reduce size of PDF file.
#' This function largely replace by get_byte_stream() to support more special 
#' characters.
#' @noRd
get_text_stream <- function(contents, startx, starty, 
                            lineheight, fontsize, fontscale) {
  
  
  ypos <- seq(from = starty, length.out = length(contents), by = -lineheight)
  
  ret <- paste0("BT /F1 ", fontsize, 
                " Tf ", fontscale, " Tz ", startx, " ", ypos, " Td (", 
                viconv(contents), ")Tj ET")
  
  # Not working
  #ret <- memCompress(ret, type = "gzip")
  
  #ret <- paste0(ret, collapse = "")
  
  return(ret)
  
}

#' Utility function to create content for an image stream.
#' Has to be a JPEG, as that is the only format that PDF supports natively.
#' @noRd
get_image_stream <- function(filename) {
  

  
  if (!file.exists(filename))
    stop(paste0("File does not exist: ", filename))
  
  # Get file size and add 10% for safety
  fz <- file.info(filename)[["size"]] * 1.1

  # Read in bytes
  f <- file(filename, open="rb", encoding = "native.enc")
  
  ret <- readBin(con = f, "raw", fz)
  
  close(f)
  
  return(ret)
  
}

#' Utility function to create text content for an image
#' @noRd
get_image_text <- function(img_ref, height, width, xpos, ypos) {
  
  ret <- c()
  
  ret[1] <- "q"
  ret[2] <- paste(width, 0, 0, height, (xpos - 5), ypos, "cm")
  ret[3] <- paste0("/X", img_ref, " Do")
  ret[4] <- "Q"
  
  
  return(ret)
  
}


#' Utility function to create pdf codes for a line segment.
#' @noRd
get_line_segment <- function(startx, starty, endx, endy) {
  
  
  ret <- paste(startx, starty, "m", endx, endy, "l S")
  
  
  return(ret)
  
}

#' Utility function to create pdf codes for a box.
#' @noRd
get_box <- function(startx, starty, pheight, pwidth, except = "") {
  
  ret <- c()
  
  if (except != "top")
    ret[length(ret) + 1] <- paste(startx, starty, "m", 
                                  startx + pwidth, starty, "l S")
  if (except != "left")
    ret[length(ret) + 1] <- paste(startx, starty, "m", startx, starty - pheight, "l S")
  if (except != "bottom")
    ret[length(ret) + 1] <- paste(startx, starty - pheight, "m", startx + pwidth, 
                  starty - pheight, "l S")
  if (except != "right")
    ret[length(ret) + 1] <- paste(startx + pwidth, starty, "m", startx + pwidth, 
                  starty - pheight, "l S")
  
  return(ret)
  
}

#' Utility function to create pdf codes for a grid.
#' @noRd
get_grid <- function(startx, starty, rows, cols, pheights, pwidths) {
  
  ret <- c()
  
  wdths <- rep(pwidths, cols/length(pwidths)) 
  hgths <- rep(pheights, rows/length(pheights))
  twidth <- sum(wdths)
  theight <- sum(hgths)

  # Start top and left lines  
  ret[1] <- paste(startx, starty, "m", startx + twidth, starty, "l S")
  ret[2] <- paste(startx, starty, "m", startx, starty - theight, "l S")

  
  # Create remaining rows
  yline <- starty
  for (rw in hgths) {
    yline <- yline - rw
    ret[length(ret) + 1] <- paste(startx, yline, "m", 
                                  startx + twidth, yline, "l S")
  
  }
  
  # Create remaining columns
  xcol <- startx
  for (cl in wdths) {
    
    xcol <- xcol + cl
    ret[length(ret) + 1] <- paste(xcol, starty, "m", 
                                  xcol, starty - theight, "l S")
    
  }
  
  return(ret)
  
}

#' @description This is a vectorized version of iconv(), which converts 
#' encodings on a string.  The CP1252 is a Windows superset of Latin1
#' which the PDF spec happens to support natively.  This is useful because
#' we get more characters out of this than Latin1.  It is basically ANSI
#' plus a few extra characters.  Also added logic to convert characters that
#' fall outside this range to a question mark, which is better than the default
#' empty box.
#' @noRd
viconv <- Vectorize(function(vstr) {
  
  
  #print(paste0(Encoding(vstr), ": ", vstr))
  
  if (Encoding(vstr) == "UTF-8") {
    ret <- iconv(vstr, from =Encoding(vstr), to = "CP1252", sub = "?")
    
    if (length(grep("???", ret, fixed = TRUE)) > 0) {
      ret <- gsub("???", "?", ret, fixed = TRUE)
      if (Encoding(ret) == "UTF-8")
        ret <- iconv(ret, from ="UTF-8", to = "CP1252", sub = "?")
    }
    
  } else {
    
    ret <- vstr
  }
  
  return(ret)
  
}, USE.NAMES = FALSE, SIMPLIFY = TRUE)

context("write_pdf Tests")

base_path <- "c:/packages/reporter/tests/testthat"
data_dir <- base_path

base_path <- tempdir()
data_dir <- "."

test_that("pdf_array and render.pdf_array work as expected.", {
  
  
  arr <- pdf_array(1, 2, "three")
  
  
  expect_equal(length(arr), 3)
  expect_equal(arr[[2]], 2)
  expect_equal(arr[[3]], "three")
  
  rnd <- render(arr)
  
  
  expect_equal(rnd, "[1 2 three]")
  
})



test_that("pdf_dictionary and render.pdf_dictionary work as expected.", {
  
  
  d <- pdf_dictionary(A=1, B = 2, C = "three")
  
  
  expect_equal(length(d), 3)
  expect_equal(d[[2]], 2)
  expect_equal(d[[3]], "three")
  
  rnd <- render(d)
  
  
  rnd
  expect_equal(rnd, "<</A 1 /B 2 /C three>>")
  
})


test_that("render.pdf_dictionary with combo elements work as expected.", {
  
  
  d <- pdf_dictionary(A=1, B = pdf_array(1, 2, 3, 4), C = "three", 
                      D = pdf_dictionary(R="Hi", Q = "Spork"))
  
  
  expect_equal(length(d), 4)
  expect_equal(d[[1]], 1)
  expect_equal(render(d[[2]]), "[1 2 3 4]")
  expect_equal(d[[3]], "three")
  
  rnd <- render(d)
  
  
  rnd
  expect_equal(rnd, "<</A 1 /B [1 2 3 4] /C three /D <</R Hi /Q Spork>>>>")
  
})



test_that("pdf_object and render.pdf_object work as expected.", {
  
  
  d <- pdf_object(1, pdf_dictionary(Type = "/Catalog", 
                                    Pages = "2 0 R"))
  
  
  expect_equal(length(d), 3)
  expect_equal(d$id, 1)
  expect_equal(d$version, 0)
  expect_equal(render(d$parameters), "<</Type /Catalog /Pages 2 0 R>>")

  
  rnd <- render(d)
  
  
  #cat(rnd)
  
  
  expect_equal(rnd, "1 0 obj<</Type /Catalog /Pages 2 0 R>>\nendobj\n")
  
})


test_that("pdf_stream and render.pdf_stream work as expected.", {
  
  
  d <- pdf_text_stream(1, "Here is some text")
  
  
  expect_equal(length(d), 2)
  expect_equal(d$id, 1)
  expect_equal(d$contents, "Here is some text")
  
  
  rnd <- render(d)
  
  
  #cat(rnd)
  
  
  expect_true(grep("1 0 obj<</Length", rnd, fixed = TRUE, value = FALSE) > 0)
  
  expect_true(grep(">>\nstream\nHere is some text\nendstream\nendobj\n",
                   rnd,
                   fixed = TRUE, value = FALSE) > 0)
  
  
})


test_that("pdf_info works as expected.", {
  
  inf <- pdf_info(25,
                  author = "David",
                  keywords = c("PDF", "document"),
                  title = "My Doc",
                  subject = "My Subject")
                  

  inf
  
  expect_equal(inf[["id"]], 25)
  expect_equal(inf$producer, paste0("reporter v", 
                                    getNamespaceVersion("reporter")))
  expect_equal(inf$keywords, c("PDF", "document"))
  expect_equal(inf$title, "My Doc")
  expect_equal(inf$subject, "My Subject")
  expect_equal(inf$author, "David")
  expect_equal(is.null(inf$create_date), FALSE)
  expect_equal(is.null(inf$mod_date), FALSE)
  
  
  res <- render.pdf_info(inf)
  
  #cat(res)
  
  expect_equal(chars(res) >= 238, TRUE)
  
})


test_that("render.xref works as expected.", {
  
  d <- render_xref(c(20, 68, 124, 224, 267, 334), 1, 10, 514)
  
  #cat(d)
  
  expect_equal(length(d), 1)
  
  expect_equal(chars(d), 208)
  
  
})

test_that("pdf_document and render.pdf_document work as expected.", {
  
  d <- pdf_document(pdf_object(1, pdf_dictionary(Type = "/Catalog", 
                                                 Pages = "2 0 R")),
                    pdf_text_stream(2, "Here is some text"))
  
  class(d)
  
  expect_equal(length(d), 2)
  expect_equal(d[[1]]$id, 1)
  expect_equal(d[[2]]$id, 2)
  expect_equal(d[[2]]$contents, "Here is some text")
  expect_equal("pdf_document" %in% class(d), TRUE) 
  
  
  rnd <- render(d)
  
  # rnd
  #cat(rawToChar(rnd))
  expect_equal(length(rnd), 245)
  
  
  # paste0("%PDF-1.7\n",
  #        "%\u0203\u00e3\u00cf\u00d3\n",
  #        "1 0 obj<</Type /Catalog /Pages 2 0 R>>\n",
  #        "endobj\n",
  #        "2 0 obj<</Length 17>>\nstream\n", 
  #        "Here is some text\nendstream\nendobj\n",
  #        "xref\n0 3\n",
  #        "0000000000 65535 f\n",
  #        "0000000020 00000 n\n",
  #        "0000000068 00000 n\n",
  #        "trailer <</Size 3 /Root 1 0 R>>\n",
  #        "startxref\n137\n%%EOF")
  # 
  

})


test_that("get_header works as expected.", {
  
  # Check one page
  hdr <- get_header(page_count = 1,
                    font_name = "Arial",
                    page_ids = c(3))
  
  expect_equal(length(hdr), 5)
  
  hdrtxt <- render(pdf_document(hdr))
  
  #cat(rawToChar(hdrtxt))
  
  # if (Sys.info()[["sysname"]] == "Windows")
    expect_equal(length(hdrtxt), 614) #369)
  # else 
  #   expect_equal(length(hdrtxt), 370)
  
  # Check two pages
  hdr <- get_header(page_count = 2,
                    page_ids = c(3, 4),
                    page_height = 500,
                    page_width = 600)
  
  expect_equal(length(hdr), 5)
  
  hdrtxt <- render(pdf_document(hdr))
  
  #cat(rawToChar(hdrtxt))
  
  # if (Sys.info()[["sysname"]] == "Windows")
    expect_equal(length(hdrtxt), 614) #377)
  # else 
  #   expect_equal(length(hdrtxt), 378)
  
  # Check three pages
  hdr <- get_header(page_count = 3,
                    page_ids = c(3, 4, 5))
  
  expect_equal(length(hdr), 5)
  
  hdrtxt <- render(pdf_document(hdr))
  
  #cat(rawToChar(hdrtxt))
  
  # if (Sys.info()[["sysname"]] == "Windows")
    expect_equal(length(hdrtxt), 620) #383)
  # else
  #   expect_equal(length(hdrtxt), 384)
  
})

test_that("pdf_page works as expected.", {
  
  pg1 <- pdf_page(5, 6)
  
  expect_equal(length(pg1), 4)
  expect_equal(pg1$id, 5)
  expect_equal(pg1$content_id, 6)
  expect_equal(unclass(pg1$parameters$Resources$ProcSet), list("/PDF", "/Text"))
  
  
  pg1r <- render(pg1)
  
  pg1r
  
  pg2 <- pdf_page(5, 6, c(7, 8))
  
  expect_equal(length(pg2), 5)
  expect_equal(pg2$id, 5)
  expect_equal(pg2$content_id, 6)
  expect_equal(unclass(pg2$parameters$Resources$ProcSet), list("/PDF", 
                                                     "/Text", 
                                                     "/ImageB", 
                                                     "/ImageC", 
                                                     "/ImageI"))
  
  pg2r <- render(pg2)
  
  pg2r
  
  expect_equal(pg2$graphic_ids, c(7, 8))

})

test_that("create full document works as expected.", {
  
  
  pg <- pdf_page(4, 5)
  
  strm <- pdf_text_stream(5, c(
                     "BT /F1 12 Tf 175 600 Td (Hello here is some more)Tj ET",
                     "BT /F1 12 Tf 175 580 Td (There I like text.)Tj ET",
                     "BT /F1 12 Tf 175 560 Td (World and more)Tj ET", 
                     "BT /F1 12 Tf 175 540 Td (And some more)Tj ET"))
  

  
  doc <- pdf_document(get_header(1, page_ids = 4), pg, strm)
  
  expect_equal(length(doc), 5)
  expect_equal("pdf_document" %in% class(doc), TRUE)
  
  res <- render(doc)
  
  #cat(rawToChar(res))
  
  fp <- file.path(base_path, "pdf/direct1.pdf")
  
  if (file.exists(fp))
    file.remove(fp)
  

  f <- file(fp, open="wb", encoding = "native.enc")
  
  
  writeBin(res, con = f, useBytes = TRUE)
  
  
  close(f)
  
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("chars function works as expected.", {
  
  
  tmp <- c("there",
           "here\n")
  
  res <- chars(tmp)
  
  expect_equal(res, 10)
    

  tmp <- c("%PDF-1.P7\n",
           "%fork\n")
  
  res <- chars(tmp)

  expect_equal(res, 16)
  
  
}) 


test_that("get_text_stream function works as expected.", {
  
  contents <- c("Hello", "goodbye", "later")
  
  res <- get_text_stream(contents, 50, 600, 20, 12, 100)
  
  
  expect_equal(length(res), 3)
  
  expect_equal(res[[3]], "BT /F1 12 Tf 100 Tz 50 560 Td (later)Tj ET")
  
})



test_that("vraw function works as expected.", {
  
  str <- list("Hello\n", "thereÏ\n", charToRaw("fork\n"))
  
  str2 <- vraw(str)
  
  expect_equal(length(str2), 3)
  
  expect_equal(length(unlist(str2)), 19)
  
  
})


test_that("get_image_text works as expected.", {
  
  res <- get_image_text(img_ref = 23, height = 200, width = 400, 
                        xpos= 100, ypos = 150)
  
  
  expect_equal(res[1], "q")
  expect_equal(res[2], "400 0 0 200 95 150 cm")
  expect_equal(res[3], "/X23 Do")
  expect_equal(res[4], "Q")
  
  
})

# Commented out because of CRAN warning.  view parameter is apparently illegal now.
# test_that("pdf_image_stream works as expected.", {
#   
#   fp <- file.path(data_dir, "data/dot.jpg")
#   
#   f <- file(fp, open="rb", encoding = "native.enc")
#   
#   png <- readBin(f, "raw", 10000)
#   
#   close(f)
#   
#   
#   strm <- pdf_image_stream(6, 60, 60, png)
#   
#   strm
#   
#   expect_equal(length(strm), 5)
#   
#   res <- render.pdf_image_stream(strm, view = TRUE)
#   
#   res
#   
#   expect_equal(length(res), 6)
#   
#   
# })




test_that("create_pdf and add_page work as expected.", {
  
  p <- create_pdf(filename = "mypath.pdf",
                  orientation = "portrait",
                  fontsize = 12,
                  fontname = "Arial",
                  page_height = 200,
                  page_width = 500,
                  units = "cm",
                  margin_top = 3,
                  margin_left = 4) %>% 
    add_page(page_text("here is some text"), 
             page_image("myimagefile.jpg", 
                        200, 600, align= "left",
                        xpos = 25, ypos = 30)) %>% 
    add_info(author = "David",
             subject = "Stuff",
             keywords = "reports",
             title = "My Title")
  
  
  
    expect_equal(p$filename, "mypath.pdf")
    expect_equal(p$page_height, 200)
    expect_equal(p$fontname, "Arial")
    expect_equal(p$fontsize, 12)
    expect_equal(p$margin_top, 3)
    expect_equal(p$margin_left, 4)
    expect_equal(p$info, TRUE)
    expect_equal(p$author, "David")
    expect_equal(p$title, "My Title")
    expect_equal(p$subject, "Stuff")
    expect_equal(p$keywords, "reports")
    expect_equal(p$orientation, "portrait")
    expect_equal(p$units, "cm")
    expect_equal(length(p$pages), 1)
    txt <- p$pages[[1]][[1]]
    expect_equal(txt$text, "here is some text")
    img <- p$pages[[1]][[2]]
    expect_equal(img$filename, "myimagefile.jpg")
    expect_equal(img$height, 200)
    expect_equal(img$width, 600)
    expect_equal(is.null(img$align), TRUE)
    expect_equal(img$xpos, 25)
    expect_equal(img$ypos, 30)
  
})



test_that("get_image_stream works as expected.", {

  fp <- file.path(data_dir, "data/dot.jpg")
  
  res <- get_image_stream(fp)



  fp <- file.path(base_path, "pdf/dot2.jpg")
  
  if (file.exists(fp))
    file.remove(fp)

  con <- file(fp, "wb", encoding = "native.enc")

  writeBin(res, con, useBytes = TRUE)

  close(con)
  
  
  expect_equal(file.exists(fp), TRUE)


})

test_that("get_pages works as expected.", {
  
  fp <- file.path(data_dir, "data/dot.jpg")
  
  
  rpt <- create_pdf() %>% 
    add_page(page_text("Hello")) %>% 
    add_page(page_text("There")) %>% 
    add_page(page_text("Image page"),
             page_image(fp, height = 60, width = 62, xpos = 100, ypos = 150))
  
  rpt 
  
  expect_equal(length(rpt$pages), 3)
  expect_equal(length(rpt$pages[[3]]), 2)
  
  res <- get_pages(rpt$pages, 
                   margin_left = rpt$margin_left, 
                   margin_top = rpt$margin_top, 
                   page_height = rpt$page_height, 
                   page_width = rpt$page_width,
                   fontsize = rpt$fontsize)
  
  
  res
  
  expect_equal(res$page_ids, c(6, 8, 10))
  expect_equal(length(res$objects), 7)
  
  
})

test_that("basic write_pdf works as expected.", {
  
  fp <- file.path(base_path, "pdf/direct2.pdf")
  
  
  r <- create_pdf(fp) %>% 
    add_page(page_text(c("Hello", "There", "Here is some text")))
  
  expect_equal(r$filename, fp)
  expect_equal(r$fontname, "Courier")
  expect_equal(length(r$pages), 1)
  expect_equal(r$pages[[1]][[1]]$text, c("Hello", "There", "Here is some text"))
  

  write_pdf(r)
  
  expect_equal(file.exists(fp), TRUE)  
  
})


test_that("write_pdf with info works as expected.", {
  
  fp <- file.path(base_path, "pdf/direct3.pdf")

  r <- create_pdf(fp) %>% 
    add_page(page_text( c("Hello", "There", "Here is some text"))) %>% 
    add_info(author = "David Bosak",
             keywords = "one two three", subject = "Reporting",
             title = "PDF 1.0")
    

  write_pdf(r)

  expect_equal(file.exists(fp), TRUE)
  
})


test_that("write_pdf with info defaults work as expected.", {
  
  fp <- file.path(base_path, "pdf/direct4.pdf")

  
  r <- create_pdf(fp, margin_top = .5, margin_left = .5) %>% 
    add_page(page_text( c("Hello", "There", "Here is some text"))) 
  
  
  write_pdf(r)

  expect_equal(file.exists(fp), TRUE)
  
})


test_that("Two page pdf works as expected.", {
  
  fp <- file.path(base_path, "pdf/direct5.pdf")

  
  r <- create_pdf(fp) %>% 
    add_page(page_text(c("Hello", "There", "Here is some text"))) %>% 
    add_page(page_text(c("Hey!", "Here is a second page!"))) 
    

  write_pdf(r)

  expect_equal(file.exists(fp), TRUE)
  
})


test_that("Three page pdf works as expected.", {
  
  fp <- file.path(base_path, "pdf/direct6.pdf")
  
  
  r <- create_pdf(fp) %>% 
    add_page(page_text(c("Hello", "There", "Here is some text"))) %>% 
    add_page(page_text(c("Hey!", "Here is a second page!"))) %>% 
    add_page(page_text(c("A third page!")))
  
  
  write_pdf(r)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("Simplest direct table works as expected.", {
  
  
  fp <- file.path(base_path, "pdf/direct7.pdf")
  
  rpt <- create_report(fp, output_type = "PDF") %>%
    add_content(create_table(mtcars[1:10, ]), align = "left") %>% 
    page_footer("Page [pg] of [tpg]") %>% 
    set_margins(top = .5) 
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  
  
})



test_that("Direct table with 2 pages works as expected.", {
  
  
  
  fp <- file.path(base_path, "pdf/direct8.pdf")
  
  rpt <- create_report(fp, output_type = "PDF") %>%
    add_content(create_table(mtcars[1:10, ]), page_break = FALSE) %>% 
    add_content(create_text("Hello")) %>% 
    add_content(create_text("There")) %>% 
    page_footer(right = "Page [pg] of [tpg]")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  
  
})




test_that("simple jpg document works as expected.", {
  
  library(ggplot2)
  
  ip <- file.path(data_dir, "data/plot.jpg")
  fp <- file.path(base_path, "pdf/direct9.pdf")
  
  # 
  # if (file.exists(ip))
  #   file.remove(ip)
  # 
  # p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  # 
  # ggsave(filename= ip, plot = p, width = 9, height = 5, units = "in")
  
  r <- create_pdf(fp) %>% 
    add_page(page_image(ip, height = 5, width = 9, xpos = 1, ypos = 1))
  
  
  write_pdf(r)
  
  expect_equal(file.exists(fp), TRUE)    
  

  
})


test_that("Test jpg alignment works as expected.", {
  
  library(ggplot2)
  
  ip <- file.path(data_dir, "data/plot.jpg")

  
  # 
  # if (file.exists(ip))
  #   file.remove(ip)
  # 
  # p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  # 
  # ggsave(filename= ip, plot = p, width = 6, height = 3, units = "in")
  
  
  fp <- file.path(base_path, "pdf/direct10a.pdf")
  
  r <- create_pdf(fp) %>% 
    add_page(page_image(ip, height = 3, width = 6, align = "left", line_start = 1))
  
  
  write_pdf(r)
  
  expect_equal(file.exists(fp), TRUE)   
  
  fp <- file.path(base_path, "pdf/direct10b.pdf")
  
  r <- create_pdf(fp) %>% 
    add_page(page_image(ip, height = 3, width = 6, align = "right", line_start = 1))
  
  
  write_pdf(r)
  
  expect_equal(file.exists(fp), TRUE)   
  
  
  fp <- file.path(base_path, "pdf/direct10c.pdf")
  
  r <- create_pdf(fp) %>% 
    add_page(page_image(ip, height = 3, width = 6, align = "center", line_start = 1))
  
  
  write_pdf(r)
  
  
  expect_equal(file.exists(fp), TRUE)    
  
  
  
})

# Work on this
test_that("simple jpg document works as expected in centimeters.", {

  library(ggplot2)

  ip <- file.path(data_dir, "data/plot.jpg")
  fp <- file.path(base_path, "pdf/direct11.pdf")


  # if (file.exists(ip))
  #   file.remove(ip)
  # 
  # p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  # 
  # ggsave(filename= ip, plot = p, width = 23, height = 13,
  #        units = "cm")

  r <- create_pdf(fp, page_height = 27.94, page_width = 21.59, units = "cm") %>%
    add_page(page_image(ip, height = 13, width = 23, xpos = 2.5,
                        ypos = 2.5, units = "cm"))


  write_pdf(r)

  expect_equal(file.exists(fp), TRUE)

})

test_that("Simplest direct table right aligned works as expected.", {
  
  
  fp <- file.path(base_path, "pdf/direct12.pdf")
  
  rpt <- create_report(fp, output_type = "PDF") %>%
    add_content(create_table(mtcars[1:10, ]), align = "right") %>% 
    set_margins(top = .5)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  
  
})


test_that("Simplest direct plot works as expected.", {


  library(ggplot2)

  fp <- file.path(base_path, "pdf/direct13.pdf")
  
  if (file.exists(fp))
    file.remove(fp)

  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()

  plt <- create_plot(p, height = 4, width = 8)


  rpt <- create_report(fp, output_type = "PDF") %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt, align = "center") %>%
    footnotes("* Motor Trend, 1974") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)



})


test_that("viconv function works as expected.", {
  
  v <- c("A B", "â ã")
  
  str <-  viconv(enc2utf8(v))

  expect_equal(Encoding(str), c("unknown", "latin1"))
  
  
  # str3 <- viconv(enc2utf8(v), as_raw = TRUE)
  # 
  # expect_equal(str3[[1]], charToRaw("A B"))
  # 
  # charToRaw(v[2])
  # 
  # r <- raw(3)
  # 
  # r[1] <- as.raw(226)
  # r[2] <- as.raw(32)
  # r[3] <- as.raw(227)
  # 
  # expect_equal(str3[[2]], r)

})


test_that("PDF with special chars works as expected.", {
  
  # 
  # str <- "Here are some special chars \u0203\u00e3\u00cf\u00d3"
  # str2 <- charToRaw(str)
  # 
  # hc <- paste0(str2, collapse = " ")
  # 
  # str3 <- paste0("<", hc, ">")
  library(magrittr)
  
  # ± ≠ ⋜ ⋝ ° ƒ ×
  str4 <- c("Here are some special chars â ã Ï Ó µ ¿ ‰ + - / %", 
            "Special symbols µ £ ¿ there  to mess things up ° ^ é ñ ± € ∆ ƒ") # ∆ ∈ ∑ √ ∫ Привет , 
               # "\xe2 \xe3 \xcf \xd3 \265 \277 \211 \234 £\n",
               # "Ω ± ∑ π α β")
  str4
  
  # str5 <- enc2utf8(str4)
  # 
  # Encoding(str5)
  # 
  # str6 <- iconv(str5, from ="UTF-8", to = "CP1252", sub = "Unicode")
  # Encoding(str6)
  
  # f <- file(file.path(base_path, "pdf/fork.txt"), open = "wt", encoding = "native.enc")
  # 
  # writeLines(enc2utf8(str4), con = f, useBytes = TRUE)
  # writeLines(enc2utf8(trimws(str4, which = "right")), con = f, useBytes = TRUE)
  # 
  # 
  # close(f)
  
  
  # stri_enc_detect(str4)
  # 
  # str5 <- stri_pad_right(str4, 100)
  # 
  # stri_enc_detect(str5)
  
  # lns <- stri_encode(str4, from = NULL,
  #                    to = "ANSI1251")
  # 
  # 
  # str5 <- charToRaw(as.character(str4, "!"))
  # stri_enc_detect(str5)
  # 
  # rawToChar(str5)
# 
#   str6 <- paste0("<", paste0(str5, collapse = " "), ">")
  # 
  fp <- file.path(base_path, "pdf/direct14")
  
  txt <- create_text(str4) %>% 
    titles(str4) %>% 
    footnotes("Ï Ó µ ¿")
  
  rpt <- create_report(fp, output_type = "PDF") %>%
    titles("Ï Ó µ ¿", borders = "bottom") %>% 
    page_header("special chars â ã Ï Ó µ") %>% #  Здраво
    page_footer("special chars â ã Ï Ó µ") %>% #  Привет
    add_content(txt, align = "left") %>% 
    footnotes(str4)
  
  
  res1 <- write_report(rpt, output_type = "PDF")
  res2 <- write_report(rpt, output_type = "TXT")
  res3 <- write_report(rpt, output_type = "RTF")

  expect_equal(file.exists(paste0(res1$file_path, ".pdf")), TRUE)
  expect_equal(file.exists(paste0(res2$file_path, ".txt")), TRUE)
  expect_equal(file.exists(paste0(res3$file_path, ".rtf")), TRUE)
  
  # cat(str4)
  

  
})


test_that("basic write_pdf with positional text works as expected.", {
  
  fp <- file.path(base_path, "pdf/direct15.pdf")
  
  l <- list(page_text("Hello", font_size = 14, 
                      xpos = 100, ypos = 100, bold = TRUE),
            page_text(c("Goodbye", "My friend"), font_size = 8, 
                        xpos = 200, ypos = 150))
  
  r <- create_pdf(fp, fontname = "Arial") %>% 
    add_page(l)

  
  write_pdf(r)
  
  expect_equal(file.exists(fp), TRUE)  
  
})

test_that("basic write_pdf with positional text works as expected.", {
  
  fp <- file.path(base_path, "pdf/direct15.pdf")
  
  l <- list(page_text("Hello", font_size = 14, 
                      xpos = 100, ypos = 100, bold = TRUE),
            page_text(c("Goodbye", "My friend"), font_size = 8, 
                      xpos = 200, ypos = 150))
  
  r <- create_pdf(fp, fontname = "Arial") %>% 
    add_page(l)
  
  
  write_pdf(r)
  
  expect_equal(file.exists(fp), TRUE)  
  
})

test_that("page_text detects page number tokens correctly", {
  
  res1 <- page_text("here is some text")
  res1
  
  expect_equal(res1$has_page_numbers, FALSE)
  
  res2 <- page_text("here is some [pg] text")
  res2
  
  expect_equal(res2$has_page_numbers, TRUE)
  
  
  res3 <- page_text("here is some [tpg] text")
  
  expect_equal(res3$has_page_numbers, TRUE)
  
  
  res4 <- page_text("here is some [PG] text")
  
  expect_equal(res4$has_page_numbers, FALSE)
  
})


test_that("basic write_pdf with page lines works as expected.", {
  
  fp <- file.path(base_path, "pdf/direct16.pdf")
  
  l <- list(page_text("Hello", font_size = 14, 
                      xpos = 100, ypos = 100, bold = TRUE),
            page_line(100, 105, 200, 105), 
            page_text("Goodbye", font_size = 14, 
                      xpos = 100, ypos = 120, bold = TRUE),
            page_hline(100, 120, 100),
            page_vline(300, 105, 100), 
            page_hline(300, 105, 100), 
            page_hline(300, 205, 100),
            page_vline(400, 105, 100))
  
  r <- create_pdf(fp, fontname = "Arial") %>% 
    add_page(l)
  
  
  write_pdf(r)
  
  expect_equal(file.exists(fp), TRUE)  
  
})

test_that("basic write_pdf with page box works as expected.", {
  
  fp <- file.path(base_path, "pdf/direct17.pdf")
  
  l <- list(page_text("Hello", font_size = 14, 
                      xpos = 100, ypos = 100, bold = TRUE),
            page_box(90, 85, 20, 50, ""))
  
  r <- create_pdf(fp, fontname = "Arial") %>% 
    add_page(l)
  
  
  write_pdf(r)
  
  expect_equal(file.exists(fp), TRUE)  
  
})

test_that("basic write_pdf with page grid works as expected.", {
  
  fp <- file.path(base_path, "pdf/direct18.pdf")
  
  l <- list(page_grid(90, 85, 5, 4, 20, c(50, 70, 50, 70)),
            page_grid(90, 250, 4, 3, 25, 25))
  
  r <- create_pdf(fp, fontname = "Arial") %>% 
    add_page(l)
  
  
  write_pdf(r)
  
  expect_equal(file.exists(fp), TRUE)  
  
})

# Write Report RTF2 -------------------------------------------------------



#' @title
#' Write a RTF2 report to the file system
#'
#' @description
#' This function writes a report_spec object to the file system, using the
#' parameters provided in the object.
#'
#' @param x The report_spec object to write.
#' @return The report spec.
#' @noRd
write_report_rtf2 <- function(rs) {
  
  orig_path <- rs$modified_path
  
  if (file.exists(orig_path))
    file.remove(orig_path)
  
  # Establish content and body sizes
  rs <- page_setup_rtf(rs)
  
  # Document header is mostly independent of content
  hdr <- get_rtf_document(rs) 
  
  # Put content in a new variable
  ls <- rs$content
  
  # Get content and break it into pages
  # Needs to return a list of pages so preview can work
  # Page numbers need to be included
  bdy <- paginate_content_rtf(rs, ls)
  
  # Get column widths
  rs$column_widths <- bdy[["widths"]]
  
  # Deal with preview
  if (!is.null(rs$preview)) {
    if (rs$preview < length(bdy[[1]]$pages))
      bdy[[1]]$pages <- bdy[[1]]$pages[seq(1, rs$preview)]
  }
  
  # Write content to file system
  # Later we can just return the stream
  rs <- write_content_rtf(rs, hdr, bdy, rs$page_template)
  
  # Update page numbers for title headers
  update_page_numbers_rtf(orig_path, rs$pages)
  
  return(rs)
}


#' @description Returns header for RTF document.  This is independent of content,
#' except for the page header and footer.
#' @noRd
get_rtf_document <- function(rs) {
  
  # Set up vectors
  ret <- c()

  conv <- rs$twip_conversion
  
  fnt <- rs$font
  if (tolower(rs$font) == "times")
    fnt <- "Times New Roman"
  
  # Prepare header
  ret[length(ret) + 1] <- paste0("{\\rtf1\\ansi\\deff0 {\\fonttbl {\\f0 ", fnt , ";}}")
  if (rs$orientation == "landscape") {
    #ret[length(ret) + 1] <- "\\landscape\\horzdoc"
    ret[length(ret) + 1] <- "\\landscape"
    ret[length(ret) + 1] <- paste0("\\paperw", round(rs$page_size[2] * conv),
                                   "\\paperh", round(rs$page_size[1] * conv))
  } else {
    #ret[length(ret) + 1] <- "\\vertdoc"
    ret[length(ret) + 1] <- paste0("\\paperw", round(rs$page_size[1] * conv),
                                   "\\paperh", round(rs$page_size[2] * conv))
  }
  
  ret[length(ret) + 1] <- paste0("\\margl", round(rs$margin_left * conv),
                                 "\\margr", round(rs$margin_right * conv),
                                 "\\margt", round(rs$margin_top * conv),
                                 "\\margb", round(rs$margin_bottom  * conv),
                                 "\\headery", round(rs$margin_top  * conv),
                                 "\\footery", round(rs$margin_bottom  * conv))
  
  ph <- get_page_header_rtf(rs)
  if (ph$rtf != "")
    ret[length(ret) + 1] <- ph$rtf
  
  pf <- get_page_footer_rtf(rs)
  if (pf$rtf != "")
    ret[length(ret) + 1] <- pf$rtf
  
  # Line spacing values determined by trial and error.
  # Needed for LibreOffice.  Appear to be ignored in Word.
  ret[length(ret) + 1] <- paste0(rs$spacing_multiplier, rs$font_rtf)


  
  return(ret)
  
}



#' @noRd
paginate_content_rtf <- function(rs, ls) {
  
  ret <- c()
  last_object <- FALSE
  last_page_lines <- 0
  table_widths <- list()
  
  hrf <- has_bottom_footnotes(rs)

  
  # Loop through content objects
  for (i in seq_along(ls)) {
    
    pgs <- list()  # list of vectors with page rtf lines
    lns <- c()
    
    # Set last object flag
    if (i == length(ls)) {
      last_object <- TRUE
    } else {
      last_object <- FALSE
    }
    
    # Put content and object in variables for convenience
    cntnt <- ls[[i]] 
    obj <- cntnt$object
    
    # Remove blank row if last content object
    cbr <- cntnt$blank_row
    if (last_object) {
      if (all(cbr == "below")) 
        cbr <- "none"
      else if (all(cbr == "all"))
        cbr <- "above"
      
    }
    
    # Break each content type into a list of pages
    if (any(class(obj) == "table_spec")) {
      
      res <- create_table_pages_rtf(rs, cntnt, last_page_lines)
      
      # Collect multiple pages and line counts
      for (j in seq_len(length(res$page_list))) {
        pgs[[length(pgs) + 1]] <- res$page_list[[j]]$rtf
        lns[[length(lns) + 1]] <- res$page_list[[j]]$lines
      }
      
      # Retrieve table widths.  These are useful for debugging.
      # Assigned to returning report object.
      table_widths[[length(table_widths) + 1]] <- res$widths
      
    } else if (any(class(obj) == "text_spec")) {
      
      res <- create_text_pages_rtf(rs, cntnt, last_page_lines, cbr)
      for (j in seq_len(length(res$rtf))) {
        pgs[[length(pgs) + 1]] <- res$rtf[[j]]
        lns[[length(lns) + 1]] <- res$lines[[j]]
      
      }
      
    } else if (any(class(obj) == "plot_spec")) {
      
      # Plots not started rtf2 conversion
      res <- create_plot_pages_rtf(rs, cntnt, last_page_lines, tempdir())
      for (j in seq_len(length(res$rtf))) {
        pgs[[length(pgs) + 1]] <- res$rtf[[j]]
        lns[[length(lns) + 1]] <- res$lines[[j]]
        
      }
    }   
    
    # Store pages and lines with content objects
    # The content settings will be used when writing content
    ls[[i]]$pages <- pgs
    ls[[i]]$lines <- lns
    
    # This section of code is appending blank lines to get
    # footnotes at the bottom of the page.  The complication
    # is when there are multiple pieces of content, and user-defined
    # page breaks.  So these can't be added earlier in
    # the process.  In short, these blanks are for in between
    # pieces of content.  Blanks within a piece of content are 
    # handled in the create_table_*, create_text_*, and create_plot_* 
    # functions.
    
    # Capture last page
    last_page <- pgs[[length(pgs)]]
    
    # Capture number of lines on the last page
    last_page_lines <- lns[[length(lns)]] + last_page_lines
    
    # print(last_page_lines)
    
    if (length(pgs) > 1)
      last_page_lines <- lns[[length(lns)]]
    
    # If there is a page break or it's the last object in the
    # content list, add the blank lines if needed.
    if ((ls[[i]]$page_break | last_object) & hrf) {
      
      
      # Add extra offsets if table has a lot of borders turned on
      # to avoid undesired page wraps
      boff <- 0
      if (any(class(obj) == "table_spec") & 
          any(obj$borders %in% c("all", "inside"))) {
        
        boff <- round(last_page_lines * rs$border_height / rs$row_height)
      }
      
      blnks <- c()
      bl <- rs$body_line_count - last_page_lines - boff
      if (bl > 0)
        blnks <- rep("\\par", bl)
      
      last_page <- append(last_page, blnks)
      last_page_lines <- 0 
      
    }
    
    if (cntnt$page_break == TRUE | last_page_lines >= rs$body_line_count)
      last_page_lines <- 0
    
    ls[[i]]$pages[[length(pgs)]] <- last_page

  }
  
  
  # Can return something else if needed here
  ret <- list(widths = table_widths, pages = ls)
  
  return(ret)
  
}



#' @title Write out content
#' @description This loop writes out pages created in paginate_content
#' Page template items added to the report (titles/footnotes/title_header)
#' are added in this step.  That means these items need to have been accounted
#' for in the previous steps.
#' @noRd
write_content_rtf <- function(rs, hdr, body, pt) {
  
  # Kill existing file
  if (file.exists(rs$modified_path))
    file.remove(rs$modified_path)
  
  counter <- 0
  page <- 0
  last_object <- FALSE
  last_page <- FALSE
  page_open <- FALSE
  
  
  f <- file(rs$modified_path, open="a", encoding = "native.enc")
  
  writeLines(hdr, con = f, useBytes = TRUE)
  

  for (cont in body$pages) {
    
    
    # Increment counter
    counter <- counter + 1
    page <- 0
    
    # Set last_object flag
    if (counter == length(body$pages))
      last_object <- TRUE
    else 
      last_object <- FALSE
    
    
    for (pg in cont$pages) {
      
      page <- page + 1
      
      if (page == length(cont$pages))
        last_page <- TRUE
      else
        last_page <- FALSE
      
      
      #print(page_open)
      if (page_open == FALSE) {

        
        if (!is.null(rs$title_hdr) & !is.null(pt$title_hdr$rtf))
          writeLines(pt$title_hdr$rtf, con = f, useBytes = TRUE)
        
        if (!is.null(rs$titles) & !is.null(pt$titles$rtf))
          writeLines(pt$titles$rtf, con = f, useBytes = TRUE)
        
      }
      
      if (!is.null(pg)) {

        writeLines(pg, con = f, useBytes = TRUE)
        
      }
      
      # Set page_open flag based on status of page_break and current objects
      if (last_object == FALSE & last_page == TRUE & cont$page_break == FALSE)
        page_open <- TRUE
      else 
        page_open <- FALSE
      
      if (page_open == FALSE) {
        
        if (!is.null(rs$footnotes) & !is.null(pt$footnotes$rtf))
          writeLines(pt$footnotes$rtf, con = f, useBytes = TRUE)
        
        if (is.null(rs$pages))
          rs$pages <- 1
        else 
          rs$pages <- rs$pages + 1 
        
        # Add form feed character for text page break
        if (last_object == FALSE | last_page == FALSE) {
          
          if (is.null(cont$object$continuous) || cont$object$continuous == FALSE) {
          
            writeLines(rs$page_break_rtf, con = f, useBytes = TRUE) 
            
          } else {
           
            if (last_page & cont$page_break == TRUE) {
              
              writeLines(rs$page_break_rtf, con = f, useBytes = TRUE) 
            }
            
          }
          
        }
      }
      
      # if (last_object == TRUE & last_page == TRUE) {
      #   
      #   rs$pages <- rs$pages + 1 
      #   
      # }
    }
    
  }
  
  writeLines("{\\pard\\fs1\\sl0  \\par}}", con = f, useBytes = TRUE)
  
  close(f)
  
  return(rs)
  
}


update_page_numbers_rtf <- function(path, tpg) {
  

  lns <- readLines(path, encoding = "UTF-8")
  
  lns <- gsub("[tpg]", tpg, lns, fixed = TRUE)
  
  f <- file(path, open = "w+", encoding = "native.enc")
  
  writeLines(lns, con = f, useBytes = TRUE)
  
  close(f)
  
}




# Setup Functions ---------------------------------------------------------


#' @description Setup page for content
#' @details  Calculates available space for content and prepares text lines
#' for page template.
#' @noRd
page_setup_rtf <- function(rs) {
  
  debug <- FALSE
  
  # Content size is the page size minus margins, in units of measure
  rs$content_size <- get_content_size(rs)
  
  # A lot of these values are guesses.  Need to test.
  # Row height and line height were defined independently in case
  # they are different.  Right now, appear to be the same.
  if (rs$font_size == 8) {
    rh <- 228 #round(.11 * 1440)
    lh <- 228 #round(.1 * 1440) 
    #pb <- "\\fs1\\sl0\\par\\pard\\fs16\\page\\fs1\\sl0\\par\\pard\\fs16"
    pb <- "{\\pard\\pagebb\\fs1\\sl0\\par}\\fs16"
    if (tolower(rs$font) == "courier")
      gtr <- .15
    else
      gtr <- .13 
    cw <- .1
    cp <- 40
   # sm <- "\\sl-180\\slmult0"
  } else if (rs$font_size == 9) {
    rh <- 244 #round(.165 * 1440) # 225
    lh <- 244 #round(.165 * 1440)  
    #pb <- "\\page\\line" #fs1\\sl0\\par\\pard\\fs20"
    pb <-  "{\\pard\\pagebb\\fs1\\sl0\\par}\\fs18"
    if (tolower(rs$font) == "courier")
      gtr <- .16
    else
      gtr <- .14
    cw <- .11
    cp <- 40
    #sm <- "\\sl-200\\slmult0"
  } else if (rs$font_size == 10) {
    rh <- 250 #round(.165 * 1440) # 225
    lh <- 250 #round(.165 * 1440)  
    #pb <- "\\page\\line" #fs1\\sl0\\par\\pard\\fs20"
    pb <-  "{\\pard\\pagebb\\fs1\\sl0\\par}\\fs20"
    if (tolower(rs$font) == "courier")
      gtr <- .17
    else
      gtr <- .15
    cw <- .11
    cp <- 40
    #sm <- "\\sl-225\\slmult0"
  } else if (rs$font_size == 11) {
    rh <- 250 #round(.165 * 1440) # 225
    lh <- 250 #round(.165 * 1440)  
    #pb <- "\\page\\line" #fs1\\sl0\\par\\pard\\fs20"
    pb <-  "{\\pard\\pagebb\\fs1\\sl0\\par}\\fs22"
    if (tolower(rs$font) == "courier")
      gtr <- .18
    else
      gtr <- .16
    cw <- .12
    cp <- 40
    #sm <- "\\sl-250\\slmult0"
  } else if (rs$font_size == 12) {
    rh <- 275 #round(.2 * 1440)
    lh <- 275 #round(.1875 * 1440) #270
    pb <- "{\\pard\\pagebb\\fs1\\sl0\\par}\\fs24"
    if (tolower(rs$font) == "courier")
      gtr <- .19
    else 
      gtr <- .17
    cw <- .12
    cp <- 40
    #sm <- "\\sl-275\\slmult0"
  }
  
  
  # Get conversion factor to twips
  if (rs$units == "inches") {
    conv <- 1440
  } else {
    conv <- 566.9291
  }
  
  rs$twip_conversion <- conv
  rs$row_height <- rh
  rs$line_height <- lh
  rs$char_width <- cw
  rs$line_size <- rs$content_size[["width"]]
  rs$cell_padding <- cp
  rs$spacing_multiplier <- get_spacing_multiplier(rs$font_size)
  rs$page_break_rtf <- paste0(pb, rs$spacing_multiplier)
  rs$border_height <- 15
  
  # Line spacing values determined by trial and error.
  # Needed for LibreOffice.  Appear to be ignored in Word.
  if (rs$font_size == 10) {
    rs$font_rtf <-  "\\f0\\fs20"
  } else if (rs$font_size == 12) {
    rs$font_rtf <-  "\\f0\\fs24"
  } else if (rs$font_size == 11) {
    rs$font_rtf <-  "\\f0\\fs22"
  } else if (rs$font_size == 9) {
    rs$font_rtf <-  "\\f0\\fs18"
  } else if (rs$font_size == 8) {
    rs$font_rtf  <- "\\f0\\fs16"
  }
  
  
  if (rs$units == "cm")
    rs$gutter_width <- ccm(gtr)
  else 
    rs$gutter_width <- gtr
  
  if (is.null(rs$user_line_count)) {
    # There is one row above the page footer that is not printable.
    # Therefore adjust by 1.
    rs$line_count <- floor(rs$content_size[[1]] * conv / rh) - 1
  } else 
    rs$line_count <- rs$user_line_count
  
  if (debug) {
    print(paste("Content Height:", rs$content_size[[1]]))
    print(paste("Content Width:", rs$content_size[[2]]))
    print(paste("Line Count:", rs$line_count))
    print(paste("Line Height:", rs$line_height))
    print(paste("Gutter Width:", rs$gutter_width))
    print(paste("Char Width:", rs$char_width))
  }
  
  # Get page template
  pt <- page_template_rtf(rs)
  rs$page_template <- pt

  # Body size in twips
  # Small adjustment by one line height
  # This gets used to determine lines on a page.
  rs$body_size <- 
    c(height = floor((rs$content_size[[1]] * conv) - pt$page_header$twips - pt$page_footer$twips - lh), 
      width = floor(rs$content_size[[2]] * conv))
  
  if (debug) {
    print(paste("Body Height:", rs$body_size[[1]]))
    print(paste("Body Width:", rs$body_size[[2]]))
  }
  
  
  # Get the page template row count
  # Include all the rows associated with the page template
  rs$page_template_header_count <- sum(pt$page_header$lines, pt$titles$lines, 
    pt$title_hdr$lines, pt$page_by$lines)
  if (debug)
    print(paste("Page Template Header Count:", rs$page_template_header_count))
  
  rs$page_template_footer_count <- sum(pt$footnotes$lines, pt$page_footer$lines)
  if (debug)
    print(paste("Page Template Footer Count:", rs$page_template_footer_count))
  
  rs$page_template_row_count <- rs$page_template_header_count + 
    rs$page_template_footer_count
  if (debug)
    print(paste("Page Template Row Count:", rs$page_template_row_count))
  
  # Body line count is the number of rows available for content on each page
  rs$body_line_count <- rs$line_count - rs$page_template_row_count
  if (debug)
    print(paste0("Body Line Count: ", rs$body_line_count))
  
  return(rs)
}

#' @noRd
get_spacing_multiplier <- function(font_size) {
  
  if (font_size == 8) {
    sm <- "\\sl-228\\slmult0" 
  } else if (font_size == 9) {
    sm <- "\\sl-244\\slmult0" 
  } else if (font_size == 10) {
    sm <- "\\sl-250\\slmult0"
  } else if (font_size == 11) {
    sm <- "\\sl-250\\slmult0" 
  } else if (font_size == 12) {
    sm <- "\\sl-275\\slmult0"
  } else if (font_size == 13) {
    sm <- "\\sl-300\\slmult0"
  } else if (font_size == 14) {
    sm <- "\\sl-325\\slmult0"
  } else {
    sm <- "\\sl-250\\slmult0" 
  }
  
  return(sm)
}

# Write Report DOCX -------------------------------------------------------



#' @title
#' Write an DOCX report to the file system
#'
#' @description
#' This function writes a report_spec object to the file system, using the
#' parameters provided in the object.
#'
#' @param x The report_spec object to write.
#' @return The report spec.
#' @noRd
write_report_docx <- function(rs) {
  
  orig_path <- rs$modified_path
  
  if (file.exists(orig_path)) {
    file.remove(orig_path)
  }
  
  # Establish content and body sizes
  rs <- page_setup_docx(rs)
  
  
  # Document header is mostly independent of content
  hdr <- get_docx_document(rs) 
  
  # Put content in a new variable
  ls <- rs$content
  
  
  # Get content and break it into pages
  # Needs to return a list of pages so preview can work
  # Page numbers need to be included
  bdy <- paginate_content_docx(rs, ls)
  
  # Get column widths
  rs$column_widths <- bdy[["widths"]]
  
  # Deal with preview
  # if (!is.null(rs$preview)) {
  #   if (rs$preview < length(bdy[[1]]$pages))
  #     bdy[[1]]$pages <- bdy[[1]]$pages[seq(1, rs$preview)]
  # }

  
  # Write content to file system
  # Later we can just return the stream
  rs <- write_content_docx(rs, hdr, bdy, rs$page_template)
  
  # Update page numbers for title headers
 # update_page_numbers_docx(orig_path, rs$pages)
  
  return(rs)
}



#' @description Returns header for DOCX document.  This is independent of content,
#' except for the page header and footer.
#' @noRd
get_docx_document <- function(rs) {
  
  # Set up vectors
  ret <- c()
  
  conv <- rs$twip_conversion
  
  fnt <- rs$font
  if (tolower(rs$font) == "times")
    fnt <- "Times New Roman"
  
  # Prepare header
  ret[length(ret) + 1] <- '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<w:document
	xmlns:wpc="http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas"
	xmlns:cx="http://schemas.microsoft.com/office/drawing/2014/chartex"
	xmlns:cx1="http://schemas.microsoft.com/office/drawing/2015/9/8/chartex"
	xmlns:cx2="http://schemas.microsoft.com/office/drawing/2015/10/21/chartex"
	xmlns:cx3="http://schemas.microsoft.com/office/drawing/2016/5/9/chartex"
	xmlns:cx4="http://schemas.microsoft.com/office/drawing/2016/5/10/chartex"
	xmlns:cx5="http://schemas.microsoft.com/office/drawing/2016/5/11/chartex"
	xmlns:cx6="http://schemas.microsoft.com/office/drawing/2016/5/12/chartex"
	xmlns:cx7="http://schemas.microsoft.com/office/drawing/2016/5/13/chartex"
	xmlns:cx8="http://schemas.microsoft.com/office/drawing/2016/5/14/chartex"
	xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006"
	xmlns:aink="http://schemas.microsoft.com/office/drawing/2016/ink"
	xmlns:am3d="http://schemas.microsoft.com/office/drawing/2017/model3d"
	xmlns:o="urn:schemas-microsoft-com:office:office"
	xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships"
	xmlns:m="http://schemas.openxmlformats.org/officeDocument/2006/math"
	xmlns:v="urn:schemas-microsoft-com:vml"
	xmlns:wp14="http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing"
	xmlns:wp="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing"
	xmlns:w10="urn:schemas-microsoft-com:office:word"
	xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main"
	xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml"
	xmlns:w15="http://schemas.microsoft.com/office/word/2012/wordml"
	xmlns:w16cex="http://schemas.microsoft.com/office/word/2018/wordml/cex"
	xmlns:w16cid="http://schemas.microsoft.com/office/word/2016/wordml/cid"
	xmlns:w16="http://schemas.microsoft.com/office/word/2018/wordml"
	xmlns:w16sdtdh="http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash"
	xmlns:w16se="http://schemas.microsoft.com/office/word/2015/wordml/symex"
	xmlns:wpg="http://schemas.microsoft.com/office/word/2010/wordprocessingGroup"
	xmlns:wpi="http://schemas.microsoft.com/office/word/2010/wordprocessingInk"
	xmlns:wne="http://schemas.microsoft.com/office/word/2006/wordml"
	xmlns:wps="http://schemas.microsoft.com/office/word/2010/wordprocessingShape" 
	mc:Ignorable="w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14">'
	
	ret[length(ret) + 1] <- '<w:body>'
  

  # Assume landscape
  pg_h <- rs$page_size[1]
  pg_w <- rs$page_size[2]
  
  # Change to portrait
  if(rs$orientation == "portrait") {
    pg_w <- rs$page_size[1]
    pg_h <- rs$page_size[2]
  }
  
  ret[length(ret) + 1] <- paste0('<w:sectPr w:rsidR="00444C49" w:rsidRPr="00444C49">
    <w:headerReference w:type="default" r:id="rId6"/>
    <w:footerReference w:type="default" r:id="rId7"/>',
    '<w:pgSz w:w="', pg_w * conv, '" w:h="', pg_h * conv, '" w:orient="', 
    rs$orientation, '"/>',
    '<w:pgMar w:top="', rs$margin_top * conv, '" w:right="', rs$margin_right * conv, '" ', 
    'w:bottom="', rs$margin_bottom * conv, '" w:left="', rs$margin_left * conv, '" ',
    'w:header="', rs$margin_top * conv, '" w:footer="', rs$margin_bottom * conv, 
    '" w:gutter="0"/>',
    '<w:cols w:space="720"/>
    <w:docGrid w:linePitch="360"/>
    </w:sectPr>')
  
  return(ret)
  
}



#' @noRd
paginate_content_docx <- function(rs, ls) {
  
  ret <- c()
  last_object <- FALSE
  last_page_lines <- 0
  table_widths <- list()
  imgCnt <- 0
  imgPaths <- c()
  
  hrf <- has_bottom_footnotes(rs)
  
  
  # Loop through content objects
  for (i in seq_along(ls)) {
    
    pgs <- list()  # list of vectors with page rtf lines
    lns <- c()
    imgs <- list()
    
    # Set last object flag
    if (i == length(ls)){
      last_object <- TRUE
    } else {
      last_object <- FALSE
    } 
    
    # Put content and object in variables for convenience
    cntnt <- ls[[i]] 
    obj <- cntnt$object
    
    # Remove blank row if last content object
    cbr <- cntnt$blank_row
    if (last_object) {
      if (all(cbr == "below")) 
        cbr <- "none"
      else if (all(cbr == "all"))
        cbr <- "above"
      
    }
    
    # Break each content type into a list of pages
    if (any(class(obj) == "table_spec")) {
      
      res <- create_table_pages_docx(rs, cntnt, last_page_lines)
      
      # Collect multiple pages and line counts
      for (j in seq_len(length(res$page_list))) {
        pgs[[length(pgs) + 1]] <- res$page_list[[j]]$docx
        lns[[length(lns) + 1]] <- res$page_list[[j]]$lines
      }
      
      # Retrieve table widths.  These are useful for debugging.
      # Assigned to returning report object.
      table_widths[[length(table_widths) + 1]] <- res$widths
      
    } else if (any(class(obj) == "text_spec")) {
      
      res <- create_text_pages_docx(rs, cntnt, last_page_lines, cbr)
      for (j in seq_len(length(res$docx))) {
        pgs[[length(pgs) + 1]] <- res$docx[[j]]
        lns[[length(lns) + 1]] <- res$lines[[j]]
        
      }
      
    } else if (any(class(obj) == "plot_spec")) {
      
      res <- create_plot_pages_docx(rs, cntnt, last_page_lines, tempdir(), imgCnt)
      for (j in seq_len(length(res$docx))) {
        pgs[[length(pgs) + 1]] <- res$docx[[j]]
        lns[[length(lns) + 1]] <- res$lines[[j]]
        imgs[[length(imgs) + 1]] <- res$images[[j]]
        imgCnt <- imgCnt + 1
        imgPaths[[length(imgPaths) + 1]] <- res$images[[j]]
      }
    } else {
      
      stop(paste("Invalid content: ", class(obj))) 
    }
    
    # Store pages and lines with content objects
    # The content settings will be used when writing content
    ls[[i]]$pages <- pgs
    ls[[i]]$lines <- lns
    ls[[i]]$images <- imgs
    
    # This section of code is appending blank lines to get
    # footnotes at the bottom of the page.  The complication
    # is when there are multiple pieces of content, and user-defined
    # page breaks.  So these can't be added earlier in
    # the process.  In short, these blanks are for in between
    # pieces of content.  Blanks within a piece of content are 
    # handled in the create_table_*, create_text_*, and create_plot_* 
    # functions.
    
    # Capture last page
    last_page <- pgs[[length(pgs)]]
    
    # Capture number of lines on the last page
    last_page_lines <- lns[[length(lns)]] + last_page_lines
    
    # print(last_page_lines)
    
    if (length(pgs) > 1)
      last_page_lines <- lns[[length(lns)]]
    
    # If there is a page break or it's the last object in the
    # content list, add the blank lines if needed.
    if (rs$paper_size != "none") {
      if ((ls[[i]]$page_break | last_object) & hrf) {


        # Add extra offsets if table has a lot of borders turned on
        # to avoid undesired page wraps
        boff <- 0
        if (any(class(obj) == "table_spec") &
            any(obj$borders %in% c("all", "inside"))) {

          boff <- round(last_page_lines * rs$border_height / rs$row_height)
          #boff <- 1
        }

        blnks <- c()
        bl <- rs$body_line_count - last_page_lines - boff
        if (bl > 0)
          blnks <- rep(rs$blank_row_below, bl)

        last_page <- append(last_page, blnks)
        last_page_lines <- 0

      } 
    }
    
    if (cntnt$page_break == TRUE | last_page_lines >= rs$body_line_count)
      last_page_lines <- 0

    ls[[i]]$pages[[length(pgs)]] <- last_page
    
  }
  
  
  # Can return something else if needed here
  ret <- list(widths = table_widths, pages = ls, imageCount = imgCnt, 
              imagePaths = imgPaths)
  
  return(ret)
  
}


# Could be consolidated with RTF.
#' @title Write out content
#' @description This loop writes out pages created in paginate_content
#' Page template items added to the report (titles/footnotes/title_header)
#' are added in this step.  That means these items need to have been accounted
#' for in the previous steps.
#' @noRd
write_content_docx <- function(rs, hdr, body, pt) {
  

  
  counter <- 0
  page <- 0
  last_object <- FALSE
  last_page <- FALSE
  page_open <- FALSE
  imgCnt <- 0
  
  
  
  # Create new document in temp location
  tf <- create_new_docx(rs$font, rs$font_size, body$imageCount, body$imagePaths, 
                        rs$page_template$page_footer$image_path, rs$page_template$page_header$image_path)
  
  # Write out header
  create_header(tf, rs$page_template$page_header$docx)
  
  # Write out footer
  create_footer(tf, rs$page_template$page_footer$docx)
  
  fp <- file.path(tf, "word/document.xml")
  
  f <- file(fp, open="a", encoding = "native.enc")
  

  writeLines(hdr, con = f, useBytes = TRUE)
  
  
 # writeLines(body, con = f, useBytes = TRUE)
  
  # Copy images to document folder for footer
  f_imgCnt <- 0
  if (!is.null(rs$page_template$page_footer$image_path)) {
    for (im in rs$page_template$page_footer$image_path) { 
      f_imgCnt <- f_imgCnt + 1
      ext <- tools::file_ext(im)
      if (ext == "jpg") {
        ext <- "jpeg"
      }
      ifp <- file.path(tf, paste0("word/media/imagef", f_imgCnt, ".", ext))
      file.copy(im, ifp)   
    }
  }
  
  # Copy images to document folder for header
  h_imgCnt <- 0
  if (!is.null(rs$page_template$page_header$image_path)) {
    for (im in rs$page_template$page_header$image_path) { 
      h_imgCnt <- h_imgCnt + 1
      ext <- tools::file_ext(im)
      if (ext == "jpg") {
        ext <- "jpeg"
      }
      ifp <- file.path(tf, paste0("word/media/imageh", h_imgCnt, ".", ext))
      file.copy(im, ifp)   
    }
  }
  
  # Derive total page for later use
  tpg <- sum(sapply(body$pages, function(x){length(x$pages)}))
  
  for (cont in body$pages) {
    
    # Copy images to document folder
    if (length(cont$images) > 0) {
      for (im in cont$images) { 
        imgCnt <- imgCnt + 1
        ext <- tools::file_ext(im)
        if (ext == "jpg") {
          ext <- "jpeg"
        }
        ifp <- file.path(tf, paste0("word/media/image", imgCnt, ".", ext))
        file.copy(im, ifp)
      
      }
    }


    # Increment counter
    counter <- counter + 1
    page <- 0

    # Set last_object flag
    if (counter == length(body$pages))
      last_object <- TRUE
    else
      last_object <- FALSE

    # ta <- "align=\"left\" "
    # if (cont$align == "right")
    #   ta <- "align=\"right\" "
    # else if (cont$align %in% c("center", "centre"))
    #   ta <- "align=\"center\" "


    for (pg in cont$pages) {

      page <- page + 1

      if (page == length(cont$pages))
        last_page <- TRUE
      else
        last_page <- FALSE


      #print(page_open)
      if (page_open == FALSE) {

        # if (!is.null(rs$page_template$page_header) &
        #     !is.null(rs$page_template$page_header$html))
        #   writeLines(update_page(rs$page_template$page_header$html,  rs$pages),
        #              con = f, useBytes = TRUE)

        # Write content div to keep page together
        # writeLines(paste0("<div ", ta, ">"), con = f, useBytes = TRUE)


        if (!is.null(rs$title_hdr) & !is.null(pt$title_hdr$docx))
          writeLines(update_page_docx(pt$title_hdr$docx,  rs$pages, tpg), con = f,
                     useBytes = TRUE)

        if (!is.null(rs$titles) & !is.null(pt$titles$docx))
          writeLines(update_page_docx(pt$titles$docx, rs$pages, tpg), con = f,
                     useBytes = TRUE)

      }

      if (!is.null(pg)) {
        
        # Replace page numbers in title/footnote of report/content
        pg <- page_replace_docx(pg, rs, "titles", rs$pages, tpg)
        pg <- page_replace_docx(pg, rs, "footnotes", rs$pages, tpg)
        pg <- page_replace_docx(pg, cont$object, "titles", rs$pages, tpg)
        pg <- page_replace_docx(pg, cont$object, "footnotes", rs$pages, tpg)
        
        # if (!is.null(rs$titles)) {
        #   for (rt in rs$titles) {
        #     for (rt_string in rt$titles) {
        #       if (rt_string != "") {
        #         raw_title <- rt_string
        #         new_title <- update_page_docx(raw_title,  rs$pages, tpg)
        #         pg <- gsub(raw_title, new_title, pg, fixed = TRUE)
        #       }
        #     }
        #   }
        # }
        
        writeLines(pg, con = f, useBytes = TRUE)
      }

      # Set page_open flag based on status of page_break and current objects
      if (last_object == FALSE & last_page == TRUE & cont$page_break == FALSE)
        page_open <- TRUE
      else
        page_open <- FALSE

      if (page_open == FALSE) {

        if (!is.null(rs$footnotes) & !is.null(pt$footnotes$docx))
          writeLines(update_page_docx(pt$footnotes$docx, rs$pages, tpg),
                     con = f, useBytes = TRUE)

        # Content div
        # writeLines("</div>", con = f, useBytes = TRUE)

        # if (!is.null(rs$page_template$page_footer) &
        #     !is.null(rs$page_template$page_footer$html))
        #   writeLines(update_page(rs$page_template$page_footer$html, rs$pages),
        #              con = f, useBytes = TRUE)


        # Add form feed character for text page break
        if (last_object == FALSE | last_page == FALSE) {

          if (is.null(rs$pages))
            rs$pages <- 1
          else
            rs$pages <- rs$pages + 1

          writeLines(rs$page_break_docx, con = f, useBytes = TRUE)

        }
      }

      if (last_object == TRUE & last_page == TRUE) {

        rs$pages <- rs$pages + 1

      }
    }

  }
  
  # Close document body
  writeLines("</w:body>\n</w:document>", con = f, useBytes = TRUE)
  
  close(f)
  
  
  # Kill existing file
  if (file.exists(rs$modified_path))
    file.remove(rs$modified_path)
  
  
  # Copy from temp path to perm path and zip
  write_docx(tf, rs$modified_path)
  
  
  return(rs)
  
}

#' @description Could be consolidated with RTF
#' @noRd
update_page_numbers_docx <- function(path, tpg) {
  
  
  lns <- readLines(path, encoding = "UTF-8")
  
  lns <- gsub("[tpg]", tpg, lns, fixed = TRUE)
  
  f <- file(path, open = "w+", encoding = "native.enc")
  
  writeLines(lns, con = f, useBytes = TRUE)
  
  close(f)
  
}

update_page_docx <- function(lns, pg, tpg = NULL) {
  
 ret <- gsub("[pg]", pg + 1, lns, fixed = TRUE) 
 
 if (!is.null(tpg)) {
   ret <- gsub("[tpg]", tpg, ret, fixed = TRUE) 
 }
 
 return(ret)
  
}


# Setup Functions ---------------------------------------------------------



#' @description Setup page for content
#' @details  Calculates available space for content and prepares text lines
#' for page template.
#' @noRd
page_setup_docx <- function(rs) {
  
  debug <- FALSE
  
  if (is.null(rs$font_size))
    rs$font_size <- 10
  
  if (is.null(rs$font))
    rs$font <- "Courier"
  
  if (rs$font == "fixed")
    rs$font <- "Courier"
  
  if (rs$font_size == 8) {
    
    gtr <- .12
    cw <- .1    # na
    radj <- 56
    
  } else if (rs$font_size == 9) {
    
    gtr <- .13
    cw <- .1  # na
    radj <- 36

    
  } else if (rs$font_size == 10) {
    
    if (tolower(rs$font) == "courier")
      gtr <- .14
    else 
      gtr <- .15
    
    cw <- .11   # na
    radj <- 8

  } else if (rs$font_size == 11) {
    
    if (tolower(rs$font) == "courier")
      gtr <- .15
    else 
      gtr <- .16
    
    cw <- .11  # na
    radj <- -16
    
  } else if (rs$font_size == 12) {
    
    if (tolower(rs$font) == "courier")
      gtr <- 0.16
    else
      gtr <- 0.16
    
    cw <- .12  #na
    radj <- -42
  }
  
  rh <- get_rh(rs$font, rs$font_size)
  
  rs$border_height <- 1/72/2
  
  if (rs$units == "cm") {
    rh <- ccm(rh)
    cw <- ccm(cw)
    rs$border_height <- ccm(rs$border_height)
    gtr <- ccm(gtr)
  }
  
  # Get conversion factor to twips
  if (rs$units == "inches") {
    conv <- 1440
  } else {
    conv <- 566.9291
  }
  
  # A zero height paragraph to break between tables.
  # Otherwise, Word will treat as one table and the column
  # widths will be messed up.
  rs$table_break <- '<w:p><w:pPr>
              				<w:spacing w:after="0" w:line="0" w:lineRule="none"/>
              				<w:rPr>
              					<w:sz w:val="0"/>
              				</w:rPr>
              			</w:pPr></w:p>\n'
  
  rs$blank_row <- paste0('<w:p><w:pPr>
              				<w:spacing w:after="8" w:line="', round(rh * conv) + radj,
              				'" w:lineRule="auto"/>
              				<w:contextualSpacing/>
              				<w:rPr>
              					<w:sz w:val="', rs$font_size * 2, '"/>
              				</w:rPr>
              			</w:pPr></w:p>\n')
  
  # This is for adding extra blanks after the table, left more buffer to prevent from
  # unexpected page break
  rs$blank_row_below <- paste0('<w:p><w:pPr>
              				<w:spacing w:after="8" w:line="', round(rh * conv) + radj - 3,
                         '" w:lineRule="auto"/>
              				<w:contextualSpacing/>
              				<w:rPr>
              					<w:sz w:val="', rs$font_size * 2, '"/>
              				</w:rPr>
              			</w:pPr></w:p>\n')
  
  rs$cell_margin <- paste0('<w:tblCellMar>
                           <w:left w:w="32" w:type="dxa"/>
                           <w:right w:w="32" w:type="dxa"/>
                           </w:tblCellMar>')
  
  # The starting point for relationship IDs.
  # This is used when adding images to the document.
  # See plot_spec.
  rs$relIndex <- 9
  
  rs$twip_conversion <- conv
  
  rs$row_height <- rh
  rs$line_height <- rh
  rs$char_width <- cw
  rs$base_indent <- 130
  
  # Content size is the page size minus margins, in units of measure
  rs$content_size <- get_content_size(rs)
  rs$line_size <- rs$content_size[["width"]]
  

  rs$gutter_width <- gtr

  
  # rs$page_break_docx <- paste0('<w:p>
  #                     <w:pPr>
  #                     <w:pStyle w:val="Normal"/>
  #             				<w:spacing w:after="0" w:before="0" w:line="120" w:lineRule="auto"/>
  #             				<w:contextualSpacing/>
  #             				  <w:rPr>
  #             					<w:sz w:val="0"/>
  #             				  </w:rPr>
  #             				</w:pPr>
  #             				<w:r>
  #             				  <w:rPr>
  #             					<w:sz w:val="0"/>
  #             				  </w:rPr>
  #             				</w:r>
  #               			<w:r>
  #               				<w:br w:type="page"/>
  #               			</w:r>
  #               		</w:p>')
  
  
  # rs$page_break_docx <- paste0('<w:p><w:pPr><w:pStyle w:val="Normal"/>',
  #                              '<w:bidi w:val="0"/><w:jc w:val="left"/><w:rPr></w:rPr>',
  #                              '</w:pPr><w:r><w:rPr></w:rPr></w:r>', 
  #                              '<w:r><w:br w:type="page"/></w:r></w:p>')
  # 
  # 
  # rs$page_break_docx <- paste0('<w:p><w:pPr><w:pStyle w:val="Normal"/>',
  #    '<w:spacing w:after="0" w:before="0" w:line="120" w:lineRule="auto"/>',
  #    '<w:bidi w:val="0"/><w:jc w:val="left"/><w:rPr></w:rPr>',
  #    '</w:pPr><w:r><w:rPr><w:sz w:val="0"/></w:rPr></w:r>', 
  #    '<w:r><w:br w:type="page"/></w:r></w:p>')
  
  
  # rs$page_break_docx <- paste0('<w:p><w:pPr>',
  #    '<w:rPr><w:sz w:val="0"/></w:rPr>',
  #    '</w:pPr><w:r><w:rPr><w:sz w:val="0"/></w:rPr></w:r>', 
  #    '<w:r><w:br w:type="page"/></w:r></w:p>')
  
  rs$page_break_docx <- paste0('<w:p><w:pPr>',
     '<w:spacing w:after="0" w:before="0" w:line="0" w:lineRule="none"/>',
     '<w:pageBreakBefore/>',
     '<w:rPr><w:sz w:val="0"/></w:rPr>',
     '</w:pPr><w:r><w:rPr><w:sz w:val="0"/></w:rPr></w:r>', 
     '</w:p>')
  

  if (is.null(rs$user_line_count)) {
    rs$line_count <- round(rs$content_size[[1]] / rh) 
  } else
    rs$line_count <- rs$user_line_count

  if (debug) {
    print(paste("Font Size:", rs$font_size))
    print(paste("Content Height:", rs$content_size[[1]]))
    print(paste("Content Width:", rs$content_size[[2]]))
    print(paste("Line Count:", rs$line_count))
    print(paste("Line Height:", rs$line_height))
    print(paste("Gutter Width:", rs$gutter_width))
    print(paste("Char Width:", rs$char_width))
  }
  
  
  # Create temp directory structure for docx file
  #rs$temp_dir_docx <- create_new_docx()

  # Get page template
  pt <- page_template_docx(rs)
  rs$page_template <- pt


  # Get the page template row count
  # Include all the rows associated with the page template
  rs$page_template_header_count <- sum(max(pt$page_header$lines_content, 1), 
                                       pt$titles$lines_content,
                                       pt$title_hdr$lines, pt$page_by$lines)
  
  # DOCX would still have one blank row on top even no page headers
  # rs$page_template_header_count <- max(c(1, rs$page_template_header_count))
  
  if (debug)
    print(paste("Page Template Header Count:", rs$page_template_header_count))

  rs$page_template_footer_count <- sum(pt$footnotes$lines_content, 
                                       pt$page_footer$lines_content)
  if (debug)
    print(paste("Page Template Footer Count:", rs$page_template_footer_count))

  rs$page_template_row_count <- rs$page_template_header_count +
    rs$page_template_footer_count
  if (debug)
    print(paste("Page Template Row Count:", rs$page_template_row_count))

  # Body line count is the number of rows available for content on each page
  # - 1 adjustment needed for footer buffer
  rs$body_line_count <- rs$line_count - rs$page_template_row_count - 1

  if (debug)
    print(paste0("Body Line Count: ", rs$body_line_count))
  
  return(rs)
}

get_rh <- function(font, font_size) {
  
  rh <- 0
  
  if (font_size == 8) {
    
    if (tolower(font) == "times")
      rh <- 0.141  
    else if (tolower(font) == "arial")
      rh <- 0.141 
    else 
      rh <- 0.141
    
    
  } else if (font_size == 9) {
    
    if (tolower(font) == "times")
      rh <- 0.156
    else if (tolower(font) == "arial")
      rh <- 0.156 
    else 
      rh <- 0.155 
    
  } else if (font_size == 10) {
    
    if (tolower(font) == "times")
      rh <- 0.178  
    else if (tolower(font) == "arial")
      rh <- 0.178 #0.182  # 0.1585366
    else
      rh <- 0.175 #0.182  # 0.1585366
    
  } else if (font_size == 11) {
    
    if (tolower(font) == "times")
      rh <- 0.195
    else if (tolower(font) == "arial")
      rh <- 0.195 # 0.168
    else 
      rh <- 0.192 # 0.168
    
    
  } else if (font_size == 12) {
    
    # inches 
    if (tolower(font) == "times")
      rh <- 0.212  # 1911765 
    else if (tolower(font) == "arial")
      rh <- 0.212 #0.212  # 1911765 
    else 
      rh <- 0.212  # 1911765 
    
  } else if (font_size == 14) {
    
    # inches 
    if (tolower(font) == "times")
      rh <- 0.25 
    else if (tolower(font) == "arial")
      rh <- 0.25 
    else 
      rh <- 0.25 
    
  }
  
  return(rh)
}


get_rh2 <- function(font, font_size) {
  
  rh <- 0
  
  if (font_size == 8) {
    
    if (tolower(font) == "times")
      rh <- 0.144  
    else if (tolower(font) == "arial")
      rh <- 0.144 
    else 
      rh <- 0.144
    
    
  } else if (font_size == 9) {
    
    if (tolower(font) == "times")
      rh <- 0.166
    else if (tolower(font) == "arial")
      rh <- 0.166 
    else 
      rh <- 0.1625 
    
  } else if (font_size == 10) {
    
    if (tolower(font) == "times")
      rh <- 0.185  
    else if (tolower(font) == "arial")
      rh <- 0.185 #0.182  # 0.1585366
    else
      rh <- 0.185 #0.182  # 0.1585366
    
  } else if (font_size == 11) {
    
    if (tolower(font) == "times")
      rh <- 0.208
    else if (tolower(font) == "arial")
      rh <- 0.208 # 0.168
    else 
      rh <- 0.203 # 0.168

    
  } else if (font_size == 12) {
    
    # inches 
    if (tolower(font) == "times")
      rh <- 0.23  # 1911765 
    else if (tolower(font) == "arial")
      rh <- 0.23 #0.212  # 1911765 
    else 
      rh <- 0.23  # 1911765 

  } else if (font_size == 14) {
    
    # inches 
    if (tolower(font) == "times")
      rh <- 0.25 
    else if (tolower(font) == "arial")
      rh <- 0.25 
    else 
      rh <- 0.25 
    
  }
  
  return(rh)
}

#' @description Replace page number in title and footnote
#' @details  Replace page number in title and footnote
#' @noRd
page_replace_docx <- function(pg, target, type = "titles", page, tpg) {
  if (!is.null(target[[type]])) {
    for (v in target[[type]]) {
      for (v_string in v[[type]]) {
        if (v_string != "") {
          raw_title <- v_string
          new_title <- update_page_docx(raw_title, page, tpg)
          pg <- gsub(raw_title, new_title, pg, fixed = TRUE)
        }
      }
    }
  }
  return(pg)
}

# Write Report PDF2 -------------------------------------------------------



#' @title
#' Write a PDF2 report to the file system
#'
#' @description
#' This function writes a report_spec object to the file system, using the
#' parameters provided in the object.
#'
#' @param x The report_spec object to write.
#' @return The report spec.
#' @noRd
write_report_pdf2 <- function(rs) {
  

  if (file.exists(rs$modified_path))
    file.remove(rs$modified_path)
  
  # Establish content and body sizes
  rs <- page_setup_pdf(rs)
  
  # Document header is mostly independent of content
  doc <- create_pdf(filename = rs$modified_path,
                  margin_top = rs$margin_top,
                  margin_left = rs$margin_left,
                  fontname = rs$font,
                  fontsize = rs$font_size,
                  page_height = rs$page_size[2],
                  page_width = rs$page_size[1],
                  orientation = rs$orientation,
                  units = rs$units,
                  conversion = rs$point_conversion,
                  info = TRUE)


  
  # Get content and break it into pages
  # Needs to return a list of pages so preview can work
  # Page numbers need to be included
  doc <- paginate_content_pdf(rs, doc)
  
  # Get column widths
  rs$column_widths <- doc[["widths"]]
  
  # Deal with preview
  if (!is.null(rs$preview)) {
    if (rs$preview < length(doc$pages))
      doc$pages <- doc$pages[seq(1, rs$preview)]
  }
  
  # Write content to file system
  # Later we can just return the stream
  doc <- write_pdf(doc)
  rs$pages <- doc$pages
  
  # Update page numbers for title headers
  #update_page_numbers_rtf(orig_path, rs$pages)
  
  return(rs)
}


#' @description Returns header for RTF document.  This is independent of content,
#' except for the page header and footer.
#' @noRd
get_pdf_document <- function(rs) {
  
  # Set up vectors
  ret <- c()
  
  conv <- rs$twip_conversion
  
  fnt <- rs$font
  if (tolower(rs$font) == "times")
    fnt <- "Times New Roman"
  
  # Prepare header
  r <- create_pdf(filename = rs$modified_path,
                  margin_top = rs$margin_top,
                  margin_left = rs$margin_left,
                  fontsize = rs$font_size,
                  page_height = rs$page_size[2],
                  page_width = rs$page_size[1],
                  orientation = rs$orientation,
                  units = rs$units,
                  info = TRUE)
  
  # ph <- get_page_header_pdf(rs)
  # if (ph$pdf != "")
  #   ret[length(ret) + 1] <- ph$pdf
  # 
  # pf <- get_page_footer_pdf(rs)
  # if (pf$pdf != "")
  #   ret[length(ret) + 1] <- pf$pdf
  

  
  
  
  return(ret)
  
}



#' @noRd
paginate_content_pdf <- function(rs, doc) {
  
  # Put content in a new variable
  ls <- rs$content
  ret <- c()
  last_object <- FALSE
  last_page_lines <- 0
  table_widths <- list()
  
  # Get report titles and footnotes
  rttls <- rs$page_template$titles
  rttl_hdr <- rs$page_template$title_hdr
  rftnts <- rs$page_template$footnotes
  rheader <- rs$page_template$page_header
  rfooter <- rs$page_template$page_footer
  
  hrf <- has_bottom_footnotes(rs)
  
  
  # Loop through content objects
  for (i in seq_along(ls)) {
    
    pgs <- list()  # list of vectors with page pdf lines
    lns <- c()
    
    # Set last object flag
    if (i == length(ls)) {
      last_object <- TRUE
    } else {
      last_object <- FALSE
    }
    
    # Put content and object in variables for convenience
    cntnt <- ls[[i]] 
    obj <- cntnt$object
    
    # Remove blank row if last content object
    cbr <- cntnt$blank_row
    if (last_object) {
      if (all(cbr == "below")) 
        cbr <- "none"
      else if (all(cbr == "all"))
        cbr <- "above"
      
    }
    
    # Break each content type into a list of pages
    if (any(class(obj) == "table_spec")) {
      
      res <- create_table_pages_pdf(rs, cntnt, last_page_lines)
      
      for (j in seq_len(length(res$page_list))) {
        
        if (j == 1 & last_page_lines > 0 & 
            (ceiling(last_page_lines + res$page_list[[j]]$lines) <= rs$body_line_count)) {
          
          doc$pages[[length(doc$pages)]] <- append(doc$pages[[length(doc$pages)]],
                                                   res$page_list[[j]]$pdf)
          
          
        } else {
        
          doc <- add_page(doc, res$page_list[[j]]$pdf, 
                          rttls$pdf, rttl_hdr$pdf, 
                          rftnts$pdf, rheader$pdf, rfooter$pdf)
        

        }
        
        if (j == length(res$page_list)) {
          
          last_page_lines <- last_page_lines + res$page_list[[j]]$lines
        } else {
          last_page_lines <- 0 
        }
      }
      # 
      # # Collect multiple pages and line counts
      # for (j in seq_len(length(res$page_list))) {
      #   pgs[[length(pgs) + 1]] <- res$page_list[[j]]$pdf
      #   lns[[length(lns) + 1]] <- res$page_list[[j]]$lines
      # }
      # 
      # Retrieve table widths.  These are useful for debugging.
      # Assigned to returning report object.
      table_widths[[length(table_widths) + 1]] <- res$widths
      
    } else if (any(class(obj) == "text_spec")) {
      
      res <- create_text_pages_pdf(rs, cntnt, last_page_lines, cbr)
      for (j in seq_len(length(res$pdf))) {
        
        if (j == 1 & last_page_lines > 0 & 
            (ceiling(last_page_lines + res$lines[[j]]) <= rs$body_line_count)) {
          
          doc$pages[[length(doc$pages)]] <- append(doc$pages[[length(doc$pages)]],
                                                   res$pdf[[j]])
        
          
        } else {
          
          doc <- add_page(doc, res$pdf[[j]], rttls$pdf, rttl_hdr$pdf, 
                          rftnts$pdf, rheader$pdf, rfooter$pdf) 
        
        }
        
        if (j == length(res$pdf)) {
          
          last_page_lines <- last_page_lines + res$lines[[j]]
        } else {
          last_page_lines <- 0 
        }
      }
      
      # for (j in seq_len(length(res$pdf))) {
      #   pgs[[length(pgs) + 1]] <- res$pdf[[j]]
      #   lns[[length(lns) + 1]] <- res$lines[[j]]
      #   
      # }
      
    } else if (any(class(obj) == "plot_spec")) {
      
      # Plots not started rtf2 conversion
      # res <- create_plot_pages_pdf(rs, cntnt, last_page_lines, tempdir())
      # for (j in seq_len(length(res$pdf))) {
      #   pgs[[length(pgs) + 1]] <- res$pdf[[j]]
      #   lns[[length(lns) + 1]] <- res$lines[[j]]
      #   
      # }
      
      res <- create_plot_pages_pdf(rs, cntnt, last_page_lines, tempdir())
      for (j in seq_len(length(res$pdf))) {
        
        if (j == 1 & last_page_lines > 0 & 
            (ceiling(last_page_lines + res$lines[[j]]) <= rs$body_line_count)) {
          
          doc$pages[[length(doc$pages)]] <- append(doc$pages[[length(doc$pages)]],
                                                   res$pdf[[j]])
          
          
          
        } else {
        
          doc <- add_page(doc, res$pdf[[j]], rttls$pdf, rttl_hdr$pdf, 
                          rftnts$pdf, rheader$pdf, rfooter$pdf) 
        
        }
        
        if (j == length(res$pdf)) {
          
          last_page_lines <- last_page_lines + res$lines[[j]]
        } else {
          last_page_lines <- 0 
        }
      }
    }
    
    if (cntnt$page_break == TRUE | last_page_lines >= rs$body_line_count)
      last_page_lines <- 0
    
    # Store pages and lines with content objects
    # The content settings will be used when writing content
    # ls[[i]]$pages <- pgs
    # ls[[i]]$lines <- lns
    
    # This section of code is appending blank lines to get
    # footnotes at the bottom of the page.  The complication
    # is when there are multiple pieces of content, and user-defined
    # page breaks.  So these can't be added earlier in
    # the process.  In short, these blanks are for in between
    # pieces of content.  Blanks within a piece of content are 
    # handled in the create_table_*, create_text_*, and create_plot_* 
    # functions.
    
    # Capture last page
    # last_page <- pgs[[length(pgs)]]
    # 
    # # Capture number of lines on the last page
    # last_page_lines <- lns[[length(lns)]] + last_page_lines
    
    # print(last_page_lines)
    # 
    # if (length(pgs) > 1)
    #   last_page_lines <- lns[[length(lns)]]
    # 
    # # If there is a page break or it's the last object in the
    # # content list, add the blank lines if needed.
    # if ((ls[[i]]$page_break | last_object) & hrf) {
    #   
    #   
    #   # Add extra offsets if table has a lot of borders turned on
    #   # to avoid undesired page wraps
    #   boff <- 0
    #   if (any(class(obj) == "table_spec") & 
    #       any(obj$borders %in% c("all", "inside"))) {
    #     
    #     boff <- round(last_page_lines * rs$border_height / rs$row_height)
    #   }
    #   
    #   blnks <- c()
    #   bl <- rs$body_line_count - last_page_lines - boff
    #   if (bl > 0)
    #     blnks <- rep("\\par", bl)
    #   
    #   last_page <- append(last_page, blnks)
    #   last_page_lines <- 0 
    #   
    # }
    
    # ls[[i]]$pages[[length(pgs)]] <- last_page
    
  }
  
  doc$widths <- table_widths
  
  # Can return something else if needed here
  # ret <- list(widths = table_widths, pages = ls)
  
  return(doc)
  
}



update_page_numbers_pdf <- function(path, tpg) {
  
  
  lns <- readLines(path, encoding = "UTF-8")
  
  lns <- gsub("[tpg]", tpg, lns, fixed = TRUE)
  
  f <- file(path, open = "w+", encoding = "native.enc")
  
  writeLines(lns, con = f, useBytes = TRUE)
  
  close(f)
  
}




# Setup Functions ---------------------------------------------------------


#' @description Setup page for content
#' @details  Calculates available space for content and prepares text lines
#' for page template.
#' @noRd
page_setup_pdf <- function(rs) {
  
  debug <- FALSE
  
  # Content size is the page size minus margins, in units of measure
  rs$content_size <- get_content_size(rs)
  

  if (rs$font_size == 8) {

    gtr <- .1 
    cw <- .1

  } else if (rs$font_size == 9) {

    gtr <- .11
    cw <- .11

  } else if (rs$font_size == 10) {
   
    gtr <- .11
    cw <- .11

  } else if (rs$font_size == 11) {
  
    gtr <- .11
    cw <- .12

  } else if (rs$font_size == 12) {

    gtr <- .11
    cw <- .12

  }
  
  # Line height is for text
  lh <- get_line_height_pdf(rs$font_size)
  
  # Row height is for tables.  Right now both the same.
  rh <- lh
  
  
  # Get conversion factor to points
  if (rs$units == "inches") {
    conv <- 72
  } else {
    conv <- 1/2.54 * 72 
  }
  
  rs$point_conversion <- conv
  rs$row_height <- rh
  rs$line_height <- lh
  rs$char_width <- cw
  rs$line_size <- rs$content_size[["width"]]
  rs$cell_padding <- 2
  rs$border_height <- 2   # How much to add to the line height if there is a border
  rs$border_spacing <- 3  # Amount to offset the border from the bottom of the line
  
  # Assume landscape
  pg_h <- rs$page_size[1]
  pg_w <- rs$page_size[2]
  
  # Change to portrait
  if(rs$orientation == "portrait") {
    pg_w <- rs$page_size[1]
    pg_h <- rs$page_size[2]
  }
  
  if (rs$units == "cm")
    rs$gutter_width <- ccm(gtr)
  else 
    rs$gutter_width <- gtr
  
  if (is.null(rs$user_line_count)) {
    rs$line_count <- floor(rs$content_size[[1]] * conv / rh) 
  } else 
    rs$line_count <- rs$user_line_count
  
  if (debug) {
    print(paste("Content Height:", rs$content_size[[1]]))
    print(paste("Content Width:", rs$content_size[[2]]))
    print(paste("Line Count:", rs$line_count))
    print(paste("Line Height:", rs$line_height))
    print(paste("Gutter Width:", rs$gutter_width))
    print(paste("Char Width:", rs$char_width))
  }
  
  # Get page template
  pt <- page_template_pdf(rs)
  rs$page_template <- pt
  
  # Body size in twips
  # Small adjustment by one line height
  # This gets used to determine lines on a page.
  rs$body_size <- 
    c(height = floor((rs$content_size[[1]] * conv) - pt$page_header$points - pt$page_footer$points), 
      width = floor(rs$content_size[[2]] * conv))
  
  if (debug) {
    print(paste("Body Height:", rs$body_size[[1]]))
    print(paste("Body Width:", rs$body_size[[2]]))
  }
  
  
  # Get the page template row count
  # Include all the rows associated with the page template
  rs$page_template_header_count <- sum(pt$page_header$lines, pt$titles$lines, 
                                       pt$title_hdr$lines, pt$page_by$lines)
  if (debug)
    print(paste("Page Template Header Count:", rs$page_template_header_count))
  
  rs$page_template_footer_count <- sum(pt$footnotes$lines, pt$page_footer$lines)
  if (debug)
    print(paste("Page Template Footer Count:", rs$page_template_footer_count))
  
  rs$page_template_row_count <- rs$page_template_header_count + 
    rs$page_template_footer_count
  if (debug)
    print(paste("Page Template Row Count:", rs$page_template_row_count))
  
  # Body line count is the number of rows available for content on each page
  rs$body_line_count <- rs$line_count - rs$page_template_row_count
  if (debug)
    print(paste0("Body Line Count: ", rs$body_line_count))
  
  return(rs)
}

# Row heights determined by trial and error on RTF and converted to points
# for PDF (divide by 20).
#' @noRd
get_line_height_pdf <- function(font_size) {

  if (font_size == 8) {
    sm <- 185 / 20
  } else if (font_size == 9) {
    sm <- 218 / 20
  } else if (font_size == 10) {
    sm <- 228 / 20
  } else if (font_size == 11) {
    sm <- 250 / 20
  } else if (font_size == 12) {
    sm <- 275 / 20  #275
  } else if (font_size == 13) {
    sm <- 300 / 20
  } else if (font_size == 14) {
    sm <- 325 /20
  } else {
    sm <- 250 / 20
  }

  return(sm)
}

# Text Sizing Functions --------------------------------------------------------

# @description Splits the data according to height
# @details Logic is to try and estimate the vertical height of each cell
# in a row to get the overall row height.  Then break rows into pages based
# on available body height.  This function is only used for non-text output.
# For text output, get_splits_text is used.
# @import graphics
# @import qlcMatrix
# @import stringi
# @noRd
# get_splits <- function(dat, col_widths, data_size, font_family) {
#   
#   # Row height fixed for now
#   # Ultimately needs to be dynamic based on font size and letting
#   rh <- .3 #strheight("Test String", units = "inches", family = font_family)
#   
#   ws <- list()
#   for (i in seq_along(dat)) {
#     
#     w_all <- strwidth(dat[[i]], units="inches", family=font_family)
#     ws[[length(ws) + 1]] <- w_all
#   }
#   
#   
#   m <- matrix(unlist(ws), ncol=length(ws), byrow=FALSE)
#   # print(m)
#   
#   #lin <- ceiling(m /col_widths)
#   lin <- ceiling(t(t(m) / col_widths))
#   
#   # print(lin)
#   
#   # print(col_widths)
#   
#   # Get counts of carriage returns per cell
#   
#   rs <- list()
#   for (col in seq_along(dat)) {
#     rs[[length(rs)+1]] <- stri_count(dat[[col]], fixed = "\n")
#   }
#   
#   r <- matrix(unlist(rs), ncol=length(rs), byrow=FALSE)
#   
#   # Add carriage returns to line counts
#   lr <- lin + r
#   # print(lr)
#   
#   lr[is.na(lr) | lr == 0] <- 1
#   
#   # print(lr)
#   # print(rowMax(lr)@x)
#   
#   row_heights <- rowMax(lr, ignore.zero = FALSE)@x * rh
#   
#   # print(row_heights)
#   # print(length(row_heights))
#   
#   row_pages <- get_pages(row_heights, data_size["height"])
#   
#   # print(nrow(dat))
#   # print(length(row_pages))
#   # print(row_pages)
#   
#   ret <- split(dat, row_pages)
#   
#   
#   return(ret)
#   
# }

#' @description Subset the data by row and column. Depends on whether
#' the user supplied a preview page number.
#' @noRd
get_data_subset <- function(dat, keys, pages) {
 
  
  if (!is.null(pages)) {
    est <- 60 * pages
    if (nrow(dat) > est) {
      ret <- dat[seq(1, est ), keys]
    } else {
      ret <- dat[ , keys]
    }
    
  } else {
    
    ret <- dat[ , keys]
  }
   
  return(ret)
}

#' Gets the page wraps for both text and variable width reports
#' @noRd
get_page_wraps <- function(content_width, ts, widths, gutter, control_cols) {
  
  defs <- ts$col_defs
  
  # Get ID variable from definitions
  # These need to be shown on each page
  # Also get page wraps
  if (!is.null(ts$stub)) 
    id_vars <- c("stub")
  else
    id_vars <- c()
  
  wraps <- c()
  nms <- c()
  for (def in defs) {
    if (!is.null(def$id_var) && def$id_var)
      id_vars[length(id_vars) + 1] <- def$var_c
    
    wraps[length(wraps) + 1] <- def$page_wrap
    nms[length(nms) + 1] <- def$var_c
  }
  names(wraps) <- nms
  
  #print(id_vars)
  
  ret <- list() # list of columns for each page
  pg <- c()     # columns on a page
  tw <- content_width  # width of the page
  
  for (nm in names(widths)) {
    
    if (!is.control(nm)) {
      #If ID vars exist, add them to list
      if (length(pg) == 0 && length(id_vars) > 0) {
        
        # Plus gutter in between columns
        pg <- widths[id_vars] + gutter
        
        if (any(is.na(pg)))
          stop(paste0("ID column width for '", 
                      paste(id_vars[is.na(pg)], sep = " ", collapse = ""),
                      " not found."))
        
        names(pg) <- id_vars
        
      }
      
      # Force a page wrap if requested in definition
      force_wrap <- FALSE
      if (is.null(wraps) == FALSE) 
        if (is.na(wraps[nm]) == FALSE) 
          if (wraps[nm] == TRUE)
            force_wrap <- TRUE
      
      # print(paste("Name:", nm))
      # print(paste("Force wrap:", force_wrap))
      # print(paste("Sum of widths:", sum(pg, widths[nm] + 1)))
      # print(paste("Page:", pg))
      # print(paste("Widths:", widths[nm]))
      # print(paste("Total width:", tw))
      # print("")
      
      
      if ((sum(pg, widths[nm]) >  tw | force_wrap) & (!nm %in% id_vars)) {
        
        # If sum of widths exceed page size, add page to list and reset pg
        # Also add control cols so downstream functions can use them
        ret[[length(ret) + 1]] <- c(names(pg), control_cols)
        pg <- c()
        
        # Add widths for ID vars
        if (length(id_vars) > 0) {
          pg <- widths[id_vars] + gutter
          names(pg) <- id_vars
        }
        
        # Add width for current column
        pg[nm] <- widths[nm] + gutter
        
      } else {
        
        # If sum of widths does not exceed page size, add to pg and keep going
        pg[nm] <- widths[nm] + gutter
        
      } 
    }
    
  }
  
  # Pick any remaining columns
  if (length(pg) > 0) {
    # Add page to list
    ret[[length(ret) + 1]] <- c(names(pg), control_cols)
  }
  
  
  return(ret)
  
}



#' @description Preps the data
#' @details 
#' This function performs a variety of tasks needed to prepare the data
#' for printing.  This function must be called before page wrapping and 
#' splitting because it will add rows and columns which must be taken into
#' account during those operations.  Tasks performed include several 
#' options on the variable define function.  These include blank_after, 
#' label_row, indenting, and creating stub columns. 
#' @noRd
prep_data <- function(dat, ts, char_width, missing_val) {
  
  defs <- ts$col_defs
  # print("Before prep data")
  # print(dat)
  
  # Get vector of columns for blank rows
  ls_after <- c()
  ls_before <- c()
  for (def in defs) {

    if (def$blank_after) {
      ls_after[length(ls_after) + 1] <- translate_invisible(def$var_c, names(dat))
    }
    
    if (def$blank_before) {
      ls_before[length(ls_before) + 1] <- translate_invisible(def$var_c, names(dat))
    }
  }

  # Add after blanks on requested columns
  if (!is.null(ls_after)) {
    if (length(ls_after) > 0) {
      
      # Reverse order so groups turn out correct
      ls_after <- ls_after[order(ls_after, decreasing = TRUE)]
      
      # Add blanks
      if (length(ls_after) > 0) {
        dat <- add_blank_rows(dat, location = "below", vars = ls_after)
      }
    }
  }

   # print("Blanks")
    #print(dat)
  
  # Set up label rows
  for (def in defs) {
    if (def$label_row) {
      
      # Convert to character if necessary
      if (all(class(dat[[def$var_c]]) != "character"))
        dat[[def$var_c]] <- as.character(dat[[def$var_c]])
      
      dat <- add_blank_rows(dat, "label", vars = def$var_c)
    }
  }
  
  # Add before blanks on requested columns
  # This step should be after label rows so that the blank rows' order are correct 
  if (!is.null(ls_before)) {
    if (length(ls_before) > 0) {
      
      # Reverse order so groups turn out correct
      ls_before <- ls_before[order(ls_before, decreasing = TRUE)]
      
      # Add blanks
      if (length(ls_before) > 0) {
        dat <- add_blank_rows(dat, location = "above", vars = ls_before)
      }
    }
  }
  
  # Add cell border indicator
  group_border_vars <- c()
  for (def in defs) {
    if (def$group_border) {
      group_border_vars <- c(group_border_vars, def$var_c)
    }
  }
  if (length(group_border_vars) > 0) {
    dat <- add_group_border_ind(dat, group_border_vars)
  }

   # print("Labels")
   # print(dat)
  
  # Indent and Dedupe variables as requested
  # Do this after adding blanks
  # So any group values in blank rows are removed
  # if (blank_indent) {
  #   for (def in defs) {
  #     if (!is.null(def$indent) | def$dedupe) {
  #       
  #       # Convert to character if necessary
  #       if (all(class(dat[[def$var_c]]) != "character"))
  #         dat[[def$var_c]] <- as.character(dat[[def$var_c]])
  #       
  #       # Actual deduping now takes place in get_splits_text, so
  #       # label appears at top of each page
  #       
  #     }
  #     
  #     # Perform Indenting of requested variables
  #     if (!is.null(def$indent)) {
  #       ind <- floor(def$indent / char_width)
  #       blnks <- paste0(rep(" ", ind), sep = "", collapse = "")
  #       dat[[def$var_c]] <- ifelse(is.na(dat[[def$var_c]]), NA, paste0(blnks, dat[[def$var_c]]))
  #     }
  #   }
  # }
   # print("Before stub")
   # print(dat)

  # Create stub
  dat <- create_stub(dat, ts)
  
   #print("After stub")
   #print(dat)
    
  # Clear out missing values 
  dat <- clear_missing(dat, missing_val)

  
  return(dat)
  
}


create_stub <- function(dat, ts) {
 
  if (!is.null(ts$stub)) {
    
    s <- ts$stub 
    v <- s$vars
    
    # Initialize with first column 
    st <- dat[[v[1]]]
    stv <- rep(names(dat)[1], nrow(dat))
    #print(st)
    
    nms <- names(dat)
    
    # For each subsequent column, if the value is not NA,
    # replace first column value.
    for (i in seq(from = 2, to = length(v), by = 1)) {
      
      if ("..blank" %in% nms) {
        st <- ifelse(is.na(dat[[v[i]]]) | (trimws(dat[[v[i]]]) == "" & dat[["..blank"]] == "L"), 
                   st, dat[[v[i]]])
        # stub var to indicate the value source
        stv <- ifelse(is.na(dat[[v[i]]]) | (trimws(dat[[v[i]]]) == "" & dat[["..blank"]] == "L"), 
                      stv, v[i])
      } else {
        st <- ifelse(is.na(dat[[v[i]]]), st, dat[[v[i]]])  
        stv <- ifelse(is.na(dat[[v[i]]]), stv, v[i])  
      }
      
      # Allow empty to remain.  Sure I put this in for a reason.  Need to test.
      #st <- ifelse(is.na(dat[[v[i]]]) | trimws(dat[[v[i]]]) == "", st, dat[[v[i]]])    
      
    }
    
    dat$..stub_var <- stv

    # Remove stub variables from data frame
    d <-   dat[ , -which(names(dat) %in% v)]
    
    # print("Names before")
    # print(names(d))
    
    # Combine with stub column
    dat <- data.frame(stub = st, d, stringsAsFactors = FALSE)
    
    # Have to restore names, otherwise data.frame will mess them up
    names(dat) <- c("stub", names(d))
    
    # Issue 373: Remove row label when it's NA. If it's "", will consider it as
    # a real value
    if ("..blank" %in% names(dat)) {
      dat <- dat[!(is.na(dat$stub) & dat$..blank == "L"),]
      # dat <- dat[!(dat$stub == "" & dat$..blank == "L"),]
    }
    
    # print("Names after")
    # print(names(dat))
  }
  
  return(dat)
  
}

#' Get the column widths
# @import graphics
#' @import stringi
#' @noRd
get_col_widths <- function(dat, ts, labels, char_width, uom, 
                           merge_label_row = TRUE) {

  defs <- ts$col_defs
  if (uom == "cm") {
    #12.7
    max_col_width = 12.7  
    min_col_width = .254
  } else if (uom == "inches") {
    max_col_width = 5  
    min_col_width = .1
  } else if (uom == "char") {
    max_col_width = 60
    min_col_width = 2
  }
  

  nms <- names(labels)
  #print(nms)
  dwidths <- c()
  mwidths <- c()
  
  # Set default widths based on length of data
  for (nm in nms) {
    
    #w <- max(strwidth(dat[[nm]], units="inches", family=font_family))
    if (is.control(nm) | all(is.na(dat[[nm]]) == TRUE))
      w <- 0
    else {
      
      # Clear out label rows, as these can mess up column width calculations.
      # Label row widths are dealt with later.
      if ("..blank" %in% names(dat) & merge_label_row) {
        colattr <- attributes(dat[[nm]])
        dat[[nm]] <- ifelse(dat[["..blank"]] %in% c("L", "B", "A"), " ", dat[[nm]])
        attributes(dat[[nm]]) <- colattr
      }
      
      w <- max(nchar(as.character(dat[[nm]])), na.rm = TRUE) * char_width
      
      sd <- stri_split(as.character(dat[[nm]]), regex=" |\n|\r|\t", simplify = TRUE)
      mwidths[[nm]]  <- max(nchar(as.character(sd)), na.rm = TRUE) * char_width 

    }
     
    if (w > max_col_width)
      w <- max_col_width
    else if (w < min_col_width)
      w <- min_col_width
    else
      w <- (ceiling(w * 100)/100) 
    
    # print(paste("w:", w))
    # print(paste("Label:", labels[[nm]]))
    
    # Determine width of words in label for this column
    s <- stri_split(labels[[nm]], regex=" |\n|\r|\t", simplify = TRUE)
    #l <- strwidth(s[[1]], units="inches", family=font_family)
    l <- max(nchar(as.character(s)), na.rm = TRUE) * char_width
    
    # print(paste("s:", s))
    # print(paste("l:", l))
    
    # If the max word width is greater than the data width,
    # set column width to max label word width
    # so as not to break any words in the label
    if (max(l) > w)
      dwidths[[nm]] <- max(l) + char_width
    else
      dwidths[[nm]] <- w + char_width
  }
  
  # Set names for easy access
  #names(dwidths) <- names(dat)
  
  # Set default widths
  ret <-  dwidths
  #print("Default Widths")
  #print(ret)
  
  # Let widths on orig df override defaults
  orig <- widths(dat)
  for (nm in names(orig)) {
    ret[[nm]] <- orig[[nm]]
  }

  defnms <- c()
  # Let user definitions override everything
  for (def in defs) {
    
    if (def$var_c %in% names(dat) & !is.null(def$width) && def$width > 0) {
      if (def$width >= mwidths[[def$var_c]])
        ret[[def$var_c]] <- def$width
      else 
        ret[[def$var_c]] <- mwidths[[def$var_c]] + char_width
      
      defnms[length(defnms) + 1] <- def$var_c
    }
    
  }
  
  # Deal with stub
  if (!is.null(ts$stub)) {
 
    # Add stub width if exists   
    if (!is.null(ts$stub$width)) {
      ret[["stub"]] <- ts$stub$width
      defnms[length(defnms) + 1] <- "stub"
    }
    
    # print(ret[["stub"]])
    # print(ts$stub$width)
  }
  
  # Turn into vector if needed
  ret <- unlist(ret)
  
  # Remove control columns
  ret <- ret[!sapply(names(ret), is.control)]
  
  # Deal with table width
  if (!is.null(ts$width)) {
     # print(paste("Table width:", ts$width))
    
     # Adjusted to keep 
     #blnkw <- (length(ret) - 0) * char_width
     # print(paste("Blank width:", blnkw))
     # print(paste("Before:", sum(ret)))
     # print(ret)
     
    if (sum(ret) < ts$width) {
      
      # Get columns not defined by user
      variable <- ret[!names(ret) %in% defnms]
      fixed <- ret[names(ret) %in% defnms]
      
      available <- ts$width - sum(fixed)
      
      variable <- variable * available / sum(variable)
      
      for (nm in names(variable))
        ret[[nm]] <- variable[[nm]]
      
      # print("diff")
      # Add one character to first column so it fills the entire page
      ret[[1]] <- ret[[1]] + char_width

    } else {
      # If table width is less than the sum of the columns.
      # This means the table should wrap to the next page, 
      # but each page should be set to the table width.
      # Having a hard time figuring out how to do this.
      # Will come back to it later.
      # print("Here")
      # ret2 <- c()
      # pg <- c() 
      # for (i in seq_along(ret)) {
      #   print(ret[i])
      #   print(paste("sumpg:", sum(pg)))
      #   print(paste("lengthpg:", length(pg)))
      #   print(paste("char_width:", char_width))
      #   blnkw <- (length(pg) - 1) * char_width
      #   tot <- sum(pg) + blnkw
      #   print(paste("tot:", tot))
      #   if (tot + ret[i] > ts$width) {
      # 
      #     pg <- pg * ((ts$width - blnkw) / sum(pg))
      #     ret2 <- c(ret2, pg)
      #     print(ret2)
      #     pg <- c(ret[[i]])
      #   } else {
      #     pg[length(pg) + 1] <- ret[[i]]
      #     print(paste("pg:", pg))
      #   }
      #   
      # }
      # 
      # if (length(pg) > 0) {
      #   blnkw <- (length(pg) - 1) * char_width
      #   pg <- pg * ((ts$width - blnkw) / sum(pg))
      #   ret2 <- c(ret2, pg)
      #   
      # }
      # 
      # names(ret2) <- names(ret)
      # print(paste("ret2:", ret2))
      # ret <- ret2
    }
    
    # print(paste("After:", sum(ret)))
    # print(ret)
     
  }
  

  
  return(ret)
}



#' Get the column widths
#' @import stringi
#' @noRd
get_col_widths_variable <- function(dat, ts, labels, font, 
                               font_size, uom, gutter_width,
                               merge_label_row = TRUE) {
  
  dat_orig <- dat
  defs <- ts$col_defs
  if (uom == "cm") {
    #12.7
    max_col_width = 12.7  
    min_col_width = .254
  } else if (uom == "inches") {
    max_col_width = 5  
    min_col_width = .1
  } else if (uom == "char") {
    max_col_width = 60
    min_col_width = 2
  }
  
  nms <- names(labels)
  #print(nms)
  dwidths <- c()
  mwidths <- c()
  
  # Set default widths based on length of data
  for (nm in nms) {
    

    if (is.control(nm) | all(is.na(dat[[nm]]) == TRUE))
      w <- 0
    else {
      
      # Clear out label rows, as these can mess up column width calculations.
      # Label row widths are dealt with later.
      blank_idx <- rep(F, dim(dat)[1])
      if ("..blank" %in% names(dat) & merge_label_row) {
        dat[[nm]] <- ifelse(dat[["..blank"]] %in% c("L", "B", "A"), " ", dat[[nm]])
        blank_idx <- dat[["..blank"]] %in% c("L", "B", "A")
      }
      
      # --------------------------------- #
      #   Maximum width of a column
      # --------------------------------- #
      w_each <-  get_text_width(dat[[nm]], units=uom, font=font, font_size = font_size)

      # Add indention width
      if (!is.null(defs[[nm]]$indent)) {
        
        w_each <- w_each + defs[[nm]]$indent
        
      } else if (!is.null(ts$stub) & nm == "stub") {
        stub_var <- ts$stub$vars
        
        for (v in stub_var) {
          if (!is.null(defs[[v]]$indent)) {
            idx <- dat$..stub_var == v & !is.na(dat$stub) & blank_idx == F
            w_each[idx] <- w_each[idx] + defs[[v]]$indent
          }
        }
      }
      
      w <-  max(w_each)
      
      # --------------------------------- #
      #   Minimum width of a column
      # --------------------------------- #
      if (!is.null(ts$stub) & nm == "stub") {
        mwidths[[nm]] <- 0
        
        # For stub column, derive maximum word width by ..stub_var because each
        # stub var may have different indenting width
        stub_var <- ts$stub$vars
        for (v in stub_var) {
          idx <- dat$..stub_var == v & !is.na(dat$stub) & blank_idx == F
          sd <- stri_split(as.character(dat[[nm]][idx]), regex=" |\n|\r|\t", simplify = TRUE)
          mwidth <- max(get_text_width(as.character(sd), units=uom, font=font, font_size = font_size))
          
          if (!is.null(defs[[v]]$indent)) {
            mwidth <- mwidth + defs[[v]]$indent
          }
          
          mwidths[[nm]] <- max(mwidths[[nm]], mwidth)
        }
      } else {
        sd <- stri_split(as.character(dat[[nm]]), regex=" |\n|\r|\t", simplify = TRUE)

        # Would prefer to not calculate this again.  w and mwidths seem redundant.
        mwidths[[nm]]  <- max(get_text_width(as.character(sd), units=uom,
                              font=font, font_size = font_size))
        
        if (!is.null(defs[[nm]]$indent)) {
          mwidths[[nm]] <- mwidths[[nm]] + defs[[nm]]$indent
        }
      }
      
    }
    
    if (w > max_col_width)
      w <- max_col_width
    else if (w < min_col_width)
      w <- min_col_width
    else
      w <- (ceiling(w * 100)/100) 
    
    # print(paste("w:", w))
    # print(paste("Label:", labels[[nm]]))
    
    # Determine width of words in label for this column
    s <- stri_split(labels[[nm]], regex=" |\n|\r|\t", simplify = TRUE)
    l <- max(get_text_width(as.character(s), units=uom, 
                            font=font, font_size = font_size))
    
    # print(paste("s:", s))
    # print(paste("l:", l))
    
    # If the max word width is greater than the data width,
    # set column width to max label word width
    # so as not to break any words in the label
    if (max(l) > w)
      dwidths[[nm]] <- max(l) + gutter_width
    else
      dwidths[[nm]] <- w + gutter_width
  }
  
  # Set names for easy access
  #names(dwidths) <- names(dat)
  
  # Set default widths
  ret <-  dwidths
  #print("Default Widths")
  #print(ret)
  
  # Let widths on orig df override defaults
  orig <- widths(dat_orig)
  for (nm in names(orig)) {
    ret[[nm]] <- orig[[nm]]
  }
  
  defnms <- c()
  # Let user definitions override everything
  for (def in defs) {
    
    if (def$var_c %in% names(dat) & !is.null(def$width) && def$width > 0) {
      if (def$width >= mwidths[[def$var_c]]) {
        ret[[def$var_c]] <- def$width
      }
      else {
        ret[[def$var_c]] <- mwidths[[def$var_c]] + gutter_width
      }
      
      defnms[length(defnms) + 1] <- def$var_c
    }
    
  }
  
  # Deal with stub
  if (!is.null(ts$stub)) {
    
    # Add stub width if exists   
    if (!is.null(ts$stub$width)) {
      if (ts$stub$width >= mwidths[["stub"]]) {
        ret[["stub"]] <- ts$stub$width
      } else {
        ret[["stub"]] <- mwidths[["stub"]] + gutter_width
      }
      defnms[length(defnms) + 1] <- "stub"
    }
    
    # print(ret[["stub"]])
    # print(ts$stub$width)
  }
  
  # Turn into vector if needed
  ret <- unlist(ret)
  
  # Remove control columns
  ret <- ret[!sapply(names(ret), is.control)]
  
  # Deal with table width
  if (!is.null(ts$width)) {
    # print(paste("Table width:", ts$width))
    
    # Adjusted to keep 
    #blnkw <- (length(ret) - 0) * char_width
    # print(paste("Blank width:", blnkw))
    # print(paste("Before:", sum(ret)))
    # print(ret)
    
    if (sum(ret) < ts$width) {
      
      # Get columns not defined by user
      variable <- ret[!names(ret) %in% defnms]
      fixed <- ret[names(ret) %in% defnms]
      
      available <- ts$width - sum(fixed)
      
      variable <- variable * available / sum(variable)
      
      for (nm in names(variable))
        ret[[nm]] <- variable[[nm]]
      
      # print("diff")
      # Add one character to first column so it fills the entire page
      #ret[[1]] <- ret[[1]] + gutter_width
      
    } else {
      # If table width is less than the sum of the columns.
      # This means the table should wrap to the next page, 
      # but each page should be set to the table width.
      # Having a hard time figuring out how to do this.
      # Will come back to it later.
      # print("Here")
      # ret2 <- c()
      # pg <- c() 
      # for (i in seq_along(ret)) {
      #   print(ret[i])
      #   print(paste("sumpg:", sum(pg)))
      #   print(paste("lengthpg:", length(pg)))
      #   print(paste("char_width:", char_width))
      #   blnkw <- (length(pg) - 1) * char_width
      #   tot <- sum(pg) + blnkw
      #   print(paste("tot:", tot))
      #   if (tot + ret[i] > ts$width) {
      # 
      #     pg <- pg * ((ts$width - blnkw) / sum(pg))
      #     ret2 <- c(ret2, pg)
      #     print(ret2)
      #     pg <- c(ret[[i]])
      #   } else {
      #     pg[length(pg) + 1] <- ret[[i]]
      #     print(paste("pg:", pg))
      #   }
      #   
      # }
      # 
      # if (length(pg) > 0) {
      #   blnkw <- (length(pg) - 1) * char_width
      #   pg <- pg * ((ts$width - blnkw) / sum(pg))
      #   ret2 <- c(ret2, pg)
      #   
      # }
      # 
      # names(ret2) <- names(ret)
      # print(paste("ret2:", ret2))
      # ret <- ret2
    }
    
    # print(paste("After:", sum(ret)))
    # print(ret)
    
  }
  
  
  
  return(ret)
}


# widths not incorporated yet
# Gets the data size
# @import graphics
# @noRd
# get_data_size <- function(body_size, widths, labels, font_family) {
#   
#   #ppi = 72
#   blank_row <- .3
#   
#   sz <- c()
#   for (n in labels) {
#     sz[length(sz) + 1] <- strheight(n, units="inches", family=font_family)
#     
#   }
#   
#   ret <- body_size
#   
#   ret["height"] <- ret["height"] - max(sz) - blank_row
#   
#   return(ret)
#   
# }


get_label_aligns <- function(ts, aligns) {
  
  defs <- ts$col_defs
  
  # Default to column aligns
  ret <- aligns

  # Let any defined value override the default
  has_cell_border <- FALSE
  for (d in defs) {
    if (!is.null(d$label_align) & d$var_c %in% names(aligns))
      ret[[d$var_c]] <-  d$label_align
    
    if (!is.null(d[["group_border"]])) {
      if (d[["group_border"]] == TRUE) {
        has_cell_border <- TRUE
      }
    }
  }
  
  # Set a value so that it won't be NA in pi
  if (has_cell_border) {
    ret[["..group_border"]] <- "left"
  }
  
  # Deal with stub
  if (!is.null(ts$stub)) {
  
    # Assing label align if exists
    if (is.null(ts$stub$label_align))
      ret[["stub"]] <- ts$stub[["align"]] 
    else
      ret[["stub"]] <- ts$stub[["label_align"]] 
    
  }
  
  return(ret)
}

#' Gets column formats from definitions
#' @noRd
get_col_formats <- function(dat, ts) {
  
  defs <- ts$col_defs
  ret <- c()
  
  
  # Get any existing formats
  if ("format" %in% ts$use_attributes) {
    ret <- formats(dat)
  }

  for (d in defs) {
    if (!is.null(d$format))
      ret[[d$var_c]] <-  d$format
  }

  
  return(ret)
}


#' Gets the alignments
#' @noRd
get_aligns <- function(dat, ts) {
  
  defs <- ts$col_defs
  nms <- names(dat)
  ret <- c()
  
  # Get default alignments
  # based on data type
  # Character will go to left
  # Others goes to right
  for (nm in nms) {
    
    if (is.character(dat[[nm]]))
      ret[length(ret) + 1] <- "left"
    else
      ret[length(ret) + 1] <- "right"
    
  }
  
  # Assign names to vector for easy access to alignment values
  names(ret) <- nms

  # Get any justification from original data frame
  if ("justify" %in% ts$use_attributes) {
    orig <- justification(dat)
    for (nm in names(orig))
      ret[[nm]] <- orig[[nm]]
  }
  
  # Assign alignments from column definitions
  has_cell_border <- FALSE
  for (d in defs) {
    if (!is.null(d$align) & d$var_c %in% nms)
      ret[d$var_c] <- d$align
    
    if (!is.null(d[["group_border"]])) {
      if (d[["group_border"]] == TRUE) {
        has_cell_border <- TRUE
      }
    }
  }
  
  # Assign cell border to let it not be NA in pi
  if (has_cell_border) {
    ret[["..group_border"]] <- "left"
  }
  
  # Deal with stub
  if (!is.null(ts$stub)) {
    
    # Remove variables associated with stub
    ret <- ret[-which(names(ret) %in% ts$stub$vars)]  
    
    # Assign alignment from stub definition
    ret[["stub"]] <- ts$stub$align 
    
    # Assign stub indicator to let it not be NA
    ret[["..stub_var"]] <- ts$stub$align
    
  }
  
  return(ret)
  
}


#' Gets the labels
#' @noRd
get_labels <- function(dat, ts){
  
  defs <- ts$col_defs
  nfmt <- ts$n_format
  
  # Get the column names from the dataframe
  #print("Names in label function")
  v1 <- names(dat)
  #print(v1)
  
  # Get the labels from the dataframe
  # Not so easy because not all columns have labels.
  # If a column has no labels, use the column name
  # as the header string.
  v2 <- c()
  counter <- 1
  
  for (col in dat) {
    if (!is.null(attr(col, "label")) & "label" %in% ts$use_attributes) {
      v2 <- c(v2, attr(col, "label"))
    } else {
      if (is.null( names(col))) {
        v2 <- c(v2, v1[counter])
      } else {
        
        # Not actually sure what the situation is when this would be viable.
        # Consider taking this whole else condition out.
        # Only time I've seen names directly on the column, 
        # something is messsed up.
        if (length(names(col)) == 1)
          v2 <- c(v2, names(col))
        else
          v2 <- c(v2, v1[counter])
      }
      
    }
    counter <- counter + 1
  }
  
  # Convert label vector to a list
  #ls <- as.list(v2)
  ls <- v2


  # Assign names to list
  names(ls) <- v1
  #print(ls)
  
  # Let any defined labels override any attribute labels
  has_cell_border <- FALSE
  for (def in defs) {

    if (!is.null(def[["label"]]))
      ls[[def$var]] <- def[["label"]]
    
    if (!is.null(def$n) ) {
      ls[[def$var]] <- paste0(ls[[def$var]],  nfmt(def$n))
    }
    
    if (!is.null(def[["group_border"]])) {
      if (def[["group_border"]] == TRUE) {
        has_cell_border <- TRUE
      }
    }
  }
  
  if (has_cell_border) {
    ls <- c(ls, ..group_border = "..group_border")
  }
  
  # Deal with stub
  if (!is.null(ts$stub)) {
    
    # Remove variables associated with stub
    ls <- ls[-which(names(ls) %in% ts$stub$vars)]  
    
    # Assign label from stub definition
    ls <- c(stub = ts$stub$label, ls)
    
    # Assign label for stub indicator
    ls <- c(ls, ..stub_var = "..stub_var")
    
  }
  
  
  return(ls)
}


#' @title Get the columns for the table
#'
#' @description Get the columns for the table
#'
#' @param x The Table spec object
#' @noRd
get_table_cols <- function(x, control_cols) {
  
  dat <- x$data
  
  ret <- c()
  show_all <- FALSE
  if (length(x$show_cols) == 1 && x$show_cols == "all") {
    ret <- names(dat)
    show_all <- TRUE
  }
  else if (length(x$show_cols) == 1 && x$show_cols == "none") {
    show_all <- FALSE
  }
  else if (all(x$show_cols %in% names(dat)))
    ret <- x$show_cols
  else {
    
    diff <- setdiff(x$show_cols, names(dat))
    if (length(diff) > 0)
      stop("Columns not found in data: ", paste(diff, collapse = ","))
    
  }
  
  # Deal with visible options
  if (!is.null(x$col_defs)) {
    for (def in x$col_defs) {
      
      if (show_all == FALSE & def$visible)
        ret[length(ret) + 1] <- def$var_c
      #else if (show_all == TRUE & def$visible == FALSE)
      else if (def$visible == FALSE) {
        if (def$var_c %in% ret) {
          ret <- ret[!ret %in% def$var_c]
        }
      }
    }
    
    ret <- unique(ret)
    
    diff <- setdiff(ret, names(dat))
    if (length(diff) > 0)
      stop("Columns not found in data: ", paste(diff, collapse = ","))
  }
  
  ret <- unique(c(ret, control_cols))
  
  return(ret)
  
}



# Text Sizing Functions ---------------------------------------------------



#' Split the data horizontally by available page size
#' @param x The data frame to split
#' @param widths The column widths
#' @param page_size The size of the available space in rows
#' @noRd
get_splits_text <- function(x, widths, page_size, lpg_rows, 
                            content_offsets, ts, count_row_var = FALSE) {
  
  defs <- ts$col_defs
  

  # Calculate where page breaks should occur
  # Based on available height, content size, and offsets
  # Function adds a ..page variable with page indicator
  if (any(grepl("..group_cohesion", names(x)))) {
    group_cohesion <- names(x)[grepl("..group_cohesion", names(x))]
    group_cohesion <- group_cohesion[order(group_cohesion)]
  } else {
    group_cohesion <- NULL
  }
  if (any(grepl("..min_page_prop",names(x)))) {
    min_page_prop_col <- names(x)[grepl("..min_page_prop", names(x))]
    min_page_prop <- c()
    for (col in min_page_prop_col) {
      min_page_prop <- c(min_page_prop, unique(x[[col]][!is.na(x[[col]])]))
    }
    
  } else {
    min_page_prop <- NULL
  }
  
  pgs <- get_page_breaks(x, page_size, lpg_rows, content_offsets, count_row_var,
                         group_cohesion = group_cohesion, 
                         min_page_prop = min_page_prop)
  
  # Eliminate pages that have only blank lines
  sb <- subset(pgs, trimws(pgs$..blank) == "")
  non_blank_pages <- unique(sb$..page)
  sbst <- subset(pgs, pgs$..page %in% non_blank_pages)
  
  # If the first blank row is from blank after, then remove it
  first_page_pos <- get_first_pos(sbst$..page)
  rm_blank <- first_page_pos & sbst$..blank == "B"
  sbst <- sbst[!rm_blank,]
  
  # Add break labels if needed
  if (any(grepl("..break_label", names(sbst)))) {
    sbst$..temp_seq <- 1:nrow(sbst)
    break_pos <- which(!is.na(sbst$..break_occur))
    
    for (p in break_pos) {
      rep_col <- as.numeric(strsplit(sbst$..break_occur[p], ",")[[1]])
      rev_order <- rep_col[order(rep_col, decreasing = T)]
      for (r in rep_col) {
        temp_df <- sbst[p,]
        
        # If first row is blank before, add after it
        if (temp_df$..blank[1] == "A") {
          temp_df$..temp_seq[1] <- temp_df$..temp_seq[1] + 1
        }
        
        # Revise to label of row
        temp_df$..blank[1] <-  "L"
        
        # Create label of row
        temp_df[1, !grepl("^\\.\\.", names(temp_df))] <- ""
        temp_df[1,1] <- temp_df[[paste0("..break_label", r)]]
        temp_df$..temp_seq[1] <- temp_df$..temp_seq[1] - rev_order[r]*0.1
        
        # Set ..break_occur for current column for later use
        temp_df$..break_occur <- r
        
        # Reset the row name for PDF process
        row.names(temp_df) <- paste0("break_label", 1:nrow(temp_df))
        
        # Stack to original data, no need to repeat if blank
        if (temp_df[1,1] != "") {
          sbst <- rbind(sbst, temp_df)
        }
      }
    }
    # re-order to put repeated label in right position
    sbst <- sbst[order(sbst$..temp_seq),]
    
    # drop temporary variable
    sbst <- sbst[, setdiff(names(sbst), c("..temp_seq"))]
  }
  
  # Split the data frame at the page indicators
  ret <- split(sbst, sbst$..page)
  
  # Perform column deduping
  ret <- dedupe_pages(ret, defs)
  
  # Dedupe stub if child column dedupe was requested.
  # Somewhat unsure of this.
  if (stub_dedupe(ts$stub, defs)) {
    sdef <- define_c("stub", dedupe = TRUE)
    ret <- dedupe_pages(ret, list(sdef))
  }

  
  return(ret)
  
}


#' Function to calculate page breaks
#' @param x Data frame to page.
#' @param page_size Available data height in number of rows.
#' @param lpg_rows Last page rows.  Subtracted from available height.
#' @param content_offsets Blank rows requested above or below content
#' @return Data frame with ..page column populated with page numbers.
#' @noRd
get_page_breaks <- function(x, page_size, lpg_rows, content_offsets, 
                            count_row_var = FALSE,
                            group_cohesion = NULL,
                            min_page_prop = NULL){
  
  pg <- 1
  counter <- 0
  offset <- lpg_rows + content_offsets[["blank_upper"]]
  
  # If the upper line has name, it means using page_by
  using_pgby_up <- F
  using_pgby_low <- F
  if (is.null(names(content_offsets[["upper"]]))) {
    ttfl <- content_offsets["upper"] + content_offsets["lower"]
  } else {
    using_pgby_up <- T
    
    if (!is.null(names(content_offsets[["lower"]]))) {
      using_pgby_low <- T
    }
  }
  
  
  # User Paging variable
  currentPage <- NA
  lastPage <- NA
  userForce <- FALSE
  
  # print(paste("Content Upper:", content_offsets["upper"]))
  # print(paste("Content Lower:", content_offsets["lower"]))
  # print(paste("Page size:", page_size))
  # print(paste("Last Page Rows:", lpg_rows))
  # print(paste("Content offset:", offset))
  
  if (!is.null(group_cohesion)) {
    # Set min page percent as 0.8 if not assigned
    if (is.null(min_page_prop)) {
      min_page_prop <- rep(0.8, length(group_cohesion))
    }

    group_change_df <- as.data.frame(matrix(NA, dim(x)[1], length(group_cohesion)))
    group_counts_df <- as.data.frame(matrix(NA, dim(x)[1], length(group_cohesion)))
    
    for (i in 1:length(group_cohesion)) {
      col <- group_cohesion[i]
      
      if (count_row_var) {
        group_counts <- get_group_count(x[[col]], x$..row)
      } else {
        group_counts <- get_group_count(x[[col]])
      }
      group_counts_df[[i]] <- group_counts
      
      group_change <- as.logical(c(TRUE, mapply(function(x, y) !identical(x, y), 
                                                x[[col]][-1], 
                                                x[[col]][-length(x[[col]])])))
      
      group_change_df[[i]] <- group_change
    }
  }
  
  # Transfer break labels into sequence to detect the change
  break_label_df <- NULL
  if (any(grepl("..break_label", names(x)))) {
    break_label_df <- x[,names(x)[grepl("..break_label", names(x))]]
    for (nm in names(break_label_df)) {
      if (grepl("..break_label\\d+", nm)){
        break_label_df[[nm]] <- get_group_seq(break_label_df[[nm]])
        break_label_num <- sub(".*?([0-9]+).*", "\\1", nm)
        break_label_df[[paste0("..break_group_first", break_label_num)]] <- get_first_pos(break_label_df[[nm]])
      }
    }
    break_label_col <- names(break_label_df)[grepl("..break_label\\d+", names(break_label_df))]
    break_label_col <- break_label_col[order(break_label_col)]
    break_label_lines_col <- names(break_label_df)[grepl("..break_label_lines\\d+", names(break_label_df))]
    break_label_lines_col <- break_label_lines_col[order(break_label_lines_col)]
    
    break_label_retain <- break_label_df[1, break_label_col]
    
    pre_pg <- pg
  }
  
  for (i in seq_len(nrow(x))){
    
    if (using_pgby_up & !using_pgby_low) {
      ttfl <- content_offsets[["lower"]] + content_offsets[["upper"]][[x$..page_by[i]]]
    } else if (using_pgby_up & using_pgby_low) {
      ttfl <- content_offsets[["lower"]][[x$..page_by[i]]] + content_offsets[["upper"]][[x$..page_by[i]]]
    }
    
    if (count_row_var) {
      if (is.na(x$..row[i])) {
        counter <- counter + 1
        cur_count <- 1
      } else {
        counter <- counter + x$..row[i]
        cur_count <- x$..row[i]
      }
    } else {
      counter <- counter + 1
      cur_count <- 1
    }
    
    if (i == nrow(x)) {
      # Exception where last line is equal to number of available lines
      # Don't put the blank row in this case.
      if (counter < (page_size - ttfl)) {  
        offset <- offset + content_offsets[["blank_lower"]]
        #print(paste("Lower Blank:", offset))
      }
    }
    # print(paste("Counter:", counter))
    # print(paste("Condition:", (page_size - offset - ttfl)))
    

    # Get current page value
    currentPage <- x$..page[i]
    
    # If last page is not equal to current page, and neither is NA,
    # force a page break
    if (!is.na(lastPage) & !is.na(currentPage) & 
        trimws(lastPage) != "NA" & trimws(currentPage) != "NA" & 
        trimws(lastPage) != "" & trimws(currentPage) != "" & 
        currentPage != lastPage) {
      userForce <- TRUE

    } else
      userForce <- FALSE
    
    # After comparison, set last page value
    lastPage <- currentPage
    
    # Force a break if the next row has wraps that exceed the page size
    # if (count_row_var & i < nrow(x)) {
    #   nextRowCount <- x$..row[i + 1]
    #   if (!is.na(nextRowCount)) {
    #     if (nextRowCount > 1) {
    #       if (page_size  - offset - ttfl - nextRowCount - counter < -1) {
    #         userForce <- TRUE
    #         # print(pg)
    #         # print(page_size  - offset - ttfl - nextRowCount - counter)
    #       }
    #     }
    #   }
    # }
    
    if (!is.null(group_cohesion)) {
      j <- 1
      while (j <= length(group_cohesion) & userForce == FALSE) {
        group_change <- group_change_df[[j]]
        group_counts <- group_counts_df[[j]]
        min_page_prop_j <- min_page_prop[j]
        
        # Rules of moving a group into next page
        # - Remaining space is not enough for a whole group
        # - After moving, the existed lines meet minimum requirement
        # - A whole group is not bigger than the capacity for a page
        if (group_change[i]) {
          pre_counter <- counter - cur_count
          if ((group_counts[i] > (page_size  - offset - ttfl - pre_counter)) &
              (pre_counter >= round((page_size  - offset - ttfl) * min_page_prop_j)) &
              (group_counts[i] <= (page_size - offset - ttfl))) {
            userForce <- TRUE
          }
        }
        
        j <- j + 1
      }
    }
    
    # print("--------------------------------------")
    # print(paste0("userForce is ", userForce))
    # print(paste0("row: ", i))
    # print(paste0("counter: ", counter))
    # print(paste0("ttfl: ", ttfl))
    # print(paste0("page_size: ", page_size))
    # print(paste0("Capacity: ", page_size  - offset - ttfl))
    # print(paste0("Page before: ", pg))

    # If line count is greater than page size, start a new page
    if ((counter > (page_size  - offset - ttfl)) | userForce) {
      
      # Don't understand why this adjustment is needed.  But it is.
      # Have to figure it out, because it seems wrong.
      # if (pg == 1)
      #   page_size <- page_size - 1
      
      # Increment page count
      pg <- pg + 1
      
      # Reset line counter
      if (count_row_var) {
        if (is.na(x$..row[i])) {
          counter <- 1
        } else {
          counter <- x$..row[i]
        }
      } else {
        counter <- 1
      }
      
      # If last record of last page is blank before, move it to new page
      if ("..blank" %in% names(x) & i > 1) {
        if (x$..blank[i-1] == "A") {
          x$..page[i-1] <- pg
          counter <- counter + 1
        }
      }
      
      # Break label process
      if (!is.null(break_label_df) & i > 1) {
        break_label_cur <- break_label_df[i, break_label_col]
        break_label_pre <- break_label_df[i-1, break_label_col]

        # In case the break label are all NA, need to detect first
        if (any(!is.na(break_label_pre == break_label_cur))) {
          # Check if same break label is spanned across pages
          if (any(break_label_pre == break_label_cur)){
            # If the break label is blank value, no need to repeat it
            break_idx <- (break_label_pre == break_label_cur) & break_label_cur != ""
            
            if ("..blank" %in% names(x)) {
              if (x$..blank[i-1] == "A") {
                break_group_first_col <- names(break_label_df)[grepl("..break_group_first\\d+", names(break_label_df))]
                break_label_first <- break_label_df[i - 1, break_group_first_col]
                break_idx <- (break_label_pre == break_label_cur) & (!break_label_first)
              }
            }
            
            if (any(break_idx)) {
              counter <- counter + sum(x[i, break_label_lines_col[break_idx]])
              x$..break_occur[i] <- paste0(which(break_idx), collapse = ", ")
            }
          }
        }
      }
      
      # If the first row is blank line from blank_after, set as 0 because it
      # will be removed later
      if ("..blank" %in% names(x)){
        if (x$..blank[i] == "B") {
          counter <- 0
        }
      }

      # Set offset to zero on first page
      offset <- 0

    }
    
    # print(paste0("Page after: ", pg))
    # print("--------------------------------------")
  
    x$..page[i] <- pg
    # print(pg)
  }

  
  # Convert ..page back to number if necessary
  if (all(class(x$..page) != "numeric"))
    x$..page <- as.numeric(x$..page)
  
  #print(paste("Page count:", counter))
  
  return(x)
}



stub_dedupe <- function(stb, defs) {
  
  ret <- FALSE
  
  for (df in defs) {
    
    if (df$var_c %in% stb$vars) {
      if (df$dedupe == TRUE & df$label_row == FALSE) {
        ret <- TRUE
      }
    }
  }
  
  return(ret)
}


#' @import stringi
#' @noRd
get_pgby_cnt <- function(pgby) {
  
  ret <- 0
  
  if (!is.null(pgby)) {
    
    upgby <- unique(pgby)
  
    for (pgb in upgby) {
    
      lns <- unlist(stri_split_fixed(pgb, "\n"))
      
      if (length(lns) > ret)
        ret <- length(lns)
    
    }
  
  }
  
  return(ret)
  
}

#' @noRd
get_pgby_value <- function(value, pgby_cnt) {
  
  ret <- ""
  if (is.null(value)) {
    
    if (!is.null(pgby_cnt)) {
      
      if (pgby_cnt > 0) {
        ret <- rep("x", pgby_cnt)
        ret <- paste0(ret, collapse = "\n")
      } 
    }

  }
  
  return(ret)
}

context("Sizing Functions Tests")

options("logr.output" = FALSE)

test_that("get_table_cols() works as expected.", {
  
  control_cols <- c("..blank", "..page", "..row", "..page_by")
  
  tbl <- create_table(mtcars[1:10, ], show_cols = "none") %>% 
    define(mpg, format = "%.1f") %>% 
    define(cyl, width = 1) %>% 
    define(hp)
  
  lst1 <- get_table_cols(tbl, control_cols)
  
  expect_equal(lst1, c("mpg", "cyl", "hp", control_cols))
  
  tbl2 <- create_table(mtcars[1:10, ], show_cols = "all") %>% 
    define(mpg, format = "%.1f") %>% 
    define(cyl, width = 1) %>% 
    define(hp)
  
  lst2 <- get_table_cols(tbl2, control_cols)
  
  expect_equal(length(lst2), 11 + length(control_cols))
  
  tbl3 <- create_table(mtcars[1:10, ], show_cols = c("mpg", "disp", "wt")) %>% 
    define(mpg, format = "%.1f") %>% 
    define(cyl, width = 1) %>% 
    define(hp)
  
  lst3 <- get_table_cols(tbl3, control_cols)
  
  expect_equal(lst3, c("mpg", "disp", "wt", "cyl", "hp", control_cols))
  
  
})



test_that("get_page_breaks works as expected", {
  
  off <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
  dat <- iris
  dat$..page <- NA
  mdm <- get_page_breaks(dat, 50, 0, off)
  
  expect_equal( unique(mdm$..page), c(1, 2, 3))
  
  
})

test_that("get_page_breaks with count_row_var works as expected", {
  
  off <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
  dat <- iris[1:10,]
  dat$..page <- NA
  dat$..row <- c(rep(1,9), 3)
  mdm <- get_page_breaks(dat, 10, 0, off, count_row_var = TRUE)
  
  expect_equal(mdm$..page, c(rep(1,9), 2))
})

test_that("get_page_breaks with group_cohesion works as expected", {
  
  off <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
  dat <- iris[1:20,]
  
  dat$..group_cohesion <- c(
    rep("A", 4),
    rep("B", 2),
    rep("C", 3),
    rep("D", 4),
    rep("E", 7)
  )
  
  dat$..row <- c(
    rep(2, 4),
    rep(2, 2),
    c(3, 3, 1),
    c(rep(2, 4)),
    c(1, rep(2, 6))
  )
  
  dat$..page <- NA
  
  mdm <- get_page_breaks(dat, 10, 0, off, count_row_var = TRUE,
                         group_cohesion = "..group_cohesion")
  
  # test results variable for review
  mdm$test_group_cnt <- get_group_count(mdm$..group_cohesion, mdm$..row)
  mdm$cum_lines <- unlist(tapply(mdm$..row, mdm$..page, cumsum))

  expect_equal(mdm$..page, c(1,1,1,1,2,2,2,2,3,3,3,3,3,3,4,4,4,4,4,5))
})

test_that("get_page_breaks with group_cohesion and min_page_prop works as expected", {
  
  off <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
  dat <- iris[1:20,]
  
  dat$..group_cohesion <- c(
    rep("A", 4),
    rep("B", 2),
    rep("C", 3),
    rep("D", 4),
    rep("E", 7)
  )
  
  dat$..row <- c(
    rep(2, 4),
    rep(2, 2),
    c(3, 3, 1),
    c(rep(2, 4)),
    c(1, rep(2, 6))
  )
  
  dat$..page <- NA
  
  mdm <- get_page_breaks(dat, 10, 0, off, count_row_var = TRUE,
                         group_cohesion = "..group_cohesion",
                         min_page_prop = 0.4)
  
  # test results variable for review
  mdm$test_group_cnt <- get_group_count(mdm$..group_cohesion, mdm$..row)
  mdm$cum_lines <- unlist(tapply(mdm$..row, mdm$..page, cumsum))
  
  expect_equal(mdm$..page, c(1,1,1,1,2,2,3,3,3,4,4,4,4,4,5,5,5,5,5,6))
})


test_that("get_splits_text works as expected", {
  
  
  off <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
  
  dat <- iris
  dat$..page <- NA
  dat$..blank <- ""
  w <- c(Sepal.Length = 1.5, Sepal.Width = 1.5, Petal.Length = 1.5,
         Petal.Width = 1.5, Species = 1.25)
  
  mdm <- get_splits_text(dat, w, 50, 0, off, list())
  
  expect_equal(length(mdm), 3)
  expect_equal(nrow(mdm[[1]]), 50)
  expect_equal(nrow(mdm[[2]]), 50)
  expect_equal(nrow(mdm[[3]]), 50)
  
  
  mdm2 <- get_splits_text(dat, w, 60, 0, off, list())
  expect_equal(length(mdm2), 3)
  expect_equal(nrow(mdm2[[1]]), 60)
  expect_equal(nrow(mdm2[[2]]), 60)
  expect_equal(nrow(mdm2[[3]]), 30)
  
  mdm <- get_splits_text(dat, w, 50, 10, off, list())
  
  expect_equal(length(mdm), 4)
  expect_equal(nrow(mdm[[1]]), 40)
  expect_equal(nrow(mdm[[2]]), 50)
  expect_equal(nrow(mdm[[3]]), 50)
  expect_equal(nrow(mdm[[4]]), 10)
    
})



test_that("prep_data works as expected", {
  
   mw <- max(nchar(as.character(iris$Species)))
  
  
  tbl <- create_table(iris) %>% 
    define(Species, blank_after = TRUE, blank_before = TRUE,
           dedupe = TRUE, indent = .25)

  
  d <- prep_data(iris, tbl, .0833333, "")

  
  expect_equal(sum(d$..blank != ""), 6)
  # Indendation is now processed after prep_data
  expect_equal(max(nchar(d$Species)) == mw, TRUE)
  
})


test_that("get_labels works as expected", {
  
  dat <- mtcars
  dat$name <- rownames(mtcars)
  
  fmt <- value(condition(x >= 20, "High"),
               condition(TRUE, "Low"))
  
  dat$mpg_cat <- fapply(dat$mpg, fmt)
  
  dat
  
  tbl <- create_table(dat) %>% 
    stub(c("mpg_cat", "name"), label = "Stub") %>% 
    define(cyl, label = "Cylinders") %>% 
    define(disp, label = "Displacement")
  
  lbls <- get_labels(dat, tbl)
  
  expect_equal(lbls[["cyl"]], "Cylinders")
  expect_equal(lbls[["hp"]], "hp")
  expect_equal(lbls[["disp"]], "Displacement")
  expect_equal(lbls[["stub"]], "Stub")
  
})

test_that("get_aligns works as expected", {
  
  dat <- mtcars
  dat$name <- rownames(mtcars)
  
  fmt <- value(condition(x >= 20, "High"),
               condition(TRUE, "Low"))
  
  dat$mpg_cat <- fapply(dat$mpg, fmt)
  
  dat
  
  tbl <- create_table(dat) %>% 
    stub(c("mpg_cat", "name"), label = "Stub") %>% 
    define(cyl, label = "Cylinders", align = "right") %>% 
    define(disp, label = "Displacement", align = "center")
  
  algn <- get_aligns(dat, tbl)
  
  expect_equal(algn[["cyl"]], "right")
  expect_equal(algn[["stub"]], "left")
  expect_equal(algn[["disp"]], "center")

  
})

test_that("get_label_aligns works as expected", {
  
  dat <- mtcars
  dat$name <- rownames(mtcars)
  
  fmt <- value(condition(x >= 20, "High"),
               condition(TRUE, "Low"))
  
  dat$mpg_cat <- fapply(dat$mpg, fmt)
  
  dat
  
  tbl <- create_table(dat) %>% 
    stub(c("mpg_cat", "name"), label = "Stub", label_align = "center") %>% 
    define(cyl, label = "Cylinders", label_align = "left") %>% 
    define(disp, label = "Displacement", label_align = "center")
  
  ls <- rep("right", 13) 
  names(ls) <- names(dat)
  
  lbls <- get_label_aligns(tbl, ls)
  
  expect_equal(lbls[["cyl"]], "left")
  expect_equal(lbls[["hp"]], "right")
  expect_equal(lbls[["disp"]], "center")
  expect_equal(lbls[["stub"]], "center")
  
})

test_that("get_col_formats works as expected", {
  
  dat <- mtcars
  dat$name <- rownames(mtcars)
  
  fmt <- value(condition(x >= 20, "High"),
               condition(TRUE, "Low"))
  
  dat$mpg_cat <- fapply(dat$mpg, fmt)
  
  dat
  
  tbl <- create_table(dat) %>% 
    define(cyl, label = "Cylinders", format = "%.1f") %>% 
    define(disp, label = "Displacement", format = "%.2f")
  
  lbls <- get_col_formats(dat, tbl)
  
  
  expect_equal(lbls[["cyl"]], "%.1f")
  expect_equal(is.null(lbls[["hp"]]), TRUE)
  expect_equal(lbls[["disp"]], "%.2f")

  
})

test_that("get_page_wraps works as expected.", {
  
  
  tbl <- create_table(mtcars) %>% 
    define(mpg, width = 1) %>% 
    define(cyl, width = 1) %>% 
    define(disp, width = 1) %>% 
    define(hp, width = 1) %>% 
    define(drat, width = 1) %>% 
    define(wt, width = 1)
  
  wdths <- rep(1, 11)
  names(wdths) <- names(mtcars)

  res <- get_page_wraps(4, tbl, wdths, .2, c())
    
  expect_equal(length(res), 4)
  expect_equal(all(c("mpg", "cyl", "disp") %in% res[[1]]), TRUE)
  expect_equal(all(c("mpg", "cyl", "disp") %in% res[[2]]), FALSE)

  
})


test_that("get_col_widths_rtf works as expected.", {
  
  # dat, ts, labels, font, 
  # font_size, uom, gutter_width
  
  df <- mtcars
  
  tbl <- create_table(df) %>% 
    define(mpg, label = "Miles Per Gallon") %>% 
    define(cyl, width = 1.5) %>% 
    define(disp, width = 2) 

  lbls <- get_labels(df, tbl)
    
  res <- get_col_widths_variable(df, tbl, lbls, "Arial", 12, "inches", .2)
  res
  
  expect_equal(res[["mpg"]] < 1, TRUE)
  expect_equal(res[["cyl"]], 1.5)
  expect_equal(res[["disp"]], 2)
  expect_equal(res[["hp"]] < .5, TRUE)
  
  
})

test_that("get_col_widths_rtf works with indentation as expected.", {
  
  # dat, ts, labels, font, 
  # font_size, uom, gutter_width
  
  df <- mtcars
  
  tbl <- create_table(df) %>% 
    define(mpg, label = "Miles Per Gallon", indent = 1) %>% 
    define(cyl, width = 1.5) %>% 
    define(disp, width = 2) 
  
  lbls <- get_labels(df, tbl)
  
  res <- get_col_widths_variable(df, tbl, lbls, "Arial", 12, "inches", .2)
  res
  
  expect_equal(res[["mpg"]] > 1, TRUE)
  expect_equal(res[["cyl"]], 1.5)
  expect_equal(res[["disp"]], 2)
  expect_equal(res[["hp"]] < .5, TRUE)
  
  
})

test_that("get_col_widths_rtf works with stub as expected.", {
  
  # dat, ts, labels, font, 
  # font_size, uom, gutter_width
  
  df <- mtcars[, c("mpg", "cyl", "disp")]
  
  tbl <- create_table(df) %>% 
    stub(c("mpg", "cyl")) %>%
    define(mpg, label = "", label_row = TRUE) %>% 
    define(cyl, label = "") %>% 
    define(disp, width = 2) 
  
  lbls <- get_labels(df, tbl)
  
  fdat <- prep_data(df, tbl, 0.11, "")
  
  res <- get_col_widths_variable(fdat, tbl, lbls, "Arial", 12, "inches", .2)
  res
  
  expect_equal(res[["stub"]] < 1, TRUE)
  expect_equal(res[["disp"]], 2)
  
})

test_that("get_col_widths_rtf works with stub and indentation as expected.", {
  
  # dat, ts, labels, font, 
  # font_size, uom, gutter_width
  
  df <- mtcars[, c("mpg", "cyl", "disp")]
  
  tbl <- create_table(df) %>% 
    stub(c("mpg", "cyl")) %>%
    define(mpg, label = "", label_row = TRUE) %>% 
    define(cyl, label = "", indent = 1) %>% 
    define(disp, width = 2) 
  
  lbls <- get_labels(df, tbl)
  
  fdat <- prep_data(df, tbl, 0.11, "")
  
  res <- get_col_widths_variable(fdat, tbl, lbls, "Arial", 12, "inches", .2)
  res
  
  expect_equal(res[["stub"]] > 1, TRUE)
  expect_equal(res[["disp"]], 2)
  
})


test_that("stub_dedupe works as expected", {
  
  df <- mtcars
  
  tbl <- create_table(df) %>% 
    stub(c("cyl", "mpg"),  width = 2, label = "Fork") %>%
    define(cyl,) %>% 
    define(mpg, dedupe = TRUE) %>% 
    define(disp, width = 2) 
  
  res <- stub_dedupe(tbl$stub, tbl$col_defs)
  
  expect_equal(res, TRUE)
  
  tbl <- create_table(df) %>% 
    stub(c("cyl", "mpg"),  width = 2, label = "Fork") %>%
    define(cyl,) %>% 
    define(mpg) %>% 
    define(disp, width = 2) 
  
  res <- stub_dedupe(tbl$stub, tbl$col_defs)
  
  expect_equal(res, FALSE)
  
})


test_that("get_pgby_cnt works as expected", {
  
  res1 <- get_pgby_cnt(NULL)
  
  expect_equal(res1, 0)
  
  res2 <- get_pgby_cnt(c("myval"))
  
  expect_equal(res2, 1)
  
  res3 <- get_pgby_cnt(c("myval", "my\nval"))
  
  expect_equal(res3, 2)
  
  res4 <- get_pgby_cnt(c("myval0", "myval1", "myval2"))
  
  expect_equal(res4, 1)
  
  
})

# Right now only png supported
#' @noRd
get_image_rtf <- function(file_path, width, height, units) {
  
  if (!units %in% c("inches", "cm")) {
    stop("Invalid units.") 
  }
  
  # Get conversion factor to twips
  if (units == "inches") {
    conv <- 1440
  } else {
    conv <- 566.9291
  }
  
  blp <- "\\pngblip"
  if (grepl(".emf", file_path, fixed = TRUE))
    blp <- "\\emfblip"
  
  # Construct rtf codes
  ret <- paste0("{\\pict", blp, "\\picwgoal",round(width*conv),"\\pichgoal",round(height*conv)," \n")
  
  # Convert image to byte codes
  ret <- paste0(ret, get_image_bytes(file_path), "\n}")

  
  return(ret)
}

#' @noRd
get_image_bytes <- function(file_path) {
  
  
  if (!file.exists(file_path))
    stop(paste("Image file not found:", file_path))

  max.bytes<-50000000

  # Read in bytes
  dat<-readBin(file_path, what="raw", size=1, 
               signed=TRUE, endian="little", n=max.bytes);

  # Split single vector into list of 80 character vectors
  sdat <- split(dat, ceiling(seq_along(dat)/40))

  # Collapse each 80 character vector into a string
  lns <- sapply(sdat, paste, collapse = "")

  # Append a line return to each line and collapse into single string
  ret <- paste0(lns, "\n", collapse = "")

  
  return(ret)

}

#' @import stringi      
#' @noRd
get_image_html <- function(file_path, report_path, plt, units) {
  

  dr <- file.path(dirname(report_path), "images")
  if (!file.exists(dr)) {
    dir.create(dr) 
  }
  
  fl <-  paste0( gsub(".html", "", basename(report_path), fixed = TRUE), "-", 
                 stri_rand_strings(1, 4, pattern = "[A-Z0-9]"), ".jpg")
    
  pth <- file.path(dr, fl)

  
  res <- file.copy(file_path, pth)
  if (all(res == TRUE)) {
    file.remove(file_path) 
  }
  
  u <- units
  if (u == "inches")
    u <- "in"
  
  ph <- round(plt$height * .99, 3)
  pw <- round(plt$width * .99, 3)
  
  ret <- paste0("<img src=\"./images/", basename(pth), "\"", 
                " style=\"height:", ph, u, ";",
                " width:", pw, u, ";\">")
  
  
  return(ret)
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  library(reporter)
#  
#  # Create temporary path
#  tmp <- file.path(tempdir(), "example3.rtf")
#  
#  # Read in prepared data
#  df <- read.table(header = TRUE, text = '
#        var     label        A             B
#        "ampg"   "N"          "19"          "13"
#        "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)"
#        "ampg"   "Median"     "16.4"        "21.4"
#        "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
#        "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
#        "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)"
#        "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)"
#        "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
#  
#  # Create table
#  tbl <- create_table(df, first_row_blank = TRUE) %>%
#    stub(c("var", "label")) %>%
#    define(var, blank_after = TRUE, label_row = TRUE,
#           format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>%
#    define(label, indent = .25) %>%
#    define(A, label = "Group A", align = "center", n = 19) %>%
#    define(B, label = "Group B", align = "center", n = 13)
#  
#  
#  # Create report and add content
#  rpt <- create_report(tmp, output_type = "RTF",
#                       font = "Times", font_size = 12) %>%
#    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>%
#    titles("Table 1.0", "MTCARS Summary Table") %>%
#    add_content(tbl) %>%
#    footnotes("* Motor Trend, 1974") %>%
#    page_footer(left = Sys.time(),
#                center = "Confidential",
#                right = "Page [pg] of [tpg]")
#  
#  # Write out report
#  write_report(rpt)
#  
#  # file.show(tmp)
#  

# The page template is everything except the content: page header/footer,
# titles, footnotes, etc.


# Page Template Functions ----------------------------------------------

#' Create a page template with header, titles, footnotes, and footer
#' @param rs The report spec
#' @return The page template object
#' @noRd
page_template_text <- function(rs) {
  
  pt <- structure(list(), class = c("page_template_text", "list"))
  
  pt$page_header <- get_page_header(rs)
  pt$title_hdr <- get_title_header(rs$title_hdr, rs$line_size, rs$line_size, 
                                   rs$uchar, rs$char_width)
  pt$titles <- get_titles(rs$titles, rs$line_size, rs$line_size, 
                          rs$uchar, rs$char_width)
  pt$footnotes <- c()
  if (!is.null(rs$footnotes)) {
    if (!is.null(rs$footnotes[[1]])) {
      if (rs$footnotes[[1]]$valign == "bottom")
        pt$footnotes <- get_footnotes(rs$footnotes, rs$line_size, rs$line_size, 
                                      rs$uchar, rs$char_width)
    }
  
  }
  pt$page_footer <- get_page_footer(rs)
  # Page by not here.  Messes up line counts.
  
  return(pt)
}

#' Get page header text strings suitable for printing
#' @param rs The report spec
#' @return A vector of strings
#' @noRd
get_page_header <- function(rs) {
  
  if (is.null(rs$line_size)) {
    stop("line_size cannot be null.") 
    
  }
  
  phdrr <- rs$page_header_right
  phdrl <- rs$page_header_left
  phdr <- rs$page_header_left
  if(length(phdrl) < length(phdrr))
    phdr <- phdrr
  
  ret <- c()
  
  if (!is.null(phdr)) {
    for (i in seq_along(phdr)) {
      
      hl <- ""
      hr <- ""
      
      if (length(phdrl) >= i)
        hl <- phdrl[[i]]
      
      if (length(phdrr) >= i)
        hr <- phdrr[[i]]
      
      gp <- rs$line_size - (nchar(hl) + nchar(hr))
      
      #print("header")
      if (gp >= 0) {
        
        lw <- rs$line_size - nchar(hr)
        ln <- paste0(pad_right(hl, lw), hr) 
      }
      
      else {
        stop(paste0("Page header exceeds available width\n", 
                    "Header Left: ", hl, "\n", 
                    "Header Right: ", hr, "\n",
                    "Header length: ", nchar(hl) + nchar(hr), "\n",
                    "Line length: ", rs$line_size, "\n"))
        ln <- ""
      }
      
      ret[i] <- ln
    }
    
    if (length(ret) > 0 & rs$page_header_blank_row == "below") 
      ret[[length(ret) + 1]] <- "" 
    
  }
  
  return(ret)
}

#' Get title text strings suitable for printing
#' @import stringi
#' @param titles A list of title objects
#' @param width The width to set the title strings to
#' @return A vector of strings
#' @noRd
get_titles <- function(titles, content_width, page_width, uchar, char_width) {
  
  if (is.null(content_width)) {
    stop("width cannot be null.") 
    
  }
  
  ret <- c()
  

  
  if (!is.null(titles)) { 
    
    for (ttl in titles) {
      
      if (!any(class(ttl) == "title_spec")) {
        stop("titles parameter value is not a title spec.")
      }
      
      cols <- ttl$columns
      
      if (ttl$width == "page")
        width <- page_width
      else if (ttl$width == "content")
        width <- content_width
      else if (is.numeric(ttl$width))
        width <- ceiling(ttl$width / char_width)
      
      cll <- round(width / cols)
      
      ll <- width 
      
      
      algn <- ttl$align
      if (algn == "centre")
        algn <- "center"
          
      if (any(ttl$blank_row %in% c("above", "both")) & length(ttl$titles) > 0)
        ret[length(ret) + 1] <- ""
      
      if (any(ttl$borders %in% c("top", "all")) & length(ttl$titles) > 0)
        ret[length(ret) + 1] <-  paste0(rep(uchar, ll), 
                                               collapse = "")
      i <- 1
      while (i <= length(ttl$titles)) {
        
        ln <- ""
        
        for (j in seq_len(cols)) {
      
          # Not all cells have titles
          if (i > length(ttl$titles))
            t <- ""
          else 
            t <- ttl$titles[[i]]
          
          # Deal with column alignments
          if (cols == 1) {
            calgn <- algn 
          } else if (cols == 2) {
            if (j == 1)
              calgn <- "left"
            else 
              calgn <- "right"
          } else if (cols == 3) {
            if (j == 1)
              calgn <- "left"
            else if (j == 2)
              calgn <- "center"
            else if (j == 3) 
              calgn <- "right"
          }
          
          gp <- cll - nchar(t)
          
          #print("titles")
          if (gp > 0) {
            
            if (calgn == "left")
              ln <- paste0(ln, pad_right(t, cll))
            else if (calgn == "right")
              ln <- paste0(ln, pad_left(t, cll))
            else if (calgn == "center")
              ln <- paste0(ln, pad_both(t, cll))
            
          } else {
            warning(paste0("Title exceeds available width.",
                        "\nTitle: ", t,
                        "\nTitle width: ", nchar(t),
                        "\nLine length: ", cll))
            
            tgp <- cll - 3
  
            if (tgp >= 0) {
              if (ttl$align == "left") {
                ln <- paste0(substr(pad_right(t, cll), 1, tgp), "...")
              } else if (ttl$align == "right") {
                ln <- paste0("...", substr(pad_left(t, cll), 1, tgp))
              } else if (ttl$align == "center" | ttl$align == "centre") {
                ln <- paste0(substr(pad_both(t, cll), 1, tgp), "...")
              }
              
            } else ln <- ""
            
          }
          
          i <- i + 1
        
        }
        
        ret[length(ret) + 1] <- ln
      }
      
      if (any(ttl$borders %in% c("bottom", "all")) & length(ttl$titles) > 0)
        ret[length(ret) + 1] <- paste0(rep(uchar, ll ),  collapse = "")
      
      if (any(ttl$blank_row %in% c("below", "both")) & length(ttl$titles) > 0)
        ret[length(ret) + 1] <- ""
    }
    
  }
  
  
  return(ret)
}



#' Get title text strings suitable for printing
#' @import stringi
#' @param titles A list of title objects
#' @param width The width to set the title strings to
#' @return A vector of strings
#' @noRd
get_titles_back <- function(titles, content_width, page_width, uchar, char_width) {
  
  if (is.null(content_width)) {
    stop("width cannot be null.") 
    
  }
  
  ret <- c()
  
  if (!is.null(titles)) { 
    
    for (ttl in titles) {
      
      if (!any(class(ttl) == "title_spec")) {
        stop("titles parameter value is not a title spec.")
      }
      
      if (ttl$width == "page")
        width <- page_width
      else if (ttl$width == "content")
        width <- content_width
      else if (is.numeric(ttl$width))
        width <- ceiling(ttl$width / char_width)
      
      ll <- width 
      
      if (ttl$blank_row %in% c("above", "both") & length(ttl$titles) > 0)
        ret[length(ret) + 1] <- ""
      
      if (any(ttl$borders %in% c("top", "all")) & length(ttl$titles) > 0)
        ret[length(ret) + 1] <-  paste0(rep(uchar, ll), 
                                        collapse = "")
      
      for (i in seq_along(ttl$titles)) {
        
        t <- ttl$titles[i]
        
        gp <- ll - nchar(t)
        
        #print("titles")
        if (gp > 0) {
          
          if (ttl$align == "left")
            ln <- pad_right(t, ll)
          else if (ttl$align == "right")
            ln <- pad_left(t, ll)
          else if (ttl$align == "center" | ttl$align == "centre")
            ln <- pad_both(t, ll)
          
        } else {
          warning(paste0("Title exceeds available width.",
                         "\nTitle: ", t,
                         "\nTitle width: ", nchar(t),
                         "\nLine length: ", ll))
          
          tgp <- ll - 3
          
          if (tgp >= 0) {
            if (ttl$align == "left") {
              ln <- paste0(substr(pad_right(t, ll), 1, tgp), "...")
            } else if (ttl$align == "right") {
              ln <- paste0("...", substr(pad_left(t, ll), 1, tgp))
            } else if (ttl$align == "center" | ttl$align == "centre") {
              ln <- paste0(substr(pad_both(t, ll), 1, tgp), "...")
            }
            
            
          } else ln <- ""
          
        }
        
        
        ret[length(ret) + 1] <- ln
      }
      
      if (any(ttl$borders %in% c("bottom", "all")) & length(ttl$titles) > 0)
        ret[length(ret) + 1] <- paste0(rep(uchar, ll ), 
                                       collapse = "")
      
      if (ttl$blank_row %in% c("below", "both") & length(ttl$titles) > 0)
        ret[length(ret) + 1] <- ""
    }
    
  }
  
  
  return(ret)
}

#' Get page by text strings suitable for printing
#' @import stringi
#' @param titles Page by object
#' @param width The width to set the page by strings to
#' @return A vector of strings
#' @import stringi
#' @noRd
get_page_by <- function(pgby, width, value, pgby_cnt = NULL) {
  
  if (is.null(width)) {
    stop("width cannot be null.") 
    
  }
  
  # Set up fake values if there is no value yet.
  # Needed for estimation of header line counts
  if (is.null(value)) {
    value <- get_pgby_value(value, pgby_cnt)
  }
  
  ll <- width
  ret <- c()
  
  if (!is.null(pgby)) { 
    
    if (!any(class(pgby) == "page_by"))
      stop("pgby parameter value is not a page_by.")
    
    if (pgby$blank_row %in% c("above", "both"))
      ret[length(ret) + 1] <- ""
  

    pb <- paste0(pgby$label, value)
    
    lns <- unlist(stri_split_fixed(pb, "\n"))
    
    for (pbln in lns) {
    
      gp <- ll - nchar(pbln) 
      
  
      if (gp > 0) {
        
        if (pgby$align == "left")
          ln <- pad_right(pbln, ll)
        else if (pgby$align == "right")
          ln <- pad_left(pbln, ll)
        else if (pgby$align == "center" | pgby$align == "centre")
          ln <- pad_both(pbln, ll)
        
      } else 
        stop("Page by exceeds available width.")
      
      
      ret[length(ret) + 1] <- ln
    
    }
    
    
    if (pgby$blank_row %in% c("below", "both"))
      ret[length(ret) + 1] <- ""
    
    
  }
  
  
  return(ret)
}

#' Get title header text strings suitable for printing
#' @import stringi
#' @param title_hdr A title_hdr object
#' @param width The width to set the title header to
#' @return A vector of strings
#' @noRd
get_title_header <- function(title_hdr, content_width, page_width, 
                             uchar = "-", char_width) {
  
  if (is.null(content_width)) {
    stop("width cannot be null.") 
    
  }
  
  ret <- c()
  
  if (!is.null(title_hdr)) { 
    
    for (ttl_hdr in title_hdr) {
      
      if (!any(class(ttl_hdr) == "title_hdr"))
        stop("title header parameter value is not a title header.")
      
      if (ttl_hdr$width == "page")
        width <- page_width
      else if (ttl_hdr$width == "content")
        width <- content_width
      else if (is.numeric(ttl_hdr$width))
        width <- ceiling(ttl_hdr$width / char_width)
      
      ll <- width 
      
      if (ttl_hdr$blank_row %in% c("above", "both") & length(ttl_hdr$titles) > 0)
        ret[length(ret) + 1] <- ""
      
      if (any(ttl_hdr$borders %in% c("top", "all")) & length(ttl_hdr$titles) > 0) {
        #ret[length(ret) + 1] <-  paste0(paste0(rep(uchar, ll), collapse = ""), " ")
        ret[length(ret) + 1] <-  paste0(rep(uchar, ll), collapse = "")
      }
      
      maxlen <- length(ttl_hdr$titles)
      if (length(ttl_hdr$right) > maxlen)
        maxlen <- length(ttl_hdr$right)
      
      hdr <- ttl_hdr$right
      
      for (i in seq_len(maxlen)) {
        
        if (i <= length(ttl_hdr$titles))
          t <- ttl_hdr$titles[i]
        else 
          t <- ""
        
        if (i <= length(hdr))
          h <- hdr[i]
        else 
          h <- ""
        
        gp <- ll - nchar(t) - nchar(h)
        
        #print("titles")
        if (gp >= 0) {
          
  
            #ln <- paste0(pad_right(t, ll - nchar(h)), h, " ")
            ln <- paste0(pad_right(t, ll - nchar(h)), h)
  
          
        } else {
          warning(paste0("Title header exceeds available width.\n",
                      "Title: ", t, "\n",
                      "Header: ", h, "\n",
                      "Title length: ", nchar(t), "\n",
                      "Header length: ", nchar(h), "\n",
                      "Line length: ", ll, "\n"))
          
          tgp <- ll - 3
          if (tgp >= 0) {
            
            ln <- paste0(substr(paste0(pad_right(t, ll - nchar(h)), h, " "), 
                                1, tgp), "...")
            
          } else ln <- ""
        }
        
        
        ret[length(ret) + 1] <- ln
      }
      
      if (any(ttl_hdr$borders %in% c("bottom", "all")) & 
          length(ttl_hdr$titles) > 0) {
        #ret[length(ret) + 1] <-  paste0(paste0(rep(uchar, ll), collapse = ""), " ")
        ret[length(ret) + 1] <-  paste0(rep(uchar, ll), collapse = "")
        
      }
      
      if (ttl_hdr$blank_row %in% c("below", "both") & 
          length(ttl_hdr$titles) > 0)
        ret[length(ret) + 1] <- ""
  
    
    }
  }
  
  
  return(ret)
}



#' Get footnote text strings suitable for printing
#' @param rs The report spec
#' @return A vector of strings
#' @noRd
get_footnotes <- function(footnotes, content_width, page_width, 
                          uchar = "-", char_width) {
  
  if (is.null(char_width)) {
    stop("width cannot be null.") 
    
  }
  
  ret <- c()
  
  if (!is.null(footnotes)) {
    for (ftn in footnotes) {
      
      if (!any(class(ftn) == "footnote_spec"))
        stop("footnotes parameter value is not a footnote spec.")
      
      cols <- ftn$columns
      
      if (ftn$width == "page")
        width <- page_width
      else if (ftn$width == "content")
        width <- content_width
      else if (is.numeric(ftn$width))
        width <- ceiling(ftn$width / char_width)
      
      cll <- round(width / cols)
      
      ll <- width 
      
      algn <- ftn$align
      if (algn == "centre")
        algn <- "center"
      
      if (ftn$blank_row %in% c("above", "both") & length(ftn$footnotes) > 0)
        ret[length(ret) + 1] <- ""
      
      if (any(ftn$borders %in% c("top", "all")) & length(ftn$footnotes) > 0) {
        # ret[length(ret) + 1] <- paste0(paste0(rep(uchar, ll), 
        #                                       collapse = ""), " ")
        
        ret[length(ret) + 1] <- paste0(rep(uchar, ll), collapse = "")
      }
      
      i <- 1
      while (i <= length(ftn$footnotes)) {
        
        ln <- ""
      
        for (j in seq_len(cols)) {
          
          # Not all cells have footnotes
          if (i > length(ftn$footnotes))
            f <- ""
          else 
            f <- ftn$footnotes[[i]]
          
          # Deal with column alignments
          if (cols == 1) {
            calgn <- algn 
          } else if (cols == 2) {
            if (j == 1)
              calgn <- "left"
            else 
              calgn <- "right"
          } else if (cols == 3) {
            if (j == 1)
              calgn <- "left"
            else if (j == 2)
              calgn <- "center"
            else if (j == 3) 
              calgn <- "right"
          }
          
          gp <- cll - nchar(f)
          
          #print("footnotes")
          if (gp > 0) {
            
            # if (ftn$align == "left")
            #   ln <- pad_right(paste0(f, " "), ll + 1)
            # else if (ftn$align == "right")
            #   ln <- pad_left(paste0(f, " "), ll + 1)
            # else if (ftn$align == "center" | ftn$align == "centre")
            #   ln <- pad_both(paste0(f, " "), ll + 1)
            
            # if (ftn$align == "left")
            #   ln <- pad_right(f, ll)
            # else if (ftn$align == "right")
            #   ln <- pad_left(f, ll)
            # else if (ftn$align == "center" | ftn$align == "centre")
            #   ln <- pad_both(f, ll)
            
            if (calgn == "left")
              ln <- paste0(ln, pad_right(f, cll))
            else if (calgn == "right")
              ln <- paste0(ln, pad_left(f, cll))
            else if (calgn == "center")
              ln <- paste0(ln, pad_both(f, cll))
            
          } else {
            warning(paste0("Footnote exceeds available width.",
                        "\nFootnote: ", f,
                        "\nFootnote length: ", nchar(f), 
                        "\nLine Length: ", ll))
            
            tln <- cll - 3
            
            if (tln >= 0) {
              
              if (ftn$align == "left") {
                ln <- paste0(substr(pad_right(f , cll), 1, tln), "...")
              } else if (ftn$align == "right") {
                ln <- paste0("...", substr(pad_left(f, cll), 1, tln))
              } else if (ftn$align == "center" | ftn$align == "centre") {
                ln <- paste0(substr(pad_both(f, cll), 1, tln), "...")
              }
            } else ln <- "" 
          }
          
          i <- i + 1
        }
        
        ret[length(ret) + 1] <- ln
      }
      
      if (any(ftn$borders %in% c("bottom", "all")) & length(ftn$footnotes) > 0) {
        # ret[length(ret) + 1] <- paste0(paste0(rep(uchar, ll), 
        #                                       collapse = ""), " ")
        ret[length(ret) + 1] <- paste0(rep(uchar, ll), collapse = "")
      }
      
      if (ftn$blank_row %in% c("below", "both") & length(ftn$footnotes) > 0)
        ret[length(ret) + 1] <- ""
    }
  }
  
  return(ret)
}


#' Get footnote text strings suitable for printing
#' @param rs The report spec
#' @return A vector of strings
#' @noRd
get_footnotes_back <- function(footnotes, content_width, page_width, 
                          uchar = "-", char_width) {
  
  if (is.null(char_width)) {
    stop("width cannot be null.") 
    
  }
  
  ret <- c()
  
  if (!is.null(footnotes)) {
    for (ftn in footnotes) {
      
      if (!any(class(ftn) == "footnote_spec"))
        stop("footnotes parameter value is not a footnote spec.")
      
      if (ftn$width == "page")
        width <- page_width
      else if (ftn$width == "content")
        width <- content_width
      else if (is.numeric(ftn$width))
        width <- ceiling(ftn$width / char_width)
      
      ll <- width 
      
      if (ftn$blank_row %in% c("above", "both") & length(ftn$footnotes) > 0)
        ret[length(ret) + 1] <- ""
      
      if (any(ftn$borders %in% c("top", "all")) & length(ftn$footnotes) > 0) {
        # ret[length(ret) + 1] <- paste0(paste0(rep(uchar, ll), 
        #                                       collapse = ""), " ")
        
        ret[length(ret) + 1] <- paste0(rep(uchar, ll), collapse = "")
      }
      
      for (i in seq_along(ftn$footnotes)) {
        
        f <- ftn$footnotes[i]
        
        gp <- ll - nchar(f)
        
        #print("footnotes")
        if (gp > 0) {
          
          # if (ftn$align == "left")
          #   ln <- pad_right(paste0(f, " "), ll + 1)
          # else if (ftn$align == "right")
          #   ln <- pad_left(paste0(f, " "), ll + 1)
          # else if (ftn$align == "center" | ftn$align == "centre")
          #   ln <- pad_both(paste0(f, " "), ll + 1)
          
          if (ftn$align == "left")
            ln <- pad_right(f, ll)
          else if (ftn$align == "right")
            ln <- pad_left(f, ll)
          else if (ftn$align == "center" | ftn$align == "centre")
            ln <- pad_both(f, ll)
          
        } else {
          warning(paste0("Footnote exceeds available width.",
                         "\nFootnote: ", f,
                         "\nFootnote length: ", nchar(f), 
                         "\nLine Length: ", ll))
          
          tln <- ll - 3
          
          if (tln >= 0) {
            
            if (ftn$align == "left") {
              ln <- paste0(substr(pad_right(f , ll), 1, tln), "...")
            } else if (ftn$align == "right") {
              ln <- paste0("...", substr(pad_left(f, ll), 1, tln))
            } else if (ftn$align == "center" | ftn$align == "centre") {
              ln <- paste0(substr(pad_both(f, ll), 1, tln), "...")
            }
          } else ln <- "" 
        }
        
        
        ret[length(ret) + 1] <- ln
      }
      
      if (any(ftn$borders %in% c("bottom", "all")) & length(ftn$footnotes) > 0) {
        # ret[length(ret) + 1] <- paste0(paste0(rep(uchar, ll), 
        #                                       collapse = ""), " ")
        ret[length(ret) + 1] <- paste0(rep(uchar, ll), collapse = "")
      }
      
      if (ftn$blank_row %in% c("below", "both") & length(ftn$footnotes) > 0)
        ret[length(ret) + 1] <- ""
    }
  }
  
  return(ret)
}

#' Get page footer text strings suitable for printing
#' @param rs The report spec
#' @return A vector of strings
#' @noRd
get_page_footer <- function(rs) {
  
  if (is.null(rs$line_size)) {
    stop("line_size cannot be null.") 
    
  }
  
  pftrr <- rs$page_footer_right
  pftrl <- rs$page_footer_left
  pftrc <- rs$page_footer_center
  
  mx <- max(c(length(pftrr), length(pftrl), length(pftrc)))
  
  if (length(pftrr) < mx)
    pftrr <- c(rep("", mx - length(pftrr)), pftrr)
  
  if (length(pftrl) < mx)
    pftrl <- c(rep("", mx - length(pftrl)), pftrl)
  
  if (length(pftrc) < mx)
    pftrc <- c(rep("", mx - length(pftrc)), pftrc)
  
  # Put blank space above page footer by default
  ret <- NULL
  
  if (mx != 0) {
    
    if (rs$page_footer_blank_row == "above") 
      ret <- c("")
    else 
      ret <- c()
    
    for (i in 1:mx) {
      
      fl <- ""
      fc <- ""
      fr <- ""
      
      if (length(pftrl) >= i)
        fl <- as.character(pftrl[[i]])
      
      if (length(pftrr) >= i)
        fr <- as.character(pftrr[[i]])
      
      if (length(pftrc) >= i)
        fc <- as.character(pftrc[[i]])
      
      l_sz <- if (is.null(fl)) 0 else nchar(fl)
      r_sz <- if (is.null(fr)) 0 else nchar(fr)
      c_sz <- if (is.null(fc)) 0 else nchar(fc)
      
      gp <- rs$line_size - (l_sz + r_sz + c_sz)
      
      #print("footer")
      if (gp >= 0) {
        if (l_sz > r_sz)
          fr <- pad_left(fr, l_sz)
        else
          fl <- pad_right(fl, r_sz)
        
        lw <- rs$line_size - nchar(fr) - nchar(fl)
        ln <- paste0(fl, pad_both(fc, lw), fr)
      }
      else {
        
        stop(paste0("Page footer exceeds available width\n", 
                    "Footer Left: ", fl, "\n", 
                    "Footer Center: ", fc, "\n",
                    "Footer Right: ", fr, "\n",
                    "Footer length: ", nchar(fl) + nchar(fc) + nchar(fr), "\n",
                    "Line length: ", rs$line_size, "\n"))
        ln <- ""
      }
      
      ret[length(ret) + 1] <- ln
    }
    

  }
  
  
  return(ret)
}



# Page Info ---------------------------------------------------------------

#' @description A internal class to collect data about a report page.  This 
#' object is created in create_table_pages_text and passed around to 
#' different functions so they can make decisions based on the information
#' contained in this object.
#' @noRd
page_info <- function(data, keys, font_name, col_width, col_align,
                      label, label_align, page_by = NULL, table_align = NULL) {
  
  ret <- structure(list(), class = c("page_info", "list"))
  
  ret$data <- data
  ret$keys <- keys
  ret$font_name <- font_name
  ret$col_width <- col_width
  ret$col_align <- col_align
  ret$label <- label
  ret$label_align <- label_align
  ret$total_pages <- 0
  ret$page_number <- 0
  ret$page_by <- page_by
  ret$table_align <- table_align
  
  return(ret)
  
}




# Utilities ---------------------------------------------------------------

#' @noRd
pad_right <- Vectorize(function(s, w) {
  
  l <- w - nchar(s)
  
  if (l < 0)
    ret <- s
  else
    ret <- paste0(s, paste0(rep_len(" ", length.out = w - nchar(s)), collapse = ""))

  return(ret)
}, USE.NAMES = FALSE)

#' @noRd
pad_left <- Vectorize(function(s, w) {
  
  l <- w - nchar(s)
  
  if (l < 0)
    ret <- s
  else 
    ret <- paste0(paste0(rep_len(" ", length.out =l), collapse = ""), s)
  
  
  return(ret)
  
}, USE.NAMES = FALSE)

#' @noRd
pad_both <- Vectorize(function(s, w) {
  
  
  l <- w - nchar(s)
  
  if (l < 0)
    ret <- s
  else {
  
    lp <- floor(l / 2)
    rp <- ceiling(l / 2)
    
    ret <- paste0(paste0(rep_len(" ", length.out = lp), collapse = ""), 
                s, paste0(rep_len(" ", length.out = rp), collapse = ""))
  
  }
  
  return(ret)
  
  
}, USE.NAMES = FALSE)

pad_any <- function(s, w, j) {
  
  if (j == "left")
    ret <- pad_right(s, w)
  else if (j == "right")
    ret <- pad_left(s, w)
  else if (j %in% c("center", "centre"))
    ret <- pad_both(s, w)
  else 
    ret <- pad_right(s, w)
  
  return(ret)
  
}

#' @title
#' Write a PDF report to the file system
#'
#' @description
#' This function writes a report_spec object to the file system, using the
#' parameters provided in the object.
#'
#' @param x The report_spec object to write.
#' @return The report spec.
#' @noRd
write_report_pdf <- function(rs) {
  
  
  if (file.exists(rs$modified_path))
    file.remove(rs$modified_path)
  
  
  # Create text output 
  res <- create_report_text(rs)
  
  rs <- res[["rs"]]
  ls <- res[["ls"]]

  # Revise text and write to pdf
  fls <- write_pdf_output(rs, ls, rs$modified_path)
  
  return(rs)
}

#' @noRd
write_pdf_output <- function(rs, ls, pdf_path) {
  
  pgs <- list()
  pg <- list(text = c(), images = c())

  # Loop through text pages
  for (i in seq_along(ls)) {
  
    
    # Replace fill lines with blanks
    if (rs$has_graphics) {
      
      ls[[i]] <- gsub("```fill```", "", ls[[i]], fixed = TRUE)
      
      # Need to split up text and images
      for (ln in ls[[i]]) {
        
        res <- grep("```([^}]*)```", ln)
        if (length(res) > 0) {
          pg$text <- append(pg$text, "")
          pi <- parse_image(ln)
          pi$line_start <- length(pg$text)
          pg$images[[length(pg$images) + 1]] <- pi
        } else {
          pg$text <- append(pg$text, ln)   
        }
        
      }
      
    } else 
      pg$text <- ls[[i]]
    

    
    pgs[[length(pgs) + 1]] <- pg
    pg <- list(text = c(), images = c())
  }
  
  
  if (file.exists(pdf_path))
    file.remove(pdf_path)
  
  r <- create_pdf(filename = pdf_path,
                  margin_top = rs$margin_top,
                  margin_left = rs$margin_left,
                  fontsize = rs$font_size,
                  fontname = rs$font,
                  page_height = rs$page_size[2],
                  page_width = rs$page_size[1],
                  orientation = rs$orientation,
                  units = rs$units,
                  info = TRUE)
  
  for (pg in pgs) {
    
    if (length(pg$images) > 0) {
      
      imgs <- list()
      for (img in pg$images) {

        imgs[[length(imgs) + 1]] <- page_image(img$filename,
                                             height = img$height,
                                             width = img$width,
                                             align = img$align,
                                             line_start = img$line_start)
                  
      }
     
      r <- add_page(r, page_text(pg$text), imgs)
                    
    } else {
      r <- add_page(r, page_text(pg$text)) 
    }
  }
  
  
  write_pdf(r)
  
  
  return(pdf_path)
  
}

parse_image <- function(image_string) {
 
  ret <- list()
  
  # Remove braces
  rw <- trimws(gsub("`", "", image_string, fixed = TRUE))  
  
  # Split on pipe
  spec <- strsplit(rw, "|", fixed = TRUE)[[1]]
  
  pth <- gsub("\\", "/", spec[[1]], fixed = TRUE)
  
  # 1 = path
  # 2 = height
  # 3 = width
  # 4 = align 
  
  ret$filename <- pth
  ret$height <- as.numeric(spec[[2]])
  ret$width <- as.numeric(spec[[3]])
  ret$align <- spec[[4]]
  
  return(ret)
  
}

# Old rmarkdown version. Keep for now.
# @noRd
# write_pdf_output2 <- function(rs, ls, rmd_path, pdf_path, tmp_dir) {
#   
#   # Set up vectors
#   hdr <- c() 
#   body <- c() 
#   ret <- c()
#   
#   # Prepare header
#   hdr[length(hdr) + 1] <- "---"
#   hdr[length(hdr) + 1] <- "output:"
#   hdr[length(hdr) + 1] <- "  pdf_document:"
#   hdr[length(hdr) + 1] <- "    latex_engine: xelatex"
#   if (rs$orientation == "landscape") {
#     hdr[length(hdr) + 1] <- "classoption: landscape"
#   } # default if portrait
#   
#   if (rs$font_size == 10)
#     hdr[length(hdr) + 1] <- "fontsize: 10pt"
#   else if (rs$font_size == 12)
#     hdr[length(hdr) + 1] <- "fontsize: 12pt"
#   # if (rs$font_size == 8)
#   #   hdr[length(hdr) + 1] <- "classoption: 8pt"
# 
# 
#   
#   # "left=3cm,right=3cm,top=2cm,bottom=2cm"
#   if (rs$units == "inches") {
#     geom <- paste0(" \"left=", sprintf("%.3f", rs$margin_left * .80), 
#                    "in, right=", sprintf("%.3f", rs$margin_right * .65), 
#                    "in, top=", sprintf( "%.3f", rs$margin_top * .85), 
#                    "in, bottom=", sprintf("%.3f", rs$margin_bottom/2), "in\"")
#   } else {
#     geom <- paste0("\"left=", rs$margin_left * .80, "cm, right=", 
#                    rs$margin_right  * .65, "cm, top=", rs$margin_top * .85, 
#                    "cm, bottom=", rs$margin_bottom/2, "cm\"")
#   }
#   hdr[length(hdr) + 1] <- paste("geometry:", geom)
#   # Figure out paper size options
#   if (rs$paper_size == "letter")
#     hdr[length(hdr) + 1] <- "papersize: letter"
#   else if (rs$paper_size == "A4")
#     hdr[length(hdr) + 1] <- "papersize: a4"
#   hdr[length(hdr) + 1] <- "header-includes:"
#   if (rs$font_size == 8)
#     hdr[length(hdr) + 1] <- "  - \\usepackage[fontsize=8pt]{scrextend}"
#   hdr[length(hdr) + 1] <- "  - \\renewcommand{\\familydefault}{\\ttdefault}"
#   hdr[length(hdr) + 1] <- "  - \\thispagestyle{empty}"
#   hdr[length(hdr) + 1] <- "---"
# 
#   hdr[length(hdr) + 1] <- "\\pagenumbering{gobble}"
#   hdr[length(hdr) + 1] <- "\\begin{verbatim}"
# 
#   # Start with all lines
#   body <- ls
#   
#   if (rs$has_graphics) {
#     # Remove fill lines
#     fill_tags <- grep("```fill```", body, fixed = TRUE)
#     body <- body[-fill_tags]
#     
#     # Replace any plot tags with latex codes
#     plt_tags <- grep("```([^}]*)```", body)
#     
#     for (i in plt_tags) {
#       
#       # Remove braces
#       rw <- trimws(gsub("`", "", body[i], fixed = TRUE))  
#       
#       # Split on pipe
#       spec <- strsplit(rw, "|", fixed = TRUE)[[1]]
#       
#       pth <- gsub("\\", "/", spec[[1]], fixed = TRUE)
#       ret[length(ret) + 1] <- pth
#   
#       # 1 = path
#       # 2 = height
#       # 3 = width
#       # 4 = align
#       
#       # Create latex codes
#       if (spec[[4]] == "left") {
#         ltx <- paste0("\\end{verbatim}\n",
#                       "\\begin{figure}[h!]\n",
#                       "\\begin{flushleft}\n", 
#                       "\\includegraphics{", pth, "}\n",
#                       "\\end{flushleft}\n",
#                       "\\end{figure}\n",
#                       "\\begin{verbatim}")
#         
#       } else if (spec[[4]] == "right") {
#         ltx <- paste0("\\end{verbatim}\n",
#                       "\\begin{figure}[h!]\n",
#                        "\\begin{flushright}\n", 
#                        "\\includegraphics{", pth, "}\n",
#                        "\\end{flushright}\n",
#                        "\\end{figure}\n",
#                        "\\begin{verbatim}")
#       } else  {
#       ltx <- paste0("\\end{verbatim}\n",
#                     "\\begin{figure}[h!]\n",
#                     "\\centering\n", 
#                     "\\includegraphics{", pth, "}\n",
#                     "\\end{figure}\n",
#                     "\\begin{verbatim}")
#       }
#       
#       # Replace original line with latex codes
#       body[[i]] <- ltx
#     }
#   }
# 
#   
#   body <- gsub("\f", "\\end{verbatim}\n\\newpage\n\\begin{verbatim}", body, fixed = TRUE)
#   body[length(body) + 1] <- "\\end{verbatim}"
#   
#   t1 <- tempfile(tmpdir = tmp_dir, fileext = ".pdf")
#   
#   if (file.exists(rmd_path))
#     file.remove(rmd_path)
# 
#   # Write to file
#   f <- file(rmd_path, open="w+", encoding = "native.enc")
# 
# 
#   writeLines(enc2utf8(hdr), con = f, useBytes = TRUE)
# 
#   writeLines(enc2utf8(body), con = f, useBytes = TRUE)
# 
#   close(f)
#   
#   # There is a known warning on rendering 8pt font.
#   # Can't figure out how to get rid of it.
#   # So going to suppress it for now.
#   # Other fonts sizes shouldn't have any warnings.
#   # if (rs$font_size == 8) {
#   #   suppressWarnings(rmarkdown::render(rmd_path, 
#   #                                      rmarkdown::pdf_document(), t1, 
#   #                                      quiet = TRUE))
#   # } else {
#   #   rmarkdown::render(rmd_path, 
#   #                     rmarkdown::pdf_document(), t1, 
#   #                     quiet = TRUE)
#   # }
#   
#   file.copy(t1, pdf_path)
# 
#   if (file.exists(t1))
#     file.remove(t1)
#   
#   return(ret)
#   
# }

# Globals -----------------------------------------------------------------

# pointsize <- 1/72
# inchsize <- 72
# in2cm <- 2.54



# Write DOCX --------------------------------------------------------------

#' @import zip
#' @import withr
#' @noRd
write_docx <- function(src, pth) {

  # print(src)
  # print(file.exists(src))
  # print(pth)
  # Sys.sleep(1)
  
  if (file.exists(src)) {
    
    npth <- suppressWarnings(normalizePath(pth))
    
    if (grepl("^\\./", npth)) {
      npth <- file.path(getwd(), sub("./", "", npth, fixed = TRUE))
    }
    
    if (file.exists(npth))
      file.remove(npth)
    
    if (!file.exists(dirname(npth)))
      dir.create(dirname(npth))
    
    
    fls <- list.files(src, recursive = TRUE, all.files = TRUE)
    
    # print("Just before withr")
    # print(src)
    # print(npth)
    # 
    # tmp <- getwd()
    # print(tmp)
    withr::with_dir(src,
   # setwd(src)
    zip::zip(npth, fls, mode = "mirror"))
    # setwd(src)
    # utils::zip(pth, fls)
    
  #  setwd(tmp)
    # print("Just after zip")
  
    
  } else {
    
   print("Nothing") 
  }
  
}



# Create DOCX --------------------------------------------------------------

#' @noRd
create_new_docx <- function(font, font_size, imageCount, imagePaths,
                            footer_imagePath = NULL, header_imagePath = NULL) {
  
  tdd <- file.path(tempdir(), stri_rand_strings(1, length = 6))
  
  dir.create(tdd)
  dir.create(file.path(tdd, "_rels"))
  dir.create(file.path(tdd, "docProps"))
  dir.create(file.path(tdd, "word"))
  dir.create(file.path(tdd, "word/_rels"))
  dir.create(file.path(tdd, "word/media"))
  
  create_content_types(tdd)
  create_web_settings(tdd)
  create_font_table(tdd)
  create_styles(tdd, font, font_size)
  create_app(tdd)
  create_core(tdd)
  create_endnotes(tdd)
  create_footnotes(tdd)
  create_document_rels(tdd, imageCount, imagePaths)
  
  if (!is.null(footer_imagePath)) {
    create_footer_rels(tdd, length(footer_imagePath), footer_imagePath)
  }
  
  if (!is.null(header_imagePath)) {
    create_header_rels(tdd, length(header_imagePath), header_imagePath)
  }
  
  create_rels(tdd)
  
  # Temporary
 # create_document(tdd) 
  # create_header(tdd)
   create_footer(tdd)

  
  
  return(tdd)
}

#' @noRd
create_content_types <- function(pth) {
  
  
cnt <- '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<Types
	xmlns="http://schemas.openxmlformats.org/package/2006/content-types">
	<Default Extension="jpeg" ContentType="image/jpeg"/>
	<Default Extension="png" ContentType="image/png"/>
	<Default Extension="emf" ContentType="image/emf"/>
	<Default Extension="rels" 
	ContentType="application/vnd.openxmlformats-package.relationships+xml"/>
	<Default Extension="xml" 
	ContentType="application/xml"/>
	<Override PartName="/word/document.xml" 
	ContentType="application/vnd.openxmlformats-officedocument.wordprocessingml.document.main+xml"/>
	<Override PartName="/word/styles.xml" 
	ContentType="application/vnd.openxmlformats-officedocument.wordprocessingml.styles+xml"/>
	<Override PartName="/word/settings.xml" 
	ContentType="application/vnd.openxmlformats-officedocument.wordprocessingml.settings+xml"/>
	<Override PartName="/word/webSettings.xml" 
	ContentType="application/vnd.openxmlformats-officedocument.wordprocessingml.webSettings+xml"/>
	<Override PartName="/word/footnotes.xml" 
	ContentType="application/vnd.openxmlformats-officedocument.wordprocessingml.footnotes+xml"/>
	<Override PartName="/word/endnotes.xml" 
	ContentType="application/vnd.openxmlformats-officedocument.wordprocessingml.endnotes+xml"/>
	<Override PartName="/word/header1.xml" 
	ContentType="application/vnd.openxmlformats-officedocument.wordprocessingml.header+xml"/>
	<Override PartName="/word/footer1.xml" 
	ContentType="application/vnd.openxmlformats-officedocument.wordprocessingml.footer+xml"/>
	<Override PartName="/word/fontTable.xml" 
	ContentType="application/vnd.openxmlformats-officedocument.wordprocessingml.fontTable+xml"/>
	<Override PartName="/docProps/core.xml" 
	ContentType="application/vnd.openxmlformats-package.core-properties+xml"/>
	<Override PartName="/docProps/app.xml" 
	ContentType="application/vnd.openxmlformats-officedocument.extended-properties+xml"/>
</Types>'


  nm <- file.path(pth, "[Content_Types].xml")

  f <- file(nm, open="w", encoding = "native.enc")

  writeLines(cnt, f,  useBytes = TRUE) 
  
  close(f)
  
  
}


create_web_settings <- function(pth) {
  
  
  cnt <- paste0('<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n',
'<w:webSettings ',
'xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006" ',
'xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships" ',
'xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main" ',
'xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml" ',
'xmlns:w15="http://schemas.microsoft.com/office/word/2012/wordml" ',
'xmlns:w16cex="http://schemas.microsoft.com/office/word/2018/wordml/cex" ',
'xmlns:w16cid="http://schemas.microsoft.com/office/word/2016/wordml/cid" ',
'xmlns:w16="http://schemas.microsoft.com/office/word/2018/wordml" ',
'xmlns:w16sdtdh="http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash" ',
'xmlns:w16se="http://schemas.microsoft.com/office/word/2015/wordml/symex" ', 
'mc:Ignorable="w14 w15 w16se w16cid w16 w16cex w16sdtdh">',
'<w:optimizeForBrowser/>',
'<w:allowPNG/>',
'</w:webSettings>')
  
  
  nm <- file.path(pth, "word/webSettings.xml")
  
  f <- file(nm, open="w", encoding = "native.enc")
  
  writeLines(cnt, f,  useBytes = TRUE) 
  
  close(f)
  
  
}

create_font_table <- function(pth) {
  
  cnt <- paste0('<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n
<w:fonts
	xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006"
	xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships"
	xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main"
	xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml"
	xmlns:w15="http://schemas.microsoft.com/office/word/2012/wordml"
	xmlns:w16cex="http://schemas.microsoft.com/office/word/2018/wordml/cex"
	xmlns:w16cid="http://schemas.microsoft.com/office/word/2016/wordml/cid"
	xmlns:w16="http://schemas.microsoft.com/office/word/2018/wordml"
	xmlns:w16sdtdh="http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash"
	xmlns:w16se="http://schemas.microsoft.com/office/word/2015/wordml/symex" 
	mc:Ignorable="w14 w15 w16se w16cid w16 w16cex w16sdtdh">
	<w:font w:name="Times New Roman">
		<w:panose1 w:val="02020603050405020304"/>
		<w:charset w:val="00"/>
		<w:family w:val="roman"/>
		<w:pitch w:val="variable"/>
		<w:sig w:usb0="E0002EFF" w:usb1="C000785B" w:usb2="00000009" 
		w:usb3="00000000" w:csb0="000001FF" w:csb1="00000000"/>
	</w:font>
	<w:font w:name="Arial">
		<w:panose1 w:val="020B0604020202020204"/>
		<w:charset w:val="00"/>
		<w:family w:val="swiss"/>
		<w:pitch w:val="variable"/>
		<w:sig w:usb0="E0002EFF" w:usb1="C000785B" w:usb2="00000009" 
		w:usb3="00000000" w:csb0="000001FF" w:csb1="00000000"/>
	</w:font>
	<w:font w:name="Courier New">
		<w:panose1 w:val="02070309020205020404"/>
		<w:charset w:val="00"/>
		<w:family w:val="modern"/>
		<w:pitch w:val="fixed"/>
		<w:sig w:usb0="E0002EFF" w:usb1="C0007843" w:usb2="00000009" 
		w:usb3="00000000" w:csb0="000001FF" w:csb1="00000000"/>
	</w:font>
</w:fonts>')
  
  
  nm <- file.path(pth, "word/fontTable.xml")
  
  f <- file(nm, open="w", encoding = "native.enc")
  
  writeLines(cnt, f,  useBytes = TRUE) 
  
  close(f)
  
}

create_settings <- function(pth) {
  
 cnt <- '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n
<w:settings
	xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006"
	xmlns:o="urn:schemas-microsoft-com:office:office"
	xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships"
	xmlns:m="http://schemas.openxmlformats.org/officeDocument/2006/math"
	xmlns:v="urn:schemas-microsoft-com:vml"
	xmlns:w10="urn:schemas-microsoft-com:office:word"
	xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main"
	xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml"
	xmlns:w15="http://schemas.microsoft.com/office/word/2012/wordml"
	xmlns:w16cex="http://schemas.microsoft.com/office/word/2018/wordml/cex"
	xmlns:w16cid="http://schemas.microsoft.com/office/word/2016/wordml/cid"
	xmlns:w16="http://schemas.microsoft.com/office/word/2018/wordml"
	xmlns:w16sdtdh="http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash"
	xmlns:w16se="http://schemas.microsoft.com/office/word/2015/wordml/symex"
	xmlns:sl="http://schemas.openxmlformats.org/schemaLibrary/2006/main" 
	mc:Ignorable="w14 w15 w16se w16cid w16 w16cex w16sdtdh">
	<w:zoom w:percent="100"/>
	<w:proofState w:spelling="clean" w:grammar="clean"/>
	<w:defaultTabStop w:val="720"/>
	<w:characterSpacingControl w:val="doNotCompress"/>
	<w:footnotePr>
		<w:footnote w:id="-1"/>
		<w:footnote w:id="0"/>
	</w:footnotePr>
	<w:endnotePr>
		<w:endnote w:id="-1"/>
		<w:endnote w:id="0"/>
	</w:endnotePr>
	<w:compat>
		<w:compatSetting w:name="compatibilityMode" 
		w:uri="http://schemas.microsoft.com/office/word" w:val="15"/>
		<w:compatSetting w:name="overrideTableStyleFontSizeAndJustification" 
		w:uri="http://schemas.microsoft.com/office/word" w:val="1"/>
		<w:compatSetting w:name="enableOpenTypeFeatures" 
		w:uri="http://schemas.microsoft.com/office/word" w:val="1"/>
		<w:compatSetting w:name="doNotFlipMirrorIndents" 
		w:uri="http://schemas.microsoft.com/office/word" w:val="1"/>
		<w:compatSetting w:name="differentiateMultirowTableHeaders" 
		w:uri="http://schemas.microsoft.com/office/word" w:val="1"/>
		<w:compatSetting w:name="useWord2013TrackBottomHyphenation" 
		w:uri="http://schemas.microsoft.com/office/word" w:val="0"/>
	</w:compat>
	<w:rsids>
		<w:rsidRoot w:val="005D1180"/>
		<w:rsid w:val="00046D2A"/>
		<w:rsid w:val="00444C49"/>
		<w:rsid w:val="005D1180"/>
		<w:rsid w:val="00BC1857"/>
		<w:rsid w:val="00F10374"/>
		<w:rsid w:val="00FC21CA"/>
	</w:rsids>
	<m:mathPr>
		<m:mathFont m:val="Cambria Math"/>
		<m:brkBin m:val="before"/>
		<m:brkBinSub m:val="--"/>
		<m:smallFrac m:val="0"/>
		<m:dispDef/>
		<m:lMargin m:val="0"/>
		<m:rMargin m:val="0"/>
		<m:defJc m:val="centerGroup"/>
		<m:wrapIndent m:val="1440"/>
		<m:intLim m:val="subSup"/>
		<m:naryLim m:val="undOvr"/>
	</m:mathPr>
	<w:themeFontLang w:val="en-US"/>
	<w:clrSchemeMapping w:bg1="light1" w:t1="dark1" w:bg2="light2" w:t2="dark2" 
	w:accent1="accent1" w:accent2="accent2" w:accent3="accent3" w:accent4="accent4" 
	w:accent5="accent5" w:accent6="accent6" w:hyperlink="hyperlink" 
	w:followedHyperlink="followedHyperlink"/>
	<w:shapeDefaults>
		<o:shapedefaults v:ext="edit" spidmax="1026"/>
		<o:shapelayout v:ext="edit">
			<o:idmap v:ext="edit" data="1"/>
		</o:shapelayout>
	</w:shapeDefaults>
	<w:decimalSymbol w:val="."/>
	<w:listSeparator w:val=","/>
	<w14:docId w14:val="0322C6E4"/>
	<w15:chartTrackingRefBased/>
	<w15:docId w15:val="{C1FAABC8-AE44-436C-B3A7-4CB20B70DB45}"/>
</w:settings>'
 
 
 nm <- file.path(pth, "word/settings.xml")
 
 f <- file(nm, open="w", encoding = "native.enc")
 
 writeLines(cnt, f,  useBytes = TRUE) 
 
 close(f)
 
}

create_styles <- function(pth, font, font_size) {
  
  fnt <- font
  if (toupper(font) == "TIMES")
    fnt <- "Times New Roman"
  else if (toupper(font) == "COURIER")
    fnt <- "Courier New"
    
  
  fs <- font_size * 2
  
  cnt <- paste0('<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n
<w:styles
	xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006"
	xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships"
	xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main"
	xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml"
	xmlns:w15="http://schemas.microsoft.com/office/word/2012/wordml"
	xmlns:w16cex="http://schemas.microsoft.com/office/word/2018/wordml/cex"
	xmlns:w16cid="http://schemas.microsoft.com/office/word/2016/wordml/cid"
	xmlns:w16="http://schemas.microsoft.com/office/word/2018/wordml"
	xmlns:w16sdtdh="http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash"
	xmlns:w16se="http://schemas.microsoft.com/office/word/2015/wordml/symex" 
	mc:Ignorable="w14 w15 w16se w16cid w16 w16cex w16sdtdh">
	<w:docDefaults>
		<w:rPrDefault>
			<w:rPr>
			  <w:rFonts w:ascii="', fnt, '" w:hAnsi="', fnt, 
			  '" w:cs="', fnt, '"/>
				<w:sz w:val="', fs, '"/>
				<w:szCs w:val="', fs, '"/>
				<w:lang w:val="en-US" w:eastAsia="en-US" w:bidi="ar-SA"/>
			</w:rPr>
		</w:rPrDefault>
		<w:pPrDefault>
			<w:pPr>
				<w:spacing w:after="160" w:line="259" w:lineRule="auto"/>
			</w:pPr>
		</w:pPrDefault>
	</w:docDefaults>
	<w:latentStyles w:defLockedState="0" w:defUIPriority="99" w:defSemiHidden="0" 
	w:defUnhideWhenUsed="0" w:defQFormat="0" w:count="376">
		<w:lsdException w:name="Normal" w:uiPriority="0" w:qFormat="1"/>
		<w:lsdException w:name="heading 1" w:uiPriority="9" w:qFormat="1"/>
		<w:lsdException w:name="heading 2" w:semiHidden="1" w:uiPriority="9" 
		w:unhideWhenUsed="1" w:qFormat="1"/>
		<w:lsdException w:name="heading 3" w:semiHidden="1" w:uiPriority="9" 
		w:unhideWhenUsed="1" w:qFormat="1"/>
		<w:lsdException w:name="heading 4" w:semiHidden="1" w:uiPriority="9" 
		w:unhideWhenUsed="1" w:qFormat="1"/>
		<w:lsdException w:name="heading 5" w:semiHidden="1" w:uiPriority="9" 
		w:unhideWhenUsed="1" w:qFormat="1"/>
		<w:lsdException w:name="heading 6" w:semiHidden="1" w:uiPriority="9" 
		w:unhideWhenUsed="1" w:qFormat="1"/>
		<w:lsdException w:name="heading 7" w:semiHidden="1" w:uiPriority="9" 
		w:unhideWhenUsed="1" w:qFormat="1"/>
		<w:lsdException w:name="heading 8" w:semiHidden="1" w:uiPriority="9" 
		w:unhideWhenUsed="1" w:qFormat="1"/>
		<w:lsdException w:name="heading 9" w:semiHidden="1" w:uiPriority="9" 
		w:unhideWhenUsed="1" w:qFormat="1"/>
		<w:lsdException w:name="index 1" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="index 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="index 3" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="index 4" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="index 5" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="index 6" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="index 7" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="index 8" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="index 9" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="toc 1" w:semiHidden="1" w:uiPriority="39" 
		w:unhideWhenUsed="1"/>
		<w:lsdException w:name="toc 2" w:semiHidden="1" w:uiPriority="39" 
		w:unhideWhenUsed="1"/>
		<w:lsdException w:name="toc 3" w:semiHidden="1" w:uiPriority="39" 
		w:unhideWhenUsed="1"/>
		<w:lsdException w:name="toc 4" w:semiHidden="1" w:uiPriority="39" 
		w:unhideWhenUsed="1"/>
		<w:lsdException w:name="toc 5" w:semiHidden="1" w:uiPriority="39" 
		w:unhideWhenUsed="1"/>
		<w:lsdException w:name="toc 6" w:semiHidden="1" w:uiPriority="39" 
		w:unhideWhenUsed="1"/>
		<w:lsdException w:name="toc 7" w:semiHidden="1" w:uiPriority="39" 
		w:unhideWhenUsed="1"/>
		<w:lsdException w:name="toc 8" w:semiHidden="1" w:uiPriority="39" 
		w:unhideWhenUsed="1"/>
		<w:lsdException w:name="toc 9" w:semiHidden="1" w:uiPriority="39" 
		w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Normal Indent" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="footnote text" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="annotation text" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="header" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="footer" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="index heading" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="caption" w:semiHidden="1" w:uiPriority="35" 
		w:unhideWhenUsed="1" w:qFormat="1"/>
		<w:lsdException w:name="table of figures" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="envelope address" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="envelope return" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="footnote reference" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="annotation reference" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="line number" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="page number" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="endnote reference" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="endnote text" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="table of authorities" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="macro" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="toa heading" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List Bullet" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List Number" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List 3" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List 4" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List 5" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List Bullet 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List Bullet 3" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List Bullet 4" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List Bullet 5" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List Number 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List Number 3" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List Number 4" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List Number 5" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Title" w:uiPriority="10" w:qFormat="1"/>
		<w:lsdException w:name="Closing" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Signature" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Default Paragraph Font" w:semiHidden="1" 
		w:uiPriority="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Body Text" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Body Text Indent" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List Continue" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List Continue 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List Continue 3" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List Continue 4" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="List Continue 5" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Message Header" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Subtitle" w:uiPriority="11" w:qFormat="1"/>
		<w:lsdException w:name="Salutation" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Date" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Body Text First Indent" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Body Text First Indent 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Note Heading" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Body Text 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Body Text 3" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Body Text Indent 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Body Text Indent 3" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Block Text" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Hyperlink" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="FollowedHyperlink" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Strong" w:uiPriority="22" w:qFormat="1"/>
		<w:lsdException w:name="Emphasis" w:uiPriority="20" w:qFormat="1"/>
		<w:lsdException w:name="Document Map" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Plain Text" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="E-mail Signature" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="HTML Top of Form" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="HTML Bottom of Form" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Normal (Web)" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="HTML Acronym" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="HTML Address" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="HTML Cite" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="HTML Code" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="HTML Definition" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="HTML Keyboard" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="HTML Preformatted" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="HTML Sample" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="HTML Typewriter" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="HTML Variable" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Normal Table" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="annotation subject" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="No List" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Outline List 1" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Outline List 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Outline List 3" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Simple 1" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Simple 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Simple 3" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Classic 1" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Classic 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Classic 3" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Classic 4" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Colorful 1" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Colorful 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Colorful 3" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Columns 1" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Columns 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Columns 3" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Columns 4" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Columns 5" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Grid 1" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Grid 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Grid 3" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Grid 4" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Grid 5" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Grid 6" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Grid 7" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Grid 8" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table List 1" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table List 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table List 3" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table List 4" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table List 5" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table List 6" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table List 7" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table List 8" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table 3D effects 1" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table 3D effects 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table 3D effects 3" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Contemporary" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Elegant" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Professional" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Subtle 1" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Subtle 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Web 1" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Web 2" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Web 3" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Balloon Text" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Table Grid" w:uiPriority="39"/>
		<w:lsdException w:name="Table Theme" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Placeholder Text" w:semiHidden="1"/>
		<w:lsdException w:name="No Spacing" w:uiPriority="1" w:qFormat="1"/>
		<w:lsdException w:name="Light Shading" w:uiPriority="60"/>
		<w:lsdException w:name="Light List" w:uiPriority="61"/>
		<w:lsdException w:name="Light Grid" w:uiPriority="62"/>
		<w:lsdException w:name="Medium Shading 1" w:uiPriority="63"/>
		<w:lsdException w:name="Medium Shading 2" w:uiPriority="64"/>
		<w:lsdException w:name="Medium List 1" w:uiPriority="65"/>
		<w:lsdException w:name="Medium List 2" w:uiPriority="66"/>
		<w:lsdException w:name="Medium Grid 1" w:uiPriority="67"/>
		<w:lsdException w:name="Medium Grid 2" w:uiPriority="68"/>
		<w:lsdException w:name="Medium Grid 3" w:uiPriority="69"/>
		<w:lsdException w:name="Dark List" w:uiPriority="70"/>
		<w:lsdException w:name="Colorful Shading" w:uiPriority="71"/>
		<w:lsdException w:name="Colorful List" w:uiPriority="72"/>
		<w:lsdException w:name="Colorful Grid" w:uiPriority="73"/>
		<w:lsdException w:name="Light Shading Accent 1" w:uiPriority="60"/>
		<w:lsdException w:name="Light List Accent 1" w:uiPriority="61"/>
		<w:lsdException w:name="Light Grid Accent 1" w:uiPriority="62"/>
		<w:lsdException w:name="Medium Shading 1 Accent 1" w:uiPriority="63"/>
		<w:lsdException w:name="Medium Shading 2 Accent 1" w:uiPriority="64"/>
		<w:lsdException w:name="Medium List 1 Accent 1" w:uiPriority="65"/>
		<w:lsdException w:name="Revision" w:semiHidden="1"/>
		<w:lsdException w:name="List Paragraph" w:uiPriority="34" w:qFormat="1"/>
		<w:lsdException w:name="Quote" w:uiPriority="29" w:qFormat="1"/>
		<w:lsdException w:name="Intense Quote" w:uiPriority="30" w:qFormat="1"/>
		<w:lsdException w:name="Medium List 2 Accent 1" w:uiPriority="66"/>
		<w:lsdException w:name="Medium Grid 1 Accent 1" w:uiPriority="67"/>
		<w:lsdException w:name="Medium Grid 2 Accent 1" w:uiPriority="68"/>
		<w:lsdException w:name="Medium Grid 3 Accent 1" w:uiPriority="69"/>
		<w:lsdException w:name="Dark List Accent 1" w:uiPriority="70"/>
		<w:lsdException w:name="Colorful Shading Accent 1" w:uiPriority="71"/>
		<w:lsdException w:name="Colorful List Accent 1" w:uiPriority="72"/>
		<w:lsdException w:name="Colorful Grid Accent 1" w:uiPriority="73"/>
		<w:lsdException w:name="Light Shading Accent 2" w:uiPriority="60"/>
		<w:lsdException w:name="Light List Accent 2" w:uiPriority="61"/>
		<w:lsdException w:name="Light Grid Accent 2" w:uiPriority="62"/>
		<w:lsdException w:name="Medium Shading 1 Accent 2" w:uiPriority="63"/>
		<w:lsdException w:name="Medium Shading 2 Accent 2" w:uiPriority="64"/>
		<w:lsdException w:name="Medium List 1 Accent 2" w:uiPriority="65"/>
		<w:lsdException w:name="Medium List 2 Accent 2" w:uiPriority="66"/>
		<w:lsdException w:name="Medium Grid 1 Accent 2" w:uiPriority="67"/>
		<w:lsdException w:name="Medium Grid 2 Accent 2" w:uiPriority="68"/>
		<w:lsdException w:name="Medium Grid 3 Accent 2" w:uiPriority="69"/>
		<w:lsdException w:name="Dark List Accent 2" w:uiPriority="70"/>
		<w:lsdException w:name="Colorful Shading Accent 2" w:uiPriority="71"/>
		<w:lsdException w:name="Colorful List Accent 2" w:uiPriority="72"/>
		<w:lsdException w:name="Colorful Grid Accent 2" w:uiPriority="73"/>
		<w:lsdException w:name="Light Shading Accent 3" w:uiPriority="60"/>
		<w:lsdException w:name="Light List Accent 3" w:uiPriority="61"/>
		<w:lsdException w:name="Light Grid Accent 3" w:uiPriority="62"/>
		<w:lsdException w:name="Medium Shading 1 Accent 3" w:uiPriority="63"/>
		<w:lsdException w:name="Medium Shading 2 Accent 3" w:uiPriority="64"/>
		<w:lsdException w:name="Medium List 1 Accent 3" w:uiPriority="65"/>
		<w:lsdException w:name="Medium List 2 Accent 3" w:uiPriority="66"/>
		<w:lsdException w:name="Medium Grid 1 Accent 3" w:uiPriority="67"/>
		<w:lsdException w:name="Medium Grid 2 Accent 3" w:uiPriority="68"/>
		<w:lsdException w:name="Medium Grid 3 Accent 3" w:uiPriority="69"/>
		<w:lsdException w:name="Dark List Accent 3" w:uiPriority="70"/>
		<w:lsdException w:name="Colorful Shading Accent 3" w:uiPriority="71"/>
		<w:lsdException w:name="Colorful List Accent 3" w:uiPriority="72"/>
		<w:lsdException w:name="Colorful Grid Accent 3" w:uiPriority="73"/>
		<w:lsdException w:name="Light Shading Accent 4" w:uiPriority="60"/>
		<w:lsdException w:name="Light List Accent 4" w:uiPriority="61"/>
		<w:lsdException w:name="Light Grid Accent 4" w:uiPriority="62"/>
		<w:lsdException w:name="Medium Shading 1 Accent 4" w:uiPriority="63"/>
		<w:lsdException w:name="Medium Shading 2 Accent 4" w:uiPriority="64"/>
		<w:lsdException w:name="Medium List 1 Accent 4" w:uiPriority="65"/>
		<w:lsdException w:name="Medium List 2 Accent 4" w:uiPriority="66"/>
		<w:lsdException w:name="Medium Grid 1 Accent 4" w:uiPriority="67"/>
		<w:lsdException w:name="Medium Grid 2 Accent 4" w:uiPriority="68"/>
		<w:lsdException w:name="Medium Grid 3 Accent 4" w:uiPriority="69"/>
		<w:lsdException w:name="Dark List Accent 4" w:uiPriority="70"/>
		<w:lsdException w:name="Colorful Shading Accent 4" w:uiPriority="71"/>
		<w:lsdException w:name="Colorful List Accent 4" w:uiPriority="72"/>
		<w:lsdException w:name="Colorful Grid Accent 4" w:uiPriority="73"/>
		<w:lsdException w:name="Light Shading Accent 5" w:uiPriority="60"/>
		<w:lsdException w:name="Light List Accent 5" w:uiPriority="61"/>
		<w:lsdException w:name="Light Grid Accent 5" w:uiPriority="62"/>
		<w:lsdException w:name="Medium Shading 1 Accent 5" w:uiPriority="63"/>
		<w:lsdException w:name="Medium Shading 2 Accent 5" w:uiPriority="64"/>
		<w:lsdException w:name="Medium List 1 Accent 5" w:uiPriority="65"/>
		<w:lsdException w:name="Medium List 2 Accent 5" w:uiPriority="66"/>
		<w:lsdException w:name="Medium Grid 1 Accent 5" w:uiPriority="67"/>
		<w:lsdException w:name="Medium Grid 2 Accent 5" w:uiPriority="68"/>
		<w:lsdException w:name="Medium Grid 3 Accent 5" w:uiPriority="69"/>
		<w:lsdException w:name="Dark List Accent 5" w:uiPriority="70"/>
		<w:lsdException w:name="Colorful Shading Accent 5" w:uiPriority="71"/>
		<w:lsdException w:name="Colorful List Accent 5" w:uiPriority="72"/>
		<w:lsdException w:name="Colorful Grid Accent 5" w:uiPriority="73"/>
		<w:lsdException w:name="Light Shading Accent 6" w:uiPriority="60"/>
		<w:lsdException w:name="Light List Accent 6" w:uiPriority="61"/>
		<w:lsdException w:name="Light Grid Accent 6" w:uiPriority="62"/>
		<w:lsdException w:name="Medium Shading 1 Accent 6" w:uiPriority="63"/>
		<w:lsdException w:name="Medium Shading 2 Accent 6" w:uiPriority="64"/>
		<w:lsdException w:name="Medium List 1 Accent 6" w:uiPriority="65"/>
		<w:lsdException w:name="Medium List 2 Accent 6" w:uiPriority="66"/>
		<w:lsdException w:name="Medium Grid 1 Accent 6" w:uiPriority="67"/>
		<w:lsdException w:name="Medium Grid 2 Accent 6" w:uiPriority="68"/>
		<w:lsdException w:name="Medium Grid 3 Accent 6" w:uiPriority="69"/>
		<w:lsdException w:name="Dark List Accent 6" w:uiPriority="70"/>
		<w:lsdException w:name="Colorful Shading Accent 6" w:uiPriority="71"/>
		<w:lsdException w:name="Colorful List Accent 6" w:uiPriority="72"/>
		<w:lsdException w:name="Colorful Grid Accent 6" w:uiPriority="73"/>
		<w:lsdException w:name="Subtle Emphasis" w:uiPriority="19" w:qFormat="1"/>
		<w:lsdException w:name="Intense Emphasis" w:uiPriority="21" w:qFormat="1"/>
		<w:lsdException w:name="Subtle Reference" w:uiPriority="31" w:qFormat="1"/>
		<w:lsdException w:name="Intense Reference" w:uiPriority="32" w:qFormat="1"/>
		<w:lsdException w:name="Book Title" w:uiPriority="33" w:qFormat="1"/>
		<w:lsdException w:name="Bibliography" w:semiHidden="1" w:uiPriority="37" 
		w:unhideWhenUsed="1"/>
		<w:lsdException w:name="TOC Heading" w:semiHidden="1" w:uiPriority="39" 
		w:unhideWhenUsed="1" w:qFormat="1"/>
		<w:lsdException w:name="Plain Table 1" w:uiPriority="41"/>
		<w:lsdException w:name="Plain Table 2" w:uiPriority="42"/>
		<w:lsdException w:name="Plain Table 3" w:uiPriority="43"/>
		<w:lsdException w:name="Plain Table 4" w:uiPriority="44"/>
		<w:lsdException w:name="Plain Table 5" w:uiPriority="45"/>
		<w:lsdException w:name="Grid Table Light" w:uiPriority="40"/>
		<w:lsdException w:name="Grid Table 1 Light" w:uiPriority="46"/>
		<w:lsdException w:name="Grid Table 2" w:uiPriority="47"/>
		<w:lsdException w:name="Grid Table 3" w:uiPriority="48"/>
		<w:lsdException w:name="Grid Table 4" w:uiPriority="49"/>
		<w:lsdException w:name="Grid Table 5 Dark" w:uiPriority="50"/>
		<w:lsdException w:name="Grid Table 6 Colorful" w:uiPriority="51"/>
		<w:lsdException w:name="Grid Table 7 Colorful" w:uiPriority="52"/>
		<w:lsdException w:name="Grid Table 1 Light Accent 1" w:uiPriority="46"/>
		<w:lsdException w:name="Grid Table 2 Accent 1" w:uiPriority="47"/>
		<w:lsdException w:name="Grid Table 3 Accent 1" w:uiPriority="48"/>
		<w:lsdException w:name="Grid Table 4 Accent 1" w:uiPriority="49"/>
		<w:lsdException w:name="Grid Table 5 Dark Accent 1" w:uiPriority="50"/>
		<w:lsdException w:name="Grid Table 6 Colorful Accent 1" w:uiPriority="51"/>
		<w:lsdException w:name="Grid Table 7 Colorful Accent 1" w:uiPriority="52"/>
		<w:lsdException w:name="Grid Table 1 Light Accent 2" w:uiPriority="46"/>
		<w:lsdException w:name="Grid Table 2 Accent 2" w:uiPriority="47"/>
		<w:lsdException w:name="Grid Table 3 Accent 2" w:uiPriority="48"/>
		<w:lsdException w:name="Grid Table 4 Accent 2" w:uiPriority="49"/>
		<w:lsdException w:name="Grid Table 5 Dark Accent 2" w:uiPriority="50"/>
		<w:lsdException w:name="Grid Table 6 Colorful Accent 2" w:uiPriority="51"/>
		<w:lsdException w:name="Grid Table 7 Colorful Accent 2" w:uiPriority="52"/>
		<w:lsdException w:name="Grid Table 1 Light Accent 3" w:uiPriority="46"/>
		<w:lsdException w:name="Grid Table 2 Accent 3" w:uiPriority="47"/>
		<w:lsdException w:name="Grid Table 3 Accent 3" w:uiPriority="48"/>
		<w:lsdException w:name="Grid Table 4 Accent 3" w:uiPriority="49"/>
		<w:lsdException w:name="Grid Table 5 Dark Accent 3" w:uiPriority="50"/>
		<w:lsdException w:name="Grid Table 6 Colorful Accent 3" w:uiPriority="51"/>
		<w:lsdException w:name="Grid Table 7 Colorful Accent 3" w:uiPriority="52"/>
		<w:lsdException w:name="Grid Table 1 Light Accent 4" w:uiPriority="46"/>
		<w:lsdException w:name="Grid Table 2 Accent 4" w:uiPriority="47"/>
		<w:lsdException w:name="Grid Table 3 Accent 4" w:uiPriority="48"/>
		<w:lsdException w:name="Grid Table 4 Accent 4" w:uiPriority="49"/>
		<w:lsdException w:name="Grid Table 5 Dark Accent 4" w:uiPriority="50"/>
		<w:lsdException w:name="Grid Table 6 Colorful Accent 4" w:uiPriority="51"/>
		<w:lsdException w:name="Grid Table 7 Colorful Accent 4" w:uiPriority="52"/>
		<w:lsdException w:name="Grid Table 1 Light Accent 5" w:uiPriority="46"/>
		<w:lsdException w:name="Grid Table 2 Accent 5" w:uiPriority="47"/>
		<w:lsdException w:name="Grid Table 3 Accent 5" w:uiPriority="48"/>
		<w:lsdException w:name="Grid Table 4 Accent 5" w:uiPriority="49"/>
		<w:lsdException w:name="Grid Table 5 Dark Accent 5" w:uiPriority="50"/>
		<w:lsdException w:name="Grid Table 6 Colorful Accent 5" w:uiPriority="51"/>
		<w:lsdException w:name="Grid Table 7 Colorful Accent 5" w:uiPriority="52"/>
		<w:lsdException w:name="Grid Table 1 Light Accent 6" w:uiPriority="46"/>
		<w:lsdException w:name="Grid Table 2 Accent 6" w:uiPriority="47"/>
		<w:lsdException w:name="Grid Table 3 Accent 6" w:uiPriority="48"/>
		<w:lsdException w:name="Grid Table 4 Accent 6" w:uiPriority="49"/>
		<w:lsdException w:name="Grid Table 5 Dark Accent 6" w:uiPriority="50"/>
		<w:lsdException w:name="Grid Table 6 Colorful Accent 6" w:uiPriority="51"/>
		<w:lsdException w:name="Grid Table 7 Colorful Accent 6" w:uiPriority="52"/>
		<w:lsdException w:name="List Table 1 Light" w:uiPriority="46"/>
		<w:lsdException w:name="List Table 2" w:uiPriority="47"/>
		<w:lsdException w:name="List Table 3" w:uiPriority="48"/>
		<w:lsdException w:name="List Table 4" w:uiPriority="49"/>
		<w:lsdException w:name="List Table 5 Dark" w:uiPriority="50"/>
		<w:lsdException w:name="List Table 6 Colorful" w:uiPriority="51"/>
		<w:lsdException w:name="List Table 7 Colorful" w:uiPriority="52"/>
		<w:lsdException w:name="List Table 1 Light Accent 1" w:uiPriority="46"/>
		<w:lsdException w:name="List Table 2 Accent 1" w:uiPriority="47"/>
		<w:lsdException w:name="List Table 3 Accent 1" w:uiPriority="48"/>
		<w:lsdException w:name="List Table 4 Accent 1" w:uiPriority="49"/>
		<w:lsdException w:name="List Table 5 Dark Accent 1" w:uiPriority="50"/>
		<w:lsdException w:name="List Table 6 Colorful Accent 1" w:uiPriority="51"/>
		<w:lsdException w:name="List Table 7 Colorful Accent 1" w:uiPriority="52"/>
		<w:lsdException w:name="List Table 1 Light Accent 2" w:uiPriority="46"/>
		<w:lsdException w:name="List Table 2 Accent 2" w:uiPriority="47"/>
		<w:lsdException w:name="List Table 3 Accent 2" w:uiPriority="48"/>
		<w:lsdException w:name="List Table 4 Accent 2" w:uiPriority="49"/>
		<w:lsdException w:name="List Table 5 Dark Accent 2" w:uiPriority="50"/>
		<w:lsdException w:name="List Table 6 Colorful Accent 2" w:uiPriority="51"/>
		<w:lsdException w:name="List Table 7 Colorful Accent 2" w:uiPriority="52"/>
		<w:lsdException w:name="List Table 1 Light Accent 3" w:uiPriority="46"/>
		<w:lsdException w:name="List Table 2 Accent 3" w:uiPriority="47"/>
		<w:lsdException w:name="List Table 3 Accent 3" w:uiPriority="48"/>
		<w:lsdException w:name="List Table 4 Accent 3" w:uiPriority="49"/>
		<w:lsdException w:name="List Table 5 Dark Accent 3" w:uiPriority="50"/>
		<w:lsdException w:name="List Table 6 Colorful Accent 3" w:uiPriority="51"/>
		<w:lsdException w:name="List Table 7 Colorful Accent 3" w:uiPriority="52"/>
		<w:lsdException w:name="List Table 1 Light Accent 4" w:uiPriority="46"/>
		<w:lsdException w:name="List Table 2 Accent 4" w:uiPriority="47"/>
		<w:lsdException w:name="List Table 3 Accent 4" w:uiPriority="48"/>
		<w:lsdException w:name="List Table 4 Accent 4" w:uiPriority="49"/>
		<w:lsdException w:name="List Table 5 Dark Accent 4" w:uiPriority="50"/>
		<w:lsdException w:name="List Table 6 Colorful Accent 4" w:uiPriority="51"/>
		<w:lsdException w:name="List Table 7 Colorful Accent 4" w:uiPriority="52"/>
		<w:lsdException w:name="List Table 1 Light Accent 5" w:uiPriority="46"/>
		<w:lsdException w:name="List Table 2 Accent 5" w:uiPriority="47"/>
		<w:lsdException w:name="List Table 3 Accent 5" w:uiPriority="48"/>
		<w:lsdException w:name="List Table 4 Accent 5" w:uiPriority="49"/>
		<w:lsdException w:name="List Table 5 Dark Accent 5" w:uiPriority="50"/>
		<w:lsdException w:name="List Table 6 Colorful Accent 5" w:uiPriority="51"/>
		<w:lsdException w:name="List Table 7 Colorful Accent 5" w:uiPriority="52"/>
		<w:lsdException w:name="List Table 1 Light Accent 6" w:uiPriority="46"/>
		<w:lsdException w:name="List Table 2 Accent 6" w:uiPriority="47"/>
		<w:lsdException w:name="List Table 3 Accent 6" w:uiPriority="48"/>
		<w:lsdException w:name="List Table 4 Accent 6" w:uiPriority="49"/>
		<w:lsdException w:name="List Table 5 Dark Accent 6" w:uiPriority="50"/>
		<w:lsdException w:name="List Table 6 Colorful Accent 6" w:uiPriority="51"/>
		<w:lsdException w:name="List Table 7 Colorful Accent 6" w:uiPriority="52"/>
		<w:lsdException w:name="Mention" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Smart Hyperlink" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Hashtag" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Unresolved Mention" w:semiHidden="1" w:unhideWhenUsed="1"/>
		<w:lsdException w:name="Smart Link" w:semiHidden="1" w:unhideWhenUsed="1"/>
	</w:latentStyles>
	<w:style w:type="paragraph" w:default="1" w:styleId="Normal">
		<w:name w:val="Normal"/>
		<w:qFormat/>
	</w:style>
	<w:style w:type="character" w:default="1" w:styleId="DefaultParagraphFont">
		<w:name w:val="Default Paragraph Font"/>
		<w:uiPriority w:val="1"/>
		<w:semiHidden/>
		<w:unhideWhenUsed/>
	</w:style>
	<w:style w:type="table" w:default="1" w:styleId="TableNormal">
		<w:name w:val="Normal Table"/>
		<w:uiPriority w:val="99"/>
		<w:semiHidden/>
		<w:unhideWhenUsed/>
		<w:tblPr>
			<w:tblInd w:w="0" w:type="dxa"/>
			<w:tblCellMar>
				<w:top w:w="0" w:type="dxa"/>
				<w:left w:w="108" w:type="dxa"/>
				<w:bottom w:w="0" w:type="dxa"/>
				<w:right w:w="108" w:type="dxa"/>
			</w:tblCellMar>
		</w:tblPr>
	</w:style>
	<w:style w:type="numbering" w:default="1" w:styleId="NoList">
		<w:name w:val="No List"/>
		<w:uiPriority w:val="99"/>
		<w:semiHidden/>
		<w:unhideWhenUsed/>
	</w:style>
	<w:style w:type="paragraph" w:styleId="Header">
		<w:name w:val="header"/>
		<w:basedOn w:val="Normal"/>
		<w:link w:val="HeaderChar"/>
		<w:uiPriority w:val="99"/>
		<w:unhideWhenUsed/>
		<w:rsid w:val="00FC21CA"/>
		<w:pPr>
			<w:tabs>
				<w:tab w:val="center" w:pos="4680"/>
				<w:tab w:val="right" w:pos="9360"/>
			</w:tabs>
			<w:spacing w:after="0" w:line="240" w:lineRule="auto"/>
		</w:pPr>
	</w:style>
	<w:style w:type="character" w:customStyle="1" w:styleId="HeaderChar">
		<w:name w:val="Header Char"/>
		<w:basedOn w:val="DefaultParagraphFont"/>
		<w:link w:val="Header"/>
		<w:uiPriority w:val="99"/>
		<w:rsid w:val="00FC21CA"/>
	</w:style>
	<w:style w:type="paragraph" w:styleId="Footer">
		<w:name w:val="footer"/>
		<w:basedOn w:val="Normal"/>
		<w:link w:val="FooterChar"/>
		<w:uiPriority w:val="99"/>
		<w:unhideWhenUsed/>
		<w:rsid w:val="00FC21CA"/>
		<w:pPr>
			<w:tabs>
				<w:tab w:val="center" w:pos="4680"/>
				<w:tab w:val="right" w:pos="9360"/>
			</w:tabs>
			<w:spacing w:after="0" w:line="240" w:lineRule="auto"/>
		</w:pPr>
	</w:style>
	<w:style w:type="character" w:customStyle="1" w:styleId="FooterChar">
		<w:name w:val="Footer Char"/>
		<w:basedOn w:val="DefaultParagraphFont"/>
		<w:link w:val="Footer"/>
		<w:uiPriority w:val="99"/>
		<w:rsid w:val="00FC21CA"/>
	</w:style>
</w:styles>')
  
  nm <- file.path(pth, "word/styles.xml")
  
  f <- file(nm, open="w", encoding = "native.enc")
  
  writeLines(cnt, f,  useBytes = TRUE) 
  
  close(f)
  
}


create_app <- function(pth) {
  
  cnt <- '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n
<Properties
	xmlns="http://schemas.openxmlformats.org/officeDocument/2006/extended-properties"
	xmlns:vt="http://schemas.openxmlformats.org/officeDocument/2006/docPropsVTypes">
	<Template>Normal.dotm</Template>
	<TotalTime>1</TotalTime>
	<Pages>1</Pages>
	<Words>2</Words>
	<Characters>16</Characters>
	<Application>Microsoft Office Word</Application>
	<DocSecurity>0</DocSecurity>
	<Lines>1</Lines>
	<Paragraphs>1</Paragraphs>
	<ScaleCrop>false</ScaleCrop>
	<Company></Company>
	<LinksUpToDate>false</LinksUpToDate>
	<CharactersWithSpaces>17</CharactersWithSpaces>
	<SharedDoc>false</SharedDoc>
	<HyperlinksChanged>false</HyperlinksChanged>
	<AppVersion>16.0000</AppVersion>
</Properties>'
  
  nm <- file.path(pth, "docProps/app.xml")
  
  f <- file(nm, open="w", encoding = "native.enc")
  
  writeLines(cnt, f,  useBytes = TRUE) 
  
  close(f)
  
}


create_core <- function(pth) {
  
  

  ts <- format(Sys.time(), format = "%Y-%m-%dT%H:%M:%SZ")
  
  cnt <- paste0('<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n
<cp:coreProperties
	xmlns:cp="http://schemas.openxmlformats.org/package/2006/metadata/core-properties"
	xmlns:dc="http://purl.org/dc/elements/1.1/"
	xmlns:dcterms="http://purl.org/dc/terms/"
	xmlns:dcmitype="http://purl.org/dc/dcmitype/"
	xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
	<dc:title></dc:title>
	<dc:subject></dc:subject>
	<dc:creator>',  Sys.info()[["user"]], '</dc:creator>
	<cp:keywords></cp:keywords>
	<dc:description></dc:description>
	<cp:lastModifiedBy>',  Sys.info()[["user"]], '</cp:lastModifiedBy>
	<cp:revision>1</cp:revision>
	<dcterms:created xsi:type="dcterms:W3CDTF">', ts, '</dcterms:created>
	<dcterms:modified xsi:type="dcterms:W3CDTF">', ts, '</dcterms:modified>
</cp:coreProperties>')
  
  nm <- file.path(pth, "docProps/core.xml")
  
  f <- file(nm, open="w", encoding = "native.enc")
  
  writeLines(cnt, f,  useBytes = TRUE) 
  
  close(f)
  
}

# create_document <- function(pth) {
#   
#  cnt <- '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n
# <w:document
# 	xmlns:wpc="http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas"
# 	xmlns:cx="http://schemas.microsoft.com/office/drawing/2014/chartex"
# 	xmlns:cx1="http://schemas.microsoft.com/office/drawing/2015/9/8/chartex"
# 	xmlns:cx2="http://schemas.microsoft.com/office/drawing/2015/10/21/chartex"
# 	xmlns:cx3="http://schemas.microsoft.com/office/drawing/2016/5/9/chartex"
# 	xmlns:cx4="http://schemas.microsoft.com/office/drawing/2016/5/10/chartex"
# 	xmlns:cx5="http://schemas.microsoft.com/office/drawing/2016/5/11/chartex"
# 	xmlns:cx6="http://schemas.microsoft.com/office/drawing/2016/5/12/chartex"
# 	xmlns:cx7="http://schemas.microsoft.com/office/drawing/2016/5/13/chartex"
# 	xmlns:cx8="http://schemas.microsoft.com/office/drawing/2016/5/14/chartex"
# 	xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006"
# 	xmlns:aink="http://schemas.microsoft.com/office/drawing/2016/ink"
# 	xmlns:am3d="http://schemas.microsoft.com/office/drawing/2017/model3d"
# 	xmlns:o="urn:schemas-microsoft-com:office:office"
# 	xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships"
# 	xmlns:m="http://schemas.openxmlformats.org/officeDocument/2006/math"
# 	xmlns:v="urn:schemas-microsoft-com:vml"
# 	xmlns:wp14="http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing"
# 	xmlns:wp="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing"
# 	xmlns:w10="urn:schemas-microsoft-com:office:word"
# 	xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main"
# 	xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml"
# 	xmlns:w15="http://schemas.microsoft.com/office/word/2012/wordml"
# 	xmlns:w16cex="http://schemas.microsoft.com/office/word/2018/wordml/cex"
# 	xmlns:w16cid="http://schemas.microsoft.com/office/word/2016/wordml/cid"
# 	xmlns:w16="http://schemas.microsoft.com/office/word/2018/wordml"
# 	xmlns:w16sdtdh="http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash"
# 	xmlns:w16se="http://schemas.microsoft.com/office/word/2015/wordml/symex"
# 	xmlns:wpg="http://schemas.microsoft.com/office/word/2010/wordprocessingGroup"
# 	xmlns:wpi="http://schemas.microsoft.com/office/word/2010/wordprocessingInk"
# 	xmlns:wne="http://schemas.microsoft.com/office/word/2006/wordml"
# 	xmlns:wps="http://schemas.microsoft.com/office/word/2010/wordprocessingShape" 
# 	mc:Ignorable="w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14">
# 	<w:body>
# 		<w:p w14:paraId="1BC7DC64" w14:textId="00AA6D14" w:rsidR="00046D2A" 
# 		w:rsidRPr="00444C49" w:rsidRDefault="00BC1857">
# 			<w:pPr>
# 				<w:rPr>
# 					<w:rFonts w:ascii="Arial" w:hAnsi="Arial" w:cs="Arial"/>
# 				</w:rPr>
# 			</w:pPr>
# 			<w:r w:rsidRPr="00444C49">
# 				<w:rPr>
# 					<w:rFonts w:ascii="Arial" w:hAnsi="Arial" w:cs="Arial"/>
# 				</w:rPr>
# 				<w:t>Fork</w:t>
# 			</w:r>
# 		</w:p>
# 		<w:p w14:paraId="25273E55" w14:textId="6D8F00D8" w:rsidR="00444C49" 
# 		w:rsidRDefault="00444C49"/>
# 		<w:p w14:paraId="22204C21" w14:textId="4434A694" w:rsidR="00444C49" 
# 		w:rsidRPr="00444C49" w:rsidRDefault="00444C49">
# 			<w:pPr>
# 				<w:rPr>
# 					<w:rFonts w:ascii="Times New Roman" w:hAnsi="Times New Roman" 
# 					w:cs="Times New Roman"/>
# 				</w:rPr>
# 			</w:pPr>
# 			<w:r w:rsidRPr="00444C49">
# 				<w:rPr>
# 					<w:rFonts w:ascii="Times New Roman" w:hAnsi="Times New Roman" 
# 					w:cs="Times New Roman"/>
# 				</w:rPr>
# 				<w:t>Bork</w:t>
# 			</w:r>
# 		</w:p>
# 		<w:p w14:paraId="32E76245" w14:textId="71B1B71F" w:rsidR="00444C49" 
# 		w:rsidRDefault="00444C49"/>
# 		<w:p w14:paraId="49CB1A18" w14:textId="256FDEDD" w:rsidR="00444C49" 
# 		w:rsidRPr="00444C49" w:rsidRDefault="00444C49">
# 			<w:pPr>
# 				<w:rPr>
# 					<w:rFonts w:ascii="Courier New" w:hAnsi="Courier New" w:cs="Courier New"/>
# 				</w:rPr>
# 			</w:pPr>
# 			<w:r w:rsidRPr="00444C49">
# 				<w:rPr>
# 					<w:rFonts w:ascii="Courier New" w:hAnsi="Courier New" w:cs="Courier New"/>
# 				</w:rPr>
# 				<w:t>Spork</w:t>
# 			</w:r>
# 		</w:p>
# 		<w:sectPr w:rsidR="00444C49" w:rsidRPr="00444C49">
# 			<w:headerReference w:type="default" r:id="rId6"/>
# 			<w:footerReference w:type="default" r:id="rId7"/>
# 			<w:pgSz w:w="12240" w:h="15840"/>
# 			<w:pgMar w:top="1440" w:right="1440" w:bottom="1440" w:left="1440" 
# 			w:header="720" w:footer="720" w:gutter="0"/>
# 			<w:cols w:space="720"/>
# 			<w:docGrid w:linePitch="360"/>
# 		</w:sectPr>
# 	</w:body>
# </w:document>'
#  
#  nm <- file.path(pth, "word/document.xml")
#  
#  f <- file(nm, open="w", encoding = "native.enc")
#  
#  writeLines(cnt, f,  useBytes = TRUE) 
#  
#  close(f)
#  
# }

create_header <- function(pth, cnt = "") {
 
  hdr <- '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<w:hdr
	xmlns:wpc="http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas"
	xmlns:cx="http://schemas.microsoft.com/office/drawing/2014/chartex"
	xmlns:cx1="http://schemas.microsoft.com/office/drawing/2015/9/8/chartex"
	xmlns:cx2="http://schemas.microsoft.com/office/drawing/2015/10/21/chartex"
	xmlns:cx3="http://schemas.microsoft.com/office/drawing/2016/5/9/chartex"
	xmlns:cx4="http://schemas.microsoft.com/office/drawing/2016/5/10/chartex"
	xmlns:cx5="http://schemas.microsoft.com/office/drawing/2016/5/11/chartex"
	xmlns:cx6="http://schemas.microsoft.com/office/drawing/2016/5/12/chartex"
	xmlns:cx7="http://schemas.microsoft.com/office/drawing/2016/5/13/chartex"
	xmlns:cx8="http://schemas.microsoft.com/office/drawing/2016/5/14/chartex"
	xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006"
	xmlns:aink="http://schemas.microsoft.com/office/drawing/2016/ink"
	xmlns:am3d="http://schemas.microsoft.com/office/drawing/2017/model3d"
	xmlns:o="urn:schemas-microsoft-com:office:office"
	xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships"
	xmlns:m="http://schemas.openxmlformats.org/officeDocument/2006/math"
	xmlns:v="urn:schemas-microsoft-com:vml"
	xmlns:wp14="http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing"
	xmlns:wp="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing"
	xmlns:w10="urn:schemas-microsoft-com:office:word"
	xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main"
	xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml"
	xmlns:w15="http://schemas.microsoft.com/office/word/2012/wordml"
	xmlns:w16cex="http://schemas.microsoft.com/office/word/2018/wordml/cex"
	xmlns:w16cid="http://schemas.microsoft.com/office/word/2016/wordml/cid"
	xmlns:w16="http://schemas.microsoft.com/office/word/2018/wordml"
	xmlns:w16sdtdh="http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash"
	xmlns:w16se="http://schemas.microsoft.com/office/word/2015/wordml/symex"
	xmlns:wpg="http://schemas.microsoft.com/office/word/2010/wordprocessingGroup"
	xmlns:wpi="http://schemas.microsoft.com/office/word/2010/wordprocessingInk"
	xmlns:wne="http://schemas.microsoft.com/office/word/2006/wordml"
	xmlns:wps="http://schemas.microsoft.com/office/word/2010/wordprocessingShape" 
	mc:Ignorable="w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14">\n'
  

  
  nm <- file.path(pth, "word/header1.xml")
  
  f <- file(nm, open="w", encoding = "native.enc")
  
  writeLines(paste0(hdr, cnt, '\n</w:hdr>\n', collapse = ""), f,  useBytes = TRUE) 
  
  close(f)
  
  
}

# 
# create_footer <- function(pth) {
#   
#   cnt <- '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n
# <w:ftr
# 	xmlns:wpc="http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas"
# 	xmlns:cx="http://schemas.microsoft.com/office/drawing/2014/chartex"
# 	xmlns:cx1="http://schemas.microsoft.com/office/drawing/2015/9/8/chartex"
# 	xmlns:cx2="http://schemas.microsoft.com/office/drawing/2015/10/21/chartex"
# 	xmlns:cx3="http://schemas.microsoft.com/office/drawing/2016/5/9/chartex"
# 	xmlns:cx4="http://schemas.microsoft.com/office/drawing/2016/5/10/chartex"
# 	xmlns:cx5="http://schemas.microsoft.com/office/drawing/2016/5/11/chartex"
# 	xmlns:cx6="http://schemas.microsoft.com/office/drawing/2016/5/12/chartex"
# 	xmlns:cx7="http://schemas.microsoft.com/office/drawing/2016/5/13/chartex"
# 	xmlns:cx8="http://schemas.microsoft.com/office/drawing/2016/5/14/chartex"
# 	xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006"
# 	xmlns:aink="http://schemas.microsoft.com/office/drawing/2016/ink"
# 	xmlns:am3d="http://schemas.microsoft.com/office/drawing/2017/model3d"
# 	xmlns:o="urn:schemas-microsoft-com:office:office"
# 	xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships"
# 	xmlns:m="http://schemas.openxmlformats.org/officeDocument/2006/math"
# 	xmlns:v="urn:schemas-microsoft-com:vml"
# 	xmlns:wp14="http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing"
# 	xmlns:wp="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing"
# 	xmlns:w10="urn:schemas-microsoft-com:office:word"
# 	xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main"
# 	xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml"
# 	xmlns:w15="http://schemas.microsoft.com/office/word/2012/wordml"
# 	xmlns:w16cex="http://schemas.microsoft.com/office/word/2018/wordml/cex"
# 	xmlns:w16cid="http://schemas.microsoft.com/office/word/2016/wordml/cid"
# 	xmlns:w16="http://schemas.microsoft.com/office/word/2018/wordml"
# 	xmlns:w16sdtdh="http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash"
# 	xmlns:w16se="http://schemas.microsoft.com/office/word/2015/wordml/symex"
# 	xmlns:wpg="http://schemas.microsoft.com/office/word/2010/wordprocessingGroup"
# 	xmlns:wpi="http://schemas.microsoft.com/office/word/2010/wordprocessingInk"
# 	xmlns:wne="http://schemas.microsoft.com/office/word/2006/wordml"
# 	xmlns:wps="http://schemas.microsoft.com/office/word/2010/wordprocessingShape" 
# 	mc:Ignorable="w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14">
# 	<w:p w14:paraId="65A3354A" w14:textId="3C459A2F" w:rsidR="00FC21CA" 
# 	w:rsidRDefault="00FC21CA">
# 		<w:pPr>
# 			<w:pStyle w:val="Footer"/>
# 		</w:pPr>
# 		<w:r>
# 			<w:t>Footer</w:t>
# 		</w:r>
# 	</w:p>
# </w:ftr>' 
#   
#   
#   nm <- file.path(pth, "word/footer1.xml")
#   
#   f <- file(nm, open="w", encoding = "native.enc")
#   
#   writeLines(cnt, f,  useBytes = TRUE) 
#   
#   close(f)
# }

create_footer <- function(pth, cnt = "") {
 
  hdr <- '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n
<w:ftr
	xmlns:wpc="http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas"
	xmlns:cx="http://schemas.microsoft.com/office/drawing/2014/chartex"
	xmlns:cx1="http://schemas.microsoft.com/office/drawing/2015/9/8/chartex"
	xmlns:cx2="http://schemas.microsoft.com/office/drawing/2015/10/21/chartex"
	xmlns:cx3="http://schemas.microsoft.com/office/drawing/2016/5/9/chartex"
	xmlns:cx4="http://schemas.microsoft.com/office/drawing/2016/5/10/chartex"
	xmlns:cx5="http://schemas.microsoft.com/office/drawing/2016/5/11/chartex"
	xmlns:cx6="http://schemas.microsoft.com/office/drawing/2016/5/12/chartex"
	xmlns:cx7="http://schemas.microsoft.com/office/drawing/2016/5/13/chartex"
	xmlns:cx8="http://schemas.microsoft.com/office/drawing/2016/5/14/chartex"
	xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006"
	xmlns:aink="http://schemas.microsoft.com/office/drawing/2016/ink"
	xmlns:am3d="http://schemas.microsoft.com/office/drawing/2017/model3d"
	xmlns:o="urn:schemas-microsoft-com:office:office"
	xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships"
	xmlns:m="http://schemas.openxmlformats.org/officeDocument/2006/math"
	xmlns:v="urn:schemas-microsoft-com:vml"
	xmlns:wp14="http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing"
	xmlns:wp="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing"
	xmlns:w10="urn:schemas-microsoft-com:office:word"
	xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main"
	xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml"
	xmlns:w15="http://schemas.microsoft.com/office/word/2012/wordml"
	xmlns:w16cex="http://schemas.microsoft.com/office/word/2018/wordml/cex"
	xmlns:w16cid="http://schemas.microsoft.com/office/word/2016/wordml/cid"
	xmlns:w16="http://schemas.microsoft.com/office/word/2018/wordml"
	xmlns:w16sdtdh="http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash"
	xmlns:w16se="http://schemas.microsoft.com/office/word/2015/wordml/symex"
	xmlns:wpg="http://schemas.microsoft.com/office/word/2010/wordprocessingGroup"
	xmlns:wpi="http://schemas.microsoft.com/office/word/2010/wordprocessingInk"
	xmlns:wne="http://schemas.microsoft.com/office/word/2006/wordml"
	xmlns:wps="http://schemas.microsoft.com/office/word/2010/wordprocessingShape" 
	mc:Ignorable="w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14">' 
  
  
  nm <- file.path(pth, "word/footer1.xml")
  
  f <- file(nm, open="w", encoding = "native.enc")
  
  writeLines(paste0(hdr, cnt, '\n</w:ftr>\n', collapse = ""), f,  useBytes = TRUE) 
  
  close(f)
}


create_document_rels <- function(pth, imgCnt, imagePaths) {
  
  imgs <- ""
  
  for (i in seq_len(imgCnt)) {
    ext <- tools::file_ext(imagePaths[[i]])
    if (ext == "jpg") {
      ext <- "jpeg" 
    }
      
    imgs <- paste0(imgs, '<Relationship Target="media/image', i, '.', ext, '"', 
' Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/image"',
 ' Id="rId', 8 + i, '"/>\n')
    
  }
  
 
  cnt <- paste0('<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n
<Relationships
	xmlns="http://schemas.openxmlformats.org/package/2006/relationships">
	<Relationship Id="rId8" 
	Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/fontTable" 
	Target="fontTable.xml"/>
	<Relationship Id="rId3" 
	Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/webSettings" 
	Target="webSettings.xml"/>
	<Relationship Id="rId7" 
	Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/footer" 
	Target="footer1.xml"/>
	<Relationship Id="rId2" 
	Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/settings" 
	Target="settings.xml"/>
	<Relationship Id="rId1" 
	Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/styles" 
	Target="styles.xml"/>
	<Relationship Id="rId6" 
	Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/header" 
	Target="header1.xml"/>
	<Relationship Id="rId5" 
	Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/endnotes" 
	Target="endnotes.xml"/>
	<Relationship Id="rId4" 
	Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/footnotes" 
	Target="footnotes.xml"/>', imgs, '</Relationships>')
  
  
  nm <- file.path(pth, "word/_rels/document.xml.rels")
  
  f <- file(nm, open="w", encoding = "native.enc")
  
  writeLines(cnt, f,  useBytes = TRUE) 
  
  close(f)
}

create_footer_rels <- function(pth, imgCnt, imagePaths) {
  
  imgs <- ""
  
  for (i in seq_len(imgCnt)) {
    ext <- tools::file_ext(imagePaths[[i]])
    if (ext == "jpg") {
      ext <- "jpeg" 
    }
    
    imgs <- paste0(imgs, '<Relationship Target="media/imagef', i, '.', ext, '"', 
                   ' Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/image"',
                   ' Id="rIdf', i, '"/>\n')
    
  }
  
  
  cnt <- paste0('<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n
<Relationships
	xmlns="http://schemas.openxmlformats.org/package/2006/relationships">', imgs, '</Relationships>')
  
  
  nm <- file.path(pth, "word/_rels/footer1.xml.rels")
  
  f <- file(nm, open="w", encoding = "native.enc")
  
  writeLines(cnt, f,  useBytes = TRUE) 
  
  close(f)
}

create_header_rels <- function(pth, imgCnt, imagePaths) {
  
  imgs <- ""
  
  for (i in seq_len(imgCnt)) {
    ext <- tools::file_ext(imagePaths[[i]])
    if (ext == "jpg") {
      ext <- "jpeg" 
    }
    
    imgs <- paste0(imgs, '<Relationship Target="media/imageh', i, '.', ext, '"', 
                   ' Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/image"',
                   ' Id="rIdh', i, '"/>\n')
    
  }
  
  
  cnt <- paste0('<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n
<Relationships
	xmlns="http://schemas.openxmlformats.org/package/2006/relationships">', imgs, '</Relationships>')
  
  
  nm <- file.path(pth, "word/_rels/header1.xml.rels")
  
  f <- file(nm, open="w", encoding = "native.enc")
  
  writeLines(cnt, f,  useBytes = TRUE) 
  
  close(f)
}

create_rels <- function(pth) {
 
  cnt <- '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<Relationships
	xmlns="http://schemas.openxmlformats.org/package/2006/relationships">
	<Relationship Id="rId3" 
	Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/extended-properties" 
	Target="docProps/app.xml"/>
	<Relationship Id="rId2" 
	Type="http://schemas.openxmlformats.org/package/2006/relationships/metadata/core-properties" 
	Target="docProps/core.xml"/>
	<Relationship Id="rId1" 
	Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/officeDocument" 
	Target="word/document.xml"/>
</Relationships>' 
  
  
  nm <- file.path(pth, "_rels/.rels")
  
  f <- file(nm, open="w", encoding = "native.enc")
  
  writeLines(cnt, f,  useBytes = TRUE) 
  
  close(f)
  
}

create_endnotes <- function(pth) {
 
  cnt <- '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<w:endnotes
	xmlns:wpc="http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas"
	xmlns:cx="http://schemas.microsoft.com/office/drawing/2014/chartex"
	xmlns:cx1="http://schemas.microsoft.com/office/drawing/2015/9/8/chartex"
	xmlns:cx2="http://schemas.microsoft.com/office/drawing/2015/10/21/chartex"
	xmlns:cx3="http://schemas.microsoft.com/office/drawing/2016/5/9/chartex"
	xmlns:cx4="http://schemas.microsoft.com/office/drawing/2016/5/10/chartex"
	xmlns:cx5="http://schemas.microsoft.com/office/drawing/2016/5/11/chartex"
	xmlns:cx6="http://schemas.microsoft.com/office/drawing/2016/5/12/chartex"
	xmlns:cx7="http://schemas.microsoft.com/office/drawing/2016/5/13/chartex"
	xmlns:cx8="http://schemas.microsoft.com/office/drawing/2016/5/14/chartex"
	xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006"
	xmlns:aink="http://schemas.microsoft.com/office/drawing/2016/ink"
	xmlns:am3d="http://schemas.microsoft.com/office/drawing/2017/model3d"
	xmlns:o="urn:schemas-microsoft-com:office:office"
	xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships"
	xmlns:m="http://schemas.openxmlformats.org/officeDocument/2006/math"
	xmlns:v="urn:schemas-microsoft-com:vml"
	xmlns:wp14="http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing"
	xmlns:wp="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing"
	xmlns:w10="urn:schemas-microsoft-com:office:word"
	xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main"
	xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml"
	xmlns:w15="http://schemas.microsoft.com/office/word/2012/wordml"
	xmlns:w16cex="http://schemas.microsoft.com/office/word/2018/wordml/cex"
	xmlns:w16cid="http://schemas.microsoft.com/office/word/2016/wordml/cid"
	xmlns:w16="http://schemas.microsoft.com/office/word/2018/wordml"
	xmlns:w16sdtdh="http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash"
	xmlns:w16se="http://schemas.microsoft.com/office/word/2015/wordml/symex"
	xmlns:wpg="http://schemas.microsoft.com/office/word/2010/wordprocessingGroup"
	xmlns:wpi="http://schemas.microsoft.com/office/word/2010/wordprocessingInk"
	xmlns:wne="http://schemas.microsoft.com/office/word/2006/wordml"
	xmlns:wps="http://schemas.microsoft.com/office/word/2010/wordprocessingShape" 
	mc:Ignorable="w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14">
	<w:endnote w:type="separator" w:id="-1">
		<w:p w14:paraId="11DB1783" w14:textId="77777777" w:rsidR="00F10374" 
		w:rsidRDefault="00F10374" w:rsidP="00FC21CA">
			<w:pPr>
				<w:spacing w:after="0" w:line="240" w:lineRule="auto"/>
			</w:pPr>
			<w:r>
				<w:separator/>
			</w:r>
		</w:p>
	</w:endnote>
	<w:endnote w:type="continuationSeparator" w:id="0">
		<w:p w14:paraId="304CAF46" w14:textId="77777777" w:rsidR="00F10374" w:rsidRDefault="00F10374" w:rsidP="00FC21CA">
			<w:pPr>
				<w:spacing w:after="0" w:line="240" w:lineRule="auto"/>
			</w:pPr>
			<w:r>
				<w:continuationSeparator/>
			</w:r>
		</w:p>
	</w:endnote>
</w:endnotes>'
  
  
  nm <- file.path(pth, "word/endnotes.xml")
  
  f <- file(nm, open="w", encoding = "native.enc")
  
  writeLines(cnt, f,  useBytes = TRUE) 
  
  close(f)
  
}

create_footnotes <- function(pth) {
 
  cnt <- '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<w:footnotes
	xmlns:wpc="http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas"
	xmlns:cx="http://schemas.microsoft.com/office/drawing/2014/chartex"
	xmlns:cx1="http://schemas.microsoft.com/office/drawing/2015/9/8/chartex"
	xmlns:cx2="http://schemas.microsoft.com/office/drawing/2015/10/21/chartex"
	xmlns:cx3="http://schemas.microsoft.com/office/drawing/2016/5/9/chartex"
	xmlns:cx4="http://schemas.microsoft.com/office/drawing/2016/5/10/chartex"
	xmlns:cx5="http://schemas.microsoft.com/office/drawing/2016/5/11/chartex"
	xmlns:cx6="http://schemas.microsoft.com/office/drawing/2016/5/12/chartex"
	xmlns:cx7="http://schemas.microsoft.com/office/drawing/2016/5/13/chartex"
	xmlns:cx8="http://schemas.microsoft.com/office/drawing/2016/5/14/chartex"
	xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006"
	xmlns:aink="http://schemas.microsoft.com/office/drawing/2016/ink"
	xmlns:am3d="http://schemas.microsoft.com/office/drawing/2017/model3d"
	xmlns:o="urn:schemas-microsoft-com:office:office"
	xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships"
	xmlns:m="http://schemas.openxmlformats.org/officeDocument/2006/math"
	xmlns:v="urn:schemas-microsoft-com:vml"
	xmlns:wp14="http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing"
	xmlns:wp="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing"
	xmlns:w10="urn:schemas-microsoft-com:office:word"
	xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main"
	xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml"
	xmlns:w15="http://schemas.microsoft.com/office/word/2012/wordml"
	xmlns:w16cex="http://schemas.microsoft.com/office/word/2018/wordml/cex"
	xmlns:w16cid="http://schemas.microsoft.com/office/word/2016/wordml/cid"
	xmlns:w16="http://schemas.microsoft.com/office/word/2018/wordml"
	xmlns:w16sdtdh="http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash"
	xmlns:w16se="http://schemas.microsoft.com/office/word/2015/wordml/symex"
	xmlns:wpg="http://schemas.microsoft.com/office/word/2010/wordprocessingGroup"
	xmlns:wpi="http://schemas.microsoft.com/office/word/2010/wordprocessingInk"
	xmlns:wne="http://schemas.microsoft.com/office/word/2006/wordml"
	xmlns:wps="http://schemas.microsoft.com/office/word/2010/wordprocessingShape" 
	mc:Ignorable="w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14">
	<w:footnote w:type="separator" w:id="-1">
		<w:p w14:paraId="32C960DE" w14:textId="77777777" w:rsidR="00F10374" 
		w:rsidRDefault="00F10374" w:rsidP="00FC21CA">
			<w:pPr>
				<w:spacing w:after="0" w:line="240" w:lineRule="auto"/>
			</w:pPr>
			<w:r>
				<w:separator/>
			</w:r>
		</w:p>
	</w:footnote>
	<w:footnote w:type="continuationSeparator" w:id="0">
		<w:p w14:paraId="7E2E8024" w14:textId="77777777" w:rsidR="00F10374" w:rsidRDefault="00F10374" w:rsidP="00FC21CA">
			<w:pPr>
				<w:spacing w:after="0" w:line="240" w:lineRule="auto"/>
			</w:pPr>
			<w:r>
				<w:continuationSeparator/>
			</w:r>
		</w:p>
	</w:footnote>
</w:footnotes>' 
  
  
  
  nm <- file.path(pth, "word/footnotes.xml")
  
  f <- file(nm, open="w", encoding = "native.enc")
  
  writeLines(cnt, f,  useBytes = TRUE) 
  
  close(f)
}





# DOCX Construction -------------------------------------------------------


#' @noRd
cell_pct <- function(txt, align = "left", width = NULL) {
  
  if (is.null(width)) {
    ret <- paste0('<w:tc>', 
                  para(txt, align),
                  "</w:tc>\n", collapse = "")
    
  } else {
    ret <- paste0('<w:tc><w:tcPr><w:tcW w:w="', width,'" w:type="pct"/></w:tcPr>', 
           para(txt, align),
           "</w:tc>\n", collapse = "")
  }
  
  return(ret)
}

#' Bottom border means it is a header row
#' @noRd
cell_abs <- function(txt, align = "left", width = NULL, borders = NULL, valign = NULL, bold = FALSE,
                     multiple = FALSE) {
  
  al <- ""
  bb <- ""
  if (!is.null(borders)) {

    bb <- paste0('<w:tcBorders>',
                 ifelse(any(borders %in% "bottom"),  
                        '<w:bottom w:val="single" w:sz="4" w:space="0" w:color="auto"/>', ""),
                 ifelse(any(borders %in% "top"),
                        '<w:top w:val="single" w:sz="4" w:space="0" w:color="auto"/>', ""),
                 '</w:tcBorders>')
  }
  
  if (!is.null(valign)) {
    if (valign == "bottom") {
      al <- '<w:vAlign w:val="bottom"/>'
    }
  }
  
  # If multiple is TRUE, use para_wp to separate different text with different styles
  if (multiple == FALSE) {
    if (is.null(width)) {
      ret <- paste0('<w:tc><w:tcPr>', bb, al, '</w:tcPr>', 
                    para(txt, align),
                    "</w:tc>\n", collapse = "")
      
    } else {
      ret <- paste0('<w:tc><w:tcPr><w:tcW w:w="', width,'"/>', bb, al, '</w:tcPr>', 
                    para(txt, align, bold = bold),
                    "</w:tc>\n", collapse = "")
    }
  } else {
    if (is.null(width)) {
      ret <- paste0('<w:tc><w:tcPr>', bb, al, '</w:tcPr>', 
                    para_wp(txt, align),
                    "</w:tc>\n", collapse = "")
      
    } else {
      ret <- paste0('<w:tc><w:tcPr><w:tcW w:w="', width,'"/>', bb, al, '</w:tcPr>', 
                    para_wp(txt, align, bold = bold),
                    "</w:tc>\n", collapse = "")
    }
  }

  
  return(ret)
}


#' @noRd
para <- function(txt, align = "left", font_size = NULL, bold = FALSE, 
                 italics = FALSE, indent_left = NA, indent_right = NA, borders = "") {
  
  ret <- ""
  
  vl <- encodeDOCX(txt)
  if (nchar(txt) == 0)
    vl <- " "

  splt <- strsplit(vl, split = "\n", fixed = TRUE)
  
  for (i in seq_len(length(splt))) {
    
    
    for (j in seq_len(length(splt[[i]]))) {
    
      b <- ""
      if (bold == TRUE)
        b <-  '<w:b/><w:bCs/>'
      
      it <- ""
      if (italics == TRUE)
        it <- '<w:i/><w:iCs/>'
      
      fs <- ""
      if (!is.null(font_size))
        fs <- paste0('<w:sz w:val="', font_size * 2, 
               '"/><w:szCs w:val="', font_size * 2, '"/>')
      
      rpr <- ""
      if (!is.null(font_size) | bold == TRUE | italics == TRUE)
        rpr <- paste0('<w:rPr>', b, it, fs, '</w:rPr>')
            
            
      if (align == "centre")
        align <- "center"
      
      indent_c <- ""
      if (!is.na(indent_left) | !is.na(indent_right)) {
        indent_left <- ifelse(is.na(indent_left), 0, indent_left)
        indent_right <- ifelse(is.na(indent_right),0,indent_right)
        
        indent_c <- sprintf('<w:ind w:left="%s" w:right="%s"/>', indent_left, indent_right)
      }
      
      # Underline only added in last line
      borders_j <- ""
      if (grepl("w:bottom", borders)) {
        borders_j <- ifelse(j == length(splt[[i]]), borders, "")
      }
      
     ret <- paste0(ret, '<w:p>',
                   '<w:pPr><w:jc w:val="', align, '"/>',
                   '<w:spacing w:after="0"/>',
                   indent_c,
                   borders_j,
                   '</w:pPr>',
                   '<w:r>', rpr, '<w:t xml:space="preserve">', splt[[i]][j], 
                   '</w:t></w:r></w:p>\n')
    }
  }
 
 return(ret)
  
}

#' @noRd
para_wp <- function(txt, align = "left", font_size = NULL, bold = FALSE, 
                    italics = FALSE, indent_left = NA, indent_right = NA, borders = "") {
  
  ret <- ""
  
  if (align == "centre") {
    align <- "center"
  }
  
  indent_c <- ""
  if (!is.na(indent_left) | !is.na(indent_right)) {
    indent_left <- ifelse(is.na(indent_left), 0, indent_left)
    indent_right <- ifelse(is.na(indent_right),0,indent_right)
    
    indent_c <- sprintf('<w:ind w:left="%s" w:right="%s"/>', indent_left, indent_right)
  }
  
  # Process multiple <wr> tags
  wr <- para_wr(txt, font_size, bold, italics)
  
  ret <- paste0(ret, '<w:p>',
                '<w:pPr><w:jc w:val="', align, '"/>',
                '<w:spacing w:after="0"/>',
                indent_c,
                borders,
                '</w:pPr>',
                wr,
                "</w:p>\n")
  
  return(ret)
  
}

#' @noRd
para_wr <- function(txt, font_size, bold, italics){
  vl <- encodeDOCX(txt)
  
  if (all(nchar(txt) == 0)) {
    vl <- " "
  }
  
  splt <- strsplit(vl, split = "\n", fixed = TRUE)
  
  ret <- ""
  
  # 1st loop is for different text
  for (i in seq_len(length(splt))) {
    b <- ""
    if (length(bold) > 1){
      if (bold[i] == TRUE) {
        b <-  '<w:b/><w:bCs/>'
      }
    } else {
      if (bold == TRUE) {
        b <-  '<w:b/><w:bCs/>'
      }
    }
    
    it <- ""
    if (length(italics) > 1) {
      if (italics[i] == TRUE) {
        it <- '<w:i/><w:iCs/>'
      }
    } else {
      if (italics == TRUE) {
        it <- '<w:i/><w:iCs/>'
      }
    }
    
    
    fs <- ""
    if (length(font_size) > 1){
      if (!is.na(font_size[i])) {
        fs <- paste0('<w:sz w:val="', font_size[i] * 2, 
                     '"/><w:szCs w:val="', font_size[i] * 2, '"/>')
      }
    } else {
      if (!is.null(font_size)) {
        fs <- paste0('<w:sz w:val="', font_size[i] * 2, 
                     '"/><w:szCs w:val="', font_size[i] * 2, '"/>')
      }
    }
    
    
    rpr <- ""
    if (any(!is.na(font_size)) | any(bold == TRUE) | any(italics == TRUE)) {
      rpr <- paste0('<w:rPr>', b, it, fs, '</w:rPr>')
    }
    
    # 2nd loop is for multiple line in same text
    for (j in seq_len(length(splt[[i]]))) {
      if (j == 1){
        wr <- paste0('<w:r>', rpr, '<w:t xml:space="preserve">', splt[[i]][j], 
                     '</w:t></w:r>')
      } else {
        line_break <- "<w:r><w:br/></w:r>"
        wr <- paste0(wr, line_break, '<w:r>', rpr, '<w:t xml:space="preserve">', splt[[i]][j], 
                     '</w:t></w:r>')
      }
    }
    
    if (substr(wr, 1, 1) != " " & substr(ret, nchar(ret), nchar(ret)) != " ") {
      sep <- " "
    } else {
      sep <- ""
    }
    ret <- paste0(ret, sep, wr)
  }
  
  return(ret)
}

#' @noRd
run <- function(txt) {
  
  ret <- paste0('<w:r><w:t xml:space="preserve">', encodeDOCX(txt), '</w:t></w:r>', collapse = "")
  
  return(ret)
  
}

#' @noRd
get_cell_borders_docx <- function(row, col, trow, tcol, brdrs, flg = NULL,
                                  cell_border = NULL) {
  
  ret <- ""
  r <- ""
  l <- ""
  b <- ""
  t <- ""
  
  # if (any(brdrs %in% c("bottom", "outside", "all", "body", "right", "left", "top"))) {
  #   
  #   if (row == 1 & any(brdrs %in% c("top", "outside", "body"))) {
  #     t <- '<w:top w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
  #   }
  #   
  #   if (row == trow & any(brdrs %in% c("bottom", "outside", "body"))) {
  #     b <- '<w:bottom w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
  #   }
  #   
  #   if (col == 1 & any(brdrs %in% c("left", "body"))) {
  #     l <- '<w:left w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
  #   }
  #   
  #   if (col == tcol & any(brdrs %in% c("right", "body"))) {
  #     r <- '<w:right w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
  #   }
  #   
  #   if (!is.null(flg)) {
  #     if (flg %in% c("L", "B", "A")) {
  #       if (col == 1 & any(brdrs %in% c("right", "body"))) {
  #         r <- '<w:right w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
  #       }
  #     }
  #   }
  #   
  #   if (b != "" | l != "" | r != "" | t!= "") {
  #             
  #     ret <-  paste0('<w:tcBorders>', r, l, t, b,
  #                    '</w:tcBorders>')
  #   }
  # }
  
  if ((row == 1 & any(brdrs %in% c("top", "outside", "body"))) |
      any(cell_border %in% c("top", "outside", "body"))
      ) {
    t <- '<w:top w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
  }
  
  if (row == trow & any(brdrs %in% c("bottom", "outside", "body")) |
      any(cell_border %in% c("bottom", "outside", "body"))
      ) {
    b <- '<w:bottom w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
  }
  
  if (col == 1 & any(brdrs %in% c("left", "body")) |
      any(cell_border %in% c("left", "body"))
      ) {
    l <- '<w:left w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
  }
  
  if (col == tcol & any(brdrs %in% c("right", "body")) |
      any(cell_border %in% c("right", "body"))
      ) {
    r <- '<w:right w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
  }
  
  if (!is.null(flg)) {
    if (flg %in% c("L", "B", "A")) {
      if (col == 1 & any(brdrs %in% c("right", "body"))) {
        r <- '<w:right w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
      }
    }
  }
  
  if (b != "" | l != "" | r != "" | t!= "") {
    
    ret <-  paste0('<w:tcBorders>', r, l, t, b,
                   '</w:tcBorders>')
  }

  
  return(ret)
}

context("write_docx Tests")

base_path <- "c:/packages/reporter/tests/testthat"
data_dir <- base_path

base_path <- tempdir()
data_dir <- "."



test_that("create_new_docx() function works as expected.", {


  
  p <- create_new_docx("Arial", 12, 0)


  expect_true(file.exists(p))
  
  expect_true(file.exists(file.path(p, "_rels")))
  
  expect_true(file.exists(file.path(p, "word")))
  
  expect_true(file.exists(file.path(p, "docProps")))
  
  expect_true(file.exists(file.path(p, "[Content_Types].xml")))
  
  expect_true(file.exists(file.path(p, "word/webSettings.xml")))
  
  expect_true(file.exists(file.path(p, "word/fontTable.xml")))
  
  expect_true(file.exists(file.path(p, "word/styles.xml")))
  
  expect_true(file.exists(file.path(p, "docProps/app.xml")))
  
  expect_true(file.exists(file.path(p, "docProps/core.xml")))
  
  expect_true(file.exists(file.path(p, "_rels/.rels")))
  
  expect_true(file.exists(file.path(p, "word/_rels/document.xml.rels")))
  
  expect_true(file.exists(file.path(p, "word/footnotes.xml")))
  
  expect_equal(file.exists(file.path(p, "word/endnotes.xml")), TRUE)
  
})



test_that("write_docx() function works as expected.", {
  
  p <- create_new_docx("Arial", 10, 0)
  
  fp <-  file.path(base_path, "docx/test.docx")
  

  write_docx(p, fp)
 
  
  expect_equal(file.exists(fp), TRUE)
  

  
})


test_that("get_docx_document() works as expected.", {
  
  fp <-  file.path(base_path, "docx/test1.docx")
  
  txt <- create_text("Fork")
  
  rpt <- create_report(fp, output_type = "DOCX", font = "Arial") %>%
    add_content(txt)
  
  rpt <- page_setup_docx(rpt)
  
  res <- get_docx_document(rpt)
  
  expect_true(length(res) > 0)
  
  
})


test_that("para() function works as expected.", {
  
  txt <- "here is \n a nice text string."
  
  
  res <- para(txt)
  
  res
  
  expect_equal(nchar(res) > 100, TRUE)
  
})

test_that("para() function works as expected with indentation.", {
  
  txt <- "here is \n a nice text string."
  
  
  res <- para(txt, indent_left = 360)
  
  res
  
  expect_true(
    any(grepl("<w:ind w:left=\"360\" w:right=\"0\"/></w:pPr><w:r><w:t xml:space=\"preserve\">here is ",
              res))
  )
  
  expect_equal(nchar(res) > 100, TRUE)
  
})
  

# Attribute Classes --------------------------------------------------------


#' @title Defines a spanning header attribute
#' @description Create a spanning header attribute object that can be
#' attached to a data frame and passed to the \code{\link{create_table}}
#' function.  This attribute is used internally by other packages in the 
#' \strong{r-sassy} system.  
#' @details 
#' A spanning header is a label and underline that spans one or more 
#' columns.  A spanning header is defined minimally by identifying the
#' column range to be spanned, and a label.  A label alignment 
#' may also be supplied.
#' 
#' The spanning column range is defined by the \code{from} and \code{to} 
#' parameters.  The range identifies a contiguous set of variables on the data.
#' Variables can be identified by position, a quoted variable name, or an 
#' unquoted variable name.
#' @param from The starting column to span.  Spanning columns are defined as
#' range of columns 'from' and 'to'. The columns may be identified by position, 
#' or by quoted or unquoted variable names. 
#' The \code{from} parameter is required.  
#' @param to The ending column to span.  Spanning columns are defined as
#' range of columns 'from' and 'to'. The columns may be identified by position,
#' or by quoted or unquoted variable names.  
#' The \code{to} parameter is required. 
#' @param label The label to apply to the spanning header.
#' @param label_align The alignment to use for the label. Valid values are 
#' "left", "right", "center", and "centre".  The default for spanning columns
#' is "center".
#' @param level The level to use for the spanning header.  The lowest
#' spanning level is level 1, the next level above is level 2, and so on.  
#' By default, the level is set to 1.
#' @param underline A TRUE or FALSE value indicating whether the spanning
#' header should be underlined.  Default is TRUE.  
#' @return The spanning header attribute object.
#' @family attributes
#' @export
span <- function(from, to, label = NULL, label_align = 'center',
                      level = 1, underline = TRUE) {
  
  # Create new structure of class "span_attr"
  x <- structure(list(), class = c("span_attr", "list"))
  x$label <- label
  x$from <- from
  x$to <-  to
  x$label_align <- label_align
  x$level <- level
  x$underline <- underline
  
  return(x)
}


#' @title
#' Creates a title attribute 
#' @description
#' This function creates a titles attribute, which can be used to
#' define titles by attaching as an attribute to a data frame.  This function is 
#' used internally to the system.
#' @param titles A vector of title strings.
#' @param align The position to align the titles.  Valid values are 'left', 
#' 'right', 'center' or 'centre'.  For titles, the default is 'center'.
#' @param blank_row Where to place a blank row.  Valid values are 'above',
#' 'below', 'both', or 'none'.  Default is "below".
#' @param borders Whether and where to place a border. Valid values are 'top',
#' 'bottom', 'left', 'right', 'outside', 'inside', 'all', or 'none'.  
#' Default is "none".  
#' @param width The width of the titles block.  
#' @param bold A parameter to bold the titles.  Valid values are TRUE and FALSE.
#' Default is FALSE.  
#' @param font_size The font size to use for the title block.  The font size
#' of the report will be used by default.  Valid values are 8, 9, 10, 11, 12,
#' 13, and 14.  
#' @return A title attribute object.
#' @seealso \code{\link{titles}} function.
#' @family attributes
#' @export
ttl <- function(titles, align = 'center',
                       blank_row = "below", borders = NULL, width = NULL,
                       bold = FALSE, font_size = NULL) {
  
  # Create new structure of class "title_attr"
  x <- structure(list(), class = c("ttl_attr", "list"))
  
  x$titles <- titles
  x$align <- align
  x$blank_row <- blank_row
  x$borders <- borders
  x$width <- width
  x$bold <- bold
  x$font_size <- font_size
  
  return(x)
}

#' @title
#' Creates a footnote attribute
#' @description
#' The \code{ftn} function creates a footnote attribute,
#' which may be attached as an attribute to a data frame and passed
#' into \code{\link{create_table}}. This function is used internally to
#' the system.
#' @param footnotes A vector of footnote strings.
#' @param align The position to align the footnotes.  Valid values are: 'left',
#' 'right', 'center', or 'centre'.
#' @param blank_row Whether to print a blank row above or below the footnote.
#' Valid values are 'above', 'below', 'both', or 'none'.  Default is 'above'.
#' @param borders Whether to print a border above or below the footnote. Valid
#' values are 'top', 'bottom', 'outside', 'inside', 'all',  or 'none'.  
#' Default is 'none'.  
#' @param valign The vertical position to align the footnotes.  Valid
#' values are: 'top' and 'bottom'.  
#' @param width The width of the footnotes block.  
#' @return The footnote attribute object.
#' @seealso \code{\link{footnotes}} to create a footnote.
#' @family attributes
#' @export
ftn <- function(footnotes, align = 'left',
                          blank_row = "above", borders = "none",
                          valign = NULL, width = NULL) {
  
  # Create new structure of class "ftn_attr"
  x <- structure(list(), class = c("ftn_attr", "list"))
  
  x$footnotes <- footnotes
  x$align <- align
  x$blank_row <- blank_row
  x$borders <- borders
  x$valign <- valign
  x$width <- width
  
  return(x)
}


# Process Attributes ------------------------------------------------------

apply_attributes <- function(x, obj) {
  
  
  ttls <- attr(obj, "titles", exact = TRUE)
  if (!is.null(ttls)) {
    if (is.list(ttls)) {
      for (i in seq_along(ttls)) {
        ttl <- ttls[[i]]
        if ("ttl_attr" %in% class(ttl)) {
          
          x <- titles(x, ttl$titles, align = ttl$align, borders = ttl$borders,
                      bold = ttl$bold, font_size = ttl$font_size, width = ttl$width,
                      blank_row = ttl$blank_row)
          
        }
      }
    }
  }
  
  ftns <- attr(obj, "footnotes", exact = TRUE)
  if (!is.null(ftns)) {
    if (is.list(ftns)) {
      for (i in seq_along(ftns)) {
        ftn <- ftns[[i]]
        if ("ftn_attr" %in% class(ftn)) {

          x <- footnotes(x, ftn$footnotes, align = ftn$align, borders = ftn$borders,
                      valign = ftn$valign, width = ftn$width,
                      blank_row = ftn$blank_row)

        }
      }
    }
  }

  spns <- attr(obj, "spans", exact = TRUE)
  if (!is.null(spns)) {
    if (is.list(spns)) {
      for (i in seq_along(spns)) {
        spn <- spns[[i]]
        if ("span_attr" %in% class(spn)) {

          x <- spanning_header(x, from = spn$from, to = spn$to, 
                               label_align = spn$label_align, 
                               label = spn$label,
                         level = spn$level, underline = spn$underline,
                         standard_eval = TRUE)

        }
      }
    }
  }
  
  return(x)
  
}

context("Attributes Tests")

base_path <- "c:/packages/reporter/tests/testthat"
data_dir <- base_path

base_path <- tempdir()
data_dir <- "."

cnt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
              "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
              "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
              "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ",
              "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
              "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
              "qui officia deserunt mollit anim id est laborum.")

fnt <- "Arial"
fsz <- 10

dev <- FALSE


# Basic Tests 1 - 10 ------------------------------------------------------


test_that("attr0: Title attribute class works.", {
  
  
  
  t1 <- ttl(c("Title1", "Title2"), bold = TRUE, font_size = 12,
                   align = "right", borders = c("top", "bottom"),
                   blank_row = "above", width = 8)
  
  t1
  expect_equal(length(t1$titles), 2)
  expect_equal(t1$align, "right")
  expect_equal(t1$blank_row, "above")
  expect_equal(t1$bold, TRUE)
  expect_equal(t1$font_size, 12)
  expect_equal(t1$borders[1], "top")
  expect_equal(t1$borders[2], "bottom")
  expect_equal(t1$width, 8)
  
})

test_that("attr1: Footnote attribute class works.", {
  
  
  
  f1 <- ftn(c("Foot1", "foot2"), valign = "top", width = 8,
                   align = "right", borders = c("top", "bottom"),
                   blank_row = "above")
  
  f1
  expect_equal(length(f1$footnotes), 2)
  expect_equal(f1$align, "right")
  expect_equal(f1$valign, "top")
  expect_equal(f1$blank_row, "above")
  expect_equal(f1$borders[1], "top")
  expect_equal(f1$borders[2], "bottom")
  expect_equal(f1$width, 8)
  
})


test_that("attr2: Span attribute class works.", {
  
  
  
  s1 <- span(from = "col1", to = "col2", label = "here",
                   label_align = "right", level = 2, underline = TRUE)
  
  s1
  expect_equal(s1$label, "here")
  expect_equal(s1$label_align, "right")
  expect_equal(s1$underline, TRUE)
  expect_equal(s1$from, "col1")
  expect_equal(s1$to, "col2")
  expect_equal(s1$level, 2)
  
})



test_that("attr3: apply_attributes works with title.", {
  
  
  dt <- mtcars
  
  tbl <- create_table(dt)
  
  attr(dt, "titles") <- list(ttl(c("Title1", "Title2")))

  
  expect_equal(is.null(attr(dt, "titles")), FALSE)
  
  tbl2 <- apply_attributes(tbl, dt)
  
  ttl <- tbl2$titles[[1]]
  
  expect_equal(length(ttl$titles), 2)
  expect_equal(ttl$width, "content")
  expect_equal(ttl$bold, FALSE)
  expect_equal(ttl$blank_row, "below")
  expect_equal(ttl$align, "center")

  
})



test_that("attr4: apply_attributes works with footnote.", {
  
  
  dt <- mtcars
  
  tbl <- create_table(dt)
  
  
  attr(dt, "footnotes") <- list(ftn(c("Foot1", "Foot2")))
  

  
  expect_equal(is.null(attr(dt, "footnotes")), FALSE)
  
  tbl2 <- apply_attributes(tbl, dt)
  
  ttl <- tbl2$footnotes[[1]]
  
  expect_equal(length(ttl$footnotes), 2)
  expect_equal(ttl$width, "content")
  expect_equal(ttl$valign, "top")
  expect_equal(ttl$blank_row, "above")
  expect_equal(ttl$align, "left")
  
  
})

test_that("attr4: apply_attributes works with spanning_header", {
  
  
  dt <- mtcars
  
  tbl <- create_table(dt)
  
  attr(dt, "spans") <- list(span(from = "mpg", to = "disp", 
                                     label = "Fork"))
  

  
  expect_equal(is.null(attr(dt, "spans")), FALSE)
  
  tbl2 <- apply_attributes(tbl, dt)
  
  spn <- tbl2$col_spans[[1]]
  
  expect_equal(length(spn$span_cols), 3)
  expect_equal(spn$label, "Fork")
  expect_equal(spn$label_align, "center")
  expect_equal(spn$underline, TRUE)
  expect_equal(spn$level, 1)
  
  
})


test_that("attr5: Attribute classes work.", {
  
  
  dt <- mtcars
  
  attr(dt, "titles") <- list(ttl(c("Title1", "Title2")))
  attr(dt, "footnotes") <- list(ftn(c("foot1", "foot2")))
  attr(dt, "spans") <- list(span("mpg", "disp", "Here"))
  
  tbl <- create_table(dt)

  tbl
  
  expect_equal(length(tbl$titles), 1)
  expect_equal(length(tbl$footnotes), 1)
  expect_equal(length(tbl$col_spans), 1)
  
})

test_that("attr6: Create report works with attributes.", {
  
  
  fp <- file.path(base_path, "attr/test6.txt")
  
  dt <- mtcars
  
  attr(dt, "titles") <- list(ttl(c("Title1", "Title2")))
  attr(dt, "footnotes") <- list(ftn(c("foot1", "foot2")))
  attr(dt, "spans") <- list(span("mpg", "disp", "Here"))
  
  tbl <- create_table(dt)
  
  rpt <- create_report(fp) %>% 
    add_content(tbl)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})

# Create Tables -----------------------------------------------------------

#' @details 
#' Order of these operations is very important.  Any change in the order 
#' will require considerable thought, and understanding of any potential
#' consequences.
#' @import fmtr
#' @noRd
create_table_pages_text <- function(rs, cntnt, lpg_rows) {

  #print("Create Table")
  
  ts <- cntnt$object
  content_blank_row <- cntnt$blank_row
  
  pgby_var <- NA
  pgby_cnt <- 0
  pgby_fmt <- NULL
  if (!is.null(rs$page_by)) {
    pgby_var <- rs$page_by$var
    pgby_fmt <- rs$page_by$format
  } else if (!is.null(ts$page_by)) {
    pgby_var <- ts$page_by$var
    pgby_fmt <- ts$page_by$format
  }
  
  
  if (all(ts$show_cols == "none") & length(ts$col_defs) == 0) {
    
    stop("ERROR: At least one column must be defined if show_cols = \"none\".")
  }

  font_name <- "Courier New"
  
  
  # Set up control columns
  dat <- as.data.frame(ts$data, stringsAsFactors = FALSE)  
  dat$..blank <- ""
  dat$..row <- NA
  dat$..page_by <- NA
  
  # If page_break variable has been defined, use it
  if (is.null(ts$page_var)) {
    if (is.na(pgby_var))
      dat$..page <- NA
    else 
      dat$..page <-  dat[[pgby_var]]
    
  } else 
    dat$..page <- dat[[ts$page_var]]
  
  # If page by is defined, use it
  if (!is.na(pgby_var)) {
    
    if (!is.null(pgby_fmt))
      dat$..page_by <-  fapply(dat[[pgby_var]], pgby_fmt)
    else 
      dat$..page_by <-  dat[[pgby_var]]
    
    pgby_cnt <- get_pgby_cnt(dat$..page_by)
    
    # Commenting out for now
    # if (is.unsorted(dat[[pgby_var]], strictly = FALSE))
    #   message("Page by variable not sorted.")
  }
  
  # Deal with invisible columns
  if (!is.null(ts$col_defs)) {
    for (def in ts$col_defs) {
      if (def$visible == FALSE) {
        nnm <- paste0("..x.", def$var_c)
        dat[[nnm]] <- dat[[def$var_c]]
      }
    }
    
    # Deal with Group Cohesion
    # Prepare ..group_cohesion and ..min_page_prop
    group_cohesion <- c()
    min_page_prop <- c()
    
    i <- 1
    for (def in ts$col_defs) {
      if (def$group_cohesion == TRUE) {
        group_cohesion <- c(group_cohesion, def$var_c)
        min_page_prop <- c(min_page_prop, def$min_page_prop)
        
        dat$..temp <- dat[[def$var_c]]
        names(dat)[names(dat) == "..temp"] <-  paste0("..group_cohesion",i)
        
        dat$..temp <- def$min_page_prop
        names(dat)[names(dat) == "..temp"] <-  paste0("..min_page_prop",i)
        i <- i + 1
      }
    }
    
    if (length(group_cohesion) == 0) {
      group_cohesion <- NULL
    }
    
    if (length(min_page_prop) == 0) {
      min_page_prop <- NULL
    }
    
    # Prepare empty ..break_label and ..break_label_lines for later use
    i <- 1
    for (def in ts$col_defs) {
      if (!is.null(def$break_label)) {
        
        dat$..temp <- ifelse(!is.na(dat[[def$var_c]]) & dat[[def$var_c]] != "",
                             paste0(dat[[def$var_c]], " ", def$break_label),
                             dat[[def$var_c]])
        names(dat)[names(dat) == "..temp"] <-  paste0("..break_label",i)
        
        dat$..temp <- rep(NA, nrow(dat))
        names(dat)[names(dat) == "..temp"] <-  paste0("..break_label_lines",i)
        i <- i + 1
      }
    }
    # Prepare repeated label occurrence for later use
    if (any(grepl("..break_label", names(dat)))) {
      dat$..break_occur <- rep(NA, nrow(dat))
    }
  }
  
  # Get control column names 
  control_cols <- names(dat)[is.controlv(names(dat))]

  # Get vector of all included column names
  # Not all columns in dataset are necessarily included
  # depends on show_cols parameter on create_table and
  # visible parameter on column definitions
  keys <- get_table_cols(ts, control_cols)
  # print("keys")
  # print(keys)

  # Filter dataset by included columns
  dat <- get_data_subset(dat, keys, rs$preview)
   # print("Key columns:")
   # print(dat)
  
  # Update column definitions with column defaults
  ts$col_defs <- set_column_defaults(ts, keys)
  # print("col_defs:")
  # print(ts$col_defs)
  
  # Get labels
  labels <- get_labels(dat, ts)
  # print("Labels:")
  # print(labels)
  
  # Get column alignments
  aligns <- get_aligns(dat, ts)
    
  # Get alignment for labels
  # Follows column alignment by default
  label_aligns <- get_label_aligns(ts, aligns)
  # print("Label Aligns:")
  # print(label_aligns)

  # Clear out existing formats
  cdat <- clear_formats(dat)
  
  # Get column formats
  formats(cdat) <- get_col_formats(dat, ts)
  # print("formats:")
  # print(formats(cdat))

  # Apply formatting
  fdat <- fdata(cdat)
  # print("fdata:")
  # print(fdat)
  
  # Prep data for blank lines, indents, and stub columns
  fdat <- prep_data(fdat, ts, rs$char_width, rs$missing)
  # print("prep_data")
  # print(fdat)
  # str(fdat)
  
  # Reset keys, since prep_data can add/remove columns for stub
  keys <- names(fdat)
  # print("Keys")
  # print(keys)
  # print("Aligns")
  # print(aligns)
  
  # Copy any width attributes to formatted data frame
  if ("width" %in% ts$use_attributes)
    widths(fdat) <- widths(dat)
  # print("Original Widths")
  # print(widths(dat))
  
  # Get column widths
  widths_uom <- get_col_widths(fdat, ts, labels, rs$char_width, rs$units)
  # print("Widths UOM")
  # print(widths_uom)
  
  # Try to figure out a way to put these widths in report spec
  #rs$column_widths[length(rs$column_widths) + 1] <- widths_uom

  # # Convert to text measurements
  # # Adjust by -1 to account for space between columns
  widths_char <- round(widths_uom / rs$char_width) - 1
  # # print("Widths Char")
  # # print(widths_char)

  # Split long text strings into multiple rows
  fdat <- split_cells(fdat, widths_char, ts, rs$char_width)
  
  # Derive lines for break label if needed
  fdat <- get_break_lines(fdat, widths_char, ts, rs$char_width)

  # print("split_cells")
  # print(fdat)

  # Apply widths and justification
  # widths(fdat) <- widths_char
  # print(widths_char)
  # justification(fdat) <- aligns
  # print(aligns)
  # fdat <- fdata(fdat)
  # print("fdata2")
  # print(fdat)
  
  fdat <- apply_widths(fdat, widths_char, aligns)
  
  # Break columns into pages
  wraps <- get_page_wraps(rs$line_size, ts, widths_char, 1, control_cols)
  # print("wraps")
  # print(wraps)

  # Create a temporary page info to pass into get_content_offsets
  tmp_pi <- list(keys = keys, col_width = widths_uom, label = labels,
                 label_align = label_aligns)
  # print("Temp PI")
  # print(tmp_pi)
  
  # Offsets are needed to calculate splits and page breaks
  content_offset <- get_content_offsets(rs, ts, tmp_pi, content_blank_row, pgby_cnt)
  
  # split rows
  splits <- get_splits_text(fdat, widths_uom, rs$body_line_count, 
                            lpg_rows, content_offset, ts)
   # print("splits")
   # print(splits)
  
  # Subset splits by preview, if requested
  if (!is.null(rs$preview)) {
    if (rs$preview < length(splits))
      splits <- splits[seq(1, rs$preview)] 
  }
  
  tot_count <- length(splits) * length(wraps)
  counter <- 0
  wrap_flag <- FALSE
  blnk_ind <- "none"
  
  pg_lst <- list()
  for(s in splits) {
    
    # Impute records for break labels
    if (any(grepl("break_label", row.names(s)))) {
      break_label_df <- s[grepl("break_label", row.names(s)),]
      break_label_df <- split_cells(break_label_df, widths_char, ts, rs$char_width)
      break_label_df[, 1] <- stri_pad_right(break_label_df[, 1], sum(widths_char))
      s <- rbind(break_label_df, s[!grepl("break_label", row.names(s)),])
    }
    
    for(pg in wraps) {
      counter <- counter + 1

      if (counter < tot_count)
        wrap_flag <- TRUE
      else 
        wrap_flag <- FALSE
      
      #print(s)
      # Ensure content blank rows are added only to the first and last pages
      blnk_ind <- get_blank_indicator(counter, tot_count, content_blank_row,
                                      rs$body_line_count, content_offset, nrow(s))
      #print(blnk_ind)
      
      if (!is.na(pgby_var))
        pgby <- trimws(s[1, "..page_by"])
      else 
        pgby <- NULL
      
      
      pi <- page_info(data= s[, pg], keys = pg, label=labels[pg],
                     col_width = widths_uom[pg], col_align = aligns[pg],
                     font_name = font_name, label_align = label_aligns[pg],
                     pgby)
      pg_lst[[length(pg_lst) + 1]] <- create_table_text(rs, ts, pi, 
                                                        blnk_ind, wrap_flag,
                                                        lpg_rows)
    }
  }
  
  ret <- list(widths = widths_uom, page_list = pg_lst)
  
  return(ret)
  
}

#' @noRd
create_table_text <- function(rs, ts, pi, content_blank_row, wrap_flag, 
                              lpg_rows) {
  
  shdrs <- c()
  hdrs <- c()
  
  if (ts$headerless == FALSE) {
    shdrs <- get_spanning_header(rs, ts, pi)   
    hdrs <- get_table_header(rs, ts, pi)  
  }
  
  rws <- get_table_body(pi$data)
  
  
  ls <- rs$line_size
  if (length(rws) > 0)
    ls <- max(nchar(rws))
  
  if (!is.null(ts$title_hdr)) {
    ttls <- get_title_header(ts$title_hdr, ls , rs$line_size, 
                             rs$uchar, rs$char_width)
  } else
    ttls <- get_titles(ts$titles, ls, rs$line_size, rs$uchar, rs$char_width) 
  

  #print("Titles")
  #print(ttls)
  
  if (!is.null(rs$page_by))
    pgby <- get_page_by(rs$page_by, rs$line_size - 1, pi$page_by)
  else if(!is.null(ts$page_by))
    pgby <- get_page_by(ts$page_by, ls, pi$page_by)
  else 
    pgby <- c()
  
  # print("Create table text:")
  # print(length(pgby))
  
  a <- NULL
  if (content_blank_row %in% c("above", "both"))
    a <- ""
  
  # Add top border if requested
  tbrdr <- NULL
  if (any(ts$borders %in% c("top", "all", "outside")))
    tbrdr <-  paste0(paste0(rep(rs$uchar,  ls - 1), collapse = ""), " ")
  
  # Add bottom border if requested
  bbrdr <- NULL
  if (any(ts$borders %in% c("bottom", "all", "outside")))
    bbrdr <-  paste0(paste0(rep(rs$uchar,  ls - 1), collapse = ""), " ")
  
  
  # Append everything together
  ret <- c(a, ttls, pgby, tbrdr, shdrs, hdrs, rws, bbrdr)
  
  # Footnotes are complicated enough that they need their own function
  ftnts <- get_page_footnotes_text(rs, ts, ls, lpg_rows, 
                                   length(ret), wrap_flag, content_blank_row)
  
  # Append footnotes 
  ret <- c(ret, ftnts)
  

  
  return(ret) 
}

#' @description Function to deal with complexities of footnotes.  Complexities
#' include valign parameter and width parameter.
#' @noRd
get_page_footnotes_text <- function(rs, spec, spec_width, 
                                    lpg_rows, row_count, wrap_flag,
                                    content_blank_row) {
  
  ftnts <- c()
  vflag <- FALSE
  
  # Deal with valign parameter
  if (!is.null(spec$footnotes)) {
    if (!is.null(spec$footnotes[[length(spec$footnotes)]])) {
      if (spec$footnotes[[length(spec$footnotes)]]$valign == "bottom") {
        vflag <- TRUE
      } 
      ftnts <- get_footnotes(spec$footnotes, spec_width, rs$line_size, 
                             rs$uchar, rs$char_width) 
      
    }
  } else {
    
    if (!is.null(rs$footnotes[[1]])) {
      if (!is.null(rs$footnotes[[1]]$valign)) {
        if (rs$footnotes[[1]]$valign == "top") {
          
          ftnts <- get_footnotes(rs$footnotes, spec_width, rs$line_size, 
                                 rs$uchar, rs$char_width) 
        } 
      }
    }
  }
  
  b <- NULL
  if (content_blank_row %in% c("below", "both"))
    b <- ""
  
  ublnks <- c()
  lblnks <- c()
  
  len_diff <- rs$body_line_count - lpg_rows - row_count - length(ftnts) - length(b)
  
  if (vflag) {
    
    # At some point will have to deal with wrap flag
    # Right now cannot put anything in between end of content
    # and the start of the footnote.  Edge case but should still
    # fix at some point.
    if (len_diff > 0) {
      ublnks <- c(b, rep("", len_diff))
    }
    
  } else { 
  
  
    if (wrap_flag & len_diff > 0) {
      lblnks <- c(rep("", len_diff), b)
    } else {
      lblnks <- b
    }
    
  }
  
  ret <- c(ublnks, ftnts, lblnks)
  
  return(ret)
  
}

# Sub-Functions -----------------------------------------------------------


#' Get content offsets for table header, titles, footnotes, and content blanks.
#' Needed to calculate page breaks accurately.
#' @return A vector of upper and lower offsets
#' @noRd
get_content_offsets <- function(rs, ts, pi, content_blank_row, pgby_cnt = NULL) {
  
  ret <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
  
  shdrs <- c()
  hdrs <- c()
  
  if (ts$headerless == FALSE) {
    shdrs <- get_spanning_header(rs, ts, pi)   
    hdrs <- get_table_header(rs, ts, pi)  
  }
  
  w <- ceiling(sum(pi$col_width) / rs$char_width)

  if (is.null(ts$title_hdr))
    ttls <- get_titles(ts$titles, w, rs$line_size, rs$uchar, rs$char_width) 
  else {
    ttls <- get_title_header(ts$title_hdr, w, rs$line_size, rs$uchar, 
                             rs$char_width)
  }
  
  pgb <- c()
  if (!is.null(ts$page_by))
    pgb <- get_page_by(ts$page_by, rs$line_size, NULL, pgby_cnt)
  else if (!is.null(rs$page_by))
    pgb <- get_page_by(rs$page_by, rs$line_size, NULL, pgby_cnt)
  
  #print(length(pgb))
  
  # print(paste("Table titles:", ttls))
  tbrdr <- 0
  if (any(ts$borders %in% c("all", "top", "outside")))
      tbrdr <- 1
      
  bbrdr <- 0
  if (any(ts$borders %in% c("all", "bottom", "outside")))
    bbrdr <- 1
  
  ret["upper"] <- length(shdrs) + length(hdrs) + length(ttls) + length(pgb) + tbrdr
  
  if (content_blank_row %in% c("above", "both"))
      ret["blank_upper"] <- 1
  
  ftnts <- get_footnotes(ts$footnotes, w, rs$line_size, rs$uchar, rs$char_width) 
  rftnts <- get_footnotes(rs$footnotes, w, rs$line_size, rs$uchar, rs$char_width) 
  
  if (has_top_footnotes(rs)) {
    ret["lower"] <- length(ftnts) + length(rftnts) + bbrdr
  } else {
    ret["lower"] <- length(ftnts) + bbrdr
  }

  if (content_blank_row %in% c("both", "below"))
    ret["blank_lower"] <- 1

  return(ret)

}

#' @description Return a vector of strings for the table header
#' @details Basic idea of this function is to create a list
#' of string vectors of label words sized according to the column 
#' widths, then combine by line/row, and concatenate everything and justify.
#' @noRd
get_table_header <- function(rs, ts, pi) {
  
  lbls <- pi$label
  lbla <- pi$label_align
  w <- round(pi$col_width / rs$char_width) - 1 # Adjust for gutter
  # print("Label A")
  # print(lbla)
  ret <- c()
  ln <- c()

  # Wrap header labels if needed
  d <- data.frame(as.list(lbls), stringsAsFactors = FALSE)
  names(d) <- names(lbls)
  d <- split_cells(d, w)
  d <- push_down(d)

  # Label justification, width, and row concatenation
  for (i in seq_len(nrow(d))) {
    
    r <- ""

    for (nm in names(d)) {
      if (!is.control(nm)) {
        if (nm %in% names(lbla)) {
          r <- paste0(r, pad_any(d[i, nm],  w[[nm]],
                               get_justify(lbla[[nm]])), " ")
        }
        
      }
    }
    
    ln[[length(ln) + 1]] <- r 
  }
  
  
  # Underline
  sep <- paste0(paste0(rep(rs$uchar, nchar(r) - 1), collapse = ""), " ")
  
  ln[[length(ln) + 1]] <- sep

  
  ret <- ln
  
  if (ts$first_row_blank)
    ret[[length(ret) + 1]] <- ""
  
  ret <- unlist(ret)
  
  return(ret)
}


#' @description Return a vector of strings for the table spanning headers
#' @details Basic idea of this function is to figure out which columns 
#' the header spans, add widths, then call get_table_header.  Everything
#' from there is the same.  
#' @import stats
#' @noRd
get_spanning_header <- function(rs, ts, pi) {
  
  # spns <- ts$col_spans
  # cols <- pi$keys
  # cols <- cols[!is.controlv(cols)]
  # w <- round(pi$col_width / rs$char_width) - 1 # Adjust for gutter
  # w <- w[cols]
  # 
  # # print("Cols:")
  # # print(cols)
  # #print(w)
  # 
  # # Figure out how many levels there are, 
  # # and organize spans by level
  # lvls <- c()     # Unique levels
  # slvl <- list()  # Will be a list of lists of spans
  # for (spn in spns) {
  # 
  #   if (!spn$level %in% lvls) {
  #     lvls[length(lvls) + 1] <- spn$level
  #     slvl[[spn$level]] <- list()
  #   }
  #   slvl[[spn$level]][[length(slvl[[spn$level]]) + 1]] <- spn
  # }
  # 
  # # Get unique levels and sort in decreasing order so we go from top down
  # lvls <- sort(unique(lvls), decreasing = TRUE)
  # # print("Levels:")
  # # print(lvls)
  # # 
  # # print("Spanning levels:")
  # # print(slvl)
  # 
  # # Create data structure to map spans to columns and columns widths by level
  # # - Seed span_num with negative index numbers to identify unspanned columns
  # # - Also add one to each column width for the blank space between columns 
  # d <- data.frame(colname = cols, colwidth = w + 1, 
  #                 span_num = seq(from = -1, to = -length(cols), by = -1), 
  #                 stringsAsFactors = FALSE)
  # 
  # wlvl <- list()  # Create one data structure for each level
  # for (l in lvls) {
  # 
  #   t <- d  # Copy to temporary variable
  #   
  #   # if column is in spanning column list, populate structure with index.
  #   # Otherwise, leave as negative value.
  #   for (i in 1:length(slvl[[l]])) {
  #     cl <- slvl[[l]][[i]]$span_cols
  #     
  #     
  #     # Span specifications can be a vector of column names or numbers
  #     if (typeof(cl) == "character")
  #       t$span_num <- ifelse(t$colname %in% cl, i, t$span_num)
  #     else 
  #       t$span_num <- ifelse(t$colname %in% cols[cl], i, t$span_num)
  #     
  # 
  #   }
  #   
  #   # Aggregate data structures to get span widths for each span
  #   s <- aggregate(x = list(width = t$colwidth), by = list(span = t$span_num), FUN = sum)
  # 
  #   # Then put back in original column order
  #   s$span <- factor(s$span, levels = unique(t$span_num))
  #   s <- s[order(s$span), ]
  #   rownames(s) <- NULL
  # 
  #   # Prep data structure
  #   s$span <- unique(t$span_num)
  #   s$label <- ""
  #   s$align <- ""
  #   s$n <- NA
  #   s$name <- ""
  #   s$underline <- TRUE
  #   
  #   # Populate data structure with labels, alignments, and n values from 
  #   # spanning column objects
  #   counter <- 1
  #   for (index in s$span) {
  #     if (index > 0) {
  #       s$label[counter] <- slvl[[l]][[index]]["label"]
  #       s$align[counter] <- slvl[[l]][[index]]$label_align
  #       s$underline[counter] <- slvl[[l]][[index]]$underline
  #       if (!is.null(slvl[[l]][[index]]$n))
  #         s$n[counter] <- slvl[[l]][[index]]$n
  # 
  #     }
  #     s$name[counter] <- paste0("Span", counter)
  #     counter <- counter + 1
  #   }
  #   
  #   # Apply n counts to labels
  #   if (!is.null(ts$n_format)) {
  #     s$label <- ifelse(is.na(s$n), s$label, paste0(s$label, ts$n_format(s$n))) 
  #   }
  # 
  #   wlvl[[l]] <- s
  # 
  # }
  
  # Put spanning info in function to eliminate redundancy with rtf version
  w <- round(pi$col_width / rs$char_width) - 1 # Adjust for gutter
  wlvl <- get_spanning_info(rs, ts, pi, w, 1)
  lvls <- sort(seq_along(wlvl), decreasing = TRUE)
  
  # At this point we have a data frame with labels, etc. and spanning
  # column widths, and are ready to create spanning header rows
  #print(wlvl)
  
  # Format labels for each level
  ln <- c()
  for (l in lvls) {
    s <- wlvl[[l]]
    
    # Wrap header labels if needed
    d <- data.frame(as.list(s$label), stringsAsFactors = FALSE)
    w <- s$width
    j <- s$align
    names(d) <- s$name
    names(w) <- s$name
    names(j) <- s$name
    
    # print("Ready for split cells")
    # print(paste("d:", d))
    # print(paste("w:", w))
    d <- split_cells(d, w)
    d <- push_down(d)
    
    #print(d)
    
    # Label justification, width, and row concatenation
    for (i in seq_len(nrow(d))) {

      r <- ""
      


      for (nm in names(d)) {
        if (!is.control(nm)) {

          if (j[[nm]] == "right") # Adjust 1 space for right alignment
            r <- paste0(r, pad_any(d[i, nm], w[[nm]] -1,
                                  get_justify(j[[nm]])), " ")
          else
            r <- paste0(r, pad_any(d[i, nm], w[[nm]],
                                   get_justify(j[[nm]])))
          

        }
      }

      ln[[length(ln) + 1]] <- r
    }
    
    r <- ""
    #print(s)
    
    for (i in seq_len(nrow(s))) {

      if (s$span[i] > 0) {

        if (s$underline[i])
          r <- paste0(r, paste0(rep(rs$uchar, s$width[i] - 1), collapse = ""), " ")
        else 
          r <- paste0(r, paste0(rep(" ", s$width[i] - 1), collapse = ""), " ")
        
          
      } else {
        r <- paste0(r, paste0(rep(" ", s$width[i]), collapse = ""))
      }
    }
    
    # r <- ifelse(s$span > 0, paste0(r, paste0(rep("-", 4), collapse = ""), " "),
    #                         paste0(r, rep(" ", 4), collapse = ""))
    ln[[length(ln) + 1]] <- r
    
    #print(ln)
  }

  ret <- unlist(ln)
  
  return(ret)
}

#' @description At this point, can just put out the cells in the page.
#' All the wrapping and splitting has been taken care of.
#' @noRd
get_table_body <- function(dat) {
  
  df <- dat
  ret <- c()
  nm <- names(df)
  
  for (i in seq_len(nrow(df))) {
    
    r <- ""
    for (j in seq_len(ncol(df))) {
      v <- ""
      p <-df[i, j]
      if (!is.null(p) & !is.na(p))
        v <- p
      

      if (!is.control(nm[j])) { 
        if ("..blank" %in% names(df)) {
          if (!trimws(df[i, "..blank"]) %in% c("B", "A")) {
  
            r <- paste0(r, v, " ")
          
          }
        } else {
          r <- paste0(r, v, " ") 
        }
      }
    }
    
    ret[length(ret) + 1] <- r
  }
  
  return(ret)
  
}




# Utilities ---------------------------------------------------------------

#' @noRd
is.control <- function(x) {
  
 ret <- FALSE
 if (!is.na(x)) {
   if (substr(x, 1, 2) == "..")
     ret <- TRUE
 }
 
 return(ret)
  
}

is.controlv <- Vectorize(function(x) {
  
  return(is.control(x))
})

#' @noRd
is.invisible <- function(x) {
  
  ret <- FALSE
  if (!is.na(x)) {
    if (substr(x, 1, 4) == "..x.")
      ret <- TRUE
  }
  
  return(ret)
  
}

is.invisiblev <- Vectorize(function(x) {
  
  return(is.invisible(x))
})


has_invisible <- function(lst) {
  
  ret <- FALSE
  
  if (length(lst[is.invisiblev(lst)]) > 0)
    ret <- TRUE
  
  return(ret)
}

get_invisible <- function(nm) {
 
  return(paste0("..x.", nm)) 
  
}

translate_invisible <- function(nm, lst) {
 
  ret <- nm
  
  if (has_invisible(lst)) {
    
    inv <- lst[is.invisiblev(lst)]
    
    if (get_invisible(nm) %in% inv) {
      ret <- get_invisible(nm)
    }
  }
  
  return(ret)
}

#' @noRd
get_justify <- function(x) {
  
  ret <- "left"
  if (is.null(x) | is.na(x))
    ret <- "left"
  else if (x == "center")
    ret <- "centre"
  else if (x %in% c("left", "right", "centre"))
    ret <- x
  
  return(ret)
}

#' @noRd
get_blank_indicator <- function(pg_num, tot_pg, content_blanks,
                                page_size, content_offset, num_rows, pgby = NULL) {
  
  if (is.null(names(content_offset[["upper"]]))) {
    content_offset_up <- content_offset[["upper"]]
  } else {
    content_offset_up <- content_offset[["upper"]][[pgby]]
  }
  
  if (is.null(names(content_offset[["lower"]]))) {
    content_offset_low <- content_offset[["lower"]]
  } else {
    content_offset_low <- content_offset[["lower"]][[pgby]]
  }
  
  if (pg_num == 1 & pg_num == tot_pg & content_blanks == "both")
    blnk_ind <- "both"
  else if (pg_num == 1 & content_blanks %in% c("both", "above"))
    blnk_ind <- "above"
  else if ((pg_num == tot_pg & content_blanks %in% c("both", "below")) &
           (num_rows < page_size - content_offset_up - content_offset_low)) {
    # Exception when number of rows exactly equals available space.
    # Then don't put a blank row below.
    # print(num_rows)
    # print( page_size )
    # print(content_offset)
    blnk_ind <- "below"
  } else 
    blnk_ind <- "none"
  
  return(blnk_ind)
}

get_break_lines <- function(dat, widths_char, ts, char_width){
  defs <- ts$col_defs
  if (any(grepl("..break_label", names(dat)))) {
    
    # For break labels, prepare the break label columns for later use
    break_label_col <- c()
    for (d in 1:length(defs)) {
      if (!is.null(defs[[d]]$break_label)){
        break_label_col <- c(break_label_col, defs[[d]]$var_c)
      }
    }
    
    for (b in names(dat)[grepl("..break_label\\d+", names(dat))]) {
      break_label_num <- sub(".*?([0-9]+).*", "\\1", b)
      # Get the indentation from original variable
      break_label_indent <- 0
      cur_break_label_col <- break_label_col[as.numeric(break_label_num)]
      if (!is.null(defs[[cur_break_label_col]]$indent)) {
        break_label_indent <- defs[[cur_break_label_col]]$indent
      }
      
      ind_blank_num <- floor(break_label_indent / char_width)
      
      get_lines <- function(input, ind = ind_blank_num, txt_width = sum(widths_char)){
        txt_wrap <- stri_wrap(unlist(
          strsplit(input, split = "\n", fixed = TRUE)),
          width = txt_width, normalize = FALSE, indent = ind, exdent = ind)
        return(length(txt_wrap))
      }
      
      dat[[paste0("..break_label_lines",break_label_num)]] <- vapply(dat[[b]], get_lines, numeric(1))
    }
  }
  return(dat)
}

context("Create Table Text Tests")

test_that("get_justify() works as expected.", {
  
  expect_equal(get_justify(NA), "left")
  expect_equal(get_justify("right"), "right")
  expect_equal(get_justify("center"), "centre")
  expect_equal(get_justify("fork"), "left")
})


test_that("get_table_header works as expected.", {
  

  rpt <- create_report("fork.out")
  rpt$line_size <- 108
  
  tbl <- create_table(mtcars)
  
  lbl <- names(mtcars)
  names(lbl) <- names(mtcars)
  lbl
  
  lbl[["vs"]] <- "vs\n(n=25)" 
  
  aln <- c(unlist(rep("left",11)))
  names(aln) <- names(mtcars)
  w <- c(unlist(rep(.5, 11)))
  names(w) <- names(mtcars)
  w[["vs"]] <- 1

  p <- page_info(mtcars, keys = names(mtcars), font_name = "Courier New", 
                 col_width = w, 
                 col_align = aln,
                 label = lbl , label_align = aln)
  
  tbl$data
  res1 <- get_table_header(rpt, tbl, p)
  
  
  expect_equal(length(res1), 3)
  expect_equal(trimws(res1[[1]]), "vs")
      
})


test_that("get_spanning_header works as expected.", {
  
  dat <- mtcars[1:10, ]
  rpt <- create_report("fork.out")
  rpt$line_size <- 108
  
  tbl <- create_table(dat) %>% 
    spanning_header("mpg","disp", label = "Span 1",
                    label_align = "center", n = 24) %>% 
    spanning_header("hp", "wt", label = "Span 2",
                    label_align = "center", n = 25)  %>%
    spanning_header("qsec", "am",
                    label = "Span 3", label_align = "center", n = 26) %>%
    spanning_header(2, 5, level = 2,
                    label = "Span 4", label_align = "center") %>%
    spanning_header(6, 9, level = 2,
                    label = "Span 5", label_align = "center")
  
  lbl <- names(mtcars)
  names(lbl) <- names(mtcars)

  
  aln <- c(unlist(rep("right",11)))
  names(aln) <- names(mtcars)
  w <- c(unlist(rep(.5, 11)))
  names(w) <- names(mtcars)
  w[["qsec"]] <- .75
  
  p <- page_info(dat, keys = names(mtcars), font_name = "Courier New", 
                 col_width = w, 
                 col_align = aln,
                 label = lbl , label_align = aln)
  

  res1 <- get_spanning_header(rpt, tbl, p)
  res1
  
  expect_equal(length(res1), 5)
  expect_equal(trimws(res1[[1]]), "Span 4                   Span 5")
  
})


test_that("get_page_footnotes_text works as expected.", {
  
  
  tbl1 <- create_table(iris)
  
  rpt1 <- create_report("") %>%
    titles("IRIS Data Frame") %>%
    add_content(tbl1) %>% 
    footnotes("Here is a footnote", valign = "top")
  
  rpt1 <- page_setup(rpt1)
  
  res1 <- get_page_footnotes_text(rpt1, tbl1, 6,
                                  0, 25, TRUE, "below")
  
  expect_equal(length(res1), 18)
  
  tbl2 <- create_table(iris)
  
  rpt2 <- create_report("") %>%
    titles("IRIS Data Frame") %>%
    add_content(tbl1) %>% 
    footnotes("Here is a footnote", valign = "bottom")
  
  rpt2 <- page_setup(rpt2)
  
  res2 <- get_page_footnotes_text(rpt2, tbl2, 6,
                                  0, 25, TRUE, "below")  
  
  expect_equal(length(res2), 16)
  
  
  tbl3 <- create_table(iris)  %>% 
    footnotes("Here is a footnote", valign = "bottom")
  
  rpt3 <- create_report("") %>%
    titles("IRIS Data Frame") %>%
    add_content(tbl1)
  
  rpt3 <- page_setup(rpt3)
  
  res3 <- get_page_footnotes_text(rpt3, tbl3, 6,
                                  0, 25, TRUE, "below")  
  
  expect_equal(length(res3), 18)
  
  tbl4 <- create_table(iris)  %>% 
    footnotes("Here is a footnote", valign = "top")
  
  rpt4 <- create_report("") %>%
    titles("IRIS Data Frame") %>%
    add_content(tbl1)
  
  rpt4 <- page_setup(rpt4)
  
  res4 <- get_page_footnotes_text(rpt4, tbl4, 26,
                                  0, 25, TRUE, "below")  
  
  expect_equal(length(res4), 18)
  
})

#' @title
#' Write a RTF report to the file system
#'
#' @description
#' This function writes a report_spec object to the file system, using the
#' parameters provided in the object.
#'
#' @param x The report_spec object to write.
#' @return The report spec.
#' @noRd
write_report_rtf <- function(rs) {
  
  debug <- FALSE
  
  orig_path <- rs$modified_path
  
  # Create temp path for text output
  if (debug) {
# 
#     b_path <- file.path(getwd(), "tests/testthat")
# 
#     tmp_path <- file.path(b_path, "output/tmp.txt")
#     rtf_path <- file.path(b_path, "output/tmp.rtf")
#     if (file.exists(tmp_path))
#       file.remove(tmp_path)
#     if (file.exists(rtf_path))
#       file.remove(rtf_path)
#     if (file.exists(orig_path))
#       file.remove(orig_path)
  } else {
    tmp_dir <- tempdir()
    if (!file.exists(tmp_dir))
      dir.create(tmp_dir)
    tmp_path <- tempfile(fileext = ".txt", tmpdir = tmp_dir)
    rtf_path <- tempfile(fileext = ".rtf", tmpdir = tmp_dir)

  }
  # print(tmp_path)
  # print(rtf_path)
  # print(getwd())

  if (file.exists(orig_path))
    file.remove(orig_path)
  
  # Replace original path
  rs$modified_path <- tmp_path
  
  # Create text output normally to temp location
  rs <- write_report_text(rs)

  # Read lines from text output
  ls <- readLines(tmp_path)

  # Revise text and write to rtf
  fls <- write_rtf_output(rs, ls, rtf_path, orig_path, tmp_dir)
  
  # Restore original path
  rs$modified_path <- orig_path

  # Clean up
  if (!debug) {
    file.remove(tmp_path)
    #file.remove(rtf_path)
    #nms <- list.files(tmp_dir, pattern = "(.*)\\.png", full.names = TRUE)
    if (length(fls) > 0)
      file.remove(fls)
    #unlink(tmp_dir, recursive = TRUE)
  }

  return(rs)
}


# May need some adjustments/sophistication/options to this function
#' @return Vector of graphic file paths
#' @noRd
write_rtf_output <- function(rs, ls, rtf_path, orig_path, tmp_dir) {
  
  # Set up vectors
  hdr <- c() 
  body <- c() 
  
  ret <- c()

  
  # Get conversion factor to twips
  if (rs$units == "inches") {
    conv <- 1440
  } else {
    conv <- 566.9291
  }
  
  
  # Prepare header
  hdr[length(hdr) + 1] <- "{\\rtf1\\ansi\\deff0 {\\fonttbl {\\f0 Courier;}}"
  if (rs$orientation == "landscape") {
    #hdr[length(hdr) + 1] <- "\\landscape\\horzdoc"
    hdr[length(hdr) + 1] <- "\\landscape"
    hdr[length(hdr) + 1] <- paste0("\\paperw", round(rs$page_size[2] * conv),
                                   "\\paperh", round(rs$page_size[1] * conv))
  } else {
    #hdr[length(hdr) + 1] <- "\\vertdoc"
    hdr[length(hdr) + 1] <- paste0("\\paperw", round(rs$page_size[1] * conv),
                                   "\\paperh", round(rs$page_size[2] * conv))
  }
  
  hdr[length(hdr) + 1] <- paste0("\\margl", round(rs$margin_left * conv),
                                "\\margr", round(rs$margin_right * conv),
                                "\\margt", round(rs$margin_top * conv),
                                "\\margb", round(rs$margin_bottom  * conv))

  # Line spacing values determined be trial and error.
  # Needed for LibreOffice.  Appear to be ignored in Word.
  if (rs$font_size == 10)
    hdr[length(hdr) + 1] <- "\\sl-225\\slmult0\\fs20"
  else if (rs$font_size == 12)
    hdr[length(hdr) + 1] <- "\\sl-275\\slmult0\\fs24"
  else if (rs$font_size == 8)
    hdr[length(hdr) + 1] <- "\\sl-180\\slmult0\\fs16"
  else if (rs$font_size == 11)
    hdr[length(hdr) + 1] <- "\\sl-250\\slmult0\\fs22"
  else if (rs$font_size == 9)
    hdr[length(hdr) + 1] <- "\\sl-200\\slmult0\\fs18"
  
  # Start with all lines
  body <- encodeRTF(ls)
  
  if (rs$has_graphics) {
    # Remove fill lines
    fill_tags <- grep("```fill```", body, fixed = TRUE)
    body <- body[-fill_tags]

    # Replace any plot tags with latex codes
    plt_tags <- grep("```([^}]*)```", body)

    for (i in plt_tags) {

      # Remove braces
      rw <- trimws(gsub("`", "", body[i], fixed = TRUE))

      # Split on pipe
      spec <- strsplit(rw, "|", fixed = TRUE)[[1]]

      # 1 = path
      # 2 = height
      # 3 = width
      # 4 = align
      
      img <- get_image_rtf(spec[[1]], as.numeric(spec[[3]]), 
                           as.numeric(spec[[2]]), rs$units)
      ret[length(ret) + 1] <- spec[[1]]
      
      # Create rtf codes
      if (spec[[4]] == "left") {
        hd <- paste0("\\par\\sl0\\ql\n"  )

      } else if (spec[[4]] == "right") {
        hd <- paste0("\\par\\sl0\\qr\n"  )
      } else  {
        hd <- paste0("\\par\\sl0\\qc\n"  )
      }
      
      # Restore sizing and alignment
      if (rs$font_size == 8) {
        ft <- "\\par\\sl-180\\ql"
      } else if (rs$font_size ==  12) {
        ft <- "\\par\\sl-275\\ql"
      } else if (rs$font_size ==  11) {
        ft <- "\\par\\sl-250\\ql"
      } else if (rs$font_size ==  9) {
        ft <- "\\par\\sl-200\\ql"
      } else {
        ft <- "\\par\\sl-225\\ql"
      }

      # Replace original line with RTF codes
      body[[i]] <- paste0(hd, img, ft)
    }
  }
  
  # Get page breaks
  pgs <- grepl("\f", body, fixed = TRUE)
  
  # Adjust by one to get previous line
  pgs <- pgs[2:length(pgs)]
  
  # Add one for end of document
  pgs[length(pgs) + 1] <- TRUE 
  # print(length(body))
  # print(length(pgs))
  # print(pgs)
  
  body <- gsub("\f", "\\page ", body, fixed = TRUE)
  body <- paste0(body, ifelse(pgs, "", "\\line"))
  
  # Write to file  
  f <- file(orig_path, open="a")
  
  writeLines(hdr, con = f)
  
  writeLines(body, con = f)
  
  writeLines("}", con = f)
  
  close(f)
  
#   file.copy(rtf_path, orig_path)
#   file.remove(rtf_path)
  
  return(ret)

}

#' @noRd
encodeRTF <- Vectorize(function(x) {
  
  
  ints <- utf8ToInt(x)
  res <- paste0(encodeRTF_internal(ints), collapse = "")
  
  return(res)
})


# These are characters that will mess up the RTF.
# Have to be translated to something else.
# Using character codes because R doesn't like 
# non-ascii characters in the code.
#' @noRd
encodeRTF_internal <- Vectorize(function(x) {
  
  ret <- intToUtf8(x, multiple = TRUE)
  if (ret == "}")
    ret <- "\\'7d"
  else if (ret == "{")
    ret <- "\\'7b"
  else if (ret == "\\")
    ret <- "\\'5c"
  else if (x == 175)  # This is a macron used for underlines
    ret <- "\\u175\\u175"
  else if (x > 127)  # Any non-alphanumeric character
    ret <- paste0("\\u", x, "  ")
  
  return(ret)
  
})

# The page template is everything except the content: page header/footer,
# titles, footnotes, etc.

# Page Template RTF Functions ---------------------------------------------


#' Create a page template with header, titles, footnotes, and footer.
#' @param rs The report spec
#' @return The page template object
#' @noRd
page_template_pdf<- function(rs) {
  
  pt <- structure(list(), class = c("page_template_pdf", "list"))
  
  pt$page_header <- get_page_header_pdf(rs)
  pt$title_hdr <- get_title_header_pdf(rs$title_hdr, rs$line_size, rs,
                                       ystart = pt$page_header$points)
  pt$titles <- get_titles_pdf(rs$titles, rs$line_size, rs, 
                              ystart = pt$page_header$points)
  
  pt$page_footer <- get_page_footer_pdf(rs)
  


  
  pt$footnotes <- c()
  if (!is.null(rs$footnotes)) {
    if (!is.null(rs$footnotes[[1]])) {
      if (rs$footnotes[[1]]$valign == "bottom")

  
        pt$footnotes <- get_footnotes_pdf(rs$footnotes, rs$line_size, rs, 
                                          footer_lines = pt$page_footer$lines)
    }
  }
  
  pt$lines <- sum(pt$page_header$lines, pt$page_footer$lines,
                  pt$title_hdr$lines, pt$titles$lines, pt$footnotes$lines)
  
  pt$points <- sum(pt$page_header$points, pt$page_footer$points,
                  pt$title_hdr$points, pt$titles$points, pt$footnotes$points)
  
  # Page by not here.  Messes up line counts.
  
  return(pt)
}

#' @import grDevices
#' @noRd
get_page_header_pdf <- function(rs) {
  
  ret <- list()
  cnt <- 0
  pnts <- 0
  
  if ((!is.null(rs$header_image_left) & is.null(rs$page_header_left)) |
      (!is.null(rs$header_image_center) & is.null(rs$page_header_center)) |
      (!is.null(rs$header_image_right) & is.null(rs$page_header_right))) {
    
    stop("`page_header` must be used when using `header_image`.")
  }
  
  # Make sure the length is 3, NA will be imputed later.
  width <- c(rs$page_header_width, rep(NA, 3 - length(rs$page_header_width)))
  
  image_left <- FALSE
  image_center <- FALSE
  image_right <- FALSE
  
  if (!is.null(rs$header_image_left)) {
    image_left <- TRUE
    hl <- rs$header_image_left
  } else{
    hl <- rs$page_header_left
  }
  
  if (!is.null(rs$header_image_right)) {
    image_right <- TRUE
    hr <- rs$header_image_right
  } else {
    hr <- rs$page_header_right
  }
  
  if (!is.null(rs$header_image_center)) {
    image_center <- TRUE
    hc <- rs$header_image_center
    
    # Default center cell is not displayed, open it when picture exists
    if (width[2] == 0) {
      width[2] <- NA
    }
  } else {
    hc <- rs$page_header_center
  }
  
  lh <- rs$row_height
  
  rb <- rs$content_size[["width"]] 
  
  # User controlled width of left column
  # lwdth <- rs$page_header_width
  # if (is.null(lwdth))
  #   lwdth <- rs$content_size[["width"]]/2
  # 
  # # Calculate right column width
  # rwdth <- rs$content_size[["width"]] - lwdth
  # lpct <- round(5000 * lwdth / rs$content_size[["width"]])
  # rpct <- round(5000 * rwdth / rs$content_size[["width"]])
  

  # maxh <- max(length(hl), length(hr))
  
  # Calculate the widths
  total_width <- sum(width, na.rm = T)
  if (total_width > rs$content_size[["width"]]) {
    
    stop(sprintf("Total width of page header %s %s cannot be greater than content width %s %s.",
                 total_width,
                 rs$units,
                 rs$content_size[["width"]],
                 rs$units))
    
  } else {
    na_num <- sum(is.na(width))
    imputed_width <- (rs$content_size[["width"]] - total_width) / na_num
    
    left_width <- ifelse(is.na(width[1]), imputed_width, width[1])
    center_width <- ifelse(is.na(width[2]), imputed_width, width[2])
    right_width <- ifelse(is.na(width[3]), imputed_width, width[3])
    
    rb1 <- left_width
    rb2 <- (left_width + center_width)
    rb3 <- (left_width + center_width + right_width)
  }
  
  hl_num <- ifelse(image_left, length(hl$image_path), length(hl))
  hc_num <- ifelse(image_center, length(hc$image_path), length(hc))
  hr_num <- ifelse(image_right, length(hr$image_path) , length(hr))
  
  maxh <- max(hl_num, hc_num, hr_num)

  if (maxh > 0) {

    pdf(NULL)
    par(family = get_font_family(rs$font), ps = rs$font_size)

    lyline <- 0 
    ryline <- 0
    cyline <- 0
    
    for (i in seq(1, maxh)) {

      lcnt <- 0
      ccnt <- 0
      rcnt <- 0
      
      if (left_width > 0) {
        if (hl_num >= i) {
          
          if (image_left == FALSE) {
            
            # Split strings if they exceed width
            tmp <- split_string_text(hl[[i]], left_width, rs$units)
            
            
            for (ln in seq_len(tmp$lines)) {
              
              ret[[length(ret) + 1]] <- page_text(tmp$text[ln], rs$font_size, 
                                                  xpos = get_points_left(0, 
                                                                         rb1,
                                                                         tmp$widths[ln],
                                                                         units = rs$units),
                                                  ypos = lyline, 
                                                  align = "left", 
                                                  alignx = 0)
              lyline <- lyline + lh
            }
            
            lcnt <- tmp$lines
          } else {
            
            # Read image
            if (i == 1) {
              lnstrt <- ceiling(lyline / lh)
              lcnt <- round((hl$height * rs$point_conversion) / lh)
            } else {
              lnstrt <- lnstrt + ceiling((hl$height * rs$point_conversion)/lh)
              # From second picture, use ceiling instead of round to leave some
              # space between pictures
              lcnt <- lcnt + ceiling((hl$height * rs$point_conversion)/lh)
            }
            
            ret[[length(ret) + 1]] <- page_image(hl$image_path[i],
                                                 height = hl$height,
                                                 width = min(hl$width, left_width),
                                                 align = "left",
                                                 line_start = lnstrt)
          }
        } else {
          
          lcnt <- 1
        }
      }
      
      if (center_width > 0) {
        if (hc_num >= i) {
          
          if (image_center == FALSE) {
            
            # Split strings if they exceed width
            tmp2 <- split_string_text(hc[[i]], center_width, rs$units)
            
            for (ln in seq_len(tmp2$lines)) {
              ret[[length(ret) + 1]] <- page_text(tmp2$text[ln], rs$font_size, 
                                                  xpos = get_points_center(rb1,
                                                                           rb2, 
                                                                           tmp2$widths[ln], 
                                                                           rs$units),
                                                  ypos = cyline, 
                                                  align = "center", 
                                                  alignx = rb1 + ((rb1 + rb2)/2))
              cyline <- cyline + lh
            }
            
            ccnt <- tmp2$lines
          } else {
            
            # Read image
            if (i == 1) {
              lnstrt <- ceiling(cyline / lh)
              ccnt <- round((hc$height * rs$point_conversion) / lh)
            } else {
              # From second picture, use ceiling instead of round to leave some
              # space between pictures
              lnstrt <- lnstrt + ceiling((hc$height * rs$point_conversion)/lh)
              ccnt <- ccnt + ceiling((hc$height * rs$point_conversion)/lh)
            }
            
            ret[[length(ret) + 1]] <- page_image(hc$image_path[i],
                                                 height = hc$height,
                                                 width = min(hc$width, center_width),
                                                 align = "center",
                                                 line_start = lnstrt)
          }
        } else {
          ccnt <- 1
        }
      }
    
      if (right_width > 0) {
        if (hr_num >= i) {
          
          if (image_right == FALSE) {
            # Split strings if they exceed width
            tmp3 <- split_string_text(hr[[i]], right_width, rs$units)
            
            for (ln in seq_len(tmp3$lines)) {
              ret[[length(ret) + 1]] <- page_text(tmp3$text[ln], rs$font_size, 
                                                  xpos = get_points_right(rb2,
                                                                          rb3, 
                                                                          tmp3$widths[ln], 
                                                                          rs$units),
                                                  ypos = ryline, 
                                                  align = "right", 
                                                  alignx = rb3)
              ryline <- ryline + lh
            }
            
            rcnt <- tmp3$lines
          } else {
            
            # Read image
            if (i == 1) {
              lnstrt <- ceiling(ryline / lh)
              rcnt <- round((hr$height * rs$point_conversion) / lh)
            } else {
              # From second picture, use ceiling instead of round to leave some
              # space between pictures
              lnstrt <- lnstrt + ceiling((hr$height * rs$point_conversion)/lh)
              rcnt <- rcnt + ceiling((hr$height * rs$point_conversion)/lh)
            }
            
            ret[[length(ret) + 1]] <- page_image(hr$image_path[i],
                                                 height = hr$height,
                                                 width = min(hr$width, right_width),
                                                 align = "right",
                                                 line_start = lnstrt)
          }
        } else {
          rcnt <- 1
        }
      }
      
      # if (lcnt > rcnt) {
      #   cnt <- cnt + lcnt
      # } else {
      #   cnt <- cnt + rcnt
      # }
      cnt <- cnt + max(lcnt, ccnt, rcnt)
      
    }

    dev.off()

    if (rs$page_header_blank_row == "below") {

      cnt <- cnt + 1
    } 

  }

  pnts <- cnt * lh

  
  res <- list(pdf = ret, 
              lines = cnt,
              points = pnts)
  
  return(res)
}

#' @import grDevices
#' @noRd
get_page_footer_pdf <- function(rs) {
  
  ret <- list()
  cnt <- 0
  pnts <- 0
  
  conv <- rs$point_conversion
  
  if ((!is.null(rs$footer_image_left) & is.null(rs$page_footer_left)) |
      (!is.null(rs$footer_image_right) & is.null(rs$page_footer_right)) |
      (!is.null(rs$footer_image_center) & is.null(rs$page_footer_center))) {
    
    stop("`page_footer` must be used when using `footer_image`.")
  }
  
  image_left <- FALSE
  image_center <- FALSE
  image_right <- FALSE
  
  if (!is.null(rs$footer_image_left)) {
    fl <- rs$footer_image_left
    image_left <- TRUE
  } else {
    fl <- rs$page_footer_left
  }
  
  if (!is.null(rs$footer_image_center)) {
    fc <- rs$footer_image_center
    image_center <- TRUE
  } else {
    fc <- rs$page_footer_center
  }
  
  if (!is.null(rs$footer_image_right)) {
    fr <- rs$footer_image_right
    image_right <- TRUE
  } else {
    fr <- rs$page_footer_right
  }
  
  lh <- rs$row_height
  
  fl_num <- ifelse(image_left, length(fl$image_path), length(fl))
  fc_num <- ifelse(image_center, length(fc$image_path), length(fc))
  fr_num <- ifelse(image_right, length(fr$image_path) , length(fr))

  maxf <- max(fl_num, fc_num, fr_num)

  if (maxf > 0) {
    
    width <- rs$page_footer_width
    
    # Make sure the length is 3
    width <- c(width, rep(NA, 3 - length(width)))
    
    total_width <- sum(width, na.rm = T)
    if (total_width > rs$content_size[["width"]]) {
      
      stop(sprintf("Total width of page footer %s %s cannot be greater than content width %s %s.",
                   total_width,
                   rs$units,
                   rs$content_size[["width"]],
                   rs$units))
      
    } else {
      na_num <- sum(is.na(width))
      imputed_width <- (rs$content_size[["width"]] - total_width) / na_num
      
      left_width <- ifelse(is.na(width[1]), imputed_width, width[1])
      center_width <- ifelse(is.na(width[2]), imputed_width, width[2])
      right_width <- ifelse(is.na(width[3]), imputed_width, width[3])
      
      rb1 <- left_width
      rb2 <- (left_width + center_width)
      rb3 <- (left_width + center_width + right_width)
    }
    
    tmp1 <- list()
    tmp2 <- list()
    tmp3 <- list()

    pdf(NULL)
    par(family = get_font_family(rs$font), ps = rs$font_size)


    # First get all wraps
    for (i in seq(1, maxf)) {

      lcnt <- 0
      ccnt <- 0
      rcnt <- 0
      
      if (fl_num >= i & left_width > 0) {

        if (image_left == FALSE) {
          # Split strings if they exceed width
          tmp1[[length(tmp1) + 1]] <- split_string_text(fl[[i]], left_width, rs$units)
          
          lcnt <- tmp1[[length(tmp1)]]$lines
        } else {
          if (i == 1) {
            # Get image lines
            lcnt <- round((fl$height * rs$point_conversion) / lh)
          } else {
            # From second picture, use ceiling instead of round to leave some
            # space between pictures
            lcnt <- lcnt + ceiling((fl$height * rs$point_conversion) / lh)
          }
        }
        
      } else {
        lcnt <- 1
      }

      if (fc_num >= i & center_width > 0) {

        if (image_center == FALSE) {
          # Split strings if they exceed width
          tmp2[[length(tmp2) + 1]] <- split_string_text(fc[[i]], center_width, rs$units)
          
          ccnt <- tmp2[[length(tmp2)]]$lines
          
        } else {
          if (i == 1) {
            # Get image lines
            ccnt <- round((fc$height * rs$point_conversion) / lh)
          } else {
            # From second picture, use ceiling instead of round to leave some
            # space between pictures
            ccnt <- ccnt + ceiling((fc$height * rs$point_conversion) / lh)
          }
        }
      } else {
        ccnt <- 1
      }

      if (fr_num >= i & right_width > 0) {

        if (image_right == FALSE) {
          tmp3[[length(tmp3) + 1]] <- split_string_text(fr[[i]], right_width, rs$units)
          
          rcnt <- tmp3[[length(tmp3)]]$lines
    
        } else {
          if (i == 1) {
            # Get image lines
            rcnt <- round((fr$height * rs$point_conversion) / lh)
          } else {
            rcnt <- rcnt + ceiling((fr$height * rs$point_conversion) / lh)
          }
        }
      } else {
        rcnt <- 1
      }

      cnt <- cnt + max(lcnt, ccnt, rcnt)

    }
    
    dev.off()
    
    # Have to determine wraps so we can calculate the height
    # of the page footer before creating any pdf.
    sy <- (rs$content_size[["height"]] * rs$point_conversion) - (cnt * lh )
    
    lyline <- sy
    cyline <- sy
    ryline <- sy
    
    # Then create pdf text or image
    
    if (left_width > 0) {
      if (image_left == FALSE) {
        # Write text
        for (i in seq(1, length(tmp1))) {
          for (ln in seq_len(tmp1[[i]]$lines)) {
            
            ret[[length(ret) + 1]] <- page_text(tmp1[[i]]$text[ln], rs$font_size, 
                                                xpos = get_points_left(0, 
                                                                       rb1,
                                                                       tmp1[[i]]$widths[ln],
                                                                       units = rs$units),
                                                ypos = lyline, 
                                                align = "left",
                                                alignx = 0)
            lyline <- lyline + lh
          }
          
        }
      } else {
        # Write image
        for (i in 1:fl_num){
          if (i == 1){
            lnstrt <- ceiling(lyline / lh)
          } else {
            lnstrt <- lnstrt + ceiling(fl$height * rs$point_conversion/lh)
          }
          
          ret[[length(ret) + 1]] <- page_image(fl$image_path[i],
                                               height = fl$height,
                                               width = min(fl$width, left_width),
                                               align = "left",
                                               line_start = lnstrt)
        }
      }
    }
    
    if (center_width > 0) {
      if (image_center == FALSE) {
        # Write text
        for (i in seq(1, length(tmp2))) {}
        for (ln in seq_len(tmp2[[i]]$lines)) {
          
          ret[[length(ret) + 1]] <- page_text(tmp2[[i]]$text[ln], rs$font_size, 
                                              xpos = get_points_center(rb1, 
                                                                       rb2,
                                                                       tmp2[[i]]$widths[ln],
                                                                       units = rs$units),
                                              ypos = cyline, 
                                              align = "center",
                                              alignx = rb1 + ((rb2 - rb1)/2))
          cyline <- cyline + lh
        }
      } else {
        # Write image
        for (i in 1:fc_num){
          if (i == 1) {
            lnstrt <- ceiling(cyline / lh)
          } else {
            lnstrt <- lnstrt + ceiling((fc$height * rs$point_conversion)/lh)
          }
          
          ret[[length(ret) + 1]] <- page_image(fc$image_path[i],
                                               height = fc$height,
                                               width = min(fc$width, center_width),
                                               align = "center",
                                               line_start = lnstrt)
        }
      }
    }
    
    if (right_width > 0) {
      if (image_right == FALSE) {
        # Write text
        for (i in seq(1, length(tmp3))) {
          for (ln in seq_len(tmp3[[i]]$lines)) {
            
            ret[[length(ret) + 1]] <- page_text(tmp3[[i]]$text[ln], rs$font_size, 
                                                xpos = get_points_right(rb2, 
                                                                        rb3,
                                                                        tmp3[[i]]$widths[ln],
                                                                        units = rs$units),
                                                ypos = ryline, 
                                                align = "right",
                                                alignx = rb3)
            ryline <- ryline + lh
          }
          
        }
      } else {
        # Write image
        for (i in 1:fr_num){
          if (i == 1) {
            lnstrt <- ceiling(ryline / lh)
          } else {
            lnstrt <- lnstrt + ceiling((fr$height * rs$point_conversion)/lh)
          }
          
          ret[[length(ret) + 1]] <- page_image(fr$image_path[i],
                                               height = fr$height,
                                               width = min(fr$width, right_width),
                                               align = "right",
                                               line_start = lnstrt)
        }
      }
    }

  }
  
  if (any(rs$page_footer_blank_row == "above"))
    cnt <- cnt + 1

  res <- list(pdf = ret,
              lines = cnt,
              points = cnt * lh)
  
  return(res)
}

#' @import grDevices
#' @noRd
get_titles_pdf <- function(ttllst, content_width, rs, 
                           talgn = "center", ystart = 0) {
  
  ret <- c()
  cnt <- 0
  pnts <- 0
  #print(ystart)
  #print("get titles")
  
  conv <- rs$point_conversion
  bs <- rs$border_spacing
  bh <- rs$border_height
  border_flag <- FALSE
  lh <- rs$line_height

  start_offset <- NULL
  
  yline <- ystart 

  if (length(ttllst) > 0) {
    
    blkcnt <- 0

    for (ttls in ttllst) {
      
      blkcnt <- blkcnt + 1
      
      cols <- ttls$columns
      

      if (ttls$width == "page")
        width <- rs$content_size[["width"]]
      else if (ttls$width == "content")
        width <- content_width
      else if (is.numeric(ttls$width))
        width <- ttls$width
      
      cwidth <- width / cols

      # Get content alignment codes
      if (talgn == "right") {
        lb <- rs$content_size[["width"]] - width
        rb <- rs$content_size[["width"]]
      } else if (talgn %in% c("center", "centre")) {
        lb <- (rs$content_size[["width"]] - width) / 2
        rb <- width + lb
      } else {
        lb <- 0
        rb <- width
      }
      
      border_flag <- FALSE
      
      fs <- rs$font_size
      if (!is.null(ttls$font_size))
        fs <- ttls$font_size
      
      lho <- get_line_height_pdf(fs)
      lh <- lho
      if (any(ttls$borders %in% c("all", "inside", "outside")))
          lh <- lh + bh

      # Open device context
      pdf(NULL)
      par(family = get_font_family(rs$font), ps = fs)
      
      brdrs <- strip_borders(ttls$borders)
      
      if (blkcnt == 1) {
        
        start_offset <- lho - rs$row_height
        
        if (any(brdrs %in% c("all", "outside", "top"))) {
          yline <- ystart + start_offset + bh
        } else {
          yline <- ystart + start_offset
        }
        
      }
      

      
      # Extra line for blank row
      if (any(ttls$blank_row %in% c("above", "both"))) {
        
        
        if (any(brdrs %in% c("all", "outside", "top"))) {
          
          if (any(brdrs %in% c("top")) & !any(brdrs %in% c("outside", "all"))){
            # Move y start up with bh to prevent from title and top border stick together
            ret[[length(ret) + 1]] <- page_hline(lb * conv,
                                                 yline - lh - bh + bs + 1,
                                                 (rb - lb) * conv)
          } else {
            ret[[length(ret) + 1]] <- page_hline(lb * conv,
                                                 yline - lh + bs + 1,
                                                 (rb - lb) * conv)
          }
          
          if (blkcnt == 1 & any(brdrs %in% c("outside")))
            pnts <- pnts + bs
        }
        
        
        
        if (any(brdrs %in% c("all", "outside", "left", "right"))) {
          ret[[length(ret) + 1]] <- page_vline(lb * conv,
                                               yline - lh + bs + 1,
                                               (1 * lh) )
          
          ret[[length(ret) + 1]] <- page_vline(rb * conv,
                                               yline - lh + bs + 1,
                                               (1 * lh))
        }
        
        yline <- yline + lh
        cnt <- cnt + 1
        pnts <- pnts + lh
        
        
        if (any(brdrs %in% c("all", "inside")) ) {

          ret[[length(ret) + 1]] <- page_hline(lb * conv,
                                               yline - lh + bs + 1,
                                               (rb - lb) * conv)
          if (blkcnt == 1)
            pnts <- pnts + bs

        }
      } else if (any(brdrs %in% c("all", "outside", "top")) ) {
        
          if (any(brdrs %in% c("top")) & !any(brdrs %in% c("outside", "all"))) {
            # Move y start up with bh to prevent from title and top border stick together
            ret[[length(ret) + 1]] <- page_hline(lb * conv,
                                                 yline - lh - bh + bs + 1,
                                                 (rb - lb) * conv)
          } else {
            ret[[length(ret) + 1]] <- page_hline(lb * conv,
                                                 yline - lh + bs + 1,
                                                 (rb - lb) * conv)
          }

          if (blkcnt == 1)
            pnts <- pnts + bs

      }
      
      
      i <- 1
      
      while (i <= length(ttls$titles)) {
        
        # Max lines per row
        mxlns <- 0
        
        # Calculate current row
        rw <- ceiling(i / cols)
        

        for (j in seq_len(cols)) { 
          
          # Not all cells have titles
          if (i > length(ttls$titles))
            vl <- ""
          else 
            vl <- ttls$titles[[i]]
          
          # Deal with column alignments
          if (cols == 1) {
            calgn <- ttls$align
          } else if (cols == 2) {
            if (j == 1)
              calgn <- "left"
            else 
              calgn <- "right"
          } else if (cols == 3) {
            if (j == 1)
              calgn <- "left"
            else if (j == 2)
              calgn <- "center"
            else if (j == 3) 
              calgn <- "right"
          }
            
          
          # Split title strings if they exceed width
          tmp <- split_string_text(vl, cwidth, rs$units)
          
          if (tmp$lines > mxlns)
            mxlns <- tmp$lines
          

          
          # Recalculate boundaries for cells
          clb <- lb + (cwidth * (j - 1))
          crb <- clb + cwidth
          
          # Row within cell pointer
          rwln <- 0
          
          for (ln in seq_len(tmp$lines)) {
            
            # print(yline) 
            ret[[length(ret) + 1]] <- page_text(tmp$text[ln], fs, 
                                                bold = ttls$bold,
                                                xpos = get_points(clb, 
                                                                  crb,
                                                                  tmp$widths[ln],
                                                                  units = rs$units,
                                                                  align = calgn),
                                                ypos = yline + rwln,
                                                titles = TRUE)
            rwln <- rwln + lh
          }
          


          
          i <- i + 1
        }
        
        
        if (any(brdrs %in% c("all", "inside")) & cols > 1) {
          
          for (z in seq_len(cols - 1)) {
            ret[[length(ret) + 1]] <- page_vline((lb + (z * cwidth)) * conv,
                                                 yline - lh + bs + 1,
                                                 mxlns * lh)
          }
        }
        
        
        # Left border
        if (any(brdrs %in% c("all", "outside", "left"))) {

          if (any(brdrs %in% c("top")) & !any(brdrs %in% c("outside", "all")) & cnt==0) {
            # For top instead of all and outside, bh is considered to connect the top border
            ret[[length(ret) + 1]] <- page_vline(lb * conv,
                                                 yline - lh - bh + bs + 1,
                                                 (mxlns * lh) + bh )
          } else {
            ret[[length(ret) + 1]] <- page_vline(lb * conv,
                                                 yline - lh + bs + 1,
                                                 (mxlns * lh) )
          }
        }

        # Right border
        if (any(brdrs %in% c("all", "outside", "right"))) {

          if (any(brdrs %in% c("top")) & !any(brdrs %in% c("outside", "all")) & cnt==0) {
            # For top instead of all and outside, bh is considered to connect the top border
            ret[[length(ret) + 1]] <- page_vline(rb * conv,
                                                 yline - lh - bh + bs + 1,
                                                 (mxlns * lh) + bh )
          } else {
            ret[[length(ret) + 1]] <- page_vline(rb * conv,
                                                 yline - lh + bs + 1,
                                                 (mxlns * lh))
          }
        }
        
        yline <- yline + (lh * mxlns)
        pnts <- pnts + (lh * mxlns)
        
        # Inside borders
        if (any(brdrs %in% c("all", "inside")) & i > 1) {
          
          ret[[length(ret) + 1]] <- page_hline(lb * conv, 
                                               yline - lh + bs + 1, 
                                               (rb - lb) * conv) 
        }

        cnt <- cnt + mxlns

      }
      
      # Extra border for blank line below
      if (any(ttls$blank_row %in% c("below", "both"))) {
        
        if (any(brdrs %in% c("all", "inside"))) {
          
          ret[[length(ret) + 1]] <- page_hline(lb * conv, 
                                               yline - lh + bs + 1, 
                                               (rb - lb) * conv) 
        }
        
        if (any(brdrs %in% c("all", "outside", "bottom"))) {
          
          ret[[length(ret) + 1]] <- page_hline(lb * conv, 
                                               yline + bs + 1, 
                                               (rb - lb) * conv) 
          
          # if (blkcnt == 1 & brdrs %in% c("outside"))
          #    pnts <- pnts + bh + bh + 1
        }
          
        if (any(brdrs %in% c("all", "outside", "left", "right"))) {
          ret[[length(ret) + 1]] <- page_vline(lb * conv,
                                               yline - lh + bs + 1,
                                               (1 * lh) )
          
          ret[[length(ret) + 1]] <- page_vline(rb * conv,
                                               yline - lh + bs + 1,
                                               (1 * lh))
          
        }
        
        
        yline <- yline + lh
        cnt <- cnt + 1
        pnts <- pnts + lh
      } else {
        #pnts <- pnts + .5 # no idea why
        if (any(brdrs %in% c("outside", "bottom"))) {
          
          ret[[length(ret) + 1]] <- page_hline(lb * conv, 
                                               yline - lh + bs + 1, 
                                               (rb - lb) * conv) 
        }
      }
      
      if (any(brdrs %in% c("outside", "all", "bottom")))
        border_flag <- TRUE
      
      dev.off()
      

      # if (blkcnt == 1) {
      #   ypos <-  ystart + start_offset - lho + bs + 1 
      #   lcnt <- cnt
      #   blkrws[blkcnt] <- cnt
      # } else {
      #   
      #   blkrws[blkcnt] <- cnt - lcnt
      #   ypos <- ypos + (blkrws[blkcnt - 1] * lh)
      #   lcnt <- cnt
      # }
      # 
      # 
      # badj <- 0
      # if (lh == lho & any(brdrs %in% c("bottom", "outside"))) {
      #   badj <- bh
      # }
      
      # print("ypos")
      # print(ypos)
      
      # Top border
      # if (any(brdrs %in% c("all", "outside", "top"))) {
      #   
      #   ret[[length(ret) + 1]] <- page_hline(lb * conv, 
      #                                        ypos , 
      #                                        (rb - lb) * conv) 
      #   if (blkcnt == 1)
      #     pnts <- pnts + bs
      #   
      # }
      # 
      # # Bottom border
      # if (any(brdrs %in% c("all", "outside", "bottom"))) {
      #   
      #     
      #     ret[[length(ret) + 1]] <- page_hline(lb * conv, 
      #                                          ypos + (blkrws * lh) + badj, 
      #                                          (rb - lb) * conv) 
      #     if (blkcnt == 1)
      #       pnts <- pnts + badj
      #     #pnts <- pnts + 2
      #     border_flag <- TRUE
      #   
      # }
      # 
      # # Left border
      # if (any(brdrs %in% c("all", "outside", "left"))) {
      #   
      #   
      #   ret[[length(ret) + 1]] <- page_vline(lb * conv, 
      #                                        ypos, 
      #                                        (blkrws * lh) + badj) 
      #   
      # }
      # 
      # # Right border
      # if (any(brdrs %in% c("all", "outside", "right"))) {
      #   
      #   
      #   ret[[length(ret) + 1]] <- page_vline(rb * conv, 
      #                                        ypos, 
      #                                        (blkrws * lh) + badj) 
      #   
      # }
      # 
      # if (any(brdrs %in% c("all", "inside")) & cols > 1) {
      #   
      #   
      #   for (k in seq_len(cols - 1)) {
      #   
      #     ret[[length(ret) + 1]] <- page_vline((lb + (cwidth * k)) * conv, 
      #                                          ypos, 
      #                                          (blkrws * lh) + badj) 
      #   }
      #   
      # }
      

    }
    
    


  }

  # pnts <- (cnt * lh) + start_offset - .5
  
  #pnts <- pnts + start_offset
  #print(pnts)
  cnts <- pnts / rs$row_height
  
  res <- list(pdf = ret, 
              lines = cnts,
              points = pnts,
              border_flag = border_flag)
  
  return(res)
}


#' @import grDevices
#' @noRd
get_titles_pdf_back <- function(ttllst, content_width, rs, 
                           talgn = "center", ystart = 0) {
  
  ret <- c()
  cnt <- 0
  pnts <- 0
  #print(ystart)
  #print("get titles")
  
  conv <- rs$point_conversion
  bs <- rs$border_spacing
  bh <- rs$border_height
  border_flag <- FALSE
  lh <- rs$line_height

  
  start_offset <- NULL
  
  yline <- ystart 
  
  if (length(ttllst) > 0) {
    
    for (ttls in ttllst) {
      
      if (ttls$width == "page")
        width <- rs$content_size[["width"]]
      else if (ttls$width == "content")
        width <- content_width
      else if (is.numeric(ttls$width))
        width <- ttls$width
      
      # Get content alignment codes
      if (talgn == "right") {
        lb <- rs$content_size[["width"]] - width
        rb <- rs$content_size[["width"]]
      } else if (talgn %in% c("center", "centre")) {
        lb <- (rs$content_size[["width"]] - width) / 2
        rb <- width + lb
      } else {
        lb <- 0
        rb <- width
      }
      
      border_flag <- FALSE
      
      fs <- rs$font_size
      if (!is.null(ttls$font_size))
        fs <- ttls$font_size
      
      lho <- get_line_height_pdf(fs)
      lh <- lho
      if (any(ttls$borders %in% c("all", "inside")))
        lh <- lh + bh
      
      # Open device context
      pdf(NULL)
      par(family = get_font_family(rs$font), ps = fs)
      
      for (i in seq_along(ttls$titles)) {
        
        brdrs <- strip_borders(ttls$borders)
        
        if (i == 1) {
          
          if (is.null(start_offset)) {
            
            start_offset <- lho - rs$row_height
            
            if (any(brdrs %in% c("all", "outside", "top"))) {
              yline <- ystart + start_offset + bh
            } else {
              yline <- ystart + start_offset
            }
            
          }
          
          # Extra line for blank row
          if (any(ttls$blank_row %in% c("above", "both"))) {
            
            
            if (any(brdrs %in% c("all", "inside"))) {
              
              ret[[length(ret) + 1]] <- page_hline(lb * conv,
                                                   yline + bs,
                                                   (rb - lb) * conv)
            }
            
            yline <- yline + lh
            cnt <- cnt + 1
            pnts <- pnts + lh
          }
        }
        
        
        # Split title strings if they exceed width
        tmp <- split_string_text(ttls$titles[[i]], width, rs$units)
        
        # Inside borders
        if (any(brdrs %in% c("all", "inside")) & i > 1) {
          
          ret[[length(ret) + 1]] <- page_hline(lb * conv, 
                                               yline - lh + bs + 1, 
                                               (rb - lb) * conv) 
        }
        
        for (ln in seq_len(tmp$lines)) {
          
          # print(yline) 
          ret[[length(ret) + 1]] <- page_text(tmp$text[ln], fs, 
                                              bold = ttls$bold,
                                              xpos = get_points(lb, 
                                                                rb,
                                                                tmp$widths[ln],
                                                                units = rs$units,
                                                                align = ttls$align),
                                              ypos = yline)
          yline <- yline + lh
          pnts <- pnts + lh
        }
        
        
        if (i == length(ttls$titles)) {
          
          # Extra border for blank line below
          if (any(ttls$blank_row %in% c("below", "both"))) {
            
            if (any(brdrs %in% c("all", "inside"))) {
              
              ret[[length(ret) + 1]] <- page_hline(lb * conv, 
                                                   yline - lh + bs + 1, 
                                                   (rb - lb) * conv) 
            }
            
            yline <- yline + lh
            cnt <- cnt + 1
            pnts <- pnts + lh
          }
          
          if (any(brdrs %in% c("outside", "all", "bottom")))
            border_flag <- TRUE
        }
        
        
        
        
        cnt <- cnt + tmp$lines
        
      }
      dev.off()
      
      ypos <-  ystart + start_offset - lho + bs + 1
      
      badj <- 0
      if (lh == lho & any(brdrs %in% c("bottom", "outside"))) {
        badj <- bh
      }
      
      # print("ypos")
      # print(ypos)
      
      # Top border
      if (any(brdrs %in% c("all", "outside", "top"))) {
        
        ret[[length(ret) + 1]] <- page_hline(lb * conv, 
                                             ypos , 
                                             (rb - lb) * conv) 
        pnts <- pnts + bs
        
      }
      
      # Bottom border
      if (any(brdrs %in% c("all", "outside", "bottom"))) {
        
        
        ret[[length(ret) + 1]] <- page_hline(lb * conv, 
                                             ypos + (cnt * lh) + badj, 
                                             (rb - lb) * conv) 
        
        pnts <- pnts + badj
        #pnts <- pnts + 2
        border_flag <- TRUE
        
      }
      
      # Left border
      if (any(brdrs %in% c("all", "outside", "left"))) {
        
        
        ret[[length(ret) + 1]] <- page_vline(lb * conv, 
                                             ypos, 
                                             (cnt * lh) + badj) 
        
      }
      
      # Right border
      if (any(brdrs %in% c("all", "outside", "right"))) {
        
        
        ret[[length(ret) + 1]] <- page_vline(rb * conv, 
                                             ypos, 
                                             (cnt * lh) + badj) 
        
      }
      
    }
    
    
    
    
  }
  
  # pnts <- (cnt * lh) + start_offset - .5
  
  #pnts <- pnts + start_offset
  #print(pnts)
  cnts <- pnts / rs$row_height
  
  res <- list(pdf = ret, 
              lines = cnts,
              points = pnts,
              border_flag = border_flag)
  
  return(res)
}

#' @import grDevices
#' @noRd
get_footnotes_pdf <- function(ftnlst, content_width, rs, 
                              talgn = "center", ystart = NULL, footer_lines = 0,
                              brdr_flag = FALSE) {
  
  ret <- c()
  cnt <- 0
  pnts <- 0
  border_flag <- FALSE
  
  olh <- rs$row_height
  lh <- olh
  conv <- rs$point_conversion
  bs <- rs$border_spacing
  bh <- rs$border_height
  pnts <- 0
  
  # print(paste0("first ystart: ", ystart))
  # print(paste0("footer lines: ", footer_lines))
  
  if (!is.null(ystart)) {
    if (brdr_flag) {
      yline <- ystart + bh - 2
      pnts <- pnts + bh - 2
    } else 
      yline <- ystart
  } else 
    yline <- 0


  if (length(ftnlst) > 0) {
    
    blocks <- list()
    blockstates <- list()
    
    for (ftnts in ftnlst) {
      
      tmp <- list() # line values
      blkst <- list() # block state

      
      cols <- ftnts$columns
      blkst$startcnt <- cnt
      blkst$cols <- cols


      if (ftnts$width == "page")
        width <- rs$content_size[["width"]]
      else if (ftnts$width == "content")
        width <- content_width
      else if (is.numeric(ftnts$width))
        width <- ftnts$width
      
      cwidth <- width / cols
      
      # Get content alignment codes
      if (talgn == "right") {
        lb <- rs$content_size[["width"]] - width
        rb <- rs$content_size[["width"]]
      } else if (talgn %in% c("center", "centre")) {
        lb <- (rs$content_size[["width"]] - width) / 2
        rb <- width + lb
      } else {
        lb <- 0
        rb <- width
      }

      alcnt <- 0
      blcnt <- 0
      border_flag <- FALSE
      brdrs <- strip_borders(ftnts$borders)

      
      if (any(brdrs %in% c("outside", "all", "top")))
        border_flag <- TRUE
      
      fs <- rs$font_size
      if (!is.null(ftnts$font_size)) {
        fs <- ftnts$font_size
      } else {
        fs <- rs$font_size
      }
        
      pdf(NULL)
      par(family = get_font_family(rs$font), ps = fs)
      
      # If all borders on, change line height to account for extra points 
      # needed for border
      if (any(brdrs %in% c("all", "inside")))
        lh <- olh + bs
      else
        lh <- olh
      
      al <- ""

      if (any(ftnts$blank_row %in% c("above", "both"))) {
        
        alcnt <- 1
        
        #yline <- yline + lh
        
        cnt <- cnt + 1
        #pnts <- pnts + lh
        
        tmp[[length(tmp) + 1]] <- list(text = "", lines = 1, widths = 0, 
                                       align = ftnts$align,
                                       line = cnt, italics = FALSE, 
                                       col = 1, clb = lb, crb = rb)
        
      }
      

      
      # b <- get_cell_borders(i + alcnt, 1,
      #                       length(ftnts$footnotes) + alcnt + blcnt,
      #                       1, ftnts$borders)
      
      i <- 1
      
      while (i <= length(ftnts$footnotes)) {
        
        # Max lines per row
        mxlns <- 0
        
        # Calculate current row
        rw <- ceiling(i / cols)

        for (j in seq_len(cols)) {
  
  
          # Not all cells have titles
          if (i > length(ftnts$footnotes))
            vl <- ""
          else 
            vl <- ftnts$footnotes[[i]]
          
          # Deal with column alignments
          if (cols == 1) {
            calgn <- ftnts$align
          } else if (cols == 2) {
            if (j == 1)
              calgn <- "left"
            else 
              calgn <- "right"
          } else if (cols == 3) {
            if (j == 1)
              calgn <- "left"
            else if (j == 2)
              calgn <- "center"
            else if (j == 3) 
              calgn <- "right"
          }
  
  
          # Split footnote strings if they exceed width
          t <- split_string_text(vl, cwidth, rs$units)
          
          if (t$lines > mxlns)
            mxlns <- t$lines
          
          # Recalculate boundaries for cells
          t$clb <- lb + (cwidth * (j - 1))
          t$crb <- t$clb + cwidth
          t$col <- j
          
          # Capture alignment for this footnote block
          t$align <- calgn
          
          # Italics setting for this footnote block
          t$italics <- ftnts$italics
          
          # Get line offset from top of footnotes block
          if (t$lines == 0) {
            t$lines <- 1
            t$line <- cnt + 1
            
          } else {
            t$line <- seq(cnt + 1, cnt + t$lines)
          }
  
  
          # Assign strings to temp variable for now
          tmp[[length(tmp) + 1]] <- t
          
          i <- i + 1
    
        }  # j cols
        
        
        # Count number of lines
        cnt <- cnt + mxlns
        
      } # i footnote strings
        
      
      bl <- ""
      
      if (any(ftnts$blank_row %in% c("below", "both"))) {
        blcnt <- 1

        cnt <- cnt + 1
        #pnts <- pnts + lh

        tmp[[length(tmp) + 1]] <- list(text = "", lines = 1, widths = 0,
                                       align = ftnts$align,
                                       line = cnt, italics = FALSE,
                                       col = 1, clb = lb, crb = rb)
      }
        
      blocks[[length(blocks) + 1]] <- tmp
      

      # Save state on these variables so they 
      # can be used to create borders
      blkst$lb <- lb
      blkst$rb <- rb
      blkst$width <- width
      blkst$cwidth <- cwidth
      blkst$lh <- lh
      blkst$brdrs <- brdrs
      blkst$alcnt <- alcnt
      blkst$blcnt <- blcnt
      blkst$endcnt <- cnt
      blockstates[[length(blockstates) + 1]] <- blkst
      
      dev.off()
      
    } # ftnts
    

    

    
    # Get starting point for report footnotes
    if (is.null(ystart)) {
      
      yline <- (rs$content_size[["height"]] * rs$point_conversion) - 
        ((cnt + footer_lines) * lh )
    }
    
    
    # Now get pdf text for each temp variable
    # for (i in seq(1, length(tmp))) {
    # 
    #   
    #   if (any(brdrs %in% c("all", "inside")) & i < length(tmp)) {
    # 
    # 
    #     if (brdr_flag)
    #       ypos <- yline - lh + bs
    #     else
    #       ypos <- yline + bh - lh + bs
    # 
    #     ret[[length(ret) + 1]] <- page_hline(lb * conv,
    #                                          ypos,
    #                                          (rb - lb) * conv)
    #   }
    #   
    #   cnt <- cnt + mxlns
    #   
    # }
    
    # Now get pdf text for each temp variable
    for (b in seq(1, length(blocks))) {
      ftnts <- ftnlst[[b]]
      blk <- blocks[[b]]
      blkst <- blockstates[[b]]
      
      mxlns <- 0
      rw <- 0
      
      for (i in seq(1, length(blk))) {
        tmp <- blk[[i]]
        
        # Track max lines per footnote row
        if (tmp$lines > mxlns | min(tmp$line) != rw) {
          mxlns <- tmp$lines
        } 
        rw <- min(tmp$line)
        
        for (ln in seq(1, tmp$lines)) {
          
          # print(paste0("lh: ", lh))
          # print(paste0("rwln: ", rwln))
          # print(paste0("yline: ", yline))
          
          if (nchar(tmp$text[ln]) > 0) {
          
            ret[[length(ret) + 1]] <- page_text(tmp$text[ln], fs, 
                                                xpos = get_points(tmp$clb, 
                                                                  tmp$crb,
                                                                  tmp$widths[ln],
                                                                  units = rs$units,
                                                                  align = tmp$align),
                                                ypos = yline + ((tmp$line[[ln]] - 1) * lh), 
                                                italics = ftnts$italics,
                                                align = tmp$align,
                                                footnotes = TRUE)
          
          }
          
        } # cell lines
        
        
        # Inside borders
        if (any(blkst$brdrs %in% c("all", "inside"))) {
          
          
          if (brdr_flag)
            ypos <- max((yline + (tmp$line * blkst$lh)) - blkst$lh + bs)
          else 
            ypos <- max((yline + (tmp$line * blkst$lh)) + bh - blkst$lh + bs)
          
          # print(paste0("line ", tmp$line))

          # Horizontal inside lines
          if (i < length(tmp) & tmp$col == 1 & max(tmp$line) < blkst$endcnt) {
            
            #print(paste0("H", i , " ", length(tmp)))
            
            ret[[length(ret) + 1]] <- page_hline(blkst$lb * conv, 
                                                 ypos, 
                                                 (blkst$rb - blkst$lb) * conv) 
          
          }
          
          # Vertical inside lines
          if (tmp$col > 1) {
            
            #print(paste0("V", i , " ", length(tmp)))
           # print(paste0("lines ", tmp$lines))
        
            ret[[length(ret) + 1]] <- page_vline(tmp$clb * conv,
                                               ypos - blkst$lh,
                                               mxlns * blkst$lh)
          
          }
        
          
        }
        
        # if (any(brdrs %in% c("all", "outside", "top"))) {
        #   
        #   ret[[length(ret) + 1]] <- page_hline(lb * conv, 
        #                                        ypos, 
        #                                        (rb - lb) * conv) 
        # }
        
      } # tmp lines 
      

      
      badj <- 0
      if (!any(blkst$brdrs %in% c("all", "inside")))
        badj <- bs
      
     # ypos <- yline  - blkst$lh # try it
      
      # Determine starting y position for borders
      if (is.null(ystart)) {
        
        ypos <- (rs$content_size[["height"]] * rs$point_conversion) - 
          ((cnt + footer_lines - blkst$alcnt) * blkst$lh ) - 
          (blkst$alcnt * blkst$lh) - blkst$lh + bs + 1 - badj
        
      } else {
        
        
        if (brdr_flag)
          ypos <- ystart - olh  
        else 
          ypos <- ystart + bh - olh  #+ badj - 4 # + bs 
      }
      
      # Top border
      if (any(blkst$brdrs %in% c("all", "outside", "top"))) {

        ret[[length(ret) + 1]] <- page_hline(lb * conv,
                                             ypos + (blkst$startcnt * blkst$lh),
                                             (blkst$rb - blkst$lb) * conv)
      }

      # Bottom border
      if (any(blkst$brdrs %in% c("all", "outside", "bottom"))) {

        ret[[length(ret) + 1]] <- page_hline(blkst$lb * conv,
                                             ypos + (blkst$endcnt * blkst$lh) + badj,
                                             (blkst$rb - blkst$lb) * conv)

        pnts <- pnts + badj
      }
    
      # Left border
      if (any(blkst$brdrs %in% c("all", "outside", "left"))) {

        ret[[length(ret) + 1]] <- page_vline(blkst$lb * conv,
                                             ypos + (blkst$startcnt * blkst$lh),
                                             ((blkst$endcnt - blkst$startcnt) * blkst$lh) + badj )
      }
      
      # Right border
      if (any(blkst$brdrs %in% c("all", "outside", "right"))) {

        ret[[length(ret) + 1]] <- page_vline(blkst$rb * conv,
                                             ypos + (blkst$startcnt * blkst$lh),
                                             ((blkst$endcnt - blkst$startcnt) * blkst$lh) + badj)
      }
      
    } # blocks
    
    yline <- yline + (lh * cnt)
    pnts <- pnts + (lh * cnt) 
        
        
        # if (is.null(ystart)) {
        #   
        #   ypos <- (rs$content_size[["height"]] * rs$point_conversion) - 
        #     ((cnt + footer_lines - alcnt) * lh ) - (alcnt * lh) - lh 
        # } else {
        # 
        # 
        #   if (brdr_flag)
        #     ypos <- ystart - olh  
        #   else 
        #     ypos <- ystart + bh - olh + bs 
        # }
        # 
        # badj <- 0
        # if (!any(brdrs %in% c("all", "inside")))
        #   badj <- bs
        # 
        # if (any(brdrs %in% c("all", "outside", "top"))) {
        #   
        #   ret[[length(ret) + 1]] <- page_hline(lb * conv, 
        #                                        ypos, 
        #                                        (rb - lb) * conv) 
        # }
        # 
        # if (any(brdrs %in% c("all", "outside", "bottom"))) {
        #   
        #   ret[[length(ret) + 1]] <- page_hline(lb * conv, 
        #                                        ypos + (cnt * lh) + badj, 
        #                                        (rb - lb) * conv) 
        #   
        #   pnts <- pnts + badj
        # }
        
        # if (any(brdrs %in% c("all", "outside", "left"))) {
        #   
        #   ret[[length(ret) + 1]] <- page_vline(lb * conv, 
        #                                        ypos, 
        #                                        (cnt * lh) + badj ) 
        # }
        
        # if (any(brdrs %in% c("all", "outside", "right"))) {
        # 
        #   ret[[length(ret) + 1]] <- page_vline(rb * conv, 
        #                                        ypos, 
        #                                        (cnt * lh) + badj) 
        # }
      
     

      
    
    
  } # length(ftnlst) > 0

  res <- list(pdf = ret,
              lines = pnts / olh,
              points = pnts,
              border_flag = border_flag)
  
  return(res)
}


#' @import grDevices
#' @noRd
get_footnotes_pdf_back <- function(ftnlst, content_width, rs, 
                              talgn = "center", ystart = NULL, footer_lines = 0,
                              brdr_flag = FALSE) {
  
  ret <- c()
  cnt <- 0
  pnts <- 0
  border_flag <- FALSE
  
  olh <- rs$row_height
  lh <- olh
  conv <- rs$point_conversion
  bs <- rs$border_spacing
  bh <- rs$border_height
  pnts <- 0
  
  print(paste0("first ystart: ", ystart))
  print(paste0("footer lines: ", footer_lines))
  
  if (!is.null(ystart)) {
    if (brdr_flag) {
      yline <- ystart + bh - 2
      pnts <- pnts + bh - 2
    } else 
      yline <- ystart
  } else 
    yline <- 0
  
  
  if (length(ftnlst) > 0) {
    
    for (ftnts in ftnlst) {
      
      tmp <- list()
      
      if (ftnts$width == "page")
        width <- rs$content_size[["width"]]
      else if (ftnts$width == "content")
        width <- content_width
      else if (is.numeric(ftnts$width))
        width <- ftnts$width
      
      # Get content alignment codes
      if (talgn == "right") {
        lb <- rs$content_size[["width"]] - width
        rb <- rs$content_size[["width"]]
      } else if (talgn %in% c("center", "centre")) {
        lb <- (rs$content_size[["width"]] - width) / 2
        rb <- width + lb
      } else {
        lb <- 0
        rb <- width
      }
      
      alcnt <- 0
      blcnt <- 0
      border_flag <- FALSE
      brdrs <- strip_borders(ftnts$borders)
      
      
      pdf(NULL)
      par(family = get_font_family(rs$font), ps = rs$font_size)
      
      for (i in seq_along(ftnts$footnotes)) {
        
        # If all borders on, change line height to account for extra points 
        # needed for border
        if (any(brdrs %in% c("all", "inside")))
          lh <- olh + bs
        else
          lh <- olh
        
        al <- ""
        if (i == 1) {
          if (any(ftnts$blank_row %in% c("above", "both"))) {
            
            alcnt <- 1
            
            yline <- yline + lh
            
            cnt <- cnt + 1
            pnts <- pnts + lh
            
          }
        }
        
        bl <- ""
        if (i == length(ftnts$footnotes)) {
          if (any(ftnts$blank_row %in% c("below", "both"))) {
            blcnt <- 1
            
            cnt <- cnt + 1
            pnts <- pnts + lh
          }
          
          if (any(brdrs %in% c("outside", "all", "top")))
            border_flag <- TRUE
          
        }
        
        # b <- get_cell_borders(i + alcnt, 1,
        #                       length(ftnts$footnotes) + alcnt + blcnt,
        #                       1, ftnts$borders)
        
        
        
        # Split footnote strings if they exceed width
        t <- split_string_text(ftnts$footnotes[[i]], width, rs$units)
        
        # Capture alignment for this footnote block
        t$align <- ftnts$align
        
        # Count number of lines
        cnt <- cnt + t$lines
        
        
        # Assign strings to temp variable for now
        tmp[[length(tmp) + 1]] <- t
      }
      
      dev.off()
      
      if (is.null(ystart)) {
        
        yline <- (rs$content_size[["height"]] * rs$point_conversion) - 
          ((cnt + footer_lines - alcnt) * lh )
      }
      
      
      
      # Now get pdf text for each temp variable
      for (i in seq(1, length(tmp))) {
        for (ln in seq_len(tmp[[i]]$lines)) {
          
          ret[[length(ret) + 1]] <- page_text(tmp[[i]]$text[ln], rs$font_size, 
                                              xpos = get_points(lb, 
                                                                rb,
                                                                tmp[[i]]$widths[ln],
                                                                units = rs$units,
                                                                align = tmp[[i]]$align),
                                              ypos = yline, 
                                              italics = ftnts$italics,
                                              align = tmp[[i]]$align,
                                              footnotes = TRUE)
          
          
          
          
          yline <- yline + lh
          pnts <- pnts + lh
        }
        
        if (any(brdrs %in% c("all", "inside")) & i < length(tmp)) {
          
          
          if (brdr_flag)
            ypos <- yline - lh + bs 
          else 
            ypos <- yline + bh - lh + bs 
          
          ret[[length(ret) + 1]] <- page_hline(lb * conv, 
                                               ypos, 
                                               (rb - lb) * conv) 
        }
        
      }
      
      
      if (is.null(ystart)) {
        
        ypos <- (rs$content_size[["height"]] * rs$point_conversion) - 
          ((cnt + footer_lines - alcnt) * lh ) - (alcnt * lh) - lh 
      } else {
        
        
        if (brdr_flag)
          ypos <- ystart - olh  
        else 
          ypos <- ystart + bh - olh + bs 
      }
      
      badj <- 0
      if (!any(brdrs %in% c("all", "inside")))
        badj <- bs
      
      if (any(brdrs %in% c("all", "outside", "top"))) {
        
        ret[[length(ret) + 1]] <- page_hline(lb * conv, 
                                             ypos, 
                                             (rb - lb) * conv) 
      }
      
      if (any(brdrs %in% c("all", "outside", "bottom"))) {
        
        ret[[length(ret) + 1]] <- page_hline(lb * conv, 
                                             ypos + (cnt * lh) + badj, 
                                             (rb - lb) * conv) 
        
        pnts <- pnts + badj
      }
      
      if (any(brdrs %in% c("all", "outside", "left"))) {
        
        ret[[length(ret) + 1]] <- page_vline(lb * conv, 
                                             ypos, 
                                             (cnt * lh) + badj ) 
      }
      
      if (any(brdrs %in% c("all", "outside", "right"))) {
        
        ret[[length(ret) + 1]] <- page_vline(rb * conv, 
                                             ypos, 
                                             (cnt * lh) + badj) 
      }
      
    }
    
  }
  
  res <- list(pdf = ret,
              lines = pnts / olh,
              points = pnts,
              border_flag = border_flag)
  
  return(res)
}

#' @import grDevices
#' @noRd
get_title_header_pdf <- function(thdrlst, content_width, rs, 
                                 talgn = "center", ystart = 0) {
  
  ret <- c()
  cnt <- 0
  border_flag <- FALSE

  lh <- rs$row_height
  conv <- rs$point_conversion
  bs <- rs$border_spacing
  bh <- rs$border_height
  startpos <- ystart
  tcnt <- 0
  hcnt <- 0
  pnts <- 0
  counter <- 0


  if (length(thdrlst) > 0) {

    for (ttlhdr in thdrlst) {
      
      counter <- counter + 1

      if (ttlhdr$width == "page")
        width <- rs$content_size[["width"]]
      else if (ttlhdr$width == "content")
        width <- content_width
      else if (is.numeric(ttlhdr$width))
        width <- ttlhdr$width
      
      # Get content alignment codes
      if (talgn == "right") {
        lb <- rs$content_size[["width"]] - width
        rb1 <- rs$content_size[["width"]]
      } else if (talgn %in% c("center", "centre")) {
        lb <- (rs$content_size[["width"]] - width) / 2
        rb1 <- width + lb
      } else {
        lb <- 0
        rb1 <- width
      }
      rb2 <- rb1 * .7
      splitx <- (lb + ((rb1 - lb) * .7))  * conv
      brdrs <- strip_borders(ttlhdr$borders)
      
      if (counter == 1) {
        if (any(brdrs %in% c("all", "top", "outside"))) {
          
          startpos <- ystart + bs - 1
          
        } else {
          
          startpos <- ystart
          
        }
        lyline <- startpos
        ryline <- startpos
      }
      
      
      lh <- rs$line_height
      if (any(ttlhdr$borders %in% c("all", "inside")))
        lh <- lh + bh


      mx <- max(length(ttlhdr$titles), length(ttlhdr$right))

      alcnt <- 0
      blcnt <- 0
      border_flag <- FALSE


      pdf(NULL)
      par(family = get_font_family(rs$font), ps = rs$font_size)
      
      for(i in seq_len(mx)) {

        al <- ""
        if (i == 1) {
          
          
          if (any(ttlhdr$blank_row %in% c("above", "both"))) {

            alcnt <- 1
            
            if (any(brdrs %in% c("all", "inside"))) {
              
              ret[[length(ret) + 1]] <- page_hline(lb * conv, 
                                                   startpos + bs, 
                                                   (rb1 - lb) * conv) 
            }

            lyline <- lyline + lh
            ryline <- ryline + lh

            cnt <- cnt + 1

          }
        }

        if (length(ttlhdr$titles) >= i) {
          # Split strings if they exceed width
          tmp1 <- split_string_text(ttlhdr$titles[[i]], width * .7, rs$units)
          
        
          
          for (ln in seq_len(tmp1$lines)) {
            
            ret[[length(ret) + 1]] <- page_text(tmp1$text[ln], rs$font_size, 
                                                xpos = get_points(lb, 
                                                                  rb2,
                                                                  tmp1$widths[ln],
                                                                  units = rs$units,
                                                                  align = "left"),
                                                ypos = lyline, 
                                                align = "left",
                                                alignx = 0)
            lyline <- lyline + lh
          }
          
          if (any(brdrs %in% c("all", "inside")) ) {
            
            ret[[length(ret) + 1]] <- page_hline(lb * conv, 
                                                 lyline - lh + bs + 1, 
                                                 splitx - (lb * conv)) 
          }
          
          tcnt <- tcnt + tmp1$lines
        } else {

          tcnt <- tcnt + 0
        }

        if (length(ttlhdr$right) >= i) {
          tmp2 <- split_string_text(ttlhdr$right[[i]],
                                   width * .3, rs$units)
          
          
          for (ln in seq_len(tmp2$lines)) {
            
            ret[[length(ret) + 1]] <- page_text(tmp2$text[ln], rs$font_size, 
                                                xpos = get_points(rb2, 
                                                                  rb1,
                                                                  tmp2$widths[ln],
                                                                  units = rs$units,
                                                                  align = "right"),
                                                ypos = ryline, 
                                                align = "right",
                                                alignx = rb1)
            ryline <- ryline + lh
          }
          
          if (any(brdrs %in% c("all", "inside"))) {

            ret[[length(ret) + 1]] <- page_hline(splitx,
                                                 ryline - lh + bs + 1,
                                                 (rb1* conv) - splitx)
          }
          
          hcnt <- hcnt + tmp2$lines
        } else {

          hcnt <- hcnt + 0
        }
        
      }
      
      if (tcnt > hcnt)
        cnt <- cnt + tcnt
      else
        cnt <- cnt + hcnt
      
    
      if (any(ttlhdr$blank_row %in% c("below", "both"))) {
        blcnt <- 1
        
        if (any(brdrs %in% c("all", "inside"))) {
          
          
          ret[[length(ret) + 1]] <- page_hline(lb * conv,
                                               (startpos + (cnt * lh) - lh) + bs + 1,
                                               (rb1 - lb) * conv)
        }
        
        lyline <- lyline + lh
        ryline <- ryline + lh
        
        cnt <- cnt + 1
      }
      
      if (any(brdrs %in% c("all", "outside", "bottom")))
        border_flag <- TRUE
      

      dev.off()
      
      if (any(brdrs %in% "bottom"))
        pnts <- (cnt * lh) + bh
      else
        pnts <- cnt * lh
      
     pnts <- pnts + (startpos - ystart) + 1
     
     ypos <-  ystart - rs$line_height + bs + 1
     
     badj <- 0
     if (lh == rs$line_height & any(brdrs %in% c("bottom", "outside"))) {
       badj <- bh
     }
     

      # Top border
      if (any(brdrs %in% c("all", "outside", "top"))) {
        
        ret[[length(ret) + 1]] <- page_hline(lb * conv, 
                                             ypos, 
                                             (rb1 - lb) * conv) 
        
      }
      
      #  Bottom border
      if (any(brdrs %in% c("all", "outside", "bottom"))) {
        
        ret[[length(ret) + 1]] <- page_hline(lb * conv, 
                                             ypos + (cnt * lh) + badj, 
                                             (rb1 - lb) * conv) 
        
        pnts <- pnts + badj
        
      }
      
      # Left border
      if (any(brdrs %in% c("all", "outside", "left"))) {
        
        
        ret[[length(ret) + 1]] <- page_vline(lb * conv, 
                                             ypos, 
                                             (cnt * lh) + badj) 
        
      }
      
      # Right border
      if (any(brdrs %in% c("all", "outside", "right"))) {
        
        
        ret[[length(ret) + 1]] <- page_vline(rb1 * conv, 
                                             ypos, 
                                             (cnt * lh) + badj) 
        
      }
      
     # Middle border
      if (any(brdrs %in% c("all", "inside"))) {
        
        bldiff <- blcnt - alcnt
        if (bldiff < 0)
          bldiff <- 0
        
        ret[[length(ret) + 1]] <- page_vline(splitx, 
                                             ypos + (alcnt * lh), 
                                             ((cnt - alcnt - blcnt) * lh) ) 
        
      }
      
    }

  }

  res <- list(pdf = ret,
              lines = cnt,
              points = pnts,
              border_flag = border_flag)

  
  return(res)
}


#' Get page by text strings suitable for printing
#' @import stringi
#' @param titles Page by object
#' @param width The width to set the page by strings to
#' @return A vector of strings
#' @noRd
get_page_by_pdf <- function(pgby, width, value, rs, talgn, ystart = 0, 
                            brdr_flag = FALSE, pgby_cnt = NULL) {
  
  if (is.null(width)) {
    stop("width cannot be null.") 
    
  }
  
  if (is.null(value))
    value <-  get_pgby_value(value, pgby_cnt)

  ret <- c()
  cnt <- 0
  border_flag <- FALSE
  lh <- rs$line_height
  conv <- rs$point_conversion
  bs <- rs$border_spacing
  bh <- rs$border_height
  pnts <- 0

  if (!is.null(pgby)) {

    if (!any(class(pgby) == "page_by"))
      stop("pgby parameter value is not a page_by.")
    
    # Get content alignment codes
    if (talgn == "right") {
      lb <- rs$content_size[["width"]] - width
      rb <- rs$content_size[["width"]]
    } else if (talgn %in% c("center", "centre")) {
      lb <- (rs$content_size[["width"]] - width) / 2
      rb <- width + lb
    } else {
      lb <- 0
      rb <- width
    }
    
    brdrs <- strip_borders(pgby$borders)
    
    if (any(brdrs %in% c("all", "inside"))) {
      lh <- lh + bh 
    }
  
    if (any(brdrs %in% c("all", "outside", "top")) & !brdr_flag) {
      yline <- ystart + bh
      pnts <- pnts + bh
    } else {
      
      yline <- ystart
    }
    
    # Open device context
    pdf(NULL)
    par(family = get_font_family(rs$font), ps = rs$font_size)


    if (pgby$blank_row %in% c("above", "both")) {

      if (any(brdrs %in% c("all", "inside"))) {
        ret[[length(ret) + 1]] <- page_hline(lb * conv,
                                             yline + bs,
                                             (rb - lb) * conv)
      }
      
      yline <- yline + lh
      cnt <- cnt + 1
      pnts <- pnts + lh
    }
    
    if (pgby$bold %in% c(TRUE, FALSE)) {
      
      if (substr(pgby$label, nchar(pgby$label), nchar(pgby$label)) != " "){
        sep <- " "
      } else {
        sep <- ""
      }
      
      vl <- paste0( pgby$label, sep, value)
      
      tmp <- split_string_text(vl, width, rs$units)
      
      
      for (ln in seq_len(tmp$lines)) {
        
        x_adjust <- 0
        
        if (pgby$bold == TRUE & pgby$align == "right") {
          x_adjust <- -3
        }
        
        ret[[length(ret) + 1]] <- page_text(tmp$text[ln], rs$font_size, 
                                            bold = pgby$bold,
                                            xpos = get_points(lb, 
                                                              rb,
                                                              tmp$widths[ln],
                                                              units = rs$units,
                                                              align = pgby$align) + x_adjust,
                                            ypos = yline)
        yline <- yline + lh
        cnt <- cnt + 1
        pnts <- pnts + lh
      }
    } else if (pgby$bold %in% c("label", "value")) {
      
      label_bold <- ifelse(pgby$bold == "label", TRUE, FALSE)
      value_bold <- ifelse(pgby$bold == "value", TRUE, FALSE)
      
      # Process label first
      tmp_label <- split_string_text(pgby$label, width, rs$units)
      last_label_width <- tmp_label$widths[length(tmp_label$widths)]
      
      # First value split with remaining width
      tmp_value <- split_string_text(value, width - last_label_width, rs$units)
      
      # If the first part is less than remaining with, it means value will 
      # be concatenated to label in a line; otherwise, the value will start a new line
      if (tmp_value$widths[1] <= width - last_label_width) {
        first_value_width <- tmp_value$widths[1]
        
        for (ln in seq_len(tmp_label$lines)) {
          
          if (ln < tmp_label$lines) {
            ret[[length(ret) + 1]] <- page_text(tmp_label$text[ln], rs$font_size, 
                                                bold = label_bold,
                                                xpos = get_points(lb, 
                                                                  rb,
                                                                  tmp_label$widths[ln],
                                                                  units = rs$units,
                                                                  align = pgby$align),
                                                ypos = yline)
          } else {
            # ----------------------------------------------------------- #
            # Last label and first value are processed together
            # ----------------------------------------------------------- #
            
            # Adjust xpos of value for some scenarios
            x_value_adjust <- 0
            
            # The second label and value are too close
            # And center label and value are too close
            if ( (ln > 1) |
                 (label_bold & pgby$align %in% c("center", "centre")) |
                 (ln == 1 & label_bold & pgby$align == "left")
                 ) {
              x_value_adjust <- 6
            }
            
            if (pgby$align == "left") {
              # ---- Label ---- # 
              ret[[length(ret) + 1]] <- page_text(tmp_label$text[ln], rs$font_size, 
                                                  bold = label_bold,
                                                  xpos = get_points(lb, 
                                                                    rb,
                                                                    tmp_label$widths[ln],
                                                                    units = rs$units,
                                                                    align = pgby$align),
                                                  ypos = yline)
              
              # ---- Value ---- #
              ret[[length(ret) + 1]] <- page_text(tmp_value$text[1], rs$font_size, 
                                                  bold = value_bold,
                                                  xpos = get_points(lb + last_label_width, 
                                                                    rb,
                                                                    tmp_value$widths[1],
                                                                    units = rs$units,
                                                                    align = pgby$align) + x_value_adjust,
                                                  ypos = yline)
              
            } else if (pgby$align == c("right")) {
              # ---- Label ---- # 
              x_label_right_adjust <- ifelse(label_bold, 4, 0)
              ret[[length(ret) + 1]] <- page_text(tmp_label$text[ln], rs$font_size, 
                                                  bold = label_bold,
                                                  xpos = get_points(lb, 
                                                                    rb,
                                                                    tmp_label$widths[ln] + first_value_width,
                                                                    units = rs$units,
                                                                    align = pgby$align) - x_label_right_adjust,
                                                  ypos = yline)
              # ---- Value ---- #
              ret[[length(ret) + 1]] <- page_text(tmp_value$text[1], rs$font_size, 
                                                  bold = value_bold,
                                                  xpos = get_points(lb, 
                                                                    rb,
                                                                    tmp_value$widths[1],
                                                                    units = rs$units,
                                                                    align = pgby$align) + x_value_adjust,
                                                  ypos = yline)
              
            } else if (pgby$align  %in% c("center", "centre")) {
              # ---- Label ---- # 
              ret[[length(ret) + 1]] <- page_text(tmp_label$text[ln], rs$font_size, 
                                                  bold = label_bold,
                                                  xpos = get_points(lb, 
                                                                    rb,
                                                                    tmp_label$widths[ln] + first_value_width,
                                                                    units = rs$units,
                                                                    align = pgby$align),
                                                  ypos = yline)
              # ---- Value ---- #
              label_value_width <- tmp_label$widths[ln] + first_value_width
              xpos_value <- cpoints(lb + ((rb - lb)/ 2) - (label_value_width/ 2) + tmp_label$widths[ln], rs$units) + 1
              
              ret[[length(ret) + 1]] <- page_text(tmp_value$text[1], rs$font_size, 
                                                  bold = value_bold,
                                                  xpos = xpos_value + x_value_adjust,
                                                  ypos = yline)
            }
          }
          yline <- yline + lh
          cnt <- cnt + 1
          pnts <- pnts + lh
        }
        
        # Second value split with whole width
        remain_value <- trimws(sub(tmp_value$text[1], "", value), which = "left")
        if (nchar(remain_value) > 0) {
          tmp_value2 <- split_string_text(remain_value, width, rs$units)
          
          tmp_value$text <- c(tmp_value$text[1], tmp_value2$text)
          tmp_value$lines <- 1 + tmp_value2$lines
          
          for (ln in 2:tmp_value$lines) {
            
            ret[[length(ret) + 1]] <- page_text(tmp_value$text[ln], rs$font_size,
                                                bold = value_bold,
                                                xpos = get_points(lb, 
                                                                  rb,
                                                                  tmp_value$widths[ln],
                                                                  units = rs$units,
                                                                  align = pgby$align),
                                                ypos = yline)
            
            yline <- yline + lh
            cnt <- cnt + 1
            pnts <- pnts + lh
          }
        }
        
      } else {
        # Case when Value starts a new line in beginning
        for (ln in seq_len(tmp_label$lines)) {
          ret[[length(ret) + 1]] <- page_text(tmp_label$text[ln], rs$font_size, 
                                              bold = label_bold,
                                              xpos = get_points(lb, 
                                                                rb,
                                                                tmp_label$widths[ln],
                                                                units = rs$units,
                                                                align = pgby$align),
                                              ypos = yline)
          yline <- yline + lh
          cnt <- cnt + 1
          pnts <- pnts + lh
        }
        
        tmp_value <- split_string_text(value, width, rs$units)
        for (ln in seq_len(tmp_value$lines)) {
          ret[[length(ret) + 1]] <- page_text(tmp_value$text[ln], rs$font_size,
                                              bold = value_bold,
                                              xpos = get_points(lb, 
                                                                rb,
                                                                tmp_value$widths[ln],
                                                                units = rs$units,
                                                                align = pgby$align),
                                              ypos = yline)
          
          yline <- yline + lh
          cnt <- cnt + 1
          pnts <- pnts + lh
        }
      }
    }
    
    dev.off()

    if (pgby$blank_row %in% c("below", "both")) {
      
      
      if (any(brdrs %in% c("all", "inside"))) {
        ret[[length(ret) + 1]] <- page_hline(lb * conv,
                                             yline - lh + bs,
                                             (rb - lb) * conv)
      }
      
      yline <- yline + lh
      cnt <- cnt + 1
      pnts <- pnts + lh
    }
    
    if (any(brdrs %in% c("all", "inside"))) {
      if (brdr_flag)
        ypos <-  ystart - lh + 3 
      else 
        ypos <- ystart - lh + bs + 2
      
    } else 
      ypos <-  ystart - lh + bs + 1
    
    # Top border
    if (any(brdrs %in% c("all", "outside", "top"))) {

      ret[[length(ret) + 1]] <- page_hline(lb * conv,
                                           ypos,
                                           (rb - lb) * conv)

    }

    # Bottom border
    if (any(brdrs %in% c("all", "outside", "bottom"))) {

      ret[[length(ret) + 1]] <- page_hline(lb * conv,
                                           ypos  + (cnt * lh),
                                           (rb - lb) * conv)
      border_flag <- TRUE
    }

    # Left border
    if (any(brdrs %in% c("all", "outside", "left"))) {


      ret[[length(ret) + 1]] <- page_vline(lb * conv,
                                           ypos,
                                           (cnt * lh)) 

    }

    # Right border
    if (any(brdrs %in% c("all", "outside", "right"))) {


      ret[[length(ret) + 1]] <- page_vline(rb * conv,
                                           ypos,
                                           (cnt * lh) )

    }

  }

  
  res <- list(pdf = ret, 
              lines = pnts / lh,
              points = pnts,
              border_flag = border_flag)
  
  return(res)
}

# Utilities ---------------------------------------------------------------



get_page_numbers_pdf <- function(txt, pg, tpg) {
  
  ret <- txt
  
  ret <- gsub("[pg]", pg, ret, fixed = TRUE)
  
  ret <- gsub("[tpg]", tpg, ret, fixed = TRUE)
  
  return(ret)
}


#' Function to temporarily minimize border specs for PDF
#' @noRd
strip_borders <- function(bspec) {
  
 ret <- bspec
 
 # if ("all" %in% ret)
 #   ret <- append(ret[ret != "all"], c("top", "bottom"))
 # if ("inside" %in% ret)
 #   ret <- ret[ret != "inside"]
 # if ("outside" %in% ret)
 #   ret <- append(ret[ret != "outside"], c("top", "bottom"))
 # if ("body" %in% ret)
 #   ret <- append(ret[ret != "body"], c("top", "bottom"))
 # if ("left" %in% ret)
 #   ret <- ret[ret != "left"]
 # if ("right" %in% ret)
 #   ret <- ret[ret != "right"]
  
 return(ret)
}

# Style class -------------------------------------------------------------

#' @title 
#' Creates a style object
#' @description 
#' This function will create a style object to control background colors and 
#' font settings on your report.  The style object can be applied to a 
#' report using the \code{\link{add_style}} function. Currently, styles may only
#' be applied to HTML reports.
#' @details
#' The style object contains style settings for a report. The style object 
#' allows you to control background colors and font specifications such as 
#' font size, font color, and font bold.  The style object can be created
#' once and reused on many reports.  See the \code{\link{add_style}} function
#' to learn how to add the style object to a report.  
#' 
#' Note that styles will be applied uniformly to the entire report.  Also note
#' that at present, styles can be used only on HTML output types. Future versions
#' of the \strong{reporter} package will provide style support for other 
#' output types.
#' 
#' On the style object, colors for background and fonts may be passed as 
#' an RGB hex code or a color name. For example, the color red may be passed as
#' the hex code "#FF0000" or as the color name "Red". Color names must conform
#' to HTML/CSS standards.  These color names can be easily discovered on 
#' the internet.  A sample of common color names is presented below.
#' 
#' @section Color Names:
#' Many of the parameters on the style object accept a color name or code.
#' The values accepted for these parameters follow standard HTML/CSS style
#' color values.  Below is a sample of common color names
#' that can be used to specify colors with the \code{create_style} function.
#' These color names should be passed as a quoted string:
#' \itemize{
#'   \item \strong{Primary and Secondary Colors:} Black, White, Red, Yellow, 
#'   Blue, Green, Orange, Purple and Brown.
#'   \item \strong{Common Shades:} Beige, Crimson, Gold, Indigo, Ivory, Lavender, 
#'   Maroon, Navy, Olive, Pink, 
#'   Plum, RoyalBlue, Silver, Teal, Turquoise, Violet
#'   \item \strong{Shades of White:} AntiqueWhite, Azure, GhostWhite, 
#'   SeaShell, Snow, WhiteSmoke
#'   \item \strong{Shades of Grey:} Grey, Gray, DarkGray, DarkGrey, DimGray, DimGrey, 
#'   LightGray, LightGrey, SlateGray, SlateGrey
#'   \item \strong{Shades of Blue:} AliceBlue, CadetBlue, CornflowerBlue, 
#'   DodgerBlue, PowderBlue, LightBlue, MidnightBlue, SkyBlue, SlateBlue, SteelBlue
#'   \item \strong{Earth Colors:} Beige, Bisque, BurlyWood, ForestGreen, Khaki,
#'    Linen, SandyBrown, SaddleBrown, Salmon, SeaGreen, Sienna, Tan, Thistle, Wheat
#'   \item \strong{Bright Colors:} Aqua, Aquamarine, BlueViolet, Cyan, Fuchia, 
#'   HotPink, Lime, Magenta, OrangeRed, SpringGreen
#' }
#' @param font_name The name of the font to use on the report.  Valid values
#' are "Courier", "Arial", or "Times".  The default is "Courier".
#' @param font_size The default font size to use for the report.  This font
#' size will be used for all text, unless overridden by another font size 
#' parameter.
#' @param text_color The default color to use for all text in the report.  This
#' parameter will apply to the entire report, unless overridden by other
#' font color settings.
#' @param background_color The color to use for the background of the report.
#' This color will appear everywhere on the document unless overridden 
#' by another color specification.  
#' @param title_font_size The font size to use for the title font in points.
#' @param title_font_bold Whether to bold the title or not.  Valid values 
#' are TRUE or FALSE. By default, the title will not be bold.
#' @param title_font_color The color to use for the title font.
#' @param title_background The background color for the title block.
#' @param footnote_font_bold Whether to bold the footnote or not.  Valid values 
#' are TRUE or FALSE. By default, the footnote will not be bold.
#' @param footnote_font_color The font color to use for footnotes.
#' @param footnote_background The color to be used for the background of 
#' footnotes.
#' @param border_color The color to use for all borders in the report.  
#' By default, the border will be black.
#' @param table_header_background The background color to use in the table header.
#' This color may be different than the background color used in the table
#' body.
#' @param table_header_font_bold Whether to bold the header labels or not.  
#' Valid values are TRUE and FALSE. By default, the header will not be bold.
#' @param table_header_font_color The font color to use on the table header.
#' @param table_body_background The background color to use in the body of any 
#' table in the report. 
#' @param table_body_stripe The background color to use for every other row
#' in a table.  The stripe color is used in conjunction with the body background
#' color to perform table striping.  The stripe color will start on the second row.
#' @param table_body_font_color The font color to use for the body of any 
#' table in the report.  
#' @param table_stub_background The background color to use for the stub column,
#' if one exists on the table. 
#' @param table_stub_font_color The font color to be used for the stub column,
#' if one exists on the table.
#' @param table_stub_font_bold Whether or not to bold the stub column.  Valid
#' values are TRUE and FALSE.
#' @param table_label_row_bold Whether or not to bold the label row on a stub
#' column.  This parameter is so you can bold the label row, but not the 
#' values in the stub column.  Valid values are TRUE and FALSE.
#' @family styles
#' @examples 
#' library(reporter)
#' library(magrittr)
#' 
#' # Prepare data
#' dat <- data.frame(stub = rownames(mtcars), mtcars)
#' dat <- dat[1:15, ]
#' 
#' # Create temp file path
#' tmp <- file.path(tempdir(), "HairAndEyes2.html")
#' 
#' # Define custom style
#' sty <- create_style(font_name = "Arial",
#'                     font_size = 10,
#'                     background_color = "WhiteSmoke",
#'                     border_color = "Grey",
#'                     title_font_size = 12,
#'                     title_font_bold = TRUE,
#'                     title_font_color = "SteelBlue",
#'                     table_header_background = "Tan",
#'                     table_header_font_bold = TRUE,
#'                     table_header_font_color = "White",
#'                     table_body_background = "White",
#'                     table_body_stripe = "Wheat", 
#'                     table_stub_background = "Tan",
#'                     table_stub_font_color = "White")
#' 
#' # Create table object
#' tbl <- create_table(dat, borders = "all") %>% 
#' titles("MTCARS Dataset With Style") %>% 
#' column_defaults(width = .5) %>% 
#' define(stub, label = "Car Make and Model", width = 1.5)
#' 
#' # Create report and add style spec
#' rpt <- create_report(tmp, output_type = "HTML") %>% 
#'        add_content(tbl) %>% 
#'        add_style(style = sty)
#'
#' # Write out the report        
#' write_report(rpt)
#' 
#' # Uncomment to View report
#' # file.show(tmp)
#' @export
create_style <- function(font_name = NULL,
                         font_size = NULL,
                         text_color = NULL,
                         background_color = NULL,
                         title_font_size = NULL,
                         title_font_bold = NULL,
                         title_font_color = NULL,
                         title_background = NULL,
                         footnote_font_bold = NULL,
                         footnote_font_color = NULL,
                         footnote_background = NULL,
                         border_color = NULL,
                         table_header_background = NULL,
                         table_header_font_bold = NULL,
                         table_header_font_color = NULL,
                         table_body_background = NULL,
                         table_body_stripe = NULL,
                         table_body_font_color = NULL,
                         table_stub_background = NULL,
                         table_stub_font_color = NULL,
                         table_stub_font_bold = NULL,
                         table_label_row_bold = NULL) {


  ret <- structure(list(), class = c("style_spec", "list"))
  
  ret$font_name <- font_name
  ret$font_size <- font_size
  ret$text_color <- text_color
  ret$background_color <- background_color
  ret$title_font_size <- title_font_size
  ret$title_font_bold <- title_font_bold
  ret$title_font_color <- title_font_color
  ret$title_background <- title_background
  ret$footnote_font_bold <- footnote_font_bold
  ret$footnote_font_color <- footnote_font_color
  ret$footnote_background <- footnote_background
  ret$border_color <- border_color
  ret$table_header_background <- table_header_background
  ret$table_header_font_bold <- table_header_font_bold
  ret$table_header_font_color <- table_header_font_color
  ret$table_body_background <- table_body_background
  ret$table_body_stripe <- table_body_stripe
  ret$table_stub_background <- table_stub_background
  ret$table_stub_font_color <- table_stub_font_color
  ret$table_stub_font_bold <- table_stub_font_bold
  ret$table_body_font_color <- table_body_font_color
  ret$table_label_row_bold <- table_label_row_bold
  
  
  return(ret)

}

#' @title 
#' Add a style object to a report.
#' @description 
#' This function will add a style object to a report specification.  The 
#' style may be added either by passing a style object to the \code{style} 
#' parameter, or by passing a theme name to the \code{theme} parameter.
#' @details
#' The \code{add_style()} function allows you to add styling to HTML reports.
#' This functionality will apply to additional output types in future
#' versons of the \strong{reporter} package.
#' 
#' Styling can be added by passing a named theme to the \code{theme} parameter,
#' or by creating a style object using the \code{\link{create_style}} function
#' and passing that object to the \code{style} parameter.  You may also export 
#' a theme as a style object using the \code{\link{get_theme}} function,
#' modify it, and pass that to the \code{style} parameter.
#' 
#' @section Style Specifications:
#' The style specification is created using the \code{\link{create_style}} function,
#' and provides the most styling flexibility.  The style object allows you
#' to control background colors, font colors, border colors, and more.  Colors
#' can be specified using an RGB hex code, or an HTML/CSS-compliant color name.
#' See the \code{\link{create_style}} documentation for additional information.
#' 
#' @section Themes:
#' There are currently seven themes available: "MidnightBlue", "SteelBlue", 
#' "DarkRed", "SeaGreen", "SlateGrey", "Plain", and "SASDefault". 
#' All themes use Arial 10pt font. These themes are intended to provide a 
#' basic set of examples on how to create styles. Themes can be used by 
#' passing the theme name to the theme parameter on the \code{add_style()}
#' function. The theme will assign a variety of style settings according 
#' to the specifics of the theme.  For example, the "MidnightBlue" theme
#' sets the title font and header background colors to "MidnightBlue" and sets
#' the border color to "Grey".
#' 
#' To view theme style settings, you can use \code{\link{get_theme}} function. 
#' This function will return the theme as a style object.  See
#' the \code{\link{get_theme}} documentation for further details.
#' @param rpt The report specification to add a style to.
#' @param style A style object which contains style settings to add to the 
#' report. This parameter is optional.  Default is NULL.  
#' @param theme A theme name to use for this report. Valid values are
#' "MidnightBlue", "SteelBlue", "DarkRed", "SeaGreen", "SlateGrey", "Plain", 
#' and "SASDefault".  Default is NULL.
#' @family styles
#' @examples 
#' library(reporter)
#' library(magrittr)
#' 
#' # Prepare data
#' dat <- as.data.frame(HairEyeColor)
#' dat <- dat[dat$Freq >= 10, ]
#' 
#' ## Example 1: Use Pre-defined Theme ##
#' 
#' # Create temp file path
#' tmp1 <- file.path(tempdir(), "HairAndEyes1.html")
#' 
#' # Create table object
#' tbl <- create_table(dat, borders = "outside") %>% 
#' titles("Hair and Eye Colors with Theme") %>% 
#' column_defaults(width = .6)
#' 
#' # Create report and add theme
#' rpt <- create_report(tmp1, output_type = "HTML") %>% 
#'        add_content(tbl) %>% 
#'        add_style(theme = "SteelBlue")
#'
#' # Write out the report        
#' write_report(rpt)
#' 
#' # Uncomment to View report
#' # file.show(tmp1)
#' 
#' #' ## Example 2: Create Custom Style ##
#' 
#' # Create temp file path
#' tmp2 <- file.path(tempdir(), "HairAndEyes2.html")
#' 
#' # Define custom style
#' sty <- create_style(font_name = "Times",
#'                     font_size = 10,
#'                     title_font_size = 12,
#'                     title_font_bold = TRUE,
#'                     title_font_color = "Blue",
#'                     table_header_background = "Blue",
#'                     table_header_font_bold = TRUE,
#'                     table_header_font_color = "White",
#'                     table_body_background = "White",
#'                     table_body_stripe = "Red")
#' 
#' # Create table object
#' tbl <- create_table(dat, borders = "outside") %>% 
#' titles("Hair and Eye Colors with Style") %>% 
#' column_defaults(width = .6)
#' 
#' # Create report and add style spec
#' rpt <- create_report(tmp2, output_type = "HTML") %>% 
#'        add_content(tbl) %>% 
#'        add_style(style = sty)
#'
#' # Write out the report        
#' write_report(rpt)
#' 
#' # Uncomment to View report
#' # file.show(tmp2)
#' @export
add_style <- function(rpt, style = NULL, theme = NULL) {
  
  if (is.null(rpt)) {
    stop("Report object cannot be NULL.") 
    
  }
  
  if (is.null(style) & is.null(theme)) {
    stop("Style object and theme cannot both be NULL.") 
    
  }
  
  if (!any(class(rpt) %in% c("report_spec"))) {
    
    stop("First parameter to add_style() must be a report spec.") 
  }
  
  if (!is.null(style)) {
    if (!any(class(style) %in% c("style_spec"))) {
      
     stop("Style parameter to add_style() must be a style spec.") 
    }
  }
  
  if (!is.null(theme)) {
    if (!any(class(theme) %in% c("character"))) {
      
      stop("Theme parameter to add_style() must be a valid theme name.") 
    }
  }
  
  # Get theme if requested
  if (!is.null(theme)) {
    
    style <- get_theme(theme)
  } 
  
  if (!is.null(style)) {
    
    rpt$style <- style
    
    # Set font name and size if requested
    if (!is.null(style$font_name))
      rpt$font <- style$font_name
    
    if (!is.null(style$font_size))
      rpt$font_size <- style$font_size
  }
  
  
    
  return(rpt)
  
}

#' @title 
#' Get a theme 
#' @description 
#' This function will return a style object for a specified theme.
#' There are currently seven themes available.  The returned object 
#' may be modifed and applied to a report using \code{\link{add_style}}.
#' @param theme A string that contains the desired theme name to return.
#' Valid values are "MidnightBlue", "SteelBlue", "DarkRed", "SeaGreen", 
#' "SlateGrey", "Plain", and "SASDefault". 
#' @family styles
#' @examples 
#' library(reporter)
#' library(magrittr)
#' 
#' # Get theme
#' tm <- get_theme("SteelBlue")
#' 
#' # View theme settings
#' print(tm)
#' ## A style specification: 
#' #- font_name: 'Arial'
#' #- font_size: 10
#' #- text_color: 'DimGrey'
#' #- title_font_size: 11
#' #- title_font_bold: TRUE
#' #- title_font_color: 'SteelBlue'
#' #- border_color: 'Grey'
#' #- table_header_background: 'SteelBlue'
#' #- table_header_font_bold: TRUE
#' #- table_header_font_color: 'LightGrey'
#' #- table_body_background: 'White'
#' #- table_body_stripe: 'WhiteSmoke'
#' #- table_stub_background: 'SteelBlue'
#' #- table_stub_font_color: 'LightGrey'
#' #- table_stub_font_bold: TRUE
#' 
#' # Modify theme
#' tm$font_size <- 12
#' tm$title_font_size <- 13
#' 
#' # Create temp file path
#' tmp <- file.path(tempdir(), "HairAndEyes.html")
#' 
#' # Get data
#' dat <- as.data.frame(HairEyeColor)
#' 
#' # Create table object
#' tbl <- create_table(dat[dat$Freq >= 10, ], 
#' borders = "outside") %>% 
#' titles("Hair and Eye Colors")
#' 
#' # Use modified theme
#' rpt <- create_report(tmp, output_type = "HTML") %>% 
#'        add_content(tbl) %>% 
#'        add_style(tm)
#'
#' # Write out the report        
#' write_report(rpt)
#' 
#' # Uncomment to View report
#' # file.show(tmp)
#' 
#' @export
get_theme <- function(theme) {
  
  ret <- NULL
  
  if (tolower(theme) == "sasdefault") {
    
    ret <- create_style(font_name = "Arial",
                        font_size = 10,
                        text_color = "DimGrey",
                        background_color = "#FAFBFE",
                        title_font_color = "MidnightBlue",
                        title_font_bold = TRUE,
                        title_font_size = 11,
                        border_color = "Grey",
                        table_header_background = "#EDF2F9",
                        table_header_font_bold = TRUE,
                        table_header_font_color = "MidnightBlue",
                        table_body_background = "White",
                        table_stub_background = "#EDF2F9",
                        table_stub_font_bold = TRUE,
                        table_stub_font_color = "MidnightBlue",
                        footnote_background = "#EDF2F9",
                        footnote_font_color = "MidnightBlue",
                        footnote_font_bold = TRUE)
    
  } else if (tolower(theme) == "midnightblue") {
      
      ret <- create_style(font_name = "Arial",
                          font_size = 10,
                          text_color = "DimGrey",
                          background_color = NULL,
                          title_font_color = "MidnightBlue",
                          title_font_bold = TRUE,
                          title_font_size = 11,
                          border_color = "Grey",
                          table_header_background = "MidnightBlue",
                          table_header_font_bold = TRUE,
                          table_header_font_color = "LightGrey",
                          table_body_background = "White",
                          table_stub_background = "MidnightBlue",
                          table_stub_font_bold = TRUE,
                          table_stub_font_color = "LightGrey",
                          table_body_stripe = "WhiteSmoke")
      
   } else if (tolower(theme) == "darkred") {
      
      ret <- create_style(font_name = "Arial",
                          font_size = 10,
                          text_color = "DimGrey",
                          background_color = NULL,
                          title_font_color = "DarkRed",
                          title_font_bold = TRUE,
                          title_font_size = 11,
                          border_color = "Grey",
                          table_header_background = "DarkRed",
                          table_header_font_bold = TRUE,
                          table_header_font_color = "LightGrey",
                          table_body_background = "White",
                          table_stub_background = "DarkRed",
                          table_stub_font_bold = TRUE,
                          table_stub_font_color = "LightGrey",
                          table_body_stripe = "WhiteSmoke")
      
   } else if (tolower(theme) == "seagreen") {
     
     ret <- create_style(font_name = "Arial",
                         font_size = 10,
                         text_color = "DimGrey",
                         background_color = NULL,
                         title_font_color = "SeaGreen",
                         title_font_bold = TRUE,
                         title_font_size = 11,
                         border_color = "Grey",
                         table_header_background = "SeaGreen",
                         table_header_font_bold = TRUE,
                         table_header_font_color = "LightGrey",
                         table_body_background = "White",
                         table_stub_background = "SeaGreen",
                         table_stub_font_bold = TRUE,
                         table_stub_font_color = "LightGrey",
                         table_body_stripe = "WhiteSmoke")
     
   } else if (tolower(theme) == "steelblue") {
     
     ret <- create_style(font_name = "Arial",
                         font_size = 10,
                         text_color = "DimGrey",
                         background_color = NULL,
                         title_font_color = "SteelBlue",
                         title_font_bold = TRUE,
                         title_font_size = 11,
                         border_color = "Grey",
                         table_header_background = "SteelBlue",
                         table_header_font_bold = TRUE,
                         table_header_font_color = "LightGrey",
                         table_body_background = "White",
                         table_stub_background = "SteelBlue",
                         table_stub_font_bold = TRUE,
                         table_stub_font_color = "LightGrey",
                         table_body_stripe = "WhiteSmoke")
     
   } else if (tolower(theme) == "slategrey") {
     
     ret <- create_style(font_name = "Arial",
                         font_size = 10,
                         text_color = "DimGrey",
                         background_color = NULL,
                         title_font_color = "SlateGrey",
                         title_font_bold = TRUE,
                         title_font_size = 11,
                         border_color = "Grey",
                         table_header_background = "SlateGrey",
                         table_header_font_bold = TRUE,
                         table_header_font_color = "LightGrey",
                         table_body_background = "White",
                         table_stub_background = "SlateGrey",
                         table_stub_font_bold = TRUE,
                         table_stub_font_color = "LightGrey",
                         table_body_stripe = "WhiteSmoke")
     
   } else if (tolower(theme) == "plain") {
     
     ret <- create_style(font_name = "Arial",
                         font_size = 10,
                         title_font_bold = TRUE,
                         title_font_size = 11,
                         border_color = "Grey",
                         table_header_font_bold = TRUE,
                         table_label_row_bold = TRUE)
  } else {
    
   stop(paste0("Theme name '", theme, "' is not available."))
    
  }
  
  return(ret)
  
}


#' @title 
#' Prints a style specification
#' @description 
#' This function will print a style object to the console.  The print function
#' will display each style setting that has been assigned, and the value
#' which is assigned.
#' @param x A style object to print.
#' @param ... Any follow-on parameters to pass to print().
#' @param verbose If verbose is TRUE, the function will print the style
#' object as a list. Otherwise, the object will print using the custom
#' print function.  The custom print is more compact than the verbose style 
#' print. Default is FALSE.
#' @family styles
#' @import crayon
#' @examples 
#' library(reporter)
#' 
#' # Get theme
#' tm <- get_theme("SteelBlue")
#' 
#' # View theme settings
#' print(tm)
#' ## A style specification: 
#' #- font_name: 'Arial'
#' #- font_size: 10
#' #- text_color: 'DimGrey'
#' #- title_font_size: 11
#' #- title_font_bold: TRUE
#' #- title_font_color: 'SteelBlue'
#' #- border_color: 'Grey'
#' #- table_header_background: 'SteelBlue'
#' #- table_header_font_bold: TRUE
#' #- table_header_font_color: 'LightGrey'
#' #- table_body_background: 'White'
#' #- table_body_stripe: 'WhiteSmoke'
#' #- table_stub_background: 'SteelBlue'
#' #- table_stub_font_color: 'LightGrey'
#' #- table_stub_font_bold: TRUE
#' @export
print.style_spec <- function(x, ..., verbose = FALSE) {
  
  if (verbose == TRUE) {
    
    
    print(as.list(unclass(x)))
    
    
    
  } else {
    
    grey60 <- make_style(grey60 = "#999999")
    
    # Print header
    cat(grey60("# A style specification: \n"))
    
    # Print values for each style parameter
    for (nm in names(x)) {
      
      if (any(class(x[[nm]]) %in% "character")) {
      
        cat(paste0("- ", nm, ": '", x[[nm]], "'\n"))
        
      } else {
        cat(paste0("- ", nm, ": ", x[[nm]], "\n"))
        
      } 

    }
    
  }
  
  invisible(x)
  
}




# A function to determine whether a style exists.
# Returns a TRUE or FALSE.
#' @noRd
has_style <- function(rs, style_name) {
 
  ret <- FALSE
  
  if (!is.null(rs$style)) {
    if (!is.null(rs$style[[style_name]])) {
      
      ret <- TRUE 
    }
  }
  
  return(ret)
  
}

# A function to get a style value if it exists.
# Returns the value or an empty string if it does not exist.
#' @noRd
get_style <- function(rs, style_name) {
  
  ret <- ""
  
  if (!is.null(rs$style)) {
    if (!is.null(rs$style[[style_name]])) {
      
      ret <- rs$style[[style_name]]
    }
  }
  
  return(ret)
}

# A function to convert a style setting to the appropriate css.
# Returns an empty string if the style setting is not found.
#' @noRd
get_style_html <- function(rs, style_name, default = NULL) {
  
  ret <- ""
  
  val <- ""
  
  if (!is.null(default))
    val <- default
  if (has_style(rs, style_name))
    val <- get_style(rs, style_name)
  

  if (val != "") {
    
    if (style_name == "text_color")
      ret <- paste0("color: ", val, ";")
    
    if (style_name == "title_font_color")
      ret <- paste0("color: ", val, ";")
    
    if (style_name == "title_background")
      ret <- paste0("background-color: ", val, ";")
    
    if (style_name == "title_font_size")
      ret <- paste0("font-size: ", val, "pt;")
    
    if (style_name == "title_font_bold") {
      if (val == TRUE)
        ret <- paste0("font-weight: bold;")
      else 
        ret <- paste0("font-weight: normal;")
      
    }
    
    if (style_name == "footnote_font_color")
      ret <- paste0("color: ", val, ";")
    
    if (style_name == "footnote_background")
      ret <- paste0("background-color: ", val, ";")
    
    if (style_name == "footnote_font_bold") {
      if (val == TRUE)
        ret <- paste0("font-weight: bold;")
      else 
        ret <- paste0("font-weight: normal;")
      
    }
    
    if (style_name == "background_color")
      ret <- paste0("background-color: ", val, ";")
    
    if (style_name == "table_body_font_color")
      ret <- paste0("color: ", val, ";")
    
    if (style_name == "table_body_background")
      ret <- paste0("background-color: ", val, ";")
    
    if (style_name == "table_body_stripe")
      ret <- paste0("background-color: ", val, ";")
    
    if (style_name == "table_header_background") 
      ret <- paste0("background-color: ", val, ";")
    
    if (style_name == "table_header_font_bold") {
      if (val == TRUE)
        ret <- paste0("font-weight: bold;")
      else 
        ret <- paste0("font-weight: normal;")
    }
    
    if (style_name == "table_header_font_color")
      ret <- paste0("color: ", val, ";")
    
    if (style_name == "table_stub_font_color")
      ret <- paste0("color: ", val, ";")
    
    if (style_name == "table_stub_background")
      ret <- paste0("background-color: ", val, ";")
    
    if (style_name == "table_stub_font_bold") {
      if (val == TRUE)
        ret <- paste0("font-weight: bold;")
      else 
        ret <- paste0("font-weight: normal;")
      
    }
    
    if (style_name == "table_label_row_bold") {
      if (val == TRUE)
        ret <- paste0("font-weight: bold;")
      else 
        ret <- paste0("font-weight: normal;")
      
    }
    
        
  }
  
  
  return(ret)
}



#' @title A style specification for a table cell
#' @description A class to define the style for a cell in a table.  This
#' class can be assigned to the "style" parameter of a \code{\link{define}}
#' function, a \code{\link{stub}} function, or a \code{\link{column_defaults}}  
#' function.  When assigned, the cell style will apply to some or all of the cells 
#' in the relevant columns.  The 
#' "indicator" parameter identifies a column in the table to trigger the style.
#' Alternatively, the "labelrow", "blankrow", or "datarow" shortcuts may be used
#' to identify cells for styling.  
#' @param indicator A keyword or column name to indicate which rows the cell
#' style should be applied to. Valid keywords are "labelrow", "blankrow", or
#' "datarow". To use an indicator column, create a column on the input dataset
#' that has a TRUE value for each row that you want the cell style applied.
#' Then pass the name of the column to the indicator parameter. If you do not
#' want to see the indicator column on the report, set the "visible" property 
#' on the \code{\link{define}} function to FALSE for that column. 
#' The default value of the indicator parameter is NULL, 
#' meaning to apply the style to all rows.
#' @param bold Whether to bold the text in the cell.  Valid values are TRUE and
#' FALSE.  The default is FALSE.
#' @examples 
#' library(reporter)
#' library(magrittr)
#'
#' # Create temporary path
#' tmp <- file.path(tempdir(), "table1.rtf")
#' 
#' # Prepare data
#' df <- data.frame(names = rownames(mtcars), mtcars[, 1:3])
#' 
#' # Set indicator variable
#' df$mpgind <- ifelse(df$mpg > 20, TRUE, FALSE)
#' 
#' # Create table
#' tbl <- create_table(df, first_row_blank = TRUE, 
#'                     header_bold = TRUE, borders = c("top", "bottom")) %>% 
#'   column_defaults(style = cell_style(bold = TRUE, indicator = "mpgind")) %>%
#'   define(names, label = "Car Name") %>%
#'   define(mpg, label = "Miles Per Gallon") %>% 
#'   define(cyl, label = "Cylinders") %>% 
#'   define(disp, label = "Displacement")  %>% 
#'   define(mpgind, visible = FALSE) %>%
#'   titles("Table 1.0", "MTCARS with Indicator Variable", 
#'          borders = "none", bold = TRUE, font_size = 11) %>% 
#'   footnotes("* Motor Trend, 1974", borders = "none", blank_row = "none")
#' 
#' # Create report and add custom style
#' rpt <- create_report(tmp, output_type = "RTF", font = "Arial") %>% 
#'   add_content(tbl) 
#' 
#' # Write out report
#' write_report(rpt)
#' 
#' # View report
#' # file.show(tmp)
#' @export
cell_style <- function(indicator = NULL, bold = FALSE) {
  
  ret <- structure(list(), class = c("cell_style", "list"))
  
  
  # Deal with single value unquoted parameter values
  oindicator <- deparse(substitute(indicator, env = environment()))
  indicator  <- tryCatch({if (typeof(indicator) %in% c("character", "NULL")) indicator else oindicator},
                 error = function(cond) {oindicator})
  
  
  ret$indicator <- indicator
  ret$bold <- bold
  
  
  return(ret)
  
}


has_cell_style <- function(colnm, styles) {
  
 ret <- FALSE  
 if (colnm %in% names(styles)) {
   
   ret <- TRUE
   
 }
  
 return(ret)
}


get_cell_style <- function(colnm, styles) {
  
  ret <- NULL
  if (has_cell_style(colnm, styles)) {
    
    ret <- styles[[colnm]]
  }
  
  
  return(ret)
}


# Return vector of strings saying which styles to apply to this cell
get_cell_styles <- function(colnm, styles, flgs, rw, tbl) {
  
  ret <- ""
  bflg <- FALSE
  
  if (has_cell_style(colnm, styles)) {
    stl <- get_cell_style(colnm, styles)
    if (stl$bold == TRUE) {
      bflg <- TRUE
      if (!is.null(stl$indicator)) {
        if ("datarow" %in% stl$indicator &&
            flgs[rw] %in% c("B", "L", "A")) {
          bflg <- FALSE
        } else if ("labelrow" %in% stl$indicator &&
                   !flgs[rw] %in% "L") {
          bflg <- FALSE
        } else if ("blankrow" %in% stl$indicator &&
                   !flgs[rw] %in% c("B", "A")) {
          bflg <- FALSE
        } else if (stl$indicator %in% names(tbl)) {
          if (!is.null(tbl[[rw, stl$indicator]])) {
            bflg <- FALSE 
            if (!is.na(tbl[[rw, stl$indicator]])) {
              if (tbl[[rw, stl$indicator]] == TRUE) {
                bflg <- TRUE   
              }
            }
          }
        }
      } 
    }
  }
  
  if (bflg) {
    ret[length(ret) + 1] <- "bold" 
  }
  
  return(ret)
  
}

context("Style Tests")

base_path <- "c:/packages/reporter/tests/testthat"
data_dir <- base_path

base_path <- tempdir()
data_dir <- "."

dev <- FALSE

test_that("style0: Style functions work as expected.", {
  
  res1 <- create_style(font_name = "Arial", font_size = 10,
                       table_header_background = "blue")
  
  expect_equal(res1$font_name, "Arial")
  expect_equal(res1$font_size, 10)
  expect_equal(res1$table_header_background, "blue")

  rpt <- create_report() %>% add_style(res1)
  
  expect_equal(get_style(rpt, "border_color"), "")
  
  
  expect_equal(is.null(rpt$style), FALSE)
  expect_equal(rpt$style$font_name, "Arial")
  
  
  rpt <- create_report() %>% add_style(create_style(border_color = "grey"))
  
  expect_equal(has_style(rpt, "border_color"), TRUE)
  expect_equal(get_style(rpt, "border_color"), "grey")
  
  rpt <- create_report() %>% add_style(theme = "SASDefault")
  
  expect_equal(get_style(rpt, "border_color"), "Grey")
  
})



test_that("style1: SAS theme works with html.", {
  
  
  fp <- file.path(base_path, "html/style1.html")
  # print(fp)
  
  dat <- data.frame(stub = rownames(mtcars)[1:15], mtcars[1:15, ])
  attr(dat[[3]], "label") <- "Cylin."
  attr(dat[[3]], "width") <- 1
  attr(dat[[3]], "justify") <- "center"
  
  tbl <- create_table(dat, borders = "all", first_row_blank = TRUE) %>%
    titles("Table 1.0", "My Nice Table", borders = c("none"), 
           width = "content") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "none", 
              align = "left", width = "content") %>% 
    define(wt, width = 1, label = "Weight", align = "center", 
           label_align = "right") %>% 
    spanning_header(from = "mpg", to = "cyl", label = "Span1")
  
  rpt <- create_report(fp, output_type = "HTML", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), 
                blank_row = "below") %>% 
    add_content(tbl, align = "center")  %>% 
    page_footer("Left1", "Center1", "Right1") %>% 
    add_style(theme = "SASDefault")
  
  res <- write_report(rpt)
  
  if (dev == TRUE)
    file.show(res$modified_path)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})




test_that("style2: add_style() works with html.", {
  
  
  fp <- file.path(base_path, "html/style2.html")
  # print(fp)
  
  dat <- data.frame(stub = rownames(mtcars[1:15, ]), mtcars[1:15, ])
  attr(dat[[2]], "label") <- "Cylin."
  attr(dat[[2]], "width") <- 1
  attr(dat[[2]], "justify") <- "center"
  
  sty <- create_style(font_name = "Arial", font_size = 12, text_color = "blue",
                      background_color = "pink", title_font_bold = TRUE,
                      title_font_color = "orange", title_background = "brown",
                      border_color = "grey", table_body_background = "yellow",
                      table_body_stripe = "blue", table_stub_background = "purple",
                      table_stub_font_bold = TRUE, table_stub_font_color = "green", 
                      table_header_background = "green", table_header_font_bold = TRUE,
                      table_header_font_color = "brown", footnote_font_bold = TRUE,
                      footnote_font_color = "brown", footnote_background = "orange",
                      table_body_font_color = "red")
  
  tbl <- create_table(dat, borders = "all", first_row_blank = FALSE) %>%
    titles("Table 1.0", "My Nice Table", borders = c("all"), 
           width = "content") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "all", 
              align = "left", width = "content") %>% 
    define(wt, width = 1, label = "Weight", align = "center", 
           label_align = "right") %>% 
    spanning_header(from = "mpg", to = "drat", label = "Span 1") %>% 
    spanning_header(from = "wt", to ="carb", label = "Span 2")
  
  rpt <- create_report(fp, output_type = "HTML", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), 
                blank_row = "below") %>% 
    add_content(tbl, align = "center")  %>% 
    page_footer("Left1", "Center1", "Right1") %>% 
    add_style(sty)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  
})

test_that("style3: Test print.style_spec() and get_theme().", {
  
  s1 <- create_style(font_name = "Arial", font_size = 12, 
                     table_body_background = "Blue",
                     table_header_background = "Red")
  if (dev)
    print(s1)
  
  
  s2 <- get_theme("SASDefault")
  
  if (dev)
    print(s2)
  
  if (dev)
    print(s1, verbose = TRUE)
  
  expect_equal(TRUE, TRUE)
  
})


test_that("style4: DarkRed theme works.", {
  
  
  fp <- file.path(base_path, "html/style4.html")
  # print(fp)
  
  dat <- data.frame(stub = rownames(mtcars)[1:15], mtcars[1:15, ])
  attr(dat[[3]], "label") <- "Cylin."
  attr(dat[[3]], "width") <- 1
  attr(dat[[3]], "justify") <- "center"
  
  tbl <- create_table(dat, borders = "all", first_row_blank = TRUE) %>%
    titles("Table 1.0", "My Nice Table", borders = c("none"), 
           width = "content") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "none", 
              align = "left", width = "content") %>% 
    define(wt, width = 1, label = "Weight", align = "center", 
           label_align = "right") %>% 
    spanning_header(from = "mpg", to = "cyl", label = "Span1")
  
  rpt <- create_report(fp, output_type = "HTML", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), 
                blank_row = "below") %>% 
    add_content(tbl, align = "center")  %>% 
    page_footer("Left1", "Center1", "Right1") %>% 
    add_style(theme = "DarkRed")
  
  res <- write_report(rpt)
  
  if (dev == TRUE)
    file.show(res$modified_path)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})

test_that("style5: SeaGreen theme works.", {
  
  
  fp <- file.path(base_path, "html/style5.html")
  # print(fp)
  
  dat <- data.frame(stub = rownames(mtcars)[1:15], mtcars[1:15, ])
  attr(dat[[3]], "label") <- "Cylin."
  attr(dat[[3]], "width") <- 1
  attr(dat[[3]], "justify") <- "center"
  
  tbl <- create_table(dat, borders = "all", first_row_blank = TRUE) %>%
    titles("Table 1.0", "My Nice Table", borders = c("none"), 
           width = "content") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "none", 
              align = "left", width = "content") %>% 
    define(wt, width = 1, label = "Weight", align = "center", 
           label_align = "right") %>% 
    spanning_header(from = "mpg", to = "cyl", label = "Span1")
  
  rpt <- create_report(fp, output_type = "HTML", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), 
                blank_row = "below") %>% 
    add_content(tbl, align = "center")  %>% 
    page_footer("Left1", "Center1", "Right1") %>% 
    add_style(theme = "SeaGreen")
  
  res <- write_report(rpt)
  
  if (dev == TRUE)
    file.show(res$modified_path)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})

test_that("style6: SlateGrey theme works.", {
  
  
  fp <- file.path(base_path, "html/style6.html")
  # print(fp)
  
  dat <- data.frame(stub = rownames(mtcars)[1:15], mtcars[1:15, ])
  attr(dat[[3]], "label") <- "Cylin."
  attr(dat[[3]], "width") <- 1
  attr(dat[[3]], "justify") <- "center"
  
  tbl <- create_table(dat, borders = "all", first_row_blank = TRUE) %>%
    titles("Table 1.0", "My Nice Table", borders = c("none"), 
           width = "content") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "none", 
              align = "left", width = "content") %>% 
    define(wt, width = 1, label = "Weight", align = "center", 
           label_align = "right") %>% 
    spanning_header(from = "mpg", to = "cyl", label = "Span1")
  
  rpt <- create_report(fp, output_type = "HTML", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), 
                blank_row = "below") %>% 
    add_content(tbl, align = "center")  %>% 
    page_footer("Left1", "Center1", "Right1") %>% 
    add_style(theme = "SlateGrey")
  
  res <- write_report(rpt)
  
  if (dev == TRUE)
    file.show(res$modified_path)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})


test_that("style7: MidnightBlue theme works with html.", {
  
  
  fp <- file.path(base_path, "html/style7.html")
  # print(fp)
  
  dat <- data.frame(stub = rownames(mtcars)[1:15], mtcars[1:15, ])
  attr(dat[[3]], "label") <- "Cylin."
  attr(dat[[3]], "width") <- 1
  attr(dat[[3]], "justify") <- "center"
  
  tbl <- create_table(dat, borders = "all", first_row_blank = TRUE) %>%
    titles("Table 1.0", "My Nice Table", borders = c("none"), 
           width = "content") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "none", 
              align = "left", width = "content") %>% 
    define(wt, width = 1, label = "Weight", align = "center", 
           label_align = "right") %>% 
    spanning_header(from = "mpg", to = "cyl", label = "Span1")
  
  rpt <- create_report(fp, output_type = "HTML", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), 
                blank_row = "below") %>% 
    add_content(tbl, align = "center")  %>% 
    page_footer("Left1", "Center1", "Right1") %>% 
    add_style(theme = "MidnightBlue")
  
  res <- write_report(rpt)
  
  if (dev == TRUE)
    file.show(res$modified_path)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})


test_that("style8: SteelBlue theme works.", {
  
  
  fp <- file.path(base_path, "html/style8.html")
  # print(fp)
  
  dat <- data.frame(stub = rownames(mtcars)[1:15], mtcars[1:15, ])
  attr(dat[[3]], "label") <- "Cylin."
  attr(dat[[3]], "width") <- 1
  attr(dat[[3]], "justify") <- "center"
  
  tbl <- create_table(dat, borders = "all", first_row_blank = TRUE) %>%
    titles("Table 1.0", "My Nice Table", borders = c("none"), 
           width = "content") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "none", 
              align = "left", width = "content") %>% 
    define(wt, width = 1, label = "Weight", align = "center", 
           label_align = "right") %>% 
    spanning_header(from = "mpg", to = "cyl", label = "Span1")
  
  rpt <- create_report(fp, output_type = "HTML", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), 
                blank_row = "below") %>% 
    add_content(tbl, align = "center")  %>% 
    page_footer("Left1", "Center1", "Right1") %>% 
    add_style(theme = "SteelBlue")
  
  res <- write_report(rpt)
  
  if (dev == TRUE)
    file.show(res$modified_path)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})

test_that("style9: Plain theme works.", {
  
  
  fp <- file.path(base_path, "html/style9.html")
  # print(fp)
  
  dat <- data.frame(stub = rownames(mtcars)[1:15], mtcars[1:15, ])
  attr(dat[[3]], "label") <- "Cylin."
  attr(dat[[3]], "width") <- 1
  attr(dat[[3]], "justify") <- "center"
  
  tbl <- create_table(dat, borders = "all", first_row_blank = TRUE) %>%
    titles("Table 1.0", "My Nice Table", borders = c("none"), 
           width = "content") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "none", 
              align = "left", width = "content") %>% 
    define(wt, width = 1, label = "Weight", align = "center", 
           label_align = "right") %>% 
    spanning_header(from = "mpg", to = "cyl", label = "Span1")
  
  rpt <- create_report(fp, output_type = "HTML", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), 
                blank_row = "below") %>% 
    add_content(tbl, align = "center")  %>% 
    page_footer("Left1", "Center1", "Right1") %>% 
    add_style(theme = "Plain")
  
  res <- write_report(rpt)
  
  if (dev == TRUE)
    file.show(res$modified_path)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})


test_that("style10: Header border works as expected.", {
  

  # Create temp file path
  fp <- file.path(base_path, "html/style10.html")
  
  dat <- as.data.frame(HairEyeColor)
  dat <- dat[dat$Freq > 10, ]

  # Define custom style
  sty <- create_style(font_name = "Times",
                      font_size = 10,
                      title_font_size = 12,
                      title_font_bold = TRUE,
                      title_font_color = "Blue",
                      table_header_background = "Grey",
                      table_header_font_bold = TRUE,
                      table_header_font_color = "White",
                      table_body_background = "White",
                      table_body_stripe = "Red")

  # Create table object
  tbl <- create_table(dat, borders = "outside") %>%
  titles("Hair and Eye Colors with Style") %>%
  column_defaults(width = .6) %>% 
  spanning_header(Hair, Freq, label = "Span")

  # Create report and add theme
  rpt <- create_report(fp, output_type = "HTML") %>%
         add_content(tbl) %>%
         add_style(style = sty)

  # Write out the report
  res <- write_report(rpt)

  if (dev)
    file.show(fp)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})

test_that("style11: stub border work as expected.", {
  
  
  library(reporter)
  
  # Create temporary path
  fp <- file.path(base_path, "html/style11.html")
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  # Create custom style
  sty <- create_style(font_name = "Arial",
                      font_size = 10,
                      background_color = "WhiteSmoke",
                      border_color = "Grey",
                      title_font_size = 12,
                      title_font_bold = TRUE,
                      title_font_color = "SteelBlue",
                      table_header_background = "Tan",
                      table_header_font_bold = TRUE,
                      table_header_font_color = "White",
                      table_body_background = "White",
                      table_body_stripe = "Wheat",
                      table_stub_background = "Tan",
                      table_stub_font_color = "White",
                      table_stub_font_bold = TRUE
                   )
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE, borders = c("all")) %>% 
    stub(c("var", "label")) %>% 
    column_defaults(width = 1.25) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center", n = 19) %>% 
    define(B, label = "Group B", align = "center", n = 13)  %>% 
    titles("Table 1.0", "MTCARS Summary Table with Custom Style", 
           borders = "none") %>% 
    footnotes("* Motor Trend, 1974", borders = "outside")
  
  # Create report and add content
  rpt <- create_report(fp, output_type = "HTML", font = "Arial", 
                       font_size = 12) %>% 
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>% 
    add_style(sty)
  
  # Write out report
  res <- write_report(rpt)
  
  if (dev)
    file.show(fp)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})



test_that("style12: Label row is one cell and styles work.", {
  
  
  fp <- file.path(base_path, "html/style12.html")
  
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  ll <- "Here is a super long label to see if it can span the entire table."
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE, borders = "all", 
                      header_bold = TRUE) %>% 
    stub(c("var", "label")) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = ll, cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center", n = 19) %>% 
    define(B, label = "Group B", align = "center", n = 13)
  
  
  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "HTML",
                       font = "Times") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = Sys.time(), 
                center = "Confidential", 
                right = "Page [pg] of [tpg]") %>%
    add_style(theme = "SteelBlue")
  
  
  
  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)
  
  
})

test_that("style13: Label row is one cell and styles work.", {
  
  res <- cell_style(indicator = fork, bold = TRUE)
  
  expect_equal(res$indicator, "fork")
  expect_equal(res$bold, TRUE)
  
})


test_that("style14: has_cell_style() works as expected.", {
  
  stls <- list(fork = cell_style(bold = TRUE),
               bork = cell_style(bold = TRUE, indicator = "sammy"))
  
  
  expect_equal(has_cell_style("fork", stls), TRUE)
  expect_equal(has_cell_style("forp", stls), FALSE)
  
})


test_that("style15: get_cell_style() works as expected.", {
  
  stls <- list(fork = cell_style(bold = TRUE),
               bork = cell_style(bold = TRUE, indicator = "sammy"))
  
  sty <- get_cell_style("fork", stls)
  
  expect_equal(is.null(sty), FALSE)
  
  sty2 <- get_cell_style("forp", stls)
  
  expect_equal(is.null(sty2), TRUE)
  
})

# Create Tables ----------------------------------------------------------


#' @noRd
create_table_pages_rtf <- function(rs, cntnt, lpg_rows) {
  
  ts <- cntnt$object
  content_blank_row <- cntnt$blank_row
  
  pgby_var <- NA
  pgby_cnt <- 0
  pgby_fmt <- NULL
  if (!is.null(rs$page_by)) {
    pgby_var <- rs$page_by$var
    pgby_fmt <- rs$page_by$format
  } else if (!is.null(ts$page_by)) {
    pgby_var <- ts$page_by$var
    pgby_fmt <- ts$page_by$format
  }

  if (all(ts$show_cols == "none") & length(ts$col_defs) == 0) {
    
    stop("ERROR: At least one column must be defined if show_cols = \"none\".")
  }
  
  font_name <- rs$font
  
  # Set up control columns
  dat <- as.data.frame(ts$data, stringsAsFactors = FALSE)  
  dat$..blank <- ""
  dat$..row <- 1
  dat$..page_by <- NA
  
  # If page_break variable has been defined, use it
  if (is.null(ts$page_var)) {
    if (is.na(pgby_var))
      dat$..page <- NA
    else 
      dat$..page <-  dat[[pgby_var]]
    
  } else {
    if (any(class(dat[[ts$page_var]]) == "factor"))
      dat$..page <- as.character(dat[[ts$page_var]])
    else 
      dat$..page <- dat[[ts$page_var]]
  }
  
  # If page by is defined, use it
  if (!is.na(pgby_var)) {

    # Clear out factors on page by if they exist
    # Needed because split cells used to convert everything to characters
    if (any(class(dat[[pgby_var]]) == "factor")) {
      dat[[pgby_var]] <- as.character(dat[[pgby_var]] )
      dat$..page_by <- dat[[pgby_var]] 
      
      if (any(class(dat$..page) == "factor")) 
        dat$..page <- as.character(dat[[pgby_var]])
      
    } else {
      
      if (!is.null(pgby_fmt))
        dat$..page_by <-  fapply(dat[[pgby_var]], pgby_fmt)
      else 
        dat$..page_by <-  dat[[pgby_var]]
    }
    
    pgby_cnt <- get_pgby_cnt(dat$..page_by)
    
    # Commenting out for now
    # if (is.unsorted(dat[[pgby_var]], strictly = FALSE))
    #   message("Page by variable not sorted.")
  }
  
  # Deal with invisible columns
  if (!is.null(ts$col_defs)) {
    for (def in ts$col_defs) {
      if (def$visible == FALSE) {
        nnm <- paste0("..x.", def$var_c)
        dat[[nnm]] <- dat[[def$var_c]]
      }
    }
    
    # Deal with Group Cohesion
    # Prepare ..group_cohesion and ..min_page_prop
    group_cohesion <- c()
    min_page_prop <- c()
    
    i <- 1
    for (def in ts$col_defs) {
      if (def$group_cohesion == TRUE) {
        group_cohesion <- c(group_cohesion, def$var_c)
        min_page_prop <- c(min_page_prop, def$min_page_prop)
        
        dat$..temp <- dat[[def$var_c]]
        names(dat)[names(dat) == "..temp"] <-  paste0("..group_cohesion",i)
        
        dat$..temp <- def$min_page_prop
        names(dat)[names(dat) == "..temp"] <-  paste0("..min_page_prop",i)
        i <- i + 1
      }
    }
    
    if (length(group_cohesion) == 0) {
      group_cohesion <- NULL
    }
    
    if (length(min_page_prop) == 0) {
      min_page_prop <- NULL
    }
    
    # Prepare empty ..break_label and ..break_label_lines for later use
    i <- 1
    for (def in ts$col_defs) {
      if (!is.null(def$break_label)) {

        dat$..temp <- ifelse(!is.na(dat[[def$var_c]]) & dat[[def$var_c]] != "",
                             paste0(dat[[def$var_c]], " ", def$break_label),
                             dat[[def$var_c]])
        names(dat)[names(dat) == "..temp"] <-  paste0("..break_label",i)
        
        dat$..temp <- rep(NA, nrow(dat))
        names(dat)[names(dat) == "..temp"] <-  paste0("..break_label_lines",i)
        i <- i + 1
      }
    }
    # Prepare repeated label occurrence for later use
    if (any(grepl("..break_label", names(dat)))) {
      dat$..break_occur <- rep(NA, nrow(dat))
    }
  }
  
  # Get control column names 
  control_cols <- names(dat)[is.controlv(names(dat))]
  
  # Get vector of all included column names
  # Not all columns in dataset are necessarily included
  # depends on show_cols parameter on create_table and
  # visible parameter on column definitions
  keys <- get_table_cols(ts, control_cols)
  # print("keys")
  # print(keys)
  
  # Filter dataset by included columns
  dat <- get_data_subset(dat, keys, rs$preview)
  # print("Key columns:")
  # print(dat)
  
  # Update column definitions with column defaults
  ts$col_defs <- set_column_defaults(ts, keys)
  # print("col_defs:")
  # print(ts$col_defs)
  
  # Deal with styles
  styles <- get_styles(ts)
  
  # Get labels
  labels <- get_labels(dat, ts)
  # print("Labels:")
  # print(labels)
  
  # Get column alignments
  aligns <- get_aligns(dat, ts)
  
  # Get alignment for labels
  # Follows column alignment by default
  label_aligns <- get_label_aligns(ts, aligns)
  # print("Label Aligns:")
  # print(label_aligns)
  
  # Clear out existing formats
  cdat <- clear_formats(dat)
  
  # Get column formats
  formats(cdat) <- get_col_formats(dat, ts)
  # print("formats:")
  # print(formats(cdat))
  
  # Apply formatting
  fdat <- fdata(cdat)
  # print("fdata:")
  # print(fdat)

  # Prep data for blank lines, indents, and stub columns
  fdat <- prep_data(fdat, ts, rs$char_width, rs$missing) # OK
  # print("prep_data")
  # print(fdat)
  # str(fdat)
  
  # Add stub_var into control_cols for adding indentation
  if ("..stub_var" %in% names(fdat)) {
    control_cols <- c(control_cols, "..stub_var")
  }
  
  # Add cell_border into control_cols for adding cell border
  if ("..group_border" %in% names(fdat)) {
    control_cols <- c(control_cols, "..group_border")
  }
  
  # Reset keys, since prep_data can add/remove columns for stub
  keys <- names(fdat)
  # print("Keys")
  # print(keys)
  # print("Aligns")
  # print(aligns)
  
  # Copy any width attributes to formatted data frame
  if ("width" %in% ts$use_attributes)
    widths(fdat) <- widths(dat)
  # print("Original Widths")
  # print(widths(dat))
  
  
  # Get column widths
  widths_uom <- get_col_widths_variable(fdat, ts, labels, 
                                        rs$font, rs$font_size, rs$units, 
                                        rs$gutter_width) 
  
  # print("Widths UOM")
  # print(widths_uom)
  
  # Split long text strings into multiple rows. Number of rows are stored in
  # ..row variable. If too slow, may need to be rewritten in C
  fdat <- split_cells_variable(fdat, widths_uom, rs$font, 
                                rs$font_size, rs$units, rs$output_type, rs$char_width, ts)$data 
  # print("split_cells")
  # print(fdat)
  
  # Break columns into pages
  wraps <- get_page_wraps(rs$line_size, ts, 
                          widths_uom, 0, control_cols)  # No gutter width for RTF
  # print("wraps")
  # print(wraps)
  
  
  # Create a temporary page info to pass into get_content_offsets
  tmp_pi <- list(keys = keys, col_width = widths_uom, label = labels,
                 label_align = label_aligns, table_align = cntnt$align, data = fdat)
  # print("Temp PI")
  # print(tmp_pi)
  

  # Offsets are needed to calculate splits and page breaks
  content_offset <- get_content_offsets_rtf(rs, ts, tmp_pi, 
                                            content_blank_row, pgby_cnt)
  
  # split rows
  splits <- get_splits_text(fdat, widths_uom, rs$body_line_count, 
                            lpg_rows, content_offset$lines, ts, TRUE)  
  # print("splits")
  # print(splits)
  
  # Subset splits by preview, if requested
  if (!is.null(rs$preview)) {
    if (rs$preview < length(splits))
      splits <- splits[seq(1, rs$preview)] 
  }
  
  tot_count <- length(splits) * length(wraps)
  counter <- 0
  wrap_flag <- FALSE
  blnk_ind <- "none"
  
  pg_lst <- list()
  for(s in splits) {
    for(pg in wraps) {
      counter <- counter + 1
      
      if (counter < tot_count){
        wrap_flag <- TRUE
      } else {
        wrap_flag <- FALSE
      }
      
      if (!is.na(pgby_var)){
        pgby <- trimws(s[1, "..page_by"])
      } else {
        pgby <- NULL
      }
      
      #print(s)
      # Ensure content blank rows are added only to the first and last pages
      blnk_ind <- get_blank_indicator(counter, tot_count, content_blank_row,
                                      rs$body_line_count, content_offset$lines, 
                                      nrow(s), pgby)
      #print(blnk_ind)
      
      pgind <- ""
      if (counter == 1) {
        pgind <- "first"
      }
      if (counter == tot_count) {
        pgind[length(pgind) + 1] <- "last"
      }
      
      # Get page info for this page
      pi <- page_info(data= s[, pg], keys = pg, label=labels[pg],
                      col_width = widths_uom[pg], col_align = aligns[pg],
                      font_name = font_name, label_align = label_aligns[pg],
                      pgby, cntnt$align)
      
      # Get rtf for this page
      pg_lst[[length(pg_lst) + 1]] <- create_table_rtf(rs, ts, pi, 
                                                        blnk_ind, wrap_flag,
                                                        lpg_rows, pgind, styles)
    }
  }
  
  ret <- list(widths = widths_uom, page_list = pg_lst)
  
  return(ret)
  
  
}



#' @noRd
create_table_rtf <- function(rs, ts, pi, content_blank_row, wrap_flag, 
                              lpg_rows, pgind, styles) {
  rh <- rs$row_height
  shdrs <- list(lines = 0, twips = 0)
  hdrs <- list(lines = 0, twips = 0)
  
  if (ts$headerless == FALSE) {
    shdrs <- get_spanning_header_rtf(rs, ts, pi)   

    hdrs <- get_table_header_rtf(rs, ts, pi$col_width, 
                                 pi$label, pi$label_align, pi$table_align)  
  }
  
  # rs, ts, widths,  algns, halgns, talgn
  rws <- get_table_body_rtf(rs, pi$data, pi$col_width, 
                            pi$col_align, pi$table_align, ts$borders, 
                            ts$first_row_blank, ts$continuous, styles, pgind, ts)
  
  # Default to content width
  ls <- rs$content_size[["width"]]
  
  # Get table width
  if (!is.null(pi$col_width))
     ls <- sum(pi$col_width, na.rm = TRUE)
  
  if (!is.null(ts$title_hdr))
    ttls <- get_title_header_rtf(ts$title_hdr, ls, rs, pi$table_align)
  else
    ttls <- get_titles_rtf(ts$titles, ls, rs, pi$table_align) 
  
  
  if (!is.null(rs$page_by)) {
    pgby <- get_page_by_rtf(rs$page_by, rs$content_size[["width"]], 
                            pi$page_by, rs, pi$table_align)
  } else if(!is.null(ts$page_by))
    pgby <- get_page_by_rtf(ts$page_by, ls, pi$page_by, rs, pi$table_align)
  else 
    pgby <- c()
  
  
  a <- NULL
  if (content_blank_row %in% c("above", "both"))
    a <- paste0("\\par", a)
  
  
  blnks <- c()
  
  # Get row count by summing ..row variable
  # if ("..row" %in% names(pi$data)) {
  #   rcnt <- sum(pi$data$..row)
  # } else {
  #   rcnt <- nrow(pi$data) 
  # }
    
  # Determine sum of all lines
  rc <- sum(ttls$lines, pgby$lines, shdrs$lines, 
           hdrs$lines, rws$lines,
           length(a))
  
  # Get footnotes, passing in sum of all current lines
  ftnts <- get_page_footnotes_rtf(rs, ts, ls, lpg_rows, rc,
                                  wrap_flag, content_blank_row,  pi$table_align)

  # Deal with cell padding.  Don't count this in line count.
  cp <- paste0("\\li", rs$cell_padding, "\\ri", rs$cell_padding, rs$spacing_multiplier)
  
  # On LibreOffice, have to protect the table from the title width or
  # the table row will inherit the title row width. Terrible problem.
  tpt <- "{\\pard\\fs1\\sl0\\par}"
  if (any(ts$borders %in% c("all", "top", "outside"))) {
    if (ttls$border_flag | rs$page_template$titles$border_flag |  
        rs$page_template$title_hdr$border_flag)
      tpt <- ""
    
    if (length(pgby) > 0) {
      if (pgby$border_flag)
        tpt <- ""
    }
  }
  if (titles_aligned(ts$titles, ts$title_hdr))
    tpt <- ""

  
  # Same thing as above with titles.  If footnote block is contiguous
  # with table and footnotes are wider than table, row width of footnotes
  # will infect row width of table.  On libre only.  So this is to protect
  # the table width.
  bpt <- "{\\pard\\fs1\\sl0\\par}"
  if (any(ts$borders %in% c("all", "top", "outside", "body"))) {
    if (!is.null(ftnts)) {
      if (ftnts$border_flag)
        bpt <- ""
    }
    
    if (!is.null(rs$page_template$footnotes)) {
      if (rs$page_template$footnotes$border_flag)
        bpt <- ""
    }
  }
  if (footnotes_aligned(ts$footnotes)) 
    bpt <- ""
  
  # Continuous tables are different
  # Titles only on first page and footnotes only on last page
  if (ts$continuous) {
   
    if ("first" %in% pgind & "last" %in% pgind) {
      
        ret <- list(rtf = c(a, cp, ttls$rtf, pgby$rtf, shdrs$rtf, 
                            hdrs$rtf, rws$rtf, ftnts$rtf),
                    lines = rc + ftnts$lines)
        
    } else if ("first" %in% pgind) {
      ret <- list(rtf = c(a, cp, ttls$rtf, pgby$rtf, shdrs$rtf, 
                          hdrs$rtf, rws$rtf),
                  lines = rc )
      
    } else if ("last" %in% pgind) {
      ret <- list(rtf = c(pgby$rtf, 
                          rws$rtf, ftnts$rtf),
                  lines = rc  + ftnts$lines - ttls$lines)
      
    } else {
      
      ret <- list(rtf = c( pgby$rtf,  
                          rws$rtf),
                  lines = rc  - ttls$lines)
      
    }
    
  } else {
    
    # By default, repeat titles and footnotes on every page
    ret <- list(rtf = c(a, cp, ttls$rtf, cp, pgby$rtf, tpt, cp, shdrs$rtf, 
                        hdrs$rtf, rws$rtf, bpt, cp, ftnts$rtf),
                lines = rc  + ftnts$lines)
  }
    
  return(ret) 
}

get_page_footnotes_rtf <- function(rs, spec, spec_width, lpg_rows, row_count,
                                   wrap_flag, content_blank_row, talgn) {
  
  ftnts <- list(lines = 0, twips = 0, border_flag = FALSE)
  vflag <- "none"
  
  # Deal with valign parameter
  if (!is.null(spec$footnotes)) {
    if (!is.null(spec$footnotes[[length(spec$footnotes)]])) {
      if (spec$footnotes[[length(spec$footnotes)]]$valign == "bottom") {
        
        vflag <- "bottom"
        ftnts <- get_footnotes_rtf(spec$footnotes, 
                                   spec_width, rs, 
                                   talgn) 
      } else {
        vflag <- "top"
        ftnts <- get_footnotes_rtf(spec$footnotes, spec_width, rs, talgn) 
      }
      
    }
  } else {
    
    if (!is.null(rs$footnotes[[1]])) {
      if (!is.null(rs$footnotes[[1]]$valign)) {
        if (rs$footnotes[[1]]$valign == "top") {
          vflag <- "top"
          ftnts <- get_footnotes_rtf(rs$footnotes, 
                                     spec_width, rs, 
                                     talgn) 
        } else {
          
          if (wrap_flag)
            vflag <- "bottom" 
        }
      }
    }
  }
  
  b <- NULL
  blen <- 0
  if (content_blank_row %in% c("below", "both")) {
    b <- "\\par"
    blen <- 1
  }
  # } else {
  #   b <- paste0("\\fs1\\sl0\\par\\pard", rs$font_rtf, rs$spacing_multiplier)
  # }
  
  # Add extra offsets if table has a lot of borders turned on
  # to avoid undesired page wraps
  boff <- 0
  if (any(class(spec) == "table_spec") &
      any(spec$borders %in% c("all", "inside"))) {

    boff <- round(row_count * rs$border_height / rs$row_height)
  }
  
  ublnks <- c()
  lblnks <- c()
  
  # Determine number of filler lines needed
  len_diff <- rs$body_line_count - row_count - ftnts$lines - lpg_rows - blen - boff


  if (vflag == "bottom" & len_diff > 0) {
  
      ublnks <- c(b, rep("\\par", len_diff))
  
  } else {
    
    if ((wrap_flag & len_diff > 0)) {
      if (vflag == "bottom" | has_bottom_footnotes(rs))
        lblnks <- c(rep("\\par", len_diff), b)
    } else {
      lblnks <- b
    }
      
    
  }
  
  tlns <- sum(ftnts$lines, length(ublnks), length(lblnks))
  ret <- list(rtf = c(ublnks, ftnts$rtf, lblnks),
              lines = tlns,
              twips = tlns * rs$twip_conversion,
              border_flag = ftnts$border_flag)
  
  return(ret)
}

# Sub-Functions ---------------------------------------------------------


#' Get content offsets for table header, titles, footnotes, and content blanks.
#' Needed to calculate page breaks accurately.
#' @return A vector of upper and lower offsets
#' @noRd
get_content_offsets_rtf <- function(rs, ts, pi, content_blank_row, pgby_cnt = NULL) {
  
  # ret <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
  # cnt <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
  
  # Width is normally the width of the table, not the page
  wdth <- rs$content_size[["width"]]
  if (!is.null(pi$col_width))
    wdth <- sum(pi$col_width)
  
  # Default to zero
  shdrs <- list(lines = 0, twips = 0)
  hdrs <- list(lines = 0, twips = 0)
  
  if (ts$headerless == FALSE) {
    shdrs <- get_spanning_header_rtf(rs, ts, pi)   

    hdrs <- get_table_header_rtf(rs, ts, pi$col_width, 
                                 pi$label, pi$label_align, 
                                 pi$table_align)  
  }
  
  # Get title headers or titles
  if (is.null(ts$title_hdr))
    ttls <- get_titles_rtf(ts$titles, wdth, rs) 
  else 
    ttls <- get_title_header_rtf(ts$title_hdr, wdth, rs)
  
  # Get page by if it exists
  # if (!is.null(ts$page_by))
  #   pgb <- get_page_by_rtf(ts$page_by, wdth, NULL, rs, pi$table_align, pgby_cnt)
  # else if (!is.null(rs$page_by))
  #   pgb <- get_page_by_rtf(rs$page_by, wdth, NULL, rs, pi$table_align, pgby_cnt)

  page_by_info <- NULL
  if (!is.null(ts$page_by)) {
    page_by_info <- ts$page_by
  } else if (!is.null(rs$page_by)) {
    page_by_info <- rs$page_by
  }
  
  # Get lines for each page by value
  if (!is.null(page_by_info)) {
    ret <- c(upper = list(), lower = 0, blank_upper = 0, blank_lower = 0)
    cnt <- c(upper = list(), lower = 0, blank_upper = 0, blank_lower = 0)
    
    pgby_unique <- unique(pi$data$..page_by)

    for (i in 1:length(pgby_unique)) {
      pgby_temp <- get_page_by_rtf(page_by_info, wdth, pgby_unique[i], rs, pi$table_align, pgby_cnt = pgby_cnt)
      
      # Add everything up
      ret[["upper"]][[i]] <- shdrs$twips + hdrs$twips + ttls$twips + pgby_temp$twips
      cnt[["upper"]][[i]] <- shdrs$lines + hdrs$lines + ttls$lines + pgby_temp$lines
    }
    
    names(ret[["upper"]]) <- pgby_unique
    names(cnt[["upper"]]) <- pgby_unique
    
  } else {
    pgb <- list(lines = 0, twips = 0)
    pgb <- get_page_by_rtf(page_by_info, wdth, NULL, rs, pi$table_align, pgby_cnt = pgby_cnt)
    
    ret <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
    cnt <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
    
    # Add everything up
    ret[["upper"]] <- shdrs$twips + hdrs$twips + ttls$twips + pgb$twips
    cnt[["upper"]] <- shdrs$lines + hdrs$lines + ttls$lines + pgb$lines
    
  }
  
  #print(length(pgb))
  
  # print(paste("Table titles:", ttls))
  
  # Add everything up
  # ret[["upper"]] <- shdrs$twips + hdrs$twips + ttls$twips + pgb$twips
  # cnt[["upper"]] <- shdrs$lines + hdrs$lines + ttls$lines + pgb$lines
  
  if (content_blank_row %in% c("above", "both")) {
    ret[["blank_upper"]] <- rs$line_height
    cnt[["blank_upper"]] <- 1 
  }
  
  ftnts <- get_footnotes_rtf(ts$footnotes, wdth, rs) 
  rftnts <- get_footnotes_rtf(rs$footnotes, wdth, rs)
  
  if (has_top_footnotes(rs)) {
    ret[["lower"]] <- ftnts$twips + rftnts$twips
    cnt[["lower"]] <- ftnts$lines + rftnts$lines
  } else {
    ret[["lower"]] <- ftnts$twips
    cnt[["lower"]] <- ftnts$lines
  }
  
  # Add extra offsets if table has a lot of borders turned on
  # to avoid undesired page wraps
  if (any(ts$borders %in% c("all", "inside"))) {
    ret[["lower"]] <- ret[["lower"]] + (rs$row_height * 2)
    cnt[["lower"]] <- cnt[["lower"]] + 2
  }
    
  if (content_blank_row %in% c("both", "below")) {
    ret[["blank_lower"]] <- rs$line_height
    cnt[["blank_lower"]] <- 1 
  }
  
  res <- list(lines = cnt,
              twips = ret)
  
  return(res)
  
}


#' @description Return a vector of strings for the table header
#' @details Basic idea of this function is to create a list
#' of string vectors of label words sized according to the column 
#' widths, then combine by line/row, and concatenate everything and justify.
#' @noRd
get_table_header_rtf <- function(rs, ts, widths, lbls, halgns, talgn) {
  
  ret <- c()
  cnt <- 0
  rh <- rs$row_height
  tbl <- ts$data
  conv <- rs$twip_conversion
  nms <- names(lbls)
  nms <- nms[!is.controlv(nms)]
  
  # Get cell widths
  sz <- c()
  for (k in seq_along(widths)) {
    if (!is.control(nms[k])) {
      if (k == 1)
        sz[k] <- round(widths[k] * conv)
      else 
        sz[k] <- round(widths[k] * conv + sz[k - 1])
    }
  }
  
  # Table alignment
  ta <- "\\trql"
  if (talgn == "right")
    ta <- "\\trqr"
  else if (talgn %in% c("center", "centre"))
    ta <- "\\trqc"

  
  brdrs <- ts$borders
  if (length(ts$col_spans) > 0) {
    
    if (any(ts$borders %in% c("outside")))
      brdrs <- c("left", "right")
    
    if (any(ts$borders %in% "top"))
      brdrs <- brdrs[!brdrs %in% "top"]
    
  }
  
  # Header Cell alignment
  ha <- c()
  for (k in seq_along(halgns)) {
    if (!is.control(nms[k])) {
      if (halgns[k] == "left")
        ha[k] <- "\\ql"
      else if (halgns[k] == "right")
        ha[k] <- "\\qr"
      else if (halgns[k] %in% c("center", "centre"))
        ha[k] <- "\\qc"
    }
  }
  
  # Table Header
  # Request from Astellas
  # Need to watch this for undesired side effects.
  if (ts$continuous) 
    ret[1] <-  paste0("\\trowd\\trkeep\\trhdr\\trgaph0", ta, "\\trrh", rh)  
  else 
    ret[1] <-  paste0("\\trowd\\trgaph0", ta, "\\trrh", rh)
  
  # Loop for cell definitions
  for(j in seq_along(nms)) {
    if (!is.control(nms[j])) {
      b <- get_cell_borders(1, j, 2, length(nms), brdrs)
      ret[1] <- paste0(ret[1], "\\clvertalb", b, "\\clbrdrb\\brdrs\\cellx", sz[j])
    }
  }
  
  cnt <-  1 
  
  # Loop for column names
  pdf(NULL)
  par(family = get_font_family(rs$font), ps = rs$font_size)
  
  for(k in seq_along(nms)) {
    if (!is.control(nms[k])) {
      
      # Split label strings if they exceed column width
      tmp <- split_string_rtf(lbls[k], widths[k], rs$units, rs$font)
      
      tb <- tmp$rtf
      if (ts$header_bold)
        tb <- paste0("\\b ", tmp$rtf, "\\b0")
      
      ret[1] <- paste0(ret[1], ha[k], " ", tb, "\\cell")

      # Add in extra lines for labels that wrap
      xtr <- tmp$lines
      if (xtr > cnt)
        cnt <- xtr
    }
  }
  dev.off()
  
  ret[1] <- paste(ret[1], "\\row")
  
  if (ts$first_row_blank == TRUE) {
    
    if (any(brdrs == "body"))
      b <- get_cell_borders(1, 1, 3, 1, c("left", "right"))
    else
      b <- get_cell_borders(1, 1, 3, 1, brdrs)
    
    w <- round(sum(widths, na.rm = TRUE) * conv)
    
    ret[1] <- paste0(ret[1], "\n\\trowd\\trgaph0", ta, 
                     "\\trrh", rh, b, "\\cellx", w,
                     "\\ql\\cell\\row")
   cnt <- cnt + 1
  }
  
  res <- list(rtf = ret,
              lines = cnt,
              twips = cnt * rh)
  
  return(res)

}

# Need to fix this
#' @description Return a vector of rtf strings for the table spanning headers
#' @details Basic idea of this function is to figure out which columns 
#' the header spans, add widths, then call get_table_header.  Everything
#' from there is the same.  
#' @import stats
#' @noRd
get_spanning_header_rtf <- function(rs, ts, pi) {
  
  spns <- ts$col_spans
  cols <- pi$keys
  cols <- cols[!is.controlv(cols)]
  w <- pi$col_width 
  w <- w[cols]
  gutter <- 0
  cnt <- c()
  
  # print("Cols:")
  # print(cols)
  #print(w)
  wlvl <- get_spanning_info(rs, ts, pi, w, gutter)
  lvls <- sort(seq_along(wlvl), decreasing = TRUE)
  
  # At this point we have a data frame with labels, etc. and spanning
  # column widths, and are ready to create spanning header rows
  #print(wlvl)
  
  conv <- rs$twip_conversion
  talgn <- pi$table_align
  
  # Detect continuous underline and insert gap, adjust widths
  # Do not create gaps for all and inside border
  if (length(wlvl) > 0 & all(!ts$borders %in% c("all", "inside"))) {
    wlvl <- get_spanning_gap_rtf(wlvl, conv)
  }

  # Table alignment
  ta <- "\\trql"
  if (talgn == "right")
    ta <- "\\trqr"
  else if (talgn %in% c("center", "centre"))
    ta <- "\\trqc"

  # Get borders
  brdrs <- ts$borders
  if (any(ts$borders %in% "bottom"))
    brdrs <- brdrs[!brdrs %in% "bottom"]
  
  rh <- rs$row_height
  
  # Format labels for each level
  ln <- c()
  for (l in lvls) {
    
    s <- wlvl[[l]]
    
    widths <- s$width
    names(widths) <- s$name
    algns <- s$align
    names(algns) <- s$name
    lbls <- s$label
    names(lbls) <- s$name
    cs <- s$col_span
    
    # Get cell widths
    sz <- c()
    for (k in seq_along(widths)) {

      if (k == 1)
        sz[k] <- round(widths[k] * conv)
      else
        sz[k] <- round(widths[k] * conv + sz[k - 1])
      
    }
    
    # Header Cell alignment
    ha <- c()
    for (k in seq_along(algns)) {

      if (algns[k] == "right")
        ha[k] <- "\\qr"
      else if (algns[k] %in% c("center", "centre"))
        ha[k] <- "\\qc"
      else
        ha[k] <- "\\ql"
      
    }

    r <- ""
    cnt[length(cnt) + 1] <- 1 
    
    # Table Header
    r <-  paste0("\\trowd\\trgaph0", ta, "\\trrh", rh)
    
    # Label justification, width, and row concatenation

    # Loop for cell definitions
    for(j in seq_along(sz)) {
      
      # Get cell borders
      b <- get_cell_borders(length(wlvl) - l + 1, j, 10, nrow(s), brdrs)
      
      # Add colspans
      vl <- lbls[j]
      
      # Special handling of borders for different situations.  
      # I hate this, but can't see a way around it.
      if (any(brdrs %in% c("all", "inside"))) {
        sflg <- "B"
        if (j > 1) {
          if (cs[j] > 1)
            sflg <- ""
        }
        
        if (vl == "") {
          bb <- get_cell_borders(length(lvls) - l + 1, j, length(lvls), 
                                      length(sz), brdrs, 
                                      flag = sflg)
        } else 
          bb <- b
      } else if (all(brdrs == "outside")) {
        
        if (vl == "")
          bb <- b
        else {
          
          bb <- b #paste0(b, "\\clbrdrb\\brdrs")
          
          
        }
        
      } else {
        
        if (vl == "") {
          bb <- get_cell_borders(length(lvls) - l + 1, j, length(lvls), 
                                      length(sz), brdrs, 
                                      flag = "")
        } else 
          bb <- b  #paste0(b, "\\clbrdrb\\brdrs")
        
      }
      
      
      if (s$span[j] > 0 & s$underline[j])
        r <- paste0(r, "\\clvertalb", bb, "\\clbrdrb\\brdrs\\cellx", sz[j])
      else 
        r <- paste0(r, "\\clvertalb", bb, "\\cellx", sz[j])
    }
    
    # Open device context
    pdf(NULL)
    par(family = get_font_family(rs$font), ps = rs$font_size)
    
    # Loop for labels
    for(k in seq_along(lbls)) {

      # Split label strings if they exceed column width
      tmp <- split_string_rtf(lbls[k], widths[k], rs$units, rs$font)
      
      if (s$bold[k])
        tb <- paste0("\\b ", tmp$rtf, "\\b0")
      else 
        tb <- paste0(" ", tmp$rtf)
      
      # Concat label
      r <- paste0(r, ha[k], tb, "\\cell")
      # print(lbls[k])
      # print(widths[k])
      # Add in extra lines for labels that wrap
      xtr <- tmp$lines
      if (xtr > cnt[length(cnt)])
        cnt[length(cnt)] <- xtr
      
    }
    dev.off()
    
    
    r <- paste0(r, "\\row")
    
    ln[[length(ln) + 1]] <- r
    
  }
  
  
  ret <- unlist(ln)
  
  res <- list(rtf = ret, 
              lines = sum(cnt), 
              twips = sum(cnt) * rh)
  
  return(res)
}

#' @description Check gap and insert blank for RTF
#' @details Check gap and insert blank for RTF
#' @noRd
get_spanning_gap_rtf <- function(wlvl, conv, gap_twips = 100){
  
  ret <- list()
  
  for (k in 1:length(wlvl)) {
    
    df <- wlvl[[k]]
    
    # Extract spanning and underline only for detection
    df_span <- df[df$span > 0 & df$underline == TRUE,] 
    
    if (nrow(df_span) < 2) {
      ret[[k]] <- df
    } else {
      gap_width <- gap_twips/conv
      
      df_span$order <- as.numeric(stri_extract_first_regex(df_span$name, "\\d+"))
      
      # Empty template for gap data
      df_gap <- na.omit(
        data.frame(
          span = NA_integer_,
          width = NA,
          label = NA_character_,
          align = NA_character_,
          n = NA_integer_,
          name = NA_character_,
          underline = NA,
          col_span = NA_integer_,
          bold = NA,
          order = NA_integer_
        )
      )
      
      for (i in 1:nrow(df_span)) {
        
        # Temporary gap data for stacking
        df_gap_temp <- data.frame(
          span = -1,
          width = NA,
          label = "",
          align = "",
          n = NA_integer_,
          name = NA_character_,
          underline = TRUE,
          col_span = 1,
          bold = FALSE,
          order = NA_integer_
        )
        
        cur_order <- df_span$order[i]
        
        # Process by first, last, and others
        if (i == 1){
          next_order <- df_span$order[i + 1]
          
          # Continuous underline detection
          if (next_order - cur_order== 1) {
            
            # Create the gap data after this spanning cell
            df_gap_temp$width <- gap_width
            df_gap_temp$order <- cur_order + 0.5
            df_gap_temp$name <- paste0("Span", df_gap_temp$order)
            df_gap <- rbind(df_gap, df_gap_temp)
            
            # Reduce the gap width for spanning cell
            df_span$width[i] <- df_span$width[i] - (gap_width/2)
          }
        } else if (i == nrow(df_span)){
          pre_order <- df_span$order[i - 1]
          
          # Continuous underline detection
          if (cur_order - pre_order == 1) {
            
            # Reduce the gap width for spanning cell
            df_span$width[i] <- df_span$width[i] - (gap_width/2)
          }
          
        } else {
          pre_order <- df_span$order[i - 1]
          next_order <- df_span$order[i + 1]
          
          # Continuous underline detection
          if (next_order - cur_order == 1 |
              cur_order - pre_order == 1) {
            
            # Create the gap data after this spanning cell
            if (next_order - cur_order == 1) {
              df_gap_temp$width <- gap_width
              df_gap_temp$order <- cur_order + 0.5
              df_gap_temp$name <- paste0("Span", df_gap_temp$order)
              df_gap <- rbind(df_gap, df_gap_temp)
              
              # Reduce the gap width for spanning cell
              df_span$width[i] <- df_span$width[i] - (gap_width/2)
            }
            
            if (cur_order - pre_order == 1) {
              # Reduce the gap width for spanning cell
              df_span$width[i] <- df_span$width[i] - (gap_width/2)
            }
          }
        }
      } # End of data row loop
      
      # Drop order
      df_span <- df_span[, setdiff(names(df_span), "order")]
      df_gap <- df_gap[, setdiff(names(df_gap), "order")]
      
      # Stack together and sort
      df_ret <- rbind(df_span, df_gap, df[!(df$span > 0 & df$underline == TRUE),])
      df_ret <- df_ret[order(df_ret$name),]
      
      ret[[k]] <- df_ret
      
    } # End of gap process
  }
  
  return(ret)
}

#' @description This function counts lines per row independently because
#' the ..row field does not account for page wrapping.  Need number
#' of lines on this particular page.
#' @noRd
get_table_body_rtf <- function(rs, tbl, widths, algns, talgn, tbrdrs, 
                               frb, continuous = FALSE, styles, pgind, ts) {
  
  if ("..blank" %in% names(tbl))
    flgs <- tbl$..blank
  else 
    flgs <- NA
  
  # Count lines per row
  rws <- c()
  
  nms <- names(widths)
  nms <- nms[!is.na(nms)]
  nms <- nms[!is.controlv(nms)]
  wdths <- widths[nms]
  
  if (length(nms) == 1) {
    t <- as.data.frame(tbl[[nms]])
    names(t) <- nms
  } else 
    t <- tbl[ , nms]

  brdrs <- tbrdrs
  if (all(tbrdrs == "body"))
    brdrs <- c("top", "bottom", "left", "right")
  
  # Deal with outside borders on continuous tables
  if (continuous & "outside" %in% brdrs) {
    if ("first" %in% pgind & !"last" %in% pgind) {
      brdrs <- c("top", "left", "right") 
    } else if (!"first" %in% pgind & "last" %in% pgind) {
      brdrs <- c("bottom", "left", "right") 
    } else if (!"first" %in% pgind & !"last" %in% pgind) {
      brdrs <- c("left", "right") 
    }
  }
  
  # Get line height.  Don't want to leave editor default.
  rh <- rs$row_height
  conv <- rs$twip_conversion
  
  
  # Get cell widths
  sz <- c()
  for (k in seq_along(wdths)) {
    if (!is.control(nms[k])) {
      if (k == 1)
        sz[k] <- round(wdths[k] * conv)
      else 
        sz[k] <- round(wdths[k] * conv + sz[k - 1])
    }
  }
  
  # Table alignment
  ta <- "\\trql"
  if (talgn == "right")
    ta <- "\\trqr"
  else if (talgn %in% c("center", "centre"))
    ta <- "\\trqc"
  
  # Cell alignment
  ca <- c()
  for (k in seq_along(algns)) {
    if (!is.control(nms[k])) {
      if (algns[k] == "left")
        ca[k] <- "\\ql"
      else if (algns[k] == "right")
        ca[k] <- "\\qr"
      else if (algns[k] %in% c("center", "centre"))
        ca[k] <- "\\qc"
    }
  }
  
  ret <- c()
  
  pdf(NULL)
  par(family = get_font_family(rs$font), ps = rs$font_size)
  
  # For break labels, prepare the break label columns for later use
  defs <- ts$col_defs
  break_label_col <- c()
  if (any(grepl("..break_label", names(tbl)))) {
    
    for (d in 1:length(defs)) {
      if (!is.null(defs[[d]]$break_label)){
        break_label_col <- c(break_label_col, defs[[d]]$var_c)
      }
    }
  }
  
  # Table Body
  for(i in seq_len(nrow(t))) {
    
    
    if (i ==  1 & continuous == FALSE)
      ret[i] <- paste0("{\\trowd\\trgaph0\\trrh", rh, ta)
    else 
      ret[i] <- paste0("\\trowd\\trgaph0\\trrh", rh, ta)
    
    

    if (flgs[i] %in% c("B", "A", "L")) {
      
      # Deal with blank and label rows
      radj <- 0
      if (frb == TRUE) 
        radj <- 1
      
      cell_border <- NULL
      if ("..group_border" %in% names(tbl)) {
        cell_border <- tbl[["..group_border"]][i]
      }
      
      # Don't draw group line for first blank row
      if (flgs[i] %in% c("B", "A") & i == 1){
        cell_border <- NULL
      }
      
      b <- get_cell_borders(i + radj, 1, nrow(t) + radj, ncol(t), brdrs, flgs[i],
                            cell_border = cell_border)
      ret[i] <- paste0(ret[i], b, "\\cellx", max(sz))
      
      
    } else {
      
      # Loop for cell definitions
      for(j in seq_len(ncol(t))) {
        if (!is.control(nms[j])) {
          radj <- 0
          if (frb == TRUE) 
            radj <- 1
          
          cell_border <- NULL
          if ("..group_border" %in% names(tbl)) {
            cell_border <- tbl[["..group_border"]][i]
          }
          
          b <- get_cell_borders(i + radj, j, nrow(t) + radj, ncol(t), brdrs, flgs[i],
                                cell_border = cell_border)
          ret[i] <- paste0(ret[i], b, "\\cellx", sz[j])
        }
      }
    }
    
    mxrw <- 1
    
    # Prepare indenting convert
    conv_twips <- rs$twip_conversion
  
    if (flgs[i] %in% c("B", "A", "L")) {
      
      # Deal with label rows
      vl <- t[i, 1]
      
      # Deal with styles
      # if (nms[1] %in% names(styles)) {
      #   stl <- styles[[nms[1]]]
      #   if (stl$bold == TRUE) {
      #     bflg <- FALSE
      #     if (!is.null(stl$indicator)) {
      #       if ("labelrow" %in% stl$indicator) {
      #         if (flgs[i] %in% c("L"))
      #           bflg <- TRUE
      #       } 
      #       if (stl$indicator %in% names(tbl)) {
      #         if (!is.null(tbl[[i, stl$indicator]] )) {
      #           if (tbl[[i, stl$indicator]] == TRUE) {
      #             bflg <- TRUE   
      #           }
      #         }
      #       } 
      #     } else {
      #       bflg <- TRUE 
      #     }
      #     if (bflg) {
      #       vl <- paste0("\\b ", vl, "\\b0")
      #     }
      #   }
      # }
      
      stl <- get_cell_styles(nms[1], styles, flgs, i, tbl)
      
      # Get indenting information
      rtf_ind <- ""
      
      # For break labels, use the indenting info from original column
      repeat_break_label <- FALSE
      if (any(grepl("..break_label", names(tbl)))) {
        if (!is.na(tbl$..break_occur[i])) {
          cur_break_label_col <- break_label_col[as.numeric(tbl$..break_occur[i])]
          if (!is.null(defs[[cur_break_label_col]]$indent)) {
            ind <- floor(defs[[cur_break_label_col]]$indent*conv_twips)
            rtf_ind <- paste0("\\li", ind, " ")
          }
          repeat_break_label <- TRUE
        }
      }
      
      if (!repeat_break_label) {
        if (!is.null(defs[[nms[1]]]$indent)) {
          
          ind <- floor(defs[[nms[1]]]$indent*conv_twips)
          rtf_ind <- paste0("\\li", ind, " ")
          
        } else if (nms[1] == "stub" & !is.null(ts$stub)) {
          
          stub_var <- tbl$..stub_var[i]
          if (!is.null(defs[[stub_var]]$indent)) {
            ind <- floor(defs[[stub_var]]$indent*conv_twips)
            rtf_ind <- paste0("\\li", ind, " ")
          }
        }
      }
      
      vl <- paste0(rtf_ind, vl)
      
      if ("bold" %in% stl) {
        vl <- paste0("\\b ", vl, "\\b0")
      }
      
      ret[i] <- paste0(ret[i], ca[1], " ", vl, "\\cell\\li", rs$cell_padding)
      
      
      cl <- strsplit(vl, "\\line", fixed = TRUE)[[1]]
      if (length(cl) > mxrw)
        mxrw <- length(cl)
      
    } else {
      
      # Loop for cell values
      for(j in seq_len(ncol(t))) {
        
  
        if (!is.control(nms[j])) {
          
          tb <- t[i, j]
          # if (nms[j] %in% names(styles)) {
          #   stl <- styles[[nms[j]]]
          #   if (stl$bold == TRUE) {
          #     bflg <- TRUE
          #     if (!is.null(stl$indicator)) {
          #       if ("datarow" %in% stl$indicator) {
          #         if (flgs[i] %in% c("B", "L"))
          #           bflg <- FALSE
          #         
          #       } else {
          #         bflg <- FALSE
          #         if (stl$indicator %in% names(tbl)) {
          #           if (!is.null(tbl[[i, stl$indicator]] )) {
          #             if (tbl[[i, stl$indicator]] == TRUE) {
          #               bflg <- TRUE   
          #             }
          #           }
          #         } 
          #         
          #       }
          #     } 
          #     if (bflg) {
          #       tb <- paste0("\\b ", tb, "\\b0")
          #     }
          #   }
          # }
          
          stl <- get_cell_styles(nms[j], styles, flgs, i, tbl)
          
          # Get indenting information
          rtf_ind <- ""
          if (!is.null(defs[[nms[j]]]$indent)) {
            
            ind <- floor(defs[[nms[j]]]$indent*conv_twips)
            rtf_ind <- paste0("\\li", ind, " ")
            
          } else if (nms[j] == "stub" & !is.null(ts$stub)) {
            
            stub_var <- tbl$..stub_var[i]
            if (!is.null(defs[[stub_var]]$indent)) {
              ind <- floor(defs[[stub_var]]$indent*conv_twips)
              rtf_ind <- paste0("\\li", ind, " ")
            }
          }
          tb <- paste0(rtf_ind, tb)
          
          if ("bold" %in% stl) {
            tb <- paste0("\\b ", tb, "\\b0") 
          }
          
          # Construct rtf
          ret[i] <- paste0(ret[i], ca[j], " ", tb, "\\cell\\li", rs$cell_padding)
          
          vl <- t[i, j]
          if (all(class(vl) != "character"))
            vl <- as.character(vl)
          
          # Count lines in cell 
          cl <- strsplit(vl, "\\line", fixed = TRUE)[[1]]
          if (length(cl) > mxrw)
              mxrw <- length(cl)
        }
        
      }
    }
    
    rws[i] <- mxrw
    
    ret[i] <- paste0(ret[i], "\\row")
    
    
  }
  
  dev.off()
  
  if (!continuous) {

    ret[length(ret)] <- paste0(ret[length(ret)], "}\\pard",
                               rs$font_rtf, rs$spacing_multiplier)
  
  } else {
    
    ret[length(ret)] <- paste0(ret[length(ret)], "")  # Not needed

  }
  
  
  res <- list(rtf = ret,
              lines = sum(rws))
  
  return(res)
  
  
}

# Utility Functions -------------------------------------------------------

titles_aligned <- function(ttls = NULL, ttlhdrs = NULL) {
 
  lst <- ttlhdrs
  if (!is.null(ttls)) 
    lst <- ttls
  
  ret <- FALSE
  
  if (!is.null(lst)) {
  
    lastttl <- lst[[length(lst)]]
  
    if (lastttl$width == "content")
      ret <- TRUE
  }
  
  return(ret)
  
}


footnotes_aligned <- function(ftnts = NULL) {
  
  lst <- ftnts
  
  ret <- FALSE
  
  if (!is.null(lst)) {
    
    lastfnt <- lst[[1]]
    
    if (lastfnt$width == "content" & lastfnt$valign == "top")
      ret <- TRUE
  }
  
  return(ret)
  
}

#' @import crayon
#' @noRd
print_title_header <- function(t_hdr) {
  
  # Print title header
  if (!is.null(t_hdr)) {
    
    ttlcnt <- 1
    
    for (i in seq_along(t_hdr)) {
      
      mx <- max(length(t_hdr[[i]]$titles), length(t_hdr[[i]]$right))
      
      for (j in seq(1, mx)) {
        
        cat(paste0("- title header ", as.character(ttlcnt), ": "))
        if (!is.na(t_hdr[[i]]$titles[j]))
          cat(paste0("'", t_hdr[[i]]$titles[j], "' "))
        
        if (!is.na(t_hdr[[i]]$right[j]))
          cat(paste0("right='", t_hdr[[i]]$right[j], "'"))
        
        cat("\n")
        
      }
      
      ttlcnt <- ttlcnt + 1
    }
  } 
  
  
}

tslst <- c("one")
is.na(tslst[2])


#' @import crayon
#' @noRd
print_titles <- function(ttls) {
  
  # Print titles
  if (!is.null(ttls)) {
    
    ttlcnt <- 1
    for (i in seq_along(ttls)) {
      
      for (j in seq_along(ttls[[i]]$titles)) {
        cat(paste0("- title ", as.character(ttlcnt), ": '" 
            , substring(ttls[[i]]$titles[[j]], 1), "'\n"))
        ttlcnt <- ttlcnt + 1
      }
      
    }
  } 
  
  
}


#' @import crayon
#' @noRd
print_footnotes <- function(ftnts) {
  
  # Print footnotes
  if (!is.null(ftnts)) {
    
    ftncnt <- 1
    # There can be more than 1 footnote block
    for (i in seq_along(ftnts)) {
      
      for (j in seq_along(ftnts[[i]]$footnotes)) {
        cat(paste0("- footnote ", as.character(ftncnt), ": '" 
            , substring(ftnts[[i]]$footnotes[[j]], 1), "'\n"))
        ftncnt <- ftncnt + 1
      }
      
    }
  } 
}

#' @title reporter: A package for creating statistical reports
#'
#' @description The \strong{reporter} package creates statistical reports in TXT, 
#' RTF, PDF, HTML, and DOCX file formats.  Features include automatic page wrapping and 
#' breaking for wide and long tables, spanning headers, titles, footnotes, 
#' page headers, page footers, and page numbering.  The package allows mixing
#' of multiple tables, text, and plots in a single report, or even on a single
#' page.     
#'     
#' @details 
#' The \strong{reporter} package creates regulatory-style, statistical reports. It 
#' was designed to create tables, listings, and figures (TLFs) 
#' for use in the pharmaceutical, biotechnology, and medical-device industries.
#' However, the functions are generalized enough to provide statistical reporting
#' for any industry.  The package is written primarily in Base R, and has 
#' few dependencies on other packages.
#' 
#' The package is intended to give R programmers 
#' flexible report layout capabilities, and a choice of output formats. 
#' The package will 
#' initially focus on printable, file-based 
#' output formats, as there are already numerous R packages that 
#' provide tabular reporting in HTML. 
#' 
#' PDF output may have limitations not associated with other output types.  
#' See \link{NotesOnPDF}
#' for more information.
#'     
#' @section Key Features:
#' The \strong{reporter} package contains the following key features:
#' \itemize{
#'   \item Titles, footnotes, page header, and page footer are repeated on each page
#'   \item Supports header labels and spanning headers 
#'   \item Calculates default columns widths automatically
#'   \item Includes automatic wrapping and splitting of wide and long tables
#'   \item Integrates with the \strong{\link[fmtr]{fmtr}} package to format 
#'         numeric, date, and character data
#'   \item Plots from the popular \strong{\link[ggplot2]{ggplot2}} package can 
#'         be added to RTF, PDF, HTML, and DOCX reports
#'   \item Allows appending multiple tables to a report, multiple tables to a page, 
#'         and intermingling of text, tables, and plots
#'   \item Supports in-report date/time stamps and "Page X of Y" page numbering
#' }
#'
#' @section Key Functions:
#' \itemize{
#'   \item \code{\link{create_report}} to define a report
#'   \item \code{\link{create_table}} to define a table
#'   \item \code{\link{create_text}} to define text content
#'   \item \code{\link{create_plot}} to define a plot 
#'   \item \code{\link{write_report}} to write out the report
#' }
#' 
#' @section Package Assumptions and Limitations:
#' Note that the \strong{reporter} package is built on several assumptions, and 
#' has some limitations.  Those
#' assumptions and limitations are as follows:
#' \itemize{
#'   \item The current version supports both a monospace, fixed-width style 
#'   report and a variable-width style report.  The monospace report is the 
#'   default.  To create a variable width report, set the font parameter on
#'   the \code{create_report} function to 'Arial', 'Courier', or 'Times'.
#'   \item If a font is selected, it will be applied uniformly to the entire
#'   report.  The package has no capabilities to mix different fonts in the 
#'   same report.
#'   \item The package assumes that, except for formatting and layout, 
#'   the incoming data is ready for 
#'   printing. The \strong{reporter} package has no capabilities to perform 
#'   calculations, summaries, grouping, or derivations.  Use the many 
#'   capabilities in R to perform these operations prior to sending data
#'   to the \strong{reporter} package.
#'   \item It is assumed that the incoming data is sorted as desired.  
#'   The \strong{reporter} package
#'   has no capabilities for sorting.  Use the sorting functionality in 
#'   Base R or supplemental packages to sort the data prior to sending to
#'   \strong{reporter}.
#'   \item For monospace reports, titles, footnotes, page headers, and page 
#'   footers must fit in the available space. 
#'   If they don't, the \strong{reporter} package will 
#'   generate a warning.  In these situations, the recommended course of 
#'   action is to split the offending string into two or more strings that 
#'   fit within the available width.
#'   \item The \strong{reporter} package will never break a word. That means you 
#'   cannot set a column width that is less than the length of the longest 
#'   word.  If you wish to break words, add the breaks with
#'   \code{\link[base]{strwrap}} or an equivalent function before reporting.  
#'   \item The package will make a best guess on column widths based on the 
#'   width of the data and column headers.  To override the best guess,
#'   set the column width directly by placing a width attribute on the variable
#'   or by using the \code{\link{column_defaults}} or 
#'   \code{\link{define}} functions. 
#'   \item The max automatic column width
#'   is 5 inches.  Longer data values will be wrapped.
#'   \item The package support plots from \strong{ggplot2}. These plots
#'   can be added to RTF, PDF, HTML, and DOCX output types.  The package does not 
#'   support Base R plots.
#'   \item The package currently supports styling for HTML reports.  
#'   This styling allows setting background colors, font colors, border
#'   colors, and some font sizing and bolding.   
#' }
#' 
#' @import common
#' @aliases reporter-package
#' @keywords internal
#' @name reporter
"_PACKAGE"

#' @title Notes on PDF output type
#' @description 
#' PDF output supports a limited character set.  Some Unicode characters
#' can cause the PDF rendering to fail.  All Latin characters and other 
#' common characters like 
#' mathematical operators and Greek letters are supported.  However, 
#' other special characters and characters from Asian languages may not 
#' be supported, depending on the operating system and installed operating 
#' system languages. If a character is not supported, it will be replaced
#' with a question mark (?).
#' 
#' As of \strong{reporter} v1.1.3, the package will 
#' generate PDF files directly, without using \strong{rmarkdown} as an 
#' intermediary. This technique eliminates dependencies, makes the package 
#' easier to install, and greatly improves performance.  It is recommended
#' to use the most recent version of the package if possible.
#' 
#' For versions of \strong{reporter} prior to v1.1.3, the package has
#' additional dependencies and installation requirements.  See details below.
#' 
#' @details 
#' For \strong{reporter} v1.1.2 and below, 
#' PDF output type requires the \strong{rmarkdown} package and a 
#' LaTeX renderer like \strong{MiKTeX}.  
#' \strong{MiKTeX} must be installed separately, and will
#' not be installed as part of the \strong{reporter} install.  This program must
#' furthermore be added to the \code{path} environment variable on the system
#' where \strong{reporter} will run, so that \strong{reporter} can find it.
#' 
#' @name NotesOnPDF
#' @keywords internal
NULL

# The page template is everything except the content: page header/footer,
# titles, footnotes, etc.

# Page Template DOCX Functions ---------------------------------------------


#' Create a page template with header, titles, footnotes, and footer.
#' @param rs The report spec
#' @return The page template object
#' @noRd
page_template_docx <- function(rs) {
  
  pt <- structure(list(), class = c("page_template_docx", "list"))
  
  pt$page_header <- get_page_header_docx(rs)
  pt$title_hdr <- get_title_header_docx(rs$title_hdr, rs$line_size, rs)
  pt$titles <- get_titles_docx(rs$titles, rs$line_size, rs)
  pt$footnotes <- c()
  if (!is.null(rs$footnotes)) {
    if (!is.null(rs$footnotes[[1]])) {
      if (rs$footnotes[[1]]$valign == "bottom")
        pt$footnotes <- get_footnotes_docx(rs$footnotes, rs$line_size, rs)
    }
    
  } 
  pt$page_footer <- get_page_footer_docx(rs)
  
  pt$lines <- sum(pt$page_header$lines, pt$page_footer$lines,
                  pt$title_hdr$lines, pt$titles$lines, pt$footnotes$lines)
  
  
  # Page by not here.  Messes up line counts.
  
  return(pt)
}

#' @import grDevices
#' @noRd
get_page_header_docx <- function(rs) {
  
  ret <- ""
  cnt <- 0
  image_path <- c()
  
  if ((!is.null(rs$header_image_left) & is.null(rs$page_header_left)) |
      (!is.null(rs$header_image_center) & is.null(rs$page_header_center)) |
      (!is.null(rs$header_image_right) & is.null(rs$page_header_right))) {
    
    stop("`page_header` must be used when using `header_image`.")
  }
  
  # Make sure the length is 3, NA will be imputed later.
  width <- c(rs$page_header_width, rep(NA, 3 - length(rs$page_header_width)))
  
  # hl <- rs$page_header_left
  # hr <- rs$page_header_right
  # maxh <- max(length(hl), length(hr))
  
  image_left <- FALSE
  image_center <- FALSE
  image_right <- FALSE
  
  if (!is.null(rs$header_image_left)) {
    image_left <- TRUE
    hl <- rs$header_image_left
  } else{
    hl <- rs$page_header_left
  }
  
  if (!is.null(rs$header_image_right)) {
    image_right <- TRUE
    hr <- rs$header_image_right
  } else {
    hr <- rs$page_header_right
  }
  
  if (!is.null(rs$header_image_center)) {
    image_center <- TRUE
    hc <- rs$header_image_center
    
    # Default center cell is not displayed, open it when picture exists
    if (width[2] == 0) {
      width[2] <- NA
    }
  } else {
    hc <- rs$page_header_center
  }
  
  conv <- rs$twip_conversion
  rht <- get_row_height(round(rs$row_height * conv))
  
  hl_num <- ifelse(image_left, length(hl$image_path), length(hl))
  hc_num <- ifelse(image_center, length(hc$image_path), length(hc))
  hr_num <- ifelse(image_right, length(hr$image_path) , length(hr))
  
  maxh <- max(hl_num, hc_num, hr_num)
  
  # # User controlled width of left column
  # lwdth <- rs$page_header_width
  # if (is.null(lwdth))
  #   lwdth <- rs$content_size[["width"]]/2
  # 
  # # Calculate right column width
  # rwdth <- rs$content_size[["width"]] - lwdth
  # lpct <- round(5000 * lwdth / rs$content_size[["width"]])
  # rpct <- round(5000 * rwdth / rs$content_size[["width"]])
  
  u <- rs$units
  if (rs$units == "inches")
    u <- "in"

  if (maxh > 0) {
    
    # ret <- paste0("<w:tbl> ",
    #               "<w:tblPr>",
    #               '<w:tblStyle w:val="TableGrid"/>',
    #               '<w:tblW w:w="5000"',
    #               ' w:type="pct"/>',
    #               "</w:tblPr>",
    #               "<w:tblGrid>",
    #               '<w:gridCol w:w="', lpct, '" w:type="pct"/>',
    #               '<w:gridCol w:w="', rpct, '" w:type="pct"/>',
    #               "</w:tblGrid>", collapse = "")
    
    total_width <- sum(width, na.rm = T)
    if (total_width > rs$content_size[["width"]]) {
      
      stop(sprintf("Total width of page footer %s %s cannot be greater than content width %s %s.",
                   total_width,
                   rs$units,
                   rs$content_size[["width"]],
                   rs$units))
      
    } else {
      na_num <- sum(is.na(width))
      imputed_width <- (rs$content_size[["width"]] - total_width) / na_num
      
      left_width <- ifelse(is.na(width[1]), imputed_width, width[1])
      center_width <- ifelse(is.na(width[2]), imputed_width, width[2])
      right_width <- ifelse(is.na(width[3]), imputed_width, width[3])
      
      left_pct <- floor(5000 * left_width/rs$content_size[["width"]])
      center_pct <- floor(5000 * center_width/rs$content_size[["width"]])
      right_pct <- floor(5000 * right_width/rs$content_size[["width"]])
      
      # Make sure the total is 5000 exactly
      pct_lst <- c(left_pct, center_pct, right_pct)
      rest_pct <- 5000 - sum(pct_lst)
      add_pos <- c(1, 3 ,2)
      
      if (rest_pct >= 1){
        for (i in 1:rest_pct) {
          pct_lst[add_pos[i%%3]] <- pct_lst[i%%3] + 1
        }
      }
    }
    
    width <- c(left_width, center_width, right_width)
    
    if (sum(width > 0) == 3) {
      ret <- paste0("<w:tbl> ",
                    "<w:tblPr>",
                    '<w:tblStyle w:val="TableGrid"/>',
                    '<w:tblW w:w="5000"',
                    ' w:type="pct"/>',
                    "</w:tblPr>",
                    "<w:tblGrid>",
                    '<w:gridCol w:w="1667" w:type="pct"/>',
                    '<w:gridCol w:w="1666" w:type="pct"/>',
                    '<w:gridCol w:w="1667" w:type="pct"/>',
                    "</w:tblGrid>", collapse = "")
    } else if (sum(width > 0) == 2) {
      ret <- paste0("<w:tbl> ",
                    "<w:tblPr>",
                    '<w:tblStyle w:val="TableGrid"/>',
                    '<w:tblW w:w="5000"',
                    ' w:type="pct"/>',
                    "</w:tblPr>",
                    "<w:tblGrid>",
                    '<w:gridCol w:w="2500" w:type="pct"/>',
                    '<w:gridCol w:w="2500" w:type="pct"/>',
                    "</w:tblGrid>", collapse = "")
    } else {
      ret <- paste0("<w:tbl> ",
                    "<w:tblPr>",
                    '<w:tblStyle w:val="TableGrid"/>',
                    '<w:tblW w:w="5000"',
                    ' w:type="pct"/>',
                    "</w:tblPr>",
                    "<w:tblGrid>",
                    '<w:gridCol w:w="5000" w:type="pct"/>',
                    "</w:tblGrid>", collapse = "")
    }

    pdf(NULL)
    par(family = get_font_family(rs$font), ps = rs$font_size)

    # index for images in footer
    rID <- 0
    iconv <- 914400  # 1 inch = 914,400 EMUs (English Metric Units)
    
    lcnt <- 0
    ccnt <- 0
    rcnt <- 0
    
    lheight <- 0
    cheight <- 0
    rheight <- 0
    max_height <- 0
    
    for (i in seq(1, maxh)) {

      cret <- ""

      if (left_width > 0) {
        if (hl_num >= i) {
          
          if (image_left == FALSE) {
            # Split strings if they exceed width
            tmp <- split_string_docx(hl[[i]], left_width, rs$units, rs$font)
            
            cret <- paste0(cret, 
                           '<w:tc><w:tcPr><w:tcW w:w="', left_pct, '" w:type="pct"/></w:tcPr>', 
                           get_page_numbers_docx(para(tmp$docx)),
                           "</w:tc>\n")
            
            lcnt <- tmp$lines
          } else {
            # Calculate image height & width
            if (rs$units == "inches") {
              hgth <- (hl$height) * iconv 
              wdth <- (min(hl$width, left_width)) * iconv 
              
            } else {
              hgth <- (cin(hl$height)) * iconv 
              wdth <- (cin(min(hl$width, left_width))) * iconv
              
            }
            
            # Get image DOCX codes
            rID <- rID + 1
            image_path <- c(image_path, hl$image_path[i])
            
            img <- get_image_docx(rID, "left", hgth, wdth, type = "h")
            cret <- paste0(cret, "<w:tc>", img, "</w:tc>")
            
            lheight <- hl$height
          }
        } else {
          cret <- paste0(cret, 
                         '<w:tc><w:tcPr><w:tcW w:w="', left_pct, '" w:type="pct"/></w:tcPr>', 
                         para(" "), "</w:tc>\n")
          lcnt <- 1
        }
      }
      
      if (center_width > 0) {
        if (hc_num >= i) {
          
          if (image_center == FALSE) {
            # Split strings if they exceed width
            tmp2 <- split_string_docx(hc[[i]], center_width, rs$units, font = rs$font)
            
            cret <- paste0(cret, 
                           '<w:tc><w:tcPr><w:tcW w:w="', center_pct, '" w:type="pct"/></w:tcPr>', 
                           get_page_numbers_docx(para(tmp2$docx)),
                           "</w:tc>\n")
            
            ccnt <- tmp2$lines
          } else {
            # Calculate image height & width
            if (rs$units == "inches") {
              hgth <- (hc$height) * iconv 
              wdth <- (min(hc$width, center_width)) * iconv 
              
            } else {
              hgth <- (cin(hc$height)) * iconv 
              wdth <- (cin(min(hc$width, center_width))) * iconv
              
            }
            
            # Get image DOCX codes
            rID <- rID + 1
            image_path <- c(image_path, hc$image_path[i])
            
            img <- get_image_docx(rID, "center", hgth, wdth, type = "h")
            cret <- paste0(cret, "<w:tc>", img, "</w:tc>")
            
            cheight <- hc$height
          }
        } else {
          cret <- paste0(cret, 
                         '<w:tc><w:tcPr><w:tcW w:w="', center_pct, '" w:type="pct"/></w:tcPr>', 
                         para(" ", "center"), "</w:tc>")
          ccnt <- 1
        }
      }

      if (right_width > 0) {
        if (hr_num >= i) {
          
          if (image_right == FALSE) {
            # Split strings if they exceed width
            tmp3 <- split_string_docx(hr[[i]], right_width, rs$units, font = rs$font)
            
            
            cret <- paste0(cret, 
                           '<w:tc><w:tcPr><w:tcW w:w="', right_pct, '" w:type="pct"/></w:tcPr>', 
                           get_page_numbers_docx(para(tmp3$docx, "right")), 
                           "</w:tc>")
            
            rcnt <- tmp3$lines 
          } else {
            # Calculate image height & width
            if (rs$units == "inches") {
              hgth <- (hr$height) * iconv 
              wdth <- (min(hr$width, right_width)) * iconv 
              
            } else {
              hgth <- (cin(hr$height)) * iconv 
              wdth <- (cin(min(hr$width, right_width))) * iconv
              
            }
            
            # Get image DOCX codes
            rID <- rID + 1
            image_path <- c(image_path, hr$image_path[i])
            
            img <- get_image_docx(rID, "right", hgth, wdth, type = "h")
            cret <- paste0(cret, "<w:tc>", img, "</w:tc>")
            
            rheight <- hr$height
          }
        } else {
          cret <- paste0(cret, 
                         '<w:tc><w:tcPr><w:tcW w:w="', right_pct, '" w:type="pct"/></w:tcPr>', 
                         para(" ", "right"), "</w:tc>")
          rcnt <- 1
        } 
      }
      
      cret <- paste0(cret, "</w:tr>\n")

      # if (lcnt > rcnt) {
      #   trht <- get_row_height(round(rs$row_height * lcnt * conv))
      #   cnt <- cnt + lcnt
      # 
      # } else {
      #   trht <- get_row_height(round(rs$row_height * rcnt * conv))
      #   cnt <- cnt + rcnt
      # }
      
      cnt <- cnt + max(lcnt, ccnt, rcnt)
      max_height <- max_height + max(lheight, cheight, rheight)
      
      trht <- get_row_height(round(max(rs$row_height * max(lcnt, ccnt, rcnt), max(lheight, cheight, rheight)) * conv)) 
      
      
      ret <- paste0(ret, '<w:tr>', trht, cret)
      
    }
    
    ret <- paste0(ret, "</w:tbl>")

    dev.off()
    
    if (max_height > 0) {
      cnt <- max(cnt, ceiling(max_height/rs$line_height))
    }
    
    if (rs$page_header_blank_row == "below") {
      ret <- paste0(ret, rs$blank_row)
      cnt <- cnt + 1
    }
    
  }
  
  # Prepare lines for converting to content font size
  cnt_content <- cnt
  if (!is.null(rs$header_titles)) {
    
    tret <- get_titles_docx(rs$header_titles, rs$content_size[["width"]], rs)
    ret <- paste0(ret, rs$table_break, tret$docx)
    cnt <- cnt + tret$lines
    cnt_content <- cnt_content + tret$lines_content
    
  }

  res <- list(docx = ret, lines = cnt, image_path = image_path,
              lines_content = cnt_content)
  
  return(res)
}

#' @import grDevices
#' @noRd
get_page_footer_docx <- function(rs) {
  
  ret <- ""
  cnt <- 0
  image_path <- c()
  
  # fl <- rs$page_footer_left
  # fc <- rs$page_footer_center
  # fr <- rs$page_footer_right
  if ((!is.null(rs$footer_image_left) & is.null(rs$page_footer_left)) |
      (!is.null(rs$footer_image_right) & is.null(rs$page_footer_right)) |
      (!is.null(rs$footer_image_center) & is.null(rs$page_footer_center))) {
    
    stop("`page_footer` must be used when using `footer_image`.")
  }
  
  image_left <- FALSE
  image_center <- FALSE
  image_right <- FALSE
  
  if (!is.null(rs$footer_image_left)) {
    fl <- rs$footer_image_left
    image_left <- TRUE
  } else {
    fl <- rs$page_footer_left
  }
  
  if (!is.null(rs$footer_image_center)) {
    fc <- rs$footer_image_center
    image_center <- TRUE
  } else {
    fc <- rs$page_footer_center
  }
  
  if (!is.null(rs$footer_image_right)) {
    fr <- rs$footer_image_right
    image_right <- TRUE
  } else {
    fr <- rs$page_footer_right
  }
  
  conv <- rs$twip_conversion
  rht <- get_row_height(round(rs$row_height * conv))
  

  # maxf <- max(length(fl), length(fc), length(fr))
  fl_num <- ifelse(image_left, length(fl$image_path), length(fl))
  fc_num <- ifelse(image_center, length(fc$image_path), length(fc))
  fr_num <- ifelse(image_right, length(fr$image_path) , length(fr))
  
  maxf <- max(fl_num, fc_num, fr_num)
  

  if (maxf > 0) {

    pdf(NULL)
    par(family = get_font_family(rs$font), ps = rs$font_size)
    
    width <- rs$page_footer_width
    
    # Make sure the length is 3
    width <- c(width, rep(NA, 3 - length(width)))
    
    total_width <- sum(width, na.rm = T)
    if (total_width > rs$content_size[["width"]]) {
      
      stop(sprintf("Total width of page footer %s %s cannot be greater than content width %s %s.",
                   total_width,
                   rs$units,
                   rs$content_size[["width"]],
                   rs$units))
      
    } else {
      na_num <- sum(is.na(width))
      imputed_width <- (rs$content_size[["width"]] - total_width) / na_num
      
      left_width <- ifelse(is.na(width[1]), imputed_width, width[1])
      center_width <- ifelse(is.na(width[2]), imputed_width, width[2])
      right_width <- ifelse(is.na(width[3]), imputed_width, width[3])
      
      left_pct <- floor(5000 * left_width/rs$content_size[["width"]])
      center_pct <- floor(5000 * center_width/rs$content_size[["width"]])
      right_pct <- floor(5000 * right_width/rs$content_size[["width"]])
      
      # Make sure the total is 5000 exactly
      pct_lst <- c(left_pct, center_pct, right_pct)
      rest_pct <- 5000 - sum(pct_lst)
      add_pos <- c(1, 3 ,2)
      
      if (rest_pct >= 1){
        for (i in 1:rest_pct) {
          pct_lst[add_pos[i%%3]] <- pct_lst[i%%3] + 1
        }
      }
    }
    
    width <- c(left_width, center_width, right_width)

    if (sum(width > 0) == 3) {
      ret <- paste0("<w:tbl> ",
                    "<w:tblPr>",
                    '<w:tblStyle w:val="TableGrid"/>',
                    '<w:tblW w:w="5000"',
                    ' w:type="pct"/>',
                    "</w:tblPr>",
                    "<w:tblGrid>",
                    '<w:gridCol w:w="1667" w:type="pct"/>',
                    '<w:gridCol w:w="1666" w:type="pct"/>',
                    '<w:gridCol w:w="1667" w:type="pct"/>',
                    "</w:tblGrid>", collapse = "")
    } else if (sum(width > 0) == 2) {
      ret <- paste0("<w:tbl> ",
                    "<w:tblPr>",
                    '<w:tblStyle w:val="TableGrid"/>',
                    '<w:tblW w:w="5000"',
                    ' w:type="pct"/>',
                    "</w:tblPr>",
                    "<w:tblGrid>",
                    '<w:gridCol w:w="2500" w:type="pct"/>',
                    '<w:gridCol w:w="2500" w:type="pct"/>',
                    "</w:tblGrid>", collapse = "")
    } else {
      ret <- paste0("<w:tbl> ",
                    "<w:tblPr>",
                    '<w:tblStyle w:val="TableGrid"/>',
                    '<w:tblW w:w="5000"',
                    ' w:type="pct"/>',
                    "</w:tblPr>",
                    "<w:tblGrid>",
                    '<w:gridCol w:w="5000" w:type="pct"/>',
                    "</w:tblGrid>", collapse = "")
    }

    # index for images in footer
    rID <- 0
    iconv <- 914400  # 1 inch = 914,400 EMUs (English Metric Units)
    
    lcnt <- 0
    ccnt <- 0
    rcnt <- 0
    
    lheight <- 0
    cheight <- 0
    rheight <- 0
    max_height <- 0
    
    for (i in seq(1, maxf)) {

      tret <- ""

      if (left_width > 0) {
        if (fl_num >= i) {
          
          if (image_left == FALSE) {
            # Split strings if they exceed width
            tmp1 <- split_string_docx(fl[[i]], left_width, rs$units, font = rs$font)
            
            
            # tret <- paste0(tret, cell_pct(tmp1$docx, "left", 1667))
            tret <- paste0(tret, cell_pct(tmp1$docx, "left", left_pct))
            
            lcnt <- tmp1$lines
          } else {
            # Calculate image height & width
            if (rs$units == "inches") {
              hgth <- (fl$height) * iconv 
              wdth <- (min(fl$width, left_width)) * iconv 
              
            } else {
              hgth <- (cin(fl$height)) * iconv 
              wdth <- (cin(min(fl$width, left_width))) * iconv
              
            }
            
            # Get image DOCX codes
            rID <- rID + 1
            image_path <- c(image_path, fl$image_path[i])
            
            img <- get_image_docx(rID, "left", hgth, wdth, type = "f")
            tret <- paste0(tret, "<w:tc>", img, "</w:tc>")
            
            lheight <- fl$height
          }
        } else {
          tret <- paste0(tret, cell_pct(" ", "left", left_pct))
          lcnt <- 1
        }
      }
      
      if (center_width > 0) {
        if (fc_num >= i) {
          
          if (image_center == FALSE) {
            # Split strings if they exceed width
            tmp2 <- split_string_docx(fc[[i]], center_width, rs$units, font = rs$font)
            
            # tret <- paste0(tret,  cell_pct(tmp2$docx, "center", 1666))
            tret <- paste0(tret,  cell_pct(tmp2$docx, "center", center_pct))
            ccnt <- tmp2$lines
          } else {
            # Calculate image height & width
            if (rs$units == "inches") {
              hgth <- (fc$height) * iconv 
              wdth <- (min(fc$width, center_width)) * iconv 
              
            } else {
              hgth <- (cin(fc$height)) * iconv 
              wdth <- (cin(min(fc$width, center_width))) * iconv
              
            }
            
            # Get image DOCX codes
            rID <- rID + 1
            image_path <- c(image_path, fc$image_path[i])
            
            img <- get_image_docx(rID, "center", hgth, wdth, type = "f")
            tret <- paste0(tret, "<w:tc>", img, "</w:tc>")
            
            cheight <- fc$height
          }
        } else {
          tret <- paste0(tret,  cell_pct(" ", "center", center_pct))
          ccnt <- 1
        }
      }
      
      if (right_width > 0) {
        if (fr_num >= i) {
          
          if (image_right == FALSE) {
            tmp3 <- split_string_docx(fr[[i]], right_width, rs$units, font = rs$font)
            
            # tret <- paste0(tret, cell_pct(tmp3$docx, "right", 1667))
            tret <- paste0(tret, cell_pct(tmp3$docx, "right", right_pct))
            
            rcnt <- tmp3$lines
          } else {
            # Calculate image height & width
            if (rs$units == "inches") {
              hgth <- (fr$height) * iconv 
              wdth <- (min(fr$width, right_width)) * iconv 
              
            } else {
              hgth <- (cin(fr$height)) * iconv 
              wdth <- (cin(min(fr$width, right_width))) * iconv
              
            }
            
            # Get image DOCX codes
            rID <- rID + 1
            image_path <- c(image_path, fr$image_path[i])
            
            img <- get_image_docx(rID, "right", hgth, wdth, type = "f")
            tret <- paste0(tret, "<w:tc>", img, "</w:tc>")
            
            rheight <- fr$height
          }
        } else {
          tret <- paste0(tret,  cell_pct(" ", "right", right_pct))
          rcnt <- 1
        }
      }

      cnt <- cnt + max(lcnt, ccnt, rcnt)
      max_height <- max_height + max(lheight, cheight, rheight)
      
      text_height_emu <- round(rs$row_height * max(lcnt, ccnt, rcnt) * conv)
      image_height_emu <- round(max(lheight, cheight, rheight) * conv)
      trht <- get_row_height(max(text_height_emu, image_height_emu)) 
      
      # <w:tr> should end here for each row
      ret <- paste0(ret, '<w:tr>', trht, tret, "</w:tr>\n")

    }
    dev.off()
    
    if (max_height > 0) {
      cnt <- max(cnt, ceiling(max_height/rs$line_height))
    }

    # ret <- paste0(get_page_numbers_docx(ret), "</w:tr>\n")
    ret <- paste0(get_page_numbers_docx(ret))
    ret <- paste0(ret, "</w:tbl>\n")
  }
  
  # Prepare lines for converting to content font size
  cnt_content <- cnt
  if (!is.null(rs$footer_footnotes)) {
    
    tret <- get_footnotes_docx(rs$footer_footnotes, rs$content_size[["width"]], rs)
    ret <- paste0(tret$docx, rs$table_break, ret)
    cnt <- cnt + tret$lines
    cnt_content <- cnt_content + tret$lines_content
    
  }


  
  res <- list(docx = paste0(ret, collapse = ""),
              lines = cnt,
              image_path = image_path,
              lines_content = cnt_content)
  
  return(res)
}

#' @import grDevices
#' @noRd
get_titles_docx <- function(ttllst, content_width, rs, talgn = "center", 
                            colspan = 0, col_widths = NULL, 
                            content_brdrs = NULL) {
  
  ret <- c()
  cnt <- 0
  border_flag <- FALSE
  conv <- rs$twip_conversion
  rht <- get_row_height(round(rs$row_height * conv))
  cflag <- FALSE
  
  # The heights for converting to content font size
  total_height <- 0
  title_size_flag <- FALSE
  cnt_content <- 0
  
  if (length(ttllst) > 0) {
    
    for (ttls in ttllst) {
      
      cols <- ttls$columns
      
      if (ttls$width == "page")
        width <- rs$content_size[["width"]]
      else if (ttls$width == "content") {
        width <- content_width
        cflag <- TRUE
      } else if (is.numeric(ttls$width))
        width <- ttls$width

      w <- round(width * conv)
      
      # Calculate cell widths
      cw <- round(w / cols)
      cwidth <- width / cols
      
      cwdths <- rep(cwidth, cols)
      names(cwdths) <- paste0("col", seq_len(cols))
      
      # Get grid for title block
      grd <- get_col_grid(cwdths, conv)
      
      
      if (ttls$align %in% c("centre", "center"))
        algn <- "center"
      else if (ttls$align == "right")
        algn <- "right"
      else
        algn <- "left"
      
      brd <- ttls$borders
      # If content borders are on, align indent with content
      if (!is.null(content_brdrs)) {
        
        if (any(content_brdrs %in% c("all", "outside", "body", "left")))
          brd <- "outside"
        
      }

      # Get indent codes for alignment
      ta <- get_indent_docx(talgn, rs$line_size, w, 
                           rs$base_indent, brd, conv)

      
      alcnt <- 0
      blcnt <- 0
      
      # Lines per each title object for converting to content size later
      cnt_ttl <- 0
      
      # Open device context
      pdf(NULL)
      
      # Set point size (ps) for strwidth to calculate string width
      if (!is.null(ttls$font_size)) {
        ttlfs <- ttls$font_size
      } else {
        ttlfs <- rs$font_size
      }
      par(family = get_font_family(rs$font), ps = ttlfs)
      
      
      tb <- get_table_borders_docx(ttls$borders)
      
      ret[length(ret) + 1] <- paste0("<w:tbl>",
                                     "<w:tblPr>",
                                     '<w:tblStyle w:val="TableGrid"/>',
                                     rs$cell_margin,
                                     "<w:tblW w:w=\"", w, "\"/>", ta, tb, 
                                     "</w:tblPr>\n", grd)
      
      # Get spanning for blank rows
      spn <- paste0('<w:gridSpan w:val="', cols , '"/>')
      
      al <- ""
      if (any(ttls$blank_row %in% c("above", "both"))) {
        
        alcnt <- 1
        
        al <- paste0("<w:tr>", rht, 
                     "<w:tc><w:tcPr>", spn, '</w:tcPr><w:p><w:r><w:t>', 
                     "</w:t></w:r></w:p></w:tc></w:tr>\n")
        
        ret <- append(ret, al)
        
        cnt <- cnt + 1
        cnt_ttl <- cnt_ttl + 1
      }
      
      
      bb <- ""
      if (any(ttls$borders %in% c("bottom", "outside"))) {
        
        bb <- get_cell_borders_docx(2, 1, 2, 1, ttls$borders)
        
      }
      
      bl <- ""
      if (any(ttls$blank_row %in% c("below", "both"))) {
        blcnt <- 1
        

        
        bl <- paste0("<w:tr>", rht, 
                     "<w:tc>'<w:tcPr>'", bb, spn, '</w:tcPr>',
                     "<w:p><w:r><w:t></w:t></w:r></w:p></w:tc></w:tr>\n")
        
        bb <- ""  # needed?
        
        cnt <- cnt + 1
        cnt_ttl <- cnt_ttl + 1
      }
      
      
      
      i <- 1
      while (i <= length(ttls$titles)) {
        
        mxlns <- 0
        
        # Calculate current row
        rwnum <- ceiling(i / cols)
        
        rw <- ""
        
        for (j in seq_len(cols)) {
          
          # Not all cells have titles
          if (i > length(ttls$titles))
            vl <- ""
          else 
            vl <- ttls$titles[[i]]
          
          # Deal with column alignments
          if (cols == 1) {
            calgn <- algn 
          } else if (cols == 2) {
            if (j == 1)
              calgn <- "left"
            else 
              calgn <- "right"
          } else if (cols == 3) {
            if (j == 1)
              calgn <- "left"
            else if (j == 2)
              calgn <- "center"
            else if (j == 3) 
              calgn <- "right"
          }

          # Split title strings if they exceed width
          tmp <- split_string_docx(vl, cwidth, rs$units, font = rs$font)
          
          # Track max lines for counting
          if (tmp$lines > mxlns)
            mxlns <- tmp$lines
          
          # Get paragraph 
          tstr <- para(tmp$docx, calgn, ttlfs, ttls$bold)

          cwdth <- paste0('<w:tcW w:w="', cw,'"/>')
          
          # Append cell
          rw <- paste0(rw, "<w:tc><w:tcPr>", cwdth, "</w:tcPr>", tstr, "</w:tc>\n")
          
          
          i <- i + 1
        }
        
        # Row height has to be determined based on max number of lines in cell
        if (is.null(ttls$font_size)) {
          
          srht <- get_row_height(round(rs$row_height * mxlns * conv))
          
        } else {
          
          srht <- get_row_height(round(get_rh(rs$font, ttls$font_size) * mxlns * conv))
          title_size_flag <- TRUE
          
        }
        
        # Construct row
        ret <- append(ret, paste0("<w:tr>", srht, rw, "</w:tr>\n"))
        
        # Track lines
        cnt <- cnt + mxlns
        cnt_ttl <- cnt_ttl + mxlns
      }
      
      # Sum of total height
      if (!is.null(ttls$font_size)) {
        total_height <- total_height + cnt_ttl * get_rh(rs$font, ttls$font_size)
      } else {
        total_height <- total_height + cnt_ttl * rs$row_height
      }
      
      
      if (bl != "")
        ret <- append(ret, bl)
      
      if (cflag) {
        ret[length(ret) + 1] <- paste0("</w:tbl>")
      } else {
        ret[length(ret) + 1] <- paste0("</w:tbl>", rs$table_break)
      }
      
      # A flag to indicate that this block has bottom borders.  
      # Used to eliminate border duplication on subsequent blocks.
      if ("bottom" %in% get_outer_borders(ttls$borders))
        border_flag <- TRUE
      
      dev.off()
      

    }
    
    if (title_size_flag) {
      cnt_content <- ceiling(total_height / rs$row_height)
    } else {
      cnt_content <- cnt
    }
  }
  

  
  res <- list(docx = paste0(ret, collapse = ""), 
              lines = cnt,
              border_flag = border_flag,
              lines_content = cnt_content)
  
  return(res)
}


#' @import grDevices
#' @noRd
get_titles_docx_back <- function(ttllst, content_width, rs, talgn = "center", 
                            colspan = 0, col_widths = NULL, 
                            content_brdrs = NULL) {
  
  ret <- c()
  cnt <- 0
  border_flag <- FALSE
  conv <- rs$twip_conversion
  rht <- get_row_height(round(rs$row_height * conv))
  cflag <- FALSE
  
  
  if (length(ttllst) > 0) {
    
    for (ttls in ttllst) {
      
      
      if (ttls$width == "page")
        width <- rs$content_size[["width"]]
      else if (ttls$width == "content") {
        width <- content_width
        cflag <- TRUE
      } else if (is.numeric(ttls$width))
        width <- ttls$width
      
      w <- round(width * conv)
      
      # Get grid for title block
      grd <- get_col_grid(c(title=width), conv)
      
      
      if (ttls$align %in% c("centre", "center"))
        algn <- "center"
      else if (ttls$align == "right")
        algn <- "right"
      else
        algn <- "left"
      
      brd <- ttls$borders
      # If content borders are on, align indent with content
      if (!is.null(content_brdrs)) {
        
        if (any(content_brdrs %in% c("all", "outside", "body", "left")))
          brd <- "outside"
        
      }
      
      # Get indent codes for alignment
      ta <- get_indent_docx(talgn, rs$line_size, w, 
                            rs$base_indent, brd, conv)
      
      
      alcnt <- 0
      blcnt <- 0
      
      # Open device context
      pdf(NULL)
      par(family = get_font_family(rs$font), ps = rs$font_size)
      
      
      tb <- get_table_borders_docx(ttls$borders)
      
      ret[length(ret) + 1] <- paste0("<w:tbl>",
                                     "<w:tblPr>",
                                     rs$cell_margin,
                                     "<w:tblW w:w=\"", w, "\"/>", ta, tb, 
                                     "</w:tblPr>\n", grd)
      
      for (i in seq_along(ttls$titles)) {
        
        
        
        al <- ""
        if (i == 1) {
          if (any(ttls$blank_row %in% c("above", "both"))) {
            
            alcnt <- 1
            
            
            al <- paste0("<w:tr>", rht, 
                         "<w:tc>", '<w:p><w:r><w:t>', 
                         "</w:t></w:r></w:p></w:tc></w:tr>\n")
            
            
            cnt <- cnt + 1
          }
        }
        
        bb <- ""
        
        
        bl <- ""
        if (i == length(ttls$titles)) {
          
          if (any(ttls$borders %in% c("bottom", "outside"))) {
            bb <- get_cell_borders_docx(2, 1, 2, 1, ttls$borders)
            
            
            bb <- paste0('<w:tcPr>', bb, '</w:tcPr>')
            
          }
          
          if (any(ttls$blank_row %in% c("below", "both"))) {
            blcnt <- 1
            
            bl <- paste0("<w:tr>", rht, 
                         "<w:tc>", bb,
                         "<w:p><w:r><w:t></w:t></w:r></w:p></w:tc></w:tr>\n")
            
            bb <- ""
            
            cnt <- cnt + 1
          }
          
        }
        
        
        # Split title strings if they exceed width
        tmp <- split_string_docx(ttls$titles[[i]], width, rs$units, font = rs$font)
        
        
        tstr <- para(tmp$docx, algn, ttls$font_size, ttls$bold)
        
        
        
        # Concatenate title string
        if (al != "")
          ret <- append(ret, al)
        
        if (is.null(ttls$font_size)) {
          
          srht <- get_row_height(round(rs$row_height * tmp$lines * conv))
          
          ret <- append(ret, paste0("<w:tr>", srht, "<w:tc>", bb, tstr, 
                                    "</w:tc></w:tr>\n"))
        } else {
          
          srht <- get_row_height(round(get_rh(rs$font, ttls$font_size) * tmp$lines * conv))
          
          ret <- append(ret, paste0("<w:tr>", srht, "<w:tc>", bb, tstr, 
                                    "</w:tc></w:tr>\n"))
        }
        
        if (bl != "")
          ret <- append(ret, bl)
        
        cnt <- cnt + tmp$lines
        
        # A flag to indicate that this block has bottom borders.  
        # Used to eliminate border duplication on subsequent blocks.
        if ("bottom" %in% get_outer_borders(ttls$borders))
          border_flag <- TRUE
      }
      
      
      if (cflag) {
        ret[length(ret) + 1] <- paste0("</w:tbl>")
      } else {
        ret[length(ret) + 1] <- paste0("</w:tbl>", rs$table_break)
      }
      
      dev.off()
      
      
    }
    
  }
  
  
  
  res <- list(docx = paste0(ret, collapse = ""), 
              lines = cnt,
              border_flag = border_flag)
  
  return(res)
}

#' @import grDevices
#' @noRd
get_footnotes_docx <- function(ftnlst, content_width, rs, talgn = "center", 
                               ex_brdr = FALSE, colspan = 0, col_widths = NULL,
                               content_brdrs = NULL) {
  
  ret <- c()
  cnt <- 0
  exclude_top <- NULL
  if (ex_brdr)
    exclude_top <- "top"
  
  conv <- rs$twip_conversion
  rht <- get_row_height(round(rs$row_height * conv))

  u <- rs$units
  
  # The heights for converting to content font size
  total_height <- 0
  footnote_size_flag <- FALSE
  cnt_content <- 0
  
  if (rs$units == "inches")
    u <- "in"
  
  
  if (length(ftnlst) > 0) {

    for (ftnts in ftnlst) {
      
      cols <- ftnts$columns

      if (ftnts$width == "page")
        width <- rs$content_size[["width"]]
      else if (ftnts$width == "content")
        width <- content_width
      else if (is.numeric(ftnts$width))
        width <- ftnts$width

      w <- round(width * conv) + 2
      
      # Calculate cell widths
      cw <- round(w / cols)
      cwidth <- width / cols
      
      cwdths <- rep(cwidth, cols)
      names(cwdths) <- paste0("col", seq_len(cols))
  
      # Get grid for footnote block
      # grd <- get_col_grid(c(all=w), 1)
      grd <- get_col_grid(cwdths, conv)
      
      if (ftnts$align %in% c("centre", "center"))
        algn <- "center"
      else if (ftnts$align == "right")
        algn <- "right"
      else
        algn <- "left"
      
      brd <- ftnts$borders
      # If content borders are on, align indent with content
      if (!is.null(content_brdrs)) {
        
        if (any(content_brdrs %in% c("all", "outside", "body", "left")))
          brd <- "outside"
        
      }
      
      # Get indent codes for alignment
      ta <- get_indent_docx(talgn, rs$line_size, w, 
                           rs$base_indent, brd, conv)

      alcnt <- 0
      blcnt <- 0

      # Lines per each footnote object for converting to content size later
      cnt_ft <- 0

      pdf(NULL)
      
      # Set point size (ps) for strwidth to calculate string width
      if (!is.null(ftnts$font_size)) {
        ftntfs <- ftnts$font_size
      } else {
        ftntfs <- rs$font_size
      }
      par(family = get_font_family(rs$font), ps = ftntfs)
      
      tb <- get_table_borders_docx(ftnts$borders)
    
      
      ret[length(ret) + 1] <- paste0("<w:tbl>",
                                     "<w:tblPr>",
                                     rs$cell_margin,
                                     "<w:tblW w:w=\"", w, "\"/>", ta, tb,
                                     "</w:tblPr>", grd)
    
      
      # Get spanning for blank rows
      spn <- paste0('<w:gridSpan w:val="', cols , '"/>')
      
      al <- ""
      if (any(ftnts$blank_row %in% c("above", "both"))) {
        
        alcnt <- 1
        
        
        # al <- paste0("<w:tr>", rht, 
        #              "<w:tc>", bb, "<w:p><w:r><w:t>", 
        #              "</w:t></w:r></w:p></w:tc></w:tr>\n")
        al <- paste0("<w:tr>", rht, 
                     "<w:tc><w:tcPr>", spn, '</w:tcPr><w:p><w:r><w:t>', 
                     "</w:t></w:r></w:p></w:tc></w:tr>\n")

        ret <- append(ret, al)
        
        cnt <- cnt + 1
        cnt_ft <- cnt_ft + 1
        bb <- ""
        
      }
      
      bb <- ""
      if (any(ftnts$borders %in% c("bottom", "outside"))) {
        bb <- get_cell_borders_docx(2, 1, 2, 1, ftnts$borders)
        
        
        bb <- paste0('<w:tcPr>', bb, '</w:tcPr>')
        
      }
      
      bl <- ""
      if (any(ftnts$blank_row %in% c("below", "both"))) {
        blcnt <- 1
        
        
        # bl <- paste0("<w:tr>", rht, 
        #              "<w:tc>", "<w:p><w:r><w:t>", 
        #              "</w:t></w:r></w:p></w:tc></w:tr>\n")
        bl <- paste0("<w:tr>", rht, 
                     "<w:tc>'<w:tcPr>'", bb, spn, '</w:tcPr>',
                     "<w:p><w:r><w:t></w:t></w:r></w:p></w:tc></w:tr>\n")
        
        bb <- "" 
        
        cnt <- cnt + 1
        cnt_ft <- cnt_ft + 1
      }

      i <- 1
      while (i <= length(ftnts$footnotes)) {
        
        mxlns <- 0
        
        # Calculate current row
        rwnum <- ceiling(i / cols)
        
        rw <- ""
        

        for (j in seq_len(cols)) {
          
          # Not all cells have titles
          if (i > length(ftnts$footnotes))
            vl <- ""
          else 
            vl <- ftnts$footnotes[[i]]
          
          # Deal with column alignments
          if (cols == 1) {
            calgn <- algn 
          } else if (cols == 2) {
            if (j == 1)
              calgn <- "left"
            else 
              calgn <- "right"
          } else if (cols == 3) {
            if (j == 1)
              calgn <- "left"
            else if (j == 2)
              calgn <- "center"
            else if (j == 3) 
              calgn <- "right"
          }
  
          # Split footnote strings if they exceed width
          tmp <- split_string_docx(vl, cwidth, rs$units, font = rs$font)
          
          # Track max lines for counting
          if (tmp$lines > mxlns)
            mxlns <- tmp$lines
          
          # Get paragraph 
          tstr <- para(tmp$docx, calgn, ftntfs, italics = ftnts$italics)
          
          cwdth <- paste0('<w:tcW w:w="', cw,'"/>')
          
          # Deal with bottom borders for when there is no blank row
          cb <- ""
          if (!any(ftnts$blank_row %in% c("below", "both"))) {
            if (any(ftnts$borders %in% c("bottom", "outside"))) {
              cb <- get_cell_borders_docx(2, 1, 2, 1, ftnts$borders)
              
              
              cb <- paste0('<w:tcPr>', cb, '</w:tcPr>')
              
            }
          }
          
          # Append cell
          rw <- paste0(rw, "<w:tc><w:tcPr>", cwdth, cb, "</w:tcPr>", tstr, "</w:tc>\n")
          
          
          i <- i + 1
          

        }
        
        # Row height has to be determined based on max number of lines in cell
        if (is.null(ftnts$font_size)) {
          
          srht <- get_row_height(round(rs$row_height * mxlns * conv))
          
        } else {
          
          srht <- get_row_height(round(get_rh(rs$font, ftnts$font_size) * mxlns * conv))
          footnote_size_flag <- TRUE
          
        }
        
        # Need to change height depending on max lines
        # srht <- get_row_height(round(rs$row_height * mxlns * conv))
        
        # Construct row
        ret <- append(ret, paste0("<w:tr>", srht, rw, "</w:tr>\n"))
        
        # Track lines
        cnt <- cnt + mxlns
        cnt_ft <- cnt_ft + mxlns
        
      }
      
      # Sum of total height
      if (!is.null(ftnts$font_size)) {
        total_height <- total_height + cnt_ft * get_rh(rs$font, ftnts$font_size)
      } else {
        total_height <- total_height + cnt_ft * rs$row_height
      }
      
      # ret <- append(ret, paste0("<w:tr>", trht, 
      #                           "<w:tc>", bb, para(tmp$docx, algn, 
      #                                              italics = ftnts$italics), 
      #                           "</w:tc></w:tr>\n"))
      
      
      if (bl != "")
        ret <- append(ret, bl)
      
      # Close table
      ret[length(ret) + 1] <- "</w:tbl>\n"
      
      dev.off()

    }
    
    if (footnote_size_flag) {
      cnt_content <- ceiling(total_height / rs$row_height)
    } else {
      cnt_content <- cnt
    }

    # For Issue 341, comment out the page process
    # ret <- get_page_numbers_docx(ret)

  }

  
  res <- list(docx = paste0(paste0(ret,  collapse = ""), rs$table_break),
              lines = cnt, lines_content = cnt_content)
  
  return(res)
}



#' @import grDevices
#' @noRd
get_footnotes_docx_back <- function(ftnlst, content_width, rs, talgn = "center", 
                               ex_brdr = FALSE, colspan = 0, col_widths = NULL,
                               content_brdrs = NULL) {
  
  ret <- c()
  cnt <- 0
  exclude_top <- NULL
  if (ex_brdr)
    exclude_top <- "top"
  
  conv <- rs$twip_conversion
  rht <- get_row_height(round(rs$row_height * conv))
  
  u <- rs$units
  if (rs$units == "inches")
    u <- "in"
  
  
  if (length(ftnlst) > 0) {
    
    for (ftnts in ftnlst) {
      
      if (ftnts$width == "page")
        width <- rs$content_size[["width"]]
      else if (ftnts$width == "content")
        width <- content_width
      else if (is.numeric(ftnts$width))
        width <- ftnts$width
      
      w <- round(width * conv) + 2
      
      
      grd <- get_col_grid(c(all=w), 1)
      
      
      if (ftnts$align %in% c("centre", "center"))
        algn <- "center"
      else if (ftnts$align == "right")
        algn <- "right"
      else
        algn <- "left"
      
      brd <- ftnts$borders
      # If content borders are on, align indent with content
      if (!is.null(content_brdrs)) {
        
        if (any(content_brdrs %in% c("all", "outside", "body", "left")))
          brd <- "outside"
        
      }
      
      # Get indent codes for alignment
      ta <- get_indent_docx(talgn, rs$line_size, w, 
                            rs$base_indent, brd, conv)
      
      alcnt <- 0
      blcnt <- 0
      
      
      pdf(NULL)
      par(family = get_font_family(rs$font), ps = rs$font_size)
      
      tb <- get_table_borders_docx(ftnts$borders)
      
      
      ret[length(ret) + 1] <- paste0("<w:tbl>",
                                     "<w:tblPr>",
                                     rs$cell_margin,
                                     "<w:tblW w:w=\"", w, "\"/>", ta, tb,
                                     "</w:tblPr>", grd)
      
      for (i in seq_along(ftnts$footnotes)) {
        
        
        al <- ""
        bb <- ""
        if (i == 1) {
          
          if (any(ftnts$borders %in% c("top", "bottom", "outside"))) {
            bb <- get_cell_borders_docx(1, 1, 2, 1, ftnts$borders)
            
            
            bb <- paste0('<w:tcPr>', bb, '</w:tcPr>')
            
          }
          
          if (any(ftnts$blank_row %in% c("above", "both"))) {
            
            alcnt <- 1
            
            
            al <- paste0("<w:tr>", rht, 
                         "<w:tc>", bb, "<w:p><w:r><w:t>", 
                         "</w:t></w:r></w:p></w:tc></w:tr>\n")
            
            
            cnt <- cnt + 1
            bb <- ""
            
          }
        }
        
        bl <- ""
        if (i == length(ftnts$footnotes)) {
          if (any(ftnts$blank_row %in% c("below", "both"))) {
            blcnt <- 1
            
            
            bl <- paste0("<w:tr>", rht, 
                         "<w:tc>", "<w:p><w:r><w:t>", 
                         "</w:t></w:r></w:p></w:tc></w:tr>\n")
            
            
            cnt <- cnt + 1
          }
          
        }
        
        
        # Split footnote strings if they exceed width
        tmp <- split_string_docx(ftnts$footnotes[[i]], width, rs$units, font = rs$font)
        
        if (al != "")
          ret <- append(ret, al)
        
        
        trht <- get_row_height(round(rs$row_height * tmp$lines * conv))
        
        ret <- append(ret, paste0("<w:tr>", trht, 
                                  "<w:tc>", bb, para(tmp$docx, algn, 
                                                     italics = ftnts$italics), 
                                  "</w:tc></w:tr>\n"))
        
        
        
        if (bl != "")
          ret <- append(ret, bl)
        
        cnt <- cnt + tmp$lines
      }
      ret[length(ret) + 1] <- "</w:tbl>\n"
      dev.off()
      
      
    }
    
    ret <- get_page_numbers_docx(ret)
    
  }
  
  
  res <- list(docx = paste0(paste0(ret,  collapse = ""), rs$table_break),
              lines = cnt)
  
  return(res)
}

#' @import grDevices
#' @noRd
get_title_header_docx <- function(thdrlst, content_width, rs, talgn = "center", 
                                  colspan = 0, col_widths = NULL) {
  
  ret <- c()
  cnt <- 0
  border_flag <- FALSE
  conv <- rs$twip_conversion
  rht <- get_row_height(round(rs$row_height * conv))
  cflag <- FALSE
  
  # ta <- "align=\"left\" "
  # if (talgn == "right")
  #   ta <- "align=\"right\" "
  # else if (talgn %in% c("center", "centre"))
  #   ta <- "align=\"center\" "
  
  u <- rs$units
  if (rs$units == "inches")
    u <- "in"

  if (length(thdrlst) > 0) {

    for (ttlhdr in thdrlst) {

      if (ttlhdr$width == "page")
        width <- rs$content_size[["width"]]
      else if (ttlhdr$width == "content") {
        width <- content_width
        cflag <- TRUE
      } else if (is.numeric(ttlhdr$width))
        width <- ttlhdr$width

      w <- round(width * conv) + 2
      wl <- round(w * .7)
      wr <- round(w * .3)
      

      grd <- get_col_grid(c(left= wl, right = wr), 1)


      mx <- max(length(ttlhdr$titles), length(ttlhdr$right))

      alcnt <- 0
      blcnt <- 0

      pdf(NULL)
      par(family = get_font_family(rs$font), ps = rs$font_size)
      
      
      
      tb <- get_table_borders_docx(ttlhdr$borders)
      
      ta <- get_indent_docx(talgn, rs$line_size, w, 
                            rs$base_indent, ttlhdr$borders, conv)

      
      ret[length(ret) + 1] <- paste0("<w:tbl>", 
                                     "<w:tblPr>",
                                     '<w:tblStyle w:val="TableGrid"/>',
                                     rs$cell_margin,
                                     '<w:tblW w:w="', w, '"/>', ta, tb,
                                     "</w:tblPr>", grd, "\n")
      
                                     

      for(i in seq_len(mx)) {

        al <- ""
        if (i == 1) {
          if (any(ttlhdr$blank_row %in% c("above", "both"))) {

            alcnt <- 1

            
            al <- paste0("<w:tr>", rht, 
                  "<w:tc><w:p><w:r><w:t></w:t></w:r></w:p></w:tc>",
                  "<w:tc><w:p><w:r><w:t></w:t></w:r></w:p></w:tc></w:tr>\n")
            cnt <- cnt + 1

          }
        }
        
        bb <- ""

        bl <- ""
        if (i == mx) {
          
          if (any(ttlhdr$borders %in% c("bottom", "outside"))) {
            bb <- get_cell_borders_docx(2, 1, 2, 1, ttlhdr$borders)
            
            
            bb <- paste0('<w:tcPr>', bb, '</w:tcPr>')
            
          }
          
          if (any(ttlhdr$blank_row %in% c("below", "both"))) {
            blcnt <- 1

            
            bl <- paste0("<w:tr>", rht, 
                  "<w:tc>", bb, "<w:p><w:r><w:t></w:t></w:r></w:p></w:tc>", 
                  "<w:tc>", bb, "<w:p><w:r><w:t></w:t></w:r></w:p></w:tc></w:tr>\n")
            cnt <- cnt + 1
            
            bb <- ""
          }

        }

        if (length(ttlhdr$titles) >= i) {
          # Split strings if they exceed width
          tmp1 <- split_string_docx(ttlhdr$titles[[i]], width * .7, rs$units, font = rs$font)

          ttl <-  tmp1$docx 
          tcnt <- tmp1$lines
        } else {
          ttl <- ""
          tcnt <- 1
        }

        if (length(ttlhdr$right) >= i) {
          tmp2 <- split_string_docx(ttlhdr$right[[i]],
                                   width * .3, rs$units, font = rs$font)

          
          hdr <- tmp2$docx
          hcnt <- tmp2$lines
        } else {
          hdr <- " "
          hcnt <- 1
        }



        if (al != "")
          ret <- append(ret, al)
        
        if (bb == "") {
          ret <- append(ret, paste0("<w:tr>", rht, 
                                     cell_abs(ttl, "left", wl), 
                                     cell_abs(hdr, "right", wr), 
                                    "</w:tr>\n"))
        } else {
          
          ret <- append(ret, paste0("<w:tr>", rht, 
                                    cell_abs(ttl, "left", wl, borders = "bottom"), 
                                    cell_abs(hdr, "right", wr, borders = "bottom"), 
                                    "</w:tr>\n"))
        }
        
        
        if (bl != "")
          ret <- append(ret, bl)

        if (tcnt > hcnt)
          cnt <- cnt + tcnt
        else
          cnt <- cnt + hcnt
        
        if ("bottom" %in% get_outer_borders(ttlhdr$borders))
          border_flag <- TRUE
      }

      if (cflag) {
        ret[length(ret) + 1] <- paste0('</w:tbl>\n')
      } else {
        ret[length(ret) + 1] <- paste0('</w:tbl>\n', rs$table_break)

      }
      dev.off()
      
    }

  }
  

  
  res <- list(docx = paste0(ret, collapse = ""),
              lines = cnt,
              border_flag = border_flag)
  
  return(res)
}


#' Get page by text strings suitable for printing
#' @import stringi
#' @param titles Page by object
#' @param width The width to set the page by strings to
#' @return A vector of strings
#' @noRd
get_page_by_docx <- function(pgby, width, value, rs, talgn, 
                             ex_brdr = FALSE, pgby_cnt = NULL) {
  
  if (is.null(width)) {
    stop("width cannot be null.") 
    
  }
  
  if (is.null(value))
    value <- get_pgby_value(value, pgby_cnt)
  
  ll <- width
  conv <- rs$twip_conversion
  ret <- c()
  cnt <- 0
  rht <- get_row_height(round(rs$row_height * conv))
  border_flag <- FALSE
  exclude_top <- NULL
  if (ex_brdr)
    exclude_top <- "top"


  if (!is.null(pgby)) {

    if (!any(class(pgby) == "page_by"))
      stop("pgby parameter value is not a page_by.")

    w <- round(width * conv)
    
    grd <- get_col_grid(c(pgby=w), 1)

    if (pgby$align %in% c("centre", "center"))
      algn <- "center"
    else if (pgby$align == "right")
      algn <- "right"
    else
      algn <- "left"
  
    
    tb <- get_table_borders_docx(pgby$borders)
    
    
    ta <- get_indent_docx(talgn, rs$line_size, w, 
                          rs$base_indent, pgby$borders, conv)
    
    ret[length(ret) + 1] <- paste0("<w:tbl>",
                                   "<w:tblPr>", 
                                   rs$cell_margin,
                                   "<w:tblW w:w=\"", w, "\"/>", ta, tb,
                                   "</w:tblPr>", grd)
    
    trows <- 1
    brow <- 1
    if (pgby$blank_row %in% c("above", "both")) {
      trows <- trows + 1
      brow <- 2
    }
    if (pgby$blank_row %in% c("below", "both"))
      trows <- trows + 1

    if (pgby$blank_row %in% c("above", "both")) {
      
      cb <- ""
      if (any(pgby$borders %in% c("outside", "top"))) {
        cb <- get_cell_borders_docx(1, 1, 2, 1, "top")
        
        if (cb != "")
          cb <- paste0("<w:tcPr>", cb, "</w:tcPr>")
      }


      ret[length(ret) + 1] <- paste0("<w:tr>", rht, 
                   "<w:tc>", cb, 
                   "<w:p><w:r><w:t></w:t></w:r></w:p></w:tc></w:tr>\n")
      
      cnt <- cnt + 1
    }

    
    vb <- c()
    if (any(pgby$borders %in% c("outside", "top", "all")) & 
        !pgby$blank_row %in% c("above", "both")) {
      vb <- "top"
    
    }
    
    if (any(pgby$borders %in% c("outside", "bottom", "all")) & 
        !pgby$blank_row %in% c("below", "both")) {
      vb <- append(vb, "bottom")
      
    }
    
    
    pdf(NULL)
    par(family = get_font_family(rs$font), ps = rs$font_size)
    
    # Bold has two scenarios
    #   (1) All TRUE, FALSE -> Directly output as all bold or not bold
    #   (2) Only label or value is bold -> Specially process the line with label and value
    
    # if (pgby$bold %in% c(T, F)) {
    #   # Account for multiple pgby lines
    #   tmp <- split_string_html(paste0(pgby$label, value), width, rs$units)
    #   
    #   vl <- tmp$html
    #   cnt <- cnt + tmp$lines
    #   vrht <- get_row_height(round(rs$row_height * conv) * tmp$lines)
    #   
    #   ret <- append(ret, paste0("<w:tr>", vrht, 
    #                             cell_abs(vl, width = w, 
    #                                      borders = vb, bold = pgby$bold), 
    #                             "</w:tr>\n"))
    #   
    # } else if (pgby$bold %in% c("value", "label", T, F)) {
    if (pgby$bold %in% c("value", "label", T, F)) {
      
      # -----  This is for test ------- #
      # pgby <- list("label" = paste0(
      #   "This is a very long\nlabel with intentionally line change and also",
      #   " long text which should take at least three lines: "
      # ),
      #              "bold" = "label")
      # # value <- "serota"
      # 
      # # pgby <- list("label" = "Flower Type: ",
      # # "bold" = "value")
      # value <- paste0(
      #   "This is a very long value without intentionally line change and also",
      #   " long text which should take at least three lines: ",
      #   "serota"
      # )
      # width <- 5
      # rs <- list("units" = "inches")
      # conv <- 1440
      # vb <- c()
      # w <- round(width * conv)
      # cnt <- 0
      tmp_cnt <- 0
      # ------------------------------- #
      if (pgby$bold == "value") {
        bold <- c(FALSE, TRUE)
      } else if (pgby$bold == "label") {
        bold <- c(TRUE, FALSE)
      } else if (pgby$bold == TRUE) {
        bold <- c(TRUE, TRUE)
      } else if (pgby$bold == FALSE) {
        bold <- c(FALSE, FALSE)
      }
      
      # Split label
      label_buffer <- ifelse(bold[1] == TRUE, 0.5, 0.2)
      label_split <- split_string_docx(pgby$label, width - label_buffer, rs$units, font = rs$font)
      cnt <- cnt + label_split$lines

      # Split value
      value_buffer <- ifelse(bold[2] == TRUE, 0.5, 0.2)
      label_last_width <- label_split$widths[length(label_split$widths)]
      remain_width <- width - label_last_width - value_buffer
      value_split <- split_string_docx(value, remain_width, rs$units, font = rs$font)

      remain_value_lines <- 0
      if (value_split$widths[1] > remain_width) {
        # If the first width is greater than remain width, it means value start a new line
        value_split <- split_string_docx(value, width, rs$units, font = rs$font)
        cnt <- cnt + value_split$lines
        remain_value_lines <- value_split$lines
        value_split_txt <- value_split$docx
      } else {
        # If not, calculate from second line with full width
        splt <- strsplit(value_split$docx, split = "\n", fixed = TRUE)
        if (length(splt[[1]]) > 1) {
          remain_value <- trimws(sub(splt[[1]][1], "", value), which = "left")
          remain_value_split <- split_string_docx(remain_value, width - value_buffer, rs$units, font = rs$font)
          cnt <- cnt + remain_value_split$lines
          remain_value_lines <- remain_value_split$lines
          
          value_split_txt <- paste0(splt[[1]][1], "\n", remain_value_split$docx)
        } else {
          value_split_txt <- value_split$docx
        }
      }
      
      vrht <- get_row_height(round(rs$row_height * conv) * (label_split$lines + remain_value_lines))
      
      # Make sure there is blank between label and value
      label_nchar <- nchar(label_split$docx)
      value_nchar <- nchar(value_split_txt)
      if (substr(label_split$docx, label_nchar, label_nchar) != " "
          & substr(value_split_txt, value_nchar, value_nchar) != " "){
        value_split_txt <- paste0(" ", value_split_txt)
      }
      
      # print(paste0("w is ", w))
      # print(paste0("line is ", tmp_cnt))
      # print(paste0("label is ", label_split$docx))
      # print(paste0("value is ", value_split_txt))
      
      ret <- append(ret, paste0("<w:tr>", vrht, 
                                cell_abs(c(label_split$docx, value_split_txt), width = w, 
                                         borders = vb, bold = bold, multiple = TRUE), 
                                "</w:tr>\n"))
    }
    
    dev.off()
    
    
    # cnt <- cnt + 1

    if (pgby$blank_row %in% c("below", "both")) {
      
      cb <- ""
      if (any(pgby$borders %in% c("outside", "bottom"))) {
        cb <- get_cell_borders_docx(2, 1, 2, 1, "bottom")
        
        if (cb != "")
          cb <- paste0("<w:tcPr>", cb, "</w:tcPr>")
      }
      
      
      ret[length(ret) + 1] <- paste0("<w:tr>", rht, 
                                     "<w:tc>", cb, 
                                     "<w:p><w:r><w:t></w:t></w:r></w:p></w:tc></w:tr>\n")
      
      cnt <- cnt + 1
    }

    # if (pgby$blank_row %in% c("below", "both")) {
    # 
    #   
    #   ret[length(ret) + 1] <- paste0("<w:tr>", rht, 
    #                 "<w:tc><w:p><w:r><w:t></w:t></w:r></w:p></w:tc></w:tr>\n")
    #   cnt <- cnt + 1
    #   
    # 
    # }

    ret[length(ret) + 1] <- paste0("</w:tbl>")
    
    if ("bottom" %in% get_outer_borders(pgby$borders))
      border_flag <- TRUE

  }
  
  
  res <- list(docx = paste0(ret, collapse = ""), 
              lines = cnt,
              border_flag = border_flag)
  
  return(res)
}

# Utilities ---------------------------------------------------------------



# <w:tblBorders>
#   <w:top w:val="single" w:sz="4" w:space="0" w:color="auto"/>
#   <w:left w:val="single" w:sz="4" w:space="0" w:color="auto"/>
#   <w:bottom w:val="single" w:sz="4" w:space="0" w:color="auto"/>
#   <w:right w:val="single" w:sz="4" w:space="0" w:color="auto"/>
#   <w:insideH w:val="single" w:sz="4" w:space="0" w:color="auto"/>
#   <w:insideV w:val="single" w:sz="4" w:space="0" w:color="auto"/>
#   </w:tblBorders>
  
#' @description Return border codes for the whole table.  
#' @noRd
get_table_borders_docx <- function(brdrs, 
                                  flag = "", exclude = NULL) {
  
  t <- ""
  b <- ""
  l <- ""
  r <- ""
  ih <- ""
  iv <- ""
  
  
  if ("all" %in% brdrs) {
    t <- '<w:top w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
    b <- '<w:bottom w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
    l <- '<w:left w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
    r <- '<w:right w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
    ih <- '<w:insideH w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
    iv <- '<w:insideV w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
    
    
  } else {
    
    if ("inside" %in% brdrs) {
      
      ih <- '<w:insideH w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
      iv <- '<w:insideV w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
      
      
    }
    
    if ( any(brdrs %in% c("outside", "top")))
      t <- '<w:top w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
    
    if (any(brdrs %in% c("bottom", "outside")))
      b <- '<w:bottom w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
    
    if (any(brdrs %in% c("outside", "left")))
      l <- '<w:left w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
    
    if (any(brdrs %in% c("outside", "right")))
      r <- '<w:right w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
    
  }
  
  
  
  if (!is.null(exclude)) {
    if (any(exclude == "top"))
      t <- ""
    if (any(exclude == "bottom"))
      b <- ""
    if (any(exclude == "left"))
      l <- ""
    if (any(exclude == "right"))
      r <- ""
  }
  
  ret <- paste0('<w:tblBorders>',
                t, b, l, r, ih, iv, 
                '</w:tblBorders>')
  
  return(ret)
  
}

#' @noRd
get_page_numbers_docx <- function(val, tpg = TRUE) {
  
  ret <- val
  
  ret <- gsub("[pg]", get_page_number_docx(), ret, fixed = TRUE)
  
  if (tpg)
    ret <- gsub("[tpg]", get_total_pages_docx(), ret, fixed = TRUE)
  
  return(ret)
}

#' @noRd
get_row_height <- function(hgt) {
  
  ret <- paste0('<w:trPr><w:trHeight w:hRule="exact" w:val="', hgt, 
                  '"/></w:trPr>')
  
  return(ret)
  
}

#' @noRd
get_page_number_docx <- function() {
  
  ret <- paste0('</w:t></w:r><w:r><w:fldChar w:fldCharType="begin"/></w:r>
     <w:r>
        <w:instrText xml:space="preserve"> PAGE   \\* MERGEFORMAT </w:instrText>
     </w:r>
     <w:r>
        <w:fldChar w:fldCharType="separate"/>
     </w:r>
     <w:r>
        <w:t>1</w:t>
     </w:r>
     <w:r>
        <w:fldChar w:fldCharType="end"/>
     </w:r><w:r><w:t xml:space="preserve">')
  
  return(ret)
}

#' @noRd
get_total_pages_docx <- function() {
  
  ret <- paste0('</w:t></w:r><w:r><w:fldChar w:fldCharType="begin"/></w:r>
     <w:r>
        <w:instrText xml:space="preserve"> NUMPAGES   \\* MERGEFORMAT </w:instrText>
     </w:r>
     <w:r>
        <w:fldChar w:fldCharType="separate"/>
     </w:r>
     <w:r>
        <w:t>1</w:t>
     </w:r>
     <w:r>
        <w:fldChar w:fldCharType="end"/>
     </w:r><w:r><w:t xml:space="preserve">')
  
  return(ret)
}

#' @noRd
get_indent_docx <- function(talgn, line_size, width, base_indent, borders, conv) {
  
  ret <- ""
  bi <- 0
  
  if (any(borders %in% c("outside", "all", "left"))) 
    bi <- base_indent
    
    if (talgn == "right")
      aw <- round(line_size * conv) - width + bi
    else if (talgn %in% c("center", "centre"))
      aw <- round(((line_size * conv) - width) / 2) + bi
    else 
      aw <- bi
    
  if (aw < 0)
    aw <- 0
  
  ret <- paste0('<w:tblInd w:w="', aw, '" w:type="dxa"/>')
  
  
  return(ret)
  
}


#' @noRd
get_col_grid <- function(widths, conv) {
  
  nms <- names(widths) 
  ret <- "<w:tblGrid>\n"
  
  
  # Get cell widths
  for (k in seq_along(widths)) {
    if (!is.control(nms[k])) {
      if (!is.na(widths[k])) {
        ret <- paste0(ret,  "<w:gridCol w:w=\"", 
                      round(widths[k] * conv), "\"/>\n")
      }
    }
  } 
  
  ret <- paste0(ret, "</w:tblGrid>\n")
  
  
  return(ret)
  
}

# Utility Functions -------------------------------------------------------


#' gen_groups
#'
#' Creates group values based on a number of items and a vector of item
#' counts per group.
#'
#' This function is used to create a vector of group values.  It can use used to
#' dynamically subset a dataframe based on row counts rather than data values.
#' The last_indices parameter can be used to return the indexes of the 
#' last item in each group.
#'
#' @param tot Total number of items in return vector.
#' @param group_cnt Number of items in each group.  This can be a single value
#' or vector of values.  Will recycle if needed.
#' @param last_indices TRUE or FALSE value indicating whether to return the
#' vector of values, or a vector of row indices where the breaks occur.
#' @return A vector of length \code{tot} broken into groups specified by
#' \code{group_cnt}.  Groups will be identified by integers from 1 to n.
#' @examples
#' gen_groups(10, 3)
#' #>[1] 1 1 1 2 2 2 3 3 3 4
#'
#' gen_groups(12, c(3, 2, 5, 2))
#' #>[1] 1 1 1 2 2 3 3 3 3 3 4 4
#'
#' gen_groups(10, 3, last_indices = TRUE)
#' #>[1] 3 6 9 10
#'
#' gen_groups(12, c(3, 2, 5, 2), last_indices = TRUE)
#' #>[1] 3 5 10 12
#' @noRd
gen_groups <- function(tot, group_cnt, last_indices = FALSE) {
  
  # Create empty return vector
  ret <- c()
  
  # Initialize cursors
  cnt <- 0
  ind <- 1
  
  # Populate return vector with group values or last indices
  while(cnt <= tot){
    for(i in seq_along(group_cnt)){
      for(j in 1:group_cnt[i]){
        
        cnt <- cnt + 1
        
        if (cnt <= tot){
          if (last_indices){
            ret[ind] <- cnt
          }else{
            ret[cnt] <- ind
          }
        }
        
      }
      ind <- ind + 1
    }
  }
  
  return(ret)
}



#' Get the font family from the font name
#' @noRd
get_font_family <- function(font_name) {
  
  # Trap missing or invalid font_name parameter
  if (!tolower(font_name) %in% c("arial", "courier", "times")) {
    
    stop(paste0("ERROR: font_name parameter on get_font_family() ",
                "function is invalid: '", font_name,
      "'\n\tValid values are: 'Arial', 'Courier', and 'Times'."
      ))
  }
  
  fam <- "mono"
  # mono, serif, sans
  if (tolower(font_name) == "arial") {
    fam <- "sans"
  } else if (tolower(font_name) == "times") {
    fam <- "serif"
  }
  
  return(fam)
  
}




#' @title
#' Add a blank row to a data frame
#'
#' @description
#' The purpose of this function is to add a blank row to the top or bottom
#' of a dataframe.  Character columns will be set to an empty string.  Numeric
#' and date columns will be set to an NA. The function allows the user to pass
#' in values for specified columns.  This feature is useful for setting key
#' values.  Note that a blank value will be added to factor levels that do not
#' contain blanks.
#' @param x The dataframe to add blanks to.
#' @param ... Column names and non-blank values to assign.
#' @param location The location to add the blank row.  Valid values are "above",
#' "below", and "both".  The default value is "below".
#' @return  The input dataset with the blank row added at the specified
#' location.
#' @examples
#' #s <- filter(iris, Species == "setosa")
#' #b <- add_blank_row(s)
#' @noRd
add_blank_row <- function(x, location="below", vars = NULL){


  # Create a blank row with the same structure as the incoming dataframe.
  rw <- x[0, ]
  
  # Get group values
  if (!is.null(vars)) {
    gv <- x[1, vars]
    names(gv) <- vars
    gn <- names(x)
  }

  # For character columns, add a blank.
  # For numeric columns, NA is generated automatically.
  # For factors, cast to vector if blank is not in level list.
  for (i in seq_along(x)) {
    rv <- NA
    if ("character" %in% class(rw[[i]])) {
      
      rv <- ""
      
    } else if("factor" %in% class(rw[[i]])) {

      if (!"" %in% levels(rw[[i]])) {
        levels(x[[i]]) <- c(levels(x[[i]]), "")
        levels(rw[[i]]) <- c(levels(rw[[i]]), "")
      }

      rv <- ""
    }
    
    if (is.null(vars))
      rw[1, i] <- rv
    else if (gn[i] %in% vars) {
      if (!is.data.frame(gv))
        rw[1, i] <- gv[1]
      else
        rw[1, i] <- gv[1, gn[[i]]]
    }
    else
      rw[1, i] <- rv
  }
  
  # Change copy position depending on location
  cpos <- 1
  if (location == "below")
    cpos <- nrow(x)

  # Set page and page by values
  if ("..page_by" %in% names(x))
    rw[1, "..page_by"] <-  x[1, "..page_by"]
  if ("..page" %in% names(x)) {
    # rw[1, "..page"] <-  x[1, "..page"]
    rw[1, "..page"] <-  x[cpos, "..page"]
  }
  # Set group together
  if (any(grepl("..group_cohesion\\d+", names(x)))){
    for (g in names(x)[grepl("..group_cohesion\\d+", names(x))]) {
      rw[1, g] <-  x[cpos, g]
    }
  }
  
  # Keep break label
  if (any(grepl("..break_label\\d+", names(x)))){
    for (b in names(x)[grepl("..break_label\\d+", names(x))]) {
      rw[1, b] <-  x[cpos, b]
    }
  }
  
  # Add the blank row to the specified location.
  ret <- x
  if ("..blank" %in% names(ret))
    ret$..blank <- ifelse(is.na(ret$..blank), "", ret$..blank)
  else 
    ret$..blank <- ""
  
  if (location == "below") {
    rw$..blank <- "B"
    ret <- rbind(ret, rw)
  } else if (location == "above") {
    rw$..blank <- "A"
    ret <- rbind(rw, ret)
  } else if (location == "label") {
    rw$..blank <- "L"
    ret <- rbind(rw, ret)
  } else if (location == "both") {
    rw2 <- rw
    rw$..blank <- "A"
    rw2$..blank <- "B"
    
    if ("..page" %in% names(x)) 
      rw2[1, "..page"] <- x[nrow(x), "..page"]
    
    ret <- rbind(rw, ret, rw2)
  }

  return(ret)
}

#' @noRd
compare_rows <- function(row1, row2) {
 
  ret <- TRUE
  for(i in seq_along(row1)) {
    if (!strong_eq(row1[[i]], row2[[i]])) {
      ret <- FALSE
      break
    }
  }
  
  return(ret)
  
}

#' @noRd
strong_eq <- Vectorize(function(x1, x2) {
  
  ret <- TRUE
  if (is.null(x1) & is.null(x2))
    ret <- TRUE
  else if (is.null(x1) & !is.null(x2))
    ret <- FALSE
  else if (!is.null(x1) & is.null(x2))
    ret <- FALSE
  else if (is.na(x1) & is.na(x2))
    ret <- TRUE
  else if (is.na(x1) & !is.na(x2))
    ret <- FALSE
  else if (!is.na(x1) & is.na(x2))
    ret <- FALSE
  else {
    ret <- x1 == x2
    
  }
  
  return(ret)
  
})

#' @noRd
get_breaks <- function(x) {
  
  counter <- 1
  lastrow <- x[1, ]
  ret <- c()
  for(i in seq_len(nrow(x))) {
    row <- x[i, ]
    if (!compare_rows(row, lastrow)) {
      counter <- counter + 1
      lastrow <- row
      
    }
    ret[length(ret) + 1] <- counter
    
  }
  
  return(ret)
}


#' @title
#' Add blank rows to a data frame after each by-group
#'
#' @description
#' The purpose of this function is to add a blank rows to a dataframe for each
#' by-group.  Character columns will be set to an empty string.  Numeric
#' and date columns will be set to an NA. The function allows the user to pass
#' in column names to group by.  Note that a blank value will be added to factor
#' levels that do not contain blanks.
#' @param x The dataframe to add blanks to.
#' @param ... Column names for group variables.
#' @param .var_list A character vector or list of column names to split by.
#' @return  The input dataset with the blank row added after each by-group.
#' @examples
#' b <- add_blank_rows(iris, Species)
#' @noRd
add_blank_rows <- function(x, location = "below", vars = NULL) {

  # Seems like this is not needed any more
  # for (nm in vars)
  #   x[[nm]] <- factor(x[[nm]], levels=unique(x[[nm]]))

  # Alternate to get rid of tidyverse dependency
  if (is.null(vars))
    lst <- list(x)
  else {
    
    lst <- split(x, get_breaks(x[vars]))
    
  }

  # print("Split list")
  # print(lst)

  # Create a new list to avoid complaints
  ret <- list()

  # Add blank row for each split
  for (i in seq_along(lst)) {
    
    # Don't append line for blank rows
    if ("..blank" %in% names(lst[[i]]) & 
        (all(lst[[i]][["..blank"]] == "B") | all(lst[[i]][["..blank"]] == "A"))) {
      
      ret[[i]] <- lst[[i]]
    } else {

      ret[[i]] <- add_blank_row(lst[[i]], location = location, vars = vars)
    }

  }

  # print("After blank row")
  # print(ret)

  # Combine splits
  ret <- do.call("rbind", ret)
  
  rownames(ret) <- NULL

  return(ret)

}


#' @noRd
get_page_size <- function(paper_size, units) {

  # If no paper size specified, 
  # make it essentially infinite.
  ret <- c(1000000, 1000000)
  
  if (units == "inches") {
    if (paper_size == "letter")
      ret <- c(8.5, 11)
    else if (paper_size == "legal")
      ret <- c(8.5, 14)
    else if (paper_size == "A4")
      ret <- c(8.27, 11.69)
    else if (paper_size == "RD4")
      ret <- c(7.7, 10.7)
    
  } else if (units == "cm") {
    if (paper_size == "letter")
      ret <- c(21.59, 27.94)
    else if (paper_size == "legal")
      ret <- c(21.59, 35.56)
    else if (paper_size == "A4")
      ret <- c(21, 29.7)
    else if (paper_size == "RD4")
      ret <- c(19.6, 27.3)
    
     # For character units, 
     # use inches and convert
  }  else if (units == "char") {
    if (paper_size == "letter")
      ret <- c(8.5, 11)
    else if (paper_size == "legal")
      ret <- c(8.5, 14)
    else if (paper_size == "A4")
      ret <- c(8.27, 11.69)
    else if (paper_size == "RD4")
      ret <- c(7.7, 10.7)
  }
  
  return(ret)
}


#' Split data frame cells into multiple rows based on expected column width.
#' This is necessary to wrap long values onto multiple lines, and have other
#' columns stay aligned horizontally.
#' @param x A data frame
#' @param col_widths A named vector of columns widths in number of characters
#' @return The data frame with long values split and added to their own
#' rows.
#' @noRd
split_cells <- function(x, col_widths, ts = NULL, char_width = NULL) {
  
  dat <- NULL           # Resulting data frame
  row_values <- list()  # A list to hold cell values for one row 
  max_length <- 0       # The maximum number of splits of a cell in that row
  wdths <- col_widths[!is.controlv(names(x))]
  defs <- ts$col_defs
  for (i in seq_len(nrow(x))) {
    for (nm in names(x)) {

      if (any(typeof(x[[nm]]) == "character") & 
          !is.control(nm) ) {
        
        ind_blank_num <- 0
        
        break_label_flg <- FALSE
        if (any(grepl("..break_label\\d+", names(x)))) {
          
          # For break labels, prepare the break label columns for later use
          break_label_col <- c()
          for (d in 1:length(defs)) {
            if (!is.null(defs[[d]]$break_label)){
              break_label_col <- c(break_label_col, defs[[d]]$var_c)
            }
          }
          
          # Use original indentation setting
          if (!is.na(x$..break_occur[i]) & x$..blank[i] == "L"){
            cur_break_label_col <- break_label_col[as.numeric(x$..break_occur[i])]
            if (!is.null(defs[[cur_break_label_col]]$indent)) {
              ind_blank_num <- floor(defs[[cur_break_label_col]]$indent / char_width)
            }
            break_label_flg <- TRUE
          }
        }
        
        if (!break_label_flg) {
          if (!is.null(defs)) {
            def <- defs[[nm]]
            if (!is.null(def$indent)) {
              
              ind_blank_num <- floor(def$indent / char_width)
              
            } else if (nm == "stub" & !is.null(ts$stub)) {
              
              stub_var <- x$..stub_var[i]
              if (!is.null(defs[[stub_var]]$indent)) {
                
                ind_blank_num <- floor(defs[[stub_var]]$indent / char_width)
                
              }
            }
          }
        }

        if ("..blank" %in% names(x) && x[[i, "..blank"]] %in% c("B", "A")) {
          
          cell <- substr(x[[i, nm]], 1, col_widths[[nm]])
          
        } else if ("..blank" %in% names(x) && x[[i, "..blank"]] == "L") {
          
          cell <- stri_wrap(unlist(
            strsplit(x[[i, nm]], split = "\n", fixed = TRUE)), 
            width = sum(wdths), normalize = FALSE, indent = ind_blank_num, exdent = ind_blank_num)
          
        } else {
          
          cell <- stri_wrap(unlist(
            strsplit(x[[i, nm]], split = "\n", fixed = TRUE)), 
            width = col_widths[[nm]], normalize = FALSE, indent = ind_blank_num, exdent = ind_blank_num)
        }
        
      
      } else {
        cell <- x[i, nm]
      }
      # print(paste("cell: ", cell))
      
      if (length(cell) > max_length)
        max_length <- length(cell)
      
      if (identical(cell, character(0)))
          cell <- ""
    
      row_values[[length(row_values) + 1]] <- cell
      # print(paste("Row:", row_values))
    }

    names(row_values) <- names(x)

    a <- align_cells(row_values, max_length)
    a$..row <- i

    if (is.null(dat))
      dat <- a
    else
      dat <- rbind(dat, a)
    max_length <- 0
    row_values <- list()
    
  }
  
  # Reset names
  if ("..row" %in% names(x)) 
    names(dat) <- c(names(x))
  else
    names(dat) <- c(names(x), "..row")

  return(dat)
}

#' @noRd
split_strings <- function(strng, width, units, multiplier = 1.03) {
 
  lnlngth <- 0
  ln <- c()
  lns <- c()
  wdths <- c()
  
  un <- "inches"
  w <- width
  if (units == "cm")
    w <- cin(width)
  
  
  if (!is.na(strng) & as.character(strng) != "") {
    
    splits <- unlist(stri_split_fixed(as.character(strng), "\n"))
    
    for (split in splits) {
      
      wrds <- strsplit(split, " ", fixed = TRUE)[[1]]
      
      # Old code
      # lngths <- (suppressWarnings(strwidth(wrds, units = un)) + 
      #              suppressWarnings(strwidth(" ", units = un))) * multiplier

      if (length(wrds) > 0) {
        lngths <- (strwdth(wrds, un) + strwdth(" ", un)) * multiplier
      } 
      
      # Loop through words and add up lines
      for (i in seq_along(wrds)) {
        
        lnlngth <- lnlngth + lngths[i] 
        if (lnlngth <= w)
          ln <- append(ln, wrds[i])
        else {
          
          if (length(ln) == 0) {
            
            lns[length(lns) + 1] <- wrds[i]
            if (units == "cm")
              wdths[length(wdths) + 1] <- ccm(lnlngth)
            else
              wdths[length(wdths) + 1] <- lnlngth
            
            lnlngth <- 0
            
          } else {
            # Assign current lines and counts
            lns[length(lns) + 1] <- paste(ln, collapse = " ")
            if (units == "cm")
              wdths[length(wdths) + 1] <- ccm(lnlngth - lngths[i])
            else
              wdths[length(wdths) + 1] <- lnlngth - lngths[i]
            
            # Assign overflow to next line
            ln <- wrds[i]
            lnlngth <- lngths[i]
          }
        }
        
        
      }
      
      # Deal with last line
      if (length(ln) > 0) {
        
        lns[length(lns) + 1] <- paste(ln, collapse = " ")
        if (units == "cm")
          wdths[length(wdths) + 1] <- ccm(lnlngth)
        else
          wdths[length(wdths) + 1] <- lnlngth
        
      }
      
      # Reset ln and lnlngth
      ln <- c()
      lnlngth <- 0
      
    }
    
    
  } else {
    
    lns <- ""
    wdths <- 0
  } 
  
  
  ret <- list(text = lns, 
              widths = wdths)
  
  return(ret)
}



strwdth <- Vectorize(function(wrd, un) {
  
  if (is.na(wrd)) {
    ret <- 0 
  } else {

    ret <- tryCatch({
    
      suppressWarnings(strwidth(wrd, units = un))
      
      
    
    }, error = function(cond) {
      
      suppressWarnings(strwidth("a", units = un)) * nchar(wrd)
        
    })
  }
  
  return(ret)
}, USE.NAMES = FALSE, SIMPLIFY = TRUE)

#' @description Calling function is responsible for opening the 
#' device context and assigning the font.  This function will use 
#' strwidth to determine number of wraps of a string within a particular
#' width.  Lines are returned as a single rtf string separated by an rtf
#' line ending.  
#' @noRd
split_string_rtf <- function(strng, width, units, font = "Arial", nm = "", 
                             char_width = 1) {
  
  
  if (tolower(font) == "courier")
    mp <- 1.01
  else 
    mp <- 1.02
  
  # Deal with indents
  blnks <- ""
  indnt <- 0
  cstrng <- strng
  indntw <- 0
  # if (nm == "stub") {
  #   
  #   bpos <- regexpr("^\\s+", strng)
  #   if (bpos > 0) {
  #     
  #     indnt <-  attr(bpos, "match.length")
  #     blnks <- paste0(rep(" ", indnt), sep = "", collapse = "")
  #     cstrng <- substr(strng, indnt + 1, nchar(strng))
  #     #indntw <- indnt * char_width / mp
  #   }
  # }
  
  res <- split_strings(cstrng, width - indntw, units, multiplier = mp)
  
  
  # Concat lines and add line ending to all but last line.
  # Also translate any special characters to a unicode rtf token
  # Doing it here handles for the entire report, as every piece runs
  # through here.
  ret <- list(rtf = paste0(blnks, encodeRTF(res$text), collapse = "\\line "),
              lines = length(res$text),
              widths = res$widths + indntw)
  
  return(ret)
}

#' @noRd
split_string_html <- function(strng, width, units, nm = "", char_width = 1) {
  
  
  # Deal with indents
  blnks <- ""
  indnt <- 0
  cstrng <- strng
  indntw <- 0
  # if (nm == "stub") {
  #   
  #   bpos <- regexpr("^\\s+", strng)
  #   if (bpos > 0) {
  #     
  #     indnt <-  attr(bpos, "match.length")
  #     blnks <- paste0(rep(" ", indnt), sep = "", collapse = "")
  #     cstrng <- substr(strng, indnt + 1, nchar(strng))
  #     indntw <- indnt * char_width 
  #   }
  # }
  
  
  res <- split_strings(cstrng, width - indntw, units, multiplier = 1)
  
  ret <- list(html = paste0(blnks, res$text, collapse = "\n"),
              lines = length(res$text),
              widths = res$widths + indntw)
  
  return(ret)
}

#' @noRd
split_string_docx <- function(strng, width, units, nm = "", char_width = 1,
                              font = "Arial") {
  
  if (tolower(font) == "courier") {
    mp <- 1.01
  } else {
    mp <- 1.02
  } 
  
  # Deal with indents
  blnks <- ""
  indnt <- 0
  cstrng <- strng
  indntw <- 0
  # if (nm == "stub") {
  #   
  #   bpos <- regexpr("^\\s+", strng)
  #   if (bpos > 0) {
  #     
  #     indnt <-  attr(bpos, "match.length")
  #     blnks <- paste0(rep(" ", indnt), sep = "", collapse = "")
  #     cstrng <- substr(strng, indnt + 1, nchar(strng))
  #     indntw <- indnt * char_width 
  #   }
  # }
  
  
  res <- split_strings(cstrng, width - indntw, units, multiplier = mp)
  
  ret <- list(docx = paste0(blnks, res$text, collapse = "\n"),
              lines = length(res$text),
              widths = res$widths + indntw)
  
  return(ret)
}


#' @noRd
split_string_text <- function(strng, width, units, nm = "", char_width = 1) {
  
  
  # Deal with indents
  blnks <- ""
  indnt <- 0
  cstrng <- strng
  indntw <- 0
  # if (nm == "stub") {
  #   
  #   bpos <- regexpr("^\\s+", strng)
  #   if (bpos > 0) {
  #     
  #     indnt <-  attr(bpos, "match.length")
  #     blnks <- paste0(rep(" ", indnt), sep = "", collapse = "")
  #     cstrng <- substr(strng, indnt + 1, nchar(strng))
  #     indntw <- indnt * char_width 
  #   }
  # }
  
  res <- split_strings(cstrng, width - indntw, units, multiplier = 1)
  
  ret <- list(text = paste0(blnks, res$text),
              lines = length(res$text),
              widths = res$widths + indntw)
  
  return(ret)
}

#' Split data frame cells based on expected column width.
#' For variable width reports, will just insert a carriage return at the 
#' split points.
#' @param x A data frame
#' @param col_widths A named vector of columns widths in the unit of measure
#' @return The data frame with long values split by carriage returns.
#' @import stringi
#' @import grDevices
#' @noRd
split_cells_variable <- function(x, col_widths, font, font_size, units, 
                                 output_type, char_width, ts) {
  
  dat <- NULL           # Resulting data frame
  wdths <- list()       # Resulting list of widths
  row_values <- list()  # A list to hold cell values for one row 
  row_widths <- list()  # A list to hold text widths for one row
  max_length <- 1       # The maximum number of splits of a cell in that row
  
  fnt <- "mono"
  if (tolower(font) == "arial")
    fnt <- "sans"
  else if (tolower(font) == "times")
    fnt <- "serif"
  
  defs <- ts$col_defs
  
  pdf(NULL)
  par(family = fnt, ps = font_size)
  
  # Parepare break_label temporary df
  break_label_df <- NULL
  if (any(grepl("..break_label\\d+", names(x)))) {
    
    break_label_df <- x[, grepl("..break_label", names(x))]
    
    # For break labels, prepare the break label columns for later use
    break_label_col <- c()
    for (d in 1:length(defs)) {
      if (!is.null(defs[[d]]$break_label)){
        break_label_col <- c(break_label_col, defs[[d]]$var_c)
      }
    }
    
    for (i in 1:nrow(break_label_df)) {
      for (nm in names(break_label_df)) {
        if (grepl("..break_label\\d+", nm)) {
          
          break_label_num <- sub(".*?([0-9]+).*", "\\1", nm)
          
          break_label_indent <- 0
          cur_break_label_col <- break_label_col[as.numeric(break_label_num)]
          if (!is.null(defs[[cur_break_label_col]]$indent)) {
            break_label_indent <- defs[[cur_break_label_col]]$indent
          }
          
          if (output_type %in% c("HTML")) {
            break_label_res <- split_string_html(x[[i, nm]], sum(col_widths) - break_label_indent, units)
            
            break_label_df[i, nm] <- break_label_res$html
            break_label_df[i, paste0("..break_label_lines",break_label_num)] <- break_label_res$lines
            
          } else if (output_type == "RTF") {
            break_label_res <- split_string_rtf(x[[i, nm]], sum(col_widths) - break_label_indent, units, font)
            
            break_label_df[i, nm] <- break_label_res$rtf
            break_label_df[i, paste0("..break_label_lines",break_label_num)] <- break_label_res$lines
          } else if (output_type == "PDF") {
            
            break_label_res <- split_string_text(x[[i, nm]], sum(col_widths) - break_label_indent, units)
            
            break_label_df[i, nm] <- paste0(break_label_res$text, collapse = "\n")
            break_label_df[i, paste0("..break_label_lines",break_label_num)] <- break_label_res$lines
          } else if (output_type == "DOCX") {
            break_label_res <- split_string_docx(x[[i, nm]], sum(col_widths) - break_label_indent, units, font = font)
            
            break_label_df[i, nm] <- break_label_res$docx
            break_label_df[i, paste0("..break_label_lines",break_label_num)] <- break_label_res$lines
          }
          
        }
      }
    }
    
    x <- x[, setdiff(names(x), names(break_label_df))]
  }
  
  
  for (i in seq_len(nrow(x))) {
    for (nm in names(x)) {
      
      nch <- 1
      res <- list(widths = 0)
      
      if (any(typeof(x[[nm]]) == "character") & 
          !is.control(nm) ) {
        
        if ("..blank" %in% names(x) && x[[i, "..blank"]] %in% c("B", "A")) {
          
          cell <- ""
          
        } else if ("..blank" %in% names(x) && x[[i, "..blank"]] == "L") {
          
          if (output_type %in% c("HTML")) {
            res <- split_string_html(x[[i, nm]], sum(col_widths), units)
            
            cell <- res$html
            
          } else if (output_type == "RTF") {
            res <- split_string_rtf(x[[i, nm]], sum(col_widths), units, font)
            
            cell <- res$rtf
          } else if (output_type == "PDF") {
            
            res <- split_string_text(x[[i, nm]], sum(col_widths), units)
            
            cell <- paste0(res$text, collapse = "\n")
            
          } else if (output_type == "DOCX") {
            res <- split_string_docx(x[[i, nm]], sum(col_widths), units, font = font)
            
            cell <- res$docx
          }
          
        } else {
          
          if (output_type %in% c("HTML")) {
            # For indenting values, the width should be (col_widths - indentation)
            if (!is.null(defs[[nm]]$indent)) {
              res <- split_string_html(x[[i, nm]], col_widths[[nm]] - defs[[nm]]$indent, 
                                       units, nm, char_width)
            } else if (nm == "stub" & !is.null(ts$stub)) {
              stub_var <- x$..stub_var[i]
              if (!is.null(defs[[stub_var]]$indent)) {
                res <- split_string_html(x[[i, nm]], col_widths[[nm]] - defs[[stub_var]]$indent, 
                                         units, nm, char_width)
              } else {
                res <- split_string_html(x[[i, nm]], col_widths[[nm]], units, nm, char_width)
              }
            } else {
              res <- split_string_html(x[[i, nm]], col_widths[[nm]], units, nm, char_width)
            }
            
            cell <- res$html
          
          } else if (output_type == "RTF") {
            # For indenting values, the width should be (col_widths - indentation)
            if (!is.null(defs[[nm]]$indent)) {
              res <- split_string_rtf(x[[i, nm]], col_widths[[nm]] - defs[[nm]]$indent, 
                                      units, font, nm, char_width)
            } else if (nm == "stub" & !is.null(ts$stub)) {
              stub_var <- x$..stub_var[i]
              if (!is.null(defs[[stub_var]]$indent)) {
                res <- split_string_rtf(x[[i, nm]], col_widths[[nm]] - defs[[stub_var]]$indent, 
                                        units, font, nm, char_width)
              } else {
                res <- split_string_rtf(x[[i, nm]], col_widths[[nm]], units, font, nm, char_width)
              }
            } else {
              res <- split_string_rtf(x[[i, nm]], col_widths[[nm]], units, font, nm, char_width)
            }
          
            cell <- res$rtf
          } else if (output_type == "PDF") {
            # For indenting values, the width should be (col_widths - indentation)
            if (!is.null(defs[[nm]]$indent)) {
              res <- split_string_text(x[[i, nm]], col_widths[[nm]] - defs[[nm]]$indent, 
                                       units, nm, char_width)
            } else if (nm == "stub" & !is.null(ts$stub)) {
              stub_var <- x$..stub_var[i]
              if (!is.null(defs[[stub_var]]$indent)) {
                res <- split_string_text(x[[i, nm]], col_widths[[nm]] - defs[[stub_var]]$indent, 
                                         units, nm, char_width)
              } else {
                res <- split_string_text(x[[i, nm]], col_widths[[nm]], units, nm, char_width)
              }
            } else {
              res <- split_string_text(x[[i, nm]], col_widths[[nm]], units, nm, char_width)
            }
            
            cell <- paste0(res$text, collapse = "\n")
            
          } else if (output_type == "DOCX") {
            # For indenting values, the width should be (col_widths - indentation)
            if (!is.null(defs[[nm]]$indent)) {
              res <- split_string_docx(x[[i, nm]], col_widths[[nm]] - defs[[nm]]$indent, 
                                       units, nm, char_width, font = font)
            } else if (nm == "stub" & !is.null(ts$stub)) {
              stub_var <- x$..stub_var[i]
              if (!is.null(defs[[stub_var]]$indent)) {
                res <- split_string_docx(x[[i, nm]], col_widths[[nm]] - defs[[stub_var]]$indent, 
                                         units, nm, char_width, font = font)
              } else {
                res <- split_string_docx(x[[i, nm]], col_widths[[nm]], units, nm, char_width, font = font)
              }
            } else {
              res <- split_string_docx(x[[i, nm]], col_widths[[nm]], units, nm, char_width, font = font)
            }
            
            cell <- res$docx
          }
          nch <- res$lines
        }
        
        
      } else {
        cell <- x[i, nm]
      }
      # print(paste("cell: ", cell))
      
      if (identical(cell, character(0)))
        cell <- ""
      
      # print(nch)
      # print(cell)
      # print(max_length)
      
      if (nch > max_length)
        max_length <- nch
      
      
      row_values[[length(row_values) + 1]] <- cell
      if (is.null(res$widths)) 
        row_widths[[length(row_widths) + 1]] <- 0
      else
        row_widths[[length(row_widths) + 1]] <- res$widths
      # print(paste("Row:", row_values))
    }
    
    # print(names(x))
    names(row_values) <- names(x)
    names(row_widths) <- names(x)
    
    row_values$..row <- max_length
    wdths[[length(wdths) + 1]] <- row_widths
    
    if (is.null(dat)) {
      dat <- as.data.frame(row_values, stringsAsFactors = FALSE, 
                           check.names = FALSE)
    } else {
      # names(dat)
      # names(row_values)
      dat <- rbind(dat, row_values)
    }
    
    max_length <- 1
    row_values <- list()
    row_widths <- list()
  }
  
  dev.off()
  
  if (!is.null(break_label_df)) {
    dat <- cbind(dat, break_label_df)
  }
  
  rownames(dat) <- NULL
  # # Reset names
  # if ("..row" %in% names(x)) 
  #   names(dat) <- c(names(x))
  # else
  #   names(dat) <- c(names(x), "..row")

  ret <- list(data = dat,
              widths = wdths)
  
  return(ret)
}



#' Given a jagged set of vectors, align to the longest by filling with 
#' empty strings
#' @param x A list of vectors of varying lengths
#' @param len The length of the longest vector
#' @return A data frame with the number of rows equal to len. Missing cell
#' values are filled with empty strings.
#' @noRd
align_cells <- function(x, len) {
  
  ret <- list() 
  
  for(nm in names(x)) {
    
    t <- len - length(x[[nm]])
    
    if (t > 0) {
      if (nm == "..blank" | grepl("..group_cohesion", nm) | grepl("..min_page_prop", nm)
          | grepl("..break_label\\d+", nm)) {
        if (is.na(x[[nm]])) 
          v <- c(rep(NA, t))
        else
          v <- c(rep(x[[nm]], t))
      } else if (any(typeof(x[[nm]]) == "character")) 
        v <- c(rep("", t))
      else
        v <- c(rep(NA, t))
      
      ret[[nm]] <- c(x[[nm]], v)
    } else {
      
      ret[[nm]] <-  x[[nm]]
    }
  }
  
  names(ret)  <- names(x)
  ret <- as.data.frame(ret, stringsAsFactors = FALSE)
  
  return(ret)
  
}

#' For any remaining non-character columns, convert to character and
#' clear the missing values.
#' @noRd
clear_missing <- function(x, missing_val) {
  
  for (nm in names(x)) {
    if (!is.control(nm)) { 
      
      if (typeof(x[[nm]]) != "character")
        x[[nm]] <- as.character(x[[nm]])
      
      x[[nm]] <- ifelse(is.na(x[[nm]]), missing_val, x[[nm]])
    }
  }
  
  return(x)
}

#' Push string values down to lowest row in data frame
#' @noRd
push_down <- function(x) {

  tot <- nrow(x)

  for (nm in names(x)) {
    tmp <- x[[nm]][x[[nm]] != ""]
    x[[nm]] <- c(rep("", tot - length(tmp)), tmp) 
  }
  
  return(x)
}

#' @description Dedupe requested columns
#' @details This function is performed in the page splitting routine
#' so that groups which span multiple pages retain a label at the top 
#' of the page.
#' @import common
#' @noRd
dedupe_pages <- function(pgs, defs) {
  
  ret <- list()

  for (dat in pgs) {
    nms <- names(dat)
    for (def in defs) {
      if (def$dedupe) {
        
        if(def$var_c %in% nms) {
        
          # Convert to character if necessary
          if (all(dat[[def$var_c]] != "character"))
            dat[[def$var_c]] <- as.character(dat[[def$var_c]])
          
          # Fill with blanks as appropriate
          w <- min(nchar(dat[[def$var_c]])) # Take min to exclude label row
          v <- paste0(rep(" ", times = w), collapse = "")
          
          if (length(dat[[def$var_c]]) > 1) {
      
            if ("..blank" %in% names(dat)) {
              non_blank_idx <- dat[["..blank"]] == ""
              dat[[def$var_c]][non_blank_idx] <- ifelse(changed(dat[[def$var_c]][non_blank_idx]),
                                                        dat[[def$var_c]][non_blank_idx], v)
            } else {
              dat[[def$var_c]] <- ifelse(changed(dat[[def$var_c]]),
                                         dat[[def$var_c]], v)
              
            }
          }
          
          # dat[[def$var_c]] <- ifelse(!duplicated(dat[[def$var_c]]), 
          #                            dat[[def$var_c]], v) 
        }
      }
    }
    ret[[length(ret) + 1]] <- dat
  }
  
  return(ret)
}


clear_formats <- function(x) {
  
  for (nm in names(x)) {
   attr(x[[nm]], "format") <- NULL 
  }
  
  return(x)
}

quote_names <- function(x) {
 
  
  if (typeof(substitute(x, env = environment())) == "language") 
    v <- substitute(x, env = environment())
  else 
    v <- substitute(list(x), env = environment())
  
  vars <- c()
  if (length(v) > 1) {
    for (i in 2:length(v)) {
      vars[[length(vars) + 1]] <- as.character(v[[i]]) 
    }
    
  }
  
  ret <- unlist(vars)
  
  return(ret)
  
}

set_column_defaults <- function(ts, keys) {
  
  ret <- ts$col_defs
  dflts <- ts$col_dflts
  if (!is.null(dflts)) {
    for (i in seq_along(dflts)) {
      
      # A vector of columns names for the defaults
      v <- c()
      
      # An individual default
      dflt <- dflts[[i]]
      
      # Need to get vector of column names whether it is 
      # passed to the default as a vector or a range
      if (!is.null(dflt$vars))
          v <- dflt$vars
      else if (!is.null(dflt$from) & !is.null(dflt$to)) {
         startpos <- which(keys == dflt$from)
         endpos <- which(keys == dflt$to)
         if (startpos > endpos) {
           stop(paste("'from' and 'to' parameters on column_defaults",
                      "must be specified from left to right"))
           
         }
         
         # Get vector of column names from range
         v <- keys[seq(from = startpos, to = endpos)]
        
      } else 
        v <- keys[!is.controlv(keys)]
      
      
      
      for (nm in v) {
        #print(nm)
        
        # if definition doesn't exist, create it
        if (nm %in% names(ret))
          def <- ret[[nm]]
        else 
          def <- define_c(nm)
        
        if (is.null(def$width))
          def$width <- dflt$width
        if (is.null(def$align))
          def$align <- dflt$align
        # Have to deal with $ partial matching
        if (is.null(def[["label"]]))
          def[["label"]] <- dflt[["label"]]
        if (is.null(def$format))
          def$format <- dflt$format
        if (is.null(def$label_align))
          def$label_align <- dflt$label_align
        if (is.null(def$n))
          def$n <- dflt$n
        if (is.null(def$style))
          def$style <- dflt$style
        
        ret[[nm]] <- def
      }
      

    }
  
  }
  
  return(ret)
}

# Create a list of style specs for all columns that have styles assigned.
get_styles <- function(ts) {
  
  ret <- list()
  if (!is.null(ts$col_defs)) {
    for (def in ts$col_defs) {
      
      if (!is.null(def$style)) {
        if (!is.null(def$style$indicator)) {
          icol <- ts$col_defs[[def$style$indicator]] 
          if (!is.null(icol)) {
            if (icol$visible == FALSE) {
              def$style$indicator <- paste0("..x.", def$style$indicator)
            }
          }
        }
        ret[[def$var_c]] <- def$style 
      }
      if (!is.null(ts$stub$style)) {
        if (!"stub" %in% names(styles)) { 
          if (!is.null(ts$stub$style$indicator)) {
            icol <- ts$col_defs[[ts$stub$style$indicator]] 
            if (!is.null(icol)) {
              if (icol$visible == FALSE) {
                ts$stub$style$indicator <- paste0("..x.", ts$stub$style$indicator)
              }
            }
          }
          
          ret[["stub"]] <- ts$stub$style
        }
      }
    }
  } 
  
  return(ret)
}


#' @noRd
getExtension <- function(file){ 
  ex <- strsplit(basename(file), split="\\.")[[1]]
  return(ex[-1])
} 


#' @noRd
log_logr <- function(x) {
 
  if (length(find.package('logr', quiet=TRUE)) > 0) {
    if (utils::packageVersion("logr") >= "1.2.0") {
      logr::log_hook(x)
    }
  }
}

#' @noRd
has_bottom_footnotes <- function(rs) {
  
  ret <- FALSE
  
  if (!is.null(rs$footnotes)) {
    
    for (ftn in rs$footnotes) {
      if (ftn$valign == "bottom") { 
        ret <- TRUE
        break()
      }
    }
    
  }
  
  return(ret)
}


#' @noRd
has_top_footnotes <- function(rs) {
  
  ret <- FALSE
  
  if (!is.null(rs$footnotes)) {
    
    for (ftn in rs$footnotes) {
      if (ftn$valign == "top") { 
        ret <- TRUE
        break()
      }
    }
    
  }
  
  return(ret)
}

#' @description  Translate border spec to a vector of border positions
#' @noRd
get_outer_borders <- function(brdr_spec) {
  
  ret <- c()
  if (any(brdr_spec == "none")) {
    ret <- c()
  } else {
    
    if (any(brdr_spec == "outside") | any(brdr_spec == "all")) {
      
      ret <- append(ret, c("top", "bottom", "left", "right")) 
    } else {
    
      if (any(brdr_spec == "top"))
        ret <- append(ret, "top")
      if (any(brdr_spec == "bottom"))
        ret <- append(ret, "bottom")
      if (any(brdr_spec == "left"))
        ret <- append(ret, "left")
      if (any(brdr_spec == "right"))
        ret <- append(ret, "right")
    }

  } 
  
  return(ret)
}

#' @noRd
has_page_footer <- function(rs) {
  
  ret <- FALSE
  
  if (!is.null(rs$page_footer_left)) {
    
        ret <- TRUE
  }
  if (!is.null(rs$page_footer_right)) {
    ret <- TRUE
  }
  if (!is.null(rs$page_footer_center)) {
    ret <- TRUE
  }
  
  return(ret)
}

get_spanning_info <- function(rs, ts, pi, widths, gutter = 1) {
  
  spns <- ts$col_spans
  cols <- pi$keys
  cols <- cols[!is.controlv(cols)]
  w <- widths
  w <- w[cols]
  
  # print("Cols:")
  # print(cols)
  #print(w)
  
  # Figure out how many levels there are, 
  # and organize spans by level
  lvls <- c()     # Unique levels
  slvl <- list()  # Will be a list of lists of spans
  for (spn in spns) {
    
    if (!spn$level %in% lvls) {
      lvls[length(lvls) + 1] <- spn$level
      slvl[[spn$level]] <- list()
    }
    slvl[[spn$level]][[length(slvl[[spn$level]]) + 1]] <- spn
  }
  
  # Get unique levels and sort in decreasing order so we go from top down
  lvls <- sort(unique(lvls), decreasing = TRUE)
  # print("Levels:")
  # print(lvls)
  # 
  # print("Spanning levels:")
  # print(slvl)
  
  # Create data structure to map spans to columns and columns widths by level
  # - Seed span_num with negative index numbers to identify unspanned columns
  # - Also add gutter to each column width for the space between columns 
  d <- data.frame(colname = cols, colwidth = w + gutter, 
                  span_num = seq(from = -1, to = -length(cols), by = -1), 
                  stringsAsFactors = FALSE)
  
  wlvl <- list()  # Create one data structure for each level
  for (l in lvls) {
    
    t <- d  # Copy to temporary variable
    # print(t)
    # print(slvl)
    col_span <- c()
    # if column is in spanning column list, populate structure with index.
    # Otherwise, leave as negative value.
    for (i in 1:length(slvl[[l]])) {
      cl <- slvl[[l]][[i]]$span_cols
      
      # Make sure only visible columns are included
      cl <- cl[cl %in% t$colname]
      
      # Span specifications can be a vector of column names or numbers
      if (typeof(cl) == "character")
        t$span_num <- ifelse(t$colname %in% cl, i, t$span_num)
      else 
        t$span_num <- ifelse(t$colname %in% cols[cl], i, t$span_num)
      
      col_span[i] <- length(cl)

    }
    
    # Aggregate data structures to get span widths for each span
    s <- aggregate(x = list(width = t$colwidth), by = list(span = t$span_num), FUN = sum)
    
    # Then put back in original column order
    s$span <- factor(s$span, levels = unique(t$span_num))
    s <- s[order(s$span), ]
    rownames(s) <- NULL
    
    # Prep data structure
    s$span <- unique(t$span_num)
    s$label <- ""
    s$align <- ""
    s$n <- NA
    s$name <- ""
    s$underline <- TRUE
    s$col_span <- 1
    s$bold <- FALSE
    
    # Populate data structure with labels, alignments, and n values from 
    # spanning column objects
    counter <- 1
    for (index in s$span) {
      if (index > 0) {
        s$label[counter] <- slvl[[l]][[index]]["label"]
        s$align[counter] <- slvl[[l]][[index]]$label_align
        s$underline[counter] <- slvl[[l]][[index]]$underline
        s$bold[counter] <- slvl[[l]][[index]]$bold
        if (!is.null(slvl[[l]][[index]]$n))
          s$n[counter] <- slvl[[l]][[index]]$n
        
      }
      s$name[counter] <- paste0("Span", counter)
      if (index > 0) {
        if (!is.na(col_span[index]))
          s$col_span[counter] <- col_span[index]
      }
      counter <- counter + 1
    }
    
    # Apply n counts to labels
    if (!is.null(ts$n_format)) {
      s$label <- ifelse(is.na(s$n), s$label, paste0(s$label, ts$n_format(s$n))) 
    }
    
    wlvl[[l]] <- s
    
  }
  
  
  return(wlvl)
  
}


#' @import stringi
apply_widths <- function(dat, wdths, algns) {
  
  w <- wdths[!is.controlv(names(wdths))]
  ret <- dat
  
  for (nm in names(ret)) {
    if (!is.control(nm)) {
      
      cw <- wdths[[nm]]
      
      # If column width is miscalculated, fix it.
      if ("..blank" %in% names(dat)) {
        mxw <- max(ifelse(ret[["..blank"]] == "L", 0, nchar(ret[[nm]])), na.rm = TRUE)
        if (mxw > cw)
          cw <- mxw
      }
        
      if (algns[nm] == "left") {
        
        ret[[nm]] <- stri_pad_right(ret[[nm]], cw)
      } else if (algns[nm] == "right") {
        
        ret[[nm]] <- stri_pad_left(ret[[nm]], cw)
      } else if (algns[nm] == "center") {
        
        ret[[nm]] <- stri_pad_both(ret[[nm]], cw) 
      }
      
      # Clear out label rows
      ret[[nm]] <- clear_labels(ret[[nm]], ret[["..blank"]], sum(w)) 

    }
    
    if (any(class(ret[[nm]]) == "factor"))
      ret[[nm]] <- as.character(ret[[nm]])
    
  }
  

   
  return(ret)
}

#' @import stringi
clear_labels <- Vectorize(function(vect, bvect, tw) {
  
  ret <- vect
  if (bvect == "L") {
    if (trimws(vect) == "") {
      ret <- ""
    
    } else {
      
      ret <- stri_pad_right(vect, tw)
    }
  }
  
  return(ret)
}, USE.NAMES = FALSE)

#' Add cell border indicator
#' @noRd
add_group_border_ind <- function(df, vars, border = "bottom"){
  
  df$..group_border <- rep("", nrow(df))
  for (i in vars) {
    last_pos <- get_last_pos(df[[i]])
    df$..group_border[last_pos] <- border
  }
  
  return(df)
}

#' Return the last position logical value of a group
#' @noRd
get_last_pos <- function(x) {
  # Initialize all as FALSE
  result <- logical(length(x))
  
  # Mark TRUE where the next value is different or at the end
  result[c(which(x[-length(x)] != x[-1]), length(x))] <- TRUE
  
  return(result)
}

#' Return the first position logical value of a group
#' @noRd
get_first_pos <- function(x){
  # Initialize all as FALSE
  result <- logical(length(x))
  
  # Mark TRUE where the next value is different
  result <- x[-length(x)] != x[-1]
  
  # The first one is always TRUE
  result <- c(TRUE, result)
  
  return(result)
}

#' Return the count of continuous group value. This the count_var is not input,
#' the function will count each record as 1
#' @noRd
get_group_count <- function(group_var, count_var = NULL) {
  if (length(group_var) == 0) {
    return(integer(0))
  } 
  
  # Find out the changing position
  change_points <- as.logical(c(TRUE, mapply(function(x, y) !identical(x, y), 
                                             group_var[-1], 
                                             group_var[-length(group_var)])))
  
  # Sum cumulatively to give ID for each group
  group_id <- cumsum(change_points)
  
  # Calculate size of each group
  if (is.null(count_var)) {
    count_var <- rep(1, length(group_var))
  }
  group_sizes <- tapply(count_var, group_id, sum)
  
  # Put frequency back to the original position
  result <- as.integer(group_sizes[group_id])
  
  return(result)
}

#' Return the sequence of continuous group value.
#' @noRd
get_group_seq <- function(x) {
  # use rle (run length encoding) to find the continuous same group
  ret <- rle(x)
  # repeat group sequences
  return(rep(seq_along(ret$lengths), ret$lengths))
}

# Sizing utilities --------------------------------------------------------

#' @noRd
get_content_size <- function(rs) {
  
  
  # Assume landscape
  pg_h <- rs$page_size[1]
  pg_w <- rs$page_size[2]
  
  # Change to portrait
  if(rs$orientation == "portrait") {
    pg_w <- rs$page_size[1]
    pg_h <- rs$page_size[2]
  }
  
  # Calculate available space for page body
  ret <- c(height = pg_h - rs$margin_top - rs$margin_bottom ,
           width = pg_w - rs$margin_right - rs$margin_left)

  
  return(ret)
  
}


#' @noRd
ccm <- function(x) {
  
  return(2.54 * x)
}

#' @noRd
cin <- function(x) {
  
  return(x / 2.54)
}


units_html <- function(u) {
  
 ret <- u  
 if (ret == "inches")
   ret <- "in"
 
 
 return(ret)
  
}

# RTF Functions -----------------------------------------------------------



#' @description Estimate number of wraps based on text, width, and a font.
#' @import stringi
#' @noRd
get_lines_rtf <- function(txt, width, font, font_size = 10, units = "inches") {
  
  
  f <- "mono"
  if (tolower(font) == "arial")
    f <- "sans"
  else if (tolower(font) == "times")
    f <- "serif"
  
  names(width) <- NULL
  
  lns <- unlist(stri_split_fixed(txt, "\n"))
  
  val <- get_text_width(lns, units = units, 
                        font = font, font_size = font_size) * .975/width
  
  # print(val)
  ret <- sum(ceiling(val))

  
  return(ret)
}


#' @description Estimate number of wraps based on text, width, and a font.
#' @import graphics
#' @noRd
get_text_width <- function(txt, font, font_size = 10, units = "inches", 
                           multiplier = .975) {
  
  
  f <- "mono"
  if (tolower(font) == "arial")
    f <- "sans"
  else if (tolower(font) == "times")
    f <- "serif"
  
  un <- "inches"

  
  #R.devices::devEval("nulldev", {
    pdf(NULL)
    par(family = f, ps = font_size)
    if (length(txt) > 0) {
      #ret <- suppressWarnings(strwidth(txt, units = un)) * multiplier 
      ret <- strwdth(txt, un) * multiplier 
    } else {
      
      ret <- 0 
    }
    dev.off()
  #})
  

  if (units == "cm")
    ret <- ccm(ret)
    
  return(ret)
}



# PDF Functions -----------------------------------------------------------

# Convert units to points for PDF x/y placement
cpoints <- function(vals, units) {
  
  if (units == "inches")
    ret <- round(vals * 72, 3)
  else if (units == "cm")
    ret <- round((vals / 2.54) * 72, 3)
  
  names(ret) <- NULL
  
  return(ret)
}

get_points_left <- function(left_bound, right_bound, widths, units) {
  
  # Three points added to the left function to prevent text from touching borders
  ret <- cpoints(rep(left_bound, length(widths)), units) + 3
  
  names(ret) <- NULL
  
  return(ret)
  
} 

get_points_right <- function(left_bound, right_bound, widths, units) {
  
  ret <- cpoints(right_bound - widths, units)
  
  names(ret) <- NULL
  
  return(ret)
} 

get_points_center <- function(left_bound, right_bound, widths, units) {
  
  # One point added to the center function to prevent text from touching borders
  # and center the text more accurately.
  ret <- cpoints(left_bound + ((right_bound - left_bound)/ 2) - (widths/ 2), units) + 1
  
  names(ret) <- NULL
  
  return(ret)
}  

get_points <- function(left_bound, right_bound, widths, units, align) {
  
  ret <- NULL
  if (tolower(align) == "left")
    ret <- get_points_left(left_bound, right_bound, widths, units)
  else if (tolower(align) == "right")
    ret <- get_points_right(left_bound, right_bound, widths, units)
  else if (tolower(align) %in% c("center", "centre"))
    ret <- get_points_center(left_bound, right_bound, widths, units)
  
  
  return(ret)
}


gluev <- function(str) {
 
  ret <- str
  
  if (!is.null(ret)) {
    if (!all(is.na(ret))) {
      if (typeof(ret) == "character") { 
        ret <- glueint(str)
      }
    }
  } 
  
  return(ret)
}

#' @import glue
glueint <- Vectorize(function(str) {
  
  ret <- glue(str, .trim = FALSE)
  
  return(ret) 
}, USE.NAMES = FALSE, SIMPLIFY = TRUE)


#' @noRd
has_glue <- function() {
  
 ret <- TRUE
 
 opts <- options("reporter.glue")[[1]]
 if (!is.null(opts)) {
   if (opts == FALSE) {
     ret <- FALSE 
   }
 }
 
 return(ret)
  
}

context("Utilities Tests")


test_that("utils1: gen_groups() works as expected.", {
  
  expect_equal(gen_groups(10, 3),
               c(1, 1, 1, 2, 2, 2, 3, 3, 3, 4))
  
  expect_equal(gen_groups(12, c(3, 2, 5, 2)),
               c(1, 1, 1, 2, 2, 3, 3, 3, 3, 3, 4, 4))

  expect_equal(gen_groups(10, 3, last_indices = TRUE), 
               c(3, 6, 9, 10))

  expect_equal(gen_groups(12, c(3, 2, 5, 2), last_indices = TRUE),
               c(3, 5, 10, 12))

})

test_that("utils2: get_font_family() works as expected.", {
  
  expect_equal(get_font_family("Arial"), "sans")
  expect_error(get_font_family("Calibri"))
  expect_equal(get_font_family("Courier"), "mono")
  expect_equal(get_font_family("Times"), "serif")
  expect_error(get_font_family("Wingdings"))
  
})


test_that("utils3: get_page_size() works as expected.", {
  ret <- get_page_size("letter", "inches")
  
  expect_equal(ret[1], 8.5)
  expect_equal(ret[2], 11)
  
  ret <- get_page_size("A4", "cm")
  expect_equal(ret[1], 21)
  expect_equal(ret[2], 29.7)
  
})


test_that("utils4: add_blank_rows() works as expected.", {
  
  tdat <- iris[order(iris$Species, decreasing = TRUE), ]
  tdat$cat <- c(rep("A", 25), rep("B", 25))
  
  res <- add_blank_rows(tdat, location = "both", vars = c("cat", "Species"))
  
  expect_equal(nrow(res), nrow(iris) + 12)
  
  expect_equal("..blank" %in% names(res), TRUE) 
  
  res2 <- add_blank_rows(tdat, location = "below", vars = c("Species"))
  
  expect_equal(nrow(res2), nrow(iris) + 3)

  res3 <- add_blank_rows(iris, location = "above")
  
  expect_equal(nrow(res3), nrow(iris) + 1)
  
})



test_that("utils5: split_cells works as expected.", {
  
  
  a1 <- c("one", "two", "three")
  b1 <- c("here is a\n long value", "Some sort of value", 
          "Another value that goes on and on")
  c1 <- c("here is a longer value", "Really", "Medium value")
  ..blank <- ""
  ..page <- NA
  ..row <- NA
  
  df <- data.frame(a1, b1, c1, ..blank, ..page, ..row, stringsAsFactors = FALSE)
  df
  
  w <- c(a1 = .5, b1 = 1.5, c1 = 1.25)
  
  conversion_factor <- .083333 
  
  wc <- floor(w / conversion_factor)

  
  res1 <- split_cells(df, wc)
  res1
  
  expect_equal(nrow(res1), 5) 
  
})

test_that("utils5-2: split_cells with indentation works as expected.", {
  
  
  a1 <- c("one", "two", "three")
  b1 <- c("here is a\n long value", "Some sort of value", 
          "Another value that goes on and on")
  c1 <- c("here is a longer value", "Really", "Medium value")
  ..blank <- ""
  ..page <- NA
  ..row <- NA
  
  df <- data.frame(a1, b1, c1, ..blank, ..page, ..row, stringsAsFactors = FALSE)
  
  w <- c(a1 = .5, b1 = 1.5, c1 = 1.25)
  
  conversion_factor <- .083333 
  
  wc <- floor(w / conversion_factor)
  
  tbl <- create_table(df) %>%
    define("b1", indent = 0.25)
  
  res1 <- split_cells(df, wc, ts = tbl, char_width = 0.11)
  res1
  
  expect_equal(nrow(res1), 7) 
  expect_equal(res1$b1, c("  here is a", "   long value", "  Some sort of", 
                          "  value", "  Another value", "  that goes on and", "  on"))
  
})

test_that("utils6: align_cells works as expected.", {
  
  
  x <- list(a1 = c("a", "b", "c"), 
            b1  = c("a"), 
            c1 = c("a", "b", "c", "d", "e"))
  
  g <- align_cells(x, 5)
  
  expect_equal(nrow(g), 5)
  
})


test_that("utils7: clear_missing works as expected.", {
  
  
  df <- data.frame(a = c(1, 2, 3), b = c("A", "B", "C"), 
                   stringsAsFactors = FALSE)
  
  df2 <- clear_missing(df)

  expect_equal(class(df2$a), "character")
  expect_equal(sum(is.na(df2$a)), 0)
  
})


test_that("utils8: push_down works as expected.", {
  
  x <- data.frame(a1 = c("a", "b", "c", "", ""), 
            b1  = c("a", rep("", 4)), 
            c1 = c("a", "b", "c", "d", "e"), stringsAsFactors = FALSE)
  
  res2 <- push_down(x)
  
  expect_equal(res2$a1, c("", "", "a", "b", "c"))
  expect_equal(res2$b1, c("", "", "", "", "a"))
  expect_equal(res2$c1, c("a", "b", "c", "d", "e"))
  
})

test_that("utils9: create_stub works as expected.", {
  
  x <- data.frame(a1 = c("a", "a", "a", "a", "a", "a"), 
                  b1  = c(NA, "b", "b", NA, "b", "b"), 
                  c1 = c(NA, NA, "c", NA, NA, "c"), 
                  d1 = c("Here", "is", "some", "stuff", "shouldn't", "touch"), 
                  stringsAsFactors = FALSE)
  x
  
  tbl <- create_table(x) %>% stub(c("a1", "b1", "c1"))
  
  d <- create_stub(x, tbl)
  
  expect_equal(d$stub, c("a", "b", "c", "a", "b", "c"))
  expect_equal(d$..stub_var, c("a1", "b1", "c1", "a1", "b1", "c1"))
})

test_that("utils9-2: create_stub works as expected.", {
  
  x <- data.frame(var = c(rep(NA, 4), rep("RACE", 5), rep("", 4)), 
                  label = c("", "txt1", "txt2", "", "", "txt3", "txt4", "txt5", "",
                            "", "txt6", "txt7", ""), 
                  A = c("", "txt1", "txt2", "", "", "txt3", "txt4", "txt5", "",
                         "", "txt6", "txt7", ""), 
                  B = c("", "txt1", "txt2", "", "", "txt3", "txt4", "txt5", "",
                         "", "txt6", "txt7", ""), 
                  ..blank = c("L", "", "", "B", "L", "", "", "", "B", "L", "",
                              "", "B"),
                  stringsAsFactors = FALSE)
  
  tbl <- create_table(x) %>% stub(c("var", "label"))
  
  d <- create_stub(x, tbl)
  row.names(d) <- 1:nrow(d)
  
  exp_d <- data.frame(
    stub = c("txt1", "txt2", "", "RACE", "txt3", "txt4", "txt5", "", "", "txt6", "txt7", ""),
    A = c("txt1", "txt2", "", "", "txt3", "txt4", "txt5", "", "", "txt6", "txt7", ""),
    B = c("txt1", "txt2", "", "", "txt3", "txt4", "txt5", "", "", "txt6", "txt7", ""),
    ..blank = c("", "", "B", "L", "", "", "", "B", "L", "", "", "B")
  )
  
  expect_equal(d[, setdiff(names(d), "..stub_var")], exp_d)
})

test_that("utils10: quote_names function works as expected.", {

  expect_equal(quote_names(c("fork")), c("fork"))
  expect_equal(quote_names(c(fork)), c("fork"))
  expect_equal(quote_names("fork"), c("fork"))
  expect_equal(quote_names("fork"), c("fork"))
  expect_equal(quote_names(c("fork", "bork")), c("fork", "bork"))
  expect_equal(quote_names(c(fork, bork)), c("fork", "bork"))
  
})

test_that("utils11: set_column_defaults function works as expected.", {
  
  tbl <- create_table(mtcars) 
  tbl <- column_defaults(tbl, c(mpg, cyl, disp), width = 2, format = "%.1f")
  tbl <- define(tbl, mpg, width = 3, format = "%.2f")
  
  res <- set_column_defaults(tbl, names(mtcars))

  
  expect_equal(length(res), 3)
  expect_equal(res[["mpg"]]$width, 3)
  expect_equal(res[["mpg"]]$format, "%.2f")
  expect_equal(res[["cyl"]]$width, 2)
  expect_equal(res[["cyl"]]$format, "%.1f")
  expect_equal(res[["disp"]]$width, 2)
  expect_equal(res[["disp"]]$format, "%.1f")
  
})
  

test_that("utils12: set_column_defaults function with range works as expected.", {
  
  tbl <- create_table(mtcars) 
  tbl <- column_defaults(tbl, from = mpg, to = drat, width = 2, format = "%.1f")
  tbl <- define(tbl, mpg, width = 3, format = "%.2f")
  
  res <- set_column_defaults(tbl, names(mtcars))
  
  
  expect_equal(length(res), 5)
  expect_equal(res[["mpg"]]$width, 3)
  expect_equal(res[["mpg"]]$format, "%.2f")
  expect_equal(res[["cyl"]]$width, 2)
  expect_equal(res[["cyl"]]$format, "%.1f")
  expect_equal(is.null(res[["vs"]]$width), TRUE)
  expect_equal(is.null(res[["vs"]]$format), TRUE)
  
})


test_that("utils13: set_column_defaults function all cols works as expected.", {
  
  tbl <- create_table(mtcars) 
  tbl <- column_defaults(tbl, width = 2, format = "%.1f")
  tbl <- define(tbl, mpg, width = 3, format = "%.2f")
  
  res <- set_column_defaults(tbl, names(mtcars))
  
  
  expect_equal(length(res), 11)
  expect_equal(res[["mpg"]]$width, 3)
  expect_equal(res[["mpg"]]$format, "%.2f")
  expect_equal(res[["cyl"]]$width, 2)
  expect_equal(res[["cyl"]]$format, "%.1f")
  expect_equal(res[["vs"]]$width, 2)
  expect_equal(res[["vs"]]$format, "%.1f")
  
})

# Commented out this test because the dm_final.rds file creates a dependency 
# on higher level of R > 3.5.
test_that("utils14: data with extra names on variables doesn't crash labels function.", {
  
  # fp <- file.path(base_path, "user/user8.out")
  # 
  # dat <- readRDS(file.path(base_path, "./data/dm_final.rds"))
  # 
  # #names(dat) 
  # 
  # #dat1 <- mtcars
  # 
  # #names(dat1$cyl) <- c("fork", "bork", "spork") 
  # 
  # names(dat$statistic)
  # tbl <- create_table(dat)
  # 
  # labels <- get_labels(dat, tbl)
  # 
  # names(labels)
  # length(labels)
  # 
  # ncol(dat)
  # 
  # names(dat)
  # 
  # labels[names(dat)]
  
  expect_equal(TRUE, TRUE)
  
})


test_that("utils15: split_cells_variable works as expected.", {
  
  # split_cells_variable <- function(x, col_widths, font, font_size, units)
  
  df <- data.frame(car = c("Dart", "Bart", "Here is another car name"),
                   cyl = c(8, 3, 4),
                   disp = c("My lovely disp and it gets longer", 
                            "Here is a big long string value to exceed 2 inches",
                            "Short"), stringsAsFactors = FALSE)
  
  tbl <- create_table(df) %>% 
    define(car, label = "Cars", width = 1) %>% 
    define(cyl, width = 1.5) %>% 
    define(disp, width = 2) 
  
  lbls <- get_labels(df, tbl)
  
  cw <- get_col_widths_variable(df, tbl, lbls, "Arial", 12, "inches", .2)
  cw
  
  res <- split_cells_variable(df, cw, "Arial", 12, "inches", "RTF", ts = tbl)
  res$widths
  
  expect_equal(any(class(res$data) == "data.frame"), TRUE)
  expect_equal(res$data[1, "..row"], 2)  
  expect_equal(res$data[3, "..row"], 3) 
  
})

test_that("utils15: split_cells_variable works with indentation as expected.", {
  
  # split_cells_variable <- function(x, col_widths, font, font_size, units)
  
  df <- data.frame(car = c("Dart", "Bart", "Here is another car name"),
                   cyl = c(8, 3, 4),
                   disp = c("My lovely disp and it gets longer", 
                            "Here is a big long string value to exceed 2 inches",
                            "Short"), stringsAsFactors = FALSE)
  
  tbl <- create_table(df) %>% 
    define(car, label = "Cars", width = 1, indent = 0.8) %>% 
    define(cyl, width = 1.5, indent = 1) %>% 
    define(disp, width = 2, indent = 1.5) 
  
  lbls <- get_labels(df, tbl)
  
  cw <- get_col_widths_variable(df, tbl, lbls, "Arial", 12, "inches", .2)
  cw
  
  res <- split_cells_variable(df, cw, "Arial", 12, "inches", "RTF", ts = tbl)
  res$widths
  
  expect_equal(any(class(res$data) == "data.frame"), TRUE)
  expect_equal(res$data[1, "..row"], 5)
  expect_equal(res$data[2, "..row"], 7)
  expect_equal(res$data[3, "..row"], 4) 
  
})


test_that("utils16: get_width function works as expected.", {
  
  res1 <- get_text_width("This is cool.", "Arial", 12)
  res1
  
  res2 <- get_text_width("This is cool.", "Arial", 10)
  res2
  
  expect_equal(res1 > res2, TRUE)
  
  
  res3 <- get_text_width("This is cool.", "Courier", 10)
  res3
  expect_equal(res2 < res3, TRUE)
  
  
  res4 <- get_text_width("This is cool.", "Courier", 10, units = "cm")
  res4
  expect_equal(res3 == cin(res4), TRUE)
  
  # get_text_width("This is cool and this is cooler and more an more.", 
  #                "Courier", 10, multiplier = .5)
  
})

test_that("utils17: has_bottom_footnotes works as expected.", {
  
  rpt <- create_report("") 
  
  expect_equal(has_bottom_footnotes(rpt), FALSE)
  
  rpt2 <- rpt %>% 
    footnotes("Hello")
  
  expect_equal(has_bottom_footnotes(rpt2), TRUE)
  
  rpt3 <- rpt %>% 
    footnotes("Hello", valign = "top")
  
  expect_equal(has_bottom_footnotes(rpt3), FALSE)
  
})

test_that("utils18: has_top_footnotes works as expected.", {
  
  rpt <- create_report("") 
  
  expect_equal(has_top_footnotes(rpt), FALSE)
  
  rpt2 <- rpt %>% 
    footnotes("Hello")
  
  expect_equal(has_top_footnotes(rpt2), FALSE)
  
  rpt3 <- rpt %>% 
    footnotes("Hello", valign = "top")
  
  expect_equal(has_top_footnotes(rpt3), TRUE)
  
})

test_that("utils19: get_spanning_info works as expected.", {
  
  
  fp <- ""
  
  dat <- mtcars[1:15, ]
  
  tbl <- create_table(dat, borders = "none") %>%
    spanning_header(mpg, disp, "Span 1") %>%
    spanning_header(hp, wt, "Span 2") %>%
    spanning_header(qsec, vs, "Span 3") %>%
    spanning_header(drat, gear, "Super span", level = 2)
  
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed",
                       font_size = 10, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(tbl) %>%
    footnotes("My footnote 1", "My footnote 2") %>%
    page_footer("Left1", "Center1", "Right1")
  
  rpt <- page_setup(rpt)
  
  cols <- rep(.6, ncol(dat))
  names(cols) <- names(dat)
  
  pi <- list(keys = names(dat), col_width = cols)
  w <- round(pi$col_width / rpt$char_width) - 1
  
  res <- get_spanning_info(rpt, tbl, pi, w)
  res
  
  expect_equal(length(res), 2)
  expect_equal(nrow(res[[1]]), 6)
  
})

test_that("utils20: get_lines_rtf works as expected.", {
  
  
  expect_equal(get_lines_rtf("Here There", 3, "Arial", 10), 1)
  expect_equal(get_lines_rtf("Here\nThere", 3, "Arial", 10), 2)
  
  expect_equal(get_lines_rtf("Here\nThere now here is something else\nanother split", 
                             2, "Arial", 12), 4)
  
  cnt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
                "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
                "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
                "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ", 
                "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
                "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
                "qui officia deserunt mollit anim id est laborum.")
  
  expect_equal(get_lines_rtf(cnt, 9, "Arial", 10), 3)
  
  

})


test_that("utils21: split_string_rtf() works as expected.", {
  
  cnt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
                "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
                "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
                "nisi ut aliquip ex ea commodo consequat.\n Duis aute irure dolor in ", 
                "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
                "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
                "qui officia deserunt mollit anim id est laborum.")
  
  pdf(NULL)
  par(family = "sans", ps = 10)
  
  res <- split_string_rtf(cnt, 4, "inches")
  res
  
  
  expect_equal(length(res$rtf), 1)
  expect_equal(res$line, 8)
  
  res2 <- split_string_rtf(NA, 4, "inches")
  res2
  
  expect_equal(length(res2$rtf), 1)
  expect_equal(res2$lines, 1)
  
  res3 <- split_string_rtf("hereisareallylongstringlongerthanoneinch", 
                           1, "inches")
  res3
  
  expect_equal(length(res3$rtf), 1)
  expect_equal(res3$lines, 1)
  
  
  res4 <- split_string_rtf("here is areallylongstringlongerthan one inch", 
                           1, "inches")
  res4
  
  expect_equal(length(res4$rtf), 1)
  expect_equal(res4$lines, 3)
  
  
  res5 <- split_string_rtf("here is areally long string longer than 3 cm", 
                           3, "cm")
  res5
  
  expect_equal(length(res4$rtf), 1)
  expect_equal(res4$lines, 3)
  
  res6 <- split_string_rtf("here \nis a\nreally \nlong string\n longer than 3 cm", 
                           3, "cm")
  res6
  
  expect_equal(length(res6$rtf), 1)
  expect_equal(res6$lines, 5)
  
  
  dev.off()
  

  
  
})

test_that("utils22: split_cells_variable() works as expected.", {
  
  dat <- data.frame(col1 = c("hello", "there", "here is a big long\nline to wrap and wrap"),
                    col2 = c(1, 2, 3),
                    col3 = c("my big string I want to wrap", "fork", "bork"),
                    ..row = c(NA, NA, NA), stringsAsFactors = FALSE)
  
  ts <- create_table(dat)
  
  res <- split_cells_variable(dat, c(col1 = 1, col2 = 1, col3 = 1), 
                              "Arial", 12, "inches", "RTF", ts = ts)
  res
  
  expect_equal(nrow(res$data), 3)
  expect_equal(ncol(res$data), 4)
  expect_equal(res$data$..row, c(3, 1, 4))
  expect_equal(length(res), 2)
  expect_equal(length(res$widths), 3)
  expect_equal(length(res$widths[[3]]$col1), 4)
  expect_equal(all(res$widths[[3]]$col1 < 1), TRUE)
  
})


test_that("utils23: get_image_html works as expected.", {
  
  
  tmp <-  tempfile(fileext = ".jpg")
  tmp2 <- tempfile(fileext = ".html")
  
  plt <- list(height = 4, width = 8)
  
  res <- get_image_html(tmp, tmp2, plt, "inches")
  res
  
  expect_equal(length(res), 1)
  expect_equal(substr(res, 1, 4), "<img")
  
})


test_that("utils24: get_outer_borders works as expected.", {
  
 
  expect_equal(is.null(get_outer_borders("none")), TRUE) 
  expect_equal(get_outer_borders("all"), c("top", "bottom", "left", "right")) 
  expect_equal(get_outer_borders(c("top", "bottom")), c("top", "bottom")) 
  expect_equal(get_outer_borders(c("outside", "inside")), 
               c("top", "bottom", "left", "right")) 
  
})

test_that("utils25: split_strings() works as expected.", {
  
  pdf(NULL)
  par(family = "sans", ps = 10)
  
  res <- split_strings("Here is a big long string that will not fit in one inch",
                       1, "inches")
  
  dev.off()
  
  res
  
  expect_equal(length(res$text), 4)
  expect_equal(length(res$widths), 4)
  
})


test_that("utils26: split_string_text() works as expected.", {
  
  pdf(NULL)
  par(family = "sans", ps = 10)
  
  res <- split_string_text("Here is a big long string that will not fit in one inch",
                       1, "inches")

  
  dev.off()
  
  res
  
  expect_equal(length(res$text), 4)
  expect_equal(res$lines, 4)
  expect_equal(length(res$widths), 4)
  
})

test_that("utils27: get_points_* functions work as expected.", {
  
  w <- c(1, .8, .5)
  
  
  res1 <- get_points_left(1, 4, w, "inches")
  res1
  
  expect_equal(res1, c(75, 75, 75))
  
  res2 <- get_points_right(1, 2, w, "inches")
  res2
  
  expect_equal(res2, c(72, 86.4, 108))
  
  res3 <- get_points_center(1, 2, w, "inches")
  res3
  
  expect_equal(res3, c(73, 80.2, 91))
})

test_that("utils28: apply_widths works as expected.", {
  
  df1 <- data.frame(col1 = c("here is a something", "fork", "bork"),
                    col2 = c("", 3, 43),
                    col3 = c("", "B", "C"), stringsAsFactors = FALSE,
                    "..blank" = c("L", "", ""))
  
  wdth <- c(col1 = 4, col2 = 3, col3 = 5)
  jst <- c(col1 = "left", col2 = "right", col3 = "center")
  
  res1 <- apply_widths(df1, wdth, jst)
  
  res1$col1
  res1$col2 
  res1$col3
  
  expect_equal(res1$col1[2], "fork")
  expect_equal(nchar(res1$col2[1]), 0)
  expect_equal(nchar(res1$col3[1]), 0)             

})

test_that("utils29: invisible functions work as expected.", {
  
  lst <- c("A", "B", "..blank", "..x.C")
  
  expect_equal(has_invisible(lst), TRUE)
  expect_equal(is.invisible("..x.C"), TRUE)
  expect_equal(is.invisible("..blank"), FALSE)
  expect_equal(is.invisiblev(lst), c(A = FALSE, B = FALSE, 
                                     ..blank = FALSE, ..x.C = TRUE))
  
  expect_equal(translate_invisible("C", lst), "..x.C")
  expect_equal(translate_invisible("D", lst), "D")
  
})

test_that("utils30: gluev() works as expected.", {
  
  library(common)
  
  v1 <- c("here{supsc('1')}", "there{subsc('a')}")
  
  
  res <- gluev(v1)
  
  res 
  
  expect_equal(res, c("here\U00B9", "there\U2090"))
  
  res2 <- gluev(NULL)
  
  expect_equal(is.null(res2), TRUE)
})

test_that("utils30: has_glue() works as expected.", {
  
  res1 <- has_glue()
  
  expect_equal(res1, TRUE)
  
  options("reporter.glue" = FALSE)
  
  res2 <- has_glue()

  expect_equal(res2, FALSE)
  
  options("reporter.glue" = NULL)
  
  
  res3 <- has_glue()
  
  expect_equal(res3, TRUE)
  
})

test_that("utils31: strwdth() works as expected.", {
  
  strngs <- c("hello", "goodbye", paste0("here is something", supsc('6')))
  
  
  
  # pdf(NULL)
  # par(family = 'mono', ps = 10)
  # 
  # res <- strwidth(strngs, un = "inches")
  # 
  # dev.off()
  
  res <- c(0.4166667, 0.5833333, 1.5000000)
  
  
  
  pdf(NULL)
  par(family = 'mono', ps = 10)
  
  res1 <- strwdth(strngs, un = "inches")
  
  dev.off()
  
  res1
  
  expect_equal(res[1] == round(res1[1], 7), TRUE) 
  expect_equal(res[2] == round(res1[2], 7), TRUE) 
   
  
  # pdf(NULL)
  # par(family = 'sans', ps = 10)
  # 
  # res3 <- strwidth(strngs, un = "inches")
  # 
  # dev.off()
  
  res3 <- c(0.2933333, 0.5272222, 1.1733333)
  
  pdf(NULL)
  par(family = 'sans', ps = 10)
  
  res4 <- strwdth(strngs, un = "inches")
  
  dev.off()
  
  res4
  
  # print(res1)
  # print(res2)
  
  expect_equal(all(res != res4), TRUE) 
  expect_equal(res3[1] == round(res4[1], 7), TRUE) 
  expect_equal(res3[2] == round(res4[2], 7), TRUE) 
  
})

test_that("utils32: get_first_pos works as expected.", {
  
  arm <- c(rep("A",3), rep("B",5), rep("C",1), rep("D",3))
  first_arm <- get_first_pos(arm)
  
  expect_equal(first_arm, c(T,F,F,T,F,F,F,F,T,T,F,F)) 
})

test_that("utils33: get_last_pos works as expected.", {
  
  arm <- c(rep("A",3), rep("B",5), rep("C",1), rep("D",3))
  last_arm <- get_last_pos(arm)
  
  expect_equal(last_arm, c(F,F,T,F,F,F,F,T,T,F,F,T)) 
})

test_that("utils34: add_group_border_ind works as expected.", {
  
  arm <- c(rep("A",3), rep("B",5), rep("C",1), rep("D",3))
  df <- data.frame("arm" = arm)
  df <- add_group_border_ind(df, "arm")
  
  df_expect <- data.frame("arm" = arm,
                          "..group_border" = c("", "", "bottom", "","","","",
                                               "bottom","bottom","","","bottom"))
  expect_equal(df, df_expect) 
})

test_that("utils35: get_group_count works as expected.", {
  
  group <- c("A", "A", "A", "B", "B", "C", "C", "C")
  count <- c(1, 1, 1, 2, 1, 3, 0, 2)
  
  group_counts <- get_group_count(group, count)
  expect_equal(group_counts, c(3, 3, 3, 3, 3, 5, 5, 5))
  
  group_counts2 <- get_group_count(group)
  expect_equal(group_counts2, c(3, 3, 3, 2, 2, 3, 3, 3))
})

# Table Spec Functions ---------------------------------------------------------
#' @title Create a table
#' @description 
#' The \code{create_table} function creates a table object to which 
#' further specifications can be added.  The object can be added to a report
#' using the \code{\link{add_content}} function. The object is implemented as an 
#' S3 object of class 'table_spec'.
#' @details 
#' A table object is a container to hold information about a table.  The 
#' only required information for a table is the table data.  All other 
#' parameters and functions are optional.
#' 
#' By default, the table will display all columns in the data frame.  To change
#' this default, use the \code{show_cols} parameter.  Setting this parameter
#' to 'none' will display none of the columns in the data, unless they are
#' explicitly defined with a \code{\link{define}} function.  
#' 
#' The \code{show_cols} parameter also accepts a vector of column positions
#' or column names. When a vector is supplied, \code{create_table} will 
#' display only those columns on the report, in the order encountered in the 
#' vector.  The \code{show_cols} parameter is the only mechanism in 
#' \code{create_table} to modify the column order. Otherwise, modify the 
#' order prior to sending the data to \code{create_table} using the many options
#' available in Base R or supplemental packages.
#' 
#' @section Setting Formatting Attributes:
#' Formatting attributes can be controlled in three ways.  By default, formatting
#' attributes assigned to the data frame will be passed through to the 
#' reporting functions.  The reporting functions will recognize the 'label',
#' 'format', 'width', and 'justify' attributes. In other words, you can control 
#' the column label, width, format, and alignment of your report columns simply by 
#' assigning those attributes to your data frame. The advantage of using
#' attributes assigned to data frame columns is that you can store those 
#' attributes permanently with the data frame, and those attributes will
#' not have to be re-specified for each report.  To ignore attributes assigned
#' to the data frame, set the \code{use_attributes} parameter to 'none'.
#' 
#' Secondly, attributes can be specified using the \code{\link{column_defaults}}
#' function.  This function allows the user to apply a default set of parameters
#' to one or more columns.  If no columns are specified in the \code{var} 
#' or \code{from} and \code{to} parameter of this function, the defaults 
#' will apply to all columns.  Any default parameter value can be overridden 
#' by the \code{\link{define}} function.
#' 
#' Lastly, the \code{\link{define}} function provides the most control over 
#' column parameters.  This function provides a significant amount of 
#' functionality that cannot be specified elsewhere.  See the 
#' \code{\link{define}} function for additional information.  The \code{define}
#' function will also override any formatting attributes assigned to the 
#' data frame, or anything set by the \code{column_defaults} function.
#' 
#' @section Additional Functionality:
#' The \code{create_table} function also provides the capabilities to create
#' a "headerless" table.  A headerless table is useful when combining two tables 
#' into one report.  The example below illustrates use of a headerless table.
#' 
#' Since the purpose of the \strong{reporter} package is to create statistical 
#' reports, the \code{create_table} function makes it easy to add population
#' counts to the table header.  These population counts are added to column
#' labels and spanning header labels using the \code{n} parameter on the 
#' \code{\link{define}} or \code{\link{spanning_header}} functions.  The 
#' population count is formatted according to the  
#' \code{n_format} parameter on \code{create_table}. The \strong{reporter} 
#' package provides four population count formatting functions.  
#' You may create your own formatting function 
#' if one of these functions does not meet your needs.  See 
#' \code{\link{upcase_parens}} for further details.
#' 
#' @param x The data frame or tibble from which to create the table object.
#' @param show_cols This parameter gives control over which columns in the 
#' input data to display on the report by default.  Valid values are
#' 'all', 'none', a vector of quoted column names, or a vector of 
#' column positions.  'all' means show all columns, 
#' unless overridden by the column definitions.  
#' 'none' means don't show any 
#' columns unless specified in the column definitions.  If a vector of column
#' names or positions is supplied, those columns will be shown in the report 
#' in the order specified, whether or not a definition is supplied. See the 
#' \code{\link{define}} function for additional information on how to
#' show/hide report columns.
#' @param use_attributes Whether or not to use any formatting attributes assigned
#' to the columns on the input data frame.  Valid values are 'all', 'none', or
#' a vector of attribute names to use.  Possible attributes that may be used
#' are 'label', 'format', 'width', and 'justify'.  By default, any of these
#' attribute values will be applied to the table.  For example, if you assign
#' a label to the 'label' attribute of a data frame column, pass that data 
#' frame into \code{create_table}, and don't override the label value on a 
#' \code{define} function, the label will appear as a column header on the
#' table.  The \code{use_attributes} parameter allows you to control this default
#' behavior, and use or ignore data frame attributes as desired.  
#' @param width The expected width of the table in the report units of 
#' measure.  By default, the width setting is NULL, and columns will be sized
#' according to the width of the data and labels.  If the width parameter is 
#' set, the function will attempt to size the table to the specified width.
#' If the sum of the column widths is less than the specified width, the 
#' function will adjust the columns widths proportionally to fit the specified
#' width.  If the sum of the column widths is wider than the table width 
#' parameter value, the table width parameter will be ignored. 
#' @param first_row_blank Whether to place a blank row under the table header.
#' Valid values are TRUE or FALSE.  Default is FALSE.
#' @param n_format The formatting function to apply to the header "N=" label. 
#' The default formatting function is \code{\link{upcase_parens}}. 
#' @param headerless Whether to create a headerless table. A headerless
#' table displays the table data only. Default is FALSE, meaning the table
#' will have a header. 
#' @param borders Whether and where to place a border. Valid values are 'top',
#' 'bottom', 'left', 'right', 'all', 'none', 'outside', 'inside', and 'body'.  
#' Default is 'none'.  The 'left', 'right', 'outside', 'inside', and 'body'
#' border specifications only apply to RTF, HTML, PDF, and DOCX reports.  
#' The 'body' border 
#' specification means put borders around only the body of the table.
#' @param header_bold Whether or not the column headers on the tables should
#' be bolded.  Valid values are TRUE and FALSE.  The default is FALSE.
#' @param continuous If a table crosses multiple pages, it is normally
#' broken into a separate table for each page, and the titles and footnotes
#' are repeated on each page. When the "continuous" parameter is TRUE, the
#' table will instead be a single table, and the titles and footnotes will
#' not be repeated on each page.  This parameter currently only works for RTF
#' outputs.  Also, this parameter only works for titles and footnotes that
#' are attached to the table body.  Titles and footnotes attached to the 
#' report will still be shown on every page. 
#' @family table
#' @seealso \code{\link{create_report}} to create a report, 
#' \code{\link{create_plot}} to create a plot,
#' \code{\link{create_text}} to create text content, and 
#' \code{\link{add_content}} to append content to a report.  Also see
#' the \code{\link{titles}}, \code{\link{footnotes}}, and \code{\link{page_by}}
#' functions to add those items to the table if desired.
#' @examples 
#' library(reporter)
#' library(magrittr)
#' 
#' # Create temp file path
#' tmp <- file.path(tempdir(), "mtcars.txt")
#' 
#' #Subset cars data
#' dat <- mtcars[1:10, 1:7]
#' 
#' # Calculate means for all columns
#' dat_sum <- data.frame(all_cars = "All cars average", as.list(sapply(dat, mean)), 
#'                       stringsAsFactors = FALSE)
#' 
#' # Get vehicle names into first column
#' dat_mod <- data.frame(vehicle = rownames(dat), dat, stringsAsFactors = FALSE)
#'                       
#' # Create table for averages
#' tbl1 <- create_table(dat_sum) %>% 
#'         titles("Table 1.0", "MTCARS Sample Data") %>% 
#'         column_defaults(width = .5) %>% 
#'         define(all_cars, label = "", width = 2) %>% 
#'         define(mpg, format = "%.1f") %>% 
#'         define(disp, format = "%.1f") %>% 
#'         define(hp, format = "%.0f") %>% 
#'         define(qsec, format = "%.2f")
#' 
#' # Create table for modified data
#' tbl2 <- create_table(dat_mod, headerless = TRUE) %>% 
#'         column_defaults(width = .5) %>% 
#'         define(vehicle, width = 2) 
#' 
#' # Create the report object
#' rpt <- create_report(tmp) %>% 
#'   add_content(tbl1, align = "left", page_break = FALSE) %>% 
#'   add_content(tbl2, align = "left") 
#' 
#' # Write the report to the file system
#' write_report(rpt)
#' 
#' # Write report to console
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' #                                 Table 1.0
#' #                             MTCARS Sample Data
#' # 
#' #                             mpg    cyl   disp     hp   drat     wt   qsec
#' # -------------------------------------------------------------------------
#' # All cars average           20.4    5.8  208.6    123  3.538  3.128  18.58
#' # 
#' # Mazda RX4                    21      6    160    110    3.9   2.62  16.46
#' # Mazda RX4 Wag                21      6    160    110    3.9  2.875  17.02
#' # Datsun 710                 22.8      4    108     93   3.85   2.32  18.61
#' # Hornet 4 Drive             21.4      6    258    110   3.08  3.215  19.44
#' # Hornet Sportabout          18.7      8    360    175   3.15   3.44  17.02
#' # Valiant                    18.1      6    225    105   2.76   3.46  20.22
#' # Duster 360                 14.3      8    360    245   3.21   3.57  15.84
#' # Merc 240D                  24.4      4  146.7     62   3.69   3.19     20
#' # Merc 230                   22.8      4  140.8     95   3.92   3.15   22.9
#' # Merc 280                   19.2      6  167.6    123   3.92   3.44   18.3
#' # 
#' @export
create_table <- function(x, show_cols = "all", use_attributes = "all",
                         width = NULL, 
                         first_row_blank=FALSE,
                         n_format = upcase_parens, headerless = FALSE,
                         borders = "none", header_bold = FALSE, 
                         continuous = FALSE) {
  if (is.null(x)) {
    stop("Data parameter 'x' missing or invalid.") 
    
  }
  
  if (!"data.frame" %in% class(x)) {
    stop(paste("ERROR: data parameter 'x' on",
               "page_template() function is invalid.",
               "\n\tValid values are a data.frame or tibble."))
  }
  
  if (is.null(use_attributes))
    stop("use_attributes parameter cannot be null.")
  else if (any(use_attributes %in%  
               c("all", "none", "label", "width", "justify", "format") == FALSE)){
    
    stop(paste("Invalid use_attributes value.  Valid values are 'all', 'none'", 
               "or a vector of any of the following attributes names: 'label',",
               "'format', 'width', or 'justify'"))
  }
  
  if (!all(borders %in% c("top", "bottom", "left", "right", 
                          "all", "none", "outside", "inside", "body")))
    stop(paste("Borders parameter invalid.  Valid values are", 
               "'top', 'bottom', 'left', 'right', 'all', ",
               "'none', 'outside', 'inside', or 'body'."))
    
  
  ret <- structure(list(), class = c("table_spec", "list"))

  ret$data <- x
  ret$dataname <- paste0(deparse(substitute(x, env = environment())))
  ret$n_format <- n_format
  ret$col_defs <- list()
  ret$col_spans <- list()
  if (is.integer(show_cols)) {
    ret$show_cols <- names(x)[show_cols]
  } else if (is.null(show_cols)) {
    ret$show_cols <- "all"
  } else {
    ret$show_cols <- show_cols
  }
  ret$first_row_blank <- first_row_blank
  ret$headerless <- headerless
  ret$stub <- NULL
  ret$width <- width
  ret$page_var <- NULL
  ret$borders <- borders
  ret$header_bold <- header_bold
  ret$continuous <- continuous
  if (any(use_attributes == "all"))
    ret$use_attributes <- c("label", "width", "justify", "format")
  else if (all(use_attributes == "none"))
    ret$use_attributes <- c("")
  else  
    ret$use_attributes <- use_attributes
  
  # Apply any titles, footnotes, or spans
  # attached to the data frame itself
  ret <- apply_attributes(ret, x)

  return(ret)

}

#' @title Defines a column 
#' @description A function to define a table column.  The
#' \code{define} function contains a variety of a parameters to control the 
#' appearance of the report.  Using the \code{define} function, you can control
#' simple options like column alignment and width, but also control more 
#' sophisticated options like page wrapping and page breaking.
#' @details 
#' Column definitions are optional.  By default, all columns in the data
#' are displayed in the order assigned to the data frame. 
#' 
#' The report will use attributes assigned to the data frame 
#' such as 'width', 'justify', 'label', and 'format'.  In other words, 
#' some control over the column 
#' formatting is available by manipulating the data frame attributes prior
#' to assigning the data frame to \code{create_table}.  See 
#' \code{\link{create_table}} for more details.
#' 
#' The \code{define} function is used to provide additional control over
#' column appearance.  For example, you may use the \code{define} function
#' to assign an "N=" population count, eliminate duplicates from the column,
#' or place a blank row after each unique value of the variable. 
#' See the parameter documentation for additional options.
#' 
#' Some of the parameters on the \code{define} function are used in the 
#' creation of a table stub.  Specifically, the \code{label_row} and 
#' \code{indent} parameters participate in the creation of the stub column.
#' See the \code{\link{stub}} function for further
#' information.
#' 
#' A single column definition may be defined for multiple variables.  
#' To create a definition for multiple variables, pass the variables as
#' a quoted or unquoted vector.  When creating a single definition for 
#' multiple variables, the parameters will be unified across those variables.
#' Note that some parameters (such as \code{page_break}) may only be set
#' once per report, and cannot be shared across multiple variables.  
#' 
#' @param x The table spec.
#' @param vars The variable name or names to define a column for.  Names may
#' be quoted or unquoted.  If defining for multiple variables, 
#' pass them as a vector of names.  If you want to pass an R variable of names,
#' set the \code{standard_eval} parameter to TRUE .
#' The \code{standard_eval} parameter is useful when writing functions that construct
#' reports dynamically. 
#' @param label The label to use for the column header.  If a label is assigned
#' to the label column attribute, it will be used as a default.  Otherwise,
#' the column name will be used.
#' @param format The format to use for the column data.  The format can 
#' be a string format, a formatting function, a lookup list, a user-defined
#' format, or a formatting list. 
#' All formatting is performed by the \code{\link[fmtr]{fapply}} function from
#' the \code{\link[fmtr]{fmtr}} package.  For 
#' a list of common formatting codes, see \link[fmtr]{FormattingStrings}.
#' @param align The column alignment.  Valid values are "left", "right", 
#' "center", and "centre".  By default, text columns will be left aligned
#' and numeric columns will be right aligned.
#' @param label_align How to align the header labels for this column.
#' Valid values are "left", "right", "center", and "centre".  By default, 
#' the label alignment will follow any alignment set on the column \code{align}
#' parameter.
#' @param width The width of the column in the specified units of measure.
#' The units of measure are specified on the \code{units} parameter of the
#' \code{\link{create_report}} function.  If no width is supplied, the
#' \code{\link{write_report}} function will assign a default width based on the 
#' width of the column data and the label.  \code{write_report} will not set a 
#' column width less than the width of the largest word in the data or label.
#' In other words, \code{write_report} will not break words. 
#' @param visible Whether or not the column should be visible on the report.
#' This parameter can be used as a simple way to drop columns from the report.
#' @param n The n value to place in the "N=" header label.  Formatting for
#' the n value will be performed by the formatting function assigned to the 
#' \code{n_format} parameter on \code{\link{create_table}}.
#' @param blank_after Whether to place a blank row after unique values of this
#' variable.  Valid values are TRUE or FALSE.  Default is FALSE.
#' @param blank_before Whether to place a blank row before unique values of this
#' variable.  Valid values are TRUE or FALSE.  Default is FALSE.
#' @param dedupe Whether to dedupe the values for this variable.  Variables
#' that are deduped only show the value on the first row of each group.  This 
#' option is commonly used for grouping variables.
#' @param id_var Whether this variable should be considered an ID variable.
#' ID variables are retained on each page when the page is wrapped. ID variables
#' are also moved to the far left of the page.
#' @param page_wrap Force a page wrap on this variable.  A page wrap is a vertical
#' page break necessary when the table is too wide to fit on a single page.
#' The excess variables will be wrapped to the next page.  Page wraps will
#' continue until all columns are displayed.  Use the \code{id_var}
#' parameter to identify rows across wrapped pages. 
#' @param page_break You may control when page breaks occur by defining
#' a page break variable yourself, and setting this parameter to TRUE for
#' that variable.  Only one page break variable can be defined per table.
#' If two or more variables are defined as a page break, an error will be 
#' generated.
#' @param indent How much to indent the column values.  The parameter takes a 
#' numeric value that will be interpreted according to the \code{units} 
#' (Unit Of Measure) setting on the report.  This parameter can be used to 
#' help create a stub column.  The default value is NULL, meaning the column
#' should not be indented.  See the \code{\link{stub}} function for additional
#' information on creating a stub column.
#' @param label_row Whether the values of the variable should be used to
#' create a label row.  Valid values are TRUE or FALSE.  Default is FALSE.
#' If \code{label_row} is set to TRUE, the dedupe parameter will also be 
#' set to TRUE.  This parameter is often used in conjunction with the 
#' \code{\link{stub}} function and \code{indent} parameter to create a 
#' stub column.
#' @param standard_eval A TRUE or FALSE value indicating whether to 
#' use standard evaluation on the \code{vars} parameter value.  Default is
#' FALSE.  Set this parameter to TRUE if you want to pass the \code{vars}
#' value(s) using a variable.
#' @param style A \code{\link{cell_style}} object that defines the desired
#' style for this column.  The cell style object can be used to define 
#' conditional styling.
#' @param group_border Adds a horizontal border at the end of a group. 
#' Valid values are TRUE or FALSE. Default is FALSE. The group border is
#' another way to separate analysis groups, and can be used in addition
#' to or instead of the "blank_before" and "blank_after" parameters.
#' @param group_cohesion Default is FALSE. If TRUE, it would avoid splitting the 
#' same group across multiple pages with default strength 0.2. Users can increase
#' the strength to make groups on the same page more likely. The accepted value
#' is 0 to 1.
#' @param break_label Decide whether to repeat the grouping value when the group 
#' spans multiple pages. If set it as TRUE, when a group spans, the break label will 
#' be displayed as a label of the first row. If set it as a text such as "(Continued)",
#' it would be used as a suffix text. E.g., Investigation (Continued).
#' @return The modified table spec.
#' @family table
#' @examples
#' library(reporter)
#' library(magrittr)
#'  
#' # Create temp file name
#' tmp <- file.path(tempdir(), "mtcars.txt")
#' 
#' # Prepare data
#' dat <- mtcars[1:10, ]
#' dat <- data.frame(vehicle = rownames(dat), dat, stringsAsFactors = FALSE)
#' 
#' # Define table
#' tbl <- create_table(dat, show_cols = 1:8) %>% 
#'   define(vehicle, label = "Vehicle", width = 3, id_var = TRUE, align = "left") %>% 
#'   define(mpg, label = "Miles per Gallon", width = 1) %>% 
#'   define(cyl, label = "Cylinders", format = "%.1f") %>% 
#'   define(disp, label = "Displacement") %>% 
#'   define(hp, label = "Horsepower", page_wrap = TRUE) %>% 
#'   define(drat, visible = FALSE) %>% 
#'   define(wt, label = "Weight") %>% 
#'   define(qsec, label = "Quarter Mile Time", width = 1.5) 
#' 
#' 
#' # Create the report
#' rpt <- create_report(tmp, orientation = "portrait") %>% 
#'   titles("Listing 2.0", "MTCARS Data Listing with Page Wrap") %>% 
#'   add_content(tbl, align = "left") %>% 
#'   page_footer(right = "Page [pg] of [tpg]")
#' 
#' # Write the report
#' write_report(rpt)
#' 
#' # Send report to console for viewing
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' #                                  Listing 2.0
#' #                       MTCARS Data Listing with Page Wrap
#' # 
#' #                                         Miles per
#' # Vehicle                                    Gallon Cylinders Displacement
#' # ------------------------------------------------------------------------
#' # Mazda RX4                                      21       6.0          160
#' # Mazda RX4 Wag                                  21       6.0          160
#' # Datsun 710                                   22.8       4.0          108
#' # Hornet 4 Drive                               21.4       6.0          258
#' # Hornet Sportabout                            18.7       8.0          360
#' # Valiant                                      18.1       6.0          225
#' # Duster 360                                   14.3       8.0          360
#' # Merc 240D                                    24.4       4.0        146.7
#' # Merc 230                                     22.8       4.0        140.8
#' # Merc 280                                     19.2       6.0        167.6
#' # 
#' # ...
#' # 
#' #                                                                    Page 1 of 2
#' #                                  Listing 2.0
#' #                       MTCARS Data Listing with Page Wrap
#' # 
#' # Vehicle                              Horsepower Weight  Quarter Mile Time
#' # -------------------------------------------------------------------------
#' # Mazda RX4                                   110   2.62              16.46
#' # Mazda RX4 Wag                               110  2.875              17.02
#' # Datsun 710                                   93   2.32              18.61
#' # Hornet 4 Drive                              110  3.215              19.44
#' # Hornet Sportabout                           175   3.44              17.02
#' # Valiant                                     105   3.46              20.22
#' # Duster 360                                  245   3.57              15.84
#' # Merc 240D                                    62   3.19                 20
#' # Merc 230                                     95   3.15               22.9
#' # Merc 280                                    123   3.44               18.3
#' # 
#' # ...
#' # 
#' #                                                                    Page 2 of 2
#' @export
define <- function(x, vars, label = NULL, format = NULL, 
                   align=NULL, label_align=NULL, width=NULL,
                   visible=TRUE, n = NULL, blank_after=FALSE, blank_before=FALSE,
                   dedupe=FALSE, id_var = FALSE, page_wrap = FALSE,
                   page_break = FALSE, indent = NULL, label_row = FALSE,
                   standard_eval = FALSE, style = NULL, group_border = FALSE,
                   group_cohesion = FALSE, break_label = NULL) {
  
  if (standard_eval) {
    if (!typeof(vars) %in% c("character", "numeric")) {
      
      stop("Type of vars parameter must be character or numeric when standard_eval is TRUE.")
    }
    
    vars_c <- vars
    
  } else {
    
    # Determine if it is a vector or not.  "language" is a vector.
    if (typeof(substitute(vars, env = environment())) == "language") 
      v <- substitute(vars, env = environment())
    else 
      v <- substitute(list(vars), env = environment())
  
    # Turn each item into a character
    vars_c <- c()
    if (length(v) > 1) {
      for (i in 2:length(v)) {
        vars_c[[length(vars_c) + 1]] <- as.character(v[[i]]) 
      }
      
    }
    
    # Convert list to vector
    vars_c <- unlist(vars_c)
    
  }
  

  
  # Deal with curly brace escape
  if (length(vars_c) > 0) {
    if (vars_c[[1]] == "{") { 
      vars_c <- get(vars_c[[2]], envir = parent.frame())
    }
  }
  
  # Check that variable exists in data frame
  if (!is.null(x$data) & !is.null(vars_c)) {
    if (!any(vars_c %in% names(x$data))) {
      for (nm in vars_c) {
        if (!nm %in% names(x$data)) 
          stop(paste0("Variable does not exist in data: ", nm))
      }
    }
  }
  
  # Parse group cohesion into min page proportion
  # If gorup_cohesion is a fraction, assign it to min_page_prop for internal using
  min_page_prop <- NULL
  if (group_cohesion != FALSE) {
    if (group_cohesion == TRUE & !is.numeric(group_cohesion)) {
      
      min_page_prop <- 0.8
      
    } else if (is.numeric(group_cohesion)) {
      
      if (group_cohesion > 1 | group_cohesion < 0) {
        stop(paste0("`group_cohesion` should be TRUE/FALSE or numeric value ",
                    "from 0 to 1 instead of ", group_cohesion, "."))
      } else {
        min_page_prop <- 1 - group_cohesion
        group_cohesion <-  TRUE
      }
      
    } else {
      stop(paste0("`group_cohesion` should be TRUE/FALSE or numeric value ",
                  "from 0 to 1 instead of '", group_cohesion, "'."))
    }
  }
  
  # Process break_label
  if (!is.null(break_label)) {
    if (break_label == TRUE) {
      break_label <- ""
    } else if (break_label == FALSE) {
      break_label <- NULL
    } else if (!is.character(break_label)) {
      stop("`break_label` should be character value or TRUE/FALSE.")
    }
  }
  
  # For each passed variable, create an individual definition
  # This make subsequent processing much easier
  for (nm in vars_c) {
    
    if (has_glue()) {
      lbl <- gluev(label) 
    } else {
      
      lbl <- label
    }
    
    
   def <- define_c(nm, label = lbl, format = format, 
                   align=align, label_align=label_align, width=width,
                   visible=visible, n = n, blank_after=blank_after,
                   blank_before = blank_before,
                   dedupe=dedupe, id_var = id_var, page_wrap = page_wrap,
                   page_break = page_break, indent = indent, 
                   label_row = label_row, style = style, group_border = group_border,
                   group_cohesion = group_cohesion, min_page_prop = min_page_prop,
                   break_label = break_label)
   
   x$col_defs[[nm]] <- def
   
   if (page_break == TRUE) {
     if (is.null(x$page_var)) {
       x$page_var <- nm
     } else
       stop("Cannot define more than one page break variable")
   }
    
  }

  return(x)
}

#' @description Define a variable with a quoted name. Used internally.
#' @noRd
define_c <- function(var, label = NULL, format = NULL, 
                   align=NULL, label_align=NULL, width=NULL,
                   visible=TRUE, n = NULL, blank_after=FALSE,
                   blank_before = FALSE,
                   dedupe=FALSE, id_var = FALSE, page_wrap = FALSE,
                   page_break = FALSE, indent = NULL, label_row = FALSE,
                   style = NULL, group_border = FALSE,
                   group_cohesion = FALSE, min_page_prop = NULL,
                   break_label = NULL) {
  
  
  if (!typeof(var) %in% c("character", "numeric"))
    stop("class of var must be character.")
 
  def <- structure(list(), class = c("col_def", "list"))
  
  def$var <- var
  def$var_c <- var
  if (visible == TRUE) {
    def$label <- label
    def$format <- format
    def$align <- align
    def$label_align <- if (is.null(label_align) & !is.null(align))
      align else label_align
    def$width <- width
  }
  def$visible <- visible
  def$n <- n
  def$blank_after <- blank_after
  def$blank_before <- blank_before
  def$dedupe <- dedupe
  def$id_var <- id_var
  def$page_wrap <- page_wrap
  def$indent <- indent
  def$label_row <- label_row
  def$page_break <- page_break
  if (label_row == TRUE)
    def$dedupe <- TRUE
  def$style <- style
  def$group_border <- group_border
  def$group_cohesion <- group_cohesion
  def$min_page_prop <- min_page_prop
  def$break_label <- break_label
  
  return(def)
}


#' @title Set default attributes for one or more columns
#' @description A function to set default attributes for columns on a table.  
#' The \code{column_defaults} function contains a subset of the parameters
#' on the \code{\link{define}} function that can be shared across variables.
#' Any attributes set by \code{column_defaults} can be overridden by 
#' the \code{define} function.  The overall purpose of the 
#' function is to minimize redundancy in column definitions. 
#' @details 
#' Column defaults can be specified for multiple variables.  By default,
#' the function will apply to all variables.  Alternately, you can 
#' specify a vector of columns on the \code{vars} parameter, or a range of
#' columns using the \code{from} and \code{to} parameters.  Both the 
#' \code{vars} parameters and the \code{from} and \code{to} parameters
#' will accept column positions, quoted variable names, or unquoted variable 
#' names.
#' 
#' The parameters that can be set with the \code{column_defaults} 
#' include the formatting attributes 'width', 'justify', 'label', and 
#' 'format'.  Any parameters set with \code{column_defaults} will override
#' any attributes set on the data frame.
#' 
#' Note that you may call the \code{column_defaults} function multiple times
#' on the same table specification.  Typically, multiple \code{column_defaults}
#' calls would be made with a different set or range of variables.
#' 
#' @param x A table spec.
#' @param vars The variable name or names to define defaults for.  Variable
#' names may be quoted or unquoted.  The parameter will also accept 
#' integer column positions instead of names.  For multiple variables, 
#' pass the names or positions as a vector. If you want to pass an R variable 
#' of names,
#' set the \code{standard_eval} parameter to TRUE.
#' The \code{standard_eval} parameter is useful when writing functions that construct
#' reports dynamically. 
#' @param from The variable name or position that starts a column range.  
#' If passed as a variable name, it may be quoted or unquoted.
#' @param to The variable name or position that ends a column range. 
#' If passed as a variable name, it 
#' may be quoted or unquoted.
#' @param label The label to use for a column header.  This label will be 
#' applied to all variables assigned to the \code{column_defaults} function.
#' @param format The format to use for the column data.  The format can 
#' be a string format, a formatting function, a lookup list, a user-defined
#' format, or a formatting list. 
#' All formatting is performed by the \code{\link[fmtr]{fmtr}} package.  For 
#' additional information, see the help for that package.
#' @param align The column alignment.  Valid values are "left", "right", 
#' "center", and "centre".
#' @param label_align How to align the header labels for this column.
#' Valid values are "left", "right", "center", and "centre".
#' @param width The width of the column in the specified units of measure.
#' The units of measure are specified on the \code{units} parameter of the
#' \code{\link{create_report}} function.  If no width is supplied, the
#' \code{\link{write_report}} function will assign a default width based on the 
#' width of the column data and the label.  \code{write_report} will not set a 
#' column width less than the width of the largest word in the data or label.
#' In other words, \code{write_report} will not break words. 
#' @param n The n value to place in the "N=" header label.  Formatting for
#' the n value will be performed by the formatting function assigned to the 
#' \code{n_format} parameter on \code{\link{create_table}}.
#' @param standard_eval A TRUE or FALSE value that indicates whether to
#' use standard or non-standard evaluation of the \code{vars}, \code{from},
#' and \code{to} parameters.  Set \code{standard_eval} to TRUE if you want
#' to pass the column names as variables. Default is FALSE, meaning it
#' will use non-standard (unquoted) evaluation.
#' @param style A \code{\link{cell_style}} object that defines a style 
#' for all columns associated with the column defaults.  
#' @return The modified table spec.
#' @family table
#' @examples
#' library(reporter)
#' library(magrittr)
#'  
#' # Create temp file name
#' tmp <- file.path(tempdir(), "mtcars.txt")
#' 
#' # Prepare data
#' dat <- mtcars[1:10, ]
#' dat <- data.frame(vehicle = rownames(dat), dat, stringsAsFactors = FALSE)
#' 
#' # Define table
#' tbl <- create_table(dat, show_cols = 1:8) %>% 
#'   column_defaults(from = mpg, to = qsec, width = .5, format = "%.1f") %>% 
#'   define(vehicle, label = "Vehicle", width = 1.5, align = "left") %>% 
#'   define(c(cyl, hp), format = "%.0f") 
#' 
#' # Create the report
#' rpt <- create_report(tmp, orientation = "portrait") %>% 
#'   titles("Table 2.5", "MTCARS Sample Report") %>% 
#'   add_content(tbl) 
#' 
#' # Write the report
#' write_report(rpt)
#' 
#' # Send report to console for viewing
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' #                                 Table 2.5
#' #                            MTCARS Sample Report
#' # 
#' #    Vehicle               mpg    cyl   disp     hp   drat     wt   qsec
#' #    -------------------------------------------------------------------
#' #    Mazda RX4            21.0      6  160.0    110    3.9    2.6   16.5
#' #    Mazda RX4 Wag        21.0      6  160.0    110    3.9    2.9   17.0
#' #    Datsun 710           22.8      4  108.0     93    3.8    2.3   18.6
#' #    Hornet 4 Drive       21.4      6  258.0    110    3.1    3.2   19.4
#' #    Hornet Sportabout    18.7      8  360.0    175    3.1    3.4   17.0
#' #    Valiant              18.1      6  225.0    105    2.8    3.5   20.2
#' #    Duster 360           14.3      8  360.0    245    3.2    3.6   15.8
#' #    Merc 240D            24.4      4  146.7     62    3.7    3.2   20.0
#' #    Merc 230             22.8      4  140.8     95    3.9    3.1   22.9
#' #    Merc 280             19.2      6  167.6    123    3.9    3.4   18.3
#' #
#' @export
column_defaults <- function(x, vars = NULL, from = NULL, to = NULL, label = NULL, 
                  format = NULL, align=NULL, label_align=NULL, width=NULL,
                   n = NULL, standard_eval = FALSE, style = NULL) {
  
  if (!"table_spec" %in% class(x))
    stop("Input object must be of class 'table_spec'.")

  if (standard_eval) {
    
    if (is.null(vars)) {
      
      vars_c <- character(0)
    } else {
      if (typeof(vars) != "character")
        stop("vars parameter must be a character if standard_eval is TRUE.")
    
      vars_c <- vars
    }
    
  } else {
  
    # Determine if it is a vector or not.  "language" is a vector.
    
    if (typeof(substitute(vars, env = environment())) == "language") {
      ret <- tryCatch({
        if (is.numeric(vars)) {
          names(x$data)[vars]
          
        }
  
      }, error = function(e) {
        FALSE
      })
      
      if (all(ret == FALSE))
        v <- substitute(vars, env = environment())
      else 
        v <- ret
      
    } else 
      v <- substitute(list(vars), env = environment())
    
    
    # Turn each item into a character
    if (!is.character(v)) {
      vars_c <- c()
      if (length(v) > 1) {
        for (i in 2:length(v)) {
          vars_c[[length(vars_c) + 1]] <- as.character(v[[i]]) 
        }
        
      }
    } else 
      vars_c <- v
    
    # Convert list to vector
    vars_c <- unlist(vars_c)
  
  }
  
  # Deal with curly brace escape
  if (length(vars_c) > 0) {
    if (vars_c[[1]] == "{") { 
      vars_c <- get(vars_c[[2]], envir = parent.frame())
    }
  }
  
  # Check that variable exists in data frame
  if (!is.null(x$data) & !is.null(vars_c)) {
    if (!any(vars_c %in% names(x$data))) {
      for (nm in vars_c) {
        if (!nm %in% names(x$data)) 
        stop(paste0("Variable does not exist in data: ", nm))
      }
    }
  }
  
  # Create column default object
  dflt <- structure(list(), class = c("col_dflt", "list"))
  
  if (!identical(vars_c, character(0)))
    dflt$vars = vars_c
  
  if (standard_eval) {
    
    if (is.null(from)) {
      
      f <- character(0)
      
    } else {
      if (typeof(from) != "character" )
        stop("from parameter must be of type character if standard_eval = TRUE.")
      
      f <- from
    }
    
  } else {
  
    # Assign from value
    f <- as.character(substitute(from, env = environment()))
  
  }
  
  if (!identical(f, character(0))) {
    
    # Deal with curly brace escape for from
    if (length(f) > 1) {
      if (f[[1]] == "{") { 
        f <- f[[2]]
        f <- gsub("{", "", f, fixed = TRUE)
        f <- gsub("}", "", f, fixed = TRUE)
        f <- gsub("\\n", "", f, fixed = TRUE)
        f <- get(trimws(f), envir = parent.frame())
      }
    }
    
    if (suppressWarnings(!is.na(as.integer(f))))
        f <- names(x$data)[as.integer(f)]
    
    if (!f %in% names(x$data))
      stop(paste("Variable does not exist in input data frame:", f))
    
    dflt$from =  f
  }
  
  
  
  if (standard_eval) {
    
    if (is.null(to)) {
      
      t <- character(0)
      
    } else {
      if (typeof(to) != "character")
        stop("to parameter must be of type character if standard_eval = TRUE.")
      
      t <- to
    }
    
  } else {
    
    # Assign to value
    t <- as.character(substitute(to, env = environment()))
    
  }
  
  if (!identical(t, character(0))) {
    
    # Deal with curly brace escape for to
    if (length(t) > 1) {
      if (t[[1]] == "{") { 
        t <- t[[2]]
        t <- gsub("{", "", t, fixed = TRUE)
        t <- gsub("}", "", t, fixed = TRUE)
        t <- gsub("\\n", "", t, fixed = TRUE)
        t <- get(trimws(t), envir = parent.frame())
      }
    }
    
    if (suppressWarnings(!is.na(as.integer(t))))
      t <- names(x$data)[as.integer(t)]
  
    if (!t %in% names(x$data))
      stop(paste("Variable does not exist in input data frame:", t))
    
    dflt$to =  t
  }
  
  # Catch range mismatches
  if (!is.null(dflt$from) & is.null(dflt$to)) {
    stop("'to' parameter cannot be null if 'from' is populated.") 
  }
  
  if (is.null(dflt$from) & !is.null(dflt$to)) {
    stop("'from' parameter cannot be null if 'to' is populated.") 
  }
  
  dflt$label = label
  dflt$format = format
  dflt$align = align
  dflt$label_align = if (is.null(label_align) & !is.null(align))
    align else label_align
  dflt$width = width
  dflt$n = n
  dflt$style = style

  x$col_dflts[[length(x$col_dflts) + 1]] <- dflt
  
  
  return(x)
}


#' @title Defines a spanning header
#' @description Create a header that spans multiple columns.  Spanning headers
#' are used to group related columns.  Such groupings are a common 
#' feature of statistical reports.
#' @details 
#' A spanning header is a label and underline that spans one or more 
#' columns.  A spanning header is defined minimally by identifying the
#' column range to be spanned, and a label.  A label alignment and "N="
#' value may also be supplied.
#' 
#' The spanning column range is defined by the \code{from} and \code{to} 
#' parameters.  The range identifies a contiguous set of variables on the data.
#' Variables can be identified by position, a quoted variable name, or an 
#' unquoted variable name.
#' @param x The table object to add spanning headers to.
#' @param from The starting column to span.  Spanning columns are defined as
#' range of columns 'from' and 'to'. The columns may be identified by position, 
#' or by quoted or unquoted variable names. If you want to pass the \code{from} 
#' value using an R variable,
#' set the \code{standard_eval} parameter to TRUE.
#' The \code{from} parameter is required.  
#' @param to The ending column to span.  Spanning columns are defined as
#' range of columns 'from' and 'to'. The columns may be identified by position,
#' or by quoted or unquoted variable names.  If you want to pass the \code{to} 
#' value using an R variable,
#' set the \code{standard_eval} parameter to TRUE.
#' The \code{to} parameter is required. 
#' @param label The label to apply to the spanning header.
#' @param label_align The alignment to use for the label. Valid values are 
#' "left", "right", "center", and "centre".  The default for spanning columns
#' is "center".
#' @param level The level to use for the spanning header.  The lowest
#' spanning level is level 1, the next level above is level 2, and so on.  
#' By default, the level is set to 1.
#' @param n The population count to use for the "N=" label on the spanning 
#' header. The "N=" label will be formatted according to the \code{n_format}
#' parameter on the \code{\link{create_table}} function.
#' @param underline A TRUE or FALSE value indicating whether the spanning
#' header should be underlined.  Default is TRUE.  
#' @param bold A TRUE or FALSE value indicating whether the spanning header
#' label should be bold.  Default is FALSE.
#' @param standard_eval A TRUE or FALSE value that indicates whether to
#' use standard or non-standard evaluation of the \code{from},
#' and \code{to} parameters.  Set \code{standard_eval} to TRUE if you want
#' to pass the column names as variables. Default is FALSE, meaning it
#' will use non-standard (unquoted) evaluation.
#' @return The modified table spec.
#' @family table
#' @examples 
#' library(reporter)
#' library(magrittr)
#' 
#' # Create a temporary file
#' tmp <- file.path(tempdir(), "iris.txt")
#' 
#' # Prepare data
#' dat <- iris[sample(1:150, 15), c(5, 1, 2, 3, 4)]
#' dat <- dat[order(dat$Species), ]
#'
#' # Define table
#' tbl <- create_table(dat) %>% 
#'   titles("Table 3.2", "IRIS Sample Report") %>% 
#'   spanning_header(2, 3, label = "Sepal") %>% 
#'   spanning_header(4, 5, label = "Petal") %>% 
#'   column_defaults(2:5, format = "%.1f") %>% 
#'   define(Species, align = "left", dedupe = TRUE, blank_after = TRUE) %>% 
#'   define(Sepal.Length, label = "Length") %>% 
#'   define(Sepal.Width, label = "Width") %>% 
#'   define(Petal.Length, label = "Length") %>% 
#'   define(Petal.Width, label = "Width") %>% 
#'   footnotes("* From Fisher's Iris Dataset")
#'        
#' # Define report
#' rpt <- create_report(tmp, orientation="portrait") %>%
#'   options_fixed(blank_margins = TRUE) %>% 
#'   set_margins(top = 1, bottom =1) %>% 
#'   add_content(tbl, align = "left") 
#' 
#' # Write the report
#' write_report(rpt)
#' 
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' #
#' #
#' #
#' #
#' #                      Table 3.2
#' #                  IRIS Sample Report
#' #
#' #                       Sepal        Petal
#' #                   ------------ ------------
#' #       Species     Length Width Length Width
#' #       -------------------------------------
#' #       setosa         5.0   3.0    1.6   0.2
#' #                      4.6   3.4    1.4   0.3
#' #                      5.0   3.4    1.6   0.4
#' #                      5.7   3.8    1.7   0.3
#' #
#' #       versicolor     5.7   2.8    4.1   1.3
#' #                      6.2   2.9    4.3   1.3
#' #                      7.0   3.2    4.7   1.4
#' #                      6.6   2.9    4.6   1.3
#' #
#' #       virginica      6.2   3.4    5.4   2.3
#' #                      7.2   3.0    5.8   1.6
#' #                      6.9   3.1    5.1   2.3
#' #                      5.6   2.8    4.9   2.0
#' #                      7.7   2.6    6.9   2.3
#' #                      6.3   2.8    5.1   1.5
#' #                      7.7   2.8    6.7   2.0
#' #
#' #
#' #       * From Fisher's Iris Dataset
#' @export
spanning_header <- function(x, from, to, label = "",
                            label_align = "center", level = 1, n = NULL,
                            underline = TRUE, bold = FALSE, standard_eval = FALSE) {
  
  if (standard_eval) {
    
    if (!typeof(from) %in% c("double", "integer", "character"))
      stop(paste0("from parameter must be a type of character, ", 
                  "integer, or double if standard_eval is TRUE."))
    
    if (!typeof(to)  %in% c("double", "integer", "character"))
      stop(paste0("to parameter must be a type of character, ",
                  "integer, or double if standard_eval is TRUE."))
    
    f <- from
    t <- to
    
  } else {
  
    f <- as.character(substitute(from, env = environment()))
    t <- as.character(substitute(to, env = environment()))
  }

  # Deal with curly brace escape for from
  if (length(f) > 1) {
    if (f[[1]] == "{") { 
      f <- f[[2]]
      f <- gsub("{", "", f, fixed = TRUE)
      f <- gsub("}", "", f, fixed = TRUE)
      f <- gsub("\\n", "", f, fixed = TRUE)
      f <- get(trimws(f), envir = parent.frame())
    }
  }
  
  # Deal with curly brace escape for to
  if (length(t) > 1) {
    if (t[[1]] == "{") { 
      t <- t[[2]]
      t <- gsub("{", "", t, fixed = TRUE)
      t <- gsub("}", "", t, fixed = TRUE)
      t <- gsub("\\n", "", t, fixed = TRUE)
      t <- get(trimws(t), envir = parent.frame())
    }
  }
  
  if (!is.na(suppressWarnings(as.numeric(f))))
    f <- as.numeric(f)
  
  if (!is.na(suppressWarnings(as.numeric(t))))
    t <- as.numeric(t)
  
  nms <- names(x$data)
  
  if (is.character(f)) {
    if (!f %in% nms) {
      stop(paste0("From variable '", f, "' does not exist in data."))

    }
  }
  
  if (is.character(t)) {
    if (!t %in% nms) {
      stop(paste0("To variable '", t, "' does not exist in data."))
      
    }
  }
  
  if (is.numeric(f)) {

      if (!(f > 0 & f <= ncol(x$data))) {
        stop(paste0("From variable position '", f, "' is invalid."))

      } else
        f <- nms[[f]]
  } 
  
  if (is.numeric(t)) {

    if (!(t > 0 & t <= ncol(x$data))) {
      stop(paste0("To variable position '", t, "' is invalid."))
      
    } else 
      t <- nms[[t]]
  } 
  
  if (!label_align %in% c("left", "right", "center", "centre")) {
   stop(paste0("label_align '", label_align, "' is invalid. ",
               "Valid values are 'left', 'right', 'center', or 'centre'."))
  }
  
  if (!is.numeric(level) | is.na(level) | is.null(level)) {
   stop(paste0("level parameter value '", level, "' is invalid.")) 
  }
  
  sh <- structure(list(), class = c("span_def", "list"))
  
  # Get spanning range
  nms <- names(x$data)
  startpos <- match(f, nms)
  endpos <- match(t, nms)
  spn <- nms[startpos:endpos]
  
  if (has_glue()) {
    sh$label <- gluev(label)
  } else {
    sh$label <- label
  }
  
  sh$span_cols <- spn
  sh$from = f
  sh$to = t
  sh$label_align = label_align
  sh$level = level
  sh$n = n
  sh$underline = underline
  sh$bold = bold

  x$col_spans[[length(x$col_spans) + 1]] <- sh

  return(x)
}

#' @title Defines a report stub
#' @description Combine columns into a nested report stub.  The report stub
#' is a common feature of statistical reports.  The stub is created with
#' the \code{stub} function, and frequently appears in combination with the 
#' \code{label_row} and \code{indent} parameters from the 
#' \code{\link{define}} function.  These elements work together to define
#' the appearance of the stub.
#' @details 
#' The table stub is a nested set of labels that identify rows 
#' on the table. The stub is created by combining two or more columns into 
#' a single stub column.  The relationship between the columns is typically 
#' visualized as a hierarchy, with lower level concepts indented under 
#' higher level concepts.  
#' 
#' A typical stub is created with the following steps:
#' \itemize{
#'   \item Prepare the data. 
#'   \item Create the table object.
#'   \item Define the stub on the table using the \code{stub} function, 
#'   and identify the variables to be combined.
#'   \item Identify higher level concepts with the \code{label_row} parameter
#'   on the \code{\link{define}} function. 
#'   \item Identify lower level concepts using the \code{indent} parameter 
#'   on the \code{\link{define}} function.
#' }
#' 
#' The stub will be automatically added as an identity variable on the report, 
#' and will always appear as the leftmost column.  There can only be one stub 
#' defined on a report.
#' 
#' If you wish to create multiple levels of nested labels, use
#' an NA value to prevent lower level labels from overwriting
#' higher level labels. 
#' 
#' For example, the following data:
#' \preformatted{
#' continent          country   state_province    
#' "North America"    NA        NA   
#' "North America"    "Canada"  NA   
#' "North America"    "Canada"  "Ontario"   
#' "North America"    "USA"     NA   
#' "North America"    "USA"     "New York"   
#' "South America"    NA        NA   
#' "South America"    "Brazil"  NA   
#' "South America"    "Brazil"  "Amazonas"   
#' "South America"    "Brazil"  "Bahia"   
#' }
#' Will produce the following stub:   
#' \preformatted{
#' North America   
#'   Canada   
#'     Ontario   
#'   USA   
#'     New York   
#' South America  
#'   Brazil 
#'     Amazonas   
#'     Bahia   
#' }
#' With the following code:
#' \preformatted{
#' tbl <- create_table(dat) \%>\% 
#'   stub(c(continent, country, state_province)) \%>\% 
#'   define(country, indent = .25) \%>\% 
#'   define(state_province, indent = .5)
#' }
#' @param x The table spec.
#' @param vars A vector of quoted or unquoted variable names from 
#' which to create the stub. If you want to pass an R variable of names,
#' escape the values with double curly braces, i.e. \code{vars = {{myvar}}}.
#' The curly brace escape is useful when writing functions that construct
#' reports dynamically. 
#' @param label The label for the report stub.  The default label is an empty
#' string.
#' @param label_align The alignment for the stub column label.  
#' Valid values are 'left', 'right', 'center', and 'centre'.  Default follows
#' the \code{align} parameter.
#' @param width The width of the stub, in report units of measure.
#' @param align How to align the stub column.  Valid values are 'left', 
#' 'right', 'center', and 'centre'.  Default is 'left'.
#' @param standard_eval A TRUE or FALSE value that indicates whether to
#' use standard or non-standard evaluation of the \code{vars}, \code{from},
#' and \code{to} parameters.  Set \code{standard_eval} to TRUE if you want
#' to pass the column names as variables. Default is FALSE, meaning it
#' will use non-standard (unquoted) evaluation.
#' @param style A \code{\link{cell_style}} object that contains the style
#' specifications for the stub.
#' @return The modified table spec.
#' @family table
#' @examples 
#' library(reporter)
#' library(magrittr)
#' 
#' # Create temporary path
#' tmp <- file.path(tempdir(), "stub.txt")
#' 
#' # Read in prepared data
#' df <- read.table(header = TRUE, text = '
#'       var      label        A             B          
#'       "ampg"   "N"          "19"          "13"         
#'       "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
#'       "ampg"   "Median"     "16.4"        "21.4"       
#'       "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
#'       "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
#'       "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
#'       "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
#'       "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
#' 
#' # Create table
#' tbl <- create_table(df, first_row_blank = TRUE) %>% 
#'   stub(c(var, label)) %>% 
#'   define(var, blank_after = TRUE, label_row = TRUE, 
#'          format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
#'   define(label, indent = .25) %>% 
#'   define(A, label = "Group A", align = "center", n = 19) %>% 
#'   define(B, label = "Group B", align = "center", n = 13)
#' 
#' 
#' # Create report and add content
#' rpt <- create_report(tmp, orientation = "portrait") %>% 
#'   page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
#'   titles("Table 1.0", "MTCARS Summary Table") %>% 
#'   add_content(tbl) %>% 
#'   footnotes("* Motor Trend, 1974") %>%
#'   page_footer(left = Sys.time(), 
#'               center = "Confidential", 
#'               right = "Page [pg] of [tpg]")
#' 
#' # Write out report
#' write_report(rpt)
#' 
#' # View report in console
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' # Client: Motor Trend                                                Study: Cars
#' #                                   Table 1.0
#' #                              MTCARS Summary Table
#' # 
#' #                                     Group A      Group B
#' #                                      (N=19)       (N=13)
#' #                 -------------------------------------------
#' # 
#' #                 Miles Per Gallon
#' #                    N                   19           13
#' #                    Mean            18.8 (6.5)   22.0 (4.9)
#' #                    Median             16.4         21.4
#' #                    Q1 - Q3        15.1 - 21.2  19.2 - 22.8
#' #                    Range          10.4 - 33.9  14.7 - 32.4
#' #
#' #                 Cylinders
#' #                    8 Cylinder     10 ( 52.6%)   4 ( 30.8%)
#' #                    6 Cylinder      4 ( 21.1%)   3 ( 23.1%)
#' #                    4 Cylinder      5 ( 26.3%)   6 ( 46.2%)
#' # 
#' # ...
#' # 
#' # 
#' # * Motor Trend, 1974
#' # 
#' # 2020-08-30 03:50:02              Confidential                      Page 1 of 1
#' #
#' @export
stub <- function(x, vars, label = "", label_align = NULL, 
                 align = "left", width = NULL, standard_eval = FALSE,
                 style = NULL) {
  
  def <- structure(list(), class = c("stub_def", "list"))
  
  if (standard_eval) {
    
    if (typeof(vars) != "character") 
      stop("vars parameter must be of type character if standard_eval is TRUE.")
    
    vars_c <- vars
    
  } else {
  
    # Determine if it is a vector or not.  "language" is a vector.
    if (typeof(substitute(vars, env = environment())) == "language") 
      v <- substitute(vars, env = environment())
    else 
      v <- substitute(list(vars), env = environment())
    
    # Turn each item into a character
    vars_c <- c()
    if (length(v) > 1) {
      for (i in 2:length(v)) {
        vars_c[[length(vars_c) + 1]] <- as.character(v[[i]]) 
      }
      
    }
  
  }
  
  # Convert list to vector
  vars_c <- unlist(vars_c)
  
  # Deal with curly brace escape
  if (length(vars_c) > 0) {
    if (vars_c[[1]] == "{") { 
      vars_c <- get(vars_c[[2]], envir = parent.frame())
    }
  }
  
  # Check that variable exists in data frame
  if (!is.null(x$data) & !is.null(vars_c)) {
    if (!any(vars_c %in% names(x$data))) {
      for (nm in vars_c) {
        if (!nm %in% names(x$data)) 
          stop(paste0("Variable does not exist in data: ", nm))
      }
    }
  }
  
  if (has_glue()) {
    def$label <- gluev(label)
  } else {
    def$label <- label
  }
  def$label_align <- label_align
  def$align <- align
  def$vars <- vars_c
  def$width <- width
  def$style <- style
  
  x$stub <- def
  
  return(x)
}


#' @title Prints the table spec
#' @description A function to print the table spec.
#' The \strong{print} function will print the table spec in summary 
#' form.  To view all parameters, set the \code{verbose} parameter to TRUE.
#' @param x The table spec.
#' @param ... Additional parameters to pass to the underlying print function.
#' @param verbose Whether to print in verbose form, which is similar to 
#' a list.  Default is FALSE, which prints in summary form.
#' @seealso 
#' \code{\link{create_table}} function to create a table specification.
#' @return The table spec, invisibly.
#' @family table
#' @examples 
#' library(magrittr)
#' 
#' # Create Table
#' tbl <- create_table(mtcars) %>% 
#'   define(mpg, label = "Miles Per Gallon", width = .5) %>% 
#'   define(cyl, label = "Cylinders") %>% 
#'   titles("Table 6.4", "MTCARS Sample Table") %>% 
#'   footnotes("* Motor Trend, 1974")
#'   
#' tbl
#'
#' # A table specification:
#' # - data: data.frame 'mtcars' 32 rows 11 cols
#' # - show_cols: all
#' # - use_attributes: all
#' # - title 1: 'Table 6.4'
#' # - title 2: 'MTCARS Sample Table'
#' # - footnote 1: '* Motor Trend, 1974'
#' # - define: mpg 'Miles Per Gallon' width=0.5 
#' # - define: cyl 'Cylinders' 
#' @import crayon
#' @export
print.table_spec <- function(x, ..., verbose = FALSE){
  

  if (verbose == TRUE) {
    
    # If verbose mode is indicated, print values as a list
    for (nm in names(x)) {
      
      cat("$", nm, "\n", sep = "")
      if (nm == "data") {
  
        m <- ncol(x[[nm]]) * 10
        print(x[[nm]], ..., max = m)
      }
      else  {
        
        print(x[[nm]], ...)
      }
      cat("\n")
    }
    
  } else {
    
    if ("tbl_df" %in% class(x$data))
      dtyp = "tibble "
    else 
      dtyp = "data.frame "
   
    grey60 <- make_style(grey60 = "#999999")
    #pcolor <- make_style("snow")
    
    # Print header
    cat(grey60("# A table specification:\n"))
    cat(paste0("- data: ", dtyp, "'", x$dataname,
                 "' ", nrow(x$data), " rows ", ncol(x$data),
                 " cols\n"))
    
    if (!is.null(x$show_cols)) {
      cat(paste0("- show_cols: ",
            paste(as.character(x$show_cols), collapse = " "), "\n"))
      
    }
    
    if (all(!is.null(x$use_attributes))) {
      
      if (all(x$use_attributes == ""))
        ua <- "none"
      else if (all(x$use_attributes %in% c("label", "width", "justify", "format")))
        ua <- "all"
      else 
        ua <- paste(x$use_attributes, collapse = " ")
      
      cat(paste0("- use_attributes: ",  ua, "\n"))
      
  
    }
    
    if (!is.null(x$width))
      cat(paste0("- width: ", as.character(x$width), "\n"))
    
    if (!is.null(x$headerless)) {
      if (x$headerless != FALSE)
        cat(paste0("- headerless: ", as.character(x$headerless), "\n"))
    }
    
    if (!is.null(x$page_by)) {
      cat(paste0("- page by: ", x$page_by$var, "\n"))
      
    }
    
    print_title_header(x$title_hdr)
    
    print_titles(x$titles)
    
    print_footnotes(x$footnotes)
    
    
    # Print spanning headers
    if (!is.null(x$col_spans)) {
      
      for (def in x$col_spans) {
        
        cat(paste0("- spanning_header: from='", def$from,
            "' to='", def$to, "' "))
        
        if (!is.null(def[["label"]]))
          cat(paste0("'", def[["label"]], "' "))
        
        if (!is.null(def$level)) 
          cat(paste0("level=", def$level, " "))

        
        cat("\n")
        
      }
    }

    # (x, vars, label = "", label_align = NULL, 
    # align = "left", width = NULL)
    
    if (!is.null(x$stub)) {
      
      stb <- x$stub
      
      cat(paste0("- stub: ", paste(stb$vars, collapse = " "), " "))
      if (stb$label != "")
        cat(paste0("'", stb$label, "' "))
      
      if (!is.null(stb$width))
        cat(paste0("width=", stb$width, " "))
      
      if (!is.null(stb$align)) 
        cat(paste0("align='", stb$align, "' "))
      
      cat("\n")
      
    }
    
    # Print column definitions
    if (!is.null(x$col_defs)) {
      
      for (def in x$col_defs) {

        cat(paste0("- define: ", def$var_c, " "))
        if (!is.null(def[["label"]]))
          cat(paste0("'", def[["label"]], "' "))

        if (!is.null(def$width)) 
          cat(paste0("width=", def$width, " "))
        
        if (!is.null(def$align)) 
          cat(paste0("align='", def$align, "' "))
        
        if (def$visible == FALSE) 
          cat(paste0("visible='", def$visible, "' "))
        
        if (def$id_var == TRUE) 
          cat(paste0("id_var='", def$id_var, "' "))
        
        if (def$dedupe == TRUE) 
          cat(paste0("dedupe='", def$dedupe, "' "))

        if (def$page_wrap == TRUE) 
          cat(paste0("page_wrap='", def$page_wrap, "' "))
        
        if (def$page_break == TRUE) 
          cat(paste0("page_break='", def$page_break, "' "))
        
        cat("\n")
        
      }
    }
    
  }

  invisible(x)
}


# Formats ----------------------------------------------------------------------


#' @title Functions to format the population label
#' @description These functions are used to format the "N=" population label
#' on column headers.  
#' @details Which function to use to format the population label is specified
#' on the \code{n_format} parameter on the \code{\link{create_table}} function.
#' These formatting functions provide several options for formatting the "N=", 
#' including whether the "N" should be upper case or lower case, and whether
#' or not to put the value in parentheses.  If one of these options does not 
#' meet the specifications for your report, you may also write your own 
#' formatting function and pass it to the \code{n_format} function.  When an 
#' N value is supplied, the output of this function will be concatenated 
#' to the header label.
#' @usage lowcase_parens(x)
#' @usage upcase_parens(x)
#' @usage lowcase_n(x)
#' @usage upcase_n(x)
#' @aliases lowcase_parens upcase_parens lowcase_n upcase_n
#' @seealso 
#' \code{\link{create_table}} function to create a table.
#' @param x Population count
#' @examples 
#' # Create test data
#' l <- "Label"
#' n <- 47
#' 
#' cat(paste0(l, lowcase_parens(n)))
#' # Label
#' # (n=47)
#' 
#' cat(paste0(l, upcase_parens(n)))
#' # Label
#' # (N=47)
#' 
#' cat(paste0(l, lowcase_n(n)))
#' # Label
#' # n=47
#' 
#' cat(paste0(l, upcase_n(n)))
#' # Label
#' # N=47
#' 
#' customN <- function(n) {
#'   return(paste0(": N=", n))
#' }
#' cat(paste0(l, customN(n)))
#' # Label: N=47
#' 
#' ## Use alternate n format ##
#' library(reporter)
#' library(magrittr)
#' 
#' tmp <- tempfile(fileext = ".txt")
#' 
#' # Prepare data
#' df <- read.table(header = TRUE, text = '
#'       Hair  Group1  Group2
#'       Black    25      16
#'       Brown    13      18
#'       Blonde   5        7
#'       Red      2        1')
#'       
#' # Create table with lowcase n formatting
#' tbl <- create_table(df, n_format=lowcase_n) %>% 
#'        titles("Hair Color") %>% 
#'        define(Group1, n = 45) %>% 
#'        define(Group2, n = 42)
#' 
#' # Create report
#' rpt <- create_report(tmp) %>% 
#'        add_content(tbl, align = "left")
#'        
#' # Write to file system
#' write_report(rpt)     
#' 
#' writeLines(readLines(tmp))
#' #      Hair Color
#' # 
#' #        Group1 Group2
#' # Hair     n=45   n=42
#' # ¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯
#' # Black      25     16
#' # Brown      13     18
#' # Blonde      5      7
#' # Red         2      1
#' @export
lowcase_parens <- function(x) {
  
  ret <- paste0("\n(n=", x, ")")
  
  return(ret)
}

#' @aliases lowcase_parens
#' @export
upcase_parens <- function(x) {
  
  ret <- paste0("\n(N=", x, ")")
  
  return(ret)
  
}

#' @aliases lowcase_parens
#' @export
lowcase_n <- function(x) {
  
  ret <- paste0("\nn=", x)
  
  return(ret)
}

#' @aliases lowcase_parens
#' @export
upcase_n <- function(x) {
  
  ret <- paste0("\nN=", x)
  
  return(ret)
  
}

context("Table Spec Tests")



test_that("table spec constructor operator works as expected.", {
  

  
  tbl <- create_table(mtcars[1:10, ], show_cols = "none",
                      first_row_blank = TRUE) %>% 
    define(mpg, format = "%.1f") %>% 
    define(cyl, width = 1) %>% 
    define(hp)
  
  expect_equal(tbl$show_cols, "none")
  expect_equal(tbl$first_row_blank, TRUE)
  expect_equal(nrow(tbl$data), 10)
  expect_equal(length(tbl$col_defs), 3)
  
  
})


test_that("spanning header constructor works as expected.", {
  
  # Unquoted names and other parameters
  tbl <- create_table(mtcars[1:10, ]) %>% 
    spanning_header(from = mpg, to = disp, 
                    label = "Span 1", n = 25, label_align = "left") %>% 
    spanning_header(from = hp, to = qsec, 
                    label = "Span 2", n = 30, underline = FALSE) %>% 
    define(mpg, format = "%.1f") %>% 
    define(cyl, width = 1) %>% 
    define(hp)
  
  s <- tbl$col_spans
  
  expect_equal(length(s), 2)
  expect_equal(s[[1]]$label, "Span 1")
  expect_equal(s[[1]]$label_align, "left")
  expect_equal(s[[1]]$n, 25)
  expect_equal(s[[1]]$from, "mpg")
  expect_equal(s[[1]]$to, "disp")
  expect_equal(s[[1]]$underline, TRUE)
  expect_equal(s[[2]]$label, "Span 2")
  expect_equal(s[[2]]$n, 30)
  expect_equal(s[[2]]$underline, FALSE)
  expect_equal(length(tbl$col_defs), 3)
  
  # Quoted Names
  tbl <- create_table(mtcars[1:10, ]) %>% 
    spanning_header(from = "mpg", to = "disp")
  s <- tbl$col_spans
  expect_equal(s[[1]]$from, "mpg")
  expect_equal(s[[1]]$to, "disp")
  
  # Positions
  tbl <- create_table(mtcars[1:10, ]) %>% 
    spanning_header(from = 1, to = 3)
  s <- tbl$col_spans
  expect_equal(s[[1]]$from, "mpg")
  expect_equal(s[[1]]$to, "disp")
  
  # Positions with standard_eval = TRUE
  tbl <- create_table(mtcars[1:10, ]) %>% 
    spanning_header(from = 1, to = 3, standard_eval = TRUE)
  s <- tbl$col_spans
  expect_equal(s[[1]]$from, "mpg")
  expect_equal(s[[1]]$to, "disp")
  
  
  # Parameter checks
  expect_error(spanning_header(tbl, "mpgg", "disP"))
  expect_error(spanning_header(tbl, 1, 25))
  expect_error(spanning_header(tbl, mpg, disp,
                               level = "one"))
  expect_error(spanning_header(tbl, "mpg", "disp",
                               label_align = "lefty"))
  
})

test_that("stub function works as expected.", {
  
  dat <- mtcars
  dat$name <- rownames(mtcars)
  
  fmt <- value(condition(x >= 20, "High"),
               condition(TRUE, "Low"))
  
  dat$mpg_cat <- fapply(dat$mpg, fmt)
  rownames(dat) <- NULL
  
  dat <- dat[order(dat$mpg_cat), c("mpg_cat", "name", "mpg", "cyl", "disp", "hp")]
  
  tbl <- create_table(dat) %>%
    stub(label = "Stub", vars = c("mpg_cat", "name")) %>% 
      define(mpg_cat, label_row = TRUE, dedupe = TRUE) %>% 
      define(name, indent = .25)
  
  
  expect_equal(tbl$col_defs[[1]]$label_row, TRUE)
  expect_equal(is.null(tbl$stub), FALSE)
  
  
})



test_that("page break parameter works as expected.", {
  
  dat <- mtcars
  rownames(dat) <- NULL
  
  dat$pg <- c(rep(1, 16), rep(2, 16))
  

  
  tbl <- create_table(dat) %>%
    define(pg, page_break = TRUE)

  
  expect_equal(tbl$col_defs[[1]]$page_break, TRUE)
  expect_equal(tbl$page_var, "pg")

  
  
})

test_that("use_attributes parameter works as expected.", {
  
  # Check default
  tbl <- create_table(mtcars) 
  expect_equal(tbl$use_attributes, c("label", "width", "justify", "format" ))
  
  # Check none
  tbl <- create_table(mtcars, use_attributes = "none") 
  expect_equal(tbl$use_attributes, c(""))
  
  # Check some
  tbl <- create_table(mtcars, use_attributes = c("justify", "format")) 
  expect_equal(tbl$use_attributes, c("justify", "format"))
  
  # Check invalid
  expect_error(create_table(mtcars, use_attributes = c("justify", "fork")) )
  
  
  
})



test_that("column_defaults function works as expected.", {
  
  
  tbl <- create_table(mtcars)
  cd <- column_defaults(tbl, c(mpg, cyl, disp)) 

  expect_equal(cd$col_dflts[[1]]$vars, c("mpg", "cyl", "disp"))
  expect_equal(is.null(cd$col_dflts[[1]]$from), TRUE)
  expect_equal(is.null(cd$col_dflts[[1]]$to), TRUE)
  
  cd <- column_defaults(tbl, from = mpg, to = disp) 
  
  expect_equal(is.null(cd$col_dflts[[1]]$vars), TRUE)
  expect_equal(cd$col_dflts[[1]]$from, "mpg")
  expect_equal(cd$col_dflts[[1]]$to, "disp")
  
  
  cd <- column_defaults(tbl, c("mpg", "cyl", "disp")) 
  
  expect_equal(cd$col_dflts[[1]]$vars, c("mpg", "cyl", "disp"))
  expect_equal(is.null(cd$col_dflts[[1]]$from), TRUE)
  expect_equal(is.null(cd$col_dflts[[1]]$to), TRUE)
  
  cd <- column_defaults(tbl, from = "mpg", to = "disp") 
  
  expect_equal(is.null(cd$col_dflts[[1]]$vars), TRUE)
  expect_equal(cd$col_dflts[[1]]$from, "mpg")
  expect_equal(cd$col_dflts[[1]]$to, "disp")
  
  
  expect_error(column_defaults(tbl, fork))
  expect_error(column_defaults(tbl, from = fork, to = disp))
  expect_error(column_defaults(tbl, from = mpg, to = fork))
  expect_error(column_defaults(tbl, from = mpg))
  expect_error(column_defaults(tbl, to = mpg))

  
  cd <- column_defaults(tbl, from = "mpg", to = "disp", width = 10,
                        align = "left", label = "fork", label_align = "right",
                        format = "%.1f", n = 12)
  expect_equal(cd$col_dflts[[1]]$width, 10)
  expect_equal(cd$col_dflts[[1]]$align, "left")
  expect_equal(cd$col_dflts[[1]]$label, "fork")
  expect_equal(cd$col_dflts[[1]]$label_align, "right")
  expect_equal(cd$col_dflts[[1]]$format, "%.1f")
  expect_equal(cd$col_dflts[[1]]$n, 12)
  
  
  expect_error(column_defaults(create_report()))
  
  
  cd <- column_defaults(tbl, 1:3) 
  
  expect_equal(cd$col_dflts[[1]]$vars, c("mpg", "cyl", "disp"))
  expect_equal(is.null(cd$col_dflts[[1]]$from), TRUE)
  expect_equal(is.null(cd$col_dflts[[1]]$to), TRUE)
  
  
  cd <- column_defaults(tbl, from = 1, to = 3) 
  
  expect_equal(is.null(cd$col_dflts[[1]]$vars), TRUE)
  expect_equal(cd$col_dflts[[1]]$from, "mpg")
  expect_equal(cd$col_dflts[[1]]$to, "disp")
  
})




test_that("define_c function works as expected.", {
  
  
  tbl <- create_table(mtcars)
  d <- define_c("mpg", width = 2)
  
  expect_equal("col_def" %in% class(d), TRUE )
  expect_equal(d$width, 2)
  expect_equal(d$var, "mpg")
  
})


test_that("print.tbl_spec() works as expected.", {
  
  tbl <- create_table(mtcars, headerless = TRUE, 
                      show_cols = c("mpg", "cyl", "disp"),
                      use_attributes = "none", width = 8) %>% 
    titles("Table 1", "My title") %>% 
    titles("Here is a much longer title to see what happens when I print it") %>% 
    footnotes("My footnote 1", "My footnote 2") %>% 
    spanning_header(from = "mpg", to = "disp", label = "My label") %>% 
    stub(c(cyl, disp), "My stub") %>% 
    define(mpg, width = 2, label = "Miles per Gallon") %>% 
    define(c(cyl, disp), align = "left")
  
  tbl
  
  tbl2 <- create_table(mtcars) 
  
  tbl2

  tbl3 <- create_table(mtcars) %>% 
    titles("Table 1", "My title") %>% 
    titles("Here is a much longer title to see what happens when I print it") %>% 
    footnotes("My footnote 1", "My footnote 2") %>% 
    spanning_header(from = "mpg", to = "disp", label = "My label") %>% 
    spanning_header(from = "vs", to = "carb", label = "My label 2") %>% 
    define(mpg, width = 2, label = "Miles per Gallon") %>% 
    define(cyl, align = "left")
  
  tbl3
  
  expect_equal(TRUE, TRUE)

})


test_that("show_cols parameter works as expected.", {
  
  
  tbl <- create_table(mtcars, show_cols = 1:3)
  
  expect_equal(tbl$show_cols, c("mpg", "cyl", "disp"))
  
  tbl <- create_table(mtcars, show_cols = c("disp", "hp", "drat"))
  
  expect_equal(tbl$show_cols, c("disp", "hp", "drat"))
  
  tbl <- create_table(mtcars)
  
  expect_equal(tbl$show_cols, "all")
  
  
  tbl <- create_table(mtcars, show_cols = "none")
  
  expect_equal(tbl$show_cols, "none")
  
  tbl <- create_table(mtcars, show_cols = NULL)
  
  expect_equal(tbl$show_cols, "all")
  
  
})
  

test_that("define escape works as expected.", {
  
  tbl <- create_table(mtcars)
  
  v1 <- c("mpg", "cyl")
  
  expect_error(define(tbl, vars = v1))
  
  
  dfn <- define(tbl, vars = {{v1}})
  
  expect_equal(length(dfn$col_defs), 2)
  expect_equal(dfn$col_defs[[1]]$var, "mpg")
  expect_equal(dfn$col_defs[[2]]$var, "cyl")
  
  mytest <- function(myvar) {
    
    tbl1 <- create_table(mtcars)
    dfn1 <- define(tbl1, {{myvar}})
    return(dfn1$col_defs[[1]])
  }
  
  
  expect_equal(mytest("disp")$var, "disp")
  
})

test_that("stub escape works as expected.", {
  
  tbl <- create_table(mtcars)
  
  v1 <- c("mpg", "cyl")
  
  expect_error(stub(tbl, vars = v1))
  
  
  dfn <- stub(tbl, vars = {{v1}})
  
  expect_equal(length(dfn$stub$vars), 2)
  expect_equal(dfn$stub$vars[[1]], "mpg")
  expect_equal(dfn$stub$vars[[2]], "cyl")
  
  mytest <- function(myvar) {
    
    tbl1 <- create_table(mtcars)
    dfn1 <- stub(tbl1, {{myvar}})
    return(dfn1$stub$vars)
  }
  
  
  expect_equal(mytest("disp"), "disp")
  
})

test_that("spanning header escape works as expected.", {
  
  tbl <- create_table(mtcars)
  
  v1 <- "mpg"
  v2 <- "cyl"
  
  expect_error(spanning_header(tbl, from = v1, to = v2))
  
  sh <- spanning_header(tbl, from = {{v1}}, to = {{v2}})
  
  expect_equal(sh$col_spans[[1]]$from, "mpg")
  expect_equal(sh$col_spans[[1]]$to, "cyl")
  
  mytest <- function(myvar1, myvar2) {
    
    tbl1 <- create_table(mtcars)
    sh1 <- spanning_header(tbl1, from = {{myvar1}}, to = {{myvar2}})
    return(sh1$col_spans[[1]])
  }
  
  cs <- mytest("mpg", "disp")
  expect_equal(cs$from, "mpg")
  expect_equal(cs$to, "disp")
  
  v1 <- 1
  v2 <- 4
  
  tbl2 <- create_table(mtcars)
  sh2 <- spanning_header(tbl2, from = {{v1}}, to = {{v2}})
  
  expect_equal(sh2$col_spans[[1]]$from, "mpg")
  expect_equal(sh2$col_spans[[1]]$to, "hp")
  
  
})


test_that("column_defaults escape works as expected.", {
  
  tbl <- create_table(mtcars)
  
  v1 <- c("mpg", "cyl")
  
  expect_error(column_defaults(tbl, vars = v1))
  
  
  dfn <- column_defaults(tbl, vars = {{v1}})
  
  expect_equal(length(dfn$col_dflts[[1]]$vars), 2)
  expect_equal(dfn$col_dflts[[1]]$vars[[1]], "mpg")
  expect_equal(dfn$col_dflts[[1]]$vars[[2]], "cyl")
  
  mytest <- function(myvar) {
    
    tbl1 <- create_table(mtcars)
    dfn1 <- column_defaults(tbl1, {{myvar}})
    return(dfn1$col_dflts[[1]]$vars)
  }
  
  
  expect_equal(mytest("disp"), "disp")
  
  v1 <- "mpg"
  v2 <- "disp"
  
  tbl <- create_table(mtcars)
  dfn <- column_defaults(tbl, from = {{v1}}, to = {{v2}})
  
  expect_equal(dfn$col_dflts[[1]]$from, "mpg")
  expect_equal(dfn$col_dflts[[1]]$to, "disp")
  
  
})

test_that("standard_eval parameter works as expected for define.", {
  
  tmp <- "mpg"
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    define(tmp, format = "%.1f", standard_eval = TRUE) 
  
  expect_equal(length(tbl$col_defs), 1)
  expect_equal(tbl$col_defs[[1]]$var_c, "mpg")
  
})

test_that("standard_eval parameter works as expected for column_defaults", {
  
  tmp <- "mpg" 
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    column_defaults(tmp, width = 2, standard_eval = TRUE)

  
  expect_equal(length(tbl$col_dflts), 1)
  expect_equal(tbl$col_dflts[[1]]$vars, "mpg")
  
  
  tmp1 <- "mpg" 
  tmp2 <- "cyl"
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    column_defaults(from = tmp1, to = tmp2, width = 2, standard_eval = TRUE)
  
  
  expect_equal(length(tbl$col_dflts), 1)
  expect_equal(tbl$col_dflts[[1]]$from, "mpg")
  expect_equal(tbl$col_dflts[[1]]$to, "cyl")
  
  
  
})

test_that("standard_eval parameter works as expected for spanning_header.", {
  
  tmp1 <- "mpg" 
  tmp2 <- "cyl"
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    spanning_header(from = tmp1, to = tmp2, standard_eval = TRUE)
  
  
  expect_equal(length(tbl$col_spans), 1)
  expect_equal(tbl$col_spans[[1]]$from, "mpg")
  expect_equal(tbl$col_spans[[1]]$to, "cyl")
  
  
  
  
})

test_that("standard_eval parameter works as expected for stub", {
  
  tmp1 <- c("mpg", "cyl") 
  
  
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    stub(vars = tmp1, standard_eval = TRUE)
  
  
  expect_equal(length(tbl$stub), 3)
  expect_equal(tbl$stub$vars, c("mpg", "cyl"))
  
  
  
  
})

test_that("create_table borders parameter works as expected.", {
  
  
  expect_error(create_table(mtcars, borders = "fork"))
  
  tbl <- create_table(mtcars, borders = "all")
  
  expect_equal(tbl$borders, "all")
  
  tbl <- create_table(mtcars, borders = c("outside", "inside"))
  
  expect_equal(tbl$borders, c("outside", "inside"))
  
})

# The page template is everything except the content: page header/footer,
# titles, footnotes, etc.

# Page Template RTF Functions ---------------------------------------------


#' Create a page template with header, titles, footnotes, and footer.
#' @param rs The report spec
#' @return The page template object
#' @noRd
page_template_rtf<- function(rs) {
  
  pt <- structure(list(), class = c("page_template_rtf", "list"))
  
  pt$page_header <- get_page_header_rtf(rs)
  pt$title_hdr <- get_title_header_rtf(rs$title_hdr, rs$line_size, rs)
  pt$titles <- get_titles_rtf(rs$titles, rs$line_size, rs)
  pt$footnotes <- c()
  if (!is.null(rs$footnotes)) {
    if (!is.null(rs$footnotes[[1]])) {
      if (rs$footnotes[[1]]$valign == "bottom")
        pt$footnotes <- get_footnotes_rtf(rs$footnotes, rs$line_size, rs)
    }
    
  }
  pt$page_footer <- get_page_footer_rtf(rs)
  
  pt$lines <- sum(pt$page_header$lines, pt$page_footer$lines,
                  pt$title_hdr$lines, pt$titles$lines, pt$footnotes$lines)
  
  pt$twips <- sum(pt$page_header$twips, pt$page_footer$twips,
                  pt$title_hdr$twips, pt$titles$twips, pt$footnotes$twips)
  
  # Page by not here.  Messes up line counts.
  
  return(pt)
}

#' @import grDevices
#' @noRd
get_page_header_rtf <- function(rs) {
  
  ret <- ""
  cnt <- 0
  twps <- 0
  
  if ((!is.null(rs$header_image_left) & is.null(rs$page_header_left)) |
      (!is.null(rs$header_image_center) & is.null(rs$page_header_center)) |
      (!is.null(rs$header_image_right) & is.null(rs$page_header_right))) {
    
    stop("`page_header` must be used when using `header_image`.")
  }
  
  # Make sure the length is 3, NA will be imputed later.
  width <- c(rs$page_header_width, rep(NA, 3 - length(rs$page_header_width)))
  
  image_left <- FALSE
  image_center <- FALSE
  image_right <- FALSE
  
  if (!is.null(rs$header_image_left)) {
    image_left <- TRUE
    hl <- rs$header_image_left
  } else{
    hl <- rs$page_header_left
  }
  
  if (!is.null(rs$header_image_right)) {
    image_right <- TRUE
    hr <- rs$header_image_right
  } else {
    hr <- rs$page_header_right
  }
  
  if (!is.null(rs$header_image_center)) {
    image_center <- TRUE
    hc <- rs$header_image_center
    
    # Default center cell is not displayed, open it when picture exists
    if (width[2] == 0) {
      width[2] <- NA
    }
  } else {
    hc <- rs$page_header_center
  }
  
  conv <- rs$twip_conversion
  lh <- rs$row_height
  
  # # User controlled width of left column
  # lwdth <- rs$page_header_width
  # if (is.null(lwdth)) {
  #   lwdth <- rs$content_size[["width"]]/2
  # } else if (lwdth > rs$content_size[["width"]]) {
  #   stop(sprintf(
  #     "Total width of page header %s cannot be greater than content width %s.",
  #     lwdth,
  #     rs$content_size[["width"]]
  #   ))
  # }
  # 
  # # Calculate right column width
  # rwdth <- rs$content_size[["width"]] - lwdth
  
  
  # Calculate the widths
  total_width <- sum(width, na.rm = T)
  if (total_width > rs$content_size[["width"]]) {
    
    stop(sprintf("Total width of page header %s %s cannot be greater than content width %s %s.",
                 total_width,
                 rs$units,
                 rs$content_size[["width"]],
                 rs$units))
    
  } else {
    na_num <- sum(is.na(width))
    imputed_width <- (rs$content_size[["width"]] - total_width) / na_num
    
    left_width <- ifelse(is.na(width[1]), imputed_width, width[1])
    center_width <- ifelse(is.na(width[2]), imputed_width, width[2])
    right_width <- ifelse(is.na(width[3]), imputed_width, width[3])
    
    width <- c(left_width, center_width, right_width)
    
    # cellx must be an integer
    c1 <- round(left_width * conv)
    c2 <- round((left_width + center_width) * conv)
    c3 <- round((left_width + center_width + right_width) * conv)
    
    c_lst <- c(c1, c2, c3)
  }
  
  hl_num <- ifelse(image_left, length(hl$image_path), length(hl))
  hc_num <- ifelse(image_center, length(hc$image_path), length(hc))
  hr_num <- ifelse(image_right, length(hr$image_path) , length(hr))
  
  maxh <- max(hl_num, hc_num, hr_num)

  if (maxh > 0 | length(rs$header_titles) > 0) {
    
    fs <- rs$font_size * 2
    
    # c1 <- round(lwdth * conv) 
    # c2 <- round(rwdth * conv) + c1
    
    ret <- paste0("{\\header \\f0\\fs", fs)
    
    if (maxh > 0) {
      pdf(NULL)
      par(family = get_font_family(rs$font), ps = rs$font_size)
      
      max_twips <- 0
      for (i in seq(1, maxh)) {
        # ret <- paste0(ret, "\\trowd\\trgaph0\\trrh", lh, 
        #               "\\cellx", c1, "\\cellx", c2, "\\cellx", c3)
        
        ret <- paste0(ret, "\\trowd\\trgaph0\\trrh", lh)
        for (j in 1:length(width)){
          if (width[j] > 0) {
            ret <- paste0(ret, "\\cellx", c_lst[j])
          }
        }
        
        lcnt <- 0
        ccnt <- 0
        rcnt <- 0
        
        ltwips <- 0
        ctwips <- 0
        rtwips <- 0
        
        if (left_width > 0) {
          if (hl_num >= i) {
            
            # If header image is assigned, the header text will be ignored
            if (image_left == FALSE) {
              
              # Split strings if they exceed width
              tmp <- split_string_rtf(hl[[i]], left_width, rs$units)
              
              ret <- paste0(ret, "\\ql ", get_page_numbers_rtf(tmp$rtf), "\\cell")
              lcnt <- tmp$lines
              
            } else {
              
              # Get image RTF codes
              img <- get_image_rtf(hl$image_path[i], min(hl$width, left_width), hl$height, rs$units)
              
              ret <- paste0(ret, "\\ql ", img, "\\cell")
              
              # Calculate the total lines of image after summarize all twips to
              # prevent from rounding bias
              ltwips <- hl$height * rs$twip_conversion
            }
            
          } else {
            ret <- paste0(ret, "\\ql \\cell")
            lcnt <- 1 
          }
        }
        
        if (center_width > 0) {
          if (hc_num >= i) {
            
            # If header image is assigned, the header text will be ignored
            if (image_center == FALSE) {
              
              # Split strings if they exceed width
              tmp <- split_string_rtf(hc[[i]], center_width, rs$units)
              
              ret <- paste0(ret, "\\qc ", get_page_numbers_rtf(tmp$rtf), "\\cell")
              ccnt <- tmp$lines
              
            } else {
              
              # Get image RTF codes
              img <- get_image_rtf(hc$image_path[i], min(hc$width, center_width), hc$height, rs$units)
              
              ret <- paste0(ret, "\\qc ", img, "\\cell")
              
              ctwips <- hc$height * rs$twip_conversion
            }
            
          } else {
            ret <- paste0(ret, "\\qc \\cell")
            ccnt <- 1 
          }
        }
        
        if (right_width > 0) {
          if (hr_num >= i) {
            
            # If header image is assigned, the header text will be ignored
            if (image_right == FALSE) {
              
              # Split strings if they exceed width
              tmp2 <- split_string_rtf(hr[[i]], right_width, rs$units)
              
              ret <- paste0(ret, "\\qr ", get_page_numbers_rtf(tmp2$rtf), "\\cell")
              rcnt <- tmp2$lines
              
            } else {
              
              # Get image RTF codes
              img <- get_image_rtf(hr$image_path[i], min(hr$width, right_width), hr$height, rs$units)
              
              ret <- paste0(ret, "\\qr ", img, "\\cell")
              
              rtwips <- hr$height * rs$twip_conversion
            }
            
          } else {
            ret <- paste0(ret, "\\qr \\cell")
            rcnt <- 1
          }
        }
        
        ret <- paste0(ret, "\\row\n")
        
        # if (lcnt > rcnt)
        #   cnt <- cnt + lcnt
        # else 
        #   cnt <- cnt + rcnt
        max_twips <- max_twips + max(ltwips, ctwips, rtwips)
        cnt <- cnt + max(lcnt, ccnt, rcnt)
      }
      
      dev.off()

      # Calculate total lines of pictures
      if (max_twips > 0) {
        cnt <- max(cnt, round(max_twips/lh))
      }
    
      if (rs$page_header_blank_row == "below") {
        ret <- paste0(ret, "\\par\\pard", rs$font_rtf, rs$spacing_multiplier)
        cnt <- cnt + 1
      } else {
        
        ret <- paste0(ret, "\\fs1\\sl0\\par\\pard", rs$font_rtf, 
                      rs$spacing_multiplier) 
      }
    }
    
    htitles <- ""
    if (!is.null(rs$header_titles)) {
    
      tret <- get_titles_rtf(rs$header_titles, rs$content_size[["width"]], rs)
      htitles <- paste0(tret$rtf, "\\fs1\\sl0\\par\\pard", rs$font_rtf, 
                                              rs$spacing_multiplier) 
      cnt <- cnt + tret$lines
    
    }
    
    ret <- paste0(ret, htitles, "}")
  }
  
  twps <- cnt * lh
  
  res <- list(rtf = ret, lines = cnt, twips = twps)
  
  return(res)
}

#' @import grDevices
#' @noRd
get_page_footer_rtf <- function(rs) {
  
  ret <- ""
  cnt <- 0
  twps <- 0
  
  if ((!is.null(rs$footer_image_left) & is.null(rs$page_footer_left)) |
      (!is.null(rs$footer_image_right) & is.null(rs$page_footer_right)) |
      (!is.null(rs$footer_image_center) & is.null(rs$page_footer_center))) {
    
    stop("`page_footer` must be used when using `footer_image`.")
  }
  
  image_left <- FALSE
  image_center <- FALSE
  image_right <- FALSE
  
  if (!is.null(rs$footer_image_left)) {
    fl <- rs$footer_image_left
    image_left <- TRUE
  } else {
    fl <- rs$page_footer_left
  }
  
  if (!is.null(rs$footer_image_center)) {
    fc <- rs$footer_image_center
    image_center <- TRUE
  } else {
    fc <- rs$page_footer_center
  }
  
  if (!is.null(rs$footer_image_right)) {
    fr <- rs$footer_image_right
    image_right <- TRUE
  } else {
    fr <- rs$page_footer_right
  }
  
  conv <- rs$twip_conversion
  lh <- rs$row_height
  
  fl_num <- ifelse(image_left, length(fl$image_path), length(fl))
  fc_num <- ifelse(image_center, length(fc$image_path), length(fc))
  fr_num <- ifelse(image_right, length(fr$image_path) , length(fr))
  
  maxf <- max(fl_num, fc_num, fr_num)
  
  if (maxf > 0 | length(rs$footer_footnotes) > 0) {
    
    fs <- rs$font_size * 2
    
    width <- rs$page_footer_width
    
    # Make sure the length is 3, NA will be imputed later.
    width <- c(width, rep(NA, 3 - length(width)))
    
    total_width <- sum(width, na.rm = T)
    if (total_width > rs$content_size[["width"]]) {
      
      stop(sprintf("Total width of page footer %s %s cannot be greater than content width %s %s.",
                   total_width,
                   rs$units,
                   rs$content_size[["width"]],
                   rs$units))
      
    } else {
      na_num <- sum(is.na(width))
      imputed_width <- (rs$content_size[["width"]] - total_width) / na_num
      
      left_width <- ifelse(is.na(width[1]), imputed_width, width[1])
      center_width <- ifelse(is.na(width[2]), imputed_width, width[2])
      right_width <- ifelse(is.na(width[3]), imputed_width, width[3])
      
      width <- c(left_width, center_width, right_width)
      
      # cellx must be an integer
      c1 <- round(left_width * conv)
      c2 <- round((left_width + center_width) * conv)
      c3 <- round((left_width + center_width + right_width) * conv)
      
      c_lst <- c(c1, c2, c3)
    }
    
    
    ret <- paste0("{\\footer \\f0\\fs", fs, "[ff]")
    
    if (maxf > 0) {
      pdf(NULL)
      par(family = get_font_family(rs$font), ps = rs$font_size)
      
      max_twips <- 0
      for (i in seq(1, maxf)) {
        
        # ret <- paste0(ret, "\\trowd\\trgaph0\\cellx", c1, 
        #               "\\cellx", c2 , "\\cellx", c3)
        ret <- paste0(ret, "\\trowd\\trgaph0")
        for (j in 1:length(width)){
          if (width[j] > 0) {
            ret <- paste0(ret, "\\cellx", c_lst[j])
          }
        }
        
        lcnt <- 0
        ccnt <- 0
        rcnt <- 0
        
        ltwips <- 0
        ctwips <- 0
        rtwips <- 0
        
        if (left_width > 0) {
          if (fl_num >= i) {
            
            # If header image is assigned, the header text will be ignored
            if (image_left == FALSE) {
              
              # Split strings if they exceed width
              tmp1 <- split_string_rtf(fl[[i]], left_width, rs$units)
              
              ret <- paste0(ret, "\\ql ", get_page_numbers_rtf(tmp1$rtf), "\\cell")
              lcnt <- tmp1$lines
              
            } else {
              
              # Get image RTF codes
              img <- get_image_rtf(fl$image_path[i], min(fl$width, left_width), fl$height, rs$units)
              
              ret <- paste0(ret, "\\ql ", img, "\\cell")
              
              # Calculate the total lines of image after summarize all twips to
              # prevent from rounding bias
              ltwips <- fl$height * rs$twip_conversion
            }
          } else {
            ret <- paste0(ret, "\\ql \\cell")
            lcnt <- 1
          }
        }
        
        if (center_width > 0) {
          if (fc_num >= i) {
            
            # If header image is assigned, the header text will be ignored
            if (image_center == FALSE) {
              
              # Split strings if they exceed width
              tmp2 <- split_string_rtf(fc[[i]], center_width, rs$units)
              
              ret <- paste0(ret, "\\qc ", get_page_numbers_rtf(tmp2$rtf), "\\cell")
              ccnt <- tmp2$lines
              
            } else {
              
              # Get image RTF codes
              img <- get_image_rtf(fc$image_path[i], min(fc$width, center_width), fc$height, rs$units)
              
              ret <- paste0(ret, "\\qc ", img, "\\cell")
              
              ctwips <- fc$height * rs$twip_conversion
            }
          } else {
            ret <- paste0(ret, "\\qc \\cell")
            ccnt <- 1
          }
        }
        
        if (right_width > 0) {
          if (fr_num >= i) {
            
            # If header image is assigned, the header text will be ignored
            if (image_right == FALSE) {
              
              # Split strings if they exceed width
              tmp3 <- split_string_rtf(fr[[i]], right_width, rs$units)
              
              ret <- paste0(ret, "\\qr ", get_page_numbers_rtf(tmp3$rtf), "\\cell")
              rcnt <- tmp3$lines
              
            } else {
              
              # Get image RTF codes
              img <- get_image_rtf(fr$image_path[i], min(fr$width, right_width), fr$height, rs$units)
              
              ret <- paste0(ret, "\\qr ", img, "\\cell")
              
              rtwips <- fr$height * rs$twip_conversion
            }
          } else {
            ret <- paste0(ret, "\\qr \\cell")
            rcnt <- 1
          }
        }
        
        ret <- paste0(ret, "\\row\n")
        max_twips <- max_twips + max(ltwips, ctwips, rtwips)
        cnt <- cnt + max(lcnt, ccnt, rcnt)
        
      }
      dev.off()
      
      if (max_twips > 0) {
        cnt <- max(cnt, round(max_twips/lh))
      }
    }
    
    
    ffootnotes <- ""
    if (!is.null(rs$footer_footnotes)) {
      
      tret <- get_footnotes_rtf(rs$footer_footnotes, rs$content_size[["width"]], rs)
      ffootnotes <- paste0(tret$rtf, "\\fs1\\sl0\\par\\pard", rs$font_rtf, 
                        rs$spacing_multiplier) 
      cnt <- cnt + tret$lines
      
    }
    
    ret <- sub("[ff]", ffootnotes, ret, fixed = TRUE)
    
    ret <- paste0(ret, "\\fs1\\sl0\\par\\pard", 
                  rs$font_rtf, rs$spacing_multiplier, "}")
  }
  

  res <- list(rtf = paste0(ret, collapse = ""),
              lines = cnt, 
              twips = cnt * lh)
  
  return(res)
}

#' @import grDevices
#' @noRd
get_titles_rtf <- function(ttllst, content_width, rs, talgn = "center") {
  
  ret <- c()
  cnt <- 0
  twps <- 0
  
  conv <- rs$twip_conversion
  lh <- rs$row_height
  border_flag <- FALSE

  
  ta <- "\\trql"
  if (talgn == "right")
    ta <- "\\trqr"
  else if (talgn %in% c("center", "centre"))
    ta <- "\\trqc"

  if (length(ttllst) > 0) {
    
    for (ttls in ttllst) {
      
      cols <- ttls$columns
      
      
      if (ttls$width == "page")
        width <- rs$content_size[["width"]] 
      else if (ttls$width == "content") 
        width <- content_width 
      else if (is.numeric(ttls$width))
        width <- ttls$width 
        
      w <- round(width * conv)
      
      cwidth <- width / cols
      cw <- round(w / cols)
      

      if (ttls$align == "center")
        algn <- "\\qc"
      else if (ttls$align == "right")
        algn <- "\\qr"
      else 
        algn <- "\\ql"
      
      border_flag <- FALSE
      alcnt <- 0
      blcnt <- 0
      
      # Open device context
      pdf(NULL)
      
      # Set point size (ps) for strwidth to calculate string width
      if (!is.null(ttls$font_size)) {
        ttlfs <- ttls$font_size
      } else {
        ttlfs <- rs$font_size
      }
      par(family = get_font_family(rs$font), ps = ttlfs)
      
      
      al <- ""
      # Get blank row above
      if (any(ttls$blank_row %in% c("above", "both"))) {
        
        alcnt <- 1
        
        tb <- get_cell_borders(1, 1, length(ttls$titles) + alcnt, 1, ttls$borders)
        
        al <- paste0("\\trowd\\trgaph0", ta, tb, "\\cellx", w, 
                     algn, "\\cell\\row\n")
        cnt <- cnt + 1 
        
      }
      
      bl <- ""
      # Get blank row below
      if (any(ttls$blank_row %in% c("below", "both"))) {
        blcnt <- 1
        
        tb <- get_cell_borders(length(ttls$titles) + alcnt + blcnt, 1, 
                               length(ttls$titles) + alcnt + blcnt, 
                               1, ttls$borders)
        
        sm <- get_spacing_multiplier(rs$font_size)
        
        bl <- paste0("\\trowd\\trgaph0", ta, tb, "\\cellx", w, 
                     algn, sm, "\\cell\\row\n")
        cnt <- cnt + 1
      }
      
      # Append blank row above
      if (al != "")
        ret <- append(ret, al)
      
      fz <- ""
      fs <- ""
      # Get font size
      if (!is.null(ttls$font_size)) {
        
        fz <- paste0("\\fs", ttls$font_size * 2, 
                     get_spacing_multiplier(ttls$font_size)) 
        fs <- paste0("\\fs", rs$font_size * 2)
      }
      
      # Reset column width accumulation
      cwa <- 0
      
      # Calculate total number of rows in this block
      rws <- ceiling(length(ttls$titles) / cols) + alcnt + blcnt

      i <- 1
      while (i <= length(ttls$titles)) {    
        
        mxlns <- 0
        
        # Calculate current row
        rw <- ceiling(i / cols)

        # Start row
        ret <- append(ret, paste0("\\trowd\\trgaph0", ta))
        
        for (j in seq_len(cols)) {
          
          # Get border specs for this cell
          b <- get_cell_borders(rw + alcnt, j, 
                                rws, 
                                cols, ttls$borders)
          
          # Not all cells have titles
          if (i > length(ttls$titles))
            vl <- ""
          else 
            vl <- ttls$titles[[i]]
          
          # Deal with column alignments
          if (cols == 1) {
            calgn <- algn 
          } else if (cols == 2) {
           if (j == 1)
             calgn <- "\\ql"
           else 
             calgn <- "\\qr"
          } else if (cols == 3) {
           if (j == 1)
             calgn <- "\\ql"
           else if (j == 2)
             calgn <- "\\qc"
           else if (j == 3) 
             calgn <- "\\qr"
          }
          
          if (j == 1) 
            cwa <- 0
        
          # RTF cell widths are absolute ending points
          cwa <- cwa + cw
          
          # Split title strings if they exceed width
          tmp <- split_string_rtf(vl, cwidth, rs$units)
          
          # Track max lines for counting
          if (tmp$lines > mxlns)
            mxlns <- tmp$lines

          # Add bold if needed
          tb <- get_page_numbers_rtf(tmp$rtf, FALSE)
          if (ttls$bold)
            tb <- paste0("\\b ", get_page_numbers_rtf(tmp$rtf, FALSE), "\\b0")
          
          # Construct cell from constituent parts
          ret <- append(ret, paste0(b, "\\cellx", cwa, 
                                    calgn, fz, " ", tb, fs, "\\cell"))
          
          i <- i + 1
          
        }
        
        # End row
        ret <- append(ret, "\\row\n")

        # Track lines
        cnt <- cnt + mxlns
      }
      
      # Append blank row below
      if (bl != "")
        ret <- append(ret, bl)
      
      if (any(ttls$borders %in% c("outside", "all", "bottom")))
        border_flag <- TRUE
      
      dev.off()
    
    }
    
  }
  
  res <- list(rtf = paste0(ret, collapse = ""), 
              lines = cnt, 
              twips = cnt * lh,
              border_flag = border_flag)
  
  return(res)
}



#' @import grDevices
#' @noRd
get_titles_rtf_back <- function(ttllst, content_width, rs, talgn = "center") {
  
  ret <- c()
  cnt <- 0
  twps <- 0
  
  conv <- rs$twip_conversion
  lh <- rs$row_height
  border_flag <- FALSE
  
  
  ta <- "\\trql"
  if (talgn == "right")
    ta <- "\\trqr"
  else if (talgn %in% c("center", "centre"))
    ta <- "\\trqc"
  
  if (length(ttllst) > 0) {
    
    for (ttls in ttllst) {
      
      
      if (ttls$width == "page")
        width <- rs$content_size[["width"]]
      else if (ttls$width == "content")
        width <- content_width
      else if (is.numeric(ttls$width))
        width <- ttls$width
      
      w <- round(width * conv)
      
      
      if (ttls$align == "center")
        algn <- "\\qc"
      else if (ttls$align == "right")
        algn <- "\\qr"
      else 
        algn <- "\\ql"
      
      border_flag <- FALSE
      alcnt <- 0
      blcnt <- 0
      
      # Open device context
      pdf(NULL)
      par(family = get_font_family(rs$font), ps = rs$font_size)
      
      for (i in seq_along(ttls$titles)) {
        
        
        
        al <- ""
        
        if (i == 1) {
          if (any(ttls$blank_row %in% c("above", "both"))) {
            
            alcnt <- 1
            
            tb <- get_cell_borders(i, 1, length(ttls$titles) + alcnt, 1, ttls$borders)
            
            al <- paste0("\\trowd\\trgaph0", ta, tb, "\\cellx", w, 
                         algn, "\\cell\\row\n")
            cnt <- cnt + 1 
            
          }
        }
        
        bl <- ""
        if (i == length(ttls$titles)) {
          if (any(ttls$blank_row %in% c("below", "both"))) {
            blcnt <- 1
            
            tb <- get_cell_borders(i + alcnt + blcnt, 1, 
                                   length(ttls$titles) + alcnt + blcnt, 
                                   1, ttls$borders)
            
            sm <- get_spacing_multiplier(rs$font_size)
            
            bl <- paste0("\\trowd\\trgaph0", ta, tb, "\\cellx", w, 
                         algn, sm, "\\cell\\row\n")
            cnt <- cnt + 1
          }
          
          if (any(ttls$borders %in% c("outside", "all", "bottom")))
            border_flag <- TRUE
        }
        
        b <- get_cell_borders(i + alcnt, 1, 
                              length(ttls$titles) + alcnt + blcnt, 
                              1, ttls$borders)
        
        # Split title strings if they exceed width
        tmp <- split_string_rtf(ttls$titles[[i]], width, rs$units)
        
        fz <- ""
        fs <- ""
        if (!is.null(ttls$font_size)) {
          
          fz <- paste0("\\fs", ttls$font_size * 2, 
                       get_spacing_multiplier(ttls$font_size)) 
          fs <- paste0("\\fs", rs$font_size * 2)
        }
        
        
        tb <- tmp$rtf
        if (ttls$bold)
          tb <- paste0("\\b ", tmp$rtf, "\\b0")
        
        # Concatenate title string
        if (al != "")
          ret <- append(ret, al)
        ret <- append(ret, paste0("\\trowd\\trgaph0", ta, b, "\\cellx", w, 
                                  algn, fz, " ", tb, fs, "\\cell\\row\n"))
        if (bl != "")
          ret <- append(ret, bl)
        
        cnt <- cnt + tmp$lines
      }
      dev.off()
      
    }
    
  }
  
  res <- list(rtf = paste0(ret, collapse = ""), 
              lines = cnt, 
              twips = cnt * lh,
              border_flag = border_flag)
  
  return(res)
}

#' @import grDevices
#' @noRd
get_footnotes_rtf <- function(ftnlst, content_width, rs, talgn = "center") {
  
  ret <- c()
  cnt <- 0
  twps <- 0
  border_flag <- FALSE
  
  conv <- rs$twip_conversion
  lh <- rs$row_height
  
  ta <- "\\trql"
  if (talgn == "right")
    ta <- "\\trqr"
  else if (talgn %in% c("center", "centre"))
    ta <- "\\trqc"
  
  if (length(ftnlst) > 0) {
    
    for (ftnts in ftnlst) {
      
      cols <- ftnts$columns
      
      if (ftnts$width == "page")
        width <- rs$content_size[["width"]]
      else if (ftnts$width == "content")
        width <- content_width
      else if (is.numeric(ftnts$width))
        width <- ftnts$width
      
      w <- round(width * conv)
      
      cwidth <- width / cols
      cw <- round(w / cols)
      
      if (ftnts$align == "center")
        algn <- "\\qc"
      else if (ftnts$align == "right")
        algn <- "\\qr"
      else 
        algn <- "\\ql"
      
      alcnt <- 0
      blcnt <- 0
      border_flag <- FALSE
      
      pdf(NULL)
      
      # Set point size (ps) for strwidth to calculate string width
      if (!is.null(ftnts$font_size)) {
        ftntfs <- ftnts$font_size
      } else {
        ftntfs <- rs$font_size
      }
      par(family = get_font_family(rs$font), ps = ftntfs)
      
      al <- ""
      # Get blank row above
      if (any(ftnts$blank_row %in% c("above", "both"))) {
        
        alcnt <- 1
        
        tb <- get_cell_borders(1, 1, length(ftnts$footnotes) + alcnt, 
                               1, ftnts$borders)
        
        al <- paste0("\\trowd\\trgaph0", ta, tb, "\\cellx", w, 
                     algn, "\\cell\\row\n")
        cnt <- cnt + 1 
        
      }
      
      
      bl <- ""
      # Get blank row below
      if (any(ftnts$blank_row %in% c("below", "both"))) {
        blcnt <- 1
        
        tb <- get_cell_borders(length(ftnts$footnotes) + alcnt + blcnt, 1, 
                               length(ftnts$footnotes) + alcnt + blcnt, 
                               1, ftnts$borders)
        
        bl <- paste0("\\trowd\\trgaph0", ta, tb, "\\cellx", w, 
                     algn, "\\cell\\row\n")
        cnt <- cnt + 1
      }
      if (any(ftnts$borders %in% c("outside", "all", "top")))
        border_flag <- TRUE
      
      # Append blank row above
      if (al != "")
        ret <- append(ret, al)
      
      fz <- ""
      fs <- ""
      # Get font size - Not used yet
      if (!is.null(ftnts$font_size)) {
        
        fz <- paste0("\\fs", ftnts$font_size * 2, 
                     get_spacing_multiplier(ftnts$font_size)) 
        fs <- paste0("\\fs", rs$font_size * 2)
      }
      
      # Reset column width accumulation
      cwa <- 0
      
      # Calculate total number of rows in this block
      rws <- ceiling(length(ftnts$footnotes) / cols) + alcnt + blcnt
      
      i <- 1
      while (i <= length(ftnts$footnotes)) {    
        
        mxlns <- 0
        
        # Calculate current row
        rw <- ceiling(i / cols)
        
        # Start row
        ret <- append(ret, paste0("\\trowd\\trgaph0", ta))
        
        for (j in seq_len(cols)) {
          
          # Get border specs for this cell
          # b <- get_cell_borders(i + alcnt, 1, 
          #                       length(ftnts$footnotes) + alcnt + blcnt, 
          #                       1, ftnts$borders)
          b <- get_cell_borders(rw + alcnt, j, 
                                rws, 
                                cols, ftnts$borders)
          
          # Not all cells have titles
          if (i > length(ftnts$footnotes)) {
            vl <- ""
          } else { 
            vl <- ftnts$footnotes[[i]]
          }
          
          # Deal with column alignments
          if (cols == 1) {
            calgn <- algn 
          } else if (cols == 2) {
            if (j == 1)
              calgn <- "\\ql"
            else 
              calgn <- "\\qr"
          } else if (cols == 3) {
            if (j == 1)
              calgn <- "\\ql"
            else if (j == 2)
              calgn <- "\\qc"
            else if (j == 3) 
              calgn <- "\\qr"
          }
          
          if (j == 1) 
            cwa <- 0
          
          # RTF cell widths are absolute ending points
          cwa <- cwa + cw
          
          
          # Split footnote strings if they exceed width
          tmp <- split_string_rtf(vl, cwidth, rs$units)
        
          # Track max lines for counting
          if (tmp$lines > mxlns)
            mxlns <- tmp$lines
          
          # Add italics if requested
          if (ftnts$italic)
            txt <- paste0("\\i ", get_page_numbers_rtf(tmp$rtf, FALSE), "\\i0") 
          else 
            txt <- paste0(" ", get_page_numbers_rtf(tmp$rtf, FALSE))
          
          # Concat footnote row
          # ret <- append(ret, paste0("\\trowd\\trgaph0", ta, b, "\\cellx", w, 
          #                           algn, txt, 
          #                           "\\cell\\row\n"))
          
          # Construct cell from constituent parts
          ret <- append(ret, paste0(b, "\\cellx", cwa, 
                                    calgn, fz, txt, fs, "\\cell"))
          
          i <- i + 1
          
        }
        
        # End row
        ret <- append(ret, "\\row\n")
        
        # Track lines
        cnt <- cnt + mxlns
        
      }
      
      if (bl != "")
        ret <- append(ret, bl)
      
      # Do I need this?
      # if (any(ftnts$borders %in% c("outside", "all", "bottom")))
      #   border_flag <- TRUE
      
      dev.off()
      
      
    }
    
  }
  
  
  res <- list(rtf = paste0(ret, collapse = ""),
              lines = cnt, 
              twips = cnt * lh,
              border_flag = border_flag)
  
  return(res)
}

#' @import grDevices
#' @noRd
get_footnotes_rtf_back <- function(ftnlst, content_width, rs, talgn = "center") {
  
  ret <- c()
  cnt <- 0
  twps <- 0
  border_flag <- FALSE
  
  conv <- rs$twip_conversion
  lh <- rs$row_height
  
  ta <- "\\trql"
  if (talgn == "right")
    ta <- "\\trqr"
  else if (talgn %in% c("center", "centre"))
    ta <- "\\trqc"
  
  if (length(ftnlst) > 0) {
    
    for (ftnts in ftnlst) {
      
      if (ftnts$width == "page")
        width <- rs$content_size[["width"]]
      else if (ftnts$width == "content")
        width <- content_width
      else if (is.numeric(ftnts$width))
        width <- ftnts$width
      
      w <- round(width * conv)
      
      
      if (ftnts$align == "center")
        algn <- "\\qc"
      else if (ftnts$align == "right")
        algn <- "\\qr"
      else 
        algn <- "\\ql"
      
      alcnt <- 0
      blcnt <- 0
      border_flag <- FALSE
      
      pdf(NULL)
      par(family = get_font_family(rs$font), ps = rs$font_size)
      
      for (i in seq_along(ftnts$footnotes)) {
        
        
        al <- ""
        if (i == 1) {
          if (any(ftnts$blank_row %in% c("above", "both"))) {
            
            alcnt <- 1
            
            tb <- get_cell_borders(i, 1, length(ftnts$footnotes) + alcnt, 
                                   1, ftnts$borders)
            
            al <- paste0("\\trowd\\trgaph0", ta, tb, "\\cellx", w, 
                         algn, "\\cell\\row\n")
            cnt <- cnt + 1 
            
          }
        }
        
        bl <- ""
        if (i == length(ftnts$footnotes)) {
          if (any(ftnts$blank_row %in% c("below", "both"))) {
            blcnt <- 1
            
            tb <- get_cell_borders(i + alcnt + blcnt, 1, 
                                   length(ftnts$footnotes) + alcnt + blcnt, 
                                   1, ftnts$borders)
            
            bl <- paste0("\\trowd\\trgaph0", ta, tb, "\\cellx", w, 
                         algn, "\\cell\\row\n")
            cnt <- cnt + 1
          }
          if (any(ftnts$borders %in% c("outside", "all", "top")))
            border_flag <- TRUE
        }
        
        b <- get_cell_borders(i + alcnt, 1, 
                              length(ftnts$footnotes) + alcnt + blcnt, 
                              1, ftnts$borders)
        
        
        
        # Split footnote strings if they exceed width
        tmp <- split_string_rtf(ftnts$footnotes[[i]], width, rs$units)
        
        if (al != "")
          ret <- append(ret, al)
        
        if (ftnts$italic)
          txt <- paste0("\\i ", get_page_numbers_rtf(tmp$rtf, FALSE), "\\i0") 
        else 
          txt <- paste0(" ", get_page_numbers_rtf(tmp$rtf, FALSE))
        
        # Concat footnote row
        ret <- append(ret, paste0("\\trowd\\trgaph0", ta, b, "\\cellx", w, 
                                  algn, txt, 
                                  "\\cell\\row\n"))
        if (bl != "")
          ret <- append(ret, bl)
        
        cnt <- cnt + tmp$lines
      }
      dev.off()
      

    }
    
  }
  
  
  res <- list(rtf = paste0(ret, collapse = ""),
              lines = cnt, 
              twips = cnt * lh,
              border_flag = border_flag)
  
  return(res)
}

#' @import grDevices
#' @noRd
get_title_header_rtf <- function(thdrlst, content_width, rs, talgn = "center") {
  
  ret <- c()
  cnt <- 0
  twps <- 0
  border_flag <- FALSE 
  
  conv <- rs$twip_conversion

  lh <- rs$row_height
  
  ta <- "\\trql"
  if (talgn == "right")
    ta <- "\\trqr"
  else if (talgn %in% c("center", "centre"))
    ta <- "\\trqc"
  
  if (length(thdrlst) > 0) {
    
    for (ttlhdr in thdrlst) {
      
      if (ttlhdr$width == "page")
        width <- rs$content_size[["width"]]
      else if (ttlhdr$width == "content")
        width <- content_width
      else if (is.numeric(ttlhdr$width))
        width <- ttlhdr$width
      
      w1 <- round(width * conv)
      w2 <- round(width * .7 * conv)
      
      mx <- max(length(ttlhdr$titles), length(ttlhdr$right))
    
      alcnt <- 0
      blcnt <- 0
      border_flag <- FALSE
      
      pdf(NULL)
      par(family = get_font_family(rs$font), ps = rs$font_size)
      
      for(i in seq_len(mx)) {
        
        
        
        al <- ""
        if (i == 1) {
          if (any(ttlhdr$blank_row %in% c("above", "both"))) {
            
            alcnt <- 1
            
            tb <- get_cell_borders(i, 1, mx + alcnt, 
                                   1, ttlhdr$borders)
            
            al <- paste0("\\trowd\\trgaph0", ta, tb, "\\cellx", w1, 
                         "\\ql\\cell\\row\n")
            cnt <- cnt + 1 
            
          }
        }
        
        bl <- ""
        if (i == mx) {
          if (any(ttlhdr$blank_row %in% c("below", "both"))) {
            blcnt <- 1
            
            tb <- get_cell_borders(i + alcnt + blcnt, 1, 
                                   mx + alcnt + blcnt, 
                                   1, ttlhdr$borders)
            
            bl <- paste0("\\trowd\\trgaph0", ta, tb, "\\cellx", w1, 
                         "\\ql\\cell\\row\n")
            cnt <- cnt + 1
          }
          
          if (any(ttlhdr$borders %in% c("all", "outside", "bottom")))
            border_flag <- TRUE
        }
        
        
      
        if (length(ttlhdr$titles) >= i) {
          # Split strings if they exceed width
          tmp1 <- split_string_rtf(ttlhdr$titles[[i]], width * .7, rs$units)
          ttl <- tmp1$rtf
          tcnt <- tmp1$lines
        } else {
          ttl <- ""
          tcnt <- 1 
        }
        
        if (length(ttlhdr$right) >= i) {
          tmp2 <- split_string_rtf(ttlhdr$right[[i]],
                                   width * .3, rs$units)
          hdr <- get_page_numbers_rtf(tmp2$rtf, FALSE)
          hcnt <- tmp2$lines
        } else {
          hdr <- ""
          hcnt <- 1
        }
        
        b1 <- get_cell_borders(i + alcnt, 1, mx + alcnt + blcnt, 2, ttlhdr$borders)
        b2 <- get_cell_borders(i + alcnt, 2, mx+ alcnt + blcnt, 2, ttlhdr$borders)
        
        
        if (al != "")
          ret <- append(ret, al)
        
        ret <- append(ret, paste0("\\trowd\\trgaph0", ta, b1, "\\cellx", w2, 
                                  b2, "\\cellx", w1,
                                  "\\ql ", ttl, "\\cell\\qr ", 
                                  hdr, "\\cell\\row\n"))
        if (bl != "")
          ret <- append(ret, bl)
        
        if (tcnt > hcnt)
          cnt <- cnt + tcnt
        else 
          cnt <- cnt + hcnt
      }
      
      dev.off()
      
    }

  }
  
  res <- list(rtf = paste0(ret, collapse = ""),
              lines = cnt,
              twips = cnt * lh,
              border_flag = border_flag)
  
  return(res)
}


#' Get page by text strings suitable for printing
#' @import stringi
#' @param titles Page by object
#' @param width The width to set the page by strings to
#' @return A vector of strings
#' @noRd
get_page_by_rtf <- function(pgby, width, value, rs, talgn, pgby_cnt = NULL) {
  
  if (is.null(width)) {
    stop("width cannot be null.") 
    
  }
  
  if (is.null(value))
    value <- get_pgby_value(value, pgby_cnt)
  
  ll <- width
  ret <- c()
  cnt <- 0
  border_flag <- FALSE
  
  ta <- "\\trql"
  if (talgn == "right")
    ta <- "\\trqr"
  else if (talgn %in% c("center", "centre"))
    ta <- "\\trqc"
  
  if (!is.null(pgby)) { 
    
    if (!any(class(pgby) == "page_by"))
      stop("pgby parameter value is not a page_by.")
    
    
    w1 <- round(width * rs$twip_conversion)
    
    algn <- "\\ql"
    if (pgby$align == "right")
      algn <- "\\qr"
    else if (pgby$align %in% c("center", "centre"))
      algn <- "\\qc"
    
    trows <- 1
    brow <- 1
    if (pgby$blank_row %in% c("above", "both")) {
      trows <- trows + 1
      brow <- 2
    }
    if (pgby$blank_row %in% c("below", "both"))
      trows <- trows + 1
    
    if (pgby$blank_row %in% c("above", "both")) {
      
      tb <- get_cell_borders(1, 1, trows, 1, pgby$borders)

      ret[length(ret) + 1] <- paste0("\\trowd\\trgaph0", ta, tb, 
                                     "\\cellx", w1, algn, 
                                  "\\cell\\row\n")
      cnt <- cnt + 1 
    }
    
    tb <- get_cell_borders(brow, 1 , trows, 1, pgby$borders)
    
    pdf(NULL)
    par(family = get_font_family(rs$font), ps = rs$font_size)
    
    # Account for multiple pgby lines
    # tmp <- split_string_rtf(value, width, rs$units)
    # vl <- tmp$rtf
    # cnt <- cnt + tmp$lines
    
    # Add bold
    if (pgby$bold %in% c(TRUE, FALSE)) {
      
      if (substr(pgby$label, nchar(pgby$label), nchar(pgby$label)) != " "){
        sep <- " "
      } else {
        sep <- ""
      }
      
      tmp <- split_string_rtf(paste0(pgby$label, sep, value), width, rs$units)
      vl <- tmp$rtf
      cnt <- cnt + tmp$lines
      
      if (pgby$bold ) {
        page_by_text <- paste0("\\b ", vl, "\\b0")
      } else {
        page_by_text <- vl
      }
      
    } else if (pgby$bold %in% c("value", "label")) {
      
      # Split label
      label_split <- split_string_rtf(pgby$label, width, rs$units)
      cnt <- cnt + label_split$lines
      
      # Use remain width to split value
      remain_width <- width - label_split$widths[length(label_split$widths)]
      value_split <- split_string_rtf(value, remain_width, rs$units)
      
      if (value_split$widths[1] > remain_width) {
        
        # If first width is bigger than remaining width, it means value starts a new line
        value_split <- split_string_rtf(value, width, rs$units)
        cnt <- cnt + value_split$lines
        value_split_txt <- value_split$rtf
        
      } else {
        
        # Otherwise, split with full width if there is a second line
        splt <- strsplit(value_split$rtf, split = "\\line", fixed = TRUE)
        
        if (length(splt[[1]]) > 1) {
          remain_value <- trimws(sub(splt[[1]][1], "", value), which = "left")
          remain_value_split <- split_string_rtf(remain_value, width, rs$units)
          cnt <- cnt + remain_value_split$lines
          value_split_txt <- paste0(splt[[1]][1], "\\line", remain_value_split$rtf)
        } else {
          value_split_txt <- value_split$rtf
        }
        
      }
      
      if (pgby$bold == "label") {
        page_by_text <- paste0("\\b ", label_split$rtf, " \\b0", value_split_txt)
      } else {
        page_by_text <- paste0(label_split$rtf, " \\b ", value_split_txt, "\\b0")
      }
    }
    
    dev.off()
    
    # Construct RTF for pageby value
    ret[length(ret) + 1] <- paste0("\\trowd\\trgaph0", ta, tb, 
                                   "\\cellx", w1, algn, " ",
                                   page_by_text, "\\cell\\row\n")
    

    # cnt <- cnt + get_lines_rtf(paste0( pgby$label, ": ", value), width,
    #                            rs$font, rs$font_size, rs$units)
    
  
    if (pgby$blank_row %in% c("below", "both")) {
      
      tb <- get_cell_borders(trows, 1, trows, 1, pgby$borders)
      
      ret[length(ret) + 1] <- paste0("\\trowd\\trgaph0", ta, tb, 
                                     "\\cellx", w1, algn, 
                                  "\\cell\\row\n")
      cnt <- cnt + 1 
    }
    
    if (any(pgby$borders %in% c("all", "outside", "bottom")))
      border_flag <- TRUE
    
  }
  
  res <- list(rtf = paste0(ret, collapse = ""), 
              lines = cnt, 
              twips = cnt * rs$line_height,
              border_flag = border_flag)
  
  return(res)
}



#' Get page by text strings suitable for printing
#' @import stringi
#' @param titles Page by object
#' @param width The width to set the page by strings to
#' @return A vector of strings
#' @noRd
get_page_by_rtf_back <- function(pgby, width, value, rs, talgn) {
  
  if (is.null(width)) {
    stop("width cannot be null.") 
    
  }
  
  if (is.null(value))
    value <- ""
  
  ll <- width
  ret <- c()
  cnt <- 0
  border_flag <- FALSE
  
  ta <- "\\trql"
  if (talgn == "right")
    ta <- "\\trqr"
  else if (talgn %in% c("center", "centre"))
    ta <- "\\trqc"
  
  if (!is.null(pgby)) { 
    
    if (!any(class(pgby) == "page_by"))
      stop("pgby parameter value is not a page_by.")
    
    
    w1 <- round(width * rs$twip_conversion)
    
    algn <- "\\ql"
    if (pgby$align == "right")
      algn <- "\\qr"
    else if (pgby$align %in% c("center", "centre"))
      algn <- "\\qc"
    
    trows <- 1
    brow <- 1
    if (pgby$blank_row %in% c("above", "both")) {
      trows <- trows + 1
      brow <- 2
    }
    if (pgby$blank_row %in% c("below", "both"))
      trows <- trows + 1
    
    if (pgby$blank_row %in% c("above", "both")) {
      
      tb <- get_cell_borders(1, 1, trows, 1, pgby$borders)
      
      ret[length(ret) + 1] <- paste0("\\trowd\\trgaph0", ta, tb, 
                                     "\\cellx", w1, algn, 
                                     "\\cell\\row\n")
      cnt <- cnt + 1 
    }
    
    tb <- get_cell_borders(brow, 1 , trows, 1, pgby$borders)
    
    #tmp <- split_string_rtf(vl, cwidth, rs$units)
    
    ret[length(ret) + 1] <- paste0("\\trowd\\trgaph0", ta, tb, 
                                   "\\cellx", w1, algn, " ",
                                   pgby$label, value, "\\cell\\row\n")
    
    
    cnt <- cnt + get_lines_rtf(paste0( pgby$label, ": ", value), width,
                               rs$font, rs$font_size, rs$units)
    
    
    if (pgby$blank_row %in% c("below", "both")) {
      
      tb <- get_cell_borders(trows, 1, trows, 1, pgby$borders)
      
      ret[length(ret) + 1] <- paste0("\\trowd\\trgaph0", ta, tb, 
                                     "\\cellx", w1, algn, 
                                     "\\cell\\row\n")
      cnt <- cnt + 1 
    }
    
    if (any(pgby$borders %in% c("all", "outside", "bottom")))
      border_flag <- TRUE
    
  }
  
  res <- list(rtf = paste0(ret, collapse = ""), 
              lines = cnt, 
              twips = cnt * rs$line_height,
              border_flag = border_flag)
  
  return(res)
}


# Utilities ---------------------------------------------------------------


get_cell_borders <- function(row, col, nrow, ncol, brdrs, flag = "", 
                             exclude = NULL, cell_border = NULL) {
  
  t <- ""
  b <- ""
  l <- ""
  r <- ""
  
  
  if ("all" %in% brdrs) {
    t <- "\\clbrdrt\\brdrs"
    b <- "\\clbrdrb\\brdrs"
    l <- "\\clbrdrl\\brdrs"
    r <- "\\clbrdrr\\brdrs"
  } else {
    
    if ("inside" %in% brdrs) {
      
      t <- "\\clbrdrt\\brdrs"
      b <- "\\clbrdrb\\brdrs"
      l <- "\\clbrdrl\\brdrs"
      r <- "\\clbrdrr\\brdrs"
      
      if (col == 1) 
        l <- ""
      
      if (col == ncol)
        r <- ""
      
      if (row == nrow)
        b <- ""
      
      if (row == 1)
        t <- ""
      
    }
    
    # Cell border is border indicator from group_line or cell_style
    if ((row == 1 & any(brdrs %in% c("outside", "top"))) |
        any(cell_border %in% c("outside", "top"))
        ) {
      t <- "\\clbrdrt\\brdrs"
    }
    
    if ((row == nrow & any(brdrs %in% c("bottom", "outside"))) |
        any(cell_border %in% c("bottom", "outside"))
        ) {
      b <- "\\clbrdrb\\brdrs"
    }
    
    if ((col == 1 & any(brdrs %in% c("outside", "left"))) |
        any(cell_border %in% c("outside", "left"))
        ) {
      l <- "\\clbrdrl\\brdrs"
    }
    
    if ((col == ncol & any(brdrs %in% c("outside", "right"))) |
        any(cell_border %in% c("outside", "right"))
        ) {
      r <- "\\clbrdrr\\brdrs"
    }

  }
  
  # Deal with flag
  if (!is.na(flag)) {
    if (flag %in% c("L", "B", "A")) {
      
      if (col == 1 & any(brdrs %in% c("outside", "right", "all")))
        r <- "\\clbrdrr\\brdrs"
      
      if (col != 1)
        l <- ""
    }
  }
  
  if (!is.null(exclude)) {
    if (any(exclude == "top"))
      t <- ""
    if (any(exclude == "bottom"))
      b <- ""
    if (any(exclude == "left"))
      l <- ""
    if (any(exclude == "right"))
      r <- ""
  }
  
  ret <- paste0(t, b, l, r)
  
  return(ret)
  
}

get_page_numbers_rtf <- function(val, tpg = TRUE) {
  
  ret <- val
  
  ret <- gsub("[pg]", "\\chpgn ", ret, fixed = TRUE)
  
  if (tpg)
    ret <- gsub("[tpg]", "{\\field{\\*\\fldinst  NUMPAGES }}", ret, fixed = TRUE)

  return(ret)
}

context("Page Template RTF Tests")



test_that("get_titles_rtf function works as expected.", {
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello", blank_row = "below") %>%
    footnotes("Goodbye", blank_row = "below")
  
  rpt <- page_setup_rtf(rpt)
  
  t <- get_titles_rtf(rpt$titles, 6, rpt)
  t
  expect_equal(t$rtf,
        paste0("\\trowd\\trgaph0\\trqc\\cellx12960\\qc Hello\\cell\\row\n",
        "\\trowd\\trgaph0\\trqc\\cellx12960\\qc\\sl-275\\slmult0\\cell\\row\n"))
  expect_equal(t$lines, 2)
  # expect_equal(t$twips, 576)
  
})

test_that("get_titles_rtf function works as expected with font_size.", {
  
  # Different title font size will change the title lines
  rpt <- create_report("", font = "Arial", font_size = 9) %>%
    titles("a a a a a a a a a", 
           width = 1)
  
  rpt <- page_setup_rtf(rpt)
  
  t9 <- get_titles_rtf(rpt$titles, 1, rpt)
  expect_equal(t9$lines, 2)
  expect_equal(t9$rtf,
               paste0("\\trowd\\trgaph0\\trqc\\cellx1440\\qc a a a a a a a a a",
                      "\\cell\\row\n\\trowd\\trgaph0\\trqc\\cellx1440\\qc",
                      "\\sl-244\\slmult0\\cell\\row\n")
  )
  
  rpt <- create_report("", font = "Arial", font_size = 9) %>%
    titles("a a a a a a a a a", 
           width = 1,
           font_size = 14)
  
  rpt <- page_setup_rtf(rpt)
  
  t14 <- get_titles_rtf(rpt$titles, 1, rpt)
  expect_equal(t14$lines, 3)
  expect_equal(t14$rtf,
               paste0("\\trowd\\trgaph0\\trqc\\cellx1440\\qc\\fs28\\sl-325",
                      "\\slmult0 a a a a a a\\line a a a\\fs18\\cell\\row\n",
                      "\\trowd\\trgaph0\\trqc\\cellx1440\\qc\\sl-244\\slmult0",
                      "\\cell\\row\n")
  )
  
})


test_that("get_footnotes_rtf function works as expected.", {
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello") %>%
    footnotes("Goodbye")
  
  rpt <- page_setup_rtf(rpt)
  
  f <- get_footnotes_rtf(rpt$footnotes, 6, rpt)
  f
  expect_equal(f$rtf,
      paste0("\\trowd\\trgaph0\\trqc\\cellx12960\\ql\\cell\\row\n", 
             "\\trowd\\trgaph0\\trqc\\cellx12960\\ql Goodbye\\cell\\row\n"))
  expect_equal(f$lines, 2)
  
})

test_that("get_footnotes_rtf function works as expected with font_size.", {
  
  # Different footnote font size will change the footnote lines
  rpt <- create_report("", font = "Arial", font_size = 9) %>%
    footnotes("a a a a a a a a a",
              width = 1)
  
  rpt <- page_setup_rtf(rpt)
  
  f9 <- get_footnotes_rtf(rpt$footnotes, 1, rpt)
  expect_equal(f9$lines, 2)
  expect_equal(f9$rtf,
               paste0("\\trowd\\trgaph0\\trqc\\cellx1440\\ql\\cell\\row\n",
                      "\\trowd\\trgaph0\\trqc\\cellx1440\\ql a a a a a a a a a",
                      "\\cell\\row\n")
  )
  
  rpt <- create_report("", font = "Arial", font_size = 9) %>%
    footnotes("a a a a a a a a a", 
              width = 1,
              font_size = 10)
  
  rpt <- page_setup_rtf(rpt)
  
  f10 <- get_footnotes_rtf(rpt$footnotes, 1, rpt)
  expect_equal(f10$lines, 3)
  expect_equal(f10$rtf,
               paste0("\\trowd\\trgaph0\\trqc\\cellx1440\\ql\\cell\\row\n",
                      "\\trowd\\trgaph0\\trqc\\cellx1440\\ql\\fs20\\sl-250\\",
                      "slmult0 a a a a a a a a\\line a\\fs18\\cell\\row\n")
  )
  
})

test_that("get_title_header_rtf function works as expected.", {

  rpt2 <- create_report("", font = "Arial", font_size = 12) %>%
    title_header("Hello", right = paste("Right here is something",
     "really long that will wrap and wrap and wrap and wrap keep wrapping"))
  
  rpt2 <- page_setup_rtf(rpt2)
  
  th <- get_title_header_rtf(rpt2$title_hdr, 6, rpt2)
  th
  expect_equal(nchar(th$rtf) > 1, TRUE)
  expect_equal(th$lines, 4)
  
})


test_that("get_page_header_rtf works as expected.", {
  
  rpt3 <- create_report("", font = "Arial", font_size = 12) %>%
    page_header(left= c("Hello"),
                right = paste("Right here is something that might wrap.",
                              "If it is long enough so let's make it longer",
                              "If it is long enough so let's make it longer",
                              "If it is long enough so let's make it longer")) %>%
    page_footer("Left", "Center", "Right here is something")
  
  rpt3 <- page_setup_rtf(rpt3)
  
  ph <- get_page_header_rtf(rpt3)
  ph
  
  expect_equal(ph$lines, 3)

})


test_that("get_page_footer_rtf works as expected.", {
  
  rpt3 <- create_report("", font = "Arial", font_size = 12) %>%
    page_header(left= c("Hello"),
                right = paste("Right here is something that might wrap.",
                              "If it is long enough so let's make it longer",
                              "If it is long enough so let's make it longer",
                              "If it is long enough so let's make it longer")) %>%
    page_footer("Left", "Center", "Right here is something")
  
  rpt3 <- page_setup_rtf(rpt3)
  
  pf <- get_page_footer_rtf(rpt3)
  pf
  
  expect_equal(pf$lines, 1)
  
  
})



test_that("get_pageby_rtf works as expected.", {
  
  tbl <- create_table(mtcars) %>% 
    page_by(cyl, "Cylinders:")
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello") %>%
    footnotes("Goodbye") %>% 
    add_content(tbl)
  
  rpt <- page_setup_rtf(rpt)
  
  rpt$line_height
  
  res <- get_page_by_rtf(tbl$page_by, 6, "fork", rpt, "left")
  res  
  
  expect_equal(res$lines, 2)
  expect_equal(res$rtf, 
    paste0("\\trowd\\trgaph0\\trql\\cellx8640\\ql Cylinders: fork\\cell\\row\n", 
           "\\trowd\\trgaph0\\trql\\cellx8640\\ql\\cell\\row\n"))
  
})


test_that("rtf2-2: get_cell_borders works as expected.", {
  
  
  expect_equal(get_cell_borders(1, 1, 4, 4, c("all")),
               "\\clbrdrt\\brdrs\\clbrdrb\\brdrs\\clbrdrl\\brdrs\\clbrdrr\\brdrs")
  
  expect_equal(get_cell_borders(4, 1, 4, 4, c("all")),
               "\\clbrdrt\\brdrs\\clbrdrb\\brdrs\\clbrdrl\\brdrs\\clbrdrr\\brdrs")
  
  expect_equal(get_cell_borders(1, 1, 4, 4, c("outside")),
               "\\clbrdrt\\brdrs\\clbrdrl\\brdrs")
  
  expect_equal(get_cell_borders(4, 1, 4, 4, c("outside")),
               "\\clbrdrb\\brdrs\\clbrdrl\\brdrs")
  
  expect_equal(get_cell_borders(2, 2, 4, 4, c("outside")),
               "")
  
  expect_equal(get_cell_borders(2, 4, 4, 4, c("right")),
               "\\clbrdrr\\brdrs")
  
  expect_equal(get_cell_borders(2, 4, 4, 4, c("inside")),
               "\\clbrdrt\\brdrs\\clbrdrb\\brdrs\\clbrdrl\\brdrs")
  
  expect_equal(get_cell_borders(2, 4, 4, 4, "all", "B"), 
               "\\clbrdrt\\brdrs\\clbrdrb\\brdrs\\clbrdrr\\brdrs")    
  
  expect_equal(get_cell_borders(2, 3, 4, 4, "all", "B"), 
               "\\clbrdrt\\brdrs\\clbrdrb\\brdrs\\clbrdrr\\brdrs")  
  
  expect_equal(get_cell_borders(2, 1, 4, 4, "all", "B"), 
               "\\clbrdrt\\brdrs\\clbrdrb\\brdrs\\clbrdrl\\brdrs\\clbrdrr\\brdrs") 
  
})


test_that("page_template_rtf works as expected.", {
  
  
  tbl <- create_table(mtcars) %>% 
    page_by(cyl, "Cylinders:")
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    page_header("left", "right") %>% 
    titles("Hello") %>%
    footnotes("Goodbye") %>% 
    add_content(tbl) %>% 
    page_footer("left", right = "right")
  
  
  rpt <- page_setup_rtf(rpt)
  

  
  res <- page_template_rtf(rpt)
  
  res
  expect_equal(res$lines, 6)
  expect_equal(res$titles$lines, 2)
  
  
})

test_that("get_page_numbers_rtf works as expected.", {
  
  res <- get_page_numbers_rtf("Page [pg] of [tpg]")
  
  res
  expect_equal(res, "Page \\chpgn  of {\\field{\\*\\fldinst  NUMPAGES }}")
  
  
})

# Special Functions for PDF and RTF ---------------------------------------


#' @title
#' Create a text report, but not on the file system
#'
#' @description
#' This function creates a report, using the
#' parameters provided in the report_spec object.
#'
#' @param x The report_spec object to write.
#' @return The report spec.
#' @noRd
create_report_text <- function(rs) {
  
  # Calculate available space for content 
  rs <- page_setup(rs)
  
  # Get page template for easy access
  pt <- rs$page_template
  
  # Put content in new variable for easy access
  ls <- rs$content
  
  # Break content into pages
  ret <- paginate_content(rs, ls)
  
  # Assign table column widths back to report spec for reference
  rs$column_widths <- ret[["widths"]]
  
  # Assign graphics flag back to report spec
  rs$has_graphics <- ret[["has_graphics"]]
  
  # Assign pages to ls and continue processing
  ls <- ret[["pages"]]
  
  # Deal with preview
  if (!is.null(rs$preview)) {
    if (rs$preview < length(ls[[1]]$pages))
      ls[[1]]$pages <- ls[[1]]$pages[seq(1, rs$preview)]
  }
  
  # Combine all content pages 
  res <- create_content(rs, ls, pt)
  rs <- res$rs
  ls <- res$ls
  
  # After report is created, fix the page numbers.
  # Reason is we don't really know how many pages there are 
  # until the report is created.
  upt <- update_page_numbers(rs, ls)
  
  res <- list(rs = rs, ls = upt)
  
  return(res)
}


#' @title Create content for PDF and RTF output type.
#' @description This loop prepares pages created in paginate_content
#' @noRd
create_content <- function(rs, ls, pt) {
  
  ret <- list()

  counter <- 0
  page <- 0
  last_object <- FALSE
  last_page <- FALSE
  page_open <- FALSE
  
  blank_margin_top <- rep("", rs$blank_margin_top)
  blank_margin_left <- paste0(rep(" ", rs$blank_margin_left), collapse = "")
  
  
  for (cont in ls) {
    
    
    # Increment counter
    counter <- counter + 1
    page <- 0
    
    
    # Set last_object flag
    if (counter == length(ls))
      last_object <- TRUE
    else 
      last_object <- FALSE
    
    
    for (pg in cont$pages) {
    
      page <- page + 1
      
      # Vector to hold lines on a page
      if (page_open == FALSE)
        lns <- c()
      
      if (page == length(cont$pages))
        last_page <- TRUE
      else
        last_page <- FALSE
      
      
      
      if (rs$blank_margins)
        lns <- append(lns, blank_margin_top)
      
      
      #print(page_open)
      if (page_open == FALSE) {
        if (!is.null(pt$page_header))
          lns <- append(lns, paste0(blank_margin_left, pt$page_header))
        
        if (!is.null(pt$title_hdr)) {
          tmp <- trimws(pad_any(pt$title_hdr, rs$line_size, get_justify(cont$align)),
                        which = "right")
          lns <- append(lns, paste0(blank_margin_left, tmp))
        }
        
        if (!is.null(pt$titles)) {
          
          tmp <- trimws(pad_any(pt$titles, rs$line_size, get_justify(cont$align)),
                        which = "right")
          lns <- append(lns, paste0(blank_margin_left, tmp))
        }
        
      }
      
      if (!is.null(pg)) {
        tmp <- trimws(pad_any(pg, rs$line_size, get_justify(cont$align)), 
                      which = "right")
        lns <- append(lns, paste0(blank_margin_left, tmp))
        
      }
      
      # Set page_open flag based on status of page_break and current objects
      if (last_object == FALSE & last_page == TRUE & cont$page_break == FALSE)
        page_open <- TRUE
      else 
        page_open <- FALSE
      
      if (page_open == FALSE) {
        
        if (!is.null(pt$footnotes)) {
          tmp <- trimws(pad_any(pt$footnotes, rs$line_size, get_justify(cont$align)),
                        which = "right")
          lns <- append(lns, paste0(blank_margin_left, tmp))
        }
        
        if (!is.null(pt$page_footer))
          lns <- append(lns, paste0(blank_margin_left, pt$page_footer))
        
        # Increment total page count
        if (is.null(rs$pages))
          rs$pages <- 1
        else 
          rs$pages <- rs$pages + 1 
        
        # If page not open, add to list
        ret[[length(ret) + 1]] <- lns 
        
        
      }
  
      
    }
    
  }
  
  res <- list(rs = rs, ls = ret)
  
  return(res)
  
}

#' @description Update page numbers in text file
#' @details Logic is to read in each line of the file, loop through and replace
#' tokens as needed.  Pages numbers are incremented every time a form feed
#' is encountered.  Total pages is retrieved from the report pages property.
#' Total pages is calculated when a page break occurs.
#' @noRd
update_page_numbers <- function(rs, lns) {
  
  
  # Set up page variables
  tpg <- rs$pages

  
  # Define vectorized function to replace tokens
  replace_tokens <- Vectorize(function(x, srch, just) {
    
    ret <- x
    
    # Replace tokens, but keep overall width the same
    if (grepl(srch, x, fixed = TRUE)) {
      #print("found it")
      
      tmp <- sub("[tpg]", tpg, srch, fixed = TRUE)
      tmp <- sub("[pg]", pg, tmp, fixed = TRUE)
      tmp <- pad_any(tmp, nchar(srch), just)
      ret <- sub(srch, tmp, x, fixed = TRUE)
    }
    
    return(ret)
  }, USE.NAMES = FALSE)
  
  
  for (pg in seq_along(lns)) {
  
    # Call vectorized function on entire page header/footer segment
    
    if (token_check(rs$page_header_left)) {
      for (i in seq_len(length(rs$page_header_left))) {
        if (token_check(rs$page_header_left[i])) {
  
          lns[[pg]] <- replace_tokens(lns[[pg]], rs$page_header_left[i], "left")
        }
      }
    }
    
    if (token_check(rs$page_header_right)) {
      for (i in seq_len(length(rs$page_header_right))) {
        if (token_check(rs$page_header_right[i])) {
  
          lns[[pg]] <- replace_tokens(lns[[pg]], rs$page_header_right[i], "right")
        }
      }
    }
    
    if (token_check(rs$page_footer_left)) {
      for (i in seq_len(length(rs$page_footer_left))) {
        if (token_check(rs$page_footer_left[i])) {
          lns[[pg]] <- replace_tokens(lns[[pg]], rs$page_footer_left[i], "left")
        }
      }
    }
    if (token_check(rs$page_footer_center)) {
      for (i in seq_len(length(rs$page_footer_center))) {
        if (token_check(rs$page_footer_center[i])) {
          lns[[pg]] <- replace_tokens(lns[[pg]], rs$page_footer_center[i], "centre")
        }
      }
    }
    if (token_check(rs$page_footer_right)) {
      
      for (i in seq_len(length(rs$page_footer_right))) {
        if (token_check( rs$page_footer_right[i])) {
          lns[[pg]] <- replace_tokens(lns[[pg]], rs$page_footer_right[i], "right")
        }
      }
    }
    if (!is.null(rs$title_hdr)) {
      if (token_check(rs$title_hdr$right)) {
        for (i in seq_len(length(rs$title_hdr$right))) {
          if (token_check(rs$title_hdr$right[i])) {
            lns[[pg]] <- replace_tokens(lns[[pg]], rs$title_hdr$right[i], "right")
          }
        }
      }
    }
    
    if (!is.null(rs$titles)) {
      for (t in rs$titles) {
        if (token_check(t$titles)) {
          lns[[pg]] <- replace_tokens(lns[[pg]], t$titles, t$align)
        }
      }
    }
    
    if (!is.null(rs$footnotes)) {
      for (f in rs$footnotes) {
        if (token_check(f$footnotes)) {
          lns[[pg]] <- replace_tokens(lns[[pg]], f$footnotes, f$align)
        }
      }
    }
    
    for (cntnt in rs$content) {
      if (!is.null(cntnt$object$title_hdr)) { 
        if (token_check(cntnt$object$title_hdr$right)) {
          for (i in seq_len(length(cntnt$object$title_hdr$right))) {
            if (token_check(cntnt$object$title_hdr$right[i])) {
              lns[[pg]] <- replace_tokens(lns[[pg]], cntnt$object$title_hdr$right[i], "right")
            }
          }
        }
        
      }
      
      if (!is.null(cntnt$object$titles)) {
        for (t in cntnt$object$titles) {
          if (token_check(t$titles)) {
            lns[[pg]] <- replace_tokens(lns[[pg]], t$titles, t$align)
          }
        }
      }
      
      if (!is.null(cntnt$object$footnotes)) {
        for (f in cntnt$object$footnotes) {
          if (token_check(f$footnotes)) {
            lns[[pg]] <- replace_tokens(lns[[pg]], f$footnotes, f$align)
          }
        }
      }
    }
    
  }
  
  
  return(lns)
}



# Write Text Driver Function ----------------------------------------------



#' @title
#' Write a text report to the file system
#'
#' @description
#' This function writes a report_spec object to the file system, using the
#' parameters provided in the object.
#'
#' @param x The report_spec object to write.
#' @return The report spec.
#' @noRd
write_report_text <- function(rs) {

  # Calculate available space for content 
  rs <- page_setup(rs)
  
  # Get page template for easy access
  pt <- rs$page_template

  # Put content in new variable for easy access
  ls <- rs$content

  # Break content into pages
  ret <- paginate_content(rs, ls)
  
  # Assign table column widths back to report spec for reference
  rs$column_widths <- ret[["widths"]]
 
  # Assign graphics flag back to report spec
  rs$has_graphics <- ret[["has_graphics"]]
  
  # Assign pages to ls and continue processing
  ls <- ret[["pages"]]

  # Deal with preview
  if (!is.null(rs$preview)) {
    if (rs$preview < length(ls[[1]]$pages))
      ls[[1]]$pages <- ls[[1]]$pages[seq(1, rs$preview)]
  }
  
  # Write pages to file
  rs <- write_content(rs, ls, pt)

  # After report is written, reopen and fix the page numbers.
  # Reason is we don't really know how many pages there are 
  # until the report is written.
  rs <- write_page_numbers(rs)

  invisible(rs)
}


#' @title Paginate content
#' @description This loop breaks long content into separate pages.
#' @details 
#' Tricky part is taking account of a partially filled page with no page break.
#' Basically we need to track how many lines are on the last page,
#' and send to next pagination call to use for an offset.
#' @noRd
paginate_content <- function(rs, ls) {
  
  last_page_lines <- 0 
  last_object <- FALSE
  table_widths <- list()
  tmp_dir <- tempdir()
  has_graphics <- FALSE
  
  for(i in seq_along(ls)){
    
    if (i == length(ls))
      last_object <- TRUE
    else 
      last_object <- FALSE
    
    # Break content into multiple pages if needed.
    # If there are multiple pages, each function is responsible for 
    # filling out the page body entirely.  Only the last page will
    # remain open, in case there is more content and no page break between.
    # Blanks on last page are handled below.
    # If a last page is left unfinished, and there is no page break, 
    # the last_page_lines variable is populated with the number of lines
    # on the last page. That is passed to the next function, so it can
    # subtract those lines from the first page of the next piece of content.
    if (class(ls[[i]]$object)[1] == "table_spec"){
      
      ret <- create_table_pages_text(rs, ls[[i]], last_page_lines)
      
      table_widths[[length(table_widths) + 1]] <- ret[["widths"]] 
      pgs <- ret[["page_list"]]
      
    } else if (class(ls[[i]]$object)[1] == "text_spec") {
      
      pgs <- create_text_pages_text(rs, ls[[i]], last_page_lines)
      
    } else if (class(ls[[i]]$object)[1] == "plot_spec") {
      
      pgs <- create_plot_pages_text(rs, ls[[i]], last_page_lines, tmp_dir)
      has_graphics <- TRUE
    }
    
    # Append pages to content page list
    # These will be written out in following step
    ls[[i]]$pages <- pgs
    
    last_page <- pgs[[length(pgs)]]
    last_page_lines <- length(last_page) + last_page_lines
    
    # If there is more than one page returned, then a page break occurred
    # and we can reset the last page lines.
    if (length(pgs) > 1)
      last_page_lines <- length(last_page)
    
    # print(paste("Last page lines:", last_page_lines))
    
    # If there is a requested page break, or it is the last object/last page,
    # then fill up the remaining page with blanks.
    if (ls[[i]]$page_break | last_object) {
      # Fill blanks on last page 
      blnks <- c()
      bl <- rs$body_line_count - last_page_lines 
      if (bl > 0)
        blnks <- rep("", bl)
      
      last_page <- append(last_page, blnks)
      last_page_lines <- 0  # Needed for requested page breaks
      #print(paste("Last Page Line Count:", length(last_page)))
    } 
    
    if (ls[[i]]$page_break == TRUE | last_page_lines >= rs$body_line_count)
      last_page_lines <- 0

    # Replace last page with any modifications
    ls[[i]]$pages[[length(pgs)]] <- last_page

  }
  
  ret <- list(widths = table_widths, pages = ls, has_graphics = has_graphics)
  
  return(ret)
}



#' @title Write out content
#' @description This loop writes out pages created in paginate_content
#' @noRd
write_content <- function(rs, ls, pt) {
  
  # Kill existing file
  if (file.exists(rs$modified_path))
    file.remove(rs$modified_path)
  
  counter <- 0
  page <- 0
  last_object <- FALSE
  last_page <- FALSE
  page_open <- FALSE
  
  blank_margin_top <- rep("", rs$blank_margin_top)
  blank_margin_left <- paste0(rep(" ", rs$blank_margin_left), collapse = "")
  

  for (cont in ls) {
    
    
    # Increment counter
    counter <- counter + 1
    page <- 0
    
    # Set last_object flag
    if (counter == length(ls))
      last_object <- TRUE
    else 
      last_object <- FALSE
    
    
    for (pg in cont$pages) {
      
      page <- page + 1
      
      if (page == length(cont$pages))
        last_page <- TRUE
      else
        last_page <- FALSE
      
      
      f <- file(rs$modified_path, open="a", encoding = "native.enc")
      
      if (rs$blank_margins)
        writeLines(enc2utf8(blank_margin_top), con = f, useBytes = TRUE)
      
      
      #print(page_open)
      if (page_open == FALSE) {
        if (!is.null(pt$page_header))
          writeLines(enc2utf8(paste0(blank_margin_left, pt$page_header)), 
                     con = f, useBytes = TRUE)
        
        if (!is.null(pt$title_hdr)) {
          tmp <- trimws(pad_any(pt$title_hdr, rs$line_size, get_justify(cont$align)),
                        which = "right")
          
          writeLines(enc2utf8(paste0(blank_margin_left, tmp)), 
                     con = f, useBytes = TRUE)
        }
        
        if (!is.null(pt$titles)) {
           tmp <- trimws(pad_any(pt$titles, rs$line_size, get_justify(cont$align)),
                        which = "right")
           writeLines(enc2utf8(paste0(blank_margin_left, tmp)), 
                      con = f, useBytes = TRUE)
        }
        
      }
      
      if (!is.null(pg)) {
        tmp <- trimws(pad_any(pg, rs$line_size, get_justify(cont$align)),
                      which = "right")
        writeLines(enc2utf8(paste0(blank_margin_left, tmp)), 
                   con = f, useBytes = TRUE)
        
      }
      
      # Set page_open flag based on status of page_break and current objects
      if (last_object == FALSE & last_page == TRUE & cont$page_break == FALSE)
        page_open <- TRUE
      else 
        page_open <- FALSE
      
      if (page_open == FALSE) {
        
        if (!is.null(pt$footnotes)) {
          tmp <- trimws(pad_any(pt$footnotes, rs$line_size, get_justify(cont$align)),
                        which = "right")
          writeLines(enc2utf8(paste0(blank_margin_left, tmp)), 
                     con = f, useBytes = TRUE)
        }
        
        if (!is.null(pt$page_footer))
          writeLines(enc2utf8(paste0(blank_margin_left, pt$page_footer)), 
                     con = f, useBytes = TRUE)
        
        # Add form feed character for text page break
        if (last_object == FALSE | last_page == FALSE) {
          
          if (is.null(rs$pages))
            rs$pages <- 1
          else 
            rs$pages <- rs$pages + 1 
          
          writeLines(enc2utf8(""), con = f, sep = "\f", useBytes = TRUE) 
        
        }
      }
      
      close(f)
      
    }
    
  }
  
  return(rs)
  
}

#' @description Setup page for content
#' @details  Calculates available space for content and prepares text lines
#' for page template.
#' @noRd
page_setup <- function(rs) {
  
  debug <- FALSE
  
  if (debug) {
    print(paste("Row height:", rs$row_height))
    print(paste("Character width:", rs$char_width))
  }
  
  # Content size is the page size minus margins, in units of measure
  rs$content_size <- get_content_size(rs)
  if (debug)
    print(paste0("Content Size: ", rs$content_size))
  
  # Line size is the number of characters that will fit in the content size width
  if (is.null(rs$user_line_size)) {
    if (rs$output_type == "RTF") {
      # 1 char adjustment to avoid occasional wrapping 
      rs$line_size <- floor(rs$content_size[["width"]] / rs$char_width) - 1 
    } else {
      rs$line_size <- floor(rs$content_size[["width"]] / rs$char_width) 
    }
  } else 
    rs$line_size <- rs$user_line_size
  if (debug)
    print(paste0("Line Size: ", rs$line_size))
  
  # Line count is the number of lines that will fit in the content size height
  if (is.null(rs$user_line_count))
    rs$line_count <- floor(rs$content_size[["height"]] / rs$row_height)
  else
    rs$line_count <- rs$user_line_count
  if (debug)
    print(paste0("Line Count: ", rs$line_count))
  
  # Get page template
  # Page template is the headers, footers, titles, and footnotes attached to report
  # Requires the line_size
  # Returns a page_template_text object
  pt <- page_template_text(rs)
  rs$page_template <- pt
  #print(pt)
  
  # Get the page template row count
  # Include all the rows associated with the page template
  rs$page_template_header_count <- length(pt$page_header) + length(pt$titles) + 
    length(pt$title_hdr) + length(pt$page_by)
  if (debug)
    print(paste("Page Template Header Count:", rs$page_template_header_count))

  rs$page_template_footer_count <- length(pt$footnotes) + length(pt$page_footer)
  if (debug)
    print(paste("Page Template Footer Count:", rs$page_template_footer_count))

  rs$page_template_row_count <- rs$page_template_header_count + 
                                rs$page_template_footer_count
  if (debug)
    print(paste("Page Template Row Count:", rs$page_template_row_count))
  
  # Body line count is the number of rows available for content on each page
  rs$body_line_count <- rs$line_count - rs$page_template_row_count
  if (debug)
    print(paste0("Body Line Count: ", rs$body_line_count))
  
  
  # Calculate blank margins, if specified
  if (rs$blank_margins) {
    rs$blank_margin_top <- floor((rs$margin_top - rs$min_margin) / rs$row_height)
  } else 
    rs$blank_margin_top <- 0
  if (debug)
    print(paste("Blank Margin Top:", rs$blank_margin_top))
  
  if (rs$blank_margins)
    rs$blank_margin_left <- ceiling((rs$margin_left - rs$min_margin) / rs$char_width)
  else 
    rs$blank_margin_left <- 0
  if (debug)
    print(paste("Blank Margin Left:", rs$blank_margin_left))
  

  
  return(rs)
}

#' @description Update page numbers in text file
#' @details Logic is to read in each line of the file, loop through and replace
#' tokens as needed.  Pages numbers are incremented every time a form feed
#' is encountered.  Total pages is retrieved from the report pages property.
#' Total pages is calculated when a page break occurs.
#' @noRd
write_page_numbers <- function(rs) {
 
  # Read file into vector
  lns <- readLines(rs$modified_path, encoding = "UTF-8")
  

  # Adjust page count
  rs$pages <- rs$pages + 1
  
  # Set up page variables
  tpg <- rs$pages 
  pg <- 1
  
  # Define vectorized function to replace tokens
  replace_tokens <- Vectorize(function(x, srch, just) {
    
    ret <- x
  
    # Update page number if hit a page break
    if (grepl("\f", x, fixed = TRUE)) {
      #print("Updated page number")
      # This double arrow <<- is updating the pg variable in the containing
      # function write_page_numbers().  It is not updating the global environment.
      pg <<- pg + 1 
    }
    
    # Replace tokens, but keep overall width the same
    if (grepl(srch, x, fixed = TRUE)) {
      #print("found it")
      
      tmp <- sub("[tpg]", tpg, srch, fixed = TRUE)
      tmp <- sub("[pg]", pg, tmp, fixed = TRUE)
      tmp <- pad_any(tmp, nchar(srch), just)
      ret <- sub(srch, tmp, x, fixed = TRUE)
    }
    
    return(ret)
  })

  # Call vectorized function on entire page header/footer segment

  if (token_check(rs$page_header_left)) {
    for (i in seq_len(length(rs$page_header_left))) {
      if (token_check(rs$page_header_left[i])) {
        pg <- 1
        lns <- replace_tokens(lns, rs$page_header_left[i], "left")
      }
    }
  }

  if (token_check(rs$page_header_right)) {
    for (i in seq_len(length(rs$page_header_right))) {
      if (token_check(rs$page_header_right[i])) {
        pg <- 1
        lns <- replace_tokens(lns, rs$page_header_right[i], "right")
      }
    }
  }
  
  if (token_check(rs$page_footer_left)) {
    for (i in seq_len(length(rs$page_footer_left))) {
      if (token_check(rs$page_footer_left[i])) {
        pg <- 1
        lns <- replace_tokens(lns, rs$page_footer_left[i], "left")
      }
    }
  }
  if (token_check(rs$page_footer_center)) {
    for (i in seq_len(length(rs$page_footer_center))) {
      if (token_check(rs$page_footer_center[i])) {
        pg <- 1
        lns <- replace_tokens(lns, rs$page_footer_center[i], "centre")
      }
    }
  }
  if (token_check(rs$page_footer_right)) {

    for (i in seq_len(length(rs$page_footer_right))) {
      if (token_check( rs$page_footer_right[i])) {
        pg <- 1
        lns <- replace_tokens(lns, rs$page_footer_right[i], "right")
      }
    }
  }
  if (!is.null(rs$title_hdr)) {
    for (th in rs$title_hdr) {
      if (token_check(th$right)) {
        for (i in seq_len(length(th$right))) {
          if (token_check(th$right[i])) {
            pg <- 1
            lns <- replace_tokens(lns, th$right[i], "right")
          }
        }
      }
    }
  }
  
  if (!is.null(rs$titles)) {
    for (t in rs$titles) {
      if (token_check(t$titles)) {
        pg <- 1
        lns <- replace_tokens(lns, t$titles, t$align)
      }
    }
  }
  
  if (!is.null(rs$footnotes)) {
    for (f in rs$footnotes) {
      if (token_check(f$footnotes)) {
        pg <- 1
        lns <- replace_tokens(lns, f$footnotes, f$align)
      }
    }
  }
  
  for (cntnt in rs$content) {
   if (!is.null(cntnt$object$title_hdr)) { 
     for (th in cntnt$object$title_hdr) {
       if (token_check(th$right)) {
         for (i in seq_len(length(th$right))) {
           if (token_check(th$right[i])) {
            pg <- 1
            lns <- replace_tokens(lns, th$right[i], "right")
           }
         }
       }
     }
   }
    
    if (!is.null(cntnt$object$titles)) {
      for (t in cntnt$object$titles) {
        if (token_check(t$titles)) {
          pg <- 1
          lns <- replace_tokens(lns, t$titles, t$align)
        }
      }
    }
    
    if (!is.null(cntnt$object$footnotes)) {
      for (f in cntnt$object$footnotes) {
        if (token_check(f$footnotes)) {
          pg <- 1
          lns <- replace_tokens(lns, f$footnotes, f$align)
        }
      }
    }
  }
  
  # Replace file with updated lines
  f <- file(rs$modified_path, open="w+", encoding = "native.enc")
  

  writeLines(enc2utf8(lns), con = f, useBytes = TRUE)
  close(f)
  
  return(rs)
}

#' @noRd
token_check <- function(x) {
  
  ret <- FALSE
  
  if (!is.null(x))
    if (any(nchar(x) > 0))
      if (any(grepl("[pg]", x, fixed = TRUE)) |
          any(grepl("[tpg]", x, fixed = TRUE)))
            ret <- TRUE
      
  return(ret)
}

# Create Tables ----------------------------------------------------------


#' @noRd
create_table_pages_docx <- function(rs, cntnt, lpg_rows) {
  
  ts <- cntnt$object
  content_blank_row <- cntnt$blank_row
  
  pgby_var <- NA
  pgby_cnt <- 0
  pgby_fmt <- NULL
  if (!is.null(rs$page_by)) {
    pgby_var <- rs$page_by$var
    pgby_fmt <- rs$page_by$format
  } else if (!is.null(ts$page_by)) {
    pgby_var <- ts$page_by$var
    pgby_fmt <- ts$page_by$format
  }
  
  
  if (all(ts$show_cols == "none") & length(ts$col_defs) == 0) {
    
    stop("ERROR: At least one column must be defined if show_cols = \"none\".")
  }
  
  font_name <- rs$font
  
  # Set up control columns
  dat <- as.data.frame(ts$data, stringsAsFactors = FALSE)  
  dat$..blank <- ""
  dat$..row <- 1
  dat$..page_by <- NA
  
  # If page_break variable has been defined, use it
  if (is.null(ts$page_var)) {
    if (is.na(pgby_var))
      dat$..page <- NA
    else 
      dat$..page <-  dat[[pgby_var]]
    
  } else {
    if (any(class(dat[[ts$page_var]]) == "factor"))
      dat$..page <- as.character(dat[[ts$page_var]])
    else 
      dat$..page <- dat[[ts$page_var]]
  }
  
  # If page by is defined, use it
  if (!is.na(pgby_var)) {
    
    # Clear out factors on page by if they exist
    # Needed because split cells used to convert everything to characters
    if (any(class(dat[[pgby_var]]) == "factor")) {
      dat[[pgby_var]] <- as.character(dat[[pgby_var]] )
      dat$..page_by <- dat[[pgby_var]] 
      
      if (any(class(dat$..page) == "factor")) 
        dat$..page <- as.character(dat[[pgby_var]])
      
    } else {
      
      if (!is.null(pgby_fmt))
        dat$..page_by <-  fapply(dat[[pgby_var]], pgby_fmt)
      else 
        dat$..page_by <-  dat[[pgby_var]]
    }
    
    pgby_cnt <- get_pgby_cnt(dat$..page_by)
    
    # Commenting out for now
    # if (is.unsorted(dat[[pgby_var]], strictly = FALSE))
    #   message("Page by variable not sorted.")
  }
  
  # Deal with invisible columns
  if (!is.null(ts$col_defs)) {
    for (def in ts$col_defs) {
      if (def$visible == FALSE) {
        nnm <- paste0("..x.", def$var_c)
        dat[[nnm]] <- dat[[def$var_c]]
      }
    }
    
    # Deal with Group Cohesion
    # Prepare ..group_cohesion and ..min_page_prop
    group_cohesion <- c()
    min_page_prop <- c()
    
    i <- 1
    for (def in ts$col_defs) {
      if (def$group_cohesion == TRUE) {
        group_cohesion <- c(group_cohesion, def$var_c)
        min_page_prop <- c(min_page_prop, def$min_page_prop)
        
        dat$..temp <- dat[[def$var_c]]
        names(dat)[names(dat) == "..temp"] <-  paste0("..group_cohesion",i)
        
        dat$..temp <- def$min_page_prop
        names(dat)[names(dat) == "..temp"] <-  paste0("..min_page_prop",i)
        i <- i + 1
      }
    }
    
    if (length(group_cohesion) == 0) {
      group_cohesion <- NULL
    }
    
    if (length(min_page_prop) == 0) {
      min_page_prop <- NULL
    }
    
    # Prepare empty ..break_label and ..break_label_lines for later use
    i <- 1
    for (def in ts$col_defs) {
      if (!is.null(def$break_label)) {
        
        dat$..temp <- ifelse(!is.na(dat[[def$var_c]]) & dat[[def$var_c]] != "",
                             paste0(dat[[def$var_c]], " ", def$break_label),
                             dat[[def$var_c]])
        names(dat)[names(dat) == "..temp"] <-  paste0("..break_label",i)
        
        dat$..temp <- rep(NA, nrow(dat))
        names(dat)[names(dat) == "..temp"] <-  paste0("..break_label_lines",i)
        i <- i + 1
      }
    }
    # Prepare repeated label occurrence for later use
    if (any(grepl("..break_label", names(dat)))) {
      dat$..break_occur <- rep(NA, nrow(dat))
    }
  }
  
  # Get control column names 
  control_cols <- names(dat)[is.controlv(names(dat))]
  
  # Get vector of all included column names
  # Not all columns in dataset are necessarily included
  # depends on show_cols parameter on create_table and
  # visible parameter on column definitions
  keys <- get_table_cols(ts, control_cols)
  # print("keys")
  # print(keys)
  
  # Filter dataset by included columns
  dat <- get_data_subset(dat, keys, rs$preview)
  # print("Key columns:")
  # print(dat)
  
  # Update column definitions with column defaults
  ts$col_defs <- set_column_defaults(ts, keys)
  # print("col_defs:")
  # print(ts$col_defs)
  
  # Deal with styles
  styles <- get_styles(ts)
  
  # Get labels
  labels <- get_labels(dat, ts)
  # print("Labels:")
  # print(labels)
  
  # Get column alignments
  aligns <- get_aligns(dat, ts)
  
  # Get alignment for labels
  # Follows column alignment by default
  label_aligns <- get_label_aligns(ts, aligns)
  # print("Label Aligns:")
  # print(label_aligns)
  
  # Clear out existing formats
  cdat <- clear_formats(dat)
  
  # Get column formats
  formats(cdat) <- get_col_formats(dat, ts)
  # print("formats:")
  # print(formats(cdat))
  
  # Apply formatting
  fdat <- fdata(cdat)
  # print("fdata:")
  # print(fdat)
  
  # Prep data for blank lines, indents, and stub columns
  fdat <- prep_data(fdat, ts, rs$char_width, rs$missing) # OK
  # print("prep_data")
  # print(fdat)
  # str(fdat)
  
  if ("..stub_var" %in% names(fdat)) {
    control_cols <- c(control_cols, "..stub_var")
  }
  
  if ("..group_border" %in% names(fdat)){
    control_cols <- c(control_cols, "..group_border")
  }
  
  # Reset keys, since prep_data can add/remove columns for stub
  keys <- names(fdat)
  # print("Keys")
  # print(keys)
  # print("Aligns")
  # print(aligns)
  
  # Copy any width attributes to formatted data frame
  if ("width" %in% ts$use_attributes)
    widths(fdat) <- widths(dat)
  # print("Original Widths")
  # print(widths(dat))
  
  
  # Get column widths
  widths_uom <- get_col_widths_variable(fdat, ts, labels, 
                                        rs$font, rs$font_size, rs$units, 
                                        rs$gutter_width) 
  # print("Widths UOM")
  # print(widths_uom)
  
  # Split long text strings into multiple rows. Number of rows are stored in
  # ..row variable. If too slow, may need to be rewritten in C
  fdat <- split_cells_variable(fdat, widths_uom, rs$font,
                               rs$font_size, rs$units, rs$output_type, rs$char_width, ts)$data
  # print("split_cells")
  # print(fdat)
  
  
  # Break columns into pages
  wraps <- get_page_wraps(rs$line_size, ts, 
                          widths_uom, 0, control_cols)  # No gutter width for RTF
  # print("wraps")
  # print(wraps)
  
  
  # Create a temporary page info to pass into get_content_offsets
  tmp_pi <- list(keys = keys, col_width = widths_uom, label = labels,
                 label_align = label_aligns, table_align = cntnt$align, data = fdat)
  # print("Temp PI")
  # print(tmp_pi)
  
  
  # Offsets are needed to calculate splits and page breaks
  content_offset <- get_content_offsets_docx(rs, ts, tmp_pi, 
                                             content_blank_row, pgby_cnt)
  
  
  # split rows
  splits <- get_splits_text(fdat, widths_uom, rs$body_line_count, 
                            lpg_rows, content_offset$lines, ts, TRUE)  
  # print("splits")
  # print(splits)
  
  # Subset splits by preview, if requested
  if (!is.null(rs$preview)) {
    if (rs$preview < length(splits))
      splits <- splits[seq(1, rs$preview)] 
  }
  
  tot_count <- length(splits) * length(wraps)
  counter <- 0
  wrap_flag <- FALSE
  blnk_ind <- "none"
  
  pg_lst <- list()
  for(s in splits) {
    for(pg in wraps) {
      counter <- counter + 1
      
      if (counter < tot_count)
        wrap_flag <- TRUE
      else 
        wrap_flag <- FALSE
      
      #print(s)
      
      if (!is.na(pgby_var))
        pgby <- trimws(s[1, "..page_by"])
      else 
        pgby <- NULL
      
      # Ensure content blank rows are added only to the first and last pages
      blnk_ind <- get_blank_indicator(counter, tot_count, content_blank_row,
                                      rs$body_line_count, content_offset$lines, 
                                      nrow(s), pgby)
      #print(blnk_ind)
      
      
      pi <- page_info(data= s[, pg], keys = pg, label=labels[pg],
                      col_width = widths_uom[pg], col_align = aligns[pg],
                      font_name = font_name, label_align = label_aligns[pg],
                      pgby, cntnt$align)
      pg_lst[[length(pg_lst) + 1]] <- create_table_docx(rs, ts, pi, 
                                                       blnk_ind, wrap_flag,
                                                       lpg_rows, styles)
    }
  }
  
  ret <- list(widths = widths_uom, page_list = pg_lst)
  
  return(ret)
  
  
}



#' @noRd
create_table_docx <- function(rs, ts, pi, content_blank_row, wrap_flag, 
                             lpg_rows, styles) {
  rh <- rs$row_height
  shdrs <- list(lines = 0, twips = 0)
  hdrs <- list(lines = 0, twips = 0)
  conv <- rs$twip_conversion
  
  # Default to content width
  ls <- rs$content_size[["width"]]
  
  ccnt <- length(pi$col_width[!is.na(pi$col_width)])
  
  # Get table width
  if (!is.null(pi$col_width))
    ls <- sum(pi$col_width, na.rm = TRUE)
  
  if (!is.null(ts$title_hdr))
    ttls <- get_title_header_docx(ts$title_hdr, ls, rs, pi$table_align)
  else
    ttls <- get_titles_docx(ts$titles, ls, rs, pi$table_align, 
                            colspan = ccnt, pi$col_width, 
                            content_brdrs = ts$borders) 
  
  
  if (!is.null(rs$page_by)) {
    pgby <- get_page_by_docx(rs$page_by, ls, 
                            pi$page_by, rs, pi$table_align, ttls$border_flag)
  } else if(!is.null(ts$page_by)) {
    pgby <- get_page_by_docx(ts$page_by, ls, pi$page_by, rs, 
                             pi$table_align, ttls$border_flag)
  } else 
    pgby <- c()
  
  exbrdr <- ttls$border_flag
  if (length(pgby) > 0)
    exbrdr <- pgby$border_flag
  
  if (ts$headerless == FALSE) {
    
    # Get table header also includes spanning header
    hdrs <- get_table_header_docx(rs, ts, pi, exbrdr)  
  }
  
  # rs, ts, widths,  algns, halgns, talgn
  rws <- get_table_body_docx(rs, pi$data, pi$col_width, 
                             pi$col_align, pi$table_align, ts$borders, 
                             exbrdr, ts$first_row_blank, styles, ts)
  
  a <- NULL
  if (content_blank_row %in% c("above", "both"))
    a <- rs$blank_row
  
  
  blnks <- c()
  
  
  # Determine sum of all lines
  rc <- sum(ttls$lines, pgby$lines, shdrs$lines, 
            hdrs$lines, rws$lines,
            length(a))
  
  # Get footnotes, passing in sum of all current lines
  ftnts <- get_page_footnotes_docx(rs, ts, ls, lpg_rows, rc,
                                  wrap_flag, content_blank_row,  pi$table_align, 
                                  rws$border_flag, 
                                  colspan = ccnt, pi$col_width,
                                  content_brdrs = ts$borders)
  

  # Get units
  u <- rs$units
  if (u == "inches")
    u <- "in"

  # Table borders
  tb <- get_table_borders_docx(ts$borders)
  
  # Table width
  tw <- round(ls * conv)
 
  
  # Get indent codes for alignment
  ta <- get_indent_docx(pi$table_align, rs$line_size, tw, 
                        rs$base_indent, ts$borders, conv)
  
  # Construct table 
  ts <- paste0("<w:tbl>", "<w:tblPr>",
               ta,
               '<w:tblStyle w:val="TableGrid"/>',
               '<w:tblW w:w="', tw,'"',
               ' w:type="dxa"/>', tb,
               rs$cell_margin,
               "</w:tblPr>")
  
  ret <- list(docx = c(a, ttls$docx, pgby$docx, ts, shdrs$docx, 
                      hdrs$docx, rws$docx, "</w:tbl>", ftnts$docx),
              lines = rc  + ftnts$lines)
  
  return(ret) 
}

#' Footnotes takes special handling because of the valign option.
#' @noRd
get_page_footnotes_docx <- function(rs, spec, spec_width, lpg_rows, row_count,
                                   wrap_flag, content_blank_row, talgn, 
                                   ex_brdr = FALSE, colspan = 0, 
                                   col_widths = NULL, 
                                   content_brdrs = NULL) {
  
  ftnts <- list(lines = 0, twips = 0, border_flag = FALSE, lines_content = 0)
  vflag <- "none"

  
  # Deal with valign parameter
  if (!is.null(spec$footnotes)) {
    if (!is.null(spec$footnotes[[length(spec$footnotes)]])) {
      if (spec$footnotes[[length(spec$footnotes)]]$valign == "bottom") {
        
        vflag <- "bottom"
        ftnts <- get_footnotes_docx(spec$footnotes, 
                                   spec_width, rs, 
                                   talgn, FALSE) 
      } else {
        vflag <- "top"
        ftnts <- get_footnotes_docx(spec$footnotes, spec_width, rs, 
                                    talgn, ex_brdr, colspan, col_widths,
                                    content_brdrs) 
      }
      
    }
  } else {
    
    if (!is.null(rs$footnotes[[1]])) {
      if (!is.null(rs$footnotes[[1]]$valign)) {
        if (rs$footnotes[[1]]$valign == "top") {
          vflag <- "top"
          ftnts <- get_footnotes_docx(rs$footnotes, 
                                     spec_width, rs, 
                                     talgn, ex_brdr,
                                     content_brdrs) 
        } else {
          
          if (wrap_flag) {
            vflag <- "bottom"
          } else {
            # If not wrapping, assign original valign for inserting table_break
            # for report footnotes
            vflag <- rs$footnotes[[1]]$valign
          }
        }
      }
    }
  }
  
  b <- NULL
  blen <- 0
  if (content_blank_row %in% c("below", "both")) {
    b <- rs$blank_row_below
    blen <- 1
  }
  

  
  
  # Add extra offsets if table has a lot of borders turned on
  # to avoid undesired page wraps
  boff <- 0
  if (any(class(spec) == "table_spec") &
      any(spec$borders %in% c("all", "inside"))) {

    boff <- round(row_count * rs$border_height / rs$row_height)
    #boff <- 1
  }

  ublnks <- c()
  lblnks <- c()

  if (rs$paper_size != "none") {
    
    # Determine number of filler lines needed
    len_diff <- rs$body_line_count - row_count - ftnts$lines_content - lpg_rows - blen - boff
    
    if (vflag == "bottom" & len_diff > 0) {
  
      ublnks <- c(b, rep(rs$blank_row_below, len_diff))
  
    } else {
  
      if ((wrap_flag & len_diff > 0)) {
        if (vflag == "bottom" | has_bottom_footnotes(rs))
          lblnks <- c(rep(rs$blank_row_below, len_diff), b)
      } else {
        lblnks <- b
      }
    }
  }

  tbr <- NULL
  if (vflag == "bottom") {
    if (length(ublnks) == 0) {
      tbr <- rs$table_break
    }
  }
  
  # Note that the returned lines have been converted to content font size
  tlns <- sum(ftnts$lines_content, length(ublnks), length(lblnks))
  ret <- list(docx = c(ublnks, tbr, ftnts$docx, lblnks),
              lines = tlns)
  
  
  return(ret)
}

# Sub-Functions ---------------------------------------------------------


#' Get content offsets for table header, titles, footnotes, and content blanks.
#' Needed to calculate page breaks accurately.
#' @return A vector of upper and lower offsets
#' @noRd
get_content_offsets_docx <- function(rs, ts, pi, content_blank_row, pgby_cnt = NULL) {
  
  cnt <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
  
  # Width is normally the width of the table, not the page
  wdth <- rs$content_size[["width"]]
  if (!is.null(pi$col_width))
    wdth <- sum(pi$col_width)
  
  # Default to zero
  shdrs <- list(lines = 0, twips = 0)
  hdrs <- list(lines = 0, twips = 0)
  
  
  if (ts$headerless == FALSE) {
    
    # Spanning headers now inside get_table_header_docx
    hdrs <- get_table_header_docx(rs, ts, pi)  
  }
  
  # Get title headers or titles
  if (is.null(ts$title_hdr))
    ttls <- get_titles_docx(ts$titles, wdth, rs) 
  else {
    ttls <- get_title_header_docx(ts$title_hdr, wdth, rs)
    ttls$lines_content <- ttls$lines
  }
  
  # Get page by if it exists
  # if (!is.null(ts$page_by))
  #   pgb <- get_page_by_docx(ts$page_by, wdth, NULL, rs, pi$table_align, pgby_cnt = pgby_cnt)
  # else if (!is.null(rs$page_by))
  #   pgb <- get_page_by_docx(rs$page_by, wdth, NULL, rs, pi$table_align, pgby_cnt = pgby_cnt)
  
  # Add everything up
  # cnt[["upper"]] <- shdrs$lines + hdrs$lines + ttls$lines + pgb$lines
  
  page_by_info <- NULL
  if (!is.null(ts$page_by)) {
    page_by_info <- ts$page_by
  } else if (!is.null(rs$page_by)) {
    page_by_info <- rs$page_by
  }
  
  # Get lines for each page by value
  if (!is.null(page_by_info)) {
    cnt <- c(upper = list(), lower = 0, blank_upper = 0, blank_lower = 0)
    
    pgby_unique <- unique(pi$data$..page_by)
    
    for (i in 1:length(pgby_unique)) {
      pgby_temp <- get_page_by_docx(page_by_info, wdth, pgby_unique[i], rs, pi$table_align, pgby_cnt = pgby_cnt)
      
      # Add everything up
      cnt[["upper"]][[i]] <- shdrs$lines + hdrs$lines + ttls$lines_content + pgby_temp$lines
    }
    
    names(cnt[["upper"]]) <- pgby_unique
    
  } else {
    pgb <- list(lines = 0)
    pgb <- get_page_by_docx(page_by_info, wdth, NULL, rs, pi$table_align, pgby_cnt = pgby_cnt)
    cnt <- c(upper = 0, lower = 0, blank_upper = 0, blank_lower = 0)
    
    # Add everything up
    cnt[["upper"]] <- shdrs$lines + hdrs$lines + ttls$lines_content + pgb$lines
    
  }
  
  if (content_blank_row %in% c("above", "both")) {
    #ret[["blank_upper"]] <- rs$line_height
    cnt[["blank_upper"]] <- 1 
  }
  
  ftnts <- get_footnotes_docx(ts$footnotes, wdth, rs) 
  rftnts <- get_footnotes_docx(rs$footnotes, wdth, rs)
  
  if (has_top_footnotes(rs)) {
    cnt[["lower"]] <- ftnts$lines_content + rftnts$lines_content
  } else {
    cnt[["lower"]] <- ftnts$lines_content
  }
  
  # Add extra offsets if table has a lot of borders turned on
  # to avoid undesired page wraps
  if (any(ts$borders %in% c("all", "inside"))) {
    #ret[["lower"]] <- ret[["lower"]] + (rs$row_height * 2)
    cnt[["lower"]] <- cnt[["lower"]] + 1
  }
  
  if (content_blank_row %in% c("both", "below")) {
    cnt[["blank_lower"]] <- 1 
  }
  
  res <- list(lines = cnt)
  
  return(res)
  
}


#' @description Return a vector of strings for the table header
#' @details Basic idea of this function is to create a list
#' of string vectors of label words sized according to the column 
#' widths, then combine by line/row, and concatenate everything and justify.
#' @noRd
get_table_header_docx <- function(rs, ts, pi, ex_brdr = FALSE) {
  
  ret <- c()
  cols <- c()
  cnt <- 0
  rh <- rs$row_height
  tbl <- ts$data
  conv <- rs$twip_conversion


  
  widths <- pi$col_width[!is.na(pi$col_width)]
  lbls <- pi$label
  halgns <- pi$label_align
  talgn <- pi$table_align
  exclude_top <- NULL
  if (ex_brdr)
    exclude_top <- "top"
  
  
  conv <- rs$twip_conversion
  nms <- names(lbls)
  
  
  # Get cell widths
  sz <- c()
  for (k in seq_along(widths)) {
    if (!is.control(nms[k])) {
        sz[k] <- round(widths[k] * conv)
    }
  }
  
  grd <- get_col_grid(widths, conv)
  

  brdrs <- ts$borders
  # If there are spanning columns, adjust border spec
  if (length(ts$col_spans) > 0) {
    
    if (any(ts$borders %in% c("outside")))
      brdrs <- c("left", "right")
      
    if (any(ts$borders %in% "top"))
      brdrs <- brdrs[!brdrs %in% "top"]
    
  }
  
  # Header Cell alignment
  ha <- c()
  for (k in seq_along(halgns)) {
    if (!is.control(nms[k])) {
      if (halgns[k] == "left")
        ha[k] <- "left"
      else if (halgns[k] == "right")
        ha[k] <- "right"
      else if (halgns[k] %in% c("center", "centre"))
        ha[k] <- "center"
    }
  }
  
  # Table Header
  ret[1] <- ""
  cols[1] <- "<w:tblGrid>\n"
  
  # Always at least one header row
  cnt <-  1 

  
  # Loop for column names
  pdf(NULL)
  par(family = get_font_family(rs$font), ps = rs$font_size)
  
  for(k in seq_along(widths)) {
    if (!is.control(nms[k])) {
      
      cols[1] <- paste0(cols[1], "<w:gridCol w:w=\"", sz[k], "\"/>\n")
      

      b <- "bottom"
      if (any(brdrs %in% c("top", "outside", "all")))
          b <- c("bottom", "top")
      
      # Split label strings if they exceed column width
      tmp <- split_string_docx(lbls[k], widths[k], rs$units, font = rs$font)

      #if (b == "") {
        ret[1] <- paste0(ret[1], cell_abs(tmp$docx, ha[k], sz[k], 
                                          borders = b, 
                                          valign = "bottom", 
                                          bold = ts$header_bold))
      
      # Add in extra lines for labels that wrap
      xtr <- tmp$lines
      if (xtr > cnt)
         cnt <- xtr
    }
  }
  
  
  dev.off()
  
  rht <- get_row_height(round(rs$row_height * cnt * conv))
  
  cols[1] <- paste0(cols[1], "</w:tblGrid>\n")
  ret[1] <- paste0("<w:tr>", rht, "\n", ret[1], "</w:tr>\n")
  
  if (ts$first_row_blank == TRUE) {
    
    b <- get_cell_borders_docx(2, 1, 3, 1, brdrs)
    
    rht <- get_row_height(round(rs$row_height  * conv))
    
    ret[1] <- paste0(ret[1], "<w:tr>", rht, "<w:tc>", 
               '<w:tcPr>', '<w:gridSpan w:val="', length(sz) , '"/>', b, 
               "</w:tcPr>", para(" "),
               "</w:tc></w:tr>")

    cnt <- cnt + 1
  }
  
  # Get spanning headers
  sphdrs <- get_spanning_header_docx(rs, ts, pi,
                                     ifelse(is.null(exclude_top), FALSE, TRUE))
  

  # Combine everything
  res <- list(docx = paste0(cols, paste0(sphdrs$docx, collapse=""),
                            ret),
              lines = cnt + sphdrs$lines)
  

  
  return(res)
  
}


#' @description Return a vector of docx strings for the table spanning headers
#' @details Basic idea of this function is to figure out which columns 
#' the header spans, add widths, then call get_table_header.  Everything
#' from there is the same.  
#' @import stats
#' @noRd
get_spanning_header_docx <- function(rs, ts, pi, ex_brdr = FALSE) {
  
  spns <- ts$col_spans
  cols <- pi$keys
  cols <- cols[!is.controlv(cols)]
  w <- pi$col_width 
  w <- w[cols]
  gutter <- 0
  cnt <- c()
  conv <- rs$twip_conversion
  
  exclude_top <- NULL
  if (ex_brdr)
    exclude_top <- "top"
  
  # print("Cols:")
  # print(cols)
  #print(w)
  wlvl <- get_spanning_info(rs, ts, pi, w, gutter)
  lvls <- sort(seq_along(wlvl), decreasing = TRUE)
  
  # At this point we have a data frame with labels, etc. and spanning
  # column widths, and are ready to create spanning header rows
  #print(wlvl)
  
  # Add indenting information for gap
  if (length(wlvl) > 0) {
    wlvl <- get_spanning_gap_docx(wlvl) 
    
    # Do not create gaps for all and inside border
    if (any(ts$borders %in% c("all", "inside"))) {
      for (i in 1:length(wlvl)) {
        wlvl[[i]]$indent_left <- NA
        wlvl[[i]]$indent_right <- NA
      }
    }
  }
  
  # Get borders
  brdrs <- ts$borders
  
  # Format labels for each level
  ln <- c()
  for (l in lvls) {
    
    s <- wlvl[[l]]
   # print(s)
    widths <- s$width
    names(widths) <- s$name
    algns <- s$align
    names(algns) <- s$name
    lbls <- s$label
    names(lbls) <- s$name
    cs <- s$col_span
    
    # Header Cell alignment
    ha <- c()
    for (k in seq_along(algns)) {
      
      if (algns[k] == "right")
        ha[k] <- "right"
      else if (algns[k] %in% c("center", "centre"))
        ha[k] <- "center"
      else
        ha[k] <- "left"
      
    }
    
    r <- ""
    cnt[length(cnt) + 1] <- 1 
    
    # Start row
    r <-  "<w:tr>\n"
    
    # Open device context
    pdf(NULL)
    par(family = get_font_family(rs$font), ps = rs$font_size)
    
    # Loop for labels
    for(k in seq_along(lbls)) {
      
      # Split label strings if they exceed column width
      tmp <- split_string_docx(lbls[k], widths[k], rs$units, font = rs$font)
      
      
      # b <- get_table_borders_docx(length(lvls) - l + 1, k, length(lvls) + 1, 
      #                            length(widths), brdrs, 
      #                            exclude = exclude_top)
      
      # Add colspans
      vl <- tmp$docx
      tb <- ""
      bb <- ""
      
      # Special handling of cell borders for different situations.  
      if (any(brdrs %in% c("all", "inside"))) {

        # If borders are on, turn off vertical borders appropriately
        if (cs[k] == 1) {
          if (vl == "") {
            
            if (k == 1) {
              
              bb <- '<w:tcBorders>
              <w:right w:val="nil"/>
                </w:tcBorders>'
              
            } else if (k == length(cs)) {
              
              bb <- '<w:tcBorders>
              <w:left w:val="nil"/>
                </w:tcBorders>'
              
            } else {
            
              bb <- '<w:tcBorders>
              <w:right w:val="nil"/>
              <w:left w:val="nil"/>
                </w:tcBorders>'
            } 
          
          }
        }

      } else {
        t <- ""
        b <- ""
        
        # Add top border for top and outside border specs.
        if (any(brdrs %in% c("top", "outside"))) {
          
          if (l == length( lvls)) {
            
            t <- '<w:top w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
            
          }
          
          
        }
        
        # If borders are off, add an underline if requested
        indent <- ""
        if (cs[k] > 1 | vl != "") {
            
          if (s$underline[k]) {
            b <- '<w:bottom w:val="single" w:sz="4" w:space="0" w:color="auto"/>'
          }
        }
        
        
        tb <- paste0('<w:tcBorders>', t, '</w:tcBorders>\n')
        bb <- paste0('<w:pBdr>', b, '</w:pBdr>')
      }
      
      
      # r <- paste0(r, "<w:tc>", 
      #             '<w:tcPr>', '<w:gridSpan w:val="', cs[k] , '"/>', bb,
      #             '<w:vAlign w:val="bottom"/>',
      #             '<w:tcW w:w="', round(widths[k] * conv), '" w:type="dxa"/>',
      #             '</w:tcPr>', 
      #             para(vl, ha[k], bold = s$bold[k]), "</w:tc>\n")
      
      r <- paste0(r, "<w:tc>", 
                  '<w:tcPr>', '<w:gridSpan w:val="', cs[k] , '"/>', tb,
                  '<w:vAlign w:val="bottom"/>',
                  '<w:tcW w:w="', round(widths[k] * conv), '" w:type="dxa"/>',
                  '</w:tcPr>', 
                  para(vl, ha[k], bold = s$bold[k],
                       indent_left = s$indent_left[k],
                       indent_right = s$indent_right[k],
                       borders = bb), 
                  "</w:tc>\n")
      
      
      # Add in extra lines for labels that wrap
      xtr <- tmp$lines
      
      if (xtr > cnt[length(cnt)])
        cnt[length(cnt)] <- xtr
      
    }
    dev.off()
    
    
    r <- paste0(r, "</w:tr>\n")
    
    ln[[length(ln) + 1]] <- r
    
  }
  
  
  ret <- unlist(ln)
  
  res <- list(docx = ret, 
              lines = sum(cnt))
  
  return(res)
}

#' @description Check gap and insert indent variables for DOCX
#' @details Check gap and insert indent variables for DOCX
#' @noRd
get_spanning_gap_docx <- function(wlvl, gap_twips = 100){
  
  ret <- list()
  
  for (k in 1:length(wlvl)) {
    
    df <- wlvl[[k]]
    
    # Prepare indent variables
    df$indent_left <- rep(NA, nrow(df))
    df$indent_right <- rep(NA, nrow(df))
    
    
    # Extract spanning and underline only for detection
    df_span <- df[df$span > 0 & df$underline == TRUE,] 
    
    if (nrow(df_span) < 2) {
      ret[[k]] <- df
    } else {
      # This would be used directly in w:ind, to need to convert
      gap_width <- gap_twips
      
      df_span$order <- as.numeric(stri_extract_first_regex(df_span$name, "\\d+"))

      for (i in 1:nrow(df_span)) {
        
        cur_order <- df_span$order[i]
        
        # Process by first, last, and others
        if (i == 1){
          next_order <- df_span$order[i + 1]
          
          # Continuous underline detection
          if (next_order - cur_order == 1) {
            
            # Add indenting information for reducing underline
            df_span$indent_right[i] <- gap_width/2
          }
        } else if (i == nrow(df_span)){
          pre_order <- df_span$order[i - 1]
          
          # Continuous underline detection
          if (cur_order - pre_order == 1) {
            
            # Add indenting information for reducing underline
            df_span$indent_left[i] <- gap_width/2
          }
          
        } else {
          pre_order <- df_span$order[i - 1]
          next_order <- df_span$order[i + 1]
          
          # Continuous underline detection
          if (next_order - cur_order == 1 |
              cur_order - pre_order == 1) {
            
            # Create the gap data after this spanning cell
            if (next_order - cur_order == 1) {
              
              # Add indenting information for reducing underline
              df_span$indent_right[i] <- gap_width/2
            }
            
            if (cur_order - pre_order == 1) {
              
              # Add indenting information for reducing underline
              df_span$indent_left[i] <- gap_width/2
            }
          }
        }
      } # End of data row loop
      
      # Drop order
      df_span <- df_span[, setdiff(names(df_span), "order")]

      # Stack together and sort
      df_ret <- rbind(df_span, df[!(df$span > 0 & df$underline == TRUE),])
      df_ret <- df_ret[order(df_ret$name),]
      
      ret[[k]] <- df_ret
      
    } # End of gap process
  }
  
  return(ret)
}

#' @description This function counts lines per row independently because
#' the ..row field does not account for page wrapping.  Need number
#' of lines on this particular page.
#' @noRd
get_table_body_docx <- function(rs, tbl, widths, algns, talgn, tbrdrs, 
                                ex_brdr = FALSE, frb = FALSE, styles, ts) {
  
  conv <- rs$twip_conversion
  rht <- get_row_height(round(rs$row_height * conv))
  
  if ("..blank" %in% names(tbl))
    flgs <- tbl$..blank
  else 
    flgs <- NA
  
  # Count lines per row
  rws <- c()
  border_flag <- FALSE
  exclude_top <- NULL
  if (ex_brdr)
    exclude_top <- "top"
  
  tb <- ifelse(is.na(tbl[["..blank"]]), "", tbl[["..blank"]])
  nms <- names(widths)
  nms <- nms[!is.na(nms)]
  nms <- nms[!is.controlv(nms)]
  wdths <- widths[nms]
  if (length(nms) == 1) {
    t <- as.data.frame(tbl[[nms]])
    names(t) <- nms
  } else 
    t <- tbl[ , nms]
  
  brdrs <- tbrdrs
  if (all(tbrdrs == "body"))
      brdrs <- c("top", "bottom", "left", "right")

  # Cell alignment
  ca <- c()
  for (k in seq_along(algns)) {
    if (!is.control(nms[k])) {
      if (algns[k] == "left")
        ca[k] <- "left"
      else if (algns[k] == "right")
        ca[k] <- "right"
      else if (algns[k] %in% c("center", "centre"))
        ca[k] <- "center"
    }
  }
  
  ret <- c()
  
  badj <- 0
  if (frb == TRUE) {
    badj <- 1
  }
  
  pdf(NULL)
  par(family = get_font_family(rs$font), ps = rs$font_size)
  
  # For break labels, prepare the break label columns for later use
  defs <- ts$col_defs
  break_label_col <- c()
  if (any(grepl("..break_label", names(tbl)))) {
    
    for (d in 1:length(defs)) {
      if (!is.null(defs[[d]]$break_label)){
        break_label_col <- c(break_label_col, defs[[d]]$var_c)
      }
    }
  }
  
  # Table Body
  for(i in seq_len(nrow(t))) {
    
    ret[i] <- paste0("<w:tr>")
    
    mxrw <- 1
    
    # Loop for cell values
    for(j in seq_len(ncol(t))) {
      
      
      if (!is.control(nms[j])) {
        
        b <- ""
        cs <- ""

        # if (any(brdrs %in% c("bottom", "left", "right", "outside", "body"))) {
        # 
        #   b <- get_cell_borders_docx(i + badj, j, nrow(t) + badj, 
        #                              length(nms), brdrs, tb[i])
        #   
        # 
        #   
        # }
        
        cell_border <- NULL
        if ("..group_border" %in% names(tbl)) {
          cell_border <- tbl[["..group_border"]][i]
        }
        
        # Don't draw group line for first blank row
        if (tb[i] %in% c("B", "A") & i == 1){
          cell_border <- NULL
        }
        
        b <- get_cell_borders_docx(i + badj, j, nrow(t) + badj, 
                                   length(nms), brdrs, tb[i],
                                   cell_border = cell_border)
        
        
        
        if (tb[i] %in% c("B", "A", "L")) {
          cs <-  paste0('<w:gridSpan w:val="', length(nms) , '"/>')
        }
        
        if (b != "" | cs != "") {
          b <- paste0('<w:tcPr>', b, cs, '</w:tcPr>')
        }
        
        vl <- t[i, j]
        if (!is.character(vl))
          vl <- as.character(vl)
        
        # if (tb[i] %in% c("B", "L")) {
        #   
        #   # Strip out line feeds for label rows
        #   vl <- gsub("\n", " ", vl, fixed = TRUE)
        #   
        #   # Redo splits
        #   vtmp <- split_string_docx(vl, sum(wdths), rs$units)
        #   vl <- vtmp$docx
        #   
        # }
        
        
        stl <- get_cell_styles(nms[j], styles, flgs, i, tbl)
        
        bflg <- FALSE
        if ("bold" %in% stl) {
          bflg <- TRUE
        }
        
        
        if (!(tb[i] %in% c("B", "A", "L") & j > 1)) {
          # Construct docx
          
          # Put indent information into paragraph properties
          if (rs$units == "inches") {
            twips_conv <- 1440 
          } else if (rs$units == "cm") {
            twips_conv <- 1440/2.54
          }
          ind_twips <- NA
          
          # For break labels, use the indenting info from original column
          repeat_break_label <- FALSE
          if (any(grepl("..break_label", names(tbl)))) {
            if (!is.na(tbl$..break_occur[i]) & tb[i] == "L") {
              cur_break_label_col <- break_label_col[as.numeric(tbl$..break_occur[i])]
              if (!is.null(defs[[cur_break_label_col]]$indent)) {
                ind_twips <- defs[[cur_break_label_col]]$indent*twips_conv
              }
              repeat_break_label <- TRUE
            }
          }
          
          if (!repeat_break_label) {
            if (!is.null(defs[[nms[j]]]$indent)) {
              
              ind_twips <- defs[[nms[j]]]$indent * twips_conv
              
            } else if (nms[j] == "stub" & !is.null(ts$stub)) {
              
              stub_var <- tbl$..stub_var[i]
              if (!is.null(defs[[stub_var]]$indent)) {
                
                ind_twips <- defs[[stub_var]]$indent * twips_conv
                
              }
            }
          }
 
          ret[i] <- paste0(ret[i], "<w:tc>", b,
                           para(vl, ca[j], bold = bflg, indent_left = ind_twips), "</w:tc>")
        }

        
        # Count lines in cell 
        cl <- strsplit(vl, "\n", fixed = TRUE)[[1]]
        if (length(cl) > mxrw)
          mxrw <- length(cl) 
      }
      
    }
    
    rws[i] <- mxrw
    
    # if (i == nrow(t))
    #   ret[i] <- paste0(ret[i], "</w:tr>\n</tbody>")
    # else 
    
      rht <- get_row_height(round(rs$row_height * mxrw * conv))
      ret[i] <- paste0(rht, ret[i], "</w:tr>")
    
    
  }
  
  dev.off()
  
  if ("bottom" %in% get_outer_borders(brdrs))
    border_flag <- TRUE
  
  res <- list(docx = ret,
              lines = sum(rws), 
              border_flag = border_flag)
  
  return(res)
  
  
}

#' @description Have to check wrapping on a lot of files.  May have unintended 
#' results.
#' @noRd
encodeDOCX <- function(strng) {
  
  ret <- strng
  
  ret <- gsub("&", "&amp;", ret , fixed = TRUE)
  ret <- gsub(">", "&gt;", ret , fixed = TRUE)
  ret <- gsub("<", "&lt;", ret , fixed = TRUE)
  #ret <- gsub("\n", "<br>", ret , fixed = TRUE)
  #ret <- gsub(" ", "&nbsp;", ret, fixed = TRUE)
  # if (ret == "")
  #   ret <- "&nbsp;"
  
  return(ret)
}

# Write Report HTML -------------------------------------------------------



#' @title
#' Write an HTML report to the file system
#'
#' @description
#' This function writes a report_spec object to the file system, using the
#' parameters provided in the object.
#'
#' @param x The report_spec object to write.
#' @return The report spec.
#' @noRd
write_report_html <- function(rs) {
  
  orig_path <- rs$modified_path
  
  if (file.exists(orig_path)) {
    file.remove(orig_path)
  }
  remove_image_files(orig_path)
  
  # Establish content and body sizes
  rs <- page_setup_html(rs)
  
  # Document header is mostly independent of content
  hdr <- get_html_document(rs) 
  
  # Put content in a new variable
  ls <- rs$content
  
  # Get content and break it into pages
  # Needs to return a list of pages so preview can work
  # Page numbers need to be included
  bdy <- paginate_content_html(rs, ls)
  
  # Get column widths
  rs$column_widths <- bdy[["widths"]]
  
  # Deal with preview
  if (!is.null(rs$preview)) {
    if (rs$preview < length(bdy[[1]]$pages))
      bdy[[1]]$pages <- bdy[[1]]$pages[seq(1, rs$preview)]
  }
  
  # Write content to file system
  # Later we can just return the stream
  rs <- write_content_html(rs, hdr, bdy, rs$page_template)
  
  # Update page numbers for title headers
  update_page_numbers_html(orig_path, rs$pages)
  
  return(rs)
}

#' @noRd
remove_image_files <- function(htmlpath) {
  
  f <- paste0(gsub(".html", "", basename(htmlpath)), "-")
  d <- dirname(htmlpath)
  im <- paste0(d, "/images")
  ret <- FALSE
  
  if (dir.exists(im)) {
    fls <- list.files(path = im, paste0("^", f), full.names = TRUE,
                      include.dirs = FALSE, no.. = TRUE)
    
    ret <- file.remove(fls)
  }

  return(ret)
}

#' @description Returns header for HTML document.  This is independent of content,
#' except for the page header and footer.
#' @noRd
get_html_document <- function(rs) {
  
  # Set up vectors
  ret <- c()
  
  sty <- rs$style
  
  conv <- rs$twip_conversion
  
  fnt <- rs$font
  if (tolower(rs$font) == "times")
    fnt <- "Times New Roman"
  
  u <- rs$units
  if (rs$units == "inches")
    u <- "in"
  
  # Prepare header
  ret[length(ret) + 1] <- paste0("<!DOCTYPE html>\n",
                                 "<html>\n", "<head>")
  
  ret[length(ret) + 1] <- "<style>"
  ret[length(ret) + 1] <- paste0("@media print{\n",
                                 "@page {size:", rs$orientation, ";",
                                  "margin:0;}\n",
                                  ".noprint{display:none;}}")
  ret[length(ret) + 1] <- paste0("body {\nfont-family: ", fnt,
                                 ";\nfont-size: ", rs$font_size, "pt;\n", 
                                 "margin-top: ", rs$margin_top, u, ";\n",
                                 "margin-bottom: ", 
                                 round(rs$margin_bottom/2, 3), u, ";\n",
                                 "margin-left: ", rs$margin_left, u, ";\n",
                                 "margin-right: ", rs$margin_right, u, ";\n",
                                 get_style_html(rs, "background_color"),
                                 get_style_html(rs, "text_color"),
                                 "}")
  
  brdrcolor <- get_style(rs, "border_color")
  if (brdrcolor == "")
    brdrcolor <- "black"
  
  ret[length(ret) + 1] <- paste0(".thdr {", 
                                  "border-bottom: thin solid ", brdrcolor, ";", 
                                  get_style_html(rs, "table_header_background"),
                                  get_style_html(rs, "table_header_font_color"),
                                  get_style_html(rs, "table_header_font_bold", FALSE),
                                  "}")
  ret[length(ret) + 1] <- paste0(".shdr {", 
                                 get_style_html(rs, "table_header_background"),
                                 get_style_html(rs, "table_header_font_color"),
                                 get_style_html(rs, "table_header_font_bold", FALSE),
                                 "}")
  ret[length(ret) + 1] <- paste0(".tdc {text-align:center;}")
  ret[length(ret) + 1] <- paste0(".tdl {text-align:left;}")
  ret[length(ret) + 1] <- paste0(".tdr {text-align:right;}")
  ret[length(ret) + 1] <- paste0(".tc {", 
                                 get_style_html(rs, "table_body_background"), 
                                 "}")

  ret[length(ret) + 1] <- paste0(".tbs {", 
                                 get_style_html(rs, "table_body_font_color"),
                                 "}")
  ret[length(ret) + 1] <- paste0(".tbstr {", 
                                 get_style_html(rs, "table_body_stripe"),
                                 "}")
  ret[length(ret) + 1] <- paste0(".ts {", 
                                 get_style_html(rs, "table_stub_background"), 
                                 get_style_html(rs, "table_stub_font_bold"),
                                 get_style_html(rs, "table_stub_font_color"),
                                 "}")
  ret[length(ret) + 1] <- paste0(".tlr {", 
                                 get_style_html(rs, "table_label_row_bold"), 
                                 "}")
  ret[length(ret) + 1] <- paste0("table {",
                                 "border-spacing: 0;",
                                 "border-collapse: collapse;",
                                 "}")
  ret[length(ret) + 1] <- paste0("td {padding:0px 2px 0px 2px;}")
  ret[length(ret) + 1] <- paste0("tbody td {vertical-align: top;}")
  ret[length(ret) + 1] <- "</style>"
  

  
  
  ret[length(ret) + 1] <- "</head>\n<body>"
  

  

  
  return(ret)
  
}



#' @noRd
paginate_content_html <- function(rs, ls) {
  
  ret <- c()
  last_object <- FALSE
  last_page_lines <- 0
  table_widths <- list()
  
  hrf <- has_bottom_footnotes(rs)
  if (hrf ==  FALSE)
    hrf <- has_page_footer(rs)
  
  
  # Loop through content objects
  for (i in seq_along(ls)) {
    
    pgs <- list()  # list of vectors with page rtf lines
    lns <- c()
    
    # Set last object flag
    if (i == length(ls))
      last_object <- TRUE
    else 
      last_object <- FALSE
    
    # Put content and object in variables for convenience
    cntnt <- ls[[i]] 
    obj <- cntnt$object
    
    # Remove blank row if last content object
    cbr <- cntnt$blank_row
    if (last_object) {
      if (all(cbr == "below")) 
        cbr <- "none"
      else if (all(cbr == "all"))
        cbr <- "above"
      
    }
    
    # Break each content type into a list of pages
    if (any(class(obj) == "table_spec")) {
      
      res <- create_table_pages_html(rs, cntnt, last_page_lines)
      
      # Collect multiple pages and line counts
      for (j in seq_len(length(res$page_list))) {
        pgs[[length(pgs) + 1]] <- res$page_list[[j]]$html
        lns[[length(lns) + 1]] <- res$page_list[[j]]$lines
      }
      
      # Retrieve table widths.  These are useful for debugging.
      # Assigned to returning report object.
      table_widths[[length(table_widths) + 1]] <- res$widths
      
    } else if (any(class(obj) == "text_spec")) {
      
      res <- create_text_pages_html(rs, cntnt, last_page_lines, cbr)
      for (j in seq_len(length(res$html))) {
        pgs[[length(pgs) + 1]] <- res$html[[j]]
        lns[[length(lns) + 1]] <- res$lines[[j]]
        
      }
      
    } else if (any(class(obj) == "plot_spec")) {
      
      res <- create_plot_pages_html(rs, cntnt, last_page_lines, tempdir())
      for (j in seq_len(length(res$html))) {
        pgs[[length(pgs) + 1]] <- res$html[[j]]
        lns[[length(lns) + 1]] <- res$lines[[j]]
        
      }
    }   
    
    # Store pages and lines with content objects
    # The content settings will be used when writing content
    ls[[i]]$pages <- pgs
    ls[[i]]$lines <- lns
    
    # This section of code is appending blank lines to get
    # footnotes at the bottom of the page.  The complication
    # is when there are multiple pieces of content, and user-defined
    # page breaks.  So these can't be added earlier in
    # the process.  In short, these blanks are for in between
    # pieces of content.  Blanks within a piece of content are 
    # handled in the create_table_*, create_text_*, and create_plot_* 
    # functions.
    
    # Capture last page
    last_page <- pgs[[length(pgs)]]
    
    # Capture number of lines on the last page
    last_page_lines <- lns[[length(lns)]] + last_page_lines
    
    # print(last_page_lines)
    
    if (length(pgs) > 1)
      last_page_lines <- lns[[length(lns)]]
    
    # If there is a page break or it's the last object in the
    # content list, add the blank lines if needed.
    if (rs$paper_size != "none") {
      if ((ls[[i]]$page_break | last_object) & hrf) {
  
  
        # Add extra offsets if table has a lot of borders turned on
        # to avoid undesired page wraps
        boff <- 0
        if (any(class(obj) == "table_spec") &
            any(obj$borders %in% c("all", "inside"))) {

          #boff <- round(last_page_lines * rs$border_height / rs$row_height)
          boff <- 1
        }
  
        blnks <- c()
        bl <- rs$body_line_count - last_page_lines - boff
        if (bl > 0)
          blnks <- rep("<br>", bl)
  
        last_page <- append(last_page, blnks)
        last_page_lines <- 0
  
      }
    }
    
    if (cntnt$page_break == TRUE | last_page_lines >= rs$body_line_count)
      last_page_lines <- 0

    ls[[i]]$pages[[length(pgs)]] <- last_page
    
  }
  
  
  # Can return something else if needed here
  ret <- list(widths = table_widths, pages = ls)
  
  return(ret)
  
}


# Could be consolidated with RTF.
#' @title Write out content
#' @description This loop writes out pages created in paginate_content
#' Page template items added to the report (titles/footnotes/title_header)
#' are added in this step.  That means these items need to have been accounted
#' for in the previous steps.
#' @noRd
write_content_html <- function(rs, hdr, body, pt) {
  
  # Kill existing file
  if (file.exists(rs$modified_path))
    file.remove(rs$modified_path)
  
  counter <- 0
  page <- 0
  last_object <- FALSE
  last_page <- FALSE
  page_open <- FALSE
  
  
  f <- file(rs$modified_path, open="a", encoding = "native.enc")
  
  writeLines(hdr, con = f, useBytes = TRUE)
  
  
  for (cont in body$pages) {
    
    
    # Increment counter
    counter <- counter + 1
    page <- 0
    
    # Set last_object flag
    if (counter == length(body$pages))
      last_object <- TRUE
    else 
      last_object <- FALSE
    
    ta <- "align=\"left\" "
    if (cont$align == "right")
      ta <- "align=\"right\" "
    else if (cont$align %in% c("center", "centre"))
      ta <- "align=\"center\" "
    
    
    for (pg in cont$pages) {
      
      page <- page + 1
      
      if (page == length(cont$pages))
        last_page <- TRUE
      else
        last_page <- FALSE
      
      
      #print(page_open)
      if (page_open == FALSE) {
        
        if (!is.null(rs$page_template$page_header) & 
            !is.null(rs$page_template$page_header$html))
          writeLines(update_page(rs$page_template$page_header$html,  rs$pages), 
                     con = f, useBytes = TRUE)
        
        # Write content div to keep page together
        writeLines(paste0("<div ", ta, ">"), con = f, useBytes = TRUE)
        
        
        if (!is.null(rs$title_hdr) & !is.null(pt$title_hdr$html))
          writeLines(update_page(pt$title_hdr$html,  rs$pages), con = f, 
                     useBytes = TRUE)
        
        if (!is.null(rs$titles) & !is.null(pt$titles$html))
          writeLines(update_page(pt$titles$html, rs$pages), con = f, useBytes = TRUE)
        
      }
      
      if (!is.null(pg)) {
        # Replace page numbers in title/footnote of report/content
        pg <- page_replace_html(pg, rs, "titles", rs$pages)
        pg <- page_replace_html(pg, rs, "footnotes", rs$pages)
        pg <- page_replace_html(pg, cont$object, "titles", rs$pages)
        pg <- page_replace_html(pg, cont$object, "footnotes", rs$pages)
        # 
        # if (!is.null(rs$titles)) {
        #   for (rt in rs$titles) {
        #     for (rt_string in rt$titles) {
        #       raw_title <- encodeHTML(rt_string)
        #       new_title <- update_page(raw_title, rs$pages)
        #       pg <- gsub(raw_title, new_title, pg, fixed = TRUE)
        #     }
        #   }
        # }
        # 
        writeLines(pg, con = f, useBytes = TRUE)
      }
      
      # Set page_open flag based on status of page_break and current objects
      if (last_object == FALSE & last_page == TRUE & cont$page_break == FALSE)
        page_open <- TRUE
      else 
        page_open <- FALSE
      
      if (page_open == FALSE) {
        
        if (!is.null(rs$footnotes) & !is.null(pt$footnotes$html))
          writeLines(update_page(pt$footnotes$html,  rs$pages), 
                     con = f, useBytes = TRUE)
        
        # Content div
        writeLines("</div>", con = f, useBytes = TRUE)
        
        if (!is.null(rs$page_template$page_footer) & 
            !is.null(rs$page_template$page_footer$html))
          writeLines(update_page(rs$page_template$page_footer$html, rs$pages), 
                     con = f, useBytes = TRUE)
        
        
        # Add form feed character for text page break
        if (last_object == FALSE | last_page == FALSE) {
          
          if (is.null(rs$pages))
            rs$pages <- 1
          else 
            rs$pages <- rs$pages + 1 
          
          writeLines(rs$page_break_html, con = f, useBytes = TRUE) 
          
        }
      }
      
      if (last_object == TRUE & last_page == TRUE) {
        
        rs$pages <- rs$pages + 1 
        
      }
    }
    
  }
  
  writeLines("</body></html>", con = f, useBytes = TRUE)
  
  close(f)
  
  return(rs)
  
}

#' @description Could be consolidated with RTF
#' @noRd
update_page_numbers_html <- function(path, tpg) {
  
  
  lns <- readLines(path, encoding = "UTF-8")
  
  lns <- gsub("[tpg]", tpg, lns, fixed = TRUE)
  
  f <- file(path, open = "w+", encoding = "native.enc")
  
  writeLines(lns, con = f, useBytes = TRUE)
  
  close(f)
  
}

update_page <- function(lns, pg) {
  
 ret <- gsub("[pg]", pg + 1, lns, fixed = TRUE) 
 
 return(ret)
  
}


# Setup Functions ---------------------------------------------------------

# # A lot of these values are guesses.  Need to test.
# # Row height and line height were defined independently in case
# # they are different.  Right now, appear to be the same.

#' @description Setup page for content
#' @details  Calculates available space for content and prepares text lines
#' for page template.
#' @noRd
page_setup_html <- function(rs) {
  
  debug <- FALSE
  
  if (is.null(rs$font_size))
    rs$font_size <- 10
  
  if (is.null(rs$font))
    rs$font <- "Courier"
  
  if (rs$font == "fixed")
    rs$font <- "Courier"
  
  if (rs$font_size == 8) {
    
    if (tolower(rs$font) == "times")
      rh <- 0.1178  
    else
      rh <- 0.127451  
        
    gtr <- .1
    cw <- .1    # na
    
  } else if (rs$font_size == 9) {
    
    if (tolower(rs$font) == "times")
      rh <- 0.158
    else 
      rh <- 0.148 
    cw <- .11  # na
    gtr <- .1
    
  } else if (rs$font_size == 10) {
    
    if (tolower(rs$font) == "times")
      rh <- 0.17  
    else 
      rh <- 0.1585366
    
    gtr <- .11
    cw <- .11   # na

  } else if (rs$font_size == 11) {
    
    if (tolower(rs$font) == "times")
      rh <- 0.18
    else 
      rh <- 0.168 # na
    gtr <- .1
    cw <- .11  # na
    
  } else if (rs$font_size == 12) {
    
    # inches 
    rh <- 0.1911765  
    gtr <- 0.11
    cw <- .12  #na
  }
  
  rs$border_height <- 1/72
  
  if (rs$units == "cm") {
    rh <- ccm(rh)
    cw <- ccm(cw)
    rs$border_height <- ccm(rs$border_height)
    gtr <- ccm(gtr)
  }
  
  rs$row_height <- rh
  rs$line_height <- rh
  rs$char_width <- cw
  
  # Content size is the page size minus margins, in units of measure
  rs$content_size <- get_content_size(rs)
  rs$line_size <- rs$content_size[["width"]]
  

  rs$gutter_width <- gtr
  if (rs$units == "cm")
    rs$gutter_width <- ccm(rs$gutter_width)
  
  rs$page_break_html <- paste0("<hr class=\"noprint\"><div style=\"page-break-before: always;",
                               "height:", rs$margin_top, 
                               units_html(rs$units), ";", "\"></div>")

  if (is.null(rs$user_line_count)) {
    rs$line_count <- round(rs$content_size[[1]] / rh) 
  } else
    rs$line_count <- rs$user_line_count

  if (debug) {
    print(paste("Font Size:", rs$font_size))
    print(paste("Content Height:", rs$content_size[[1]]))
    print(paste("Content Width:", rs$content_size[[2]]))
    print(paste("Line Count:", rs$line_count))
    print(paste("Line Height:", rs$line_height))
    print(paste("Gutter Width:", rs$gutter_width))
    print(paste("Char Width:", rs$char_width))
  }
  # 
  # Get page template
  pt <- page_template_html(rs)
  rs$page_template <- pt


  # Get the page template row count
  # Include all the rows associated with the page template
  rs$page_template_header_count <- sum(pt$page_header$lines, pt$titles$lines,
                                       pt$title_hdr$lines, pt$page_by$lines)
  if (debug)
    print(paste("Page Template Header Count:", rs$page_template_header_count))

  rs$page_template_footer_count <- sum(pt$footnotes$lines, pt$page_footer$lines)
  if (debug)
    print(paste("Page Template Footer Count:", rs$page_template_footer_count))

  rs$page_template_row_count <- rs$page_template_header_count +
    rs$page_template_footer_count
  if (debug)
    print(paste("Page Template Row Count:", rs$page_template_row_count))

  # Body line count is the number of rows available for content on each page
  rs$body_line_count <- rs$line_count - rs$page_template_row_count
  if (debug)
    print(paste0("Body Line Count: ", rs$body_line_count))
  
  return(rs)
}

#' @description Replace page number in title and footnote
#' @details  Replace page number in title and footnote
#' @noRd
page_replace_html <- function(pg, target, type = "titles", page) {
  if (!is.null(target[[type]])) {
    for (v in target[[type]]) {
      for (v_string in v[[type]]) {
        if (v_string != "") {
          raw_title <- encodeHTML(v_string)
          new_title <- update_page(raw_title, page)
          pg <- gsub(raw_title, new_title, pg, fixed = TRUE)
        }
      }
    }
  }
  return(pg)
}

# Create Plot Constructor     ---------------------------------------------


#' @title Create plot content
#' @description Function to create a plot specification that can be 
#' added as content to a report. The \code{create_plot} function can 
#' be used to include charts, graphs, and figures on a statistical report.  
#' The function supports plot objects returned by 
#' \code{\link[ggplot2]{ggplot}} or \code{\link[survminer]{ggsurvplot}}.  
#' It does not support the Base R 
#' \code{plot} function. 
#' @details 
#' To add a plot to a report, use the \code{create_plot} function.  The 
#' function allows you to set a width and height for the plot.  The
#' function will preserve any other geometries you apply to the plot.  See
#' the \code{\link{add_content}} function to control page breaking and 
#' blanks spaces above or below the plot.    
#' 
#' A plot specification accepts a \code{\link{page_by}} function.  If a page by
#' is applied to the plot, the plot data will be subset by the page by 
#' variable, and re-run for each subset.
#' 
#' The plot specification also accepts titles and footnotes.  See the 
#' \code{\link{titles}} and \code{\link{footnotes}} functions for further 
#' details.
#' 
#' The function 
#' also accepts a path to an image file stored on the file system instead of a 
#' plot object.  This
#' functionality was added to allow the user to create figures from other
#' plotting packages.  If you pass an image path, the image will be inserted
#' into the report at the location specified. The system supports JPEG images 
#' for all output types that support graphics. EMF images are also supported for
#' DOCX and RTF outputs. Note that EMF format may not render properly in 
#' all operating environments, and JPEG is the preferred image type.
#' 
#' @param x The plot to create.  Specifically, this parameter should be 
#' set to an object returned from a call to \code{\link[ggplot2]{ggplot}}
#' or \code{\link[survminer]{ggsurvplot}}.  This parameter also accepts 
#' a path to a JPEG file.  If a path is specified, the image will be appended
#' to the report at the point the content object is added. For RTF and DOCX files, 
#' the function additionally accepts a path to an EMF file.
#' @param height The height of the plot in the specified units of measure. 
#' @param width The width of the plot in the specified units of measure. 
#' @param borders Whether and where to place a border. Valid values are 'top',
#' 'bottom', 'left', 'right', 'all', 'none', and 'outside'.  
#' Default is 'none'.  The 'left', 'right', and 'outside' 
#' border specifications do not apply to TXT reports.
#' @return The plot specification.
#' @family plot
#' @seealso 
#' \code{\link{titles}} to add a title block to the plot,  
#' \code{\link{footnotes}} to add footnotes, and \code{\link{add_content}} 
#' to add the plot object to a report.
#' @examples
#' library(reporter)
#' library(ggplot2)
#' library(magrittr)
#' 
#' # Create temp file path
#' tmp <- file.path(tempdir(), "mtcars.rtf")
#' 
#' # Create ggplot
#' p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
#'
#' # Create plot object
#' plt <- create_plot(p, height = 4, width = 8)
#'
#' rpt <- create_report(tmp, font = "Arial", output_type = "RTF") %>%
#'   page_header("Client", "Study: XYZ") %>% 
#'   titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
#'   set_margins(top = 1, bottom = 1) %>%
#'   add_content(plt) %>%
#'   footnotes("* Motor Trend, 1974") %>% 
#'   page_footer("Time", "Confidential", "Page [pg] of [tpg]")
#' 
#' # Write out report
#' write_report(rpt)
#'
#' # Uncomment to view RTF file
#' # shell.exec(tmp)
#' @export
create_plot <- function(x, height, width, borders = "none") {
  
  
  if (!any(class(x) %in% c("gg", "ggplot", "ggcoxzph", "ggsurv", "character")))
    stop("plot object must be of type 'ggplot', 'ggsurv', or a JPG file path.")
  
  
  if (!all(borders %in% c("top", "bottom", "left", "right", 
                          "all", "none", "outside")))
    stop(paste("Borders parameter invalid.  Valid values are", 
               "'top', 'bottom', 'left', 'right', 'all', ",
               "'none', or 'outside'."))
  
  ret <- structure(list(), class = c("plot_spec", "list"))
  
  ret$plot <- x
  ret$height <- height 
  ret$width <- width
  ret$borders <- borders

  return(ret)
}



# Utilities ---------------------------------------------------------------

#' @title Prints the plot spec
#' @description A function to print the plot spec.
#' The \strong{print} function will print the plot spec in summary 
#' form.  To view all parameters, set the \code{verbose} parameter to TRUE.
#' @param x The plot spec.
#' @param ... Additional parameters to pass to the underlying print function.
#' @param verbose Whether to print in verbose form.  Default is FALSE.
#' @seealso 
#' \code{\link{create_plot}} function to create a plot specification.
#' @return The plot spec, invisibly.
#' @family plot
#' @examples 
#' txt <- create_text("Lorem ipsum dolor sit amet, consectetur...")
#' print(txt)
#'
#' # A text specification:
#' # - text: data.frame 'mtcars' 32 rows 11 cols
#' @import crayon
#' @export
print.plot_spec <- function(x, ..., verbose = FALSE){
  
  
  if (verbose == TRUE) {
    
    # If verbose mode is indicated, print values as a list
    for (nm in names(x)) {
      
      cat("$", nm, "\n", sep = "")
      if (nm == "plot") {
        
        print(unclass(x[[nm]]), ...)
      }
      else  {
        
        print(x[[nm]], ...)
      }
      cat("\n")
    }
    
    
  } else {
    
    
    grey60 <- make_style(grey60 = "#999999")
    
    
    cat(grey60("# A plot specification: \n"))
    
    if (!is.null(x$plot)) {
      
      if (typeof(x$plot) == "character") {
        
        cat(paste0("- path: '", x$plot, "'\n"))
        
        
      } else {
      
        if ("ggplot" %in% class(x$plot) & !"patchwork" %in% class(x$plot)) {
          if (!is.null(x$plot[["data"]])) {
            dat <- x$plot[["data"]]
      
            cat("- data: ")
            cat(paste0(nrow(dat), " rows, ", ncol(dat), " cols\n"))
      
            cat(paste0("- layers: ", length(x$plot[["layers"]]), "\n"))
          
          }
        }
      
      }
    } 
    
      
    cat(paste0("- height: ", x$height, "\n"))
    cat(paste0("- width: ", x$width, "\n"))
    
    if (!is.null(x$page_by)) {
      cat(paste0("- page by: ", x$page_by$var, "\n"))
      
    }
    
    print_title_header(x$title_hdr)
    
    print_titles(x$titles)
    
    print_footnotes(x$footnotes)
    
  }
  
  invisible(x)
}


# Write Text Functions -------------------------------------------------------

#' @description A function to output strings for plot content
#' @details Basic logic is to write the plot as a png file to a temporary
#' location, then put a token in the text that references the temp location.
#' Later code will pick up the image and insert it into the report.  Blank
#' lines are generated to fill up the page based on a calculation using
#' the plot height and width.
#' @param rs The Report Spec
#' @param cntnt The text content to output
#' @param lpg_rows Last page rows.
#' @noRd
create_plot_pages_text <- function(rs, cntnt, lpg_rows, tmp_dir) {
  
  if (!"report_spec" %in% class(rs))
    stop("Report spec expected for parameter rs")
  
  if (!"report_content" %in% class(cntnt))
    stop("Report Content expected for parameter cntnt")
  
  # Get plot spec 
  plt <- cntnt$object
  
  p <- plt$plot
  
  
  if (any(class(p) %in% c("character"))) {
    
    tmp_nm <- tempfile(tmpdir = tmp_dir, fileext = ".jpg")
    
    file.copy(p, tmp_nm, overwrite = TRUE)
    
    ret <- list()
    pgs <- get_plot_body(plt, tmp_nm, cntnt$align, rs,
                         lpg_rows, cntnt$blank_row, NULL, NULL, 
                         FALSE)
    
    for (pg in pgs) {
      
      ret <- c(ret, list(pg))
    }
  
    
  } else {
  
    # Get data in case we need it for page by
    raw <- p$data
    
    # Determine if there is a page by
    pgby <- NULL
    if (!is.null(rs$page_by))
      pgby <- rs$page_by
    if (!is.null(plt$page_by))
      pgby <- plt$page_by
    
    if (!is.null(pgby)) {
      if (!pgby$var %in% names(raw))
        stop("Page by variable not found in plot data.")
      
      dat_lst <- split(raw, raw[[pgby$var]])
    } else {
      dat_lst <- list(raw) 
    }
    
    u <- ifelse(rs$units == "inches", "in", rs$units)
  
    ret <- list()
    cntr <- 1
    
    for (dat in dat_lst) {
    
      tmp_nm <- tempfile(tmpdir = tmp_dir, fileext = ".jpg")
      
      p$data <- dat
      
      pgval <- NULL
      if (!is.null(pgby)) {
        # print(pgby$var)
        # print(dat)
        pgval <- dat[1, c(pgby$var)]
        
        if (!is.null(pgby$format)) {
          pgval <- fapply(pgval, pgby$format)
        }
      }
      
      # Save plot to temp file
      if (any(class(p) %in% c("ggcoxzph", "ggsurv"))) {
  
        # Deal with survival plots
        ggplot2::ggsave(tmp_nm, gridExtra::arrangeGrob(grobs = p), 
                        width =  plt$width, height = plt$height, 
                        dpi = 300, units = u)
        
        
      } else {
  
        # Any other type of plots
        ggplot2::ggsave(tmp_nm, p, width =  plt$width, height = plt$height, 
               dpi = 300, units = u)
      }
      
      # Get rtf page bodies
      # Can return multiple pages if it breaks across pages
      # Not sure what that means for a plot, but that is the logic
      pgs <- get_plot_body(plt, tmp_nm, cntnt$align, rs,
                           lpg_rows, cntnt$blank_row, pgby, pgval, 
                           cntr < length(dat_lst))
      
      # Within a content creation function, assumed that it will take 
      # care of filling out blanks for each page.  Only the last page
      # can have empty rows.
      for (pg in pgs) {
  
        ret <- c(ret, list(pg))
      }
      
      cntr <- cntr + 1
    
    }
  
  }
  
  return(ret)
}

#' Create list of vectors of strings for each page 
#' @noRd
get_plot_body <- function(plt, plot_path, align, rs,
                          lpg_rows, content_blank_row, pgby, pgval, wrap_flag) {
  
  # Get titles and footnotes
  w <- ceiling(plt$width / rs$char_width)
  ttls <- get_titles(plt$titles, w, rs$line_size, rs$uchar, rs$char_width) 
  ttl_hdr <- get_title_header(plt$title_hdr, w, rs$line_size, rs$uchar, 
                              rs$char_width)
  
  pgbys <- get_page_by(pgby, w, pgval)
  
  pltpth <- gsub("\\", "/", plot_path, fixed = TRUE)
  
  
  s <- c(paste0("```", pltpth, "|", plt$height, 
                  "|", plt$width, "|", align, "```"))
  
  h <- ceiling(plt$height / rs$row_height) + 1  # adjustment needed? Appears so.
  w <- ceiling(plt$width / rs$char_width) 

  fill <- rep("```fill```", h) 
  s <- c(s, fill)
  
  
  # Add blank above content if requested
  a <- NULL
  if (content_blank_row %in% c("both", "above"))
    a <- ""
  
  # Add top border if requested
  tbrdr <- NULL
  if (any(plt$borders %in% c("top", "all")))
    tbrdr <-  paste0(rep(rs$uchar, w), collapse = "")
  
  # Add bottom border if requested
  bbrdr <- NULL
  if (any(plt$borders %in% c("bottom", "all", "outside")))
    bbrdr <-  paste0(rep(rs$uchar, w), collapse = "")
  
  # Add blank below content if requested
  b <- NULL
  if (content_blank_row %in% c("both", "below", "outside"))
    b <- ""
  
  # Combine titles, blanks, body, and footnotes
  rws <- c(a, ttls, ttl_hdr, pgbys, tbrdr, s, bbrdr)
  
  
  ftnts <- get_page_footnotes_text(rs, plt, w, lpg_rows, length(rws),
                                   wrap_flag, content_blank_row)
  
  rws <- c(rws, ftnts)
  
  
  # Page list
  ret <- list()  
  
  # Create tmp variable for 1 page of content
  tmp <- c()
  
  # Offset the first page with remaining rows from the 
  # last page of the previous content
  offset <- lpg_rows 
  #print(paste("Offset:", offset))
  
  # Assign content to pages
  for (i in seq_along(rws)) {
    if (length(tmp) < (rs$body_line_count - offset)) {
      
      # Append to existing page
      tmp[length(tmp) + 1] <- rws[i]
      
    } else {
      
      # Start a new page
      ret[[length(ret) + 1]] <- pad_any(tmp, rs$line_size, 
                                        get_justify(align))
      tmp <- rws[i]
      
      # Set to zero on second page and leave it that way
      offset <- 0  
    }
  }
  
  # Deal with last page
  if (length(tmp) > 0 ) {
    
    
    # If page is not empty
    if (max(nchar(trimws(tmp))) > 0) {
      
      # Add last page
      ret[[length(ret) + 1]] <- pad_any(tmp,  rs$line_size, 
                                        get_justify(align))
    }
    
  }
  
  return(ret)
  
}



# Write RTF Functions -----------------------------------------------------


#' @description A function to output strings for plot content
#' @details Basic logic is to write the plot as a png file to a temporary
#' location, then put a token in the text that references the temp location.
#' Later code will pick up the image and insert it into the report.  Blank
#' lines are generated to fill up the page based on a calculation using
#' the plot height and width.
#' @param rs The Report Spec
#' @param cntnt The text content to output
#' @param lpg_rows Last page rows.
#' @noRd
create_plot_pages_rtf <- function(rs, cntnt, lpg_rows, tmp_dir) {
  
  if (!"report_spec" %in% class(rs))
    stop("Report spec expected for parameter rs")
  
  if (!"report_content" %in% class(cntnt))
    stop("Report Content expected for parameter cntnt")
  
  pgs <- list()
  cnts <- c()
  
  # Get plot spec 
  plt <- cntnt$object
  
  
  p <- plt$plot
  
  
  if (any(class(p) %in% c("character"))) {
    
    fx <- ".jpg"
    if (grepl(".emf", p, fixed = TRUE))
      fx <- ".emf"
    
    tmp_nm <- tempfile(tmpdir = tmp_dir, fileext = fx)
    
    file.copy(p, tmp_nm, overwrite = TRUE)
    
    # Get rtf page bodies
    res <- get_plot_body_rtf(plt, tmp_nm, cntnt$align, rs,
                             lpg_rows, cntnt$blank_row, NULL, NULL, 
                             FALSE)
    
    pgs[[length(pgs) + 1]] <- res$rtf
    cnts[[length(cnts) + 1]] <- res$lines
  
    
    
  } else {
    
    
    # Get data in case we need it for page by
    raw <- plt$plot$data
    
    # Determine if there is a page by
    pgby <- NULL
    if (!is.null(rs$page_by))
      pgby <- rs$page_by
    if (!is.null(plt$page_by))
      pgby <- plt$page_by
    
    if (!is.null(pgby)) {
      if (!pgby$var %in% names(raw))
        stop("Page by variable not found in plot data.")
      
      dat_lst <- split(raw, raw[[pgby$var]])
    } else {
      dat_lst <- list(raw) 
    }
    
    u <- ifelse(rs$units == "inches", "in", rs$units)
    p <- plt$plot
    ret <- list()
    cntr <- 1

    
    for (dat in dat_lst) {
      
      tmp_nm <- tempfile(tmpdir = tmp_dir, fileext = ".jpg")
      
      p$data <- dat
      
      pgval <- NULL
      if (!is.null(pgby)) {
        # print(pgby$var)
        # print(dat)
        pgval <- dat[1, c(pgby$var)]
        
        if (!is.null(pgby$format)) {
          pgval <- fapply(pgval, pgby$format)
        }
      }
      
      # Save plot to temp file
      if (any(class(p) %in% c("ggcoxzph", "ggsurv"))) {
        
        # Deal with survival plots
        ggplot2::ggsave(tmp_nm, gridExtra::arrangeGrob(grobs = p), 
                        width =  plt$width, height = plt$height, 
                        dpi = 300, units = u )
        
        
      } else if (any(class(p) %in% c("character"))) {
        
        
        file.copy(p, tmp_nm, overwrite = TRUE)
        
        
      } else {
        
        # Any other type of plots
        ggplot2::ggsave(tmp_nm, p, width =  plt$width, height = plt$height, 
                        dpi = 300, units = u)
      }
      
      # Get rtf page bodies
      res <- get_plot_body_rtf(plt, tmp_nm, cntnt$align, rs,
                           lpg_rows, cntnt$blank_row, pgby, pgval, 
                           cntr < length(dat_lst))
      
      pgs[[length(pgs) + 1]] <- res$rtf
      cnts[[length(cnts) + 1]] <- res$lines
      
      cntr <- cntr + 1
      
    }
  
  }
  
  ret <- list(rtf = pgs,
              lines = cnts)
  
  return(ret)
}

#' Create list of vectors of strings for each page 
#' @noRd
get_plot_body_rtf <- function(plt, plot_path, talign, rs,
                          lpg_rows, content_blank_row, pgby, pgval, wrap_flag) {
  
  # Default to content width
  wth <- rs$content_size[["width"]] 
  
  # If user supplies a width, override default
  if (!is.null(plt$width))
    wth <- plt$width
  


  # Get titles and footnotes
  ttls <- get_titles_rtf(plt$titles, wth, rs, talign) 
  ttl_hdr <- get_title_header_rtf(plt$title_hdr, wth, rs, talign)
  pgbys <- get_page_by_rtf(pgby, wth, pgval, rs, talign)
  
  # Get image RTF codes
  img <- get_image_rtf(plot_path, plt$width, plt$height, rs$units)
  
  # Assign table alignment codes
  if (talign == "left") {
    talgn <- "\\trql" 
  } else if (talign == "right") {
    talgn <- "\\trqr"
  } else  {
    talgn <- "\\trqc"
  }
  
  algn <- "\\qc" 

  # Convert width to twips
  w <- round(wth * rs$twip_conversion)
  
  # Get border codes
  b <- get_cell_borders(1, 1, 1, 1, plt$borders)
  
  # Concat all header codes
  hd <- paste0("\\sl0\\trowd\\trgaph0", talgn, b, "\\cellx", w, algn, " \n")
  
  ft <- paste0("\\cell\\row\n\\ql", rs$font_rtf, rs$spacing_multiplier)
  
  
  
  # Concat RTF codes for image
  img <- paste0(hd, img, ft)
  imght <- round((plt$height * rs$twip_conversion) / rs$line_height)
  
  # Add blank above content if requested
  a <- NULL
  if (content_blank_row %in% c("both", "above"))
    a <- "\\par"
  
  

  
  # Get sum of all items to this point
  lns <- sum(length(a), ttls$lines, ttl_hdr$lines, pgbys$lines, imght)
  
  # Get footnotes, filler, and content blank line
  ftnts <- get_page_footnotes_rtf(rs, plt, wth, lpg_rows, lns,
                                   wrap_flag, content_blank_row, talign)
  
  
  # On LibreOffice, have to protect the table from the title width or
  # the table row will inherit the title row width. Terrible problem.
  tpt <- "{\\pard\\fs1\\sl0\\par}"
  if (any(plt$borders %in% c("all", "top", "outside"))) {
    if (ttls$border_flag | rs$page_template$titles$border_flag |  
        rs$page_template$title_hdr$border_flag)
      tpt <- ""
  }
  if (titles_aligned(plt$titles, plt$title_hdr))
    tpt <- ""
  
  # Prevent infection of widths on LibreOffice.
  bpt <- "{\\pard\\fs1\\sl0\\par}"
  if (any(plt$borders %in% c("all", "top", "outside"))) {
    if (!is.null(ftnts)) {
      if (ftnts$border_flag)
        bpt <- ""
    }
    
    if (!is.null(rs$page_template$footnotes)) {
      if (rs$page_template$footnotes$border_flag)
        bpt <- ""
    }
  }
  if (footnotes_aligned(plt$footnotes)) 
    bpt <- ""
  
  # Combine titles, blanks, body, and footnotes
  rws <- c(a, ttls$rtf, ttl_hdr$rtf, pgbys$rtf, tpt, img, bpt)
  
  # Combine everything
  rws <- c(rws, ftnts$rtf)
  lns <- sum(lns, ftnts$lines)
  

  # Page list
  ret <- list(rtf = rws,
              lines = lns)  
  
  
  return(ret)
  
}


# Write HTML Functions -----------------------------------------------------


#' @description A function to output strings for plot content
#' @details Basic logic is to write the plot as a png file to a temporary
#' location, then put a token in the text that references the temp location.
#' Later code will pick up the image and insert it into the report.  Blank
#' lines are generated to fill up the page based on a calculation using
#' the plot height and width.
#' @param rs The Report Spec
#' @param cntnt The text content to output
#' @param lpg_rows Last page rows.
#' @noRd
create_plot_pages_html <- function(rs, cntnt, lpg_rows, tmp_dir) {
  
  if (!"report_spec" %in% class(rs))
    stop("Report spec expected for parameter rs")
  
  if (!"report_content" %in% class(cntnt))
    stop("Report Content expected for parameter cntnt")
  
  pgs <- list()
  cnts <- c()
  
  # Get plot spec 
  plt <- cntnt$object
  
  p <- plt$plot
  
  
  if (any(class(p) %in% c("character"))) {
    
    tmp_nm <- tempfile(tmpdir = tmp_dir, fileext = ".jpg")
    
    file.copy(p, tmp_nm, overwrite = TRUE)

    # Get rtf page bodies
    res <- get_plot_body_html(plt, tmp_nm, cntnt$align, rs,
                              lpg_rows, cntnt$blank_row, NULL, NULL, 
                              FALSE)
    
    pgs[[length(pgs) + 1]] <- res$html
    cnts[[length(cnts) + 1]] <- res$lines
    
    
  } else {
    

    # Get data in case we need it for page by
    raw <- plt$plot$data
    
    # Determine if there is a page by
    pgby <- NULL
    if (!is.null(rs$page_by))
      pgby <- rs$page_by
    if (!is.null(plt$page_by))
      pgby <- plt$page_by
    
    if (!is.null(pgby)) {
      if (!pgby$var %in% names(raw))
        stop("Page by variable not found in plot data.")
      
      dat_lst <- split(raw, raw[[pgby$var]])
    } else {
      dat_lst <- list(raw) 
    }
    
    u <- ifelse(rs$units == "inches", "in", rs$units)
    p <- plt$plot
    ret <- list()
    cntr <- 1

    
    for (dat in dat_lst) {
      
      tmp_nm <- tempfile(tmpdir = tmp_dir, fileext = ".jpg")
      
      p$data <- dat
      
      pgval <- NULL
      if (!is.null(pgby)) {
        # print(pgby$var)
        # print(dat)
        pgval <- dat[1, c(pgby$var)]
        
        if (!is.null(pgby$format)) {
          pgval <- fapply(pgval, pgby$format)
        }
      }
      
      # Save plot to temp file
      if (any(class(p) %in% c("ggcoxzph", "ggsurv"))) {
        
        # Deal with survival plots
        ggplot2::ggsave(tmp_nm, gridExtra::arrangeGrob(grobs = p), 
                        width =  plt$width, height = plt$height, 
                        dpi = 300, units = u )
        
        
      } else {
        
        # Any other type of plots
        ggplot2::ggsave(tmp_nm, p, width =  plt$width, height = plt$height, 
                        dpi = 300, units = u)
      }
      
      # Get rtf page bodies
      res <- get_plot_body_html(plt, tmp_nm, cntnt$align, rs,
                               lpg_rows, cntnt$blank_row, pgby, pgval, 
                               cntr < length(dat_lst))
      
      pgs[[length(pgs) + 1]] <- res$html
      cnts[[length(cnts) + 1]] <- res$lines
      
      cntr <- cntr + 1
      
    }
  }
  
  ret <- list(html = pgs,
              lines = cnts)
  
  return(ret)
}

#' Create list of vectors of strings for each page 
#' @noRd
get_plot_body_html <- function(plt, plot_path, talign, rs,
                              lpg_rows, content_blank_row, pgby, pgval, wrap_flag) {
  
  # Default to content width
  wth <- rs$content_size[["width"]] 
  
  # If user supplies a width, override default
  if (!is.null(plt$width))
    wth <- plt$width
  

  # Get titles and footnotes
  ttls <- get_titles_html(plt$titles, wth, rs, talign) 
  ttl_hdr <- get_title_header_html(plt$title_hdr, wth, rs, talign)
  
  exclude_top <- NULL
  if (ttls$border_flag == TRUE | ttl_hdr$border_flag == TRUE) {
    exclude_top <- "top"
  
    pgbys <- get_page_by_html(pgby, wth, pgval, rs, talign, TRUE)
  } else {
    
    pgbys <- get_page_by_html(pgby, wth, pgval, rs, talign, FALSE)
  }
  
  if (is.null(exclude_top)) {
    if (pgbys$border_flag)
      exclude_top <- "top"
    
  }
  
  # Get image RTF codes
  img <- get_image_html(plot_path, rs$modified_path, plt, rs$units)

  
  # algn <- "\\qc" 
  u <- rs$units
  if (u == "inches")
    u <- "in"
  
  # Convert width to twips
  w <- paste0("width:", round(wth, 3), u, ";")
  
  # Get border codes
  b <- get_cell_borders_html(1, 1, 1, 1, plt$borders, exclude = exclude_top)
  
  # Concat all header codes
  hd <- paste0("<table style =\"", w, "\">\n", 
               "<tr><td style=\"", b, "\">\n")
  
  ft <- "</td></tr></table>\n"

  
  # Concat RTF codes for image
  img <- paste0(hd, img, ft)
  
  # ** Do something with this **
  imght <- floor(plt$height / rs$line_height)
  
  # Add blank above content if requested
  a <- NULL
  if (content_blank_row %in% c("both", "above"))
    a <- "<br>"
  
  
  # Get sum of all items to this point
  lns <- sum(length(a), ttls$lines, ttl_hdr$lines, pgbys$lines, imght)
  
  if ("bottom" %in% get_outer_borders(plt$borders))
    extp <- TRUE
  else 
    extp <- FALSE 
  
  # Get footnotes, filler, and content blank line
  ftnts <- get_page_footnotes_html(rs, plt, wth, lpg_rows, lns,
                                  wrap_flag, content_blank_row, talign, extp)
  
  # Combine titles, blanks, body, and footnotes
  rws <- c(a, ttls$html, ttl_hdr$html, pgbys$html, img)
  
  
  # Combine everything
  rws <- c(rws, ftnts$html)
  lns <- sum(lns, ftnts$lines)
  
  
  # Page list
  ret <- list(html = rws,
              lines = lns)  
  
  
  return(ret)
  
}




# Write PDF Functions -----------------------------------------------------


#' @description A function to output strings for plot content
#' @details Basic logic is to write the plot as a png file to a temporary
#' location, then put a token in the text that references the temp location.
#' Later code will pick up the image and insert it into the report.  Blank
#' lines are generated to fill up the page based on a calculation using
#' the plot height and width.
#' @param rs The Report Spec
#' @param cntnt The text content to output
#' @param lpg_rows Last page rows.
#' @noRd
create_plot_pages_pdf <- function(rs, cntnt, lpg_rows, tmp_dir) {
  
  if (!"report_spec" %in% class(rs))
    stop("Report spec expected for parameter rs")
  
  if (!"report_content" %in% class(cntnt))
    stop("Report Content expected for parameter cntnt")
  
  pgs <- list()
  cnts <- c()
  pnts <- c()
  
  # Get plot spec 
  plt <- cntnt$object
  p <- plt$plot
  
  if (any(class(p) %in% c("character"))) {
    
    tmp_nm <- tempfile(tmpdir = tmp_dir, fileext = ".jpg")
    
    file.copy(p, tmp_nm, overwrite = TRUE)
    
    ys <- sum(rs$page_template$titles$points, rs$page_template$title_hdr$points,
              rs$page_template$page_header$points, (lpg_rows * rs$line_height))
    
    # Get pdf page bodies
    res <- get_plot_body_pdf(plt, tmp_nm, cntnt$align, rs, row_offset,
                             cntnt$blank_row, NULL, NULL, 
                             FALSE, ystart = ys)
    
    pgs[[length(pgs) + 1]] <- res$pdf
    cnts[length(cnts) + 1] <- res$lines
    pnts[length(pnts) + 1] <- res$lines * rs$line_height
    
    
  } else {
  
    # Get data in case we need it for page by
    raw <- plt$plot$data
    
    # Determine if there is a page by
    pgby <- NULL
    if (!is.null(rs$page_by))
      pgby <- rs$page_by
    if (!is.null(plt$page_by))
      pgby <- plt$page_by
    
    if (!is.null(pgby)) {
      if (!pgby$var %in% names(raw))
        stop("Page by variable not found in plot data.")
      
      dat_lst <- split(raw, raw[[pgby$var]])
    } else {
      dat_lst <- list(raw) 
    }
    
    u <- ifelse(rs$units == "inches", "in", rs$units)
    p <- plt$plot
    ret <- list()
    cntr <- 1

    row_offset <- lpg_rows
    
    for (dat in dat_lst) {
      
      if (cntr > 1)
        row_offset <- 0
      
      tmp_nm <- tempfile(tmpdir = tmp_dir, fileext = ".jpg")
      
      p$data <- dat
      
      pgval <- NULL
      if (!is.null(pgby)) {
        # print(pgby$var)
        # print(dat)

        pgval <- dat[1, c(pgby$var)]
        
        if (!is.null(pgby$format)) {
          pgval <- fapply(pgval, pgby$format)
        }
      }
      
      # Save plot to temp file
      if (any(class(p) %in% c("ggcoxzph", "ggsurv"))) {
        
        # Deal with survival plots
        ggplot2::ggsave(tmp_nm, gridExtra::arrangeGrob(grobs = p), 
                        width =  plt$width, height = plt$height, 
                        dpi = 300, units = u )
        
        
      } else {
        
        # Any other type of plots
        ggplot2::ggsave(tmp_nm, p, width =  plt$width, height = plt$height, 
                        dpi = 300, units = u)
      }
      ys <- sum(rs$page_template$titles$points, rs$page_template$title_hdr$points,
                rs$page_template$page_header$points, (row_offset * rs$line_height))
      
      # Get pdf page bodies
      res <- get_plot_body_pdf(plt, tmp_nm, cntnt$align, rs, row_offset,
                               cntnt$blank_row, pgby, pgval, 
                               cntr < length(dat_lst), ystart = ys)
      
      pgs[[length(pgs) + 1]] <- res$pdf
      cnts[length(cnts) + 1] <- res$lines
      pnts[length(pnts) + 1] <- res$lines * rs$line_height
      cntr <- cntr + 1
      
  }
  }
  
  ret <- list(pdf = pgs,
              lines = cnts, 
              points = pnts)
  
  return(ret)
}

#' Create list of page_text or page_image objects for a single page 
#' @noRd
get_plot_body_pdf <- function(plt, plot_path, talign, rs,
                              lpg_rows, content_blank_row, pgby, pgval, 
                              wrap_flag, ystart = 0) {
  cnt <- 0
  lh <- rs$line_height
  conv <- rs$point_conversion
  bh <- rs$border_height
  
  # Default to content width
  wth <- rs$content_size[["width"]] 
  
  # If user supplies a width, override default
  if (!is.null(plt$width))
    wth <- plt$width
  
  
  # Add blank above content if requested
  aln <- 0
  if (content_blank_row %in% c("both", "above")) {
    aln <- 1
    ystart <- ystart + lh
  }
  
  # Get titles and footnotes
  ttls <- get_titles_pdf(plt$titles, wth, rs, talign, ystart = ystart) 
  ttl_hdr <- get_title_header_pdf(plt$title_hdr, wth, rs, 
                                  talign, ystart = ystart)
  
  if (ttls$border_flag | ttl_hdr$border_flag) {
    pgbys <- get_page_by_pdf(pgby, wth, pgval, rs, talign, 
                           ystart = sum(ystart, ttls$points, ttl_hdr$points) - 2)
  
  } else {
    
    pgbys <- get_page_by_pdf(pgby, wth, pgval, rs, talign, 
                             ystart = sum(ystart, ttls$points, ttl_hdr$points))
  }
  
  # Get image PDF codes
  #img <- get_image_pdf(plot_path, plt$width, plt$height, rs$units)
  
  if (talign == "right") {
    lb <- rs$content_size[["width"]] - wth
    rb <- rs$content_size[["width"]]
  } else if (talign %in% c("center", "centre")) {
    lb <- (rs$content_size[["width"]] - wth) / 2
    rb <- wth + lb
  } else {
    lb <- 0
    rb <- wth
  }
  # 
  # algn <- "\\qc" 
  
  # Convert width to twips
  w <- round(wth * rs$point_conversion)

  if ((ttls$border_flag | ttl_hdr$border_flag) & pgbys$points > 0)
    ypos <- sum(ystart, ttls$points, ttl_hdr$points, pgbys$points) - 2
  else 
    ypos <- sum(ystart, ttls$points, ttl_hdr$points, pgbys$points) 
  
  lnstrt <- ceiling(ypos / rs$line_height)
  

  imght <- round((plt$height * rs$point_conversion) / lh)
  

  rws <- list()
  rws[[length(rws) + 1]] <- page_image(plot_path,
                             height = plt$height,
                             width = plt$width,
                             align = talign,
                             line_start = lnstrt + aln)

  
  yline <- ceiling(ypos + (plt$height * rs$point_conversion)) 
  brdrs <- strip_borders(plt$borders)
  
  ylpos <- ypos - lh + bh - 1
  
  # Top border
  if (any(brdrs %in% c("all", "outside", "top"))) {
    
    rws[[length(rws) + 1]] <- page_hline(lb * conv, 
                                         ylpos, 
                                         (rb - lb) * conv) 
    
  }
  
  # Bottom border
  if (any(brdrs %in% c("all", "outside", "bottom"))) {
    
    rws[[length(rws) + 1]] <- page_hline(lb * conv, 
                                         ylpos + (yline - ypos), 
                                         (rb - lb) * conv) 
    
  }
  
  # Left border
  if (any(brdrs %in% c("all", "outside", "left"))) {
    
    
    rws[[length(rws) + 1]] <- page_vline(lb * conv, 
                                         ylpos, 
                                         yline - ypos) 
    
  }
  
  # Right border
  if (any(brdrs %in% c("all", "outside", "right"))) {
    
    
    rws[[length(rws) + 1]] <- page_vline(rb * conv, 
                                         ylpos, 
                                         yline - ypos) 
    
  }
  
  

  
  # Get footnotes, filler, and content blank line
  
  # The y position of plot footnote should consider plot bottom border height
  if (any(brdrs %in% c("all", "outside", "bottom"))) {
    ft_ystart <- yline - 1
  } else {
    ft_ystart <- yline - bh - 1
  }
  
  ftnts <- get_page_footnotes_pdf(rs, plt, wth, lpg_rows, ft_ystart,
                                  wrap_flag, content_blank_row, talign)
  
  bln <- 0
  if (content_blank_row %in% c("both", "below")) {
    bln <-  1
  }
  
  # Add remaining page content.
  # This needs to be done now so everything is on the page
  # Different than all other output types.
  if (ttls$lines > 0) {
    rws<- append(rws, ttls$pdf)
  }
  if (ttl_hdr$lines > 0) {
    rws <- append(rws, ttl_hdr$pdf)
  }
  if (pgbys$lines > 0) {
    rws <- append(rws, pgbys$pdf)
  }
  if (ftnts$lines > 0) {
    rws <- append(rws, ftnts$pdf)
  }
  

  # Get sum of all items to this point
  lns <- sum(aln, ttls$lines, ttl_hdr$lines, pgbys$lines, 
             imght, ftnts$lines, bln)
  
  # Page list
  ret <- list(pdf = rws,
              lines = lns,
              points =  ylpos - ystart)  
  
  
  return(ret)
  
}



# Write DOCX Functions -----------------------------------------------------


#' @description A function to output strings for plot content
#' @details Basic logic is to write the plot as a png file to a temporary
#' location, then put a token in the text that references the temp location.
#' Later code will pick up the image and insert it into the report.  Blank
#' lines are generated to fill up the page based on a calculation using
#' the plot height and width.
#' @param rs The Report Spec
#' @param cntnt The text content to output
#' @param lpg_rows Last page rows.
#' @param imgCnt The number of images in document.
#' @noRd
create_plot_pages_docx<- function(rs, cntnt, lpg_rows, tmp_dir, imgCnt) {
  
  if (!"report_spec" %in% class(rs))
    stop("Report spec expected for parameter rs")
  
  if (!"report_content" %in% class(cntnt))
    stop("Report Content expected for parameter cntnt")

  pgs <- list()
  cnts <- c()
  imgs <- list()
  
  relIndex <- rs$relIndex + imgCnt
  
  # Get plot spec 
  plt <- cntnt$object
  p <- plt$plot
  
  
  if (any(class(p) %in% c("character"))) {
    
    fx <- ".jpg"
    if (grepl(".emf", p, fixed = TRUE))
      fx <- ".emf"
    
    tmp_nm <- tempfile(tmpdir = tmp_dir, fileext = fx)
    
    file.copy(p, tmp_nm, overwrite = TRUE)
    
    # Get docx page bodies
    res <- get_plot_body_docx(plt, tmp_nm, cntnt$align, rs,
                              lpg_rows, cntnt$blank_row, NULL, NULL, 
                              FALSE, relIndex)
    
    pgs[[length(pgs) + 1]] <- res$docx
    cnts[[length(cnts) + 1]] <- res$lines
    
    relID <- paste0("rId", relIndex)
    imgs[[relID]] <- tmp_nm
    relIndex <- relIndex + 1
    
  } else {
    
    
    # Get data in case we need it for page by
    raw <- plt$plot$data
    
    # Determine if there is a page by
    pgby <- NULL
    if (!is.null(rs$page_by))
      pgby <- rs$page_by
    if (!is.null(plt$page_by))
      pgby <- plt$page_by
    
    if (!is.null(pgby)) {
      if (!pgby$var %in% names(raw))
        stop("Page by variable not found in plot data.")
      
      dat_lst <- split(raw, raw[[pgby$var]])
    } else {
      dat_lst <- list(raw) 
    }
    
    u <- ifelse(rs$units == "inches", "in", rs$units)
    p <- plt$plot
    ret <- list()
    cntr <- 1
    
    
    for (dat in dat_lst) {
      
      tmp_nm <- tempfile(tmpdir = tmp_dir, fileext = ".jpg")
      
      p$data <- dat
      
      pgval <- NULL
      if (!is.null(pgby)) {
        # print(pgby$var)
        # print(dat)
        pgval <- dat[1, c(pgby$var)]
        
        if (!is.null(pgby$format)) {
          pgval <- fapply(pgval, pgby$format)
        }
      }
      
      # Save plot to temp file
      if (any(class(p) %in% c("ggcoxzph", "ggsurv"))) {
        
        # Deal with survival plots
        ggplot2::ggsave(tmp_nm, gridExtra::arrangeGrob(grobs = p), 
                        width =  plt$width, height = plt$height, 
                        dpi = 300, units = u )
        
        
      } else {
        
        # Any other type of plots
        ggplot2::ggsave(tmp_nm, p, width =  plt$width, height = plt$height, 
                        dpi = 300, units = u)
      }
      
      # Get rtf page bodies
      res <- get_plot_body_docx(plt, tmp_nm, cntnt$align, rs,
                                lpg_rows, cntnt$blank_row, pgby, pgval, 
                                cntr < length(dat_lst), relIndex)
      
      pgs[[length(pgs) + 1]] <- res$docx
      cnts[[length(cnts) + 1]] <- res$lines
      
      relID <- paste0("rId", relIndex)
      imgs[[relID]] <- tmp_nm
      relIndex <- relIndex + 1
      
      cntr <- cntr + 1
      
    }
  }
  
  ret <- list(docx = pgs,
              lines = cnts,
              images = imgs)
  
  return(ret)
}

#' Create list of vectors of strings for each page 
#' @noRd
get_plot_body_docx <- function(plt, plot_path, talign, rs,
                               lpg_rows, content_blank_row, pgby, pgval, 
                               wrap_flag, rID) {
  
  conv <- rs$twip_conversion
  
  # Default to content width
  wth <- rs$content_size[["width"]] 
  
  # If user supplies a width, override default
  if (!is.null(plt$width))
    wth <- plt$width
  
  
  # Get titles and footnotes
  ttls <- get_titles_docx(plt$titles, wth, rs, talign,
                          content_brdrs = plt$borders) 
  ttl_hdr <- get_title_header_docx(plt$title_hdr, wth, rs, talign)
  
  exclude_top <- NULL
  if (ttls$border_flag == TRUE | ttl_hdr$border_flag == TRUE) {
    exclude_top <- "top"
    
    pgbys <- get_page_by_docx(pgby, wth, pgval, rs, talign, TRUE)
  } else {
    
    pgbys <- get_page_by_docx(pgby, wth, pgval, rs, talign, FALSE)
  }
  
  if (is.null(exclude_top)) {
    if (pgbys$border_flag)
      exclude_top <- "top"
    
  }
  
  algn <- talign
  if (talign == "centre")
    algn <- "center"
  
  # algn <- "\\qc" 
  # This height and width is for the image scaling.
  u <- rs$units
  iconv <- 914400  # Image unit conversion. Not even sure what units this is.
  if (u == "inches") {
    hgth <- (plt$height - .2) * iconv 
    wdth <- (plt$width - .2) * iconv 
    
  } else {
    hgth <- (cin(plt$height) - .2) * iconv 
    wdth <- (cin(plt$width) - .2) * iconv
    
  }
  
  # Get image DOCX codes
  img <- get_image_docx(rID, algn, hgth, wdth)
  
  
  # Get indent codes for alignment
  ta <- get_indent_docx(talign, rs$line_size, round(wth * conv), 
                        rs$base_indent, plt$borders, conv)
  
  grd <- get_col_grid(c(img = round(wth * conv)), 1)
  

  # Get border codes
  tb <- get_table_borders_docx(plt$borders)
  
  tcb <- paste('<w:tcBorders>', 
               ifelse(any(plt$borders %in% c("all", "outside", "top")),
							'<w:top w:val="single" w:sz="4" w:space="0" w:color="auto"/>', ""),
							ifelse(any(plt$borders %in% c("all", "outside", "bottom")),
							'<w:bottom w:val="single" w:sz="4" w:space="0" w:color="auto"/>', ""),
						  '</w:tcBorders>')
  
  # Concat all header codes
  hd <- paste0("<w:tbl>",  "<w:tblPr>",
               '<w:tblStyle w:val="TableGrid"/>',
               rs$cell_margin,
               '<w:tblW w:w="', round(wth * conv),'"',
               ' w:type="dxa"/>', ta, tb,
               "</w:tblPr>\n", grd,
               "<w:tr><w:tc>\n",
               '<w:tcPr><w:tcW w:w="', round(wth * conv),'"/>', tcb,'</w:tcPr>')
  
  ft <- paste0("</w:tc></w:tr></w:tbl>\n")
  
  
  # Concat DOCX codes for image
  img <- paste0(hd, img, ft)
  
  # ** Do something with this **
  imght <- floor(plt$height / rs$line_height)
  
  # Add blank above content if requested
  a <- NULL
  if (content_blank_row %in% c("both", "above"))
    a <- rs$blank_row
  
  
  # Get sum of all items to this point
  lns <- sum(length(a), ttls$lines, ttl_hdr$lines, pgbys$lines, imght)
  
  if ("bottom" %in% get_outer_borders(plt$borders))
    extp <- TRUE
  else 
    extp <- FALSE 
  
  # Get footnotes, filler, and content blank line
  ftnts <- get_page_footnotes_docx(rs, plt, wth, lpg_rows, lns,
                                   wrap_flag, content_blank_row, talign, extp,
                                   content_brdrs = plt$borders)
  
  # Combine titles, blanks, body, and footnotes
  rws <- c(a, ttls$docx, ttl_hdr$docx, pgbys$docx, img)
  
  
  # Combine everything
  rws <- c(rws, ftnts$docx)
  lns <- sum(lns, ftnts$lines)
  
  
  # Page list
  ret <- list(docx = rws,
              lines = lns)  
  
  
  return(ret)
  
}


get_image_docx <- function(id, algn, hgth, wdth, type = "") {
 
  
 ret <- paste0('<w:p>',
               '<w:pPr><w:jc w:val="', algn, '"/></w:pPr>',
    '<w:r>
    <w:rPr>
    <w:noProof/>
    </w:rPr>
    <w:drawing>
    <wp:inline distT="0" distB="0" distL="0" distR="0" 
       wp14:anchorId="53259B4E" wp14:editId="79F9BC2E">
      <wp:extent cx="', wdth, '" cy="', hgth, '"/>
        <wp:effectExtent l="0" t="0" r="0" b="0"/>
          <wp:docPr id="', id, '" name="Picture ', id, '"/>
            <wp:cNvGraphicFramePr>
            <a:graphicFrameLocks
          xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/main" noChangeAspect="1"/>
            </wp:cNvGraphicFramePr>
            <a:graphic
          xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/main">
            <a:graphicData uri="http://schemas.openxmlformats.org/drawingml/2006/picture">
            <pic:pic
          xmlns:pic="http://schemas.openxmlformats.org/drawingml/2006/picture">
            <pic:nvPicPr>
            <pic:cNvPr id="', id, '" name="Picture ', id, '"/>
            <pic:cNvPicPr/>
            </pic:nvPicPr>
            <pic:blipFill>
            <a:blip r:embed="rId', type, id, '" cstate="print">
            <a:extLst>
            <a:ext uri="{28A0092B-C50C-407E-A947-70E740481C1C}">
            <a14:useLocalDpi
          xmlns:a14="http://schemas.microsoft.com/office/drawing/2010/main" val="0"/>
            </a:ext>
            </a:extLst>
            </a:blip>
            <a:stretch>
            <a:fillRect/>
            </a:stretch>
            </pic:blipFill>
            <pic:spPr>
            <a:xfrm>
            <a:off x="0" y="0"/>
            <a:ext cx="', wdth, '" cy="', hgth, '"/>
            </a:xfrm>
            <a:prstGeom prst="rect">
            <a:avLst/>
            </a:prstGeom>
            </pic:spPr>
            </pic:pic>
            </a:graphicData>
            </a:graphic>
            </wp:inline>
            </w:drawing>
            </w:r>
            </w:p>')
 
 #ret <- paste0('<w:p><w:r><w:t>Hello</w:t></w:r></w:p>')
 
 
 return(ret)
  
}

# Create Text Constructor     ---------------------------------------------

#' @title Create text content
#' @description Function to create a text specification that can be 
#' added as content to a report.   The text content can be used to include 
#' analysis on a statistical report.  A text specification is an S3 
#' object of class 'text_spec'. 
#' @details 
#' To add plain text to a report, use the \code{create_text} function.  The 
#' function allows you to set a width and alignment for the text.  The
#' function will preserve any other formatting you apply to the text.  See
#' the \code{\link{add_content}} function to control page breaking and 
#' blanks spaces above or below the text.  
#' 
#' The text specification also accepts titles and footnotes.  See the 
#' \code{\link{titles}} and \code{\link{footnotes}} functions for further 
#' details.
#' 
#' @param txt The text to create.
#' @param width The width of the text in the specified units of measure.  If 
#' no width is specified, the full page width will be used.
#' @param align How to align the text within the content area.  Valid values
#' are 'left', 'right', 'center', or 'centre'.  Default is 'left'.
#' @param borders Whether and where to place a border. Valid values are 'top',
#' 'bottom', 'left', 'right', 'all', 'none', and 'outside'.  
#' Default is 'none'.  The 'left', 'right', and 'outside' 
#' border specifications only apply to RTF reports.
#' @return The text specification.
#' @family text
#' @seealso 
#' \code{\link{titles}} to add a title block to the text,  
#' \code{\link{footnotes}} to add footnotes, and \code{\link{add_content}} 
#' to add the text object to a report.
#' @examples 
#' library(reporter)
#' library(magrittr)
#' 
#' # Create temp file path
#' tmp <- file.path(tempdir(), "mtcars.txt")
#' 
#' # Create dummy text
#' dt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
#'   "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
#'   "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
#'   "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ", 
#'   "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
#'   "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
#'   "qui officia deserunt mollit anim id est laborum.")
#' 
#' # Create the text object
#' txt <- create_text(dt) %>% 
#'   titles("Text Content 1.0", "Sample Text Report") %>% 
#'   footnotes("* Cicero, 1st century BCE")
#' 
#' # Create the report object
#' rpt <- create_report(tmp, orientation = "portrait") %>% 
#'   add_content(txt) 
#' 
#' # Write the report to the file system
#' write_report(rpt)
#' 
#' # Write the report to console
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' #                                Text Content 1.0
#' #                               Sample Text Report
#' # 
#' # Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor
#' # incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis
#' # nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.
#' # Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore
#' # eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt
#' # in culpa qui officia deserunt mollit anim id est laborum.
#' # 
#' # * Cicero, 1st century BCE
#' #
#' @export
create_text <- function(txt, width = NULL, align = "left", borders = "none") {
  
  if (!"character" %in% class(txt))
    stop("value must be of class 'character'")
  
  if (is.null(align))
    align = "left"
  
  if (!align %in% c("left", "right", "center", "centre"))
    stop(paste0("align value invalid.  Valid values are 'left', 'right',",
                "'center', or 'centre'."))
  
  if (!is.null(width)) {
    
    if (!is.numeric(width))
      stop("width must be a number.")
    
    if (width <= 0)
      stop("width must be greater than zero.")
    
  }
  
  if (!all(borders %in% c("top", "bottom", "left", "right", 
                          "all", "none", "outside")))
    stop(paste("Borders parameter invalid.  Valid values are", 
               "'top', 'bottom', 'left', 'right', 'all', ",
               "'none', or 'outside'."))
  
  ret <- structure(list(), class = c("text_spec", "list"))
  
  ret$text <- txt
  ret$align <- align
  ret$width <- width
  ret$borders <- borders
  
  return(ret)
  
}


# Utilities ---------------------------------------------------------------




#' @title Prints the text spec
#' @description A function to print the text spec.
#' The \strong{print} function will print the text spec in summary 
#' form.  To view all parameters, set the \code{verbose} parameter to TRUE.
#' @param x The text spec.
#' @param ... Additional parameters to pass to the underlying print function.
#' @param verbose Whether to print in verbose form.  Default is FALSE.
#' @seealso 
#' \code{\link{create_text}} function to create a text specification.
#' @return The text spec, invisibly.
#' @family text
#' @examples 
#' txt <- create_text("Lorem ipsum dolor sit amet, consectetur...",
#'                    align = "left", width = 3)
#' txt
#'
#' # A text specification: 6 words
#' # - text: Lorem ipsum dolor sit amet, consectetur...
#' # - width: 3
#' # - align: left
#' @import crayon
#' @export
print.text_spec <- function(x, ..., verbose = FALSE){
  
  
  if (verbose == TRUE) {
    
    print(unclass(x))
    
  } else {
    
    
    grey60 <- make_style(grey60 = "#999999")

    # Print header
    cat(grey60("# A text specification: "))
    
    # Print 100 characters of text 
    if (!is.null(x$text)) {
      
      wc <- lengths(strsplit(x$text, " "))
      cat(grey60(paste0(as.character(wc), " words\n")))
      
      cat("- text: ")
      if (nchar(x$text) > 100)
        cat(paste0(substr(x$text, 1, 100), "..."))
      else 
        cat(x$text)
      
      cat("\n")
            
    } else {
      
     cat("\n") 
    }
    
    if (!is.null(x$width)) 
      cat(paste0("- width: ", x$width, "\n"))
    
    if (!is.null(x$align)) 
      cat(paste0("- align: ", x$align, "\n"))
    
    
    print_title_header(x$title_hdr)

    print_titles(x$titles)
    
    print_footnotes(x$footnotes)
    
    
  }
  
  invisible(x)
}


# Write Text Functions -------------------------------------------------------

#' @description A function to output strings for plain text content
#' @details Basic logic is to wrap any text to the available line width, 
#' then then just dump it out.  All formatting is left to the user.
#' @param rs The Report Spec
#' @param cntnt The text content to output
#' @param lpg_rows Last page rows.

#' @noRd
create_text_pages_text <- function(rs, cntnt, lpg_rows) {
  
  if (!"report_spec" %in% class(rs))
    stop("Report spec expected for parameter rs")
  
  if (!"report_content" %in% class(cntnt))
    stop("Report Content expected for parameter cntnt")
  
  txt <- cntnt$object
  
  # Default width to the overall line size
  w <- rs$line_size
  
  # If user supplies a width, override default
  if (!is.null(txt$width))
    w <- round(txt$width / rs$char_width)
  
  rws <- get_text_body(rs, txt, w, rs$body_line_count, lpg_rows, cntnt$blank_row)
  
  return(rws)
}



#' Create list of vectors of strings for each page 
#' @import stringi
#' @noRd
get_text_body <- function(rs, txt, line_width, line_count, lpg_rows, 
                          content_blank_row) {
  
  # Get titles and footnotes
  ttls <- get_titles(txt$titles, line_width, rs$line_size, 
                     rs$uchar, rs$char_width) 
  ftnts <- get_footnotes(txt$footnotes, line_width,  rs$line_size, 
                         rs$uchar, rs$char_width) 
  ttl_hdr <- get_title_header(txt$title_hdr, line_width, rs$line_size, 
                              rs$uchar, rs$char_width)
  
  # Wrap the text 
  s <- stri_wrap(unlist(
    strsplit(txt$text, split = "\n", fixed = TRUE)), 
    width = line_width, normalize = FALSE)
  
  
  # Make sure rows are the same length
  s <- pad_any(s, line_width, 
               get_justify(txt$align))
  
  # Add blank above content if requested
  a <- NULL
  if (content_blank_row %in% c("both", "above"))
      a <- ""
  
  # Add top border if requested
  tbrdr <- NULL
  if (any(txt$borders %in% c("top", "all", "outside")))
    tbrdr <-  paste0(rep(rs$uchar, line_width), collapse = "")
  
  # Add bottom border if requested
  bbrdr <- NULL
  if (any(txt$borders %in% c("bottom", "all", "outside")))
    bbrdr <- paste0(rep(rs$uchar, line_width), collapse = "")
  
  # Add blank below content if requested
  b <- NULL
  if (content_blank_row %in% c("both", "below"))
    b <- ""
  
  # Combine titles, blanks, body, and footnotes
  rws <- c(a, ttls, ttl_hdr, tbrdr, s, bbrdr)
  
  # Set to true for now.  Need to fix text paging below.
  wrap_flag <- FALSE
  
  # Get footnotes
  ftnts <- get_page_footnotes_text(rs, txt, line_width, lpg_rows, 
                                   length(rws), wrap_flag, content_blank_row)
  # Append footnotes 
  rws <- c(rws, ftnts)
  
  # Page list
  ret <- list()  
  
  # Create tmp variable for 1 page of content
  tmp <- c()
  
  # Offset the first page with remaining rows from the 
  # last page of the previous content
  offset <- lpg_rows 
  #print(paste("Offset:", offset))
  
  # Assign content to pages
  for (i in seq_along(rws)) {
    if (length(tmp) < (line_count - offset)) {
      
      # Append to existing page
      tmp[length(tmp) + 1] <- rws[i]
      
    } else {
      
      # Start a new page
      ret[[length(ret) + 1]] <- pad_any(tmp, line_width, 
                                        get_justify(txt$align))
      tmp <- rws[i]
      
      # Set to zero on second page and leave it that way
      offset <- 0  
    }
  }
  
  # Deal with last page
  if (length(tmp) > 0 ) {
    

    # If page is not empty
    if (max(nchar(trimws(tmp))) > 0) {
    
      # Add last page
      ret[[length(ret) + 1]] <- pad_any(tmp, line_width, 
                                       get_justify(txt$align))
    }
    
  }
  
  return(ret)
  
}


# Write RTF Functions -------------------------------------------------------


#' @noRd
create_text_pages_rtf <- function(rs, cntnt, lpg_rows, content_blank_row) {
  
  if (!"report_spec" %in% class(rs))
    stop("Report spec expected for parameter rs")
  
  if (!"report_content" %in% class(cntnt))
    stop("Report Content expected for parameter cntnt")
  
  txt <- cntnt$object
  
  # Default to content width
  w <- rs$content_size[["width"]] 

  # If user supplies a width, override default
  if (!is.null(txt$width))
    w <- txt$width
  
  res <- get_text_body_rtf(rs, txt, w, rs$body_line_count, 
                           lpg_rows, content_blank_row, cntnt$align)
  

  return(res)
  
  
}

#' Create list of vectors of strings for each page 
#' @import stringi
#' @noRd
get_text_body_rtf <- function(rs, txt, width, line_count, lpg_rows, 
                              content_blank_row, talgn) {


  # Get content titles and footnotes
  ttls <- get_titles_rtf(txt$titles, width, rs, talgn) 
  ttl_hdr <- get_title_header_rtf(txt$title_hdr, width, rs, talgn)
  ftnts <- get_footnotes_rtf(txt$footnotes, width, rs, talgn) 
  
  t <- sum(ttls$lines, ftnts$lines, ttl_hdr$lines)
  hgt <- rs$body_line_count - t

  # Break text content into pages if necessary
  tpgs <- split_text(txt$text, hgt, width, rs$font, 
                 rs$font_size, rs$units, lpg_rows)
  
  # Capture rtf pages and line counts
  txtpgs <- tpgs$text
  lns <- tpgs$lines
  
  # Calculate text width in twips
  w <- round(width * rs$twip_conversion)
  
  # Get content alignment codes
  if (talgn == "right") 
    tgn <- "\\trqr"
  else if (talgn %in% c("center", "centre"))
    tgn <- "\\trqc"
  else 
    tgn <- "\\trql"
  
  # Get text alignment codes
  if (txt$align == "right") 
    algn <- "\\qr"
  else if (txt$align %in% c("center", "centre"))
    algn <- "\\qc"
  else 
    algn <- "\\ql"

  # Get cell border codes
  b <- get_cell_borders(1, 1, 1, 1, txt$borders)  
  
  # Prepare row header and footer
  rwhd <- paste0("\\trowd\\trgaph0", tgn, b, "\\cellx", w, algn, " ")
  rwft <- paste0("\\cell\\row")
  
  ret <- list()
  cnt <- c()
  
  # Gather rtf and line counts for each page
  for (i in seq_along(txtpgs)) {
    
    if (i == length(txtpgs))
      wrap_flag <- FALSE
    else 
      wrap_flag <- TRUE
    
    pg <- txtpgs[[i]]
    
    # Put line ending on all but last line
    if (length(pg) > 1) {
      s <- paste0(pg[seq(1, length(pg) - 1)], "\\line ")
      s <- c(s, pg[length(pg)])
    } else
      s <- pg

    # Add blank above content if requested
    a <- NULL
    if (i == 1 & content_blank_row %in% c("both", "above"))
      a <- "\\par"
  
    # Deal with cell padding.  Don't count this in line count.
    cp <- paste0("\\li", rs$cell_padding, "\\ri", rs$cell_padding)
    

    # Sum up lines
    cnts <- sum(length(a),  ttls$lines, ttl_hdr$lines, lns[[i]])
    
    # Get footnotes
    ftnts <- get_page_footnotes_rtf(rs, txt, width, lpg_rows, cnts,
                                    wrap_flag, content_blank_row, talgn)
    
    # On LibreOffice, have to protect the table from the title width or
    # the table row will inherit the title row width. Terrible problem.
    tpt <- "{\\pard\\fs1\\sl0\\par}"
    if (any(txt$borders %in% c("all", "top", "outside"))) {
      if (ttls$border_flag | rs$page_template$titles$border_flag |  
          rs$page_template$title_hdr$border_flag)
        tpt <- ""
    }
    if (titles_aligned(txt$titles, txt$title_hdr))
      tpt <- ""
    
    
    # Prevent infection of widths on LibreOffice.
    bpt <- "{\\pard\\fs1\\sl0\\par}"
    if (any(txt$borders %in% c("all", "top", "outside"))) {
      if (!is.null(ftnts)) {
        if (ftnts$border_flag)
          bpt <- ""
      }
      
      if (!is.null(rs$page_template$footnotes)) {
        if (rs$page_template$footnotes$border_flag)
          bpt <- ""
      }
    }
    if (footnotes_aligned(txt$footnotes)) 
      bpt <- ""
    
    # Combine titles, blanks, body, and footnotes
    rws <- c(a, cp, ttls$rtf, ttl_hdr$rtf, tpt, cp, rwhd, s, rwft, bpt)
    
    ret[[length(ret) + 1]] <- c(rws, cp, ftnts$rtf)
    cnt[[length(cnt) + 1]] <- sum(cnts, ftnts$lines)
    
  }
  
  res <- list(rtf = ret, lines = cnt)

  return(res)
  
}


#' @description lines is the number of lines per page or cell before breaking.
#' Width is the width of the page or cell.
#' @import stringi
#' @noRd
split_text <- function(txt, lines, width, font, 
                           font_size, units, offset = 0) {
  
  pgs <- c()
  lnlngth <- 0
  ln <- c()
  cnt <- 0
  lns <- c()
  cnts <- c()
  wdths <- c()
  wdth <- c()
  
  # Break string into paragraphs defined by user
  paras <- unlist(stri_split_fixed(txt, "\n"))
  
  for (k in seq_along(paras)) {
  
    # Split text into words
    wrds <- strsplit(trimws(paras[k]), " ", fixed = TRUE)[[1]]
    
    if (length(wrds) == 0)
      wrds <- paras[k]
    
    # Set font
    f <- "mono"
    if (tolower(font) == "arial")
      f <- "sans"
    else if (tolower(font) == "times")
      f <- "serif"
    
    lngths <- c()
    
    
    lngths <- (get_text_width(wrds, units = units, font = font, font_size = font_size) + 
       get_text_width(" ", units = units, font = font, font_size = font_size)) * 1.03
    
    # Loop through words and add up lines
    for (i in seq_along(wrds)) {
      
      lnlngth <- lnlngth + lngths[i] 
      if (lnlngth <= width)
        ln <- append(ln, wrds[i])
      else {
        cnt <- cnt + 1
        
        # If cnt exceeds allowed lines per page, start a new page
        if (cnt <= lines - offset) {
          lns <- append(lns, paste(ln, collapse = " "))
          wdth[length(wdth) + 1] <- lnlngth - lngths[i] 
          ln <- wrds[i]
          lnlngth <- lngths[i]
        } else {
          
          # Assign current lines and counts
          pgs[[length(pgs) + 1]] <- lns
          cnts[[length(cnts) + 1]] <- length(lns)
          wdths[[length(wdths) + 1]] <- wdth
          wdth <- lnlngth  - lngths[i] 
          
          # Assign overflow to next page
          lns <- paste(ln, collapse = " ")
          ln <- wrds[i]
          lnlngth <- lngths[i]
  
          # After first page, set this to zero.
          offset <- 0
          cnt <- 1
        }
      }
    }
    
    if (length(lns) > 0 | length(ln) > 0) {
      cnt <- cnt + 1
      if (cnt <= lines - offset) {
        # Not a new page
        lns <- append(lns, paste(ln, collapse = " "))
        wdth[length(wdth) + 1] <- lnlngth
        ln <- c()
        lnlngth <- 0
       } else {
         
        # Add to page 
        pgs[[length(pgs) + 1]] <- lns
        cnts[[length(cnts) + 1]] <- length(lns)
        wdths[[length(wdths) + 1]] <- wdth

        # Deal with overflow
        lns <-  paste(ln, collapse = " ")
        wdth <- lnlngth  
        
        # Reset line so it won't append any more
        ln <- c()
        lnlngth <- 0
        
        #wdth <- lnlngth
        #lnlngth <- width
        
        # Assign overflow to next page
        # lns <- paste(ln, collapse = " ")
        # ln <- wrds[i]
        # lnlngth <- lngths[i]
        
        # After first page, set this to zero.
        offset <- 0
        cnt <- 1
      }
    }
  
  }
  
  # This block covers any remaining text left over 
  # from the previous block.
  if (length(lns) > 0 | length(ln) > 0) {
    cnt <- cnt + 1
    if (cnt <= lines - offset) {
      if (!is.null(ln)) {
        lns <- append(lns, paste(ln, collapse = " "))
        wdth[length(wdth) + 1] <- lnlngth

      }
      
      pgs[[length(pgs) + 1]] <- lns
      cnts[[length(cnts) + 1]] <- length(lns)
      wdths[[length(wdths) + 1]] <- wdth
      
    } else {
      pgs[[length(pgs) + 1]] <- lns
      cnts[[length(cnts) + 1]] <- length(lns)
      wdths[[length(wdths) + 1]] <- wdth
      
      if (!is.null(ln)) {
        lns <-  paste(ln, collapse = " ")
        wdth[length(wdth) + 1] <- lnlngth
        pgs[[length(pgs) + 1]] <- lns
        cnts[[length(cnts) + 1]] <- length(lns)
        wdths[[length(wdths) + 1]] <- wdth
      }
    }

  }
  
  res <- list(text = pgs, 
              lines = cnts,
              widths = wdths)
  
  return(res)
  
}


# Write HTML Functions -------------------------------------------------------


#' @noRd
create_text_pages_html <- function(rs, cntnt, lpg_rows, content_blank_row) {
  
  if (!"report_spec" %in% class(rs))
    stop("Report spec expected for parameter rs")
  
  if (!"report_content" %in% class(cntnt))
    stop("Report Content expected for parameter cntnt")
  
  txt <- cntnt$object
  
  # Default to content width
  w <- rs$content_size[["width"]] 
  
  # If user supplies a width, override default
  if (!is.null(txt$width))
    w <- txt$width
  
  res <- get_text_body_html(rs, txt, w, rs$body_line_count, 
                           lpg_rows, content_blank_row, cntnt$align)
  
  
  return(res)
  
  
}

#' Create list of vectors of strings for each page 
#' @import stringi
#' @noRd
get_text_body_html <- function(rs, txt, width, line_count, lpg_rows, 
                              content_blank_row, talgn) {
  
  
  # Get content titles and footnotes
  ttls <- get_titles_html(txt$titles, width, rs, talgn) 
  ttl_hdr <- get_title_header_html(txt$title_hdr, width, rs, talgn)
  
  ftnts <- get_footnotes_html(txt$footnotes, width, rs, talgn, FALSE) 
  
  exclude_top <- NULL
  if (ttls$border_flag | ttl_hdr$border_flag)
    exclude_top <- "top"
  
  t <- sum(ttls$lines, ftnts$lines, ttl_hdr$lines)
  hgt <- line_count - t
  
  # Break text content into pages if necessary
  tpgs <- split_text(txt$text, hgt, width, rs$font, 
                         rs$font_size, rs$units, lpg_rows)
  
  # Capture rtf pages and line counts
  txtpgs <- tpgs$text
  lns <- tpgs$lines
  
  u <- rs$units
  if (u == "inches")
    u <- "in"
  
  # Calculate text width in twips
  w <- paste0("width:", round(width, 3), u, ";")
  
  
  # Get text alignment codes
  if (txt$align == "right") 
    algn <- "text-align:right;"
  else if (txt$align %in% c("center", "centre"))
    algn <- "text-align:center;"
  else 
    algn <- "text-align:left;"
  
  # Get cell border codes
  b <- get_cell_borders_html(1, 1, 1, 1, txt$borders, exclude = exclude_top)  
  
  # Prepare row header and footer
  rwhd <- paste0("<table ", 
                 " style=\"", w, algn, "\">\n<tr><td style=\"", b, "\">")
  rwft <- paste0("</td></tr>\n</table>\n")
  
  ret <- list()
  cnt <- c()
  
  # Gather rtf and line counts for each page
  for (i in seq_along(txtpgs)) {
    
    if (i == length(txtpgs))
      wrap_flag <- FALSE
    else 
      wrap_flag <- TRUE
    
    pg <- txtpgs[[i]]
    
    # Put line ending on all but last line
    if (length(pg) > 1) {
      s <- paste0(pg[seq(1, length(pg) - 1)], "<br>\n")
      s <- c(s, pg[length(pg)])
    } else
      s <- pg
    
    # Add blank above content if requested
    a <- NULL
    if (i == 1 & content_blank_row %in% c("both", "above"))
      a <- "<br>\n"
    
    
    # Sum up lines
    cnts <- sum(length(a),  ttls$lines, ttl_hdr$lines, lns[[i]])
    
    # Get footnotes
    if ("bottom" %in% get_outer_borders(txt$borders))
      ftnts <- get_page_footnotes_html(rs, txt, width, lpg_rows, cnts,
                                       wrap_flag, content_blank_row, talgn, TRUE)
    else 
      ftnts <- get_page_footnotes_html(rs, txt, width, lpg_rows, cnts,
                                    wrap_flag, content_blank_row, talgn, FALSE)
    
    
    # Combine titles, blanks, body, and footnotes
    rws <- c(a, ttls$html, ttl_hdr$html, 
             rwhd, s, rwft)
    
    ret[[length(ret) + 1]] <- c(rws, ftnts$html)
    cnt[[length(cnt) + 1]] <- sum(cnts, ftnts$lines)
    
  }
  
  res <- list(html = ret, lines = cnt)
  
  return(res)
  
}


# Write PDF Functions -------------------------------------------------------


#' @noRd
create_text_pages_pdf <- function(rs, cntnt, lpg_rows, content_blank_row) {
  
  if (!"report_spec" %in% class(rs))
    stop("Report spec expected for parameter rs")
  
  if (!"report_content" %in% class(cntnt))
    stop("Report Content expected for parameter cntnt")
  
  txt <- cntnt$object
  
  # Default to content width
  w <- rs$content_size[["width"]] 
  
  # If user supplies a width, override default
  if (!is.null(txt$width))
    w <- txt$width
  
  ys <- sum(rs$page_template$titles$points, rs$page_template$title_hdr$points,
            rs$page_template$page_header$points)

  
  res <- get_text_body_pdf(rs, txt, w, rs$body_line_count, 
                           lpg_rows, content_blank_row, 
                           cntnt$align, ystart = ys)
  
  
  return(res)
  
  
}

#' Create list of vectors of strings for each page 
#' @import stringi
#' @noRd
get_text_body_pdf <- function(rs, txt, width, line_count, lpg_rows, 
                              content_blank_row, talgn, ystart = 0) {
  
  lh <- rs$line_height
  bh <- rs$border_height
  bs <- rs$border_spacing
  conv <- rs$point_conversion
  
  # Get content titles and footnotes
  ttls <- get_titles_pdf(txt$titles, width, rs, talgn, 
                         ystart = ystart) 
  ttl_hdr <- get_title_header_pdf(txt$title_hdr, width, rs, talgn, 
                                  ystart = ystart)
  ftnts <- get_footnotes_pdf(txt$footnotes, width, rs, talgn) 
  
  # Nice that if titles are taller because of the font size,
  # this will account for it.
  p <- sum(ttls$points, ftnts$points, ttl_hdr$points)
  t <- ceiling(p/lh)

  hgt <- line_count - t
  
  # If last page already filled up, start a new page
  if (ceiling(lpg_rows) >= floor(hgt))
    lpg_rows <- 0
  
  if (content_blank_row %in% c("above", "all"))
    lpg_rows <- lpg_rows + 1
  
  # Break text content into pages if necessary
  tpgs <- split_text(txt$text, hgt, width, rs$font, 
                     rs$font_size, rs$units, lpg_rows)
  
  # Capture pages, line counts, and widths
  txtpgs <- tpgs$text
  lns <- tpgs$lines
  wdths <- tpgs$widths
  
  # Calculate text width in points
  w <- round(width * rs$point_conversion)
  
  # Get content alignment codes
  if (talgn == "right") {
    lb <- rs$content_size[["width"]] - width
    rb <- rs$content_size[["width"]]
  } else if (talgn %in% c("center", "centre")) {
    lb <- (rs$content_size[["width"]] - width) / 2
    rb <- width + lb
  } else {
    lb <- 0
    rb <- width
  }
  
  fp_ttls <- list(pdf = NULL, lines = 0)
  fp_ttl_hdr <- list(pdf = NULL, lines = 0)

  ret <- list()
  cnt <- c()
  pnt <- c()

  # Gather pdf and line counts for each page
  for (i in seq_along(txtpgs)) {
    
    cnts <- t
    pnts <- p
    rws <- list()
    
    if (i == length(txtpgs))
      wrap_flag <- FALSE
    else 
      wrap_flag <- TRUE
    
    pg <- txtpgs[[i]]
    pw <- wdths[[i]]
    
    if (lpg_rows > 0 & i == 1) {
      
      # Get starting y coordinate
      yline <- sum(ystart, (lpg_rows * lh))
      
      if (content_blank_row %in% c("above", "both")) {
        yline <- yline + lh
        cnts <- cnts + 1 
      }
      
      fp_ttls <- get_titles_pdf(txt$titles, width, rs, talgn, 
                             ystart = yline) 
      fp_ttl_hdr <- get_title_header_pdf(txt$title_hdr, width, rs, talgn, 
                                      ystart = yline)
      

      yline <- sum(yline, fp_ttls$points, fp_ttl_hdr$points)
      
      
    } else {
      
      fp_ttls <- list(lines = 0)
      fp_ttl_hdr <- list(lines = 0)
    
      # Get starting y coordinate
      yline <- sum(ystart, ttls$points, ttl_hdr$points)
    
    }
    
    
    if (ttls$border_flag | ttl_hdr$border_flag)
      ypos <- yline - lh + 1
    else 
      ypos <- yline - lh + bs
    
    for (ln in seq_along(pg)) {
      
      rws[[length(rws) + 1]] <- page_text(pg[ln], rs$font_size, 
                                          xpos = get_points(lb, 
                                                            rb,
                                                            pw[ln],  
                                                            units = rs$units,
                                                            align = txt$align),
                                          ypos = yline)
      yline <- yline + lh
      cnts <- cnts + 1
    }
    
    yend <- yline - lh + bs
    
    #yline <- ceiling(ypos + (plt$height * rs$point_conversion)) 
    brdrs <- strip_borders(txt$borders)
    
    # Top border
    if (any(brdrs %in% c("all", "outside", "top"))) {
      
      rws[[length(rws) + 1]] <- page_hline(lb * conv, 
                                           ypos, 
                                           (rb - lb) * conv) 
      
    }
    
    border_flag <- FALSE
    # Bottom border
    if (any(brdrs %in% c("all", "outside", "bottom"))) {
      
      rws[[length(rws) + 1]] <- page_hline(lb * conv, 
                                           yend , 
                                           (rb - lb) * conv) 
      border_flag <- TRUE
      
    }
    
    # Left border
    if (any(brdrs %in% c("all", "outside", "left"))) {
      
      
      rws[[length(rws) + 1]] <- page_vline(lb * conv, 
                                           ypos, 
                                           yend - ypos) 
      
    }
    
    # Right border
    if (any(brdrs %in% c("all", "outside", "right"))) {
      
      
      rws[[length(rws) + 1]] <- page_vline(rb * conv, 
                                           ypos, 
                                           yend - ypos) 
      
    }
    
    # rs, spec, spec_width, lpg_rows, ystart,
    # wrap_flag, content_blank_row, talgn,
    # brdr_flag = FALSE
    
    if (border_flag)
      ys <- yline + bs
    else 
      ys <- yline + 1
    
    # Get footnotes
    ftnts <- get_page_footnotes_pdf(rs, txt, width, lpg_rows, ys,
                                    wrap_flag, content_blank_row, talgn,
                                    brdr_flag = border_flag)
    
    # Add blank above content if requested
    if (i ==length(txtpgs) & content_blank_row %in% c("both", "below")) {
      cnts <- cnts + 1
      yline <- yline + lh

    }

    pnts <- pnts + (cnts * lh)
    
    
    # Add remaining page content.
    if (fp_ttls$lines > 0) {
      rws<- append(rws, fp_ttls$pdf)
    } else {
      if (ttls$lines > 0) {
        rws<- append(rws, ttls$pdf)
      } 
    }
    if (fp_ttl_hdr$lines > 0) {
      rws <- append(rws, fp_ttl_hdr$pdf)
    } else {
      if (ttl_hdr$lines > 0) {
        rws <- append(rws, ttl_hdr$pdf)
      }
    }


    if (ftnts$lines > 0) {
        rws <- append(rws, ftnts$pdf)
    }
    
    ret[[length(ret) + 1]] <- rws
    cnt[length(cnt) + 1]  <- cnts
    pnt[length(pnt) + 1]  <- pnts
  }
  
  res <- list(pdf = ret, 
              lines = cnt,
              points = pnt)
  
  return(res)
  
}




# Write DOCX Functions -------------------------------------------------------


#' @noRd
create_text_pages_docx <- function(rs, cntnt, lpg_rows, content_blank_row) {
  
  if (!"report_spec" %in% class(rs))
    stop("Report spec expected for parameter rs")
  
  if (!"report_content" %in% class(cntnt))
    stop("Report Content expected for parameter cntnt")
  
  txt <- cntnt$object
  
  # Default to content width
  w <- rs$content_size[["width"]] 
  
  # If user supplies a width, override default
  if (!is.null(txt$width))
    w <- txt$width
  
  res <- get_text_body_docx(rs, txt, w, rs$body_line_count, 
                            lpg_rows, content_blank_row, cntnt$align)
  
  
  return(res)
  
  
}

#' Create list of vectors of strings for each page 
#' @import stringi
#' @noRd
get_text_body_docx <- function(rs, txt, width, line_count, lpg_rows, 
                               content_blank_row, talgn) {
  
  conv <- rs$twip_conversion
  
  # Get content titles and footnotes
  ttls <- get_titles_docx(txt$titles, width, rs, talgn, 
                          content_brdrs = txt$borders) 
  ttl_hdr <- get_title_header_docx(txt$title_hdr, width, rs, talgn)
  
  ftnts <- get_footnotes_docx(txt$footnotes, width, rs, talgn, FALSE) 
  
  exclude_top <- NULL
  if (ttls$border_flag | ttl_hdr$border_flag)
    exclude_top <- "top"
  
  t <- sum(ttls$lines, ftnts$lines, ttl_hdr$lines)
  hgt <- line_count - t
  
  # Break text content into pages if necessary
  tpgs <- split_text(txt$text, hgt, width, rs$font, 
                     rs$font_size, rs$units, lpg_rows)
  
  # Capture rtf pages and line counts
  txtpgs <- tpgs$text
  lns <- tpgs$lines
  
  u <- rs$units
  if (u == "inches")
    u <- "in"
  
  # Get text alignment codes
  if (txt$align == "right")
    algn <- "right"
  else if (txt$align %in% c("center", "centre"))
    algn <- "center"
  else
    algn <- "left"
  
  
  tw <- round(width * conv)
  
  # Get indents for alignment
  ta <- get_indent_docx(talgn, rs$line_size, tw, 
                        rs$base_indent, txt$borders, conv)
  
  grd <- get_col_grid(c(text=tw), 1)
   
  # Get cell border codes
  b <- get_table_borders_docx(txt$borders, exclude = exclude_top)  
  
  cb <- ""
  if (any(txt$borders %in% c("outside", "top", "bottom"))) {
    cb <- get_cell_borders_docx(1, 1, 1, 1, txt$borders)
    
  }
   
  # Prepare row header and footer
  rwhd <- paste0("<w:tbl>\n", '<w:tblPr>',
                 rs$cell_margin,
                 '<w:tblW w:w="', tw, '"/>', 
                 ta,
                 b,
                 "</w:tblPr>", grd,
                 "<w:tr><w:tc>",
                 '<w:tcPr><w:tcW w:w="', tw, '"/>', cb, '</w:tcPr>',
                 '<w:p>',
                 '<w:pPr><w:jc w:val="', algn, '"/></w:pPr>')
  rwft <- paste0("</w:p></w:tc></w:tr>\n</w:tbl>\n")
  
  ret <- list()
  cnt <- c()
  
  # Gather docx and line counts for each page
  for (i in seq_along(txtpgs)) {
    
    if (i == length(txtpgs))
      wrap_flag <- FALSE
    else 
      wrap_flag <- TRUE
    
    pg <- txtpgs[[i]]
    
    # Put line ending on all but last line
    # if (length(pg) > 1) {
    #   s <- paste0(pg[seq(1, length(pg) - 1)], "<br>\n")
    #   s <- c(s, pg[length(pg)])
    # } else
      s <- run(pg)
    
    # Add blank above content if requested
    a <- NULL
    if (i == 1 & content_blank_row %in% c("both", "above"))
      a <- rs$blank_row
    
    
    # Sum up lines
    cnts <- sum(length(a),  ttls$lines, ttl_hdr$lines, lns[[i]])
    
    # Get footnotes
    if ("bottom" %in% get_outer_borders(txt$borders))
      ftnts <- get_page_footnotes_docx(rs, txt, width, lpg_rows, cnts,
                                       wrap_flag, content_blank_row, talgn, TRUE, 
                                       content_brdrs = txt$borders)
    else 
      ftnts <- get_page_footnotes_docx(rs, txt, width, lpg_rows, cnts,
                                       wrap_flag, content_blank_row, talgn, FALSE, 
                                       content_brdrs = txt$borders)
    
    
    # Combine titles, blanks, body, and footnotes
    rws <- c(a, ttls$docx, ttl_hdr$docx, 
             rwhd, s, rwft)
    
    ret[[length(ret) + 1]] <- c(rws, ftnts$docx, rs$table_break)
    cnt[[length(cnt) + 1]] <- sum(cnts, ftnts$lines)
    
  }
  
  res <- list(docx = ret, lines = cnt)
  
  return(res)
  
}

context("Text Spec Tests")


test_that("create_text parameter checks work as expected.", {
  
  expect_error(create_text(NULL))
  
  expect_error(create_text(1))
  
  expect_error(create_text("hello", align = "fork"))
  
  expect_error(create_text("Hello", width = - 4))
               
               
  
})

test_that("text print() function works as expected.", {
  
  txt <- create_text("Hello", width = 2, align = "center") %>% 
    titles("Test Title") %>% 
    footnotes("Test footnote")
  
  #print(txt)
  
  #print(txt, verbose = TRUE)
  
  expect_equal(TRUE, TRUE)
  
})


test_that("split_text() works as expected.", {
  
  cnt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
                "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
                "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
                "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ", 
                "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
                "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
                "qui officia deserunt mollit anim id est laborum.")
  
  
  res <- split_text(cnt, 4, 5, "Arial", 12, "inches")
  
  res
  expect_equal(length(res$text), 2)
  expect_equal(length(res$widths), 2)
  expect_equal(length(res$widths[[1]]), 4)
  expect_equal(length(res$widths[[2]]), 3)
  
  res <- split_text(cnt, 4, 3, "Arial", 12, "inches", 2)
  
  res
  expect_equal(length(res$text), 4)
  expect_equal(length(res$lines), 4)
  expect_equal(length(res$widths), 4)
  expect_equal(res$lines[[1]], 2)
  expect_equal(res$lines[[2]], 4)
  expect_equal(res$lines[[3]], 4)
  expect_equal(res$lines[[4]], 3)
  
  
  cnt2 <- paste0("Lorem ipsum dolor sit amet,\n consectetur adipiscing elit, ",
                "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
                "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
                "nisi ut aliquip ex ea commodo consequat.\n Duis aute irure dolor \nin ", 
                "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
                "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
                "qui officia deserunt\n mollit anim id est laborum.")


  res <- split_text(cnt2, 4, 5, "Arial", 12, "inches")

  res
  
  expect_equal(length(res), 3)
  expect_equal(length(res$text), 3)
  expect_equal(length(res$lines), 3)
  expect_equal(length(res$widths), 3)
  expect_equal(res$lines[[1]], 4)
  expect_equal(res$lines[[2]], 4)
  expect_equal(res$lines[[3]], 2)
  
  res <- split_text(cnt2, 4, 5, "Arial", 12, "inches", 2)
  
  res
  
  expect_equal(length(res), 3)
  expect_equal(length(res$text), 3)
  expect_equal(length(res$lines), 3)
  expect_equal(length(res$widths), 3)
  expect_equal(res$lines[[1]], 2)
  expect_equal(res$lines[[2]], 4)
  expect_equal(res$lines[[3]], 4)
  
})


test_that("get_text_body_pdf() works as expected.", {
  
  cnt2 <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
                 "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
                 "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
                 "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ", 
                 "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
                 "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
                 "qui officia deserunt mollit anim id est laborum.")
  
  txt <- create_text(cnt2) %>% 
    titles("Here is a title")
  
  rpt <- create_report("", "PDF", font = "Arial", font_size = 12)
  
  rpt <- page_setup_pdf(rpt)
  
  res <- get_text_body_pdf(rpt, txt, 6, 20, 0, FALSE, "center", 25) 
  res
  
  expect_equal(length(res$lines), 1)
  expect_equal(res$lines[1], 8)
  
  
  res <- get_text_body_pdf(rpt, txt, 6, 5, 0, FALSE, "center", 25) 
  res
  
  expect_equal(length(res$lines), 2)
  expect_equal(res$lines[1], 5)
  expect_equal(res$lines[2], 5)
  
})

# The page template is everything except the content: page header/footer,
# titles, footnotes, etc.

# Page Template HTML Functions ---------------------------------------------


#' Create a page template with header, titles, footnotes, and footer.
#' @param rs The report spec
#' @return The page template object
#' @noRd
page_template_html <- function(rs) {
  
  pt <- structure(list(), class = c("page_template_html", "list"))
  
  pt$page_header <- get_page_header_html(rs)
  pt$title_hdr <- get_title_header_html(rs$title_hdr, rs$line_size, rs)
  pt$titles <- get_titles_html(rs$titles, rs$line_size, rs)
  pt$footnotes <- c()
  if (!is.null(rs$footnotes)) {
    if (!is.null(rs$footnotes[[1]])) {
      if (rs$footnotes[[1]]$valign == "bottom")
        pt$footnotes <- get_footnotes_html(rs$footnotes, rs$line_size, rs)
    }
    
  }
  pt$page_footer <- get_page_footer_html(rs)
  
  pt$lines <- sum(pt$page_header$lines, pt$page_footer$lines,
                  pt$title_hdr$lines, pt$titles$lines, pt$footnotes$lines)
  
  
  # Page by not here.  Messes up line counts.
  
  return(pt)
}

#' @import grDevices
#' @noRd
get_page_header_html <- function(rs) {
  
  ret <- ""
  cnt <- 0
  
  # hl <- rs$page_header_left
  # hr <- rs$page_header_right
  # maxh <- max(length(hl), length(hr))
  
  # # User controlled width of left column
  # lwdth <- rs$page_header_width
  # if (is.null(lwdth))
  #   lwdth <- rs$content_size[["width"]]/2
  # 
  # # Calculate right column width
  # rwdth <- rs$content_size[["width"]] - lwdth
  # lpct <- round(lwdth / rs$content_size[["width"]] * 100)
  # rpct <- round(rwdth / rs$content_size[["width"]] * 100)
  
  if ((!is.null(rs$header_image_left) & is.null(rs$page_header_left)) |
      (!is.null(rs$header_image_center) & is.null(rs$page_header_center)) |
      (!is.null(rs$header_image_right) & is.null(rs$page_header_right))) {
    
    stop("`page_header` must be used when using `header_image`.")
  }
  
  # Make sure the length is 3, NA will be imputed later.
  width <- c(rs$page_header_width, rep(NA, 3 - length(rs$page_header_width)))
  
  image_left <- FALSE
  image_center <- FALSE
  image_right <- FALSE
  
  if (!is.null(rs$header_image_left)) {
    image_left <- TRUE
    hl <- rs$header_image_left
  } else{
    hl <- rs$page_header_left
  }
  
  if (!is.null(rs$header_image_right)) {
    image_right <- TRUE
    hr <- rs$header_image_right
  } else {
    hr <- rs$page_header_right
  }
  
  if (!is.null(rs$header_image_center)) {
    image_center <- TRUE
    hc <- rs$header_image_center
    
    # Default center cell is not displayed, open it when picture exists
    if (width[2] == 0) {
      width[2] <- NA
    }
  } else {
    hc <- rs$page_header_center
  }
  
  hl_num <- ifelse(image_left, length(hl$image_path), length(hl))
  hc_num <- ifelse(image_center, length(hc$image_path), length(hc))
  hr_num <- ifelse(image_right, length(hr$image_path) , length(hr))
  
  maxh <- max(hl_num, hc_num, hr_num)
  
  # Calculate the widths
  total_width <- sum(width, na.rm = T)
  if (total_width > rs$content_size[["width"]]) {
    
    stop(sprintf("Total width of page footer %s %s cannot be greater than content width %s %s.",
                 total_width,
                 rs$units,
                 rs$content_size[["width"]],
                 rs$units))
    
  } else {
    na_num <- sum(is.na(width))
    imputed_width <- (rs$content_size[["width"]] - total_width) / na_num
    
    left_width <- ifelse(is.na(width[1]), imputed_width, width[1])
    center_width <- ifelse(is.na(width[2]), imputed_width, width[2])
    right_width <- ifelse(is.na(width[3]), imputed_width, width[3])
    
    left_pct <- 100 * left_width/rs$content_size[["width"]]
    center_pct <- 100 * center_width/rs$content_size[["width"]]
    right_pct <- 100 * right_width/rs$content_size[["width"]]
  }

  if (maxh > 0) {

    # ret <- paste0("<table ",
    #               "style=\"width:100%\">",
    #               "<colgroup>\n<col style=\"width:", left_pct, "%\">\n",
    #               "<col style=\"width:", right_pct,"%\">\n</colgroup>\n")
    
    ret <- paste0("<table ",
                  "style=\"width:100%\">",
                  "<colgroup>\n",
                  ifelse(left_pct > 0, paste0("<col style=\"width:", left_pct, "%\">\n"), ""),
                  ifelse(center_pct > 0, paste0("<col style=\"width:", center_pct, "%\">\n"), ""),
                  ifelse(right_pct > 0, paste0("<col style=\"width:", right_pct,"%\">\n"), ""),
                  "</colgroup>\n")

    pdf(NULL)
    par(family = get_font_family(rs$font), ps = rs$font_size)

    u <- rs$units
    if (rs$units == "inches") {
      u <- "in"
    }
    
    lcnt <- 0
    ccnt <- 0
    rcnt <- 0
    
    lheight <- 0
    cheight <- 0
    rheight <- 0
    
    max_height <- 0
    
    for (i in seq(1, maxh)) {
      ret <- paste0(ret, "<tr>")

      if (left_width > 0) {
        if (hl_num >= i) {
          
          if (image_left == FALSE) {
            # Split strings if they exceed width
            tmp <- split_string_html(hl[[i]], left_width, rs$units)
            
            ret <- paste0(ret, "<td style=\"text-align:left\">", encodeHTML(tmp$html),
                          "</td>\n")
            
            lcnt <- tmp$lines
          } else {
            # Get image code
            tmp_nm <- tempfile(tmpdir = tempdir(), fileext = ".jpg")
            file.copy(hl$image_path[i], tmp_nm, overwrite = TRUE)
            
            hl$width <- min(hl$width, left_width)
            img <- get_image_html(tmp_nm, rs$modified_path, hl, rs$units)
            ret <- paste0(ret, "<td style=\"text-align:left\">", img, "</td>")
            lheight <- hl$height
          }
        } else {
          ret <- paste0(ret, "<td style=\"text-align:left\">&nbsp</td>\n")
          lcnt <- 1
        }
      }
      
      if (center_width > 0) {
        if (hc_num >= i) {
          
          if (image_center == FALSE) {
            # Split strings if they exceed width
            tmp3 <- split_string_html(hc[[i]], center_width, rs$units)
            
            ret <- paste0(ret, "<td style=\"text-align:center\">", encodeHTML(tmp3$html),
                          "</td>\n")
            
            ccnt <- tmp3$lines
          } else {
            # Get image code
            tmp_nm <- tempfile(tmpdir = tempdir(), fileext = ".jpg")
            file.copy(hc$image_path[i], tmp_nm, overwrite = TRUE)
            
            hc$width <- min(hc$width, center_width)
            img <- get_image_html(tmp_nm, rs$modified_path, hc, rs$units)
            ret <- paste0(ret, "<td style=\"text-align:center\">", img, "</td>")
            cheight <- hc$height
          }
        } else {
          ret <- paste0(ret, "<td style=\"text-align:center\">&nbsp</td>\n")
          ccnt <- 1
        }
      }

      if (right_width) {
        if (hr_num >= i) {
          
          if (image_right == FALSE) {
            # Split strings if they exceed width
            tmp2 <- split_string_html(hr[[i]], right_width, rs$units)
            
            ret <- paste0(ret, "<td style=\"text-align:right\">", encodeHTML(tmp2$html), 
                          "</td></tr>\n")
            
            rcnt <- tmp2$lines 
          } else {
            # Get image code
            tmp_nm <- tempfile(tmpdir = tempdir(), fileext = ".jpg")
            file.copy(hr$image_path[i], tmp_nm, overwrite = TRUE)
            
            hr$width <- min(hr$width, right_width)
            img <- get_image_html(tmp_nm, rs$modified_path, hr, rs$units)
            ret <- paste0(ret, "<td style=\"text-align:right\">", img, "</td>")
            rheight <- hr$height
          }
        } else {
          ret <- paste0(ret, "<td style=\"text-align:right\">&nbsp</td></tr>\n")
          rcnt <- 1
        }
      }

      # if (lcnt > rcnt)
      #   cnt <- cnt + lcnt
      # else
      #   cnt <- cnt + rcnt
      cnt <- cnt + max(lcnt, ccnt, rcnt)
      max_height <- max_height + max(lheight, cheight, rheight)
    }
    
    ret <- paste0(ret, "</table>")

    dev.off()
    
    if (max_height > 0) {
      cnt <- max(cnt, round(max_height/ rs$line_height))
    }

    if (rs$page_header_blank_row == "below") {
      ret <- paste0(ret, "<br>")
      cnt <- cnt + 1
    }
  }

  
  res <- list(html = ret, lines = cnt)
  
  return(res)
}

#' @import grDevices
#' @noRd
get_page_footer_html <- function(rs) {
  
  ret <- ""
  cnt <- 1

  # fl <- rs$page_footer_left
  # fc <- rs$page_footer_center
  # fr <- rs$page_footer_right
  # 
  # maxf <- max(length(fl), length(fc), length(fr))
  
  if ((!is.null(rs$footer_image_left) & is.null(rs$page_footer_left)) |
      (!is.null(rs$footer_image_right) & is.null(rs$page_footer_right)) |
      (!is.null(rs$footer_image_center) & is.null(rs$page_footer_center))) {
    
    stop("`page_footer` must be used when using `footer_image`.")
  }
  
  image_left <- FALSE
  image_center <- FALSE
  image_right <- FALSE
  
  if (!is.null(rs$footer_image_left)) {
    fl <- rs$footer_image_left
    image_left <- TRUE
  } else {
    fl <- rs$page_footer_left
  }
  
  if (!is.null(rs$footer_image_center)) {
    fc <- rs$footer_image_center
    image_center <- TRUE
  } else {
    fc <- rs$page_footer_center
  }
  
  if (!is.null(rs$footer_image_right)) {
    fr <- rs$footer_image_right
    image_right <- TRUE
  } else {
    fr <- rs$page_footer_right
  }
  
  fl_num <- ifelse(image_left, length(fl$image_path), length(fl))
  fc_num <- ifelse(image_center, length(fc$image_path), length(fc))
  fr_num <- ifelse(image_right, length(fr$image_path) , length(fr))
  
  maxf <- max(fl_num, fc_num, fr_num)

  if (maxf > 0) {
    
    width <- rs$page_footer_width
    
    # Make sure the length is 3, NA will be imputed later.
    width <- c(width, rep(NA, 3 - length(width)))
    
    total_width <- sum(width, na.rm = T)
    if (total_width > rs$content_size[["width"]]) {
      
      stop(sprintf("Total width of page footer %s %s cannot be greater than content width %s %s.",
                   total_width,
                   rs$units,
                   rs$content_size[["width"]],
                   rs$units))
      
    } else {
      na_num <- sum(is.na(width))
      imputed_width <- (rs$content_size[["width"]] - total_width) / na_num
      
      left_width <- ifelse(is.na(width[1]), imputed_width, width[1])
      center_width <- ifelse(is.na(width[2]), imputed_width, width[2])
      right_width <- ifelse(is.na(width[3]), imputed_width, width[3])
      
      left_pct <- 100 * left_width/rs$content_size[["width"]]
      center_pct <- 100 * center_width/rs$content_size[["width"]]
      right_pct <- 100 * right_width/rs$content_size[["width"]]
    }

    # ret <- paste0("<br>\n<table ",
    #               "style=\"width:100%\">\n",
    #               paste0("<colgroup>\n<col style=\"width:", left_pct,"%\">\n"),
    #               paste0("<col style=\"width:", center_pct,"%\">\n"),
    #               paste0("<col style=\"width:", right_pct,"%\">\n</colgroup>\n"))
    
    ret <- paste0("<br>\n<table ",
                  "style=\"width:100%\">\n",
                  "<colgroup>\n",
                  ifelse(left_pct > 0, paste0("<col style=\"width:", left_pct,"%\">\n"), ""),
                  ifelse(center_pct > 0, paste0("<col style=\"width:", center_pct,"%\">\n"), ""),
                  ifelse(right_pct > 0, paste0("<col style=\"width:", right_pct,"%\">\n"), ""),
                  "</colgroup>\n")

    pdf(NULL)
    par(family = get_font_family(rs$font), ps = rs$font_size)

    u <- rs$units
    if (u == "inches") {
      u <- "in"
    }
    
    lcnt <- 0
    ccnt <- 0
    rcnt <- 0
    
    lheight <- 0
    cheight <- 0
    rheight <- 0
    
    max_height <- 0
    
    for (i in seq(1, maxf)) {

      ret <- paste0(ret, "<tr>")

      if (left_width > 0) {
        if (fl_num >= i) {
          
          if (image_left == FALSE) {
            # Split strings if they exceed width
            tmp1 <- split_string_html(fl[[i]], left_width, rs$units)
            
            ret <- paste0(ret, "<td style=\"text-align:left\">", encodeHTML(tmp1$html),
                          "</td>")
            lcnt <- tmp1$lines
          } else {
            # Get image code
            tmp_nm <- tempfile(tmpdir = tempdir(), fileext = ".jpg")
            file.copy(fl$image_path[i], tmp_nm, overwrite = TRUE)
            
            fl$width <- min(fl$width, left_width)
            img <- get_image_html(tmp_nm, rs$modified_path, fl, rs$units)
            ret <- paste0(ret, "<td style=\"text-align:left\">", img, "</td>")
            lheight <- fl$height
          }
        } else {
          ret <- paste0(ret, "<td style=\"text-align:left\">&nbsp;</td>")
          lcnt <- 1
        }
      }

      if (center_width > 0) {
        if (fc_num >= i) {
          
          if (image_center == FALSE) {
            # Split strings if they exceed width
            tmp2 <- split_string_html(fc[[i]], center_width, rs$units)
            
            ret <- paste0(ret, "<td style=\"text-align:center\">", encodeHTML(tmp2$html),
                          "</td>")
            ccnt <- tmp2$lines
          } else {
            # Get image code
            tmp_nm <- tempfile(tmpdir = tempdir(), fileext = ".jpg")
            file.copy(fc$image_path[i], tmp_nm, overwrite = TRUE)
            
            fc$width <- min(fc$width, center_width)
            img <- get_image_html(tmp_nm, rs$modified_path, fc, rs$units)
            ret <- paste0(ret, "<td style=\"text-align:center\">", img, "</td>")
            cheight <- fc$height
          }
        } else {
          ret <- paste0(ret, "<td style=\"text-align:center\">&nbsp;</td>")
          ccnt <- 1
        }
      }

      if (right_width > 0) {
        if (fr_num >= i) {
          
          if (image_right == FALSE) {
            tmp3 <- split_string_html(fr[[i]], right_width, rs$units)
            
            ret <- paste0(ret, "<td style=\"text-align:right\">", encodeHTML(tmp3$html),
                          "</td>")
            
            rcnt <- tmp3$lines
          } else {
            # Get image code
            tmp_nm <- tempfile(tmpdir = tempdir(), fileext = ".jpg")
            file.copy(fr$image_path[i], tmp_nm, overwrite = TRUE)
            
            fr$width <- min(fr$width, right_width)
            img <- get_image_html(tmp_nm, rs$modified_path, fr, rs$units)
            ret <- paste0(ret, "<td style=\"text-align:right\">", img, "</td>")
            rheight <- fr$height
          }
        } else {
          ret <- paste0(ret, "<td style=\"text-align:right\">&nbsp;</td>")
          rcnt <- 1
        }
      }

      cnt <- cnt + max(lcnt, ccnt, rcnt)
      max_height <- max_height + max(lheight, cheight, rheight)

    }
    dev.off()

    if (max_height > 0) {
      cnt <- max(cnt, round(max_height/ rs$line_height))
    }
    
    ret <- paste0(ret, "</tr>\n")
  }

  ret <- paste0(ret, "</table>\n")
  
  res <- list(html = paste0(ret, collapse = ""),
              lines = cnt)
  
  return(res)
}

#' @import grDevices
#' @noRd
get_titles_html <- function(ttllst, content_width, rs, talgn = "center") {
  
  ret <- c()
  cnt <- 0
  border_flag <- FALSE

  # ta <- "align=\"left\" "
  # if (talgn == "right")
  #   ta <- "align=\"right\" "
  # else if (talgn %in% c("center", "centre"))
  #   ta <- "align=\"center\" "
  
  sty <- paste0(get_style_html(rs, "title_font_color"),
                get_style_html(rs, "title_background"),
                get_style_html(rs, "title_font_bold"),
                get_style_html(rs, "title_font_size"))
  
  u <- rs$units
  if (rs$units == "inches")
    u <- "in"
  
  if (length(ttllst) > 0) {
    
    for (ttls in ttllst) {
      
      cols <- ttls$columns
      
      if (ttls$width == "page")
        width <- rs$content_size[["width"]]
      else if (ttls$width == "content")
        width <- content_width
      else if (is.numeric(ttls$width))
        width <- ttls$width

      w <- round(width, 3)
      
      cwidth <- width / cols
      cw <- w / cols
      
      if (ttls$align %in% c("centre", "center"))
        algn <- "text-align: center;"
      else if (ttls$align == "right")
        algn <- "text-align: right;"
      else
        algn <- "text-align: left;"
      
      alcnt <- 0
      blcnt <- 0
      
      # Open device context
      pdf(NULL)
      if (!is.null(ttls$font_size)) {
        ttlfs <- ttls$font_size
      } else {
        ttlfs <- rs$font_size
      }
      par(family = get_font_family(rs$font), ps = ttlfs)
      
      ret[length(ret) + 1] <- paste0("<table ",
                                     "style=\"width:", w, u, ";", 
                                      sty,
                                     "\">\n")
      
      al <- ""
      if (any(ttls$blank_row %in% c("above", "both"))) {
        
        alcnt <- 1
        
        tb <- get_cell_borders_html(1, 1, length(ttls$titles) + alcnt, 
                                    1, ttls$borders, 
                                    border_color = get_style(rs, "border_color"))
        
        if (tb == "")
          al <- paste0("<tr><td colspan=\"", cols, "\">&nbsp;</td></tr>\n")
        else 
          al <- paste0("<tr><td style=\"", tb, "\" colspan=\"", cols, 
                       "\">&nbsp;</td></tr>\n")
        
        # Can append now, since it is first
        ret[length(ret) + 1] <- al
        
        cnt <- cnt + 1
      }
      
      bl <- ""
      if (any(ttls$blank_row %in% c("below", "both"))) {
        blcnt <- 1
        
        tb <- get_cell_borders_html(length(ttls$titles) + alcnt + blcnt, 1,
                                    length(ttls$titles) + alcnt + blcnt,
                                    1, ttls$borders,
                                    border_color = get_style(rs, "border_color"))
        
        if (tb == "")
          bl <- paste0("<tr><td colspan=\"", cols, "\">&nbsp;</td></tr>\n")
        else 
          bl <- paste0("<tr><td style=\"", tb, "\" colspan=\"", cols, 
                       "\">&nbsp;</td></tr>\n")
        
        # Wait to append until after title rows
        
        cnt <- cnt + 1
      }
        
    
      i <- 1
      while (i <= length(ttls$titles)) {
        
        
        # Calculate current row
        rwnum <- ceiling(i / cols)
        
        mxlns <- 0
        rw <- "<tr>"
        
        for (j in seq_len(cols)) {
          
          b <- get_cell_borders_html(rwnum + alcnt, j,
                                length(ttls$titles) + alcnt + blcnt,
                                cols, ttls$borders,
                                border_color = get_style(rs, "border_color"))
          
          # Not all cells have titles
          if (i > length(ttls$titles))
            vl <- ""
          else 
            vl <- ttls$titles[[i]]
          
          # Deal with column alignments
          if (cols == 1) {
            calgn <- algn 
          } else if (cols == 2) {
            if (j == 1)
              calgn <- "text-align: left;"
            else 
              calgn <- "text-align: right;"
          } else if (cols == 3) {
            if (j == 1)
              calgn <- "text-align: left;"
            else if (j == 2)
              calgn <- "text-align: center;"
            else if (j == 3) 
              calgn <- "text-align: right;"
          }
          
          cws <- paste0("width:", cw, u, ";")
          valgn <- "vertical-align:text-top;"
          
          # Split title strings if they exceed width
          tmp <- split_string_html(vl, cwidth, rs$units)
          
          # Track max lines for counting
          if (tmp$lines > mxlns)
            mxlns <- tmp$lines
          
          if (ttls$bold)
            tstr <- paste0("<b>", encodeHTML(tmp$html), "</b>")
          else 
            tstr <- encodeHTML(tmp$html)
          
          fz <- ""
          if (!is.null(ttls$font_size)){
            
            fz <- paste0("font-size:", ttls$font_size, "pt;") 
          }
          
          # Concatenate title string
          rw <- paste0(rw, paste0("<td style=\"", cws, b, fz, calgn, valgn, "\">",  
                                  tstr, "</td>\n"))
          
          
          i <- i + 1
        }
        
        ret <- append(ret, paste0(rw, "</tr>\n"))

        # Keep track of lines
        cnt <- cnt + mxlns
        
        # A flag to indicate that this block has bottom borders.  
        # Used to eliminate border duplication on subsequent blocks.
        if ("bottom" %in% get_outer_borders(ttls$borders))
          border_flag <- TRUE
        

      }
      
      # Append blank row at bottom
      if (bl != "")
        ret <- append(ret, bl)
      
      ret[length(ret) + 1] <- "</table>"
      dev.off()
      

    }
    
  }
  

  
  res <- list(html = paste0(ret, collapse = ""), 
              lines = cnt,
              border_flag = border_flag)
  
  return(res)
}


#' @import grDevices
#' @noRd
get_titles_html_back <- function(ttllst, content_width, rs, talgn = "center") {
  
  ret <- c()
  cnt <- 0
  border_flag <- FALSE
  
  # ta <- "align=\"left\" "
  # if (talgn == "right")
  #   ta <- "align=\"right\" "
  # else if (talgn %in% c("center", "centre"))
  #   ta <- "align=\"center\" "
  
  sty <- paste0(get_style_html(rs, "title_font_color"),
                get_style_html(rs, "title_background"),
                get_style_html(rs, "title_font_bold"),
                get_style_html(rs, "title_font_size"))
  
  u <- rs$units
  if (rs$units == "inches")
    u <- "in"
  
  if (length(ttllst) > 0) {
    
    for (ttls in ttllst) {
      
      
      if (ttls$width == "page")
        width <- rs$content_size[["width"]]
      else if (ttls$width == "content")
        width <- content_width
      else if (is.numeric(ttls$width))
        width <- ttls$width
      
      w <- round(width, 3)
      
      if (ttls$align %in% c("centre", "center"))
        algn <- "text-align: center;"
      else if (ttls$align == "right")
        algn <- "text-align: right;"
      else
        algn <- "text-align: left;"
      
      alcnt <- 0
      blcnt <- 0
      
      # Open device context
      pdf(NULL)
      par(family = get_font_family(rs$font), ps = rs$font_size)
      
      ret[length(ret) + 1] <- paste0("<table ",
                                     "style=\"width:", w, u, ";", 
                                     algn, sty,
                                     "\">\n")
      
      for (i in seq_along(ttls$titles)) {
        
        
        
        al <- ""
        if (i == 1) {
          if (any(ttls$blank_row %in% c("above", "both"))) {
            
            alcnt <- 1
            
            tb <- get_cell_borders_html(i, 1, length(ttls$titles) + alcnt, 
                                        1, ttls$borders, 
                                        border_color = get_style(rs, "border_color"))
            
            if (tb == "")
              al <- "<tr><td>&nbsp;</td></tr>\n"
            else 
              al <- paste0("<tr><td style=\"", tb, "\">&nbsp;</td></tr>\n")
            
            cnt <- cnt + 1
          }
        }
        
        bl <- ""
        if (i == length(ttls$titles)) {
          if (any(ttls$blank_row %in% c("below", "both"))) {
            blcnt <- 1
            
            tb <- get_cell_borders_html(i + alcnt + blcnt, 1,
                                        length(ttls$titles) + alcnt + blcnt,
                                        1, ttls$borders,
                                        border_color = get_style(rs, "border_color"))
            
            if (tb == "")
              bl <- "<tr><td>&nbsp;</td></tr>\n"
            else 
              bl <- paste0("<tr><td style=\"", tb, "\">&nbsp;</td></tr>\n")
            
            cnt <- cnt + 1
          }
          
        }
        
        b <- get_cell_borders_html(i + alcnt, 1,
                                   length(ttls$titles) + alcnt + blcnt,
                                   1, ttls$borders,
                                   border_color = get_style(rs, "border_color"))
        
        # Split title strings if they exceed width
        tmp <- split_string_html(ttls$titles[[i]], width, rs$units)
        
        if (ttls$bold)
          tstr <- paste0("<b>", encodeHTML(tmp$html), "</b>")
        else 
          tstr <- encodeHTML(tmp$html)
        
        fz <- ""
        if (!is.null(ttls$font_size)){
          
          fz <- paste0("font-size:", ttls$font_size, "pt;") 
        }
        
        
        # Concatenate title string
        if (al != "")
          ret <- append(ret, al)
        
        if (b == "" & fz == "") {
          ret <- append(ret, paste0("<tr><td>", tstr, 
                                    "</td></tr>\n"))
        } else {
          
          ret <- append(ret, paste0("<tr><td style=\"", b, fz, "\">", 
                                    tstr, 
                                    "</td></tr>\n"))
        }
        
        if (bl != "")
          ret <- append(ret, bl)
        
        cnt <- cnt + tmp$lines
        
        # A flag to indicate that this block has bottom borders.  
        # Used to eliminate border duplication on subsequent blocks.
        if ("bottom" %in% get_outer_borders(ttls$borders))
          border_flag <- TRUE
      }
      
      ret[length(ret) + 1] <- "</table>"
      dev.off()
      
      
    }
    
  }
  
  
  
  res <- list(html = paste0(ret, collapse = ""), 
              lines = cnt,
              border_flag = border_flag)
  
  return(res)
}


#' @import grDevices
#' @noRd
get_footnotes_html <- function(ftnlst, content_width, rs, talgn = "center", 
                               ex_brdr = FALSE) {
  
  ret <- c()
  cnt <- 0
  exclude_top <- NULL
  if (ex_brdr)
    exclude_top <- "top"

  u <- rs$units
  if (rs$units == "inches")
    u <- "in"
  
  sty <- paste0(get_style_html(rs, "footnote_font_color"),
                get_style_html(rs, "footnote_background"),
                get_style_html(rs, "footnote_font_bold"))
  
  
  if (length(ftnlst) > 0) {

    for (ftnts in ftnlst) {
      
      cols <- ftnts$columns

      if (ftnts$width == "page")
        width <- rs$content_size[["width"]]
      else if (ftnts$width == "content")
        width <- content_width
      else if (is.numeric(ftnts$width))
        width <- ftnts$width

      w <- round(width, 3)
      
      cwidth <- width / cols
      cw <- w / cols


      if (ftnts$align %in% c("centre", "center"))
        algn <- "text-align: center;"
      else if (ftnts$align == "right")
        algn <- "text-align: right;"
      else
        algn <- "text-align: left;"

      alcnt <- 0
      blcnt <- 0


      pdf(NULL)
      if (!is.null(ftnts$font_size)) {
        ftntfs <- ftnts$font_size
      } else {
        ftntfs <- rs$font_size
      }
      par(family = get_font_family(rs$font), ps = ftntfs)
      ret[length(ret) + 1] <- paste0("<table ",
                                     "style=\"width:", w, u, ";",
                                     algn, sty,
                                     "\">\n")
      
      al <- ""
      if (any(ftnts$blank_row %in% c("above", "both"))) {
        
        alcnt <- 1
        
        tb <- get_cell_borders_html(1, 1, length(ftnts$footnotes) + alcnt,
                                    1, ftnts$borders, exclude = exclude_top,
                                    border_color = get_style(rs, "border_color"))

        
        if (tb == "")
          al <- paste0("<tr><td colspan=\"", cols, "\">&nbsp;</td></tr>\n")
        else 
          al <- paste0("<tr><td style=\"", tb, "\" colspan=\"", cols, 
                       "\">&nbsp;</td></tr>\n")
        
        # Can append now, since it is first
        ret[length(ret) + 1] <- al
        
        cnt <- cnt + 1
        
      }

      
      bl <- ""
      if (any(ftnts$blank_row %in% c("below", "both"))) {
        blcnt <- 1
        
        tb <- get_cell_borders_html(length(ftnts$footnotes) + alcnt + blcnt, 1,
                                    length(ftnts$footnotes) + alcnt + blcnt,
                                    1, ftnts$borders,
                                    border_color = get_style(rs, "border_color"))
        
        if (tb == "")
          bl <- paste0("<tr><td colspan=\"", cols, "\">&nbsp;</td></tr>\n")
        else 
          bl <- paste0("<tr><td style=\"", tb, "\" colspan=\"", cols, 
                       "\">&nbsp;</td></tr>\n")
        
        cnt <- cnt + 1
      }

      i <- 1
      while (i <= length(ftnts$footnotes)) {
        
        
        # Calculate current row
        rwnum <- ceiling(i / cols)
        
        mxlns <- 0
        rw <- "<tr>"

        for (j in seq_len(cols)) {
  

          b <- get_cell_borders_html(rwnum + alcnt, j,
                                length(ftnts$footnotes) + alcnt + blcnt,
                                cols, ftnts$borders, exclude = exclude_top,
                                border_color = get_style(rs, "border_color"))
  
          # Not all cells have titles
          if (i > length(ftnts$footnotes))
            vl <- ""
          else 
            vl <- ftnts$footnotes[[i]]
          
          # Deal with column alignments
          if (cols == 1) {
            calgn <- algn 
          } else if (cols == 2) {
            if (j == 1)
              calgn <- "text-align: left;"
            else 
              calgn <- "text-align: right;"
          } else if (cols == 3) {
            if (j == 1)
              calgn <- "text-align: left;"
            else if (j == 2)
              calgn <- "text-align: center;"
            else if (j == 3) 
              calgn <- "text-align: right;"
          }
          
          cws <- paste0("width:", cw, u, ";")
          valgn <- "vertical-align:text-top;"
  
          # Split footnote strings if they exceed width
          tmp <- split_string_html(vl, cwidth, rs$units)
  
          # Track max lines for counting
          if (tmp$lines > mxlns)
            mxlns <- tmp$lines
          
          
          if (ftnts$italics)
            txt <- paste0("<i>", encodeHTML(tmp$html), "</i>")
          else
            txt <- encodeHTML(tmp$html)
  
          # if (b == "")
          #   ret <- append(ret, paste0("<tr><td>", txt, 
          #                             "</td></tr>\n"))
          # else {
          #   ret <- append(ret, paste0("<tr><td style=\"", b, "\">", 
          #                             txt, 
          #                             "</td></tr>\n"))
          #}
          
          fz <- ""
          if (!is.null(ftnts$font_size)){
            
            fz <- paste0("font-size:", ftnts$font_size, "pt;") 
          }
          
          # Concat tags and footnote content
          rw <- paste0(rw, paste0("<td style=\"", cws, b, fz, calgn, valgn, "\">",  
                                  txt, "</td>\n"))
  
  
          i <- i + 1

        }
        
        ret <- append(ret, paste0(rw, "</tr>\n"))
        
        # Keep track of lines
        cnt <- cnt + mxlns
        
        
      }
      
      
      if (bl != "")
        ret <- append(ret, bl)
      
      ret[length(ret) + 1] <- "</table>"
      dev.off()


    }

  }

  
  res <- list(html = paste0(ret, collapse = ""),
              lines = cnt)
  
  return(res)
}


#' @import grDevices
#' @noRd
get_footnotes_html_back <- function(ftnlst, content_width, rs, talgn = "center", 
                               ex_brdr = FALSE) {
  
  ret <- c()
  cnt <- 0
  exclude_top <- NULL
  if (ex_brdr)
    exclude_top <- "top"
  
  u <- rs$units
  if (rs$units == "inches")
    u <- "in"
  
  sty <- paste0(get_style_html(rs, "footnote_font_color"),
                get_style_html(rs, "footnote_background"),
                get_style_html(rs, "footnote_font_bold"))
  
  
  if (length(ftnlst) > 0) {
    
    for (ftnts in ftnlst) {
      
      if (ftnts$width == "page")
        width <- rs$content_size[["width"]]
      else if (ftnts$width == "content")
        width <- content_width
      else if (is.numeric(ftnts$width))
        width <- ftnts$width
      
      w <- round(width, 3)
      
      
      if (ftnts$align %in% c("centre", "center"))
        algn <- "text-align: center;"
      else if (ftnts$align == "right")
        algn <- "text-align: right;"
      else
        algn <- "text-align: left;"
      
      alcnt <- 0
      blcnt <- 0
      
      
      pdf(NULL)
      par(family = get_font_family(rs$font), ps = rs$font_size)
      ret[length(ret) + 1] <- paste0("<table ",
                                     "style=\"width:", w, u, ";",
                                     algn, sty,
                                     "\">\n")
      
      for (i in seq_along(ftnts$footnotes)) {
        
        
        al <- ""
        if (i == 1) {
          if (any(ftnts$blank_row %in% c("above", "both"))) {
            
            alcnt <- 1
            
            tb <- get_cell_borders_html(i, 1, length(ftnts$footnotes) + alcnt,
                                        1, ftnts$borders, exclude = exclude_top,
                                        border_color = get_style(rs, "border_color"))
            
            if (tb == "")
              al <- "<tr><td>&nbsp;</td></tr>\n"
            else 
              al <- paste0("<tr><td style=\"", tb, "\">&nbsp;</td></tr>\n")
            
            cnt <- cnt + 1
            
          }
        }
        
        bl <- ""
        if (i == length(ftnts$footnotes)) {
          if (any(ftnts$blank_row %in% c("below", "both"))) {
            blcnt <- 1
            
            tb <- get_cell_borders_html(i + alcnt + blcnt, 1,
                                        length(ftnts$footnotes) + alcnt + blcnt,
                                        1, ftnts$borders,
                                        border_color = get_style(rs, "border_color"))
            
            if (tb == "")
              bl <- "<tr><td>&nbsp;</td></tr>\n"
            else 
              bl <- paste0("<tr><td style=\"", tb, "\">&nbsp;</td></tr>\n")
            
            cnt <- cnt + 1
          }
          
        }
        
        b <- get_cell_borders_html(i + alcnt, 1,
                                   length(ftnts$footnotes) + alcnt + blcnt,
                                   1, ftnts$borders, exclude = exclude_top,
                                   border_color = get_style(rs, "border_color"))
        
        
        
        # Split footnote strings if they exceed width
        tmp <- split_string_html(ftnts$footnotes[[i]], width, rs$units)
        
        if (al != "")
          ret <- append(ret, al)
        
        if (ftnts$italics)
          txt <- paste0("<i>", encodeHTML(tmp$html), "</i>")
        else
          txt <- encodeHTML(tmp$html)
        
        if (b == "")
          ret <- append(ret, paste0("<tr><td>", txt, 
                                    "</td></tr>\n"))
        else {
          ret <- append(ret, paste0("<tr><td style=\"", b, "\">", 
                                    txt, 
                                    "</td></tr>\n"))
        }
        
        
        if (bl != "")
          ret <- append(ret, bl)
        
        cnt <- cnt + tmp$lines
      }
      ret[length(ret) + 1] <- "</table>"
      dev.off()
      
      
    }
    
  }
  
  
  res <- list(html = paste0(ret, collapse = ""),
              lines = cnt)
  
  return(res)
}

#' @import grDevices
#' @noRd
get_title_header_html <- function(thdrlst, content_width, rs, talgn = "center") {
  
  ret <- c()
  cnt <- 0
  border_flag <- FALSE
  
  
  u <- rs$units
  if (rs$units == "inches")
    u <- "in"

  if (length(thdrlst) > 0) {

    for (ttlhdr in thdrlst) {

      if (ttlhdr$width == "page")
        width <- rs$content_size[["width"]]
      else if (ttlhdr$width == "content")
        width <- content_width
      else if (is.numeric(ttlhdr$width))
        width <- ttlhdr$width

      w <- round(width, 3)

      mx <- max(length(ttlhdr$titles), length(ttlhdr$right))

      alcnt <- 0
      blcnt <- 0

      pdf(NULL)
      par(family = get_font_family(rs$font), ps = rs$font_size)
      
      ret[length(ret) + 1] <- paste0("<table ",
                                     "style=\"width:", w, u, ";", 
                                     "\">\n", 
                                     "<colgroup><col style=\"width:70%;\">\n",
                                     "<col style=\"width:30%;\"></colgroup>\n")
                                     

      for(i in seq_len(mx)) {

        al <- ""
        if (i == 1) {
          if (any(ttlhdr$blank_row %in% c("above", "both"))) {

            alcnt <- 1

            tb1 <- get_cell_borders_html(i, 1, mx + alcnt,
                                   2, ttlhdr$borders,
                                   border_color = get_style(rs, "border_color"))
            tb2 <- get_cell_borders_html(i, 2, mx + alcnt,
                                        2, ttlhdr$borders,
                                        border_color = get_style(rs, "border_color"))

            al <- paste0("<tr><td style=\"text-align:left;", tb1, "\">&nbsp;</td>", 
                         "<td style=\"text-align:right;", tb2, 
                         "\">&nbsp;</td></tr>\n")
            cnt <- cnt + 1

          }
        }

        bl <- ""
        if (i == mx) {
          if (any(ttlhdr$blank_row %in% c("below", "both"))) {
            blcnt <- 1

            tb1 <- get_cell_borders_html(i + alcnt + blcnt, 1,
                                   mx + alcnt + blcnt,
                                   2, ttlhdr$borders,
                                   border_color = get_style(rs, "border_color"))
            tb2 <- get_cell_borders_html(i + alcnt + blcnt, 2,
                                        mx + alcnt + blcnt,
                                        2, ttlhdr$borders,
                                        border_color = get_style(rs, "border_color"))
            
            bl <- paste0("<tr><td style=\"text-align:left;", tb1, "\">&nbsp;</td>", 
                         "<td style=\"text-align:right;", tb2, 
                         "\">&nbsp;</td></tr>\n")
            cnt <- cnt + 1
          }

        }

        if (length(ttlhdr$titles) >= i) {
          # Split strings if they exceed width
          tmp1 <- split_string_html(ttlhdr$titles[[i]], width * .7, rs$units)

          ttl <-  tmp1$html 
          tcnt <- tmp1$lines
        } else {
          ttl <- ""
          tcnt <- 1
        }

        if (length(ttlhdr$right) >= i) {
          tmp2 <- split_string_html(ttlhdr$right[[i]],
                                   width * .3, rs$units)
          # hdr <- get_page_numbers_html(tmp2$html, FALSE)
          
          hdr <- tmp2$html
          hcnt <- tmp2$lines
        } else {
          hdr <- ""
          hcnt <- 1
        }

        b1 <- get_cell_borders_html(i + alcnt, 1, mx + alcnt + blcnt, 
                                    2, ttlhdr$borders,
                                    border_color = get_style(rs, "border_color"))
        b2 <- get_cell_borders_html(i + alcnt, 2, mx+ alcnt + blcnt, 
                                    2, ttlhdr$borders,
                                    border_color = get_style(rs, "border_color"))


        if (al != "")
          ret <- append(ret, al)
        
        ret <- append(ret, paste0("<tr><td style=\"text-align:left;", b1, "\">",
                                  encodeHTML(ttl), 
                                  "</td><td style=\"text-align:right;", b2, "\">", 
                                  encodeHTML(hdr), 
                                  "</td></tr>\n"))
        
        if (bl != "")
          ret <- append(ret, bl)

        if (tcnt > hcnt)
          cnt <- cnt + tcnt
        else
          cnt <- cnt + hcnt
        
        if ("bottom" %in% get_outer_borders(ttlhdr$borders))
          border_flag <- TRUE
      }

      ret[length(ret) + 1] <- "</table>\n"
      dev.off()
      


    }

  }
  

  
  res <- list(html = paste0(ret, collapse = ""),
              lines = cnt,
              border_flag = border_flag)
  
  return(res)
}


#' Get page by text strings suitable for printing
#' @import stringi
#' @return A vector of strings
#' @noRd
get_page_by_html <- function(pgby, width, value, rs, talgn, ex_brdr = FALSE, pgby_cnt = NULL) {
  
  if (is.null(width)) {
    stop("width cannot be null.") 
    
  }
  
  if (is.null(value))
    value <- get_pgby_value(value, pgby_cnt)
  
  ll <- width
  ret <- c()
  cnt <- 0
  border_flag <- FALSE
  exclude_top <- NULL
  if (ex_brdr)
    exclude_top <- "top"


  if (!is.null(pgby)) {

    if (!any(class(pgby) == "page_by"))
      stop("pgby parameter value is not a page_by.")

    w <- paste0("width:", round(width, 3), units_html(rs$units), ";")

    if (pgby$align %in% c("centre", "center"))
      algn <- "text-align: center;"
    else if (pgby$align == "right")
      algn <- "text-align: right;"
    else
      algn <- "text-align: left;"

    ret[length(ret) + 1] <- paste0("<table style=\"", algn, w, "\">\n")
    
    trows <- 1
    brow <- 1
    if (pgby$blank_row %in% c("above", "both")) {
      trows <- trows + 1
      brow <- 2
    }
    if (pgby$blank_row %in% c("below", "both"))
      trows <- trows + 1

    if (pgby$blank_row %in% c("above", "both")) {

      tb <- get_cell_borders_html(1, 1, trows, 1, pgby$borders, 
                                  exclude = exclude_top, 
                                  border_color = get_style(rs, "border_color"))

      ret[length(ret) + 1] <- paste0("<tr><td style=\"", tb, 
                                     "\">&nbsp;</td></tr>\n")
      cnt <- cnt + 1
    }

    tb <- get_cell_borders_html(brow, 1 , trows, 1, pgby$borders, 
                                exclude = exclude_top, 
                                border_color = get_style(rs, "border_color"))
    
    
    pdf(NULL)
    par(family = get_font_family(rs$font), ps = rs$font_size)
    
    # # Account for multiple pgby lines
    # tmp <- split_string_html(value, width, rs$units)
    # 
    # dev.off()
    # 
    # vl <- tmp$html
    # cnt <- cnt + tmp$lines
    # 
    # # Construct HTML for page by
    # ret[length(ret) + 1] <- paste0("<tr><td style=\"", tb, "\">",
    #                                pgby$label, encodeHTML(vl), "</td></tr>\n")
    # #cnt <- cnt + 1
    
    # Add bold
    if (pgby$bold %in% c(TRUE, FALSE)) {
      
      if (substr(pgby$label, nchar(pgby$label), nchar(pgby$label)) != " "){
        sep <- " "
      } else {
        sep <- ""
      }
      
      tmp <- split_string_html(paste0(pgby$label, sep, value), width, rs$units)
      vl <- tmp$html
      cnt <- cnt + tmp$lines
      
      if (pgby$bold ) {
        page_by_text <- paste0("<b>", encodeHTML(vl), "</b>")
      } else {
        page_by_text <- encodeHTML(vl)
      }
      
    } else if (pgby$bold %in% c("value", "label")) {
      
      # Split label
      label_split <- split_string_html(pgby$label, width, rs$units)
      cnt <- cnt + label_split$lines
      
      # Use remain width to split value
      remain_width <- width - label_split$widths[length(label_split$widths)]
      value_split <- split_string_html(value, remain_width, rs$units)
      
      if (value_split$widths[1] > remain_width) {
        
        # If first width is bigger than remaining width, it means value starts a new line
        value_split <- split_string_html(value, width, rs$units)
        cnt <- cnt + value_split$lines
        value_split_txt <- value_split$html
        
      } else {
        
        # Otherwise, split with full width if there is a second line
        splt <- strsplit(value_split$html, split = "\n", fixed = TRUE)
        
        if (length(splt[[1]]) > 1) {
          remain_value <- trimws(sub(splt[[1]][1], "", value), which = "left")
          remain_value_split <- split_string_html(remain_value, width, rs$units)
          cnt <- cnt + remain_value_split$lines
          value_split_txt <- paste0(splt[[1]][1], "\n", remain_value_split$html)
        } else {
          value_split_txt <- value_split$html
        }
        
      }
      
      if (pgby$bold == "label") {
        page_by_text <- paste0("<b>", encodeHTML(paste0(label_split$html, " ")), "</b>", encodeHTML(value_split_txt))
      } else {
        page_by_text <- paste0(encodeHTML(paste0(label_split$html, " ")), "<b>", encodeHTML(value_split_txt), "</b>")
      }
    }
    
    dev.off()
    
    # Construct HTML for page by
    ret[length(ret) + 1] <- paste0("<tr><td style=\"", tb, "\">",
                                   page_by_text, "</td></tr>\n")
  

    if (pgby$blank_row %in% c("below", "both")) {

      tb <- get_cell_borders_html(trows, 1, trows, 1, pgby$borders, 
                                  border_color = get_style(rs, "border_color"))

      ret[length(ret) + 1] <- paste0("<tr><td style=\"", tb, 
                                     "\">&nbsp;</td></tr>\n")
      cnt <- cnt + 1
      

    }

    ret[length(ret) + 1] <- "</table>"
    
    if ("bottom" %in% get_outer_borders(pgby$borders))
      border_flag <- TRUE

  }
  
  
  res <- list(html = paste0(ret, collapse = ""), 
              lines = cnt,
              border_flag = border_flag)
  
  return(res)
}

# Utilities ---------------------------------------------------------------

#' @description Return border code for a particular cell.  Idea is 
#' you pass in the size of the table and the particular cell position,
#' and this function will return the correct border codes.  System works 
#' great.
#' @noRd
get_cell_borders_html <- function(row, col, nrow, ncol, brdrs, 
                                  flag = "", exclude = NULL, 
                                  border_color = "",
                                  stub_flag = FALSE,
                                  cell_border = NULL) {
  
  t <- ""
  b <- ""
  l <- ""
  r <- ""
  
  if (border_color == "")
    border_color <- "black"
  
  
  if ("all" %in% brdrs) {
    t <- paste0("border-top:thin solid ", border_color, ";")
    b <- paste0("border-bottom:thin solid ", border_color, ";")
    l <- paste0("border-left:thin solid ", border_color, ";")
    r <- paste0("border-right:thin solid ", border_color, ";")
    
    if (row > 1)
      t <- ""

    if (col < ncol & stub_flag == FALSE)
      r <- ""
    
    
  } else {
    
    if ("inside" %in% brdrs) {
      
      t <- ""
      b <- paste0("border-bottom:thin solid ", border_color, ";")
      l <- paste0("border-left:thin solid ", border_color, ";")
      r <- ""
      
      if (col == 1) 
        l <- ""
      
      if (col == ncol)
        r <- ""
      
      if (row == nrow)
        b <- ""
      
      if (row == 1)
        t <- ""
      
    }
    
    if (row == 1 & any(brdrs %in% c("outside", "top")) |
        any(cell_border %in% c("outside", "top"))
        ) {
      t <- paste0("border-top:thin solid ", border_color, ";")
    }
      
    if (row == nrow & any(brdrs %in% c("bottom", "outside")) |
        any(cell_border %in% c("bottom", "outside"))
        ) {
      b <- paste0("border-bottom:thin solid ", border_color, ";")
    }
    
    if (col == 1 & any(brdrs %in% c("outside", "left")) |
        any(cell_border %in% c("outside", "left"))
        ) {
      l <- paste0("border-left:thin solid ", border_color, ";")
    }
    
    if (col == ncol & any(brdrs %in% c("outside", "right")) |
        any(cell_border %in% c("outside", "right"))
        ) {
      r <- paste0("border-right:thin solid ", border_color, ";")
    }
    
  }
  
  # Deal with flag
  # Flag is for special rows like blanks or labels
  if (!is.null(flag)) {
    if (flag %in% c("L", "B", "A")) {
      
      if (stub_flag == FALSE & col == 1 & any(brdrs %in% c("outside", "all", "right")))
        r <- paste0("border-right:thin solid ", border_color, ";")
      else if (col != ncol & stub_flag == FALSE)
        r <- ""
      
      if (col != 1)
        l <- ""
    }
  }
  
  
  if (!is.null(exclude)) {
    if (any(exclude == "top"))
      t <- ""
    if (any(exclude == "bottom"))
      b <- ""
    if (any(exclude == "left"))
      l <- ""
    if (any(exclude == "right"))
      r <- ""
  }
  
  ret <- paste0(t, b, l, r)
  
  return(ret)
  
}

get_page_numbers_html <- function(val, tpg = TRUE) {
  
  ret <- val
  
  ret <- gsub("[pg]", "\\chpgn ", ret, fixed = TRUE)
  
  if (tpg)
    ret <- gsub("[tpg]", "{\\field{\\*\\fldinst  NUMPAGES }}", ret, fixed = TRUE)
  
  return(ret)
}

# Report Spec Constructor -----------------------------------------------


#' @title
#' Create a report
#'
#' @description
#' Creates a report shell to which you may add titles, footnotes, content, etc.
#' 
#' @details
#' This function is the constructor for the report object.  The report
#' object contains information needed to create a report. The object is
#' defined as an S3 object, and has a class of 'report_spec'.
#' 
#' The report object holds information concerning report page size, orientation, 
#' titles, footnotes, page header, page footer, margins, and other options.  
#' Use the \code{\link{add_content}} function to add content to the report.  
#' The report may be written to a file using the \code{\link{write_report}} 
#' function. 
#' 
#' @section Report family of functions:
#' The report is the primary container for report specifications.  The
#' following functions add additional specifications to the report object 
#' initialized with \code{create_report}.
#' \itemize{
#'   \item \code{\link{titles}} to add titles to the report.
#'   \item \code{\link{footnotes}} to add footnotes to the report.
#'   \item \code{\link{title_header}} to add a title header to the report.
#'   \item \code{\link{page_header}} to add a page header to the report. 
#'   \item \code{\link{page_footer}} to add a page_footer to the report. 
#'   \item \code{\link{add_content}} to add content to the report.
#'   \item \code{\link{options_fixed}} to set options for fixed-width output.
#'   \item \code{\link{add_content}} to add content to the report.
#'   \item \code{\link{write_report}} to write the report to the file system.
#' }
#' 
#' The report family of functions are pipe-friendly.  After creating the 
#' report, you may pipe the object to any of the above functions to append
#' additional options.
#'
#' @param file_path The output path of the desired report. Either a full path or
#' a relative path is acceptable.  This parameter is not required to create the
#' report_spec object, but will be required to write the report.  In addition, 
#' the file extension is not required.  If the file extension is not supplied,
#' the \code{\link{write_report}} function will add a file extension based
#' on the \code{output_type} specified.
#' @param output_type The report output type.  Default is "TXT".  Valid
#' values are "TXT", "RTF", "PDF", "HTML", and "DOCX".
#' @param orientation The page orientation of the desired report.  Valid values
#' are "landscape" or "portrait".  The default page orientation is "landscape".
#' @param units Specifies the units of measurement.  This setting will 
#' indicate the units for columns widths, margins, paper size, and other 
#' measurements. Valid values are "inches" or "cm" (centimeters).  
#' Default value is "inches".
#' @param paper_size The expected paper size on which the report may be 
#' printed.  The \code{paper_size} will determine how much text can fit on
#' one page.  Valid values are "letter", "legal", "A4", and "RD4".  Default is 
#' "letter".  For the HTML output type, a paper size of "none" is also valid. 
#' That means the HTML will be generated in an unbounded manner as a typical
#' web page.  For a custom page size, the parameter also accepts a double vector
#' with the page width and height.  The width and height should be in the report
#' units of measure, and assume a portrait orientation.
#' @param missing How to display missing values in the report.  Default is
#' to replace them with an empty string, which removes them from the report.
#' To display missing values as is, set the missing parameter to NULL.  To
#' replace missing values with a character string (such as ".", "-", or "<NA>")
#' pass the desired character string to the missing parameter.
#' @param font The font to use on the report.  The font specified will be
#' used for the entire report.  Valid values are "Courier", "Arial", "Times",
#' and "fixed".  The value of "fixed" will create a fixed-width, text style
#' report in Courier font.  The \code{font} parameter only applies to 
#' RTF, HTML, PDF, and DOCX reports.  The default value is "Courier".
#' @param font_size The size of the font to use on the report. The \code{font_size}
#' specified will be used for the entire report.  Valid values are 8, 9, 10, 11,
#' and 12.  The \code{font_size} parameter only applies to RTF, PDF, HTML, and  
#' DOCX output types. The default value is 10.
#' @return A new report_spec object.
#' @family report
#' @seealso \code{\link{create_table}}, \code{\link{create_text}}, and
#' \code{\link{create_plot}} functions
#' to create content for the report.
#' @examples
#' library(reporter)
#' library(magrittr)
#' 
#' # Create temp file path
#' tmp <- file.path(tempdir(), "airquality.txt")
#' 
#' # Prepare Data
#' dat <- airquality[sample(1:153, 15), ]
#' dat$Month <-  as.Date(paste0("1973-", dat$Month, "-01"))
#' 
#' # Define table
#' tbl <- create_table(dat, show_cols = c("Month", "Day", "Wind", "Temp", "Ozone")) %>% 
#'   titles("Table 9.6", "Air Quality Sample Report") %>% 
#'   column_defaults(width = .5) %>% 
#'   define(Month, format = "%B", align = "left", width = 1) %>% 
#'   define(Temp, format = "%.0f") %>% 
#'   footnotes("* New York, May to September 1973")
#' 
#' # Define report 
#' rpt <- create_report(tmp, orientation = "portrait",  missing = "-") %>% 
#'   add_content(tbl) 
#' 
#' # Write the report to the file system
#' write_report(rpt)
#' 
#' # Write the report to the console
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' #                      Table 9.6
#' #              Air Quality Sample Report
#' #      
#' #      Month           Day   Wind   Temp  Ozone
#' #      ----------------------------------------
#' #      July              8    6.3     92     97
#' #      July              9    5.7     92     97
#' #      August            1    6.9     81     39
#' #      July             23   11.5     82      -
#' #      June              9   13.8     90     71
#' #      July             12   14.3     73     10
#' #      July              4   10.9     84      -
#' #      May              31    7.4     76     37
#' #      September        30   11.5     68     20
#' #      June             25      8     75      -
#' #      June             28   11.5     80      -
#' #      August           18    7.4     76     23
#' #      June             20   10.3     76     13
#' #      July              1    4.1     84    135
#' #      May              23    9.7     61      4
#' #      
#' #      * New York, May to September 1973
#' @export
create_report <- function(file_path = "", output_type = "TXT", 
                          orientation ="landscape", units = "inches",
                          paper_size = "letter", missing = "",
                          font = "Courier", font_size = NULL) {
  # font = "fixed"
  # font_size = NULL

  x <- structure(list(), class = c("report_spec", "list"))

  # Trap missing or invalid output_type parameter
  if (!toupper(output_type) %in% c("TXT", "PDF", "RTF", "HTML", "DOCX")) {
    
    stop(paste0("output_type parameter on create_report() ",
                "function is invalid: '", output_type,
                "'\n\tValid values are: 'TXT', 'PDF', 'RTF', 'HTML', and 'DOCX'."))
  } else {
    
    output_type <- toupper(output_type) 
  }
  
  # Trap missing or invalid orientation parameter.
  if (!orientation %in% c("landscape", "portrait")) {

    stop(paste0("orientation parameter on ",
                "create_report() function is invalid: '", orientation,
                "'\n\tValid values are: 'landscape' or 'portrait'."))
  }
  
  # Trap missing or invalid units parameter.
  if (!units %in% c("inches", "cm")) {
    
    stop(paste0("units parameter on ",
                "create_report() function is invalid: '", units,
                "'\n\tValid values are: 'inches' or 'cm'."))
  }
  
  # if (!units %in% c("inches", "cm", "char")) {
  #   
  #   stop(paste0("units parameter on ",
  #               "create_report() function is invalid: '", units,
  #               "'\n\tValid values are: 'inches', 'cm', or 'char'."))
  # }
  
  
  # Trap missing or invalid paper_size parameter.
  if (typeof(paper_size) == "character") {
    if (!paper_size %in% c("letter", "legal", "A4", "RD4", "none")) {
      
      stop(paste0("paper_size parameter on ",
                  "create_report() function is invalid: '", paper_size,
                  "'\n\tValid values are: 'letter', 'legal', 'A4', 'RD4', 'none'."))
    }
  } else if (typeof(paper_size) == "double") {
   if (length(paper_size) != 2) { 
    stop("Invalid page_size parameter.")
   }
  } else {
    stop("Invalid page_size parameter.") 
  }
  
  # Trap missing or invalid font parameter
  if (is.null(font))
    font <- "fixed"
  else {
    if (!tolower(font) %in% c("fixed", "courier", "arial", "times"))
      stop(paste0("font value invalid.  ", 
                  "Valid values are 'Courier', 'Arial', 'Times', and 'fixed'."))
  }
  
  # Trap invalid font_size parameter
  if (!is.null(font_size)) {
    if (!font_size %in% c(8, 9, 10, 11, 12)) {
      stop("font_size parameter invalid.  Valid values are 8, 9, 10, 11, and 12.") 
    }
  }
    
  # Populate report_spec fields
  x$file_path <- file_path
  x$output_type <- output_type
  x$orientation <- orientation
  x$content <- list()           # Initialize content list
  x$units <- units              # Unit of measure
  if (typeof(paper_size) == "character") {
    x$paper_size <- paper_size
    x$page_size <- get_page_size(paper_size, units)
  } else {
    x$paper_size <- "custom"
    x$page_size <- paper_size
  }
  x$pages <- 0                  # Track # of pages in report
  x$column_widths <- list()      # Capture table column widths for reference
  x$missing <- missing
  x$font <- font      
  x$font_size <- font_size

  # Not needed for DOCX and HTML because there is no text version
  if (output_type %in% c("TXT", "PDF", "RTF")) {
    
    # Set default options for text
    # This sets row_height and char_width
    # which are needed for all conversions from 
    # units to text
    x <- options_fixed(x)
    
  } 
    
  # else if (output_type == "docx") {
  #   
  #   # Set default options for docx
  #   #x <- options_docx(x)
  #   
  # }

  # Set default margins
  x <- set_margins(x)

  return(x)

}


# Options -----------------------------------------------------------------


#' @description This is a lookup table to get standard settings for various 
#' editors.
#' @noRd
editor_settings <- read.table(header = TRUE, text = '
                    editor          cpi     cpcm     lpi     lpcm    mmi   mmcm
                    editplus       14.2     5.75     6.1     2.62      0      0
                    notepad          12     4.75    5.65     2.22      0      0
                    notepad++        12     4.75    6.50     2.55  0.393      1
                    word           11.2      4.4       6     2.35      0      0
                    wordpad        10.8      4.2       6     2.35      0      0
                    pdf12            12     4.70       5    2.000  .1967     .5
                    pdf11            13     5.11     5.6     2.25  .1967     .5
                    pdf10       14.2222     5.58    6.10      2.4  .1967     .5
                    pdf9           15.8     6.22     6.8     2.67  .1967     .5
                    pdf8           17.5     6.88    7.55     2.95  .1967     .5
                    rtf12            10   3.9473     5.3     2.05      0      0
                    rtf11            11   4.3307     5.7    2.244      0      0
                    rtf10            12   4.7619    6.38      2.5      0      0
                    rtf9           13.5   5.3149    7.12      2.8      0      0
                    rtf8             15      5.9    7.95     3.05      0      0
                               ') 


#' @title
#' Set options for a fixed-width report
#'
#' @description
#' This function sets the options for a report  
#' with a fixed width font. 
#' 
#' @details The \code{options_fixed} function sets options for reports 
#' with a fixed-width, monospace font.  These reports are based off a 
#' text report, but may be output as type "RTF" or "PDF".  
#' 
#' @section Text Reports:
#' The \code{options_fixed} function sets
#' the characters per 
#' unit of measure (\code{cpuom}) and lines per unit of measure
#' (\code{lpuom}) settings for the report.  These settings determine how 
#' many characters and lines will fit within one unit of measure (uom), as 
#' specified on the \code{\link{create_report}} function.  These settings are
#' important to ensure the report content stays within the available page size 
#' and margins.  Because every editor allows a different number of 
#' characters and lines on a page, these settings must be adjusted depending
#' on the editor.  
#' 
#' The \code{options_fixed} function provides a shortcut 
#' \code{editor} parameter
#' to directly specify a popular editor.  If this parameter is specified, the
#' function will set the characters per unit of measure and lines per
#' unit of measure for you.  If the editor is not available in the 
#' \code{editor} parameter selections, for best results, you should 
#' set the \code{cpuom} and 
#' \code{lpuom} parameters manually.  To determine your \code{cpuom}
#' and \code{lpuom}, see the help for \code{\link{write_registration_file}}.
#' 
#' Alternatively, using the \code{options_fixed} function, 
#' you may set the \code{line_size} and \code{line_count} directly.  Note that
#' the \code{line_size} and \code{line_count} may be different for different
#' output types and editors. 
#'
#' The \code{min_margin} parameter is used to set the minimum margin allowed
#' by the editor.  This value will be subtracted from the margin settings 
#' when the \code{blank_margins} option is used. It is useful for 
#' editors that do not calculate margins from the edge of the page.
#' 
#' As some editors do not support Unicode characters, it may be necessary 
#' to change the character used for the header and spanning header underlines.
#' The default character is a Unicode #U00AF macron.  The macron is sometimes
#' referred to as an "overline", since it is located at the top of the 
#' character area.  If your editor does not support Unicode, the macron
#' will not be displayed properly.  In this case, change the underline character
#' to a dash ("-") or an underscore ("_") using the \code{uchar} parameter.
#' 
#' @section RTF and PDF Reports:
#' For RTF and PDF reports, most of the parameters on the \code{options_fixed}
#' function do not apply.  For RTF and PDF reports, these parameters will
#' be set automatically, and cannot be changed.  
#' 
#' Some of the \code{options_fixed} function apply only to RTF and PDF.
#' In particular, the \code{font_size} parameter applies only to RTF and PDF
#' reports.  Valid font size options are 8, 9, 10, 11, and 12.  The font size may
#' also be set on the \code{\link{create_report}} function.
#' 
#' @param x The report spec.
#' @param editor The expected text editor to use for printing text reports.  
#' Assigning this parameter will set the \code{cpuom} and \code{lpuom} 
#' parameters appropriately for the text editor.  Valid values are 'notepad',
#' 'word', 'wordpad', 'notepad++', and 'editplus'.  If the editor parameter 
#' is used, any settings for \code{cpuom} and \code{lpuom} will be 
#' ignored. It is not necessary to set this parameter for RTF and PDF reports.
#' @param cpuom Characters per unit of measure of printed text.    
#' If units is inches, the default is 12.  If units is centimeters (cm), the 
#' default is 4.687.  This value will be used to 
#' determine how many characters can fit on a line.  
#' @param lpuom Lines per unit of measure of the printed text. Default for 
#' inches is 6. The default for centimeters (cm) is 2.55.  This value 
#' will be used to determine the number of lines that can fit on a page. 
#' @param min_margin The editor minimum margin.  This parameter normally
#' defaults to 0, but may be set for some types of editors.  
#' @param blank_margins When this option is TRUE, \strong{reporter} will use blank 
#' spaces and blank rows to create left and top margins, rather than rely 
#' on the editor to set margins.  When used, editor margins
#' should be set to zero.  Valid values are TRUE and FALSE. Default is
#' FALSE.  This option is only valid for \code{output_type = 'TXT'}.
#' @param font_size The size of the font in points.  Default is 10pt.  This
#' option is only valid for output types RTF and PDF.  Valid values are 8, 9, 10, 
#' 11, and 12.
#' @param line_size The number of characters that will fit on a line.  Normally,
#' the \code{line_size} is calculated based on the page size, font size, and cpuom.
#' You can override the calculated value by setting the \code{line_size}
#' directly.  
#' @param line_count The number of lines that will fit on page.  Normally,
#' the \code{line_count} is calculated based on the page size, font size, and lpuom.
#' You can override the calculated value by setting the \code{line_count}
#' directly.
#' @param uchar The character to use for underlines on the table 
#' header and spanning headers.  Default is a Unicode macron character #U00AF.
#' You may use a dash or underscore if your editor does not support
#' Unicode.  The \code{uchar} is forced to a dash for PDF output, 
#' as the LaTeX converter does not support the macron character.
#' @return The updated report spec.
#' @seealso \code{\link{create_report}} to create a report and set the unit
#' of measure, \code{\link{write_registration_file}} to determine the 
#' characters and lines per unit of measure manually.
#' @family report
#' @encoding UTF-8
#' @examples
#' library(reporter)
#' library(magrittr)
#' 
#' # Create a temporary file
#' tmp <- file.path(tempdir(), "bod.txt")
#'
#' # Define table
#' tbl <- create_table(BOD, width = 2.5) %>% 
#'   titles("Table 3.6", "BOD* Sample Report") %>% 
#'   define(Time, format = "Day %s", align = "left") %>% 
#'   define(demand, format = "%2.1f mg/l", label = "Demand") %>% 
#'   footnotes("* Biochemical Oxygen Demand")
#'        
#' # Define report #1 - No blank margins
#' rpt <- create_report(tmp, orientation="portrait") %>%
#'   add_content(tbl, align = "left") 
#' 
#' # Write the report
#' write_report(rpt)
#' 
#' # Write report to console
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' #           Table 3.6
#' #       BOD* Sample Report
#' # 
#' # Time                  Demand
#' # ----------------------------
#' # Day 1               8.3 mg/l
#' # Day 2              10.3 mg/l
#' # Day 3              19.0 mg/l
#' # Day 4              16.0 mg/l
#' # Day 5              15.6 mg/l
#' # Day 7              19.8 mg/l
#' # 
#' # * Biochemical Oxygen Demand
#' 
#' 
#' # Define report #2 - blank margins
#' rpt <- create_report(tmp, orientation="portrait") %>%
#'   options_fixed(blank_margins = TRUE) %>% 
#'   set_margins(top = .5, left = 1) %>% 
#'   add_content(tbl, align = "left") 
#' 
#' # Write the report
#' write_report(rpt)
#' 
#' # Write report to console
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' # 
#' # 
#' # 
#' #                       Table 3.6
#' #                   BOD* Sample Report
#' # 
#' #              Time                  Demand
#' #              ----------------------------
#' #              Day 1               8.3 mg/l
#' #              Day 2              10.3 mg/l
#' #              Day 3              19.0 mg/l
#' #              Day 4              16.0 mg/l
#' #              Day 5              15.6 mg/l
#' #              Day 7              19.8 mg/l
#' # 
#' #              * Biochemical Oxygen Demand
#' @export
options_fixed <- function(x, editor = NULL, cpuom = NULL, lpuom = NULL,
                          min_margin = NULL, blank_margins = FALSE,
                          font_size = NULL, line_size = NULL, line_count = NULL,
                          uchar = "\U00AF") {
  
  # Set font_size defaults
  if (is.null(x$font_size) & is.null(font_size)) {
    font_size <- 10
  } else if (!is.null(x$font_size) & is.null(font_size)) {
    font_size <- x$font_size
  }

  if (!"report_spec" %in% class(x)) {
    stop("Input object must be of class 'report_spec'.") 
  }
  
  if (x$output_type == "TXT") {
    if (is.null(editor)) {
      # Trap missing or invalid cpuom parameter.
      if (is.null(cpuom)) {
        if (x$units == "inches")
          x$cpuom <- 12 
        else if (x$units == "cm")
          x$cpuom <- 4.687
        else if (x$units == "char")
          x$cpuom <- 1
      } else if (!(cpuom >= 1 & cpuom <= 15)) {
        
        stop(paste0("cpi parameter on create_report() ",
                    "function is invalid: '", cpuom,
                    "'\n\tValue must be between 0 and 15."))
      }
      else
        x$cpuom <- cpuom
        
      # Trap missing or invalid lpuom parameter.
      if (is.null(lpuom)) {
        if (x$units == "inches")
          x$lpuom <-  6 
        else if (x$units == "cm")
          x$lpuom <- 2.55
        else if (x$units == "char")
          x$lpuom <- 1
      } else if (!(lpuom > 0 & lpuom <= 10)) {
        
        stop(paste0("lpuom parameter on create_report() ",
                    "function is invalid: '", lpuom,
                    "'\n\tValue must be between 0 and 10."))
      }
      else
        x$lpuom <- lpuom
        
      if (!is.null(min_margin)) {
        if (is.na(min_margin) | min_margin < 0 | !is.numeric(min_margin)){
          stop("ERROR: invalid value for min_margin")
        }
        else
          x$min_margin = min_margin
      } else
        x$min_margin = 0
  
    } else {
      
      if (!tolower(editor) %in% c("notepad", "word", "wordpad", "notepad++",
                                  "editplus")) {
        
       stop(paste("editor parameter invalid.  Valid values are:", 
                  "'notepad', 'word', 'wordpad', 'notepad++','editplus'"))
      }
      
      e <- editor_settings[editor_settings$editor == tolower(editor), ]
      
      x$editor <- editor
      
      # Set characters per unit of measure
      # and lines per unit of measure
      if (x$units == "inches") {
        x$cpuom <- e$cpi
        x$lpuom <- e$lpi
        x$min_margin <- e$mmi
      } else if (x$units == "cm") {
        x$cpuom <- e$cpcm
        x$lpuom <- e$lpcm
        x$min_margin <- e$mmcm
      } else if (x$units == "char") {
        
        x$cpuom <- 1
        x$lpuom <- 1
        x$min_margin <- round(e$mmi / 12)
      }
      
      # print(paste("cpuom:", x$cpuom))
      # print(paste("lpuom:", x$lpuom))
    }
    
    
    x$uchar <- uchar
    x$blank_margins <- blank_margins
  
  } else if (x$output_type == "PDF") {
    
    
    if (font_size == 12)
      e <- editor_settings[editor_settings$editor == "pdf12", ]
    else if (font_size == 10)
      e <- editor_settings[editor_settings$editor == "pdf10", ]
    else if (font_size == 8)
      e <- editor_settings[editor_settings$editor == "pdf8", ]
    else if (font_size == 9)
      e <- editor_settings[editor_settings$editor == "pdf9", ]
    else if (font_size == 11)
      e <- editor_settings[editor_settings$editor == "pdf11", ]
    else 
      stop("Invalid font_size setting.  Valid values are 8, 9, 10, 11, and 12")
    
    # Set cpuom and lpuom
    if (x$units == "inches") {
      x$cpuom <- e$cpi
      x$lpuom <- e$lpi
      x$min_margin <- e$mmi
    } else if (x$units == "cm") {
      x$cpuom <- e$cpcm
      x$lpuom <- e$lpcm
      x$min_margin <- e$mmcm
    } else if (x$units == "char") {
      
      x$cpuom <- 1
      x$lpuom <- 1
      x$min_margin <- round(e$mmi / 12)
    }
    
    if (uchar == "\U00AF")
      x$uchar <- "-"
    else 
      x$uchar <- uchar
    
    x$blank_margins <- FALSE
    
  } else if (x$output_type == "RTF") {

    
    if (font_size == 12)
      e <- editor_settings[editor_settings$editor == "rtf12", ]
    else if (font_size == 10)
      e <- editor_settings[editor_settings$editor == "rtf10", ]
    else if (font_size == 11)
      e <- editor_settings[editor_settings$editor == "rtf11", ]
    else if (font_size == 9)
      e <- editor_settings[editor_settings$editor == "rtf9", ]
    else if (font_size == 8)
      e <- editor_settings[editor_settings$editor == "rtf8", ]
    else 
      stop("Invalid font_size setting.  Valid values are 8, 9, 10, 11, and 12")
    
    # Set cpuom and lpuom
    if (x$units == "inches") {
      x$cpuom <- e$cpi
      x$lpuom <- e$lpi
      x$min_margin <- e$mmi
    } else if (x$units == "cm") {
      x$cpuom <- e$cpcm
      x$lpuom <- e$lpcm
      x$min_margin <- e$mmcm
    } else if (x$units == "char") {
      x$cpuom <- 1
      x$lpuom <- 1
      x$min_margin <- round(e$mmi / 12)
    }
    
    x$blank_margins <- FALSE
    
    if (uchar == "\U00AF") {
      if (Sys.info()[["sysname"]] != "Windows")
        x$uchar <- "-"
      else 
        x$uchar <- uchar
    } else 
      x$uchar <- uchar
    
  }
  
  if (!is.null(line_size)) {
    if (!is.numeric(line_size))
      stop("line_size must be a number.")
    if (line_size <= 0)
      stop("line_size must be greater than zero.")
  }
  
  if (!is.null(line_count)) {
    if (!is.numeric(line_count))
      stop("line_count must be a number.")
    if (line_count <= 0)
      stop("line_count must be greater than zero.")
  }

  x$font_size <- font_size
  x$char_width <- 1 / x$cpuom
  x$row_height <- 1 / x$lpuom
  x$user_line_size <- line_size
  x$user_line_count <- line_count
  
  return(x)
  
}

#' @title
#' Set page margins
#' @description Sets the page margins for the report.  The units for this 
#' parameter can be inches or centimeters, depending on the units of measure 
#' specified on the \code{\link{create_report}} function.  
#' @details
#' The margins set with \code{set_margins} will be used for the entire report.  
#' Units for the margins
#' are specified by the \code{units} parameter on the 
#' \code{\link{create_report}} function.  Available units are 'inches' and 'cm'.
#' When the unit of measure is inches, default margins are 1 inch on the left and 
#' right, and .5 inches on top and bottom.  When the unit of measure is 
#' centimeters, default margins are 2.54 cm on left and right, and 1.27 cm
#' on top and bottom.
#'
#' Note that when using output type of TXT, and not using the 
#' \code{blank_margins} option, setting the margins only reduces
#' the area available for content on a page.  You must still set the actual
#' margins on the available editor to match those specified in 
#' \code{set_margins}.  Any mismatch may result in content not fitting properly
#' on the page. For best results, set the right and bottom margins to zero 
#' to allow for slight overflow without causing a page break or wrapping lines.
#' @param x The report spec object.
#' @param top The top margin.
#' @param bottom The bottom margin.
#' @param left The left margin.
#' @param right The right margin.
#' @return The report_spec with margins set as desired.
#' @family report
#' @examples
#' library(reporter)
#' library(magrittr)
#' 
#' # Create a temporary file
#' tmp <- file.path(tempdir(), "bod.txt")
#'
#' # Define table
#' tbl <- create_table(BOD, width = 2.5) %>% 
#'   titles("Table 3.6", "BOD¹ Sample Report") %>% 
#'   define(Time, format = "Day %s", align = "left") %>% 
#'   define(demand, format = "%2.1f mg/l", label = "Demand") %>% 
#'   footnotes("¹ Biochemical Oxygen Demand")
#'        
#' # Define report #1 - No blank margins
#' rpt <- create_report(tmp, orientation="portrait") %>%
#'   add_content(tbl, align = "left") 
#' 
#' # Write the report
#' write_report(rpt)
#' 
#' # Write report to console
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' #           Table 3.6
#' #       BOD* Sample Report
#' # 
#' # Time                  Demand
#' # ----------------------------
#' # Day 1               8.3 mg/l
#' # Day 2              10.3 mg/l
#' # Day 3              19.0 mg/l
#' # Day 4              16.0 mg/l
#' # Day 5              15.6 mg/l
#' # Day 7              19.8 mg/l
#' # 
#' # * Biochemical Oxygen Demand
#' 
#' 
#' # Define report #2 - blank margins
#' rpt <- create_report(tmp, orientation="portrait") %>%
#'   options_fixed(blank_margins = TRUE) %>% 
#'   set_margins(top = .5, left = 1) %>% 
#'   add_content(tbl, align = "left") 
#' 
#' # Write the report
#' write_report(rpt)
#' 
#' # Write report to console
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' # 
#' # 
#' # 
#' #                       Table 3.6
#' #                   BOD* Sample Report
#' # 
#' #              Time                  Demand
#' #              ----------------------------
#' #              Day 1               8.3 mg/l
#' #              Day 2              10.3 mg/l
#' #              Day 3              19.0 mg/l
#' #              Day 4              16.0 mg/l
#' #              Day 5              15.6 mg/l
#' #              Day 7              19.8 mg/l
#' # 
#' #              * Biochemical Oxygen Demand
#' @export
set_margins <- function(x, top=NULL, bottom=NULL,
                           left=NULL, right=NULL) {

  if (!is.null(top)) {
    if (is.na(top) | top < 0 | !is.numeric(top)){
      stop("ERROR: invalid value for top")
    } 
    else
      x$margin_top = top
  }
  else {
    if (x$units == "inches")
      x$margin_top <- .5 
    else if (x$units == "cm")
      x$margin_top <- 1.27
    # else if (x$units == "char")
    #   x$margin_top <- set_char_margins(x, "top")
  }
  
  if (!is.null(bottom)) {
    if (is.na(bottom) | bottom < 0| !is.numeric(bottom)){
      stop("ERROR: invalid value for bottom")
    }
    else
      x$margin_bottom = bottom
  }
  else {
    if (x$units == "inches")
      x$margin_bottom <- .5 
    else if (x$units == "cm")
      x$margin_bottom <- 1.27
    # else if (x$units == "char")
    #   x$margin_bottom <- set_char_margins(x, "bottom")
  }
  
  if (!is.null(left)) {
    if (is.na(left) | left < 0| !is.numeric(left)){
      stop("ERROR: invalid value for left")
    }
    else
      x$margin_left = left
  }
  else {
    if (x$units == "inches")
      x$margin_left <-  1 
    else if (x$units == "cm")
      x$margin_left <- 2.54
    # else if (x$units == "char")
    #   x$margin_left <- set_char_margins(x, "left")
    
  }
  
  if (!is.null(right)) {
    if (is.na(right) | right < 0| !is.numeric(right)){
      stop("ERROR: invalid value for right")
    }
    else 
      x$margin_right = right
  }
  else {
    if (x$units == "inches")
      x$margin_right <- 1 
    else if (x$units == "cm")
      x$margin_right <- 2.54
    # else if (x$units == "char")
    #   x$margin_right <- set_char_margins(x, "right")
  }


  return(x)
}

# Not used now.  May be used in future.
# @noRd
# set_char_margins <- function(rs, margin) {
#   
#   ret <- NULL
#   if (!is.null(rs$font_size)) {
#     if (rs$font_size == 8) {
#       if (margin %in% c("right", "left"))
#         ret <- 15
#       else if (margin %in% c("top", "bottom"))
#         ret <- 4
#     }  
#     if (rs$font_size == 10) {
#       if (margin %in% c("right", "left"))
#         ret <- 12
#       else if (margin %in% c("top", "bottom"))
#         ret <- 3
#     }  
#     if (rs$font_size == 12) {
#       if (margin %in% c("right", "left"))
#         ret <- 10
#       else if (margin %in% c("top", "bottom"))
#         ret <- 3
#     }  
#   }
#   
#   return(ret)
# }

# Page Template Items -----------------------------------------------



#' @title
#' Add a page header
#'
#' @description
#' This function adds a page header to the report.  The page header will appear
#' at the top of each page of the report.  
#'
#' @details
#' The page header may contain text on the left or right. Use the appropriate
#' parameters to specify the desired text.  Only one page header is allowed
#' on a report. The page header will be repeated on every page of the report.
#' Multiple text values for each side
#' may be specified as a vector of strings.  
#' 
#' If the width of the page header
#' string exceeds the available space, an error will be generated. There is 
#' also a limit of 5 page header strings per each side.
#' 
#' There are two special tokens to generate page numbers: [pg] and [tpg]. 
#' Use [pg] to indicate the current page number.  Use [tpg] to indicate the
#' total number of pages in the report.  These tokens may be placed anywhere 
#' in the page header or page footer. 
#' 
#' Each header string must fit within the available space.  The \strong{reporter}
#' package will not wrap headers.  If a header string does not fit within the 
#' available space, an error will be generated.  In these situations, either
#' shorten the header string or split it into multiple headers that each fit 
#' within the available space. 
#' 
#' @param x The report object.
#' @param left The left page header text.  May be a single string or a vector
#' of strings.
#' @param right The right page header text.  May be a single string or a vector
#' of strings.
#' @param blank_row Whether to create a blank row below the page header.
#' Valid values are 'below' and 'none'.  Default is 'none'.
#' @param width Widths for left, center, and right columns of the page header, 
#' passed as a vector of double values in the report unit of measure.  If the vector
#' contains fewer than three widths, the widths passed will be interpreted from
#' left to right, and any unassigned column widths will be calculated from
#' the remaining width on the page.  For example, if two values are assigned, they will be
#' interpreted as left and center, and the right column width will be calculated.
#' Setting a cell width to zero (0) will remove that column from the page header
#' entirely. To set the left and right widths only, pass a zero for the center
#' cell, i.e. \code{width = c(6, 0, 3)} for a total width of 9 inches.
#' @param center The center page header text.  May be a single string or a vector
#' of strings. By default, the center column is not displayed.  To display it,
#' either assign a value on this parameter, or assign a width on the "width" parameter.
#' @return The modified report specification.
#' @family report
#' @examples
#' library(magrittr)
#' 
#' # Create temp file path
#' tmp <- file.path(tempdir(), "mtcars.txt")
#' 
#' dat <- data.frame(name = rownames(mtcars[1:10, ]), mtcars[1:10, 1:5], 
#'                   stringsAsFactors = FALSE)
#' 
#' # Create the report object
#' rpt <- create_report(tmp, orientation = "portrait") %>% 
#'   page_header("Client: Motor Trend", "Study: Cars") %>% 
#'   titles("MTCARS Sample Report") %>% 
#'   add_content(create_table(dat)) %>% 
#'   page_footer(Sys.time(), right = "Page [pg] of [tpg]")
#' 
#' # Write the report to the file system
#' write_report(rpt)
#' 
#' # Write report to console
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' # Client: Motor Trend                                                Study: Cars
#' #                              MTCARS Sample Report
#' # 
#' #                name                 mpg cyl   disp   hp  drat
#' #                ----------------------------------------------
#' #                Mazda RX4             21   6    160  110   3.9
#' #                Mazda RX4 Wag         21   6    160  110   3.9
#' #                Datsun 710          22.8   4    108   93  3.85
#' #                Hornet 4 Drive      21.4   6    258  110  3.08
#' #                Hornet Sportabout   18.7   8    360  175  3.15
#' #                Valiant             18.1   6    225  105  2.76
#' #                Duster 360          14.3   8    360  245  3.21
#' #                Merc 240D           24.4   4  146.7   62  3.69
#' #                Merc 230            22.8   4  140.8   95  3.92
#' #                Merc 280            19.2   6  167.6  123  3.92
#' # 
#' # ...
#' # 
#' # 2020-10-17 11:53:51                                                Page 1 of 1
#' @export
page_header <- function(x, left="", right="", blank_row = "none", 
                        width = NULL, center = ""){

  if (!"report_spec" %in% class(x))
    stop("Page header can only be assigned to an object of class 'report_spec'")
  
  if (length(left) > 5 | length(right) > 5 | length(center) > 5){
    stop("Header string count exceeds limit of 5 strings per side.")
  }
  
  if (is.null(blank_row))
    blank_row <- "none"
  
  if (!blank_row %in% c("below", "none")) {
    stop(paste("blank_row parameter value invalid:", blank_row, "\n", 
      "Valid values are 'below' and 'none'."))
  }
  
  if (!is.null(x$title_hdr))
    stop("Cannot add both a page header and a title header.")
  
  if (has_glue()) {
    x$page_header_left <- gluev(left)
    x$page_header_right <- gluev(right)
    x$page_header_center <- gluev(center)

  } else {
    
    x$page_header_left <- left
    x$page_header_right <- right
    x$page_header_right <- center

  }
  x$page_header_blank_row <- blank_row
  
  
  if (is.null(width)) {
    # If no length is assigned, set center as 0 for not displaying
    width <- c(NA, 0, NA)
  } else if (length(width) == 1) {
    # If only left length is assigned, set center as 0 for not displaying
    width <- c(width, 0, NA)
  }
  x$page_header_width <- width

  return(x)
}


#' @title
#' Adds a title header block 
#'
#' @description
#' This function adds a title header to an object.  A title header is a special
#' type of title layout that has titles on the left and header information
#' on the right.  
#' @details
#' The \code{title_header} function accepts a set of strings of the desired 
#' title text, and a vector of header strings. The titles will appear on the 
#' left of the title header, and the header strings on the right. To
#' specify multiple titles for the block, pass them to the function 
#' as separate strings.
#' 
#' Title headers may be assigned to a report, a table, a text specification, 
#' or a plot. If assigned to the report, the title header will appear 
#' at the top of the page, and be repeated for every page of the report.  
#' If the title header is assigned to  
#' content, the titles will appear above the content, and be repeated if the 
#' content breaks to the next page.  
#' 
#' One title header function accepts up to 10 titles. Blank rows above or below 
#' the title block may be controlled using the 
#' \code{blank_row} parameter.
#'
#' Each title string must fit within the available space.  The \strong{reporter}
#' package will not wrap titles on fixed-width reports.  
#' If a title does not fit within the 
#' available space, an error will be generated.  In these situations, either
#' shorten the title or split it into multiple titles that each fit within the
#' available space. 
#' 
#' @param x The object to assign titles to.  Valid objects are a report,
#' table, text, or plot specification.
#' @param ... A set of title strings.
#' @param right A set of header strings to be shown on the right side of the
#' title header.  Pass the header strings as a vector of strings.
#' @param blank_row Where to place a blank row.  Valid values are 'above',
#' 'below', 'both', or 'none'.  Default is 'below'.
#' @param borders Whether and where to place a border. Valid values are 'top',
#' 'bottom', 'left', 'right', 'outside', 'inside', 'all', or 'none'.  
#' Default is 'none'.   The 
#' 'left', 'right', 'outside', and 'inside' border specifications only 
#' apply to RTF, HTML, PDF and DOCX reports.
#' @param width The width of the title header.  If the title header is attached
#' to the report, valid values are 'page' or a numeric width, and the default
#' is 'page'. If the title header is attached to the
#' table, plot, or text content, the valid values are 'page', 'content' or a 
#' numeric value, and the default is 'content'.  The value 'content' means the 
#' footnotes will be aligned to the width of the table, plot, or text
#' content.  The value
#' 'page' means the footnotes will be aligned to the width of the page. 
#' In addition to these two convenience settings, you 
#' may also specify a specific width in the current units of measure.  The
#' units of measure is determined by the 'units' parameter on 
#' \code{\link{create_report}}.
#' @return The modified report.
#' @family report
#' @examples
#' library(reporter)
#' library(magrittr)
#' 
#' # Create a temporary file
#' tmp <- file.path(tempdir(), "expenses.txt")
#' 
#' # Prepare data
#' dat <- data.frame(category = rownames(USPersonalExpenditure),
#'                   USPersonalExpenditure, stringsAsFactors = FALSE)
#' 
#' # Define table
#' tbl <- create_table(dat) %>% 
#'   title_header("Table 1.0", "US Personal Expenditures from 1940 - 1960",
#'                right = c("Page [pg] of [tpg]", "World Almanac")) %>% 
#'   column_defaults(from = X1940, to = X1960, width = .6, format = "$%.2f") %>%
#'   define(category, label = "Category") %>% 
#'   define(X1940, label = "1940") %>% 
#'   define(X1945, label = "1945") %>% 
#'   define(X1950, label = "1950") %>% 
#'   define(X1955, label = "1955") %>% 
#'   define(X1960, label = "1960") %>% 
#'   footnotes("* In billions of dollars")
#' 
#' # Define report
#' rpt <- create_report(tmp, orientation="portrait") %>%
#'   add_content(tbl) 
#' 
#' # Write the report
#' write_report(rpt)
#' 
#' # Display in console
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' #     Table 1.0                                        Page 1 of 1
#' #     US Personal Expenditures from 1940 - 1960      World Almanac
#' # 
#' #     Category                1940    1945    1950    1955    1960
#' #     ------------------------------------------------------------
#' #     Food and Tobacco      $22.20  $44.50  $59.60  $73.20  $86.80
#' #     Household Operation   $10.50  $15.50  $29.00  $36.50  $46.20
#' #     Medical and Health     $3.53   $5.76   $9.71  $14.00  $21.10
#' #     Personal Care          $1.04   $1.98   $2.45   $3.40   $5.40
#' #     Private Education      $0.34   $0.97   $1.80   $2.60   $3.64
#' # 
#' #     * In billions of dollars
#' @export
title_header <- function(x, ..., right = "", 
                         blank_row = "below", borders = "none", 
                         width = NULL) {
  

  diff <- setdiff(class(x), c("list"))
  
  if (identical(diff, character(0)) | 
    all(!diff %in% c("report_spec", "table_spec", "plot_spec", "text_spec"))) {
    stop(paste("Title header can only be assigned ", 
                "to an objects of class 'report_spec', 'title_spec', 'plot_spec'",
                "or 'text_spec'."))
  }
  
  # Create title structure
  ttl_hdr <- structure(list(), class = c("title_hdr", "list"))
  
  if (length(c(...)) > 10)
    stop("Limit of 10 titles reached.") 
  
  if (!is.null(x$titles)){
    x$titles
    stop("Cannot add both titles and a title header.")
  }
  
  if (!is.null(x$page_header_left) | !is.null(x$page_header_right))
    stop("Cannot add both a page header and a title header.")
  
  if (!blank_row %in% c("above", "below", "both", "none"))
    stop(paste("Blank row parameter invalid.  Valid values are", 
               "'above', 'below', 'both', or 'none'."))
  
  if (!all(borders %in% c("top", "bottom", "left", "right", 
                          "outside", "inside", "all", "none"))) {
    stop(paste("Borders parameter invalid.  Valid values are", 
      "'top', 'bottom', 'left', 'right', 'outside', 'inside, 'all', or 'none'."))
  }
  
  if (is.null(width)) {
    if (any(class(x) %in% c("report_spec"))) 
      width <- "page"
    else 
      width <- "content"
  } else {
    if (any(class(x) %in% c("report_spec"))) {
      if (!width %in% c("page") & !is.numeric(width))
        stop("Width parameter invalid.  Valid values are 'page' or a number.")
    } else {
      if (!width %in% c("page", "content") & !is.numeric(width))
        stop(paste("Width parameter invalid.  Valid values are 'page',",
                   "'content' or a number."))
    }
    
  }
  
  ttls <- c(...)
  
  # Assign attributes
  if (has_glue()) {
    ttl_hdr$titles <- gluev(ttls)
    ttl_hdr$right <- gluev(right)
  } else {
    ttl_hdr$titles <-  ttls
    ttl_hdr$right <- right
  }
  
  ttl_hdr$blank_row <- blank_row
  ttl_hdr$borders <- borders

  ttl_hdr$width <- width
  
  x$title_hdr[[length(x$title_hdr) + 1]] <- ttl_hdr
  
  return(x)
  
}

#' @title
#' Adds a title block 
#'
#' @description
#' This function adds one or more titles to an object as a title block.  
#' If added to a report, 
#' the titles will be added to
#' the page template, and thus appear on each page of the report. Titles may 
#' also be added to a table, text, or plot object.
#'
#' @details
#' The titles function accepts a set of strings of the desired title text. To
#' specify multiple titles for the block, pass them to the function 
#' as separate strings.
#' 
#' The titles may be aligned center, left or right using the align parameter.
#' The alignment will be applied to all titles contained in the 
#' block.  To control alignment of titles separately for each title, use 
#' multiple titles functions.
#' 
#' Titles may be assigned to a report, a table, a text specification, or a plot. 
#' If assigned to the report, the title will appear at the top of the page, and
#' be repeated for every page of the report.  If the titles are assigned to  
#' content, the titles will appear above the content, and be repeated if the 
#' content breaks to the next page.  
#' 
#' If titles are assigned to the report,
#' alignment will be oriented to the page body.  If titles are assigned to
#' content, alignment will be oriented to the edge of the content.
#' 
#' One title function accepts up to 10 titles.  However, multiple title 
#' blocks may be added to the same object if needed.  
#' 
#' Blank rows above or below the title block may be controlled using the 
#' \code{blank_row} parameter.
#' 
#' Each title string must fit within the available space.  The \strong{reporter}
#' package will not wrap titles on fixed-width reports.  
#' If a title does not fit within the 
#' available space, a warning will be generated and the title will be 
#' truncated.  In these situations, either
#' shorten the title or split it into multiple titles that each fit within the
#' available space. 
#'
#' @param x The object to assign titles to.  Valid objects are a report, or
#' a table, text, or plot specification.
#' @param ... A set of title strings.
#' @param align The position to align the titles.  Valid values are 'left', 
#' 'right', 'center' or 'centre'.  For titles, the default is 'center'.
#' @param blank_row Where to place a blank row.  Valid values are 'above',
#' 'below', 'both', or 'none'.  Default is "below".
#' @param borders Whether and where to place a border. Valid values are 'top',
#' 'bottom', 'left', 'right', 'outside', 'inside', 'all', or 'none'.  
#' Default is "none".  
#' The 'left', 'right', 'outside', and 'inside' border specifications 
#' only apply to RTF, HTML, PDF and DOCX reports.
#' @param width The width of the titles block.  If the titles are attached
#' to the report, valid values are 'page' or a numeric width, and the default
#' is 'page'. If the titles are attached to the
#' table, plot, or text content, the valid values are 'page', 'content' or a 
#' numeric value, and the default is 'content'.  The value 'content' means the 
#' titles will be aligned to the width of the table, plot, or text
#' content.  The value
#' 'page' means the titles will be aligned to the width of the page. 
#' In addition to these two convenience settings, you 
#' may also specify a specific width in the current units of measure.  The
#' units of measure is determined by the 'units' parameter on 
#' \code{\link{create_report}}.
#' @param bold A parameter to bold the titles.  Valid values are TRUE and FALSE.
#' Default is FALSE.  This parameter only applies to variable-width RTF, HTML, 
#' PDF, and DOCX output types.
#' @param font_size The font size to use for the title block.  The font size
#' of the report will be used by default.  Valid values are 8, 9, 10, 11, 12,
#' 13, and 14.  This parameter only applies to variable-width RTF, HTML, PDF,
#' and DOCX output types.
#' @param header Whether to put the titles in the page header. Valid values
#' are TRUE and FALSE.  Default is FALSE. This option only works on the RTF 
#' and DOCX output types, and only applies to titles assigned 
#' to the report object.  Titles in the page header will appear on every page, 
#' and be the same throughout the report.
#' @param columns The number of columns for the title block. Valid values
#' are 1, 2, and 3.  Default is 1.  If this parameter is set to 2, the title
#' block will be split into two columns, each aligned to the outside.  If 
#' this parameter is set to 3, the title block will be split into 3 columns,
#' with the outer columns aligned to the outside and the middle column
#' aligned center.  Titles are assigned to cells from top left to bottom right.
#' @return The modified report.
#' @family report
#' @examples
#' library(reporter)
#' library(magrittr)
#' 
#' # Create a temporary file
#' tmp <- file.path(tempdir(), "expenses.txt")
#' 
#' # Prepare data
#' dat <- data.frame(category = rownames(USPersonalExpenditure),
#'                   USPersonalExpenditure, stringsAsFactors = FALSE)
#' 
#' # Define table
#' tbl <- create_table(dat) %>% 
#'   titles("Table 1.0", "US Personal Expenditures from 1940 - 1960") %>% 
#'   column_defaults(from = X1940, to = X1960, width = .6, format = "$%.2f") %>%
#'   define(category, label = "Category") %>% 
#'   define(X1940, label = "1940") %>% 
#'   define(X1945, label = "1945") %>% 
#'   define(X1950, label = "1950") %>% 
#'   define(X1955, label = "1955") %>% 
#'   define(X1960, label = "1960") %>% 
#'   footnotes("* In billions of dollars")
#' 
#' # Define report
#' rpt <- create_report(tmp, orientation="portrait") %>%
#'   add_content(tbl) 
#' 
#' # Write the report
#' write_report(rpt)
#' 
#' # Display in console
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' #                               Table 1.0
#' #               US Personal Expenditures from 1940 - 1960
#' # 
#' #     Category                1940    1945    1950    1955    1960
#' #     ------------------------------------------------------------
#' #     Food and Tobacco      $22.20  $44.50  $59.60  $73.20  $86.80
#' #     Household Operation   $10.50  $15.50  $29.00  $36.50  $46.20
#' #     Medical and Health     $3.53   $5.76   $9.71  $14.00  $21.10
#' #     Personal Care          $1.04   $1.98   $2.45   $3.40   $5.40
#' #     Private Education      $0.34   $0.97   $1.80   $2.60   $3.64
#' # 
#' #     * In billions of dollars
#' @export
titles <- function(x, ..., align = "center", blank_row = "below", 
                   borders = "none", width = NULL, bold = FALSE, 
                   font_size = NULL, header = FALSE, columns = 1){

  # Create title structure
  ttl <- structure(list(), class = c("title_spec", "list"))
  
  if (length(c(...)) > 10)
    stop("Limit of 10 titles reached.") 
  
  if (!is.null(x$title_hdr))
    stop("Cannot add both titles and a title header.")
  
  if (!all(blank_row %in% c("above", "below", "both", "none")))
    stop(paste("Blank row parameter invalid.  Valid values are", 
               "'above', 'below', 'both', or 'none'."))
  
  if (!all(borders %in% c("top", "bottom", "left", "right", "all", "outside", 
                          "inside", "none")))
    stop(paste("Borders parameter invalid.  Valid values are", 
      "'top', 'bottom', 'left', 'right', 'outside', 'inside', 'all', or 'none'."))

  if (is.null(width)) {
    if (any(class(x) %in% c("report_spec"))) 
      width <- "page"
    else 
      width <- "content"
    
  } else {
    if (any(class(x) %in% c("report_spec"))) {
      if (!width %in% c("page") & !is.numeric(width))
        stop("Width parameter invalid.  Valid values are 'page' or a number.")
    } else {
      if (!width %in% c("page", "content") & !is.numeric(width))
        stop(paste("Width parameter invalid.  Valid values are 'page',",
                   "'content' or a number."))
    }
    
  }
  
  # Trap invalid font_size parameter
  if (!is.null(font_size)) {
    if (!font_size %in% c(8, 9, 10, 11, 12, 13, 14)) {
      stop(paste0("font_size parameter invalid.  ", 
                  "Valid values are 8, 9, 10, 11, 12, 13, and 14."))
    }
  }
  
  if (is.null(columns)) {
    columns <- 1 
  } else {
    if (!columns %in% c(1, 2, 3)) {
      stop("Columns paramter is invalid. Valid values are 1, 2, or 3.") 
    }
    
  }
  
  if (header == TRUE & !"report_spec" %in% class(x)) {
    
    stop("Header titles can only be assigned to a report.")
    
  }
  
  ttls <- c(...)
  
  # Assign attributes
  if (has_glue()) {
    ttl$titles <- gluev(ttls)
  } else {
    ttl$titles <- ttls
  }
  ttl$blank_row <- blank_row
  ttl$borders <- borders
  ttl$align <- align
  ttl$width <- width
  ttl$bold <- bold
  ttl$font_size <- font_size
  ttl$header <- header
  ttl$columns <- columns

  if (header == TRUE)
    x$header_titles[[length(x$header_titles) + 1]] <- ttl
  else 
    x$titles[[length(x$titles) + 1]] <- ttl
  
  return(x)

}


#' @title
#' Adds a footnote block
#'
#' @description
#' The \code{footnotes} function adds one or more footnotes to the report.  
#' If added to the report specification, the footnotes will
#' be added to the page template, and thus appear on each page of the report.
#' Footnotes may also be added directly to table, text, or plot content.
#'
#' @details
#' The \code{footnotes} function accepts a set of strings of the desired 
#' footnote text. The footnotes may be aligned center, left or right using 
#' the align parameter. The user is responsible for adding desired footnote 
#' symbols. Footnote symbols will not be generated automatically.
#' 
#' If footnotes are assigned to the report,
#' alignment will be oriented to the page body.  If footnotes are assigned to
#' a table or text, alignment will be oriented to the edge of the content.
#' 
#' One footnote function accepts up to 25 footnotes.  However, multiple footnote 
#' blocks may be added to the same object.  
#' 
#' Blank rows above or below the footnote block may be controlled using the 
#' \code{blank_row} parameter.
#' 
#' Each footnote string must fit within the available space.  The \strong{reporter}
#' package will not wrap footnotes on fixed-width reports.  
#' If a footnote does not fit within the 
#' available space, a warning will be generated and the footnote will be 
#' truncated.  In these situations, either
#' shorten the footnote or split it into multiple footnotes that each fit within 
#' the available space. 
#' 
#' @param x The object to assign footnotes to.
#' @param ... A set of footnote strings.
#' @param align The position to align the footnotes.  Valid values are: 'left',
#' 'right', 'center', or 'centre'.
#' @param blank_row Whether to print a blank row above or below the footnote.
#' Valid values are 'above', 'below', 'both', or 'none'.  Default is 'above'.
#' @param borders Whether to print a border above or below the footnote. Valid
#' values are 'top', 'bottom', 'outside', 'inside', 'all',  or 'none'.  
#' Default is 'none'.  
#' For fixed width reports, the 
#' border character will be taken from the value of the \code{uchar} parameter
#' on the \code{\link{options_fixed}} function.  The 
#' 'left', 'right', 'outside', and 'inside' border specifications only apply 
#' to RTF, HTML, PDF, and DOCX reports.
#' @param valign The vertical position to align the footnotes.  Valid
#' values are: 'top' and 'bottom'.  For footnotes attached to a report,
#' default is 'bottom'.  For footnotes attached to content, default is 'top'.
#' @param width The width of the footnotes block.  If the footnotes are attached
#' to the report, valid values are 'page' or a numeric width, and the default
#' is 'page'. If the footnotes are attached to the
#' table, plot, or text content, the valid values are 'page', 'content' or a 
#' numeric value, and the default is 'content'.  The value 'content' means the 
#' footnotes will be aligned to the width of the table, plot, or text
#' content.  The value
#' 'page' means the footnotes will be aligned to the width of the page. 
#' In addition to these two convenience settings, you 
#' may also specify a specific width in the current unit of measure.  The
#' unit of measure is determined by the 'units' parameter on 
#' \code{\link{create_report}}.
#' @param footer Whether to put the footnotes in the page footer. Valid
#' values are TRUE and FALSE.  Default is FALSE.  This parameter only works
#' with RTF and DOCX output types.  It also only applies to footnotes
#' assigned to the report object.  Footnotes in the page footer will 
#' appear on every page, and be the same throughout the report.
#' @param italics A TRUE or FALSE value indicating whether the footnote 
#' text shoud be in italics font.  If TRUE, the entire footnote will be 
#' in italics.
#' @param columns The number of columns for the footnote block. Valid values
#' are 1, 2, and 3.  Default is 1.  If this parameter is set to 2, the footnote
#' block will be split into two columns, each aligned to the outside.  If 
#' this parameter is set to 3, the title block will be split into 3 columns,
#' with the outer columns aligned to the outside and the middle column
#' aligned center.  Footnotes are assigned to cells from top left to bottom right.
#' @param font_size The font size to use for the footnote block.  The font size
#' of the report will be used by default.  Valid values are 8, 9, 10, 11, 12
#' This parameter only applies to variable-width RTF, HTML, PDF, and DOCX 
#' output types.
#' @return The modified report.
#' @family report
#' @examples
#' library(reporter)
#' library(magrittr)
#' 
#' # Create a temporary file
#' tmp <- file.path(tempdir(), "expenses.txt")
#' 
#' # Prepare data
#' dat <- data.frame(category = rownames(USPersonalExpenditure),
#'                   USPersonalExpenditure, stringsAsFactors = FALSE)
#' 
#' # Define table
#' tbl <- create_table(dat) %>% 
#'   titles("Table 1.0", "US Personal Expenditures from 1940 - 1960") %>% 
#'   column_defaults(from = X1940, to = X1960, width = .6, format = "$%.2f") %>%
#'   define(category, label = "Category") %>% 
#'   define(X1940, label = "1940") %>% 
#'   define(X1945, label = "1945") %>% 
#'   define(X1950, label = "1950") %>% 
#'   define(X1955, label = "1955") %>% 
#'   define(X1960, label = "1960") %>% 
#'   footnotes("* In billions of dollars")
#' 
#' # Define report
#' rpt <- create_report(tmp, orientation="portrait") %>%
#'   add_content(tbl) 
#' 
#' # Write the report
#' write_report(rpt)
#' 
#' # Display in console
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' #                               Table 1.0
#' #               US Personal Expenditures from 1940 - 1960
#' # 
#' #     Category                1940    1945    1950    1955    1960
#' #     ------------------------------------------------------------
#' #     Food and Tobacco      $22.20  $44.50  $59.60  $73.20  $86.80
#' #     Household Operation   $10.50  $15.50  $29.00  $36.50  $46.20
#' #     Medical and Health     $3.53   $5.76   $9.71  $14.00  $21.10
#' #     Personal Care          $1.04   $1.98   $2.45   $3.40   $5.40
#' #     Private Education      $0.34   $0.97   $1.80   $2.60   $3.64
#' # 
#' #     * In billions of dollars
#' @export
footnotes <- function(x, ..., align = "left", blank_row = "above", 
                      borders = "none", valign = NULL, width = NULL, 
                      footer = FALSE, italics = FALSE, columns = 1,
                      font_size = NULL){

  # Create footnote structure
  ftn <- structure(list(), class = c("footnote_spec", "list"))
  
  ft <- c(...)
  
  if (length(ft) > 25){
    stop("footnotes function is limited to a maximum of 25 footnotes.")
  }
  
  if (!align %in% c("left", "right", "center", "centre"))
    stop(paste("Align parameter invalid. Valid values are 'left', 'right',",
               "'center', and 'centre'"))
  
  if (!is.null(valign)) {
    if (!all(valign %in% c("top", "bottom")))
      stop(paste("Valign parameter invalid. Valid values are 'top' and 'bottom'"))
  }
  
  if (!all(blank_row %in% c("above", "below", "both", "none")))
    stop(paste("Blank row parameter invalid.  Valid values are", 
               "'above', 'below', 'both', or 'none'."))
  
  if (any(!borders %in% c("top", "bottom", "left", "right", "all", 
                          "outside", "inside", "none")))
    stop(paste("Borders parameter invalid.  Valid values are", 
      "'top', 'bottom', 'left', 'right', 'outside', 'inside', 'all', or 'none'."))
  
  # Trap invalid font_size parameter
  if (!is.null(font_size)) {
    if (!font_size %in% 8:12) {
      stop(paste0("Footnote font_size parameter invalid.  ", 
                  "Valid values are 8, 9, 10, 11 and 12."))
    }
  }
  
  if (is.null(width)) {
    if (any(class(x) %in% c("report_spec"))) 
      width <- "page"
    else {
      width <- "content"
      if (!is.null(valign))
        if (valign == "bottom")
          width <- "page"
    }
  } else {
    if (any(class(x) %in% c("report_spec"))) {
      if (!width %in% c("page") & !is.numeric(width))
        stop("Width parameter invalid.  Valid values are 'page' or a number.")
    } else {
      if (!width %in% c("page", "content") & !is.numeric(width))
        stop(paste("Width parameter invalid.  Valid values are 'page',",
                   "'content' or a number."))
    }
  }
  
  if (footer == TRUE & !"report_spec" %in% class(x)) {
    
    stop("Footer footnotes can only be assigned to a report.")
    
  }
  
  if (is.null(columns)) {
    columns <- 1 
  } else {
    if (!columns %in% c(1, 2, 3)) {
      stop("Columns paramter is invalid. Valid values are 1, 2, or 3.") 
    }
    
  }

  if (has_glue()) {
    ftn$footnotes <- gluev(ft)
  } else {
    ftn$footnotes <- ft
  }
  
  
  ftn$blank_row <- blank_row
  ftn$align <- align
  ftn$borders <- borders
  ftn$width <- width  
  ftn$footer <- footer
  ftn$italics <- italics
  ftn$font_size <- font_size
  ftn$columns <- columns

  if (is.null(valign)) {
    if ("report_spec" %in% class(x))
      ftn$valign <- "bottom"
    else
      ftn$valign <- "top"
  } else {
    
   ftn$valign <- valign 
  }
  
  if (footer == TRUE)
    x$footer_footnotes[[length(x$footer_footnotes) + 1]] <- ftn
  else
    x$footnotes[[length(x$footnotes) + 1]] <- ftn

  return(x)

}

#' @title
#' Adds a page footer 
#'
#' @description
#' This function adds a page footer to the report.  The page footer will appear
#' on each page of the report, at the bottom of the page.  The page footer
#' contains three sections: left, center, and right.  Content for each section
#' may be specified with the appropriate parameter.
#'
#' @details
#' Only one page footer is allowed per report.  The page footer will appear 
#' on all pages of the report.  The page footer may contain text on the 
#' left, right, or center. Use the appropriate parameters to specify the 
#' desired text for each section. Multiple strings may be passed to each section 
#' as a vector of strings. 
#' 
#' If the width of the page header
#' string exceeds the available space, an error will be generated. In addition,
#' there is a limit of 5 strings for each page footer section.  
#' 
#' There are two special tokens to generate page numbers: [pg] and [tpg]. 
#' Use [pg] to indicate the current page number.  Use [tpg] to indicate the
#' total number of pages in the report.  These tokens may be placed anywhere 
#' in the page header or page footer. 
#' 
#' Use the \code{blank_row} parameter to control the blank space above the 
#' page footer.
#' 
#' Each footer string must fit within the available space.  The \strong{reporter}
#' package will not wrap footer.  If a footer string does not fit within the 
#' available space, an error will be generated.  In these situations, either
#' shorten the footer string or split it into multiple footers that each fit 
#' within the available space. 
#'
#' @param x The report spec object.
#' @param left The left page footer text.  May be a single string or a vector
#' of strings.
#' @param center The center page footer text.  May be a single string or a
#' vector of strings.
#' @param right The right page footer text.  May be a single string or a vector
#' of strings.
#' @param blank_row Whether to create a blank row above the page footer.
#' Valid values are 'above' and 'none'.  Default is 'above'.
#' @param width Widths for left, center, and right columns of the page footer, 
#' passed as a vector of double values in the report unit of measure.  If the vector
#' contains fewer than three widths, the widths passed will be interpreted from
#' left to right, and any unassigned column widths will be calculated from
#' the remaining width on the page.  For example, if two values are assigned, they will be
#' interpreted as left and center, and the right column width will be calculated.
#' Setting a cell width to zero (0) will remove that column from the page footer
#' entirely. To set the left and right widths only, pass a zero for the center
#' cell, i.e. \code{width = c(6, 0, 3)} for a total width of 9 inches.
#' @return The modified report.
#' @family report
#' @examples
#' library(reporter)
#' library(magrittr)
#' 
#' # Create temp file path
#' tmp <- file.path(tempdir(), "mtcars.txt")
#' 
#' dat <- data.frame(name = rownames(mtcars[1:10, ]), mtcars[1:10, 1:5], 
#'                   stringsAsFactors = FALSE)
#' 
#' # Create the report object
#' rpt <- create_report(tmp, orientation = "portrait") %>% 
#'   page_header("Client: Motor Trend", "Study: Cars") %>% 
#'   titles("MTCARS Sample Report") %>% 
#'   add_content(create_table(dat)) %>% 
#'   page_footer(Sys.time(), right = "Page [pg] of [tpg]")
#' 
#' # Write the report to the file system
#' write_report(rpt)
#' 
#' # Write report to console
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' # Client: Motor Trend                                                Study: Cars
#' #                              MTCARS Sample Report
#' # 
#' #                name                 mpg cyl   disp   hp  drat
#' #                ----------------------------------------------
#' #                Mazda RX4             21   6    160  110   3.9
#' #                Mazda RX4 Wag         21   6    160  110   3.9
#' #                Datsun 710          22.8   4    108   93  3.85
#' #                Hornet 4 Drive      21.4   6    258  110  3.08
#' #                Hornet Sportabout   18.7   8    360  175  3.15
#' #                Valiant             18.1   6    225  105  2.76
#' #                Duster 360          14.3   8    360  245  3.21
#' #                Merc 240D           24.4   4  146.7   62  3.69
#' #                Merc 230            22.8   4  140.8   95  3.92
#' #                Merc 280            19.2   6  167.6  123  3.92
#' # 
#' # ...
#' # 
#' # 2020-10-17 11:53:51                                                Page 1 of 1
#' @export
page_footer <- function(x, left="",  center="", right="", blank_row = "above",
                        width = NULL){

  if (!"report_spec" %in% class(x))
    stop("Page header can only be assigned to an object of class 'report_spec'")
  
  if (length(left) > 5 | length(right) > 5 | length(center) > 5){
    stop("Footer string count exceeds limit of 5 strings per section.")
  }
  
  if (length(width) > 3){
    stop("Width should be a vector with maximum three numeric values.")
  }
  
  if (is.null(blank_row))
    blank_row <- "none"
  
  if (!blank_row %in% c("none", "above"))
    stop("Invalid value for blank_row.  Valid values are 'above' or 'none'.")
  
  if (has_glue()) {
    x$page_footer_left <- gluev(left)
    x$page_footer_right <- gluev(right)
    x$page_footer_center <- gluev(center)
    
  } else {
    x$page_footer_left <- left
    x$page_footer_right <- right
    x$page_footer_center <- center
  }
  x$page_footer_blank_row <- blank_row
  
  # Assign width for left, center, and right
  x$page_footer_width <- width

  return(x)
}


#' @title
#' Adds a page by variable  
#'
#' @description
#' The \code{page_by} function adds a page by variable to a report, table,
#' or plot. 
#' The page by will generate a page break for each value of the page by variable.
#' A label will appear above the content showing the page
#' by variable value. You must be sort the data by the
#' page by variable prior to reporting.
#'
#' @details
#' Only one page by function is allowed per report, table, or plot.  The page by 
#' label will 
#' appear on all pages of the object.  The page by label may be aligned on the 
#' left, right, or center. Use the \code{align} parameter to specify the 
#' alignment. 
#' 
#' You must sort the data by the page by variable prior to reporting.
#' The page by labels will appear in the sorted order.  Failure to sort 
#' the page by variable prior to reporting may produce unexpected results.
#' 
#' @param x The report specification to assign the page by to.
#' @param var The page by variable.  There can be only one page by per report, 
#' and one page by variable.  The page by can be passed either quoted or 
#' unquoted.
#' @param label A label to be used as a prefix to the page by variable value.
#' By default, the label will be assigned to the variable name.  Alternatively,
#' you may specify a string value to use for the label.
#' @param align How to align the page by.  Default value is 'left'. Valid
#' values are 'left', 'right', 'center', or 'centre'.
#' @param blank_row Indicates whether a blank row is desired above or below
#' the page by.  Default value is 'none'.  Valid values are 'above', 'below',
#' 'both', or 'none'.
#' @param borders Whether and where to place a border. Valid values are 'top',
#' 'bottom', 'left', 'right', 'all', 'outside', or 'none'.  Default is "none".  
#' The 'left' and 'right' border specifications only apply to RTF, HTML, PDF,
#' and DOCX reports.
#' @param format The format to use for the page by column data.  The format can 
#' be a string format, a formatting function, a lookup list, a user-defined
#' format, or a formatting list. All formatting is performed by the 
#' \code{\link[fmtr]{fapply}} function from
#' the \code{\link[fmtr]{fmtr}} package.  For 
#' a list of common formatting codes, see \link[fmtr]{FormattingStrings}.
#' @param bold A parameter to bold the page by text. Valid values are TRUE, FALSE, 
#' 'label' or 'value'. Default is FALSE. A value of TRUE bolds both the label 
#' and the page by value. The value 'label' bolds only the label, while a value
#' of 'value' bolds only the page by value. This parameter 
#' applies to variable-width RTF, HTML, PDF, and DOCX output types only.  
#' @family report
#' @seealso \code{\link{create_table}} to create a table, and 
#' \code{\link{create_plot}} to create a plot.  
#' @examples 
#' library(reporter)
#' library(magrittr)
#' 
#' # Create temp file path
#' tmp <- file.path(tempdir(), "iris.txt")
#' 
#' # Sample and sort iris data frame
#' dat <- iris[sample(1:150, 50), ]
#' dat <- dat[order(dat$Species), ]
#' 
#' # Create table
#' tbl <- create_table(dat) %>% 
#'   page_by(Species, "Species: ") %>%  
#'   define(Species, visible = FALSE)
#'   
#' # Create the report object
#' rpt <- create_report(tmp, orientation = "portrait") %>% 
#'   page_header("Sponsor: Iris Society", "Study: flowers") %>% 
#'   titles("Table 2.0", "IRIS Sample Report with Page By") %>% 
#'   add_content(tbl) %>% 
#'   page_footer(Sys.time(), right = "Page [pg] of [tpg]")
#' 
#' # Write the report to the file system
#' write_report(rpt)
#' 
#' # Write report to console
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' # Sponsor: Iris Society                                           Study: flowers
#' #                                  Table 2.0
#' #                       IRIS Sample Report with Page By
#' # 
#' #              Species: setosa
#' # 
#' #              Sepal.Length Sepal.Width Petal.Length Petal.Width
#' #              -------------------------------------------------
#' #                       5.4         3.9          1.7         0.4
#' #                       4.9         3.1          1.5         0.1
#' #                       4.8         3.1          1.6         0.2
#' #                       5.1         3.5          1.4         0.3
#' #                         5         3.5          1.6         0.6
#' #                         5         3.3          1.4         0.2
#' #                       4.4           3          1.3         0.2
#' #                       5.1         3.5          1.4         0.2
#' #                       5.4         3.4          1.5         0.4
#' #                       4.9         3.6          1.4         0.1
#' #                       4.6         3.1          1.5         0.2
#' #                       4.6         3.2          1.4         0.2
#' #                       5.1         3.3          1.7         0.5
#' # ...
#' # 2020-10-25 19:33:35                                                Page 1 of 3
#' #
#' # Sponsor: Iris Society                                           Study: flowers
#' #                                  Table 2.0
#' #                       IRIS Sample Report with Page By
#' # 
#' #              Species: versicolor
#' # 
#' #              Sepal.Length Sepal.Width Petal.Length Petal.Width
#' #              -------------------------------------------------
#' #                       4.9         2.4          3.3           1
#' #                       6.3         3.3          4.7         1.6
#' #                       6.1         2.8          4.7         1.2
#' #                         6         2.9          4.5         1.5
#' #                       6.7           3            5         1.7
#' #                       5.6           3          4.5         1.5
#' #                       5.8         2.7          4.1           1
#' #                       6.7         3.1          4.7         1.5
#' #                       6.1         2.9          4.7         1.4
#' #                         5           2          3.5           1
#' #                       5.9         3.2          4.8         1.8
#' #                       5.5         2.5            4         1.3
#' #                         7         3.2          4.7         1.4
#' #                       6.3         2.3          4.4         1.3
#' #                       6.1         2.8            4         1.3
#' #                         6         2.2            4           1
#' #                       5.5         2.6          4.4         1.2
#' #                         6         3.4          4.5         1.6
#' #                         5         2.3          3.3           1
#' #                       5.5         2.4          3.7           1
#' # ...
#' # 2020-10-25 19:33:35                                                Page 2 of 3
#' #
#' # Sponsor: Iris Society                                           Study: flowers
#' #                                  Table 2.0
#' #                       IRIS Sample Report with Page By
#' # 
#' #              Species: versicolor
#' # 
#' #              Sepal.Length Sepal.Width Petal.Length Petal.Width
#' #              -------------------------------------------------
#' #                       6.3         3.4          5.6         2.4
#' #                       7.9         3.8          6.4           2
#' #                       6.7         3.1          5.6         2.4
#' #                       6.2         2.8          4.8         1.8
#' #                       6.7         3.3          5.7         2.5
#' #                       6.2         3.4          5.4         2.3
#' #                       5.6         2.8          4.9           2
#' #                       7.7         3.8          6.7         2.2
#' #                       7.7         2.6          6.9         2.3
#' #                       6.9         3.1          5.4         2.1
#' #                       6.5         3.2          5.1           2
#' #                       6.1         2.6          5.6         1.4
#' #                       5.7         2.5            5           2
#' #                       6.5           3          5.8         2.2
#' #                       6.3         2.8          5.1         1.5
#' #                       7.6           3          6.6         2.1
#' #                       6.3         2.5            5         1.9
#' # ...
#' # 2020-10-25 19:33:35                                                Page 3 of 3 
#' @export
page_by <- function(x, var, label = NULL, align = "left",
                    blank_row = "below", borders = "none", format = NULL,
                    bold = FALSE) {
  
  
  # Create page by structure
  pb <- structure(list(), class = c("page_by", "list"))

  if (!align %in% c("left", "right", "center", "centre" )){
    stop(paste("align value is invalid. Valid values are 'left', 'right',",
               "'center', or 'centre'."))
  }
  
  if (!blank_row %in% c("above", "below", "both", "none" )){
    stop(paste("blank_row value is invalid. Valid values are 'above', 'below',",
               "'both', or 'none'."))
  }
  
  
  if (!all(borders %in% c("top", "bottom", "left", "right", "all", "outside", "none")))
    stop(paste("Borders parameter invalid.  Valid values are", 
               "'top', 'bottom', 'left', 'right', 'all', 'outside', or 'none'."))
  
  if (!bold %in% c(TRUE, FALSE, "label", "value")) {
    stop("Bold parameter invalid. Valid values are TRUE, FALSE, 'label', or 'value'.")
  }
  
  var_c <- as.character(substitute(var, env = environment()))
  
  pb$var <- var_c
  if (is.null(label))
    pb$label <- paste0(var_c, ": ")
  else
    pb$label <- label
  
  pb$align <- align
  pb$blank_row <- blank_row
  pb$borders <- borders
  pb$format <- format
  pb$bold <- bold

  x$page_by <- pb
  
  return(x)
  
}

#' @title
#' Insert an image into the page header
#'
#' @description
#' The \code{header_image} function inserts an image into the page header. The function
#' can be used to add a logo or other graphical element to the report.
#' The image will then appear at the top of each page.
#' @details
#' The function assumes 
#' the existence of a page header inserted by the \code{\link{page_header}} function. 
#' If there is no page header, an error will be generated during the rendering
#' process in \code{\link{write_report}}. If there is text
#' in the same cell as the image, the text will be disregarded. You may add text 
#' to other cells in the header.
#' 
#' The "align" parameter determines which page header cell the image will be inserted.
#' That is, if you align an image "left", it will be inserted into the left
#' column of the page header. Options are "left", "right", and "center".  To
#' insert images in multiple cells, you may use multiple calls to 
#' \code{image_header}.
#' 
#' The page header cell width will take priority over the image width.  That is,
#' if you set the image width larger than the cell width, the image width
#' will be reduced to fit the cell.  To make a wider image, increase the cell
#' width on the page header.  
#' 
#' Note that page header cells can be removed
#' entirely by setting the width to zero.  This feature can give you more room
#' for a wide image. For instance, to create a banner image over the entire report, 
#' align the image "center" and set the left and right page header column 
#' widths to zero.
#' 
#' See the \code{\link{page_header}} function for more information on
#' setting header column widths.  For a demonstration of 
#' a header image insertion, see \code{vignette("reporter-hfimage")}.
#' @param x The report object.
#' @param image_path The path of image file. Function accepts JPG images 
#' for RTF, DOCX, PDF, and HTML output types. Function also accepts 
#' PNG images for RTF, DOCX, and HTML output types.
#' @param height Desired height of the image, in the report units of measure.
#' @param width Desired width of the image, in the report units of measure.
#' @param align Alignment of the image. Valid values are "left", "right", 
#' "center", or "centre". The default value is "left".
#' @return The modified report specification.
#' @family report
#' @export
header_image <- function(x, image_path, height, width, align = "left") {
  
  
  if (!any(class(image_path) %in% c("character"))) {
    stop("image_path object must an image file path.")
  }
  
  if (!align %in% c("left", "right", "center", "centre")) {
    stop("align must be left or right.")
  }
  
  header_image <- list()
  header_image$image_path <- image_path
  header_image$height <- height
  header_image$width <- width
  header_image$align <- align
  
  if (align == "left") {
    x$header_image_left <- header_image
  } else if (align == "right") {
    x$header_image_right <- header_image
  } else {
    x$header_image_center <- header_image
  }
  
  return(x)
}

#' @title
#' Insert an image into the page footer
#'
#' @description
#' The \code{footer_image} function inserts an image into the page footer. The function
#' can be used to add a logo or other graphical element to the report.
#' The image will then appear at the bottom of each page.
#' @details
#' The function assumes 
#' the existence of a page footer inserted by the \code{\link{page_footer}} function. 
#' If there is no page footer, an error will be generated during the rendering
#' process in \code{\link{write_report}}. If there is text
#' in the same cell as the image, the text will be disregarded. You may add text 
#' to other cells in the footer.
#' 
#' The "align" parameter determines which page footer cell the image will be inserted.
#' That is, if you align an image "left", it will be inserted into the left
#' column of the page footer. Options are "left", "right", and "center".  To
#' insert images in multiple cells, you may use multiple calls to 
#' \code{image_footer}.
#' 
#' The page footer cell width will take priority over the image width.  That is,
#' if you set the image width larger than the cell width, the image width
#' will be reduced to fit the cell.  To make a wider image, increase the cell
#' width on the page footer.  
#' 
#' Note that page footer cells can be removed
#' entirely by setting the width to zero.  This feature can give you more room
#' for a wide image. For instance, to create a banner image under the entire report, 
#' align the image "center" and set the left and right page header column 
#' widths to zero.
#' 
#' See the \code{\link{page_footer}} function for more information on
#' setting footer column widths.  For a demonstration of 
#' a header image insertion, see \code{vignette("reporter-hfimage")}.
#' 
#' @param x The report object to insert the image into.
#' @param image_path The path of image file. Function accepts JPG images 
#' for RTF, DOCX, PDF, and HTML output types. Function also accepts 
#' PNG images for RTF, DOCX, and HTML output types.
#' @param height Desired height of the image, in the report units of measure.
#' @param width Desired width of the image, in the report units of measure.
#' @param align Alignment of the image in the page footer. Valid values are 
#' "left", "right", "center", or "centre". The default value is "left".
#' @return The modified report specification.
#' @family report
#' @export
footer_image <- function(x, image_path, height, width, align = "left") {
  if (!any(class(image_path) %in% c("character"))) {
    stop("image_path object must an image file path.")
  }
  
  if (!align %in% c("left", "right", "center", "centre")) {
    stop("align must be left, right, center, or centre.")
  }
  
  footer_image <- list()
  footer_image$image_path <- image_path
  footer_image$height <- height
  footer_image$width <- width
  footer_image$align <- align
  
  if (align == "left") {
    x$footer_image_left <- footer_image
  } else if (align == "right") {
    x$footer_image_right <- footer_image
  } else {
    x$footer_image_center <- footer_image
  }
  
  return(x)
}

# Functions ---------------------------------------------------------------


#' @title Add content to a report
#' @description This function adds an object to the report content list. A 
#' report will accept multiple pieces of content.  The \code{add_content} 
#' function also controls overall alignment of the content on the page, and
#' whether there is a page break before or after.
#' @details 
#' The \code{add_content} function adds a piece of content to a report. For a 
#' text report, valid objects are a table or text object.  For an RTF, PDF,  
#' HTML, or DOCX report, valid objects are a table, text, or plot object.  See 
#' \code{\link{create_table}}, \code{\link{create_text}}, or 
#' \code{\link{create_plot}} for further 
#' information on how to create content objects.  
#' 
#' Content will be
#' appended to the report in the order it is added.  By default, a page break
#' is added after the content.  You can stack two pieces of content together
#' closely by setting the \code{page_break} parameter to FALSE, and 
#' the \code{blank_row} parameter to "none".
#' 
#' @param x The report_spec to append content to.
#' @param object The object to append.
#' @param page_break Whether to add a page break after the object. 
#' Valid values are TRUE or FALSE.  You can manipulate the \code{page_break} 
#' parameter to add multiple objects to the same page.  
#' @param align How to align the content.  Valid values are 'left', 'right',
#' 'center', and 'centre'.
#' @param blank_row Whether to put a blank row above or below the content.
#' Valid values are 'above', 'below', 'both', or 'none'.
#' @return The modified report_spec.
#' @family report
#' @seealso \code{\link{create_table}}, \code{\link{create_text}}, and
#' \code{\link{create_plot}} to create content for a report. 
#' @examples
#' library(reporter)
#' library(magrittr)
#' 
#' # Create temp file path
#' tmp <- file.path(tempdir(), "mtcars.txt")
#' 
#' # Create first table 
#' tbl1 <- create_table(mtcars[1:5, 1:6]) %>% 
#'   column_defaults(width = .5) 
#' 
#' # Create second table
#' tbl2 <- create_table(mtcars[6:10, 1:6], headerless=TRUE) %>% 
#'   column_defaults(width = .5) 
#' 
#' # Create the report object
#' rpt <- create_report(tmp) %>%
#'   titles("MTCARS Sample Data", align = "left") %>%
#'   add_content(tbl1, page_break = FALSE, align = "left", blank_row = "none") %>%
#'   add_content(tbl2, page_break = FALSE, align = "left") %>%
#'   add_content(create_text("* NOTE: Above table is actually two tables stacked."))
#' 
#' # Write the report to the file system
#' res <- write_report(rpt)
#' 
#' # Write report to console
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' # MTCARS Sample Data
#' # 
#' #    mpg    cyl   disp     hp   drat     wt
#' # -----------------------------------------
#' #     21      6    160    110    3.9   2.62
#' #     21      6    160    110    3.9  2.875
#' #   22.8      4    108     93   3.85   2.32
#' #   21.4      6    258    110   3.08  3.215
#' #   18.7      8    360    175   3.15   3.44
#' #   18.1      6    225    105   2.76   3.46
#' #   14.3      8    360    245   3.21   3.57  
#' #   24.4      4  146.7     62   3.69   3.19
#' #   22.8      4  140.8     95   3.92   3.15
#' #   19.2      6  167.6    123   3.92   3.44
#' # 
#' # * NOTE: Above table is actually two tables stacked.
#' @export
add_content <- function(x, object, page_break=TRUE, align = "center",
                        blank_row = "below") {
  
  if (!page_break %in% c(TRUE, FALSE)) {
   stop(paste("Page break value invalid.",
              "Valid values are TRUE and FALSE."))
  }
  
  cont <- structure(list(), class = c("report_content", "list"))
  
  cont$object <- object
  cont$page_break <- page_break
  cont$blank_row <- blank_row
  cont$align <- align
  cont$pages <- list()

  # Add object to the content list
  x$content[[length(x$content) + 1]] <- cont


  return(x)
}


#' @title
#' Write a report to the file system
#' @description
#' This function writes a report_spec object to the file system, using the
#' specifications provided in the object.
#' @details 
#' The function renders the report in the requested format, and writes it
#' to the location specified in the report \code{file_path} parameter.  Attempts
#' to write an object that is not of class "report_spec" will generate an error.
#' 
#' The \code{write_report} function is a driver for very complex set of 
#' rendering functions. The rendering functions
#' perform most of the advanced functionality of the \strong{reporter} package:
#' generating spanning headers, page wrapping and breaking, creating stub 
#' columns, etc.  When things go wrong, they will usually go wrong during this
#' function call.  For that reason, although this function can be part of 
#' the pipeline that creates the report object, it is best to call 
#' \code{write_report} independently, to help isolate any issues from the 
#' report definition procedure.
#' 
#' @param x The report object to write.
#' @param file_path The file name and path to write the report to.  If supplied,
#' this parameter overrides the \code{file_path} parameter on the 
#' \code{create_report} function. Default is NULL.
#' @param output_type The output file type.  This parameter will override
#' the \code{output_type} on the \code{create_report} function.  This 
#' parameter can be used to output the same report object to 
#' multiple output types. Default value is NULL, meaning it will not override
#' the \code{create_report} value.  Valid values are 'TXT', 'RTF', 'DOCX', 
#' 'PDF' and 'HTML'.
#' @param preview Whether to write the entire report, or a report preview.
#' A report preview is a subset of pages of the report.  The default value is 
#' NULL, meaning the entire report will be written.  You may also pass 
#' a number of pages to write.  For example, passing the number 1 will print
#' the first page, while passing a 5 will print the first five pages.
#' @param log Whether the function should log itself automatically, if
#' the \strong{logr} package is available.  This parameter is used internally.
#' Default is TRUE.
#' @return The report spec, with settings modified during rendering.  These 
#' modified settings can sometimes be useful for documentation, and for 
#' debugging issues with the procedure.
#' @family report
#' @examples 
#' library(reporter)
#' library(fmtr)
#' library(magrittr)
#' 
#' # Create temp file path
#' tmp <- file.path(tempdir(), "beaver2.txt")
#' 
#' # Take Sample of Data
#' dat <- beaver2[sample(1:100, 15), ]
#' 
#' # Create format for active variable
#' fmt <- value(condition(x == 0, "No"),
#'              condition(x == 1, "Yes"))
#' 
#' # Create the table
#' tbl <- create_table(dat) %>% 
#'   titles("Table 1.0", "BEAVERS Sample Report") %>% 
#'   column_defaults(width = .75) %>% 
#'   define(day, label = "Day", format = "Day %s") %>% 
#'   define(time, label = "Time") %>% 
#'   define(temp, label = "Temperature", width = 1, format = "%.1f") %>% 
#'   define(activ,label = "Active", format = fmt) %>% 
#'   footnotes("* NOTE: Data on beaver habits")
#' 
#' # Create the report object
#' rpt <- create_report(tmp) %>% 
#'   add_content(tbl, align = "left") 
#' 
#' # Write the report to the file system
#' res <- write_report(rpt)
#' 
#' # Write the modified report object to the console
#' print(res)
#' 
#' # Write the report to console
#' writeLines(readLines(tmp, encoding = "UTF-8"))
#' 
#' #                 Table 1.0
#' #           BEAVERS Sample Report
#' # 
#' #      Day      Time  Temperature    Active
#' # -----------------------------------------
#' #  Day 307      1020         37.2        No
#' #  Day 307      1030         37.2        No
#' #  Day 307       940         36.7        No
#' #  Day 307      1340         37.1        No
#' #  Day 307      1410         37.2        No
#' #  Day 307      1400         37.1        No
#' #  Day 307      1130         36.9        No
#' #  Day 307      1140         37.0        No
#' #  Day 307      1120         37.0        No
#' #  Day 307      1000         37.1        No
#' #  Day 307      1250         37.0        No
#' #  Day 307      2100         37.9       Yes
#' #  Day 307      1210         37.0        No
#' #  Day 307      1740         38.0       Yes
#' #  Day 308       130         37.8       Yes
#' # 
#' # * NOTE: Data on beaver habits
#' @export
write_report <- function(x, file_path = NULL, 
                         output_type = NULL, preview = NULL, 
                         log = TRUE) {
  
  
  if (!"report_spec" %in% class(x)) {
    
    stop(paste0("report object missing or invalid."))
  }
  
  if (!is.null(file_path)) {
    
    x$file_path <- file_path

  }
  
  # Trap missing or invalid output_type parameter
  if (!is.null(output_type)) {
    if (!toupper(output_type) %in% c("TXT", "PDF", "RTF", "HTML", "DOCX")) {
      
      stop(paste0("output_type parameter on create_report() ",
                  "function is invalid: '", output_type,
                  "'\n\tValid values are: 'TXT', 'PDF', 'RTF', 'HTML', 'DOCX'."))
    }
    x$output_type <- toupper(output_type)

    x <- options_fixed(x, editor = x$editor, 
                       cpuom = x$cpuom, lpuom = x$lpuom,
                       min_margin = x$min_margin, blank_margins = x$blank_margins,
                       font_size = x$font_size, line_size = x$user_line_size, 
                       line_count = x$user_line_count,
                       uchar = x$uchar)
  } 
  
  if (tolower(x$paper_size) == "none" & toupper(x$output_type) != "HTML") {
    stop("Paper size of 'none' is only valid on HTML reports.") 
  }
                                         
                                        

  if (nchar(x$file_path) > 0 & length(getExtension(x$file_path)) == 0) {
      x$modified_path <- paste0(x$file_path, ".", tolower(x$output_type))
      x$path <- paste0(x$file_path, ".", tolower(x$output_type))
  } else {
    x$modified_path <- x$file_path
    x$path <- x$file_path
  }
  
  #print(x$modified_path)
  
  if (!is.null(preview)) {
    if (is.numeric(preview)) {
      if (preview > 0)
        x$preview <- preview
      else
        stop("Preview value must be greater than zero.")
    } else
      stop("Preview value must be a number")
  }
  
  # Trap missing or invalid output_type parameter
  if (x$modified_path == "") {
    
    stop(paste0("report file_path missing or invalid."))
  } else {
    
    if (file.exists(x$modified_path))
      file.remove(x$modified_path)
    
    if (!dir.exists(dirname(x$modified_path)))
      dir.create(dirname(x$modified_path))
    
  }
  
  # If there are header titles or footer footnotes, 
  # reassign to regular title and footnotes lists if not RTF.
  # This is the lowest risk way to deal with it.
  if (!x$output_type %in% c("RTF", "DOCX") | x$font == "fixed") {
    if (length(x$header_titles) > 0) {
      x$titles <- append(x$header_titles, x$titles)
      x$header_titles <- NULL
    }
    
    if (length(x$footer_footnotes) > 0) {
      x$footnotes <- append(x$footer_footnotes, x$footnotes)
      x$footer_footnotes <- NULL
    }
  }

  ret <- ""

  if (toupper(x$output_type) == "TXT") {
    
    ret <- write_report_text(x)
    
  } else if (toupper(x$output_type) == "RTF") {
    
    if (tolower(x$font) == "fixed")
      ret <- write_report_rtf(x)
    else 
      ret <- write_report_rtf2(x)
  
  } else if (toupper(x$output_type) == "PDF") {
    
    if (tolower(x$font) == "fixed")
      ret <- write_report_pdf(x)
    else 
      ret <- write_report_pdf2(x)

  } else if (toupper(x$output_type) == "HTML") {
  
    ret <- write_report_html(x)
    
  } else if (toupper(x$output_type) == "DOCX") {
    
    ret <- write_report_docx(x)
    
  } else {
   stop(paste("Output type currently not supported:", x$output_type))
  }

  if (log)
    log_logr(ret)
  
  return(ret)
}


#' @title Prints the report specification
#' @description A function to print the report specification.
#' The \strong{print} function will print the report spec in summary 
#' form by default.  To print in list form, set the \code{verbose} parameter
#' to TRUE.
#' @param x The report spec.
#' @param ... Additional parameters to pass to the underlying print function.
#' @param verbose Whether to print the report object in verbose (list) form
#' or summary form.  The default is FALSE.
#' @seealso 
#' \code{\link{create_report}} function to create a report specification.
#' @return The report spec, invisibly.
#' @family report
#' @examples 
#' library(reporter)
#' library(magrittr)
#' 
#' # Create temp file path
#' tmp <- file.path(tempdir(), "mtcars.txt")
#' 
#' # Create the table
#' tbl <- create_table(mtcars) %>% 
#'   titles("Table 1.0", "MTCARS Sample Report") %>% 
#'   footnotes("* NOTE: Data from 1974")
#' 
#' # Create the report object
#' rpt <- create_report(tmp) %>% 
#'   add_content(tbl, align = "left") 
#' 
#' # Write the report to the file system
#' res <- write_report(rpt)
#' 
#' # Write the modified report object to the console
#' print(res)
#' 
#' # # A report specification: 1 pages
#' # - file_path: 'C:/Users/User/AppData/Local/Temp/RtmpWQybXs/mtcars.txt'
#' # - output_type: TXT
#' # - units: inches
#' # - orientation: landscape
#' # - line size/count: 108/45
#' # - content: 
#' # # A table specification:
#' # - data: data.frame 'mtcars' 32 rows 11 cols
#' # - show_cols: all
#' # - use_attributes: all
#' # - title 1: 'Table 1.0'
#' # - title 2: 'MTCARS Sample Report'
#' # - footnote 1: '* NOTE: Data from 1974'
#' @import crayon
#' @export
print.report_spec <- function(x, ..., verbose = FALSE){
  
  
  if (verbose == TRUE) {
    
    # If verbose mode is requested, print in list form
    for (nm in names(x)) {
      
      cat("$", nm, "\n", sep = "")
      if (nm == "content") {
        
        for (i in seq_along(x[[nm]]))
          print(x[[nm]][[i]], ..., verbose = TRUE)
      }
      else  {
        
        print(x[[nm]], ...)
      }
      cat("\n")
    }
    
    
  } else {
    
    grey60 <- make_style(grey60 = "#999999")

    # Print header
    cat(grey60("# A report specification: "))
    if (!is.null(x$pages)) {
      if (!is.null(x$preview)) {
        if (x$preview > 0)
          cat(grey60(paste0(as.character(x$pages), " pages"), red(" PREVIEW")))
        else
          cat(grey60(paste0(as.character(x$pages), " pages")))
      } else {
        
        cat(grey60(paste0(as.character(x$pages), " pages")))
      }
    }
    
    # Print key attributes
    cat("\n")
    if (is.null(x$modified_path)) {
      cat(paste0("- file_path: '", x$file_path, "'\n"))
    } else if (x$modified_path == "") {
      cat(paste0("- file_path: '", x$file_path, "'\n"))
    } else {
      cat(paste0("- file_path: '", x$modified_path, "'\n"))
    }
    cat(paste0("- output_type: ", x$output_type, "\n"))
    cat(paste0("- units: ", x$units, "\n"))
    cat(paste0("- orientation: ", x$orientation, "\n"))
    cat(paste0("- margins: top ", x$margin_top, " bottom ", 
               x$margin_bottom, " left ", x$margin_left, 
               " right ", x$margin_right, "\n"))
    if (!is.null(x$line_size)) 
      cat(paste0("- line size/count: ", x$line_size, "/", x$line_count, "\n"))
    
    # if (!is.null(x$column_widths)) {
    #   cat("- column widths: \n")
    #   cat(x$column_widths)
    #   
    # }
    if (!is.null(x$page_by)) {
      cat(paste0("- page by: ", x$page_by$var, "\n"))
      
    }
    
    # Print page header
    if (!is.null(x$page_header_left) |
        !is.null(x$page_header_right)) {
      cat("- page_header:")
      if (!is.null(x$page_header_left)) {
        cat(" left=")
        cat(paste0(x$page_header_left, collapse = ", "))
      }
      if (!is.null(x$page_header_right)) {
        cat(" right=")
        cat(paste0(x$page_header_right, collapse = ", "))
      }
      cat("\n")
    }
    
    print_title_header(x$title_hdr)

    
    print_titles(x$titles)
    
    print_footnotes(x$footnotes)
    

    
    # Print page footer
    if (!is.null(x$page_footer_left) | 
        !is.null(x$page_footer_center) |
        !is.null(x$page_footer_right)) {
      cat("- page_footer:")
      if (!is.null(x$page_footer_left)) {
        cat(" left=")
        cat(paste0(x$page_footer_left, collapse = ", "))
      }
      if (!is.null(x$page_footer_center)) {
        cat(" center=")
        cat(paste0(x$page_footer_center, collapse = ", "))
      }
      if (!is.null(x$page_footer_right)) {
        cat(" right=")
        cat(paste0(x$page_footer_right, collapse = ", "))
      }
      cat("\n")
    }
    if (!is.null(x$content)) {
      cat("- content: \n")
      for (i in seq_along(x$content)) {
        
        print(x$content[[i]]$object)
        
      }
    }
    
  }
  
  invisible(x)
}


# Write Registration File -------------------------------------------------

#' @title 
#' Create a registration file 
#' @description 
#' This function will create a registration file to help determine
#' the correct \code{cpuom} and \code{lpuom} for your editor/printer.  
#' @details
#' The \code{cpi} and \code{lpi} are 
#' used in \code{output_type = "TXT"} to determine available space on
#' the page. The registration file can help determine the correct settings
#' for the target text editor and printer.  Failure to set the 
#' correct characters per 
#' unit of measure (cpuom) and lines per unit of measure (lpuom) may result
#' in misalignment of content on the page when printing text output.
#' 
#' @section How to Use the Registration File:
#' To use the registration file, first decide the units of measure you
#' wish to use, inches or centimeters.  Next, create the registration file
#' by calling the \code{write_registration_file} function.  Then print the 
#' registration file.
#' 
#' Once the registration file is printed, take a ruler and measure both the 
#' horizontal and vertical registration lines from zero to 60 in the desired
#' units of measure.  For example, if your units of measure is 'inches', measure
#' the registration lines in inches.    
#' 
#' Record the 
#' distance measured in each direction.  For each direction, divide 60 by the 
#' distance measured, and round to three decimal places.  The horizontal 
#' result is the characters per unit of measure (cpuom).  The vertical result
#' is the lines per unit of measure (lpuom). To get an accurate printing 
#' of text reports, 
#' assign these values to the \code{cpuom} and \code{lpuom} parameters
#' on the \code{\link{options_fixed}} function.
#' 
#' For best results, test the calculated values by printing some reports and
#' checking for undesired page breaks or wrapped lines.  If necessary, adjust
#' the calculated cpuom and lpuom values until all content stays within 
#' the available space without wrapping or breaking.
#' 
#' @param file_path The full or relative file name and path to create the 
#' registration file.
#' @examples 
#' library(reporter)
#' 
#' # Create temp file path
#' tmp <- file.path(tempdir(), "reg.txt")
#' 
#' # Create the registration file 
#' write_registration_file(tmp)
#' 
#' # Write registration file to the console
#' writeLines(readLines(tmp))
#' 
#' # 0--------+---------+---------+---------+---------+---------+
#' # -       10        20        30        40        50        60
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # + 10
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # + 20
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # + 30
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # + 40
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # + 50
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # -
#' # + 60
#' @export
write_registration_file <- function(file_path) {
  
  
  f <- file(file_path, open="w")
  
  ln1 <- "0--------+---------+---------+---------+---------+---------+"
  
  writeLines(ln1, con = f)
  
  ln2 <- "-       10        20        30        40        50        60"
  
  writeLines(ln2, con = f)
  
  ln3 <- c("-", "-", "-", "-", "-", "-", "-", "+ 10")
  
  writeLines(ln3, con = f)
  
  ln4 <- c("-", "-","-", "-", "-", "-", "-", "-", "-", "+ 20")
  
  writeLines(ln4, con = f)
  
  ln5 <- c("-", "-","-", "-", "-", "-", "-", "-", "-", "+ 30")
  
  writeLines(ln5, con = f)
  
  ln6 <- c("-", "-","-", "-", "-", "-", "-", "-", "-", "+ 40")
  
  writeLines(ln6, con = f)
  
  ln7 <- c("-", "-","-", "-", "-", "-", "-", "-", "-", "+ 50")
  
  writeLines(ln7, con = f)
  
  ln8 <- c("-", "-","-", "-", "-", "-", "-", "-", "-", "+ 60")
  
  writeLines(ln8, con = f)
  
  close(f)
}

context("Create Table HTML Tests")

test_that("get_table_body_html works as expected.", {
  
  fp <- ""
  
  dat <- mtcars[1:10, 1:5]
  
  tbl <- create_table(dat) %>%
    define("mpg", indent = 0.25)
  
  rpt <- create_report(fp, output_type = "HTML", font = "Arial",
                       font_size = 12) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")
  
  rpt <- page_setup_html(rpt)
  
  wdth <- rep(1, 5)
  names(wdth) <- names(dat)
  
  algns <- rep("center", 5)
  names(algns) <- names(dat)
  
  
  res <- get_table_body_html(rpt, dat, wdth, algns,  "center", "none", 
                             styles = list(), ts = tbl)
  
  res  
  
  # Check indenting code is added
  expect_true(any(grepl("<td class=\"tc tdc \" style=\"padding-left: 0.25in;\">21</td>", res$html)))
  
  expect_equal(length(res$html), 10)
  expect_equal(res$lines, 10)
  
})

test_that("get_table_header_html works as expected.", {
  
  fp <- ""
  
  dat <- mtcars[1:10, 1:5]
  
  tbl <- create_table(dat) 
  
  rpt <- create_report(fp, output_type = "HTML", font = "Arial",
                       font_size = 12) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")
  
  rpt <- page_setup_html(rpt)
  
  wdth <- rep(1, 5)
  names(wdth) <- names(dat)
  
  algns <- rep("center", 5)
  names(algns) <- names(dat)
  
  nms <- names(dat)
  names(nms) <- nms
  nms[1] <- "Miles Per Gallon"
  nms[2] <- "Wrap One Two"
  
  pi <- list(keys = names(dat), col_width = wdth, label = nms,
             label_align = algns, table_align = "center")
  
  res <- get_table_header_html(rpt, tbl, pi)
  
  res  
  
  expect_equal(length(res$html), 1)
  expect_equal(res$lines, 2)
  
  
})

test_that("get_content_offsets_html works as expected.", {
  
  fp <- ""
  
  dat <- mtcars[1:10, 1:5]
  
  tbl <- create_table(dat) %>% 
    titles("Table 1.0", "My Nice Table") %>% 
    footnotes("Here is a footnote")
  
  rpt <- create_report(fp, output_type = "HTML", font = "Arial",
                       font_size = 12) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")
  
  rpt <- page_setup_html(rpt)
  
  wdth <- rep(1, 5)
  names(wdth) <- names(dat)
  
  algns <- rep("center", 5)
  names(algns) <- names(dat)
  
  nms <- names(dat)
  names(nms) <- nms
  nms[1] <- "Miles/Gallon"
  
  pi <- list(keys = names(dat), col_width = wdth, label = nms,
             label_align = algns, table_align = "center")
  
  res <- get_content_offsets_html(rpt, tbl, pi, "both")
  res
  
  expect_equal(res$lines[["upper"]], 4)
  expect_equal(res$lines[["lower"]], 2)
  expect_equal(res$lines[["blank_upper"]], 1)
  expect_equal(res$lines[["blank_lower"]], 1)
  
})

test_that("create_table_html works as expected.", {
  
  fp <- ""
  
  dat <- mtcars[1:10, 1:5]
  
  tbl <- create_table(dat) %>% 
    titles("Table 1.0", "My Nice Table") %>% 
    footnotes("Here is a footnote")
  
  rpt <- create_report(fp, output_type = "HTML", font = "Arial",
                       font_size = 12) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")
  
  rpt <- page_setup_html(rpt)
  
  wdth <- rep(1, 5)
  names(wdth) <- names(dat)
  
  algns <- rep("center", 5)
  names(algns) <- names(dat)
  
  nms <- names(dat)
  names(nms) <- nms
  nms[1] <- "Miles/Gallon"
  
  pi <- list(keys = names(dat), col_width = wdth, label = nms,
             label_align = algns, table_align = "center", data = dat)
  
  res <- create_table_html(rpt, tbl, pi, "below", FALSE, 0, styles = list())
  res
  
  # Doesn't seem right.  Come back to this.
  expect_equal(length(res$html), 16)
  expect_equal(res$lines, 17) 
  
})

test_that("get_spanning_header_html works as expected.", {
  
  
  fp <- ""

  dat <- mtcars[1:15, ]

  tbl <- create_table(dat, borders = "none") %>%
    spanning_header(mpg, disp, "Span 1", n = 10) %>%
    spanning_header(hp, wt, "Span 2") %>%
    spanning_header(qsec, vs, "Span 3") %>%
    spanning_header(drat, gear, "Super span", level = 2)


  rpt <- create_report(fp, output_type = "HTML", font = "Arial",
                       font_size = 10, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(tbl) %>%
    footnotes("My footnote 1", "My footnote 2") %>%
    page_footer("Left1", "Center1", "Right1")

  rpt <- page_setup_html(rpt)

  cols <- rep(.6, ncol(dat))
  names(cols) <- names(dat)

  pi <- list(keys = names(dat), col_width = cols, table_align = "center")

  res <- get_spanning_header_html(rpt, tbl, pi)
  res

  expect_equal(length(res), 2)
  expect_equal(length(res$html), 2)
  expect_equal(res$lines, 3) 
})


# Haven't quite decided what to do with this function in HTML
# For now, just return the footnote and any blank lines above or below
test_that("get_page_footnotes_html works as expected.", {
  
  
  tbl1 <- create_table(iris)
  
  rpt1 <- create_report("", font = "Courier") %>%
    titles("IRIS Data Frame") %>%
    add_content(tbl1) %>% 
    footnotes("Here is a footnote", valign = "top")
  
  rpt1 <- page_setup_html(rpt1)
  
  res1 <- get_page_footnotes_html(rpt1, tbl1, 6,
                                 0, 25, TRUE, "below", "center")
  res1
  expect_equal(res1$lines, 2)
  
  tbl2 <- create_table(iris)
  
  rpt2 <- create_report("", font = "Arial") %>%
    titles("IRIS Data Frame") %>%
    add_content(tbl1) %>% 
    footnotes("Here is a footnote", valign = "bottom")
  
  rpt2 <- page_setup_html(rpt2)
  
  res2 <- get_page_footnotes_html(rpt2, tbl2, 6,
                                 0, 25, TRUE, "below", "center")  
  res2
  expect_equal(res2$lines, 17)
  
  
  tbl3 <- create_table(iris)  %>% 
    footnotes("Here is a footnote", valign = "bottom")
  
  rpt3 <- create_report("", font = "Times") %>%
    titles("IRIS Data Frame") %>%
    add_content(tbl1)
  
  rpt3 <- page_setup_html(rpt3)
  
  res3 <- get_page_footnotes_html(rpt3, tbl3, 6,
                                 0, 25, TRUE, "below", "center")  
  
  res3
  expect_equal(res3$lines, 16)
  
  tbl4 <- create_table(iris)  %>% 
    footnotes("Here is a footnote", valign = "top")
  
  rpt4 <- create_report("", font = "Times") %>%
    titles("IRIS Data Frame") %>%
    add_content(tbl1)
  
  rpt4 <- page_setup_html(rpt4)
  
  res4 <- get_page_footnotes_html(rpt4, tbl4, 26,
                                 0, 25, TRUE, "below", "center")  
  
  expect_equal(res4$lines, 2)
  
})

context("Page Template HTML Tests")



test_that("get_titles_html function works as expected.", {
  
  
  rpt <- create_report("", output_type = "HTML", font = "Arial", 
                       font_size = 12) %>%
    titles("Hello", "here is \n something", blank_row = "below") %>%
    footnotes("Goodbye", blank_row = "below")
  
  rpt <- page_setup_html(rpt)
  
  t <- get_titles_html(rpt$titles, 6, rpt)
  t
  expect_equal(length(t$html), 1)
  expect_equal(t$lines, 4)

})

test_that("get_titles_html function works as expected with font_size.", {
  
  #Different title font size will change title lines
  rpt <- create_report("", output_type = "HTML", font = "Arial", 
                       font_size = 9) %>%
    titles("a a a a a a a a a",
           width = 1)
  
  rpt <- page_setup_html(rpt)
  t9 <- get_titles_html(rpt$titles, 1, rpt)
  expect_equal(length(t9$html), 1)
  expect_equal(t9$lines, 2)
  
  rpt <- create_report("", output_type = "HTML", font = "Arial", 
                       font_size = 9) %>%
    titles("a a a a a a a a a",
           font_size = 14,
           width = 1)
  
  rpt <- page_setup_html(rpt)
  t14 <- get_titles_html(rpt$titles, 1, rpt)
  expect_equal(length(t14$html), 1)
  expect_equal(t14$lines, 3)
})



test_that("get_footnotes_html function works as expected.", {
  
  
  rpt <- create_report("", output_type = "HTML", font = "Arial", 
                       font_size = 12) %>%
    titles("Hello", blank_row = "below") %>%
    footnotes("Goodbye", blank_row = "below")
  
  rpt <- page_setup_html(rpt)
  
  t <- get_footnotes_html(rpt$footnotes, 6, rpt)
  t
  expect_equal(t$html,
               paste0("<table ",
                      "style=\"width:9in;text-align: left;\">", 
                      "\n<tr><td style=\"width:9in;text-align: left;", 
                      "vertical-align:text-top;\">Goodbye</td>\n",
                      "</tr>\n<tr><td colspan=\"1\">&nbsp;</td></tr>\n</table>"))
  expect_equal(t$lines, 2)
  
})

test_that("get_footnotes_html function works as expected with font_size.", {
  
  rpt <- create_report("", output_type = "HTML", font = "Arial", 
                       font_size = 12) %>%
    footnotes("a a a a a a a a a",
              width = 1)
  
  rpt <- page_setup_html(rpt)
  f12 <- rpt$page_template$footnotes
  
  expect_equal(f12$lines, 3)
  expect_equal(f12$html,
               paste0("<table style=\"width:1in;text-align: left;",
                      "\">\n<tr><td colspan=\"1\">&nbsp;",
                      "</td></tr>\n<tr><td style=\"width:1in;",
                      "text-align: left;vertical-align:text-top;",
                      "\">a&nbsp;a&nbsp;a&nbsp;a&nbsp;a&nbsp;a&nbsp;",
                      "a<br>a&nbsp;a</td>\n</tr>\n</table>")
               )
  
  rpt <- create_report("", output_type = "HTML", font = "Arial", 
                       font_size = 12) %>%
    footnotes("a a a a a a a a a",
              width = 1, font_size = 8)
  
  rpt <- page_setup_html(rpt)
  f8 <- rpt$page_template$footnotes
  
  expect_equal(f8$lines, 2)
  expect_equal(f8$html,
               paste0("<table style=\"width:1in;text-align: left;",
                      "\">\n<tr><td colspan=\"1\">&nbsp;",
                      "</td></tr>\n<tr><td style=\"width:1in;",
                      "font-size:8pt;text-align: left;vertical-align:text-top;",
                      "\">a&nbsp;a&nbsp;a&nbsp;a&nbsp;a&nbsp;a&nbsp;a&nbsp;",
                      "a&nbsp;a</td>\n</tr>\n</table>")
               )
  
})


test_that("get_title_header_html function works as expected.", {

  rpt2 <- create_report("", font = "Arial", font_size = 12) %>%
    title_header("Hello", right = paste("Right here is something",
    "really long that will wrap and wrap and wrap and wrap keep wrapping"))

  rpt2 <- page_setup_html(rpt2)

  th <- get_title_header_html(rpt2$title_hdr, 6, rpt2)
  th
  expect_equal(nchar(th$html) > 1, TRUE)
  expect_equal(th$lines, 4) 

})


test_that("get_page_header_html works as expected.", {

  rpt3 <- create_report("", font = "Arial", font_size = 12) %>%
    page_header(left= c("Hello"),
                right = paste("Right here is something that might wrap.",
                              "If it is long enough so let's make it longer",
                              "If it is long enough so let's make it longer",
                              "If it is long enough so let's make it longer")) %>%
    page_footer("Left", "Center", "Right here is something")

  rpt3 <- page_setup_html(rpt3)

  ph <- get_page_header_html(rpt3)
  ph

  expect_equal(ph$lines, 3) 

})


test_that("get_page_footer_html works as expected.", {

  rpt3 <- create_report("", font = "Arial", font_size = 12) %>%
    page_header(left= c("Hello"),
                right = paste("Right here is something that might wrap.",
                              "If it is long enough so let's make it longer",
                              "If it is long enough so let's make it longer",
                              "If it is long enough so let's make it longer")) %>%
    page_footer("Left", "Center", "Right here is something")

  rpt3 <- page_setup_html(rpt3)

  pf <- get_page_footer_html(rpt3)
  pf

  expect_equal(pf$lines, 2)
  
})



test_that("get_pageby_html works as expected.", {

  tbl <- create_table(mtcars) %>%
    page_by(cyl, "Cylinders:")

  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello") %>%
    footnotes("Goodbye") %>%
    add_content(tbl)

  rpt <- page_setup_rtf(rpt)

  rpt$line_height

  res <- get_page_by_html(tbl$page_by, 6, "fork", rpt, "left")
  res

  expect_equal(res$lines, 2)
  expect_equal(res$html,
               paste0("<table style=\"text-align: left;width:6in;\">\n",
                      "<tr><td style=\"\">Cylinders:&nbsp;fork</td></tr>\n", 
                      "<tr><td style=\"\">&nbsp;</td></tr>\n</table>"))

})


test_that("get_cell_borders_html works as expected.", {
  
  expect_equal(get_cell_borders_html(1, 1, 1, 1, c("all")),
      paste0("border-top:thin solid black;border-bottom:thin solid black;",
             "border-left:thin solid black;border-right:thin solid black;"))
  
  expect_equal(get_cell_borders_html(1, 1, 1, 1, c("all"), exclude = "bottom"),
               paste0("border-top:thin solid black;",
                      "border-left:thin solid black;border-right:thin solid black;"))
  
})

# 
# test_that("rtf2-2: get_cell_borders works as expected.", {
#   
#   
#   expect_equal(get_cell_borders(1, 1, 4, 4, c("all")),
#                "\\clbrdrt\\brdrs\\clbrdrb\\brdrs\\clbrdrl\\brdrs\\clbrdrr\\brdrs")
#   
#   expect_equal(get_cell_borders(4, 1, 4, 4, c("all")),
#                "\\clbrdrt\\brdrs\\clbrdrb\\brdrs\\clbrdrl\\brdrs\\clbrdrr\\brdrs")
#   
#   expect_equal(get_cell_borders(1, 1, 4, 4, c("outside")),
#                "\\clbrdrt\\brdrs\\clbrdrl\\brdrs")
#   
#   expect_equal(get_cell_borders(4, 1, 4, 4, c("outside")),
#                "\\clbrdrb\\brdrs\\clbrdrl\\brdrs")
#   
#   expect_equal(get_cell_borders(2, 2, 4, 4, c("outside")),
#                "")
#   
#   expect_equal(get_cell_borders(2, 4, 4, 4, c("right")),
#                "\\clbrdrr\\brdrs")
#   
#   expect_equal(get_cell_borders(2, 4, 4, 4, c("inside")),
#                "\\clbrdrt\\brdrs\\clbrdrb\\brdrs\\clbrdrl\\brdrs")
#   
#   expect_equal(get_cell_borders(2, 4, 4, 4, "all", "B"), 
#                "\\clbrdrt\\brdrs\\clbrdrb\\brdrs\\clbrdrr\\brdrs")    
#   
#   expect_equal(get_cell_borders(2, 3, 4, 4, "all", "B"), 
#                "\\clbrdrt\\brdrs\\clbrdrb\\brdrs")  
#   
#   expect_equal(get_cell_borders(2, 1, 4, 4, "all", "B"), 
#                "\\clbrdrt\\brdrs\\clbrdrb\\brdrs\\clbrdrl\\brdrs") 
#   
# })
# 
# 
# test_that("page_template_rtf works as expected.", {
#   
#   
#   tbl <- create_table(mtcars) %>% 
#     page_by(cyl, "Cylinders:")
#   
#   rpt <- create_report("", font = "Arial", font_size = 12) %>%
#     page_header("left", "right") %>% 
#     titles("Hello") %>%
#     footnotes("Goodbye") %>% 
#     add_content(tbl) %>% 
#     page_footer("left", right = "right")
#   
#   
#   rpt <- page_setup_rtf(rpt)
#   
#   
#   
#   res <- page_template_rtf(rpt)
#   
#   res
#   expect_equal(res$lines, 6)
#   expect_equal(res$titles$lines, 2)
#   
#   
# })
# 
# test_that("get_page_numbers_rtf works as expected.", {
#   
#   res <- get_page_numbers_rtf("Page [pg] of [tpg]")
#   
#   res
#   expect_equal(res, "Page \\chpgn  of {\\field{\\*\\fldinst  NUMPAGES }}")
#   
#   
# })
# 
# 
# 

context("PDF2 Tests")

base_path <- paste0(getwd(),"/tests/testthat")
data_dir <- base_path

base_path <- tempdir()
data_dir <- "."

cnt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
              "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
              "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
              "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ",
              "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
              "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
              "qui officia deserunt mollit anim id est laborum.")

fnt <- "Arial"
fsz <- 10

dev <- FALSE


# Basic Tests 1 - 10 ------------------------------------------------------



test_that("pdf2-0a: Fixed report is correct.", {
  
  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf2/test0a.pdf")
  
  rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
    titles("Table 0.0", "Baseline Characteristics") %>%
    add_content(create_table(mtcars[1:10, ]))
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
  
})


test_that("pdf2-0b: Fixed report with font_size is correct.", {

  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf2/test0b.pdf")

  rpt <- create_report(fp, output_type = "PDF", font = "fixed",
                       font_size = 12) %>%
    titles("Table 0.0", "Baseline Characteristics") %>%
    add_content(create_table(mtcars[1:10, ]))

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else 
    expect_equal(TRUE, TRUE)

})

test_that("pdf2-0c: Fixed report with font_size options is correct.", {

  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf2/test0c.pdf")

  rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
    options_fixed(font_size = 12) %>%
    titles("Table 0.0", "Baseline Characteristics") %>%
    add_content(create_table(mtcars[1:10, ]))

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else 
    expect_equal(TRUE, TRUE)

})

test_that("pdf2-0d: Fixed report with conflicting font size is correct.", {

  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf2/test0d.pdf")

  rpt <- create_report(fp, output_type = "PDF", font = "fixed",
                       font_size = 8) %>%
    options_fixed(font_size = 12) %>%
    titles("Table 0.0", "Baseline Characteristics") %>%
    add_content(create_table(mtcars[1:10, ]))

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else 
    expect_equal(TRUE, TRUE)

})





test_that("pdf2-1a: Simple report with titles on report is correct.", {
  
  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf2/test1a.pdf")
  
  rpt <- create_report(fp, output_type = "PDF", font = "Arial", 
                       font_size = 12) %>%
    set_margins(top = 1, left = 1, right = 1, bottom = 1) %>% 
    page_header(c("Left1", "Another left"), 
                c("Right1", "Right2", "Another right"),  blank_row = "below") %>% 
    titles("Table 0.0", "Baseline Characteristics", align = "center",
           blank_row = "below", bold = TRUE) %>%
    add_content(create_text(cnt, width = 6, align = "left"), align = "center") %>% 
    footnotes("My footnote1", "My footnote2", valign = "top", blank_row = "none",
              width = "page", align = "left") %>% 
    page_footer(c("Left1", "Left2"), "Center", "Right")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
  
})

test_that("pdf2-1b: Report with title header is correct.", {
  
  
  fp <- file.path(base_path, "pdf2/test1b.pdf")
  
  rpt <- create_report(fp, output_type = "PDF", font = "Arial", 
                       font_size = 12) %>%
    set_margins(top = 1, left = 1, right = 1, bottom = 1) %>% 
    title_header("Table 0.0", "Baseline Characteristics", right = "right") %>%
    add_content(create_text(cnt, width = 6)) %>% 
    page_footer(c("Left1", "Left2"), "Center", c("Right1", "Right2", "Right3"))
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  
})

test_that("pdf2-1c: Report with titles on content is correct.", {
  
  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf2/test1c.pdf")
  
  txt <- create_text(cnt, width = 6) %>% 
    titles("Table 0.0", "Baseline Characteristics", align = "center") %>%
    footnotes("My footnote1", "My footnote2") 
  
  rpt <- create_report(fp, output_type = "PDF", font = "Arial", 
                       font_size = 12) %>%
    set_margins(top = 1, left = 1, right = 1, bottom = 1) %>% 
    page_header(c("Left1", "Another left"), 
                c("Right1", "Right2", "Another right"),  blank_row = "below") %>% 
    add_content(txt) %>% 
    page_footer(c("Left1", "Left2"), "Center", "Right")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
  
})

test_that("pdf2-1d: One page text spec with borders works as expected.", {
  
  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf2/test1d.pdf")
  
  txt <- create_text(cnt, width = 6, borders = "outside", align = "right") %>%
    titles("Text 1.0", "My Nice Text", borders = "outside", font_size = 12) %>%
    footnotes("My footnote 1", "My footnote 2", borders = "outside")
  
  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = 10) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(txt, align = "right") %>%
    page_footer("Left1", "Center1", "Right1")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
  
})

test_that("pdf2-2: Two page text spec works as expected in 12pt font.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "pdf2/test2.pdf")

  cnttxt <- paste(rep(cnt, 12), collapse = "")

  txt <- create_text(cnttxt) %>%
    titles("Text 1.0", "My Nice Text")

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = 12) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(txt) %>%
    page_footer("Left1", "Center1", "Right1") %>%
    footnotes("My footnote 1", "My footnote 2")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)

  } else 
    expect_equal(TRUE, TRUE)
  
})


test_that("pdf2-3: Three page text spec increased margins works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "pdf2/test3.pdf")

  cnttxt <- paste(rep(cnt, 15), collapse = "")

  txt <- create_text(cnttxt, width = 6) %>%
    titles("Text 1.0", "My Nice Text", align = "left") %>%
    titles("Text 1.0", "My Nice Text", align = "right") %>%
    footnotes("My footnote 1", "My footnote 2", align = "left") %>% 
    footnotes("My footnote 3", "My footnote 4", align = "right")

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz) %>%
    set_margins(top = 2, bottom = 2) %>%
    page_header("Left", c("Right1", "Page [pg] of [tpg]")) %>%
    add_content(txt) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 6)

  } else 
    expect_equal(TRUE, TRUE)
  
})


test_that("pdf2-4: Two page text spec works as expected in 10pt font.", {

  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf2/test4.pdf")

  cnttxt <- paste(rep(cnt, 12), collapse = "")

  txt <- create_text(cnttxt, width = 6) %>% 
    title_header("Text 1.0", right = "My Nice Text", blank_row = "above",
                 width = "page") %>%
    title_header("Text 2.0", right = "My Nice Text2", blank_row = "below", 
                 width = "page") %>%
    footnotes("My footnote 1", "My footnote 2", width = "page")

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = 10) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(txt) %>%
    page_footer("Left1", "Center1", "Right1") 


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)

  } else 
    expect_equal(TRUE, TRUE)

})


test_that("pdf2-5: Two page text spec works as expected in 8pt font.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "pdf2/test5.pdf")

  cnttxt <- paste(rep(cnt, 20), collapse = "")

  txt <- create_text(cnttxt)

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = 8) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(txt) %>%
    page_footer("Left1", "Center1", "Page [pg] of [tpg]") %>%
    titles("Text 1.0", "My Nice Text", bold = TRUE) %>%
    footnotes("My footnote 1", "My footnote 2")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)

  } else 
    expect_equal(TRUE, TRUE)

})


test_that("pdf2-6: One page table works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "pdf2/test6.pdf")

  dat <- mtcars[1:15, ]
  attr(dat[[2]], "label") <- "Cylin."

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape", 
                       units = "inches") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table", borders = "top", bold = TRUE) %>%
    add_content(create_table(dat, borders = "outside")) %>%
    footnotes("My footnote 1", "My footnote 2", borders = c("top", "bottom"), 
              blank_row = "above", valign = "bottom") %>%
    page_footer("Left1", "Center1", "Right1", blank_row = "above")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else 
    expect_equal(TRUE, TRUE)

})


test_that("pdf2-7: Multi page table works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "pdf2/test7.pdf")

  dat <- iris


  tbl <- create_table(dat, borders = "none") %>%
    titles("Table 1.0", "My Nice Irises", "Another Title") %>%
    define(Sepal.Length, label = "Sepal Length", width = 1, align = "center") %>%
    define(Sepal.Width, label = "Sepal Width", width = 1, align = "centre") %>%
    define(Species, blank_after = TRUE)

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = 12, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1")) %>%
    add_content(tbl, blank_row = "none") %>%
    page_footer("Page [pg] of [tpg]", "Page [pg] of [tpg]", "Page [pg] of [tpg]") %>%
    footnotes("My footnote 1", "My footnote 2")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 8)

  } else 
    expect_equal(TRUE, TRUE)

})



test_that("pdf2-8: Portrait table works as expected.", {

  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf2/test8.pdf")

  dat <- mtcars[1:15, ]

  tbl <- create_table(dat)

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = 8, orientation = "portrait") %>%
    # set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(tbl) %>%
    footnotes("My footnote 1", "My footnote 2") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 11)
  expect_equal(res$pages, 1)

  } else 
    expect_equal(TRUE, TRUE)

})


test_that("pdf2-9: Wide table works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "pdf2/test9.pdf")

  dat <- mtcars[1:15, ]

  tbl <- create_table(dat, first_row_blank = TRUE) %>%
    column_defaults(width = 1)


  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "portrait") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(tbl) %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top") %>% # Works!
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)

  } else 
    expect_equal(TRUE, TRUE)

})


test_that("pdf2-10: Preview works as expected.", {


  fp <- file.path(base_path, "pdf2/test10.pdf")

  dat <- iris

  tbl <- create_table(dat, borders = "none") %>%
    titles("Table 1.0", "My Nice Irises", "Another Title") %>%
    footnotes("My footnote 1", "My footnote 2")

  rpt <- create_report(fp, output_type = "PDF", font = "Arial",
                       font_size = 12, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1")) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt, preview = 2)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)


})

# Basic Tests 11 - 20 ------------------------------------------------------

# This is awesome. Shows cell wrapping, page break, and valign
# Very good for testing.
test_that("pdf2-11: Forced page wrap works as expected.", {


  fp <- file.path(base_path, "pdf2/test11.pdf")

  dat <- data.frame(labels = rownames(mtcars), mtcars)

  tbl <- create_table(dat, borders = "all") %>%
    titles("Table 1.0", "My Nice Irises", "Another Title") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "bottom") %>%
    define(labels, id_var = TRUE) %>%
    define(wt, page_wrap = TRUE)

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = 12, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1")) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Page [pg] of [tpg]")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 4)


})

# Fix borders
test_that("pdf2-12: Table Borders work as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "pdf2/test12.pdf")

  dat <- mtcars[1:14, ]

  #  c("left", "right", "bottom", "top")
  
  tbl <- create_table(dat, borders = c("all")) %>%
    define(mpg, label = "Miles Per Gallon") %>% 
    titles("Table 1.0", "My Nice Table", 
           borders = "outside", blank_row = "above", width = "content") %>% 
    footnotes("My footnote 1", "My footnote 2", borders = "outside")

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = 12, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "none") %>%
    add_content(tbl, align = "right") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 11)

  } else 
    expect_equal(TRUE, TRUE)

})

# Nice.
test_that("pdf2-13: Spanning headers work as expected.", {


  fp <- file.path(base_path, "pdf2/test13.pdf")

  dat <- mtcars[1:15, ]

  tbl <- create_table(dat, borders = c("top", "bottom")) %>%
    spanning_header(cyl, disp, "Span 1 and go an go", n = 10, label_align = "left") %>%
    spanning_header(hp, wt, "Span 2", n = 11, underline = TRUE) %>%
    spanning_header(qsec, vs, "Span 3") %>%
    spanning_header(drat, gear, "Super Duper\n Big Ol' Span", n = 25, level = 2)


  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(tbl) %>%
    footnotes("My footnote 1", "My footnote 2") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

})


test_that("pdf2-14: Labels and show_cols work as expected.", {


  fp <- file.path(base_path, "pdf2/test14.pdf")

  dat <- mtcars[1:15, ]

  attr(dat$mpg, "label") <- "Miles Per Gallon"
  attr(dat$cyl, "label") <- "Cylinders"

  tbl <- create_table(dat, show_cols = c("mpg", "hp", "cyl", "disp", "drat")) %>%
    define(mpg, width = 1.25) %>%
    define(disp, label = "Displacement")

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(tbl) %>%
    footnotes("My footnote 1", "My footnote 2") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res


  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

})


test_that("pdf2-15: Valign on report footnotes works as expected.", {


  fp <- file.path(base_path, "pdf2/test15.pdf")

  dat <- iris[1:100, ]

  tbl <- create_table(dat) %>%
    define(Species, page_break = TRUE)

  rpt <- create_report(fp, output_type = "PDF", font = "Arial",
                       font_size = 9, orientation = "portrait") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(tbl) %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 4)
  expect_equal(length(res$column_widths[[1]]), 5)

  class(dat$Species)
})

# Works
test_that("pdf2-16: Valign on table footnotes works as expected.", {


  fp <- file.path(base_path, "pdf2/test16.pdf")

  dat <- iris[1:100, ]

  tbl <- create_table(dat) %>%
    footnotes("My footnote 1", "My footnote 2", valign = "bottom") %>%
    define(Species, page_break = TRUE)

  rpt <- create_report(fp, output_type = "PDF", font = "Arial",
                       font_size = 10, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 4)
  expect_equal(length(res$column_widths[[1]]), 5)


})


test_that("pdf2-17: Title header on table works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "pdf2/test17.pdf")

  dat <- iris[1:25, ]

  tbl <- create_table(dat, width = 9) %>%
    title_header("Table 1.0", "My Nice Table",
                 right = c("Right1",
                           "Right2", "Page [pg] of [tpg]")) %>%
    footnotes("My footnote 1", "My footnote 2")

  rpt <- create_report(fp, output_type = "PDF", font = "Arial",
                       font_size = 10, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  } else 
    expect_equal(TRUE, TRUE)


})


test_that("pdf2-18: Title header on report works as expected.", {

  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf2/test18.pdf")

  dat <- iris[1:50, ]

  tbl <- create_table(dat)

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Page [pg] of [tpg]") %>%
    title_header("Table 1.0", "My Nice Table", right = c("Page [pg] of [tpg]",
                                                         "Right2", "Right3")) %>%
    footnotes("My footnote 1", "My footnote 2", valign = "bottom")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  } else 
    expect_equal(TRUE, TRUE)


})


test_that("pdf2-19: Title and Footnote borders work as expected.", {
  
  if (dev == TRUE) {


  fp <- file.path(base_path, "pdf2/test19.pdf")

  dat <- iris[1:20, ]

  tbl <- create_table(dat, borders = c("all"), first_row_blank = FALSE) %>%
    titles("Table 1.0", "My Nice Report with Borders", 
          # "A third titleg what happens when it wraps around I want to know what happens will it work",
           borders = "outside", #c("top", "bottom"), #c("top", "bottom", "left", "right"),
           blank_row = "both", align = "left", font_size = 10) %>%
    #titles("Just to mess it up", borders = "outside", blank_row = "both") %>% 
    footnotes("My footnote 1", "My footnote 2", 
             # "A third footnoteg what happens when it wraps around I want to know what happens will it work",
              valign = "top", align = "right",
              borders = "all", #c("top", "bottom", "left", "right"),
              blank_row = "none")

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    page_header("Left", "Right", blank_row = "none") %>% 
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl, align = "right") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  } else 
    expect_equal(TRUE, TRUE)


})

# Best to try different combinations of blank rows.
test_that("pdf2-20: Title Header borders work as expected.", {


  fp <- file.path(base_path, "pdf2/test20.pdf")

  dat <- iris[1:25, ]

  tbl <- create_table(dat, borders = "all") %>%
    title_header("Table 1.0", 
                 "My Nice Report with Borders ", # that keep going and going
                 right = c("Right2", "Right3"),
                 borders = c("outside"),
                 blank_row = "none") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top",
              borders = c("top", "bottom", "left", "right"),
              blank_row = "none")

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    options_fixed(line_size = 40) %>% 
    add_content(tbl, align = "right") %>%
    page_header("Left", "Right") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)


})

# # Basic Tests 21 - 30 ------------------------------------------------------


# Works unless drat column is too narrow and spanning label wraps unexpectedly
test_that("pdf2-21: Page wrap with spanning header works as expected.", {
  
  fp <- file.path(base_path, "pdf2/test21.pdf")
  
  
  df <- data.frame(vehicle = rownames(mtcars), mtcars, stringsAsFactors = FALSE)
  rownames(df) = NULL
  
  tbl <- create_table(df) %>% 
    spanning_header(2, 5,
                    label = "Span 1", label_align = "center", n = 10) %>%
    spanning_header(6, 8,
                    label = "Span 2", label_align = "center", n = 10) %>%
    spanning_header(9, 12,
                    label = "Span 3", label_align = "center", n = 10) %>%
    spanning_header(6, 12, label = "Super Span",
                    label_align = "center",
                    level = 2) %>%
    define(vehicle, label = "Vehicle", id_var = TRUE) %>% 
    define(mpg, format = "%.1f") %>% 
    define(drat, width = 1) %>% 
    define(wt, page_wrap = TRUE) %>% 
    define(vs, page_wrap = TRUE)
  
  rpt <- create_report(fp, output_type = "PDF", 
                       orientation = "portrait", font = fnt) %>%
    add_content(tbl) %>% 
    titles("Table 1.0", "MTCARS Subset Test") %>% 
    footnotes("My footnote") %>% 
    page_header("Left") %>% 
    page_footer("Left", right = "Right")
  
  #print(rpt)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  expect_equal(res$pages, 3)
})

# This is a good one for testing height calculations.  Looks good for now.
test_that("pdf2-22: Page by works as expected.", {


  fp <- file.path(base_path, "pdf2/test22.pdf")

  dat <- iris
  dat$Species <- as.character(dat$Species)
  
  library(fmtr)
  
  fmt <- value(condition(x == "setosa", "Setosa"),
               condition(x == "versicolor", "Versicolor"),
               condition(x == "virginica", "Virginica"))

  tbl <- create_table(dat, borders = "all") %>%
    titles("Table 1.0", "My Nice Report with a Page By", borders = "outside", 
           blank_row = "none") %>%
    page_by(Species, label = "Species", align = "center", borders = "all", 
            blank_row = "none", format = fmt) %>%
    footnotes("My footnote 1", "Page [pg] of [tpg]", borders = "none", align = "right")

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_header("Page [pg] of [tpg]", "Page [pg] of [tpg]") %>%
    page_footer("Page [pg] of [tpg]", 
                "Page [pg] of [tpg]", 
                "Page [pg] of [tpg]") 

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 7)
  expect_equal(length(res$column_widths[[1]]), 5)


})


test_that("pdf2-23: Two contents on one page works as expected.", {


  fp <- file.path(base_path, "pdf2/test23.pdf")

  dat <- mtcars[1:15, ]

  tbl <- create_table(dat) %>%
    titles("Table 1.0", "My Nice Table") %>%
    footnotes("My footnote 1", "My footnote 2")

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), blank_row = "below") %>%
    add_content(tbl, page_break = FALSE, blank_row = "below") %>%
    add_content(create_text(cnt, width = 5), align = "center") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)


})

# Working.
test_that("pdf2-24: Two tables one headerless works as expected.", {


  fp <- file.path(base_path, "pdf2/test24.pdf")

  dat <- mtcars
  dat2 <- mtcars[16:20, ]

  tbl <- create_table(dat) %>%
    titles("Table 1.0", "My Nice Table") %>%
    column_defaults(width = .5)

  tbl2 <- create_table(dat2, headerless = TRUE) %>%
    column_defaults(width = .5)

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), blank_row = "below") %>%
    add_content(tbl, page_break = FALSE, blank_row = "none") %>%
    add_content(tbl2, align = "center") %>%
    page_footer("Left1", "Center1", "Right1") %>%
    footnotes("My footnote 1", "My footnote 2")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)


})


# OK except plot not centered in borders.
# And footnotes have double line above.
test_that("pdf2-25: Simplest PDF2 Plot works as expected.", {

  if (dev) {
    library(ggplot2)
  
    fp <- file.path(base_path, "pdf2/test25.pdf")
  
    p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  
    plt <- create_plot(p, height = 4, width = 8, borders = c("top", "bottom", "all")) %>%
      titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", 
             borders = "outside", blank_row = "none") %>% #, font_size = 12, bold = TRUE) %>%
      footnotes("* Motor Trend, 1974", borders = "outside", blank_row = "above", 
                valign = "top", width = "content", align = "center")
  
  
    rpt <- create_report(fp, output_type = "PDF", font = fnt, font_size =fsz) %>%
      page_header("Client", "Study: XYZ") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(plt, align = "center") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
    res <- write_report(rpt)
  
    #print(res)
  
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)
  
  } else {
    expect_equal(TRUE, TRUE) 
  }


})


test_that("pdf2-26: Plot, Table and Text on same page works as expected.", {

  if (dev) {
    library(ggplot2)
  
    fp <- file.path(base_path, "pdf2/test26.pdf")
  
    p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  
    plt <- create_plot(p, height = 3.5, width = 7) %>% 
      titles("My plot title")
    tbl <- create_table(mtcars[1:3, ]) %>% 
      titles("My table Title")
    txt <- create_text("Here is some text", align = "center")
  
  
    rpt <- create_report(fp, output_type = "PDF", font = fnt, font_size = fsz) %>%
      page_header("Client", "Study: XYZ") %>%
      titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", blank_row = "below") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(plt, page_break = FALSE, blank_row = "none") %>%
      add_content(tbl, page_break = FALSE) %>%
      add_content(txt) %>% 
      footnotes("* Motor Trend, 1974") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
    res <- write_report(rpt)
  
    #print(res)
  
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)

  
  } else {
    expect_equal(TRUE, TRUE) 
  }

})


# Works
test_that("pdf2-27: Plot with page by on plot works as expected.", {

  
  if (dev) {
    library(ggplot2)
    library(fmtr)
    
    fmt <- value(condition(x == 4, "4 Cylinder"),
                 condition(x == 6, "6 Cylinder"),
                 condition(x == 8, "8 Cylinder"))
  
    fp <- file.path(base_path, "pdf2/test27.pdf")
  
  
    dat <- mtcars[order(mtcars$cyl), ]
  
    p <- ggplot(dat, aes(x=disp, y=mpg)) + geom_point()
  
  
    #dats <- split(p$data, p$data$grp)
    #tbl <- create_table(dat[1:3, ])
  
    plt <- create_plot(p, height = 4, width = 8, borders = "all") %>%
      titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", blank_row = "none",
             borders = "all") %>%
      page_by(cyl, "Cylinders2: ", blank_row = "below", 
              borders = "all", format = fmt) %>%
      footnotes("* Motor Trend, 1974", borders = "all")
  
    rpt <- create_report(fp, output_type = "PDF", font = fnt, font_size = fsz) %>%
      page_header("Client", "Study: XYZ") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(plt, align = "center") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
    res <- write_report(rpt)
  
    #print(res)
  
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 3)
  
  } else {
    expect_equal(TRUE, TRUE) 
  }

})
 
# Also works
test_that("pdf2-28: Plot with page by on report works as expected.", {

  
  if (dev) {
    library(ggplot2)
  
    fp <- file.path(base_path, "pdf2/test28.pdf")
  
  
    dat <- mtcars[order(mtcars$cyl), ]
  
    p <- ggplot(dat, aes(x=disp, y=mpg)) + geom_point()
  
  
    #dats <- split(p$data, p$data$grp)
    #tbl <- create_table(dat[1:3, ])
  
    plt <- create_plot(p, height = 4, width = 8)
  
  
    rpt <- create_report(fp, output_type = "PDF", font = fnt, font_size = fsz) %>%
      page_header("Client", "Study: XYZ") %>%
      titles("Figure 1.0", "MTCARS Miles per Cylinder Plot",
             blank_row = "none", borders = "none") %>%
      set_margins(top = 1, bottom = 1) %>%
      page_by(cyl, "Cylinders: ") %>%
      add_content(plt) %>%
      footnotes("* Motor Trend, 1974", borders = "none") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
    res <- write_report(rpt)
  
    #print(res)
  
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 3)

  
  } else {
    expect_equal(TRUE, TRUE) 
  }
  
  
})

test_that("pdf2-29: Simplest Plot with valign top works as expected.", {

  
  if (dev) {
    library(ggplot2)
  
    fp <- file.path(base_path, "pdf2/test29.pdf")
  
    p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  
    plt <- create_plot(p, height = 4, width = 8)
  
  
    rpt <- create_report(fp, output_type = "PDF", font = fnt, font_size = fsz) %>%
      page_header("Client", "Study: XYZ") %>%
      titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(plt, align = "center") %>%
      footnotes("* Motor Trend, 1974", valign = "top") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
    res <- write_report(rpt)
  
    #print(res)
  
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)
  
  
  } else {
    expect_equal(TRUE, TRUE) 
  }


})



test_that("pdf2-30: Simplest PDF Plot with valign bottom works as expected.", {

  library(ggplot2)

  fp <- file.path(base_path, "pdf2/test30.pdf")

  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()

  plt <- create_plot(p, height = 4, width = 8) %>%
    footnotes("* Motor Trend, 1974", valign = "bottom")


  rpt <- create_report(fp, output_type = "PDF", font = fnt, font_size = fsz) %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt, align = "center") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")


  res <- write_report(rpt)

  #print(res)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)


})

# # Basic Tests 31 - 40 ------------------------------------------------------

test_that("pdf2-31: Simplest Text with valign top works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "pdf2/test31.pdf")

  txt <- create_text(cnt, width = 6, borders = "none") %>%
    titles("Text 1.0", "MTCARS Miles per Cylinder Text", 
           borders = "none", blank_row = "none") %>%
    footnotes("* Motor Trend, 1974", valign = "top", borders = "none",
              blank_row = "none")


  rpt <- create_report(fp, output_type = "PDF", font = fnt, font_size = fsz) %>%
    page_header("Client", "Study: XYZ") %>%

    set_margins(top = 1, bottom = 1) %>%
    add_content(txt, align = "center") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")


  res <- write_report(rpt)

  #print(res)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else 
    expect_equal(TRUE, TRUE)
  
  
})

test_that("pdf2-32: Simplest Text with valign bottom works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "pdf2/test32.pdf")

  txt <- create_text(cnt, width = 6) %>%
    footnotes("* Motor Trend, 1974", valign = "bottom")

  rpt <- create_report(fp, output_type = "PDF", font = fnt, font_size = fsz) %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Text 1.0", "MTCARS Miles per Cylinder Plot") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(txt, align = "center") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")


  res <- write_report(rpt)

  #print(res)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)


})

# Works
test_that("pdf2-33: Table with long cell and label values wraps as expected.", {

  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf2/test33.pdf")


  # Setup
  arm <- c(rep("A", 5), rep("B", 5))
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa \nand more and more", "Schaffer, Ashley", "Laner, Tahma",
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)


  # Create data frame
  df1 <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)


  tbl1 <- create_table(df1, first_row_blank = FALSE, borders = "all") %>%
    define(subjid, label = "Subject ID for a patient", n = 10, align = "left",
           width = 1) %>%
    define(name, label = "Subject Name", width = 1) %>%
    define(sex, label = "Sex", n = 10, align = "center") %>%
    define(age, label = "Age", n = 10) %>%
    define(arm, label = "Arm",
           blank_after = TRUE,
           dedupe = TRUE) %>% 
    titles("Table 1.0", align = "center", borders = "all") %>%
    footnotes("Here", borders = "all") 


  rpt <- create_report(fp, output_type = "PDF", font = "Courier",
                       font_size = fsz) %>%
    add_content(tbl1)


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  
  } else 
    expect_equal(TRUE, TRUE)



})

# Working
test_that("pdf2-34: Table with break between sections works as expected.", {


  fp <- file.path(base_path, "pdf2/test34.pdf")


  # Setup
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma",
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
  arm <- c(rep("A", 5), rep("B", 5))

  # Create data frame
  df1 <- data.frame(subjid, name, sex, age, arm)


  tbl1 <- create_table(df1, first_row_blank = TRUE, borders = "all") %>%
    define(subjid, label = "Subject ID", align = "right", 
           label_align = "right", width = 1) %>%
    define(name, label = "Subject Name", width = 1, align = "right") %>%
    define(sex, label = "Sex", align = "right") %>%
    define(age, label = "Age", align = "right") %>%
    define(arm, label = "Arm",
           blank_after = TRUE,
           dedupe = TRUE,
           align = "right") #%>%
  # spanning_header(sex, arm, label = "Here is a spanning header")


  rpt <- create_report(fp, output_type = "PDF", font = fnt, font_size = fsz) %>%
    page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
    # options_fixed(line_count = 46) %>%
    titles("Table 1.0", "Analysis Data Subject Listing\n And more stuff",
           "Safety Population", align = "center") %>%
    footnotes("Program Name: table1_0.R",
              "Here is a big long footnote that is going to wrap\n at least once") %>%
    page_footer(left = "Time", center = "Confidential",
                right = "Page [pg] of [tpg]") %>%
    add_content(tbl1)


  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)


})


test_that("pdf2-35: Title Header and page header/footer wrapping work as expected.", {


  fp <- file.path(base_path, "pdf2/test35.pdf")

  dat <- iris[1:10, ]

  tbl <- create_table(dat, borders = "all") %>%
    title_header("Table 1.0", "My Nice Report with Borders that will go on and on",
                 right = c("Right1", "Right2",
                           "Right3 long enough to wrap around at least once"),
                 borders = "none",
                 blank_row = "none") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top",
              borders = "none",
              blank_row = "none")

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl, align = "left") %>%
    page_header(c("Left1", "Left2\nwrap"), "Right 1") %>%
    page_footer("Left1",
                "Center1 here is a whole bunch of stuff to try and make it wrap",
                "Right1\nwrap\n and wrap again")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)


})


test_that("pdf2-36: Title and Footnote widths work as expected on table.", {


  fp <- file.path(base_path, "pdf2/test36.pdf")

  dat <- iris[1:20, ]

  tbl <- create_table(dat) %>%
    titles("Table 1.0", "My Nice Report with Widths",
           width = "page", align = "left") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top",
              width = "page")

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)


})


test_that("pdf2-37: Title and Footnote widths work as expected on report", {


  fp <- file.path(base_path, "pdf2/test37.pdf")

  dat <- iris[1:20, ]

  tbl <- create_table(dat, borders = "all")

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1") %>%
    titles("Table 1.0", "My Nice Report with Widths",
           borders = c("top", "bottom", "left", "right"),
           width = 7, align = "left") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top",
              borders = c("top", "bottom", "left", "right"),
              width = 7)

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)


})


test_that("pdf2-38: Title and Footnote specific widths work as expected.", {


  fp <- file.path(base_path, "pdf2/test38.pdf")

  dat <- iris[1:20, ]

  tbl <- create_table(dat, borders = "all") %>%
    titles("Table 1.0", "My Nice Report with Borders",
           borders = c("outside"),
           width = 7, align = "left") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top",
              borders = c("top", "bottom", "left", "right"),
              width = 7)

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)


})

# Works
test_that("pdf2-39: One page table works as expected in centimeters and times.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "pdf2/test39.pdf")

  dat <- mtcars[1:15, ]
  attr(dat[[2]], "label") <- "Cylin."

  rpt <- create_report(fp, output_type = "PDF", font = "Times",
                       font_size = fsz, orientation = "landscape",
                       units = "cm") %>%
    set_margins(top = 3, bottom = 3) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(create_table(dat)) %>%
    footnotes("My footnote 1", "My footnote 2") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)


})

# Title not aligned properly
test_that("pdf2-40: One page table works as expected in courier and cm.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "pdf2/test40.pdf")

  dat <- mtcars[1:15, ]
  attr(dat[[2]], "label") <- "Cylin."

  rpt <- create_report(fp, output_type = "PDF", font = "Courier",
                       font_size = fsz, orientation = "landscape",
                       units = "cm") %>%
    set_margins(top = 3, bottom = 3) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table", align = "center") %>%
    add_content(create_table(dat)) %>%
    footnotes("My footnote 1", "My footnote 2") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)


})

# # Basic Tests 41 - 50 ------------------------------------------------------

# Good for testing borders and spacing are working correctly  ***
test_that("pdf2-41: Page by with borders works as expected.", {


  fp <- file.path(base_path, "pdf2/test41.pdf")

  dat <- iris

  brdrs <- "none"

  tbl <- create_table(dat, borders = "all", first_row_blank = TRUE) %>%
    titles("Table 1.0", "My Nice Report with a Page By", borders = "all", #c("top", "bottom"), 
           blank_row = "none", align = "left", font_size = 12) %>%
    page_by(Species, label = "Species: ", align = "left", borders = "all", # c("top", "bottom"),
            blank_row = "none") %>%
    footnotes("My footnote 1", "My footnote 2", borders = c("all"), 
              blank_row = "none")

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = 10, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 7)
  expect_equal(length(res$column_widths[[1]]), 5)


})

# Test for borders and page wraps
test_that("pdf2-42: Long table with borders and footnotes on report.", {


  fp <- file.path(base_path, "pdf2/test42.pdf")

  dat <- data.frame(labels = rownames(mtcars), mtcars)

  tbl <- create_table(dat, borders = "all") %>%
    define(labels, id_var = TRUE) %>%
    define(wt, page_wrap = TRUE)

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = 12, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1")) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Page [pg] of [tpg]") %>%
    titles("Table 1.0", "My Nice Irises", "Another Title") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "bottom")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 4)


})


# Works
test_that("pdf2-43: use_attributes parameter table works as expected.", {


  fp1 <- file.path(base_path, "pdf2/test43a.pdf")
  fp2 <- file.path(base_path, "pdf2/test43b.pdf")
  fp3 <- file.path(base_path, "pdf2/test43c.pdf")


  dat <- mtcars[1:10, ]
  attr(dat$mpg, "label") <- "Miles per gallon"
  attr(dat$cyl, "format") <- "%.1f"
  attr(dat$hp, "width") <- 2
  fattr(dat$vs) <- list(width = 2, justify = "center")

  tbl <- create_table(dat)

  # Test default
  rpt <- create_report(fp1, output_type = "PDF", font = fnt) %>%
    add_content(tbl)

  res <- write_report(rpt)

  expect_equal(file.exists(fp1), TRUE)
  expect_equal(res$pages, 1)

  # Test none
  tbl <- create_table(dat, use_attributes = "none")

  rpt <- create_report(fp2, output_type = "PDF", font = fnt) %>%
    add_content(tbl)

  res <- write_report(rpt)

  expect_equal(file.exists(fp2), TRUE)


  # Test some
  tbl <- create_table(dat, use_attributes = c("format", "width"))

  rpt <- create_report(fp3, output_type = "PDF", font = fnt) %>%
    add_content(tbl)

  res <- write_report(rpt)

  expect_equal(file.exists(fp3), TRUE)
  expect_equal(res$pages, 1)

})

# Works
test_that("pdf2-45: Title bold and font size works as expected.", {


  fp <- file.path(base_path, "pdf2/test45.pdf")

  dat <- mtcars[1:15, ]
  attr(dat[[2]], "label") <- "Cylin."
  attr(dat[[2]], "width") <- 1
  attr(dat[[2]], "justify") <- "center"

  tbl <- create_table(dat, borders = "outside") %>%
    titles("Table 1.0", 
           paste("My Nice Table that will go on and on and eventually will",
                 "wrap around I hope so maybe I can test this feature of the ",
                 "application"), borders = c("none"),
           width = "content", font_size = 14, bold = TRUE) %>%
    footnotes("My footnote 1", "My footnote 2", borders = "none",
              align = "left", width = "content") %>%
    define(wt, width = 1, label = "Weight", align = "center",
           label_align = "right")

  rpt <- create_report(fp, output_type = "PDF", font = "Arial",
                       font_size = 9, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"),
                blank_row = "below") %>%
    add_content(tbl, align = "center")  %>%
    page_footer("Left1", "Center1", 
                paste("Right1 here is one that is really, ", 
                "really long that I hope will wrap around"))

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

})



test_that("pdf2-46: 9 pt font inches works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "pdf2/test46.pdf")

  rpt <- create_report(fp, output_type = "PDF", font_size = 9,
                       font = "Courier",
                       orientation = "portrait") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>%
    set_margins(top = 1, bottom = 1)


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  
  } else 
    expect_equal(TRUE, TRUE)


})

# Working 
test_that("pdf2-47: 9 pt font cm works as expected.", {

  if (dev == TRUE) {


    fp <- file.path(base_path, "pdf2/test47.pdf")
  
    rpt <- create_report(fp, output_type = "PDF", font_size = 9,
                         font = "Courier",
                         orientation = "portrait", 
                         units = "cm") %>%
      page_header("left", "right") %>%
      titles("IRIS Data Frame") %>%
      add_content(create_table(iris), align = "left") %>%
      page_footer("left", "center", "Page [pg] of [tpg]") %>%
      set_margins(top = 1, bottom = 1)
  
  
    res <- write_report(rpt)
  
    expect_equal(file.exists(fp), TRUE)
  
  } else 
    expect_equal(TRUE, TRUE)


})

test_that("pdf2-48: 11 pt font inches works as expected.", {

  if (dev == TRUE) {

  
    fp <- file.path(base_path, "pdf2/test48.pdf")
  
    rpt <- create_report(fp, output_type = "PDF", font_size = 11,
                         font = "Courier",
                         orientation = "portrait") %>%
      page_header("left", "right") %>%
      titles("IRIS Data Frame") %>%
      add_content(create_table(iris)) %>%
      page_footer("left", "center", "Page [pg] of [tpg]") %>%
      set_margins(top = 1, bottom = 1)
  
  
    res <- write_report(rpt)
  
    expect_equal(file.exists(fp), TRUE)

  } else 
    expect_equal(TRUE, TRUE)
})


test_that("pdf2-49: 11 pt font cm works as expected.", {

  if (dev == TRUE) {

  
    fp <- file.path(base_path, "pdf2/test49.pdf")
  
    rpt <- create_report(fp, output_type = "PDF", font_size = 11,
                         font = "Courier",
                         orientation = "portrait", 
                         units = "cm") %>%
      page_header("left", "right") %>%
      titles("IRIS Data Frame") %>%
      add_content(create_table(iris), align = "center") %>%
      page_footer("left", "center", "Page [pg] of [tpg]") %>%
      set_margins(top = 1, bottom = 1)
  
  
    res <- write_report(rpt)
  
    res$column_widths
    
    expect_equal(file.exists(fp), TRUE)

  } else 
    expect_equal(TRUE, TRUE)

})

test_that("pdf2-50: Spanning headers borders work as expected.", {


  fp <- file.path(base_path, "pdf2/test50.pdf")

  dat <- mtcars[1:15, ]

  tbl <- create_table(dat, borders = c("all")) %>%
    spanning_header(cyl, disp, "Span 1", label_align = "left") %>%
    spanning_header(hp, wt, "Span 2", underline = FALSE) %>%
    spanning_header(qsec, vs, "Span 3", n = 10) %>%
    spanning_header(drat, gear, "Super Duper\nWrapped Span", n = 11, level = 2) %>%
    titles("Table 1.0", "My Nice Table", borders = "outside") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "outside")

  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl)


  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

})







# Basic Tests 51 - 60 ------------------------------------------------------

# Very good report for testing multiple content wraps
test_that("pdf2-51: Plot, Long Table and Long Text on same report works as expected.", {
  
  if (dev) {
    library(ggplot2)
    
    fp <- file.path(base_path, "pdf2/test51.pdf")
    
    p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
    
    plt <- create_plot(p, height = 3.5, width = 7) %>% 
      titles("My plot title", blank_row = "above")
    tbl <- create_table(mtcars[1:17, ]) %>% 
      titles("My table Title", blank_row = "above") %>% 
      footnotes("Table Footnote", blank_row = "above")
    txt <- create_text(cnt, align = "center", width = 6) %>% 
      titles("My Text Title", blank_row = "none") %>% 
      footnotes("Text Footnotes", blank_row = "above")
    
    
    rpt <- create_report(fp, output_type = "PDF", font = fnt, font_size = fsz) %>%
      page_header("Client", "Study: XYZ") %>%
      titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", blank_row = "none") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(plt, page_break = FALSE, blank_row = "none") %>%
      add_content(tbl, page_break = FALSE, blank_row = "none") %>%
      add_content(txt, blank_row = "below", page_break = FALSE) %>% 
      add_content(txt, blank_row = "none") %>% 
      footnotes("* Motor Trend, 1974") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")
    
    
    res <- write_report(rpt)
    
    #print(res)
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 3)
   
  
  } else {
    expect_equal(TRUE, TRUE) 
  }
  
  
})



test_that("pdf2-52: PDF Image file works as expected.", {
  
  if (dev == TRUE) {

    
    library(ggplot2)
    
    fp <- file.path(base_path, "pdf2/test52.pdf")
    
    p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
    
    pltpath <- file.path(base_path, "pdf2/test52.jpg")
    ggsave(pltpath, width = 8, height = 4, 
           units = "in",
           dpi = 300)
    
    plt <- create_plot(pltpath, height = 4, width = 8)
    
    rpt <- create_report(fp, output_type = "PDF", font = "Arial") %>%
      page_header("Client", "Study: XYZ") %>%
      titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(plt, align = "center") %>%
      footnotes("* Motor Trend, 1974") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")
    
    res <- write_report(rpt)
    
    #print(res)
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
})

test_that("pdf2-53: Header_bold works as expected", {
  
  dat <- mtcars[1:10, 1:3]
  
  fp <- file.path(base_path, "pdf2/test53.pdf")
  
  tbl <- create_table(dat, header_bold = TRUE, borders = "all") %>%
    column_defaults(width = 1) %>%
    titles("Report 1.0", "Simple Report", borders = "outside", 
           blank_row = "none", bold = TRUE) %>%
    footnotes("My footnote", blank_row = "none")
  
  rpt <- create_report(fp, orientation = "portrait",
                       output_type = "PDF", font = "Arial") %>%
    add_content(tbl) %>%
    set_margins(top = 1)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})

test_that("pdf2-54: Titles and footnotes in header and footer works as expected", {
  
  dat <- mtcars[1:10, 1:3]
  
  fp <- file.path(base_path, "pdf2/test54.pdf")
  
  tbl <- create_table(dat) %>%
    column_defaults(width = 1) %>%
    titles("My table title") %>%
    footnotes("My table footnote")
  
  rpt <- create_report(fp, orientation = "landscape",
                       output_type = "PDF", font = "Arial") %>%
    add_content(tbl) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    titles("My report title") %>%
    titles("Report 1.0", "Simple Report", 
           blank_row = "none", header = TRUE) %>%
    footnotes("My report footnote") %>%
    footnotes("My footer footnote", "Another footer footnote", "And another", 
              blank_row = "none", footer = TRUE) %>%
    page_footer("Left", "Center", "Right")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})

test_that("pdf2-55: Label row is one cell.", {
  
  
  fp <- file.path(base_path, "pdf2/test55.pdf")
  
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  ll <- "Here is a super long label to see if it can span the entire table."
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE, borders = c("all")) %>% 
    stub(c("var", "label"), width = 1) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = ll, cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center", n = 19) %>% 
    define(B, label = "Group B", align = "center", n = 13)
  
  
  # Create report and add content
  rpt <- create_report(fp, orientation = "landscape", output_type = "PDF",
                       font = "Times") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = Sys.time(), 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  
  
  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)
  
  
})

test_that("pdf2-56: Blank after on invisible column.", {
  
  fp <- file.path(base_path, "pdf2/test56.pdf")
  
  tbl <- create_table(iris, borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "PDF", font = "Courier") %>%
    page_header("Left", "Right") %>%
    add_content(tbl) %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("pdf2-57: Page header width works.", {
  
  fp <- file.path(base_path, "pdf2/test57.pdf")
  
  tbl <- create_table(iris[1:10, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "PDF", font = "Courier") %>%
    page_header(paste0("Left and here is a really long left ",
                       "cell text to put it and more and more"), 
                "Right", width = 8) %>%
    add_content(tbl) %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("pdf2-58: Glue feature works.", {
  
  if (dev) {
    library(common)
    
    fp <- file.path(base_path, "pdf2/test58.pdf")
    
    tbl <- create_table(mtcars[1:10, ], borders = "outside") %>%
      spanning_header(1, 4, label = "My span{subsc('4')}") %>%
      define(mpg, label = "Mpg{subsc('3')}")
    
    #myvar <- "23"
    assign("myvar", "23", envir = .GlobalEnv)
    
    rpt <- create_report(fp, output_type = "PDF", font = "Courier") %>%
      page_header(c("Left {supsc('2')}really long left ",
                    "cell text to put it{supsc('3')} and more and more"), 
                  "Right{supsc('x')}") %>%
      add_content(tbl) %>%
      page_footer(c("left1{supsc('5')}", "left2{supsc('6')}"), "", 
                  "right{supsc('7')}") %>%
      titles("Table 1.0{supsc('1')}", "IRIS Data Frame{myvar}",
             blank_row = "below") %>%
      footnotes("Here is a footnote{subsc('a')}", "And another{subsc('9')}")
    
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  
  } else {
    
    expect_equal(TRUE, TRUE) 
    
  }
  
  
  
})



test_that("pdf2-59: Carriage return in label row works.", {
  
  
  fp <- file.path(base_path, "pdf2/test59.pdf")
  
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - \n21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  ll <- "Here is a super long label to \nsee if it can span the entire table."
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE, borders = c("all")) %>% 
    stub(c("var", "label")) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = ll, cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center", n = 19) %>% 
    define(B, label = "Group B", align = "center", n = 13)
  
  
  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "PDF",
                       font = "Times") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left", 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  
  
  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)
  
  
})

test_that("pdf2-60: Spanning headers borders work as expected with no title border.", {
  
  
  fp <- file.path(base_path, "pdf2/test60.pdf")
  
  dat <- mtcars[1:15, ]
  
  tbl <- create_table(dat, borders = c("top", "bottom")) %>%
    spanning_header(cyl, disp, "Span 1", label_align = "left") %>%
    spanning_header(hp, wt, "Span 2", underline = FALSE) %>%
    spanning_header(qsec, vs, "Span 3", n = 10) %>%
    spanning_header(drat, gear, "Super Duper\nWrapped Span", n = 11, level = 2) %>%
    titles("Table 1.0", "My Nice Table", borders = c("none"), blank_row = "none", 
           columns = 2) %>%
    footnotes("My footnote 1", "My footnote 2", borders = "none", blank_row = "none")
  
  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    page_header("Left", "Right") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl)
  
  
  res <- write_report(rpt)
  res
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})

# Basic Tests 61 - 70 ------------------------------------------------------

test_that("pdf2-61: Title columns work 1 column.", {
  
  fp <- file.path(base_path, "pdf2/test61.pdf")
  
  tbl <- create_table(iris[1:15, ], borders = "all")  %>%
    titles("Table 1.0\nsecond row", "IRIS Data Frame2",
           blank_row = "above", columns =  1, align = "center",
           borders = "outside") 
  
  rpt <- create_report(fp, output_type = "PDF", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("pdf2-62: Title columns work 2 columns.", {
  
  fp <- file.path(base_path, "pdf2/test62.pdf")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    titles("Table 1.0\nsecond row", "IRIS Data Frame", "Left", "Right", "mo\nre",
           blank_row = "below", columns =  2, borders = "all")
  
  rpt <- create_report(fp, output_type = "PDF", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right")  %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("pdf2-63: Title columns work 3 columns.", {
  
  fp <- file.path(base_path, "pdf2/test63.pdf")
  
  rght <- paste("Here is a big long text string to see how the automatic", 
                "wrapping is happing in a reduced size cell on the right.")
  
  
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE) %>%
    titles("Table 1.0\nsecond row", "IRIS Data Frame", 
           "      My right thing", "", "Center", rght,
           blank_row = "below", columns =  3, borders = "all")
  
  

  
  rpt <- create_report(fp, output_type = "PDF", font = "Courier", 
                       font_size = 10) %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%

    footnotes("Here is a footnote", "And another", "A",
      "Here is a longer footnote to see if I can figure out the alignment pattern.",
              align = "right")
  
  
  res <- write_report(rpt)
  res
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("pdf2-64: Multiple title blocks work as expected.", {
  
  fp <- file.path(base_path, "pdf2/test64.pdf")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE) %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "none", columns =  1, borders = "all") %>%
    titles("Table 2.0", "IRIS Data Frame2", "Left", "Right",
           blank_row = "none", columns =  2, borders = "all") %>%
    titles("Table 3.0", "IRIS Data Frame3", "My right thing", "", "Center",
           blank_row = "both", columns =  3, borders = "outside")
  
  rpt <- create_report(fp, output_type = "PDF", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})



test_that("pdf2-65: Custom page size works as expected.", {
  
  fp <- file.path(base_path, "pdf2/test65.pdf")
  
  tbl <- create_table(iris[1:15, ]) %>%
    define(Species, visible = FALSE)
  
  ttl <- c("Title1", "Title2", "Title3")
  
  rpt <- create_report(fp, output_type = "PDF", 
                       font = "Courier",
                       paper_size = c(6.5, 7.5),
                       orientation = "portrait") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles(ttl,
           blank_row = "below", columns =  1, align = "center",
           borders = "none") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("pdf2-66: Basic cell style bold works as expected.", {
  
  if (dev) {
    
    fp <- file.path(base_path, "pdf2/test66.pdf")
    
    dat <- mtcars[, 1:5]
    dat$hpflg <- ifelse(dat$hp > 100, TRUE, FALSE)
    
    
    tbl <- create_table(dat, width = 7) %>%
      titles("My title") %>%
      column_defaults(width = .75) %>%
      define(mpg, width = 1, style = cell_style(bold = TRUE)) %>%
      define(cyl, width = 1) %>%
      define(disp, style = cell_style(bold = TRUE)) %>%
      define(hp, style = cell_style(bold = TRUE, indicator = hpflg)) %>%
      define(hpflg, visible = FALSE) %>%
      footnotes("My footnotes", blank_row = "none")
    
    
    rpt <- create_report(fp, output_type = "PDF", 
                         font = "Arial", orientation = "portrait") %>%
      add_content(tbl) %>%
      footnotes("Here", footer = TRUE)
    
    res <- write_report(rpt)
    
    
    # file.show(res$modified_path)
    
    res$column_widths
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)
    
    
  } else {
    
    expect_equal(TRUE, TRUE)
    
  }
  
})

test_that("pdf2-67: Bolding works with stub.", {
  
  
  fp <- file.path(base_path, "pdf2/test67.pdf")
  
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  df$cylflg <- ifelse(df$var == "cyl", TRUE, FALSE)
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE) %>% 
    column_defaults(vars = c("stub", "A"), 
                    style = cell_style(bold = TRUE, indicator = cylflg)) %>%
    stub(c("var", "label"), 
         style = cell_style(bold = TRUE, indicator = "labelrow")) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center", n = 19,
           style = cell_style(bold = TRUE, indicator = cylflg)) %>% 
    define(B, label = "Group B", align = "center", n = 13, 
           style = cell_style(bold = TRUE, indicator = "datarow")) %>%
    define(cylflg, visible = FALSE)
  
  
  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "PDF",
                       font = "Times") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left", 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  
  
  res <- write_report(rpt)
  
  # file.show(res$modified_path)
  res
  expect_equal(file.exists(fp), TRUE)
  
  
})


test_that("pdf2-68: Bold cell style with column defaults.", {
  
  if (dev) {
    
    fp <- file.path(base_path, "pdf2/test68.pdf")
    
    dat <- mtcars[, 1:5]
    dat$hpflg <- ifelse(dat$hp > 100, TRUE, FALSE)
    
    
    tbl <- create_table(dat) %>%
      titles("My title") %>%
      column_defaults(width = .75, vars = c("cyl", "disp", "hp"),
                      style = cell_style(bold=TRUE, indicator = hpflg)) %>%
      define(mpg) %>%
      define(cyl) %>%
      define(disp) %>%
      define(hp) %>%
      define(hpflg, visible = FALSE) %>%
      footnotes("My footnotes", blank_row = "none")
    
    
    rpt <- create_report(fp, output_type = "PDF", 
                         font = "Arial", orientation = "portrait") %>%
      add_content(tbl) %>%
      footnotes("Here", footer = TRUE)
    
    res <- write_report(rpt)
    
    
    # file.show(res$modified_path)
    
    res$column_widths
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)
    
    
  } else {
    
    expect_equal(TRUE, TRUE)
    
  }
  
})


test_that("pdf2-69: Bolding, column defaults, and stub works.", {
  
  
  fp <- file.path(base_path, "pdf2/test69.pdf")
  
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  df$cylflg <- ifelse(df$var == "cyl", FALSE, TRUE)
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE) %>% 
    column_defaults(style = cell_style(bold = TRUE, indicator = cylflg)) %>%
    stub(c("var", "label"), width = 2,
         style = cell_style(bold = TRUE, indicator = cylflg)) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center", n = 19) %>% 
    define(B, label = "Group B", align = "center", n = 13, 
           style = cell_style(bold = TRUE, indicator = "datarow")) %>%
    define(cylflg, visible = FALSE)
  
  
  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "PDF",
                       font = "Times") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left", 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  res <- write_report(rpt)
  
  # file.show(res$modified_path)
  res
  expect_equal(file.exists(fp), TRUE)
  
  
})


test_that("pdf2-70: Spanning header bold work as expected.", {
  
  
  fp <- file.path(base_path, "pdf2/test70.pdf")
  
  dat <- mtcars[1:15, ]
  
  tbl <- create_table(dat, borders = c( "none")) %>%
    spanning_header(cyl, disp, "Span 1", label_align = "left") %>%
    spanning_header(hp, wt, "Span 2", underline = FALSE, bold = TRUE) %>%
    spanning_header(qsec, vs, "Span 3", n = 10) %>%
    spanning_header(drat, gear, "Super Duper\nWrapped Span",
                    n = 11, level = 2, bold = TRUE) %>%
    titles("Table 1.0", "My Nice Table", blank_row = "none", 
           borders = c("top", "bottom")) %>%
    footnotes("My footnote 1", "My footnote 2", 
              blank_row = "none", borders = c("top", "bottom"))
  
  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(res$modified_path), TRUE)
  expect_equal(res$pages, 1)
  
})

# Basic Tests 71 - 80 ------------------------------------------------------

test_that("pdf2-71: Italic footnotes work as expected.", {
  
  
  fp <- file.path(base_path, "pdf2/test71.pdf")
  
  dat <- mtcars[1:15, ]
  
  tbl <- create_table(dat, borders = c( "none")) %>%
    spanning_header(cyl, disp, "Span 1", label_align = "left") %>%
    spanning_header(hp, wt, "Span 2", underline = FALSE, bold = FALSE) %>%
    spanning_header(qsec, vs, "Span 3", n = 10) %>%
    spanning_header(drat, gear, "Super Duper\nWrapped Span",
                    n = 11, level = 2, bold = FALSE) %>%
    titles("Table 1.0", "My Nice Table", blank_row = "none", 
           borders = c("top", "bottom")) %>%
    footnotes("My italic footnote1", "My italic footnote2", italics = TRUE, 
              blank_row = "none", borders = "top") %>%
    footnotes("My italic footnote", "Page [pg] of [tpg]", 
              blank_row = "none", borders = c("bottom"))
  
  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(res$modified_path), TRUE)
  expect_equal(res$pages, 1)
  
})

test_that("pdf2-72: Stub indent.", {
  
  
  fp <- file.path(base_path, "pdf2/test72.pdf")
  
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder and more perhaps more" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  ll <- "Here is a super long label to see if it can span the entire table."
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE, borders = c("all")) %>% 
    stub(c("var", "label"), width = .9) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = ll, cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center", n = 19) %>% 
    define(B, label = "Group B", align = "center", n = 13)
  
  
  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "PDF",
                       font = "Times") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left", 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  
  
  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)
  
  
})



test_that("pdf2-73: Footnote columns work 1 column.", {
  
  fp <- file.path(base_path, "pdf2/test73.pdf")
  
  tbl <- create_table(iris[1:15, ], borders = "all")  %>%
    titles("Table 1.0\nsecond row", "IRIS Data Frame2",
           blank_row = "above", columns =  1, align = "center",
           borders = "outside") %>%
     footnotes("Table foot 1", "Table foot 2", 
               borders = c("all"), columns = 1, align = "right") %>%
     footnotes("Table foot 3", "Table foot 4", 
              borders = c("all"), columns = 2)
  
  rpt <- create_report(fp, output_type = "PDF", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    footnotes("Report foot 1", "Report foot 2", columns = 1, 
              borders = c("all", "bottom"), blank_row = "both", align = "right")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("pdf2-74: Footnote columns work 2 columns.", {
  
  fp <- file.path(base_path, "pdf2/test74.pdf")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    titles("Table 1.0\nsecond row", "IRIS Data Frame", "Left", "Right", "mo\nre",
           blank_row = "below", columns =  2, borders = "all") %>%
    footnotes("Table footnote 1", "Table footnote 2", "Table footnote 3",
              columns = 2, borders = "outside", blank_row = "both")
  
  rpt <- create_report(fp, output_type = "PDF", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right")  %>%
    footnotes("Report footnote 1", "Report footnote 2", columns = 1,
              borders = c("top", "bottom"))
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("pdf2-75: Footnote columns work 3 columns.", {
  
  fp <- file.path(base_path, "pdf2/test75.pdf")
  
  rght <- paste("Here is a big long text string to see how the automatic", 
                "wrapping is happing in a reduced size cell on the right.")
  
  
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE) %>%
    titles("Table 1.0\nsecond row", "IRIS Data Frame", 
           "      My right thing", "", "Center", rght,
           blank_row = "below", columns =  3, borders = "all") %>%
    footnotes("Table foot 1", "Table foot 2", "Table foot 3",
              borders = "all", columns = 3)
  
  
  
  
  rpt <- create_report(fp, output_type = "PDF", font = "Courier", 
                       font_size = 10) %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    footnotes("Report foot 1", "Report foot 2", "Report foot 3",
              "Report foot 4",  "Report foot 5",  
              columns = 3)
  
  
  res <- write_report(rpt)
  res
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("pdf2-76: Multiple footnote blocks work as expected.", {
  
  fp <- file.path(base_path, "pdf2/test76.pdf")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE) %>%
    titles("Here is a title", "And another", borders = "all") %>%
    footnotes("Footnote 1", "Footnote 2",
           blank_row = "above", columns =  1, borders = "all") %>%
    footnotes("Footnote 3 but make it really long so it wraps", 
              "Footnote 4", "Foot Left", "Foot Right",
           blank_row = "above", columns =  2, borders = "all") %>%
    footnotes("Footnote 5", "Footnote 6", "Footnote 7", "", "Footnote 8",
           blank_row = "both", columns =  3, borders = "all")
  
  rpt <- create_report(fp, output_type = "PDF", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    footnotes("Left 2 column footnote need to make it really long so it wraps",
              "Right 2 column footnote", columns = 2, borders = "none") %>%
    footnotes("Here is a report footnote", "And another report footnote", "more",
              columns = 3, borders = "all")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("pdf2-77: Page by with wrap works as expected.", {
  
  
  fp <- file.path(base_path, "pdf2/test77.pdf")
  
  dat <- iris
  dat$Pgby <- as.character(dat$Species)
  dat$Pgby <- paste0("Flower Type\n", dat$Pgby)
  
  
  tbl <- create_table(dat, borders = "none") %>% 
    titles("Table 1.0", "My Nice Report with a Page By", borders = "none") %>%
    page_by(Pgby, label = "Species: ", align = "left", borders = "none") %>%
    define(Pgby, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "PDF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_header("Left", "Right") %>% 
    page_footer("Left1", "Center1", "Right1") %>% 
    footnotes("My footnote 1", "My footnote 2", borders = "none")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 6)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  
})

test_that("pdf2-78: Page by with format works as expected.", {
  
  
  fp <- file.path(base_path, "pdf2/test78.pdf")
  
  fmt1 <- c(setosa = 1, versicolor = 2, virginica = 3)
  fmt2 <- value(condition(x == 1, "Setosa"),
                condition(x == 2, "Versicolor"),
                condition(x == 3, "Virginica"))
  
  dat <- iris
  fmtval <- fmt1[dat$Species]
  names(fmtval) <- NULL
  dat$Pgby <- fmtval
  
  tbl <- create_table(dat, borders = "none") %>% 
    titles("Table 1.0", "My Nice Report with a Page By", borders = "none") %>%
    page_by(Pgby, align = "left", label = "Flower:", borders = "none", format = fmt2) %>%
    define(Pgby, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "PDF", font = "Courier",
                       orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_header("Left", "Right") %>% 
    page_footer("Left1", "Center1", "Right1") %>% 
    footnotes("My footnote 1", "My footnote 2", borders = "none")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 6)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  
})

# Failing right now.  Reminder to fix wrap.rmd
test_that("pdf2-79: show_cols does not override define", {
  
  if (dev) {
    
    # Create temp file name
    fp <- file.path(base_path, "pdf2/test79.pdf")
    
    # Prepare data
    dat <- mtcars[1:10, ]
    dat <- data.frame(vehicle = rownames(dat), dat)
    
    # Define table
    tbl <- create_table(dat, show_cols = 1:8)  |> 
      define(vehicle, label = "Vehicle", width = 3, id_var = TRUE, align = "left")  |> 
      define(mpg, label = "Miles per Gallon", width = 1) |> 
      define(cyl, label = "Cylinders", format = "%.1f")  |> 
      define(disp, label = "Displacement")  |>  
      define(hp, label = "Horsepower", page_wrap = TRUE) |> 
      define(drat, visible = FALSE)  |> 
      define(wt, label = "Weight")  |> 
      define(qsec, label = "Quarter Mile Time", width = 1.5) 
    
    
    # Create the report
    rpt <- create_report(fp, output_type = "PDF", 
                         font = "Courier", font_size = 12)  |> 
      titles("Listing 2.0", "MTCARS Data Listing with Page Wrap")  |> 
      set_margins(top = 1, bottom = 1)  |> 
      add_content(tbl)  |> 
      page_footer(right = "Page [pg] of [tpg]")
    
    # Write the report
    res <- write_report(rpt)
    
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 2)
    expect_equal(length(res$column_widths[[1]]), 7)
    expect_equal("drat" %in% names(res$column_widths[[1]]), FALSE)
  
  } else {
    
    expect_equal(TRUE, TRUE) 
  }
  
})

test_that("pdf2-80: spanning header works on show_cols none and defined cols", {
  
  if (dev) {
    
    # Create temp file name
    fp <- file.path(base_path, "pdf2/test80.pdf")
    
    # Prepare data
    dat <- mtcars[1:10, ]
    dat <- data.frame(vehicle = rownames(dat), dat)
    
    # Define table
    tbl <- create_table(dat, show_cols = "none")  |> 
      spanning_header(mpg, qsec, label = "fork") |> 
      column_defaults(from = mpg, to = qsec, width = 1) |> 
      define(vehicle, label = "Vehicle", width = 3, id_var = TRUE, align = "left")  |> 
      define(mpg, label = "Miles per Gallon", width = 1) |> 
      define(cyl, label = "Cylinders", format = "%.1f")  |> 
      define(disp, label = "Displacement", width = 1.25)  |>  
      define(hp, label = "Horsepower", page_wrap = TRUE) |> 
      define(drat, visible = FALSE)  |> 
      define(wt, label = "Weight")  |> 
      define(qsec, label = "Quarter Mile Time", width = 1.5) 
    
    
    # Create the report
    rpt <- create_report(fp, output_type = "PDF", 
                         font = "Courier", font_size = 12)  |> 
      titles("Listing 2.0", "MTCARS Data Listing with Page Wrap")  |> 
      set_margins(top = 1, bottom = 1)  |> 
      add_content(tbl)  |> 
      page_footer(right = "Page [pg] of [tpg]")
    
    # Write the report
    res <- write_report(rpt)
    
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 2)
    expect_equal(length(res$column_widths[[1]]), 7)
    expect_equal("drat" %in% names(res$column_widths[[1]]), FALSE)
    
  } else {
    
    expect_equal(TRUE, TRUE) 
  }
  
})

# Basic Tests 81 - 90 ------------------------------------------------------

test_that("pdf2-81: Plot outside border works as expected.", {
  
  if (dev) {
    
    library(ggplot2)
    library(common)
    
    indata <- data.frame(
      VAR = c("ARM", "ARM", "ARM", "ARM"),
      CAT = c("ARM A", "ARM B", "ARM C", "ARM D"),
      CNT = c(20, 21, 21, 23)
    )
    
    common::labels(indata) <- setNames(
      list("", "ARM", "Frequency"),
      names(indata)
    )
    
    plt <- ggplot(data = indata, aes(x = CAT, y = CNT)) +
      geom_col(fill = "#0000A0") +
      geom_text(aes(label = CNT), vjust = 1.5, colour = "white") +
      labs(x = "Treatment Group", y = "Number of Subjects (n)")
    
    # Create temp file name
    fp <- file.path(base_path, "pdf2/test81.pdf")
    
    page1 <- create_plot(plt, 4.5, 7, borders = "outside") |> 
      titles("Figure 1.1", "Distribution of Subjects by Treatment Group", 
             bold = TRUE, font_size = 11)
    
    
    rpt <- create_report(fp, output_type = "PDF", font = "Arial") |> 
      set_margins(top = 1, bottom = 1) |> 
      page_header("Sponsor: Company", "Study: ABC") |> 
      add_content(page1) |> 
      footnotes("Program: DM_Figure.R") |> 
      page_footer(paste0("Date Produced: ", fapply(Sys.time(), "%d%b%y %H:%M")), 
                  right = "Page [pg] of [tpg]")
    
    
    # Write the report
    res <- write_report(rpt)
    
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)
  } else {
    
    expect_equal(TRUE, TRUE) 
  }
  
})

test_that("pdf2-82: Table with blank_before as expected.", {
  
  if (dev == TRUE) {
    
    
    fp <- file.path(base_path, "pdf2/test82.pdf")
    
    
    # Setup
    arm <- c(rep("A", 5), rep("B", 5))
    subjid <- 100:109
    name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
              "al-Kazemi, Najwa \nand more and more", "Schaffer, Ashley", "Laner, Tahma",
              "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
              "Smith, Shaileigh")
    sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
    age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
    
    
    # Create data frame
    df1 <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)
    
    
    tbl1 <- create_table(df1, first_row_blank = FALSE, borders = "all") %>%
      define(subjid, label = "Subject ID for a patient", n = 10, align = "left",
             width = 1) %>%
      define(name, label = "Subject Name", width = 1) %>%
      define(sex, label = "Sex", n = 10, align = "center") %>%
      define(age, label = "Age", n = 10) %>%
      define(arm, label = "Arm",
             blank_before = TRUE,
             dedupe = TRUE) %>% 
      titles("Table 1.0", align = "center", borders = "all") %>%
      footnotes("Here", borders = "all", blank_row = "none") 
    
    
    rpt <- create_report(fp, output_type = "PDF", font = "Arial",
                         font_size = fsz) %>%
      add_content(tbl1)
    
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
    
  } else 
    expect_equal(TRUE, TRUE)
  
  
  
})

test_that("pdf2-83: Three level stub and indentation work as expected.", {
  
  if (dev == TRUE) {
    fp <- file.path(base_path, "pdf2/test83.pdf")
    
    
    # Setup
    cat <- c(rep("Kaplan-Meier estimates", 6), 
             rep("Cox PH estimates and some more really long stuff", 6))
    grp <- c("25th percentile", "25th percentile", 
             "median (weeks)", "median (weeks)",
             "75th percentile", "75th percentile",
             "25th percentile", "25th percentile", 
             "median (weeks)", "median (weeks)",
             "75th percentile", "75th percentile")
    ci <- c(NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval")
    values <- c(41, 53, 43, 39, 47, 52, 38, 25, 37, 23, 78, 21)
    
    # Create data frame
    df <- data.frame(cat, grp, ci, values, stringsAsFactors = FALSE)
    
    tbl1 <- create_table(df) %>%
      stub(c(cat, grp, ci), "Estimates", width = 1.5) %>% 
      define(cat, label_row = TRUE, blank_after = TRUE, indent = 0.15) %>%
      define(grp, indent = .25) %>%
      define(ci, indent = .5) %>%
      define(values, label = "Values")
    
    rpt <- create_report(fp, output_type = "PDF", font = fnt,
                         font_size = fsz) %>%
      titles("Table 3.0", "Analysis of Time to Initial PSGA Success in Weeks") %>% 
      page_header("Sponsor", "Study") %>% 
      add_content(tbl1) %>% 
      page_footer("Time", "Confidential", "Page")
    
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_true(TRUE)
  }
})

test_that("pdf2-84: Page footers with one assigned width work as expected.",{
  if (dev) {
    fp <- file.path(base_path, "pdf2/test84.pdf")
    
    dat <- mtcars[,c("mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      footnotes("This is testing footnote", blank_row = "none")
    
    rpt <- create_report(fp, output_type = "PDF", font = fnt,
                         font_size = fsz) %>%
      set_margins(top = 1, bottom = 1) %>%
      titles("Table 1.0") %>%
      add_content(tbl) %>%
      page_footer(left = "this is a very long sentence whose length exceeds the limitation of the left part", 
                  center = "Center footer", right = "Right footer",
                  width = c(5.5))
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
}
)

test_that("pdf2-85: Page footers with multiple assigned widths work as expected.",{
  if (dev) {
    fp <- file.path(base_path, "pdf2/test85.pdf")
    
    dat <- mtcars[,c("mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      footnotes("This is testing footnote", blank_row = "none")
    
    rpt <- create_report(fp, output_type = "PDF", font = fnt,
                         font_size = fsz) %>%
      set_margins(top = 1, bottom = 1) %>%
      titles("Table 1.0") %>%
      add_content(tbl) %>%
      page_footer(left = "this is a very long sentence whose length exceeds the limitation of the left part", 
                  center = "", right = "Right Footer",
                  width = c(5.5, 0 , 2))
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
}
)

test_that("pdf2-86: Page by with bold label and value works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test86.pdf")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type:", bold = TRUE, blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-87: Page by with bold label works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test87.pdf")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type:", bold = "label", blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-88: Page by with bold value works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test88.pdf")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type:", bold = "value", blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-89: Page by with bold label and value at right works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test89.pdf")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type:", bold = TRUE, blank_row = "none",
              align = "right") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-90: Page by with bold label at right works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test90.pdf")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type:", bold = "label", blank_row = "none",
              align = "right") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

# Basic Tests 91 - 100 ------------------------------------------------------

test_that("pdf2-91: Page by with bold value at right works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test91.pdf")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type:", bold = "value", blank_row = "none",
              align = "right") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-92: Page by with bold label and value at center works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test92.pdf")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type:", bold = TRUE, blank_row = "none",
              align = "center") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-93: Page by with bold label at center works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test93.pdf")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type:", bold = "label", blank_row = "none",
              align = "center") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-94: Page by with bold value at center works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test94.pdf")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type:", bold = "value", blank_row = "none",
              align = "center") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-95: Page by with bold long label and value works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test95.pdf")
    
    dat <- iris
    long_string <- "This is long page by value which should take more than one sentence - "
    dat$Pgby <- paste0(long_string, dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    long_label <- "This is a long page by label which should take more than one sentence: "
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = long_label, bold = TRUE, blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-96: Page by with bold long label works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test96.pdf")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    long_label <- "This is a long page by label which should take more than one sentence: "
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = long_label, bold = "label", blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-97: Page by with bold long value works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test97.pdf")
    
    dat <- iris
    long_string <- "This is long page by value which should take more than one sentence - "
    dat$Pgby <- paste0(long_string, dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    long_label <- "This is a long page by label which should take more than one sentence:"
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type:", bold = "value", blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-98: Page by with bold long label and value at center worksas expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test98.pdf")
    
    dat <- iris
    long_string <- "This is long page by value which should take more than one sentence - "
    dat$Pgby <- paste0(long_string, dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    long_label <- "This is a long page by label which should take more than one sentence: "
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = long_label, bold = TRUE, blank_row = "none", align = "center") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-99: Page by with bold long label at center works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test99.pdf")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    long_label <- "This is a long page by label which should take more than one sentence: "
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = long_label, bold = "label", blank_row = "none", align = "center") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-100: Page by with bold long value at center works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test100.pdf")
    
    dat <- iris
    long_string <- "This is long page by value which should take more than one sentence - "
    dat$Pgby <- paste0(long_string, dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    long_label <- "This is a long page by label which should take more than one sentence:"
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type:", bold = "value", blank_row = "none", align = "center") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})
# Basic Tests 101 - 110 ------------------------------------------------------

test_that("pdf2-101: Page by with bold long label filling first line works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test101.pdf")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    long_label <- "This is a long page by label which should take more: "
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = long_label, bold = "label", blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-102: Group border works as as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "pdf2/test102.pdf")
    
    # Setup
    arm <- c(rep("A", 3), rep("B", 2), rep("C", 3), rep("D", 2))
    subjid <- 100:109
    name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
              "al-Kazemi, Najwa \nand more and more", "Schaffer, Ashley", "Laner, Tahma",
              "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
              "Smith, Shaileigh")
    sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
    age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
    
    
    # Create data frame
    df <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)
    df <- rbind(df, df, df, df)
    
    tbl1 <- create_table(df, first_row_blank = FALSE, borders = "outside") %>%
      define(subjid, label = "Subject ID for a patient", n = 10, align = "left",
             width = 1) %>%
      define(name, label = "Subject Name", width = 1) %>%
      define(sex, label = "Sex", n = 10, align = "center") %>%
      define(age, label = "Age", n = 10) %>%
      define(arm, label = "Arm",
             dedupe = TRUE,
             group_border = TRUE) %>%
      footnotes("This is the footnote")
    
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz) %>%
      titles(c("Table 1.0", "This is a table with group border"), align = "center") %>%
      add_content(tbl1) %>%
      footnotes(c("This is the footnote 1")) %>%
      page_header(left = "Test header", right = "Test header") %>%
      set_margins(top = 1, bottom = 1)
    
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-103: Group border with blank after works as as expected.", {
  
  if (dev == TRUE) {
    fp <- file.path(base_path, "pdf2/test103.pdf")
    
    # Setup
    arm <- c(rep("A", 3), rep("B", 2), rep("C", 3), rep("D", 2))
    subjid <- 100:109
    name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
              "al-Kazemi, Najwa \nand more and more", "Schaffer, Ashley", "Laner, Tahma",
              "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
              "Smith, Shaileigh")
    sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
    age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)    
    
    # Create data frame
    df <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)
    df <- rbind(df, df, df, df)
    
    tbl1 <- create_table(df, first_row_blank = FALSE, borders = "outside") %>%
      define(subjid, label = "Subject ID for a patient", n = 10, align = "left",
             width = 1) %>%
      define(name, label = "Subject Name", width = 1) %>%
      define(sex, label = "Sex", n = 10, align = "center") %>%
      define(age, label = "Age", n = 10) %>%
      define(arm, label = "Arm",
             dedupe = TRUE,
             blank_after = TRUE,
             group_border = TRUE) %>%
      footnotes("This is the footnote")
    
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz) %>%
      titles(c("Table 1.0", "This is a table with group border"), align = "center") %>%
      add_content(tbl1) %>%
      footnotes(c("This is the footnote 1")) %>%
      page_header(left = "Test header", right = "Test header") %>%
      set_margins(top = 1, bottom = 1)
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-104: Group border with table all borders works as as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "pdf2/test104.pdf")
    
    # Setup
    arm <- c(rep("A", 3), rep("B", 2), rep("C", 3), rep("D", 2))
    subjid <- 100:109
    name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
              "al-Kazemi, Najwa \nand more and more", "Schaffer, Ashley", "Laner, Tahma",
              "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
              "Smith, Shaileigh")
    sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
    age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
    
    
    # Create data frame
    df <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)
    df <- rbind(df, df, df, df)
    
    tbl1 <- create_table(df, first_row_blank = FALSE, borders = "all") %>%
      define(subjid, label = "Subject ID for a patient", n = 10, align = "left",
             width = 1) %>%
      define(name, label = "Subject Name", width = 1) %>%
      define(sex, label = "Sex", n = 10, align = "center") %>%
      define(age, label = "Age", n = 10) %>%
      define(arm, label = "Arm",
             dedupe = TRUE,
             group_border = TRUE) %>%
      footnotes("This is the footnote")
    
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz) %>%
      titles(c("Table 1.0", "This is a table with group border"), align = "center") %>%
      add_content(tbl1) %>%
      footnotes(c("This is the footnote 1")) %>%
      page_header(left = "Test header", right = "Test header") %>%
      set_margins(top = 1, bottom = 1)
    
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-105: Center page header works as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "pdf2/test105.pdf")
    
    dat <- iris[1:50,]
    dat <- dat[, c("Species" ,"Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") 
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with a center header", header = T) %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none") %>%
      page_header(left = "Left", center = "center", right = "right",
                  width = c(1, 4, 4)) %>%
      page_footer(left = "Left", center = "center", right = "right",
                  width = c(3, 4, 2))
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-106: Images in page header/footer work as expected.", {
  
  if (dev == TRUE) {
    fp <- file.path(base_path, "pdf2/test106.pdf")
    
    dat <- iris[1:50,]
    dat <- dat[, c("Species" ,"Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")]
    
    image_path <- file.path(base_path, "data/logo.jpg")
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a page header picture")
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none") %>%
      page_header() %>%
      header_image(image_path, height = 0.5, width = 0.8, align = "left") %>%
      header_image(image_path, height = 0.6, width = 0.85, align = "right") %>%
      header_image(c(image_path, image_path), height = 0.38, width = 0.7, align = "center") %>%
      page_footer() %>%
      footer_image(c(image_path, image_path), height = 0.38, width = 0.7, align = "left") %>%
      footer_image(image_path, height = 0.5, width = 0.8, align = "right") %>%
      footer_image(image_path, height = 0.6, width = 0.85, align = "center")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-107: Spanned images in page header/footer work as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "pdf2/test107.pdf")
    
    dat <- iris[1:50,]
    dat <- dat[, c("Species" ,"Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")]
    
    image_path <- file.path(base_path, "data/span_logo.jpg")
    
    tbl <- create_table(dat, borders = "outside") 
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with a page header picture", header = T) %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none") %>%
      page_header(width = c(0, 9, 0), left = "No display because width is 0") %>%
      header_image(image_path, height = 1, width = 7.5, align = "centre") %>%
      page_footer(width = c(0, 0, 9), center = "No display because width is 0") %>%
      footer_image(image_path, height = 0.8, width = 7, align = "right")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-108: Return error when header_image/footer_image is used without page_header/page_footer.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "pdf2/test108.pdf")
    
    dat <- iris[1:50,]
    dat <- dat[, c("Species" ,"Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")]
    
    image_path <- file.path(base_path, "data/span_logo.png")
    
    tbl <- create_table(dat, borders = "outside") 
    
    # page_header is not set
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with a page header picture", header = T) %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none") %>%
      header_image(image_path, height = 1, width = 7.5, align = "centre")
    
    expect_error(
      res <- write_report(rpt),
      "`page_header` must be used when using `header_image`."
    )
    
    # page_footer is not set
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with a page header picture", header = T) %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none") %>%
      footer_image(image_path, height = 0.8, width = 7, align = "right")
    
    expect_error(
      res <- write_report(rpt),
      "`page_footer` must be used when using `footer_image`."
    )
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-109: group_cohesion puts the group in same page as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "pdf2/test109.pdf")
    
    dat <- iris[1:91,]
    
    dat$test_group <- c(
      rep("A", 14),
      rep("B", 6),
      rep("C", 22),
      rep("D", 18),
      rep("E", 31)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 14), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 6), # 12 lines + 1
      c(rep("Flower C", 22)),
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E", 31)) # 31 lines + 1
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, group_cohesion = TRUE, label = "Group", blank_after = T)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    # page size 34.6
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-110: group_cohesion with min page proportion works as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "pdf2/test110.pdf")
    
    dat <- iris[1:95,]
    
    dat$test_group <- c(
      rep("A", 14),
      rep("B", 6),
      rep("C", 22),
      rep("D", 18),
      rep("E", 35)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 14), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 6), # 12 lines + 1
      c(rep("Flower C", 22)),
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E", 35)) # 35 lines + 1 -> Still break because more than page size 34.6
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, group_cohesion = 0.7, label = "Group", blank_after = T)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-111: Multiple group_cohesion work as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "pdf2/test111.pdf")
    
    dat <- iris[1:91,]
    
    dat$test_group <- c(
      rep("A", 14),
      rep("B", 6),
      rep("C", 22),
      rep("D", 18),
      rep("E", 31)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 14), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 3), # 12 lines + 1
      rep("Flower B\nSubgroup B2", 3),
      c(rep("Flower C1", 17)),
      c(rep("Flower C2", 5)),
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E1", 2)), # 31 lines + 1
      c(rep("Flower E2", 29)) # 31 lines + 1
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, group_cohesion = TRUE, label = "Group", blank_after = T) %>%
      define(test_string, group_cohesion = TRUE)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-112: Multiple group_cohesion with different cohesion strengths work as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "pdf2/test112.pdf")
    
    dat <- iris[1:91,]
    
    dat$test_group <- c(
      rep("A", 14),
      rep("B", 6),
      rep("C", 22),
      rep("D", 18),
      rep("E", 31)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 14), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 3), # 12 lines + 1
      rep("Flower B\nSubgroup B2", 3),
      c(rep("Flower C1", 17)),
      c(rep("Flower C2", 5)),
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E1", 8)), # 31 lines + 1
      c(rep("Flower E2", 23)) # 31 lines + 1
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, group_cohesion = TRUE, label = "Group", blank_after = T) %>%
      define(test_string, group_cohesion = 0.1)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    # C2 is not moved but E2 is moved.
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-113: break_label works as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "pdf2/test113.pdf")
    
    dat <- iris[1:86,]
    
    dat$test_group <- c(
      rep("A", 12),
      rep("B", 6),
      rep("C", 19),
      rep("D", 18),
      rep("E", 31)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 12), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 6), # 12 lines + 1
      c(rep("Flower C", 19)), # 19 lines + 1
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E", 31)) # 31 lines + 1
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, break_label = "(Continued)", blank_before = T) %>%
      define(test_string, 
             break_label = paste0("(This is a long text which should take more ",
                                  "than one line. It's to test if the package is able to handle this ",
                                  "situation.)"),
             indent = 0.16)
    
    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Break Label") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf2-114: Page numbers work every in title, footnote, header, and footer.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf2/test114.pdf")
    
    dat <- iris

    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table Title: Page [pg] of [tpg]") %>%
      footnotes("Table footnote: Page [pg] of [tpg]")

    rpt <- create_report(fp, output_type = "pdf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      add_content(tbl) %>%
      titles("Report Title: Page [pg] of [tpg]") %>%
      footnotes("Report footnote: Page [pg] of [tpg]", borders = "none") %>%
      page_header("Header: Page [pg] of [tpg]") %>%
      page_footer("Footer: Page [pg] of [tpg]")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

# # User Tests --------------------------------------------------------------

# Lots of special characters not working
test_that("pdf2-user1: demo table works.", {

  if (dev) {
    library(tidyr)
    library(dplyr)

    # Data Filepath
    dir_data <- file.path(data_dir, "data")

    fp <- file.path(base_path, "pdf2/user1.pdf")


    # Load Data
    data_demo   <- file.path(dir_data, "dm.csv") %>%
      read.csv()


    data_demo <- subset(data_demo, data_demo$ARM != "SCREEN FAILURE")


    sex_decode <- c("M" = "Male",
                    "F" = "Female")

    race_decode <- c("WHITE" = "White",
                     "BLACK OR AFRICAN AMERICAN" = "Black or African American",
                     "ASIAN" = "Asian or Pacific Islander",
                     "NATIVE AMERICAN" = "Native American",
                     "UNKNOWN" = "Unknown")

    arm_pop <- table(data_demo$ARM)


    demo_age <-
      data_demo %>%
      group_by(ARM) %>%
      summarise(across(.cols = AGE,
                       .fns = list(N      = ~ fmt_n(.),
                                   Mean   = ~ fmt_mean_sd(.),
                                   Median = ~ fmt_median(.),
                                   `Q1 - Q3` = ~ fmt_quantile_range(.),
                                   Range  = ~ fmt_range(.)
                       ))) %>%
      pivot_longer(-ARM,
                   names_to  = c("var", "label"),
                   names_sep = "_",
                   values_to = "value") %>%
      pivot_wider(names_from = ARM,
                  values_from = "value")



    demo_sex <-
      data_demo %>%
      add_count(ARM, SEX,  name = "n_SEX") %>%
      select(ARM, SEX, n_SEX) %>%
      distinct() %>%
      pivot_longer(cols = c(SEX),
                   names_to  = "var",
                   values_to = "label") %>%
      pivot_wider(names_from  = ARM,
                  values_from = n_SEX,
                  values_fill = 0) %>%
      mutate(label = factor(label, levels = names(sex_decode),
                            labels = sex_decode),
             `ARM A` = fmt_cnt_pct(`ARM A`, arm_pop["ARM A"]),
             `ARM B` = fmt_cnt_pct(`ARM B`, arm_pop["ARM B"]),
             `ARM C` = fmt_cnt_pct(`ARM C`, arm_pop["ARM C"]),
             `ARM D` = fmt_cnt_pct(`ARM D`, arm_pop["ARM D"]))



    demo_race <-
      data_demo %>%
      add_count(ARM, RACE, name = "n_RACE") %>%
      select(ARM, RACE, n_RACE) %>%
      distinct() %>%
      pivot_longer(cols = RACE,
                   names_to  = "var",
                   values_to = "label") %>%
      pivot_wider(names_from  = ARM,
                  values_from = n_RACE,
                  values_fill = 0) %>%
      mutate(label = factor(label, levels = names(race_decode),
                            labels = race_decode),
             `ARM A` = fmt_cnt_pct(`ARM A`, arm_pop["ARM A"]),
             `ARM B` = fmt_cnt_pct(`ARM B`, arm_pop["ARM B"]),
             `ARM C` = fmt_cnt_pct(`ARM C`, arm_pop["ARM C"]),
             `ARM D` = fmt_cnt_pct(`ARM D`, arm_pop["ARM D"])) %>%
      arrange(var, label)


    demo <- bind_rows(demo_age, demo_sex, demo_race)


    #View(demo)


    # Stub decode
    block_fmt <- c(AGE = "Age", SEX = "Sex", RACE = "Race")

    # Define table
    tbl <- create_table(demo, first_row_blank = TRUE) %>%
      column_defaults(from = "ARM A", to = "ARM D", width = 1.25) %>%
      define(var, blank_after = TRUE, dedupe = TRUE, blank_before = TRUE,
             format = block_fmt, label = "") %>%
      define(label, label = "") %>%
      define(`ARM A`, align = "center", label = "Placebo", n = 36) %>%
      define(`ARM B`, align = "center", label = "Drug 10mg", n = 38) %>%
      define(`ARM C`, align = "center", label = "Drug 20mg", n = 38) %>%
      define(`ARM D`, align = "center", label = "Competitor", n = 38)

    # Define Report
    rpt <- create_report(fp, output_type = "PDF", font = fnt, font_size = fsz) %>%
      set_margins(top = 1, bottom = 1) %>%
      options_fixed(font_size = 10) %>%
      titles("Table 14.1/4",
             "Demographics and Baseline to Characteristics",
             "Specify Population Ω µ β ¥ ∑ ≠ ≤ £ ∞ ؈ ლ  \Ub8a 鬼") %>%
      add_content(tbl) %>%
      footnotes("Special symbols \U221e to mess things up: Ω µ β ¥ ∑ ≠ ≤ £ ∞ ؈ ლ  \Ub8a 鬼") %>%
      footnotes("Special symbols µ Ω £ there to mess things up: ", "Page [pg] of [tpg]", 
                align = "left", italics = TRUE) %>%
      page_header("Left µ Ω £ ", "Right") %>%
      page_footer("Time µ Ω £ ", right = "Page [pg] of [tpg]")

    # Write out report
    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)



  } else
    expect_equal(TRUE, TRUE)


})

# Borders and spacing need some tweaking
test_that("pdf2-user2: demo table with stub works.", {


  if (dev) {

    library(tidyr)
    library(dplyr)



    # Data Filepath
    dir_data <- file.path(data_dir, "data")

    fp <- file.path(base_path, "pdf2/user2.pdf")


    # Load Data
    data_demo   <- file.path(dir_data, "dm.csv") %>%
      read.csv()

    data_demo <- subset(data_demo, data_demo$ARM != "SCREEN FAILURE")


    sex_decode <- c("M" = "Male",
                    "F" = "Female")

    race_decode <- c("WHITE" = "White",
                     "BLACK OR AFRICAN AMERICAN" = "Black or African American",
                     "ASIAN" = "Asian or Pacific Islander",
                     "NATIVE AMERICAN" = "Native American",
                     "UNKNOWN" = "Unknown")

    arm_pop <- table(data_demo$ARM)


    demo_age <-
      data_demo %>%
      group_by(ARM) %>%
      summarise(across(.cols = AGE,
                       .fns = list(N      = ~ fmt_n(.),
                                   Mean   = ~ fmt_mean_sd(.),
                                   Median = ~ fmt_median(.),
                                   `Q1 - Q3` = ~ fmt_quantile_range(.),
                                   Range  = ~ fmt_range(.)
                       ))) %>%
      pivot_longer(-ARM,
                   names_to  = c("var", "label"),
                   names_sep = "_",
                   values_to = "value") %>%
      pivot_wider(names_from = ARM,
                  values_from = "value")



    demo_sex <-
      data_demo %>%
      add_count(ARM, SEX,  name = "n_SEX") %>%
      select(ARM, SEX, n_SEX) %>%
      distinct() %>%
      pivot_longer(cols = c(SEX),
                   names_to  = "var",
                   values_to = "label") %>%
      pivot_wider(names_from  = ARM,
                  values_from = n_SEX,
                  values_fill = 0) %>%
      mutate(label = factor(label, levels = names(sex_decode),
                            labels = sex_decode),
             `ARM A` = fmt_cnt_pct(`ARM A`, arm_pop["ARM A"]),
             `ARM B` = fmt_cnt_pct(`ARM B`, arm_pop["ARM B"]),
             `ARM C` = fmt_cnt_pct(`ARM C`, arm_pop["ARM C"]),
             `ARM D` = fmt_cnt_pct(`ARM D`, arm_pop["ARM D"]))



    demo_race <-
      data_demo %>%
      add_count(ARM, RACE, name = "n_RACE") %>%
      select(ARM, RACE, n_RACE) %>%
      distinct() %>%
      pivot_longer(cols = RACE,
                   names_to  = "var",
                   values_to = "label") %>%
      pivot_wider(names_from  = ARM,
                  values_from = n_RACE,
                  values_fill = 0) %>%
      mutate(label = factor(label, levels = names(race_decode),
                            labels = race_decode),
             `ARM A` = fmt_cnt_pct(`ARM A`, arm_pop["ARM A"]),
             `ARM B` = fmt_cnt_pct(`ARM B`, arm_pop["ARM B"]),
             `ARM C` = fmt_cnt_pct(`ARM C`, arm_pop["ARM C"]),
             `ARM D` = fmt_cnt_pct(`ARM D`, arm_pop["ARM D"])) %>%
      arrange(var, label)


    demo <- bind_rows(demo_age, demo_sex, demo_race)


    #View(demo)


    # Stub decode
    block_fmt <- c(AGE = "Age", SEX = "Sex", RACE2 = "Race")

    # Define table
    tbl <- create_table(demo, first_row_blank = TRUE, borders = "all") %>%
      stub(c("var", "label"), width = 1.5) %>%
      column_defaults(width = 1) %>%
      spanning_header(`ARM B`, `ARM D`, "Treatments", n = 10) %>% 
      define(var, blank_after = TRUE,
             format = block_fmt, label = "", label_row = TRUE) %>%
      define(label, label = "", indent = .25) %>%
      define(`ARM A`, align = "left", label = "Placebo", n = 36) %>%
      define(`ARM B`, align = "center", label = "Drug 10mg", n = 38) %>%
      define(`ARM C`, align = "right", label = "Drug 20mg", n = 38) %>%
      define(`ARM D`, align = "center", label = "Competitor", n = 38) %>%
      titles("Table 14.1/4",
             "Demographics and Baseline Characteristics",
             "Specify Population", borders = c("all"), blank_row = "both",
             align = "center", font_size = 14, bold = TRUE) %>%
      footnotes("Here is a footnote", "Here is another footnote",
                borders = "outside", blank_row = "both", align = "right")

    # Define Report
    rpt <- create_report(fp, output_type = "PDF",
                         font = "Arial", font_size = 10) %>%
      add_content(tbl, align = "right") %>%
      page_header("Sponsor", "Drug") %>%
      page_footer(left = "Time", right = "Page [pg] of [tpg]") #%>%
    #page_by(var = "var", label = "Variable: ")

    # Write out report
    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)


  } else
    expect_equal(TRUE, TRUE)

})

test_that("pdf2-user3: listings works.", {
  if (dev == TRUE) {
    # Data Filepath
    dir_data <- file.path(data_dir, "data")

    fp <- file.path(base_path, "pdf2/user3.pdf")

    # Removing to make last page exactly equal to available rows on page.
    # In this case, any added blank rows should be skipped.
    fil <- c("ABC-14-124",
             "ABC-15-153",
             "ABC-15-154",
             "ABC-15-155",
             "ABC-15-156",
             "ABC-16-045",
             "ABC-16-046",
             "ABC-16-047",
             "ABC-16-157",
             "ABC-16-158",
             "ABC-16-159",
             "ABC-16-160")

    # Load Data
    data_demo   <- file.path(dir_data, "dm.csv") %>%
      read.csv()

    data_demo <- data_demo[!data_demo$USUBJID %in% fil, ]


    # Test that any assigned formats are applied
    attr(data_demo$SUBJID, "width") <- 1
    attr(data_demo$SUBJID, "justify") <- "left"
    attr(data_demo$SUBJID, "format") <- "S:%s"
    #print(widths(data_demo))
    names(data_demo)
    # Define table
    tbl <- create_table(data_demo) %>%
      define(USUBJID, id_var = TRUE)


    # Define Report
    rpt <- create_report(fp, font = "Arial", font_size = 12,
                         orientation = "portrait") %>%
      titles("Listing 1.0",
             "Demographics Dataset") %>%
      add_content(tbl, align = "left") %>%
      page_header("Sponsor", "Drug") %>%
      page_footer(left = "Time", right = "Page [pg] of [tpg]") %>%
      footnotes("My footnotes")

    stm <- Sys.time()
    #Write out report
    res <- write_report(rpt, output_type = "PDF")
    
    etm <- Sys.time()
    etm - stm
    
    expect_equal(file.exists(fp), TRUE)


  } else
    expect_equal(TRUE, TRUE)


})


test_that("pdf2-user4: listing in cm and times works.", {
  if (dev == TRUE) {
    # Data Filepath
    dir_data <- file.path(data_dir, "data")

    fp <- file.path(base_path, "pdf2/user4.pdf")

    # Removing to make last page exactly equal to available rows on page.
    # In this case, any added blank rows should be skipped.
    fil <- c("ABC-14-124",
             "ABC-15-153",
             "ABC-15-154",
             "ABC-15-155",
             "ABC-15-156",
             "ABC-16-045",
             "ABC-16-046",
             "ABC-16-047",
             "ABC-16-157",
             "ABC-16-158",
             "ABC-16-159",
             "ABC-16-160")

    # Load Data
    data_demo   <- file.path(dir_data, "dm.csv") %>%
      read.csv()

    data_demo <- data_demo[!data_demo$USUBJID %in% fil, ]


    # Test that any assigned formats are applied
    attr(data_demo$SUBJID, "width") <- 2.54
    attr(data_demo$SUBJID, "justify") <- "left"
    attr(data_demo$SUBJID, "format") <- "S:%s"
    #print(widths(data_demo))

    # Define table
    tbl <- create_table(data_demo) %>%
      define(USUBJID, id_var = TRUE)


    # Define Report
    rpt <- create_report(fp, font = "Times", font_size = 10, units = "cm") %>%
      titles("Listing 1.0",
             "Demographics Dataset") %>%
      add_content(tbl, align = "left") %>%
      page_header("Sponsor", "Drug") %>%
      page_footer(left = "Time", right = "Page [pg] of [tpg]") %>%
      footnotes("My footnote")

    #Write out report
    res <- write_report(rpt, output_type = "PDF")

    #print(res$column_widths)

    expect_equal(file.exists(fp), TRUE)


  } else
    expect_equal(TRUE, TRUE)


})

test_that("pdf2-user5: Portrait in 12pt Arial works as expected.", {

  if (dev == TRUE) {

    dir_data <- file.path(data_dir, "data")

    fp <- file.path(base_path, "pdf2/user5.pdf")

    # Read in prepared data
    df <- read.table(header = TRUE, text = '
        var     label        A             B
        "ampg"   "N"          "19"          "13"
        "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)"
        "ampg"   "Median"     "16.4"        "21.4"
        "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
        "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
        "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)"
        "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)"
        "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')

    # Create table
    tbl <- create_table(df, first_row_blank = TRUE) %>%
      define(var, label = "Variable", blank_after = TRUE, dedupe = TRUE,
             format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>%
      define(label, label = "") %>%
      define(A, label = "Group A", align = "center", n = 19) %>%
      define(B, label = "Group B", align = "center", n = 13)


    # Create report and add content
    rpt <- create_report(fp, orientation = "portrait", output_type = "PDF",
                         font = "Arial", font_size = 12) %>%
      page_header(left = "Client: Motor Trend", right = "Study: Cars") %>%
      titles("Table 1.0", "MTCARS Summary Table") %>%
      add_content(tbl) %>%
      footnotes("* Motor Trend, 1974") %>%
      page_footer(left = Sys.time(),
                  center = "Confidential",
                  right = "Page [pg] of [tpg]")

    # Write out report
    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)

  } else
    expect_equal(TRUE, TRUE)

})


test_that("pdf2-user6: User program works as expected for PDF.", {
  
  if (dev == TRUE) {
    
    fp <- file.path(base_path, "pdf2/user6.pdf")
    dir_data <- file.path(data_dir, "data/S")
    
    dat <- readRDS(file.path(dir_data, "for_david_to_debug.RDS"))
    
    
    
    tbl <- create_table(dat, borders = "outside", first_row_blank = TRUE) |> 
      titles("Table 14-2.2.1.2. Baseline Demographics", "(Safety Analysis Set)", 
             bold = TRUE, font_size = 11) |> 
      stub(c("row_label1", "row_label2"), width = 3.4) |> 
      define(row_label1, label_row = TRUE, blank_after = TRUE) |> 
      define(row_label2, indent = .25) |> 
      define(var1_1, label = "Cohort 1\n0.001 mg\nQW\n(N=11)\nn(%)") |> 
      define(var1_2, label = "Cohort 2\n0.003 mg\nQW\n(N=11)\nn(%)") |> 
      define(var1_3, label = "Cohort 3\n0.001 mg\nQW\n(N=11)\nn(%)") |> 
      define(var1_4, label = "Cohort 4\n0.003 mg\nQW\n(N=11)\nn(%)") |> 
      define(var1_5, label = "Cohort 5\n0.1 mg\nQW\n(N=11)\nn(%)") |> 
      define(var1_6, label = "Cohort 6\n0.3 mg\nQW\n(N=11)\nn(%)") |> 
      define(var1_Total, label = "Part 1\nNo Step\nDosing\nTotal\n(N=46)\nn(%)") |> 
      define(ord_layer_index, visible = FALSE) |> 
      define(paging_column, page_break = TRUE, visible = FALSE) |> 
      column_defaults(width = .825, align = "center") |> 
      footnotes("Page [pg] of [tpg]", align = "right", blank_row = "none") |> 
      footnotes("Data cutoff date: 07DEC2022. N = Number of subjects in the analysis set. n = Number of subjects with observed data. SD = Standard Deviation, Q1 = First Quartile, Q3 = Third Quartile. S = Step Dosing.",
                "Safety analysis set includes all subjects that are enrolled and receive at least 1 dose of AMG 509.",
                paste("Cohort 7a = C1D1: 0.1 mg, C1D8: 0.3 mg, C1D15: 0.3 mg, C1D22: 0.3 mg; Cohort 7b = C1D1: 0.1 mg, C1D8: 0.3 mg, C1D15: 1 mg, C1D22: 1 mg; Cohort 7c =",
                      "C1D1: 0.1 mg, C1D8: 0.3 mg, C1D15: 1 mg; Cohort 8 = C1D1: 0.3 mg, C1D8: 1 mg, C1D15: 1 mg, C1D22: 1 mg; Cohort 9 = C1D1: 0.1 mg, C1D8: 0.3 mg, C1D15:",
                      "0.75 mg, C1D22: 0.75 mg; Cohort 10 = C1D1: 0.1 mg, C1D8: 1 mg, C1D15: 1 mg, C1D22: 1 mg; Cohort 11 = C1D1: 0.1 mg, C1D8: 0.3 mg, C1D15: 1 mg,",
                      "C1D22:1.5 mg; Cohort 12 = C1D1: 0.1 mg, C1D8: 0.3 mg, C1D15: 0.75 mg, C1D22: 1.5 mg; Cohort 13 = C1D1: 0.1 mg, C1D8: 0.3 mg, C1D15: 1 mg, C1D22: 2 mg."),
                "Two subjects received reduced dose on C1D8 in Cohort 10.", blank_row = "none") |>
      footnotes("Program: /userdata/stat/amg509/onc/20180146/analysis/mantl_pilot/tables/sme_table_project/t-dm-base-demo-saf-sd.sas",
                "Output: t14-02-002-001-002-dm-base-demo-saf-sd.rtf (Date Generated:31MAR23:13:37:59) Source: adam.adsl", blank_row = "above")
    
    
    rpt <- create_report(fp, output_type = "PDF", 
                         font = "Arial", font_size = 9) |> 
      add_content(tbl, align = "left") |> 
      set_margins(top = 1.5, bottom = 1, left = .7, right = 1)
    
    
    res <- write_report(rpt)
    
   # file.show(res$modified_path)
    
    #res$column_widths
    
    expect_equal(file.exists(res$modified_path), TRUE)
  
  } else {
    
    expect_equal(TRUE, TRUE)
    
  } 
  
})

context("HTML Tests")

base_path <- paste0(getwd(),"/tests/testthat")
data_dir <- base_path

base_path <- tempdir()
data_dir <- "."

cnt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
              "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
              "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
              "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ", 
              "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
              "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
              "qui officia deserunt mollit anim id est laborum.")

fnt <- "Arial"
fsz <- 10

dev <- FALSE

# Basic Tests -------------------------------------------------------------


test_that("html1: Basic table works as expected.", {
  
  
  fp <- file.path(base_path, "html/test1.html")
 # print(fp)
  
  dat <- mtcars[1:15, ]
  attr(dat[[2]], "label") <- "Cylin."
  attr(dat[[2]], "width") <- 1
  attr(dat[[2]], "justify") <- "center"
  
  tbl <- create_table(dat, borders = "outside", first_row_blank = TRUE) %>%
    titles("Table 1.0", "My Nice Table", borders = c("outside"), 
           width = "content") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "outside", 
              align = "left", width = "content") %>% 
    define(wt, width = 1, label = "Weight", align = "center", 
           label_align = "right")
  
  rpt <- create_report(fp, output_type = "HTML", font = "Arial",
                       font_size = 12, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), 
                blank_row = "below") %>% 
    add_content(tbl, align = "center")  %>% 
    page_footer("Left1", "Center1", "Right1")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  
})

test_that("html2: Basic table with title header works as expected.", {
  
  
  fp <- file.path(base_path, "html/test2.html")
  
  dat <- mtcars[1:15, ]
  attr(dat[[2]], "label") <- "Cylin."
  attr(dat[[2]], "width") <- 1
  
  tbl <- create_table(dat, borders = c("outside")) %>%
    title_header("Table 1.0", "My Nice Table", right = "Right", 
                 borders = c("outside"), blank_row = "both",
           width = "content") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "outside", 
              align = "left", width = "content")
  
  rpt <- create_report(fp, output_type = "HTML", font = "Arial",
                       font_size = 12, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), 
                blank_row = "below") %>%
    add_content(tbl)  %>%
    page_footer("Left1", "Center1", "Right1")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})

test_that("html3: Spanning headers work as expected.", {
  
  
  fp <- file.path(base_path, "html/test3")
  
  dat <- mtcars[1:15, ]
  
  tbl <- create_table(dat, borders = c("all")) %>%
    spanning_header(cyl, disp, "Span 1", label_align = "left") %>% 
    spanning_header(hp, wt, "Span 2", underline = FALSE) %>%
    spanning_header(qsec, vs, "Span 3", n = 10) %>%
    spanning_header(drat, gear, "Super Duper\nWrapped Span", 
                    n = 11, level = 2) %>%
    titles("Table 1.0", "My Nice Table", borders = c("outside")) %>%
    footnotes("My footnote 1", "My footnote 2", borders = c("outside")) 
  
  rpt <- create_report(fp, output_type = "HTML", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    add_content(tbl) %>% 
    page_footer("Left1", "Center1", "Right1")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(res$modified_path), TRUE)
  expect_equal(res$pages, 1)
  
})


test_that("html4: Multi page table works as expected.", {


  fp <- file.path(base_path, "html/test4.html")

  dat <- iris


  tbl <- create_table(dat, borders = "none") %>%
    titles("Table 1.0", "My Nice Irises", "Another Title") %>%
    define(Sepal.Length, label = "Sepal Length", width = 1.5, align = "center") %>%
    define(Sepal.Width, label = "Sepal Width", width = 1.25, align = "centre") %>%
    define(Species, blank_after = TRUE)

  rpt <- create_report(fp, output_type = "HTML", font = "Arial",
                       font_size = 12, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1")) %>%
    add_content(tbl, blank_row = "none") %>%
    page_footer("Left1", "Center1", "Page [pg] of [tpg]") %>%
    footnotes("My footnote 1", "My footnote 2")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 7)


})



test_that("html5: Basic text works as expected.", {

  fp <- file.path(base_path, "html/test5")

  txt <- create_text(cnt, width = 6, borders = "outside", align = "center") %>%
    titles("Text 1.0", "My Nice Text", borders = "outside", width = "content") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "outside")

  rpt <- create_report(fp, output_type = "HTML", font = "Courier",
                       font_size = 12, paper_size = "letter") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(txt, align = "center") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)

  expect_equal(file.exists(res$modified_path), TRUE)
  expect_equal(res$pages, 1)

})


# Note that image may be deleted in next test case
test_that("html6: Basic plot works as expected.", {


  library(ggplot2)

  fp <- file.path(base_path, "html/test6.html")

  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()

  plt <- create_plot(p, height = 4, width = 8, borders = c("all")) %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", borders = "outside",
           font_size = 12) %>%
    footnotes("* Motor Trend, 1974", borders = "outside")


  rpt <- create_report(fp, output_type = "HTML", font = fnt, font_size =fsz) %>%
    page_header("Client", "Study: XYZ") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt, align = "center") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")


  res <- write_report(rpt)

  #print(res)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

})


test_that("remove_image_files works as expected.", {
  
  if (dev & FALSE) {
    pth <- file.path(base_path, "html/test6.html")
    
    
    remove_image_files(pth)
    
    # Hard to test. Will just check for error.
    # And use this test interactively.
    expect_equal(TRUE, TRUE)
  } else 
    expect_equal(TRUE, TRUE)
})


test_that("html7: Multi page table paper_size none works as expected.", {
  
  
  fp <- file.path(base_path, "html/test7.html")
  
  dat <- iris
  
  
  tbl <- create_table(dat, borders = "none") %>%
    titles("Table 1.0", "My Nice Irises", "Another Title") %>%
    define(Sepal.Length, label = "Sepal Length", width = 1.5, align = "center") %>%
    define(Sepal.Width, label = "Sepal Width", width = 1.25, align = "centre") %>%
    define(Species, blank_after = TRUE)
  
  rpt <- create_report(fp, output_type = "HTML", font = "Times",
                       font_size = 9, orientation = "portrait",
                       paper_size = "none") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1")) %>%
    add_content(tbl, blank_row = "none") %>%
    page_footer("Left1", "Center1", "Page [pg] of [tpg]") %>%
    footnotes("My footnote 1", "My footnote 2")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  
})

# Good for testing
# Borders throw off line counts.  Made rudimentary adjustment.
test_that("html8: Page by works as expected.", {
  
  
  fp <- file.path(base_path, "html/test8")
  
  dat <- iris
  
  tbl <- create_table(dat, borders = "all") %>%
    titles("Table 1.0", "My Nice Irises", "Another Title", 
           borders = "outside") %>%
    page_by(Species, label = "Species: ", borders = "outside", 
            align = "center") %>% 
    define(Sepal.Length, label = "Sepal Length", width = 1.5, align = "center") %>%
    define(Sepal.Width, label = "Sepal Width", width = 1.25, align = "centre")  %>%
    define(Species, visible = FALSE) %>% 
    footnotes("My footnote 1", "My footnote 2", borders = "outside")
  
  rpt <- create_report(fp, output_type = "HTML", font = "Arial",
                       font_size = 12, orientation = "portrait",
                       paper_size = "letter") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1")) %>%
    add_content(tbl, blank_row = "none") %>%
    page_footer("Left1", "Center1", "Page [pg] of [tpg]")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(res$modified_path), TRUE)
  expect_equal(res$pages, 6)
  
  
})


# Good for testing
test_that("html9: Page by on report works as expected.", {
  
  
  fp <- file.path(base_path, "html/test9.html")
  
  dat <- iris
  
  
  tbl <- create_table(dat, borders = "none") %>%
    titles("Table 1.0", "My Nice Irises", "Another Title") %>%
    define(Sepal.Length, label = "Sepal Length", width = 1.5, align = "center") %>%
    define(Sepal.Width, label = "Sepal Width", width = 1.25, align = "centre") 
  
  rpt <- create_report(fp, output_type = "HTML", font = "Arial",
                       font_size = 12, orientation = "landscape",
                       paper_size = "letter") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1")) %>%
    page_by(Species, label = "Species: ") %>% 
    add_content(tbl, blank_row = "none") %>%
    page_footer("Left1", "Center1", "Page [pg] of [tpg]") %>%
    footnotes("My footnote 1", "My footnote 2")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 9)
  
  
})

test_that("html10: Title Header and page header/footer wrapping work as expected.", {
  
  
  fp <- file.path(base_path, "html/test10.html")
  
  dat <- iris[1:10, ] 
  
  tbl <- create_table(dat, borders = "none") %>% 
    title_header("Table 1.0", "My Nice Report with Borders",
                 right = c("Right1", "Right2", 
                           "Right3 long enough to wrap around at least once"),
                 borders = "none",
                 blank_row = "none") %>%
    footnotes("My footnote 1", 
        paste("My footnote 2 Center1 here is a whole bunch of stuff to try and make it wrap", 
              "like more down here note 2 Center1 here is a whole bunch of stuff." ),
              valign = "top",
              borders = "none", 
              blank_row = "above") %>% 
    define(Sepal.Width, label = "Here is a rather long header label")
  
  rpt <- create_report(fp, output_type = "HTML", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_header(c("Left1", "Left2\nwrap"), "Right 1") %>% 
    page_footer("Left1", 
                "Center1 here is a whole bunch of stuff to try and make it wrap", 
                "Right1\nwrap\n and wrap again")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  
})


test_that("html11: Basic plot with titles on report works as expected.", {
  
  
  library(ggplot2)
  
  fp <- file.path(base_path, "html/test11.html")
  
  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  
  plt <- create_plot(p, height = 4, width = 8, borders = c("none"))
  
  
  rpt <- create_report(fp, output_type = "HTML", font = fnt, font_size =fsz) %>%
    page_header("Client", "Study: XYZ") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt, align = "center") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", borders = "none") %>%
    footnotes("* Motor Trend, 1974", borders = "none")
  
  
  res <- write_report(rpt)
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})


test_that("html12: Text with titles on report works as expected.", {
  
  fp <- file.path(base_path, "html/test12")
  
  txt <- create_text(cnt, width = 6, borders = "none", align = "center") 
  
  rpt <- create_report(fp, output_type = "HTML", font = "Courier",
                       font_size = 12, paper_size = "letter") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(txt, align = "center") %>%
    page_footer("Left1", "Center1", "Right1")%>%
    titles("Text 1.0", "My Nice Text", borders = "none", width = "page") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "none")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(res$modified_path), TRUE)
  expect_equal(res$pages, 1)
  
})


# Works on all combinations on font and font size.  Needed adjustments to row height.
test_that("html13: Table with break between sections works as expected.", {
  
  
  fp <- file.path(base_path, "html/test13.html")
  
  
  # Setup
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma", 
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon", 
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
  arm <- c(rep("A", 5), rep("B", 5))
  
  # Create data frame
  df <- data.frame(subjid, name, sex, age, arm)
  
  
  tbl1 <- create_table(df, first_row_blank = TRUE) %>%
    define(subjid, label = "Subject ID", align = "left", width = 1) %>% 
    define(name, label = "Subject Name", width = 1) %>% 
    define(sex, label = "Sex") %>% 
    define(age, label = "Age") %>% 
    define(arm, label = "Arm", 
           blank_after = FALSE, 
           dedupe = TRUE, 
           align = "right") #%>% 
  # spanning_header(sex, arm, label = "Here is a spanning header")
  
  
  rpt <- create_report(fp, output_type = "HTML", font = fnt, font_size = fsz) %>%
    page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
    # options_fixed(line_count = 46) %>% 
    titles("Table 1.0", "Analysis Data Subject Listing\n And more stuff", 
           "Safety Population", align = "center", bold = TRUE) %>%
    footnotes("Program Name: table1_0.R", 
              "Here is a big long footnote that is going to wrap\n at least once") %>%
    page_footer(left = "Time", center = "Confidential", 
                right = "Page [pg] of [tpg]") %>%
    add_content(tbl1) 
  
  
  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)
  
  
})



test_that("html14: Plot with page by on plot works as expected.", {
  
  library(ggplot2)
  
  fp <- file.path(base_path, "html/test14.html")
  
  
  fmt <- value(condition(x == 4, "4 Cylinder"),
               condition(x == 6, "6 Cylinder"),
               condition(x == 8, "8 Cylinder"))
  
  dat <- mtcars[order(mtcars$cyl), ]
  
  p <- ggplot(dat, aes(x=disp, y=mpg)) + geom_point()
  
  brdrs <- "outside"
  
  #dats <- split(p$data, p$data$grp)
  #tbl <- create_table(dat[1:3, ])
  
  plt <- create_plot(p, height = 4, width = 8, borders = brdrs) %>% 
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", 
           borders = brdrs, 
           blank_row = "none") %>%
    page_by(cyl, "Cylinders: ", borders = brdrs, format = fmt) %>% 
    footnotes("* Motor Trend, 1974", borders = brdrs) 
  
  rpt <- create_report(fp, output_type = "HTML", font = fnt, font_size = fsz) %>%
    page_header("Client", "Study: XYZ") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt) %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 3)
  
  
})

test_that("html15: Title bold and font size works as expected.", {
  
  
  fp <- file.path(base_path, "html/test15.html")
  
  dat <- mtcars[1:15, ]
  attr(dat[[2]], "label") <- "Cylin."
  attr(dat[[2]], "width") <- 1
  attr(dat[[2]], "justify") <- "center"
  
  tbl <- create_table(dat, borders = "outside") %>%
    titles("Table 1.0", "My Nice Table", borders = c("none"), 
           width = "content", font_size = 14, bold = TRUE) %>%
    footnotes("My footnote 1", "My footnote 2", borders = "none", 
              align = "left", width = "content") %>% 
    define(wt, width = 1, label = "Weight", align = "center", 
           label_align = "right")
  
  rpt <- create_report(fp, output_type = "HTML", font = "Arial",
                       font_size = 9, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), 
                blank_row = "below") %>% 
    add_content(tbl, align = "center")  %>% 
    page_footer("Left1", "Center1", "Right1")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})


test_that("html16: 9 pt font inches works as expected.", {
  
  
  fp <- file.path(base_path, "html/test16.html")
  
  rpt <- create_report(fp, output_type = "HTML", font_size = 9, 
                       font = "Courier",
                       orientation = "portrait") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>% 
    set_margins(top = 1, bottom = 1)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  
})

test_that("html17: 9 pt font cm works as expected.", {
  
  
  fp <- file.path(base_path, "html/test17.html")
  
  rpt <- create_report(fp, output_type = "HTML", font_size = 9, 
                       font = "Courier",
                       orientation = "portrait") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>% 
    set_margins(top = 1, bottom = 1)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  
})

test_that("html18: 11 pt font inches works as expected.", {
  
  
  fp <- file.path(base_path, "html/test18.html")
  
  rpt <- create_report(fp, output_type = "HTML", font_size = 11, 
                       font = "Courier",
                       orientation = "portrait") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>% 
    set_margins(top = 1, bottom = 1)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  
})

test_that("html19: 11 pt font cm works as expected.", {
  
  
  fp <- file.path(base_path, "html/test19.html")
  
  rpt <- create_report(fp, output_type = "HTML", font_size = 11, 
                       font = "Courier",
                       orientation = "portrait") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>% 
    set_margins(top = 1, bottom = 1)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  
})


test_that("html20: JPG Image file works as expected.", {
  
  library(ggplot2)
  
  fp <- file.path(base_path, "html/test20.html")
  
  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  
  pltpath <- file.path(base_path, "html/test20.jpg")
  ggsave(pltpath, width = 8, height = 4, 
         units = "in",
         dpi = 300)
  
  plt <- create_plot(pltpath, height = 4, width = 8)
  
  
  rpt <- create_report(fp, output_type = "HTML", font = "Arial") %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt, align = "center") %>%
    footnotes("* Motor Trend, 1974") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  
})


test_that("html21: Header_bold works as expected", {
  
  dat <- mtcars[1:10, 1:3]
  
  fp <- file.path(base_path, "html/test21.html")
  
  tbl <- create_table(dat, header_bold = TRUE, borders = "all") %>%
    column_defaults(width = 1) %>%
    titles("Report 1.0", "Simple Report", borders = "outside", 
           blank_row = "none", bold = TRUE) %>%
    footnotes("My footnote", blank_row = "none")
  
  rpt <- create_report(fp, orientation = "portrait",
                       output_type = "HTML", font = "Arial") %>%
    add_content(tbl) %>%
    set_margins(top = 1)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})



test_that("html22: Label row is one cell.", {
  
  
  fp <- file.path(base_path, "html/test22.html")
  
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  ll <- "Here is a super long label to see if it can span the entire table."
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE, borders = "all", 
                      header_bold = TRUE) %>% 
    stub(c("var", "label")) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = ll, cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center", n = 19) %>% 
    define(B, label = "Group B", align = "center", n = 13)
  
  
  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "HTML",
                       font = "Times") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = Sys.time(), 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  
  
  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)
  
  
})


test_that("html23: Blank after on invisible column.", {
  
  fp <- file.path(base_path, "html/test23.html")
  
  tbl <- create_table(iris, borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "HTML") %>%
    page_header("Left", "Right") %>%
    add_content(tbl) %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("html24: Page header width works.", {
  
  fp <- file.path(base_path, "html/test24.html")
  
  tbl <- create_table(iris[1:10, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "HTML", font = "Courier") %>%
    page_header("Left and here is a really long left cell text to put it", 
                "Right", width = 8) %>%
    add_content(tbl) %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("html25: Carriage return in label row works.", {
  
  
  fp <- file.path(base_path, "html/test25.html")
  
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - \n21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  ll <- "Here is a super long label to \nsee if it can span the entire table."
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE, borders = c("all")) %>% 
    stub(c("var", "label")) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = ll, cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center", n = 19) %>% 
    define(B, label = "Group B", align = "center", n = 13)
  
  
  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "HTML",
                       font = "Times") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left", 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  
  
  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)
  
  
})



test_that("html26: Title columns work 1 column.", {
  
  fp <- file.path(base_path, "html/test26.html")
  
  tbl <- create_table(iris[1:15, ], borders = "all")  %>%
    titles("Table 1.0\nsecond row", "IRIS Data Frame2",
           blank_row = "both", columns =  1, align = "center",
           borders = c("outside")) 
  
  rpt <- create_report(fp, output_type = "HTML", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("html27: Title columns work 2 columns.", {
  
  fp <- file.path(base_path, "html/test27.html")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    titles("Table 1.0\nsecond row", "IRIS Data Frame", "Left", "Right",
           blank_row = "both", columns =  2, borders = "all")
  
  rpt <- create_report(fp, output_type = "HTML", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right")  %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("html28: Title columns work 3 columns.", {
  
  fp <- file.path(base_path, "html/test28.html")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)
  
  
  rght <- paste("Here is a big long text string to see how the automatic", 
                "wrapping is happening in a reduced size cell on the right.")
  
  rpt <- create_report(fp, output_type = "HTML", font = "Courier", 
                       font_size = 10) %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0\nsecond row", "IRIS Data Frame", 
           "      My right thing", "", "Center", rght,
           blank_row = "below", columns =  3, borders = "none") %>%
    footnotes("Here is a footnote", "And another", "A",
              "Here is a longer footnote to see if I can figure out the alignment pattern.",
              align = "right")
  
  
  res <- write_report(rpt)
  res
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("html29: Multiple title blocks work as expected.", {
  
  fp <- file.path(base_path, "html/test29.html")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "HTML", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below", columns =  1, align = "center", width = 7,
           borders = "all") %>%
    titles("Table 2.0", "IRIS Data Frame2", "Left", "Right",
           blank_row = "below", columns =  2, borders = "all") %>%
    titles("Table 3.0", "IRIS Data Frame3", "My right thing", "", "Center",
           blank_row = "below", columns =  3, borders = "all") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("html30: Custom page size works as expected.", {
  
  fp <- file.path(base_path, "html/test30.html")
  
  tbl <- create_table(iris[1:15, ]) %>%
    define(Species, visible = FALSE)
  
  ttl <- c("Title1", "Title2", "Title3")
  
  rpt <- create_report(fp, output_type = "HTML", 
                       font = "Courier",
                       paper_size = c(6.5, 7.5),
                       orientation = "portrait") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles(ttl,
           blank_row = "below", columns =  1, align = "center",
           borders = "none") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})



test_that("html31: Basic cell style bold works as expected.", {
  
  if (dev) {
    
    fp <- file.path(base_path, "html/test31.html")
    
    dat <- mtcars[, 1:5]
    dat$hpflg <- ifelse(dat$hp > 100, TRUE, FALSE)
    
    
    tbl <- create_table(dat, width = 7) %>%
      titles("My title") %>%
      column_defaults(width = .75) %>%
      define(mpg, width = 1, style = cell_style(bold = TRUE)) %>%
      define(cyl, width = 1) %>%
      define(disp, style = cell_style(bold = TRUE)) %>%
      define(hp, style = cell_style(bold = TRUE, indicator = hpflg)) %>%
      define(hpflg, visible = FALSE) %>%
      footnotes("My footnotes", blank_row = "none")
    
    
    rpt <- create_report(fp, output_type = "HTML", 
                         font = "Arial", orientation = "portrait") %>%
      add_content(tbl) %>%
      footnotes("Here", footer = TRUE)
    
    res <- write_report(rpt)
    
    
    # file.show(res$modified_path)
    
    res$column_widths
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)
    
    
  } else {
    
    expect_equal(TRUE, TRUE)
    
  }
  
})

test_that("html32: Bolding works with stub.", {
  
  
  fp <- file.path(base_path, "html/test32.html")
  
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  df$cylflg <- ifelse(df$var == "cyl", TRUE, FALSE)
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE) %>% 
    column_defaults(vars = c("stub", "A"), 
                    style = cell_style(bold = TRUE, indicator = cylflg)) %>%
    stub(c("var", "label"), 
         style = cell_style(bold = TRUE, indicator = "labelrow")) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center", n = 19,
           style = cell_style(bold = TRUE, indicator = cylflg)) %>% 
    define(B, label = "Group B", align = "center", n = 13, 
           style = cell_style(bold = TRUE, indicator = "datarow")) %>%
    define(cylflg, visible = FALSE)
  
  
  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "HTML",
                       font = "Times") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left", 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  
  
  res <- write_report(rpt)
  
  #file.show(res$modified_path)
  res
  expect_equal(file.exists(fp), TRUE)
  
  
})


test_that("html33: Bold cell style with column defaults.", {
  
  if (dev) {
    
    fp <- file.path(base_path, "html/test33.html")
    
    dat <- mtcars[, 1:5]
    dat$hpflg <- ifelse(dat$hp > 100, TRUE, FALSE)
    
    
    tbl <- create_table(dat) %>%
      titles("My title") %>%
      column_defaults(width = .75, vars = c("cyl", "disp", "hp"),
                      style = cell_style(bold=TRUE, indicator = hpflg)) %>%
      define(mpg) %>%
      define(cyl) %>%
      define(disp) %>%
      define(hp) %>%
      define(hpflg, visible = FALSE) %>%
      footnotes("My footnotes", blank_row = "none")
    
    
    rpt <- create_report(fp, output_type = "HTML", 
                         font = "Arial", orientation = "portrait") %>%
      add_content(tbl) %>%
      footnotes("Here", footer = TRUE)
    
    res <- write_report(rpt)
    
    
    # file.show(res$modified_path)
    
    res$column_widths
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)
    
    
  } else {
    
    expect_equal(TRUE, TRUE)
    
  }
  
})


test_that("html34: Bolding, column defaults, and stub works.", {
  
  
  fp <- file.path(base_path, "html/test34.html")
  
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  df$cylflg <- ifelse(df$var == "cyl", FALSE, TRUE)
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE) %>% 
    column_defaults(style = cell_style(bold = TRUE, indicator = cylflg)) %>%
    stub(c("var", "label"), width = 2,
         style = cell_style(bold = TRUE, indicator = cylflg)) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center", n = 19) %>% 
    define(B, label = "Group B", align = "center", n = 13, 
           style = cell_style(bold = TRUE, indicator = "datarow")) %>%
    define(cylflg, visible = FALSE)
  
  
  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "HTML",
                       font = "Times") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left", 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  res <- write_report(rpt)
  
  # file.show(res$modified_path)
  res
  expect_equal(file.exists(fp), TRUE)
  
  
})


test_that("html35: Spanning header bold work as expected.", {
  
  
  fp <- file.path(base_path, "html/test35.html")
  
  dat <- mtcars[1:15, ]
  
  tbl <- create_table(dat, borders = c( "none")) %>%
    spanning_header(cyl, disp, "Span 1", label_align = "left") %>%
    spanning_header(hp, wt, "Span 2", underline = FALSE, bold = TRUE) %>%
    spanning_header(qsec, vs, "Span 3", n = 10) %>%
    spanning_header(drat, gear, "Super Duper\nWrapped Span",
                    n = 11, level = 2, bold = TRUE) %>%
    titles("Table 1.0", "My Nice Table", blank_row = "none", 
           borders = c("top", "bottom")) %>%
    footnotes("My footnote 1", "My footnote 2", 
              blank_row = "none", borders = c("top", "bottom"))
  
  rpt <- create_report(fp, output_type = "HTML", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(res$modified_path), TRUE)
  expect_equal(res$pages, 1)
  
})


test_that("html36: Italic footnotes work as expected.", {
  
  
  fp <- file.path(base_path, "html/test36.html")
  
  dat <- mtcars[1:15, ]
  
  tbl <- create_table(dat, borders = c( "outside")) %>%
    spanning_header(cyl, disp, "Span 1", label_align = "left") %>%
    spanning_header(hp, wt, "Span 2", underline = FALSE, bold = FALSE) %>%
    spanning_header(qsec, vs, "Span 3", n = 10) %>%
    spanning_header(drat, gear, "Super Duper\nWrapped Span",
                    n = 11, level = 2, bold = FALSE) %>%
    titles("Table 1.0", "My Nice Table", blank_row = "none", 
           borders = c("top", "bottom")) %>%
    footnotes("My italic footnote1", "My italic footnote2", italics = TRUE, 
              blank_row = "none", borders = "top") %>%
    footnotes("My italic footnote", "My footnote 2", 
              blank_row = "none", borders = c("bottom"))
  
  rpt <- create_report(fp, output_type = "HTML", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(res$modified_path), TRUE)
  expect_equal(res$pages, 1)
  
})


test_that("html37: Stub indent.", {
  
  
  fp <- file.path(base_path, "html/test37.html")
  
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder and more perhaps more" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  ll <- "Here is a super long label to see if it can span the entire table."
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE, borders = c("all")) %>% 
    stub(c("var", "label"), width = .9) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = ll, cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center", n = 19) %>% 
    define(B, label = "Group B", align = "center", n = 13)
  
  
  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "HTML",
                       font = "Times") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left", 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  
  
  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)
  
  
})



test_that("html38: Footnotes columns work 1 column.", {
  
  fp <- file.path(base_path, "html/test38.html")
  
  tbl <- create_table(iris[1:15, ], borders = "all")  %>%
    titles("Table 1.0\nsecond row", "IRIS Data Frame2",
           blank_row = "both", columns =  1, align = "center",
           borders = c("outside")) %>%
    footnotes("Here is a footnote", "And another", 
              borders = "all", columns = 1, blank_row = "both")
  
  rpt <- create_report(fp, output_type = "HTML", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    footnotes("Here is a footnote", "And another", columns = 1)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("html39: Footnote columns work 2 columns.", {
  
  fp <- file.path(base_path, "html/test39.html")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    titles("Table 1.0\nsecond row", "IRIS Data Frame", "Left", "Right",
           blank_row = "both", columns =  2, borders = "all") %>%
    footnotes("Here is a footnote", "And another", 
              borders = "all", columns = 2, blank_row = "both")
  
  rpt <- create_report(fp, output_type = "HTML", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right")  %>%
    footnotes("Here is a footnote", "And another", 
              "footnote left", "footnote right", columns = 2)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("html40: Footnote columns work 3 columns.", {
  
  fp <- file.path(base_path, "html/test40.html")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = FALSE, visible = FALSE) %>%
    footnotes("Here is a footnote", "And another", "And more",
              "",  "centered",
              borders = "all", columns = 3, blank_row = "both")
  
  
  rght <- paste("Here is a big long text string to see how the automatic", 
                "wrapping is happening in a reduced size cell on the right.")
  
  rpt <- create_report(fp, output_type = "HTML", font = "Courier", 
                       font_size = 10) %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0\nsecond row", "IRIS Data Frame", 
           "      My right thing", "", "Center", rght,
           blank_row = "below", columns =  3, borders = "none") %>%
    footnotes("Here is a footnote", "And another", "A",
              "Here is a longer footnote",  "to see if I can figure",  
              "out the alignment pattern.", columns = 3)
  
  
  res <- write_report(rpt)
  res
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("html41: Multiple footnote blocks work as expected.", {
  
  fp <- file.path(base_path, "html/test41.html")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = FALSE, visible = FALSE)  %>%
    footnotes("Here is a footnote", "And another", "And more",
              "",  "centered",
              borders = "all", columns = 3, blank_row = "both")
  
  rpt <- create_report(fp, output_type = "HTML", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    footnotes("Table 1.0", "IRIS Data Frame",
           blank_row = "below", columns =  1, align = "center", width = 7,
           borders = "all") %>%
    footnotes("Table 2.0", "IRIS Data Frame2", "Left", "Right",
           blank_row = "below", columns =  2, borders = "all") %>%
    footnotes("Table 3.0", "IRIS Data Frame3", "My right thing", "", "Center",
           blank_row = "below", columns =  3, borders = "all") %>%
    titles("Here is a title", "And another", borders = "all")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("html42: Page by with wrap works as expected.", {
  
  
  fp <- file.path(base_path, "html/test42.html")
  
  dat <- iris
  dat$Pgby <- as.character(dat$Species)
  dat$Pgby <- paste0("Flower Type\n", dat$Pgby)
  
  
  tbl <- create_table(dat, borders = "none") %>% 
    titles("Table 1.0", "My Nice Report with a Page By", borders = "none") %>%
    page_by(Pgby, label = "Species: ", align = "right", borders = "none") %>%
    define(Pgby, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "HTML", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_header("Left", "Right") %>% 
    page_footer("Left1", "Center1", "Right1") %>% 
    footnotes("My footnote 1", "My footnote 2", borders = "none")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 6)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  
})

test_that("html43: Page by with format works as expected.", {
  
  
  fp <- file.path(base_path, "html/test43.html")
  
  fmt1 <- c(setosa = 1, versicolor = 2, virginica = 3)
  fmt2 <- value(condition(x == 1, "Setosa"),
                condition(x == 2, "Versicolor"),
                condition(x == 3, "Virginica"))
  
  dat <- iris
  fmtval <- fmt1[dat$Species]
  names(fmtval) <- NULL
  dat$Pgby <- fmtval
  
  tbl <- create_table(dat, borders = "none") %>% 
    titles("Table 1.0", "My Nice Report with a Page By", borders = "none") %>%
    page_by(Pgby, align = "left", label = "Flower:", borders = "none", format = fmt2) %>%
    define(Pgby, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "HTML", 
                       orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_header("Left", "Right") %>% 
    page_footer("Left1", "Center1", "Right1") %>% 
    footnotes("My footnote 1", "My footnote 2", borders = "none")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 6)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  
})

test_that("html-44: Multi page table removes blank spaces.", {
  
  if (dev == TRUE) {
    
    
    fp <- file.path(base_path, "html/test44.html")
    
    dat1 <- iris[1:10, ]
    dat2 <- iris[11:20, ]
    dat3 <- iris[21:30, ]
    
    
    tbl1 <- create_table(dat1, borders = "none") %>%
      titles("Table 1.0", "My Nice Irises1") 
    
    tbl2 <- create_table(dat2, borders = "none") %>%
      titles("Table 1.0", "My Nice Irises2")
    
    tbl3 <- create_table(dat3, borders = "none") %>%
      titles("Table 1.0", "My Nice Irises3")
    
    rpt <- create_report(fp, output_type = "HTML", font = fnt,
                         font_size = 12, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl1, blank_row = "none") |> 
      add_content(tbl2, blank_row = "none") |> 
      add_content(tbl3, blank_row = "none") 
    
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 3)
    
  } else
    expect_equal(TRUE, TRUE)
})

test_that("html-45: Spanning header gap works as expected.", {
  
  if (dev == TRUE) {
    fp <- file.path(base_path, "html/test45.html")
    
    dat <- mtcars[1:15, ]
    
    tbl <- create_table(dat, borders = c("outside")) %>%
      spanning_header(cyl, disp, "Span 1", label_align = "left") %>%
      spanning_header(hp, wt, "Span 2", underline = TRUE) %>%
      spanning_header(qsec, vs, "Span 3", n = 10) %>%
      spanning_header(cyl, hp, "Super Span", n = 11, level = 2) |> 
      spanning_header(drat, gear, "Super Duper\nWrapped Span", n = 11, level = 2)
    
    rpt <- create_report(fp, output_type = "HTML", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
      titles("Table 1.0", "My Nice Table") %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2") %>%
      page_footer("Left1", "Center1", "Right1")
    
    res <- write_report(rpt)
    res
    res$column_widths
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)
    
  } else {
    expect_true(TRUE)
  }
  
})

test_that("html-45-2: Spanning header gap should not exist with all border.", {
  
  if (dev == TRUE) {
    
    fp <- file.path(base_path, "html/test45-2.html")
    
    dat <- mtcars[1:15, ]
    
    tbl <- create_table(dat, borders = c("all")) %>%
      spanning_header(cyl, disp, "Span 1", label_align = "left") %>%
      spanning_header(hp, wt, "Span 2", underline = TRUE) %>%
      spanning_header(qsec, vs, "Span 3", n = 10) %>%
      spanning_header(cyl, hp, "Super Span", n = 11, level = 2) |> 
      spanning_header(drat, gear, "Super Duper\nWrapped Span", n = 11, level = 2)
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
      titles("Table 1.0", "My Nice Table") %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2") %>%
      page_footer("Left1", "Center1", "Right1")
    
    res <- write_report(rpt)
    res
    res$column_widths
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)
    
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-45-3: Spanning header gap should not exist with inside border.", {
  
  if (dev == TRUE) {
    
    fp <- file.path(base_path, "html/test45-3.html")
    
    dat <- mtcars[1:15, ]
    
    tbl <- create_table(dat, borders = c("inside")) %>%
      spanning_header(cyl, disp, "Span 1", label_align = "left") %>%
      spanning_header(hp, wt, "Span 2", underline = TRUE) %>%
      spanning_header(qsec, vs, "Span 3", n = 10) %>%
      spanning_header(cyl, hp, "Super Span", n = 11, level = 2) |> 
      spanning_header(drat, gear, "Super Duper\nWrapped Span", n = 11, level = 2)
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
      titles("Table 1.0", "My Nice Table") %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2") %>%
      page_footer("Left1", "Center1", "Right1")
    
    res <- write_report(rpt)
    res
    res$column_widths
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)
    
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-46-4: Spanning header with invisible columns works as expected.", {
  
  if (dev == TRUE) {
    
    fp <- file.path(base_path, "html/test46-4.html")
    
    dat <- mtcars[1:20,]
    
    tbl <- create_table(dat, borders = c("all")) %>%
      spanning_header(mpg, qsec, "Span 1", label_align = "left") %>%
      spanning_header(vs, carb, "Span 2", underline = TRUE) %>%
      define(cyl, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
      titles("Table 1.0", "My Nice Table") %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2") %>%
      page_footer("Left1", "Center1", "Right1")
    
    # Make sure spanning header width is matched.
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)
    
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-46: Table with blank_before works as expected.", {
  
  if (dev == TRUE) {
    fp <- file.path(base_path, "html/test46.html")
    
    
    # Setup
    subjid <- 100:109
    name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
              "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma", 
              "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon", 
              "Smith, Shaileigh")
    sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
    age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
    arm <- c(rep("A", 5), rep("B", 5))
    
    # Create data frame
    df <- data.frame(subjid, name, sex, age, arm)
    
    
    tbl1 <- create_table(df, first_row_blank = FALSE) %>%
      define(subjid, label = "Subject ID", align = "left", width = 1) %>% 
      define(name, label = "Subject Name", width = 1) %>% 
      define(sex, label = "Sex") %>% 
      define(age, label = "Age") %>% 
      define(arm, label = "Arm", 
             blank_before = TRUE, 
             dedupe = TRUE, 
             align = "right") #%>% 
    # spanning_header(sex, arm, label = "Here is a spanning header")
    
    
    rpt <- create_report(fp, output_type = "HTML", font = fnt, font_size = fsz) %>%
      page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
      # options_fixed(line_count = 46) %>% 
      titles("Table 1.0", "Analysis Data Subject Listing\n And more stuff", 
             "Safety Population", align = "center", bold = TRUE) %>%
      footnotes("Program Name: table1_0.R", 
                "Here is a big long footnote that is going to wrap\n at least once") %>%
      page_footer(left = "Time", center = "Confidential", 
                  right = "Page [pg] of [tpg]") %>%
      add_content(tbl1) 
    
    
    res <- write_report(rpt)
    res
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_true(TRUE)
  }
})

test_that("html-47: Three level stub and indentation work as expected.", {
  
  if (dev == TRUE) {
    fp <- file.path(base_path, "html/test47.html")
    
    
    # Setup
    cat <- c(rep("Kaplan-Meier estimates", 6), 
             rep("Cox PH estimates and some more really long stuff", 6))
    grp <- c("25th percentile", "25th percentile", 
             "median (weeks)", "median (weeks)",
             "75th percentile", "75th percentile",
             "25th percentile", "25th percentile", 
             "median (weeks)", "median (weeks)",
             "75th percentile", "75th percentile")
    ci <- c(NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval")
    values <- c(41, 53, 43, 39, 47, 52, 38, 25, 37, 23, 78, 21)
    
    # Create data frame
    df <- data.frame(cat, grp, ci, values, stringsAsFactors = FALSE)
    
    tbl1 <- create_table(df) %>%
      stub(c(cat, grp, ci), "Estimates", width = 1.5) %>% 
      define(cat, label_row = TRUE, blank_after = TRUE, indent = 0.15) %>%
      define(grp, indent = .25) %>%
      define(ci, indent = .5) %>%
      define(values, label = "Values")
    
    rpt <- create_report(fp, output_type = "HTML", font = fnt,
                         font_size = fsz) %>%
      titles("Table 3.0", "Analysis of Time to Initial PSGA Success in Weeks") %>% 
      page_header("Sponsor", "Study") %>% 
      add_content(tbl1) %>% 
      page_footer("Time", "Confidential", "Page")
    
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_true(TRUE)
  }
})

test_that("html-48: Page footers with one assigned width work as expected.",{
  if (dev) {
    fp <- file.path(base_path, "html/test48.html")
    
    dat <- mtcars[,c("mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      footnotes("This is testing footnote", blank_row = "none")
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz) %>%
      set_margins(top = 1, bottom = 1) %>%
      titles("Table 1.0") %>%
      add_content(tbl) %>%
      page_footer(left = "this is a very long sentence whose length exceeds the limitation of the left part", 
                  center = "Center footer", right = "Right footer",
                  width = c(5.5))
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
}
)

test_that("html-49: Page footers with multiple assigned widths work as expected.",{
  if (dev) {
    fp <- file.path(base_path, "html/test49.html")
    
    dat <- mtcars[,c("mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      footnotes("This is testing footnote", blank_row = "none")
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz) %>%
      set_margins(top = 1, bottom = 1) %>%
      titles("Table 1.0") %>%
      add_content(tbl) %>%
      page_footer(left = "this is a very long sentence whose length exceeds the limitation of the left part", 
                  center = "", right = "Right Footer",
                  width = c(5.5, 0 , 2))
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
}
)

test_that("html-50: Page by with bold label and value works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "html/test50.html")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type: ", bold = TRUE, blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-51: Page by with bold label as expected.", {
  
  if (dev) {
    # Label and Value are bold
    fp <- file.path(base_path, "html/test51.html")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type: ", bold = "label", blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-52: Page by with bold value as expected.", {
  
  if (dev) {
    # Label and Value are bold
    fp <- file.path(base_path, "html/test52.html")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type: ", bold = "value", blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-53: Page by with bold long label and value works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "html/test53.html")
    
    dat <- iris
    long_string <- "This is long page by value which should take more than one sentence - "
    dat$Pgby <- paste0(long_string, dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    long_label <- "This is a long page by label which should take more than one sentence: "
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = long_label, bold = TRUE, blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-54: Page by with bold long label works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "html/test54.html")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    long_label <- "This is a long page by label which should take more than one sentence: "
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = long_label, bold = "label", blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-55: Page by with bold long value works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "html/test55.html")
    
    dat <- iris
    long_string <- "This is long page by value which should take more than one sentence - "
    dat$Pgby <- paste0(long_string, dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    long_label <- "This is a long page by label which should take more than one sentence:"
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type:", bold = "value", blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-56: Page by with bold long label filling one line works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "html/test56.html")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    long_label <- "This is a long page by label which should take more- "
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = long_label, bold = "label", blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-57: Group border works as as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "html/test57.html")
    
    # Setup
    arm <- c(rep("A", 3), rep("B", 2), rep("C", 3), rep("D", 2))
    subjid <- 100:109
    name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
              "al-Kazemi, Najwa \nand more and more", "Schaffer, Ashley", "Laner, Tahma",
              "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
              "Smith, Shaileigh")
    sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
    age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
    
    
    # Create data frame
    df <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)
    df <- rbind(df, df, df, df)
    
    tbl1 <- create_table(df, first_row_blank = FALSE, borders = "outside") %>%
      define(subjid, label = "Subject ID for a patient", n = 10, align = "left",
             width = 1) %>%
      define(name, label = "Subject Name", width = 1) %>%
      define(sex, label = "Sex", n = 10, align = "center") %>%
      define(age, label = "Age", n = 10) %>%
      define(arm, label = "Arm",
             dedupe = TRUE,
             group_border = TRUE) %>%
      footnotes("This is the footnote")
    
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz) %>%
      titles(c("Table 1.0", "This is a table with group border"), align = "center") %>%
      add_content(tbl1) %>%
      footnotes(c("This is the footnote 1")) %>%
      page_header(left = "Test header", right = "Test header") %>%
      page_footer(left = "Test footer", right = "Test footer") %>%
      set_margins(top = 1, bottom = 1)
    
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-58: Group border with blank after works as as expected.", {
  
  if (dev == TRUE) {
    fp <- file.path(base_path, "html/test58.html")
    
    # Setup
    arm <- c(rep("A", 3), rep("B", 2), rep("C", 3), rep("D", 2))
    subjid <- 100:109
    name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
              "al-Kazemi, Najwa \nand more and more", "Schaffer, Ashley", "Laner, Tahma",
              "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
              "Smith, Shaileigh")
    sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
    age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)    
    
    # Create data frame
    df <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)
    df <- rbind(df, df, df, df)
    
    tbl1 <- create_table(df, first_row_blank = FALSE, borders = "outside") %>%
      define(subjid, label = "Subject ID for a patient", n = 10, align = "left",
             width = 1) %>%
      define(name, label = "Subject Name", width = 1) %>%
      define(sex, label = "Sex", n = 10, align = "center") %>%
      define(age, label = "Age", n = 10) %>%
      define(arm, label = "Arm",
             dedupe = TRUE,
             blank_after = TRUE,
             group_border = TRUE) %>%
      footnotes("This is the footnote")
    
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz) %>%
      titles(c("Table 1.0", "This is a table with group border"), align = "center") %>%
      add_content(tbl1) %>%
      footnotes(c("This is the footnote 1")) %>%
      page_header(left = "Test header", right = "Test header") %>%
      page_footer(left = "Test footer", right = "Test footer") %>%
      set_margins(top = 1, bottom = 1)
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-59: Center page header works as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "html/test59.html")
    
    dat <- iris[1:50,]
    dat <- dat[, c("Species" ,"Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") 
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with a center header", header = T) %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none") %>%
      page_header(left = "l", center = "center", right = "right",
                  width = c(1, 4, 4)) %>%
      page_footer(left = "Left", center = "center", right = "right",
                  width = c(3, 4, 2))
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-60: Images in page header/footer work as expected.", {
  
  if (dev == TRUE) {
    fp <- file.path(base_path, "html/test60.html")
    
    dat <- iris[1:50,]
    dat <- dat[, c("Species" ,"Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")]
    
    image_path <- file.path(base_path, "data/logo.jpg")
    image_path2 <- file.path(base_path, "data/logo.png")
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a page header picture")
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none") %>%
      page_header() %>%
      header_image(image_path, height = 0.5, width = 0.8, align = "left") %>%
      header_image(image_path, height = 0.6, width = 0.85, align = "right") %>%
      header_image(c(image_path, image_path2), height = 0.38, width = 0.7, align = "center") %>%
      page_footer() %>%
      footer_image(c(image_path, image_path), height = 0.38, width = 0.7, align = "left") %>%
      footer_image(image_path, height = 0.5, width = 0.8, align = "right") %>%
      footer_image(image_path, height = 0.6, width = 0.85, align = "center")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-61: Spanned images in page header/footer work as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "html/test61.html")
    
    dat <- iris[1:50,]
    dat <- dat[, c("Species" ,"Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")]
    
    image_path <- file.path(base_path, "data/span_logo.png")
    
    tbl <- create_table(dat, borders = "outside") 
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with a page header picture", header = T) %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none") %>%
      page_header(width = c(0, 9, 0), left = "No display because width is 0") %>%
      header_image(image_path, height = 1, width = 7.5, align = "centre") %>%
      page_footer(width = c(0, 0, 9), center = "No display because width is 0") %>%
      footer_image(image_path, height = 0.8, width = 7, align = "right")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-62: Return error when header_image/footer_image is used without page_header/page_footer.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "html/test62.html")
    
    dat <- iris[1:50,]
    dat <- dat[, c("Species" ,"Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")]
    
    image_path <- file.path(base_path, "data/span_logo.png")
    
    tbl <- create_table(dat, borders = "outside") 
    
    # page_header is not set
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with a page header picture", header = T) %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none") %>%
      header_image(image_path, height = 1, width = 7.5, align = "centre")
    
    expect_error(
      res <- write_report(rpt),
      "`page_header` must be used when using `header_image`."
    )
    
    # page_footer is not set
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with a page header picture", header = T) %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none") %>%
      footer_image(image_path, height = 0.8, width = 7, align = "right")
    
    expect_error(
      res <- write_report(rpt),
      "`page_footer` must be used when using `footer_image`."
    )
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-63: group_cohesion puts the group in same page as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "html/test63.html")
    
    dat <- iris[1:91,]
    
    dat$test_group <- c(
      rep("A", 14),
      rep("B", 6),
      rep("C", 22),
      rep("D", 18),
      rep("E", 31)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 14), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 6), # 12 lines + 1
      c(rep("Flower C", 22)),
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E", 31)) # 31 lines + 1
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    # Page Size: 34
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, group_cohesion = TRUE, label = "Group", blank_after = T)
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-64: group_cohesion with min page proportion works as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "html/test64.html")
    
    dat <- iris[1:95,]
    
    dat$test_group <- c(
      rep("A", 14),
      rep("B", 6),
      rep("C", 22),
      rep("D", 18),
      rep("E", 35)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 14), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 6), # 12 lines + 1
      c(rep("Flower C", 22)),
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E", 35)) # 35 lines + 1 -> Still break because more than page size 34.6
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, group_cohesion = 0.65,
             label = "Group", blank_after = T) %>%
      define(test_string, label = "Flower")
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-65: Multiple group_cohesion work as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "html/test65.html")
    
    dat <- iris[1:91,]
    
    dat$test_group <- c(
      rep("A", 14),
      rep("B", 6),
      rep("C", 22),
      rep("D", 28),
      rep("E", 21)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 14), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 3), # 12 lines + 1
      rep("Flower B\nSubgroup B2", 3),
      c(rep("Flower C1", 17)),
      c(rep("Flower C2", 5)),
      c(rep("Flower D", 19), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      rep("Flower D2", 7),
      c(rep("Flower E1", 2)),
      c(rep("Flower E2", 19))
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, group_cohesion = TRUE, label = "Group", blank_after = T) %>%
      define(test_string, group_cohesion = TRUE)
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    # C2 would be moved to next page
    # E2 would be moved to next page
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-66: Multiple group_cohesion with different cohesion strengths work as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "html/test66.html")
    
    dat <- iris[1:107,]
    
    dat$test_group <- c(
      rep("A", 14),
      rep("B", 6),
      rep("C", 22),
      rep("D", 28),
      rep("E", 37)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 14), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 3), # 12 lines + 1
      rep("Flower B\nSubgroup B2", 3),
      c(rep("Flower C1", 17)),
      c(rep("Flower C2", 5)),
      c(rep("Flower D", 19), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      rep("Flower D2", 7),
      c(rep("Flower E1", 31)),
      c(rep("Flower E2", 6))
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, group_cohesion = TRUE, label = "Group", blank_after = T) %>%
      define(test_string, group_cohesion = 0.1)
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    # C2 would not be moved to next page
    # E2 would be moved to next page because previous lines are 31, which meets 34*0.9
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-67: break_label works as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "html/test67.html")
    
    dat <- iris[1:91,]
    
    dat$test_group <- c(
      rep("A", 14),
      rep("B", 6),
      rep("C", 22),
      rep("D", 18),
      rep("E", 31)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 14), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 6), # 12 lines + 1
      c(rep("Flower C", 22)),
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E", 31)) # 31 lines + 1
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    # Page Size: 34
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, break_label = "(Continued)", blank_before = T) %>%
      define(test_string, 
             break_label = paste0("(This is a long text which should take more ",
                                  "than one line. It's to test if the package is able to handle this ",
                                  "situation.)"),
             indent = 0.16)
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("html-68: Page numbers work every in title, footnote, header, and footer.", {
  
  if (dev) {
    fp <- file.path(base_path, "html/test68.html")
    
    dat <- iris
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table Title: Page [pg] of [tpg]", "Second table title") %>%
      footnotes("Table footnote: Page [pg] of [tpg]", "Second table footnote")
    
    rpt <- create_report(fp, output_type = "html", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      add_content(tbl) %>%
      titles("Report Title: Page [pg] of [tpg]") %>%
      footnotes("Report footnote: Page [pg] of [tpg]", borders = "none") %>%
      page_header("Header: Page [pg] of [tpg]") %>%
      page_footer("Footer: Page [pg] of [tpg]")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})
# User Tests --------------------------------------------------------------


test_that("html-user1: demo table works.", {
  
  if (dev) {
    library(tidyr)
    library(dplyr)
    
    # Data Filepath
    dir_data <- file.path(data_dir, "data")
    
    fp <- file.path(base_path, "html/user1.html")
    
    
    # Load Data
    data_demo   <- file.path(dir_data, "dm.csv") %>%
      read.csv() 
    
    
    data_demo <- subset(data_demo, data_demo$ARM != "SCREEN FAILURE")
    
    
    sex_decode <- c("M" = "Male",
                    "F" = "Female")
    
    race_decode <- c("WHITE" = "White",
                     "BLACK OR AFRICAN AMERICAN" = "Black or African American",
                     "ASIAN" = "Asian or Pacific Islander",
                     "NATIVE AMERICAN" = "Native American",
                     "UNKNOWN" = "Unknown")
    
    arm_pop <- table(data_demo$ARM) 
    
    
    demo_age <-
      data_demo %>%
      group_by(ARM) %>%
      summarise(across(.cols = AGE,
                       .fns = list(N      = ~ fmt_n(.),
                                   Mean   = ~ fmt_mean_sd(.),
                                   Median = ~ fmt_median(.),
                                   `Q1 - Q3` = ~ fmt_quantile_range(.),
                                   Range  = ~ fmt_range(.)
                       ))) %>%
      pivot_longer(-ARM,
                   names_to  = c("var", "label"),
                   names_sep = "_",
                   values_to = "value") %>%
      pivot_wider(names_from = ARM,
                  values_from = "value")
    
    
    
    demo_sex <-
      data_demo %>%
      add_count(ARM, SEX,  name = "n_SEX") %>%
      select(ARM, SEX, n_SEX) %>%
      distinct() %>%
      pivot_longer(cols = c(SEX),
                   names_to  = "var",
                   values_to = "label") %>%
      pivot_wider(names_from  = ARM,
                  values_from = n_SEX,
                  values_fill = 0) %>%
      mutate(label = factor(label, levels = names(sex_decode),
                            labels = sex_decode),
             `ARM A` = fmt_cnt_pct(`ARM A`, arm_pop["ARM A"]),
             `ARM B` = fmt_cnt_pct(`ARM B`, arm_pop["ARM B"]),
             `ARM C` = fmt_cnt_pct(`ARM C`, arm_pop["ARM C"]),
             `ARM D` = fmt_cnt_pct(`ARM D`, arm_pop["ARM D"]))
    
    
    
    demo_race <-
      data_demo %>%
      add_count(ARM, RACE, name = "n_RACE") %>%
      select(ARM, RACE, n_RACE) %>%
      distinct() %>%
      pivot_longer(cols = RACE,
                   names_to  = "var",
                   values_to = "label") %>%
      pivot_wider(names_from  = ARM,
                  values_from = n_RACE,
                  values_fill = 0) %>%
      mutate(label = factor(label, levels = names(race_decode),
                            labels = race_decode),
             `ARM A` = fmt_cnt_pct(`ARM A`, arm_pop["ARM A"]),
             `ARM B` = fmt_cnt_pct(`ARM B`, arm_pop["ARM B"]),
             `ARM C` = fmt_cnt_pct(`ARM C`, arm_pop["ARM C"]),
             `ARM D` = fmt_cnt_pct(`ARM D`, arm_pop["ARM D"])) %>%
      arrange(var, label)
    
    
    demo <- bind_rows(demo_age, demo_sex, demo_race)
    
    
    #View(demo)
    
    
    # Stub decode
    block_fmt <- c(AGE = "Age", SEX = "Sex", RACE = "Race")
    
    # Define table
    tbl <- create_table(demo, first_row_blank = TRUE) %>%
      column_defaults(from = "ARM A", to = "ARM D", width = 1.25) %>% 
      define(var, blank_after = TRUE, dedupe = TRUE,
             format = block_fmt, label = "") %>%
      define(label, label = "") %>%
      define(`ARM A`, align = "center", label = "Placebo", n = 36) %>%
      define(`ARM B`, align = "center", label = "Drug 10mg", n = 38) %>%
      define(`ARM C`, align = "center", label = "Drug 20mg", n = 38) %>%
      define(`ARM D`, align = "center", label = "Competitor", n = 38)
    
    # Define Report
    rpt <- create_report(fp, output_type = "HTML", font = fnt, font_size = fsz) %>%
      set_margins(top = 1, bottom = 1) %>% 
      options_fixed(font_size = 10) %>% 
      titles("Table 14.1/4",
             "Demographics and Baseline to Characteristics",
             "Specify Population Ω µ β ¥ ∑ ≠ ≤ £ ∞ ؈ ლ  \Ub8a 鬼") %>%
      add_content(tbl) %>% 
      footnotes("Special symbols \U221e to mess things up: Ω µ β ¥ ∑ ≠ ≤ £ ∞ ؈ ლ  \Ub8a 鬼") %>%   
      footnotes("Special symbols µ Ω £ there to mess things up: ", "Page [pg] of [tpg]") %>% 
      page_header("Left µ Ω £ ", "Right") %>% 
      page_footer("Time µ Ω £ ", right = "Page [pg] of [tpg]")
    
    # Write out report
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
    
    
    
  } else 
    expect_equal(TRUE, TRUE)
  
  
})

test_that("html-user2: demo table with stub works.", {
  
  
  if (dev) {
    
    library(tidyr)
    library(dplyr)
    
    
    
    # Data Filepath
    dir_data <- file.path(data_dir, "data")
    
    fp <- file.path(base_path, "html/user2.html")
    
    
    # Load Data
    data_demo   <- file.path(dir_data, "dm.csv") %>%
      read.csv() 
    
    data_demo <- subset(data_demo, data_demo$ARM != "SCREEN FAILURE")
    
    
    sex_decode <- c("M" = "Male",
                    "F" = "Female")
    
    race_decode <- c("WHITE" = "White",
                     "BLACK OR AFRICAN AMERICAN" = "Black or African American",
                     "ASIAN" = "Asian or Pacific Islander",
                     "NATIVE AMERICAN" = "Native American",
                     "UNKNOWN" = "Unknown")
    
    arm_pop <- table(data_demo$ARM) 
    
    
    demo_age <-
      data_demo %>%
      group_by(ARM) %>%
      summarise(across(.cols = AGE,
                       .fns = list(N      = ~ fmt_n(.),
                                   Mean   = ~ fmt_mean_sd(.),
                                   Median = ~ fmt_median(.),
                                   `Q1 - Q3` = ~ fmt_quantile_range(.),
                                   Range  = ~ fmt_range(.)
                       ))) %>%
      pivot_longer(-ARM,
                   names_to  = c("var", "label"),
                   names_sep = "_",
                   values_to = "value") %>%
      pivot_wider(names_from = ARM,
                  values_from = "value")
    
    
    
    demo_sex <-
      data_demo %>%
      add_count(ARM, SEX,  name = "n_SEX") %>%
      select(ARM, SEX, n_SEX) %>%
      distinct() %>%
      pivot_longer(cols = c(SEX),
                   names_to  = "var",
                   values_to = "label") %>%
      pivot_wider(names_from  = ARM,
                  values_from = n_SEX,
                  values_fill = 0) %>%
      mutate(label = factor(label, levels = names(sex_decode),
                            labels = sex_decode),
             `ARM A` = fmt_cnt_pct(`ARM A`, arm_pop["ARM A"]),
             `ARM B` = fmt_cnt_pct(`ARM B`, arm_pop["ARM B"]),
             `ARM C` = fmt_cnt_pct(`ARM C`, arm_pop["ARM C"]),
             `ARM D` = fmt_cnt_pct(`ARM D`, arm_pop["ARM D"]))
    
    
    
    demo_race <-
      data_demo %>%
      add_count(ARM, RACE, name = "n_RACE") %>%
      select(ARM, RACE, n_RACE) %>%
      distinct() %>%
      pivot_longer(cols = RACE,
                   names_to  = "var",
                   values_to = "label") %>%
      pivot_wider(names_from  = ARM,
                  values_from = n_RACE,
                  values_fill = 0) %>%
      mutate(label = factor(label, levels = names(race_decode),
                            labels = race_decode),
             `ARM A` = fmt_cnt_pct(`ARM A`, arm_pop["ARM A"]),
             `ARM B` = fmt_cnt_pct(`ARM B`, arm_pop["ARM B"]),
             `ARM C` = fmt_cnt_pct(`ARM C`, arm_pop["ARM C"]),
             `ARM D` = fmt_cnt_pct(`ARM D`, arm_pop["ARM D"])) %>%
      arrange(var, label)
    
    
    demo <- bind_rows(demo_age, demo_sex, demo_race)
    
    
    #View(demo)
    
    
    # Stub decode
    block_fmt <- c(AGE = "Age", SEX = "Sex", RACE2 = "Race")
    
    # Define table
    tbl <- create_table(demo, first_row_blank = TRUE, borders = "all") %>%
      stub(c("var", "label"), width = 2.5) %>% 
      column_defaults(width = 1) %>% 
      define(var, blank_after = TRUE, 
             format = block_fmt, label = "", label_row = TRUE) %>%
      define(label, label = "", indent = .25) %>%
      define(`ARM A`, align = "center", label = "Placebo", n = 36) %>%
      define(`ARM B`, align = "center", label = "Drug 10mg", n = 38) %>%
      define(`ARM C`, align = "center", label = "Drug 20mg", n = 38) %>%
      define(`ARM D`, align = "center", label = "Competitor", n = 38) %>% 
      titles("Table 14.1/4", bold = TRUE, blank_row = "above", 
             align = "center", borders = c("top", "left", "right")) %>% 
      titles( "Demographics and Baseline Characteristics",
             "Specify Population", borders = c("left", "right", "bottom"), 
             blank_row = "below",
             align = "center") %>%
      footnotes("Here is a footnote", "Here is another footnote", valign = "top",
                borders = "outside", blank_row = "both", align = "left")  
    
    # Define Report
    rpt <- create_report(fp, output_type = "HTML", 
                         font = "Arial", font_size = 10) %>%
      add_content(tbl, align = "center") %>% 
      page_header("Sponsor", "Drug") %>% 
      page_footer(left = "Time", right = "Page [pg] of [tpg]") #%>% 
    #page_by(var = "var", label = "Variable: ")
    
    # Write out report
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
    
    
  } else
    expect_equal(TRUE, TRUE)
  
})

test_that("html-user3: listings works.", {
  if (dev == TRUE) {
    # Data Filepath
    dir_data <- file.path(data_dir, "data")
    
    fp <- file.path(base_path, "html/user3.html")
    
    # Removing to make last page exactly equal to available rows on page.
    # In this case, any added blank rows should be skipped.
    fil <- c("ABC-14-124",
             "ABC-15-153",
             "ABC-15-154",
             "ABC-15-155",
             "ABC-15-156",
             "ABC-16-045",
             "ABC-16-046",
             "ABC-16-047",
             "ABC-16-157",
             "ABC-16-158",
             "ABC-16-159", 
             "ABC-16-160")
    
    # Load Data
    data_demo   <- file.path(dir_data, "dm.csv") %>%
      read.csv() 
    
    data_demo <- data_demo[!data_demo$USUBJID %in% fil, ]
    
    
    # Test that any assigned formats are applied
    attr(data_demo$SUBJID, "width") <- 1
    attr(data_demo$SUBJID, "justify") <- "left"
    attr(data_demo$SUBJID, "format") <- "S:%s"
    #print(widths(data_demo))
    names(data_demo)
    # Define table
    tbl <- create_table(data_demo) %>% 
      define(USUBJID, id_var = TRUE) 
    
    
    # Define Report
    rpt <- create_report(fp, font = "Arial", font_size = 12, 
                         orientation = "portrait") %>%
      titles("Listing 1.0",
             "Demographics Dataset") %>%
      add_content(tbl, align = "left") %>% 
      page_header("Sponsor", "Drug") %>% 
      page_footer(left = "Time", right = "Page [pg] of [tpg]") %>% 
      footnotes("My footnotes")
    
    #Write out report
    res <- write_report(rpt, output_type = "HTML")
    
    expect_equal(file.exists(fp), TRUE)
    
    
    
    # pdfpth <- file.path(base_path, "user/user3.pdf")
    # write_report(rpt, pdfpth, output_type = "PDF")
    # expect_equal(file.exists(pdfpth), TRUE)
  } else 
    expect_equal(TRUE, TRUE)
  
  
})


test_that("html-user4: listing in cm and times works.", {
  if (dev == TRUE) {
    # Data Filepath
    dir_data <- file.path(data_dir, "data")
    
    fp <- file.path(base_path, "html/user4.html")
    
    # Removing to make last page exactly equal to available rows on page.
    # In this case, any added blank rows should be skipped.
    fil <- c("ABC-14-124",
             "ABC-15-153",
             "ABC-15-154",
             "ABC-15-155",
             "ABC-15-156",
             "ABC-16-045",
             "ABC-16-046",
             "ABC-16-047",
             "ABC-16-157",
             "ABC-16-158",
             "ABC-16-159", 
             "ABC-16-160")
    
    # Load Data
    data_demo   <- file.path(dir_data, "dm.csv") %>%
      read.csv() 
    
    data_demo <- data_demo[!data_demo$USUBJID %in% fil, ]
    
    
    # Test that any assigned formats are applied
    attr(data_demo$SUBJID, "width") <- 2.54
    attr(data_demo$SUBJID, "justify") <- "left"
    attr(data_demo$SUBJID, "format") <- "S:%s"
    #print(widths(data_demo))
    
    # Define table
    tbl <- create_table(data_demo) %>% 
      define(USUBJID, id_var = TRUE) 
    
    
    # Define Report
    rpt <- create_report(fp, font = "Times", font_size = 10, 
                         units = "cm", orientation = "landscape") %>%
      titles("Listing 1.0",
             "Demographics Dataset") %>%
      add_content(tbl, align = "left") %>% 
      page_header("Sponsor", "Drug") %>% 
      page_footer(left = "Time", right = "Page [pg] of [tpg]") %>% 
      footnotes("My footnote")
    
    #Write out report
    res <- write_report(rpt, output_type = "HTML")
    
    expect_equal(file.exists(fp), TRUE)
    
    #print(res$column_widths)
    
    # pdfpth <- file.path(base_path, "user/user3.pdf")
    # write_report(rpt, pdfpth, output_type = "PDF")
    # expect_equal(file.exists(pdfpth), TRUE)
  } else 
    expect_equal(TRUE, TRUE)
  
  
})

test_that("html-user5: Portrait in 12pt Arial works as expected.", {
  
  if (dev == TRUE) {
    
    dir_data <- file.path(data_dir, "data")
    
    fp <- file.path(base_path, "html/user5.html")
    
    # Read in prepared data
    df <- read.table(header = TRUE, text = '
        var     label        A             B          
        "ampg"   "N"          "19"          "13"         
        "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
        "ampg"   "Median"     "16.4"        "21.4"       
        "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
        "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
        "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
        "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
        "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
    
    # Create table
    tbl <- create_table(df, first_row_blank = TRUE) %>% 
      define(var, label = "Variable", blank_after = TRUE, dedupe = TRUE,
             format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
      define(label, label = "") %>% 
      define(A, label = "Group A", align = "center", n = 19) %>% 
      define(B, label = "Group B", align = "center", n = 13)
    
    
    # Create report and add content
    rpt <- create_report(fp, orientation = "portrait", output_type = "HTML",
                         font = "Arial", font_size = 12) %>% 
      page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
      titles("Table 1.0", "MTCARS Summary Table") %>% 
      add_content(tbl) %>% 
      footnotes("* Motor Trend, 1974") %>%
      page_footer(left = Sys.time(), 
                  center = "Confidential", 
                  right = "Page [pg] of [tpg]")
    
    # Write out report
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
    
  } else 
    expect_equal(TRUE, TRUE)
  
})

context("Negative System Tests")


base_path <- "c:/packages/reporter/tests/testthat"

base_path <- tempdir()

# This warning was being triggered unnecessarily.  Not working properly.
# Comment out for now as it's really not that important.
# User is responsible for their own sort.
# test_that("neg1: Unsorted Page by generates warning.", {
#   
#   fp <- file.path(base_path, "neg/neg1.out")
#   
#   
#   dat <- iris[order(iris$Sepal.Width), ]
#   tbl <- create_table(dat) %>% 
#     define(Species, visible = FALSE)
#   
#   rpt <- create_report(fp) %>% 
#     page_header("Client", "Study") %>% 
#     titles("Table 1.0", "IRIS Data Frame") %>% 
#     page_by(Species, "Species: ") %>% 
#     footnotes("Here is a footnote") %>% 
#     page_footer("Time", right = "Page [pg] of [tpg]") %>% 
#     add_content(tbl) 
#   
#   
#   expect_message(write_report(rpt))
#   
#   expect_equal(file.exists(fp), TRUE)
# 
#   
# })


test_that("neg2:  Page by variable not on plot generates error.", {
  
  library(ggplot2)
  
  fp <- file.path(base_path, "neg/neg2.rtf")
  

  dat <- mtcars[order(mtcars$cyl), ]

  p <- ggplot(mtcars, aes(x=mpg, y=disp)) + geom_point()
  
  plt <- create_plot(p, 5, 7)
  
  rpt <- create_report(fp, output_type = "RTF") %>% 
    page_header("Client", "Study") %>% 
    titles("Table 1.0", "MTCARS Data Frame") %>% 
    page_by(Species, "Cylinders: ") %>% 
    footnotes("Here is a footnote") %>% 
    page_footer("Time", right = "Page [pg] of [tpg]") %>% 
    add_content(plt) 
  
  #res <- write_report(rpt)
  
  expect_error(write_report(rpt))
  
  expect_equal(file.exists(fp), FALSE)
  
  
})

test_that("neg3: Invalid column names in show_cols generate error.", {

  fp <- file.path(base_path, "output/neg3.out")
  
  tbl <- create_table(mtcars[1:10, ], show_cols = c("mpg", "cyl", "fork", "bork"))
  
  rpt <- create_report(fp) %>% 
    add_content(tbl)
  
  expect_error(write_report(rpt))
  
})


test_that("neg4: Invalid column names in define functions generate error.", {
  
  expect_error(create_table(mtcars[1:10, ]) %>% 
    define(mpg) %>% 
    define(cyl) %>% 
    define(fork))

})


test_that("neg5: Page header and footer generate error when added to table.", {
  
  expect_error(create_table(iris) %>% 
    page_header("Client", "Study"))
    
  expect_error(create_table(iris) %>% 
                 page_footer("Client", "Study"))
    
})

test_that("neg6: Title header generates error when added to invalid object", {
  
  expect_error(list() %>% 
                 title_header("Client", "Study"))
  
  expect_error(c() %>% 
                 title_header("Client", "Study"))
  
})


test_that("neg7: Column doesn't exist in show_cols does not generate error.", {
  
  
  fp <- file.path(base_path, "neg/neg7.txt")
  
  
  tbl <- create_table(mtcars, show_cols = c("mpg", "cyl", "disp")) %>% 
                 define(mpg) %>% 
                 define(cyl) %>% 
                 define(hp)
  
  rpt <- create_report(fp) %>% add_content(tbl)
  
  expect_error(write_report(rpt), NA)
  
  
  
})



test_that("neg8: Long page header generates error.", {
  
  
  fp <- file.path(base_path, "neg/neg8.txt")
  
  
  tbl <- create_table(mtcars, show_cols = c("mpg", "cyl", "disp")) %>% 
    define(mpg) %>% 
    define(cyl) %>% 
    define(hp)
  
  rpt <- create_report(fp) %>% add_content(tbl) %>% 
    page_header("Here is a a whole bunch of stuff intended to go over the page",
                "width because it is way too long. Way too long. Way too long.")
  
  expect_error(write_report(rpt))
  
  
  
})


test_that("neg9: Long page footer generates error.", {
  
  
  fp <- file.path(base_path, "neg/neg9.txt")
  
  
  tbl <- create_table(mtcars, show_cols = c("mpg", "cyl", "disp")) %>% 
    define(mpg) %>% 
    define(cyl) %>% 
    define(hp)
  
  rpt <- create_report(fp) %>% add_content(tbl) %>% 
    page_footer("Here is a a whole bunch of stuff intended to go over the page",
                "width because it is way too long. Way too long. Way too long.")
  
  expect_error(write_report(rpt))
  
  
  
})

context("RTF Tests")

base_path <- paste0(getwd(),"/tests/testthat")

base_path <- tempdir()

cnt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
              "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
              "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
              "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ", 
              "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
              "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
              "qui officia deserunt mollit anim id est laborum.")

dev <- FALSE

test_that("rtf1: Simplest table works as expected.", {
  
  if (dev == TRUE) {
    
  fp <- file.path(base_path, "rtf/test1.rtf")
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>% 
    add_content(create_table(mtcars[1:10, ]), align = "left")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
  
})

test_that("rtf2: Simplest table with title works as expected.", {
  
  if (dev == TRUE) {
    
  fp <- file.path(base_path, "rtf/test2.rtf")
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    define(vs, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>% 
    options_fixed(font_size = 10) %>% 
    titles("MTCARS Data Frame", align = "left") %>% 
    add_content(tbl)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
})


test_that("rtf3: Table with break between sections works as expected.", {
  
  if (dev == TRUE) {
  
  fp <- file.path(base_path, "rtf/test3.rtf")
  
  # Setup
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma", 
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon", 
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
  arm <- c(rep("A", 5), rep("B", 5))
  
  # Create data frame
  df <- data.frame(subjid, name, sex, age, arm, stringsAsFactors = FALSE)
  
  
  tbl1 <- create_table(df, first_row_blank = TRUE) %>%
    define(subjid, label = "Subject ID", align = "left") %>% 
    define(name, label = "Subject Name") %>% 
    define(sex, label = "Sex") %>% 
    define(age, label = "Age") %>% 
    define(arm, label = "Arm", 
           blank_after = TRUE, 
           dedupe = TRUE, 
           align = "right")
  
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    options_fixed(font_size = 10) %>% 
    page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
    titles("Table 1.0", "Analysis Data Subject Listing", 
           "Safety Population", align = "center") %>%
    footnotes("Program Name: table1_0.R") %>%
    page_footer(left = "Time", center = "Confidential", 
                right = "Page [pg] of [tpg]") %>%
    add_content(tbl1) 
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
  
})

test_that("rtf4: Table that spans multiple pages breaks as expected.", {
  
  fp <- file.path(base_path, "rtf/test4.rtf")
  
  rpt <- create_report(fp, output_type = "RTF") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) 
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 4)
  #write_registration_file(file.path(base_path,"./rtf/reg.txt"))
})

test_that("rtf5: Table with long cell and label values wraps as expected.", {
  
  
  fp <- file.path(base_path, "rtf/test5.rtf")
  
  # Setup
  arm <- c(rep("A", 5), rep("B", 5))
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa \nand more and more", "Schaffer, Ashley", "Laner, Tahma", 
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon", 
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
  
  
  # Create data frame
  df <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)
  
  
  tbl1 <- create_table(df, first_row_blank = TRUE) %>%
    define(subjid, label = "Subject ID for a patient", n = 10, align = "left", 
           width = 1) %>%
    define(name, label = "Subject Name", width = 1) %>%
    define(sex, label = "Sex", n = 10, align = "center") %>%
    define(age, label = "Age", n = 10) %>%
    define(arm, label = "Arm",
           blank_after = TRUE,
           dedupe = TRUE)
  
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    titles("Table 1.0", align = "center") %>%
    
    add_content(tbl1)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})

test_that("rtf6: Table with spanning headers works as expected.", {
  
  fp <- file.path(base_path, "rtf/test6.rtf")
  
  
  df <- data.frame(vehicle = rownames(mtcars), mtcars, stringsAsFactors = FALSE)
  rownames(df) = NULL
  
  df$qsec <- fattr(df$qsec, format = "%.1f")
  df$wt <- fattr(df$wt, justify = "center", width = .75)
  
  tbl <- create_table(df) %>% 
    spanning_header("mpg", "hp",
                    label = "Span 1", label_align = "center", n = 10) %>% 
    spanning_header("drat", "qsec",
                    label = "Span 2", label_align = "center", n = 10) %>%
    spanning_header("vs", "carb",
                    label = "Span 3", label_align = "center", n = 10) %>%
    spanning_header(from = "drat", to = "carb", label = "Super Span",
                    label_align = "center",
                    level = 2) %>%
    define(vehicle, label = "Vehicle") %>% 
    define(mpg, format = "%.1f")
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>% 
    add_content(tbl) %>% 
    titles("Table 1.0", "MTCARS Subset Test")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
})




test_that("rtf7: Simplest RTF report with 1 in margins works as expected.", {

  if (dev == TRUE) {
    
  fp <- file.path(base_path, "rtf/test7.rtf")

  tbl <- create_table(mtcars[1:10, ]) %>%
    column_defaults(width = .5) %>%
    define(vs, visible = FALSE)

  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    page_header("Client", "Study") %>%
    titles("MTCARS Data Frame") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl, align = "left") %>%
    page_footer("Time", right = "Page [pg] of [tpg]")


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else 
    expect_equal(TRUE, TRUE)

})


test_that("rtf8: Two page RTF report works as expected.", {

  if (dev == TRUE) {
    
  fp <- file.path(base_path, "rtf/test8.rtf")

  # Setup
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma",
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
  arm <- c(rep("A", 5), rep("B", 5))

  # Create data frame
  df <- data.frame(subjid, name, sex, age, arm, stringsAsFactors = FALSE)

  df1 <- df[df$arm == "A", ]
  df2 <- df[df$arm == "B", ]

  afmt <- value(condition(x == "A", "Placebo"),
                condition(x == "B", "Treatment 1"))

  sfmt1 <- value(condition(x == "M", "Male"),
                 condition(x == "F", "Female"),
                 condition(TRUE, "Other"))

  sfmt2 <- c(M = "Male", F = "Female")

  tbl1 <- create_table(df1, width = 7) %>%
    define(sex, width = 1, format = sfmt1) %>%
    define(name, width = 2) %>%
    define(age, width = 1, align = "left")

  tbl2 <- create_table(df2, width = 7) %>%
    define(sex, width = 1, format = sfmt2) %>%
    define(age, format = "%0d%%", align = "left") %>%
    define(name, width = 2) %>%
    define(arm, format = afmt, width = 2, align = "right")


  rpt <- create_report(fp, output_type = "RTF") %>%
    options_fixed(font_size = 10) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
    titles("Table 1.0", "Analysis Data Subject Listing",
           "Safety Population", align = "center") %>%
    footnotes("Program Name: table1_0.R") %>%
    page_footer(left = "Time", center = "Confidential",
                right = "Page [pg] of [tpg]") %>%
    add_content(tbl1) %>%
    add_content(tbl2)


  res <- write_report(rpt)

  #print(res)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)

  } else 
    expect_equal(TRUE, TRUE)
})




test_that("rtf9: Simplest RTF Plot works as expected.", {

  if (dev == TRUE) {
    
  library(ggplot2)

  fp <- file.path(base_path, "rtf/test9.rtf")

  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()

  plt <- create_plot(p, height = 4, width = 8)


  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt, align = "center") %>%
    footnotes("* Motor Trend, 1974") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")


  res <- write_report(rpt)

  #print(res)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else 
    expect_equal(TRUE, TRUE)
})




test_that("rtf10: RTF Table with Plot works as expected.", {

  library(ggplot2)

  fp <- file.path(base_path, "rtf/test10.rtf")

  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()


  plt <- create_plot(p, height = 4, width = 8)
  tbl <- create_table(mtcars[1:10, ])


  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    add_content(plt, align = "center") %>%
    footnotes("* Motor Trend, 1974") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")


  res <- write_report(rpt)

  #print(res)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)


})




test_that("rtf11: RTF Table with Plot on same page works as expected.", {
  
  library(ggplot2)
  
  fp <- file.path(base_path, "rtf/test11.rtf")
  
  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  
  plt <- create_plot(p, height = 4, width = 8)
  tbl <- create_table(mtcars[1:3, ])
  
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", blank_row = "none") %>%
    set_margins(top = 1, bottom = 1) %>%

    add_content(plt, page_break = FALSE, blank_row = "none") %>%
    add_content(tbl) %>%
    footnotes("* Motor Trend, 1974") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  
})


test_that("rtf12: Table and Text output works as expected.", {
  
  fp <- file.path(base_path, "rtf/test12.rtf")
  
  tbl1 <- mtcars[1:10, ]
  tbl2 <- mtcars[11:20, ]
  
  rpt <- create_report(fp, orientation = "portrait", output_type = "RTF", 
                       font = "fixed") %>%
    titles("Report 5.0", "Table and Text Report") %>% 
    page_header(left = "Client: ABC", right = "Study: 123") %>% 
    add_content(create_table(tbl1), page_break = FALSE) %>% 
    add_content(create_text("* NOTE: Car information from 1971."), 
                align = "left") %>% 
    add_content(create_table(tbl2), page_break = FALSE) %>% 
    add_content(create_text("* NOTE: Car information from 1972."), 
                align = "left") %>% 
    page_footer(left = "Time", 
                center = "Confidential", 
                right ="Page [pg] of [tpg]")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)

})


test_that("rtf13: Very Long text output works as expected.", {
  
  
  if (dev == TRUE) {
    
    fp <- file.path(base_path, "rtf/test13.rtf")
    
    l <- paste(rep(cnt, 1000), collapse = "\n\n")
    
    rpt <- create_report(fp, orientation = "portrait", output_type = "RTF", 
                         font = "fixed") %>%
      titles("Report 6.0", "Very long Text Report") %>% 
      page_header(left = "Client: ABC", right = "Study: 123") %>% 
      add_content(create_text(l)) %>% 
      page_footer(left = "Time", 
                  center = "Confidential", 
                  right ="Page [pg] of [tpg]")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 123)

    
  } else {
    expect_equal(TRUE, TRUE) 
  }
  
})


test_that("rtf14: Simplest portrait table works as expected.", {
  
  if (dev == TRUE) {
  
  fp <- file.path(base_path, "rtf/test14.rtf")
  
  rpt <- create_report(fp, output_type = "RTF", orientation = "portrait", 
                       font = "fixed") %>% 
    page_header("left", "right") %>% 
    titles("Table 1.0", "MTCARS Data Frame", borders = "all") %>% 
    add_content(create_table(mtcars)) %>% 
    page_footer("Left", right = "right")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
  
})


test_that("rtf15: Simplest landscape table works as expected.", {
  
  if (dev == TRUE) {
  
  fp <- file.path(base_path, "rtf/test15.rtf")
  
  rpt <- create_report(fp, output_type = "RTF", orientation = "landscape", 
                       font = "fixed") %>% 
    page_header("left", "right") %>% 
    titles("Table 1.0", "MTCARS Data Frame", borders = "all") %>% 
    add_content(create_table(mtcars)) %>% 
    page_footer("Left", right = "right")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
})

test_that("test16: 10 pt report with units in cm works as expected.", {
  
  
  if (dev == TRUE) {
    
  fp <- file.path(base_path, "rtf/test16.rtf")
  
  
  rpt <- create_report(fp, units = "cm", output_type = "RTF", 
                       font = "fixed") %>%
    page_header("Client: Experis", "Study: ABC") %>% 
    titles("IRIS Data Frame") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]") %>% 
    add_content(create_table(iris)) 
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 4)
  
  } else 
    expect_equal(TRUE, TRUE)
  
})


test_that("test17: 12 pt report with units in cm works as expected.", {
  
  if (dev == TRUE) {
    
  fp <- file.path(base_path, "rtf/test17.rtf")
  
  
  rpt <- create_report(fp, units = "cm", output_type = "RTF", 
                       font = "fixed") %>%
    options_fixed(font_size = 12) %>% 
    page_header("Client: Experis", "Study: ABC") %>% 
    titles("IRIS Data Frame") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]") %>% 
    add_content(create_table(iris)) 
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 5)
  
  } else 
    expect_equal(TRUE, TRUE)
  
})



test_that("rtf18: Plot with page by on report works as expected.", {
  
  library(ggplot2)
  
  fp <- file.path(base_path, "rtf/test18.rtf")


  dat <- mtcars[order(mtcars$cyl), ]
  
  p <- ggplot(dat, aes(x=disp, y=mpg)) + geom_point()

  
  #dats <- split(p$data, p$data$grp)
  #tbl <- create_table(dat[1:3, ])
  
  plt <- create_plot(p, height = 4, width = 8)
  

  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", 
           blank_row = "none", borders = "all") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_by(cyl, "Cylinders: ") %>% 
    add_content(plt) %>%
    footnotes("* Motor Trend, 1974", borders = "all") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 3)
  
  
})




test_that("rtf19: Plot with page by on plot works as expected.", {
  
  library(ggplot2)
  
  fp <- file.path(base_path, "rtf/test19.rtf")
  
  
  dat <- mtcars[order(mtcars$cyl), ]
  
  
  fmt <- value(condition(x == 4, "4 Cylinder"),
               condition(x == 6, "6 Cylinder"),
               condition(x == 8, "8 Cylinder"))
  
  p <- ggplot(dat, aes(x=disp, y=mpg)) + geom_point()
  
  
  #dats <- split(p$data, p$data$grp)
  #tbl <- create_table(dat[1:3, ])
  
  plt <- create_plot(p, height = 4, width = 8) %>% 
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", blank_row = "none") %>%
    page_by(cyl, "Cylinders: ", format = fmt) %>% 
    footnotes("* Motor Trend, 1974") 
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt) %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 3)
  
  
})


test_that("test20: Title Header on Plot works as expected.", {
  
  if (dev == TRUE) {
    
    
  fp <- file.path(base_path, "rtf/test20.rtf")
  
  
  p <- ggplot(mtcars, aes(x=disp, y=mpg)) + geom_point()
  
  plt <- create_plot(p, height = 5, width = 7) %>% 
    title_header("Figure 1.0", "MTCARS Plot", 
                 right = c("Client", "Page", "More"), borders = "all") %>% 
    footnotes("* Motor Trend, 1974", borders = "all")
  
  rpt <- create_report(fp, units = "inches", output_type = "RTF", 
                       font = "fixed") %>%
    options_fixed(font_size = 12) %>% 
    add_content(plt) 
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
  
})

# 
# test_that("rtf19: Plot and table with page by works as expected.", {
#   
#   library(ggplot2)
#   
#   fp <- file.path(base_path, "rtf/test19.rtf")
#   
#   
#   dat <- mtcars[order(mtcars$cyl), ]
#   
#   p <- ggplot(dat, aes(x=disp, y=mpg)) + geom_point()
#   
#   
#   #dats <- split(p$data, p$data$grp)
#   #tbl <- create_table(dat[1:3, ])
#   
#   plt <- create_plot(p, height = 4, width = 8)
#   
#   
#   
#   rpt <- create_report(fp, output_type = "RTF") %>%
#     page_header("Client", "Study: XYZ") %>%
#     titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", blank_row = "none") %>%
#     set_margins(top = 1, bottom = 1) %>%
#     #page_by(grp, "Mileage Category: ") %>% 
#     add_content(plt) %>%
#     # add_content(tbl) %>%
#     footnotes("* Motor Trend, 1974") %>%
#     page_footer("Time", "Confidential", "Page [pg] of [tpg]")
#   
#   
#   res <- write_report(rpt)
#   
#   #print(res)
#   
#   expect_equal(file.exists(fp), TRUE)
#   
#   
#   
# })


test_that("test21: 8 pt report with units in inches works as expected.", {
  
  if (dev == TRUE) {
    
  fp <- file.path(base_path, "rtf/test21.rtf")
  
  tbl <- create_table(iris) 
  
  rpt <- create_report(fp, units = "inches", output_type = "RTF", 
                       font = "fixed") %>%
    page_header("Client: Experis", "Study: ABC") %>% 
    titles("IRIS Data Frame") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]") %>% 
    options_fixed(font_size = 8) %>% 
    add_content(tbl) 
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 3)
  
  } else 
    expect_equal(TRUE, TRUE)
  
})

test_that("test22: 8 pt report with units in cm works as expected.", {
  
  if (dev == TRUE) {
    
  fp <- file.path(base_path, "rtf/test22.rtf")
  
  
  rpt <- create_report(fp, units = "cm", output_type = "RTF", 
                       font = "fixed") %>%
    page_header("Client: Experis", "Study: ABC") %>% 
    titles("IRIS Data Frame") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]") %>% 
    options_fixed(font_size = 8) %>% 
    add_content(create_table(iris)) 
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 3)
  
  
  } else 
    expect_equal(TRUE, TRUE)
  
})




test_that("rtf23: RTF Table with Plot and borders works as expected.", {
  
  if (dev == TRUE) {
    
  library(ggplot2)
  
  fp <- file.path(base_path, "rtf/test23.rtf")
  
  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  
  
  plt <- create_plot(p, height = 4, width = 8)
  tbl <- create_table(mtcars[1:10, ])
  
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", borders = "all") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    add_content(plt, align = "center") %>%
    footnotes("* Motor Trend, 1974", borders = "all") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)
  
  } else 
    expect_equal(TRUE, TRUE)
  
})



test_that("rtf24: RTF Table with Plot and borders works as expected.", {
  
  library(ggplot2)
  
  fp <- file.path(base_path, "rtf/test24.rtf")
  
  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  
  
  plt <- create_plot(p, height = 4, width = 8) %>% 
    titles("My plot", borders = "all") %>% 
    footnotes("My plot footnotes", borders = "all")
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    titles("My table", borders = "all") %>% 
    footnotes("My table footnotes", borders = "all", align = "right")
  
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    add_content(plt, align = "center") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)
})


test_that("rtf25: RTF Table with custom options works as expected.", {
  
  if (dev == TRUE) {
  
  fp <- file.path(base_path, "rtf/test25.rtf")
  

  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    titles("My table", borders = "all") %>% 
    footnotes("My table footnotes", borders = "all", align = "right")
  
  tbl2 <- create_table(mtcars[11:20, ]) %>% 
    titles("My table", borders = "all") %>% 
    footnotes("My table footnotes", borders = "all", align = "right")
  
  
  rpt <- create_report(fp, font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    add_content(tbl2, align = "center") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]") %>% 
    options_fixed(font_size = 12, line_size = 80, line_count = 30,
                  uchar = "-")
  
  
  res <- write_report(rpt, output_type = "RTF")
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)
  
  } else 
    expect_equal(TRUE, TRUE)
})

test_that("rtf26: line_size and line_count overrides work as expected.", {
  
  if (dev == TRUE) {
    
  fp <- file.path(base_path, "rtf/test26.rtf")
  
  
  rpt <- create_report(fp, font = "fixed") %>%
    options_fixed(line_size = 40, line_count = 30) %>% 
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) 
  
  
  res <- write_report(rpt, output_type = "RTF")
  res
  
  expect_equal(file.exists(fp), TRUE)
  
  } else 
    expect_equal(TRUE, TRUE)
  
})


test_that("rtf27: Plot Borders work as expected.", {
  
  if (dev == TRUE) {
    
  library(ggplot2)
  
  fp <- file.path(base_path, "rtf/test27.rtf")
  
  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  
  
  plt <- create_plot(p, height = 4, width = 8, borders = "all") %>% 
    titles("My plot", borders = "none") %>% 
    footnotes("My plot footnotes", borders = "none")
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt, align = "right") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
})

test_that("rtf28: Table Borders that spans multiple pages work as expected.", {
  
  fp <- file.path(base_path, "rtf/test28.rtf")
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris, borders = "all")) %>% 
    footnotes("Here is a footnote")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 4)
  #write_registration_file(file.path(base_path,"./rtf/reg.txt"))
})

test_that("rtf29: Simplest RTF Plot with valign top works as expected.", {
  
  if (dev == TRUE) {
    
  library(ggplot2)
  
  fp <- file.path(base_path, "rtf/test29.rtf")
  
  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  
  plt <- create_plot(p, height = 4, width = 8)
  
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt, align = "center") %>%
    footnotes("* Motor Trend, 1974", valign = "top") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
  
  
})

test_that("rtf30: Simplest RTF Plot with valign bottom works as expected.", {
  
  if (dev == TRUE) {
    
  library(ggplot2)
  
  fp <- file.path(base_path, "rtf/test30.rtf")
  
  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  
  plt <- create_plot(p, height = 4, width = 8) %>%
    footnotes("* Motor Trend, 1974", valign = "bottom")
  
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt, align = "center") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
  
  
})

test_that("rtf31: Simplest RTF Text with valign top works as expected.", {
  
  if (dev == TRUE) {

  fp <- file.path(base_path, "rtf/test31.rtf")

  txt <- create_text(cnt, width = 6)
  
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Text 1.0", "MTCARS Miles per Cylinder Text") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(txt, align = "center") %>%
    footnotes("* Motor Trend, 1974", valign = "top") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
  
  
})

test_that("rtf32: Simplest RTF Text with valign bottom works as expected.", {
  
  if (dev == TRUE) {

  fp <- file.path(base_path, "rtf/test32.rtf")
  
  txt <- create_text(cnt, width = 6) %>%
    footnotes("* Motor Trend, 1974", valign = "bottom")
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(txt, align = "center") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
  
  
})


test_that("rtf33: use_attributes parameter table works as expected.", {
  
  if (dev == TRUE) {
  
  fp1 <- file.path(base_path, "rtf/test33a.rtf")
  fp2 <- file.path(base_path, "rtf/test33b.rtf")
  fp3 <- file.path(base_path, "rtf/test33c.rtf")
  
  
  dat <- mtcars[1:10, ]
  attr(dat$mpg, "label") <- "Miles per gallon"
  attr(dat$cyl, "format") <- "%.1f"
  attr(dat$hp, "width") <- 2
  fattr(dat$vs) <- list(width = 2, justify = "center")
  
  tbl <- create_table(dat) 
  
  # Test default 
  rpt <- create_report(fp1, output_type = "RTF", font = "fixed") %>% 
    add_content(tbl)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp1), TRUE)
  expect_equal(res$pages, 1)
  
  # Test none
  tbl <- create_table(dat, use_attributes = "none") 
  
  rpt <- create_report(fp2, output_type = "RTF", font = "fixed") %>% 
    add_content(tbl)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp2), TRUE)
  
  
  # Test some
  tbl <- create_table(dat, use_attributes = c("format", "width")) 
  
  rpt <- create_report(fp3, output_type = "RTF", font = "fixed") %>% 
    add_content(tbl)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp3), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
  
})



test_that("rtf34: 9 pt font inches works as expected.", {
  
  if (dev == TRUE) {
  
  fp <- file.path(base_path, "rtf/test34.rtf")
  
  rpt <- create_report(fp, output_type = "RTF", font_size = 9, 
                       orientation = "portrait", font = "fixed") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>% 
    set_margins(top = 1, bottom = 1)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  } else 
    expect_equal(TRUE, TRUE)
})

test_that("rtf35: 9 pt font cm works as expected.", {
  
  if (dev == TRUE) {
  
  fp <- file.path(base_path, "rtf/test35.rtf")
  
  rpt <- create_report(fp, output_type = "RTF", font_size = 9, 
                       orientation = "portrait", font = "fixed") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>% 
    set_margins(top = 1, bottom = 1)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  } else 
    expect_equal(TRUE, TRUE)
})

test_that("rtf36: 11 pt font inches works as expected.", {
  
  if (dev == TRUE) {
    
  fp <- file.path(base_path, "rtf/test36.rtf")
  
  rpt <- create_report(fp, output_type = "RTF", font_size = 11, 
                       orientation = "portrait", font = "fixed") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>% 
    set_margins(top = 1, bottom = 1)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  } else 
    expect_equal(TRUE, TRUE)
})

test_that("rtf37: 11 pt font cm works as expected.", {
  
  if (dev == TRUE) {
  
  fp <- file.path(base_path, "rtf/test37.rtf")
  
  rpt <- create_report(fp, output_type = "RTF", font_size = 11, 
                       orientation = "portrait", font = "fixed") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>% 
    set_margins(top = 1, bottom = 1)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)

  
  } else 
    expect_equal(TRUE, TRUE)
})

test_that("rtf38: RTF Image file works as expected.", {
  
  if (dev == TRUE) {
    
  library(ggplot2)
  
  fp <- file.path(base_path, "rtf/test38.rtf")
  
  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  
  pltpath <- file.path(base_path, "rtf/test38.jpg")
  ggsave(pltpath, width = 8, height = 4, 
         units = "in",
         dpi = 300)
  
  plt <- create_plot(pltpath, height = 4, width = 8)
  
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt, align = "center") %>%
    footnotes("* Motor Trend, 1974") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
})

test_that("rtf39: Blank after on invisible column.", {
  
  fp <- file.path(base_path, "rtf/test39.rtf")
  
  tbl <- create_table(iris, borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    page_header("Left", "Right") %>%
    add_content(tbl) %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("rtf40: Page header width works as expected.", {
  
  fp <- file.path(base_path, "rtf/test40.rtf")
  
  tbl <- create_table(iris[1:10, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    page_header("Left here is some stuff and more stuff trying to get out in the middle", 
                "Right", width = 8) %>%
    add_content(tbl) %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below", header = TRUE) %>%
    footnotes("Here is a footnote", "And another", footer = TRUE)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})



test_that("rtf41: Custom page size works as expected.", {
  
  fp <- file.path(base_path, "rtf/test41.rtf")
  
  tbl <- create_table(iris[1:15, ]) %>%
    define(Species, visible = FALSE)
  
  ttl <- c("Title1", "Title2", "Title3")
  
  rpt <- create_report(fp, output_type = "RTF", 
                       paper_size = c(6.5, 7.5),
                       orientation = "portrait", font = "fixed") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles(ttl,
           blank_row = "below", columns =  1, align = "center",
           borders = "none") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("rtf42: Symbols are proper orientation on portrait.", {
  
  if (dev) {
  
    fp <- file.path(base_path, "rtf/test42.rtf")
    
    dat <- mtcars[1:10, ]
    
    
    tbl <- create_table(dat) %>%
      titles("My Table{symbol('dagger')}")
    
    
    rpt <- create_report(fp, output_type = "RTF", 
                         orientation = "portrait", font = "fixed") %>%
      add_content(tbl)
    
    res <- write_report(rpt)
    
    
    # file.show(res$modified_path)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    
    expect_equal(TRUE, TRUE)
    
  }
  
})


test_that("rtf43: Symbols are proper orientation on landscape.", {
  
  if (dev) {
    
    fp <- file.path(base_path, "rtf/test43.rtf")
    
    dat <- mtcars[1:10, ]
    
    
    tbl <- create_table(dat) %>%
      titles("My Table{symbol('dagger')}")
    
    
    rpt <- create_report(fp, output_type = "RTF", 
                         orientation = "landscape", font = "fixed") %>%
      add_content(tbl)
    
    res <- write_report(rpt)
    
    
    file.show(res$modified_path)
    
    expect_equal(file.exists(fp), TRUE)
    
    
  } else {
    
    expect_equal(TRUE, TRUE)
    
  }
  
})

test_that("rtf44: Table with blank_before works as expected.", {
  
  if (dev == TRUE) {
    
    fp <- file.path(base_path, "rtf/test44.rtf")
    
    # Setup
    subjid <- 100:109
    name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
              "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma", 
              "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon", 
              "Smith, Shaileigh")
    sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
    age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
    arm <- c(rep("A", 5), rep("B", 5))
    
    # Create data frame
    df <- data.frame(subjid, name, sex, age, arm, stringsAsFactors = FALSE)
    
    
    tbl1 <- create_table(df, first_row_blank = FALSE) %>%
      define(subjid, label = "Subject ID", align = "left") %>% 
      define(name, label = "Subject Name") %>% 
      define(sex, label = "Sex") %>% 
      define(age, label = "Age") %>% 
      define(arm, label = "Arm", 
             blank_before = TRUE, 
             dedupe = TRUE, 
             align = "right")
    
    
    rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
      options_fixed(font_size = 10) %>% 
      page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
      titles("Table 1.0", "Analysis Data Subject Listing", 
             "Safety Population", align = "center") %>%
      footnotes("Program Name: table1_0.R") %>%
      page_footer(left = "Time", center = "Confidential", 
                  right = "Page [pg] of [tpg]") %>%
      add_content(tbl1) 
    
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)
    
  } else 
    expect_equal(TRUE, TRUE)
  
})

test_that("rtf45: Three level stub and indentation work as expected.", {
  
  if (dev == TRUE) {
    fp <- file.path(base_path, "rtf/test45.rtf")
    
    
    # Setup
    cat <- c(rep("Kaplan-Meier estimates", 6), 
             rep("Cox PH estimates and some more really long stuff", 6))
    grp <- c("25th percentile", "25th percentile", 
             "median (weeks)", "median (weeks)",
             "75th percentile", "75th percentile",
             "25th percentile", "25th percentile", 
             "median (weeks)", "median (weeks)",
             "75th percentile", "75th percentile")
    ci <- c(NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval")
    values <- c(41, 53, 43, 39, 47, 52, 38, 25, 37, 23, 78, 21)
    
    # Create data frame
    df <- data.frame(cat, grp, ci, values, stringsAsFactors = FALSE)
    
    tbl1 <- create_table(df) %>%
      stub(c(cat, grp, ci), "Estimates", width = 1.5) %>% 
      define(cat, label_row = TRUE, blank_after = TRUE, indent = 0.15) %>%
      define(grp, indent = .25) %>%
      define(ci, indent = .5) %>%
      define(values, label = "Values")
    
    rpt <- create_report(fp, output_type = "RTF", font = "fixed",
                         font_size = 10) %>%
      titles("Table 3.0", "Analysis of Time to Initial PSGA Success in Weeks") %>% 
      page_header("Sponsor", "Study") %>% 
      add_content(tbl1) %>% 
      page_footer("Time", "Confidential", "Page")
    
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})


test_that("rtf46: Path variable is returned properly from write_report().", {
  
  if (dev) {
    
    fp <- file.path(base_path, "rtf/test46")
    
    dat <- mtcars[1:10, ]
    
    
    tbl <- create_table(dat) %>%
      titles("My Table{symbol('dagger')}")
    
    
    rpt <- create_report(fp, output_type = "RTF", 
                         orientation = "landscape", font = "fixed") %>%
      add_content(tbl)
    
    res <- write_report(rpt)
    
    
    # file.show(res$path)
    
    expect_equal(file.exists(res$modified_path), TRUE)
    expect_equal(is.null(res$path), FALSE)
    
    
  } else {
    
    expect_equal(TRUE, TRUE)
    
  }
  
})

test_that("rtf47: Page numbers work every in title, footnote, header, and footer.", {
  
  if (dev) {
    fp <- file.path(base_path, "rtf/test47.rtf")
    
    dat <- iris
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table Title: Page [pg] of [tpg]") %>%
      titles("Second table title") %>%
      footnotes("Table footnote: Page [pg] of [tpg]") %>%
      footnotes("Second table footnote")
    
    rpt <- create_report(fp, output_type = "RTF", font = "fixed",
                         orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      add_content(tbl) %>%
      titles("Report Title: Page [pg] of [tpg]") %>%
      titles("Second report title") %>%
      footnotes("Report footnote: Page [pg] of [tpg]", borders = "none") %>%
      footnotes("Second report footnote") %>%
      page_header("Header: Page [pg] of [tpg]") %>%
      page_footer("Footer: Page [pg] of [tpg]")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

context("Page Template DOCX Tests")



test_that("get_titles_docx function works as expected.", {


  rpt <- create_report("", output_type = "DOCX", font = "Arial",
                       font_size = 12) %>%
    titles("Hello", "here is \n something", blank_row = "below") %>%
    footnotes("Goodbye", blank_row = "below")

  rpt <- page_setup_docx(rpt)

  t <- get_titles_docx(rpt$titles, 6, rpt)
  t
  expect_equal(length(t$docx), 1)
  expect_equal(t$lines, 4)

})

test_that("get_titles_docx function works as expected with font_size.", {
  
  rpt <- create_report("", output_type = "DOCX", font = "Arial",
                       font_size = 9) %>%
    titles("a a a a a a a a a",
           width = 1)
  
  rpt <- page_setup_docx(rpt)
  
  t9 <- get_titles_docx(rpt$titles, 1, rpt)
  expect_equal(length(t9$docx), 1)
  expect_equal(t9$lines, 2)
  
  # Lines_content returns corrent lines in content size
  rpt <- create_report("", output_type = "DOCX", font = "Arial",
                       font_size = 9) %>%
    titles("a a a a a a a a a",
           width = 1, font_size = 14)
  
  rpt <- page_setup_docx(rpt)
  
  t14 <- get_titles_docx(rpt$titles, 1, rpt)
  expect_equal(length(t14$docx), 1)
  expect_equal(t14$lines, 3)
  expect_equal(t14$lines_content, 5)
  
})


test_that("get_footnotes_docx function works as expected.", {


  rpt <- create_report("", output_type = "DOCX", font = "Arial",
                       font_size = 12) %>%
    titles("Hello", blank_row = "below") %>%
    footnotes("Goodbye", blank_row = "below")

  rpt <- page_setup_docx(rpt)

  f <- get_footnotes_docx(rpt$footnotes, 6, rpt)
  f
  expect_equal(length(f$docx), 1)
  expect_equal(f$lines, 2)

})

test_that("get_footnotes_docx function works as expected with font_size.", {
  
  rpt <- create_report("", output_type = "DOCX", font = "Arial",
                       font_size = 12) %>%
    footnotes("a a a a a a a a a",
              width = 1)
  
  rpt <- page_setup_docx(rpt)
  
  f12 <- get_footnotes_docx(rpt$footnotes, 1, rpt)
  expect_equal(length(f12$docx), 1)
  expect_equal(f12$lines, 3)
  
  rpt <- create_report("", output_type = "DOCX", font = "Arial",
                       font_size = 12) %>%
    footnotes("a a a a a a a a a",
              width = 1, font_size = 8)
  
  rpt <- page_setup_docx(rpt)
  
  f8 <- get_footnotes_docx(rpt$footnotes, 1, rpt)
  expect_equal(length(f8$docx), 1)
  expect_equal(f8$lines, 2)
  
  # Lines_content returns corrent lines in content size
  rpt <- create_report("", output_type = "DOCX", font = "Arial",
                       font_size = 8) %>%
    footnotes("a a a a a a a a a",
              width = 1, font_size = 12)
  
  rpt <- page_setup_docx(rpt)
  
  f12_8 <- get_footnotes_docx(rpt$footnotes, 1, rpt)
  expect_equal(length(f12_8$docx), 1)
  expect_equal(f12_8$lines, 3)
  expect_equal(f12_8$lines_content, 5)
})




test_that("get_title_header_docx function works as expected.", {

  rpt2 <- create_report("", output_type = "DOCX", font = "Arial", 
                        font_size = 12) %>%
    title_header("Hello", right = paste("Right here is something",
    "really long that will wrap and wrap and wrap and wrap keep wrapping"))

  rpt2 <- page_setup_docx(rpt2)

  th <- get_title_header_docx(rpt2$title_hdr, 6, rpt2)
  th
  expect_equal(nchar(th$docx) > 1, TRUE)
  expect_equal(th$lines, 4)

})

 
test_that("get_page_header_docx works as expected.", {

  rpt3 <- create_report("", output_type = "DOCX", font = "Arial", 
                        font_size = 12) %>%
    page_header(left= c("Left"), right = "Right") %>%
                # right = paste("Right here is something that might wrap.",
                #               "If it is long enough so let's make it longer",
                #               "If it is long enough so let's make it longer",
                #               "If it is long enough so let's make it longer")
                # ) %>%
    page_footer("Left", "Center", "Right here is something")

  rpt3 <- page_setup_docx(rpt3)

  ph <- get_page_header_docx(rpt3)
  ph

  expect_equal(ph$lines, 1)

})

 
test_that("get_page_footer_docx works as expected.", {

  rpt3 <- create_report("", font = "Arial", font_size = 12) %>%
    page_header(left= c("Hello"),
                right = paste("Right here is something that might wrap.",
                              "If it is long enough so let's make it longer",
                              "If it is long enough so let's make it longer",
                              "If it is long enough so let's make it longer")) %>%
    page_footer("Left", "Center", "Right here is something")

  rpt3 <- page_setup_docx(rpt3)

  pf <- get_page_footer_docx(rpt3)
  pf

  expect_equal(pf$lines, 1)

})



test_that("get_pageby_docx works as expected.", {

  tbl <- create_table(mtcars) %>%
    page_by(cyl, "Cylinders:")

  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello") %>%
    footnotes("Goodbye") %>%
    add_content(tbl)

  rpt <- page_setup_docx(rpt)

  rpt$line_height

  res <- get_page_by_docx(tbl$page_by, 6, "fork", rpt, "left")
  res

  expect_equal(res$lines, 2)
})


# test_that("get_cell_borders_docx works as expected.", {
#   
#   expect_equal(get_cell_borders_html(1, 1, 1, 1, c("all")),
#       paste0("border-top:thin solid;border-bottom:thin solid;",
#              "border-left:thin solid;border-right:thin solid;"))
#   
#   expect_equal(get_cell_borders_html(1, 1, 1, 1, c("all"), exclude = "bottom"),
#                paste0("border-top:thin solid;",
#                       "border-left:thin solid;border-right:thin solid;"))
#   
# })

context("Plot Spec Tests")


test_that("create_plot parameter checks work as expected.", {
  
  expect_error(create_plot(NULL))
  
  expect_error(create_plot("hello"))
  
})

test_that("plot print() function works as expected.", {
  
  library(ggplot2)
  
  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  
  
  plt <- create_plot(p, height = 4, width = 2) %>% 
    titles("Test Title") %>% 
    footnotes("Test footnote")
  
  res <- capture.output(print(plt))
  
  #print(plt)
  
  expect_equal(length(res) > 0, TRUE)
  
  res <- capture.output(print(plt, verbose = TRUE))
  
  #print(plt, verbose = TRUE)
  
  expect_equal(length(res) > 0, TRUE)
  
  
  
  
})
  
test_that("plot accepts survival plots as expected.", {
  
  p <- list()
  class(p) <- c("ggcoxzph", "ggsurv", "list")
  
  plt <- create_plot(p, height = 4, width = 2)
  
  expect_equal(plt$plot, p) 
  
  class(p) <- c("fork", "list")
  
  expect_error(create_plot(p, 4, 2))
  
  class(p) <- c("gg", "list")
  
  plt <- create_plot(p, height = 4, width = 2)
  
  expect_equal(plt$plot, p)
})

context("Text Tests")

base_path <- "c:/packages/reporter/tests/testthat"

base_path <- tempdir()

cnt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
    "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
    "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
    "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ", 
    "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
    "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
    "qui officia deserunt mollit anim id est laborum.")

test_that("create_text function output works as expected.", {
  
  

  # Create the report object
  rpt <- create_report("dummy.txt", orientation="portrait") %>% 
    add_content(create_text(cnt))
  rpt$line_size <- 104
  rpt$body_line_count <- 55
  
  res1 <- create_text_pages_text(rpt, rpt$content[[1]] , 10)
  

  expect_equal(length(res1), 1)
  expect_equal(length(res1[[1]]), 6)
  
})


test_that("text1: Simplest text output works as expected.", {
  
  fp <- file.path(base_path, "text/text1.txt")
  
  rpt <- create_report(fp, orientation = "portrait",
                       output_type = "TXT") %>%
    titles("Report 1.0", "Simple Text Report") %>% 
    add_content(create_text(cnt)) 
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
})



test_that("text2: Harder text output works as expected.", {
  
  fp <- file.path(base_path, "text/text2.out")
  
  rpt <- create_report(fp, orientation = "portrait") %>%
    titles("Report 2.0", "Harder Text Report") %>% 
    page_header(left = "Client: ABC", right = "Study: 123") %>% 
    add_content(create_text(cnt)) %>% 
    page_footer(left = "Time", 
                center = "Confidential", 
                right ="Page [pg] of [tpg]")

  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
})


test_that("text3: Even harder text output works as expected.", {
  
  fp <- file.path(base_path, "text/text3.out")
  
  rpt <- create_report(fp, orientation = "portrait") %>%
    titles("Report 3.0", "Even Harder Text Report") %>% 
    page_header(left = "Page [pg] of [tpg]", right = "Study: 123") %>% 
    add_content(create_text(cnt)) %>% 
    add_content(create_text(cnt)) %>% 
    add_content(create_text(cnt)) %>% 
    add_content(create_text(cnt)) %>% 
    page_footer(left = "Time", 
                center = "Confidential", 
                right =c("Total Pages [tpg]"))
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)

  expect_equal(length(lns), res$pages * 60)
})


test_that("text4: Long text output works as expected.", {
  
  fp <- file.path(base_path, "text/text4.out")

  l <- paste(rep(cnt, 10), collapse = "\n\n")
  
  rpt <- create_report(fp, orientation = "portrait") %>%
    titles("Report 4.0", "Long Text Report") %>% 
    page_header(left = "Client: ABC", right = "Study: 123") %>% 
    add_content(create_text(l)) %>% 
    page_footer(left = "Page [pg] of [tpg]", 
                center = "Confidential", 
                right = "Time")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * 60)
  
})

test_that("text5: Table and Text output works as expected.", {
  
  fp <- file.path(base_path, "text/text5.out")
  
  tbl1 <- mtcars[1:10, ]
  tbl2 <- mtcars[11:20, ]
  
  rpt <- create_report(fp, orientation = "portrait") %>%
    titles("Report 5.0", "Table and Text Report") %>% 
    page_header(left = "Client: ABC", right = "Study: 123") %>% 
    add_content(create_table(tbl1), page_break = FALSE) %>% 
    add_content(create_text("* NOTE: Car information from 1971."), 
                align = "left") %>% 
    add_content(create_table(tbl2), page_break = FALSE) %>% 
    add_content(create_text("* NOTE: Car information from 1972."), 
                align = "left") %>% 
    page_footer(left = "Time", 
                center = "Confidential", 
                right ="Page [pg] of [tpg]")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
})


test_that("text6: Very Long text output works as expected.", {
  
  debug <- TRUE
  
  if (debug) {
    fp <- file.path(base_path, "text/text6.out")
  
    l <- paste(rep(cnt, 1000), collapse = "\n\n")
    
    rpt <- create_report(fp, orientation = "portrait") %>%
      titles("Report 6.0", "Very long Text Report") %>% 
      page_header(left = "Client: ABC", right = "Study: 123") %>% 
      add_content(create_text(l)) %>% 
      page_footer(left = "Time", 
                  center = "Confidential", 
                  right ="Page [pg] of [tpg]")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
    
    lns <- readLines(fp)
    
    expect_equal(length(lns), res$pages * res$line_count)
  
  } else {
   expect_equal(TRUE, TRUE) 
  }
    
})


test_that("text7: Text page setup landscape works as expected.", {
  
  fp <- file.path(base_path, "text/text7.out")
  
  txt1 <- create_text(cnt) %>% 
    titles("Test Title ", "Test Titles 2", "Test Title 3") %>% 
    footnotes("Footnote 1", "Foonote 2", "Footnote 3", "Footnote 4", "Footnote 5")
  
  rpt <- create_report(fp, orientation = "landscape") %>%
    titles("Report 5.0", "Table and Text Report") %>% 
    page_header(left = "Client: ABC", right = "Study: 123") %>% 
    add_content(txt1, page_break = FALSE) %>% 
    add_content(txt1, page_break = FALSE) %>% 
    add_content(txt1, page_break = FALSE) %>% 
    add_content(txt1, page_break = FALSE) %>% 
    add_content(txt1, page_break = FALSE) %>% 
    page_footer(left = "Time", 
                center = "Confidential", 
                right ="Page [pg] of [tpg]")
  
  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  expect_equal(length(lns), 135)
})

# This test also puts the table footnote block on a page boundary 
test_that("text8: Text page setup portrait works as expected.", {
  
  fp <- file.path(base_path, "text/text8.out")
  
  txt1 <- create_text(cnt) %>% 
    titles("Test Title ", "Test Titles 2", "Test Title 3") %>% 
    footnotes("Footnote 1", "Foonote 2", "Footnote 3", "Footnote 4", "Footnote 5")
  
  rpt <- create_report(fp, orientation = "portrait") %>%
    titles("Report 5.0", "Table and Text Report") %>% 
    page_header(left = "Client: ABC", right = "Study: 123") %>% 
    add_content(txt1, page_break = FALSE) %>% 
    add_content(txt1, page_break = FALSE) %>% 
    add_content(txt1, page_break = FALSE) %>% 
    add_content(txt1, page_break = FALSE) %>% 
    add_content(txt1, page_break = FALSE) %>% 
    page_footer(left = "Time", 
                center = "Confidential", 
                right ="Page [pg] of [tpg]")
  
  write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  expect_equal(length(lns), 120)
})


test_that("text9: Title header with text output works as expected.", {
  
  fp <- file.path(base_path, "text/text9.out")
  
  rpt <- create_report(fp, orientation = "portrait") %>%
    title_header("Report 1.0", "Simple Text Report", 
                 right = c("Client", "Study"), blank_row = "below") %>% 
    add_content(create_text(cnt)) 
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
})


test_that("text10: Title header with text output works as expected.", {
  
  fp <- file.path(base_path, "text/text10.out")
  
  txt <- create_text(cnt, width = 5) %>% 
    title_header("Report 1.0", "Simple Text Report", 
                 right = c("Client", "Study"), blank_row = "below") 
  
  rpt <- create_report(fp, orientation = "portrait") %>%
    add_content(txt)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
})


test_that("text11: Title header with text output and border works as expected.", {
  
  fp <- file.path(base_path, "text/text11.out")
  
  rpt <- create_report(fp, orientation = "portrait") %>%
    title_header("Report 1.0", "Simple Text Report", 
                 right = c("Client", "Study"), blank_row = "below",
                 borders = "all") %>% 
    add_content(create_text(cnt)) 
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
})


test_that("text12: Title and footnote with borders on text output works as expected.", {
  
  fp <- file.path(base_path, "text/text12.out")
  
  rpt <- create_report(fp, orientation = "portrait") %>%
    page_header("Left", "Right") %>% 
    titles("Report 1.0", "Simple Text Report", blank_row = "below",
           borders = "all") %>%
    footnotes("Footnote one", "Footnote 2", borders = "all") %>% 
    add_content(create_text(cnt)) 
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
})


test_that("text13: Title header with text output and border works as expected.", {
  
  fp <- file.path(base_path, "text/text13.out")
  
  
  txt <- create_text(cnt) %>% 
    title_header("Report 1.0", "Simple Text Report", 
                 right = c("Client", "Study"), blank_row = "below",
                 borders = "all") %>% 
    footnotes("Here is a footnote", borders = "all")
    
  rpt <- create_report(fp, orientation = "portrait") %>%
    add_content(txt) 
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
})


test_that("text14: Title and footnote with borders on text output works as expected.", {
  
  fp <- file.path(base_path, "text/text14.out")
  
  txt <- create_text(cnt) %>% 
    titles("Report 1.0", "Simple Text Report", blank_row = "below",
           borders = "all") %>%
    footnotes("Footnote one", "Footnote 2", borders = "all") 
    
  rpt <- create_report(fp, orientation = "portrait") %>%
    page_header("Left", "Right") %>% 
    page_footer("Other left", "", "Other Right") %>% 
    add_content(txt) 
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
})


test_that("text15: Text borders work as expected.", {
  
  fp <- file.path(base_path, "text/text15.out")
  
  txt <- create_text(cnt, borders = "all") %>% 
    titles("Report 1.0", "Simple Text Report", blank_row = "below",
           borders = "none") %>%
    footnotes("Footnote one", "Footnote 2", borders = "none") 
  
  rpt <- create_report(fp, orientation = "portrait") %>%
    page_header("Left", "Right") %>% 
    page_footer("Other left", "", "Other Right") %>% 
    add_content(txt) 
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
})


test_that("text16: Two page text spec with titles and footnotes.", {
  
  
  fp <- file.path(base_path, "text/text16.out")
  
  cnttxt <- paste(rep(cnt, 20), collapse = "")
  
  txt <- create_text(cnttxt)  %>%
    titles("Text 1.0", "My Nice Text") %>%
    footnotes("My footnote 1", "My footnote 2")
  
  rpt <- create_report(fp, output_type = "TXT") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(txt) %>%
    page_footer("Left1", "Center1", "Right1")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(res$modified_path), TRUE)
  
  lns <- readLines(res$modified_path)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
  
})

context("PDF Tests")


base_path <- paste0(getwd(),"/tests/testthat")

base_path <- tempdir()

cnt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
              "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
              "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
              "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ", 
              "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
              "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
              "qui officia deserunt mollit anim id est laborum.")

dev <- FALSE



test_that("pdf1: Simplest table works as expected.", {

  if (dev == TRUE) {


  # if (dev & rmarkdown::pandoc_available("1.12.3")) {

  fp <- file.path(base_path, "pdf/test1.pdf")

  rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
    add_content(create_table(mtcars[1:10, ]), align = "left")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  
  } else 
    expect_equal(TRUE, TRUE)



})

test_that("pdf2: Simplest table with title works as expected.", {

  if (dev == TRUE) {


    fp <- file.path(base_path, "pdf/test2.pdf")

    tbl <- create_table(mtcars[1:10, ]) %>%
      define(vs, visible = FALSE)

    rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
      options_fixed(font_size = 10) %>%
      titles("MTCARS Data Frame", align = "left") %>%
      add_content(tbl)


    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)
    
  } else 
    expect_equal(TRUE, TRUE)


})

test_that("pdf3: Table with break between sections works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "pdf/test3.pdf")

  # Setup
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma",
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
  arm <- c(rep("A", 5), rep("B", 5))

  # Create data frame
  df <- data.frame(subjid, name, sex, age, arm, stringsAsFactors = FALSE)


  tbl1 <- create_table(df, first_row_blank = TRUE) %>%
    define(subjid, label = "Subject ID", align = "left") %>%
    define(name, label = "Subject Name") %>%
    define(sex, label = "Sex") %>%
    define(age, label = "Age") %>%
    define(arm, label = "Arm",
           blank_after = TRUE,
           dedupe = TRUE,
           align = "right")


  rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
    options_fixed(font_size = 10) %>%
    page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
    titles("Table 1.0", "Analysis Data Subject Listing",
           "Safety Population", align = "center") %>%
    footnotes("Program Name: table1_0.R") %>%
    page_footer(left = "Time", center = "Confidential",
                right = "Page [pg] of [tpg]") %>%
    add_content(tbl1)


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  
  
  } else 
    expect_equal(TRUE, TRUE)


})

test_that("pdf4: Table that spans multiple pages breaks as expected.", {

  if (dev == TRUE) {

  
    fp <- file.path(base_path, "pdf/test4.pdf")

    rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
      page_header("left", "right") %>%
      titles("IRIS Data Frame") %>%
      add_content(create_table(iris)) %>%
      page_footer("left", "center", "Page [pg] of [tpg]")


    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)

  #write_registration_file(file.path(base_path,"./rtf/reg.txt"))
    
  } else 
    expect_equal(TRUE, TRUE)
})

test_that("pdf5: Table with long cell and label values wraps as expected.", {


    fp <- file.path(base_path, "pdf/test5.pdf")

    # Setup
    arm <- c(rep("A", 5), rep("B", 5))
    subjid <- 100:109
    name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
              "al-Kazemi, Najwa \nand more and more", "Schaffer, Ashley", "Laner, Tahma",
              "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
              "Smith, Shaileigh")
    sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
    age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)


    # Create data frame
    df <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)


    tbl1 <- create_table(df, first_row_blank = TRUE) %>%
      define(subjid, label = "Subject ID for a patient", n = 10, align = "left",
             width = 1) %>%
      define(name, label = "Subject Name", width = 1) %>%
      define(sex, label = "Sex", n = 10, align = "center") %>%
      define(age, label = "Age", n = 10) %>%
      define(arm, label = "Arm",
             blank_after = TRUE,
             dedupe = TRUE)


    rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
      titles("Table 1.0", align = "center") %>%

      add_content(tbl1)


    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)


})

test_that("pdf6: Table with spanning headers works as expected.", {


    fp <- file.path(base_path, "pdf/test6.pdf")


    df <- data.frame(vehicle = rownames(mtcars), mtcars, stringsAsFactors = FALSE)
    rownames(df) = NULL

    df$qsec <- fattr(df$qsec, format = "%.1f")
    df$wt <- fattr(df$wt, justify = "center", width = .75)

    tbl <- create_table(df) %>%
      spanning_header("mpg", "hp",
                      label = "Span 1", label_align = "center", n = 10) %>%
      spanning_header("drat", "qsec",
                      label = "Span 2", label_align = "center", n = 10) %>%
      spanning_header("vs", "carb",
                      label = "Span 3", label_align = "center", n = 10) %>%
      spanning_header(from = "drat", to = "carb", label = "Super Span",
                      label_align = "center",
                      level = 2) %>%
      define(vehicle, label = "Vehicle") %>%
      define(mpg, format = "%.1f")

    rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
      add_content(tbl) %>%
      titles("Table 1.0", "MTCARS Subset Test")

    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)


})




test_that("pdf7: Simplest PDF report with 1 in margins works as expected.", {

  if (dev == TRUE) {


    fp <- file.path(base_path, "pdf/test7.pdf")

    tbl <- create_table(mtcars[1:10, ]) %>%
      column_defaults(width = .5) %>%
      define(vs, visible = FALSE)

    rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
      page_header("Client", "Study") %>%
      titles("MTCARS Data Frame") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl, align = "left") %>%
      page_footer("Time", right = "Page [pg] of [tpg]")



    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)

  } else 
    expect_equal(TRUE, TRUE)
  
})


test_that("pdf8: Two page PDF report works as expected.", {

  if (dev == TRUE) {


    fp <- file.path(base_path, "pdf/test8.pdf")

    # Setup
    subjid <- 100:109
    name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
              "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma",
              "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
              "Smith, Shaileigh")
    sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
    age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
    arm <- c(rep("A", 5), rep("B", 5))

    # Create data frame
    df <- data.frame(subjid, name, sex, age, arm, stringsAsFactors = FALSE)

    df1 <- df[df$arm == "A", ]
    df2 <- df[df$arm == "B", ]

    afmt <- value(condition(x == "A", "Placebo"),
                  condition(x == "B", "Treatment 1"))

    sfmt1 <- value(condition(x == "M", "Male"),
                   condition(x == "F", "Female"),
                   condition(TRUE, "Other"))

    sfmt2 <- c(M = "Male", F = "Female")

    tbl1 <- create_table(df1, width = 7) %>%
      define(sex, width = 1, format = sfmt1) %>%
      define(name, width = 2) %>%
      define(age, width = 1, align = "left")

    tbl2 <- create_table(df2, width = 7) %>%
      define(sex, width = 1, format = sfmt2) %>%
      define(age, format = "%0d%%", align = "left") %>%
      define(name, width = 2) %>%
      define(arm, format = afmt, width = 2, align = "right")


    rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
      options_fixed(font_size = 10) %>%
      set_margins(top = 1, bottom = 1) %>%
      page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
      titles("Table 1.0", "Analysis Data Subject Listing",
             "Safety Population", align = "center") %>%
      footnotes("Program Name: table1_0.R") %>%
      page_footer(left = "Time", center = "Confidential",
                  right = "Page [pg] of [tpg]") %>%
      add_content(tbl1) %>%
      add_content(tbl2)


    res <- write_report(rpt)

    #print(res)

    expect_equal(file.exists(fp), TRUE)
    
  } else 
    expect_equal(TRUE, TRUE)


})




test_that("pdf9: Simplest PDF Plot works as expected.", {


    library(ggplot2)

    fp <- file.path(base_path, "pdf/test9.pdf")

    p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()

    plt <- create_plot(p, height = 4, width = 8)


    rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
      page_header("Client", "Study: XYZ") %>%
      titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(plt, align = "center") %>%
      footnotes("* Motor Trend, 1974") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")


    res <- write_report(rpt)

    #print(res)

    expect_equal(file.exists(fp), TRUE)


})


test_that("pdf10: PDF Table with Plot works as expected.", {


    library(ggplot2)

    fp <- file.path(base_path, "pdf/test10.pdf")

    p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()


    plt <- create_plot(p, height = 4, width = 8)
    tbl <- create_table(mtcars[1:10, ])


    rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
      page_header("Client", "Study: XYZ") %>%
      titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      add_content(plt, align = "center") %>%
      footnotes("* Motor Trend, 1974") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")


    res <- write_report(rpt)

    #print(res)

    expect_equal(file.exists(fp), TRUE)


})




test_that("pdf11: PDF Table with Plot on same page works as expected.", {
  

  library(ggplot2)

  fp <- file.path(base_path, "pdf/test11.pdf")

  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()

  plt <- create_plot(p, height = 4, width = 8)
  tbl <- create_table(mtcars[1:3, ])


  rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", blank_row = "none") %>%
    set_margins(top = 1, bottom = 1) %>%

    add_content(plt, page_break = FALSE, blank_row = "none") %>%
    add_content(tbl) %>%
    footnotes("* Motor Trend, 1974") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")


  res <- write_report(rpt)

  #print(res)

  expect_equal(file.exists(fp), TRUE)


})


test_that("pdf12: Table and Text output works as expected.", {


  fp <- file.path(base_path, "pdf/test12.pdf")

  tbl1 <- mtcars[1:10, ]
  tbl2 <- mtcars[11:20, ]

  rpt <- create_report(fp, orientation = "portrait", output_type = "PDF", 
                       font = "fixed") %>%
    titles("Report 5.0", "Table and Text Report") %>%
    page_header(left = "Client: ABC", right = "Study: 123") %>%
    add_content(create_table(tbl1), page_break = FALSE) %>%
    add_content(create_text("* NOTE: Car information from 1971."),
                align = "left") %>%
    add_content(create_table(tbl2), page_break = FALSE) %>%
    add_content(create_text("* NOTE: Car information from 1972."),
                align = "left") %>%
    page_footer(left = "Time",
                center = "Confidential",
                right ="Page [pg] of [tpg]")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)


})


test_that("pdf13: Very Long text output works as expected.", {

  if (dev) {

    fp <- file.path(base_path, "pdf/test13.pdf")

    l <- paste(rep(cnt, 1000), collapse = "\n\n")

    rpt <- create_report(fp, orientation = "portrait", output_type = "PDF", 
                         font = "fixed") %>%
      titles("Report 6.0", "Very long Text Report") %>%
      page_header(left = "Client: ABC", right = "Study: 123") %>%
      add_content(create_text(l)) %>%
      page_footer(left = "Time",
                  center = "Confidential",
                  right ="Page [pg] of [tpg]")

    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)

  } else
    expect_equal(TRUE, TRUE)

})


test_that("pdf14: Simplest portrait table works as expected.", {



    fp <- file.path(base_path, "pdf/test14.pdf")

    rpt <- create_report(fp, output_type = "PDF", orientation = "portrait", 
                         font = "fixed") %>%
      page_header("left", "right") %>%
      titles("Table 1.0", "MTCARS Data Frame") %>%
      add_content(create_table(mtcars)) %>%
      page_footer("Left", right = "right")

    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)



})


test_that("pdf15: Simplest landscape table works as expected.", {

  if (dev == TRUE) {


    fp <- file.path(base_path, "pdf/test15.pdf")

    rpt <- create_report(fp, output_type = "PDF", orientation = "landscape", 
                         font = "fixed") %>%
      page_header("left", "right") %>%
      titles("Table 1.0", "MTCARS Data Frame") %>%
      add_content(create_table(mtcars)) %>%
      page_footer("Left", right = "right")

    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)
    
  } else 
    expect_equal(TRUE, TRUE)


})

test_that("pdf16: 10 pt report with units in cm works as expected.", {

  if (dev == TRUE) {

  
    fp <- file.path(base_path, "pdf/test16.pdf")


    rpt <- create_report(fp, units = "cm", output_type = "PDF", 
                         font = "fixed") %>%
      page_header("Client: Experis", "Study: ABC") %>%
      titles("IRIS Data Frame") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]") %>%
      add_content(create_table(iris))


    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)
    
  } else 
    expect_equal(TRUE, TRUE)



})


test_that("pdf17: 12 pt report with units in cm works as expected.", {

  if (dev == TRUE) {


    fp <- file.path(base_path, "pdf/test17.pdf")


    rpt <- create_report(fp, units = "cm", output_type = "PDF", 
                         font = "fixed") %>%
      options_fixed(font_size = 12) %>%
      page_header("Client: Experis", "Study: ABC") %>%
      titles("IRIS Data Frame") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]") %>%
      add_content(create_table(iris))


    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)
    
  } else 
    expect_equal(TRUE, TRUE)


})


test_that("pdf18: Plot with page by on report works as expected.", {

  if (dev == TRUE) {

  
    library(ggplot2)

    fp <- file.path(base_path, "pdf/test18.pdf")


    dat <- mtcars[order(mtcars$cyl), ]

    p <- ggplot(dat, aes(x=disp, y=mpg)) + geom_point()


    #dats <- split(p$data, p$data$grp)
    #tbl <- create_table(dat[1:3, ])

    plt <- create_plot(p, height = 4, width = 8)


    rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
      page_header("Client", "Study: XYZ") %>%
      titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", blank_row = "none") %>%
      set_margins(top = 1, bottom = 1) %>%
      page_by(cyl, "Cylinders: ", align = "right", blank_row = "none") %>%
      add_content(plt) %>%
      footnotes("* Motor Trend, 1974") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")


    res <- write_report(rpt)

    #print(res)

    expect_equal(file.exists(fp), TRUE)
    
  } else 
    expect_equal(TRUE, TRUE)


})



test_that("pdf19: Plot with page by on plot works as expected.", {

  if (dev == TRUE) {

  
    library(ggplot2)

    fp <- file.path(base_path, "pdf/test19.pdf")

    fmt <- value(condition(x == 4, "4 Cylinder"),
                 condition(x == 6, "6 Cylinder"),
                 condition(x == 8, "8 Cylinder"))
    
    dat <- mtcars[order(mtcars$cyl), ]

    p <- ggplot(dat, aes(x=disp, y=mpg)) + geom_point()


    #dats <- split(p$data, p$data$grp)
    #tbl <- create_table(dat[1:3, ])

    plt <- create_plot(p, height = 4, width = 8) %>%
      titles("Figure 1.0", "MTCARS Mileage By Displacement", blank_row = "none") %>%
      page_by(cyl, "Cylinders: ", align = "left", 
              blank_row = "none", format = fmt) %>%
      footnotes("* Motor Trend, 1974")

    rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
      page_header("Sponsor", "Study: cars") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(plt) %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")


    res <- write_report(rpt)

    #print(res)

    expect_equal(file.exists(fp), TRUE)
    
  } else 
    expect_equal(TRUE, TRUE)


})



test_that("test21: 8 pt report with units in inches works as expected.", {

  if (dev == TRUE) {


    fp <- file.path(base_path, "pdf/test21.pdf")

    tbl <- create_table(iris)

    rpt <- create_report(fp, units = "inches", output_type = "PDF", 
                         font = "fixed") %>%
      page_header("Client: Experis", "Study: ABC") %>%
      titles("IRIS Data Frame") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]") %>%
      options_fixed(font_size = 8) %>%
      add_content(tbl)


    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)
    
  } else 
    expect_equal(TRUE, TRUE)


})

test_that("test22: 8 pt report with units in cm works as expected.", {

  if (dev == TRUE) {

  
    fp <- file.path(base_path, "pdf/test22.pdf")


    rpt <- create_report(fp, units = "cm", output_type = "PDF", 
                         font = "fixed") %>%
      page_header("Client: Experis", "Study: ABC") %>%
      titles("IRIS Data Frame") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]") %>%
      options_fixed(font_size = 8) %>%
      add_content(create_table(iris))


    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)
    
    
  } else 
    expect_equal(TRUE, TRUE)


})




test_that("pdf23: PDF Table with Plot and borders works as expected.", {

  if (dev == TRUE) {

  
    library(ggplot2)

    fp <- file.path(base_path, "pdf/test23.pdf")

    p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()


    plt <- create_plot(p, height = 4, width = 8)
    tbl <- create_table(mtcars[1:10, ])


    rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
      page_header("Client", "Study: XYZ") %>%
      titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", borders = "all") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      add_content(plt, align = "center") %>%
      footnotes("* Motor Trend, 1974", borders = "all") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]") %>% 
      options_fixed(uchar = "-")


    res <- write_report(rpt)

    res

    expect_equal(file.exists(fp), TRUE)
    
    
  } else 
    expect_equal(TRUE, TRUE)

})



test_that("pdf24: PDF Table with Plot and borders on content works as expected.", {

    library(ggplot2)

    fp <- file.path(base_path, "pdf/test24.pdf")

    p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()


    plt <- create_plot(p, height = 4, width = 8) %>%
      titles("My plot", borders = "all") %>%
      footnotes("My plot footnotes", borders = "all")

    tbl <- create_table(mtcars[1:10, ]) %>%
      titles("My table", borders = "all") %>%
      footnotes("My table footnotes", borders = "all", align = "right")


    rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
      page_header("Client", "Study: XYZ") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      add_content(plt, align = "center") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")


    res <- write_report(rpt)

    #print(res)

    expect_equal(file.exists(fp), TRUE)


})

test_that("pdf25: PDF Table with custom options works as expected.", {
  
  if (dev == TRUE) {

  fp <- file.path(base_path, "pdf/test25.pdf")
  
  
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    titles("My table", borders = "all") %>% 
    footnotes("My table footnotes", borders = "all", align = "right")
  
  tbl2 <- create_table(mtcars[11:20, ]) %>% 
    titles("My table", borders = "all") %>% 
    footnotes("My table footnotes", borders = "all", align = "right")
  
  
  rpt <- create_report(fp, font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    add_content(tbl2, align = "center") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]") %>% 
    options_fixed(font_size = 12, line_size = 80, line_count = 30,
                  uchar = "~")
  
  
  res <- write_report(rpt, output_type = "pdf")
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)
  
  } else 
    expect_equal(TRUE, TRUE)
})

test_that("pdf26: Table Borders that spans multiple pages work as expected.", {
  
  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf/test26.pdf")
  
  rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris, borders = "all")) %>% 
    footnotes("Here is a footnote")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 5)
  #write_registration_file(file.path(base_path,"./rtf/reg.txt"))
  
  } else 
    expect_equal(TRUE, TRUE)
})


test_that("pdf27: Table Borders with ttls/fnts on table works as expected.", {
  
  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf/test27.pdf")
  
  tbl <- create_table(iris, borders = "all") %>% 
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below") %>% 
    footnotes("Here is a footnote", "And another")
  
  rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
    page_header("Left", "Right") %>% 
    add_content(tbl) %>% 
    page_footer("left", "", "right")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  } else 
    expect_equal(TRUE, TRUE)
  
})

test_that("pdf28: Table Borders with ttls/fnts on report works as expected.", {
  
  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf/test28.pdf")
  
  tbl <- create_table(iris, borders = "all") 
  
  rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
    page_header("Left", "Right") %>% 
    add_content(tbl) %>% 
    page_footer("left", "", "right") %>% 
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below") %>% 
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  
  } else 
    expect_equal(TRUE, TRUE)
  

  
})

test_that("pdf29: 9 pt font inches works as expected.", {
  
  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf/test29.pdf")
  
  rpt <- create_report(fp, output_type = "PDF", font_size = 9, 
                       orientation = "portrait", font = "fixed") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>% 
    set_margins(top = 1, bottom = 1)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  #write_registration_file(file.path(base_path,"./rtf/reg.txt"))
  
  } else 
    expect_equal(TRUE, TRUE)
})

test_that("pdf30: 9 pt font cm works as expected.", {
  
  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf/test30.pdf")
  
  rpt <- create_report(fp, output_type = "PDF", font_size = 9, 
                       orientation = "portrait", font = "fixed") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>% 
    set_margins(top = 1, bottom = 1)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  #write_registration_file(file.path(base_path,"./rtf/reg.txt"))
  
  } else 
    expect_equal(TRUE, TRUE)
})

test_that("pdf31: 11 pt font inches works as expected.", {
  
  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf/test31.pdf")
  
  rpt <- create_report(fp, output_type = "PDF", font_size = 11, 
                       orientation = "portrait", font = "fixed") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>% 
    set_margins(top = 1, bottom = 1)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  #write_registration_file(file.path(base_path,"./rtf/reg.txt"))
  
  } else 
    expect_equal(TRUE, TRUE)
})

test_that("pdf32: 11 pt font cm works as expected.", {
  
  if (dev == TRUE) {

  
  fp <- file.path(base_path, "pdf/test32.pdf")
  
  rpt <- create_report(fp, output_type = "PDF", font_size = 11, 
                       orientation = "portrait", font = "fixed") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>% 
    set_margins(top = 1, bottom = 1)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  #write_registration_file(file.path(base_path,"./rtf/reg.txt"))
  
  } else 
    expect_equal(TRUE, TRUE)
})


test_that("pdf33: PDF Image file works as expected.", {
  
  if (dev == TRUE) {

  
  library(ggplot2)
  
  fp <- file.path(base_path, "pdf/test33.pdf")
  
  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  
  pltpath <- file.path(base_path, "pdf/test33.jpg")
  ggsave(pltpath, width = 8, height = 4, 
         units = "in",
         dpi = 300)
  
  plt <- create_plot(pltpath, height = 4, width = 8)
  
  
  rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt, align = "center") %>%
    footnotes("* Motor Trend, 1974") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  } else 
    expect_equal(TRUE, TRUE)
})

test_that("pdf34: Blank after on invisible column.", {
  
  fp <- file.path(base_path, "pdf/test34.pdf")
  
  tbl <- create_table(iris, borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
    page_header("Left", "Right") %>%
    add_content(tbl) %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("pdf35: Page header width works as expected.", {
  
  fp <- file.path(base_path, "pdf/test35.pdf")
  
  tbl <- create_table(iris[1:10, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
    page_header("Left here is some stuff and more stuff trying to get out in the middle", 
                "Right", width = 8) %>%
    add_content(tbl) %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("pdf36: Custom page size works as expected.", {
  
  fp <- file.path(base_path, "pdf/test36.pdf")
  
  tbl <- create_table(iris[1:15, ]) %>%
    define(Species, visible = FALSE)
  
  ttl <- c("Title1", "Title2", "Title3")
  
  rpt <- create_report(fp, output_type = "PDF", 
                       paper_size = c(6.5, 7.5),
                       orientation = "portrait", font = "fixed") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles(ttl,
           blank_row = "below", columns =  1, align = "center",
           borders = "none") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("pdf37: Stub indent.", {
  
  
  fp <- file.path(base_path, "pdf/test37.pdf")
  
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder and more perhaps more" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  ll <- "Here is a super long label to see if it can span the entire table."
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE, borders = c("all")) %>% 
    stub(c("var", "label"), width = 1) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = ll, cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center", n = 19) %>% 
    define(B, label = "Group B", align = "center", n = 13)
  
  
  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "PDF",
                       font = "fixed") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left", 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  
  
  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)
  
  
})

test_that("pdf38: Title and footnotes Columns work as expected.", {
  
  fp <- file.path(base_path, "pdf/test38.pdf")
  
  tbl <- create_table(iris[1:15, ]) %>%
    define(Species, visible = FALSE)
  
  ttl <- c("Title1", "Title2", "Title3")
  
  rpt <- create_report(fp, output_type = "PDF", 
                       orientation = "portrait", font = "fixed") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles(ttl,
           blank_row = "below", columns =  3, align = "center",
           borders = "none") %>%
    footnotes("Here is a footnote", "And another", columns = 2, borders = "top")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("pdf39: Table with blank_before works as expected.", {
  
  if (dev == TRUE) {
    
    
    fp <- file.path(base_path, "pdf/test39.pdf")
    
    # Setup
    subjid <- 100:109
    name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
              "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma",
              "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
              "Smith, Shaileigh")
    sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
    age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
    arm <- c(rep("A", 5), rep("B", 5))
    
    # Create data frame
    df <- data.frame(subjid, name, sex, age, arm, stringsAsFactors = FALSE)
    
    
    tbl1 <- create_table(df, first_row_blank = FALSE) %>%
      define(subjid, label = "Subject ID", align = "left") %>%
      define(name, label = "Subject Name") %>%
      define(sex, label = "Sex") %>%
      define(age, label = "Age") %>%
      define(arm, label = "Arm",
             blank_before = TRUE,
             dedupe = TRUE,
             align = "right")
    
    
    rpt <- create_report(fp, output_type = "PDF", font = "fixed") %>%
      options_fixed(font_size = 10) %>%
      page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
      titles("Table 1.0", "Analysis Data Subject Listing",
             "Safety Population", align = "center") %>%
      footnotes("Program Name: table1_0.R") %>%
      page_footer(left = "Time", center = "Confidential",
                  right = "Page [pg] of [tpg]") %>%
      add_content(tbl1)
    
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
    
    
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf40: Three level stub and indentation work as expected.", {
  
  if (dev == TRUE) {
    fp <- file.path(base_path, "pdf/test40.pdf")
    
    
    # Setup
    cat <- c(rep("Kaplan-Meier estimates", 6), 
             rep("Cox PH estimates and some more really long stuff", 6))
    grp <- c("25th percentile", "25th percentile", 
             "median (weeks)", "median (weeks)",
             "75th percentile", "75th percentile",
             "25th percentile", "25th percentile", 
             "median (weeks)", "median (weeks)",
             "75th percentile", "75th percentile")
    ci <- c(NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval")
    values <- c(41, 53, 43, 39, 47, 52, 38, 25, 37, 23, 78, 21)
    
    # Create data frame
    df <- data.frame(cat, grp, ci, values, stringsAsFactors = FALSE)
    
    tbl1 <- create_table(df) %>%
      stub(c(cat, grp, ci), "Estimates", width = 1.5) %>% 
      define(cat, label_row = TRUE, blank_after = TRUE, indent = 0.15) %>%
      define(grp, indent = .25) %>%
      define(ci, indent = .5) %>%
      define(values, label = "Values")
    
    rpt <- create_report(fp, output_type = "PDF", font = "fixed",
                         font_size = 10) %>%
      titles("Table 3.0", "Analysis of Time to Initial PSGA Success in Weeks") %>% 
      page_header("Sponsor", "Study") %>% 
      add_content(tbl1) %>% 
      page_footer("Time", "Confidential", "Page")
    
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("pdf41: Page numbers work every in title, footnote, header, and footer.", {
  
  if (dev) {
    fp <- file.path(base_path, "pdf/test41.pdf")
    
    dat <- iris
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table Title: Page [pg] of [tpg]") %>%
      footnotes("Table footnote: Page [pg] of [tpg]")
    
    rpt <- create_report(fp, output_type = "PDF", font = "fixed",
                         orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      add_content(tbl) %>%
      titles("Report Title: Page [pg] of [tpg]") %>%
      footnotes("Report footnote: Page [pg] of [tpg]", borders = "none") %>%
      page_header("Header: Page [pg] of [tpg]") %>%
      page_footer("Footer: Page [pg] of [tpg]")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

context("Print Utilities Tests")

base_path <- "c:/packages/reporter/tests/testthat"

base_path <- tempdir()


test_that("title and footnote printing functions work as expected.", {
  
  txt <- create_text("Test") %>% 
    titles("One") %>% 
    footnotes("Three")
  
  
  res1 <- capture.output(print(txt))
  expect_equal(length(res1) > 0, TRUE)
  
  txt2 <- create_text("Test") %>% 
    title_header("One") %>% 
    footnotes("Three")
  
  
  res2 <- capture.output(print(txt2))
  expect_equal(length(res2) > 0, TRUE)
  
  
})


test_that("report printing functions work as expected.", {
  
  txt <- create_text("Test") %>% 
    titles("One") %>% 
    footnotes("Three")
  
  
  rpt <- create_report("") %>% 
    add_content(txt)
  
  
  res1 <- capture.output(print(rpt))
  expect_equal(length(res1) > 0, TRUE)
  
  txt2 <- create_table(mtcars) %>% 
    title_header("One") %>% 
    footnotes("Three")
  
  rpt2 <- create_report("") %>% 
    add_content(txt2)
  
  
  res2 <- capture.output(print(rpt2))
  expect_equal(length(res2) > 0, TRUE)
  
  
  res3 <- capture.output(print(rpt2, verbose = TRUE))
  expect_equal(length(res3) > 0, TRUE)
  
  
})


test_that("write_registration_file works as expected", {
  
  pth <- file.path(base_path, "reg.txt")
  
  write_registration_file(pth)
  
  expect_equal(TRUE, TRUE)
  
})

context("Page Template Tests")


test_that("titles function works as expected.", {
  
  
  ttl <- titles(list(), "My title 1", "My title 2", 
                align = "left", blank_row = "both")
  
  
  expect_equal(ttl$titles[[1]]$align, "left")
  expect_equal(ttl$titles[[1]]$blank_row, "both")
  expect_equal(length(ttl$titles[[1]]$titles), 2)
  
})
  

test_that("get_titles function works as expected.", {
  
  
  ttl <- titles(list(), "My title 1", "My title 2", 
                align = "left", blank_row = "both")
  
  expect_error(get_titles(ttl, 30))
  
  res3 <- get_titles(ttl$titles, 30, 50, "-", 1)
  
  res3
  
  expect_equal(length(res3), 4)
  expect_equal(nchar(res3[2]), 30)
  
  ttl <- titles(list(), "My title 1", "My title 2", 
                align = "left", blank_row = "both", borders = "all")
  
  
  res3 <- get_titles(ttl$titles, 30, 50, "-", 1)
  
  res3
  
  expect_equal(length(res3), 6)
  expect_equal(nchar(res3[2]), 30)
  
  
})
  

test_that("footnotes function works as expected.", {
  
  
  ftn <- footnotes(list(), "My footnote 1", "My footnote 2", 
                align = "left", blank_row = "both")
  
  
  expect_equal(ftn$footnotes[[1]]$align, "left")
  expect_equal(ftn$footnotes[[1]]$blank_row, "both")
  expect_equal(length(ftn$footnotes[[1]]$footnotes), 2)
  
})


test_that("get_footnotes function works as expected.", {
  
  
  ftn <- footnotes(list(), "My footnote 1", "My footnote 2", 
                   align = "left", blank_row = "both")
  
  expect_error(get_footnotes(ftn, 30))
  
  res3 <- get_footnotes(ftn$footnotes, 30, 50, "-", 1)
  
  res3
  
  expect_equal(length(res3), 4)
  expect_equal(nchar(res3[2]), 30)
  
})

test_that("get_footnotes function with borders works as expected.", {
  
  
  ftn <- footnotes(list(), "My footnote 1", "My footnote 2", 
                   align = "left", blank_row = "both", borders = "all")
  
  expect_error(get_footnotes(ftn, 30))
  
  res3 <- get_footnotes(ftn$footnotes, 30, 50, "-", 1)
  
  res3
  
  expect_equal(length(res3), 6)
  expect_equal(nchar(res3[2]), 30)
  
})


test_that("page_header function works as expected.", {
  
  
  ph <- page_header(create_report(), left = c("Left 1", "Left 2"), 
                right = c("Right 1", "Right 2", "Right 3"),
                blank_row = "below")
  
  
  expect_equal(length(ph$page_header_left), 2)
  expect_equal(length(ph$page_header_right), 3)
  expect_equal(ph$page_header_blank_row, "below")
  
  expect_error(page_header(create_report(), left = c("a", "b", "c", "d", "e", "f")))
  expect_error(page_header(create_report(), blank_row = "above"))
               
  
})


test_that("page_footer function works as expected.", {
  
  
  ph <- page_footer(create_report(), left = c("Left 1", "Left 2"), 
                    center = "center",
                    right = c("Right 1", "Right 2", "Right 3"),
                    blank_row = "above")
  
  
  expect_equal(length(ph$page_footer_left), 2)
  expect_equal(length(ph$page_footer_center), 1)
  expect_equal(length(ph$page_footer_right), 3)
  expect_equal(ph$page_footer_blank_row, "above")
  
  expect_error(page_header(create_report(), left = c("a", "b", "c", "d", "e", "f")))
  expect_error(page_header(create_report(), blank_row = "above"))
  
})


test_that("title_header function works as expected.", {
  
  th <- title_header(create_report(), "Title One", "Title Two",
                     right = c("One", "Two", "Three"),
                     blank_row = "both")
  
  
  rws <- get_title_header(th$title_hdr, 50, 60, "-", 1)
  
  expect_equal(length(rws), 5)
  expect_equal(rws[1], "")
  
  th <- title_header(create_table(mtcars), 
                     right = c("One", "Two", "Three"))
  
  
  rws <- get_title_header(th$title_hdr, 50, 60, "-", 1)
  
  expect_equal(length(rws), 3)
  expect_equal(trimws(rws[1]), "One")
  
  
  th <- title_header(create_text("hello"), "Title One", "Title Two",
                     blank_row = "above")
  
  rws <- get_title_header(th$title_hdr, 50, 60, "-", 1)
  
  expect_equal(length(rws), 3)
  expect_equal(trimws(rws[2]), "Title One")
  
  th <- title_header(create_text("hello"), "Title One", "Title Two",
                     blank_row = "above", borders = "all")
  
  rws <- get_title_header(th$title_hdr, 50, 60, "-", 1)
  
  expect_equal(length(rws), 5)
  expect_equal(trimws(rws[3]), "Title One")
  
})


test_that("page_by function works as expected.", {
  
  pb <- page_by(list(), mpg, "MPG: ",
                     align = "left",
                     blank_row = "both")
  
  
  rws <- get_page_by(pb$page_by, 50, NULL)
  
  expect_equal(length(rws), 3)
  expect_equal(rws[1], "")
  
  pb <- page_by(list(), mpg, "MPG: ",
                align = "center",
                blank_row = "above")
  
  rws <- get_page_by(pb$page_by, 50, "fork")
  
  expect_equal(length(rws), 2)
  expect_equal(trimws(rws[2]), "MPG: fork")
  
  pb <- page_by(list(), mpg,
                align = "right",
                blank_row = "below")
  
  rws <- get_page_by(pb$page_by, 50, NULL)
  
  expect_equal(length(rws), 2)
  expect_equal(trimws(rws[1]), "mpg:")
  
  pb <- page_by(list(), mpg, "Label:")
  
  rws <- get_page_by(pb$page_by, 50, "NULL")
  
  expect_equal(length(rws), 2)
  expect_equal(trimws(rws[1]), "Label:NULL")
  
})


test_that("page_by function with count works as expected.", {
  
  pb <- page_by(list(), mpg, "MPG: ",
                align = "left",
                blank_row = "both")
  
  
  rws <- get_page_by(pb$page_by, 50, NULL, 2)
  
  expect_equal(length(rws), 4)
  expect_equal(rws[1], "")
  expect_equal(trimws(rws[2]), "MPG: x")
  expect_equal(trimws(rws[3]), "x")
  expect_equal(rws[4], "")
  
})

context("RTF2 Tests")

base_path <- paste0(getwd(),"/tests/testthat")
data_dir <- base_path

base_path <- tempdir()
data_dir <- "."

cnt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
              "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
              "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
              "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ", 
              "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
              "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
              "qui officia deserunt mollit anim id est laborum.")

fnt <- "Arial"
fsz <- 10

dev <- FALSE


# Basic Tests 1 - 10 ------------------------------------------------------


test_that("rtf2-0a: Fixed report is correct.", {


  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test0a.rtf")

  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    titles("Table 0.0", "Baseline Characteristics") %>%
    add_content(create_table(mtcars[1:10, ]))

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)


})

test_that("rtf2-0b: Fixed report with font_size is correct.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test0b.rtf")

  rpt <- create_report(fp, output_type = "RTF", font = "fixed",
                       font_size = 12) %>%
    titles("Table 0.0", "Baseline Characteristics") %>%
    add_content(create_table(mtcars[1:10, ]))

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)

})

test_that("rtf2-0c: Fixed report with font_size options is correct.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test0c.rtf")

  rpt <- create_report(fp, output_type = "RTF", font = "fixed") %>%
    options_fixed(font_size = 12) %>%
    titles("Table 0.0", "Baseline Characteristics") %>%
    add_content(create_table(mtcars[1:10, ]))

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)

})

test_that("rtf2-0d: Fixed report with conflicting font size is correct.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test0d.rtf")

  rpt <- create_report(fp, output_type = "RTF", font = "fixed",
                       font_size = 8) %>%
    options_fixed(font_size = 12) %>%
    titles("Table 0.0", "Baseline Characteristics") %>%
    add_content(create_table(mtcars[1:10, ]))

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)

})



test_that("rtf2-1: One page text spec works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test1.rtf")

  txt <- create_text(cnt, width = 6, borders = "outside", align = "right") %>%
    titles("Text 1.0", "My Nice Text", borders = "outside", font_size = 12) %>%
    footnotes("My footnote 1", "My footnote 2", borders = "outside")

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = 10) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(txt, align = "right") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)

})



test_that("rtf2-2: Two page text spec works as expected in 12pt font.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test2.rtf")

  cnttxt <- paste(rep(cnt, 12), collapse = "")

  txt <- create_text(cnttxt) %>%
    titles("Text 1.0", "My Nice Text")

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = 12) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(txt) %>%
    page_footer("Left1", "Center1", "Right1") %>%
    footnotes("My footnote 1", "My footnote 2")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)

  } else
    expect_equal(TRUE, TRUE)

})


test_that("rtf2-3: Three page text spec increased margins works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test3.rtf")

  cnttxt <- paste(rep(cnt, 15), collapse = "")

  txt <- create_text(cnttxt) %>%
    titles("Text 1.0", "My Nice Text") %>%
    footnotes("My footnote 1", "My footnote 2")

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz) %>%
    set_margins(top = 2, bottom = 2) %>%
    page_header("Left", c("Right1", "Right2")) %>%
    add_content(txt) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 4)

  } else
    expect_equal(TRUE, TRUE)

})


test_that("rtf2-4: Two page text spec works as expected in 10pt font.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test4.rtf")

  cnttxt <- paste(rep(cnt, 12), collapse = "")

  txt <- create_text(cnttxt)

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = 10) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(txt) %>%
    page_footer("Left1", "Center1", "Right1") %>%
    titles("Text 1.0", "My Nice Text") %>%
    footnotes("My footnote 1", "My footnote 2")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)

  } else
    expect_equal(TRUE, TRUE)

})


test_that("rtf2-5: Two page text spec works as expected in 8pt font.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test5.rtf")

  cnttxt <- paste(rep(cnt, 20), collapse = "")

  txt <- create_text(cnttxt)

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = 8) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(txt) %>%
    page_footer("Left1", "Center1", "Right1") %>%
    titles("Text 1.0", "My Nice Text") %>%
    footnotes("My footnote 1", "My footnote 2")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)

  } else
    expect_equal(TRUE, TRUE)

})


test_that("rtf2-6: One page table works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test6.rtf")

  dat <- mtcars[1:15, ]
    attr(dat[[2]], "label") <- "Cylin."

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table", borders = "outside", bold = TRUE) %>%
    add_content(create_table(dat, borders = "outside")) %>%
    footnotes("My footnote 1", "My footnote 2", borders = "outside") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)

})


test_that("rtf2-7: Multi page table works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test7.rtf")

  dat <- iris


  tbl <- create_table(dat, borders = "none") %>%
    titles("Table 1.0", "My Nice Irises", "Another Title") %>%
    define(Sepal.Length, label = "Sepal Length", width = 1, align = "center") %>%
    define(Sepal.Width, label = "Sepal Width", width = 1, align = "centre") %>%
    define(Species, blank_after = TRUE)

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = 12, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1")) %>%
    add_content(tbl, blank_row = "none") %>%
    page_footer("Left1", "Center1", "Page [pg] of [tpg]") %>%
    footnotes("My footnote 1", "My footnote 2")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 7)

  } else
    expect_equal(TRUE, TRUE)
})



test_that("rtf2-8: Portrait table works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test8.rtf")

  dat <- mtcars[1:15, ]

  tbl <- create_table(dat)

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = 10, orientation = "portrait") %>%
   # set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(tbl) %>%
    footnotes("My footnote 1", "My footnote 2") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 11)
  expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)

})


test_that("rtf2-9: Wide table works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test9.rtf")

  dat <- mtcars[1:15, ]

  tbl <- create_table(dat) %>%
  column_defaults(width = 1)


  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "portrait") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(tbl) %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top") %>% # Works!
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)

  } else
    expect_equal(TRUE, TRUE)
})


test_that("rtf2-10: Preview works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test10.rtf")

  dat <- iris

  tbl <- create_table(dat, borders = "none") %>%
    titles("Table 1.0", "My Nice Irises", "Another Title") %>%
    footnotes("My footnote 1", "My footnote 2")

  rpt <- create_report(fp, output_type = "RTF", font = "Arial",
                       font_size = 12, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1")) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt, preview = 2)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)

  } else
    expect_equal(TRUE, TRUE)

})

# Basic Tests 11 - 20 ------------------------------------------------------

# This is awesome. Shows cell wrapping, page break, and valign
# Very good for testing.
test_that("rtf2-11: Forced page wrap works as expected.", {


  fp <- file.path(base_path, "rtf2/test11.rtf")

  dat <- data.frame(labels = rownames(mtcars), mtcars)

  tbl <- create_table(dat, borders = "all") %>%
    titles("Table 1.0", "My Nice Irises", "Another Title") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "bottom") %>%
    define(labels, id_var = TRUE) %>%
    define(wt, page_wrap = TRUE)

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = 12, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1")) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Page [pg] of [tpg]")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 4)


})


test_that("rtf2-12: Table Borders work as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test12.rtf")

  dat <- mtcars[1:15, ]

  tbl <- create_table(dat, borders = c("left", "right", "bottom", "top")) %>%
    define(mpg, label = "Miles Per Gallon")

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(tbl) %>%
    footnotes("My footnote 1", "My footnote 2") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 11)
  expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)

})

# Nice.
test_that("rtf2-13: Spanning headers work as expected.", {


  fp <- file.path(base_path, "rtf2/test13.rtf")

  dat <- mtcars[1:15, ]

  tbl <- create_table(dat, borders = c("top", "bottom")) %>%
    spanning_header(cyl, disp, "Span 1", label_align = "left") %>%
    spanning_header(hp, wt, "Span 2", underline = FALSE) %>%
    spanning_header(qsec, vs, "Span 3", n = 10) %>%
    spanning_header(drat, gear, "Super Duper\nWrapped Span", n = 11, level = 2)


  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(tbl) %>%
    footnotes("My footnote 1", "My footnote 2") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

})


test_that("rtf2-14: Labels and show_cols work as expected.", {


  fp <- file.path(base_path, "rtf2/test14.rtf")

  dat <- mtcars[1:15, ]

  attr(dat$mpg, "label") <- "Miles Per Gallon"
  attr(dat$cyl, "label") <- "Cylinders"

  tbl <- create_table(dat, show_cols = c("mpg", "hp", "cyl", "disp", "drat")) %>%
    define(mpg, width = 1.25) %>%
    define(disp, label = "Displacement")

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(tbl) %>%
    footnotes("My footnote 1", "My footnote 2") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res


  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

})


test_that("rtf2-15: Valign on report footnotes works as expected.", {


  fp <- file.path(base_path, "rtf2/test15.rtf")

  dat <- iris[1:100, ]

  tbl <- create_table(dat) %>%
    define(Species, page_break = TRUE)

  rpt <- create_report(fp, output_type = "RTF", font = "Arial",
                       font_size = 9, orientation = "portrait") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(tbl) %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 4)
  expect_equal(length(res$column_widths[[1]]), 5)

  class(dat$Species)
})

# Works
test_that("rtf2-16: Valign on table footnotes works as expected.", {


  fp <- file.path(base_path, "rtf2/test16.rtf")

  dat <- iris[1:100, ]

  tbl <- create_table(dat) %>%
    footnotes("My footnote 1", "My footnote 2", valign = "bottom") %>%
    define(Species, page_break = TRUE)

  rpt <- create_report(fp, output_type = "RTF", font = "Arial",
                       font_size = 10, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 6)
  expect_equal(length(res$column_widths[[1]]), 5)


})

test_that("rtf2-17: Title header on table works as expected.", {


  fp <- file.path(base_path, "rtf2/test17.rtf")

  dat <- iris[1:25, ]

  tbl <- create_table(dat, width = 8.9) %>%
    title_header("Table 1.0", "My Nice Table",
                 right = c("Right1",
                           "Right2", "Page [pg] of [tpg]")) %>%
    footnotes("My footnote 1", "My footnote 2") #%>%
    #define(Petal.Length, page_wrap = TRUE)

  rpt <- create_report(fp, output_type = "RTF", font = "Arial",
                       font_size = 10, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)


})


test_that("rtf2-18: Title header on report works as expected.", {


  fp <- file.path(base_path, "rtf2/test18.rtf")

  dat <- iris[1:50, ]

  tbl <- create_table(dat)

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Page [pg] of [tpg]") %>%
    title_header("Table 1.0", "My Nice Table", right = c("Page [pg] of [tpg]",
                                                         "Right2", "Right3")) %>%
    footnotes("My footnote 1", "My footnote 2", valign = "bottom")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)
  expect_equal(length(res$column_widths[[1]]), 5)


})


test_that("rtf2-19: Title and Footnote borders work as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test19.rtf")

  dat <- iris[1:20, ]

  tbl <- create_table(dat, borders = "all") %>%
    titles("Table 1.0", "My Nice Report with Borders",
                 borders = c("top", "bottom", "left", "right"),
           blank_row = "both") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top",
              borders = c("top", "bottom", "left", "right"),
              blank_row = "both")

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)

  } else
    expect_equal(TRUE, TRUE)

})


test_that("rtf2-20: Title Header borders work as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test20.rtf")

  dat <- iris[1:25, ]

  tbl <- create_table(dat, borders = "all") %>%
    title_header("Table 1.0", "My Nice Report with Borders",
                 right = c("Right1", "Right2", "Right3"),
           borders = c("outside"),
           blank_row = "both") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top",
              borders = c("top", "bottom", "left", "right"),
              blank_row = "both")

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)
  expect_equal(length(res$column_widths[[1]]), 5)

  } else
    expect_equal(TRUE, TRUE)


})

# Basic Tests 21 - 30 ------------------------------------------------------

# Works unless drat column is too narrow and spanning label wraps unexpectedly
test_that("rtf2-21: Page wrap with spanning header works as expected.", {

  fp <- file.path(base_path, "rtf2/test21.rtf")


  df <- data.frame(vehicle = rownames(mtcars), mtcars, stringsAsFactors = FALSE)
  rownames(df) = NULL

  tbl <- create_table(df) %>%
    spanning_header(2, 5,
                    label = "Span 1", label_align = "center", n = 10) %>%
    spanning_header(6, 8,
                    label = "Span 2", label_align = "center", n = 10) %>%
    spanning_header(9, 12,
                    label = "Span 3", label_align = "center", n = 10) %>%
    spanning_header(6, 12, label = "Super Span",
                    label_align = "center",
                    level = 2) %>%
    define(vehicle, label = "Vehicle", id_var = TRUE) %>%
    define(mpg, format = "%.1f") %>%
    define(drat, width = 1) %>%
    define(wt, page_wrap = TRUE) %>%
    define(vs, page_wrap = TRUE)

  rpt <- create_report(fp, output_type = "RTF",
                       orientation = "portrait", font = fnt) %>%
    add_content(tbl) %>%
    titles("Table 1.0", "MTCARS Subset Test") %>%
    footnotes("My footnote") %>%
    page_header("Left") %>%
    page_footer("Left", right = "Right")

  #print(rpt)

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

  expect_equal(res$pages, 3)
})

# This is a good one for testing height calculations.  Looks good for now.
test_that("rtf2-22: Page by works as expected.", {


  fp <- file.path(base_path, "rtf2/test22.rtf")

  dat <- iris

  tbl <- create_table(dat, borders = "none") %>%
    titles("Table 1.0", "My Nice Report with a Page By", borders = "none") %>%
    page_by(Species, label = "Species: ", align = "right", borders = "none")

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_header("Left", "Right") %>%
    page_footer("Left1", "Center1", "Right1") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "none")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 6)
  expect_equal(length(res$column_widths[[1]]), 5)


})

# Works!
test_that("rtf2-23: Two contents on one page works as expected.", {


  fp <- file.path(base_path, "rtf2/test23.rtf")

  dat <- mtcars[1:15, ]

  tbl <- create_table(dat) %>%
    titles("Table 1.0", "My Nice Table") %>%
    footnotes("My footnote 1", "My footnote 2")

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), blank_row = "below") %>%
    add_content(tbl, page_break = FALSE, blank_row = "below") %>%
    add_content(create_text(cnt, width = 5), align = "center") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)


})

# Working.
test_that("rtf2-24: Two tables one headerless works as expected.", {


  fp <- file.path(base_path, "rtf2/test24.rtf")

  dat <- mtcars
  dat2 <- mtcars[16:20, ]

  tbl <- create_table(dat) %>%
    titles("Table 1.0", "My Nice Table") %>%
    column_defaults(width = .5)

  tbl2 <- create_table(dat2, headerless = TRUE) %>%
    column_defaults(width = .5)

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), blank_row = "below") %>%
    add_content(tbl, page_break = FALSE, blank_row = "none") %>%
    add_content(tbl2, align = "center") %>%
    page_footer("Left1", "Center1", "Right1") %>%
    footnotes("My footnote 1", "My footnote 2")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 2)


})

# Works!
test_that("rtf2-25: Simplest RTF Plot works as expected.", {

  if (dev == TRUE) {


  library(ggplot2)

  fp <- file.path(base_path, "rtf2/test25.rtf")

  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()

  plt <- create_plot(p, height = 4, width = 8, borders = c("top", "bottom", "all")) %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", borders = "none") %>%
    footnotes("* Motor Trend, 1974", borders = "none")


  rpt <- create_report(fp, output_type = "RTF", font = fnt, font_size =fsz) %>%
    page_header("Client", "Study: XYZ") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt, align = "right") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")


  res <- write_report(rpt)

  #print(res)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)
})


test_that("rtf2-26: RTF Table with Plot on same page works as expected.", {

  library(ggplot2)

  fp <- file.path(base_path, "rtf2/test26.rtf")

  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()

  plt <- create_plot(p, height = 4, width = 8)
  tbl <- create_table(mtcars[1:3, ])


  rpt <- create_report(fp, output_type = "RTF", font = fnt, font_size = fsz) %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", blank_row = "none") %>%
    set_margins(top = 1, bottom = 1) %>%

    add_content(plt, page_break = FALSE, blank_row = "none") %>%
    add_content(tbl) %>%
    footnotes("* Motor Trend, 1974") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")


  res <- write_report(rpt)

  #print(res)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)


})



test_that("rtf2-27: Plot with page by on plot works as expected.", {

  library(ggplot2)

  fp <- file.path(base_path, "rtf2/test27.rtf")

  fmt <- value(condition(x == 4, "4 Cylinder"),
               condition(x == 6, "6 Cylinder"),
               condition(x == 8, "8 Cylinder"))

  dat <- mtcars[order(mtcars$cyl), ]

  p <- ggplot(dat, aes(x=disp, y=mpg)) + geom_point()


  #dats <- split(p$data, p$data$grp)
  #tbl <- create_table(dat[1:3, ])

  plt <- create_plot(p, height = 4, width = 8) %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", blank_row = "none") %>%
    page_by(cyl, "Cylinders: ", format = fmt) %>%
    footnotes("* Motor Trend, 1974")

  rpt <- create_report(fp, output_type = "RTF", font = fnt, font_size = fsz) %>%
    page_header("Client", "Study: XYZ") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt) %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")


  res <- write_report(rpt)

  #print(res)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 3)


})


test_that("rtf2-28: Plot with page by on report works as expected.", {

  if (dev == TRUE) {


  library(ggplot2)

  fp <- file.path(base_path, "rtf2/test28.rtf")


  dat <- mtcars[order(mtcars$cyl), ]

  p <- ggplot(dat, aes(x=disp, y=mpg)) + geom_point()


  #dats <- split(p$data, p$data$grp)
  #tbl <- create_table(dat[1:3, ])

  plt <- create_plot(p, height = 4, width = 8)


  rpt <- create_report(fp, output_type = "RTF", font = fnt, font_size = fsz) %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot",
           blank_row = "none", borders = "none") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_by(cyl, "Cylinders: ") %>%
    add_content(plt) %>%
    footnotes("* Motor Trend, 1974", borders = "none") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")


  res <- write_report(rpt)

  #print(res)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 3)

  } else
    expect_equal(TRUE, TRUE)

})

test_that("rtf2-29: Simplest RTF Plot with valign top works as expected.", {

  if (dev) {
    library(ggplot2)

    fp <- file.path(base_path, "rtf2/test29.rtf")

    p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()

    plt <- create_plot(p, height = 4, width = 8)


    rpt <- create_report(fp, output_type = "RTF", font = fnt, font_size = fsz) %>%
      page_header("Client", "Study: XYZ") %>%
      titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(plt, align = "center") %>%
      footnotes("* Motor Trend, 1974", valign = "top") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")


    res <- write_report(rpt)

    #print(res)

    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)

  } else {

   expect_equal(TRUE, TRUE)
  }


})



test_that("rtf2-30: Simplest RTF Plot with valign bottom works as expected.", {

  if (dev) {
    library(ggplot2)

    fp <- file.path(base_path, "rtf2/test30.rtf")

    p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()

    plt <- create_plot(p, height = 4, width = 8) %>%
      footnotes("* Motor Trend, 1974", valign = "bottom")


    rpt <- create_report(fp, output_type = "RTF", font = fnt, font_size = fsz) %>%
      page_header("Client", "Study: XYZ") %>%
      titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(plt, align = "center") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")


    res <- write_report(rpt)

    #print(res)

    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)

  } else {

    expect_equal(TRUE, TRUE)
  }


})

# Basic Tests 31 - 40 ------------------------------------------------------

test_that("rtf2-31: Simplest RTF Text with valign top works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test31.rtf")

  txt <- create_text(cnt, width = 6, borders = "outside")


  rpt <- create_report(fp, output_type = "RTF", font = fnt, font_size = fsz) %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Text 1.0", "MTCARS Miles per Cylinder Text", borders = "outside") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(txt, align = "center") %>%
    footnotes("* Motor Trend, 1974", valign = "top", borders = "outside") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")


  res <- write_report(rpt)

  #print(res)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)

})

test_that("rtf2-32: Simplest RTF Text with valign bottom works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test32.rtf")

  txt <- create_text(cnt, width = 6) %>%
    footnotes("* Motor Trend, 1974", valign = "bottom")

  rpt <- create_report(fp, output_type = "RTF", font = fnt, font_size = fsz) %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(txt, align = "center") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")


  res <- write_report(rpt)

  #print(res)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)

})

test_that("rtf2-33: Table with long cell and label values wraps as expected.", {


  fp <- file.path(base_path, "rtf2/test33.rtf")


  # Setup
  arm <- c(rep("A", 5), rep("B", 5))
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa \nand more and more", "Schaffer, Ashley", "Laner, Tahma",
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)


  # Create data frame
  df <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)


  tbl1 <- create_table(df, first_row_blank = TRUE) %>%
    define(subjid, label = "Subject ID for a patient", n = 10, align = "left",
           width = 1) %>%
    define(name, label = "Subject Name", width = 1) %>%
    define(sex, label = "Sex", n = 10, align = "center") %>%
    define(age, label = "Age", n = 10) %>%
    define(arm, label = "Arm",
           blank_after = TRUE,
           dedupe = TRUE)


  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz) %>%
    titles("Table 1.0", align = "center") %>%

    add_content(tbl1)


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)



})

# Works on all combinations on font and font size.  Needed adjustments to row height.
test_that("rtf2-34: Table with break between sections works as expected.", {


  fp <- file.path(base_path, "rtf2/test34.rtf")


  # Setup
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma",
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
  arm <- c(rep("A", 5), rep("B", 5))

  # Create data frame
  df <- data.frame(subjid, name, sex, age, arm)


  tbl1 <- create_table(df, first_row_blank = FALSE) %>%
    define(subjid, label = "Subject ID", align = "left", width = 1) %>%
    define(name, label = "Subject Name", width = 1) %>%
    define(sex, label = "Sex") %>%
    define(age, label = "Age") %>%
    define(arm, label = "Arm",
           blank_after = FALSE,
           dedupe = TRUE,
           align = "right") #%>%
   # spanning_header(sex, arm, label = "Here is a spanning header")


  rpt <- create_report(fp, output_type = "RTF", font = fnt, font_size = fsz) %>%
    page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
   # options_fixed(line_count = 46) %>%
    titles("Table 1.0", "Analysis Data Subject Listing\n And more stuff",
           "Safety Population", align = "center") %>%
    footnotes("Program Name: table1_0.R",
              "Here is a big long footnote that is going to wrap\n at least once") %>%
    page_footer(left = "Time", center = "Confidential",
                right = "Page [pg] of [tpg]") %>%
    add_content(tbl1)


  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)


})


test_that("rtf2-35: Title Header and page header/footer wrapping work as expected.", {


  fp <- file.path(base_path, "rtf2/test35.rtf")

  dat <- iris[1:10, ]

  tbl <- create_table(dat, borders = "none") %>%
    title_header("Table 1.0", "My Nice Report with Borders",
                 right = c("Right1", "Right2",
                           "Right3 long enough to wrap around at least once"),
                 borders = "none",
                 blank_row = "none") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "bottom",
              borders = "none",
              blank_row = "above")

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_header(c("Left1", "Left2\nwrap"), "Right 1") %>%
    page_footer("Left1",
                "Center1 here is a whole bunch of stuff to try and make it wrap",
                "Right1\nwrap\n and wrap again")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)


})


test_that("rtf2-36: Title and Footnote widths work as expected on table.", {



  fp <- file.path(base_path, "rtf2/test36.rtf")

  dat <- iris[1:20, ]

  tbl <- create_table(dat) %>%
    titles("Table 1.0", "My Nice Report with Widths",
           width = "page") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top",
              width = "page")

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)


})


test_that("rtf2-37: Title and Footnote widths work as expected on report", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test37.rtf")

  dat <- iris[1:20, ]

  tbl <- create_table(dat, borders = "all")

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1") %>%
    titles("Table 1.0", "My Nice Report with Widths",
           borders = c("top", "bottom", "left", "right"),
           width = 7) %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top",
              borders = c("top", "bottom", "left", "right"),
               width = 7)

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)

  } else
    expect_equal(TRUE, TRUE)

})


test_that("rtf2-38: Title and Footnote specific widths work as expected.", {


  fp <- file.path(base_path, "rtf2/test38.rtf")

  dat <- iris[1:20, ]

  tbl <- create_table(dat, borders = "all") %>%
    titles("Table 1.0", "My Nice Report with Borders",
           borders = c("top", "bottom", "left", "right"),
            width = 7) %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top",
              borders = c("top", "bottom", "left", "right"),
              width = 7)

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)


})

test_that("rtf2-39: One page table works as expected in centimeters and times.", {


  fp <- file.path(base_path, "rtf2/test39.rtf")

  dat <- mtcars[1:15, ]
  attr(dat[[2]], "label") <- "Cylin."

  rpt <- create_report(fp, output_type = "RTF", font = "Times",
                       font_size = fsz, orientation = "landscape",
                       units = "cm") %>%
    set_margins(top = 3, bottom = 3) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(create_table(dat)) %>%
    footnotes("My footnote 1", "My footnote 2") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)


})

test_that("rtf2-40: One page table works as expected in courier.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test40.rtf")

  dat <- mtcars[1:15, ]
  attr(dat[[2]], "label") <- "Cylin."

  rpt <- create_report(fp, output_type = "RTF", font = "Courier",
                       font_size = fsz, orientation = "landscape",
                       units = "cm") %>%
    set_margins(top = 3, bottom = 3) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(create_table(dat)) %>%
    footnotes("My footnote 1", "My footnote 2") %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)
})

# Basic Tests 41 - 50 ------------------------------------------------------

# Good for testing borders and spacing are working correctly
test_that("rtf2-41: Page by with borders works as expected.", {


  fp <- file.path(base_path, "rtf2/test41.rtf")

  dat <- iris

  brdrs <- "outside"

  tbl <- create_table(dat, borders = "all") %>%
    titles("Table 1.0", "My Nice Report with a Page By", borders = brdrs) %>%
    page_by(Species, label = "Species: ", align = "left", borders = brdrs,
            blank_row = "below") %>%
    footnotes("My footnote 1", "My footnote 2", borders = brdrs)

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 9)
  expect_equal(length(res$column_widths[[1]]), 5)


})

# Test for borders and page wraps
test_that("rtf2-42: Long table with borders and footnotes on report.", {


  fp <- file.path(base_path, "rtf2/test42.rtf")

  dat <- data.frame(labels = rownames(mtcars), mtcars)

  tbl <- create_table(dat, borders = "all") %>%
    define(labels, id_var = TRUE) %>%
    define(wt, page_wrap = TRUE)

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = 12, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1")) %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Page [pg] of [tpg]") %>%
    titles("Table 1.0", "My Nice Irises", "Another Title") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "bottom")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 4)


})



test_that("rtf2-43: use_attributes parameter table works as expected.", {

  if (dev == TRUE) {


  fp1 <- file.path(base_path, "rtf2/test43a.rtf")
  fp2 <- file.path(base_path, "rtf2/test43b.rtf")
  fp3 <- file.path(base_path, "rtf2/test43c.rtf")


  dat <- mtcars[1:10, ]
  attr(dat$mpg, "label") <- "Miles per gallon"
  attr(dat$cyl, "format") <- "%.1f"
  attr(dat$hp, "width") <- 2
  fattr(dat$vs) <- list(width = 2, justify = "center")

  tbl <- create_table(dat)

  # Test default
  rpt <- create_report(fp1, output_type = "RTF", font = fnt) %>%
    add_content(tbl)

  res <- write_report(rpt)

  expect_equal(file.exists(fp1), TRUE)
  expect_equal(res$pages, 1)

  # Test none
  tbl <- create_table(dat, use_attributes = "none")

  rpt <- create_report(fp2, output_type = "RTF", font = fnt) %>%
    add_content(tbl)

  res <- write_report(rpt)

  expect_equal(file.exists(fp2), TRUE)


  # Test some
  tbl <- create_table(dat, use_attributes = c("format", "width"))

  rpt <- create_report(fp3, output_type = "RTF", font = fnt) %>%
    add_content(tbl)

  res <- write_report(rpt)

  expect_equal(file.exists(fp3), TRUE)
  expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)

})

test_that("rtf2-45: Title bold and font size works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test45.rtf")

  dat <- mtcars[1:15, ]
  attr(dat[[2]], "label") <- "Cylin."
  attr(dat[[2]], "width") <- 1
  attr(dat[[2]], "justify") <- "center"

  tbl <- create_table(dat, borders = "outside") %>%
    titles("Table 1.0", "My Nice Table", borders = c("none"),
           width = "content", font_size = 14, bold = TRUE) %>%
    footnotes("My footnote 1", "My footnote 2", borders = "none",
              align = "left", width = "content") %>%
    define(wt, width = 1, label = "Weight", align = "center",
           label_align = "right")

  rpt <- create_report(fp, output_type = "RTF", font = "Arial",
                       font_size = 9, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Page [pg] of [tpg]"),
                blank_row = "below") %>%
    add_content(tbl, align = "center")  %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)
})



test_that("rtf2-46: 9 pt font inches works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test46.rtf")

  rpt <- create_report(fp, output_type = "RTF", font_size = 9,
                       font = "Courier",
                       orientation = "portrait") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>%
    set_margins(top = 1, bottom = 1)


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

  } else
    expect_equal(TRUE, TRUE)


})

test_that("rtf2-47: 9 pt font cm works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test47.rtf")

  rpt <- create_report(fp, output_type = "RTF", font_size = 9,
                       font = "Courier",
                       orientation = "portrait",
                       units = "cm") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>%
    set_margins(top = 1, bottom = 1)


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

  } else
    expect_equal(TRUE, TRUE)

})

test_that("rtf2-48: 11 pt font inches works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test48.rtf")

  rpt <- create_report(fp, output_type = "RTF", font_size = 11,
                       font = "Courier",
                       orientation = "portrait") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>%
    set_margins(top = 1, bottom = 1)


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

  } else
    expect_equal(TRUE, TRUE)
})

test_that("rtf2-49: 11 pt font cm works as expected.", {

  if (dev == TRUE) {


  fp <- file.path(base_path, "rtf2/test49.rtf")

  rpt <- create_report(fp, output_type = "RTF", font_size = 11,
                       font = "Courier",
                       orientation = "portrait",
                       units = "cm") %>%
    page_header("left", "right") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) %>%
    page_footer("left", "center", "Page [pg] of [tpg]") %>%
    set_margins(top = 1, bottom = 1)


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

  } else
    expect_equal(TRUE, TRUE)

})

test_that("rtf2-50: Spanning headers borders work as expected.", {


  fp <- file.path(base_path, "rtf2/test50.rtf")

  dat <- mtcars[1:15, ]

  tbl <- create_table(dat, borders = c("outside")) %>%
    spanning_header(cyl, disp, "Span 1", label_align = "left") %>%
    spanning_header(hp, wt, "Span 2", underline = FALSE) %>%
    spanning_header(qsec, vs, "Span 3", n = 10) %>%
    spanning_header(drat, gear, "Super Duper\nWrapped Span", n = 11, level = 2) %>%
    titles("Table 1.0", "My Nice Table", borders = "outside") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "outside")

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl)


  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

})


# Basic Tests 51-60 -------------------------------------------------------


test_that("rtf2-51: RTF Image file works as expected.", {

  if (dev == TRUE) {


  library(ggplot2)

  fp <- file.path(base_path, "rtf2/test51.rtf")

  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()

  pltpath <- file.path(base_path, "rtf2/test51.jpg")
  ggsave(pltpath, width = 8, height = 4,
         units = "in",
         dpi = 300)

  plt <- create_plot(pltpath, height = 4, width = 8)


  rpt <- create_report(fp, output_type = "RTF", font = "Arial") %>%
    page_header("Client", "Study: XYZ") %>%
    titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt, align = "center") %>%
    footnotes("* Motor Trend, 1974") %>%
    page_footer("Time", "Confidential", "Page [pg] of [tpg]")


  res <- write_report(rpt)

  #print(res)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)
})

test_that("rtf2-52: Table with break between sections works as expected.", {


  fp <- file.path(base_path, "rtf2/test52.rtf")


  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B
      "ampg"   "\\fi-87\\li267 N"          "19"          "13"
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)"
      "ampg"   "Median"     "16.4"        "21.4"
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)"
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)"
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')

  # Create table
  tbl <- create_table(df, first_row_blank = TRUE) %>%
    stub(c("var", "label")) %>%
    define(var, blank_after = TRUE, label_row = TRUE,
           format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>%
    define(label, indent = .25) %>%
    define(A, label = "Group A", align = "center", n = 19) %>%
    define(B, label = "Group B", align = "center", n = 13)


  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "RTF",
                       font = "Times") %>%
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>%
    titles("Table 1.0", "MTCARS Summary Table") %>%
    add_content(tbl) %>%
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left",
                center = "Confidential",
                right = "Page [pg] of [tpg]")



  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)


})

test_that("rtf2-53: Text with line feed works as expected.", {

  fp <- file.path(base_path, "rtf2/test53.rtf")

  cnt2 <- paste0("Hello here \nis something ", cnt)

  txt <- create_text(cnt2) %>%
    titles("Report 1.0", "Simple Text Report") %>%
    footnotes("My footnote")

  rpt <- create_report(fp, orientation = "portrait",
                       output_type = "RTF", font = "Arial") %>%
    add_content(txt)

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
})


test_that("rtf2-54: Header_bold works as expected", {

  dat <- mtcars[1:10, 1:3]

  fp <- file.path(base_path, "rtf2/test54.rtf")

  tbl <- create_table(dat, header_bold = TRUE, borders = "all") %>%
    column_defaults(width = 1) %>%
    titles("Report 1.0", "Simple Report", borders = "outside",
           blank_row = "none", bold = TRUE) %>%
    footnotes("My footnote", blank_row = "none")

  rpt <- create_report(fp, orientation = "portrait",
                       output_type = "RTF", font = "Arial") %>%
    add_content(tbl) %>%
    set_margins(top = 1)

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

})

test_that("rtf2-55: Titles and footnotes in header and footer works as expected", {

  dat <- mtcars[1:10, 1:3]

  fp <- file.path(base_path, "rtf2/test55.rtf")

  tbl <- create_table(dat) %>%
    column_defaults(width = 1)

  rpt <- create_report(fp, orientation = "landscape",
                       output_type = "RTF", font = "Arial") %>%
    add_content(tbl) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    titles("Report 1.0", "Simple Report",
           blank_row = "none", header = TRUE, columns = 1) %>%
    footnotes("My footnote", "Another footnote", "And another",
              blank_row = "none", footer = TRUE) %>%
    page_footer("Left", "Center", "Right")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

})

test_that("rtf2-56: Titles and footnotes variations in header and footer work as expected", {

  dat <- mtcars[1:10, 1:3]

  fp <- file.path(base_path, "rtf2/test56.rtf")

  tbl <- create_table(dat) %>%
    column_defaults(width = 1)

  rpt <- create_report(fp, orientation = "landscape",
                       output_type = "RTF", font = "Arial") %>%
    add_content(tbl) %>%
    set_margins(top = 1) %>%
    page_header("Left", "Right") %>%
    titles("Report 1.0", "Simple Report", align = "left", width = 6,
           blank_row = "none", header = TRUE) %>%
    titles("Report 1.0", "Simple Report", align = "right", width = 6,
           blank_row = "below", header = TRUE, borders = "bottom") %>%
    footnotes("My footnote", blank_row = "none", footer = TRUE, borders = "top") %>%
    footnotes("My footnote2", blank_row = "none", footer = TRUE, align = "right") %>%
    footnotes("My footnote3", blank_row = "none", footer = TRUE, align = "center") %>%
    page_footer("Left", "Center", "Right")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

})

test_that("rtf2-57: Titles and footnotes in header and footer no page header/footer works as expected", {

  dat <- mtcars[1:10, 1:3]

  fp <- file.path(base_path, "rtf2/test57.rtf")

  tbl <- create_table(dat) %>%
    column_defaults(width = 1)

  rpt <- create_report(fp, orientation = "landscape",
                       output_type = "RTF", font = "Arial") %>%
    add_content(tbl) %>%
    set_margins(top = 1, bottom = 1) %>%
    titles("Report 1.0 here is a big long title", "Simple Report", align = "left",
           blank_row = "none", header = TRUE) %>%
    titles("Report 1.0", "Simple Report", align = "center", width = 6,
           blank_row = "below", header = TRUE) %>%
    footnotes("My footnote1", "My footnote2", blank_row = "none", footer = TRUE)

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

})

test_that("rtf2-58: Label row is one cell.", {


  fp <- file.path(base_path, "rtf2/test58.rtf")


  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B
      "ampg"   "N"          "19"          "13"
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)"
      "ampg"   "Median"     "16.4"        "21.4"
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)"
      "cyl"    "6 Cylinder and more perhaps more" "4 ( 21.1%)"  "3 ( 23.1%)"
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')

  ll <- "Here is a super long label to see if it can span the entire table."

  # Create table
  tbl <- create_table(df, first_row_blank = TRUE, borders = c("all")) %>%
    stub(c("var", "label"), width = .8) %>%
    define(var, blank_after = TRUE, label_row = TRUE,
           format = c(ampg = ll, cyl = "Cylinders")) %>%
    define(label, indent = .25) %>%
    define(A, label = "Group A", align = "center", n = 19) %>%
    define(B, label = "Group B", align = "center", n = 13)


  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "RTF",
                       font = "Times") %>%
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>%
    titles("Table 1.0", "MTCARS Summary Table") %>%
    add_content(tbl) %>%
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left",
                center = "Confidential",
                right = "Page [pg] of [tpg]")



  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)


})

test_that("rtf2-59: Blank after on invisible column.", {

  fp <- file.path(base_path, "rtf2/test59.rtf")

  tbl <- create_table(iris, borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)

  rpt <- create_report(fp, output_type = "RTF", font = "Courier") %>%
    page_header("Left", "Right") %>%
    add_content(tbl) %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below") %>%
    footnotes("Here is a footnote", "And another")


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

})


test_that("rtf2-60: Blank nested stub works as expected.", {


  fp <- file.path(base_path, "rtf2/test60.rtf")


  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B
      "ampg"   "Stats"    "19"          "13"
      "ampg"   "Stats"    "18.8 (6.5)"  "22.0 (4.9)"
      "ampg"   "Stats"    "16.4"        "21.4"
      "ampg"   "Stats"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Stats"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)"
      "cyl"    "" "4 ( 21.1%)"  "3 ( 23.1%)"
      "cyl"    "" "5 ( 26.3%)"  "6 ( 46.2%)"')

  ll <- "Here is a super long label to see if it can span the entire table."

  # Create table
  tbl <- create_table(df, first_row_blank = TRUE, borders = c("all")) %>%
    stub(c("var", "label")) %>%
    define(var, blank_after = TRUE, label_row = TRUE,
           format = c(ampg = ll, cyl = "Cylinders")) %>%
    define(label, indent = .25, dedupe = TRUE) %>%
    define(A, label = "Group A", align = "center", n = 19) %>%
    define(B, label = "Group B", align = "center", n = 13)


  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "RTF",
                       font = "Times") %>%
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>%
    titles("Table 1.0", "MTCARS Summary Table") %>%
    add_content(tbl) %>%
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left",
                center = "Confidential",
                right = "Page [pg] of [tpg]")



  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)

})


# Basic Tests 61-70 -------------------------------------------------------



test_that("rtf2-61: Page header width works.", {

  fp <- file.path(base_path, "rtf2/test61.rtf")

  tbl <- create_table(iris, borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)

  rpt <- create_report(fp, output_type = "RTF", font = "Courier") %>%
    page_header("Left and here is a really long left cell text to put it",
                "Right", width = 8) %>%
    add_content(tbl) %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below") %>%
    footnotes("Here is a footnote", "And another")


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

})

test_that("rtf2-62: Carriage return in label row works.", {


  fp <- file.path(base_path, "rtf2/test62.rtf")


  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B
      "ampg"   "N"          "19"          "13"
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)"
      "ampg"   "Median"     "16.4"        "21.4"
      "ampg"   "Q1 - Q3"    "15.1 - \n21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)"
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)"
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')

  ll <- "Here is a super long label to \nsee if it can span the entire table."

  # Create table
  tbl <- create_table(df, first_row_blank = TRUE, borders = c("all")) %>%
    stub(c("var", "label")) %>%
    define(var, blank_after = TRUE, label_row = TRUE,
           format = c(ampg = ll, cyl = "Cylinders")) %>%
    define(label, indent = .25) %>%
    define(A, label = "Group A", align = "center", n = 19) %>%
    define(B, label = "Group B", align = "center", n = 13)


  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "RTF",
                       font = "Times") %>%
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>%
    titles("Table 1.0", "MTCARS Summary Table") %>%
    add_content(tbl) %>%
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left",
                center = "Confidential",
                right = "Page [pg] of [tpg]")



  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)


})


test_that("rtf2-63: Glue works.", {

  # if (dev) {

    fp <- file.path(base_path, "rtf2/test63.rtf")


    # Read in prepared data
    df <- read.table(header = TRUE, text = '
        var     label        A             B
        "ampg"   "N"          "19"          "13"
        "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)"
        "ampg"   "Median"     "16.4"        "21.4"
        "ampg"   "Q1 - Q3"    "15.1 - \n21.2" "19.2 - 22.8"
        "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
        "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)"
        "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)"
        "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')

    ll <- "Here is a super long label to \nsee if it can span the entire table."

    # Create table
    tbl <- create_table(df, first_row_blank = TRUE, borders = c("all")) %>%
      stub(c("var", "label"), label = "Stub{supsc('2')}") %>%
      define(var, blank_after = TRUE, label_row = TRUE,
             format = c(ampg = ll, cyl = "Cylinders")) %>%
      define(label, indent = .25) %>%
      define(A, label = "Group A{supsc('1')}", align = "center", n = 19) %>%
      define(B, label = "Group B", align = "center", n = 13)


    # Create report and add content
    rpt <- create_report(fp, orientation = "portrait", output_type = "RTF",
                         font = "Times") %>%
      page_header(left = c("Client: Motor Trend{supsc('6')}",
                           "Test{supsc('8')}"), right = "Study: Cars") %>%
      titles("Table 1.0", "MTCARS Summary Table{supsc('3')}") %>%
      add_content(tbl) %>%
      footnotes("* Motor Trend, 1974{supsc('4')}") %>%
      page_footer(left = "Left{supsc('6')}",
                  center = "Confidential",
                  right = "Page [pg] of [tpg]")



    res <- write_report(rpt)
    res
    expect_equal(file.exists(fp), TRUE)

  # } else {
  # 
  #   expect_equal(TRUE, TRUE)
  # 
  # }


})


test_that("rtf2-64: Title columns work 1 column.", {

  fp <- file.path(base_path, "rtf2/test64.rtf")

  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)

  rpt <- create_report(fp, output_type = "RTF", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below", columns =  1, align = "left") %>%
    footnotes("Here is a footnote", "And another")


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

})

test_that("rtf2-65: Title columns work 2 columns.", {

  fp <- file.path(base_path, "rtf2/test65.rtf")

  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)

  rpt <- create_report(fp, output_type = "RTF", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame", "Left", "Right",
           blank_row = "below", columns =  2) %>%
    footnotes("Here is a footnote", "And another")


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

})

test_that("rtf2-66: Title columns work 3 columns.", {

  fp <- file.path(base_path, "rtf2/test66.rtf")

  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)

  rpt <- create_report(fp, output_type = "RTF", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame", "My right thing", "", "Center",
           blank_row = "below", columns =  3) %>%
    footnotes("Here is a footnote", "And another")


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

})

test_that("rtf2-67: Multiple title blocks work as expected.", {

  fp <- file.path(base_path, "rtf2/test67.rtf")

  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)

  rpt <- create_report(fp, output_type = "RTF", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below", columns =  1, align = "center", width = 7,
           borders = "all") %>%
    titles("Table 2.0", "IRIS Data Frame2", "Left", "Right",
           blank_row = "below", columns =  2, borders = "all") %>%
    titles("Table 3.0", "IRIS Data Frame3", "My right thing", "", "Center",
           blank_row = "below", columns =  3, borders = "all") %>%
    footnotes("Here is a footnote", "And another")


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

})

test_that("rtf2-68: Custom page size works as expected.", {

  fp <- file.path(base_path, "rtf2/test68.rtf")

  tbl <- create_table(iris[1:15, ]) %>%
    define(Species, visible = FALSE)

  ttl <- c("Title1", "Title2", "Title3")

  rpt <- create_report(fp, output_type = "RTF",
                       font = "Courier",
                       paper_size = c(6.5, 7.5),
                       orientation = "portrait") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles(ttl,
           blank_row = "below", columns =  1, align = "center",
           borders = "none") %>%
    footnotes("Here is a footnote", "And another")


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

})


test_that("rtf2-69: Breaks removed after titles and footnotes.", {

  fp <- file.path(base_path, "rtf2/test69.rtf")

  tbl <- create_table(iris[1:15, ], borders = "outside") %>%
    define(Species, visible = FALSE) %>%
    titles("Here is a title") %>%
    footnotes("Here is a footnote")


  rpt <- create_report(fp, output_type = "RTF",
                       font = "Courier",
                       orientation = "landscape") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right")



  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

})


test_that("rtf2-70: Breaks removed after 2 titles and footnotes.", {

  fp <- file.path(base_path, "rtf2/test70.rtf")

  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, visible = FALSE) %>%
    titles("Here is a title") %>%
    titles("There is a title") %>%
    footnotes("Here is a footnote") %>%
    footnotes("There is a footnote")


  rpt <- create_report(fp, output_type = "RTF",
                       font = "Courier",
                       orientation = "landscape") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right")



  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

})


# Basic Tests 71-80 -------------------------------------------------------


test_that("rtf2-71: RTF Plot with path works as expected.", {

  if (dev == TRUE) {


    fp <- file.path(base_path, "rtf2/test71.rtf")

    tmp <- file.path(base_path, "plot.jpg")

    jpeg(tmp, width = 600, height = 500)

    plot(mtcars$mpg)


    dev.off()

    plt <- create_plot(tmp, height = 5, width = 6, borders = c("top", "bottom", "all")) %>%
      titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", borders = "none") %>%
      footnotes("* Motor Trend, 1974", borders = "none")


    rpt <- create_report(fp, output_type = "RTF", font = fnt, font_size =fsz) %>%
      page_header("Client", "Study: XYZ") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(plt) %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")


    res <- write_report(rpt)

    #print(res)

    #print(res)

    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)
})



test_that("rtf2-72: Symbols are proper orientation on portrait.", {

  if (dev) {

    fp <- file.path(base_path, "rtf2/test72.rtf")

    dat <- mtcars[1:10, ]


    tbl <- create_table(dat) %>%
      titles("My Table{symbol('dagger')}")


    rpt <- create_report(fp, output_type = "RTF",
                         font = "Arial", orientation = "portrait") %>%
      add_content(tbl)

    res <- write_report(rpt)


    #file.show(res$modified_path)

    expect_equal(file.exists(fp), TRUE)


  } else {

    expect_equal(TRUE, TRUE)

  }

})


test_that("rtf2-73: Symbols are proper orientation on landscape.", {

  if (dev) {

    fp <- file.path(base_path, "rtf2/test73.rtf")

    dat <- mtcars[1:10, ]


    tbl <- create_table(dat) %>%
      titles("My Table{symbol('dagger')}")


    rpt <- create_report(fp, output_type = "RTF",
                         font = "Arial", orientation = "landscape") %>%
      add_content(tbl)

    res <- write_report(rpt)


    #file.show(res$modified_path)

    expect_equal(file.exists(fp), TRUE)


  } else {

    expect_equal(TRUE, TRUE)

  }

})


test_that("rtf2-74: Multi-page continuous tables work as expected.", {

  if (dev) {

    fp <- file.path(base_path, "rtf2/test74.rtf")

    dat <- iris


    tbl <- create_table(dat, continuous = TRUE, borders = "all") %>%
     # titles("My Title") %>%
      footnotes("My footnotes", blank_row = "none")


    rpt <- create_report(fp, output_type = "RTF",
                         font = "Arial", orientation = "portrait") %>%
      add_content(tbl) %>%
      footnotes("Here", footer = TRUE)

    res <- write_report(rpt)


   # file.show(res$modified_path)

    expect_equal(file.exists(fp), TRUE)
   # expect_equal(res$pages, 3)


  } else {

    expect_equal(TRUE, TRUE)

  }

})

# Works!
test_that("rtf2-75: Simplest EMF Plot works as expected.", {

  if (dev == TRUE) {

    library(devEMF)

    fp <- file.path(base_path, "rtf2/test75.rtf")

    tmp <- file.path(base_path, "rtf2/plot.emf")

    emf(tmp)

    plot(1, 1)


    dev.off()

    plt <- create_plot(tmp, height = 5, width = 6, borders = c("top", "bottom", "all")) %>%
      titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", borders = "none") %>%
      footnotes("* Motor Trend, 1974", borders = "none")


    rpt <- create_report(fp, output_type = "RTF", font = fnt, font_size =fsz) %>%
      page_header("Client", "Study: XYZ") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(plt) %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")


    res <- write_report(rpt)

    #print(res)

    #print(res)

    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)
})

# Works!
test_that("rtf2-76: Patchwork Plot works as expected.", {

  if (dev == TRUE) {


    library(ggplot2)
    library(patchwork)

    fp <- file.path(base_path, "rtf2/test76.rtf")

    p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()

    p2 <-  ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()

    ptch <- (p | p2)

    plt <- create_plot(ptch, height = 4, width = 8, borders = c("top", "bottom", "all")) %>%
      titles("Figure 1.0", "MTCARS Miles per Cylinder Plot", borders = "none") %>%
      footnotes("* Motor Trend, 1974", borders = "none")


    rpt <- create_report(fp, output_type = "RTF", font = fnt, font_size =fsz) %>%
      page_header("Client", "Study: XYZ") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(plt, align = "right") %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")


    res <- write_report(rpt)

    #print(res)

    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)

  } else
    expect_equal(TRUE, TRUE)

})


test_that("rtf2-77: One-page continuous tables work as expected.", {

  if (dev) {

    fp <- file.path(base_path, "rtf2/test77.rtf")

    dat <- iris[1:15, ]


    tbl <- create_table(dat, continuous = TRUE, borders = "all") %>%
      titles("My title") %>%
      footnotes("My footnotes", blank_row = "none")


    rpt <- create_report(fp, output_type = "RTF",
                         font = "Arial", orientation = "portrait") %>%
      add_content(tbl) %>%
      footnotes("Here", footer = TRUE)

    res <- write_report(rpt)


   # file.show(res$modified_path)

    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)


  } else {

    expect_equal(TRUE, TRUE)

  }

})


test_that("rtf2-78: Two one-page continuous tables work as expected.", {

  if (dev) {

    fp <- file.path(base_path, "rtf2/test78.rtf")

    dat <- iris[1:15, ]


    tbl1 <- create_table(dat, continuous = TRUE, borders = "all") %>%
      titles("My title 1") %>%
      footnotes("My footnotes 1", blank_row = "none")

    tbl2 <- create_table(dat, continuous = TRUE, borders = "all") %>%
      titles("My title 2") %>%
      footnotes("My footnotes 2", blank_row = "none")


    rpt <- create_report(fp, output_type = "RTF",
                         font = "Arial", orientation = "portrait") %>%
      add_content(tbl1) %>%
      add_content(tbl2) %>%
      footnotes("Here", footer = TRUE)

    res <- write_report(rpt)


   # file.show(res$modified_path)

    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 2)


  } else {

    expect_equal(TRUE, TRUE)

  }

})


test_that("rtf2-79: Two multi-page continuous tables work as expected.", {

  if (dev) {

    fp <- file.path(base_path, "rtf2/test79.rtf")

    dat <- iris[1:125, ]


    tbl1 <- create_table(dat, continuous = TRUE, borders = "all") %>%
      titles("My title 1") %>%
      footnotes("My footnotes 1", blank_row = "none")

    tbl2 <- create_table(dat, continuous = TRUE, borders = "all") %>%
      titles("My title 2") %>%
      footnotes("My footnotes 2", blank_row = "none")


    rpt <- create_report(fp, output_type = "RTF",
                         font = "Arial", orientation = "portrait") %>%
      add_content(tbl1) %>%
      add_content(tbl2) %>%
      footnotes("Here", footer = TRUE) %>%
      page_header("Left", "Right") %>%
      page_footer(right = "Page [pg] of [tpg]")

    res <- write_report(rpt)


   # file.show(res$modified_path)

    expect_equal(file.exists(fp), TRUE)
    #expect_equal(res$pages, 6)


  } else {

    expect_equal(TRUE, TRUE)

  }

})

test_that("rtf2-80: Two tables one continuous works as expected.", {

  if (dev) {

    fp <- file.path(base_path, "rtf2/test80.rtf")

    dat <- iris[1:125, ]


    tbl1 <- create_table(dat, continuous = TRUE, borders = "all") %>%
      titles("My title 1") %>%
      footnotes("My footnotes 1", blank_row = "none")

    tbl2 <- create_table(dat, continuous = FALSE, borders = "all") %>%
      titles("My title 2") %>%
      footnotes("My footnotes 2", blank_row = "none")


    rpt <- create_report(fp, output_type = "RTF",
                         font = "Arial", orientation = "portrait") %>%
      add_content(tbl1) %>%
      add_content(tbl2) %>%
      footnotes("Here", footer = TRUE) %>%
      page_header("Left", "Right") %>%
      page_footer(right = "Page [pg] of [tpg]")

    res <- write_report(rpt)


 #   file.show(res$modified_path)

    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 6)


  } else {

    expect_equal(TRUE, TRUE)

  }

})


# Basic Tests 81-90 -------------------------------------------------------



# Can't reproduce Raphael's problem
test_that("rtf2-81: Column widths work as expected.", {

  if (dev) {

    fp <- file.path(base_path, "rtf2/test81.rtf")

    dat <- mtcars[, 1:9]


    tbl <- create_table(dat, width = 8.84) %>%
      titles("My title") %>%
      column_defaults(width = .75) %>%
      define(mpg, width = 2.09) %>%
      define(cyl, width = 1.5) %>%
      footnotes("My footnotes", blank_row = "none")


    rpt <- create_report(fp, output_type = "RTF",
                         font = "Arial", orientation = "portrait",
                         paper_size = c(9.5, 11.5)) %>%
      add_content(tbl) %>%
      footnotes("Here", footer = TRUE) %>%
      set_margins(top = .25, bottom = .25, left = .25, right = .25)

    res <- write_report(rpt)


    #file.show(res$modified_path)

    res$column_widths

    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)


  } else {

    expect_equal(TRUE, TRUE)

  }

})


test_that("rtf2-82: Basic cell style bold works as expected.", {

  if (dev) {

    fp <- file.path(base_path, "rtf2/test82.rtf")

    dat <- mtcars[, 1:5]
    dat$hpflg <- ifelse(dat$hp > 100, TRUE, FALSE)


    tbl <- create_table(dat, width = 7) %>%
      titles("My title") %>%
      column_defaults(width = .75) %>%
      define(mpg, width = 1, style = cell_style(bold = TRUE)) %>%
      define(cyl, width = 1) %>%
      define(disp, style = cell_style(bold = TRUE)) %>%
      define(hp, style = cell_style(bold = TRUE, indicator = hpflg)) %>%
      define(hpflg, visible = FALSE) %>%
      footnotes("My footnotes", blank_row = "none")


    rpt <- create_report(fp, output_type = "RTF",
                         font = "Arial", orientation = "portrait") %>%
      add_content(tbl) %>%
      footnotes("Here", footer = TRUE)

    res <- write_report(rpt)


    # file.show(res$modified_path)

    res$column_widths

    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)


  } else {

    expect_equal(TRUE, TRUE)

  }

})

test_that("rtf2-83: Bolding works with stub.", {


  fp <- file.path(base_path, "rtf2/test83.rtf")


  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B
      "ampg"   "N"          "19"          "13"
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)"
      "ampg"   "Median"     "16.4"        "21.4"
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)"
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)"
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')

  df$cylflg <- ifelse(df$var == "cyl", TRUE, FALSE)

  # Create table
  tbl <- create_table(df, first_row_blank = TRUE) %>%
    column_defaults(vars = c("stub", "A"),
                    style = cell_style(bold = TRUE, indicator = cylflg)) %>%
    stub(c("var", "label"),
         style = cell_style(bold = TRUE, indicator = "labelrow")) %>%
    define(var, blank_after = TRUE, label_row = TRUE,
           format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>%
    define(label, indent = .25) %>%
    define(A, label = "Group A", align = "center", n = 19,
           style = cell_style(bold = TRUE, indicator = cylflg)) %>%
    define(B, label = "Group B", align = "center", n = 13,
           style = cell_style(bold = TRUE, indicator = "datarow")) %>%
    define(cylflg, visible = FALSE)


  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "RTF",
                       font = "Times") %>%
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>%
    titles("Table 1.0", "MTCARS Summary Table") %>%
    add_content(tbl) %>%
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left",
                center = "Confidential",
                right = "Page [pg] of [tpg]")



  res <- write_report(rpt)

  # file.show(res$modified_path)
  res
  expect_equal(file.exists(fp), TRUE)


})


test_that("rtf2-84: Bold cell style with column defaults.", {

  if (dev) {

    fp <- file.path(base_path, "rtf2/test84.rtf")

    dat <- mtcars[, 1:5]
    dat$hpflg <- ifelse(dat$hp > 100, TRUE, FALSE)


    tbl <- create_table(dat) %>%
      titles("My title") %>%
      column_defaults(width = .75, vars = c("cyl", "disp", "hp"),
                      style = cell_style(bold=TRUE, indicator = hpflg)) %>%
      define(mpg) %>%
      define(cyl) %>%
      define(disp) %>%
      define(hp) %>%
      define(hpflg, visible = FALSE) %>%
      footnotes("My footnotes", blank_row = "none")


    rpt <- create_report(fp, output_type = "RTF",
                         font = "Arial", orientation = "portrait") %>%
      add_content(tbl) %>%
      footnotes("Here", footer = TRUE)

    res <- write_report(rpt)


   # file.show(res$modified_path)

    res$column_widths

    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)


  } else {

    expect_equal(TRUE, TRUE)

  }

})


test_that("rtf2-85: Bolding, column defaults, and stub works.", {


  fp <- file.path(base_path, "rtf2/test85.rtf")


  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B
      "ampg"   "N"          "19"          "13"
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)"
      "ampg"   "Median"     "16.4"        "21.4"
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)"
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)"
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')

  df$cylflg <- ifelse(df$var == "cyl", FALSE, TRUE)

  # Create table
  tbl <- create_table(df, first_row_blank = TRUE) %>%
    column_defaults(style = cell_style(bold = TRUE, indicator = cylflg)) %>%
    stub(c("var", "label"), width = 2,
         style = cell_style(bold = TRUE, indicator = cylflg)) %>%
    define(var, blank_after = TRUE, label_row = TRUE,
           format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>%
    define(label, indent = .25) %>%
    define(A, label = "Group A", align = "center", n = 19) %>%
    define(B, label = "Group B", align = "center", n = 13,
           style = cell_style(bold = TRUE, indicator = "datarow")) %>%
    define(cylflg, visible = FALSE)


  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait", output_type = "RTF",
                       font = "Times") %>%
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>%
    titles("Table 1.0", "MTCARS Summary Table") %>%
    add_content(tbl) %>%
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left",
                center = "Confidential",
                right = "Page [pg] of [tpg]")

  res <- write_report(rpt)

  # file.show(res$modified_path)
  res
  expect_equal(file.exists(fp), TRUE)


})



test_that("rtf2-86: Spanning header bold work as expected.", {


  fp <- file.path(base_path, "rtf2/test86.rtf")

  dat <- mtcars[1:15, ]

  tbl <- create_table(dat, borders = c( "none")) %>%
    spanning_header(cyl, disp, "Span 1", label_align = "left") %>%
    spanning_header(hp, wt, "Span 2", underline = FALSE, bold = TRUE) %>%
    spanning_header(qsec, vs, "Span 3", n = 10) %>%
    spanning_header(drat, gear, "Super Duper\nWrapped Span",
                    n = 11, level = 2, bold = TRUE) %>%
    titles("Table 1.0", "My Nice Table", blank_row = "none",
           borders = c("top", "bottom")) %>%
    footnotes("My footnote 1", "My footnote 2",
              blank_row = "none", borders = c("top", "bottom"))

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(res$modified_path), TRUE)
  expect_equal(res$pages, 1)

})


test_that("rtf2-87: Italic footnotes work as expected.", {


  fp <- file.path(base_path, "rtf2/test87.rtf")

  dat <- mtcars[1:15, ]

  tbl <- create_table(dat, borders = c( "none")) %>%
    spanning_header(cyl, disp, "Span 1", label_align = "left") %>%
    spanning_header(hp, wt, "Span 2", underline = FALSE, bold = FALSE) %>%
    spanning_header(qsec, vs, "Span 3", n = 10) %>%
    spanning_header(drat, gear, "Super Duper\nWrapped Span",
                    n = 11, level = 2, bold = FALSE) %>%
    titles("Table 1.0", "My Nice Table", blank_row = "none",
           borders = c("top", "bottom")) %>%
    footnotes("My italic footnote1", "My italic footnote2", italics = TRUE,
              blank_row = "none", borders = "top") %>%
    footnotes("My italic footnote2", "My footnote 2",
              blank_row = "none", borders = c("bottom"))

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(res$modified_path), TRUE)
  expect_equal(res$pages, 1)

})


test_that("rtf2-88: Footnote columns work 1 column.", {

  fp <- file.path(base_path, "rtf2/test88.rtf")

  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)  %>%
    footnotes("Here is a footnote", "And another",
               borders = "all", columns = 1)

  rpt <- create_report(fp, output_type = "RTF", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below", columns =  1, align = "left") %>%
    footnotes("Here is a footnote", "And another", columns = 1)


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

})

test_that("rtf2-89: Footnote columns work 2 columns.", {

  fp <- file.path(base_path, "rtf2/test89.rtf")

  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = FALSE, visible = FALSE) %>%
    footnotes("Here is a footnote", "And another",
              borders = "all", columns = 2)

  rpt <- create_report(fp, output_type = "RTF", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame", "Left", "Right",
           blank_row = "below", columns =  2) %>%
    footnotes("Here is a footnote that is really long and is going to wrap",
              "And another",
              "Another left", "Another right", columns = 2)


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

})

test_that("rtf2-90: Footnote columns work 3 columns.", {

  fp <- file.path(base_path, "rtf2/test90.rtf")

  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = FALSE, visible = FALSE)  %>%
    footnotes("Here is a footnote", "And another", "And more",
              "", "centered",
              borders = "all", columns = 3, blank_row = "both")

  rpt <- create_report(fp, output_type = "RTF", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame", "My right thing", "", "Center",
           blank_row = "below", columns =  3) %>%
    footnotes("Here is a footnote that is really long an is going to wrap a few times",
              "And another", "and more",
              "", "Footnote Center", columns = 3)


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

})


# Basic Tests 91-100 ------------------------------------------------------



test_that("rtf2-91: Multiple footnote blocks work as expected.", {

  fp <- file.path(base_path, "rtf2/test91.rtf")

  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE) %>%
    footnotes("Footnote left", "Footnote right", columns = 2)

  rpt <- create_report(fp, output_type = "RTF", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    footnotes("Table 1.0", "IRIS Data Frame",
           blank_row = "below", columns =  1, align = "center", width = 7,
           borders = "all") %>%
    footnotes("Table 2.0", "IRIS Data Frame2", "Left", "Right",
           blank_row = "below", columns =  2, borders = "all") %>%
    footnotes("Table 3.0", "IRIS Data Frame3", "My right thing", "", "Center",
           blank_row = "below", columns =  3, borders = "all") %>%
    titles("Here is a title", "And another")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

})

test_that("rtf2-92: Page by with wrap works as expected.", {


  fp <- file.path(base_path, "rtf2/test92.rtf")

  dat <- iris
  dat$Pgby <- as.character(dat$Species)
  dat$Pgby <- paste0("Flower Type\n", dat$Pgby)


  tbl <- create_table(dat, borders = "none") %>%
    titles("Table 1.0", "My Nice Report with a Page By", borders = "none") %>%
    page_by(Pgby, label = "Species: ", align = "right", borders = "none") %>%
    define(Pgby, visible = FALSE)

  rpt <- create_report(fp, output_type = "RTF", font = fnt,
                       font_size = fsz, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_header("Left", "Right") %>%
    page_footer("Left1", "Center1", "Right1") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "none")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 9)
  expect_equal(length(res$column_widths[[1]]), 5)


})

test_that("rtf2-93: Page by with format works as expected.", {


  fp <- file.path(base_path, "rtf2/test93.rtf")

  fmt1 <- c(setosa = 1, versicolor = 2, virginica = 3)
  fmt2 <- value(condition(x == 1, "Setosa"),
                condition(x == 2, "Versicolor"),
                condition(x == 3, "Virginica"))

  dat <- iris
  fmtval <- fmt1[dat$Species]
  names(fmtval) <- NULL
  dat$Pgby <- fmtval

  tbl <- create_table(dat, borders = "none") %>%
    titles("Table 1.0", "My Nice Report with a Page By", borders = "none") %>%
    page_by(Pgby, align = "left", label = "Flower:", borders = "none", format = fmt2) %>%
    define(Pgby, visible = FALSE)

  rpt <- create_report(fp, output_type = "RTF",
                       orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_header("Left", "Right") %>%
    page_footer("Left1", "Center1", "Right1") %>%
    footnotes("My footnote 1", "My footnote 2", borders = "none")

  res <- write_report(rpt)
  res
  res$column_widths

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 6)
  expect_equal(length(res$column_widths[[1]]), 5)


})

test_that("rtf2-94: Label with invisible column works as expected.", {

  fp <- file.path(base_path, "rtf2/test94.rtf")

  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE, label = "Fork") %>%
    footnotes("Left", "right", columns = 2)

  rpt <- create_report(fp, output_type = "RTF") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    footnotes("Footnote1", "IRIS Data Frame",
              blank_row = "below", columns =  1, align = "center", width = 7,
              borders = "all") %>%
    titles("Table 1.0", "My little title")


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

})


test_that("rtf2-95: Page break with blank row after works as expected.", {

  fp <- file.path(base_path, "rtf2/test95.rtf")

  dat <- sort(iris, by = c("Species", "Petal.Width"))

  PG <- c(rep(1, 25), rep(2, 25),
               rep(4, 25), rep(5, 25),
               rep(7, 25), rep(8, 25))
  dat$PG <- PG


  tbl <- create_table(dat, borders = "all") %>%
    stub(c("Species", "Petal.Width"), label = "My stuff") %>%
    define(Species, visible = TRUE, label_row = TRUE) %>%
    define(Petal.Width, blank_after = TRUE, indent = .25) %>%
    define(PG, page_break = TRUE, visible = TRUE) %>%
    footnotes("Left", "right", columns = 2)

  rpt <- create_report(fp, output_type = "RTF", font = "Courier") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "Page [pg] of [tpg]") %>%
    footnotes("Footnote1", "IRIS Data Frame",
              blank_row = "below", columns =  1, align = "center", width = 7,
              borders = "all") %>%
    titles("Table 1.0", "My little title")


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 11)

})


test_that("rtf2-96: Outside borders on continuous tables work as expected.", {
  
  if (dev) {
    
    fp <- file.path(base_path, "rtf2/test96.rtf")
    
    dat <- iris
    
    
    tbl <- create_table(dat, continuous = TRUE, borders = "outside") %>%
      # titles("My Title") %>%
      footnotes("My footnotes", blank_row = "none")
    
    
    rpt <- create_report(fp, output_type = "RTF",
                         font = "Arial", orientation = "portrait") %>%
      add_content(tbl) %>%
      footnotes("Here", footer = TRUE)
    
    res <- write_report(rpt)
    
    
    # file.show(res$modified_path)
    
    expect_equal(file.exists(fp), TRUE)
    # expect_equal(res$pages, 3)
    
    
  } else {
    
    expect_equal(TRUE, TRUE)
    
  }
  
})

test_that("rtf2-97: Page X of Y works as expected.", {
  
  if (dev == TRUE) {
    
    
    fp <- file.path(base_path, "rtf2/test97.rtf")
    
    dat <- iris
    
    
    tbl <- create_table(dat, borders = "none") %>%
      titles("Table 1.0", "My Nice Irises", "Another Title", "Page [pg] of [tpg]") %>%
      define(Sepal.Length, label = "Sepal Length", width = 1, align = "center") %>%
      define(Sepal.Width, label = "Sepal Width", width = 1, align = "centre") %>%
      define(Species, blank_after = TRUE) %>%
      footnotes("Table footnote Page [pg] of [tpg]")
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = 12, orientation = "landscape") %>%
      titles("Report Title", "Page [pg] of [tpg]") %>%
      set_margins(top = 1, bottom = 1) %>%
      page_header("Left", c("Page [pg] of [tpg]")) %>%
      add_content(tbl, blank_row = "none") %>%
      page_footer("Left1", "Center1", "Page [pg] of [tpg]") %>%
      footnotes("My footnote 1", "My footnote 2", "Page [pg] of [tpg]")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 11)
    
  } else
    expect_equal(TRUE, TRUE)
})

test_that("rtf2-98: Title header blank rows work as expected.", {
  
  
  fp <- file.path(base_path, "rtf2/test98.rtf")
  
  dat <- iris
  
  tbl <- create_table(dat, width = 8.9, borders = "all") %>%
    title_header("Table 1.0", "My Nice Table",
                 right = c("Right1",
                           "Right2", "Page [pg] of [tpg]"),
                 blank_row = "both", borders = "all") 

  
  rpt <- create_report(fp, output_type = "RTF", font = "Arial",
                       font_size = 10, orientation = "landscape") %>%
    page_header("Left", "Right", blank_row = "below") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    footnotes("My footnote 1", "My footnote 2") %>%
    page_footer("Left1", "Center1", "Right1")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 7)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  
})


test_that("rtf2-99: Multi page table removes blank spaces.", {
  
  if (dev == TRUE) {
    
    
    fp <- file.path(base_path, "rtf2/test99.rtf")
    
    dat1 <- iris[1:10, ]
    dat2 <- iris[11:20, ]
    dat3 <- iris[21:30, ]
    
    
    tbl1 <- create_table(dat1, borders = "none") %>%
      titles("Table 1.0", "My Nice Irises1") 
    
    tbl2 <- create_table(dat2, borders = "none") %>%
      titles("Table 1.0", "My Nice Irises2")

    tbl3 <- create_table(dat3, borders = "none") %>%
      titles("Table 1.0", "My Nice Irises3")
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = 12, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl1, blank_row = "none") |> 
      add_content(tbl2, blank_row = "none") |> 
      add_content(tbl3, blank_row = "none") 
      
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 3)
    
  } else
    expect_equal(TRUE, TRUE)
})

test_that("rtf2-100: Spanning header gap works as expected.", {
  
  if (dev == TRUE) {
    
    fp <- file.path(base_path, "rtf2/test100.rtf")
    
    dat <- mtcars[1:15, ]
    
    tbl <- create_table(dat, borders = c("outside")) %>%
      spanning_header(cyl, disp, "Span 1", label_align = "left") %>%
      spanning_header(hp, wt, "Span 2", underline = TRUE) %>%
      spanning_header(qsec, vs, "Span 3", n = 10) %>%
      spanning_header(cyl, hp, "Super Span", n = 11, level = 2) |> 
      spanning_header(drat, gear, "Super Duper\nWrapped Span", n = 11, level = 2)
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
      titles("Table 1.0", "My Nice Table") %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2") %>%
      page_footer("Left1", "Center1", "Right1")
    
    res <- write_report(rpt)
    res
    res$column_widths
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)
    
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-100-2: Spanning header gap should not exist with all border.", {
  
  if (dev == TRUE) {
    
    fp <- file.path(base_path, "rtf2/test100-2.rtf")
    
    dat <- mtcars[1:15, ]
    
    tbl <- create_table(dat, borders = c("all")) %>%
      spanning_header(cyl, disp, "Span 1", label_align = "left") %>%
      spanning_header(hp, wt, "Span 2", underline = TRUE) %>%
      spanning_header(qsec, vs, "Span 3", n = 10) %>%
      spanning_header(cyl, hp, "Super Span", n = 11, level = 2) |> 
      spanning_header(drat, gear, "Super Duper\nWrapped Span", n = 11, level = 2)
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
      titles("Table 1.0", "My Nice Table") %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2") %>%
      page_footer("Left1", "Center1", "Right1")
    
    res <- write_report(rpt)
    res
    res$column_widths
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)
    
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-100-3: Spanning header gap should not exist with inside border.", {
  
  if (dev == TRUE) {
    
    fp <- file.path(base_path, "rtf2/test100-3.rtf")
    
    dat <- mtcars[1:15, ]
    
    tbl <- create_table(dat, borders = c("inside")) %>%
      spanning_header(cyl, disp, "Span 1", label_align = "left") %>%
      spanning_header(hp, wt, "Span 2", underline = TRUE) %>%
      spanning_header(qsec, vs, "Span 3", n = 10) %>%
      spanning_header(cyl, hp, "Super Span", n = 11, level = 2) |> 
      spanning_header(drat, gear, "Super Duper\nWrapped Span", n = 11, level = 2)
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
      titles("Table 1.0", "My Nice Table") %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2") %>%
      page_footer("Left1", "Center1", "Right1")
    
    res <- write_report(rpt)
    res
    res$column_widths
    
    expect_equal(file.exists(fp), TRUE)
    expect_equal(res$pages, 1)
    
  } else {
    expect_equal(TRUE, TRUE)
  }
})

# Basic Tests 101-110 ------------------------------------------------------


test_that("rtf2-101: Table with blank_before as expected.", {
  
  if (dev == TRUE) {
    fp <- file.path(base_path, "rtf2/test101.rtf")
    
    
    # Setup
    arm <- c(rep("A", 5), rep("B", 5))
    subjid <- 100:109
    name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
              "al-Kazemi, Najwa \nand more and more", "Schaffer, Ashley", "Laner, Tahma",
              "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
              "Smith, Shaileigh")
    sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
    age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
    
    
    # Create data frame
    df <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)
    
    
    tbl1 <- create_table(df, first_row_blank = FALSE) %>%
      define(subjid, label = "Subject ID for a patient", n = 10, align = "left",
             width = 1) %>%
      define(name, label = "Subject Name", width = 1) %>%
      define(sex, label = "Sex", n = 10, align = "center") %>%
      define(age, label = "Age", n = 10) %>%
      define(arm, label = "Arm",
             blank_before = TRUE,
             dedupe = TRUE)
    
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz) %>%
      titles("Table 1.0", align = "center") %>%
      
      add_content(tbl1)
    
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-102: Three level stub and indentation work as expected.", {
  
  if (dev == TRUE) {
    fp <- file.path(base_path, "rtf2/test102.rtf")
    
    
    # Setup
    cat <- c(rep("Kaplan-Meier estimates", 6), 
             rep("Cox PH estimates and some more really long stuff", 6))
    grp <- c("25th percentile", "25th percentile", 
             "median (weeks)", "median (weeks)",
             "75th percentile", "75th percentile",
             "25th percentile", "25th percentile", 
             "median (weeks)", "median (weeks)",
             "75th percentile", "75th percentile")
    ci <- c(NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval",
            NA, "95% confidence interval")
    values <- c(41, 53, 43, 39, 47, 52, 38, 25, 37, 23, 78, 21)
    
    # Create data frame
    df <- data.frame(cat, grp, ci, values, stringsAsFactors = FALSE)
    
    tbl1 <- create_table(df) %>%
      stub(c(cat, grp, ci), "Estimates", width = 1.5) %>% 
      define(cat, label_row = TRUE, blank_after = TRUE, indent = 0.15) %>%
      define(grp, indent = .25) %>%
      define(ci, indent = .5) %>%
      define(values, label = "Values")
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz) %>%
      titles("Table 3.0", "Analysis of Time to Initial PSGA Success in Weeks") %>% 
      page_header("Sponsor", "Study") %>% 
      add_content(tbl1) %>% 
      page_footer("Time", "Confidential", "Page")
    
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})


test_that("rtf2-103: Page footers with one assigned width work as expected.",{
  if (dev) {
    fp <- file.path(base_path, "rtf2/test103.rtf")
    
    dat <- mtcars[,c("mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      footnotes("This is testing footnote", blank_row = "none")
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz) %>%
      set_margins(top = 1, bottom = 1) %>%
      titles("Table 1.0") %>%
      add_content(tbl) %>%
      page_footer(left = "this is a very long sentence whose length exceeds the limitation of the left part", 
                  center = "Center footer", right = "Right footer",
                  width = c(5.5))
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
}
)

test_that("rtf2-104: Page footers with multiple assigned widths work as expected.",{
  if (dev) {
    fp <- file.path(base_path, "rtf2/test104.rtf")
    
    dat <- mtcars[,c("mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      footnotes("This is testing footnote", blank_row = "none")
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz) %>%
      set_margins(top = 1, bottom = 1) %>%
      titles("Table 1.0") %>%
      add_content(tbl) %>%
      page_footer(left = "this is a very long sentence whose length exceeds the limitation of the left part", 
                  center = "", right = "Right Footer",
                  width = c(5.5, 0 , 2))
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
}
)

test_that("rtf2-105: Page by with bold label and value works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "rtf2/test105.rtf")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type: ", bold = TRUE, blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-106: Page by with bold label as expected.", {
  
  if (dev) {
    # Label and Value are bold
    fp <- file.path(base_path, "rtf2/test106.rtf")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type: ", bold = "label", blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-107: Page by with bold value as expected.", {
  
  if (dev) {
    # Label and Value are bold
    fp <- file.path(base_path, "rtf2/test107.rtf")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type: ", bold = "value", blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-108: Page by with bold long label and value works as expected.", {

  if (dev) {
    fp <- file.path(base_path, "rtf2/test108.rtf")

    dat <- iris
    long_string <- "This is long page by value which should take more than one sentence - "
    dat$Pgby <- paste0(long_string, dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]

    long_label <- "This is a long page by label which should take more than one sentence: "

    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = long_label, bold = TRUE, blank_row = "none") %>%
      define(Pgby, visible = FALSE)

    rpt <- create_report(fp, output_type = "rtf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")

    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-109: Page by with bold long label works as expected.", {

  if (dev) {
    fp <- file.path(base_path, "rtf2/test109.rtf")

    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]

    long_label <- "This is a long page by label which should take more than one sentence: "

    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = long_label, bold = "label", blank_row = "none") %>%
      define(Pgby, visible = FALSE)

    rpt <- create_report(fp, output_type = "rtf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")

    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-110: Page by with bold long value works as expected.", {

  if (dev) {
    fp <- file.path(base_path, "rtf2/test110.rtf")

    dat <- iris
    long_string <- "This is long page by value which should take more than one sentence - "
    dat$Pgby <- paste0(long_string, dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]

    long_label <- "This is a long page by label which should take more than one sentence:"

    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = "Flower Type:", bold = "value", blank_row = "none") %>%
      define(Pgby, visible = FALSE)

    rpt <- create_report(fp, output_type = "rtf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")

    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

# Basic Tests 111-120 ------------------------------------------------------


test_that("rtf2-111: Page by with bold long label filling one line works as expected.", {
  
  if (dev) {
    fp <- file.path(base_path, "rtf2/test111.rtf")
    
    dat <- iris
    dat$Pgby <- as.character(dat$Species)
    dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
    
    long_label <- "This is a long page by label which should take more: "
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a Page By") %>%
      page_by(Pgby, label = long_label, bold = "label", blank_row = "none") %>%
      define(Pgby, visible = FALSE)
    
    rpt <- create_report(fp, output_type = "rtf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})


test_that("rtf2-112: Group border works as as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "rtf2/test112.rtf")
    
    # Setup
    arm <- c(rep("A", 3), rep("B", 2), rep("C", 3), rep("D", 2))
    subjid <- 100:109
    name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
              "al-Kazemi, Najwa \nand more and more", "Schaffer, Ashley", "Laner, Tahma",
              "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
              "Smith, Shaileigh")
    sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
    age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
    
    
    # Create data frame
    df <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)
    df <- rbind(df, df, df, df)
    
    tbl1 <- create_table(df, first_row_blank = FALSE, borders = "outside") %>%
      define(subjid, label = "Subject ID for a patient", n = 10, align = "left",
             width = 1) %>%
      define(name, label = "Subject Name", width = 1) %>%
      define(sex, label = "Sex", n = 10, align = "center") %>%
      define(age, label = "Age", n = 10) %>%
      define(arm, label = "Arm",
             dedupe = TRUE,
             group_border = TRUE) %>%
      footnotes("This is the footnote")
    
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz) %>%
      titles(c("Table 1.0", "This is a table with group border"), align = "center") %>%
      add_content(tbl1) %>%
      footnotes(c("This is the footnote 1")) %>%
      page_header(left = "Test header", right = "Test header") %>%
      set_margins(top = 1, bottom = 1)
    
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-113: Group border with blank after works as as expected.", {
  
  if (dev == TRUE) {
    fp <- file.path(base_path, "rtf2/test113.rtf")
    
    # Setup
    arm <- c(rep("A", 3), rep("B", 2), rep("C", 3), rep("D", 2))
    subjid <- 100:109
    name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
              "al-Kazemi, Najwa \nand more and more", "Schaffer, Ashley", "Laner, Tahma",
              "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
              "Smith, Shaileigh")
    sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
    age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)    
    
    # Create data frame
    df <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)
    df <- rbind(df, df, df, df)
    
    tbl1 <- create_table(df, first_row_blank = FALSE, borders = "outside") %>%
      define(subjid, label = "Subject ID for a patient", n = 10, align = "left",
             width = 1) %>%
      define(name, label = "Subject Name", width = 1) %>%
      define(sex, label = "Sex", n = 10, align = "center") %>%
      define(age, label = "Age", n = 10) %>%
      define(arm, label = "Arm",
             dedupe = TRUE,
             blank_after = TRUE,
             group_border = TRUE) %>%
      footnotes("This is the footnote")
    
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz) %>%
      titles(c("Table 1.0", "This is a table with group border"), align = "center") %>%
      add_content(tbl1) %>%
      footnotes(c("This is the footnote 1")) %>%
      page_header(left = "Test header", right = "Test header") %>%
      set_margins(top = 1, bottom = 1)
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-114: Center page header works as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "rtf2/test114.rtf")
    
    dat <- iris[1:50,]
    dat <- dat[, c("Species" ,"Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") 
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with a center header", header = T) %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none") %>%
      page_header(left = "", center = "center", right = "right",
                  width = c(0, 5, 4)) %>%
      page_footer(left = "Left", center = "center", right = "right",
                  width = c(3, 4, 2))
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-115: Images in page header/footer work as expected.", {
  
  if (dev == TRUE) {
    fp <- file.path(base_path, "rtf2/test115.rtf")
    
    dat <- iris[1:50,]
    dat <- dat[, c("Species" ,"Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")]
    
    image_path <- file.path(base_path, "data/logo.png")
    image_path2 <- file.path(base_path, "data/logo.jpg")
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table 1.0", "My Nice Report with a page header picture")
      
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none") %>%
      page_header() %>%
      header_image(image_path, height = 0.5, width = 0.8, align = "left") %>%
      header_image(image_path, height = 0.6, width = 0.85, align = "right") %>%
      header_image(c(image_path, image_path2), height = 0.38, width = 0.7, align = "center") %>%
      page_footer() %>%
      footer_image(c(image_path2, image_path), height = 0.38, width = 0.7, align = "left") %>%
      footer_image(image_path, height = 0.5, width = 0.8, align = "right") %>%
      footer_image(image_path, height = 0.6, width = 0.85, align = "center")

    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-116: Spanned images in page header/footer work as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "rtf2/test116.rtf")
    
    dat <- iris[1:50,]
    dat <- dat[, c("Species" ,"Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")]
    
    image_path <- file.path(base_path, "data/span_logo.png")

    tbl <- create_table(dat, borders = "outside") 
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with a page header picture", header = T) %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none") %>%
      page_header(width = c(0, 9, 0), left = "No display because width is 0") %>%
      header_image(image_path, height = 1, width = 7.5, align = "centre") %>%
      page_footer(width = c(0, 0, 9), center = "No display because width is 0") %>%
      footer_image(image_path, height = 0.8, width = 7, align = "right")

    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-117: Return error when header_image/footer_image is used without page_header/page_footer.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "rtf2/test117.rtf")
    
    dat <- iris[1:50,]
    dat <- dat[, c("Species" ,"Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")]
    
    image_path <- file.path(base_path, "data/span_logo.png")
    
    tbl <- create_table(dat, borders = "outside") 
    
    # page_header is not set
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with a page header picture", header = T) %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none") %>%
      header_image(image_path, height = 1, width = 7.5, align = "centre")

    expect_error(
      res <- write_report(rpt),
      "`page_header` must be used when using `header_image`."
    )
    
    # page_footer is not set
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with a page header picture", header = T) %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none") %>%
      footer_image(image_path, height = 0.8, width = 7, align = "right")
    
    expect_error(
      res <- write_report(rpt),
      "`page_footer` must be used when using `footer_image`."
    )
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-118: group_cohesion puts the group in same page as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "rtf2/test118.rtf")
    
    dat <- iris[1:86,]
    
    dat$test_group <- c(
      rep("A", 12),
      rep("B", 6),
      rep("C", 19),
      rep("D", 18),
      rep("E", 31)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 12), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 6), # 12 lines + 1
      c(rep("Flower C", 19)), # 19 lines + 1
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E", 31)) # 31 lines + 1
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, group_cohesion = TRUE, label = "Group", blank_after = T)
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-119: group_cohesion with min page proportion works as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "rtf2/test119.rtf")
    
    dat <- iris[1:86,]
    
    dat$test_group <- c(
      rep("A", 12),
      rep("B", 6),
      rep("C", 19),
      rep("D", 18),
      rep("E", 31)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 12), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 6), # 12 lines + 1
      c(rep("Flower C", 19)), # 19 lines + 1
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E", 31)) # 31 lines + 1
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, group_cohesion = 0.6,
             label = "Group", blank_after = T) %>%
      define(test_string, label = "Flower")
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-120: group_cohesion/min_page_prop return error properly.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "rtf2/test120.rtf")
    
    dat <- iris[1:86,]
    
    dat$test_group <- c(
      rep("A", 12),
      rep("B", 6),
      rep("C", 19),
      rep("D", 18),
      rep("E", 31)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 12), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 6), # 12 lines + 1
      c(rep("Flower C", 19)), # 19 lines + 1
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E", 31)) # 31 lines + 1
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    # ++++++++++++++++++++++
    # group_cohesion is not in right range
    # ++++++++++++++++++++++
    myfun <- function(dat){
      tbl <- create_table(dat, borders = "outside") %>%
        define(test_group, group_cohesion = 2,
               label = "Group", blank_after = T) %>%
        define(test_string, label = "Flower")
      
    }
    
    expect_error(
      res <- myfun(dat),
      paste0("`group_cohesion` should be TRUE/FALSE or numeric value from 0 to 1 instead of 2.")
    )
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-121: Multiple group_cohesion work as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "rtf2/test121.rtf")
    
    dat <- iris[1:86,]
    
    dat$test_group <- c(
      rep("A", 12),
      rep("B", 6),
      rep("C", 19),
      rep("D", 18),
      rep("E", 31)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 12), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 3), # 12 lines + 1
      rep("Flower B\nSubgroup B2", 3),
      c(rep("Flower C1", 11)), # C1 ends at 24 lines, which is the minimum requirement
      c(rep("Flower C2", 8)), # C2 would be moved to next page
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E1", 24), rep("Flower E2", 7)) # E2 would be moved to next page
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, group_cohesion = TRUE, label = "Group", blank_after = T) %>%
      define(test_string, group_cohesion = TRUE)
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    # C2 would be moved to next page
    # E2 would be moved to next page
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-122: Multiple group_cohesion with different cohesion strengths work as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "rtf2/test122.rtf")
    
    dat <- iris[1:92,]
    
    dat$test_group <- c(
      rep("A", 12),
      rep("B", 6),
      rep("C", 19),
      rep("D", 18),
      rep("E", 37)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 12), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 6), # 12 lines + 1
      c(rep("Flower C1", 11)), # C1 ends at 24 lines, which is the minimum requirement
      c(rep("Flower C2", 8)), # C2 would be not moved to next page
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E1", 33), rep("Flower E2", 4)) # E2 would be moved to next page
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, group_cohesion = TRUE, label = "Group", blank_after = T) %>%
      define(test_string, group_cohesion = 0.1)
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    # C2 would not be moved to next page because previous lines are only 24
    # E2 would be moved to next page because previous lines are 27, which meets 30*0.9
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-123: break_label works as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "rtf2/test123.rtf")
    
    dat <- iris[1:86,]
    
    dat$test_group <- c(
      rep("A", 12),
      rep("B", 6),
      rep("C", 19),
      rep("D", 18),
      rep("E", 31)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 12), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 6), # 12 lines + 1
      c(rep("Flower C", 19)), # 19 lines + 1
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E", 31)) # 31 lines + 1
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, break_label = "(Continued)", blank_before = T) %>%
      define(test_string, 
             break_label = paste0("(This is a long text which should take more ",
             "than one line. It's to test if the package is able to handle this ",
             "situation.)"),
             indent = 0.16)
    
    rpt <- create_report(fp, output_type = "RTF", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Break Label") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("rtf2-124: Page numbers work every in title, footnote, header, and footer.", {
  
  if (dev) {
    fp <- file.path(base_path, "rtf2/test124.rtf")
    
    dat <- iris
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table Title: Page [pg] of [tpg]") %>%
      footnotes("Table footnote: Page [pg] of [tpg]")
    
    rpt <- create_report(fp, output_type = "rtf", font = fnt,
                         font_size = fsz, orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      add_content(tbl) %>%
      titles("Report Title: Page [pg] of [tpg]") %>%
      footnotes("Report footnote: Page [pg] of [tpg]", borders = "none") %>%
      page_header("Header: Page [pg] of [tpg]") %>%
      page_footer("Footer: Page [pg] of [tpg]")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})


test_that("rtf2-125: Missing stub row label won't be displayed.", {
  
  if (dev) {
    fp <- file.path(base_path, "rtf2/test125.rtf")
    
    df <- read.table(header = TRUE, text = '
      var label A B
      "A" "Some text 1" "11" "8"
      "A" "Some text 2" "10" "7"
      "RACE" " White" "6 (54.5)" "4 (50.0)"
      "RACE" " Black" "4 (36.4)" "3 (37.5)"
      "RACE" " Others" "1 (9.1)" "1 (12.5)"
      "B" "Some text 3" "6 (54.5)" "4 (50.0)"
      "B" "Some text 4" "4 (36.4)" "3 (37.5)"
      ')
    
    df$var <- ifelse(df$var == "A", NA, df$var)
    df$var <- ifelse(df$var == "B", "", df$var)
    
    tbl <- create_table(df, first_row_blank = FALSE, borders = c("top", "bottom")) %>%
      titles("Table 1.0 Demographics", font_size = 10) %>%
      stub(vars = c("var", "label"), "", width = 2.5) %>%
      define(var, blank_after = TRUE, label_row = TRUE) %>%
      define(A, align = "center", width=1.2, label = "Placebo", n=11) %>%
      define(B, align = "center", width=1.2, label = "Drug", n=8)
    
    rpt <- create_report(fp, output_type = "rtf", font = "Arial") %>%
      add_content(tbl)
    
    # "" will still be row label. Only remove NA
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})
# User Tests --------------------------------------------------------------

test_that("user1: demo table works.", {

  if (dev) {
    library(tidyr)
    library(dplyr)

    # Data Filepath
    dir_data <- file.path(data_dir, "data")

    fp <- file.path(base_path, "rtf2/user1.rtf")


    # Load Data
    data_demo   <- file.path(dir_data, "dm.csv") %>%
      read.csv()


    data_demo <- subset(data_demo, data_demo$ARM != "SCREEN FAILURE")


    sex_decode <- c("M" = "Male",
                    "F" = "Female")

    race_decode <- c("WHITE" = "White",
                     "BLACK OR AFRICAN AMERICAN" = "Black or African American",
                     "ASIAN" = "Asian or Pacific Islander",
                     "NATIVE AMERICAN" = "Native American",
                     "UNKNOWN" = "Unknown")

    arm_pop <- table(data_demo$ARM)


    demo_age <-
      data_demo %>%
      group_by(ARM) %>%
      summarise(across(.cols = AGE,
                       .fns = list(N      = ~ fmt_n(.),
                                   Mean   = ~ fmt_mean_sd(.),
                                   Median = ~ fmt_median(.),
                                   `Q1 - Q3` = ~ fmt_quantile_range(.),
                                   Range  = ~ fmt_range(.)
                       ))) %>%
      pivot_longer(-ARM,
                   names_to  = c("var", "label"),
                   names_sep = "_",
                   values_to = "value") %>%
      pivot_wider(names_from = ARM,
                  values_from = "value")



    demo_sex <-
      data_demo %>%
      add_count(ARM, SEX,  name = "n_SEX") %>%
      select(ARM, SEX, n_SEX) %>%
      distinct() %>%
      pivot_longer(cols = c(SEX),
                   names_to  = "var",
                   values_to = "label") %>%
      pivot_wider(names_from  = ARM,
                  values_from = n_SEX,
                  values_fill = 0) %>%
      mutate(label = factor(label, levels = names(sex_decode),
                            labels = sex_decode),
             `ARM A` = fmt_cnt_pct(`ARM A`, arm_pop["ARM A"]),
             `ARM B` = fmt_cnt_pct(`ARM B`, arm_pop["ARM B"]),
             `ARM C` = fmt_cnt_pct(`ARM C`, arm_pop["ARM C"]),
             `ARM D` = fmt_cnt_pct(`ARM D`, arm_pop["ARM D"]))



    demo_race <-
      data_demo %>%
      add_count(ARM, RACE, name = "n_RACE") %>%
      select(ARM, RACE, n_RACE) %>%
      distinct() %>%
      pivot_longer(cols = RACE,
                   names_to  = "var",
                   values_to = "label") %>%
      pivot_wider(names_from  = ARM,
                  values_from = n_RACE,
                  values_fill = 0) %>%
      mutate(label = factor(label, levels = names(race_decode),
                            labels = race_decode),
             `ARM A` = fmt_cnt_pct(`ARM A`, arm_pop["ARM A"]),
             `ARM B` = fmt_cnt_pct(`ARM B`, arm_pop["ARM B"]),
             `ARM C` = fmt_cnt_pct(`ARM C`, arm_pop["ARM C"]),
             `ARM D` = fmt_cnt_pct(`ARM D`, arm_pop["ARM D"])) %>%
      arrange(var, label)


    demo <- bind_rows(demo_age, demo_sex, demo_race)


    #View(demo)


    # Stub decode
    block_fmt <- c(AGE = "Age", SEX = "Sex", RACE = "Race")

    # Define table
    tbl <- create_table(demo, first_row_blank = TRUE, borders = c("outside")) %>%
      column_defaults(from = "ARM A", to = "ARM D", width = 1.25) %>%
      define(var, blank_after = TRUE, dedupe = TRUE,
             format = block_fmt, label = "") %>%
      define(label, label = "") %>%
      define(`ARM A`, align = "center", label = "Placebo", n = 36) %>%
      define(`ARM B`, align = "center", label = "Drug 10mg", n = 38) %>%
      define(`ARM C`, align = "center", label = "Drug 20mg", n = 38) %>%
      define(`ARM D`, align = "center", label = "Competitor", n = 38)

    # Define Report
    rpt <- create_report(fp, output_type = "RTF", font = fnt, font_size = fsz) %>%
      set_margins(top = 1, bottom = 1) %>%
      options_fixed(font_size = 10) %>%
      titles("Table 14.1/4",
             "Demographics and Baseline to Characteristics",
             "Specify Population Ω µ β ¥ ∑ ≠ ≤ £ ∞ ؈ ლ  \Ub8a 鬼") %>%
      add_content(tbl) %>%
      footnotes("Special symbols \U221e to mess things up: Ω µ β ¥ ∑ ≠ ≤ £ ∞ ؈ ლ  \Ub8a 鬼 标题") %>%
      footnotes("Special symbols µ Ω £ there to mess things up: ", "Page [pg] of [tpg]",
                align = 'right', italics = TRUE) %>%
      page_header("Left µ Ω £ ", "Right") %>%
      page_footer("Time µ Ω £ ", right = "Page [pg] of [tpg]")

    # Write out report
    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)



  } else
    expect_equal(TRUE, TRUE)


})

test_that("user2: demo table with stub works.", {


  if (dev) {

    library(tidyr)
    library(dplyr)



    # Data Filepath
    dir_data <- file.path(data_dir, "data")

    fp <- file.path(base_path, "rtf2/user2.rtf")


    # Load Data
    data_demo   <- file.path(dir_data, "dm.csv") %>%
      read.csv()

    data_demo <- subset(data_demo, data_demo$ARM != "SCREEN FAILURE")


    sex_decode <- c("M" = "Male",
                    "F" = "Female")

    race_decode <- c("WHITE" = "White more",
                     "BLACK OR AFRICAN AMERICAN" = "Black or African American",
                     "ASIAN" = "Asian or Pacific Islander",
                     "NATIVE AMERICAN" = "Native American",
                     "UNKNOWN" = "Unknown")

    arm_pop <- table(data_demo$ARM)


    demo_age <-
      data_demo %>%
      group_by(ARM) %>%
      summarise(across(.cols = AGE,
                       .fns = list(N      = ~ fmt_n(.),
                                   Mean   = ~ fmt_mean_sd(.),
                                   Median = ~ fmt_median(.),
                                   `Q1 - Q3` = ~ fmt_quantile_range(.),
                                   Range  = ~ fmt_range(.)
                       ))) %>%
      pivot_longer(-ARM,
                   names_to  = c("var", "label"),
                   names_sep = "_",
                   values_to = "value") %>%
      pivot_wider(names_from = ARM,
                  values_from = "value")



    demo_sex <-
      data_demo %>%
      add_count(ARM, SEX,  name = "n_SEX") %>%
      select(ARM, SEX, n_SEX) %>%
      distinct() %>%
      pivot_longer(cols = c(SEX),
                   names_to  = "var",
                   values_to = "label") %>%
      pivot_wider(names_from  = ARM,
                  values_from = n_SEX,
                  values_fill = 0) %>%
      mutate(label = factor(label, levels = names(sex_decode),
                            labels = sex_decode),
             `ARM A` = fmt_cnt_pct(`ARM A`, arm_pop["ARM A"]),
             `ARM B` = fmt_cnt_pct(`ARM B`, arm_pop["ARM B"]),
             `ARM C` = fmt_cnt_pct(`ARM C`, arm_pop["ARM C"]),
             `ARM D` = fmt_cnt_pct(`ARM D`, arm_pop["ARM D"]))



    demo_race <-
      data_demo %>%
      add_count(ARM, RACE, name = "n_RACE") %>%
      select(ARM, RACE, n_RACE) %>%
      distinct() %>%
      pivot_longer(cols = RACE,
                   names_to  = "var",
                   values_to = "label") %>%
      pivot_wider(names_from  = ARM,
                  values_from = n_RACE,
                  values_fill = 0) %>%
      mutate(label = factor(label, levels = names(race_decode),
                            labels = race_decode),
             `ARM A` = fmt_cnt_pct(`ARM A`, arm_pop["ARM A"]),
             `ARM B` = fmt_cnt_pct(`ARM B`, arm_pop["ARM B"]),
             `ARM C` = fmt_cnt_pct(`ARM C`, arm_pop["ARM C"]),
             `ARM D` = fmt_cnt_pct(`ARM D`, arm_pop["ARM D"])) %>%
      arrange(var, label)


    demo <- bind_rows(demo_age, demo_sex, demo_race)


    #View(demo)


    # Stub decode
    block_fmt <- c(AGE = "Age", SEX = "Sex", RACE2 = "Race")

    # Define table
    tbl <- create_table(demo, first_row_blank = TRUE, borders = "all") %>%
      stub(c("var", "label"), width = 1.5) %>%
      column_defaults(width = 1) %>%
      define(var, blank_after = TRUE, blank_before = TRUE,
             format = block_fmt, label = "", label_row = TRUE) %>%
      define(label, label = "", indent = .25) %>%
      define(`ARM A`, align = "center", label = "Placebo", n = 36) %>%
      define(`ARM B`, align = "center", label = "Drug 10mg", n = 38) %>%
      define(`ARM C`, align = "center", label = "Drug 20mg", n = 38) %>%
      define(`ARM D`, align = "center", label = "Competitor", n = 38) %>%
      titles("Table 14.1/4",
             "Demographics and Baseline Characteristics",
             "Specify Population", borders = "outside", blank_row = "both",
             align = "right") %>%
      footnotes("Here is a footnote", "Here is another footnote",
                borders = "outside", blank_row = "both", align = "right")

    # Define Report
    rpt <- create_report(fp, output_type = "RTF",
                         font = "Arial", font_size = 10) %>%
      add_content(tbl, align = "right") %>%
      page_header("Sponsor", "Drug") %>%
      page_footer(left = "Time", right = "Page [pg] of [tpg]") #%>%
      #page_by(var = "var", label = "Variable: ")

    # Write out report
    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)


  } else
    expect_equal(TRUE, TRUE)

})

test_that("user3: listings works.", {
  if (dev == TRUE) {
    # Data Filepath
    dir_data <- file.path(data_dir, "data")

    fp <- file.path(base_path, "rtf2/user3.rtf")

    # Removing to make last page exactly equal to available rows on page.
    # In this case, any added blank rows should be skipped.
    fil <- c("ABC-14-124",
             "ABC-15-153",
             "ABC-15-154",
             "ABC-15-155",
             "ABC-15-156",
             "ABC-16-045",
             "ABC-16-046",
             "ABC-16-047",
             "ABC-16-157",
             "ABC-16-158",
             "ABC-16-159",
             "ABC-16-160")

    # Load Data
    data_demo   <- file.path(dir_data, "dm.csv") %>%
      read.csv()

    data_demo <- data_demo[!data_demo$USUBJID %in% fil, ]


    # Test that any assigned formats are applied
    attr(data_demo$SUBJID, "width") <- 1
    attr(data_demo$SUBJID, "justify") <- "left"
    attr(data_demo$SUBJID, "format") <- "S:%s"
    #print(widths(data_demo))
    names(data_demo)
    # Define table
    tbl <- create_table(data_demo) %>%
      define(USUBJID, id_var = TRUE)


    # Define Report
    rpt <- create_report(fp, font = "Times", font_size = 8,
                         orientation = "landscape") %>%
      titles("Listing 1.0",
             "Demographics Dataset") %>%
      add_content(tbl, align = "left") %>%
      page_header("Sponsor", "Drug") %>%
      page_footer(left = "Time", right = "Page [pg] of [tpg]") %>%
      footnotes("My footnotes")

    #Write out report
    res <- write_report(rpt, output_type = "RTF")

    expect_equal(file.exists(fp), TRUE)



    # pdfpth <- file.path(base_path, "user/user3.pdf")
    # write_report(rpt, pdfpth, output_type = "PDF")
    # expect_equal(file.exists(pdfpth), TRUE)
  } else
    expect_equal(TRUE, TRUE)


})


test_that("user4: listing in cm and times works.", {
  if (dev == TRUE) {
    # Data Filepath
    dir_data <- file.path(data_dir, "data")

    fp <- file.path(base_path, "rtf2/user4.rtf")

    # Removing to make last page exactly equal to available rows on page.
    # In this case, any added blank rows should be skipped.
    fil <- c("ABC-14-124",
             "ABC-15-153",
             "ABC-15-154",
             "ABC-15-155",
             "ABC-15-156",
             "ABC-16-045",
             "ABC-16-046",
             "ABC-16-047",
             "ABC-16-157",
             "ABC-16-158",
             "ABC-16-159",
             "ABC-16-160")

    # Load Data
    data_demo   <- file.path(dir_data, "dm.csv") %>%
      read.csv()

    data_demo <- data_demo[!data_demo$USUBJID %in% fil, ]


    # Test that any assigned formats are applied
    attr(data_demo$SUBJID, "width") <- 2.54
    attr(data_demo$SUBJID, "justify") <- "left"
    attr(data_demo$SUBJID, "format") <- "S:%s"
    #print(widths(data_demo))

    # Define table
    tbl <- create_table(data_demo) %>%
      define(USUBJID, id_var = TRUE)


    # Define Report
    rpt <- create_report(fp, font = "Arial", font_size = 10, units = "cm",
                         orientation = "portrait") %>%
      titles("Listing 1.0",
             "Demographics Dataset") %>%
      add_content(tbl, align = "left") %>%
      page_header("Sponsor", "Drug") %>%
      page_footer(left = "Time", right = "Page [pg] of [tpg]") %>%
      footnotes("My footnote")

    #Write out report
    res <- write_report(rpt, output_type = "RTF")

    #print(res$column_widths)

    expect_equal(file.exists(fp), TRUE)



    # pdfpth <- file.path(base_path, "user/user3.pdf")
    # write_report(rpt, pdfpth, output_type = "PDF")
    # expect_equal(file.exists(pdfpth), TRUE)
  } else
    expect_equal(TRUE, TRUE)


})

test_that("user5: Portrait in 12pt Arial works as expected.", {

  if (dev == TRUE) {

    dir_data <- file.path(data_dir, "data")

    fp <- file.path(base_path, "rtf2/user5.rtf")

    # Read in prepared data
    df <- read.table(header = TRUE, text = '
        var     label        A             B
        "ampg"   "N"          "19"          "13"
        "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)"
        "ampg"   "Median"     "16.4"        "21.4"
        "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
        "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
        "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)"
        "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)"
        "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')

    # Create table
    tbl <- create_table(df, first_row_blank = TRUE) %>%
      define(var, label = "Variable", blank_after = TRUE, dedupe = TRUE,
             format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>%
      define(label, label = "") %>%
      define(A, label = "Group A", align = "center", n = 19) %>%
      define(B, label = "Group B", align = "center", n = 13)


    # Create report and add content
    rpt <- create_report(fp, orientation = "portrait", output_type = "RTF",
                         font = "Arial", font_size = 12) %>%
      page_header(left = "Client: Motor Trend", right = "Study: Cars") %>%
      titles("Table 1.0", "MTCARS Summary Table") %>%
      add_content(tbl) %>%
      footnotes("* Motor Trend, 1974") %>%
      page_footer(left = "Left",
                  center = "Confidential",
                  right = "Page [pg] of [tpg]")

    # Write out report
    res <- write_report(rpt)

    expect_equal(file.exists(fp), TRUE)

  } else
    expect_equal(TRUE, TRUE)

})


test_that("user6: Borders with spanning headers work as expected.", {

  if (dev == TRUE) {

  df <- read.table(header = TRUE, text = '
      var     label           A             B             C
      "AGE"   "n"             "19"          "13"          "32"
      "AGE"   "Mean"          "18.8 (6.5)"  "22.0 (4.9)"  "20.0 (5.9)"
      "AGE"   "Median"        "16.4"        "21.4"        "20.1"
      "AGE"   "Q1 - Q3"       "15.1 - 21.2" "19.2 - 22.8" "15.2 - 21.8"
      "RACE"  "White"         "10 (52.6)"   "4 (30.8)"    "14 (43.8)"
      "RACE"  "Black"         "4 (21.1)"    "3 (23.1)"    "7 (21.9)"
      "RACE"  "Others\U1D47"  "5 (26.3)"    "6 (46.2)"    "11 (34.4.2)"
      ')

  var_fmt <- c(AGE = "Age (yrs)", RACE = "Race - n (%)")


  fp <- file.path(base_path, "rtf2/user6.rtf")

  tbl <- create_table(df, first_row_blank = TRUE, borders = c("outside")) %>%
    stub(vars = c("var", "label"), " ", width = 2.5) %>%
    spanning_header(from = "A", to = "B", label = "Treatments\U1D43") %>%
    define(var, blank_after = TRUE, format = var_fmt, label_row = TRUE) %>%
    define(label, indent=0.25) %>%
    define(A,  align = "center", label = "Placebo\n(N = 19)") %>%
    define(B,  align = "center", label = "Drug\n(N = 13)") %>%
    define(C,  align = "center", label = "Total\n(N = 32)") %>%
    footnotes("\U1D43 study drug", blank_row="none" ) %>%
    footnotes("\U1D47 Asian, Japanese and Chinese", blank_row="none")

  rpt <- create_report(fp, output_type = "RTF",
                       font = "Arial") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_footer(left = paste("Date:", "Left"), right = "Page [pg] of [tpg]", blank_row="none") %>%
    footnotes("Program: C:/Users/Home/AppData/Local/Temp/tdemo.R", blank_row="above")

  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  # file.show(fp)


  } else
    expect_equal(TRUE, TRUE)

})

# This is good
test_that("user7: Check footnotes on page by.", {

  if (dev == TRUE) {

  fp <- file.path(base_path, "rtf2/user7")


  df <- read.table(header = TRUE, text = '
      var     stat        A             B
      "Age (yrs)"   "n"          "19"          "13"
      " "   "Mean"       "18.8 (6.5)"  "22.0 (4.9)"
      " "   "Median"     "16.4"        "21.4"
      " "   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      " "   " "          " "           " "
      "Race"  "White" "10 ( 52.6%)" "4 ( 30.8%)"
      " "     "Black" "4 ( 21.1%)"  "3 ( 23.1%)"
      " "     "Others\U1D47" "5 ( 26.3%)"  "6 ( 46.2%)"')

  df1 <- df
  df2 <- df
  df3 <- df

  df1$SEX <- 'Female'
  df2$SEX <- 'Male'
  df3$SEX <- 'Other'

  df <- rbind(df1, df2, df3)

  # Create table
  tbl <- create_table(df, first_row_blank = TRUE, borders=c("top")) %>%
    page_by(SEX, "Sex: ", align = "left", blank_row="none") %>%
    # stub(c("var", "label")) %>%
    column_defaults(width = 1.25) %>%

    spanning_header(from = "A", to = "B", label = "Treatments\U1D43") %>%

    define(SEX, visible = FALSE) %>%
    define(var, label = " ", align = "left") %>%
    define(stat,label = " ", align = "left") %>%
    define(A,   label = "Treament A", align = "center", n = 19) %>%

    define(B,   label = "Treament B", align = "center", n = 13)  %>%
    titles("Table 1.1 Demographics", "Randomised Population", font_size = 10) %>%

    footnotes("Page [pg] of [tpg]", align = "right", blank_row="none", borders=c("top")) %>%
    footnotes("\U1D43 study drug treatments", blank_row="none", borders=c("top")  ) %>%
    footnotes("\U1D47 Asian, Japanese and Chinese", blank_row="none", borders="bottom")

  rpt <- create_report(fp, output_type = "RTF", font = "Arial", font_size = 10) %>%

    #This is page header and it goes into the header of the table
    page_header("Protocol: 9999") %>%

    add_content(tbl) %>%

    page_footer(left = paste("Date:", "Left"), right = "Page [pg] of [tpg]", blank_row="none") %>%
    footnotes("Program: C:/Users/Home/AppData/Local/Temp/tdemo.R", blank_row="above", valign = "bottom")



    res <- write_report(rpt)

    expect_equal(file.exists(res$modified_path), TRUE)
    expect_equal(res$pages, 3)

  } else
    expect_equal(TRUE, TRUE)

})

context("Report Spec Tests")


test_that("create_report sets default values appropriately", {
  
  ret <- create_report()
  
  expect_equal(ret$orientation, "landscape")
  expect_equal(ret$output_type, "TXT")
  
})


test_that("create_report changes parameters appropriately", {
  
  ret <- create_report(output_type = "txt", orientation = "portrait")
  
  expect_equal(ret$orientation, "portrait")
  expect_equal(ret$output_type, "TXT")
  
})



test_that("create_report traps invalid parameters", {
  
  expect_error(create_report(orientation = "porait"))
  expect_error(create_report(output_type = "text"))
  
  
})


test_that("options_fixed sets default parameters appropriately", {
  
  ret <- create_report()
  
  ret <- options_fixed(ret)
  
  expect_equal(ret$cpuom, 12)
  
})


test_that("options_fixed changes parameters appropriately", {
  
  ret <- create_report()
  
  ret <- options_fixed(ret, cpuom = 10)
  
  expect_equal(ret$cpuom, 10)
  
})


test_that("options_fixed traps invalid parameters appropriately", {
  
  ret <- create_report()
  
  expect_error(options_fixed(ret, cpuom = 16))
  
  
})

test_that("Titles, footnotes, header, and footer limits work as expected.", {
    
  rpt <- create_report("fork.out")
  st <- rep("W", 50)
  
  expect_error(titles(rpt, st))
  expect_error(footnotes(rpt, st))
  expect_error(page_header(rpt, left=st))
  expect_error(page_footer(rpt, left=st))
  
  
})



test_that("Title/footnotes parameter checks work as expected.", {
  
  rpt <- create_report("fork.out")
  
  
  # Parameter checks
  expect_error(page_footer(rpt, "Hello", blank_row = "fork"))
  expect_error(titles(rpt, "Hello", blank_row = "fork"))
  expect_error(titles(rpt, "Hello", borders  = "fork"))
  expect_error(title_header(titles(rpt, "Hello"), "Hello"))
  expect_error(titles(title_header(rpt, "Hello"), "Hello"))
  expect_error(title_header(rpt, "Hello", blank_row = "fork"))
  expect_error(title_header(rpt, "Hello", borders = "fork"))
  expect_error(title_header(page_header(rpt, "Hello"), "Hello"))
  expect_error(page_header(title_header(rpt, "Hello"), "Hello"))
})


test_that("Footnotes traps invalid parameter as expected.", {
  
  rpt <- create_report("fork.out")
  
  expect_error(footnotes(rpt, align = "error"))
  # expect_error(footnotes(rpt, valign = "error"))
  expect_error(footnotes(rpt, blank_row = "error"))
  expect_error(footnotes(rpt, borders = "error"))
  
})

test_that("add_content works as expected.", {
  
  rpt <- create_report("fork.out")
  
  rpt <- add_content(rpt, "", page_break = FALSE)
  
  # Should put a page break token before the content
  expect_equal(length(rpt$content), 1)
  expect_equal(rpt$content[[1]]$page_break, FALSE)
  
  # Invalid value
  expect_error(add_content(rpt, "", page_break = "sam"))

})


test_that("create_report parameter checks work as expected.", {
  
  
  expect_error(create_report(units = "fork"))
  expect_error(create_report(output_type = "fork"))
  expect_error(create_report(orientation = "fork"))
  expect_error(create_report(paper_size = "fork"))
  
  
  rpt <- create_report()
  expect_error(write_report(rpt))
  
  expect_error(write_report("fork"))
  expect_error(write_report(NA))
  
})

test_that("line_size and line_count parameter checks work as expected.", {

  rpt <- create_report()
  expect_error(options_fixed(rpt, line_size = "a"))  
  expect_error(options_fixed(rpt, line_size = -35))
  expect_error(options_fixed(rpt, line_count = "a"))
  expect_error(options_fixed(rpt, line_count = -876))
})




test_that("options_fixed parameter checks work as expected.", {
  
  
  rpt <- create_report()
  
  expect_error(options_fixed(rpt, editor = "fork"))
  expect_error(options_fixed(rpt, cpuom = -2))
  expect_error(options_fixed(rpt, lpuom = 2356))
  
})


test_that("preview parameter checks work as expected.", {
  
  rpt <- create_report()
  
  expect_error(write_report(rpt, preview = - 1))
  expect_error(write_report(rpt, preview = "a"))
  
})


test_that("font parameter checks work as expected.", {
  
  expect_error(create_report(font = "fork"))
  expect_error(create_report(font_size = 14))
  
})


test_that("title_header function works as expected.", {
  
  tbl <- create_table(mtcars)
  
  th <- tbl %>% title_header("Table 1.0", "MTCARS Sample Data",
                             right = c("One", "Two"), blank_row = "below")
  
  expect_equal(is.null(th$title_hdr[[1]]), FALSE)
  expect_equal(length(th$title_hdr[[1]]$titles), 2)
  expect_equal(length(th$title_hdr[[1]]$right), 2)
  expect_equal(th$title_hdr[[1]]$blank_row, "below")

  
})


test_that("page_by function works as expected.", {
  
  tbl <- create_table(mtcars)
  
  pg <- tbl %>% page_by(mpg, "MPG: ", "right", blank_row = "below")
  
  expect_equal(is.null(pg$page_by), FALSE)
  expect_equal(pg$page_by$var, "mpg")
  expect_equal(pg$page_by$label, "MPG: ")
  expect_equal(pg$page_by$align, "right")
  expect_equal(pg$page_by$blank_row, "below")
  
  pg <- tbl %>% page_by("mpg")
  
  expect_equal(is.null(pg$page_by), FALSE)
  expect_equal(pg$page_by$var, "mpg")
  expect_equal(is.null(pg$page_by$label), FALSE)
  expect_equal(pg$page_by$label, "mpg: ")
  expect_equal(pg$page_by$align, "left")
  expect_equal(pg$page_by$blank_row, "below")
  
  # Parameter checks
  expect_error(page_by(tbl, align = "fork"))
  expect_error(page_by(tbl, blank_row = "fork"))
  
  
})

test_that("Width parameter on titles and footnotes works as expected.", {
  
  tbl <- create_table(mtcars)
  
  th <- title_header(tbl, "Table 1.0", right = c("One", "Two"), 
                             width = "content")
  
  expect_equal(th$title_hdr[[1]]$width, "content")
  expect_error( title_header(tbl, "Table 1.0", right = c("One", "Two"), 
                             width = "bork"))
  
  tbl2 <- create_table(mtcars)
  ttl <- titles(tbl2, "Table 1.0", 
                     width = "page") 
  
  expect_equal(ttl$titles[[1]]$width, "page")
  expect_error( titles(tbl2, "Table 1.0", width = as.Date("2021-09-16")))
  
  ftnt <- footnotes(tbl2, "Table 1.0", 
                width = 6)
  
  expect_equal(ftnt$footnotes[[1]]$width, 6)
  expect_error( footnotes(tbl2, "Table 1.0", width = "fork"))
  
  tbl3 <- create_table(mtcars)
  ttl <- titles(tbl3, "Table 1.0") 
  
  expect_equal(ttl$titles[[1]]$width, "content")

})

test_that("output_type parameter checks work as expected.", {
  
  res <- create_report(output_type = "TXT")
  expect_equal(res$output_type, "TXT")
  
  res <- create_report(output_type = "PDF")
  expect_equal(res$output_type, "PDF")
  
  res <- create_report(output_type = "RTF")
  expect_equal(res$output_type, "RTF")
  
  res <- create_report(output_type = "HTML")
  expect_equal(res$output_type, "HTML")
  

})

context("Write Report Text Tests")

base_path <- "c:/packages/reporter/tests/testthat"

base_path <- tempdir()

test_that("page_template_text() works as expected.", {
  
  
  rpt <- create_report("fork.out") %>% 
    titles("Here is a test title.") %>% 
    footnotes("Test footnote")
  
  rpt$line_size <- 140
  
  
  ptt <- page_template_text(rpt)
  
  expect_equal(length(ptt$titles), 2)
  expect_equal(length(ptt$footnotes), 2)
  expect_equal(length(ptt$page_header), 0)
  expect_equal(length(ptt$page_footer), 0)
  
  
  rpt1 <- create_report("fork.out") 
  
  rpt1$line_size <- 140
  
  
  ptt1 <- page_template_text(rpt1)
  
  expect_equal(length(ptt1$titles), 0)
  expect_equal(length(ptt1$footnotes), 0)
  expect_equal(length(ptt1$page_header), 0)
  expect_equal(length(ptt1$page_footer), 0)
  
  
  
  rpt2 <- create_report("fork.out") %>% 
    page_header(left = c("Something1", "Something2"), right = "Nothing") %>% 
    titles("Here is a test title.", "There is a title") %>% 
    footnotes("Test footnote", "Test footnote 2") %>% 
    page_footer(center = "confidential")
  
  rpt2$line_size <- 140
  
  ptt2 <- page_template_text(rpt2)
  
  expect_equal(length(ptt2$titles), 3)
  expect_equal(length(ptt2$footnotes), 3)
  expect_equal(length(ptt2$page_header), 2)
  expect_equal(length(ptt2$page_footer), 2)
  
  # No line size.  Should get error.
  rpt3 <- create_report("fork.out") 
  expect_error(page_template_text(rpt3))
  
  rpt4 <- create_report("fork.out") %>% 
    titles(paste0(rep("W", 200), collapse =""))
  rpt4$line_size <- 140
  expect_warning(page_template_text(rpt4))
  
})


test_that("write_report output_type parameter checks work as expected.", {
  
  fp <- file.path(base_path, "text/wreport.out")
  
  rpt2 <- create_report(fp) %>% 
    page_header(left = c("Something1", "Something2"), right = "Nothing") %>% 
    titles("Here is a test title.", "There is a title") %>% 
    add_content(create_text("Hello")) %>% 
    footnotes("Test footnote", "Test footnote 2") %>% 
    page_footer(center = "confidential")
  
  expect_error(write_report(rpt2, output_type = "text"))
  res <- write_report(rpt2, output_type = "txt")
  
  expect_equal(res$output_type, "TXT")
  
  expect_equal(file.exists(fp), TRUE)
  
  if (file.exists(fp))
    file.remove(fp)
  
  
})


test_that("write_report paper size parameter check works as expected.", {
  
  fp <- file.path(base_path, "text/wreport.out")
  
  rpt2 <- create_report(fp, paper_size = "none") %>% 
    page_header(left = c("Something1", "Something2"), right = "Nothing") %>% 
    titles("Here is a test title.", "There is a title") %>% 
    add_content(create_text("Hello")) %>% 
    footnotes("Test footnote", "Test footnote 2") %>% 
    page_footer(center = "confidential")
  
  expect_error(write_report(rpt2))
  expect_error(write_report(rpt2, output_type = "RTF"))
  
  
  expect_equal(file.exists(fp), FALSE)
  
  if (file.exists(fp))
    file.remove(fp)
  
  
})


# test_that("write_page_numbers() works as expected.", {
#   
#   fp <- file.path(base_path, "text/text4.out")
#   
#   rtp <- create_report(fp) %>% 
#     page_footer(right = "Page [pg] of [tpg]")
#   
#   res2 <- write_page_numbers(rtp)
#   
#   res2
#   
#   expect_equal(grepl("[pg]", 
#   
#   
# })

context("Page Template PDF Tests")

test_that("split_string_text and page_text produce proper results.", {
  
  cs <- c("Left")
  
  pdf(NULL)
  par(family = "sans", ps = 10)
  
  
  tmp <- split_string_text(cs, 6.5, "inches")
  tmp
  
  ret <- page_text(tmp$text, 12, 
                   xpos = get_points_left(1, 
                   6.5,
                   tmp$widths,
                   units = "inches"),
                   ypos = 72)
  ret
  
  expect_equal(ret$text, "Left")
  expect_equal(ret$xpos, 75)
  expect_equal(ret$ypos, 72)
  
  cs <- c("Right")
  
  tmp <- split_string_text(cs, 6.5, "inches")
  tmp
  
  ret <- page_text(tmp$text, 12, 
                   xpos = get_points_right(1, 
                                          6.5,
                                          tmp$widths,
                                          units = "inches"),
                   ypos = 72)
  ret
  
  expect_equal(ret$text, "Right")
  expect_equal(round(ret$xpos), 442)
  expect_equal(ret$ypos, 72)
  
  dev.off()
  
  
})

test_that("page_template_pdf is working as expected.", {
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    page_header("Left", "Right") %>% 
    titles("Hello", blank_row = "below") %>%
    footnotes("Goodbye", blank_row = "below")
  
  rpt <- page_setup_pdf(rpt)
  
  expect_equal(rpt$page_template$page_header$lines, 1)
  expect_equal(length(rpt$page_template$page_header$pdf), 2)
  
  
})
          
        
test_that("get_titles_pdf function works as expected.", {
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello", blank_row = "below") %>%
    footnotes("Goodbye", blank_row = "below")
  
  rpt <- page_setup_pdf(rpt)
  rpt$page_template$titles
  
  expect_equal(rpt$page_template$titles$lines, 2) 
  expect_equal(rpt$page_template$titles$points, rpt$row_height * 2)
  
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello", "Goodbye", blank_row = "below") %>%
    footnotes("Goodbye", blank_row = "below")
  
  rpt <- page_setup_pdf(rpt)
  rpt$page_template$titles
  
  expect_equal(rpt$page_template$titles$lines, 3) 
  expect_equal(rpt$page_template$titles$points, rpt$row_height * 3)
  
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello", "Goodbye", blank_row = "both") %>%
    footnotes("Goodbye", blank_row = "below")
  
  rpt <- page_setup_pdf(rpt)
  rpt$page_template$titles
  
  expect_equal(rpt$page_template$titles$lines, 4) 
  expect_equal(rpt$page_template$titles$points, rpt$row_height * 4)
  
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello", "Goodbye", blank_row = "above") %>%
    footnotes("Goodbye", blank_row = "below")
  
  rpt <- page_setup_pdf(rpt)
  rpt$page_template$titles
  
  expect_equal(rpt$page_template$titles$lines, 3) 
  expect_equal(rpt$page_template$titles$points, rpt$row_height * 3)

  
})

test_that("get_titles_pdf function works as expected with borders.", {
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello", blank_row = "below", borders = "all", font_size = 14) %>%
    footnotes("Goodbye", blank_row = "below")
  
  rpt <- page_setup_pdf(rpt)
  rpt$page_template$titles
  
  rh <- get_line_height_pdf(14) + rpt$border_height 
  
  expect_equal(rpt$page_template$titles$points, (rh * 2) + rpt$border_spacing)
  
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello", "Goodbye", blank_row = "below", font_size = 14, 
           borders = "all") %>%
    footnotes("Goodbye", blank_row = "below")
  
  rpt <- page_setup_pdf(rpt)
  rpt$page_template$titles
  

  expect_equal(rpt$page_template$titles$points, (rh * 3) + rpt$border_spacing)
  
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello", "Goodbye", blank_row = "both", font_size = 14, 
           borders = "all") %>%
    footnotes("Goodbye", blank_row = "below")
  
  rpt <- page_setup_pdf(rpt)
  rpt$page_template$titles
  
  expect_equal(rpt$page_template$titles$points, (rh * 4) + rpt$border_spacing)
  
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello", "Goodbye", blank_row = "above", font_size = 14, 
           borders = "all") %>%
    footnotes("Goodbye", blank_row = "below")
  
  rpt <- page_setup_pdf(rpt)
  rpt$page_template$titles
  
  expect_equal(rpt$page_template$titles$points, (rh * 3) + rpt$border_spacing)
  
  
})


test_that("get_titles_pdf function works as expected with font_size.", {
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello", blank_row = "below", font_size = 14) %>%
    footnotes("Goodbye", blank_row = "below")
  
  rpt <- page_setup_pdf(rpt)
  rpt$page_template$titles
  
  rh <- get_line_height_pdf(14)
  
  expect_equal(rpt$page_template$titles$lines, 
               rpt$page_template$titles$points / rpt$row_height) 
  expect_equal(rpt$page_template$titles$points, rh * 2)
  
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello", "Goodbye", blank_row = "below", font_size = 14) %>%
    footnotes("Goodbye", blank_row = "below")
  
  rpt <- page_setup_pdf(rpt)
  rpt$page_template$titles
  

  expect_equal(rpt$page_template$titles$points, rh * 3)
  
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello", "Goodbye", blank_row = "both", font_size = 14) %>%
    footnotes("Goodbye", blank_row = "below")
  
  rpt <- page_setup_pdf(rpt)
  rpt$page_template$titles
  
  expect_equal(rpt$page_template$titles$points, rh * 4)
  
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello", "Goodbye", blank_row = "above", font_size = 14) %>%
    footnotes("Goodbye", blank_row = "below")
  
  rpt <- page_setup_pdf(rpt)
  rpt$page_template$titles
  
  expect_equal(rpt$page_template$titles$points, rh * 3)
  
  
})


test_that("get_footnotes_pdf function works as expected.", {

  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello") %>%
    footnotes("Goodbye")

  rpt <- page_setup_pdf(rpt)

  f <- rpt$page_template$footnotes
  f

  expect_equal(f$lines, 2)
})

test_that("get_titles_footnote function works as expected with font_size.", {
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    footnotes("a a a a a a a a a a",
              width = 1)
  rpt <- page_setup_pdf(rpt)
  f12 <- rpt$page_template$footnotes
  
  rh <- get_line_height_pdf(12)
  
  expect_equal(f12$lines, 3)
  expect_equal(f12$lines, 
               f12$points / rpt$row_height) 
  expect_equal(f12$points, rh * 3)
  
  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    footnotes("a a a a a a a a a a",
              width = 1, font_size = 8)
  rpt <- page_setup_pdf(rpt)
  f8 <- rpt$page_template$footnotes
  
  expect_equal(f8$lines, 2)
  expect_equal(f8$lines, 
               f8$points / rpt$row_height) 
  expect_equal(f8$points, rh * 2)
  
})

test_that("get_title_header_pdf function works as expected.", {

  rpt2 <- create_report("", font = "Arial", font_size = 12) %>%
    title_header("Hello", "Goodbye", right = paste("Right "))
  
  rpt2 <- page_setup_pdf(rpt2)
  
  th <-rpt2$page_template$title_hdr
  th

  expect_equal(th$lines, 3)
  
  rpt2 <- create_report("", font = "Arial", font_size = 12) %>%
    title_header("Hello", right = c("Right1", "Right2"))
  
  rpt2 <- page_setup_pdf(rpt2)
  
  th <-rpt2$page_template$title_hdr
  th

  expect_equal(th$lines, 3)
  
  
  rpt2 <- create_report("", font = "Arial", font_size = 12) %>%
    title_header("Hello", right = c("Right1", "Right2"), blank_row = "both")
  
  rpt2 <- page_setup_pdf(rpt2)
  
  th <-rpt2$page_template$title_hdr
  th
  
  rh <- get_line_height_pdf(12)
  
  expect_equal(th$lines, 4)
  expect_equal(th$points, rh * 4 + 1)
  
  rpt2 <- create_report("", font = "Arial", font_size = 12) %>%
    title_header("Hello", right = paste("Right here is something",
                                        "really long that will wrap and wrap", 
                                        "and wrap and wrap keep wrapping"))

  rpt2 <- page_setup_pdf(rpt2)

  th <-rpt2$page_template$title_hdr
  th
  expect_equal(length(th$pdf), 4)
  expect_equal(th$lines, 4)

})

# test_that("get_title_header_pdf function works as expected with borders.", {
# 
#   rpt2 <- create_report("", font = "Arial", font_size = 12) %>%
#     title_header("Hello", "Goodbye", right = paste("Right "), borders = "all")
# 
#   rpt2 <- page_setup_pdf(rpt2)
# 
#   rh <- get_line_height_pdf(12)
# 
#   th <-rpt2$page_template$title_hdr
#   th
# 
#   expect_equal(th$lines, 3)
#   expect_equal(th$points, rh * 3 + rpt2$border_height )
# 
#   rpt2 <- create_report("", font = "Arial", font_size = 12) %>%
#     title_header("Hello", right = c("Right1", "Right2"), borders = "all")
# 
#   rpt2 <- page_setup_pdf(rpt2)
# 
#   th <-rpt2$page_template$title_hdr
#   th
# 
#   expect_equal(th$lines, 3)
#   expect_equal(th$points, rh * 3 + rpt2$border_height )
# 
# 
#   rpt2 <- create_report("", font = "Arial", font_size = 12) %>%
#     title_header("Hello", right = c("Right1", "Right2"), borders = "all",
#                  blank_row = "both")
# 
#   rpt2 <- page_setup_pdf(rpt2)
# 
#   th <-rpt2$page_template$title_hdr
#   th
# 
# 
# 
#   expect_equal(th$lines, 4)
#   expect_equal(th$points, rh * 4 + rpt2$border_height )
# 
#   rpt2 <- create_report("", font = "Arial", font_size = 12) %>%
#     title_header("Hello", right = paste("Right here is something",
#                                         "really long that will wrap and wrap",
#                                         "and wrap and wrap keep wrapping"),
#                  borders = "all")
# 
#   rpt2 <- page_setup_pdf(rpt2)
# 
#   th <-rpt2$page_template$title_hdr
#   th
#   #expect_equal(length(th$pdf), 12)
#   expect_equal(length(th$pdf), 6)
#   expect_equal(th$lines, 4)
#   expect_equal(th$points, rh * 4 + rpt2$border_height )
# 
# })


test_that("get_page_header_pdf works as expected.", {

  rpt1 <- create_report("", font = "Arial", font_size = 12) %>%
    page_header(left= c("Hello"), right = "there")

  rpt1 <- page_setup_pdf(rpt1)

  ph <- get_page_header_pdf(rpt1)
  ph

  expect_equal(ph$lines, 1)
  
  
  rpt2 <- create_report("", font = "Arial", font_size = 12) %>%
    page_header(left= c("Hello", "Goodbye"), right = "there") 
  
  rpt2 <- page_setup_pdf(rpt2)
  
  ph <- get_page_header_pdf(rpt2)
  ph
  
  expect_equal(ph$lines, 2)
  
  
  rpt3 <- create_report("", font = "Arial", font_size = 12) %>%
    page_header(left= c("Hello\nthere", "Goodbye"), right = "Right") 
  
  rpt3 <- page_setup_pdf(rpt3)
  
  ph <- get_page_header_pdf(rpt3)
  ph
  
  expect_equal(ph$lines, 3)

})


test_that("get_page_footer_pdf works as expected.", {

  rpt3 <- create_report("", font = "Arial", font_size = 12) %>%
    page_footer("Left", "Center", c("Right here", "is something"))

  rpt3 <- page_setup_pdf(rpt3)

  pf <- rpt3$page_template$page_footer
  pf

  expect_equal(pf$lines, 3)


})



test_that("get_pageby_pdf works as expected.", {

  tbl <- create_table(mtcars) %>%
    page_by(cyl, "Cylinders:")

  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    titles("Hello") %>%
    footnotes("Goodbye") %>%
    add_content(tbl)

  rpt <- page_setup_pdf(rpt)

  rpt$line_height

  res <- get_page_by_pdf(tbl$page_by, 6, "fork", rpt, "left")
  res

  expect_equal(res$lines, 2)
  expect_equal(length(res$pdf), 1)

})




test_that("page_template_pdf works as expected.", {


  tbl <- create_table(mtcars) %>%
    page_by(cyl, "Cylinders:")

  rpt <- create_report("", font = "Arial", font_size = 12) %>%
    page_header("left", "right") %>%
    titles("Hello") %>%
    footnotes("Goodbye") %>%
    add_content(tbl) %>%
    page_footer("left", right = "right")


  rpt <- page_setup_pdf(rpt)



  res <- page_template_pdf(rpt)

  res
  expect_equal(res$lines, 7)
  expect_equal(res$titles$lines, 2)


})

context("User Tests")

base_path <- paste0(getwd(),"/tests/testthat")
data_dir <- base_path


base_path <- tempdir()
data_dir <- "."

dev <- FALSE


test_that("user1: demo table works.", {

  if (dev) {
    library(tidyr)
    library(dplyr)
  
    # Data Filepath
    dir_data <- file.path(data_dir, "data")
    
    fp <- file.path(base_path, "user/user1.out")
    
  
    # Load Data
    data_demo   <- file.path(dir_data, "dm.csv") %>%
      read.csv() 
    
    
    data_demo <- subset(data_demo, data_demo$ARM != "SCREEN FAILURE")
  
  
    sex_decode <- c("M" = "Male",
                    "F" = "Female")
  
    race_decode <- c("WHITE" = "White",
                     "BLACK OR AFRICAN AMERICAN" = "Black or African American",
                     "ASIAN" = "Asian or Pacific Islander",
                     "NATIVE AMERICAN" = "Native American",
                     "UNKNOWN" = "Unknown")
  
    arm_pop <- table(data_demo$ARM) 
  
  
    demo_age <-
      data_demo %>%
      group_by(ARM) %>%
      summarise(across(.cols = AGE,
                       .fns = list(N      = ~ fmt_n(.),
                                   Mean   = ~ fmt_mean_sd(.),
                                   Median = ~ fmt_median(.),
                                   `Q1 - Q3` = ~ fmt_quantile_range(.),
                                   Range  = ~ fmt_range(.)
                       ))) %>%
      pivot_longer(-ARM,
                   names_to  = c("var", "label"),
                   names_sep = "_",
                   values_to = "value") %>%
      pivot_wider(names_from = ARM,
                  values_from = "value")
  
  
  
    demo_sex <-
      data_demo %>%
      add_count(ARM, SEX,  name = "n_SEX") %>%
      select(ARM, SEX, n_SEX) %>%
      distinct() %>%
      pivot_longer(cols = c(SEX),
                   names_to  = "var",
                   values_to = "label") %>%
      pivot_wider(names_from  = ARM,
                  values_from = n_SEX,
                  values_fill = 0) %>%
      mutate(label = factor(label, levels = names(sex_decode),
                            labels = sex_decode),
             `ARM A` = fmt_cnt_pct(`ARM A`, arm_pop["ARM A"]),
             `ARM B` = fmt_cnt_pct(`ARM B`, arm_pop["ARM B"]),
             `ARM C` = fmt_cnt_pct(`ARM C`, arm_pop["ARM C"]),
             `ARM D` = fmt_cnt_pct(`ARM D`, arm_pop["ARM D"]))
  
  
  
    demo_race <-
      data_demo %>%
      add_count(ARM, RACE, name = "n_RACE") %>%
      select(ARM, RACE, n_RACE) %>%
      distinct() %>%
      pivot_longer(cols = RACE,
                   names_to  = "var",
                   values_to = "label") %>%
      pivot_wider(names_from  = ARM,
                  values_from = n_RACE,
                  values_fill = 0) %>%
      mutate(label = factor(label, levels = names(race_decode),
                            labels = race_decode),
             `ARM A` = fmt_cnt_pct(`ARM A`, arm_pop["ARM A"]),
             `ARM B` = fmt_cnt_pct(`ARM B`, arm_pop["ARM B"]),
             `ARM C` = fmt_cnt_pct(`ARM C`, arm_pop["ARM C"]),
             `ARM D` = fmt_cnt_pct(`ARM D`, arm_pop["ARM D"])) %>%
      arrange(var, label)
  
  
    demo <- bind_rows(demo_age, demo_sex, demo_race)
  
  
    #View(demo)
  
  
    # Stub decode
    block_fmt <- c(AGE = "Age", SEX = "Sex", RACE = "Race")
  
    # Define table
    tbl <- create_table(demo, first_row_blank = TRUE) %>%
      column_defaults(from = "ARM A", to = "ARM D", width = 1) %>% 
      define(var, blank_after = TRUE, dedupe = TRUE,
             format = block_fmt, label = "") %>%
      define(label, label = "") %>%
      define(`ARM A`, align = "center", label = "Placebo", n = 36) %>%
      define(`ARM B`, align = "center", label = "Drug 10mg", n = 38) %>%
      define(`ARM C`, align = "center", label = "Drug 20mg", n = 38) %>%
      define(`ARM D`, align = "center", label = "Competitor", n = 38)
  
    # Define Report
    rpt <- create_report(fp) %>%
      set_margins(top = 1, bottom = 1) %>% 
      options_fixed(font_size = 10) %>% 
      titles("Table 14.1/4",
             "Demographics and Baseline to Characteristics",
             "Specify Population") %>%
      add_content(tbl) %>% 
      #footnotes("Special symbols \U221e to mess things up: Ω µ β ¥ ∑ ≠ ≤ £ ∞ ؈ ლ  \Ub8a 鬼") %>%   
      footnotes("Special symbols µ Ω £ there to mess things up: ") %>% 
      page_footer("Time", right = "Page [pg] of [tpg]")
  
    # Write out report
    res <- write_report(rpt)
  
    expect_equal(file.exists(fp), TRUE)
    
    lns <- readLines(fp, encoding = "native.enc")
    
    expect_equal(length(lns), res$pages * res$line_count)
    
    if (TRUE) {
      rtfpth <- file.path(base_path, "user/user1.rtf")
      write_report(rpt, rtfpth, output_type = "RTF")
      expect_equal(file.exists(rtfpth), TRUE)
      

      pdfpth <- file.path(base_path, "user/user1.pdf")
      write_report(rpt, pdfpth, output_type = "PDF")
      expect_equal(file.exists(pdfpth), TRUE)
      
      
      docxpth <- file.path(base_path, "user/user1.docx")
      write_report(rpt, docxpth, output_type = "DOCX")
      expect_equal(file.exists(docxpth), TRUE)
    }
  
  } else 
    expect_equal(TRUE, TRUE)
  
  
})

test_that("user2: demo table with stub works.", {
  
  
  if (dev) {
    
    library(tidyr)
    library(dplyr)
    
    
    
    # Data Filepath
    dir_data <- file.path(data_dir, "data")
    
    fp <- file.path(base_path, "user/user2.out")
    
    
    # Load Data
    data_demo   <- file.path(dir_data, "dm.csv") %>%
      read.csv() 
    
    data_demo <- subset(data_demo, data_demo$ARM != "SCREEN FAILURE")
    
    
    sex_decode <- c("M" = "Male",
                    "F" = "Female")
    
    race_decode <- c("WHITE" = "White",
                     "BLACK OR AFRICAN AMERICAN" = "Black or African American",
                     "ASIAN" = "Asian or Pacific Islander",
                     "NATIVE AMERICAN" = "Native American",
                     "UNKNOWN" = "Unknown")
    
    arm_pop <- table(data_demo$ARM) 
    
    
    demo_age <-
      data_demo %>%
      group_by(ARM) %>%
      summarise(across(.cols = AGE,
                       .fns = list(N      = ~ fmt_n(.),
                                   Mean   = ~ fmt_mean_sd(.),
                                   Median = ~ fmt_median(.),
                                   `Q1 - Q3` = ~ fmt_quantile_range(.),
                                   Range  = ~ fmt_range(.)
                       ))) %>%
      pivot_longer(-ARM,
                   names_to  = c("var", "label"),
                   names_sep = "_",
                   values_to = "value") %>%
      pivot_wider(names_from = ARM,
                  values_from = "value")
    
    
    
    demo_sex <-
      data_demo %>%
      add_count(ARM, SEX,  name = "n_SEX") %>%
      select(ARM, SEX, n_SEX) %>%
      distinct() %>%
      pivot_longer(cols = c(SEX),
                   names_to  = "var",
                   values_to = "label") %>%
      pivot_wider(names_from  = ARM,
                  values_from = n_SEX,
                  values_fill = 0) %>%
      mutate(label = factor(label, levels = names(sex_decode),
                            labels = sex_decode),
             `ARM A` = fmt_cnt_pct(`ARM A`, arm_pop["ARM A"]),
             `ARM B` = fmt_cnt_pct(`ARM B`, arm_pop["ARM B"]),
             `ARM C` = fmt_cnt_pct(`ARM C`, arm_pop["ARM C"]),
             `ARM D` = fmt_cnt_pct(`ARM D`, arm_pop["ARM D"]))
    
    
    
    demo_race <-
      data_demo %>%
      add_count(ARM, RACE, name = "n_RACE") %>%
      select(ARM, RACE, n_RACE) %>%
      distinct() %>%
      pivot_longer(cols = RACE,
                   names_to  = "var",
                   values_to = "label") %>%
      pivot_wider(names_from  = ARM,
                  values_from = n_RACE,
                  values_fill = 0) %>%
      mutate(label = factor(label, levels = names(race_decode),
                            labels = race_decode),
             `ARM A` = fmt_cnt_pct(`ARM A`, arm_pop["ARM A"]),
             `ARM B` = fmt_cnt_pct(`ARM B`, arm_pop["ARM B"]),
             `ARM C` = fmt_cnt_pct(`ARM C`, arm_pop["ARM C"]),
             `ARM D` = fmt_cnt_pct(`ARM D`, arm_pop["ARM D"])) %>%
      arrange(var, label)
    
    
    demo <- bind_rows(demo_age, demo_sex, demo_race)
    
    
    #View(demo)
    
    
    # Stub decode
    block_fmt <- c(AGE = "Age", SEX = "Sex", RACE2 = "Race")
    
    # Define table
    tbl <- create_table(demo, first_row_blank = TRUE) %>%
      stub(c("var", "label")) %>% 
      define(var, blank_after = TRUE, 
             format = block_fmt, label = "", label_row = TRUE) %>%
      define(label, label = "", indent = .25) %>%
      define(`ARM A`, align = "center", label = "Placebo", n = 36) %>%
      define(`ARM B`, align = "center", label = "Drug 10mg", n = 38) %>%
      define(`ARM C`, align = "center", label = "Drug 20mg", n = 38) %>%
      define(`ARM D`, align = "center", label = "Competitor", n = 38)
    
    # Define Report
    rpt <- create_report(fp) %>%
      titles("Table 14.1/4",
             "Demographics and Baseline Characteristics",
             "Specify Population") %>%
      add_content(tbl)
    
    # Write out report
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
    
    lns <- readLines(fp)
    
    expect_equal(length(lns), res$pages * res$line_count)
    
    rtfpth <- file.path(base_path, "user/user2.rtf")
    write_report(rpt, rtfpth, output_type = "RTF")
    expect_equal(file.exists(rtfpth), TRUE)
  
  } else
    expect_equal(TRUE, TRUE)
})


test_that("user3: listings works.", {
  
    # Data Filepath
    dir_data <- file.path(data_dir, "data")
    
    fp <- file.path(base_path, "user/user3.out")
    
    # Removing to make last page exactly equal to available rows on page.
    # In this case, any added blank rows should be skipped.
    fil <- c("ABC-14-124",
    "ABC-15-153",
    "ABC-15-154",
    "ABC-15-155",
    "ABC-15-156",
    "ABC-16-045",
    "ABC-16-046",
    "ABC-16-047",
    "ABC-16-157",
    "ABC-16-158",
    "ABC-16-159", 
    "ABC-16-160")
    
    # Load Data
    data_demo   <- file.path(dir_data, "dm.csv") %>%
      read.csv() 
    
    data_demo <- data_demo[!data_demo$USUBJID %in% fil, ]
    
    
    # Test that any assigned formats are applied
    attr(data_demo$SUBJID, "width") <- 1
    attr(data_demo$SUBJID, "justify") <- "left"
    attr(data_demo$SUBJID, "format") <- "S:%s"
    #print(widths(data_demo))
  
    # Define table
    tbl <- create_table(data_demo) %>% 
      define(USUBJID, id_var = TRUE) 
  
  
    # Define Report
    rpt <- create_report(fp) %>%
      options_fixed(editor = "notepad", font_size = 10) %>% 
      titles("Listing 1.0",
             "Demographics Dataset") %>%
      add_content(tbl, align = "left") %>% 
      page_footer(left = "Time", right = "Page [pg] of [tpg]")
    
    #Write out report
    res <- write_report(rpt)
  
    expect_equal(file.exists(fp), TRUE)
  
    lns <- readLines(fp)
  
    expect_equal(length(lns), res$pages * res$line_count)
    
    
    rtfpth <- file.path(base_path, "user/user3.rtf")
    write_report(rpt, file_path = rtfpth, output_type = "RTF")
    
    expect_equal(file.exists(rtfpth), TRUE)
    
    

    pdfpth <- file.path(base_path, "user/user3.pdf")
    write_report(rpt, pdfpth, output_type = "PDF")
    expect_equal(file.exists(pdfpth), TRUE)
    
    
    docxpth <- file.path(base_path, "user/user3.docx")
    write_report(rpt, docxpth, output_type = "DOCX")
    expect_equal(file.exists(docxpth), TRUE)
    
})

test_that("user4: Adverse Events table works.", {

  if (dev) {
    #devtools::install_github("https://github.com/dbosak01/fmtr")
    library(dplyr)
    library(tidyr)
  
    # Data Filepath
    dir_data <- file.path(data_dir, "data")
  
    fp <- file.path(base_path, "user/user4.out")
  
    dp <- file.path(dir_data, "ADAE.csv")
    
    dat <- read.csv(dp)
    
    
    # Get population counts
    arm_pop <- table(dat$TRTA)  
    
    # Subset ADSL for needed rows and columns
    df_sub <- dat %>% 
      select(TRTA, AESEV, AESEVN, AEREL, AESOC, AEDECOD) 
    
    
    # Get counts and percents 
    df1 <- df_sub %>% 
      select(-AESEV, -AEREL) %>% 
      group_by(TRTA, AESOC, AEDECOD, AESEVN) %>% 
      summarize(cnt = n()) %>% 
      pivot_wider(names_from = c(TRTA, AESEVN),
                  values_from = cnt, 
                  values_fill = 0) %>% 
      transmute(AESOC = AESOC, 
                AEDECOD = stri_trans_totitle(AEDECOD),
                `ARM A_1` = fmt_cnt_pct(`ARM A_1`, arm_pop["ARM A"]),
                `ARM A_2` = fmt_cnt_pct(`ARM A_2`, arm_pop["ARM A"]),
                `ARM A_3` = fmt_cnt_pct(0, arm_pop["ARM A"]),
                `ARM B_1` = fmt_cnt_pct(`ARM B_1`, arm_pop["ARM B"]),
                `ARM B_2` = fmt_cnt_pct(`ARM B_2`, arm_pop["ARM B"]),
                `ARM B_3` = fmt_cnt_pct(0, arm_pop["ARM B"]),
                `ARM C_1` = fmt_cnt_pct(`ARM C_1`, arm_pop["ARM C"]),
                `ARM C_2` = fmt_cnt_pct(`ARM C_2`, arm_pop["ARM C"]),
                `ARM C_3` = fmt_cnt_pct(0, arm_pop["ARM C"]),
                `ARM D_1` = fmt_cnt_pct(`ARM D_1`, arm_pop["ARM D"]), 
                `ARM D_2` = fmt_cnt_pct(`ARM D_2`, arm_pop["ARM D"]), 
                `ARM D_3` = fmt_cnt_pct(`ARM D_3`, arm_pop["ARM D"])) %>% 
      ungroup() 
    
    
    # Get counts and percents for All Adverse Events
    df2 <- df_sub %>% 
      select(-AESEV, -AEREL,-AESOC, -AEDECOD,) %>% 
      group_by(TRTA, AESEVN) %>% 
      summarize(cnt = n()) %>% 
      pivot_wider(names_from = c(TRTA, AESEVN),
                  values_from = cnt, 
                  values_fill = 0) %>% 
      ungroup() 
    
    col_template <- paste0(c(rep("ARM A_", 3), rep("ARM B_", 3), rep("ARM C_", 3),
                             rep("ARM D_", 3)), rep(c(1, 2, 3), 3))
    
    for (nm in col_template) {
      if (!nm %in% names(df2))
        df2[[nm]] <- 0
    }
    
    df2 <- df2 %>% 
      transmute(AESOC = "All System Organ Classes",
                AEDECOD = "All Adverse Events", 
                `ARM A_1` = fmt_cnt_pct(`ARM A_1`, arm_pop["ARM A"]),
                `ARM A_2` = fmt_cnt_pct(`ARM A_2`, arm_pop["ARM A"]),
                `ARM A_3` = fmt_cnt_pct(`ARM A_3`, arm_pop["ARM A"]),
                `ARM B_1` = fmt_cnt_pct(`ARM B_1`, arm_pop["ARM B"]),
                `ARM B_2` = fmt_cnt_pct(`ARM B_2`, arm_pop["ARM B"]),
                `ARM B_3` = fmt_cnt_pct(`ARM B_3`, arm_pop["ARM B"]),
                `ARM C_1` = fmt_cnt_pct(`ARM C_1`, arm_pop["ARM C"]),
                `ARM C_2` = fmt_cnt_pct(`ARM C_2`, arm_pop["ARM C"]),
                `ARM C_3` = fmt_cnt_pct(`ARM C_3`, arm_pop["ARM C"]),
                `ARM D_1` = fmt_cnt_pct(`ARM D_1`, arm_pop["ARM D"]), 
                `ARM D_2` = fmt_cnt_pct(`ARM D_2`, arm_pop["ARM D"]), 
                `ARM D_3` = fmt_cnt_pct(`ARM D_3`, arm_pop["ARM D"]))
    
    
    final <- bind_rows(df2, df1) %>% arrange(AESOC, AEDECOD)
    
    tbl <- create_table(final, first_row_blank = TRUE) %>% 
      column_defaults(from = `ARM A_1`, to = `ARM D_3`, width = 1) %>% 
      spanning_header("ARM A_1", "ARM A_3", label = "ARM A", n = arm_pop["ARM A"]) %>%
      spanning_header("ARM B_1", "ARM B_3", label = "ARM B", n = arm_pop["ARM B"]) %>%
      spanning_header("ARM C_1", "ARM C_3", label = "ARM C", n = arm_pop["ARM C"]) %>%
      spanning_header("ARM D_1", "ARM D_3", label = "ARM D", n = arm_pop["ARM D"]) %>%
      stub(vars = c("AESOC", "AEDECOD"), label = "System Organ Class\n   Preferred Term", width = 5) %>% 
      define(AESOC, blank_after = TRUE, label_row = TRUE, group_cohesion = 1, break_label = "Cont.") %>% 
      define(AEDECOD, indent = .25) %>% 
      define(`ARM A_1`, align = "center", label = "Mild") %>% 
      define(`ARM A_2`, align = "center", label = "Mod**") %>% 
      define(`ARM A_3`, align = "center", label = "Severe") %>% 
      define(`ARM B_1`, align = "center", label = "Mild", page_wrap = TRUE) %>% 
      define(`ARM B_2`, align = "center", label = "Mod**") %>% 
      define(`ARM B_3`, align = "center", label = "Severe") %>% 
      define(`ARM C_1`, align = "center", label = "Mild", page_wrap = TRUE) %>% 
      define(`ARM C_2`, align = "center", label = "Mod**") %>% 
      define(`ARM C_3`, align = "center", label = "Severe") %>% 
      define(`ARM D_1`, align = "center", label = "Mild", page_wrap = TRUE) %>% 
      define(`ARM D_2`, align = "center", label = "Mod**") %>% 
      define(`ARM D_3`, align = "center", label = "Severe") 
    
    rpt <- create_report(fp) %>% 
      options_fixed(font_size = 10) %>% 
      page_header("Client: Experis", "Study: BBC") %>% 
      titles("Table 1.0", "Adverse Events by Severity", "Safety Population < 25") %>% 
      add_content(tbl) %>% 
      footnotes(paste("Date Produced:", "Time", ";  Program: Table3_0.R"),
                paste("* Total Reporting is defined as number of subjects",
                      "who reported at least one adverse event."),
                "** Mod = Moderate",
                paste("# Episodes is defined as the total number of occurances",
                      "of adverse events"),
                paste("% is defined as Number of Subjects divided by Total Reporting"),
                "Note: Adverse events were coded using MedDRA Version 9.1",
                font_size = 9) %>%
      page_footer("Time", "Confidential", "Page [pg] of [tpg]")
    
    res <- write_report(rpt)
  
  
    expect_equal(file.exists(fp), TRUE)
  
    lns <- readLines(fp)
  
    expect_equal(length(lns), res$pages * res$line_count)
    
    
    rtfpth <- file.path(base_path, "user/user4.rtf")
    res <- write_report(rpt, file_path = rtfpth, output_type = "RTF")
    
    rtfpth <- file.path(base_path, "user/user4.rtf")
    res <- write_report(rpt, rtfpth, output_type = "RTF")
    expect_equal(file.exists(rtfpth), TRUE)
    #print(res)

    pdfpth <- file.path(base_path, "user/user4.pdf")
    res <- write_report(rpt, pdfpth, output_type = "PDF")
    expect_equal(file.exists(pdfpth), TRUE)
    
    docxpth <- file.path(base_path, "user/user4.docx")
    res <- write_report(rpt, docxpth, output_type = "DOCX")
    expect_equal(file.exists(docxpth), TRUE)
      #print(res)
    
    htmlpth <- file.path(base_path, "user/user4.html")
    res <- write_report(rpt, htmlpth, output_type = "HTML")
    expect_equal(file.exists(htmlpth), TRUE)
    
  } else
    expect_equal(TRUE, TRUE)
})



test_that("user5: large listing works.", {
  
  test <- FALSE
  
  # Skip except for special testing because it takes too long (+ 5 minutes)
  if (test) { 
    
    startTime <- Sys.time()  
  
    # Data Filepath
    dir_data <- file.path(base_path, "data")
    
    fp <- file.path(base_path, "user/user5.out")
    
    
    # Load Data
    data_lb   <- file.path(dir_data, "ADLB.csv") %>%
      read.csv() 
  
  
    # Define table
    tbl <- create_table(data_lb) %>% 
      define(USUBJID, id_var = TRUE) 
    
    # Define Report
    rpt <- create_report(fp) %>%
      options_fixed(font_size = 10) %>% 
      titles("Listing 2.0",
             "Analysis Dataset Labs") %>%
      add_content(tbl, align = "left") %>% 
      page_footer(left = "Time", right = "Page [pg] of [tpg]")
    
    # Write out report
    res <- write_report(rpt)
    
    endTime <- Sys.time()
    
    print(endTime - startTime)
    
    expect_equal(file.exists(fp), TRUE)
    
    lns <- readLines(fp)
  
    expect_equal(length(lns), res$pages * res$line_count)
    
    rtfpth <- file.path(base_path, "user/user5.rtf")
    write_report(rpt, rtfpth, output_type = "RTF")
    expect_equal(file.exists(rtfpth), TRUE)
    
    # Very special testing case.
    if (FALSE) {
      pdfpth <- file.path(base_path, "user/user5.pdf")
      write_report(rpt, pdfpth, output_type = "PDF")
      expect_equal(file.exists(pdfpth), TRUE)
    }
  
  } else {
    
   expect_equal(TRUE, TRUE) 
  }
  
})



test_that("user6: listings with page break works as expected.", {
  
  if (dev) {
    # Data Filepath
    dir_data <- file.path(data_dir, "data")
  
    
    # Load Data
    data_demo   <- file.path(dir_data, "dm.csv") %>%
      read.csv() 
    
    data_demo <- data_demo[order(data_demo$ARMCD), ]
    
    # Define table
    tbl <- create_table(data_demo) %>% 
      define(USUBJID, id_var = TRUE) %>% 
      define(ARMCD, page_break = TRUE, id_var = TRUE) %>% 
      define(ARM, id_var = TRUE)
    
    fp <- file.path(base_path, "user/user6")
    
    
    # Define Report
    rpt <- create_report(fp) %>%
      options_fixed(editor = "notepad") %>% 
      titles("Listing 1.0",
             "Demographics Dataset") %>%
      add_content(tbl, align = "left") %>% 
      page_footer(left = "Time", right = "Page [pg] of [tpg]")
    
    # Write out report
    res <- write_report(rpt)
    
    expect_equal(file.exists(res$modified_path), TRUE)
    
    lns <- readLines(res$modified_path)
    
    expect_equal(length(lns), res$pages * res$line_count)
    
    res <- write_report(rpt, output_type = "RTF")
    expect_equal(file.exists(res$modified_path), TRUE)
    
    res <- write_report(rpt, output_type = "PDF")
    expect_equal(file.exists(res$modified_path), TRUE)
    
  
  } else
    expect_equal(TRUE, TRUE)
  
})




test_that("user7: listings with NA values works.", {
  
  if (dev) {
    library(readr)
    
    # Data Filepath
    dir_data <- file.path(data_dir, "data")
    
    fp <- file.path(base_path, "user/user7")
    
    # Load Data
    dat <- file.path(dir_data, "ADSL.csv") %>%
      read_csv() 
    
    # Define table
    tbl <- create_table(dat) %>% 
      define(USUBJID, id_var = TRUE) 
    
    # Define Report
    rpt <- create_report(fp) %>%
      titles("Listing 1.0",
             "ADSL Dataset") %>%
      add_content(tbl, align = "left") %>% 
      page_footer(left = "Time", right = "Page [pg] of [tpg]")
    
    # Write out report
    res <- write_report(rpt)
    
    expect_equal(file.exists(res$modified_path), TRUE)
    
    lns <- readLines(res$modified_path)
    
    expect_equal(length(lns), res$pages * res$line_count)
    
  
    res <- write_report(rpt, output_type = "RTF")
    expect_equal(file.exists(res$modified_path), TRUE)
  
  } else
    expect_equal(TRUE, TRUE)
  
})

# This is a special case, and I don't want to test it every time.
# One reason is to not create a dependancy on > R 3.5 just because
# of the .rds test file. The .rds test file has been appended with 
# a .txt extension to fool the checker. Turn back to .rds if needed.
# test_that("user8: table with spaces in column names works.", {
#   
#   fp <- file.path(base_path, "user/user8.out")
#   
#   dat <- readRDS(file.path(base_path, "./data/dm_final.rds"))
# 
#   tbl <- create_table(dat)  
#   
#   rpt <- create_report(fp) %>% 
#     add_content(tbl)
#   
#   res <- write_report(rpt)
#   
#   expect_equal(file.exists(fp), TRUE)
#   
# })

# Also a special case
# test_that("user9: table with stub and page by works as expected.", {
#   
#   # Data Filepath
#   dir_data <- file.path(base_path, "data")
#   
#   fp <- file.path(base_path, "user/user9")
#   
#   # Load Data
#   dat <- file.path(dir_data, "final.rds") %>% readRDS() 
#   
#   dat <- subset(dat, dat$label != "Q1 - Q3")
#   
#   arm_pop <- c("ARM A" = 20, "ARM B" = 21, "ARM C" = 19, "ARM D" = 22) 
#   
#   # Create Table
#   tbl <- create_table(dat, width = 9) %>% 
#     column_defaults(from = `ARM A`, to = `ARM D`, align = "center", width = 1.2) %>% 
#     page_by(AVISIT, label = "Visit: ", blank_row = "none") %>% 
#     stub(vars = c(PARAM, label), label  = "Parameter", width = 2) %>% 
#     define(AVISIT, visible = FALSE) %>% 
#     define(PARAMCD, visible = FALSE) %>% 
#     define(PARAM, label_row = TRUE, blank_after = TRUE) %>% 
#     define(label, label = "Statistic", indent = .25) %>% 
#     define(`ARM A`,  n = arm_pop["ARM A"]) %>% 
#     define(`ARM B`,  n = arm_pop["ARM B"]) %>% 
#     define(`ARM C`,  n = arm_pop["ARM C"]) %>% 
#     define(`ARM D`,  n = arm_pop["ARM D"]) 
#   
#   
#   rpt <- create_report(fp, output_type = "RTF") %>% 
#     set_margins(top = 1, bottom = .9) %>% 
#     page_header("Sponsor: Experis", "Study: ABC") %>% 
#     titles("Table 3.0", "Summary of Vital Sign Parameters by Visit", 
#            "Safety Population") %>% 
#     add_content(tbl) %>% 
#     footnotes("R Program: VA_Table.R") %>% 
#     page_footer(paste0("Date Produced: ", fapply(Sys.time(), "%d%b%y %H:%M")), 
#                 right = "Page [pg] of [tpg]")
#   
#   res <- write_report(rpt) 
#   
#   expect_equal(file.exists(res$modified_path), TRUE)
#   
# })

# Another special case
# test_that("user10: Combined report works as expected", {
#   
#   
#   dm_table <- readRDS(file.path(base_path, "data/DM_Table.rds"))
#   ae_table <- readRDS(file.path(base_path, "data/AE_Table.rds"))
#   lb_table <- readRDS(file.path(base_path, "data/LB_Table.rds"))
#   vs_table <- readRDS(file.path(base_path, "data/VS_Table.rds"))
#   
#   
#   rpt <- create_report("rtf/Combined_Table", output_type = "RTF") %>% 
#     set_margins(top = 1, bottom = .9) %>% 
#     page_header("Sponsor: Experis", "Study: ABC") %>% 
#     add_content(dm_table) %>% 
#     add_content(ae_table) %>% 
#     add_content(vs_table) %>% 
#     add_content(lb_table) %>% 
#     page_footer(paste0("Date Produced: ", fapply(Sys.time(), "%d%b%y %H:%M")), 
#                 right = "Page [pg] of [tpg]")
#   
#   
#   res <- write_report(rpt) 
#   
#   
#   
# })

# Special Case
# test_that("mismatched format works as expected.", {
#   
#   
#   fp <- file.path(base_path, "data/final.rds")
#   
#   final <- readRDS(fp)
#   
#   var_fmt <- c("AGE" = "Age", "AGEGR1" = "Age Group", "SEX" = "Sex", "RACE" = "Race")
#   
#   arm_pop <- c("ARM A" = 20, "ARM B" = 21,  "ARM C" = 21, "ARM D" = 23) 
#   
#   # Create Table
#   tbl <- create_table(final, first_row_blank = TRUE, width = 9) %>% 
#     column_defaults(from = `ARM A`, to = `ARM D`, align = "center", width = 1) %>% 
#     stub(vars = c("var", "label"), "Variable", width = 2.5) %>% 
#     define(var, blank_after = TRUE, dedupe = TRUE, label = "Variable",
#            format = var_fmt,label_row = TRUE) %>% 
#     define(label, indent = .25, label = "Demographic Category") %>% 
#     define(`ARM A`,  n = arm_pop["ARM A"]) %>% 
#     define(`ARM B`,  n = arm_pop["ARM B"]) %>% 
#     define(`ARM C`,  n = arm_pop["ARM C"]) %>% 
#     define(`ARM D`,  n = arm_pop["ARM D"]) %>% 
#     define(stat, label = "Tests of Association*\nValue (P-value)",  
#            width = 2, dedupe = TRUE, align = "center") %>% 
#     titles("Table 1.0", "Analysis of Demographic Characteristics", 
#            "Safety Population") %>% 
#     footnotes("R Program: Table1_0.R",
#               "NOTE: Denominator based on number of non-missing responses.",
#               "*Pearsons's Chi-Square tests will be used for categorical variables",
#               "   and ANOVA tests for continuous variables.") 
#   
# 
#   
#   rpt <- create_report("output/DM_Table2.rtf", output_type = "RTF") %>% 
#     set_margins(top = 1, bottom = 1) %>% 
#     page_header("Sponsor: Experis", "Study: ABC") %>% 
#     add_content(tbl) %>% 
#     page_footer(paste0("Date Produced: ", fapply(Sys.time(), "%d%b%y %H:%M")), 
#                 right = "Page [pg] of [tpg]")
#   
#   write_report(rpt) 
#   
#   
# })



test_that("user12: Complex table works as expected.", {
  
  if (dev) {
    library(readr)
    library(fmtr)
    
    # Data Filepath
    dir_data <- file.path(data_dir, "data")
    
    fp <- file.path(base_path, "user/user12")
    
    # Load Data
    dat <- file.path(dir_data, "kk.csv") %>%
      read_csv() 
    
    
    mfmt <- value(condition(is.na(x), "-"),
                  condition(x == 0, "%d"),
                  condition(TRUE, "%5.3f"))
    
    nfmt <- value(condition(is.na(x), " "), 
                  condition(TRUE, "%2d"))
    
    pfmt <- value(condition(is.na(x), " "),
                  condition(TRUE, "(%4.1f)"))
    
    
    # Define table
    tbl <- create_table(dat, width = 9) %>% 
      spanning_header(COL1, COL2, "LLY10", n = 70, underline = FALSE, label_align = "center") %>% 
      spanning_header(COL3, COL4, "LLY20", n = 49, underline = FALSE, label_align = "center") %>% 
      spanning_header(COL5, COL6, "LLY75", n = 19, underline = FALSE, label_align = "center") %>% 
      spanning_header(COL8, COL10, "Pairwise p-values*b") %>% 
      spanning_header(COL11, COL11, "Odds\nratios*c") %>% 
      column_defaults(from = COL1, to = COL11, align = "center", width = .45) %>% 
      stub(c(CATEGORY, LABEL), width = 3.25) %>% 
      define(CATEGORY, label_row = TRUE) %>% 
      define(LABEL, indent = .25) %>% 
      define(COL1, label = "n") %>% 
      define(COL2, label = "(%)", format = pfmt) %>% 
      define(COL3, label = "n") %>% 
      define(COL4, label = "(%)", format = pfmt) %>% 
      define(COL5, label = "n") %>% 
      define(COL6, label = "(%)", format = pfmt) %>% 
      define(COL7, label = "Overall\np-value*a", format = mfmt, width = .7) %>% 
      define(COL8, label = "LLY20\nvs\nLLY10", format = mfmt, width = 0.48) %>% 
      define(COL9, label = "LLY75\nvs\nLLY10", format = mfmt, width = 0.48) %>% 
      define(COL10, label = "LLY75\nvs\nLLY20", format = mfmt, width = 0.48) %>% 
      define(COL11, label = "LLY75/\nLLY20", width = .8, format = mfmt) %>% 
      titles("Testing Odds, Pairwise and Overall with 3 Trts",
             "SAMPLE TEXT FOR title5", align = "left", borders = "bottom", blank_row = "none") %>%
      footnotes("Abbreviations: N = number of subjects in population; n = number of subjects within category.",
                "LLY10=LILLY_DRUG_10_mg; LLY20=LILLY_DRUG_20_mg; LLY75=LILLY_DRUG_75_mg.",
                "*a - p-value for overall treatment effect were computed using Fisher's Exact test.",
                "*b - p-values for pairwise treatment comparisons were computed using Chi-Square test.",
                "*c - odds ratios based on comparator LILLY_DRUG_20_mg as denominator.",
                "Program Location: /lillyce/qa/vct/common/rums/taffy_rums/dev_r/c_ds_dispsum/c_ds_dispsum_4.R",
                "Output Location: /lillyce/qa/vct/common/rums/taffy_rums/dev_r/c_ds_dispsum/validation/output/odds_pair_overall_h_test.docx",
                "Data Set Location: /lillyce/qa/vct/common/rums/taffy_rums/data/multi arms",
               borders = "top", blank_row = "none")
    
    # Define Report
    rpt <- create_report(fp, output_type = "RTF") %>%
      options_fixed(font_size = 8) %>% 
      set_margins(top = 1, bottom = .5, left = 1, right = 1) %>% 
      page_header(right = c("Page [pg] of [tpg]",
                            format(Sys.time(), "%H:%M %d-%b-%Y"),
                            "DDDL"
                            )) %>%
      add_content(tbl, align = "left") 
    
    # Write out report
    res <- write_report(rpt)
    res
    
    #file.show(res$modified_path)
    
    expect_equal(file.exists(res$modified_path), TRUE)

    
  } else
    expect_equal(TRUE, TRUE)
  
})

# Also testing alignments and line lengths
test_that("user13: Simple demographic report with 12 pt font wraps as expected.", {
  

  
  # Create temporary path
  fp <- file.path(base_path, "user/user13")
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE) %>% 
    stub(c("var", "label")) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center", n = 19) %>% 
    define(B, label = "Group B", align = "center", n = 13)
  
  
  # Create report and add content
  rpt <- create_report(fp, output_type = "RTF", units = "inches") %>% 
    set_margins(top = 1, bottom = 1) %>%
    options_fixed(font_size = 12) %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Test Title Right Aligned", align = "right") %>% 
    titles("Table 1.0", "MTCARS Summary Table", 
           "----------------------------------------", borders = "all", 
           align = "center") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974", 
              "----------------------------------------", 
              borders = "all", align = "right") %>%
    page_footer(left = Sys.time(), 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  # Write out report
  res <-  write_report(rpt)
  
  # View report
  # file.show(fp)
  
  expect_equal(file.exists(res$modified_path), TRUE)
  
})
  

# Also testing alignments and line lengths
test_that("user14: Title header alignment works as expected.", {
  

  # Create temporary path
  fp <- file.path(base_path, "user/user14")
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE) %>% 
    stub(c("var", "label")) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center", n = 19) %>% 
    define(B, label = "Group B", align = "center", n = 13)
  
  
  # Create report and add content
  rpt <- create_report(fp, output_type = "RTF", units = "inches") %>% 
    set_margins(top = 1, bottom = 1) %>%
    options_fixed(font_size = 12) %>% 
    title_header("Test Title Left Aligned", right = "Testme", 
                 borders = "all") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974", 
              borders = "all", align = "right") %>%
    page_footer(left = Sys.time(), 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  # Write out report
  res <-  write_report(rpt)
  
  # View report
  # file.show(fp)
  
  expect_equal(file.exists(res$modified_path), TRUE)
  
})


test_that("user15: Titles and footnotes only on first last page.", {
  
  # Create temporary path
  fp <- file.path(base_path, "user/user15")

  ttls <- create_text("  ") %>%
    titles("My Title Only on First Page", blank_row = "none")
  
  tbl <- create_table(iris)
  
  ftnts <- create_text("  ") %>%
    footnotes("My Footnote Only on Last Page",
              "Here is another footnote")
  
  rpt <- create_report(fp, font = "Arial", output_type = "RTF") %>%
    add_content(ttls, page_break = FALSE, blank_row = "none")  %>% 
    add_content(tbl, page_break = FALSE, blank_row = "none") %>%
    add_content(ftnts, page_break = FALSE, blank_row = "none")
  
  res <-  write_report(rpt)
  
  expect_equal(file.exists(res$modified_path), TRUE)
  
})

test_that("user16: Label row does not create extra blank spaces.", {
  
  library(libr)
  library(fmtr)
  
  # Create temporary path
  fp <- file.path(base_path, "user/user16")
  
  adsl <- read.table(header = TRUE, text = '
    SITEID SITENAME         USUBJID AGE SEX RACE WEIGHT 
    10003  "Advance ENT" 1000310001 22 M "WHITE" 74.843827
    10003  "Advance ENT" 1000310002 56 F "ASIAN" 60.312484')
  
  adsl
  
  
  adot <- read.table(header = TRUE, text = '
  USUBJID    COHORT      TREATMENT                DOSE  DOSENUM  DOSEDATE   EAR   VISDAY VISDATE      ANAVISIT OBSRESP IMPRESP
  1000310001 "COHORT 1"  "Single Dose Escalation" 0.03  1      "2020-09-29" L     2     "2020-09-30" "Day 2"   No       NA
  1000310001 "COHORT 1"  "Single Dose Escalation" 0.03  1      "2020-09-29" L     4     "2020-10-03" "Day 3"   No       NA  
  1000310001 "COHORT 1"  "Single Dose Escalation" 0.03  1      "2020-09-29" L     8     "2020-10-06" "Week 1"  No       NA 
  1000310001 "COHORT 1"  "Single Dose Escalation" 0.03  1      "2020-09-29" L     14    "2020-10-12" "Week 2"  No       NA 
  1000310001 "COHORT 1"  "Single Dose Escalation" 0.03  1      "2020-09-29" L     28    "2020-10-26" "Week 4"  No       NA
  1000310001 "COHORT 1"  "Single Dose Escalation" 0.03  1      "2020-09-29" L     57    "2020-11-24" "Week 8"  No       No 
  1000310002 "COHORT 1"  "Single Dose Escalation" 0.03  1      "2020-09-29" R     2     "2020-09-30" "Day 2"   No       NA
  1000310002 "COHORT 1"  "Single Dose Escalation" 0.03  1      "2020-09-29" R     4     "2020-10-03" "Day 3"   No       NA  
  1000310002 "COHORT 1"  "Single Dose Escalation" 0.03  1      "2020-09-29" R     8     "2020-10-06" "Week 1"  No       NA 
  1000310002 "COHORT 1"  "Single Dose Escalation" 0.03  1      "2020-09-29" R     14    "2020-10-12" "Week 2"  No       NA 
  1000310002 "COHORT 1"  "Single Dose Escalation" 0.03  1      "2020-09-29" R     28    "2020-10-26" "Week 4"  Yes      NA 
  1000310002 "COHORT 1"  "Single Dose Escalation" 0.03  1      "2020-09-29" R     57    "2020-11-24" "Week 8"  Yes      NA 
                   ')
  
  adot
  
  

  fc <- fcat(SEX = c("M" = "Male", "F" = "Female"),
             AGE = "%d Years",
             RACE = value(condition(x == "WHITE", "White"),
                          condition(x == "BLACK", "Black or African American"),
                          condition(x == "ASIAN", "Asian or Pacific Islander"),
                          condition(TRUE, "Other")),
             WEIGHT = "%6.2f kg",
             EAR = c("L" = "Left", "R" = "Right"),
             DOSE = "%4.2fug")
  
  
  # Merge and assign formats
  acomb <- 
    datastep(adsl, format = fc,
             merge = adot, merge_by = USUBJID, {})
  
  # Apply formats
  acombf <- fdata(acomb) 
  
  acombf
  
  # Prepare final data for reporting
  final <- 
    datastep(acombf, by = USUBJID,
             {
               
               BASELINE <- paste0("Investigator Site = ", SITEID,
                                 " - ", SITENAME, ",\nSubject ID=", USUBJID,
                                ",Age=", AGE, ",Sex=", SEX, ",Race=", RACE,
                                  ",Weight=", WEIGHT)
               
               if (first.) {
                 GROUP <- paste0(COHORT, " - ", TREATMENT, " - ", DOSE)
                 DOSELBL <- paste0(DOSENUM, "/\n", DOSEDATE, "/\n", EAR)
               } else {
                 
                 GROUP <- "" 
               }
               VSTLBL <- paste0(VISDAY, "/", VISDATE)
               
             })
  
  final
  
  tbl <- create_table(final, show_cols = "none", 
                      width = 9, first_row_blank = TRUE) %>% 
    stub(v(BASELINE, GROUP), label = "Cohort") %>% 
    define(BASELINE, label_row = TRUE) %>% 
    define(GROUP) %>% 
    define(DOSELBL, label = "Dose\nDay/\nDate/\nTreated Ear", width = 1.5) %>% 
    define(VSTLBL, label = "Visit\nDay/Date", width = 1.5) %>% 
    define(ANAVISIT, label = "Analysis\nVisit", width = .75) %>% 
    define(OBSRESP, label = "Observed\nResponse", width = .75) %>% 
    define(IMPRESP, label = "Imputed\nResponse", width = .75) %>% 
    define(USUBJID, blank_after = TRUE, visible = FALSE)
  
  rpt <- create_report(fp, font = "Courier", font_size = 9) %>% 
    add_content(tbl) %>% 
    set_margins(top = 1, bottom = 1) %>% 
    page_header("Program:" %p% Sys.path(),  
                right = "Draft", width = 7) %>% 
    titles("Study: 0598-CL-0101", "Appendix 10.2.6.1.2.1", 
           "Source: ADAE, ADOT", columns = 3, header = TRUE, blank_row = "none") %>%
    titles( "TMP Complete Closure Response - Single Ascending Dose (SAD)",
            "All Randomized Patients", align = "center", header = TRUE, blank_row = "below") %>% 
    footnotes("# Time to First Complete Closure of TMP.",
              "Values flagged with '@' were excluded from the by-visit " %p% 
                "analysis in tables showing the qualitative test results.",
              blank_row = "none", footer = TRUE) %>% 
    page_footer("Date: " %p% toupper(fapply(Sys.time(), "%d%b%Y %H:%M:%S")),
                "Astellas", "Page [pg] of [tpg]")
  
  res1 <- write_report(rpt, output_type = "RTF")
  res2 <- write_report(rpt, output_type = "PDF")
  
  
  #file.show(res1$modified_path)
  #file.show(res2$modified_path)
  
  
  expect_equal(file.exists(res1$modified_path), TRUE)
  expect_equal(file.exists(res2$modified_path), TRUE)
  
  
})



test_that("user17: Dedupe works as expected on hemo table", {
  
  if (dev) {
  # Sample Data
  hemo <- read.table(header = TRUE, sep = ",", text = '
labtest,tmtnc,tmtn,swmgrade,variable,c0,c1,c2,c3,c4,c7,pagebrk
HEMO-HIGH,ARM A,1,7,NG,76  98.7%,75  98.7%,0,0,0,1   1.3%,1
HEMO-HIGH,ARM A,1,8,Missing,1   1.3%,1 100.0%,0,0,0,0,1
HEMO-HIGH,ARM A,1,9,Total,77 100.0%,76  98.7%,0,0,0,1   1.3%,1
HEMO-HIGH,ARM B,2,7,NG,84  96.6%,82  97.6%,0,0,0,2   2.4%,1
HEMO-HIGH,ARM B,2,8,Missing,3   3.4%,3 100.0%,0,0,0,0,1
HEMO-HIGH,ARM B,2,9,Total,87 100.0%,85  97.7%,0,0,0,2   2.3%,1
HEMO-HIGH,ARM C,3,7,NG,45 100.0%,41  91.1%,0,0,0,4   8.9%,2
HEMO-HIGH,ARM C,3,8,Missing,0,0,0,0,0,0,2
HEMO-HIGH,ARM C,3,9,Total,45 100.0%,41  91.1%,0,0,0,4   8.9%,2
HEMO-LOW,ARM A,1,1,Grade 0,22  28.6%,2   9.1%,7  31.8%,6  27.3%,7  31.8%,0,1
HEMO-LOW,ARM A,1,3,Grade 1,38  49.4%,0,14  36.8%,14  36.8%,10  26.3%,0,1
HEMO-LOW,ARM A,1,4,Grade 2,13  16.9%,0,1   7.7%,0,11  84.6%,1   7.7%,1
HEMO-LOW,ARM A,1,5,Grade 3,3   3.9%,0,0,1  33.3%,2  66.7%,0,1
HEMO-LOW,ARM A,1,8,Missing,1   1.3%,0,0,1 100.0%,0,0,1
HEMO-LOW,ARM A,1,9,Total,77 100.0%,2   2.6%,22  28.6%,22  28.6%,30  39.0%,1   1.3%,1
HEMO-LOW,ARM B,2,1,Grade 0,23  26.4%,2   8.7%,3  13.0%,11  47.8%,7  30.4%,0,1
HEMO-LOW,ARM B,2,3,Grade 1,41  47.1%,0,4   9.8%,18  43.9%,18  43.9%,1   2.4%,1
HEMO-LOW,ARM B,2,4,Grade 2,17  19.5%,0,1   5.9%,2  11.8%,13  76.5%,1   5.9%,1
HEMO-LOW,ARM B,2,5,Grade 3,3   3.4%,0,0,0,3 100.0%,0,1
HEMO-LOW,ARM B,2,8,Missing,3   3.4%,0,1  33.3%,0,2  66.7%,0,1
HEMO-LOW,ARM B,2,9,Total,87 100.0%,2   2.3%,9  10.3%,31  35.6%,43  49.4%,2   2.3%,1
HEMO-LOW,ARM C,3,1,Grade 0,14  31.1%,2  14.3%,4  28.6%,4  28.6%,2  14.3%,2  14.3%,2
HEMO-LOW,ARM C,3,3,Grade 1,21  46.7%,1   4.8%,5  23.8%,4  19.0%,10  47.6%,1   4.8%,2
HEMO-LOW,ARM C,3,4,Grade 2,6  13.3%,0,0,1  16.7%,4  66.7%,1  16.7%,2
HEMO-LOW,ARM C,3,5,Grade 3,4   8.9%,0,0,1  25.0%,3  75.0%,0,2
HEMO-LOW,ARM C,3,8,Missing,0,0,0,0,0,0,2
HEMO-LOW,ARM C,3,9,Total,45 100.0%,3   6.7%,9  20.0%,10  22.2%,19  42.2%,4   8.9%,2')
  

  library(dplyr)
  library(fmtr)
  
  # Set variables
  program.name <- "t_ctcshift_hem"
  program.output <- "user17"
  program.timestamp <- "2001-01-01 12:00"
  program.dir <- base_path

  
  # Change column names to lower case
  colnames(hemo) <- tolower(colnames(hemo)) 
  

  
  fmt1 <- value(condition(x == "HEMO-HIGH", "Hemoglobin (G/L) - HIGH DIRECTION"),
                condition(x == "HEMO-LOW", "Hemoglobin (g/L) - LOW DIRECTION"),
                condition(x == "LEUK-HIGH", "Leukocytes (GI/L) - HIGH DIRECTION"),
                condition(x == "LEUK-LOW", "Leukocytes (GI/L) - LOW DIRECTION"))
  
  
  fmt2 <- value(condition(x == "ARM A", "Ruxolitinib 15 mg BID (N=77)"),
                condition(x == "ARM B", "Ruxolitinib 5 mg BID (N=87)"),
                condition(x == "ARM C", "Placebo (N=45)"))
  

  ftnts <- list()
  
  base_ftnt <-
    c("[1] The percentages were calculated using the baseline total as the denominator.",                                      
      paste("[2] For each row, the percentages were calculated using the number of participants",
            "with given grade at baseline as the denominator; worst value on study is the worst",
            "grade observed post-baseline for a given participant."))
  
  ftnts[["HEMO-HIGH"]] <- 
    c(base_ftnt,              
      "-  Grade 0 = Below Grade 1 and any grade in the other direction.",                                                     
      "-  For baseline NG means that grade does not apply at baseline.",                                                     
      "-  Grade 1 = Greater than ULN and increase from baseline of  >0 - 2 g/dL;",                                           
      "Grade 2 = Greater than ULN and increase from baseline of  >2 - 4 g/dL;",                                            
      "Grade 3 = Greater than ULN and increase from baseline of  >4 g/dL. ")
  
  
  ftnts[["HEMO-LOW"]] <-   
    c(base_ftnt,           
      "-  Grade 0 = Below Grade 1 and any grade in the other direction.")
  


  pth <- file.path(program.dir, "user", program.output)  
  

  rpt <- create_report(pth, font = "Courier", font_size = 9)  %>% 
    set_margins(top = 1.0, left = 1, right = 1, bottom = .5) %>%
    options_fixed(line_count = 51) %>% 
    
    titles("Table 3.3.3.1", 
           paste("Shift Summary of Hematology Laboratory Values", 
                 "in CTC Grade - to the Worst Abnormal Value"),
           "(Safety Population)", bold = TRUE, font_size = 9) %>%
    page_header(left = c("PROTOCOL: DIDA 00001-123", 
                         "DRUG/INDICATION: DIDA00001/COMPOUND-ASSOCIATED STUDY", 
                         "TLF Version: Final Database Lock (21APR2021)"),
                right = c("Page [pg] of [tpg]",  "DATABASE VERSION: 10MAY2023", 
                          "TASK: Primary Analysis")) %>% 
    footnotes(paste0("Program: ", program.name, sep=""), 
              "DATE(TIME): 2001-12-01",
              blank_row = "none", borders = "top", columns = 2, footer = TRUE)  %>%
    footnotes("Laboratory grading is based on CTCAE Version 5.",
              "Reference: Listing 2.8.1.1, 2.8.1.2", footer = TRUE )
  
  labtests <- names(table(hemo$labtest))
  

  for (i in seq_len(length(labtests))) {
    

    lb <- labtests[i] 
    

    ftnt <- ftnts[[lb]]
    

    table_hemo <- hemo %>% 
      dplyr::filter(labtest==lb) %>%
      mutate(labtest = fapply(labtest, fmt1),
             tmtnc = fapply(tmtnc, fmt2)) %>%
      select(labtest, tmtnc, tmtn, swmgrade,variable, c0,c1,c2,c3,c4,c7, pagebrk) %>% 
      arrange(labtest, tmtn, swmgrade) 
    
    tbl <- create_table(table_hemo,
                        show_cols = c("none"), 
                        borders = "top",
                        width = 9) %>%
      # page_by(labtest, label = "Laboratory Test (unit):", borders = "none", 
      #         blank_row = "none") %>% 
    #  footnotes (ftnt, blank_row ="above" ) %>% 
      column_defaults(width=.1) %>%
      # spanning_header(variable, c0, label="Baseline [1]") %>%
      # spanning_header(c1, c7, label="Worst Post-Baseline Value [2]") %>%
      define(tmtnc, dedupe = TRUE, align = "left", label = "Treatment Group", 
             width=3) %>%
      define(tmtn, blank_after = TRUE, visible = FALSE) %>%
      define(variable, align="left", label="Grade", width=.8) %>%
      define(c0, align="left", label="n  (%)", width=1)# %>%
      #define(c1,   align = "left", label = "Grade 0\n n  (%)", width=1)  %>%
     # define(c2,   align = "left", label = "Grade 1\n n  (%)", width=1) # %>%
     # define(c3,   align = "left", label = "Grade 2\n n  (%)", width=1) # %>%
      #define(c4,   align = "left", label = "Grade 3\n n  (%)", width=1) # %>%
      #define(c7,   align = "left", label = "Missing\n n  (%)", width=1)
    
    rpt <- rpt |> add_content(tbl, blank_row = "none", page_break = TRUE) 
    
  }
  

  res <- write_report(rpt, output_type = "TXT")
  
  expect_equal(file.exists(res$modified_path), TRUE)
  # View the report
  # file.show(res$modified_path)
  # file.show(logpth)
  
  } else {
    
    
    expect_equal(TRUE, TRUE) 
  }
  
})

test_that("user18: Adverse Events table with break labels works.", {
  
  if (dev) {
    library(dplyr)
    library(tidyr)
    
    # Data File path
    dir_data <- file.path(data_dir, "data")
    
    dp <- file.path(dir_data, "ADAE.csv")
    
    dat <- read.csv(dp)
    
    # ------------------------------------------------------- #
    #           Step 1. Prepare subject unique data
    # ------------------------------------------------------- #
    # Subset ADSL for needed rows and columns
    df_pre <- dat %>% 
      select(TRTAN, AESEVN, AESOC, AEDECOD, USUBJID) %>%
      distinct(TRTAN, AESEVN, AESOC, AEDECOD, USUBJID)
    
    df_pre$TRTAN[df_pre$TRTAN == 4] <- 3
    df_pre$total <- "Number of subjects with adverse events"
    
    # Get population counts
    df_dis <- df_pre %>%
      distinct(TRTAN, USUBJID, .keep_all = T)
    arm_pop <- table(df_dis$TRTAN)  
    
    # ------------------------------------------------------- #
    #           Step 2. Prepare data for every level
    # ------------------------------------------------------- #
    
    # SOC data
    df_soc <- df_pre
    df_soc$AEDECOD <- ""
    df_soc$AESEVN <- 0
    df_soc <- df_soc %>%
      distinct(TRTAN, AESOC, USUBJID, .keep_all = T)
    
    df_soc_any <- df_soc
    df_soc_any$AESOC <- "Any system organ class"
    df_soc_any <- df_soc_any %>%
      distinct(TRTAN, AESOC, USUBJID, .keep_all = T)
    
    df_soc <- rbind(df_soc_any, df_soc)
    
    # Preferred Term data
    df_pt <- df_pre
    df_pt$AESEVN <- 0
    df_pt <- df_pt %>%
      distinct(TRTAN, AESOC, AEDECOD, USUBJID, .keep_all = T)
    
    df_pt_any_pt <- df_pt
    df_pt_any_pt$AEDECOD <- "Any preferred term"
    df_pt_any_pt <- df_pt_any_pt %>%
      distinct(TRTAN, AESOC, AEDECOD, USUBJID, .keep_all = T)
    
    df_pt_any_soc <- df_pt
    df_pt_any_soc$AESOC <- "Any system organ class"
    df_pt_any_soc$AEDECOD <- "Any preferred term"
    df_pt_any_soc <- df_pt_any_soc %>%
      distinct(TRTAN, AESOC, AEDECOD, USUBJID, .keep_all = T)
    
    df_pt <- rbind(df_pt_any_soc, df_pt_any_pt, df_pt)
    
    # SEV data
    df_sev <- df_pre
    
    df_sev_any_pt <- df_sev
    df_sev_any_pt$AEDECOD <- "Any preferred term"
    df_sev_any_pt <- df_sev_any_pt %>%
      distinct(TRTAN, AESOC, AEDECOD, AESEVN, USUBJID, .keep_all = T)
    
    df_sev_any_soc <- df_sev
    df_sev_any_soc$AESOC <- "Any system organ class"
    df_sev_any_soc$AEDECOD <- "Any preferred term"
    df_sev_any_soc <- df_sev_any_soc %>%
      distinct(TRTAN, AESOC, AEDECOD, AESEVN, USUBJID, .keep_all = T)
    
    df_sev <- rbind(df_sev_any_soc, df_sev_any_pt, df_sev)
    
    # Sub data
    df_sub <- df_dis
    df_sub$AESOC <- ""
    df_sub$AEDECOD <- ""
    df_sub$AESEVN <- 0
    
    # Stack all
    df_all <- rbind(df_sub, df_soc, df_pt, df_sev)
    
    # ------------------------------------------------------- #
    #           Step 2. Calculate Frequency
    # ------------------------------------------------------- #
    # Set up severity format
    sev_fmt <- value(
      condition(x == 0, ""),
      condition(x == 1, "Mild"),
      condition(x == 2, "Moderate"),
      condition(x == 3, "Severe")
    )
    
    # Get counts and percents
    df_freq <- df_all %>% 
      group_by(TRTAN, total, AESOC, AEDECOD, AESEVN) %>% 
      summarize(cnt = n()) %>% 
      pivot_wider(names_from = c(TRTAN),
                  values_from = cnt, 
                  values_fill = 0,
                  names_prefix = "trt") %>% 
      transmute(total = total,
                AESOC = AESOC, 
                AEDECOD = AEDECOD,
                AESEVN = AESEVN,
                `trt1` = fmt_cnt_pct(`trt1`, arm_pop[1]),
                `trt2` = fmt_cnt_pct(`trt2`, arm_pop[2]),
                `trt3` = fmt_cnt_pct(`trt3`, arm_pop[3])) %>%
      ungroup() 
    
    df_freq$AESEV <- fapply(df_freq$AESEVN, sev_fmt)
    
    # Sorting
    df_freq$AEDECOD_any <- 1
    df_freq$AEDECOD_any[df_freq$AEDECOD == "Any preferred term"] <- 0
    df_freq$AEDECOD_any[df_freq$AEDECOD == ""] <- -1
    
    df_freq <- df_freq[order(df_freq$AESOC, df_freq$AEDECOD_any, df_freq$AEDECOD, df_freq$AESEVN),]
    
    # Final process
    df_freq <- df_freq[, c("total", "AESOC", "AEDECOD", "AESEV", "trt1", "trt2", "trt3")]
    
    df_freq$blank_grp <- paste(df_freq$total, 
                                       df_freq$AESOC, 
                                       ifelse(df_freq$AEDECOD == "", "Any preferred term", df_freq$AEDECOD), 
                                       sep = "|")
    
    for (nm in names(df_freq)) {
      df_freq[[nm]][df_freq[[nm]] == ""] <- NA 
    }
    
    # ------------------------------------------------------- #
    #           Step 3. Output
    # ------------------------------------------------------- #
    custom_n_format <- function(x){
      return(paste0("\n(N = ",x,")\nn(%)"))
    }
    
    # Footnote setting: Output Date and Time
    syslocal <- Sys.setlocale("LC_TIME", "C") # Let date be English
    current_date <- gsub(" ","",toupper(format(Sys.Date(),"%d %b %Y")))
    current_time <- substr(Sys.time(),12,19)
    output_name <- "user18.rtf"
    program_path <- "/dummy/tables/t-teae-soc-pt-gr-saf.R"
    last_footnote <- paste0("Output: ",output_name," (Date Generated: ",current_date,":",current_time,") Source: adae")
    
    # Capable width: 6 inches
    total_width <- 6
    trt_width <- 1.04
    # trt_width <- 1.2
    item_width <- total_width - (trt_width*3)
    
    tbl_1 <- create_table(df_freq, 
                        borders = "outside",
                        n_format = custom_n_format) %>%
      
      # ----- Column setting -----#
      column_defaults(from = trt1, to = trt3, align = "center", width = trt_width) %>%
      define(blank_grp, blank_before = T, visible = FALSE) %>%
      stub(vars = c("total", "AESOC", "AEDECOD", "AESEV"), 
           label = "System Organ Class\n  Preferred Term\n    Grade", 
           width = item_width) %>%
      define(AESOC, break_label = "(Continued)") %>%
      # define(AESOC, break_label = "(This sentence is very long so that the break label will take at least two lines and the page is still good.)") %>%
      # define(AESOC, break_label = TRUE) %>%
      define(AEDECOD, indent = 0.16, break_label = "(Cont.)") %>%
      define(AESEV, indent = 0.32) %>%
      define(trt1, label = "Treatment A", n=arm_pop[1]) %>%
      define(trt2, label = "Treatment B", n=arm_pop[2]) %>%
      define(trt3, label = "Treatment C", n=arm_pop[3])

      # # ----- Footnote setting -----#
      # footnotes("Page [pg] of [tpg]", align = "right", blank_row = "none", valign = "top") %>%
      # footnotes("Safety Analysis Set includes all subjects who are received at least one dose of study drug.",
      #           align = "left", blank_row = "none", valign = "top") %>%
      # footnotes("The analysis is performed on the maximum severity grade.",
      #           align = "left", blank_row = "none", valign = "top") %>%
      # footnotes("Coded using MedDRA version 22.1.",
      #           align = "left", blank_row = "none", valign = "top") %>%
      # footnotes("Severity of adverse event will be graded using CTCAE version 4.0 criteria.",
      #           align = "left", blank_row = "none", valign = "top") %>%
      # footnotes("{supsc('*')} Reported Verbatim Term are displayed for Uncoded AEs.",
      #           align = "left", blank_row = "none", valign = "top") %>%
      # footnotes(paste0("Program: ", program_path), italics = TRUE) %>%
      # footnotes(last_footnote, blank_row = "none", italics = TRUE ) 
    
    set_footnote <- function(tbl, program_path, last_footnote, font_size = 9, 
                             valign = "top", footer = FALSE){
      
      ret <- tbl %>%
        # ----- Footnote setting -----#
        footnotes("Page [pg] of [tpg]", align = "right", font_size = font_size,
                  blank_row = "none", valign = valign, footer = footer) %>%
        footnotes("Safety Analysis Set includes all subjects who are received at least one dose of study drug.",
                  align = "left", blank_row = "none", valign = valign,
                  font_size = font_size, footer = footer) %>%
        footnotes("The analysis is performed on the maximum severity grade.",
                  align = "left", blank_row = "none", valign = valign,
                  font_size = font_size, footer = footer) %>%
        footnotes("Coded using MedDRA version 22.1.",
                  align = "left", blank_row = "none", valign = valign,
                  font_size = font_size, footer = footer) %>%
        footnotes("Severity of adverse event will be graded using CTCAE version 4.0 criteria.",
                  align = "left", blank_row = "none", valign = valign,
                  font_size = font_size, footer = footer) %>%
        footnotes("{supsc('*')} Reported Verbatim Term are displayed for Uncoded AEs.",
                  align = "left", blank_row = "none", valign = valign,
                  font_size = font_size, footer = footer) %>%
        footnotes(paste0("Program: ", program_path), italics = TRUE,
                  font_size = font_size, footer = footer) %>%
        footnotes(last_footnote, blank_row = "none", italics = TRUE,
                  font_size = font_size, footer = footer) 
      return(ret)
      
    }
    
    tbl <- set_footnote(tbl_1, program_path, last_footnote)
    
    # Define custom style
    sty <- create_style(font_name = "Arial",
                        font_size = 9)
    
    # Create report and add content
    
    # -------------------------------------------- #
    #                   RTF                        #
    # -------------------------------------------- #
    fp <- file.path(base_path, "user/user18.rtf")
    rpt <- create_report(fp, 
                         orientation = "portrait", 
                         output_type = "RTF")  %>%
      add_style(style = sty) %>%
      set_margins(top = 1, bottom = 0.75, right = 1, left = 1.5) %>%
      add_content(tbl) %>%
      
      # ----- Title setting -----#
      titles("Table 14-6.1.3.1.  Treatment-emergent Adverse Events by System Organ Class,",
             "Preferred Term and Grade",
             "(Safety Analysis Set)",
             bold = T,
             font_size = 11)
    
    # Write the report
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
    
    # -------------------------------------------- #
    #                   PDF                        #
    # -------------------------------------------- #
    fp <- file.path(base_path, "user/user18.pdf")
    rpt <- create_report(fp, 
                         orientation = "portrait", 
                         output_type = "PDF")  %>%
      add_style(style = sty) %>%
      set_margins(top = 1, bottom = 0.75, right = 1, left = 1.5) %>%
      add_content(tbl) %>%
      
      # ----- Title setting -----#
      titles("Table 14-6.1.3.1.  Treatment-emergent Adverse Events by System Organ Class,",
             "Preferred Term and Grade",
             "(Safety Analysis Set)",
             bold = T,
             font_size = 11) %>%
      footnotes("All the data is from dummy database.")
    
    # Write the report
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
    
    # -------------------------------------------- #
    #             DOCX (Title in report)           #
    # -------------------------------------------- #
    fp <- file.path(base_path, "user/user18.docx")
    rpt <- create_report(fp, 
                         font = "Arial",
                         font_size = 9,
                         orientation = "portrait", 
                         output_type = "DOCX")  %>%
      set_margins(top = 1, bottom = 0.75, right = 1, left = 1.5) %>%
      add_content(tbl) %>%
      
      # ----- Title setting -----#
      titles("Table 14-6.1.3.1.  Treatment-emergent Adverse Events by System Organ Class,",
             "Preferred Term and Grade",
             "(Safety Analysis Set)",
             bold = T,
             font_size = 11)
    
    # Write the report
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
    
    # -------------------------------------------- #
    #             DOCX (Title in table)            #
    # -------------------------------------------- #
    fp <- file.path(base_path, "user/user18_2.docx")
    tbl_2 <- tbl %>%
      # ----- Title setting -----#
      titles("Table 14-6.1.3.1.  Treatment-emergent Adverse Events by System Organ Class,",
             "Preferred Term and Grade",
             "(Safety Analysis Set)",
             bold = T,
             font_size = 11)
      
    rpt <- create_report(fp, 
                         font = "Arial",
                         font_size = 9,
                         orientation = "portrait", 
                         output_type = "DOCX")  %>%
      set_margins(top = 1, bottom = 0.75, right = 1, left = 1.5) %>%
      add_content(tbl_2)
    
    # Write the report
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
    
    # -------------------------------------------- #
    #             DOCX (Title in header)           #
    # -------------------------------------------- #
    fp <- file.path(base_path, "user/user18_3.docx")
    rpt <- create_report(fp, 
                         font = "Arial",
                         font_size = 9,
                         orientation = "portrait", 
                         output_type = "DOCX")  %>%
      set_margins(top = 1, bottom = 0.75, right = 1, left = 1.5) %>%
      add_content(tbl) %>%
      
      # ----- Title setting -----#
      titles("Table 14-6.1.3.1.  Treatment-emergent Adverse Events by System Organ Class,",
             "Preferred Term and Grade",
             "(Safety Analysis Set)",
             bold = T,
             font_size = 11,
             header = TRUE)
    
    # Write the report
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
    
    # -------------------------------------------- #
    #        DOCX (bigger footnote in report)      #
    # -------------------------------------------- #
    fp <- file.path(base_path, "user/user18_4.docx")

    rpt <- create_report(fp, 
                         font = "Arial",
                         font_size = 9,
                         orientation = "portrait", 
                         output_type = "DOCX")  %>%
      set_margins(top = 1, bottom = 0.75, right = 1, left = 1.5) %>%
      add_content(tbl_1) %>%
      
      # ----- Title setting -----#
      titles("Table 14-6.1.3.1.  Treatment-emergent Adverse Events by System Organ Class,",
             "Preferred Term and Grade",
             "(Safety Analysis Set)",
             bold = T,
             font_size = 11)
    
    rpt <- set_footnote(rpt, program_path, last_footnote, font_size = 10, valign = "bottom")
    
    # Write the report
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
    
    # -------------------------------------------- #
    #        DOCX (bigger footnote in table)       #
    # -------------------------------------------- #
    fp <- file.path(base_path, "user/user18_5.docx")
    tbl_3 <- set_footnote(tbl_1, program_path, last_footnote, font_size = 10)
    rpt <- create_report(fp, 
                         font = "Arial",
                         font_size = 9,
                         orientation = "portrait", 
                         output_type = "DOCX")  %>%
      set_margins(top = 1, bottom = 0.75, right = 1, left = 1.5) %>%
      add_content(tbl_3) %>%
      
      # ----- Title setting -----#
      titles("Table 14-6.1.3.1.  Treatment-emergent Adverse Events by System Organ Class,",
             "Preferred Term and Grade",
             "(Safety Analysis Set)",
             bold = T,
             font_size = 11)
    
    # Write the report
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
    
    # -------------------------------------------- #
    #        DOCX (bigger footnote in footer)      #
    # -------------------------------------------- #
    fp <- file.path(base_path, "user/user18_6.docx")
    
    rpt <- create_report(fp, 
                         font = "Arial",
                         font_size = 9,
                         orientation = "portrait", 
                         output_type = "DOCX")  %>%
      set_margins(top = 1, bottom = 0.75, right = 1, left = 1.5) %>%
      add_content(tbl_1) %>%
      
      # ----- Title setting -----#
      titles("Table 14-6.1.3.1.  Treatment-emergent Adverse Events by System Organ Class,",
             "Preferred Term and Grade",
             "(Safety Analysis Set)",
             bold = T,
             font_size = 11)
    
    rpt <- set_footnote(rpt, program_path, last_footnote, font_size = 10, 
                        valign = "bottom", footer = TRUE)
    
    # Write the report
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
    
    
    # -------------------------------------------- #
    #                   HTML                       #
    # -------------------------------------------- #
    fp <- file.path(base_path, "user/user18.html")
    rpt <- create_report(fp, 
                         orientation = "portrait", 
                         output_type = "HTML")  %>%
      add_style(style = sty) %>%
      set_margins(top = 1, bottom = 0.75, right = 1, left = 1.5) %>%
      add_content(tbl) %>%
      
      # ----- Title setting -----#
      titles("Table 14-6.1.3.1.  Treatment-emergent Adverse Events by System Organ Class,",
             "Preferred Term and Grade",
             "(Safety Analysis Set)",
             bold = T,
             font_size = 11)
    
    # Write the report
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
    
    # -------------------------------------------- #
    #                   TXT                        #
    # -------------------------------------------- #
    tbl <- create_table(df_freq, 
                        borders = "outside",
                        n_format = custom_n_format) %>%
      
      # ----- Column setting -----#
      column_defaults(from = trt1, to = trt3, align = "center", width = trt_width) %>%
      define(blank_grp, blank_before = T, visible = FALSE) %>%
      stub(vars = c("total", "AESOC", "AEDECOD", "AESEV"), 
           label = "System Organ Class\n  Preferred Term\n    Grade", 
           width = item_width) %>%
      define(AESOC, break_label = "(Continued)") %>%
      define(AEDECOD, indent = 0.16, break_label = "(Cont.)") %>%
      define(AESEV, indent = 0.32) %>%
      define(trt1, label = "Treatment A", n=arm_pop[1]) %>%
      define(trt2, label = "Treatment B", n=arm_pop[2]) %>%
      define(trt3, label = "Treatment C", n=arm_pop[3]) %>%
      
      # ----- Footnote setting -----#
      # footnotes("Safety Analysis Set includes all subjects.",
      #           align = "left", blank_row = "above", valign = "top") %>%
      footnotes("The analysis is performed on the maximum severity grade.",
                align = "left", blank_row = "none", valign = "top") %>%
      footnotes("Coded using MedDRA version 22.1.",
                align = "left", blank_row = "none", valign = "top") %>%
      footnotes("Severity of adverse event will be graded using CTCAE version 4.0.",
                align = "left", blank_row = "none", valign = "top") %>%
      footnotes("{supsc('*')} Reported Verbatim Term are displayed for Uncoded AEs.",
                align = "left", blank_row = "none", valign = "top") %>%
      footnotes(paste0("Program: ", program_path)) %>%
      footnotes(last_footnote, blank_row = "none") 
    
    fp <- file.path(base_path, "user/user18.txt")
    rpt <- create_report(fp, 
                         orientation = "portrait", 
                         output_type = "TXT")  %>%
      set_margins(top = 1, bottom = 0.75, right = 1, left = 1.5) %>%
      add_content(tbl) %>%
      
      # ----- Title setting -----#
      titles("Table 14.1  Treatment-emergent Adverse Events by System Organ Class,",
             "Preferred Term and Grade",
             "(Safety Analysis Set)",
             bold = T,
             font_size = 11)
    
    # Write the report
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
    
  } else
    expect_equal(TRUE, TRUE)
})

context("Create Table DOCX Tests")

test_that("get_table_body_docx works as expected.", {

  fp <- ""

  dat <- mtcars[1:10, 1:5]

  tbl <- create_table(dat) %>%
    define("mpg", indent = 0.25)

  rpt <- create_report(fp, output_type = "DOCX", font = "Arial",
                       font_size = 12) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  rpt <- page_setup_docx(rpt)

  wdth <- rep(1, 5)
  names(wdth) <- names(dat)

  algns <- rep("center", 5)
  names(algns) <- names(dat)


  res <- get_table_body_docx(rpt, dat, wdth, algns,  "center", "none", 
                             styles = list(), ts = tbl)

  res
  
  # Check if indenting code is added
  expect_true(
    any(grepl("<w:ind w:left=\"360\" w:right=\"0\"/></w:pPr><w:r><w:t xml:space=\"preserve\">21",
              res$docx))
  )

  expect_equal(length(res$docx), 10)
  expect_equal(res$lines, 10)

})

test_that("get_table_header_docx works as expected.", {

  fp <- ""

  dat <- mtcars[1:10, 1:5]

  tbl <- create_table(dat)

  rpt <- create_report(fp, output_type = "DOCX", font = "Arial",
                       font_size = 12) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  rpt <- page_setup_docx(rpt)

  wdth <- rep(1, 5)
  names(wdth) <- names(dat)

  algns <- rep("center", 5)
  names(algns) <- names(dat)

  nms <- names(dat)
  names(nms) <- nms
  nms[1] <- "Miles Per Gallon"
  nms[2] <- "Wrap One Two"

  pi <- list(keys = names(dat), col_width = wdth, label = nms,
             label_align = algns, table_align = "center")

  res <- get_table_header_docx(rpt, tbl, pi)

  res

  expect_equal(length(res$docx), 1)
  expect_equal(res$lines, 2)

})

test_that("get_content_offsets_docx works as expected.", {

  fp <- ""

  dat <- mtcars[1:10, 1:5]

  tbl <- create_table(dat) %>%
    titles("Table 1.0", "My Nice Table") %>%
    footnotes("Here is a footnote")

  rpt <- create_report(fp, output_type = "DOCX", font = "Arial",
                       font_size = 12) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  rpt <- page_setup_docx(rpt)

  wdth <- rep(1, 5)
  names(wdth) <- names(dat)

  algns <- rep("center", 5)
  names(algns) <- names(dat)

  nms <- names(dat)
  names(nms) <- nms
  nms[1] <- "Miles/Gallon"

  pi <- list(keys = names(dat), col_width = wdth, label = nms,
             label_align = algns, table_align = "center")

  res <- get_content_offsets_docx(rpt, tbl, pi, "both")
  res

  expect_equal(res$lines[["upper"]], 4)
  expect_equal(res$lines[["lower"]], 2)
  expect_equal(res$lines[["blank_upper"]], 1)
  expect_equal(res$lines[["blank_lower"]], 1)

})

test_that("get_content_offsets_docx with page by works as expected.", {
  
  fp <- ""
  
  dat <- iris[1:60,]
  dat$Pgby <- as.character(dat$Species)
  dat <- dat[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Pgby")]
  
  tbl <- create_table(dat) %>%
    titles("Table 1.0", "My Nice Table") %>%
    page_by(Pgby, label = "Flower: ", blank_row = "none") %>%
    footnotes("Here is a footnote")
  
  rpt <- create_report(fp, output_type = "DOCX", font = "Arial",
                       font_size = 12) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")
  
  rpt <- page_setup_docx(rpt)
  
  varname <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
  wdth <- rep(1, 5)
  names(wdth) <- names(dat)
  
  algns <- rep("center", 5)
  names(algns) <- names(dat)
  
  nms <- names(dat)
  names(nms) <- nms
  
  dat$..page_by <- dat$Pgby
  
  pgby_cnt <- get_pgby_cnt(dat$..page_by)
  

  pi <- list(keys = names(dat), col_width = wdth, label = nms,
             label_align = algns, table_align = "center", data = dat)
  
  res <- get_content_offsets_docx(rpt, tbl, pi, "both", pgby_cnt)
  res
  
  expect_equal(res$lines[["upper"]]$setosa, 5)
  expect_equal(res$lines[["upper"]]$versicolor, 5)
  expect_equal(res$lines[["lower"]], 2)
  expect_equal(res$lines[["blank_upper"]], 1)
  expect_equal(res$lines[["blank_lower"]], 1)
  
})

test_that("create_table_docx works as expected.", {

  fp <- ""

  dat <- mtcars[1:10, 1:5]

  tbl <- create_table(dat) %>%
    titles("Table 1.0", "My Nice Table") %>%
    footnotes("Here is a footnote")

  rpt <- create_report(fp, output_type = "DOCX", font = "Arial",
                       font_size = 12) %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", "Right") %>%
    add_content(tbl) %>%
    page_footer("Left1", "Center1", "Right1")

  rpt <- page_setup_docx(rpt)

  wdth <- rep(1, 5)
  names(wdth) <- names(dat)

  algns <- rep("center", 5)
  names(algns) <- names(dat)

  nms <- names(dat)
  names(nms) <- nms
  nms[1] <- "Miles/Gallon"

  pi <- list(keys = names(dat), col_width = wdth, label = nms,
             col_align = algns,
             label_align = algns, table_align = "center", data = dat)

  res <- create_table_docx(rpt, tbl, pi, "below", FALSE, 0, styles = list())
  res

  # Doesn't seem right.  Come back to this.
  expect_equal(length(res$docx), 16) 
  expect_equal(res$lines, 17)    # Should be 13

})

test_that("get_spanning_header_docx works as expected.", {


  fp <- ""

  dat <- mtcars[1:15, ]

  tbl <- create_table(dat, borders = "none") %>%
    spanning_header(mpg, disp, "Span 1", n = 10) %>%
    spanning_header(hp, wt, "Span 2") %>%
    spanning_header(qsec, vs, "Span 3") %>%
    spanning_header(drat, gear, "Super span", level = 2)


  rpt <- create_report(fp, output_type = "DOCX", font = "Arial",
                       font_size = 10, orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    page_header("Left", c("Right1", "Right2", "Right3"), blank_row = "below") %>%
    titles("Table 1.0", "My Nice Table") %>%
    add_content(tbl) %>%
    footnotes("My footnote 1", "My footnote 2") %>%
    page_footer("Left1", "Center1", "Right1")

  rpt <- page_setup_docx(rpt)

  cols <- rep(.6, ncol(dat))
  names(cols) <- names(dat)

  pi <- list(keys = names(dat), col_width = cols, table_align = "center")

  res <- get_spanning_header_docx(rpt, tbl, pi)
  res

  expect_equal(length(res), 2)
  expect_equal(length(res$docx), 2)
  expect_equal(res$lines, 3)
})


test_that("get_page_footnotes_docx works as expected.", {


  tbl1 <- create_table(iris)

  rpt1 <- create_report("", font = "Courier") %>%
    titles("IRIS Data Frame") %>%
    add_content(tbl1) %>%
    footnotes("Here is a footnote", valign = "top")

  rpt1 <- page_setup_docx(rpt1)

  res1 <- get_page_footnotes_docx(rpt1, tbl1, 6,
                                 0, 25, TRUE, "below", "center")
  res1
  expect_equal(res1$lines, 2)

  tbl2 <- create_table(iris)

  rpt2 <- create_report("", font = "Arial") %>%
    titles("IRIS Data Frame") %>%
    add_content(tbl1) %>%
    footnotes("Here is a footnote", valign = "bottom")

  rpt2 <- page_setup_docx(rpt2)

  res2 <- get_page_footnotes_docx(rpt2, tbl2, 6,
                                 0, 25, TRUE, "below", "center")
  res2
  # expect_equal(res2$lines, 12)   # Need to do something with this
  # 20260422: After updating top blank line of header and row height, 
  # the imputed blank lines reduce.
  expect_equal(res2$lines, 11)


  tbl3 <- create_table(iris)  %>%
    footnotes("Here is a footnote", valign = "bottom")

  rpt3 <- create_report("", font = "Times") %>%
    titles("IRIS Data Frame") %>%
    add_content(tbl1)

  rpt3 <- page_setup_docx(rpt3)

  res3 <- get_page_footnotes_docx(rpt3, tbl3, 6,
                                 0, 25, TRUE, "below", "center")

  res3
  # expect_equal(res3$lines, 14)
  # 20260422: After updating top blank line of header and row height, 
  # the imputed blank lines reduce.
  expect_equal(res3$lines, 13)

  tbl4 <- create_table(iris)  %>%
    footnotes("Here is a footnote", valign = "top")

  rpt4 <- create_report("", font = "Times") %>%
    titles("IRIS Data Frame") %>%
    add_content(tbl1)

  rpt4 <- page_setup_docx(rpt4)

  res4 <- get_page_footnotes_docx(rpt4, tbl4, 26,
                                 0, 25, TRUE, "below", "center")

  expect_equal(res4$lines, 2)

})

test_that("get_col_grid() works as expected.", {
  
  
  w <- c(col1 = 1.5, col2 = 2.3, col3 = 3.6)
  
  res <- get_col_grid(w, 1440)
  
  expect_equal(length(res), 1)
  
})
  

context("System Tests")

base_path <- paste0(getwd(),"/tests/testthat")

base_path <- tempdir()

options("logr.output" = FALSE)

dev <- FALSE

test_that("test1: Simplest table works as expected.", {
  
  
  fp <- file.path(base_path, "output/test1.out")
  
  
  rpt <- create_report(fp) %>% 
    add_content(create_table(mtcars[1:10, ]), align = "left")
  
  #print(rpt, verbose = TRUE)
  res <- write_report(rpt)
  
  res
  #print(res, verbose = TRUE)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test2: Simplest table with title works as expected.", {
  
  fp <- file.path(base_path, "output/test2.out")

  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    define(vs, visible = FALSE)
  
  rpt <- create_report(fp) %>% 
    titles("MTCARS Data Frame", align = "left") %>% 
    add_content(tbl)

  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test3: Simple table with formats works as expected.", {
  
  
  fp <- file.path(base_path, "output/test3.out")
  
  
  # Setup
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma", 
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon", 
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
  arm <- c(rep("A", 5), rep("B", 5))
  
  # Create data frame
  df <- data.frame(subjid, name, sex, age, arm, stringsAsFactors = FALSE)
  
  
  afmt <- value(condition(x == "A", "Placebo"),
                condition(x == "B", "Treatment 1"))

  
  sfmt2 <- c(M = "Male", F = "Female")
  
  tbl1 <- create_table(df, first_row_blank = TRUE) %>%
    define(subjid, align = "left") %>% 
    define(sex, width = 1, format = sfmt2) %>%
    define(age, width = .5) %>% 
    define(arm, format = afmt, width = 1.5, align = "right", dedupe = TRUE)
  

  
  rpt <- create_report(fp) %>%
    options_fixed(editor = "notepad++") %>%
    page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
    titles("Table 1.0", "Analysis Data Subject Listing", "Safety Population", 
           align = "center") %>%
    footnotes("Program Name: table1_0.R") %>%
    page_footer(left = "Time", center = "Confidential", 
                right = "Page [pg] of [tpg]") %>%
    add_content(tbl1) 
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test4: Two page report works as expected.", {
  
  
  fp <- file.path(base_path, "output/test4.out")
  
  
  # Setup
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma", 
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon", 
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
  arm <- c(rep("A", 5), rep("B", 5))
  
  # Create data frame
  df <- data.frame(subjid, name, sex, age, arm, stringsAsFactors = FALSE)

  df1 <- df[df$arm == "A", ]
  df2 <- df[df$arm == "B", ]

  afmt <- value(condition(x == "A", "Placebo"),
                condition(x == "B", "Treatment 1"))
  
  sfmt1 <- value(condition(x == "M", "Male"),
                 condition(x == "F", "Female"),
                 condition(TRUE, "Other"))
  
  sfmt2 <- c(M = "Male", F = "Female")
  
  tbl1 <- create_table(df1, first_row_blank = TRUE) %>%
    define(sex, width = 2, format = sfmt1) %>%
    define(age, width = 2)
  
  tbl2 <- create_table(df2) %>%
    define(sex, width = .25, format = sfmt2) %>%
    define(age, format = "%0d%%") %>%
    define(arm, format = afmt, width = 2)
  

  rpt <- create_report(fp, units = "inches", paper_size = "letter") %>%
    options_fixed(editor = "notepad++") %>%
    page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
    titles("Table 1.0", "Analysis Data Subject Listing", 
           "Safety Population", align = "center") %>%
    footnotes("Program Name: table1_0.R") %>%
    page_footer(left = "Time", center = "Confidential", 
                right = "Page [pg] of [tpg]") %>%
    add_content(tbl1) %>%
    add_content(tbl2)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test5: Table with break between sections works as expected.", {
  
  
  fp <- file.path(base_path, "output/test5.out")

  
  # Setup
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma", 
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon", 
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
  arm <- c(rep("A", 5), rep("B", 3), "A", "A")
  
  # Create data frame
  df <- data.frame(subjid, name, sex, age, arm, stringsAsFactors = FALSE)

  
  tbl1 <- create_table(df, first_row_blank = TRUE) %>%
    define(subjid, label = "Subject ID", align = "left") %>% 
    define(name, label = "Subject Name") %>% 
    define(sex, label = "Sex") %>% 
    define(age, label = "Age") %>% 
    define(arm, label = "Arm", 
           blank_after = TRUE, 
           dedupe = TRUE, 
           align = "right")


  rpt <- create_report(fp) %>%
    page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
    titles("Table 1.0", "Analysis Data Subject Listing", 
           "Safety Population", align = "center") %>%
    footnotes("Program Name: table1_0.R") %>%
    page_footer(left = "Time", center = "Confidential", 
                right = "Page [pg] of [tpg]") %>%
    add_content(tbl1) 
  

  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test6: Table that spans multiple pages breaks as expected.", {

  fp <- file.path(base_path, "output/test6.out")


  rpt <- create_report(fp) %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) 


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)

})

test_that("test7: Table with long cell and label values wraps as expected.", {
  
  
  fp <- file.path(base_path, "output/test7.out")
  
  
  # Setup
  arm <- c(rep("A", 5), rep("B", 5))
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa \nand more and more", "Schaffer, Ashley", "Laner, Tahma", 
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon", 
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)

  
  # Create data frame
  df <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)
  

  tbl1 <- create_table(df, first_row_blank = TRUE) %>%
    define(subjid, label = "Subject ID for a patient", n = 10, align = "left", 
           width = 1) %>%
    define(name, label = "Subject Name", width = 1) %>%
    define(sex, label = "Sex", n = 10, align = "center") %>%
    define(age, label = "Age", n = 10) %>%
    define(arm, label = "Arm",
           blank_after = TRUE,
           dedupe = TRUE)


  rpt <- create_report(fp) %>%
    titles("Table 1.0", align = "center") %>%

    add_content(tbl1)


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)

})

test_that("test8: Table with spanning headers works as expected.", {
  
  fp <- file.path(base_path, "output/test8.out")
  

  df <- data.frame(vehicle = rownames(mtcars), mtcars, stringsAsFactors = FALSE)
  rownames(df) = NULL
  
  df$qsec <- fattr(df$qsec, format = "%.1f")
  df$wt <- fattr(df$wt, justify = "center", width = .75)
  
  tbl <- create_table(df) %>% 
    spanning_header("mpg", "hp",
                    label = "Span 1", label_align = "center", n = 10) %>% 
    spanning_header("drat", "qsec",
                    label = "Span 2", label_align = "center", n = 10) %>%
    spanning_header("vs", "carb",
                    label = "Span 3", label_align = "center", n = 10) %>%
    spanning_header(from = "drat", to = "carb", label = "Super Span",
                    label_align = "center",
                    level = 2) %>%
    define(vehicle, label = "Vehicle") %>% 
    define(mpg, format = "%.1f")
  
  rpt <- create_report(fp) %>% 
    add_content(tbl) %>% 
    titles("Table 1.0", "MTCARS Subset Test")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test9: Page wrap works as expected.", {
  
  fp <- file.path(base_path, "output/test9.out")
  
  dat <- mtcars[1:10, ]
  df <- data.frame(vehicle = rownames(dat), dat, stringsAsFactors = FALSE)
  
  tbl <- create_table(df) %>% 
    spanning_header(mpg, hp,
                    label = "Span 1", label_align = "center", n = 10) %>%
    spanning_header(drat, qsec,
                    label = "Span 2", label_align = "center", n = 10) %>%
    spanning_header(vs, carb,
                    label = "Span 3", label_align = "center", n = 10) %>%
    spanning_header("drat","carb", label = "Super Span",
                    label_align = "center",
                    level = 2) %>%
    define(vehicle, label = "Vehicle", id_var = TRUE) %>% 
    define(mpg, format = "%.1f") %>% 
    define(am, visible = FALSE) %>% 
    define(vs, page_wrap = TRUE)
  
  rpt <- create_report(fp, orientation = "portrait") %>%
    titles("Table 1.0", "MTCARS Spanning Header") %>% 
    add_content(tbl) 
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
})

test_that("test10: Page wrap with spanning header works as expected.", {
  
  fp <- file.path(base_path, "output/test10.out")
  
  
  df <- data.frame(vehicle = rownames(mtcars), mtcars, stringsAsFactors = FALSE)
  rownames(df) = NULL
  
  tbl <- create_table(df) %>% 
    spanning_header(2, 5,
                    label = "Span 1", label_align = "center", n = 10) %>%
    spanning_header(6, 8,
                    label = "Span 2", label_align = "center", n = 10) %>%
    spanning_header(9, 12,
                    label = "Span 3", label_align = "center", n = 10) %>%
    spanning_header(6, 12, label = "Super Span",
                    label_align = "center",
                    level = 2) %>%
    define(vehicle, label = "Vehicle", id_var = TRUE) %>% 
    define(mpg, format = "%.1f") %>% 
    define(wt, page_wrap = TRUE) %>% 
    define(vs, page_wrap = TRUE)
  
  rpt <- create_report(fp, orientation = "portrait") %>%
    options_fixed(editor = "wordpad") %>% 
    add_content(tbl) %>% 
    titles("Table 1.0", "MTCARS Subset Test")
  
  #print(rpt)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
})

test_that("test11: Table with break between sections works as expected.", {
  
  
  fp <- file.path(base_path, "output/test11.out")
  
  
  # Setup
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma", 
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon", 
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
  arm <- c(rep("A", 5), rep("B", 5))
  
  # Create data frame
  df <- data.frame(subjid, name, sex, age, arm, stringsAsFactors = FALSE)
  
  
  tbl1 <- create_table(df, first_row_blank = TRUE) %>%
    define(subjid, label = "Subject ID", align = "left") %>% 
    define(name, label = "Subject Name") %>% 
    define(sex, label = "Sex") %>% 
    define(age, label = "Age") %>% 
    define(arm, label = "Arm", 
           blank_after = TRUE, 
           dedupe = TRUE, 
           align = "right")
  
  
  rpt <- create_report(fp) %>%
    page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
    titles("Table 1.0", "Analysis Data Subject Listing", 
           "Safety Population", align = "center") %>%
    footnotes("Program Name: table1_0.R") %>%
    page_footer(left = "Time", center = "Confidential", 
                right = "Page [pg] of [tpg]") %>%
    add_content(tbl1) 
  
  
  res2 <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)

  expect_equal(length(lns), res2$pages * res2$line_count)
  expect_equal(nchar(lns[1]), res2$line_size)
  
})

test_that("test12: Headerless table with title works as expected.", {
  
  fp <- file.path(base_path, "output/test12.out")
  
  
  tbl <- create_table(mtcars[1:10, ], headerless = TRUE) %>% 
    define(vs, visible = FALSE)
  
  rpt <- create_report(fp) %>% 
    add_content(tbl)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test13: Combination with Headerless table works as expected.", {
  
  fp <- file.path(base_path, "output/test13.out")

  
  tbl1 <- create_table(mtcars[1:10, ],  headerless = FALSE) %>% 
    titles("MTCARS Combined Table 1.0")
  tbl2 <- create_table(mtcars[11:20, ], headerless = TRUE) 
  tbl3 <- create_table(mtcars, headerless =  TRUE) %>% 
    footnotes("Full cars table") 
  txt1 <- create_text("These tables are combined!")
  
  rpt <- create_report(fp) %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    page_footer(center = "Page [pg] of [tpg]") %>% 
    add_content(tbl1, page_break = FALSE, blank_row = "none") %>% 
    add_content(tbl2, page_break = FALSE) %>% 
    add_content(txt1) %>% 
    add_content(tbl1, page_break = FALSE, blank_row = "none") %>% 
    add_content(tbl3, page_break = FALSE) %>% 
    add_content(txt1, page_break = FALSE) # Also test open page at end
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test15: Multi-page table with Titles and Footnotes breaks as expected.", {
  
  fp <- file.path(base_path, "output/test15.out")
  
  
  cnt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
                "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
                "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
                "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ", 
                "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
                "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
                "qui officia deserunt mollit anim id est laborum.")
                
  txt <- create_text(cnt, width = 6.25, align = "left") %>% 
    titles("Introduction to Irises")
  
  tbl <- create_table(iris) %>% 
    titles("My little Iris Table") %>% 
    footnotes("* Better Gardening, 1973") %>% 
    define(Species, blank_after = TRUE, dedupe = TRUE)
  
  rpt <- create_report(fp) %>%
    add_content(txt, page_break = FALSE) %>% 
    add_content(tbl, blank_row = "none")
  
  
  res2 <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res2$pages * res2$line_count)
  
})



test_that("test16: Simple regulatory listing works as expected.", {
  
  fp <- file.path(base_path, "output/test16.out")
  
  # Create mtcars listing
  rpt <- create_report(fp, orientation = "portrait") %>% 
    page_header(left = "Client: Motor Trend", right = c("Study: Cars", 
                                                        "Something else")) %>% 
    titles("Listing 1.0", "MTCARS Data Listing") %>% 
    add_content(create_table(mtcars)) %>% 
    footnotes("* Motor Trend, 1973") %>%
    page_footer(left = "Time", 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  rpt
  res <- write_report(rpt)
  
  res
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test17: Simple regulatory table works as expected.", {
  
  library(tidyr)
  library(dplyr)
  
  fp <- file.path(base_path, "output/test17.out")
  
  dat <- mtcars
  
  # Hard coded this so report would come out the same every time.
  #dat$group <- replicate(nrow(dat), sample(c("A", "B"), 1), simplify = TRUE)
  dat$group <- c("B", "B", "B", "B", "B", "B", "A", "A", "B", "A", "A", 
                 "A", "A", "B", "A", "B", "B", "A", "B", "B", "B", "A",
                 "A", "A", "A", "A", "A", "B", "B", "A", "B", "B")
  dat$cyl <- factor(dat$cyl, levels = c(8, 6, 4), 
                    labels = c("8 Cylinder", "6 Cylinder", "4 Cylinder")) 
  group_pop <- table(dat$group)
  
  
  dat_mpg <-
    dat %>%
    group_by(group) %>%
    summarise(across(.cols = mpg,
                     .fns = list(N      = ~ fmt_n(.),
                                 Mean   = ~ fmt_mean_sd(.),
                                 Median = ~ fmt_median(.),
                                 `Q1 - Q3` = ~ fmt_quantile_range(.),
                                 Range  = ~ fmt_range(.)
                     ))) %>%
    pivot_longer(-group,
                 names_to  = c("var", "label"),
                 names_sep = "_",
                 values_to = "value") %>%
    pivot_wider(names_from = group,
                values_from = "value")
  
  
  dat_cyl <-
    dat %>%
    add_count(group, cyl,  name = "n_cyl") %>%
    select(group, cyl, n_cyl) %>%
    distinct() %>%
    pivot_longer(cols = c(cyl),
                 names_to  = "var",
                 values_to = "label") %>%
    pivot_wider(names_from  = group,
                values_from = n_cyl,
                values_fill = 0) %>%
    mutate(A = fmt_cnt_pct(A, group_pop["A"]),
           B = fmt_cnt_pct(B, group_pop["B"])) %>% 
    arrange(label)
  
  
  
  
  final <- bind_rows(dat_mpg, dat_cyl)
  #print(final)
  
  tbl <- create_table(final, first_row_blank = TRUE) %>% 
    stub(c("var", "label")) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(mpg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center") %>% 
    define(B, label = "Group B", align = "center")
  
  
  # Create mtcars table
  rpt <- create_report(fp, orientation = "portrait") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1973") %>%
    page_footer(left = "Time", 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test18: Text and table with page breaks works as expected.", {
  
  fp <- file.path(base_path, "output/test18.out")
  
  cnt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
                "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
                "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
                "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ", 
                "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
                "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
                "qui officia deserunt mollit anim id est laborum.")
  
  # Create text content
  txt <- create_text(cnt) %>% 
    titles("My Analysis of the previous table")
  
  # Prepare data
  dat <- mtcars
  dat$name <- rownames(dat)
  dat <- mtcars[1:10, ]
  
  # Create table content
  tbl <- create_table(dat) %>% 
    titles("Table 1.0", "MTCARS Sample Data") %>% 
    footnotes("* Motor Trend, 1973")
  
  
  # Create report and add both table and text content
  rpt <- create_report(fp, orientation = "portrait") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    add_content(tbl, page_break = FALSE) %>% 
    add_content(txt) %>% 
    add_content(tbl) %>% 
    add_content(txt) %>% 
    page_footer(left = "Time", 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  # Write the report
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})



test_that("test19: show_cols 'none' parameter on table works as expected.", {
  
  fp <- file.path(base_path, "output/test19.out")
  
  tbl <- create_table(mtcars[1:10, ], show_cols = "none") %>%
    define(mpg) %>% 
    define(cyl) %>% 
    define(vs) %>%
    define(drat, visible = FALSE)
  
  rpt <- create_report(fp) %>% 
    titles("MTCARS Data Frame", align = "left") %>% 
    add_content(tbl)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})



test_that("test20: show_cols 'some' parameter on table works as expected.", {
  
  fp <- file.path(base_path, "output/test20.out")

  tbl <- create_table(mtcars[1:10, ], 
                      show_cols = c("vs", "mpg", "cyl", "disp", "drat")) %>% 
    define(mpg, label = "Miles Per Gallon") %>% 
    define(vs, label = "") %>%
    define(drat, visible = FALSE)
  
  
  rpt <- create_report(fp) %>% 
    titles("MTCARS Data Frame", align = "left") %>% 
    add_content(tbl)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test21: Multiple page headers and footers work as expected.", {
  
  fp <- file.path(base_path, "output/test21.out")
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    define(vs, visible = FALSE)
  
  rpt <- create_report(fp) %>% 
    page_header(c("Line 1", "Line 2"), "Right Line") %>%  
    titles("MTCARS Data Frame", align = "left") %>% 
    page_footer(c("Footer Line 1", "Footer Line 2"), right = "Footer Right Line", 
                center = c("A", "B", "C")) %>%  
    add_content(tbl)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test22: Multiple id_var parameters work as expected.", {
  
  fp <- file.path(base_path, "output/test22.out")
  
  dat <- data.frame(vehicle = rownames(mtcars), mtcars, stringsAsFactors = FALSE)
  
  tbl <- create_table(dat) %>%
    define(vehicle, id_var = TRUE, width = 3) %>% 
    define(mpg, id_var = TRUE, width = 2, align = "left") %>% 
    define(vs, id_var = TRUE, width = 2, align = "left")
  
  rpt <- create_report(fp) %>% 
    options_fixed(editor = "word") %>% 
    page_header(c("Line 1", "Line 2"), "Right Line") %>%  
    titles("MTCARS Data Frame", align = "left") %>% 
    page_footer(c("Footer Line 1", "Footer Line 2"), 
                right = "Footer Right Line") %>%  
    add_content(tbl)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)


})

test_that("test23: Blank margins setting works as expected.", {
  
  # Compare output to test1
  
  fp <- file.path(base_path, "output/test23.out")
  
  rpt <- create_report(fp, output_type = "TXT") %>% 
    options_fixed( blank_margins = TRUE) %>% 
    set_margins(top = 1) %>% 
    add_content(create_table(mtcars[1:10, ]), align = "left")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), (res$pages * res$line_count) + res$blank_margin_top)
  
})


test_that("test24: Table width parameter works as expected for full width.", {
  
  
  fp <- file.path(base_path, "output/test24.out")
  
  tbl1 <- create_table(mtcars[1:10, 1:6], width = 9) 
  tbl2 <- create_table(mtcars[11:20, 1:6]) 
  
  rpt <- create_report(fp) %>% 
    page_header("Client", "Study") %>% 
    add_content(tbl1, align = "left", page_break = FALSE) %>% 
    add_content(tbl2, align = "left")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), 1 * res$line_count)
  
})

test_that("test25: page_break parameter simple case works as expected.", {
  
  
  fp <- file.path(base_path, "output/test25.out")
  
  dat <- mtcars
  rownames(dat) <- NULL
  
  dat$pg <- c(rep(1, 16), rep(2, 16))
  
  tbl <- create_table(dat) %>%
    define(pg, page_break = TRUE)
  
  rpt <- create_report(fp) %>% 
    page_header("Client", "Study") %>% 
    titles("MTCARS Sample Report") %>% 
    add_content(tbl)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), 2 * res$line_count)
  
})


test_that("test26: page_break parameter harder case works as expected.", {
  
  
  fp <- file.path(base_path, "output/test26.out")
 
  dat <- iris

  #dat$pg <- c(rep(1, 16), rep(2, 16))
  
  tbl <- create_table(dat) %>%
    define(Species, page_break = TRUE)
  
  rpt <- create_report(fp) %>% 
    page_header("Client", "Study") %>% 
    titles("IRIS Sample Report") %>% 
    add_content(tbl)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages  * res$line_count)
  
})



test_that("test27: page_break parameter even harder case works as expected.", {
  
  
  fp <- file.path(base_path, "output/test27.out")
  
  dat <- iris
  
  
  dat$pg <- c(rep(1, 20), rep(2, 20), rep(3, 20), rep(4, 20), rep(5, 10),
              rep(6, 20), rep(7, 20), rep(8, 20))
  
  tbl <- create_table(dat) %>%
    define(Species, blank_after = TRUE) %>% 
    define(pg, page_break = TRUE)
  
  rpt <- create_report(fp) %>% 
    page_header("Client", "Study") %>% 
    titles("IRIS Sample Report") %>% 
    add_content(tbl)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages  * res$line_count)
  
})



test_that("test28: use_attributes parameter table works as expected.", {
  
  
  fp1 <- file.path(base_path, "output/test28a.out")
  fp2 <- file.path(base_path, "output/test28b.out")
  fp3 <- file.path(base_path, "output/test28c.out")
  
  
  dat <- mtcars[1:10, ]
  attr(dat$mpg, "label") <- "Miles per gallon"
  attr(dat$cyl, "format") <- "%.1f"
  attr(dat$hp, "width") <- 2
  fattr(dat$vs) <- list(width = 2, justify = "center")
  
  tbl <- create_table(dat) 
  
  # Test default 
  rpt <- create_report(fp1) %>% 
    add_content(tbl)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp1), TRUE)
  
  lns <- readLines(fp1)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
  # Test none
  tbl <- create_table(dat, use_attributes = "none") 
  
  rpt <- create_report(fp2) %>% 
    add_content(tbl)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp2), TRUE)
  
  lns <- readLines(fp2)
  
  # Test some
  tbl <- create_table(dat, use_attributes = c("format", "width")) 
  
  rpt <- create_report(fp3) %>% 
    add_content(tbl)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp3), TRUE)
  
  lns <- readLines(fp3)

  expect_equal(length(lns), res$pages * res$line_count)
  
})




test_that("test29: column_defaults work as expected.", {
  
  
  fp <- file.path(base_path, "output/test29.out")
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    column_defaults(width = .5, align = "right", format = "%.1f",
                    n = 5) %>% 
    define(mpg, width = 2, format = "%.2f", align = "left",
           label_align = "right") %>% 
    define(wt, width = 2, format = "%.4f", align = "left", n = 6,
           label_align = "center")
  
  
  rpt <- create_report(fp) %>% 
    add_content(tbl, align = "left")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test30: multiple vars on define work as expected.", {


  fp <- file.path(base_path, "output/test30.out")
  
  tbl <- create_table(mtcars[1:10, ]) %>%
    define(c(mpg, wt), width = 2, format = "%.2f", align = "left",
           label_align = "right")


  rpt <- create_report(fp) %>%
    add_content(tbl, align = "left")


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

  lns <- readLines(fp)

  expect_equal(length(lns), res$pages * res$line_count)

})





test_that("test31: Table width parameter works for less than full width.", {
  
  
  fp <- file.path(base_path, "output/test31.out")
  
  tbl1 <- create_table(mtcars[1:10, 1:6], width = 7) 
  tbl2 <- create_table(mtcars[11:20, 1:6]) 
  
  rpt <- create_report(fp) %>% 
    page_header("Client", "Study") %>% 
    add_content(tbl1, align = "left", page_break = FALSE) %>% 
    add_content(tbl2, align = "left")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), 1 * res$line_count)
  
})


test_that("test36: Report with NAs in data works as expected.", {
  
  
  fp <- file.path(base_path, "output/test36.txt")
  
  # Setup
  subjid <- 100:109
  name <- c("Quintana, Gabriel", NA, "Minniear, Presley",
            "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma", 
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon", 
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", NA, "M", "F", "F", "M")
  age <- c(41, 53, 43, NA, 47, 52, 21, 38, 62, 26)
  arm <- c(rep("A", 5), rep("B", 4), NA)
  
  # Create data frame
  df <- data.frame(subjid, name, sex, age, arm, stringsAsFactors = FALSE)
  
  
  afmt <- value(condition(x == "A", "Placebo"),
                condition(x == "B", "Treatment 1"))
  
  sfmt1 <- value(condition(x == "M", "Male"),
                 condition(x == "F", "Female"))
  
  
  tbl1 <- create_table(df, width = 7) %>%
    define(sex, width = 1, format = sfmt1) %>%
    define(name, width = 2) %>% 
    define(age)

  
  
  rpt <- create_report(fp, output_type = "TXT", missing = "-") %>%
    options_fixed(font_size = 12) %>% 
    set_margins(top = 1, bottom = 1) %>% 
    page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
    titles("Table 1.0", "Analysis Data Subject Listing", 
           "Safety Population", align = "center") %>%
    footnotes("Program Name: table1_0.R") %>%
    page_footer(left = "Time", center = "Confidential", 
                right = "Page [pg] of [tpg]") %>%
    add_content(tbl1) 
  
  
  res <- write_report(rpt)
  
  #print(res)
  
  expect_equal(file.exists(fp), TRUE)
  
  
  
})

test_that("test37: line_size and line_count overrides work as expected.", {
  
  fp <- file.path(base_path, "output/test37.out")
  
  
  rpt <- create_report(fp) %>%
    options_fixed(line_size = 40, line_count = 30) %>% 
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) 
  
  
  res <- write_report(rpt)
  res
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test38: column_defaults works as expected with multiple tables.", {
  
  fp <- file.path(base_path, "output/test38.out")
  
  tbl1 <- create_table(mtcars[1:5, ]) %>% 
    column_defaults(width = .5) 
  
  tbl2 <- create_table(mtcars[6:10, ], headerless=TRUE) %>% 
    column_defaults(width = .5) 
  
  # Create the report object
  rpt <- create_report(fp) %>%
    titles("MTCARS Sample Data", align = "left") %>%
    add_content(tbl1, page_break = FALSE, align = "left", blank_row = "none") %>%
    add_content(tbl2, page_break = FALSE, align = "left") %>%
    add_content(create_text("* NOTE: Above table is actually two tables stacked."))
  
  # Write the report to the file system
  res <- write_report(rpt)
  res
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)

})



test_that("test39: column_defaults works as expected with column positions.", {
  
  fp <- file.path(base_path, "output/test39.out")
  
  tbl1 <- create_table(mtcars[1:5, ]) %>% 
    column_defaults(vars = 1:5, width = .5) %>% 
    column_defaults(6:11, width = .6)
  
  
  # Create the report object
  rpt <- create_report(fp) %>%
    titles("MTCARS Sample Data", align = "left") %>%
    add_content(tbl1, align = "left", blank_row = "none")
  
  # Write the report to the file system
  res <- write_report(rpt)
  res
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


# 
# test_that("test28: Table width parameter more than sum of columns works.", {
#   
#   
#   fp <- file.path(base_path, "output/test28.out")
#   
#   if (file.exists(fp))
#     file.remove(fp)
#   
#   tbl1 <- create_table(mtcars[1:10, ], width = 9) %>% 
#     define(mpg, width = 2) %>% 
#     define(cyl, width = 2) %>% 
#     define(disp, width = 2) %>% 
#     define(hp, width = 2) %>% 
#     define(am, width = 2) %>% 
#     define(carb, width = 2) 
#   
#   rpt <- create_report(fp) %>% 
#     page_header("Client", "Study") %>% 
#     add_content(tbl1, align = "left") 
# 
#   
#   
#   res <- write_report(rpt)
#   
#   expect_equal(file.exists(fp), TRUE)
#   
#   lns <- readLines(fp)
#   
#   expect_equal(length(lns), res$pages * res$line_count)
#   
# })


test_that("test40: Report with units in cm works as expected.", {
  
  fp <- file.path(base_path, "output/test40.out")
  
  
  rpt <- create_report(fp, units = "cm") %>%
    titles("IRIS Data Frame") %>%
    add_content(create_table(iris)) 
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test41: Title header on report works as expected.", {
  
  fp <- file.path(base_path, "output/test41.out")
  
  
  rpt <- create_report(fp) %>%
    title_header("Table 1.0", "IRIS Data Frame", 
                 right = c("Study ABC", "Client A", "Page [pg] of [tpg]")) %>%
    add_content(create_table(iris)) %>% 
    page_footer("DateTime", right = "Page")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test42: Title header on table works as expected.", {
  
  fp <- file.path(base_path, "output/test42.out")
  
  tbl <- create_table(iris) %>% 
    title_header("Table 1.0", "IRIS Data Frame",
                 right = c("Study ABC", "Client A", "Page [pg] of [tpg]"),
                 blank_row = "below")
  
  rpt <- create_report(fp) %>%
    add_content(tbl) 
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test43: Page by on table works as expected.", {
  
  fp <- file.path(base_path, "output/test43.out")
  
  tbl <- create_table(iris) %>% 
    titles("Table 1.0", "IRIS Data Frame") %>% 
    page_by(Species, "Species: ", align = "left") %>% 
    define(Species, visible = FALSE) %>% 
    footnotes("Here is a footnote")
  
  rpt <- create_report(fp) %>%
    page_header("Client", "Study") %>%
    add_content(tbl) %>% 
    page_footer("Time", right = "Page [pg] of [tpg]") 
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test44: Page by on report works as expected.", {
  
  fp <- file.path(base_path, "output/test44.out")
  
  tbl <- create_table(iris) %>% 
    define(Species, visible = FALSE)
  
  rpt <- create_report(fp) %>% 
    page_header("Client", "Study") %>% 
    titles("Table 1.0", "IRIS Data Frame") %>% 
    page_by(Species, "Species: ") %>% 
    footnotes("Here is a footnote") %>% 
    page_footer("Time", right = "Page [pg] of [tpg]") %>% 
    add_content(tbl) 
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})



test_that("test45: Report without page by works as expected.", {
  
  fp <- file.path(base_path, "output/test45.out")
  
  tbl <- create_table(iris) %>% 
    define(Species, page_break = TRUE)
  
  rpt <- create_report(fp, orientation = "portrait", units = "inches") %>% 
    options_fixed(editor = "word", uchar = "-") %>% 
    set_margins(top = 1, bottom = 1, left = 1, right = 1) %>% 
    # 2.54 1.27
    page_header("Client", "Study") %>% 
    titles("Table 1.0", "IRIS Data Frame") %>% 
    footnotes("Here is a footnote") %>% 
    page_footer("Time", right = "Page [pg] of [tpg]") %>% 
    add_content(tbl) 
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test46: Another title_header test to check alignment.", {
  
  
  # Create a temporary file
  fp <- file.path(base_path, "output/test46.out")

  # Prepare data
  dat <- data.frame(category = rownames(USPersonalExpenditure),
                    USPersonalExpenditure, stringsAsFactors = FALSE)

  # Define table
  tbl <- create_table(dat) %>%
    title_header("Table 1.0", "US Personal Expenditures from 1940 - 1960",
                 right = c("Page [pg] of [tpg]", "World Almanac")) %>%
    column_defaults(from = X1940, to = X1960, format = "$%.2f") %>%
    define(category, label = "Category", width = 2) %>%
    define(X1940, label = "1940") %>%
    define(X1945, label = "1945") %>%
    define(X1950, label = "1950") %>%
    define(X1955, label = "1955") %>%
    define(X1960, label = "1960") %>%
    footnotes("* In billions of dollars")

  # Define report
  rpt <- create_report(fp, orientation="portrait") %>%
    add_content(tbl)

  # Write the report
  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

# Alignment is off on this one.
test_that("test47: CM Table with long cell and label values wraps as expected.", {
  
  
  fp <- file.path(base_path, "output/test47.out")
  
  
  # Setup
  arm <- c(rep("A", 5), rep("B", 5))
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa and more and more and more and more and more and more", 
            "Schaffer, Ashley", "Laner, Tahma", 
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon", 
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
  
  
  # Create data frame
  df <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)
  
  
  tbl1 <- create_table(df, first_row_blank = TRUE) %>%
    define(subjid, label = "Subject ID for a patient", n = 10, align = "left", 
           width = 2) %>%
    define(name, label = "Subject Name") %>%
    define(sex, label = "Sex", n = 10, align = "center") %>%
    define(age, label = "Age", n = 10) %>%
    define(arm, label = "Arm",
           blank_after = TRUE,
           dedupe = TRUE)
  
  
  rpt <- create_report(fp, units = "cm") %>%
    page_header("Sponsor", "Study") %>% 
    titles("Table 1.0", align = "right") %>%
    add_content(tbl1, align = "right") %>% 
    page_footer("Time", "Confidential", "Page")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})



test_that("test48: Three level stub and indentation work as expected.", {
  
  
  fp <- file.path(base_path, "output/test48.out")
  
  
  # Setup
  cat <- c(rep("Kaplan-Meier estimates", 6), 
           rep("Cox PH estimates and some more really long stuff", 6))
  grp <- c("25th percentile", "25th percentile", 
           "median (weeks)", "median (weeks)",
           "75th percentile", "75th percentile",
           "25th percentile", "25th percentile", 
           "median (weeks)", "median (weeks)",
           "75th percentile", "75th percentile")
  ci <- c(NA, "95% confidence interval",
          NA, "95% confidence interval",
          NA, "95% confidence interval",
          NA, "95% confidence interval",
          NA, "95% confidence interval",
          NA, "95% confidence interval and more and more and still more")
  values <- c(41, 53, 43, 39, 47, 52, 38, 25, 37, 23, 78, 21)

  # Create data frame
  df <- data.frame(cat, grp, ci, values, stringsAsFactors = FALSE)

  tbl1 <- create_table(df) %>%
    stub(c(cat, grp, ci), "Estimates", width = 2.5) %>% 
    define(cat, label_row = TRUE, blank_after = TRUE, indent = 0.15) %>%
    define(grp, indent = .25) %>%
    define(ci, indent = .5) %>%
    define(values, label = "Values")

  rpt <- create_report(fp) %>%
    titles("Table 3.0", "Analysis of Time to Initial PSGA Success in Weeks") %>% 
    page_header("Sponsor", "Study") %>% 
    add_content(tbl1) %>% 
    page_footer("Time", "Confidential", "Page")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})



test_that("test49: Stub and width settings works as expected.", {
  
  
  fp <- file.path(base_path, "output/test49.out")
  
  
  # Setup
  cat <- c(rep("Kaplan-Meier estimates", 6), 
           rep("Cox PH estimates and some more really long stuff", 6))
  grp <- c("25th percentile", "25th percentile", 
           "median (weeks)", "median (weeks)",
           "75th percentile", "75th percentile",
           "25th percentile", "25th percentile", 
           "median (weeks)", "median (weeks)",
           "75th percentile", "75th percentile")
  ci <- c(NA, "95% confidence interval",
          NA, "95% confidence interval",
          NA, "95% confidence interval",
          NA, "95% confidence interval",
          NA, "95% confidence interval",
          NA, "95% confidence interval")
  values1 <- c(41, 53, 43, 39, 47, 52, 38, 25, 37, 23, 78, 21)
  values2 <- c(32, 28, 94, 32, 09, 42, 67, 65, 56, 12, 91, 32)
  values3 <- c(41, 53, 43, 39, 47, 52, 38, 25, 37, 23, 78, 21)
  values4 <- c(32, 28, 94, 32, 09, 42, 67, 65, 56, 12, 91, 32)
  
  # Create data frame
  df <- data.frame(cat, grp, ci, values1, values2, values3, values4, 
                   stringsAsFactors = FALSE)
  
  tbl1 <- create_table(df, width = 9) %>%
    stub(c(cat, grp, ci), "Estimates", width = 3) %>% 
    define(cat, label_row = TRUE, blank_after = TRUE) %>%
    define(grp, indent = .25) %>%
    define(ci, indent = .5) %>%
    define(values1, label = "Values1", width = .70) %>% 
    define(values2, label = "Values2") %>% 
    define(values3, label = "Values3") %>% 
    define(values4, label = "Values4") 
  
  rpt <- create_report(fp) %>%
    titles("Table 3.0", "Analysis of Time to Initial PSGA Success in Weeks") %>% 
    page_header("Sponsor", "Study") %>% 
    add_content(tbl1) %>% 
    page_footer("Time", "Confidential", "Page")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

# This is still messed up, but not going to fix it for now.
test_that("test50: Table with long numeric values sizes and ident works as expected.", {
  
  fp <- file.path(base_path, "output/test50.out")
  
  # Setup
  arm <- c(rep("A", 5), rep("B", 5))
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa and more and more", "Schaffer, Ashley", "Laner, Tahma", 
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon", 
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
  rnd <- c(0.7077520, 0.3180838, 0.9174493, 0.4395207, 0.5307518, 
           0.8501806, 0.4584808, 0.1016370, 0.7061505, 0.7819369)
  #rnd <- runif(10)
  calc <- age * rnd
  
  # Create data frame
  df <- data.frame(arm, subjid, name, sex, age, rnd, calc, 
                   stringsAsFactors = FALSE)
  
  
  tbl1 <- create_table(df, first_row_blank = TRUE) %>%
    define(subjid, label = "Subject ID for a patient", n = 10) %>%
    define(name, label = "Subject Name", width = 1, indent = 0.15) %>%
    define(sex, label = "Sex", n = 10, align = "left") %>%
    define(age, label = "Age", n = 10) %>%
    define(arm, label = "Arm",
           blank_after = TRUE,
           dedupe = TRUE) %>% 
    define(rnd, width = .5) %>% 
    define(calc, width = .6)
  
  
  rpt <- create_report(fp) %>%
    titles("Table 1.0", align = "center") %>%
    
    add_content(tbl1)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test51: Page break on invisible columns works as expected.", {
  
  fp <- file.path(base_path, "output/test51.out")
  
  tbl <- create_table(iris) %>% 
    titles("Table 1.0", "My Irises Data Sample") %>% 
    define(Species, visible = FALSE, page_break = TRUE) 

   
  
  rpt <- create_report(fp) %>%
    add_content(tbl) %>% 
    page_header("Client", "Study") %>% 
    page_footer("Time", right = "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test52: Column default values for label and n work as expected.", {
  
  fp <- file.path(base_path, "output/test52.out")
  
  tbl <- create_table(iris) %>% 
    titles("Table 1.0", "My Irises Data Sample") %>% 
    column_defaults(label = "Data", n = "100") %>% 
    define("Species", label = "Species")
  
  
  
  rpt <- create_report(fp) %>%
    add_content(tbl) %>% 
    page_header("Client", "Study") %>% 
    page_footer("Time", right = "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})



test_that("test53: Multiple Column defaults work as expected for quoted vars.", {
  
  fp <- file.path(base_path, "output/test53.out")
  
  tbl <- create_table(iris) %>% 
    titles("Table 1.0", "My Irises Data Sample") %>% 
    column_defaults(vars = c("Sepal.Length", "Petal.Length"), label = "Length") %>%
    column_defaults(vars = c("Sepal.Width", "Petal.Width"), label = "Width") %>%
    define("Species", label = "Species")
  
  
  
  rpt <- create_report(fp) %>%
    add_content(tbl) %>% 
    page_header("Client", "Study") %>% 
    page_footer("Time", right = "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test54: Multiple Column defaults work as expected for positions.", {
  
  fp <- file.path(base_path, "output/test54.out")
  
  tbl <- create_table(iris) %>% 
    titles("Table 1.0", "My Irises Data Sample") %>% 
    column_defaults(vars = c(1, 3), label = "Length") %>% 
    column_defaults(vars = c(2, 4), label = "Width") %>% 
    define("Species", label = "Species")
  
  
  
  rpt <- create_report(fp) %>%
    add_content(tbl) %>% 
    page_header("Client", "Study") %>% 
    page_footer("Time", right = "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test55: Multiple Column defaults work as expected for unquoted vars.", {
  
  fp <- file.path(base_path, "output/test55.out")
  
  tbl <- create_table(iris) %>% 
    titles("Table 1.0", "My Irises Data Sample") %>% 
    column_defaults(vars = c(Sepal.Length, Petal.Length), label = "Length") %>%
    column_defaults(vars = c(Sepal.Width, Petal.Width), label = "Width") %>%
    define("Species", label = "Species")
  
  
  
  rpt <- create_report(fp) %>%
    add_content(tbl) %>% 
    page_header("Client", "Study") %>% 
    page_footer("Time", right = "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test56: Multiple Column defaults work as expected for quoted from/to", {
  
  fp <- file.path(base_path, "output/test56.out")
  
  tbl <- create_table(iris) %>% 
    titles("Table 1.0", "My Irises Data Sample") %>% 
    column_defaults(from = "Sepal.Length", to = "Sepal.Width", label = "Sepal") %>%
    column_defaults(from = "Petal.Length", to = "Petal.Width", label = "Petal") %>%
    define("Species", label = "Species")
  
  
  
  rpt <- create_report(fp) %>%
    add_content(tbl) %>% 
    page_header("Client", "Study") %>% 
    page_footer("Time", right = "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test57: Multiple Column defaults work as expected for unquoted from/to", {
  
  fp <- file.path(base_path, "output/test57.out")
  
  tbl <- create_table(iris) %>% 
    titles("Table 1.0", "My Irises Data Sample") %>% 
    column_defaults(from = Sepal.Length, to = Sepal.Width, label = "Sepal") %>%
    column_defaults(from = Petal.Length, to = Petal.Width, label = "Petal") %>%
    define("Species", label = "Species")
  
  
  
  rpt <- create_report(fp) %>%
    add_content(tbl) %>% 
    page_header("Client", "Study") %>% 
    page_footer("Time", right = "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})



test_that("test58: Multiple Column defaults work as expected for positional from/to", {
  
  fp <- file.path(base_path, "output/test58.out")
  
  tbl <- create_table(iris) %>% 
    titles("Table 1.0", "My Irises Data Sample") %>% 
    column_defaults(from = 1, to = 2, label = "Sepal") %>%
    column_defaults(from = 3, to = 4, label = "Petal") %>%
    define("Species", label = "Species")
  
  
  
  rpt <- create_report(fp) %>%
    add_content(tbl) %>% 
    page_header("Client", "Study") %>% 
    page_footer("Time", right = "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test59: Formatting Attributes on data frame work as expected.", {
  
  fp <- file.path(base_path, "output/test59.out")
  
  
  dat <- mtcars
  
  attr(dat$mpg, "label") <- "Miles per gallon"
  dat$mpg <- fattr(dat$mpg, format = "%.2f", width = 2, justify = "left")
  attr(dat$cyl, "label") <- "Cylinders"
  dat$cyl <- fattr(dat$cyl, format = "%.1f", width = 1, justify = "center")
  attr(dat$disp, "label") <- "Displacement"
  dat$disp <- fattr(dat$disp, format = "%.2f", width = 1, justify = "right")
  
  tbl <- create_table(dat, show_cols = c(1:3))
  
  rpt <- create_report(fp) %>% add_content(tbl)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test60: Attributes on data frame are overridden by column defaults.", {
  
  fp <- file.path(base_path, "output/test60.out")
  
  
  dat <- mtcars
  
  attr(dat$mpg, "label") <- "Miles per gallon"
  dat$mpg <- fattr(dat$mpg, format = "%.2f", width = 2, justify = "left")
  attr(dat$cyl, "label") <- "Cylinders"
  dat$cyl <- fattr(dat$cyl, format = "%.1f", width = 1, justify = "center")
  attr(dat$disp, "label") <- "Displacement"
  dat$disp <- fattr(dat$disp, format = "%.2f", width = 1, justify = "right")
  
  tbl <- create_table(dat, show_cols = c(1:3)) %>% 
    column_defaults(width = .75, align = "left", label = "Data", format = "%.0f")
  
  rpt <- create_report(fp) %>% add_content(tbl)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})



test_that("test61: Parameters on column defaults are overridden by define vector.", {
  
  fp <- file.path(base_path, "output/test61.out")
  
  
  dat <- mtcars
  
  attr(dat$mpg, "label") <- "Miles per gallon"
  dat$mpg <- fattr(dat$mpg, format = "%.2f", width = 2, justify = "left")
  attr(dat$cyl, "label") <- "Cylinders"
  dat$cyl <- fattr(dat$cyl, format = "%.1f", width = 1, justify = "center")
  attr(dat$disp, "label") <- "Displacement"
  dat$disp <- fattr(dat$disp, format = "%.2f", width = 1, justify = "right")
  
  tbl <- create_table(dat, show_cols = c(1:3)) %>% 
    column_defaults(width = .75, align = "left", label = "Data",
                    format = "%.0f") %>% 
    define(c(mpg, cyl, disp), label = "Data1", width = 2, 
           align = "right", format = "%.2f")
  
  rpt <- create_report(fp) %>% add_content(tbl)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})



test_that("test62: Parameters on column defaults are overridden by single defines.", {
  
  fp <- file.path(base_path, "output/test62.out")
  
  
  dat <- mtcars
  
  attr(dat$mpg, "label") <- "Miles per gallon"
  dat$mpg <- fattr(dat$mpg, format = "%.2f", width = 2, justify = "left")
  attr(dat$cyl, "label") <- "Cylinders"
  dat$cyl <- fattr(dat$cyl, format = "%.1f", width = 1, justify = "center")
  attr(dat$disp, "label") <- "Displacement"
  dat$disp <- fattr(dat$disp, format = "%.2f", width = 1, justify = "right")
  
  tbl <- create_table(dat, show_cols = c(1:3)) %>% 
    column_defaults(width = .75, align = "left", label = "Data",
                    format = "%.0f") %>% 
    define(mpg, label = "Data1", width = 2, 
           align = "right", format = "%.4f") %>% 
    define(cyl, label = "Data2", width = 1, 
           align = "left", format = "%.3f") %>% 
    define(disp, label = "Data3", width = .75, 
           align = "center", format = "%.1f")
  
  rpt <- create_report(fp) %>% add_content(tbl)
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})



test_that("test63: Simple regulatory table with borders works as expected.", {
  
  library(tidyr)
  library(dplyr)
  
  fp <- file.path(base_path, "output/test63.out")
  
  dat <- mtcars
  
  # Hard coded this so report would come out the same every time.
  #dat$group <- replicate(nrow(dat), sample(c("A", "B"), 1), simplify = TRUE)
  dat$group <- c("B", "B", "B", "B", "B", "B", "A", "A", "B", "A", "A", 
                 "A", "A", "B", "A", "B", "B", "A", "B", "B", "B", "A",
                 "A", "A", "A", "A", "A", "B", "B", "A", "B", "B")
  dat$cyl <- factor(dat$cyl, levels = c(8, 6, 4), 
                    labels = c("8 Cylinder", "6 Cylinder", "4 Cylinder")) 
  group_pop <- table(dat$group)
  
  
  dat_mpg <-
    dat %>%
    group_by(group) %>%
    summarise(across(.cols = mpg,
                     .fns = list(N      = ~ fmt_n(.),
                                 Mean   = ~ fmt_mean_sd(.),
                                 Median = ~ fmt_median(.),
                                 `Q1 - Q3` = ~ fmt_quantile_range(.),
                                 Range  = ~ fmt_range(.)
                     ))) %>%
    pivot_longer(-group,
                 names_to  = c("var", "label"),
                 names_sep = "_",
                 values_to = "value") %>%
    pivot_wider(names_from = group,
                values_from = "value")
  
  
  dat_cyl <-
    dat %>%
    add_count(group, cyl,  name = "n_cyl") %>%
    select(group, cyl, n_cyl) %>%
    distinct() %>%
    pivot_longer(cols = c(cyl),
                 names_to  = "var",
                 values_to = "label") %>%
    pivot_wider(names_from  = group,
                values_from = n_cyl,
                values_fill = 0) %>%
    mutate(A = fmt_cnt_pct(A, group_pop["A"]),
           B = fmt_cnt_pct(B, group_pop["B"])) %>% 
    arrange(label)
  
  
  
  
  final <- bind_rows(dat_mpg, dat_cyl)
  #print(final)
  
  tbl <- create_table(final, first_row_blank = TRUE) %>% 
    stub(c("var", "label")) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(mpg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center") %>% 
    define(B, label = "Group B", align = "center")
  
  
  # Create mtcars table
  rpt <- create_report(fp, orientation = "portrait") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table", borders = "all") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1973", borders = "all") %>%
    page_footer(left = "Time", 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test64: Title header with border on report works as expected.", {
  
  fp <- file.path(base_path, "output/test64.out")
  
  
  rpt <- create_report(fp) %>%
    title_header("Table 1.0", "IRIS Data Frame", 
                 right = c("Study ABC", "Client A", "Page [pg] of [tpg]"),
                 borders = "all") %>%
    add_content(create_table(iris)) %>% 
    page_footer("DateTime", right = "Page")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test65: Title header with borders on table works as expected.", {
  
  fp <- file.path(base_path, "output/test65.out")
  
  tbl <- create_table(iris) %>% 
    title_header("Table 1.0", "IRIS Data Frame",
                 right = c("Study ABC", "Client A", "Page [pg] of [tpg]"),
                 blank_row = "below", borders = "all")
  
  rpt <- create_report(fp) %>%
    add_content(tbl) 
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test66: Titles and footnotes with border on table works as expected.", {
  
  fp <- file.path(base_path, "output/test66.out")
  
  tbl <- create_table(iris) %>% 
    titles("Table 1.0", "IRIS Data Frame",
                 blank_row = "below", borders = "all") %>% 
    footnotes("Here is a footnote", "And another",
              borders = "all")
  
  rpt <- create_report(fp) %>%
    page_header("Left", "Right") %>% 
    add_content(tbl) %>% 
    page_footer("left", "", "right")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test67: Title header with border on report works as expected.", {
  
  fp <- file.path(base_path, "output/test67.out")
  
  
  rpt <- create_report(fp) %>%
    page_header("Left", "Right") %>% 
    titles("Table 1.0", "IRIS Data Frame",
                 borders = "all") %>%
    add_content(create_table(iris)) %>% 
    footnotes("Something", borders = "all") %>% 
    page_footer("DateTime", right = "Page")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})




test_that("test68: show_cols 'some' parameter and label works as expected.", {
  
  fp <- file.path(base_path, "output/test68.out")
  
  mycols <- c("vs", "mpg", "cyl", "disp")
  myvar1 = "vs"
  myvar2 = "mpg"
  tbl <- create_table(mtcars[1:10, ], 
                      show_cols = mycols, use_attributes = "none") %>% 
    define({{myvar2}}, label = "n", width = 1.5) %>% 
    define({{myvar1}}, label = "(%)", width = 1.5)
  
  
  rpt <- create_report(fp, missing = NULL) %>% 
    titles("MTCARS Data Frame", align = "center") %>% 
    add_content(tbl)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test69: Table with spanning headers no underline works as expected.", {
  
  fp <- file.path(base_path, "output/test69.out")
  
  
  df <- data.frame(vehicle = rownames(mtcars), mtcars, stringsAsFactors = FALSE)
  rownames(df) = NULL
  
  df$qsec <- fattr(df$qsec, format = "%.1f")
  df$wt <- fattr(df$wt, justify = "center", width = .75)
  
  tbl <- create_table(df) %>% 
    spanning_header("mpg", "hp",
                    label = "Span 1", label_align = "center", n = 10) %>% 
    spanning_header("drat", "qsec",
                    label = "Span 2", label_align = "center", n = 10, 
                    underline = FALSE) %>%
    spanning_header("vs", "carb",
                    label = "Span 3", label_align = "center", n = 10) %>%
    spanning_header(from = "drat", to = "carb", label = "Super Span",
                    label_align = "center",
                    level = 2) %>%
    define(vehicle, label = "Vehicle") %>% 
    define(mpg, format = "%.1f")
  
  rpt <- create_report(fp) %>% 
    add_content(tbl) %>% 
    titles("Table 1.0", "MTCARS Subset Test")
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test70: Column widths work as expected.", {
  
  
  fp <- file.path(base_path, "output/test70.out")
  
  dat <- unlist(rep("AAAAAAAAA B", 10))
  
  df <- data.frame(col1 = dat, col2 = dat, col3 = dat, col4 = dat, col5 = dat,
                   col6 = dat, col7 = dat, col8 = dat, col9 = dat, 
                   stringsAsFactors = FALSE)
  
  tbl <- create_table(df) %>% 
    spanning_header(1, 2, "Span 1") %>% 
    spanning_header(3, 5, "Span 2") %>% 
    spanning_header(1, 5, "Super Span", level = 2) %>% 
    column_defaults(width = 1) %>% 
    footnotes("Here is a really long footnote that will need a few columns to contain.")
  
  rpt <- create_report(fp, output_type = "TXT") %>% 
    set_margins(left = 1, right = 1) %>% 
    add_content(tbl, align = "left") 
  

  res <- write_report(rpt)
  
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  expect_equal(res$pages, 1)
})


test_that("test71: Column widths work as expected with fractional widths.", {
  
  
  fp <- file.path(base_path, "output/test71.out")
  
  dat <- unlist(rep("AAA", 10))
  
  df <- data.frame(col1 = dat, col2 = dat, col3 = dat, col4 = dat, col5 = dat,
                   col6 = dat, col7 = dat, col8 = dat, col9 = dat, 
                   stringsAsFactors = FALSE)
  
  tbl <- create_table(df) %>% 
    spanning_header(1, 2, "Span 1") %>% 
    spanning_header(3, 5, "Span 2") %>% 
    spanning_header(1, 5, "Super Span", level = 2) %>% 
    column_defaults(width = .5) %>% 
    define(col1, width = 3.5) %>% 
    define(col2, width = 1) %>% 
    define(col7, width = .7) %>%
    define(col8, width = 1.3) %>% 
    footnotes("Here is a really long footnote that will need a few columns to contain.")
  
  rpt <- create_report(fp, output_type = "TXT") %>% 
    set_margins(left = 1, right = 1) %>% 
    add_content(tbl, align = "left")
  
  
  res <- write_report(rpt)
  
  
  # print(res, verbose = TRUE)

  lns <- readLines(fp)

  expect_equal(length(lns), res$pages * res$line_count)
  expect_equal(res$pages, 1)
})


test_that("test72: Check column alignment", {
  
  
  fp <- file.path(base_path, "output/test72.out")
  
  # Sample data
  df <- read.table(header = TRUE, text = '
  Country         N1   Mean1   SD1   N2   Mean2    SD2 
  Bangladesh      52   25.03   0.02  43   22.31   0.08
  Germany         42   95.01   1.02  53   9.31    0.09
  Italy            2   20.22   0.00  11   8.09    1.11')
  
  
  tbl <- create_table(df) %>% 
    column_defaults(width = .5) %>% 
    spanning_header(N1, SD1, label = "Treatment A") %>%
    spanning_header(N2, SD2, label = "Treatment B") %>%
    define(Country, width = 1) %>% 
    titles("Table 1.0", "Treatments by Country") %>% 
    footnotes("* Country of Site")
  
  rpt <- create_report(fp) %>% 
    options_fixed(uchar = "-") %>% 
    add_content(tbl)
  
  res <- write_report(rpt) 
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  expect_equal(res$pages, 1)
  
})



test_that("test73: Table Borders with ttls/fnts on table works as expected.", {
  
  fp <- file.path(base_path, "output/test73.out")
  
  tbl <- create_table(iris, borders = "all") %>% 
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below") %>% 
    footnotes("Here is a footnote", "And another")
  
  rpt <- create_report(fp) %>%
    page_header("Left", "Right") %>% 
    add_content(tbl) %>% 
    page_footer("left", "", "right")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test74: Table Borders with ttls/fnts on report works as expected.", {
  
  fp <- file.path(base_path, "output/test74.out")
  
  tbl <- create_table(iris, borders = "all") 
  
  rpt <- create_report(fp) %>%
    page_header("Left", "Right") %>% 
    add_content(tbl) %>% 
    page_footer("left", "", "right") %>% 
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below") %>% 
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test75: Label row is one cell.", {
  
  
  fp <- file.path(base_path, "output/test75.out")
  
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  ll <- "Here is a super long label to see if it can span the entire table."
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE) %>% 
    stub(c("var", "label")) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = ll, cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", width = 1.25, align = "center", n = 19) %>% 
    define(B, label = "Group B",  width = 1.25,  align = "center", n = 13)
  
  
  # Create report and add content
  rpt <- create_report(fp, orientation = "landscape", output_type = "TXT") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left", 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  
  
  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)
  
  
})

test_that("test76: Page break on invisible column.", {
  
  fp <- file.path(base_path, "output/test76.out")
  
  dat <- iris
  dat$Sequence <- seq(1, nrow(iris))
  
  tbl <- create_table(dat, borders = "all") %>%
    define(Species, page_break = TRUE, visible = FALSE)
  
  rpt <- create_report(fp) %>%
    page_header("Left", "Right") %>% 
    add_content(tbl) %>% 
    page_footer("left", "", "right") %>% 
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below") %>% 
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test77: Blank after on invisible column.", {

  fp <- file.path(base_path, "output/test77.out")

  tbl <- create_table(iris, borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)

  rpt <- create_report(fp) %>%
    page_header("Left", "Right") %>%
    add_content(tbl) %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below") %>%
    footnotes("Here is a footnote", "And another")


  res <- write_report(rpt)

  expect_equal(file.exists(fp), TRUE)

  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
})



test_that("test78: Blank nested stub works as expected.", {
  
  
  fp <- file.path(base_path, "output/test78.out")
  
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "Stats"    "19"          "13"         
      "ampg"   "Stats"    "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Stats"    "16.4"        "21.4"       
      "ampg"   "Stats"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Stats"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  ll <- "Here is a super long label to see if it can span the entire table."
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE, borders = c("all")) %>% 
    stub(c("var", "label")) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = ll, cyl = "Cylinders")) %>% 
    define(label, indent = .25, dedupe = TRUE) %>% 
    define(A, label = "Group A", align = "center", n = 19) %>% 
    define(B, label = "Group B", align = "center", n = 13)
  
  
  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left", 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  
  
  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("test79: Page header width works.", {
  
  fp <- file.path(base_path, "output/test79.out")
  
  tbl <- create_table(iris[1:10, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)
  
  rpt <- create_report(fp) %>%
    page_header(paste0("Left and here is a really long left ",
                       "cell text to put it and more and more"), 
                "Right", width = 8) %>%
    add_content(tbl) %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("test80: Carriage return in label row works.", {
  
  
  fp <- file.path(base_path, "output/test80.out")
  
  
  # Read in prepared data
  df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - \n21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')
  
  ll <- "Here is a super long label to \nsee if it can span\nthe entire table."
  
  # Create table
  tbl <- create_table(df, first_row_blank = TRUE, borders = c("all")) %>% 
    stub(c("var", "label")) %>% 
    define(var, blank_after = TRUE, label_row = TRUE, 
           format = c(ampg = ll, cyl = "Cylinders")) %>% 
    define(label, indent = .25) %>% 
    define(A, label = "Group A", align = "center", n = 19) %>% 
    define(B, label = "Group B", align = "center", n = 13)
  
  
  # Create report and add content
  rpt <- create_report(fp, orientation = "portrait") %>% 
    page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
    titles("Table 1.0", "MTCARS Summary Table") %>% 
    add_content(tbl) %>% 
    footnotes("* Motor Trend, 1974") %>%
    page_footer(left = "Left", 
                center = "Confidential", 
                right = "Page [pg] of [tpg]")
  
  
  
  res <- write_report(rpt)
  res
  expect_equal(file.exists(fp), TRUE)
  
  
})

test_that("test81: Glue feature works.", {
  
  if (dev) {
    library(common)
    
    fp <- file.path(base_path, "output/test81.out")
    
    tbl <- create_table(mtcars[1:10, ], borders = "all") %>%
      spanning_header(1, 4, label = "My span{subsc('4')}") %>%
      define(mpg, label = "Mpg{subsc('3')}")
    
    myvar <- "23"
    
    rpt <- create_report(fp) %>%
      page_header(c("Left {supsc('2')}really long left ",
                         "cell text to put it{supsc('3')} and more and more"), 
                  "Right{supsc('x')}") %>%
      add_content(tbl) %>%
      page_footer(c("left1{supsc('5')}", "left2{supsc('6')}"), "", 
                   "right{supsc('7')}") %>%
       titles("Table 1.0{supsc('1')}", "IRIS Data Frame{{myvar}}",
              blank_row = "below") %>%
       footnotes("Here is a footnote{subsc('a')}", "And another{subsc('9')}")
    
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  
  } else {
    
    expect_equal(TRUE, TRUE) 
  }
    
})



test_that("test82: Title columns work 1 column.", {
  
  fp <- file.path(base_path, "output/test82.out")
  
  tbl <- create_table(iris[1:15, ], borders = "all")  %>%
    titles("Table 1.0 second row", "IRIS Data Frame3",
           blank_row = "both", columns =  1, align = "center",
           borders = c("outside")) 
  
  rpt <- create_report(fp, output_type = "TXT") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("test83: Title columns work 2 columns.", {
  
  fp <- file.path(base_path, "output/test83.out")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    titles("Table 1.0", "IRIS Data Frame", "Left", "Right",
           blank_row = c("below"), columns =  2, borders = "none")
  
  rpt <- create_report(fp, output_type = "TXT") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right")  %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("test84: Title columns work 3 columns.", {
  
  fp <- file.path(base_path, "output/test84.out")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)
  
  rght <- paste("Here")
  
  rpt <- create_report(fp, output_type = "TXT", 
                       font_size = 10) %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame", 
           "My right thing", "", "Center", rght,
           blank_row = "below", columns =  3, borders = "top") %>%
    footnotes("Here is a footnote", "And another", "A",
              "Here is a longer footnote to see if I can figure out the alignment pattern.",
              align = "right")
  
  
  res <- write_report(rpt)
  res
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("test85: Multiple title blocks work as expected.", {
  
  fp <- file.path(base_path, "output/test85.out")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "TXT") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame",
           blank_row = "below", columns =  1, align = "center", width = 7,
           borders = "all") %>%
    titles("Table 2.0", "IRIS Data Frame2", "Left", "Right",
           blank_row = "below", columns =  2, borders = "all") %>%
    titles("Table 3.0", "IRIS Data Frame3", "My right thing", "", "Center",
           blank_row = "below", columns =  3, borders = "all") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("test86: Titles can be passed as a vector.", {
  
  fp <- file.path(base_path, "output/test86.out")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)
  
  ttl <- c("Title1", "Title2", "Title3")
  
  rpt <- create_report(fp, output_type = "TXT") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles(ttl,
           blank_row = "below", columns =  1, align = "center", width = 7,
           borders = "none") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("test87: Custom page size works as expected.", {
  
  fp <- file.path(base_path, "output/test87.out")
  
  tbl <- create_table(iris[1:15, ]) %>%
    define(Species, visible = FALSE)
  
  ttl <- c("Title1", "Title2", "Title3")
  
  rpt <- create_report(fp, output_type = "TXT", paper_size = c(6.5, 7.5),
                       orientation = "portrait") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles(ttl,
           blank_row = "below", columns =  1, align = "center",
           borders = "none") %>%
    footnotes("Here is a footnote", "And another")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})



test_that("test88: Footnotes columns work 1 column.", {
  
  fp <- file.path(base_path, "output/test88.out")
  
  tbl <- create_table(iris[1:15, ], borders = "all")  %>%
    titles("Table 1.0 second row", "IRIS Data Frame3",
           blank_row = "both", columns =  1, align = "center",
           borders = c("outside")) 
  
  rpt <- create_report(fp, output_type = "TXT") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    footnotes("Here is a footnote", "And another", columns = 1)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("test89: Footnotes columns work 2 columns.", {
  
  fp <- file.path(base_path, "output/test89.out")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    titles("Table 1.0", "IRIS Data Frame", "Left", "Right",
           blank_row = c("below"), columns =  2, borders = "none") %>%
    footnotes("Table left", "Table right", columns = 2)
  
  rpt <- create_report(fp, output_type = "TXT") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right")  %>%
    footnotes("Here is a footnote", "And another", columns = 2)
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("test90: Footnotes columns work 3 columns.", {
  
  fp <- file.path(base_path, "output/test90.out")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE)
  
  rght <- paste("Here")
  
  rpt <- create_report(fp, output_type = "TXT", 
                       font_size = 10) %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    titles("Table 1.0", "IRIS Data Frame", 
           "My right thing", "", "Center", rght,
           blank_row = "below", columns =  3, borders = "top") %>%
    footnotes("Here is a footnote", "And another", "A",
              "Here is a longer footnote", "to see if I can figure out",  "the alignment pattern",
              align = "right", columns = 3)
  
  
  res <- write_report(rpt)
  res
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("test91: Multiple footnotes blocks work as expected.", {
  
  fp <- file.path(base_path, "output/test91.out")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE) %>%
    footnotes("Left", "right", columns = 2)
  
  rpt <- create_report(fp, output_type = "TXT") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    footnotes("Footnote1", "IRIS Data Frame",
           blank_row = "below", columns =  1, align = "center", width = 7,
           borders = "all") %>%
    footnotes("Footnotes2", "IRIS Data Frame2", "Left", "Right",
           blank_row = "below", columns =  2, borders = "all") %>%
    footnotes("Footnotes3", "IRIS Data Frame3", "My right thing", "", "Center",
           blank_row = "below", columns =  3, borders = "all") %>%
    titles("Table 1.0", "My little title")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("test92: Page by with wrap works as expected.", {
  
  
  fp <- file.path(base_path, "output/test92.out")
  
  dat <- iris
  dat$Pgby <- as.character(dat$Species)
  dat$Pgby <- paste0("Flower Type\n", dat$Pgby)
  
  
  tbl <- create_table(dat, borders = "none") %>% 
    titles("Table 1.0", "My Nice Report with a Page By", borders = "none") %>%
    page_by(Pgby, align = "left", label = "", borders = "none") %>%
    define(Pgby, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "TXT", 
                       orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_header("Left", "Right") %>% 
    page_footer("Left1", "Center1", "Right1") %>% 
    footnotes("My footnote 1", "My footnote 2", borders = "none")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 6)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  
})


test_that("test93: Page by with wrap works as expected.", {
  
  
  fp <- file.path(base_path, "output/test93.out")
  
  fmt1 <- c(setosa = 1, versicolor = 2, virginica = 3)
  fmt2 <- value(condition(x == 1, "Setosa"),
                condition(x == 2, "Versicolor"),
                condition(x == 3, "Virginica"))
  
  dat <- iris
  fmtval <- fmt1[dat$Species]
  names(fmtval) <- NULL
  dat$Pgby <- fmtval
  
  tbl <- create_table(dat, borders = "none") %>% 
    titles("Table 1.0", "My Nice Report with a Page By", borders = "none") %>%
    page_by(Pgby, align = "left", label = "Flower:", borders = "none", format = fmt2) %>%
    define(Pgby, visible = FALSE)
  
  rpt <- create_report(fp, output_type = "TXT", 
                       orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_header("Left", "Right") %>% 
    page_footer("Left1", "Center1", "Right1") %>% 
    footnotes("My footnote 1", "My footnote 2", borders = "none")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 6)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  
})


test_that("test94: Label with invisible column works as expected.", {
  
  fp <- file.path(base_path, "output/test94.out")
  
  tbl <- create_table(iris[1:15, ], borders = "all") %>%
    define(Species, blank_after = TRUE, visible = FALSE, label = "Fork") %>%
    footnotes("Left", "right", columns = 2)
  
  rpt <- create_report(fp, output_type = "TXT") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    footnotes("Footnote1", "IRIS Data Frame",
              blank_row = "below", columns =  1, align = "center", width = 7,
              borders = "all") %>%
    titles("Table 1.0", "My little title")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("test95: Page break with blank row after works as expected.", {
  
  fp <- file.path(base_path, "output/test95.out")
  
  dat <- sort(iris, by = c("Species", "Petal.Width"))

  
  tbl <- create_table(dat, borders = "all") %>%
    define(Petal.Width, blank_after = TRUE) %>%
    define(Species, page_break = TRUE, visible = TRUE) %>%
    footnotes("Left", "right", columns = 2)
  
  rpt <- create_report(fp, output_type = "TXT") %>%
    add_content(tbl) %>%
    page_header("left", "right") %>%
    page_footer("left", "", "right") %>%
    footnotes("Footnote1", "IRIS Data Frame",
              blank_row = "below", columns =  1, align = "center", width = 7,
              borders = "all") %>%
    titles("Table 1.0", "My little title")
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("test96: Table with blank_before works as expected.", {
  
  
  fp <- file.path(base_path, "output/test96.out")
  
  
  # Setup
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma", 
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon", 
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
  arm <- c(rep("A", 5), rep("B", 3), "A", "A")
  
  # Create data frame
  df <- data.frame(subjid, name, sex, age, arm, stringsAsFactors = FALSE)
  
  
  tbl1 <- create_table(df, first_row_blank = FALSE) %>%
    define(subjid, label = "Subject ID", align = "left") %>% 
    define(name, label = "Subject Name") %>% 
    define(sex, label = "Sex") %>% 
    define(age, label = "Age") %>% 
    define(arm, label = "Arm", 
           blank_before = TRUE, 
           dedupe = TRUE, 
           align = "right")
  
  
  rpt <- create_report(fp) %>%
    page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
    titles("Table 1.0", "Analysis Data Subject Listing", 
           "Safety Population", align = "center") %>%
    footnotes("Program Name: table1_0.R") %>%
    page_footer(left = "Time", center = "Confidential", 
                right = "Page [pg] of [tpg]") %>%
    add_content(tbl1) 
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})


test_that("test97: Increase max column width to 5 inches.", {
  
  
  fp <- file.path(base_path, "output/test97.out")
  
  
  # Setup
  subjid <- 100:109
  name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
            "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma", 
            "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon", 
            "Smith, Shaileigh")
  sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
  age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)
  arm <- c(rep("A", 5), rep("B", 3), "A", "A")
  
  # Create data frame
  df <- data.frame(subjid, name, sex, age, arm, stringsAsFactors = FALSE)
  
  
  tbl1 <- create_table(df, first_row_blank = FALSE) %>%
    define(subjid, label = "Subject ID", align = "left", width = 5) %>% 
    define(name, label = "Subject Name") %>% 
    define(sex, label = "Sex") %>% 
    define(age, label = "Age") %>% 
    define(arm, label = "Arm", 
           blank_before = TRUE, 
           dedupe = TRUE, 
           align = "right")
  
  
  rpt <- create_report(fp) %>%
    page_header(left = "Experis", right = c("Study ABC", "Status: Closed")) %>%
    titles("Table 1.0", "Analysis Data Subject Listing", 
           "Safety Population", align = "center") %>%
    footnotes("Program Name: table1_0.R") %>%
    page_footer(left = "Time", center = "Confidential", 
                right = "Page [pg] of [tpg]") %>%
    add_content(tbl1) 
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * res$line_count)
  
})

test_that("test98: group_cohesion puts the group in same page as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "output/test98.txt")
    
    dat <- iris[1:86,]
    
    dat$test_group <- c(
      rep("A", 12),
      rep("B", 6),
      rep("C", 19),
      rep("D", 18),
      rep("E", 31)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 12), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 6), # 12 lines + 1
      c(rep("Flower C", 19)), # 19 lines + 1
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E", 31)) # 31 lines + 1
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, group_cohesion = TRUE, label = "Group", blank_after = T)
    
    rpt <- create_report(fp, orientation = "landscape", output_type = "TXT") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("test99: group_cohesion with min page proportion works as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "output/test99.txt")
    
    dat <- iris[1:86,]
    
    dat$test_group <- c(
      rep("A", 12),
      rep("B", 6),
      rep("C", 19),
      rep("D", 18),
      rep("E", 31)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 12), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 6), # 12 lines + 1
      c(rep("Flower C", 19)), # 19 lines + 1
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E", 31)) # 31 lines + 1
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, group_cohesion = 0.6,
             label = "Group", blank_after = T) %>%
      define(test_string, label = "Flower")
    
    rpt <- create_report(fp, output_type = "txt", orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("test100: Multiple group_cohesion work as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "output/test100.txt")
    
    dat <- iris[1:86,]
    
    dat$test_group <- c(
      rep("A", 12),
      rep("B", 6),
      rep("C", 19),
      rep("D", 18),
      rep("E", 31)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 12), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 3), # 12 lines + 1
      rep("Flower B\nSubgroup B2", 3),
      c(rep("Flower C1", 11)), # C1 ends at 24 lines, which is the minimum requirement
      c(rep("Flower C2", 8)), # C2 would be moved to next page
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E1", 24), rep("Flower E2", 7)) # E2 would be moved to next page
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, group_cohesion = TRUE, label = "Group", blank_after = T) %>%
      define(test_string, group_cohesion = TRUE)
    
    rpt <- create_report(fp, output_type = "txt", orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    # C2 would be moved to next page
    # E2 would be moved to next page
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("test101: Multiple group_cohesion with different cohesion strengths work as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "output/test101.txt")
    
    dat <- iris[1:92,]
    
    dat$test_group <- c(
      rep("A", 12),
      rep("B", 6),
      rep("C", 19),
      rep("D", 18),
      rep("E", 37)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 12), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 6), # 12 lines + 1
      c(rep("Flower C1", 11)), # C1 ends at 24 lines, which is the minimum requirement
      c(rep("Flower C2", 8)), # C2 would be not moved to next page
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E1", 33), rep("Flower E2", 4)) # E2 would be moved to next page
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, group_cohesion = TRUE, label = "Group", blank_after = T) %>%
      define(test_string, group_cohesion = 0.1)
    
    rpt <- create_report(fp, output_type = "txt", orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Group Cohesion") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    # C2 would not be moved to next page because previous lines are only 24
    # E2 would be moved to next page because previous lines are 27, which meets 30*0.9
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("test102: break_label works as expected.", {
  if (dev == TRUE) {
    fp <- file.path(base_path, "output/test102.txt")
    
    dat <- iris[1:86,]
    
    dat$test_group <- c(
      rep("A", 12),
      rep("B", 6),
      rep("C", 19),
      rep("D", 18),
      rep("E", 31)
    )
    
    dat$test_string <- c(
      rep("Flower A\nSubgroup A1", 12), # 24 lines + 1
      rep("Flower B\nSubgroup B1", 6), # 12 lines + 1
      c(rep("Flower C", 19)), # 19 lines + 1
      c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)), # 20 lines + 1
      c(rep("Flower E", 31)) # 31 lines + 1
    )
    
    dat <- dat[, c("test_group", "test_string", "Sepal.Length", 
                   "Sepal.Width", "Petal.Length","Petal.Width")]
    
    tbl <- create_table(dat, borders = "outside") %>%
      define(test_group, break_label = "(Continued)", blank_before = T) %>%
      define(test_string, 
             break_label = paste0("(This is a long text which should take more ",
                                  "than one line. It's to test if the package is able to handle this ",
                                  "situation.)"),
             indent = 0.16)
    
    rpt <- create_report(fp, output_type = "txt", orientation = "landscape") %>%
      titles("Table 1.0", "My Nice Report with Break Label") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      footnotes("My footnote 1", "My footnote 2", borders = "none",
                blank_row = "none")
    
    res <- write_report(rpt)
    
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

test_that("text103: Page numbers work every in title, footnote, header, and footer.", {
  
  if (dev) {
    fp <- file.path(base_path, "output/test103.txt")
    
    dat <- iris
    
    tbl <- create_table(dat, borders = "outside") %>%
      titles("Table Title: Page [pg] of [tpg]") %>%
      footnotes("Table footnote: Page [pg] of [tpg]")
    
    rpt <- create_report(fp, output_type = "txt", orientation = "landscape") %>%
      set_margins(top = 1, bottom = 1) %>%
      add_content(tbl) %>%
      add_content(tbl) %>%
      titles("Report Title: Page [pg] of [tpg]") %>%
      footnotes("Report footnote: Page [pg] of [tpg]", borders = "none") %>%
      page_header("Header: Page [pg] of [tpg]") %>%
      page_footer("Footer: Page [pg] of [tpg]")
    
    res <- write_report(rpt)
    expect_equal(file.exists(fp), TRUE)
  } else {
    expect_equal(TRUE, TRUE)
  }
})

context("Title and Footnote Tests")

base_path <- "c:/packages/reporter/tests/testthat"

base_path <- tempdir()

cnt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
              "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
              "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
              "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ", 
              "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
              "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
              "qui officia deserunt mollit anim id est laborum.")

test_that("ttfn1: single title right aligned works.", {
  
  fp <- file.path(base_path, "titles/ttfn1.out")
  
  tbl <- create_table(mtcars[1:10, ]) 
  
  rpt <- create_report(fp) %>% 
    titles("MTCARS Data Frame", align = "right") %>% 
    add_content(tbl)
  
  
  write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})
  

test_that("ttfn2: two titles opposite aligned works.", {
  
  fp <- file.path(base_path, "titles/ttfn2.out")
  
  tbl <- create_table(mtcars[1:10, ]) 
  
  rpt <- create_report(fp) %>% 
    titles("MTCARS Data Frame 1", align = "right", blank_row = "above") %>% 
    titles("MTCARS Data Frame 2", align = "left", blank_row = "below") %>% 
    add_content(tbl)
  
  
  write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("ttfn3: single footnote right aligned works.", {
  
  fp <- file.path(base_path, "titles/ttfn3.out")
  
  tbl <- create_table(mtcars[1:10, ]) 
  
  rpt <- create_report(fp) %>% 
    footnotes("MTCARS Data Frame", align = "right") %>% 
    add_content(tbl)
  
  
  write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("ttfn4: two footnotes opposite aligned works.", {
  
  fp <- file.path(base_path, "titles/ttfn4.out")
  
  tbl <- create_table(mtcars[1:10, ]) 
  
  rpt <- create_report(fp) %>% 
    footnotes("MTCARS Data Frame 1", align = "right", blank_row = "above") %>% 
    footnotes("MTCARS Data Frame 2", align = "left", blank_row = "below") %>% 
    add_content(tbl)
  
  
  write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("ttfn5: title and footnote assigned to table works.", {
  
  fp <- file.path(base_path, "titles/ttfn5.out")
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    titles("MTCARS Data Frame 1", align = "center") %>% 
    footnotes("MTCARS Data Frame 2", align = "left") 
  
  rpt <- create_report(fp) %>% 
    add_content(tbl, align = "center")
  
  
  write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("ttfn6: no blank_rows on header and footer work as expected.", {
  
  fp <- file.path(base_path, "titles/ttfn6.out")
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    footnotes("Table footnote", align = "left", blank_row = "none") 
  
  rpt <- create_report(fp) %>% 
    page_header(left = "Study: Cars", blank_row = "none") %>% 
    titles("MTCARS Data Frame 1", align = "left", blank_row = "none") %>% 
    footnotes("Footer footnote", align = "left", blank_row = "none") %>% 
    page_footer(left = "Footer", blank_row = "none") %>% 
    add_content(tbl, align = "left", blank_row = "none", page_break = FALSE) %>% 
    add_content(create_text("Text content", align = "left"))
  
  
  write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("ttfn7: blank_row on header and footer work as expected.", {
  
  fp <- file.path(base_path, "titles/ttfn7.out")

  tbl <- create_table(mtcars[1:10, ]) %>% 
    footnotes("Table footnote", align = "left", blank_row = "both") 
  
  rpt <- create_report(fp) %>% 
    page_header(left = "Study: Cars", blank_row = "below") %>% 
    titles("MTCARS Data Frame 1", align = "left", blank_row = "below") %>% 
    footnotes("Footer footnote", align = "left", blank_row = "none") %>% 
    page_footer(left = "Footer", blank_row = "above") %>% 
    add_content(tbl, align = "left", page_break = FALSE) %>% 
    add_content(create_text("Here is some text content", align = "left"))
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * 45)
  
})


test_that("ttfn8: blank_row on content works as expected.", {
  
  fp <- file.path(base_path, "titles/ttfn8.out")
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    footnotes("Table footnote", align = "left", blank_row = "none") 
  
  rpt <- create_report(fp) %>% 
    page_header(left = "Study: Cars", blank_row = "none") %>% 
    titles("MTCARS Data Frame 1", align = "left", blank_row = "both") %>% 
    footnotes("Footer footnote", align = "left", blank_row = "both") %>% 
    page_footer(left = "Footer", blank_row = "none") %>% 
    add_content(tbl, align = "left", page_break = FALSE, blank_row = "none") %>% 
    add_content(create_text("Here is some text content", align = "left"))
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns), res$pages * 45)
  
})

test_that("ttfn9: default blank_rows on header and footer work as expected.", {
  
  fp <- file.path(base_path, "titles/ttfn9.out")
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    footnotes("Table footnote",
              align = "left") 
  
  rpt <- create_report(fp) %>% 
    page_header(left = "Study: Cars") %>% 
    titles("MTCARS Data Frame 1", align = "left") %>% 
    footnotes("Footer footnote", align = "left") %>% 
    page_footer(left = "Footer") %>% 
    add_content(tbl, align = "left", page_break = FALSE) %>% 
    add_content(create_text("Text content", align = "left"), 
                page_break = FALSE) %>% 
    add_content(create_text("Space check", align = "left"))
  
  
  write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  expect_equal(length(lns), 45)
  expect_equal(nchar(lns[1]), 108)
  
})

test_that("ttfn10: text blank rows work as expected.", {
  
  fp <- file.path(base_path, "titles/ttfn10.out")
  
  cnt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
                "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
                "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
                "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ", 
                "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
                "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
                "qui officia deserunt mollit anim id est laborum.")
  
  tbl <- create_table(mtcars) %>% 
    footnotes("Table footnote", "T2", "T3", "T4", "T5", "T6", "T7", 
              align = "left", blank_row = "both") 
  
  rpt <- create_report(fp) %>% 
    page_header(left = "Study: Cars") %>% 
    titles("MTCARS Data Frame 1", align = "left") %>% 
    footnotes("Footer footnote", align = "left") %>% 
    page_footer(left = "Footer") %>% 
    add_content(tbl, align = "left", page_break = FALSE, blank_row = "below") %>% 
    add_content(create_text("Text content", align = "left"))
  
  
  rs <- write_report(rpt)
  
  expect_equal(rs$page_template_row_count, 7)
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("ttfn11: blank hitting page break works as expected.", {
  
  fp <- file.path(base_path, "titles/ttfn11.out")
  
  tbl <- create_table(mtcars) %>% 
    footnotes("Table footnote",
              align = "left") 
  
  rpt <- create_report(fp) %>% 
    page_header(left = "Study: Cars") %>% 
    titles("MTCARS Data Frame 1", align = "left") %>% 
    footnotes("Footer footnote", align = "left") %>% 
    page_footer(left = "Footer") %>% 
    add_content(tbl, align = "left", page_break = FALSE) %>% 
    add_content(create_text("Text content", align = "left"))
  
  
  res <- write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
  lns <- readLines(fp)
  
  expect_equal(length(lns),  45)
  
})


test_that("ttfn12: title and footnote with bottom borders assigned to table works.", {
  
  fp <- file.path(base_path, "titles/ttfn12.out")
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    titles("MTCARS Data Frame 1", align = "center", 
           borders = "bottom") %>% 
    footnotes("MTCARS Data Frame 2", align = "left", 
              borders = "bottom") 
  
  rpt <- create_report(fp) %>% 
    add_content(tbl, align = "center")
  
  
  write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("ttfn13: title and footnote with top borders assigned to table works.", {
  
  fp <- file.path(base_path, "titles/ttfn13.out")
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    titles("MTCARS Data Frame 1", align = "center", 
           borders = "top") %>% 
    footnotes("MTCARS Data Frame 2", align = "left", 
              borders = "top") 
  
  rpt <- create_report(fp) %>% 
    add_content(tbl, align = "center")
  
  
  write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("ttfn14: title_header with bottom borders assigned to table works.", {
  
  fp <- file.path(base_path, "titles/ttfn14.out")
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    title_header("MTCARS Data Frame 1", 
           borders = "bottom") %>% 
    footnotes("MTCARS Data Frame 2", align = "left", 
              borders = "bottom") 
  
  rpt <- create_report(fp) %>% 
    add_content(tbl, align = "center")
  
  
  write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("ttfn15: title_header with top borders assigned to table works.", {
  
  fp <- file.path(base_path, "titles/ttfn15.out")
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    title_header("MTCARS Data Frame 1", 
           borders = "top") %>% 
    footnotes("MTCARS Data Frame 2", align = "left", 
              borders = "top") 
  
  rpt <- create_report(fp) %>% 
    add_content(tbl, align = "center")
  
  
  write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("ttfn16: Long title generates warning not error.", {
  
  fp <- file.path(base_path, "titles/ttfn16.out")
  
  tbl <- create_table(mtcars[1:10, 1:3]) %>% 
    titles("MTCARS Data Frame 1 MTCARS Data Frame 1 ", align = "left") %>% 
    footnotes("MTCARS", align = "left") 
  
  rpt <- create_report(fp) %>% 
    add_content(tbl, align = "center")
  
  
  expect_warning(write_report(rpt))
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("ttfn17: Long footnote generates warning not error.", {
  
  fp <- file.path(base_path, "titles/ttfn17.out")
  
  tbl <- create_table(mtcars[1:10, 1:3]) %>% 
    titles("MTCARS Data", align = "left") %>% 
    footnotes("MTCARS Data Frame 1 MTCARS Data Frame 2", align = "left") 
  
  rpt <- create_report(fp) %>% 
    add_content(tbl, align = "center")
  
  
  expect_warning(write_report(rpt))
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("ttfn18: Long title_header generates warning not error.", {
  
  fp <- file.path(base_path, "titles/ttfn18.out")
  
  tbl <- create_table(mtcars[1:10, 1:3]) %>% 
    title_header("MTCARS Data Frame 1 MTCARS Data Frame 1 ", right = "Header") %>% 
    footnotes("MTCARS", align = "left") 
  
  rpt <- create_report(fp) %>% 
    add_content(tbl, align = "center")
  
  
  expect_warning(write_report(rpt))
  
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("ttfn19: footnote valign parameter checks work as expected.", {
  
  ftnt <- footnotes(list(), "MTCARS Data Frame 2",  valign = "top")
  expect_equal(ftnt$footnotes[[1]]$valign, "top")
   

  expect_error(footnotes("MTCARS Data Frame 2",  valign = "fork"))
  
  
})

test_that("ttfn20: footnote valign parameter on report works as expected.", {
  
  fp <- file.path(base_path, "titles/ttfn20.out")
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    titles("MTCARS Data Frame 1", align = "center") 
  
  rpt <- create_report(fp) %>% 
    add_content(tbl, align = "center") %>% 
    footnotes("MTCARS Data Frame 2", align = "left", valign = "top") 
  
  
  write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("ttfn21: footnote valign parameter on table works as expected.", {
  
  fp <- file.path(base_path, "titles/ttfn21.out")
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    titles("MTCARS Data Frame 1", align = "center") %>% 
    footnotes("MTCARS Data Frame 2", align = "left", valign = "bottom") 
  
  rpt <- create_report(fp) %>% 
    add_content(tbl, align = "center") 
  
  
  write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("ttfn22: Multiple title_headers works as expected.", {
  
  fp <- file.path(base_path, "titles/ttfn22.out")
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    title_header("MTCARS Data Frame 1", right ="One", blank_row = "none") %>% 
    title_header("MTCARS Data Frame 2", right ="Two", blank_row = "none") %>% 
    title_header("MTCARS Data Frame 3", right ="Three") %>% 
    footnotes("MTCARS Data Frame 2", align = "left", 
              borders = "top") 

  
  rpt <- create_report(fp) %>% 
    add_content(tbl, align = "center")
  
  
  write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})

test_that("ttfn23: Multiple title_headers with multiple titles works as expected.", {
  
  fp <- file.path(base_path, "titles/ttfn23.out")
  
  tbl <- create_table(mtcars[1:10, ]) %>% 
    title_header("MTCARS Data Frame 1", "MTCARS Data Frame 1a",
                 right ="One", blank_row = "none") %>% 
    title_header("MTCARS Data Frame 2", right ="Two", blank_row = "none") %>% 
    title_header("MTCARS Data Frame 3", right =c("Three", "Three A")) %>% 
    footnotes("MTCARS Footnote 1", align = "left", 
              borders = "top") 
  
  
  rpt <- create_report(fp) %>% 
    add_content(tbl, align = "center")
  
  
  write_report(rpt)
  
  expect_equal(file.exists(fp), TRUE)
  
})


test_that("ttfn24: Title and Footnote widths work as expected on table.", {
  
  
  fp <- file.path(base_path, "titles/ttfn24.out")
  
  dat <- iris[1:20, ] 
  
  tbl <- create_table(dat) %>% 
    titles("Table 1.0", "My Nice Report with Widths",
           width = "page", align = "right") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top",
              width = "page", align = "right")
  
  rpt <- create_report(fp, output_type = "TXT", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl) %>%
    page_header("Left", "Right") %>% 
    page_footer("Left1", "Center1", "Right1")
  
  res <- write_report(rpt)
  res
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  
})


test_that("ttfn25: Title and Footnote widths work as expected on report", {
  
  
  fp <- file.path(base_path, "titles/ttfn25.out")
  
  dat <- iris[1:20, ] 
  
  tbl <- create_table(dat, borders = "all") 
  
  rpt <- create_report(fp, output_type = "TXT", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl, align = "center") %>%
    page_footer("Left1", "Center1", "Right1") %>% 
    titles("Table 1.0", "My Nice Report with Widths",
           borders = c("top", "bottom"),
           width = 7, align = "right") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "bottom",
              borders = c("top", "bottom"), 
              width = 7, align = "right")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  
})


test_that("ttfn26: Title and Footnote specific widths work as expected.", {
  
  
  fp <- file.path(base_path, "titles/ttfn26.out")
  
  dat <- iris[1:20, ] 
  
  tbl <- create_table(dat, borders = "all") %>% 
    titles("Table 1.0", "My Nice Report with Borders",
           borders = c("top", "bottom"),
           width = 7, align = "right") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top",
              borders = c("top", "bottom"), 
              width = 7, align = "right")
  
  rpt <- create_report(fp, output_type = "TXT", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl, align = "center") %>%
    page_footer("Left1", "Center1", "Right1") 
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  
})


test_that("ttfn27: Title and Footnote widths work as expected on RTF report", {
  
  
  fp <- file.path(base_path, "titles/ttfn27.rtf")
  
  dat <- iris[1:20, ] 
  
  tbl <- create_table(dat, borders = "all") 
  
  rpt <- create_report(fp, output_type = "RTF", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl, align = "center") %>%
    page_footer("Left1", "Center1", "Right1") %>% 
    titles("Table 1.0", "My Nice Report with Widths",
           borders = c("top", "bottom"),
           width = 7, align = "right") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "bottom",
              borders = c("top", "bottom"), 
              width = 7, align = "right")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  
})


test_that("ttfn28: Title and Footnote widths work as expected on PDF report", {
  
  
  fp <- file.path(base_path, "titles/ttfn28.pdf")
  
  dat <- iris[1:20, ] 
  
  tbl <- create_table(dat, borders = "all") 
  
  rpt <- create_report(fp, output_type = "PDF", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl, align = "center") %>%
    page_footer("Left1", "Center1", "Right1") %>% 
    titles("Table 1.0", "My Nice Report with Widths",
           borders = c("top", "bottom"),
           width = 7, align = "left") %>%
    footnotes("My footnote 1", "My footnote 2", valign = "bottom",
              borders = c("top", "bottom"), 
              width = 7, align = "left")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  
})


test_that("ttfn29: Title Header widths work as expected on PDF report", {
  
  
  fp <- file.path(base_path, "titles/ttfn29.pdf")
  
  dat <- iris[1:20, ] 
  
  tbl <- create_table(dat, borders = "all") 
  
  rpt <- create_report(fp, output_type = "PDF", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl, align = "center") %>%
    page_footer("Left1", "Center1", "Right1") %>% 
    title_header("Table 1.0", "My Nice Report with Widths", right = c("Right"),
           borders = c("top", "bottom"),
           width = 7) %>%
    footnotes("My footnote 1", "My footnote 2", valign = "bottom",
              borders = c("top", "bottom"), 
              width = 7, align = "right")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  
})



test_that("ttfn30: Title Header widths work as expected on RTF report", {
  
  
  fp <- file.path(base_path, "titles/ttfn30.rtf")
  
  dat <- iris[1:20, ] 
  
  tbl <- create_table(dat, borders = "all") 
  
  rpt <- create_report(fp, output_type = "RTF", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl, align = "right") %>%
    page_footer("Left1", "Center1", "Right1") %>% 
    title_header("Table 1.0", "My Nice Report with Widths", right = "Right",
           borders = c("top", "bottom"),
           width = 7) %>%
    footnotes("My footnote 1", "My footnote 2", valign = "bottom",
              borders = c("top", "bottom"), 
              width = 7, align = "right")
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  
})


test_that("ttfn31: Title Header specific widths work as expected.", {
  
  
  fp <- file.path(base_path, "titles/ttfn31.out")
  
  dat <- iris[1:20, ] 
  
  tbl <- create_table(dat, borders = "all", width = 7.9) %>% 
    title_header("Table 1.0", "My Nice Report with Borders", right = "Right",
           borders = c("top", "bottom"),
           width = 8) %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top",
              borders = c("top", "bottom"), 
              width = 8, align = "left")
  
  rpt <- create_report(fp, output_type = "TXT", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(tbl, align = "center") %>%
    page_header("Left", "Right") %>% 
    page_footer("Left1", "Center1", "Right1") 
  
  res <- write_report(rpt)
  res
  res$column_widths
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  expect_equal(length(res$column_widths[[1]]), 5)
  
  
})


test_that("ttfn32: Text Spec specific widths work as expected.", {
  
  
  fp <- file.path(base_path, "titles/ttfn32.out")

  
  txt <- create_text(cnt, borders = "all", width = 7) %>% 
    title_header("Table 1.0", "My Nice Report with Width", right = "Right",
                 borders = c("top", "bottom"),
                 width = 7) %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top",
              borders = c("top", "bottom"), 
              width = 7, align = "left")
  
  rpt <- create_report(fp, output_type = "TXT", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(txt, align = "center") %>%
    page_header("Left", "Right") %>% 
    page_footer("Left1", "Center1", "Right1") 
  
  res <- write_report(rpt)
  res

  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)

  
  
})

test_that("ttfn33: Plot Spec specific widths work as expected on RTF.", {
  
  library(ggplot2)
  
  fp <- file.path(base_path, "titles/ttfn33.rtf")
  
  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  
  
  plt <- create_plot(p, borders = "all", height = 3, width = 7) %>% 
    title_header("Table 1.0", "My Nice Report with Width", right = "Right",
                 borders = c("top", "bottom"),
                 width = 7) %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top",
              borders = c("top", "bottom"), 
              width = 7, align = "left")
  
  rpt <- create_report(fp, output_type = "RTF", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt, align = "center") %>%
    page_header("Left", "Right") %>% 
    page_footer("Left1", "Center1", "Right1") 
  
  res <- write_report(rpt)
  res
  
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  
  
})

test_that("ttfn34: Plot Spec specific widths work as expected on PDF.", {
  
  library(ggplot2)
  
  fp <- file.path(base_path, "titles/ttfn34.pdf")
  
  p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()
  
  
  plt <- create_plot(p, borders = "all", height = 3, width = 7) %>% 
    title_header("Table 1.0", "My Nice Report with Width", right = "Right",
                 borders = c("top", "bottom"),
                 width = 7) %>%
    footnotes("My footnote 1", "My footnote 2", valign = "top",
              borders = c("top", "bottom"), 
              width = 7, align = "left")
  
  rpt <- create_report(fp, output_type = "PDF", orientation = "landscape") %>%
    set_margins(top = 1, bottom = 1) %>%
    add_content(plt, align = "center") %>%
    page_header("Left", "Right") %>% 
    page_footer("Left1", "Center1", "Right1") 
  
  res <- write_report(rpt)
  res
  
  
  expect_equal(file.exists(fp), TRUE)
  expect_equal(res$pages, 1)
  
  
  
})

---
title: "Example 8: Title Header"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 8: Title Header}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

### Table with a Title Header

A title header is a special type of title block.  The title header displays
titles on the left side, and header information on the right side.  The 
title header can be defined with the `title_header()` function.  This style of 
header is preferred by some organizations.

```{r eval=FALSE, echo=TRUE} 
library(reporter)

# Create temp file name
tmp <- file.path(tempdir(), "example8.rtf")

tbl <- create_table(mtcars[1:10, ], width = 7) %>% 
       title_header("Listing 4.0", "MTCARS Data Listing with Title Header", 
               right = c("Page [pg] of [tpg]",  "Study: Cars"),
               blank_row = "both") 

# Create the report
rpt <- create_report(tmp, output_type = "RTF",
                     font = "Courier", font_size = 12) %>% 
  page_header(left = "Client: Motor Trend", right = "Study: Cars",
              blank_row = "below") %>% 
  add_content(tbl) %>% 
  set_margins(top = 1, bottom = 1) %>% 
  footnotes("* Motor Trend, 1974") %>% 
  page_footer(left = Sys.time(), 
        center = "Confidential", 
        right = "Page [pg] of [tpg]")

# Write the report
write_report(rpt)

# file.show(tmp)
```
<img src="../man/images/e8.png" alt="Title Header example image"/>

Next: [Example 9: RTF, PDF, HTML, and DOCX](reporter-rtf.html)

---
title: "reporter"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{reporter}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

The **reporter** package creates regulatory-style statistical reports. It was designed
to produce Tables, Listings, and Figures (TLFs) for the 
the pharmaceutical, biotechnology, and medical-device industries.
However, the functions are generalized enough to provide statistical reporting
for any industry.  The package is written in Base R, and has no dependencies on
any other reporting package.

The package is intended to give R programmers flexible 
report layout capabilities 
and a choice of output formats. The package will 
initially focus on printable, file-based 
output formats. 
The current version supports TXT, RTF, PDF, HTML, and DOCX output types. Note that, 
unlike other packages, all outputs from the **reporter** package
are paged by default.  

## Key Features

The **reporter** package contains the following key features:

* Titles, footnotes, page header, and page footer are repeated on each page
* Supports header labels and spanning headers 
* Calculates default columns widths automatically
* Includes automatic wrapping and splitting of wide and long tables
* Integrates with the **fmtr** package to format numeric, date, and character data
* Plots from the popular **ggplot2** package can be added to RTF, PDF, HTML, 
and DOCX reports
* Allows appending multiple tables to a report, multiple tables to a page, 
and intermingling of text, tables, and plots
* Supports in-report date/time stamps and "Page X of Y" page numbering


## How to use **reporter**
There are four steps to creating a report:

* Create report content
* Create report 
* Add content to the report
* Write out the report 

You can create the report with the `create_report()` function.  Content is
created with the `create_table()`, `create_text()`, or `create_plot()`
functions.  Add content to the report with the `add_content()` function. 
Finally, the report can be written to a file with the `write_report()` function.  

Let's look at a simple example:
```{r eval=FALSE, echo=TRUE}
library(reporter)
library(magrittr)

# Create temp file name 
tmp <- file.path(tempdir(), "example.pdf")

# Create report content
tbl <- create_table(mtcars) %>% 
  titles("MTCARS Sample Data") %>% 
  footnotes("* Motor Trend, 1974")

# Create report and add content  
rpt <- create_report(tmp, output_type = "PDF", 
                     font = "Courier", font_size = 12) %>% 
  add_content(tbl)
  
# Write out the report
write_report(rpt)

# Un-comment to view report
# file.show(tmp)
```
<img src="../man/images/reporter1.png" alt="Reporter example image 1"/>

As you can see, using the **reporter** package, you can 
create a useful report with just a few lines of code.  By default, 
the package creates a text report.  But you can create reports in 
RTF, PDF, HTML, and DOCX with just a single parameter change.  

Besides the functions shown above, there are additional functions to create
page headers and footers, spanning headers, a report stub, a by-variable, 
and more.  The package also allows you to add text and plots to a report.
All of these capabilities are demonstrated in the examples below.

## Next Steps

For next steps, it is recommended that you 
review some examples of the **reporter** package in action.  The examples
show a variety of common use cases.  You will see
that the **reporter** functions are very intuitive and easy to use.  Here is a 
list of examples:

* [Example 1: Listing](reporter-listing.html)
* [Example 2: Table](reporter-table.html)
* [Example 3: Table Stub](reporter-stub.html) 
* [Example 4: Table and Text](reporter-text.html)
* [Example 5: Spanning Headers](reporter-spanning.html)
* [Example 6: Page Wrap](reporter-wrap.html)
* [Example 7: Page By](reporter-pageby.html)
* [Example 8: Title Header](reporter-title_header.html)
* [Example 9: RTF, PDF, HTML, and DOCX](reporter-rtf.html)
* [Example 10: Fonts and Borders](reporter-fonts.html)
* [Example 11: Figure](reporter-figure.html)
* [Example 12: Styles](reporter-styles.html)
* [Example 13: Superscripts and Subscripts](reporter-super.html)
* [More Examples](https://sassy.r-sassy.org/articles/sassy.html)

---
title: "Example 15: Header and Footer Images"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 15: Header and Footer Images}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

### Create a Report with a Header or Footer Image

The `header_image()` and `footer_image()` functions allow you to insert an image
into the report page header or footer.  These functions are useful if you want to
put a company logo or other identifying graphic into the page header or footer.

To add an image to a report header, first define the header using the 
`page_header()` function. For the column you want to insert the image, 
leave the text blank.  Then use the "align" parameter of the `header_image()`
function to insert the image into the desired column. 

Here is an example:

```{r eval=FALSE, echo=TRUE} 
library(reporter)
library(ggplot2)

# Get image path
dir <- system.file("extdata", package = "reporter")
ipth <- file.path(dir, "Logo_full.png")

# create dummy data
data <- data.frame(
  name=letters[1:5],
  value=sample(seq(4,15),5),
  sd=c(1,0.2,3,2,4)
)

# Define Revenue Plot
p <- ggplot(data) +
  geom_bar( aes(x=name, y=value), stat="identity", fill="maroon", alpha=0.7) +
  geom_errorbar( aes(x=name, ymin=value-sd, ymax=value+sd), width=0.4, colour="orange",
                 alpha=0.9, linewidth=1.3)


# Create plot object
plt <- create_plot(p, height = 4.5, width = 6) |>
  titles("Figure 1.0", "Projected Revenue by Quarter", bold = TRUE, font_size = 12)

# Create report
rpt <- create_report("./output/myreport.docx", output_type = "DOCX",
                     font = "Arial", font_size = 10) |>
  page_header("",                          # Put nothing in left column
    c("Archytas Solutions\nChicago, IL, 60601"), # Combine right column text into same cell
              width = c(6, 0, 3)) |>       # Remove middle cell with zero width
  header_image(ipth,                       # Assign image file path
               height = .75, width = 3.5,  # Set height and width
               align = "left") |>          # Align to left header column
  add_content(plt) |>
  page_footer(paste0("Date: ", Sys.Date()), right = "Page [pg] of [tpg]")

# Write report to file
res <- write_report(rpt)

# View report
# file.show(res$path)

```

And here is the generated report:
<img src="../man/images/ImageHeader.png" alt="Image in page header"/>


Next: [More Examples](reporter-more.html)

---
title: "Example 7: Page By"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 7: Page By}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

### Create a Table with a By-Group

A page by variable will split the report into pages by a variable, and
label each page according the variable value.  The page by can be assigned to
the report, to a table, or to a plot.  The page by label will appear between
the titles and the content body.  It can be aligned left, right, or center.
The data should be sorted by the page by variable prior to reporting.

```{r eval=FALSE, echo=TRUE}  
library(reporter)

# Create temp file name
tmp <- file.path(tempdir(), "example7.pdf")

# Prepare data
dat <- mtcars[order(mtcars$cyl), ]
dat <- data.frame(vehicle = rownames(dat), dat)

# Define table
tbl <- create_table(dat, show_cols = 1:8) %>% 
  page_by(cyl, label="Cylinders: ") 

# Create the report
rpt <- create_report(tmp, output_type = "PDF", 
                     font = "Courier", font_size = 12) %>% 
  page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
  titles("Listing 3.0", "MTCARS Data Listing with Page By") %>% 
  set_margins(top = 1, bottom = 1) %>% 
  add_content(tbl) %>% 
  page_footer(left = Sys.time(), 
          center = "Confidential", 
          right = "Page [pg] of [tpg]")


# Write the report
write_report(rpt)

# file.show(tmp)

```
<img src="../man/images/e7a.png" alt="Page by example page 1"/>
<img src="../man/images/e7b.png" alt="Page by example page 2"/>
<img src="../man/images/e7c.png" alt="Page by example page 3"/>


Next: [Example 8: Title Header](reporter-title_header.html)

---
title: "Example 16: Grouping"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 16: Grouping}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```
Data on reports is frequently grouped in some way. The **reporter** package offers 
several features to identify and manage these groups.  Here are some of the 
highlights.

## Group Border

The **reporter** package can draw border lines to separate different groups.
To create the border line, set `group_border = TRUE` on the `define()` function.

Here is an example:
```{r eval=FALSE, echo=TRUE}
library(reporter)

# Create temp file name
tmp <- file.path(tempdir(), "example16a.pdf")

# Create data
arm <- c(rep("A", 3), rep("B", 2), rep("C", 3), rep("D", 2))
subjid <- 100:109
name <- c("Quintana, Gabriel", "Allison, Blas", "Minniear, Presley",
          "al-Kazemi, Najwa", "Schaffer, Ashley", "Laner, Tahma",
          "Perry, Sean", "Crews, Deshawn Joseph", "Person, Ladon",
          "Smith, Shaileigh")
sex <- c("M", "F", "F", "M", "M", "F", "M", "F", "F", "M")
age <- c(41, 53, 43, 39, 47, 52, 21, 38, 62, 26)

df <- data.frame(arm, subjid, name, sex, age, stringsAsFactors = FALSE)
df <- rbind(df, df, df, df)

# Output with group_border
tbl1 <- create_table(df, first_row_blank = FALSE, borders = "outside") %>%
  define(subjid, label = "Subject ID for a patient", n = 10, align = "left",
         width = 1) %>%
  define(name, label = "Subject Name", width = 1) %>%
  define(sex, label = "Sex", n = 10, align = "center") %>%
  define(age, label = "Age", n = 10) %>%
  define(arm, label = "Arm",
         dedupe = TRUE,
         group_border = TRUE) 


rpt <- create_report(tmp, output_type = "pdf", font = "Arial",
                     font_size = 10) %>%
  titles(c("Table 1.0", "This is a table with group border"), align = "center") %>%
  add_content(tbl1) %>%
  set_margins(top = 1, bottom = 1)


res <- write_report(rpt)

```

<img src="../man/images/example16a.png" alt="Example image A"/>


## Break Label

For some types of reports,
it is desirable to have a label that indicates when a group has been broken across
pages. This feature is supported with the "break_label" parameter
on the `define()` function.  In the example below, the "break_label" is assigned
a value of "(Continued)":

```{r eval=FALSE, echo=TRUE} 
library(reporter)

fp <- file.path(tempdir(), "example16b.pdf")

# Prepare data
adae <- data.frame(
  total = rep("Number of subject with adverse events", 53),
  AESOC = c(NA, 
            rep("General disorders and administration site conditions", 9),
            rep("Hepatobiliary disorders", 5),
            rep("Infections and infestations", 25),
            rep("Injury, poisoning and procedural complications", 13)),
  AEDECOD = c(NA,
              NA, rep("APPLICATION SITE ERYTHEMA", 4), rep("INFLUENZA LIKE ILLNESS", 4),
              NA, rep("GALLBLADDER DISORDER", 4),
              NA, rep("BRONCHITIS", 4), rep("NAIL INFECTION", 4), rep("PARONYCHIA", 4),
              rep("RASH PUSTULAR", 4), rep("RESPIRATORY TRACT INFECTION", 4),
              rep("RHINITIS", 4),
              NA, rep("FALL", 4), rep("HEAT CRAMPS", 4), rep("MUSCLE INJURY", 4)
  )
)

adae$aedecod_seq <- rep(NA, nrow(adae))
non_na_idx <- !is.na(adae$AEDECOD)
adae$aedecod_seq[non_na_idx] <- ave(adae$AEDECOD[non_na_idx], adae$AEDECOD[non_na_idx], FUN = seq_along)

adae$AESEV <- adae$aedecod_seq
adae$AESEV[adae$AESEV == "1"] <- NA
adae$AESEV[adae$AESEV == "2"] <- "Mild"
adae$AESEV[adae$AESEV == "3"] <- "Moderate"
adae$AESEV[adae$AESEV == "4"] <- "Severe"

adae$trt1 <- "x ( x.x%)"
adae$trt2 <- "x ( x.x%)"
adae$trt3 <- "x ( x.x%)"

adae$aedecod_id <- rep(NA, nrow(adae))
adae$aedecod_id[non_na_idx] <- ave(adae$AEDECOD[non_na_idx], adae$AESOC[non_na_idx],
                                   FUN = function(x){as.numeric(factor(x, levels = unique(x)))})

adae$blank_grp <- paste(adae$total, 
                        adae$AESOC, 
                        ifelse(adae$aedecod_id == 1,NA,adae$AEDECOD), 
                        sep = "|")

adae <- adae[, setdiff(names(adae), c("aedecod_seq", "aedecod_id"))]

# Output
custom_n_format <- function(x){
  return(paste0("\n(N = ",x,")\nn(%)"))
}

current_date <- gsub(" ","",toupper(format(Sys.Date(),"%d %b %Y")))
current_time <- substr(Sys.time(),12,19)
output_name <- "example16c.pdf"
program_path <- "/dummy/tables/t-teae-soc-pt-gr-saf.R"

total_width <- 6
trt_width <- 1.04
item_width <- total_width - (trt_width*3)

tbl <- create_table(adae, 
                    borders = "outside",
                    n_format = custom_n_format) %>%
  
  # ----- Column setting -----#
  column_defaults(from = trt1, to = trt3, align = "center", width = trt_width) %>%
  define(blank_grp, blank_before = T, visible = FALSE) %>%
  stub(vars = c("total", "AESOC", "AEDECOD", "AESEV"), 
       label = "System Organ Class\n  Preferred Term\n    Grade", 
       width = item_width) %>%
  define(AESOC, break_label = "(Continued)") %>%
  define(AEDECOD, indent = 0.16, break_label = "(Continued)") %>%
  define(AESEV, indent = 0.32) %>%
  define(trt1, label = "Treatment A", n=11) %>%
  define(trt2, label = "Treatment B", n=7) %>%
  define(trt3, label = "Treatment C", n=15) %>%
  
  # ----- Footnote setting -----#
  footnotes("Page [pg] of [tpg]", align = "right", blank_row = "none", valign = "top") %>%
  footnotes("Safety Analysis Set includes all subjects who are received at least one dose of study drug.",
            align = "left", blank_row = "none", valign = "top") %>%
  footnotes("The analysis is performed on the maximum severity grade.",
            align = "left", blank_row = "none", valign = "top") %>%
  footnotes("Coded using MedDRA version 22.1.",
            align = "left", blank_row = "none", valign = "top") %>%
  footnotes(paste0("Program: ", program_path), italics = TRUE) 

sty <- create_style(font_name = "Arial",
                    font_size = 9)

rpt <- create_report(fp, 
                     orientation = "landscape", 
                     output_type = "PDF")  %>%
  add_style(style = sty) %>%
  set_margins(top = 1, bottom = 0.75, right = 1, left = 1.5) %>%
  add_content(tbl) %>%
  
  # ----- Title setting -----#
  titles("Table 14-6.1.3.1.  Treatment-emergent Adverse Events by System Organ Class,",
         "Preferred Term and Grade",
         "(Safety Analysis Set)",
         bold = T,
         font_size = 11) 

# Write the report
res <- write_report(rpt)

# Uncomment to view
# file.show(res$path)

```


<img src="../man/images/example16b_p1.png" alt="Example B page 1"/>

<br>


<img src="../man/images/example16b_p2.png" alt="Example B page 2"/>

<br>
<img src="../man/images/example16b_p3.png" alt="Example B page 3"/>

## Group Cohesion

You many also want to avoid splitting groups 
across pages altogether. To keep groups together on the same page, 
you can use the "group_cohesion" parameter on the `define()` function. 

To illustrate, we will make a small modification to the code above to turn on
group cohesion:
```{r eval=FALSE, echo=TRUE} 
define(AESOC, break_label = "(Continued)", group_cohesion = TRUE)
define(AEDECOD, indent = 0.16, break_label = "(Continued)", group_cohesion = TRUE)
```
Now the result is as follows:

<img src="../man/images/example16c_p1.png" alt="Example C page 1"/>

<br>

<img src="../man/images/example16c_p2.png" alt="Example C page 2"/>

<br>

<img src="../man/images/example16c_p3.png" alt="Example C page 3"/>

<br>

With the above modification, the group *"Infections and infestations"* is still split
because it is a large group, and there is no way to avoid splitting it across
multiple pages.  However, the other breaks have all been avoided by moving 
groups to the next page.  In this way, the report looks nicer and 
becomes more readable.

## Multiple Group Cohesion

The group cohesion feature also works with multiple columns.  To illustrate, 
let's look at another example: 

```{r eval=FALSE, echo=TRUE} 
library(reporter)

fp <- file.path(tempdir(), "example16b.pdf")

# Create data
dat <- iris[1:91,]

dat$group_1 <- c(
  rep("A", 14),
  rep("B", 6),
  rep("C", 22),
  rep("D", 18),
  rep("E", 31)
)

dat$group_2 <- c(
  rep("Flower A\nSubgroup A1", 14),
  rep("Flower B\nSubgroup B1", 3),
  rep("Flower B\nSubgroup B2", 3),
  c(rep("Flower C1", 17)),
  c(rep("Flower C2", 5)),
  c(rep("Flower D", 16), rep("Flower D\nSubgroup D1", 2)),
  c(rep("Flower E1", 2)),
  c(rep("Flower E2", 29))
)

dat <- dat[, c("group_1", "group_2", "Sepal.Length", 
               "Sepal.Width", "Petal.Length","Petal.Width")]

# Create table
tbl <- create_table(dat, borders = "outside") %>%
  define(group_1, group_cohesion = TRUE, label = "Region", blank_after = T, 
         width = 1.2) %>%
  define(group_2, group_cohesion = TRUE, label = "Sector")

# Create report
rpt <- create_report(fp, output_type = "pdf", font = "Arial",
                     font_size = 10, orientation = "landscape") %>%
  titles("Table 1.0", "Analysis of Iris Flowers") %>%
  set_margins(top = 1, bottom = 1) %>%
  add_content(tbl) %>%
  footnotes("Program: /test/example16b.R", 
            "Output: /test/example16b.pdf")

# Output report
res <- write_report(rpt)

```

<img src="../man/images/example16d_p1.png" alt="Example D page 1"/>

<br>
<img src="../man/images/example16d_p2.png" alt="Example D page 2"/>


<br>

<img src="../man/images/example16d_p3.png" alt="Example D page 3"/>

<br>

We can see above that there is still space at the end of page 1, but it's not enough
for all of region B. To keep region B together, it is therefore moved to 
page 2.

After region B moves to page 2, the remaining space is not enough for all of region C.
And we cannot move all of region C to page 3, because it is too big. 

As a result, region C will be split between pages 2 and 3. However, because we
engaged group cohesion for both Region and Sector,
**reporter** will try to keep the sectors on the same page. Therefore, all
of sector "Flower C2" is moved to page 3.

#### Logic of Group Cohesion

The logic of group cohesion is very simple. If a group is to be split, the 
group will be moved to next page only if the current page is at least 80% full.
Accordingly, Region B moves to the next page because Region A occupies at least 80% 
of the page. Similarly, Sector C2 moves to the next page because the combined data 
from Region B and Sector C1 reaches the 80% threshold.

When `group_cohesion = TRUE`, the default strength of group cohesion is 0.2, 
which means the required proportion is `1 - 0.2 = 0.8`. Users can change the 
required proportion by assigning more strength, such as `group_cohesion = 0.6`, 
which means the requirement of proportion is 0.4. Once the previous group has reached 
40% of page, the next group can be moved to a new page if the remaining space is not enough.
That is to say, the stronger the group cohesion, the lower the required proportion,
and the less likely the group is to split.

---
title: "Example 3: Table Stub"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 3: Table Stub}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

### Create a Summary Table with Stub Column

Here is an example of the same table of summary statistics,
but with row labels merged into a stub column.  Creation of a stub
column is accomplished by the `stub()` function, which identifies
two or more variables to merge into a single column.  The `label_row`
parameter tells `write_report()` to use that variable as a label row,
and the `indent` parameter nests the second variable value under the first:

```{r eval=FALSE, echo=TRUE} 
library(reporter)

# Create temporary path
tmp <- file.path(tempdir(), "example3.rtf")

# Read in prepared data
df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')

# Create table
tbl <- create_table(df, first_row_blank = TRUE) %>% 
  stub(c("var", "label")) %>% 
  define(var, blank_after = TRUE, label_row = TRUE, 
         format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
  define(label, indent = .25) %>% 
  define(A, label = "Group A", align = "center", n = 19) %>% 
  define(B, label = "Group B", align = "center", n = 13)


# Create report and add content
rpt <- create_report(tmp, output_type = "RTF",
                     font = "Times", font_size = 12) %>% 
  page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
  titles("Table 1.0", "MTCARS Summary Table") %>% 
  add_content(tbl) %>% 
  footnotes("* Motor Trend, 1974") %>%
  page_footer(left = Sys.time(), 
              center = "Confidential", 
              right = "Page [pg] of [tpg]")

# Write out report
write_report(rpt)

# file.show(tmp)

```
<img src="../man/images/e3.png" alt="Stub example image"/>

Next: [Example 4: Table and Text](reporter-text.html)

---
title: "Example 6: Page Wrap"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 6: Page Wrap}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

### Handle Wide Tables with Page Wrap

Page wrapping will occur 
automatically if the table width exceeds the available page width.  But 
page wrapping can also be controlled using the `page_wrap` parameter on the 
`define()` function.  Below is an example.  Also note the use of the `id_var` 
option to cause the vehicle column to be retained on each wrapped page. 

```{r eval=FALSE, echo=TRUE}  
library(reporter)

# Create temp file name
tmp <- file.path(tempdir(), "example6.pdf")

# Prepare data
dat <- mtcars[1:10, ]
dat <- data.frame(vehicle = rownames(dat), dat)

# Define table
tbl <- create_table(dat, show_cols = 1:8) %>% 
  define(vehicle, label = "Vehicle", width = 3, id_var = TRUE, align = "left") %>% 
  define(mpg, label = "Miles per Gallon", width = 1) %>% 
  define(cyl, label = "Cylinders", format = "%.1f") %>% 
  define(disp, label = "Displacement") %>% 
  define(hp, label = "Horsepower", page_wrap = TRUE) %>% 
  define(drat, visible = FALSE) %>% 
  define(wt, label = "Weight") %>% 
  define(qsec, label = "Quarter Mile Time", width = 1.5) 

  
# Create the report
rpt <- create_report(tmp, output_type = "PDF", 
                     font = "Courier", font_size = 12) %>% 
  titles("Listing 2.0", "MTCARS Data Listing with Page Wrap") %>% 
  set_margins(top = 1, bottom = 1) %>% 
  add_content(tbl) %>% 
  page_footer(right = "Page [pg] of [tpg]")

# Write the report
write_report(rpt)

# file.show(tmp)

```
<img src="../man/images/e6a.png"  alt="Page wrap example image 1"/>
<img src="../man/images/e6b.png"  alt="Page wrap example image 2"/>

Next: [Example 7: Page By](reporter-pageby.html)

---
title: "Example 11: Figure"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 11: Figure}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

### Create a Report with a Figure

RTF, PDF, HTML, and DOCX also support plots produced with the ggplot2 package.  Use 
the `create_plot()` function to create plot content for a report.  The 
plot can be added to the report with the `add_content()` function, just
like tables and text. 

If you want to add a custom image from another charting
package, you may also pass a path to a JPEG file on the `create_plot()` function.
The `write_report()` function will then get the image and insert it into 
the report at the specified location.

```{r eval=FALSE, echo=TRUE} 
library(reporter)
library(ggplot2)
  
# Create temporary path
tmp <- file.path(tempdir(), "example10.pdf")
  
# Prepare data
dat <- mtcars[order(mtcars$cyl), ]

# Generate plot
p <- ggplot(dat, aes(x=disp, y=mpg)) + geom_point()

# Define plot object
plt <- create_plot(p, height = 4, width = 8) %>% 
  titles("Figure 1.0", "MTCARS Mileage By Displacement", blank_row = "none") %>%
  footnotes("* Motor Trend, 1974")

# Add plot to report
rpt <- create_report(tmp, output_type = "PDF") %>%
  set_margins(top = 1, bottom = 1) %>%
  options_fixed(font_size = 12) %>% 
  page_header("Sponsor", "Study: cars") %>%
  add_content(plt) %>%
  page_footer(Sys.time(), "Confidential", "Page [pg] of [tpg]")

# Write out report
res <- write_report(rpt)


# View report
# file.show(tmp)
```
<img src="../man/images/plot.png" alt="Plot example image"/>


Next: [Example 12: Styles and Themes](reporter-styles.html)

---
title: "Example 12: Styles"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 12: Styles}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Styles and Themes

The **reporter** package now supports styles and themes for HTML reports.
Styling for more output types will be supported in future versions.

## Custom Style
To create a custom style, use the `create_style()` function.  This function
allows you to control colors and bolding for many common elements of a report. 
The `create_style()` function returns a style object, which can then
be passed to the style parameter of the `add_style()` function.  Here 
is an example of a custom style:

```{r eval=FALSE, echo=TRUE} 
library(reporter)

# Create temporary path
tmp <- file.path(tempdir(), "example12a.html")

# Read in prepared data
df <- read.table(header = TRUE, text = '
    var     label        A             B          
    "ampg"   "N"          "19"          "13"         
    "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
    "ampg"   "Median"     "16.4"        "21.4"       
    "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
    "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
    "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
    "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
    "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')

# Create custom style
sty <- create_style(font_name = "Arial",
                    font_size = 10,
                    background_color = "WhiteSmoke",
                    border_color = "Grey",
                    title_font_size = 12,
                    title_font_bold = TRUE,
                    title_font_color = "SteelBlue",
                    table_header_background = "Tan",
                    table_header_font_bold = TRUE,
                    table_header_font_color = "White",
                    table_body_background = "White",
                    table_body_stripe = "Wheat",
                    table_stub_background = "Tan",
                    table_stub_font_color = "White",
                    table_stub_font_bold = TRUE
                 )

# Create table
tbl <- create_table(df, first_row_blank = TRUE, borders = c("all")) %>% 
  stub(c("var", "label")) %>% 
  column_defaults(width = 1.25) %>% 
  define(var, blank_after = TRUE, label_row = TRUE, 
         format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
  define(label, indent = .25) %>% 
  define(A, label = "Group A", align = "center", n = 19) %>% 
  define(B, label = "Group B", align = "center", n = 13)  %>% 
  titles("Table 1.0", "MTCARS Summary Table with Custom Style", 
         borders = "none") %>% 
  footnotes("* Motor Trend, 1974", borders = "outside")

# Create report and add custom style
rpt <- create_report(tmp, output_type = "HTML", font = "Arial") %>% 
  add_content(tbl) %>% 
  add_style(sty)

# Write out report
write_report(rpt)

# View report
# file.show(tmp)

```
<img src="../man/images/example12a.png" alt="Styles example image A"/>

## Themes
A Theme may be applied to a report using the function `add_style()` 
and passing a theme name to the theme parameter.  Currently there are
seven themes available: "MidnightBlue", "SteelBlue", 
"DarkRed", "SeaGreen", "SlateGrey", "Plain", and "SASDefault".  Here is 
an example using a named theme:

```{r eval=FALSE, echo=TRUE} 
library(reporter)

# Create temporary path
tmp <- file.path(tempdir(), "example12b.html")

# Read in prepared data
df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')

# Create table
tbl <- create_table(df, first_row_blank = FALSE, borders = "all") %>% 
  stub(c("var", "label")) %>% 
  column_defaults(width = 1.25) %>% 
  define(var, blank_after = TRUE, label_row = TRUE, 
         format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
  define(label, indent = .25) %>% 
  define(A, label = "Group A", align = "center", n = 19) %>% 
  define(B, label = "Group B", align = "center", n = 13)  %>% 
  titles("Table 1.0", "MTCARS Summary Table with SteelBlue Theme", 
         borders = "none") %>% 
  footnotes("* Motor Trend, 1974", borders = "outside")

# Create report and add theme
rpt <- create_report(tmp, output_type = "HTML", font = "Arial", 
                     font_size = 12) %>% 
  set_margins(top = 1, bottom = 1) %>%
  add_content(tbl) %>% 
  add_style(theme = "SteelBlue")

# Write out report
write_report(rpt)

# View report
# file.show(tmp)

```
<img src="../man/images/example12b.png" alt="Styles example image B"/>

## View/Modify a Theme
Sometimes you may want to look at or modify an existing theme.  The function
`get_theme()` allows you to do this.  Simply pass the theme name to 
`get_theme()` and it will return the theme as a style object.  From there you 
can print the style object to see how it was defined.  You can also
modify it to suit your own preferences.  Here is an example of how to 
view and modify a theme.

```{r eval=FALSE, echo=TRUE} 
library(reporter)
library(magrittr)

tm <- get_theme("SteelBlue")

print(tm)
# # A style specification: 
# - font_name: 'Arial'
# - font_size: 10
# - text_color: 'DimGrey'
# - title_font_size: 11
# - title_font_bold: TRUE
# - title_font_color: 'SteelBlue'
# - border_color: 'Grey'
# - table_header_background: 'SteelBlue'
# - table_header_font_bold: TRUE
# - table_header_font_color: 'LightGrey'
# - table_body_background: 'White'
# - table_body_stripe: 'WhiteSmoke'
# - table_stub_background: 'SteelBlue'
# - table_stub_font_color: 'LightGrey'
# - table_stub_font_bold: TRUE

# Modify theme
tm$table_stub_background <- "Orange"
tm$table_stub_font_color <- "White"

# Create temporary path
tmp <- file.path(tempdir(), "example12c.html")

# Read in prepared data
df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')

# Create table
tbl <- create_table(df, first_row_blank = FALSE, borders = "all") %>% 
  stub(c("var", "label")) %>% 
  column_defaults(width = 1.25) %>% 
  define(var, blank_after = TRUE, label_row = TRUE, 
         format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
  define(label, indent = .25) %>% 
  define(A, label = "Group A", align = "center", n = 19) %>% 
  define(B, label = "Group B", align = "center", n = 13)  %>% 
  titles("Table 1.0", "MTCARS Summary Table with Modified Theme", 
         borders = "none") %>% 
  footnotes("* Motor Trend, 1974", borders = "outside")

# Create report and add modified theme
rpt <- create_report(tmp, output_type = "HTML", font = "Arial", 
                     font_size = 12) %>% 
  set_margins(top = 1, bottom = 1) %>%
  add_content(tbl) %>% 
  add_style(style = tm)

# Write out report
write_report(rpt)

# View report
# file.show(tmp)


```
<img src="../man/images/example12i.png" alt="Styles example image I"/>

#### Theme Gallery
In addition to the SteelBlue theme shown above, there are six more themes.
Here is a sample of each:

<img src="../man/images/example12c.png" alt="Styles example image C"/>

<img src="../man/images/example12d.png" alt="Styles example image D"/>

<img src="../man/images/example12e.png" alt="Styles example image E"/>

<img src="../man/images/example12f.png" alt="Styles example image F"/>

<img src="../man/images/example12g.png" alt="Styles example image G"/>

<img src="../man/images/example12h.png" alt="Styles example image H"/>

## Cell Styles

To modify styles for a particular cell in a table, use the `cell_style()`
object.  The `cell_style()` object can be passed into the style parameter on
the `define()`, `stub()`, or `column_defaults()` functions.  This object has 
parameters to set available style settings, and an indicator parameter to 
specify which rows to apply the style to.  The indicator can be one
of four values:

* **labelrow**: Applies the style to all label rows.
* **blankrow**: Applies the style to all blank rows.
* **datarow**: Applies the style to all rows that are not label rows or blank rows.
* **[variable name]**: Applies the style to all rows that have a TRUE value
in a specified variable on the incoming dataset.

First let's look at an example to set the label rows bold:
```{r eval=FALSE, echo=TRUE} 
library(reporter)

# Create temporary path
tmp <- file.path(tempdir(), "example12j.rtf")

# Read in prepared data
df <- read.table(header = TRUE, text = '
    var     label        A             B          
    "ampg"   "N"          "19"          "13"         
    "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
    "ampg"   "Median"     "16.4"        "21.4"       
    "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
    "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
    "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
    "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
    "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')


# Create table
tbl <- create_table(df, first_row_blank = TRUE, borders = c("all"), 
                    header_bold = TRUE) %>% 
  stub(c("var", "label"), 
       style = cell_style(bold = TRUE, indicator = "labelrow")) %>% 
  column_defaults(width = 1.25) %>% 
  define(var, blank_after = TRUE, label_row = TRUE, 
         format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
  define(label, indent = .25) %>% 
  define(A, label = "Group A", align = "center", n = 19) %>% 
  define(B, label = "Group B", align = "center", n = 13)  %>% 
  titles("Table 1.0", "MTCARS Summary Table with Cell Style", 
         borders = "none", bold = TRUE, font_size = 11) %>% 
  footnotes("* Motor Trend, 1974", borders = "none", blank_row = "none")

# Create report and add custom style
rpt <- create_report(tmp, output_type = "RTF", font = "Arial") %>% 
  add_content(tbl) 

# Write out report
write_report(rpt)

# View report
# file.show(tmp)

```
<img src="../man/images/example12j.png" alt="Styles example image J"/>

When a variable name is passed as the indicator value, this variable should 
exist on the input dataset.  The indicator column should have TRUE or FALSE
values for each row where you want the style applied.  This feature gives you
complete control over cell styling.  

Now let's look at and example of setting a style on a variable based on the
variable value:
```{r eval=FALSE, echo=TRUE} 
library(reporter)

# Create temporary path
tmp <- file.path(tempdir(), "example12k.rtf")

# Prepare data
df <- data.frame(names = rownames(mtcars), mtcars[, 1:3])

# Set style indicator variable
df$mpgind <- ifelse(df$mpg > 20, TRUE, FALSE)

# Create table
tbl <- create_table(df, first_row_blank = TRUE, 
                    header_bold = TRUE, borders = c("top", "bottom")) %>% 
  define(names, label = "Car Name") %>%
  define(mpg, label = "Miles Per Gallon", 
         style = cell_style(bold = TRUE, indicator = "mpgind")) %>% 
  define(cyl, label = "Cylinders") %>% 
  define(disp, label = "Displacement")  %>% 
  define(mpgind, visible = FALSE) %>%
  titles("Table 1.0", "MTCARS with Indicator Variable", 
         borders = "none", bold = TRUE, font_size = 11) %>% 
  footnotes("* Motor Trend, 1974", borders = "none", blank_row = "none")

# Create report and add custom style
rpt <- create_report(tmp, output_type = "RTF", font = "Arial") %>% 
  add_content(tbl) 

# Write out report
write_report(rpt)

# View report
# file.show(tmp)

```
<img src="../man/images/example12k.png" alt="Styles example image K"/>

Note that cell styling is available for all output types except TXT.
Also note that the `cell_style()` object currently only supports "bold" styling.
More styling options will be added in future releases of **reporter**.


Next: [Example 13: Superscripts and Subscripts](reporter-super.html)

---
title: "Example 10: Fonts and Borders"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 10: Fonts and Borders}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

### Specify Fonts and Borders

By default, the **reporter** package will create fixed-width style reports 
in Courier font. The package supports 
two additional fonts: Arial and 
Times New Roman.  The package also supports the ability 
to turn on borders for your content.  To demonstrate, the simple demographics 
table seen previously is reproduced below in Arial font with borders 
around the table, titles, and footnotes.

```{r eval=FALSE, echo=TRUE} 
library(reporter)

# Create temporary path
tmp <- file.path(tempdir(), "example10.rtf")

# Read in prepared data
df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')

# Create table
tbl <- create_table(df, first_row_blank = TRUE, borders = "all") %>% 
  stub(c("var", "label")) %>% 
  column_defaults(width = 1.25) %>% 
  define(var, blank_after = TRUE, label_row = TRUE, 
         format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
  define(label, indent = .25) %>% 
  define(A, label = "Group A", align = "center", n = 19) %>% 
  define(B, label = "Group B", align = "center", n = 13)  %>% 
  titles("Table 1.0", "MTCARS Summary Table", borders = "outside",
         bold = TRUE, font_size = 14) %>% 
  footnotes("* Motor Trend, 1974", borders = "outside")

# Create report and add content
rpt <- create_report(tmp, output_type = "RTF", font = "Arial", 
                     font_size = 12) %>% 
  set_margins(top = 1, bottom = 1) %>%
  add_content(tbl)

# Write out report
write_report(rpt)

# View report
# file.show(tmp)

```
<img src="../man/images/fonts.png" alt="Font example image"/>


Next: [Example 11: Figure](reporter-figure.html)

---
title: "Example 9: RTF, PDF, HTML, and DOCX"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 9: RTF, PDF, HTML, and DOCX}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

### Output to RTF, PDF, HTML, and DOCX

The **reporter** package supports five types of output: TXT, RTF, PDF, HTML, 
and DOCX. 
The package produces TXT reports by default.  If you want to produce RTF, PDF, 
HTML, or DOCX output, simply change the output type.  Below is a portion
of the Iris data listing from above, but the `output_type`
is now set to DOCX.

```{r eval=FALSE, echo=TRUE} 
library(reporter)

# Create temporary path
tmp <- file.path(tempdir(), "example9.docx")

# Read in prepared data
df <- iris[1:25, ]

# Create table
tbl <- create_table(df, borders = "bottom") 


# Create report and add content
rpt <- create_report(tmp, output_type = "DOCX", orientation = "portrait", 
                     font = "Times", font_size = 12) %>% 
  set_margins(top = 1, bottom = 1) %>%
  page_header(left = "Client: 1-800-FLOWERS", right = "Study: Iris") %>% 
  titles("Listing 1.0", "Iris Data Listing") %>% 
  add_content(tbl) %>% 
  footnotes("* Bulletin of the American Iris Society, 1935") %>%
  page_footer(left = Sys.time(), 
              center = "Confidential", 
              right = "Page [pg] of [tpg]")

# Write out report
write_report(rpt)

# View report
# file.show(tmp)

```
<img src="../man/images/rtf.png" alt="RTF example image"/>


Next: [Example 10: Fonts and Borders](reporter-fonts.html)

---
title: "Example 2: Table"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 2: Table}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

### Create a Summary Table

Here is an example of a regulatory-style table of summary statistics. This
example also introduces the `define()` function on `create_table()`.  The
`define()` function allows you to control many aspects of a column, including
the label, format, width, alignment, and much more:

```{r eval=FALSE, echo=TRUE} 
library(reporter)

# Create temporary path
tmp <- file.path(tempdir(), "example2.rtf")

# Read in prepared data
df <- read.table(header = TRUE, text = '
      var     label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6%)" "4 ( 30.8%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1%)"  "3 ( 23.1%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3%)"  "6 ( 46.2%)"')

# Create table
tbl <- create_table(df, first_row_blank = TRUE) %>% 
  define(var, label = "Variable", blank_after = TRUE, dedupe = TRUE,
         format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) %>% 
  define(label, label = "") %>% 
  define(A, label = "Group A", align = "center", n = 19) %>% 
  define(B, label = "Group B", align = "center", n = 13)


# Create report and add content
rpt <- create_report(tmp, output_type = "RTF", 
                     font = "Times", font_size = 12) %>% 
  page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
  titles("Table 1.0", "MTCARS Summary Table") %>% 
  add_content(tbl) %>% 
  footnotes("* Motor Trend, 1974") %>%
  page_footer(left = Sys.time(), 
              center = "Confidential", 
              right = "Page [pg] of [tpg]")

# Write out report
write_report(rpt)

# file.show(tmp)

```
<img src="../man/images/e2.png"  alt="Table example image"/>

Next: [Example 3: Table Stub](reporter-stub.html)

---
title: "Example 14: Page Break"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 14: Page Break}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
  
```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

### Handle Long Tables with Page Break

Page breaking will occur 
automatically if the table length exceeds the available page length.  But 
page wrapping can also be controlled using the `page_break` parameter on the 
`define()` function.  The `page_break` parameter identifies a column on which 
to break the page.  The page will break every time the page break variable changes.
You can use an existing variable or create a special variable just for the purpose
of controlling the page breaks.  If you create a special variable just for breaking
the page, it is often desirable to make that column invisible.  Here is an 
example:

```{r eval=FALSE, echo=TRUE}  
library(reporter)

# Create temp file name
tmp <- file.path(tempdir(), "example6.pdf")

# Prepare data
dat <- mtcars
dat <- data.frame(vehicle = rownames(dat), dat, 
                  page = c(rep(1, 10), rep(2, 12), rep(3, 10)))
rownames(dat) <- NULL

# View 
dat
#                vehicle  mpg cyl  disp  hp drat    wt  qsec vs am gear carb page
# 1            Mazda RX4 21.0   6 160.0 110 3.90 2.620 16.46  0  1    4    4    1
# 2        Mazda RX4 Wag 21.0   6 160.0 110 3.90 2.875 17.02  0  1    4    4    1
# 3           Datsun 710 22.8   4 108.0  93 3.85 2.320 18.61  1  1    4    1    1
# 4       Hornet 4 Drive 21.4   6 258.0 110 3.08 3.215 19.44  1  0    3    1    1
# 5    Hornet Sportabout 18.7   8 360.0 175 3.15 3.440 17.02  0  0    3    2    1
# 6              Valiant 18.1   6 225.0 105 2.76 3.460 20.22  1  0    3    1    1
# 7           Duster 360 14.3   8 360.0 245 3.21 3.570 15.84  0  0    3    4    1
# 8            Merc 240D 24.4   4 146.7  62 3.69 3.190 20.00  1  0    4    2    1
# 9             Merc 230 22.8   4 140.8  95 3.92 3.150 22.90  1  0    4    2    1
# 10            Merc 280 19.2   6 167.6 123 3.92 3.440 18.30  1  0    4    4    1
# 11           Merc 280C 17.8   6 167.6 123 3.92 3.440 18.90  1  0    4    4    2
# 12          Merc 450SE 16.4   8 275.8 180 3.07 4.070 17.40  0  0    3    3    2
# 13          Merc 450SL 17.3   8 275.8 180 3.07 3.730 17.60  0  0    3    3    2
# 14         Merc 450SLC 15.2   8 275.8 180 3.07 3.780 18.00  0  0    3    3    2
# 15  Cadillac Fleetwood 10.4   8 472.0 205 2.93 5.250 17.98  0  0    3    4    2
# 16 Lincoln Continental 10.4   8 460.0 215 3.00 5.424 17.82  0  0    3    4    2
# 17   Chrysler Imperial 14.7   8 440.0 230 3.23 5.345 17.42  0  0    3    4    2
# 18            Fiat 128 32.4   4  78.7  66 4.08 2.200 19.47  1  1    4    1    2
# 19         Honda Civic 30.4   4  75.7  52 4.93 1.615 18.52  1  1    4    2    2
# 20      Toyota Corolla 33.9   4  71.1  65 4.22 1.835 19.90  1  1    4    1    2
# 21       Toyota Corona 21.5   4 120.1  97 3.70 2.465 20.01  1  0    3    1    2
# 22    Dodge Challenger 15.5   8 318.0 150 2.76 3.520 16.87  0  0    3    2    2
# 23         AMC Javelin 15.2   8 304.0 150 3.15 3.435 17.30  0  0    3    2    3
# 24          Camaro Z28 13.3   8 350.0 245 3.73 3.840 15.41  0  0    3    4    3
# 25    Pontiac Firebird 19.2   8 400.0 175 3.08 3.845 17.05  0  0    3    2    3
# 26           Fiat X1-9 27.3   4  79.0  66 4.08 1.935 18.90  1  1    4    1    3
# 27       Porsche 914-2 26.0   4 120.3  91 4.43 2.140 16.70  0  1    5    2    3
# 28        Lotus Europa 30.4   4  95.1 113 3.77 1.513 16.90  1  1    5    2    3
# 29      Ford Pantera L 15.8   8 351.0 264 4.22 3.170 14.50  0  1    5    4    3
# 30        Ferrari Dino 19.7   6 145.0 175 3.62 2.770 15.50  0  1    5    6    3
# 31       Maserati Bora 15.0   8 301.0 335 3.54 3.570 14.60  0  1    5    8    3
# 32          Volvo 142E 21.4   4 121.0 109 4.11 2.780 18.60  1  1    4    2    3

# Define table
tbl <- create_table(dat) %>% 
  define(vehicle, label = "Vehicle", width = 2, align = "left") |>
  define(page, page_break = TRUE, visible = FALSE) 


# Create the report
rpt <- create_report(tmp, output_type = "PDF", 
                     font = "Courier", font_size = 12) |>
  titles("Listing 2.0", "MTCARS Data Listing with Page Break") |>
  set_margins(top = 1, bottom = 1) |>
  add_content(tbl) |> 
  page_footer(right = "Page [pg] of [tpg]")

# Write the report
write_report(rpt)

# file.show(tmp)

```
<img src="../man/images/e14a.png" alt="Page break pg 1"/>
<img src="../man/images/e14b.png" alt="Page break pg 2"/>
<img src="../man/images/e14c.png" alt="Page break pg 3"/>
  
Next: [Example 15: Header and Footer Images](reporter-hfimage.html)

---
title: "Example 1: Listing"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 1: Listing}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

### Create a Simple Listing

Here is an example of 
a regulatory-style listing using **reporter** and the iris sample data frame.
Notice how the `write_report()` function breaks the page automatically
and appropriately:

```{r eval=FALSE, echo=TRUE} 
library(reporter)

# Create temp file name
tmp <- file.path(tempdir(), "example1.pdf")

# Create the report
rpt <- create_report(tmp, orientation = "portrait", output_type = "PDF") %>% 
  page_header(left = "Client: 1-800-Flowers", right = "Study: Iris") %>% 
  titles("Listing 1.0", "IRIS Data Listing") %>% 
  add_content(create_table(iris)) %>% 
  footnotes("* Bulletin of the American Iris Society, 1935") %>%
  page_footer(left = Sys.time(), 
              center = "Confidential", 
              right = "Page [pg] of [tpg]")

# Write the report
write_report(rpt)

# file.show(tmp)

```
<img src="../man/images/e1a.png" alt="Listing example page 1"/>
<img src="../man/images/e1b.png" alt="Listing example page 2"/>
<img src="../man/images/e1c.png" alt="Listing example page 3"/>

Next: [Example 2: Table](reporter-table.html)

---
title: "Example 5: Spanning Headers"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 5: Spanning Headers}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

### Create a Table with Spanning Headers

Spanning headers are a common requirement of regulatory tables, yet few 
reporting packages support them.  The **reporter** package supports any 
number of spanning levels, along with a simple and flexible 
syntax to define them.

```{r eval=FALSE, echo=TRUE}  
library(reporter)

# Create temporary path
tmp <- file.path(tempdir(), "example5.pdf")

# Prepare Data
dat <- mtcars[1:10, ]
df <- data.frame(vehicle = rownames(dat), dat)

# Define Table with spanning headers
tbl <- create_table(df) %>% 
  titles("Table 1.0", "MTCARS Spanning Headers") %>% 
  spanning_header(from = "mpg", to = "hp", label = "Span 1", n = 10) %>%
  spanning_header(from = "drat", to = "qsec", label = "Span 2", n = 10) %>%
  spanning_header(from = "vs", to = "carb", label = "Span 3", n = 10) %>%
  spanning_header(from = "drat", to = "carb", 
                  label = "Super Span", level = 2) %>%
  define(vehicle, label = "Vehicle") %>% 
  define(mpg, format = "%.1f") %>% 
  define(disp, visible = FALSE) %>% 
  define(am, visible = FALSE) %>% 
  footnotes("* Motor Trend, 1974")

# Create Report and add table 
rpt <- create_report(tmp, output_type = "PDF", 
                     font = "Courier", font_size = 12) %>%
  page_header(left = "Client: Motor Trend", right = "Study: Cars") %>% 
  add_content(tbl) %>% 
  page_footer(left = Sys.time(), 
            center = "Confidential", 
            right = "Page [pg] of [tpg]")

# Write the report
res <- write_report(rpt)

# file.show(tmp)

```
<img src="../man/images/e5.png" alt="Spanning header example image"/>

Next: [Example 6: Page Wrap](reporter-wrap.html)

---
title: "More Examples"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{More Examples}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

options(rmarkdown.html_vignette.check_title = FALSE)

```
## Report Gallery

The previous examples in the **reporter** documentation were intentionally
simplified to focus on the workings of a particular function.  It is helpful,
however, to also view **reporter** functions in the context of a complete 
program.  There is a gallery of examples included in the  
[sassy](https://sassy.r-sassy.org) package that illustrate
how **reporter** functions work together with other **sassy** functions to
create a complete report.  You can find the report gallery 
[here](https://sassy.r-sassy.org/articles/sassy.html).

---
title: "Example 13: Superscripts, Subscripts, and Symbols"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 13: Superscripts, Subscripts, and Symbols}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

### Adding Superscripts, Subscripts, and Symbols to a Report

It is very common to include superscripts, subscripts, and special symbols 
on a statistical report. The **reporter** package 
supports such features. You may add superscripts and subscripts
to your titles, footnotes, page header, page footer, and labels for your
column headers and spanning headers.  You can add them using a curly brace 
delimiter and some useful functions from the **common** package.  The **common**
package is loaded automatically with **reporter**.  Here is an example:

```{r eval=FALSE, echo=TRUE} 
library(reporter)
library(magrittr)

# Create temp file path
tmp <- file.path(tempdir(), "example13.rtf")

# Prepare Data
dat <- airquality[sample(1:153, 15), ]
dat$Month <-  as.Date(paste0("1973-", dat$Month, "-01"))

# Define table
tbl <- create_table(dat, show_cols = c("Month", "Day", "Wind", 
                                       "Temp", "Ozone")) %>%
  titles("Table 9.6", "Air Quality Sample Report{supsc('1')}",   # Superscript 1
         borders = c("top", "bottom"), blank_row = "none") %>%
  column_defaults(width = .7) %>%
  define(Month, format = "%B", align = "left", width = 1) %>%
  define(Temp, format = "%.0f") %>%
  footnotes("{supsc('1')}New York, May to September 1973",       # Superscript 1
            borders = c("top", "bottom"), blank_row = "none") 

# Define report
rpt <- create_report(tmp, output_type = "RTF", font = "Arial", 
                     font_size = 12, missing = "-") %>%
  page_header("Sponsor: EPA", "Study: B34958", blank_row = "below") %>% 
  add_content(tbl) %>% 
  page_footer(Sys.Date(), right = "Page [pg] of [tpg]")

# Write the report to the file system
write_report(rpt)

# View report
# file.show(tmp)

```
<img src="../man/images/super.png" alt="Superscript example image"/>


### Superscript and Subscript Functions

The `supsc()` function shown above is a superscript function. This function
will translate a normal character into a superscript character.  It
works by looking up a UTF-8 superscript character based on the passed parameter
string.  The function will attempt to translate each character into a UTF-8 
superscript equivalent.  All numbers and most Latin letters are available
in a superscript version.

The **common** package includes two more useful functions: `subsc()` and `symbol()`.
The `subsc()` function is similar to the superscript function, except it looks
up UTF-8 subscript characters. While UTF-8 subscript characters are more limited
than the superscripts, the function will translate all numbers appropriately.

Note that the `supsc()` and `subsc()` functions will translate the entire string.
Translating the entire string means multiple characters can be combined
to create complex superscript or subscript expressions.  Here is an
example of such an expression:

```{r eval=FALSE, echo=TRUE} 

# Create complex superscript expression
ex1 <- "x" %p% supsc("(n+4)")

# View expression
ex1
# [1] "x⁽ⁿ⁺⁴⁾"

```
### Symbol Function

Finally, the `symbol()` function will look up UTF-8 symbols.  The function
uses HTML style keywords to identify the symbol to look up.  The function 
supports keywords for trademarks, currencies, mathematical symbols, 
logical symbols, Greek letters, and more. Here are a few 
examples:

<img src="../man/images/symbols.png" alt="Symbols example image"/>


### Superscripts and Subscripts in Table Cells

To get superscripts or subscripts into table cells, they must
be added to the table data prior to reporting.  Simply use the standard R 
subset operators 
to append the subscript or superscript to the desired cell value.

Here is an example appending the superscript "two" into a cell in the second 
row, column "A", using the MTCARS data frame from the previous example:

```{r eval=FALSE, echo=TRUE} 

# Append superscript using paste0()
df[2, "A"] <- df[2, "A"] %p% supsc('2')

# View data frame
df
#    var      label           A           B
# 1 ampg          N          19          13
# 2 ampg       Mean 18.8 (6.5)²  22.0 (4.9)  # Observe superscript in column A
# 3 ampg     Median        16.4        21.4
# 4 ampg    Q1 - Q3 15.1 - 21.2 19.2 - 22.8
# 5 ampg      Range 10.4 - 33.9 14.7 - 32.4
# 6  cyl 8 Cylinder 10 ( 52.6%)  4 ( 30.8%)
# 7  cyl 6 Cylinder  4 ( 21.1%)  3 ( 23.1%)
# 8  cyl 4 Cylinder  5 ( 26.3%)  6 ( 46.2%)

```

### Additional Unicode Characters

The functions from the common package make it easy to add superscripts, 
subscripts, and many symbols to your reports.  However, these functions
do not include all possible special characters. If you need more characters, 
you may add any UTF-8 character to a report using the Unicode escape "\\U".
For example, the Unicode character for a superscript one (1) is "\\U00B9".
There are millions of Unicode characters.
These codes may be discovered via an internet search, and can be embedded
in any string.

### Output Type Limitations

The Unicode characters will work on all output types, with
one exception.  The PDF output type only supports superscripts for the numbers
one, two, and three, and a few other special characters. 
PDF will render unsupported characters
as a question mark.  If you need to output an unsupported character
with the PDF output type, it is recommended that you use parenthesis () or 
the carrot ^ character to signify a superscript or subscript, or use a another
supported symbol from the ASCII character set.


Next: [Page Break](reporter-break.html)

---
title: "Frequently Asked Questions"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Frequently Asked Questions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

options(rmarkdown.html_vignette.check_title = FALSE)

```

Below are some frequently asked questions about the **reporter** package. Click
on the links below to navigate to the full question and answer content.

## Index{#top}

* [What file formats does reporter support?](#types)
* [What fonts does reporter support?](#fonts)
* [What styling options are available in the reporter package?](#styles)
* [Is there a way to control page orientation?](#orientation)
* [Is there a way to control the units of measure?](#units)
* [Can I perform statistics with reporter?](#statistics)
* [Can I perform grouping with reporter?](#grouping)
* [Does reporter have a way to sort the report data?](#sort)
* [Does reporter have a way to filter the report data?](#filter)
* [Does reporter support superscripts and subscripts?](#super)
* [Do I have to include all the columns in my data on the final report?](#visible)
* [Is there a way to order the columns on my report?](#order)
* [Is there a way to format a variable?](#format)
* [Is there a way to control page breaks?](#breaks)
* [Is there a way to control page wraps?](#wraps)
* [Is there a way to set the width of a table?](#width)
* [Why am I getting an error when using curly braces in my footnotes?](#curly)
* [How can I get reporter to break/wrap long words?](#wordbreak)


## Content

### What file formats does reporter support? {#types}

**Q:** I can see from the examples that the **reporter** package can create
text reports.  Can it create any other types of reports?  PDF?  DOCX?


**A:** The **reporter** package supports TXT, RTF, PDF, HTML, and DOCX
output types.  

[top](#top) 

******


### What fonts does reporter support? {#fonts}

**Q:** I see a lot of Courier font in the examples.  Does **reporter** support
any other fonts?

**A:** The supported fonts depend on the output type selected.  Obviously for
TXT reports, no fonts are supported.  
The RTF, PDF, HTML, and DOCX output types support Courier, Arial, and Times New Roman.

[top](#top)

******

### What styling options are available in the reporter package? {#styles}

**Q:** The examples show little in the way of styling: fonts, colors, bolding, 
etc.  What options are there for adding styles to the report?

**A:** Styles and themes are now available for the HTML output type. 
Take a look at the [styles vignette](reporter-styles.html) for more information
about this exciting new feature.

For all other output types, the `titles()` function has parameters to bold the title
block, and set the font size independently of the main report.  Styling such
as that for HTML will
become available for other output types in the near future.

[top](#top) 

******

### Is there a way to control the page orientation? {#orientation}

**Q:** All of the examples I saw show a landscape page orientation.  Is there
a way to have a portrait page orientation?

**A:** Yes.  The `orientation` parameter on the `create_report()` function
accepts both "landscape" and "portrait" values.

[top](#top) 

******
### Is there a way to control the units of measure? {#units}

**Q:** All of the examples I saw show column widths specified in inches. 
Is there a way to specify them in metric?  What about by percentage?

**A:** The `units` parameter on the `create_report()` function allows you
to control the units of measure.  Valid values are "inches" and "cm" for 
centimeters.  The parameter accepts no other value.  The issue of setting
column widths by percentage has been raised, and is under consideration as
a future enhancement.

Note that this parameter sets the units of measure for every measurement 
on the report, not just column widths.  The units for margins, table width, 
plot width, etc. will all be understood to be in the units specified on the 
`units` parameter.  In other words, the whole report will be in either inches
or centimeters. You cannot mix measurement systems on a report.

[top](#top) 

******

### Can I perform statistics with reporter? {#statistics}

**Q:** With SAS® proc report, I can add statistics to my report.  Does the
**reporter** package have any statistical capabilities?

**A:** No.  The focus of the reporter package is on creating a printable
report in multiple output types.  There are many R packages to generate
statistics.  If you want statistics in your report, run your desired statistical
functions, and add the result to the input data.

[top](#top)

******

### Can I perform grouping with reporter? {#grouping}

**Q:** With SAS® proc report, I can group my data in the reporting procedure.
Does the **reporter** package have grouping capabilities?

**A:** No.  There are already several ways to group data in R.  It is best
to group data with one of the existing functions before sending into the 
**reporter** package.  Note that group rows can be emphasized on the
report using the `dedupe`, `blank_after`, and `label_row` parameters on 
the `define()` function. 
See [example 2](reporter-table.html) and [example 3](reporter-stub.html) 
from the vignettes for a demonstration 
on how to use these parameters.

[top](#top) 

******


### Does reporter have a way to sort the report data? {#sort}

**Q:** I have some data to report on, but it is not sorted the way I want 
for reporting. Does **reporter** have any way to sort the data?

**A:** No.  There are many ways to sort data in R. It is therefore 
recommended to sort the data before sending it into the **reporter** package.

Generally speaking, the input data should be fully prepared for reporting.
The **reporter** package will take your input data pretty much as-is and 
create a report in the desired output file format.  Any data manipulation
should be done prior to reporting.

[top](#top)

******

### Does reporter have a way to filter the report data? {#filter}

**Q:** I have some data to report on, but it is not filtered the way I want.
Does **reporter** have any way to filter the data?


**A:** No. There are many ways to filter data in R. It is therefore 
recommended to filter the data before sending it into the **reporter** package.

[top](#top) 

******

### Does reporter support superscripts and subscripts? {#super}

**Q:** I want to add some superscripts and subscripts to my report.  Does
reporter have a way to do that?


**A:** Yes. The reporting functions recognize Unicode superscript and subscript
characters.  These characters may be entered as a 4-digit Unicode code. 
For instance, the 
code "\\U00B2" specifies a superscript two.  See the 
[superscript and subscript vignette](reporter-super.html)
for further explanation on this topic.

[top](#top) 

******

### Do I have to include all the columns in my data on the final report? {#visible}

**Q:** I have some data to report on, but it has more columns than I want to
show in the report. Does **reporter** have any way to remove some columns
from the report?

**A:** Yes.  There are two ways to remove a column from the report.  The first
is to use the `show_cols` parameter on the `create_table()` function.  This 
parameter accepts a vector of column names to include on the table.  

The second way is to use the `visible` parameter on the `define()` function.
Setting this parameter to FALSE will hide that variable on the report.

[top](#top)

******

### Is there a way to order the columns on my report? {#order}

**Q:** The variables in my data are not ordered exactly as I want them
to appear on the report.  Is there a way to control column order?

**A:** Generally it is better to prepare your input data so that everything
is ordered the way you want prior to reporting. But there is a way
to order the columns if they are not in the desired order.  You can use
the `show_cols` parameter on the `create_table()` function to do this.
Simply pass a vector of column names to the `show_cols` parameter, and put
the column names in the order you want them to appear on the report.  

Note that if you are using `define()` functions, the order of these functions 
have no affect on the order of columns on the report. 

[top](#top)

******

### Is there a way to format a variable? {#format}

**Q:** I have variable in my input data that is not formatted.  I want to apply 
a format to it; say, round to 1 decimal place and add a percent sign.  Is there
a way to format a variable?

**A:** Yes. You can use the `format` parameter on the `define()` 
function to identify a format for a variable. The format passed to this 
parameter should be a valid format for the 
[`fapply()`](https://fmtr.r-sassy.org/reference/fapply.html) function from the 
[**fmtr**](https://fmtr.r-sassy.org) package. 
See the [`fapply()`](https://fmtr.r-sassy.org/reference/fapply.html)
documentation for additional information.

[top](#top)

******

### Is there a way to control page breaks? {#breaks}

**Q:** Overall, the **reporter** package does a good job of determining
when to break to the next page.  However, sometimes I want to control exactly
when the page break occurs.  Is there a way to do that?

**A:** Yes. You can add a variable to your input data that has the desired
page number for each row.  Then use the `page_break` parameter on the `define()` 
function to identify that variable as a page break variable.  The package
will then use that variable to break pages, instead of the automatic 
paging variable.

You can also set the number of lines on a page using the `line_count`
parameter on the `options_fixed()` function.  When this parameter is set,
it will use the number of lines specified to determine when to break a page, 
instead of using the automatically calculated value. 

[top](#top)

******

### Is there a way to control page wraps? {#wraps}

**Q:** Overall, the **reporter** package does a good job of determining
when to wrap a page when it is too wide.  However, sometimes I want to 
control exactly when the page wrap occurs.  Is there a way to do that?

**A:** Yes. You can use the `page_wrap` parameter on the `define()` 
function to identify a variable to wrap on. See [example 6](reporter-wrap.html) 
in the documentation for an illustration of how to wrap pages.

[top](#top)

******

### Is there a way to set the width of a table? {#width}

**Q:** I see there are several ways to set column widths.  However, I just 
want to set the width of the entire table, and have **reporter** figure out
the column widths proportionally.  Is there a way to do that?

**A:** Yes. You can use the `width` parameter on the `create_table()` 
function to set a width for the entire table. The width should be specified
in the units identified on the `units` parameter on the `create_report()`
function.  See the `create_table()` function documentation for more information
on the `width` parameter.

[top](#top)

******

### Why am I getting an error when using curly braces in my footnotes? {#curly}

**Q:** I'm trying to use curly braces in my footnotes.  For some reason, I
keep getting an error.  Why?

**A:** The curly braces are delimiter characters for the glue function, which
is automatically run on titles, footnotes, and labels. You can use the curly
braces to reference a variable or run a function from within string. To avoid
the error, you can escape the glue functionality by doubling the curly braces.
In other words, use "{{mystring}}" instead of "{mystring}.

You can also turn off the glue functionality globally by setting 
`globals("reporter.glue" = FALSE)`.

[top](#top)

******

### How can I get reporter to break/wrap long words? {#wordbreak}

**Q:** I have some very long words in my data.  I want **reporter** to break them,
and wrap to the next line.  Is there a way to do that? 

**A:** No.  The **reporter** package will not break words.
What you should do is pre-process the column with the long words, 
and break it yourself.  You can use the `nchar()` function to check the character
length of the words.  Then insert a new line or blank space at the point where
you want to break. Once the words are broken where you want, send the data
into **reporter**.

[top](#top)

******

<!-- ### Question 1? {#q1} -->

<!-- **Q:** I have a question? -->

<!-- **A:** Answer here. -->

<!-- [top](#top) -->

<!-- ****** -->

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_spec.R
\name{print.plot_spec}
\alias{print.plot_spec}
\title{Prints the plot spec}
\usage{
\method{print}{plot_spec}(x, ..., verbose = FALSE)
}
\arguments{
\item{x}{The plot spec.}

\item{...}{Additional parameters to pass to the underlying print function.}

\item{verbose}{Whether to print in verbose form.  Default is FALSE.}
}
\value{
The plot spec, invisibly.
}
\description{
A function to print the plot spec.
The \strong{print} function will print the plot spec in summary 
form.  To view all parameters, set the \code{verbose} parameter to TRUE.
}
\examples{
txt <- create_text("Lorem ipsum dolor sit amet, consectetur...")
print(txt)

# A text specification:
# - text: data.frame 'mtcars' 32 rows 11 cols
}
\seealso{
\code{\link{create_plot}} function to create a plot specification.

Other plot: 
\code{\link{create_plot}()}
}
\concept{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_spec.r
\name{page_header}
\alias{page_header}
\title{Add a page header}
\usage{
page_header(
  x,
  left = "",
  right = "",
  blank_row = "none",
  width = NULL,
  center = ""
)
}
\arguments{
\item{x}{The report object.}

\item{left}{The left page header text.  May be a single string or a vector
of strings.}

\item{right}{The right page header text.  May be a single string or a vector
of strings.}

\item{blank_row}{Whether to create a blank row below the page header.
Valid values are 'below' and 'none'.  Default is 'none'.}

\item{width}{Widths for left, center, and right columns of the page header, 
passed as a vector of double values in the report unit of measure.  If the vector
contains fewer than three widths, the widths passed will be interpreted from
left to right, and any unassigned column widths will be calculated from
the remaining width on the page.  For example, if two values are assigned, they will be
interpreted as left and center, and the right column width will be calculated.
Setting a cell width to zero (0) will remove that column from the page header
entirely. To set the left and right widths only, pass a zero for the center
cell, i.e. \code{width = c(6, 0, 3)} for a total width of 9 inches.}

\item{center}{The center page header text.  May be a single string or a vector
of strings. By default, the center column is not displayed.  To display it,
either assign a value on this parameter, or assign a width on the "width" parameter.}
}
\value{
The modified report specification.
}
\description{
This function adds a page header to the report.  The page header will appear
at the top of each page of the report.
}
\details{
The page header may contain text on the left or right. Use the appropriate
parameters to specify the desired text.  Only one page header is allowed
on a report. The page header will be repeated on every page of the report.
Multiple text values for each side
may be specified as a vector of strings.  

If the width of the page header
string exceeds the available space, an error will be generated. There is 
also a limit of 5 page header strings per each side.

There are two special tokens to generate page numbers: [pg] and [tpg]. 
Use [pg] to indicate the current page number.  Use [tpg] to indicate the
total number of pages in the report.  These tokens may be placed anywhere 
in the page header or page footer. 

Each header string must fit within the available space.  The \strong{reporter}
package will not wrap headers.  If a header string does not fit within the 
available space, an error will be generated.  In these situations, either
shorten the header string or split it into multiple headers that each fit 
within the available space.
}
\examples{
library(magrittr)

# Create temp file path
tmp <- file.path(tempdir(), "mtcars.txt")

dat <- data.frame(name = rownames(mtcars[1:10, ]), mtcars[1:10, 1:5], 
                  stringsAsFactors = FALSE)

# Create the report object
rpt <- create_report(tmp, orientation = "portrait") \%>\% 
  page_header("Client: Motor Trend", "Study: Cars") \%>\% 
  titles("MTCARS Sample Report") \%>\% 
  add_content(create_table(dat)) \%>\% 
  page_footer(Sys.time(), right = "Page [pg] of [tpg]")

# Write the report to the file system
write_report(rpt)

# Write report to console
writeLines(readLines(tmp, encoding = "UTF-8"))

# Client: Motor Trend                                                Study: Cars
#                              MTCARS Sample Report
# 
#                name                 mpg cyl   disp   hp  drat
#                ----------------------------------------------
#                Mazda RX4             21   6    160  110   3.9
#                Mazda RX4 Wag         21   6    160  110   3.9
#                Datsun 710          22.8   4    108   93  3.85
#                Hornet 4 Drive      21.4   6    258  110  3.08
#                Hornet Sportabout   18.7   8    360  175  3.15
#                Valiant             18.1   6    225  105  2.76
#                Duster 360          14.3   8    360  245  3.21
#                Merc 240D           24.4   4  146.7   62  3.69
#                Merc 230            22.8   4  140.8   95  3.92
#                Merc 280            19.2   6  167.6  123  3.92
# 
# ...
# 
# 2020-10-17 11:53:51                                                Page 1 of 1
}
\seealso{
Other report: 
\code{\link{add_content}()},
\code{\link{create_report}()},
\code{\link{footer_image}()},
\code{\link{footnotes}()},
\code{\link{header_image}()},
\code{\link{options_fixed}()},
\code{\link{page_by}()},
\code{\link{page_footer}()},
\code{\link{print.report_spec}()},
\code{\link{set_margins}()},
\code{\link{title_header}()},
\code{\link{titles}()},
\code{\link{write_report}()}
}
\concept{report}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_spec.r
\name{add_content}
\alias{add_content}
\title{Add content to a report}
\usage{
add_content(
  x,
  object,
  page_break = TRUE,
  align = "center",
  blank_row = "below"
)
}
\arguments{
\item{x}{The report_spec to append content to.}

\item{object}{The object to append.}

\item{page_break}{Whether to add a page break after the object. 
Valid values are TRUE or FALSE.  You can manipulate the \code{page_break} 
parameter to add multiple objects to the same page.}

\item{align}{How to align the content.  Valid values are 'left', 'right',
'center', and 'centre'.}

\item{blank_row}{Whether to put a blank row above or below the content.
Valid values are 'above', 'below', 'both', or 'none'.}
}
\value{
The modified report_spec.
}
\description{
This function adds an object to the report content list. A 
report will accept multiple pieces of content.  The \code{add_content} 
function also controls overall alignment of the content on the page, and
whether there is a page break before or after.
}
\details{
The \code{add_content} function adds a piece of content to a report. For a 
text report, valid objects are a table or text object.  For an RTF, PDF,  
HTML, or DOCX report, valid objects are a table, text, or plot object.  See 
\code{\link{create_table}}, \code{\link{create_text}}, or 
\code{\link{create_plot}} for further 
information on how to create content objects.  

Content will be
appended to the report in the order it is added.  By default, a page break
is added after the content.  You can stack two pieces of content together
closely by setting the \code{page_break} parameter to FALSE, and 
the \code{blank_row} parameter to "none".
}
\examples{
library(reporter)
library(magrittr)

# Create temp file path
tmp <- file.path(tempdir(), "mtcars.txt")

# Create first table 
tbl1 <- create_table(mtcars[1:5, 1:6]) \%>\% 
  column_defaults(width = .5) 

# Create second table
tbl2 <- create_table(mtcars[6:10, 1:6], headerless=TRUE) \%>\% 
  column_defaults(width = .5) 

# Create the report object
rpt <- create_report(tmp) \%>\%
  titles("MTCARS Sample Data", align = "left") \%>\%
  add_content(tbl1, page_break = FALSE, align = "left", blank_row = "none") \%>\%
  add_content(tbl2, page_break = FALSE, align = "left") \%>\%
  add_content(create_text("* NOTE: Above table is actually two tables stacked."))

# Write the report to the file system
res <- write_report(rpt)

# Write report to console
writeLines(readLines(tmp, encoding = "UTF-8"))

# MTCARS Sample Data
# 
#    mpg    cyl   disp     hp   drat     wt
# -----------------------------------------
#     21      6    160    110    3.9   2.62
#     21      6    160    110    3.9  2.875
#   22.8      4    108     93   3.85   2.32
#   21.4      6    258    110   3.08  3.215
#   18.7      8    360    175   3.15   3.44
#   18.1      6    225    105   2.76   3.46
#   14.3      8    360    245   3.21   3.57  
#   24.4      4  146.7     62   3.69   3.19
#   22.8      4  140.8     95   3.92   3.15
#   19.2      6  167.6    123   3.92   3.44
# 
# * NOTE: Above table is actually two tables stacked.
}
\seealso{
\code{\link{create_table}}, \code{\link{create_text}}, and
\code{\link{create_plot}} to create content for a report.

Other report: 
\code{\link{create_report}()},
\code{\link{footer_image}()},
\code{\link{footnotes}()},
\code{\link{header_image}()},
\code{\link{options_fixed}()},
\code{\link{page_by}()},
\code{\link{page_footer}()},
\code{\link{page_header}()},
\code{\link{print.report_spec}()},
\code{\link{set_margins}()},
\code{\link{title_header}()},
\code{\link{titles}()},
\code{\link{write_report}()}
}
\concept{report}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_spec.r
\name{create_report}
\alias{create_report}
\title{Create a report}
\usage{
create_report(
  file_path = "",
  output_type = "TXT",
  orientation = "landscape",
  units = "inches",
  paper_size = "letter",
  missing = "",
  font = "Courier",
  font_size = NULL
)
}
\arguments{
\item{file_path}{The output path of the desired report. Either a full path or
a relative path is acceptable.  This parameter is not required to create the
report_spec object, but will be required to write the report.  In addition, 
the file extension is not required.  If the file extension is not supplied,
the \code{\link{write_report}} function will add a file extension based
on the \code{output_type} specified.}

\item{output_type}{The report output type.  Default is "TXT".  Valid
values are "TXT", "RTF", "PDF", "HTML", and "DOCX".}

\item{orientation}{The page orientation of the desired report.  Valid values
are "landscape" or "portrait".  The default page orientation is "landscape".}

\item{units}{Specifies the units of measurement.  This setting will 
indicate the units for columns widths, margins, paper size, and other 
measurements. Valid values are "inches" or "cm" (centimeters).  
Default value is "inches".}

\item{paper_size}{The expected paper size on which the report may be 
printed.  The \code{paper_size} will determine how much text can fit on
one page.  Valid values are "letter", "legal", "A4", and "RD4".  Default is 
"letter".  For the HTML output type, a paper size of "none" is also valid. 
That means the HTML will be generated in an unbounded manner as a typical
web page.  For a custom page size, the parameter also accepts a double vector
with the page width and height.  The width and height should be in the report
units of measure, and assume a portrait orientation.}

\item{missing}{How to display missing values in the report.  Default is
to replace them with an empty string, which removes them from the report.
To display missing values as is, set the missing parameter to NULL.  To
replace missing values with a character string (such as ".", "-", or "<NA>")
pass the desired character string to the missing parameter.}

\item{font}{The font to use on the report.  The font specified will be
used for the entire report.  Valid values are "Courier", "Arial", "Times",
and "fixed".  The value of "fixed" will create a fixed-width, text style
report in Courier font.  The \code{font} parameter only applies to 
RTF, HTML, PDF, and DOCX reports.  The default value is "Courier".}

\item{font_size}{The size of the font to use on the report. The \code{font_size}
specified will be used for the entire report.  Valid values are 8, 9, 10, 11,
and 12.  The \code{font_size} parameter only applies to RTF, PDF, HTML, and  
DOCX output types. The default value is 10.}
}
\value{
A new report_spec object.
}
\description{
Creates a report shell to which you may add titles, footnotes, content, etc.
}
\details{
This function is the constructor for the report object.  The report
object contains information needed to create a report. The object is
defined as an S3 object, and has a class of 'report_spec'.

The report object holds information concerning report page size, orientation, 
titles, footnotes, page header, page footer, margins, and other options.  
Use the \code{\link{add_content}} function to add content to the report.  
The report may be written to a file using the \code{\link{write_report}} 
function.
}
\section{Report family of functions}{

The report is the primary container for report specifications.  The
following functions add additional specifications to the report object 
initialized with \code{create_report}.
\itemize{
  \item \code{\link{titles}} to add titles to the report.
  \item \code{\link{footnotes}} to add footnotes to the report.
  \item \code{\link{title_header}} to add a title header to the report.
  \item \code{\link{page_header}} to add a page header to the report. 
  \item \code{\link{page_footer}} to add a page_footer to the report. 
  \item \code{\link{add_content}} to add content to the report.
  \item \code{\link{options_fixed}} to set options for fixed-width output.
  \item \code{\link{add_content}} to add content to the report.
  \item \code{\link{write_report}} to write the report to the file system.
}

The report family of functions are pipe-friendly.  After creating the 
report, you may pipe the object to any of the above functions to append
additional options.
}

\examples{
library(reporter)
library(magrittr)

# Create temp file path
tmp <- file.path(tempdir(), "airquality.txt")

# Prepare Data
dat <- airquality[sample(1:153, 15), ]
dat$Month <-  as.Date(paste0("1973-", dat$Month, "-01"))

# Define table
tbl <- create_table(dat, show_cols = c("Month", "Day", "Wind", "Temp", "Ozone")) \%>\% 
  titles("Table 9.6", "Air Quality Sample Report") \%>\% 
  column_defaults(width = .5) \%>\% 
  define(Month, format = "\%B", align = "left", width = 1) \%>\% 
  define(Temp, format = "\%.0f") \%>\% 
  footnotes("* New York, May to September 1973")

# Define report 
rpt <- create_report(tmp, orientation = "portrait",  missing = "-") \%>\% 
  add_content(tbl) 

# Write the report to the file system
write_report(rpt)

# Write the report to the console
writeLines(readLines(tmp, encoding = "UTF-8"))

#                      Table 9.6
#              Air Quality Sample Report
#      
#      Month           Day   Wind   Temp  Ozone
#      ----------------------------------------
#      July              8    6.3     92     97
#      July              9    5.7     92     97
#      August            1    6.9     81     39
#      July             23   11.5     82      -
#      June              9   13.8     90     71
#      July             12   14.3     73     10
#      July              4   10.9     84      -
#      May              31    7.4     76     37
#      September        30   11.5     68     20
#      June             25      8     75      -
#      June             28   11.5     80      -
#      August           18    7.4     76     23
#      June             20   10.3     76     13
#      July              1    4.1     84    135
#      May              23    9.7     61      4
#      
#      * New York, May to September 1973
}
\seealso{
\code{\link{create_table}}, \code{\link{create_text}}, and
\code{\link{create_plot}} functions
to create content for the report.

Other report: 
\code{\link{add_content}()},
\code{\link{footer_image}()},
\code{\link{footnotes}()},
\code{\link{header_image}()},
\code{\link{options_fixed}()},
\code{\link{page_by}()},
\code{\link{page_footer}()},
\code{\link{page_header}()},
\code{\link{print.report_spec}()},
\code{\link{set_margins}()},
\code{\link{title_header}()},
\code{\link{titles}()},
\code{\link{write_report}()}
}
\concept{report}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_style.R
\name{add_style}
\alias{add_style}
\title{Add a style object to a report.}
\usage{
add_style(rpt, style = NULL, theme = NULL)
}
\arguments{
\item{rpt}{The report specification to add a style to.}

\item{style}{A style object which contains style settings to add to the 
report. This parameter is optional.  Default is NULL.}

\item{theme}{A theme name to use for this report. Valid values are
"MidnightBlue", "SteelBlue", "DarkRed", "SeaGreen", "SlateGrey", "Plain", 
and "SASDefault".  Default is NULL.}
}
\description{
This function will add a style object to a report specification.  The 
style may be added either by passing a style object to the \code{style} 
parameter, or by passing a theme name to the \code{theme} parameter.
}
\details{
The \code{add_style()} function allows you to add styling to HTML reports.
This functionality will apply to additional output types in future
versons of the \strong{reporter} package.

Styling can be added by passing a named theme to the \code{theme} parameter,
or by creating a style object using the \code{\link{create_style}} function
and passing that object to the \code{style} parameter.  You may also export 
a theme as a style object using the \code{\link{get_theme}} function,
modify it, and pass that to the \code{style} parameter.
}
\section{Style Specifications}{

The style specification is created using the \code{\link{create_style}} function,
and provides the most styling flexibility.  The style object allows you
to control background colors, font colors, border colors, and more.  Colors
can be specified using an RGB hex code, or an HTML/CSS-compliant color name.
See the \code{\link{create_style}} documentation for additional information.
}

\section{Themes}{

There are currently seven themes available: "MidnightBlue", "SteelBlue", 
"DarkRed", "SeaGreen", "SlateGrey", "Plain", and "SASDefault". 
All themes use Arial 10pt font. These themes are intended to provide a 
basic set of examples on how to create styles. Themes can be used by 
passing the theme name to the theme parameter on the \code{add_style()}
function. The theme will assign a variety of style settings according 
to the specifics of the theme.  For example, the "MidnightBlue" theme
sets the title font and header background colors to "MidnightBlue" and sets
the border color to "Grey".

To view theme style settings, you can use \code{\link{get_theme}} function. 
This function will return the theme as a style object.  See
the \code{\link{get_theme}} documentation for further details.
}

\examples{
library(reporter)
library(magrittr)

# Prepare data
dat <- as.data.frame(HairEyeColor)
dat <- dat[dat$Freq >= 10, ]

## Example 1: Use Pre-defined Theme ##

# Create temp file path
tmp1 <- file.path(tempdir(), "HairAndEyes1.html")

# Create table object
tbl <- create_table(dat, borders = "outside") \%>\% 
titles("Hair and Eye Colors with Theme") \%>\% 
column_defaults(width = .6)

# Create report and add theme
rpt <- create_report(tmp1, output_type = "HTML") \%>\% 
       add_content(tbl) \%>\% 
       add_style(theme = "SteelBlue")

# Write out the report        
write_report(rpt)

# Uncomment to View report
# file.show(tmp1)

#' ## Example 2: Create Custom Style ##

# Create temp file path
tmp2 <- file.path(tempdir(), "HairAndEyes2.html")

# Define custom style
sty <- create_style(font_name = "Times",
                    font_size = 10,
                    title_font_size = 12,
                    title_font_bold = TRUE,
                    title_font_color = "Blue",
                    table_header_background = "Blue",
                    table_header_font_bold = TRUE,
                    table_header_font_color = "White",
                    table_body_background = "White",
                    table_body_stripe = "Red")

# Create table object
tbl <- create_table(dat, borders = "outside") \%>\% 
titles("Hair and Eye Colors with Style") \%>\% 
column_defaults(width = .6)

# Create report and add style spec
rpt <- create_report(tmp2, output_type = "HTML") \%>\% 
       add_content(tbl) \%>\% 
       add_style(style = sty)

# Write out the report        
write_report(rpt)

# Uncomment to View report
# file.show(tmp2)
}
\seealso{
Other styles: 
\code{\link{create_style}()},
\code{\link{get_theme}()},
\code{\link{print.style_spec}()}
}
\concept{styles}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_spec.r
\name{set_margins}
\alias{set_margins}
\title{Set page margins}
\usage{
set_margins(x, top = NULL, bottom = NULL, left = NULL, right = NULL)
}
\arguments{
\item{x}{The report spec object.}

\item{top}{The top margin.}

\item{bottom}{The bottom margin.}

\item{left}{The left margin.}

\item{right}{The right margin.}
}
\value{
The report_spec with margins set as desired.
}
\description{
Sets the page margins for the report.  The units for this 
parameter can be inches or centimeters, depending on the units of measure 
specified on the \code{\link{create_report}} function.
}
\details{
The margins set with \code{set_margins} will be used for the entire report.  
Units for the margins
are specified by the \code{units} parameter on the 
\code{\link{create_report}} function.  Available units are 'inches' and 'cm'.
When the unit of measure is inches, default margins are 1 inch on the left and 
right, and .5 inches on top and bottom.  When the unit of measure is 
centimeters, default margins are 2.54 cm on left and right, and 1.27 cm
on top and bottom.

Note that when using output type of TXT, and not using the 
\code{blank_margins} option, setting the margins only reduces
the area available for content on a page.  You must still set the actual
margins on the available editor to match those specified in 
\code{set_margins}.  Any mismatch may result in content not fitting properly
on the page. For best results, set the right and bottom margins to zero 
to allow for slight overflow without causing a page break or wrapping lines.
}
\examples{
library(reporter)
library(magrittr)

# Create a temporary file
tmp <- file.path(tempdir(), "bod.txt")

# Define table
tbl <- create_table(BOD, width = 2.5) \%>\% 
  titles("Table 3.6", "BOD¹ Sample Report") \%>\% 
  define(Time, format = "Day \%s", align = "left") \%>\% 
  define(demand, format = "\%2.1f mg/l", label = "Demand") \%>\% 
  footnotes("¹ Biochemical Oxygen Demand")
       
# Define report #1 - No blank margins
rpt <- create_report(tmp, orientation="portrait") \%>\%
  add_content(tbl, align = "left") 

# Write the report
write_report(rpt)

# Write report to console
writeLines(readLines(tmp, encoding = "UTF-8"))

#           Table 3.6
#       BOD* Sample Report
# 
# Time                  Demand
# ----------------------------
# Day 1               8.3 mg/l
# Day 2              10.3 mg/l
# Day 3              19.0 mg/l
# Day 4              16.0 mg/l
# Day 5              15.6 mg/l
# Day 7              19.8 mg/l
# 
# * Biochemical Oxygen Demand


# Define report #2 - blank margins
rpt <- create_report(tmp, orientation="portrait") \%>\%
  options_fixed(blank_margins = TRUE) \%>\% 
  set_margins(top = .5, left = 1) \%>\% 
  add_content(tbl, align = "left") 

# Write the report
write_report(rpt)

# Write report to console
writeLines(readLines(tmp, encoding = "UTF-8"))

# 
# 
# 
#                       Table 3.6
#                   BOD* Sample Report
# 
#              Time                  Demand
#              ----------------------------
#              Day 1               8.3 mg/l
#              Day 2              10.3 mg/l
#              Day 3              19.0 mg/l
#              Day 4              16.0 mg/l
#              Day 5              15.6 mg/l
#              Day 7              19.8 mg/l
# 
#              * Biochemical Oxygen Demand
}
\seealso{
Other report: 
\code{\link{add_content}()},
\code{\link{create_report}()},
\code{\link{footer_image}()},
\code{\link{footnotes}()},
\code{\link{header_image}()},
\code{\link{options_fixed}()},
\code{\link{page_by}()},
\code{\link{page_footer}()},
\code{\link{page_header}()},
\code{\link{print.report_spec}()},
\code{\link{title_header}()},
\code{\link{titles}()},
\code{\link{write_report}()}
}
\concept{report}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_spec.r
\name{print.table_spec}
\alias{print.table_spec}
\title{Prints the table spec}
\usage{
\method{print}{table_spec}(x, ..., verbose = FALSE)
}
\arguments{
\item{x}{The table spec.}

\item{...}{Additional parameters to pass to the underlying print function.}

\item{verbose}{Whether to print in verbose form, which is similar to 
a list.  Default is FALSE, which prints in summary form.}
}
\value{
The table spec, invisibly.
}
\description{
A function to print the table spec.
The \strong{print} function will print the table spec in summary 
form.  To view all parameters, set the \code{verbose} parameter to TRUE.
}
\examples{
library(magrittr)

# Create Table
tbl <- create_table(mtcars) \%>\% 
  define(mpg, label = "Miles Per Gallon", width = .5) \%>\% 
  define(cyl, label = "Cylinders") \%>\% 
  titles("Table 6.4", "MTCARS Sample Table") \%>\% 
  footnotes("* Motor Trend, 1974")
  
tbl

# A table specification:
# - data: data.frame 'mtcars' 32 rows 11 cols
# - show_cols: all
# - use_attributes: all
# - title 1: 'Table 6.4'
# - title 2: 'MTCARS Sample Table'
# - footnote 1: '* Motor Trend, 1974'
# - define: mpg 'Miles Per Gallon' width=0.5 
# - define: cyl 'Cylinders' 
}
\seealso{
\code{\link{create_table}} function to create a table specification.

Other table: 
\code{\link{column_defaults}()},
\code{\link{create_table}()},
\code{\link{define}()},
\code{\link{spanning_header}()},
\code{\link{stub}()}
}
\concept{table}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/text_spec.R
\name{create_text}
\alias{create_text}
\title{Create text content}
\usage{
create_text(txt, width = NULL, align = "left", borders = "none")
}
\arguments{
\item{txt}{The text to create.}

\item{width}{The width of the text in the specified units of measure.  If 
no width is specified, the full page width will be used.}

\item{align}{How to align the text within the content area.  Valid values
are 'left', 'right', 'center', or 'centre'.  Default is 'left'.}

\item{borders}{Whether and where to place a border. Valid values are 'top',
'bottom', 'left', 'right', 'all', 'none', and 'outside'.  
Default is 'none'.  The 'left', 'right', and 'outside' 
border specifications only apply to RTF reports.}
}
\value{
The text specification.
}
\description{
Function to create a text specification that can be 
added as content to a report.   The text content can be used to include 
analysis on a statistical report.  A text specification is an S3 
object of class 'text_spec'.
}
\details{
To add plain text to a report, use the \code{create_text} function.  The 
function allows you to set a width and alignment for the text.  The
function will preserve any other formatting you apply to the text.  See
the \code{\link{add_content}} function to control page breaking and 
blanks spaces above or below the text.  

The text specification also accepts titles and footnotes.  See the 
\code{\link{titles}} and \code{\link{footnotes}} functions for further 
details.
}
\examples{
library(reporter)
library(magrittr)

# Create temp file path
tmp <- file.path(tempdir(), "mtcars.txt")

# Create dummy text
dt <- paste0("Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
  "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
  "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris ",
  "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ", 
  "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla ",
  "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa ",
  "qui officia deserunt mollit anim id est laborum.")

# Create the text object
txt <- create_text(dt) \%>\% 
  titles("Text Content 1.0", "Sample Text Report") \%>\% 
  footnotes("* Cicero, 1st century BCE")

# Create the report object
rpt <- create_report(tmp, orientation = "portrait") \%>\% 
  add_content(txt) 

# Write the report to the file system
write_report(rpt)

# Write the report to console
writeLines(readLines(tmp, encoding = "UTF-8"))

#                                Text Content 1.0
#                               Sample Text Report
# 
# Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor
# incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis
# nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.
# Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore
# eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt
# in culpa qui officia deserunt mollit anim id est laborum.
# 
# * Cicero, 1st century BCE
#
}
\seealso{
\code{\link{titles}} to add a title block to the text,  
\code{\link{footnotes}} to add footnotes, and \code{\link{add_content}} 
to add the text object to a report.

Other text: 
\code{\link{print.text_spec}()}
}
\concept{text}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_spec.r
\name{titles}
\alias{titles}
\title{Adds a title block}
\usage{
titles(
  x,
  ...,
  align = "center",
  blank_row = "below",
  borders = "none",
  width = NULL,
  bold = FALSE,
  font_size = NULL,
  header = FALSE,
  columns = 1
)
}
\arguments{
\item{x}{The object to assign titles to.  Valid objects are a report, or
a table, text, or plot specification.}

\item{...}{A set of title strings.}

\item{align}{The position to align the titles.  Valid values are 'left', 
'right', 'center' or 'centre'.  For titles, the default is 'center'.}

\item{blank_row}{Where to place a blank row.  Valid values are 'above',
'below', 'both', or 'none'.  Default is "below".}

\item{borders}{Whether and where to place a border. Valid values are 'top',
'bottom', 'left', 'right', 'outside', 'inside', 'all', or 'none'.  
Default is "none".  
The 'left', 'right', 'outside', and 'inside' border specifications 
only apply to RTF, HTML, PDF and DOCX reports.}

\item{width}{The width of the titles block.  If the titles are attached
to the report, valid values are 'page' or a numeric width, and the default
is 'page'. If the titles are attached to the
table, plot, or text content, the valid values are 'page', 'content' or a 
numeric value, and the default is 'content'.  The value 'content' means the 
titles will be aligned to the width of the table, plot, or text
content.  The value
'page' means the titles will be aligned to the width of the page. 
In addition to these two convenience settings, you 
may also specify a specific width in the current units of measure.  The
units of measure is determined by the 'units' parameter on 
\code{\link{create_report}}.}

\item{bold}{A parameter to bold the titles.  Valid values are TRUE and FALSE.
Default is FALSE.  This parameter only applies to variable-width RTF, HTML, 
PDF, and DOCX output types.}

\item{font_size}{The font size to use for the title block.  The font size
of the report will be used by default.  Valid values are 8, 9, 10, 11, 12,
13, and 14.  This parameter only applies to variable-width RTF, HTML, PDF,
and DOCX output types.}

\item{header}{Whether to put the titles in the page header. Valid values
are TRUE and FALSE.  Default is FALSE. This option only works on the RTF 
and DOCX output types, and only applies to titles assigned 
to the report object.  Titles in the page header will appear on every page, 
and be the same throughout the report.}

\item{columns}{The number of columns for the title block. Valid values
are 1, 2, and 3.  Default is 1.  If this parameter is set to 2, the title
block will be split into two columns, each aligned to the outside.  If 
this parameter is set to 3, the title block will be split into 3 columns,
with the outer columns aligned to the outside and the middle column
aligned center.  Titles are assigned to cells from top left to bottom right.}
}
\value{
The modified report.
}
\description{
This function adds one or more titles to an object as a title block.  
If added to a report, 
the titles will be added to
the page template, and thus appear on each page of the report. Titles may 
also be added to a table, text, or plot object.
}
\details{
The titles function accepts a set of strings of the desired title text. To
specify multiple titles for the block, pass them to the function 
as separate strings.

The titles may be aligned center, left or right using the align parameter.
The alignment will be applied to all titles contained in the 
block.  To control alignment of titles separately for each title, use 
multiple titles functions.

Titles may be assigned to a report, a table, a text specification, or a plot. 
If assigned to the report, the title will appear at the top of the page, and
be repeated for every page of the report.  If the titles are assigned to  
content, the titles will appear above the content, and be repeated if the 
content breaks to the next page.  

If titles are assigned to the report,
alignment will be oriented to the page body.  If titles are assigned to
content, alignment will be oriented to the edge of the content.

One title function accepts up to 10 titles.  However, multiple title 
blocks may be added to the same object if needed.  

Blank rows above or below the title block may be controlled using the 
\code{blank_row} parameter.

Each title string must fit within the available space.  The \strong{reporter}
package will not wrap titles on fixed-width reports.  
If a title does not fit within the 
available space, a warning will be generated and the title will be 
truncated.  In these situations, either
shorten the title or split it into multiple titles that each fit within the
available space.
}
\examples{
library(reporter)
library(magrittr)

# Create a temporary file
tmp <- file.path(tempdir(), "expenses.txt")

# Prepare data
dat <- data.frame(category = rownames(USPersonalExpenditure),
                  USPersonalExpenditure, stringsAsFactors = FALSE)

# Define table
tbl <- create_table(dat) \%>\% 
  titles("Table 1.0", "US Personal Expenditures from 1940 - 1960") \%>\% 
  column_defaults(from = X1940, to = X1960, width = .6, format = "$\%.2f") \%>\%
  define(category, label = "Category") \%>\% 
  define(X1940, label = "1940") \%>\% 
  define(X1945, label = "1945") \%>\% 
  define(X1950, label = "1950") \%>\% 
  define(X1955, label = "1955") \%>\% 
  define(X1960, label = "1960") \%>\% 
  footnotes("* In billions of dollars")

# Define report
rpt <- create_report(tmp, orientation="portrait") \%>\%
  add_content(tbl) 

# Write the report
write_report(rpt)

# Display in console
writeLines(readLines(tmp, encoding = "UTF-8"))
#                               Table 1.0
#               US Personal Expenditures from 1940 - 1960
# 
#     Category                1940    1945    1950    1955    1960
#     ------------------------------------------------------------
#     Food and Tobacco      $22.20  $44.50  $59.60  $73.20  $86.80
#     Household Operation   $10.50  $15.50  $29.00  $36.50  $46.20
#     Medical and Health     $3.53   $5.76   $9.71  $14.00  $21.10
#     Personal Care          $1.04   $1.98   $2.45   $3.40   $5.40
#     Private Education      $0.34   $0.97   $1.80   $2.60   $3.64
# 
#     * In billions of dollars
}
\seealso{
Other report: 
\code{\link{add_content}()},
\code{\link{create_report}()},
\code{\link{footer_image}()},
\code{\link{footnotes}()},
\code{\link{header_image}()},
\code{\link{options_fixed}()},
\code{\link{page_by}()},
\code{\link{page_footer}()},
\code{\link{page_header}()},
\code{\link{print.report_spec}()},
\code{\link{set_margins}()},
\code{\link{title_header}()},
\code{\link{write_report}()}
}
\concept{report}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_style.R
\name{print.style_spec}
\alias{print.style_spec}
\title{Prints a style specification}
\usage{
\method{print}{style_spec}(x, ..., verbose = FALSE)
}
\arguments{
\item{x}{A style object to print.}

\item{...}{Any follow-on parameters to pass to print().}

\item{verbose}{If verbose is TRUE, the function will print the style
object as a list. Otherwise, the object will print using the custom
print function.  The custom print is more compact than the verbose style 
print. Default is FALSE.}
}
\description{
This function will print a style object to the console.  The print function
will display each style setting that has been assigned, and the value
which is assigned.
}
\examples{
library(reporter)

# Get theme
tm <- get_theme("SteelBlue")

# View theme settings
print(tm)
## A style specification: 
#- font_name: 'Arial'
#- font_size: 10
#- text_color: 'DimGrey'
#- title_font_size: 11
#- title_font_bold: TRUE
#- title_font_color: 'SteelBlue'
#- border_color: 'Grey'
#- table_header_background: 'SteelBlue'
#- table_header_font_bold: TRUE
#- table_header_font_color: 'LightGrey'
#- table_body_background: 'White'
#- table_body_stripe: 'WhiteSmoke'
#- table_stub_background: 'SteelBlue'
#- table_stub_font_color: 'LightGrey'
#- table_stub_font_bold: TRUE
}
\seealso{
Other styles: 
\code{\link{add_style}()},
\code{\link{create_style}()},
\code{\link{get_theme}()}
}
\concept{styles}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_spec.r
\encoding{UTF-8}
\name{options_fixed}
\alias{options_fixed}
\title{Set options for a fixed-width report}
\usage{
options_fixed(
  x,
  editor = NULL,
  cpuom = NULL,
  lpuom = NULL,
  min_margin = NULL,
  blank_margins = FALSE,
  font_size = NULL,
  line_size = NULL,
  line_count = NULL,
  uchar = "¯"
)
}
\arguments{
\item{x}{The report spec.}

\item{editor}{The expected text editor to use for printing text reports.
Assigning this parameter will set the \code{cpuom} and \code{lpuom} 
parameters appropriately for the text editor.  Valid values are 'notepad',
'word', 'wordpad', 'notepad++', and 'editplus'.  If the editor parameter 
is used, any settings for \code{cpuom} and \code{lpuom} will be 
ignored. It is not necessary to set this parameter for RTF and PDF reports.}

\item{cpuom}{Characters per unit of measure of printed text.
If units is inches, the default is 12.  If units is centimeters (cm), the 
default is 4.687.  This value will be used to 
determine how many characters can fit on a line.}

\item{lpuom}{Lines per unit of measure of the printed text. Default for 
inches is 6. The default for centimeters (cm) is 2.55.  This value 
will be used to determine the number of lines that can fit on a page.}

\item{min_margin}{The editor minimum margin.  This parameter normally
defaults to 0, but may be set for some types of editors.}

\item{blank_margins}{When this option is TRUE, \strong{reporter} will use blank 
spaces and blank rows to create left and top margins, rather than rely 
on the editor to set margins.  When used, editor margins
should be set to zero.  Valid values are TRUE and FALSE. Default is
FALSE.  This option is only valid for \code{output_type = 'TXT'}.}

\item{font_size}{The size of the font in points.  Default is 10pt.  This
option is only valid for output types RTF and PDF.  Valid values are 8, 9, 10, 
11, and 12.}

\item{line_size}{The number of characters that will fit on a line.  Normally,
the \code{line_size} is calculated based on the page size, font size, and cpuom.
You can override the calculated value by setting the \code{line_size}
directly.}

\item{line_count}{The number of lines that will fit on page.  Normally,
the \code{line_count} is calculated based on the page size, font size, and lpuom.
You can override the calculated value by setting the \code{line_count}
directly.}

\item{uchar}{The character to use for underlines on the table 
header and spanning headers.  Default is a Unicode macron character #U00AF.
You may use a dash or underscore if your editor does not support
Unicode.  The \code{uchar} is forced to a dash for PDF output, 
as the LaTeX converter does not support the macron character.}
}
\value{
The updated report spec.
}
\description{
This function sets the options for a report  
with a fixed width font.
}
\details{
The \code{options_fixed} function sets options for reports 
with a fixed-width, monospace font.  These reports are based off a 
text report, but may be output as type "RTF" or "PDF".
}
\section{Text Reports}{

The \code{options_fixed} function sets
the characters per 
unit of measure (\code{cpuom}) and lines per unit of measure
(\code{lpuom}) settings for the report.  These settings determine how 
many characters and lines will fit within one unit of measure (uom), as 
specified on the \code{\link{create_report}} function.  These settings are
important to ensure the report content stays within the available page size 
and margins.  Because every editor allows a different number of 
characters and lines on a page, these settings must be adjusted depending
on the editor.  

The \code{options_fixed} function provides a shortcut 
\code{editor} parameter
to directly specify a popular editor.  If this parameter is specified, the
function will set the characters per unit of measure and lines per
unit of measure for you.  If the editor is not available in the 
\code{editor} parameter selections, for best results, you should 
set the \code{cpuom} and 
\code{lpuom} parameters manually.  To determine your \code{cpuom}
and \code{lpuom}, see the help for \code{\link{write_registration_file}}.

Alternatively, using the \code{options_fixed} function, 
you may set the \code{line_size} and \code{line_count} directly.  Note that
the \code{line_size} and \code{line_count} may be different for different
output types and editors. 

The \code{min_margin} parameter is used to set the minimum margin allowed
by the editor.  This value will be subtracted from the margin settings 
when the \code{blank_margins} option is used. It is useful for 
editors that do not calculate margins from the edge of the page.

As some editors do not support Unicode characters, it may be necessary 
to change the character used for the header and spanning header underlines.
The default character is a Unicode #U00AF macron.  The macron is sometimes
referred to as an "overline", since it is located at the top of the 
character area.  If your editor does not support Unicode, the macron
will not be displayed properly.  In this case, change the underline character
to a dash ("-") or an underscore ("_") using the \code{uchar} parameter.
}

\section{RTF and PDF Reports}{

For RTF and PDF reports, most of the parameters on the \code{options_fixed}
function do not apply.  For RTF and PDF reports, these parameters will
be set automatically, and cannot be changed.  

Some of the \code{options_fixed} function apply only to RTF and PDF.
In particular, the \code{font_size} parameter applies only to RTF and PDF
reports.  Valid font size options are 8, 9, 10, 11, and 12.  The font size may
also be set on the \code{\link{create_report}} function.
}

\examples{
library(reporter)
library(magrittr)

# Create a temporary file
tmp <- file.path(tempdir(), "bod.txt")

# Define table
tbl <- create_table(BOD, width = 2.5) \%>\% 
  titles("Table 3.6", "BOD* Sample Report") \%>\% 
  define(Time, format = "Day \%s", align = "left") \%>\% 
  define(demand, format = "\%2.1f mg/l", label = "Demand") \%>\% 
  footnotes("* Biochemical Oxygen Demand")
       
# Define report #1 - No blank margins
rpt <- create_report(tmp, orientation="portrait") \%>\%
  add_content(tbl, align = "left") 

# Write the report
write_report(rpt)

# Write report to console
writeLines(readLines(tmp, encoding = "UTF-8"))

#           Table 3.6
#       BOD* Sample Report
# 
# Time                  Demand
# ----------------------------
# Day 1               8.3 mg/l
# Day 2              10.3 mg/l
# Day 3              19.0 mg/l
# Day 4              16.0 mg/l
# Day 5              15.6 mg/l
# Day 7              19.8 mg/l
# 
# * Biochemical Oxygen Demand


# Define report #2 - blank margins
rpt <- create_report(tmp, orientation="portrait") \%>\%
  options_fixed(blank_margins = TRUE) \%>\% 
  set_margins(top = .5, left = 1) \%>\% 
  add_content(tbl, align = "left") 

# Write the report
write_report(rpt)

# Write report to console
writeLines(readLines(tmp, encoding = "UTF-8"))

# 
# 
# 
#                       Table 3.6
#                   BOD* Sample Report
# 
#              Time                  Demand
#              ----------------------------
#              Day 1               8.3 mg/l
#              Day 2              10.3 mg/l
#              Day 3              19.0 mg/l
#              Day 4              16.0 mg/l
#              Day 5              15.6 mg/l
#              Day 7              19.8 mg/l
# 
#              * Biochemical Oxygen Demand
}
\seealso{
\code{\link{create_report}} to create a report and set the unit
of measure, \code{\link{write_registration_file}} to determine the 
characters and lines per unit of measure manually.

Other report: 
\code{\link{add_content}()},
\code{\link{create_report}()},
\code{\link{footer_image}()},
\code{\link{footnotes}()},
\code{\link{header_image}()},
\code{\link{page_by}()},
\code{\link{page_footer}()},
\code{\link{page_header}()},
\code{\link{print.report_spec}()},
\code{\link{set_margins}()},
\code{\link{title_header}()},
\code{\link{titles}()},
\code{\link{write_report}()}
}
\concept{report}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter.R
\name{NotesOnPDF}
\alias{NotesOnPDF}
\title{Notes on PDF output type}
\description{
PDF output supports a limited character set.  Some Unicode characters
can cause the PDF rendering to fail.  All Latin characters and other 
common characters like 
mathematical operators and Greek letters are supported.  However, 
other special characters and characters from Asian languages may not 
be supported, depending on the operating system and installed operating 
system languages. If a character is not supported, it will be replaced
with a question mark (?).

As of \strong{reporter} v1.1.3, the package will 
generate PDF files directly, without using \strong{rmarkdown} as an 
intermediary. This technique eliminates dependencies, makes the package 
easier to install, and greatly improves performance.  It is recommended
to use the most recent version of the package if possible.

For versions of \strong{reporter} prior to v1.1.3, the package has
additional dependencies and installation requirements.  See details below.
}
\details{
For \strong{reporter} v1.1.2 and below, 
PDF output type requires the \strong{rmarkdown} package and a 
LaTeX renderer like \strong{MiKTeX}.  
\strong{MiKTeX} must be installed separately, and will
not be installed as part of the \strong{reporter} install.  This program must
furthermore be added to the \code{path} environment variable on the system
where \strong{reporter} will run, so that \strong{reporter} can find it.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_spec.r
\name{lowcase_parens}
\alias{lowcase_parens}
\alias{upcase_parens}
\alias{lowcase_n}
\alias{upcase_n}
\title{Functions to format the population label}
\usage{
lowcase_parens(x)

upcase_parens(x)

lowcase_n(x)

upcase_n(x)
}
\arguments{
\item{x}{Population count}
}
\description{
These functions are used to format the "N=" population label
on column headers.
}
\details{
Which function to use to format the population label is specified
on the \code{n_format} parameter on the \code{\link{create_table}} function.
These formatting functions provide several options for formatting the "N=", 
including whether the "N" should be upper case or lower case, and whether
or not to put the value in parentheses.  If one of these options does not 
meet the specifications for your report, you may also write your own 
formatting function and pass it to the \code{n_format} function.  When an 
N value is supplied, the output of this function will be concatenated 
to the header label.
}
\examples{
# Create test data
l <- "Label"
n <- 47

cat(paste0(l, lowcase_parens(n)))
# Label
# (n=47)

cat(paste0(l, upcase_parens(n)))
# Label
# (N=47)

cat(paste0(l, lowcase_n(n)))
# Label
# n=47

cat(paste0(l, upcase_n(n)))
# Label
# N=47

customN <- function(n) {
  return(paste0(": N=", n))
}
cat(paste0(l, customN(n)))
# Label: N=47

## Use alternate n format ##
library(reporter)
library(magrittr)

tmp <- tempfile(fileext = ".txt")

# Prepare data
df <- read.table(header = TRUE, text = '
      Hair  Group1  Group2
      Black    25      16
      Brown    13      18
      Blonde   5        7
      Red      2        1')
      
# Create table with lowcase n formatting
tbl <- create_table(df, n_format=lowcase_n) \%>\% 
       titles("Hair Color") \%>\% 
       define(Group1, n = 45) \%>\% 
       define(Group2, n = 42)

# Create report
rpt <- create_report(tmp) \%>\% 
       add_content(tbl, align = "left")
       
# Write to file system
write_report(rpt)     

writeLines(readLines(tmp))
#      Hair Color
# 
#        Group1 Group2
# Hair     n=45   n=42
# ¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯
# Black      25     16
# Brown      13     18
# Blonde      5      7
# Red         2      1
}
\seealso{
\code{\link{create_table}} function to create a table.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_spec.r
\name{write_registration_file}
\alias{write_registration_file}
\title{Create a registration file}
\usage{
write_registration_file(file_path)
}
\arguments{
\item{file_path}{The full or relative file name and path to create the 
registration file.}
}
\description{
This function will create a registration file to help determine
the correct \code{cpuom} and \code{lpuom} for your editor/printer.
}
\details{
The \code{cpi} and \code{lpi} are 
used in \code{output_type = "TXT"} to determine available space on
the page. The registration file can help determine the correct settings
for the target text editor and printer.  Failure to set the 
correct characters per 
unit of measure (cpuom) and lines per unit of measure (lpuom) may result
in misalignment of content on the page when printing text output.
}
\section{How to Use the Registration File}{

To use the registration file, first decide the units of measure you
wish to use, inches or centimeters.  Next, create the registration file
by calling the \code{write_registration_file} function.  Then print the 
registration file.

Once the registration file is printed, take a ruler and measure both the 
horizontal and vertical registration lines from zero to 60 in the desired
units of measure.  For example, if your units of measure is 'inches', measure
the registration lines in inches.    

Record the 
distance measured in each direction.  For each direction, divide 60 by the 
distance measured, and round to three decimal places.  The horizontal 
result is the characters per unit of measure (cpuom).  The vertical result
is the lines per unit of measure (lpuom). To get an accurate printing 
of text reports, 
assign these values to the \code{cpuom} and \code{lpuom} parameters
on the \code{\link{options_fixed}} function.

For best results, test the calculated values by printing some reports and
checking for undesired page breaks or wrapped lines.  If necessary, adjust
the calculated cpuom and lpuom values until all content stays within 
the available space without wrapping or breaking.
}

\examples{
library(reporter)

# Create temp file path
tmp <- file.path(tempdir(), "reg.txt")

# Create the registration file 
write_registration_file(tmp)

# Write registration file to the console
writeLines(readLines(tmp))

# 0--------+---------+---------+---------+---------+---------+
# -       10        20        30        40        50        60
# -
# -
# -
# -
# -
# -
# -
# + 10
# -
# -
# -
# -
# -
# -
# -
# -
# -
# + 20
# -
# -
# -
# -
# -
# -
# -
# -
# -
# + 30
# -
# -
# -
# -
# -
# -
# -
# -
# -
# + 40
# -
# -
# -
# -
# -
# -
# -
# -
# -
# + 50
# -
# -
# -
# -
# -
# -
# -
# -
# -
# + 60
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_spec.r
\name{page_footer}
\alias{page_footer}
\title{Adds a page footer}
\usage{
page_footer(
  x,
  left = "",
  center = "",
  right = "",
  blank_row = "above",
  width = NULL
)
}
\arguments{
\item{x}{The report spec object.}

\item{left}{The left page footer text.  May be a single string or a vector
of strings.}

\item{center}{The center page footer text.  May be a single string or a
vector of strings.}

\item{right}{The right page footer text.  May be a single string or a vector
of strings.}

\item{blank_row}{Whether to create a blank row above the page footer.
Valid values are 'above' and 'none'.  Default is 'above'.}

\item{width}{Widths for left, center, and right columns of the page footer, 
passed as a vector of double values in the report unit of measure.  If the vector
contains fewer than three widths, the widths passed will be interpreted from
left to right, and any unassigned column widths will be calculated from
the remaining width on the page.  For example, if two values are assigned, they will be
interpreted as left and center, and the right column width will be calculated.
Setting a cell width to zero (0) will remove that column from the page footer
entirely. To set the left and right widths only, pass a zero for the center
cell, i.e. \code{width = c(6, 0, 3)} for a total width of 9 inches.}
}
\value{
The modified report.
}
\description{
This function adds a page footer to the report.  The page footer will appear
on each page of the report, at the bottom of the page.  The page footer
contains three sections: left, center, and right.  Content for each section
may be specified with the appropriate parameter.
}
\details{
Only one page footer is allowed per report.  The page footer will appear 
on all pages of the report.  The page footer may contain text on the 
left, right, or center. Use the appropriate parameters to specify the 
desired text for each section. Multiple strings may be passed to each section 
as a vector of strings. 

If the width of the page header
string exceeds the available space, an error will be generated. In addition,
there is a limit of 5 strings for each page footer section.  

There are two special tokens to generate page numbers: [pg] and [tpg]. 
Use [pg] to indicate the current page number.  Use [tpg] to indicate the
total number of pages in the report.  These tokens may be placed anywhere 
in the page header or page footer. 

Use the \code{blank_row} parameter to control the blank space above the 
page footer.

Each footer string must fit within the available space.  The \strong{reporter}
package will not wrap footer.  If a footer string does not fit within the 
available space, an error will be generated.  In these situations, either
shorten the footer string or split it into multiple footers that each fit 
within the available space.
}
\examples{
library(reporter)
library(magrittr)

# Create temp file path
tmp <- file.path(tempdir(), "mtcars.txt")

dat <- data.frame(name = rownames(mtcars[1:10, ]), mtcars[1:10, 1:5], 
                  stringsAsFactors = FALSE)

# Create the report object
rpt <- create_report(tmp, orientation = "portrait") \%>\% 
  page_header("Client: Motor Trend", "Study: Cars") \%>\% 
  titles("MTCARS Sample Report") \%>\% 
  add_content(create_table(dat)) \%>\% 
  page_footer(Sys.time(), right = "Page [pg] of [tpg]")

# Write the report to the file system
write_report(rpt)

# Write report to console
writeLines(readLines(tmp, encoding = "UTF-8"))

# Client: Motor Trend                                                Study: Cars
#                              MTCARS Sample Report
# 
#                name                 mpg cyl   disp   hp  drat
#                ----------------------------------------------
#                Mazda RX4             21   6    160  110   3.9
#                Mazda RX4 Wag         21   6    160  110   3.9
#                Datsun 710          22.8   4    108   93  3.85
#                Hornet 4 Drive      21.4   6    258  110  3.08
#                Hornet Sportabout   18.7   8    360  175  3.15
#                Valiant             18.1   6    225  105  2.76
#                Duster 360          14.3   8    360  245  3.21
#                Merc 240D           24.4   4  146.7   62  3.69
#                Merc 230            22.8   4  140.8   95  3.92
#                Merc 280            19.2   6  167.6  123  3.92
# 
# ...
# 
# 2020-10-17 11:53:51                                                Page 1 of 1
}
\seealso{
Other report: 
\code{\link{add_content}()},
\code{\link{create_report}()},
\code{\link{footer_image}()},
\code{\link{footnotes}()},
\code{\link{header_image}()},
\code{\link{options_fixed}()},
\code{\link{page_by}()},
\code{\link{page_header}()},
\code{\link{print.report_spec}()},
\code{\link{set_margins}()},
\code{\link{title_header}()},
\code{\link{titles}()},
\code{\link{write_report}()}
}
\concept{report}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_spec.r
\name{spanning_header}
\alias{spanning_header}
\title{Defines a spanning header}
\usage{
spanning_header(
  x,
  from,
  to,
  label = "",
  label_align = "center",
  level = 1,
  n = NULL,
  underline = TRUE,
  bold = FALSE,
  standard_eval = FALSE
)
}
\arguments{
\item{x}{The table object to add spanning headers to.}

\item{from}{The starting column to span.  Spanning columns are defined as
range of columns 'from' and 'to'. The columns may be identified by position, 
or by quoted or unquoted variable names. If you want to pass the \code{from} 
value using an R variable,
set the \code{standard_eval} parameter to TRUE.
The \code{from} parameter is required.}

\item{to}{The ending column to span.  Spanning columns are defined as
range of columns 'from' and 'to'. The columns may be identified by position,
or by quoted or unquoted variable names.  If you want to pass the \code{to} 
value using an R variable,
set the \code{standard_eval} parameter to TRUE.
The \code{to} parameter is required.}

\item{label}{The label to apply to the spanning header.}

\item{label_align}{The alignment to use for the label. Valid values are 
"left", "right", "center", and "centre".  The default for spanning columns
is "center".}

\item{level}{The level to use for the spanning header.  The lowest
spanning level is level 1, the next level above is level 2, and so on.  
By default, the level is set to 1.}

\item{n}{The population count to use for the "N=" label on the spanning 
header. The "N=" label will be formatted according to the \code{n_format}
parameter on the \code{\link{create_table}} function.}

\item{underline}{A TRUE or FALSE value indicating whether the spanning
header should be underlined.  Default is TRUE.}

\item{bold}{A TRUE or FALSE value indicating whether the spanning header
label should be bold.  Default is FALSE.}

\item{standard_eval}{A TRUE or FALSE value that indicates whether to
use standard or non-standard evaluation of the \code{from},
and \code{to} parameters.  Set \code{standard_eval} to TRUE if you want
to pass the column names as variables. Default is FALSE, meaning it
will use non-standard (unquoted) evaluation.}
}
\value{
The modified table spec.
}
\description{
Create a header that spans multiple columns.  Spanning headers
are used to group related columns.  Such groupings are a common 
feature of statistical reports.
}
\details{
A spanning header is a label and underline that spans one or more 
columns.  A spanning header is defined minimally by identifying the
column range to be spanned, and a label.  A label alignment and "N="
value may also be supplied.

The spanning column range is defined by the \code{from} and \code{to} 
parameters.  The range identifies a contiguous set of variables on the data.
Variables can be identified by position, a quoted variable name, or an 
unquoted variable name.
}
\examples{
library(reporter)
library(magrittr)

# Create a temporary file
tmp <- file.path(tempdir(), "iris.txt")

# Prepare data
dat <- iris[sample(1:150, 15), c(5, 1, 2, 3, 4)]
dat <- dat[order(dat$Species), ]

# Define table
tbl <- create_table(dat) \%>\% 
  titles("Table 3.2", "IRIS Sample Report") \%>\% 
  spanning_header(2, 3, label = "Sepal") \%>\% 
  spanning_header(4, 5, label = "Petal") \%>\% 
  column_defaults(2:5, format = "\%.1f") \%>\% 
  define(Species, align = "left", dedupe = TRUE, blank_after = TRUE) \%>\% 
  define(Sepal.Length, label = "Length") \%>\% 
  define(Sepal.Width, label = "Width") \%>\% 
  define(Petal.Length, label = "Length") \%>\% 
  define(Petal.Width, label = "Width") \%>\% 
  footnotes("* From Fisher's Iris Dataset")
       
# Define report
rpt <- create_report(tmp, orientation="portrait") \%>\%
  options_fixed(blank_margins = TRUE) \%>\% 
  set_margins(top = 1, bottom =1) \%>\% 
  add_content(tbl, align = "left") 

# Write the report
write_report(rpt)

writeLines(readLines(tmp, encoding = "UTF-8"))

#
#
#
#
#                      Table 3.2
#                  IRIS Sample Report
#
#                       Sepal        Petal
#                   ------------ ------------
#       Species     Length Width Length Width
#       -------------------------------------
#       setosa         5.0   3.0    1.6   0.2
#                      4.6   3.4    1.4   0.3
#                      5.0   3.4    1.6   0.4
#                      5.7   3.8    1.7   0.3
#
#       versicolor     5.7   2.8    4.1   1.3
#                      6.2   2.9    4.3   1.3
#                      7.0   3.2    4.7   1.4
#                      6.6   2.9    4.6   1.3
#
#       virginica      6.2   3.4    5.4   2.3
#                      7.2   3.0    5.8   1.6
#                      6.9   3.1    5.1   2.3
#                      5.6   2.8    4.9   2.0
#                      7.7   2.6    6.9   2.3
#                      6.3   2.8    5.1   1.5
#                      7.7   2.8    6.7   2.0
#
#
#       * From Fisher's Iris Dataset
}
\seealso{
Other table: 
\code{\link{column_defaults}()},
\code{\link{create_table}()},
\code{\link{define}()},
\code{\link{print.table_spec}()},
\code{\link{stub}()}
}
\concept{table}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_spec.r
\name{print.report_spec}
\alias{print.report_spec}
\title{Prints the report specification}
\usage{
\method{print}{report_spec}(x, ..., verbose = FALSE)
}
\arguments{
\item{x}{The report spec.}

\item{...}{Additional parameters to pass to the underlying print function.}

\item{verbose}{Whether to print the report object in verbose (list) form
or summary form.  The default is FALSE.}
}
\value{
The report spec, invisibly.
}
\description{
A function to print the report specification.
The \strong{print} function will print the report spec in summary 
form by default.  To print in list form, set the \code{verbose} parameter
to TRUE.
}
\examples{
library(reporter)
library(magrittr)

# Create temp file path
tmp <- file.path(tempdir(), "mtcars.txt")

# Create the table
tbl <- create_table(mtcars) \%>\% 
  titles("Table 1.0", "MTCARS Sample Report") \%>\% 
  footnotes("* NOTE: Data from 1974")

# Create the report object
rpt <- create_report(tmp) \%>\% 
  add_content(tbl, align = "left") 

# Write the report to the file system
res <- write_report(rpt)

# Write the modified report object to the console
print(res)

# # A report specification: 1 pages
# - file_path: 'C:/Users/User/AppData/Local/Temp/RtmpWQybXs/mtcars.txt'
# - output_type: TXT
# - units: inches
# - orientation: landscape
# - line size/count: 108/45
# - content: 
# # A table specification:
# - data: data.frame 'mtcars' 32 rows 11 cols
# - show_cols: all
# - use_attributes: all
# - title 1: 'Table 1.0'
# - title 2: 'MTCARS Sample Report'
# - footnote 1: '* NOTE: Data from 1974'
}
\seealso{
\code{\link{create_report}} function to create a report specification.

Other report: 
\code{\link{add_content}()},
\code{\link{create_report}()},
\code{\link{footer_image}()},
\code{\link{footnotes}()},
\code{\link{header_image}()},
\code{\link{options_fixed}()},
\code{\link{page_by}()},
\code{\link{page_footer}()},
\code{\link{page_header}()},
\code{\link{set_margins}()},
\code{\link{title_header}()},
\code{\link{titles}()},
\code{\link{write_report}()}
}
\concept{report}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter.R
\docType{package}
\name{reporter}
\alias{reporter}
\alias{reporter-package}
\title{reporter: A package for creating statistical reports}
\description{
The \strong{reporter} package creates statistical reports in TXT, 
RTF, PDF, HTML, and DOCX file formats.  Features include automatic page wrapping and 
breaking for wide and long tables, spanning headers, titles, footnotes, 
page headers, page footers, and page numbering.  The package allows mixing
of multiple tables, text, and plots in a single report, or even on a single
page.
}
\details{
The \strong{reporter} package creates regulatory-style, statistical reports. It 
was designed to create tables, listings, and figures (TLFs) 
for use in the pharmaceutical, biotechnology, and medical-device industries.
However, the functions are generalized enough to provide statistical reporting
for any industry.  The package is written primarily in Base R, and has 
few dependencies on other packages.

The package is intended to give R programmers 
flexible report layout capabilities, and a choice of output formats. 
The package will 
initially focus on printable, file-based 
output formats, as there are already numerous R packages that 
provide tabular reporting in HTML. 

PDF output may have limitations not associated with other output types.  
See \link{NotesOnPDF}
for more information.
}
\section{Key Features}{

The \strong{reporter} package contains the following key features:
\itemize{
  \item Titles, footnotes, page header, and page footer are repeated on each page
  \item Supports header labels and spanning headers 
  \item Calculates default columns widths automatically
  \item Includes automatic wrapping and splitting of wide and long tables
  \item Integrates with the \strong{\link[fmtr]{fmtr}} package to format 
        numeric, date, and character data
  \item Plots from the popular \strong{\link[ggplot2]{ggplot2}} package can 
        be added to RTF, PDF, HTML, and DOCX reports
  \item Allows appending multiple tables to a report, multiple tables to a page, 
        and intermingling of text, tables, and plots
  \item Supports in-report date/time stamps and "Page X of Y" page numbering
}
}

\section{Key Functions}{

\itemize{
  \item \code{\link{create_report}} to define a report
  \item \code{\link{create_table}} to define a table
  \item \code{\link{create_text}} to define text content
  \item \code{\link{create_plot}} to define a plot 
  \item \code{\link{write_report}} to write out the report
}
}

\section{Package Assumptions and Limitations}{

Note that the \strong{reporter} package is built on several assumptions, and 
has some limitations.  Those
assumptions and limitations are as follows:
\itemize{
  \item The current version supports both a monospace, fixed-width style 
  report and a variable-width style report.  The monospace report is the 
  default.  To create a variable width report, set the font parameter on
  the \code{create_report} function to 'Arial', 'Courier', or 'Times'.
  \item If a font is selected, it will be applied uniformly to the entire
  report.  The package has no capabilities to mix different fonts in the 
  same report.
  \item The package assumes that, except for formatting and layout, 
  the incoming data is ready for 
  printing. The \strong{reporter} package has no capabilities to perform 
  calculations, summaries, grouping, or derivations.  Use the many 
  capabilities in R to perform these operations prior to sending data
  to the \strong{reporter} package.
  \item It is assumed that the incoming data is sorted as desired.  
  The \strong{reporter} package
  has no capabilities for sorting.  Use the sorting functionality in 
  Base R or supplemental packages to sort the data prior to sending to
  \strong{reporter}.
  \item For monospace reports, titles, footnotes, page headers, and page 
  footers must fit in the available space. 
  If they don't, the \strong{reporter} package will 
  generate a warning.  In these situations, the recommended course of 
  action is to split the offending string into two or more strings that 
  fit within the available width.
  \item The \strong{reporter} package will never break a word. That means you 
  cannot set a column width that is less than the length of the longest 
  word.  If you wish to break words, add the breaks with
  \code{\link[base]{strwrap}} or an equivalent function before reporting.  
  \item The package will make a best guess on column widths based on the 
  width of the data and column headers.  To override the best guess,
  set the column width directly by placing a width attribute on the variable
  or by using the \code{\link{column_defaults}} or 
  \code{\link{define}} functions. 
  \item The max automatic column width
  is 5 inches.  Longer data values will be wrapped.
  \item The package support plots from \strong{ggplot2}. These plots
  can be added to RTF, PDF, HTML, and DOCX output types.  The package does not 
  support Base R plots.
  \item The package currently supports styling for HTML reports.  
  This styling allows setting background colors, font colors, border
  colors, and some font sizing and bolding.   
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://reporter.r-sassy.org}
  \item \url{https://github.com/dbosak01/reporter}
  \item Report bugs at \url{https://github.com/dbosak01/reporter/issues}
}

}
\author{
\strong{Maintainer}: David Bosak \email{dbosak01@gmail.com}

Authors:
\itemize{
  \item Bill Huang \email{Bill.Huang@Toastmasters.org.tw}
}

Other contributors:
\itemize{
  \item Kevin Kramer \email{kkrame02@amgen.com} [contributor]
  \item Duong Tran \email{trand000@aol.com} [contributor]
  \item Raphael Huang \email{chi-hua.huang@astellas.com} [contributor]
  \item Archytas Clinical Solutions [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_spec.r
\name{column_defaults}
\alias{column_defaults}
\title{Set default attributes for one or more columns}
\usage{
column_defaults(
  x,
  vars = NULL,
  from = NULL,
  to = NULL,
  label = NULL,
  format = NULL,
  align = NULL,
  label_align = NULL,
  width = NULL,
  n = NULL,
  standard_eval = FALSE,
  style = NULL
)
}
\arguments{
\item{x}{A table spec.}

\item{vars}{The variable name or names to define defaults for.  Variable
names may be quoted or unquoted.  The parameter will also accept 
integer column positions instead of names.  For multiple variables, 
pass the names or positions as a vector. If you want to pass an R variable 
of names,
set the \code{standard_eval} parameter to TRUE.
The \code{standard_eval} parameter is useful when writing functions that construct
reports dynamically.}

\item{from}{The variable name or position that starts a column range.
If passed as a variable name, it may be quoted or unquoted.}

\item{to}{The variable name or position that ends a column range. 
If passed as a variable name, it 
may be quoted or unquoted.}

\item{label}{The label to use for a column header.  This label will be 
applied to all variables assigned to the \code{column_defaults} function.}

\item{format}{The format to use for the column data.  The format can 
be a string format, a formatting function, a lookup list, a user-defined
format, or a formatting list. 
All formatting is performed by the \code{\link[fmtr]{fmtr}} package.  For 
additional information, see the help for that package.}

\item{align}{The column alignment.  Valid values are "left", "right", 
"center", and "centre".}

\item{label_align}{How to align the header labels for this column.
Valid values are "left", "right", "center", and "centre".}

\item{width}{The width of the column in the specified units of measure.
The units of measure are specified on the \code{units} parameter of the
\code{\link{create_report}} function.  If no width is supplied, the
\code{\link{write_report}} function will assign a default width based on the 
width of the column data and the label.  \code{write_report} will not set a 
column width less than the width of the largest word in the data or label.
In other words, \code{write_report} will not break words.}

\item{n}{The n value to place in the "N=" header label.  Formatting for
the n value will be performed by the formatting function assigned to the 
\code{n_format} parameter on \code{\link{create_table}}.}

\item{standard_eval}{A TRUE or FALSE value that indicates whether to
use standard or non-standard evaluation of the \code{vars}, \code{from},
and \code{to} parameters.  Set \code{standard_eval} to TRUE if you want
to pass the column names as variables. Default is FALSE, meaning it
will use non-standard (unquoted) evaluation.}

\item{style}{A \code{\link{cell_style}} object that defines a style 
for all columns associated with the column defaults.}
}
\value{
The modified table spec.
}
\description{
A function to set default attributes for columns on a table.  
The \code{column_defaults} function contains a subset of the parameters
on the \code{\link{define}} function that can be shared across variables.
Any attributes set by \code{column_defaults} can be overridden by 
the \code{define} function.  The overall purpose of the 
function is to minimize redundancy in column definitions.
}
\details{
Column defaults can be specified for multiple variables.  By default,
the function will apply to all variables.  Alternately, you can 
specify a vector of columns on the \code{vars} parameter, or a range of
columns using the \code{from} and \code{to} parameters.  Both the 
\code{vars} parameters and the \code{from} and \code{to} parameters
will accept column positions, quoted variable names, or unquoted variable 
names.

The parameters that can be set with the \code{column_defaults} 
include the formatting attributes 'width', 'justify', 'label', and 
'format'.  Any parameters set with \code{column_defaults} will override
any attributes set on the data frame.

Note that you may call the \code{column_defaults} function multiple times
on the same table specification.  Typically, multiple \code{column_defaults}
calls would be made with a different set or range of variables.
}
\examples{
library(reporter)
library(magrittr)
 
# Create temp file name
tmp <- file.path(tempdir(), "mtcars.txt")

# Prepare data
dat <- mtcars[1:10, ]
dat <- data.frame(vehicle = rownames(dat), dat, stringsAsFactors = FALSE)

# Define table
tbl <- create_table(dat, show_cols = 1:8) \%>\% 
  column_defaults(from = mpg, to = qsec, width = .5, format = "\%.1f") \%>\% 
  define(vehicle, label = "Vehicle", width = 1.5, align = "left") \%>\% 
  define(c(cyl, hp), format = "\%.0f") 

# Create the report
rpt <- create_report(tmp, orientation = "portrait") \%>\% 
  titles("Table 2.5", "MTCARS Sample Report") \%>\% 
  add_content(tbl) 

# Write the report
write_report(rpt)

# Send report to console for viewing
writeLines(readLines(tmp, encoding = "UTF-8"))

#                                 Table 2.5
#                            MTCARS Sample Report
# 
#    Vehicle               mpg    cyl   disp     hp   drat     wt   qsec
#    -------------------------------------------------------------------
#    Mazda RX4            21.0      6  160.0    110    3.9    2.6   16.5
#    Mazda RX4 Wag        21.0      6  160.0    110    3.9    2.9   17.0
#    Datsun 710           22.8      4  108.0     93    3.8    2.3   18.6
#    Hornet 4 Drive       21.4      6  258.0    110    3.1    3.2   19.4
#    Hornet Sportabout    18.7      8  360.0    175    3.1    3.4   17.0
#    Valiant              18.1      6  225.0    105    2.8    3.5   20.2
#    Duster 360           14.3      8  360.0    245    3.2    3.6   15.8
#    Merc 240D            24.4      4  146.7     62    3.7    3.2   20.0
#    Merc 230             22.8      4  140.8     95    3.9    3.1   22.9
#    Merc 280             19.2      6  167.6    123    3.9    3.4   18.3
#
}
\seealso{
Other table: 
\code{\link{create_table}()},
\code{\link{define}()},
\code{\link{print.table_spec}()},
\code{\link{spanning_header}()},
\code{\link{stub}()}
}
\concept{table}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_spec.r
\name{write_report}
\alias{write_report}
\title{Write a report to the file system}
\usage{
write_report(
  x,
  file_path = NULL,
  output_type = NULL,
  preview = NULL,
  log = TRUE
)
}
\arguments{
\item{x}{The report object to write.}

\item{file_path}{The file name and path to write the report to.  If supplied,
this parameter overrides the \code{file_path} parameter on the 
\code{create_report} function. Default is NULL.}

\item{output_type}{The output file type.  This parameter will override
the \code{output_type} on the \code{create_report} function.  This 
parameter can be used to output the same report object to 
multiple output types. Default value is NULL, meaning it will not override
the \code{create_report} value.  Valid values are 'TXT', 'RTF', 'DOCX', 
'PDF' and 'HTML'.}

\item{preview}{Whether to write the entire report, or a report preview.
A report preview is a subset of pages of the report.  The default value is 
NULL, meaning the entire report will be written.  You may also pass 
a number of pages to write.  For example, passing the number 1 will print
the first page, while passing a 5 will print the first five pages.}

\item{log}{Whether the function should log itself automatically, if
the \strong{logr} package is available.  This parameter is used internally.
Default is TRUE.}
}
\value{
The report spec, with settings modified during rendering.  These 
modified settings can sometimes be useful for documentation, and for 
debugging issues with the procedure.
}
\description{
This function writes a report_spec object to the file system, using the
specifications provided in the object.
}
\details{
The function renders the report in the requested format, and writes it
to the location specified in the report \code{file_path} parameter.  Attempts
to write an object that is not of class "report_spec" will generate an error.

The \code{write_report} function is a driver for very complex set of 
rendering functions. The rendering functions
perform most of the advanced functionality of the \strong{reporter} package:
generating spanning headers, page wrapping and breaking, creating stub 
columns, etc.  When things go wrong, they will usually go wrong during this
function call.  For that reason, although this function can be part of 
the pipeline that creates the report object, it is best to call 
\code{write_report} independently, to help isolate any issues from the 
report definition procedure.
}
\examples{
library(reporter)
library(fmtr)
library(magrittr)

# Create temp file path
tmp <- file.path(tempdir(), "beaver2.txt")

# Take Sample of Data
dat <- beaver2[sample(1:100, 15), ]

# Create format for active variable
fmt <- value(condition(x == 0, "No"),
             condition(x == 1, "Yes"))

# Create the table
tbl <- create_table(dat) \%>\% 
  titles("Table 1.0", "BEAVERS Sample Report") \%>\% 
  column_defaults(width = .75) \%>\% 
  define(day, label = "Day", format = "Day \%s") \%>\% 
  define(time, label = "Time") \%>\% 
  define(temp, label = "Temperature", width = 1, format = "\%.1f") \%>\% 
  define(activ,label = "Active", format = fmt) \%>\% 
  footnotes("* NOTE: Data on beaver habits")

# Create the report object
rpt <- create_report(tmp) \%>\% 
  add_content(tbl, align = "left") 

# Write the report to the file system
res <- write_report(rpt)

# Write the modified report object to the console
print(res)

# Write the report to console
writeLines(readLines(tmp, encoding = "UTF-8"))

#                 Table 1.0
#           BEAVERS Sample Report
# 
#      Day      Time  Temperature    Active
# -----------------------------------------
#  Day 307      1020         37.2        No
#  Day 307      1030         37.2        No
#  Day 307       940         36.7        No
#  Day 307      1340         37.1        No
#  Day 307      1410         37.2        No
#  Day 307      1400         37.1        No
#  Day 307      1130         36.9        No
#  Day 307      1140         37.0        No
#  Day 307      1120         37.0        No
#  Day 307      1000         37.1        No
#  Day 307      1250         37.0        No
#  Day 307      2100         37.9       Yes
#  Day 307      1210         37.0        No
#  Day 307      1740         38.0       Yes
#  Day 308       130         37.8       Yes
# 
# * NOTE: Data on beaver habits
}
\seealso{
Other report: 
\code{\link{add_content}()},
\code{\link{create_report}()},
\code{\link{footer_image}()},
\code{\link{footnotes}()},
\code{\link{header_image}()},
\code{\link{options_fixed}()},
\code{\link{page_by}()},
\code{\link{page_footer}()},
\code{\link{page_header}()},
\code{\link{print.report_spec}()},
\code{\link{set_margins}()},
\code{\link{title_header}()},
\code{\link{titles}()}
}
\concept{report}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_spec.r
\name{stub}
\alias{stub}
\title{Defines a report stub}
\usage{
stub(
  x,
  vars,
  label = "",
  label_align = NULL,
  align = "left",
  width = NULL,
  standard_eval = FALSE,
  style = NULL
)
}
\arguments{
\item{x}{The table spec.}

\item{vars}{A vector of quoted or unquoted variable names from 
which to create the stub. If you want to pass an R variable of names,
escape the values with double curly braces, i.e. \code{vars = {{myvar}}}.
The curly brace escape is useful when writing functions that construct
reports dynamically.}

\item{label}{The label for the report stub.  The default label is an empty
string.}

\item{label_align}{The alignment for the stub column label.
Valid values are 'left', 'right', 'center', and 'centre'.  Default follows
the \code{align} parameter.}

\item{align}{How to align the stub column.  Valid values are 'left', 
'right', 'center', and 'centre'.  Default is 'left'.}

\item{width}{The width of the stub, in report units of measure.}

\item{standard_eval}{A TRUE or FALSE value that indicates whether to
use standard or non-standard evaluation of the \code{vars}, \code{from},
and \code{to} parameters.  Set \code{standard_eval} to TRUE if you want
to pass the column names as variables. Default is FALSE, meaning it
will use non-standard (unquoted) evaluation.}

\item{style}{A \code{\link{cell_style}} object that contains the style
specifications for the stub.}
}
\value{
The modified table spec.
}
\description{
Combine columns into a nested report stub.  The report stub
is a common feature of statistical reports.  The stub is created with
the \code{stub} function, and frequently appears in combination with the 
\code{label_row} and \code{indent} parameters from the 
\code{\link{define}} function.  These elements work together to define
the appearance of the stub.
}
\details{
The table stub is a nested set of labels that identify rows 
on the table. The stub is created by combining two or more columns into 
a single stub column.  The relationship between the columns is typically 
visualized as a hierarchy, with lower level concepts indented under 
higher level concepts.  

A typical stub is created with the following steps:
\itemize{
  \item Prepare the data. 
  \item Create the table object.
  \item Define the stub on the table using the \code{stub} function, 
  and identify the variables to be combined.
  \item Identify higher level concepts with the \code{label_row} parameter
  on the \code{\link{define}} function. 
  \item Identify lower level concepts using the \code{indent} parameter 
  on the \code{\link{define}} function.
}

The stub will be automatically added as an identity variable on the report, 
and will always appear as the leftmost column.  There can only be one stub 
defined on a report.

If you wish to create multiple levels of nested labels, use
an NA value to prevent lower level labels from overwriting
higher level labels. 

For example, the following data:
\preformatted{
continent          country   state_province    
"North America"    NA        NA   
"North America"    "Canada"  NA   
"North America"    "Canada"  "Ontario"   
"North America"    "USA"     NA   
"North America"    "USA"     "New York"   
"South America"    NA        NA   
"South America"    "Brazil"  NA   
"South America"    "Brazil"  "Amazonas"   
"South America"    "Brazil"  "Bahia"   
}
Will produce the following stub:   
\preformatted{
North America   
  Canada   
    Ontario   
  USA   
    New York   
South America  
  Brazil 
    Amazonas   
    Bahia   
}
With the following code:
\preformatted{
tbl <- create_table(dat) \%>\% 
  stub(c(continent, country, state_province)) \%>\% 
  define(country, indent = .25) \%>\% 
  define(state_province, indent = .5)
}
}
\examples{
library(reporter)
library(magrittr)

# Create temporary path
tmp <- file.path(tempdir(), "stub.txt")

# Read in prepared data
df <- read.table(header = TRUE, text = '
      var      label        A             B          
      "ampg"   "N"          "19"          "13"         
      "ampg"   "Mean"       "18.8 (6.5)"  "22.0 (4.9)" 
      "ampg"   "Median"     "16.4"        "21.4"       
      "ampg"   "Q1 - Q3"    "15.1 - 21.2" "19.2 - 22.8"
      "ampg"   "Range"      "10.4 - 33.9" "14.7 - 32.4"
      "cyl"    "8 Cylinder" "10 ( 52.6\%)" "4 ( 30.8\%)" 
      "cyl"    "6 Cylinder" "4 ( 21.1\%)"  "3 ( 23.1\%)" 
      "cyl"    "4 Cylinder" "5 ( 26.3\%)"  "6 ( 46.2\%)"')

# Create table
tbl <- create_table(df, first_row_blank = TRUE) \%>\% 
  stub(c(var, label)) \%>\% 
  define(var, blank_after = TRUE, label_row = TRUE, 
         format = c(ampg = "Miles Per Gallon", cyl = "Cylinders")) \%>\% 
  define(label, indent = .25) \%>\% 
  define(A, label = "Group A", align = "center", n = 19) \%>\% 
  define(B, label = "Group B", align = "center", n = 13)


# Create report and add content
rpt <- create_report(tmp, orientation = "portrait") \%>\% 
  page_header(left = "Client: Motor Trend", right = "Study: Cars") \%>\% 
  titles("Table 1.0", "MTCARS Summary Table") \%>\% 
  add_content(tbl) \%>\% 
  footnotes("* Motor Trend, 1974") \%>\%
  page_footer(left = Sys.time(), 
              center = "Confidential", 
              right = "Page [pg] of [tpg]")

# Write out report
write_report(rpt)

# View report in console
writeLines(readLines(tmp, encoding = "UTF-8"))

# Client: Motor Trend                                                Study: Cars
#                                   Table 1.0
#                              MTCARS Summary Table
# 
#                                     Group A      Group B
#                                      (N=19)       (N=13)
#                 -------------------------------------------
# 
#                 Miles Per Gallon
#                    N                   19           13
#                    Mean            18.8 (6.5)   22.0 (4.9)
#                    Median             16.4         21.4
#                    Q1 - Q3        15.1 - 21.2  19.2 - 22.8
#                    Range          10.4 - 33.9  14.7 - 32.4
#
#                 Cylinders
#                    8 Cylinder     10 ( 52.6\%)   4 ( 30.8\%)
#                    6 Cylinder      4 ( 21.1\%)   3 ( 23.1\%)
#                    4 Cylinder      5 ( 26.3\%)   6 ( 46.2\%)
# 
# ...
# 
# 
# * Motor Trend, 1974
# 
# 2020-08-30 03:50:02              Confidential                      Page 1 of 1
#
}
\seealso{
Other table: 
\code{\link{column_defaults}()},
\code{\link{create_table}()},
\code{\link{define}()},
\code{\link{print.table_spec}()},
\code{\link{spanning_header}()}
}
\concept{table}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/attributes.R
\name{ttl}
\alias{ttl}
\title{Creates a title attribute}
\usage{
ttl(
  titles,
  align = "center",
  blank_row = "below",
  borders = NULL,
  width = NULL,
  bold = FALSE,
  font_size = NULL
)
}
\arguments{
\item{titles}{A vector of title strings.}

\item{align}{The position to align the titles.  Valid values are 'left', 
'right', 'center' or 'centre'.  For titles, the default is 'center'.}

\item{blank_row}{Where to place a blank row.  Valid values are 'above',
'below', 'both', or 'none'.  Default is "below".}

\item{borders}{Whether and where to place a border. Valid values are 'top',
'bottom', 'left', 'right', 'outside', 'inside', 'all', or 'none'.  
Default is "none".}

\item{width}{The width of the titles block.}

\item{bold}{A parameter to bold the titles.  Valid values are TRUE and FALSE.
Default is FALSE.}

\item{font_size}{The font size to use for the title block.  The font size
of the report will be used by default.  Valid values are 8, 9, 10, 11, 12,
13, and 14.}
}
\value{
A title attribute object.
}
\description{
This function creates a titles attribute, which can be used to
define titles by attaching as an attribute to a data frame.  This function is 
used internally to the system.
}
\seealso{
\code{\link{titles}} function.

Other attributes: 
\code{\link{ftn}()},
\code{\link{span}()}
}
\concept{attributes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_style.R
\name{get_theme}
\alias{get_theme}
\title{Get a theme}
\usage{
get_theme(theme)
}
\arguments{
\item{theme}{A string that contains the desired theme name to return.
Valid values are "MidnightBlue", "SteelBlue", "DarkRed", "SeaGreen", 
"SlateGrey", "Plain", and "SASDefault".}
}
\description{
This function will return a style object for a specified theme.
There are currently seven themes available.  The returned object 
may be modifed and applied to a report using \code{\link{add_style}}.
}
\examples{
library(reporter)
library(magrittr)

# Get theme
tm <- get_theme("SteelBlue")

# View theme settings
print(tm)
## A style specification: 
#- font_name: 'Arial'
#- font_size: 10
#- text_color: 'DimGrey'
#- title_font_size: 11
#- title_font_bold: TRUE
#- title_font_color: 'SteelBlue'
#- border_color: 'Grey'
#- table_header_background: 'SteelBlue'
#- table_header_font_bold: TRUE
#- table_header_font_color: 'LightGrey'
#- table_body_background: 'White'
#- table_body_stripe: 'WhiteSmoke'
#- table_stub_background: 'SteelBlue'
#- table_stub_font_color: 'LightGrey'
#- table_stub_font_bold: TRUE

# Modify theme
tm$font_size <- 12
tm$title_font_size <- 13

# Create temp file path
tmp <- file.path(tempdir(), "HairAndEyes.html")

# Get data
dat <- as.data.frame(HairEyeColor)

# Create table object
tbl <- create_table(dat[dat$Freq >= 10, ], 
borders = "outside") \%>\% 
titles("Hair and Eye Colors")

# Use modified theme
rpt <- create_report(tmp, output_type = "HTML") \%>\% 
       add_content(tbl) \%>\% 
       add_style(tm)

# Write out the report        
write_report(rpt)

# Uncomment to View report
# file.show(tmp)

}
\seealso{
Other styles: 
\code{\link{add_style}()},
\code{\link{create_style}()},
\code{\link{print.style_spec}()}
}
\concept{styles}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/text_spec.R
\name{print.text_spec}
\alias{print.text_spec}
\title{Prints the text spec}
\usage{
\method{print}{text_spec}(x, ..., verbose = FALSE)
}
\arguments{
\item{x}{The text spec.}

\item{...}{Additional parameters to pass to the underlying print function.}

\item{verbose}{Whether to print in verbose form.  Default is FALSE.}
}
\value{
The text spec, invisibly.
}
\description{
A function to print the text spec.
The \strong{print} function will print the text spec in summary 
form.  To view all parameters, set the \code{verbose} parameter to TRUE.
}
\examples{
txt <- create_text("Lorem ipsum dolor sit amet, consectetur...",
                   align = "left", width = 3)
txt

# A text specification: 6 words
# - text: Lorem ipsum dolor sit amet, consectetur...
# - width: 3
# - align: left
}
\seealso{
\code{\link{create_text}} function to create a text specification.

Other text: 
\code{\link{create_text}()}
}
\concept{text}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_style.R
\name{create_style}
\alias{create_style}
\title{Creates a style object}
\usage{
create_style(
  font_name = NULL,
  font_size = NULL,
  text_color = NULL,
  background_color = NULL,
  title_font_size = NULL,
  title_font_bold = NULL,
  title_font_color = NULL,
  title_background = NULL,
  footnote_font_bold = NULL,
  footnote_font_color = NULL,
  footnote_background = NULL,
  border_color = NULL,
  table_header_background = NULL,
  table_header_font_bold = NULL,
  table_header_font_color = NULL,
  table_body_background = NULL,
  table_body_stripe = NULL,
  table_body_font_color = NULL,
  table_stub_background = NULL,
  table_stub_font_color = NULL,
  table_stub_font_bold = NULL,
  table_label_row_bold = NULL
)
}
\arguments{
\item{font_name}{The name of the font to use on the report.  Valid values
are "Courier", "Arial", or "Times".  The default is "Courier".}

\item{font_size}{The default font size to use for the report.  This font
size will be used for all text, unless overridden by another font size 
parameter.}

\item{text_color}{The default color to use for all text in the report.  This
parameter will apply to the entire report, unless overridden by other
font color settings.}

\item{background_color}{The color to use for the background of the report.
This color will appear everywhere on the document unless overridden 
by another color specification.}

\item{title_font_size}{The font size to use for the title font in points.}

\item{title_font_bold}{Whether to bold the title or not.  Valid values 
are TRUE or FALSE. By default, the title will not be bold.}

\item{title_font_color}{The color to use for the title font.}

\item{title_background}{The background color for the title block.}

\item{footnote_font_bold}{Whether to bold the footnote or not.  Valid values 
are TRUE or FALSE. By default, the footnote will not be bold.}

\item{footnote_font_color}{The font color to use for footnotes.}

\item{footnote_background}{The color to be used for the background of 
footnotes.}

\item{border_color}{The color to use for all borders in the report.
By default, the border will be black.}

\item{table_header_background}{The background color to use in the table header.
This color may be different than the background color used in the table
body.}

\item{table_header_font_bold}{Whether to bold the header labels or not.
Valid values are TRUE and FALSE. By default, the header will not be bold.}

\item{table_header_font_color}{The font color to use on the table header.}

\item{table_body_background}{The background color to use in the body of any 
table in the report.}

\item{table_body_stripe}{The background color to use for every other row
in a table.  The stripe color is used in conjunction with the body background
color to perform table striping.  The stripe color will start on the second row.}

\item{table_body_font_color}{The font color to use for the body of any 
table in the report.}

\item{table_stub_background}{The background color to use for the stub column,
if one exists on the table.}

\item{table_stub_font_color}{The font color to be used for the stub column,
if one exists on the table.}

\item{table_stub_font_bold}{Whether or not to bold the stub column.  Valid
values are TRUE and FALSE.}

\item{table_label_row_bold}{Whether or not to bold the label row on a stub
column.  This parameter is so you can bold the label row, but not the 
values in the stub column.  Valid values are TRUE and FALSE.}
}
\description{
This function will create a style object to control background colors and 
font settings on your report.  The style object can be applied to a 
report using the \code{\link{add_style}} function. Currently, styles may only
be applied to HTML reports.
}
\details{
The style object contains style settings for a report. The style object 
allows you to control background colors and font specifications such as 
font size, font color, and font bold.  The style object can be created
once and reused on many reports.  See the \code{\link{add_style}} function
to learn how to add the style object to a report.  

Note that styles will be applied uniformly to the entire report.  Also note
that at present, styles can be used only on HTML output types. Future versions
of the \strong{reporter} package will provide style support for other 
output types.

On the style object, colors for background and fonts may be passed as 
an RGB hex code or a color name. For example, the color red may be passed as
the hex code "#FF0000" or as the color name "Red". Color names must conform
to HTML/CSS standards.  These color names can be easily discovered on 
the internet.  A sample of common color names is presented below.
}
\section{Color Names}{

Many of the parameters on the style object accept a color name or code.
The values accepted for these parameters follow standard HTML/CSS style
color values.  Below is a sample of common color names
that can be used to specify colors with the \code{create_style} function.
These color names should be passed as a quoted string:
\itemize{
  \item \strong{Primary and Secondary Colors:} Black, White, Red, Yellow, 
  Blue, Green, Orange, Purple and Brown.
  \item \strong{Common Shades:} Beige, Crimson, Gold, Indigo, Ivory, Lavender, 
  Maroon, Navy, Olive, Pink, 
  Plum, RoyalBlue, Silver, Teal, Turquoise, Violet
  \item \strong{Shades of White:} AntiqueWhite, Azure, GhostWhite, 
  SeaShell, Snow, WhiteSmoke
  \item \strong{Shades of Grey:} Grey, Gray, DarkGray, DarkGrey, DimGray, DimGrey, 
  LightGray, LightGrey, SlateGray, SlateGrey
  \item \strong{Shades of Blue:} AliceBlue, CadetBlue, CornflowerBlue, 
  DodgerBlue, PowderBlue, LightBlue, MidnightBlue, SkyBlue, SlateBlue, SteelBlue
  \item \strong{Earth Colors:} Beige, Bisque, BurlyWood, ForestGreen, Khaki,
   Linen, SandyBrown, SaddleBrown, Salmon, SeaGreen, Sienna, Tan, Thistle, Wheat
  \item \strong{Bright Colors:} Aqua, Aquamarine, BlueViolet, Cyan, Fuchia, 
  HotPink, Lime, Magenta, OrangeRed, SpringGreen
}
}

\examples{
library(reporter)
library(magrittr)

# Prepare data
dat <- data.frame(stub = rownames(mtcars), mtcars)
dat <- dat[1:15, ]

# Create temp file path
tmp <- file.path(tempdir(), "HairAndEyes2.html")

# Define custom style
sty <- create_style(font_name = "Arial",
                    font_size = 10,
                    background_color = "WhiteSmoke",
                    border_color = "Grey",
                    title_font_size = 12,
                    title_font_bold = TRUE,
                    title_font_color = "SteelBlue",
                    table_header_background = "Tan",
                    table_header_font_bold = TRUE,
                    table_header_font_color = "White",
                    table_body_background = "White",
                    table_body_stripe = "Wheat", 
                    table_stub_background = "Tan",
                    table_stub_font_color = "White")

# Create table object
tbl <- create_table(dat, borders = "all") \%>\% 
titles("MTCARS Dataset With Style") \%>\% 
column_defaults(width = .5) \%>\% 
define(stub, label = "Car Make and Model", width = 1.5)

# Create report and add style spec
rpt <- create_report(tmp, output_type = "HTML") \%>\% 
       add_content(tbl) \%>\% 
       add_style(style = sty)

# Write out the report        
write_report(rpt)

# Uncomment to View report
# file.show(tmp)
}
\seealso{
Other styles: 
\code{\link{add_style}()},
\code{\link{get_theme}()},
\code{\link{print.style_spec}()}
}
\concept{styles}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_spec.R
\name{create_plot}
\alias{create_plot}
\title{Create plot content}
\usage{
create_plot(x, height, width, borders = "none")
}
\arguments{
\item{x}{The plot to create.  Specifically, this parameter should be 
set to an object returned from a call to \code{\link[ggplot2]{ggplot}}
or \code{\link[survminer]{ggsurvplot}}.  This parameter also accepts 
a path to a JPEG file.  If a path is specified, the image will be appended
to the report at the point the content object is added. For RTF and DOCX files, 
the function additionally accepts a path to an EMF file.}

\item{height}{The height of the plot in the specified units of measure.}

\item{width}{The width of the plot in the specified units of measure.}

\item{borders}{Whether and where to place a border. Valid values are 'top',
'bottom', 'left', 'right', 'all', 'none', and 'outside'.  
Default is 'none'.  The 'left', 'right', and 'outside' 
border specifications do not apply to TXT reports.}
}
\value{
The plot specification.
}
\description{
Function to create a plot specification that can be 
added as content to a report. The \code{create_plot} function can 
be used to include charts, graphs, and figures on a statistical report.  
The function supports plot objects returned by 
\code{\link[ggplot2]{ggplot}} or \code{\link[survminer]{ggsurvplot}}.  
It does not support the Base R 
\code{plot} function.
}
\details{
To add a plot to a report, use the \code{create_plot} function.  The 
function allows you to set a width and height for the plot.  The
function will preserve any other geometries you apply to the plot.  See
the \code{\link{add_content}} function to control page breaking and 
blanks spaces above or below the plot.    

A plot specification accepts a \code{\link{page_by}} function.  If a page by
is applied to the plot, the plot data will be subset by the page by 
variable, and re-run for each subset.

The plot specification also accepts titles and footnotes.  See the 
\code{\link{titles}} and \code{\link{footnotes}} functions for further 
details.

The function 
also accepts a path to an image file stored on the file system instead of a 
plot object.  This
functionality was added to allow the user to create figures from other
plotting packages.  If you pass an image path, the image will be inserted
into the report at the location specified. The system supports JPEG images 
for all output types that support graphics. EMF images are also supported for
DOCX and RTF outputs. Note that EMF format may not render properly in 
all operating environments, and JPEG is the preferred image type.
}
\examples{
library(reporter)
library(ggplot2)
library(magrittr)

# Create temp file path
tmp <- file.path(tempdir(), "mtcars.rtf")

# Create ggplot
p <- ggplot(mtcars, aes(x=cyl, y=mpg)) + geom_point()

# Create plot object
plt <- create_plot(p, height = 4, width = 8)

rpt <- create_report(tmp, font = "Arial", output_type = "RTF") \%>\%
  page_header("Client", "Study: XYZ") \%>\% 
  titles("Figure 1.0", "MTCARS Miles per Cylinder Plot") \%>\%
  set_margins(top = 1, bottom = 1) \%>\%
  add_content(plt) \%>\%
  footnotes("* Motor Trend, 1974") \%>\% 
  page_footer("Time", "Confidential", "Page [pg] of [tpg]")

# Write out report
write_report(rpt)

# Uncomment to view RTF file
# shell.exec(tmp)
}
\seealso{
\code{\link{titles}} to add a title block to the plot,  
\code{\link{footnotes}} to add footnotes, and \code{\link{add_content}} 
to add the plot object to a report.

Other plot: 
\code{\link{print.plot_spec}()}
}
\concept{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_spec.r
\name{footnotes}
\alias{footnotes}
\title{Adds a footnote block}
\usage{
footnotes(
  x,
  ...,
  align = "left",
  blank_row = "above",
  borders = "none",
  valign = NULL,
  width = NULL,
  footer = FALSE,
  italics = FALSE,
  columns = 1,
  font_size = NULL
)
}
\arguments{
\item{x}{The object to assign footnotes to.}

\item{...}{A set of footnote strings.}

\item{align}{The position to align the footnotes.  Valid values are: 'left',
'right', 'center', or 'centre'.}

\item{blank_row}{Whether to print a blank row above or below the footnote.
Valid values are 'above', 'below', 'both', or 'none'.  Default is 'above'.}

\item{borders}{Whether to print a border above or below the footnote. Valid
values are 'top', 'bottom', 'outside', 'inside', 'all',  or 'none'.  
Default is 'none'.  
For fixed width reports, the 
border character will be taken from the value of the \code{uchar} parameter
on the \code{\link{options_fixed}} function.  The 
'left', 'right', 'outside', and 'inside' border specifications only apply 
to RTF, HTML, PDF, and DOCX reports.}

\item{valign}{The vertical position to align the footnotes.  Valid
values are: 'top' and 'bottom'.  For footnotes attached to a report,
default is 'bottom'.  For footnotes attached to content, default is 'top'.}

\item{width}{The width of the footnotes block.  If the footnotes are attached
to the report, valid values are 'page' or a numeric width, and the default
is 'page'. If the footnotes are attached to the
table, plot, or text content, the valid values are 'page', 'content' or a 
numeric value, and the default is 'content'.  The value 'content' means the 
footnotes will be aligned to the width of the table, plot, or text
content.  The value
'page' means the footnotes will be aligned to the width of the page. 
In addition to these two convenience settings, you 
may also specify a specific width in the current unit of measure.  The
unit of measure is determined by the 'units' parameter on 
\code{\link{create_report}}.}

\item{footer}{Whether to put the footnotes in the page footer. Valid
values are TRUE and FALSE.  Default is FALSE.  This parameter only works
with RTF and DOCX output types.  It also only applies to footnotes
assigned to the report object.  Footnotes in the page footer will 
appear on every page, and be the same throughout the report.}

\item{italics}{A TRUE or FALSE value indicating whether the footnote 
text shoud be in italics font.  If TRUE, the entire footnote will be 
in italics.}

\item{columns}{The number of columns for the footnote block. Valid values
are 1, 2, and 3.  Default is 1.  If this parameter is set to 2, the footnote
block will be split into two columns, each aligned to the outside.  If 
this parameter is set to 3, the title block will be split into 3 columns,
with the outer columns aligned to the outside and the middle column
aligned center.  Footnotes are assigned to cells from top left to bottom right.}

\item{font_size}{The font size to use for the footnote block.  The font size
of the report will be used by default.  Valid values are 8, 9, 10, 11, 12
This parameter only applies to variable-width RTF, HTML, PDF, and DOCX 
output types.}
}
\value{
The modified report.
}
\description{
The \code{footnotes} function adds one or more footnotes to the report.  
If added to the report specification, the footnotes will
be added to the page template, and thus appear on each page of the report.
Footnotes may also be added directly to table, text, or plot content.
}
\details{
The \code{footnotes} function accepts a set of strings of the desired 
footnote text. The footnotes may be aligned center, left or right using 
the align parameter. The user is responsible for adding desired footnote 
symbols. Footnote symbols will not be generated automatically.

If footnotes are assigned to the report,
alignment will be oriented to the page body.  If footnotes are assigned to
a table or text, alignment will be oriented to the edge of the content.

One footnote function accepts up to 25 footnotes.  However, multiple footnote 
blocks may be added to the same object.  

Blank rows above or below the footnote block may be controlled using the 
\code{blank_row} parameter.

Each footnote string must fit within the available space.  The \strong{reporter}
package will not wrap footnotes on fixed-width reports.  
If a footnote does not fit within the 
available space, a warning will be generated and the footnote will be 
truncated.  In these situations, either
shorten the footnote or split it into multiple footnotes that each fit within 
the available space.
}
\examples{
library(reporter)
library(magrittr)

# Create a temporary file
tmp <- file.path(tempdir(), "expenses.txt")

# Prepare data
dat <- data.frame(category = rownames(USPersonalExpenditure),
                  USPersonalExpenditure, stringsAsFactors = FALSE)

# Define table
tbl <- create_table(dat) \%>\% 
  titles("Table 1.0", "US Personal Expenditures from 1940 - 1960") \%>\% 
  column_defaults(from = X1940, to = X1960, width = .6, format = "$\%.2f") \%>\%
  define(category, label = "Category") \%>\% 
  define(X1940, label = "1940") \%>\% 
  define(X1945, label = "1945") \%>\% 
  define(X1950, label = "1950") \%>\% 
  define(X1955, label = "1955") \%>\% 
  define(X1960, label = "1960") \%>\% 
  footnotes("* In billions of dollars")

# Define report
rpt <- create_report(tmp, orientation="portrait") \%>\%
  add_content(tbl) 

# Write the report
write_report(rpt)

# Display in console
writeLines(readLines(tmp, encoding = "UTF-8"))

#                               Table 1.0
#               US Personal Expenditures from 1940 - 1960
# 
#     Category                1940    1945    1950    1955    1960
#     ------------------------------------------------------------
#     Food and Tobacco      $22.20  $44.50  $59.60  $73.20  $86.80
#     Household Operation   $10.50  $15.50  $29.00  $36.50  $46.20
#     Medical and Health     $3.53   $5.76   $9.71  $14.00  $21.10
#     Personal Care          $1.04   $1.98   $2.45   $3.40   $5.40
#     Private Education      $0.34   $0.97   $1.80   $2.60   $3.64
# 
#     * In billions of dollars
}
\seealso{
Other report: 
\code{\link{add_content}()},
\code{\link{create_report}()},
\code{\link{footer_image}()},
\code{\link{header_image}()},
\code{\link{options_fixed}()},
\code{\link{page_by}()},
\code{\link{page_footer}()},
\code{\link{page_header}()},
\code{\link{print.report_spec}()},
\code{\link{set_margins}()},
\code{\link{title_header}()},
\code{\link{titles}()},
\code{\link{write_report}()}
}
\concept{report}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/attributes.R
\name{span}
\alias{span}
\title{Defines a spanning header attribute}
\usage{
span(
  from,
  to,
  label = NULL,
  label_align = "center",
  level = 1,
  underline = TRUE
)
}
\arguments{
\item{from}{The starting column to span.  Spanning columns are defined as
range of columns 'from' and 'to'. The columns may be identified by position, 
or by quoted or unquoted variable names. 
The \code{from} parameter is required.}

\item{to}{The ending column to span.  Spanning columns are defined as
range of columns 'from' and 'to'. The columns may be identified by position,
or by quoted or unquoted variable names.  
The \code{to} parameter is required.}

\item{label}{The label to apply to the spanning header.}

\item{label_align}{The alignment to use for the label. Valid values are 
"left", "right", "center", and "centre".  The default for spanning columns
is "center".}

\item{level}{The level to use for the spanning header.  The lowest
spanning level is level 1, the next level above is level 2, and so on.  
By default, the level is set to 1.}

\item{underline}{A TRUE or FALSE value indicating whether the spanning
header should be underlined.  Default is TRUE.}
}
\value{
The spanning header attribute object.
}
\description{
Create a spanning header attribute object that can be
attached to a data frame and passed to the \code{\link{create_table}}
function.  This attribute is used internally by other packages in the 
\strong{r-sassy} system.
}
\details{
A spanning header is a label and underline that spans one or more 
columns.  A spanning header is defined minimally by identifying the
column range to be spanned, and a label.  A label alignment 
may also be supplied.

The spanning column range is defined by the \code{from} and \code{to} 
parameters.  The range identifies a contiguous set of variables on the data.
Variables can be identified by position, a quoted variable name, or an 
unquoted variable name.
}
\seealso{
Other attributes: 
\code{\link{ftn}()},
\code{\link{ttl}()}
}
\concept{attributes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_spec.r
\name{page_by}
\alias{page_by}
\title{Adds a page by variable}
\usage{
page_by(
  x,
  var,
  label = NULL,
  align = "left",
  blank_row = "below",
  borders = "none",
  format = NULL,
  bold = FALSE
)
}
\arguments{
\item{x}{The report specification to assign the page by to.}

\item{var}{The page by variable.  There can be only one page by per report, 
and one page by variable.  The page by can be passed either quoted or 
unquoted.}

\item{label}{A label to be used as a prefix to the page by variable value.
By default, the label will be assigned to the variable name.  Alternatively,
you may specify a string value to use for the label.}

\item{align}{How to align the page by.  Default value is 'left'. Valid
values are 'left', 'right', 'center', or 'centre'.}

\item{blank_row}{Indicates whether a blank row is desired above or below
the page by.  Default value is 'none'.  Valid values are 'above', 'below',
'both', or 'none'.}

\item{borders}{Whether and where to place a border. Valid values are 'top',
'bottom', 'left', 'right', 'all', 'outside', or 'none'.  Default is "none".  
The 'left' and 'right' border specifications only apply to RTF, HTML, PDF,
and DOCX reports.}

\item{format}{The format to use for the page by column data.  The format can 
be a string format, a formatting function, a lookup list, a user-defined
format, or a formatting list. All formatting is performed by the 
\code{\link[fmtr]{fapply}} function from
the \code{\link[fmtr]{fmtr}} package.  For 
a list of common formatting codes, see \link[fmtr]{FormattingStrings}.}

\item{bold}{A parameter to bold the page by text. Valid values are TRUE, FALSE, 
'label' or 'value'. Default is FALSE. A value of TRUE bolds both the label 
and the page by value. The value 'label' bolds only the label, while a value
of 'value' bolds only the page by value. This parameter 
applies to variable-width RTF, HTML, PDF, and DOCX output types only.}
}
\description{
The \code{page_by} function adds a page by variable to a report, table,
or plot. 
The page by will generate a page break for each value of the page by variable.
A label will appear above the content showing the page
by variable value. You must be sort the data by the
page by variable prior to reporting.
}
\details{
Only one page by function is allowed per report, table, or plot.  The page by 
label will 
appear on all pages of the object.  The page by label may be aligned on the 
left, right, or center. Use the \code{align} parameter to specify the 
alignment. 

You must sort the data by the page by variable prior to reporting.
The page by labels will appear in the sorted order.  Failure to sort 
the page by variable prior to reporting may produce unexpected results.
}
\examples{
library(reporter)
library(magrittr)

# Create temp file path
tmp <- file.path(tempdir(), "iris.txt")

# Sample and sort iris data frame
dat <- iris[sample(1:150, 50), ]
dat <- dat[order(dat$Species), ]

# Create table
tbl <- create_table(dat) \%>\% 
  page_by(Species, "Species: ") \%>\%  
  define(Species, visible = FALSE)
  
# Create the report object
rpt <- create_report(tmp, orientation = "portrait") \%>\% 
  page_header("Sponsor: Iris Society", "Study: flowers") \%>\% 
  titles("Table 2.0", "IRIS Sample Report with Page By") \%>\% 
  add_content(tbl) \%>\% 
  page_footer(Sys.time(), right = "Page [pg] of [tpg]")

# Write the report to the file system
write_report(rpt)

# Write report to console
writeLines(readLines(tmp, encoding = "UTF-8"))

# Sponsor: Iris Society                                           Study: flowers
#                                  Table 2.0
#                       IRIS Sample Report with Page By
# 
#              Species: setosa
# 
#              Sepal.Length Sepal.Width Petal.Length Petal.Width
#              -------------------------------------------------
#                       5.4         3.9          1.7         0.4
#                       4.9         3.1          1.5         0.1
#                       4.8         3.1          1.6         0.2
#                       5.1         3.5          1.4         0.3
#                         5         3.5          1.6         0.6
#                         5         3.3          1.4         0.2
#                       4.4           3          1.3         0.2
#                       5.1         3.5          1.4         0.2
#                       5.4         3.4          1.5         0.4
#                       4.9         3.6          1.4         0.1
#                       4.6         3.1          1.5         0.2
#                       4.6         3.2          1.4         0.2
#                       5.1         3.3          1.7         0.5
# ...
# 2020-10-25 19:33:35                                                Page 1 of 3
#
# Sponsor: Iris Society                                           Study: flowers
#                                  Table 2.0
#                       IRIS Sample Report with Page By
# 
#              Species: versicolor
# 
#              Sepal.Length Sepal.Width Petal.Length Petal.Width
#              -------------------------------------------------
#                       4.9         2.4          3.3           1
#                       6.3         3.3          4.7         1.6
#                       6.1         2.8          4.7         1.2
#                         6         2.9          4.5         1.5
#                       6.7           3            5         1.7
#                       5.6           3          4.5         1.5
#                       5.8         2.7          4.1           1
#                       6.7         3.1          4.7         1.5
#                       6.1         2.9          4.7         1.4
#                         5           2          3.5           1
#                       5.9         3.2          4.8         1.8
#                       5.5         2.5            4         1.3
#                         7         3.2          4.7         1.4
#                       6.3         2.3          4.4         1.3
#                       6.1         2.8            4         1.3
#                         6         2.2            4           1
#                       5.5         2.6          4.4         1.2
#                         6         3.4          4.5         1.6
#                         5         2.3          3.3           1
#                       5.5         2.4          3.7           1
# ...
# 2020-10-25 19:33:35                                                Page 2 of 3
#
# Sponsor: Iris Society                                           Study: flowers
#                                  Table 2.0
#                       IRIS Sample Report with Page By
# 
#              Species: versicolor
# 
#              Sepal.Length Sepal.Width Petal.Length Petal.Width
#              -------------------------------------------------
#                       6.3         3.4          5.6         2.4
#                       7.9         3.8          6.4           2
#                       6.7         3.1          5.6         2.4
#                       6.2         2.8          4.8         1.8
#                       6.7         3.3          5.7         2.5
#                       6.2         3.4          5.4         2.3
#                       5.6         2.8          4.9           2
#                       7.7         3.8          6.7         2.2
#                       7.7         2.6          6.9         2.3
#                       6.9         3.1          5.4         2.1
#                       6.5         3.2          5.1           2
#                       6.1         2.6          5.6         1.4
#                       5.7         2.5            5           2
#                       6.5           3          5.8         2.2
#                       6.3         2.8          5.1         1.5
#                       7.6           3          6.6         2.1
#                       6.3         2.5            5         1.9
# ...
# 2020-10-25 19:33:35                                                Page 3 of 3 
}
\seealso{
\code{\link{create_table}} to create a table, and 
\code{\link{create_plot}} to create a plot.

Other report: 
\code{\link{add_content}()},
\code{\link{create_report}()},
\code{\link{footer_image}()},
\code{\link{footnotes}()},
\code{\link{header_image}()},
\code{\link{options_fixed}()},
\code{\link{page_footer}()},
\code{\link{page_header}()},
\code{\link{print.report_spec}()},
\code{\link{set_margins}()},
\code{\link{title_header}()},
\code{\link{titles}()},
\code{\link{write_report}()}
}
\concept{report}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/attributes.R
\name{ftn}
\alias{ftn}
\title{Creates a footnote attribute}
\usage{
ftn(
  footnotes,
  align = "left",
  blank_row = "above",
  borders = "none",
  valign = NULL,
  width = NULL
)
}
\arguments{
\item{footnotes}{A vector of footnote strings.}

\item{align}{The position to align the footnotes.  Valid values are: 'left',
'right', 'center', or 'centre'.}

\item{blank_row}{Whether to print a blank row above or below the footnote.
Valid values are 'above', 'below', 'both', or 'none'.  Default is 'above'.}

\item{borders}{Whether to print a border above or below the footnote. Valid
values are 'top', 'bottom', 'outside', 'inside', 'all',  or 'none'.  
Default is 'none'.}

\item{valign}{The vertical position to align the footnotes.  Valid
values are: 'top' and 'bottom'.}

\item{width}{The width of the footnotes block.}
}
\value{
The footnote attribute object.
}
\description{
The \code{ftn} function creates a footnote attribute,
which may be attached as an attribute to a data frame and passed
into \code{\link{create_table}}. This function is used internally to
the system.
}
\seealso{
\code{\link{footnotes}} to create a footnote.

Other attributes: 
\code{\link{span}()},
\code{\link{ttl}()}
}
\concept{attributes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_spec.r
\name{header_image}
\alias{header_image}
\title{Insert an image into the page header}
\usage{
header_image(x, image_path, height, width, align = "left")
}
\arguments{
\item{x}{The report object.}

\item{image_path}{The path of image file. Function accepts JPG images 
for RTF, DOCX, PDF, and HTML output types. Function also accepts 
PNG images for RTF, DOCX, and HTML output types.}

\item{height}{Desired height of the image, in the report units of measure.}

\item{width}{Desired width of the image, in the report units of measure.}

\item{align}{Alignment of the image. Valid values are "left", "right", 
"center", or "centre". The default value is "left".}
}
\value{
The modified report specification.
}
\description{
The \code{header_image} function inserts an image into the page header. The function
can be used to add a logo or other graphical element to the report.
The image will then appear at the top of each page.
}
\details{
The function assumes 
the existence of a page header inserted by the \code{\link{page_header}} function. 
If there is no page header, an error will be generated during the rendering
process in \code{\link{write_report}}. If there is text
in the same cell as the image, the text will be disregarded. You may add text 
to other cells in the header.

The "align" parameter determines which page header cell the image will be inserted.
That is, if you align an image "left", it will be inserted into the left
column of the page header. Options are "left", "right", and "center".  To
insert images in multiple cells, you may use multiple calls to 
\code{image_header}.

The page header cell width will take priority over the image width.  That is,
if you set the image width larger than the cell width, the image width
will be reduced to fit the cell.  To make a wider image, increase the cell
width on the page header.  

Note that page header cells can be removed
entirely by setting the width to zero.  This feature can give you more room
for a wide image. For instance, to create a banner image over the entire report, 
align the image "center" and set the left and right page header column 
widths to zero.

See the \code{\link{page_header}} function for more information on
setting header column widths.  For a demonstration of 
a header image insertion, see \code{vignette("reporter-hfimage")}.
}
\seealso{
Other report: 
\code{\link{add_content}()},
\code{\link{create_report}()},
\code{\link{footer_image}()},
\code{\link{footnotes}()},
\code{\link{options_fixed}()},
\code{\link{page_by}()},
\code{\link{page_footer}()},
\code{\link{page_header}()},
\code{\link{print.report_spec}()},
\code{\link{set_margins}()},
\code{\link{title_header}()},
\code{\link{titles}()},
\code{\link{write_report}()}
}
\concept{report}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_spec.r
\name{title_header}
\alias{title_header}
\title{Adds a title header block}
\usage{
title_header(
  x,
  ...,
  right = "",
  blank_row = "below",
  borders = "none",
  width = NULL
)
}
\arguments{
\item{x}{The object to assign titles to.  Valid objects are a report,
table, text, or plot specification.}

\item{...}{A set of title strings.}

\item{right}{A set of header strings to be shown on the right side of the
title header.  Pass the header strings as a vector of strings.}

\item{blank_row}{Where to place a blank row.  Valid values are 'above',
'below', 'both', or 'none'.  Default is 'below'.}

\item{borders}{Whether and where to place a border. Valid values are 'top',
'bottom', 'left', 'right', 'outside', 'inside', 'all', or 'none'.  
Default is 'none'.   The 
'left', 'right', 'outside', and 'inside' border specifications only 
apply to RTF, HTML, PDF and DOCX reports.}

\item{width}{The width of the title header.  If the title header is attached
to the report, valid values are 'page' or a numeric width, and the default
is 'page'. If the title header is attached to the
table, plot, or text content, the valid values are 'page', 'content' or a 
numeric value, and the default is 'content'.  The value 'content' means the 
footnotes will be aligned to the width of the table, plot, or text
content.  The value
'page' means the footnotes will be aligned to the width of the page. 
In addition to these two convenience settings, you 
may also specify a specific width in the current units of measure.  The
units of measure is determined by the 'units' parameter on 
\code{\link{create_report}}.}
}
\value{
The modified report.
}
\description{
This function adds a title header to an object.  A title header is a special
type of title layout that has titles on the left and header information
on the right.
}
\details{
The \code{title_header} function accepts a set of strings of the desired 
title text, and a vector of header strings. The titles will appear on the 
left of the title header, and the header strings on the right. To
specify multiple titles for the block, pass them to the function 
as separate strings.

Title headers may be assigned to a report, a table, a text specification, 
or a plot. If assigned to the report, the title header will appear 
at the top of the page, and be repeated for every page of the report.  
If the title header is assigned to  
content, the titles will appear above the content, and be repeated if the 
content breaks to the next page.  

One title header function accepts up to 10 titles. Blank rows above or below 
the title block may be controlled using the 
\code{blank_row} parameter.

Each title string must fit within the available space.  The \strong{reporter}
package will not wrap titles on fixed-width reports.  
If a title does not fit within the 
available space, an error will be generated.  In these situations, either
shorten the title or split it into multiple titles that each fit within the
available space.
}
\examples{
library(reporter)
library(magrittr)

# Create a temporary file
tmp <- file.path(tempdir(), "expenses.txt")

# Prepare data
dat <- data.frame(category = rownames(USPersonalExpenditure),
                  USPersonalExpenditure, stringsAsFactors = FALSE)

# Define table
tbl <- create_table(dat) \%>\% 
  title_header("Table 1.0", "US Personal Expenditures from 1940 - 1960",
               right = c("Page [pg] of [tpg]", "World Almanac")) \%>\% 
  column_defaults(from = X1940, to = X1960, width = .6, format = "$\%.2f") \%>\%
  define(category, label = "Category") \%>\% 
  define(X1940, label = "1940") \%>\% 
  define(X1945, label = "1945") \%>\% 
  define(X1950, label = "1950") \%>\% 
  define(X1955, label = "1955") \%>\% 
  define(X1960, label = "1960") \%>\% 
  footnotes("* In billions of dollars")

# Define report
rpt <- create_report(tmp, orientation="portrait") \%>\%
  add_content(tbl) 

# Write the report
write_report(rpt)

# Display in console
writeLines(readLines(tmp, encoding = "UTF-8"))

#     Table 1.0                                        Page 1 of 1
#     US Personal Expenditures from 1940 - 1960      World Almanac
# 
#     Category                1940    1945    1950    1955    1960
#     ------------------------------------------------------------
#     Food and Tobacco      $22.20  $44.50  $59.60  $73.20  $86.80
#     Household Operation   $10.50  $15.50  $29.00  $36.50  $46.20
#     Medical and Health     $3.53   $5.76   $9.71  $14.00  $21.10
#     Personal Care          $1.04   $1.98   $2.45   $3.40   $5.40
#     Private Education      $0.34   $0.97   $1.80   $2.60   $3.64
# 
#     * In billions of dollars
}
\seealso{
Other report: 
\code{\link{add_content}()},
\code{\link{create_report}()},
\code{\link{footer_image}()},
\code{\link{footnotes}()},
\code{\link{header_image}()},
\code{\link{options_fixed}()},
\code{\link{page_by}()},
\code{\link{page_footer}()},
\code{\link{page_header}()},
\code{\link{print.report_spec}()},
\code{\link{set_margins}()},
\code{\link{titles}()},
\code{\link{write_report}()}
}
\concept{report}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_spec.r
\name{define}
\alias{define}
\title{Defines a column}
\usage{
define(
  x,
  vars,
  label = NULL,
  format = NULL,
  align = NULL,
  label_align = NULL,
  width = NULL,
  visible = TRUE,
  n = NULL,
  blank_after = FALSE,
  blank_before = FALSE,
  dedupe = FALSE,
  id_var = FALSE,
  page_wrap = FALSE,
  page_break = FALSE,
  indent = NULL,
  label_row = FALSE,
  standard_eval = FALSE,
  style = NULL,
  group_border = FALSE,
  group_cohesion = FALSE,
  break_label = NULL
)
}
\arguments{
\item{x}{The table spec.}

\item{vars}{The variable name or names to define a column for.  Names may
be quoted or unquoted.  If defining for multiple variables, 
pass them as a vector of names.  If you want to pass an R variable of names,
set the \code{standard_eval} parameter to TRUE .
The \code{standard_eval} parameter is useful when writing functions that construct
reports dynamically.}

\item{label}{The label to use for the column header.  If a label is assigned
to the label column attribute, it will be used as a default.  Otherwise,
the column name will be used.}

\item{format}{The format to use for the column data.  The format can 
be a string format, a formatting function, a lookup list, a user-defined
format, or a formatting list. 
All formatting is performed by the \code{\link[fmtr]{fapply}} function from
the \code{\link[fmtr]{fmtr}} package.  For 
a list of common formatting codes, see \link[fmtr]{FormattingStrings}.}

\item{align}{The column alignment.  Valid values are "left", "right", 
"center", and "centre".  By default, text columns will be left aligned
and numeric columns will be right aligned.}

\item{label_align}{How to align the header labels for this column.
Valid values are "left", "right", "center", and "centre".  By default, 
the label alignment will follow any alignment set on the column \code{align}
parameter.}

\item{width}{The width of the column in the specified units of measure.
The units of measure are specified on the \code{units} parameter of the
\code{\link{create_report}} function.  If no width is supplied, the
\code{\link{write_report}} function will assign a default width based on the 
width of the column data and the label.  \code{write_report} will not set a 
column width less than the width of the largest word in the data or label.
In other words, \code{write_report} will not break words.}

\item{visible}{Whether or not the column should be visible on the report.
This parameter can be used as a simple way to drop columns from the report.}

\item{n}{The n value to place in the "N=" header label.  Formatting for
the n value will be performed by the formatting function assigned to the 
\code{n_format} parameter on \code{\link{create_table}}.}

\item{blank_after}{Whether to place a blank row after unique values of this
variable.  Valid values are TRUE or FALSE.  Default is FALSE.}

\item{blank_before}{Whether to place a blank row before unique values of this
variable.  Valid values are TRUE or FALSE.  Default is FALSE.}

\item{dedupe}{Whether to dedupe the values for this variable.  Variables
that are deduped only show the value on the first row of each group.  This 
option is commonly used for grouping variables.}

\item{id_var}{Whether this variable should be considered an ID variable.
ID variables are retained on each page when the page is wrapped. ID variables
are also moved to the far left of the page.}

\item{page_wrap}{Force a page wrap on this variable.  A page wrap is a vertical
page break necessary when the table is too wide to fit on a single page.
The excess variables will be wrapped to the next page.  Page wraps will
continue until all columns are displayed.  Use the \code{id_var}
parameter to identify rows across wrapped pages.}

\item{page_break}{You may control when page breaks occur by defining
a page break variable yourself, and setting this parameter to TRUE for
that variable.  Only one page break variable can be defined per table.
If two or more variables are defined as a page break, an error will be 
generated.}

\item{indent}{How much to indent the column values.  The parameter takes a 
numeric value that will be interpreted according to the \code{units} 
(Unit Of Measure) setting on the report.  This parameter can be used to 
help create a stub column.  The default value is NULL, meaning the column
should not be indented.  See the \code{\link{stub}} function for additional
information on creating a stub column.}

\item{label_row}{Whether the values of the variable should be used to
create a label row.  Valid values are TRUE or FALSE.  Default is FALSE.
If \code{label_row} is set to TRUE, the dedupe parameter will also be 
set to TRUE.  This parameter is often used in conjunction with the 
\code{\link{stub}} function and \code{indent} parameter to create a 
stub column.}

\item{standard_eval}{A TRUE or FALSE value indicating whether to 
use standard evaluation on the \code{vars} parameter value.  Default is
FALSE.  Set this parameter to TRUE if you want to pass the \code{vars}
value(s) using a variable.}

\item{style}{A \code{\link{cell_style}} object that defines the desired
style for this column.  The cell style object can be used to define 
conditional styling.}

\item{group_border}{Adds a horizontal border at the end of a group. 
Valid values are TRUE or FALSE. Default is FALSE. The group border is
another way to separate analysis groups, and can be used in addition
to or instead of the "blank_before" and "blank_after" parameters.}

\item{group_cohesion}{Default is FALSE. If TRUE, it would avoid splitting the 
same group across multiple pages with default strength 0.2. Users can increase
the strength to make groups on the same page more likely. The accepted value
is 0 to 1.}

\item{break_label}{Decide whether to repeat the grouping value when the group 
spans multiple pages. If set it as TRUE, when a group spans, the break label will 
be displayed as a label of the first row. If set it as a text such as "(Continued)",
it would be used as a suffix text. E.g., Investigation (Continued).}
}
\value{
The modified table spec.
}
\description{
A function to define a table column.  The
\code{define} function contains a variety of a parameters to control the 
appearance of the report.  Using the \code{define} function, you can control
simple options like column alignment and width, but also control more 
sophisticated options like page wrapping and page breaking.
}
\details{
Column definitions are optional.  By default, all columns in the data
are displayed in the order assigned to the data frame. 

The report will use attributes assigned to the data frame 
such as 'width', 'justify', 'label', and 'format'.  In other words, 
some control over the column 
formatting is available by manipulating the data frame attributes prior
to assigning the data frame to \code{create_table}.  See 
\code{\link{create_table}} for more details.

The \code{define} function is used to provide additional control over
column appearance.  For example, you may use the \code{define} function
to assign an "N=" population count, eliminate duplicates from the column,
or place a blank row after each unique value of the variable. 
See the parameter documentation for additional options.

Some of the parameters on the \code{define} function are used in the 
creation of a table stub.  Specifically, the \code{label_row} and 
\code{indent} parameters participate in the creation of the stub column.
See the \code{\link{stub}} function for further
information.

A single column definition may be defined for multiple variables.  
To create a definition for multiple variables, pass the variables as
a quoted or unquoted vector.  When creating a single definition for 
multiple variables, the parameters will be unified across those variables.
Note that some parameters (such as \code{page_break}) may only be set
once per report, and cannot be shared across multiple variables.
}
\examples{
library(reporter)
library(magrittr)
 
# Create temp file name
tmp <- file.path(tempdir(), "mtcars.txt")

# Prepare data
dat <- mtcars[1:10, ]
dat <- data.frame(vehicle = rownames(dat), dat, stringsAsFactors = FALSE)

# Define table
tbl <- create_table(dat, show_cols = 1:8) \%>\% 
  define(vehicle, label = "Vehicle", width = 3, id_var = TRUE, align = "left") \%>\% 
  define(mpg, label = "Miles per Gallon", width = 1) \%>\% 
  define(cyl, label = "Cylinders", format = "\%.1f") \%>\% 
  define(disp, label = "Displacement") \%>\% 
  define(hp, label = "Horsepower", page_wrap = TRUE) \%>\% 
  define(drat, visible = FALSE) \%>\% 
  define(wt, label = "Weight") \%>\% 
  define(qsec, label = "Quarter Mile Time", width = 1.5) 


# Create the report
rpt <- create_report(tmp, orientation = "portrait") \%>\% 
  titles("Listing 2.0", "MTCARS Data Listing with Page Wrap") \%>\% 
  add_content(tbl, align = "left") \%>\% 
  page_footer(right = "Page [pg] of [tpg]")

# Write the report
write_report(rpt)

# Send report to console for viewing
writeLines(readLines(tmp, encoding = "UTF-8"))

#                                  Listing 2.0
#                       MTCARS Data Listing with Page Wrap
# 
#                                         Miles per
# Vehicle                                    Gallon Cylinders Displacement
# ------------------------------------------------------------------------
# Mazda RX4                                      21       6.0          160
# Mazda RX4 Wag                                  21       6.0          160
# Datsun 710                                   22.8       4.0          108
# Hornet 4 Drive                               21.4       6.0          258
# Hornet Sportabout                            18.7       8.0          360
# Valiant                                      18.1       6.0          225
# Duster 360                                   14.3       8.0          360
# Merc 240D                                    24.4       4.0        146.7
# Merc 230                                     22.8       4.0        140.8
# Merc 280                                     19.2       6.0        167.6
# 
# ...
# 
#                                                                    Page 1 of 2
#                                  Listing 2.0
#                       MTCARS Data Listing with Page Wrap
# 
# Vehicle                              Horsepower Weight  Quarter Mile Time
# -------------------------------------------------------------------------
# Mazda RX4                                   110   2.62              16.46
# Mazda RX4 Wag                               110  2.875              17.02
# Datsun 710                                   93   2.32              18.61
# Hornet 4 Drive                              110  3.215              19.44
# Hornet Sportabout                           175   3.44              17.02
# Valiant                                     105   3.46              20.22
# Duster 360                                  245   3.57              15.84
# Merc 240D                                    62   3.19                 20
# Merc 230                                     95   3.15               22.9
# Merc 280                                    123   3.44               18.3
# 
# ...
# 
#                                                                    Page 2 of 2
}
\seealso{
Other table: 
\code{\link{column_defaults}()},
\code{\link{create_table}()},
\code{\link{print.table_spec}()},
\code{\link{spanning_header}()},
\code{\link{stub}()}
}
\concept{table}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_style.R
\name{cell_style}
\alias{cell_style}
\title{A style specification for a table cell}
\usage{
cell_style(indicator = NULL, bold = FALSE)
}
\arguments{
\item{indicator}{A keyword or column name to indicate which rows the cell
style should be applied to. Valid keywords are "labelrow", "blankrow", or
"datarow". To use an indicator column, create a column on the input dataset
that has a TRUE value for each row that you want the cell style applied.
Then pass the name of the column to the indicator parameter. If you do not
want to see the indicator column on the report, set the "visible" property 
on the \code{\link{define}} function to FALSE for that column. 
The default value of the indicator parameter is NULL, 
meaning to apply the style to all rows.}

\item{bold}{Whether to bold the text in the cell.  Valid values are TRUE and
FALSE.  The default is FALSE.}
}
\description{
A class to define the style for a cell in a table.  This
class can be assigned to the "style" parameter of a \code{\link{define}}
function, a \code{\link{stub}} function, or a \code{\link{column_defaults}}  
function.  When assigned, the cell style will apply to some or all of the cells 
in the relevant columns.  The 
"indicator" parameter identifies a column in the table to trigger the style.
Alternatively, the "labelrow", "blankrow", or "datarow" shortcuts may be used
to identify cells for styling.
}
\examples{
library(reporter)
library(magrittr)

# Create temporary path
tmp <- file.path(tempdir(), "table1.rtf")

# Prepare data
df <- data.frame(names = rownames(mtcars), mtcars[, 1:3])

# Set indicator variable
df$mpgind <- ifelse(df$mpg > 20, TRUE, FALSE)

# Create table
tbl <- create_table(df, first_row_blank = TRUE, 
                    header_bold = TRUE, borders = c("top", "bottom")) \%>\% 
  column_defaults(style = cell_style(bold = TRUE, indicator = "mpgind")) \%>\%
  define(names, label = "Car Name") \%>\%
  define(mpg, label = "Miles Per Gallon") \%>\% 
  define(cyl, label = "Cylinders") \%>\% 
  define(disp, label = "Displacement")  \%>\% 
  define(mpgind, visible = FALSE) \%>\%
  titles("Table 1.0", "MTCARS with Indicator Variable", 
         borders = "none", bold = TRUE, font_size = 11) \%>\% 
  footnotes("* Motor Trend, 1974", borders = "none", blank_row = "none")

# Create report and add custom style
rpt <- create_report(tmp, output_type = "RTF", font = "Arial") \%>\% 
  add_content(tbl) 

# Write out report
write_report(rpt)

# View report
# file.show(tmp)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_spec.r
\name{create_table}
\alias{create_table}
\title{Create a table}
\usage{
create_table(
  x,
  show_cols = "all",
  use_attributes = "all",
  width = NULL,
  first_row_blank = FALSE,
  n_format = upcase_parens,
  headerless = FALSE,
  borders = "none",
  header_bold = FALSE,
  continuous = FALSE
)
}
\arguments{
\item{x}{The data frame or tibble from which to create the table object.}

\item{show_cols}{This parameter gives control over which columns in the 
input data to display on the report by default.  Valid values are
'all', 'none', a vector of quoted column names, or a vector of 
column positions.  'all' means show all columns, 
unless overridden by the column definitions.  
'none' means don't show any 
columns unless specified in the column definitions.  If a vector of column
names or positions is supplied, those columns will be shown in the report 
in the order specified, whether or not a definition is supplied. See the 
\code{\link{define}} function for additional information on how to
show/hide report columns.}

\item{use_attributes}{Whether or not to use any formatting attributes assigned
to the columns on the input data frame.  Valid values are 'all', 'none', or
a vector of attribute names to use.  Possible attributes that may be used
are 'label', 'format', 'width', and 'justify'.  By default, any of these
attribute values will be applied to the table.  For example, if you assign
a label to the 'label' attribute of a data frame column, pass that data 
frame into \code{create_table}, and don't override the label value on a 
\code{define} function, the label will appear as a column header on the
table.  The \code{use_attributes} parameter allows you to control this default
behavior, and use or ignore data frame attributes as desired.}

\item{width}{The expected width of the table in the report units of 
measure.  By default, the width setting is NULL, and columns will be sized
according to the width of the data and labels.  If the width parameter is 
set, the function will attempt to size the table to the specified width.
If the sum of the column widths is less than the specified width, the 
function will adjust the columns widths proportionally to fit the specified
width.  If the sum of the column widths is wider than the table width 
parameter value, the table width parameter will be ignored.}

\item{first_row_blank}{Whether to place a blank row under the table header.
Valid values are TRUE or FALSE.  Default is FALSE.}

\item{n_format}{The formatting function to apply to the header "N=" label. 
The default formatting function is \code{\link{upcase_parens}}.}

\item{headerless}{Whether to create a headerless table. A headerless
table displays the table data only. Default is FALSE, meaning the table
will have a header.}

\item{borders}{Whether and where to place a border. Valid values are 'top',
'bottom', 'left', 'right', 'all', 'none', 'outside', 'inside', and 'body'.  
Default is 'none'.  The 'left', 'right', 'outside', 'inside', and 'body'
border specifications only apply to RTF, HTML, PDF, and DOCX reports.  
The 'body' border 
specification means put borders around only the body of the table.}

\item{header_bold}{Whether or not the column headers on the tables should
be bolded.  Valid values are TRUE and FALSE.  The default is FALSE.}

\item{continuous}{If a table crosses multiple pages, it is normally
broken into a separate table for each page, and the titles and footnotes
are repeated on each page. When the "continuous" parameter is TRUE, the
table will instead be a single table, and the titles and footnotes will
not be repeated on each page.  This parameter currently only works for RTF
outputs.  Also, this parameter only works for titles and footnotes that
are attached to the table body.  Titles and footnotes attached to the 
report will still be shown on every page.}
}
\description{
The \code{create_table} function creates a table object to which 
further specifications can be added.  The object can be added to a report
using the \code{\link{add_content}} function. The object is implemented as an 
S3 object of class 'table_spec'.
}
\details{
A table object is a container to hold information about a table.  The 
only required information for a table is the table data.  All other 
parameters and functions are optional.

By default, the table will display all columns in the data frame.  To change
this default, use the \code{show_cols} parameter.  Setting this parameter
to 'none' will display none of the columns in the data, unless they are
explicitly defined with a \code{\link{define}} function.  

The \code{show_cols} parameter also accepts a vector of column positions
or column names. When a vector is supplied, \code{create_table} will 
display only those columns on the report, in the order encountered in the 
vector.  The \code{show_cols} parameter is the only mechanism in 
\code{create_table} to modify the column order. Otherwise, modify the 
order prior to sending the data to \code{create_table} using the many options
available in Base R or supplemental packages.
}
\section{Setting Formatting Attributes}{

Formatting attributes can be controlled in three ways.  By default, formatting
attributes assigned to the data frame will be passed through to the 
reporting functions.  The reporting functions will recognize the 'label',
'format', 'width', and 'justify' attributes. In other words, you can control 
the column label, width, format, and alignment of your report columns simply by 
assigning those attributes to your data frame. The advantage of using
attributes assigned to data frame columns is that you can store those 
attributes permanently with the data frame, and those attributes will
not have to be re-specified for each report.  To ignore attributes assigned
to the data frame, set the \code{use_attributes} parameter to 'none'.

Secondly, attributes can be specified using the \code{\link{column_defaults}}
function.  This function allows the user to apply a default set of parameters
to one or more columns.  If no columns are specified in the \code{var} 
or \code{from} and \code{to} parameter of this function, the defaults 
will apply to all columns.  Any default parameter value can be overridden 
by the \code{\link{define}} function.

Lastly, the \code{\link{define}} function provides the most control over 
column parameters.  This function provides a significant amount of 
functionality that cannot be specified elsewhere.  See the 
\code{\link{define}} function for additional information.  The \code{define}
function will also override any formatting attributes assigned to the 
data frame, or anything set by the \code{column_defaults} function.
}

\section{Additional Functionality}{

The \code{create_table} function also provides the capabilities to create
a "headerless" table.  A headerless table is useful when combining two tables 
into one report.  The example below illustrates use of a headerless table.

Since the purpose of the \strong{reporter} package is to create statistical 
reports, the \code{create_table} function makes it easy to add population
counts to the table header.  These population counts are added to column
labels and spanning header labels using the \code{n} parameter on the 
\code{\link{define}} or \code{\link{spanning_header}} functions.  The 
population count is formatted according to the  
\code{n_format} parameter on \code{create_table}. The \strong{reporter} 
package provides four population count formatting functions.  
You may create your own formatting function 
if one of these functions does not meet your needs.  See 
\code{\link{upcase_parens}} for further details.
}

\examples{
library(reporter)
library(magrittr)

# Create temp file path
tmp <- file.path(tempdir(), "mtcars.txt")

#Subset cars data
dat <- mtcars[1:10, 1:7]

# Calculate means for all columns
dat_sum <- data.frame(all_cars = "All cars average", as.list(sapply(dat, mean)), 
                      stringsAsFactors = FALSE)

# Get vehicle names into first column
dat_mod <- data.frame(vehicle = rownames(dat), dat, stringsAsFactors = FALSE)
                      
# Create table for averages
tbl1 <- create_table(dat_sum) \%>\% 
        titles("Table 1.0", "MTCARS Sample Data") \%>\% 
        column_defaults(width = .5) \%>\% 
        define(all_cars, label = "", width = 2) \%>\% 
        define(mpg, format = "\%.1f") \%>\% 
        define(disp, format = "\%.1f") \%>\% 
        define(hp, format = "\%.0f") \%>\% 
        define(qsec, format = "\%.2f")

# Create table for modified data
tbl2 <- create_table(dat_mod, headerless = TRUE) \%>\% 
        column_defaults(width = .5) \%>\% 
        define(vehicle, width = 2) 

# Create the report object
rpt <- create_report(tmp) \%>\% 
  add_content(tbl1, align = "left", page_break = FALSE) \%>\% 
  add_content(tbl2, align = "left") 

# Write the report to the file system
write_report(rpt)

# Write report to console
writeLines(readLines(tmp, encoding = "UTF-8"))

#                                 Table 1.0
#                             MTCARS Sample Data
# 
#                             mpg    cyl   disp     hp   drat     wt   qsec
# -------------------------------------------------------------------------
# All cars average           20.4    5.8  208.6    123  3.538  3.128  18.58
# 
# Mazda RX4                    21      6    160    110    3.9   2.62  16.46
# Mazda RX4 Wag                21      6    160    110    3.9  2.875  17.02
# Datsun 710                 22.8      4    108     93   3.85   2.32  18.61
# Hornet 4 Drive             21.4      6    258    110   3.08  3.215  19.44
# Hornet Sportabout          18.7      8    360    175   3.15   3.44  17.02
# Valiant                    18.1      6    225    105   2.76   3.46  20.22
# Duster 360                 14.3      8    360    245   3.21   3.57  15.84
# Merc 240D                  24.4      4  146.7     62   3.69   3.19     20
# Merc 230                   22.8      4  140.8     95   3.92   3.15   22.9
# Merc 280                   19.2      6  167.6    123   3.92   3.44   18.3
# 
}
\seealso{
\code{\link{create_report}} to create a report, 
\code{\link{create_plot}} to create a plot,
\code{\link{create_text}} to create text content, and 
\code{\link{add_content}} to append content to a report.  Also see
the \code{\link{titles}}, \code{\link{footnotes}}, and \code{\link{page_by}}
functions to add those items to the table if desired.

Other table: 
\code{\link{column_defaults}()},
\code{\link{define}()},
\code{\link{print.table_spec}()},
\code{\link{spanning_header}()},
\code{\link{stub}()}
}
\concept{table}
